/* s7, a Scheme interpreter
 *
 *    derived from:
 *
 * --------------------------------------------------------------------------------
 * T I N Y S C H E M E    1 . 3 9
 *   Dimitrios Souflis (dsouflis@acm.org)
 *   Based on MiniScheme (original credits follow)
 * (MINISCM)               coded by Atsushi Moriwaki (11/5/1989)
 * (MINISCM)           E-MAIL :  moriwaki@kurims.kurims.kyoto-u.ac.jp
 * (MINISCM) This version has been modified by R.C. Secrist.
 * (MINISCM)
 * (MINISCM) Mini-Scheme is now maintained by Akira KIDA.
 * (MINISCM)
 * (MINISCM) This is a revised and modified version by Akira KIDA.
 * (MINISCM)	current version is 0.85k4 (15 May 1994)
 * --------------------------------------------------------------------------------
 *
 * apparently tinyScheme is under the BSD license, so I guess s7 is too.
 * Here is Snd's verbiage which can apply here:
 *
 *     The authors hereby grant permission to use, copy, modify, distribute,
 *     and license this software and its documentation for any purpose.  No
 *     written agreement, license, or royalty fee is required.  Modifications
 *     to this software may be copyrighted by their authors and need not
 *     follow the licensing terms described here.
 *
 * followed by the usual all-caps shouting about liability.
 *
 * --------------------------------------------------------------------------------
 *
 * s7, Bill Schottstaedt, Aug-08, bil@ccrma.stanford.edu
 *
 * Mike Scholz provided the FreeBSD support (complex trig funcs, etc)
 * Rick Taube, Andrew Burnson, Donny Ward, and Greg Santucci provided the MS Visual C++ support
 *
 * Documentation is in s7.h and s7.html.
 * s7test.scm is a regression test.
 * glistener.c is a gtk-based listener.
 * repl.scm is a vt100-based listener.
 * cload.scm and lib*.scm tie in various C libraries.
 * lint.scm checks Scheme code for infelicities.
 * r7rs.scm implements some of r7rs (small).
 * write.scm currrently has pretty-print.
 * mockery.scm has the mock-data definitions.
 * stuff.scm has some stuff.
 *
 * s7.c is organized as follows:
 *
 *    structs and type flags
 *    constants
 *    GC
 *    stacks
 *    symbols and keywords
 *    environments
 *    continuations
 *    numbers
 *    characters
 *    strings
 *    ports
 *    format
 *    lists
 *    vectors
 *    hash-tables
 *    c-objects
 *    functions
 *    equal?
 *    generic length, copy, reverse, fill!, append
 *    error handlers
 *    sundry leftovers
 *    the optimizers
 *    multiple-values, quasiquote
 *    eval
 *    multiprecision arithmetic
 *    *s7* environment
 *    initialization
 *    repl
 *
 * naming conventions: s7_* usually are C accessible (s7.h), g_* are scheme accessible (FFI), 
 *   H_* are documentation strings, Q_* are procedure signatures,
 *   *_1 are auxilliary functions, big_* refer to gmp, 
 *   scheme "?" corresponds to C "is_", scheme "->" to C "_to_".
 *
 * ---------------- compile time switches ----------------
 */

#include "mus-config.h"

/*
 * Your config file goes here, or just replace that #include line with the defines you need.
 * The compile-time switches involve booleans, complex numbers, and multiprecision arithmetic.
 * Currently we assume we have setjmp.h (used by the error handlers).
 *
 * Complex number support which is problematic in C++, Solaris, and netBSD
 *   is on the HAVE_COMPLEX_NUMBERS switch. In OSX or Linux, if you're not using C++,
 *
 *   #define HAVE_COMPLEX_NUMBERS 1
 *   #define HAVE_COMPLEX_TRIG 1
 *
 *   In C++ I use:
 *
 *   #define HAVE_COMPLEX_NUMBERS 1
 *   #define HAVE_COMPLEX_TRIG 0
 *
 *   In windows, both are 0.
 *
 *   Some systems (FreeBSD) have complex.h, but some random subset of the trig funcs, so
 *   HAVE_COMPLEX_NUMBERS means we can find
 *      cimag creal cabs csqrt carg conj
 *   and HAVE_COMPLEX_TRIG means we have
 *      cacos cacosh casin casinh catan catanh ccos ccosh cexp clog cpow csin csinh ctan ctanh
 *
 * When HAVE_COMPLEX_NUMBERS is 0, the complex functions are stubs that simply return their
 *   argument -- this will be very confusing for the s7 user because, for example, (sqrt -2)
 *   will return something bogus (it will not signal an error).
 *
 * so the incoming (non-s7-specific) compile-time switches are
 *     HAVE_COMPLEX_NUMBERS, HAVE_COMPLEX_TRIG, SIZEOF_VOID_P
 * if SIZEOF_VOID_P is not defined, we look for __SIZEOF_POINTER__ instead
 *   the default is to assume that we're running on a 64-bit machine.
 *
 * To get multiprecision arithmetic, set WITH_GMP to 1.
 *   You'll also need libgmp, libmpfr, and libmpc (version 0.8.0 or later)
 *   In highly numerical contexts, the gmp version of s7 is about 50(!) times slower than the non-gmp version.
 *
 * and we use these predefined macros: __cplusplus, _MSC_VER, __GNUC__, __clang__, __ANDROID__
 *
 * if WITH_SYSTEM_EXTRAS is 1 (default is 1 unless _MSC_VER), various OS and file related functions are included.
 * in openBSD I think you need to include -ftrampolines in CFLAGS.
 * if you want this file to compile into a stand-alone interpreter, define WITH_MAIN
 *
 * -O3 is often slower, sometimes faster (at least according to callgrind), similarly for -finline-functions and -ftree-vectorize
 * -march=native -fomit-frame-pointer -m64 -funroll-loops gains about .1%
 * -ffast-math makes a mess of NaNs, and does not appear to be faster
 * for timing tests, I use: -O2 -march=native -fomit-frame-pointer -funroll-loops
 * according to callgrind, clang is normally about 10% slower than gcc, and vectorization either doesn't work or is much worse than gcc's
 *   also g++ appears to be slightly slower than gcc (though it takes forever to compile s7.c: gcc s7 compile time 37 secs, g++ 256!)
 */

#if (defined(__GNUC__) || defined(__clang__))
  #define WITH_GCC 1
#else
  #define WITH_GCC 0
#endif


/* ---------------- initial sizes ---------------- */

#ifndef INITIAL_HEAP_SIZE
#define INITIAL_HEAP_SIZE 128000
/* the heap grows as needed, this is its initial size.
 * If the initial heap is small, s7 can run in about 2.5 Mbytes of memory. There are (many) cases where a bigger heap is faster.
 * The heap size must be a multiple of 32.  Each object takes about 60 bytes.
 *
 * repl runs in    4Mb (18v) (64bit) if heap is 8192
 *                11Mb (25v)         if 128k heap
 *  snd (no gui)  15Mb (151v)
 *  snd (no gui)  17Mb (33v)         (no gsl, fftw, audio)
 *                21Mb (155v)        same but with ALSA!?
 *  snd (motif)   12Mb (285v)
 *  snd (gtk)     32Mb (515v!)
 */
#endif

#ifndef SYMBOL_TABLE_SIZE
#define SYMBOL_TABLE_SIZE 13567
/* names are hashed into the symbol table (a vector) and collisions are chained as lists.
 */
#endif

#define INITIAL_STACK_SIZE 512
/* the stack grows as needed, each frame takes 4 entries, this is its initial size.
 *   this needs to be big enough to handle the eval_c_string's at startup (ca 100)
 *   In s7test.scm, the maximum stack size is ca 440.  In snd-test.scm, it's ca 200.
 *   This number matters only because call/cc copies the stack, which requires filling
 *   the unused portion of the new stack, which requires memcpy of #<unspecified>'s.
 */

#define INITIAL_PROTECTED_OBJECTS_SIZE 16
/* a vector of objects that are (semi-permanently) protected from the GC, grows as needed */

#define GC_TEMPS_SIZE 256
/* the number of recent objects that are temporarily gc-protected; 8 works for s7test and snd-test.
 *    For the FFI, this sets the lag between a call on s7_cons and the first moment when its result
 *    might be vulnerable to the GC.
 */


/* ---------------- scheme choices ---------------- */

#ifndef WITH_GMP
  #define WITH_GMP 0
  /* this includes multiprecision arithmetic for all numeric types and functions, using gmp, mpfr, and mpc
   * WITH_GMP adds the following functions: bignum and bignum?, and (*s7* 'bignum-precision)
   * using gmp with precision=128 is about 50 times slower than using C doubles and int64_ts.
   */
#endif

#define DEFAULT_BIGNUM_PRECISION 128

#ifndef WITH_PURE_S7
  #define WITH_PURE_S7 0
#endif
#if WITH_PURE_S7
  #define WITH_EXTRA_EXPONENT_MARKERS 0
  #define WITH_IMMUTABLE_UNQUOTE 1
  /* also omitted: *-ci* functions, char-ready?, cond-expand, multiple-values-bind|set!, call-with-values, defmacro(*)
   *   and a lot more (inexact/exact, integer-length,  etc) -- see s7.html.
   */
#endif

#ifndef WITH_EXTRA_EXPONENT_MARKERS
  #define WITH_EXTRA_EXPONENT_MARKERS 0
  /* if 1, s7 recognizes "d", "f", "l", and "s" as exponent markers, in addition to "e" (also "D", "F", "L", "S") */
#endif

#ifndef WITH_SYSTEM_EXTRAS
  #define WITH_SYSTEM_EXTRAS (!_MSC_VER)
  /* this adds several functions that access file info, directories, times, etc
   *    this may be replaced by the cload business below
   */
#endif

#ifndef WITH_IMMUTABLE_UNQUOTE
  #define WITH_IMMUTABLE_UNQUOTE 0
  /* this removes the name "unquote" */
#endif

#ifndef WITH_C_LOADER
  #if WITH_GCC
    #define WITH_C_LOADER 1
  /* (load file.so [e]) looks for (e 'init_func) and if found, calls it
   *   as the shared object init function.  If WITH_SYSTEM_EXTRAS is 0, the caller
   *   needs to supply system and delete-file so that cload.scm works.
   */
  #else
    #define WITH_C_LOADER 0
  #endif
#endif

#ifndef WITH_HISTORY
  #define WITH_HISTORY 0
  /* this includes a circular buffer of previous evaluations for debugging, ((owlet) 'error-history) and (*s7* 'history-size) */
#endif

#ifndef DEFAULT_HISTORY_SIZE
  #define DEFAULT_HISTORY_SIZE 8
  /* this is the default length of the eval history buffer */
#endif

#ifndef WITH_PROFILE
  #define WITH_PROFILE 0
  /* this includes profiling data collection accessible from scheme via the hash-table (*s7* 'profile-info) */
#endif

/* in case mus-config.h forgets these */
#ifdef _MSC_VER
  #ifndef HAVE_COMPLEX_NUMBERS
    #define HAVE_COMPLEX_NUMBERS 0
  #endif
  #ifndef HAVE_COMPLEX_TRIG
    #define HAVE_COMPLEX_TRIG 0
  #endif
#else
  #ifndef HAVE_COMPLEX_NUMBERS
    #define HAVE_COMPLEX_NUMBERS 1
  #endif
  #if __cplusplus
    #ifndef HAVE_COMPLEX_TRIG
      #define HAVE_COMPLEX_TRIG 0
    #endif
  #else
    #ifndef HAVE_COMPLEX_TRIG
      #define HAVE_COMPLEX_TRIG 1
    #endif
  #endif
#endif

/* -------------------------------------------------------------------------------- */

#ifndef WITH_MULTITHREAD_CHECKS
  #define WITH_MULTITHREAD_CHECKS 0
  /* debugging aid if using s7 in a multithreaded program -- this code courtesy of Kjetil Matheussen */
#endif
 
#ifndef DEBUGGING
  #define DEBUGGING 0
#endif
#ifndef OP_NAMES
  #define OP_NAMES 0
#endif

#ifndef _MSC_VER
  #include <unistd.h>
  #include <sys/param.h>
  #include <strings.h>
  #include <errno.h>
  #include <locale.h>
#else
  /* in Snd these are in mus-config.h */
  #ifndef MUS_CONFIG_H_LOADED
    #define snprintf _snprintf 
    #if _MSC_VER > 1200
      #define _CRT_SECURE_NO_DEPRECATE 1
      #define _CRT_NONSTDC_NO_DEPRECATE 1
      #define _CRT_SECURE_CPP_OVERLOAD_STANDARD_NAMES 1
    #endif
  #endif
  #include <io.h>
  #pragma warning(disable: 4244)
#endif

#include <stdio.h>
#include <limits.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <time.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdint.h>
#include <inttypes.h>
#include <setjmp.h>

#if WITH_MULTITHREAD_CHECKS
  #include <pthread.h>
#endif

#if __cplusplus
  #include <cmath>
#else
  #include <math.h>
#endif

#if HAVE_COMPLEX_NUMBERS
  #if __cplusplus
    #include <complex>
  #else
    #include <complex.h>
    #ifndef __SUNPRO_C
      #if defined(__sun) && defined(__SVR4)
        #undef _Complex_I
        #define _Complex_I 1.0fi
      #endif
    #endif
  #endif
#ifndef CMPLX
  /* c11 addition? */
  #define CMPLX(r, i) ((r) + ((i) * _Complex_I))
#endif
#endif

#include "s7.h"

#ifndef M_PI
  #define M_PI 3.1415926535897932384626433832795029L
#endif

#ifndef INFINITY
  #define INFINITY (-log(0.0))
  /* 1.0 / 0.0 is also used, there is sometimes a function, infinity(), MSC apparently uses HUGE_VALF */
#endif

#ifndef NAN
  #define NAN (INFINITY / INFINITY)
#endif

#define BOLD_TEXT "\033[1m"
#define UNBOLD_TEXT "\033[22m"

#define WRITE_REAL_PRECISION 16
static int32_t float_format_precision = WRITE_REAL_PRECISION;

#if ((!__NetBSD__) && ((_MSC_VER) || (!defined(__STC__)) || (defined(__STDC_VERSION__) && (__STDC_VERSION__ < 199901L))))
  #define __func__ __FUNCTION__
#endif

#define DISPLAY(Obj) s7_object_to_c_string(sc, Obj)
#define DISPLAY_80(Obj) object_to_truncated_string(sc, Obj, 80)

typedef intptr_t opcode_t;

#define PRINT_NAME_PADDING 16
#if (((defined(SIZEOF_VOID_P)) && (SIZEOF_VOID_P == 4)) || ((defined(__SIZEOF_POINTER__)) && (__SIZEOF_POINTER__ == 4)))
  #define PRINT_NAME_SIZE (20 - PRINT_NAME_PADDING - 2) /* pointless */
  #define POINTER_32 true
#else
  #define PRINT_NAME_SIZE (40 - PRINT_NAME_PADDING - 2)
  #define POINTER_32 false
#endif


/* types */
#define T_FREE                 0
#define T_PAIR                 1
#define T_NIL                  2
#define T_UNDEFINED            3
#define T_UNSPECIFIED          4
#define T_EOF_OBJECT           5
#define T_BOOLEAN              6
#define T_CHARACTER            7
#define T_SYNTAX               8
#define T_SYMBOL               9

#define T_INTEGER             10
#define T_RATIO               11
#define T_REAL                12
#define T_COMPLEX             13

#define T_BIG_INTEGER         14 /* these four used only if WITH_GMP -- order matters */
#define T_BIG_RATIO           15
#define T_BIG_REAL            16
#define T_BIG_COMPLEX         17  

#define T_STRING              18
#define T_C_OBJECT            19
#define T_VECTOR              20
#define T_INT_VECTOR          21
#define T_FLOAT_VECTOR        22

#define T_CATCH               23
#define T_DYNAMIC_WIND        24
#define T_HASH_TABLE          25
#define T_LET                 26
#define T_ITERATOR            27
#define T_STACK               28
#define T_COUNTER             29
#define T_SLOT                30
#define T_C_POINTER           31
#define T_OUTPUT_PORT         32
#define T_INPUT_PORT          33
#define T_BAFFLE              34
#define T_RANDOM_STATE        35
#define T_OPTLIST             36

#define T_GOTO                37
#define T_CONTINUATION        38
#define T_CLOSURE             39
#define T_CLOSURE_STAR        40
#define T_C_MACRO             41
#define T_MACRO               42
#define T_MACRO_STAR          43
#define T_BACRO               44
#define T_BACRO_STAR          45
#define T_C_FUNCTION_STAR     46
#define T_C_FUNCTION          47
#define T_C_ANY_ARGS_FUNCTION 48
#define T_C_OPT_ARGS_FUNCTION 49
#define T_C_RST_ARGS_FUNCTION 50

#define NUM_TYPES        51

/* T_STACK, T_SLOT, T_BAFFLE, T_DYNAMIC_WIND, T_OPTLIST, and T_COUNTER are internal */

typedef enum {TOKEN_EOF, TOKEN_LEFT_PAREN, TOKEN_RIGHT_PAREN, TOKEN_DOT, TOKEN_ATOM, TOKEN_QUOTE, TOKEN_DOUBLE_QUOTE,
	      TOKEN_BACK_QUOTE, TOKEN_COMMA, TOKEN_AT_MARK, TOKEN_SHARP_CONST,
	      TOKEN_VECTOR, TOKEN_BYTE_VECTOR, TOKEN_INT_VECTOR, TOKEN_FLOAT_VECTOR} token_t;

typedef enum {FILE_PORT, STRING_PORT, FUNCTION_PORT} port_type_t;

typedef struct {
  bool needs_free;
  FILE *file;
  char *filename;
  int32_t filename_length, gc_loc; /* gc_loc uses -1 as unset flag -- kinda ugly */
  void *next;
  s7_pointer (*input_function)(s7_scheme *sc, s7_read_t read_choice, s7_pointer port);
  void (*output_function)(s7_scheme *sc, unsigned char c, s7_pointer port);
  /* a version of string ports using a pointer to the current location and a pointer to the end
   *   (rather than an integer for both, indexing from the base string) was not faster.
   */
  s7_pointer orig_str;                                                   /* GC protection for string port string */
  int32_t (*read_character)(s7_scheme *sc, s7_pointer port);             /* function to read a character */
  void (*write_character)(s7_scheme *sc, int32_t c, s7_pointer port);    /* function to write a character */
  void (*write_string)(s7_scheme *sc, const char *str, int32_t len, s7_pointer port); /* function to write a string of known length */
  token_t (*read_semicolon)(s7_scheme *sc, s7_pointer port);             /* internal skip-to-semicolon reader */
  int32_t (*read_white_space)(s7_scheme *sc, s7_pointer port);           /* internal skip white space reader */
  s7_pointer (*read_name)(s7_scheme *sc, s7_pointer pt);                 /* internal get-next-name reader */
  s7_pointer (*read_sharp)(s7_scheme *sc, s7_pointer pt);                /* internal get-next-sharp-constant reader */
  s7_pointer (*read_line)(s7_scheme *sc, s7_pointer pt, bool eol_case, bool copied);  /* function to read a string up to \n */
  void (*display)(s7_scheme *sc, const char *s, s7_pointer pt);
} port_t;


typedef struct opt_funcs {
  int32_t typ;
  void *func;
  struct opt_funcs *next;
} opt_funcs;

typedef struct {
  const char *name;
  int32_t name_length;
  uint32_t id;
  char *doc;
  s7_pointer generic_ff;
  s7_pointer signature;
  s7_pointer (*chooser)(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops);
  s7_pointer *arg_defaults, *arg_names;
  s7_pointer call_args;
  opt_funcs *opt_data;
} c_proc_t;


typedef struct {               /* call/cc */
  uint32_t stack_size, op_stack_loc, op_stack_size;
  int32_t local_key;               /* for with-baffle */
} continuation_t;


typedef struct vdims_t {
  uint32_t ndims;
  bool elements_allocated, dimensions_allocated; /* these are allocated as bytes, not ints, so the struct size is 32 */
  s7_int *dims, *offsets;
  s7_pointer original;
} vdims_t;


typedef struct {
  int32_t type;
  uint32_t outer_type;
  const char *name;
  s7_pointer scheme_name;
  char *(*print)(s7_scheme *sc, void *value);
  void (*free)(void *value);
  bool (*equal)(void *val1, void *val2);
  void (*gc_mark)(void *val);
  s7_pointer (*ref)(s7_scheme *sc, s7_pointer obj, s7_pointer args);
  s7_pointer (*set)(s7_scheme *sc, s7_pointer obj, s7_pointer args);
  s7_pointer (*length)(s7_scheme *sc, s7_pointer obj);
  s7_pointer (*copy)(s7_scheme *sc, s7_pointer args);
  s7_pointer (*reverse)(s7_scheme *sc, s7_pointer obj);
  s7_pointer (*fill)(s7_scheme *sc, s7_pointer args);
  char *(*print_readably)(s7_scheme *sc, void *value);
  s7_pointer (*direct_ref)(s7_scheme *sc, s7_pointer obj, s7_int index);
  s7_pointer (*direct_set)(s7_scheme *sc, s7_pointer obj, s7_int index, s7_pointer val);
} c_object_t;


typedef struct hash_entry_t {
  s7_pointer key, value;
  struct hash_entry_t *next;
  uint32_t raw_hash;
} hash_entry_t;

typedef uint32_t (*hash_map_t)(s7_scheme *sc, s7_pointer table, s7_pointer key);    /* hash-table object->location mapper */
typedef hash_entry_t *(*hash_check_t)(s7_scheme *sc, s7_pointer table, s7_pointer key); /* hash-table object equality function */
static hash_map_t *default_hash_map;

typedef s7_int (*s7_i_pi_t)(s7_pointer p, s7_int i1);
typedef s7_int (*s7_i_pii_t)(s7_pointer p, s7_int i1, s7_int i2);
typedef s7_int (*s7_i_iii_t)(s7_int i1, s7_int i2, s7_int i3);
typedef s7_pointer (*s7_p_p_t)(s7_pointer p);
typedef s7_pointer (*s7_p_t)(void);
typedef bool (*s7_b_pp_t)(s7_pointer p1, s7_pointer p2);
typedef bool (*s7_b_pi_t)(s7_pointer p1, s7_int i2);
typedef bool (*s7_b_d_t)(s7_double p1);
typedef bool (*s7_b_i_t)(s7_int p1);
typedef bool (*s7_b_ii_t)(s7_int p1, s7_int p2);
typedef bool (*s7_b_dd_t)(s7_double p1, s7_double p2);
typedef s7_pointer (*s7_p_pp_t)(s7_pointer p1, s7_pointer p2);
typedef s7_pointer (*s7_p_ppi_t)(s7_pointer p1, s7_pointer p2, s7_int i1);
typedef s7_pointer (*s7_p_ppp_t)(s7_pointer p1, s7_pointer p2, s7_pointer p3);
typedef s7_pointer (*s7_p_pi_t)(s7_pointer p1, s7_int i1);
typedef s7_pointer (*s7_p_pip_t)(s7_pointer p1, s7_int i1, s7_pointer p2);
typedef s7_pointer (*s7_p_ii_t)(s7_int i1, s7_int i2);
typedef s7_pointer (*s7_p_dd_t)(s7_double x1, s7_double x2);

typedef union {
  s7_int i;
  s7_double x;
  s7_pointer p;
  void *obj;
  s7_function cf;
  s7_double (*d_f)(void);
  s7_double (*d_d_f)(s7_double x);
  s7_double (*d_dd_f)(s7_double x1, s7_double x2);
  s7_double (*d_ddd_f)(s7_double x1, s7_double x2, s7_double x3);
  s7_double (*d_dddd_f)(s7_double x1, s7_double x2, s7_double x3, s7_double x4);
  s7_double (*d_v_f)(void *obj);
  s7_double (*d_vd_f)(void *obj, s7_double fm);
  s7_double (*d_vdd_f)(void *obj, s7_double x1, s7_double x2);
  s7_double (*d_vid_f)(void *obj, s7_int i, s7_double fm);
  s7_double (*d_id_f)(s7_int i, s7_double fm);
  s7_double (*d_pi_f)(s7_pointer obj, s7_int i1);
  s7_double (*d_ip_f)(s7_int i1, s7_pointer p);
  s7_double (*d_pd_f)(s7_pointer obj, s7_double x);
  s7_double (*d_pid_f)(s7_pointer obj, s7_int i1, s7_double x);
  s7_double (*d_p_f)(s7_pointer p);
  s7_int (*i_d_f)(s7_double i1);
  s7_int (*i_i_f)(s7_int i1);
  s7_int (*i_ii_f)(s7_int i1, s7_int i2);
  s7_int (*i_iii_f)(s7_int i1, s7_int i2, s7_int i3);
  s7_int (*i_p_f)(s7_pointer p);
  s7_int (*i_pi_f)(s7_pointer p, s7_int i1);
  s7_int (*i_pii_f)(s7_pointer p, s7_int i1, s7_int i2);
  bool (*b_i_f)(s7_int p);
  bool (*b_d_f)(s7_double p);
  bool (*b_p_f)(s7_pointer p);
  bool (*b_pp_f)(s7_pointer p1, s7_pointer p2);
  bool (*b_pi_f)(s7_pointer p1, s7_int i2);
  bool (*b_ii_f)(s7_int i1, s7_int i2);
  bool (*b_dd_f)(s7_double x1, s7_double x2);
  s7_pointer (*p_f)(void);
  s7_pointer (*p_p_f)(s7_pointer p);
  s7_pointer (*p_pp_f)(s7_pointer p1, s7_pointer p2);
  s7_pointer (*p_ppp_f)(s7_pointer p, s7_pointer p2, s7_pointer p3);
  s7_pointer (*p_pi_f)(s7_pointer p1, s7_int i1);
  s7_pointer (*p_ppi_f)(s7_pointer p1, s7_pointer p2, s7_int i1);
  s7_pointer (*p_pip_f)(s7_pointer p1, s7_int i1, s7_pointer p2);
  s7_pointer (*p_ii_f)(s7_int x1, s7_int x2);
  s7_pointer (*p_dd_f)(s7_double x1, s7_double x2);
  s7_pointer (*all_f)(s7_scheme *sc, s7_pointer expr);
  s7_double (*fd)(void *o);
  s7_int (*fi)(void *o);
  bool (*fb)(void *o);
  s7_pointer (*fp)(void *o);
} vunion;

typedef struct {
  vunion v1, v2, v3, v4, v5, v6, v7, v8;
#if DEBUGGING
  s7_pointer expr;
#endif
} opt_info;


/* cell structure */
typedef struct s7_cell {
  union {
    uint64_t flag;
    uint8_t type_field;
    uint16_t sflag;
  } tf;
  int64_t hloc;
  union {

    union {                       /* integers, floats */
      s7_int integer_value;
      s7_double real_value;

      struct {
	char padding[PRINT_NAME_PADDING];
	char name[PRINT_NAME_SIZE + 2];
      } pval;

      struct {                    /* ratios */
	s7_int numerator;
	s7_int denominator;
      } fraction_value;

      struct {                    /* complex numbers */
	s7_double rl;
	s7_double im;
      } complex_value;

      unsigned long ul_value;    /* these two are not used by s7 in any way -- "unsigned long" for backwards compatibility */
      uint64_t ull_value;

#if WITH_GMP
      mpz_t big_integer;          /* bignums */
      mpq_t big_ratio;
      mpfr_t big_real;
      mpc_t big_complex;
      /* using free_lists here was not faster, and avoiding the extra init/clear too tricky.  These make up
       *   no more than ca. 5% of the gmp computation -- it is totally dominated by stuff like __gmpz_mul,
       *   so I can't see much point in optimizing the background noise.  In a very numerical context,
       *   gmp slows us down by a factor of 50.
       */
#endif
    } number;

    struct {                       /* ports */
      port_t *port;
      unsigned char *data;
      uint32_t size, point;        /* these limit the in-core portion of a string-port to 2^32-1 bytes */
      uint32_t line_number, file_number;
      bool is_closed;
      port_type_t ptype;
    } prt;

    struct{                        /* characters */
      unsigned char c, up_c;
      int32_t length;
      bool alpha_c, digit_c, space_c, upper_c, lower_c;
      char c_name[12];
    } chr;

    struct {                       /* c-pointers */
      void *c_pointer; 
      s7_pointer c_type, info;
      /* if a gc_free function were included, the pointer could be freed (etc) via the GC-cache (sweep function), but then
       *   every possible explicit free function would need to warn the GC not to try to free the pointer.
       *   How to recognize an unfreed pointer at sweep time?
       */
    } cptr;

    int32_t baffle_key;            /* baffles */

    struct {                       /* vectors */
      s7_int length;
      union {
	s7_pointer *objects;
	s7_int *ints;
	s7_double *floats;
      } elements;
      vdims_t *dim_info;
      s7_pointer (*vget)(s7_scheme *sc, s7_pointer vec, s7_int loc);
      s7_pointer (*vset)(s7_scheme *sc, s7_pointer vec, s7_int loc, s7_pointer val);
    } vector;

    struct {                        /* stacks (internal) */
      s7_int length;
      s7_pointer *objects;
      vdims_t *dim_info;
      int64_t top;
    } stk;

    struct {                        /* hash-tables */
      uint32_t mask, entries;
      hash_entry_t **elements;
      hash_check_t hash_func;
      hash_map_t *loc;
      s7_pointer dproc;             /* user-supplied list of hashing functions */
    } hasher;

    struct {                        /* iterators */
      s7_pointer obj, cur;
      union {
	s7_int loc;
	s7_pointer lcur;
      } lc;
      union {
	s7_int len;
	s7_pointer slow;
	hash_entry_t *hcur;
      } lw;
      s7_pointer (*next)(s7_scheme *sc, s7_pointer iterator);
    } iter;

    struct {
      c_proc_t *c_proc;              /* C functions, macros */
      s7_function ff;
      s7_pointer setter;
      uint32_t required_args, optional_args, all_args;
      bool rest_arg;
    } fnc;

    struct {                         /* pairs */
      s7_pointer car, cdr, opt1, opt2, opt3;
    } cons;

    struct {                         /* pairs */
      s7_pointer car, cdr, opt1, opt2;
      uint8_t opt_type;
    } cons_ext;

    struct {                         /* special purpose pairs (symbol-table etc) */
      s7_pointer unused_car, unused_cdr;
      uint64_t hash;
      const char *fstr;
      uint32_t op, line;             /* op=optimize_op, line=pair_line or saved symbol_ctr */
    } sym_cons;

    struct {                        /* scheme functions */
      s7_pointer args, body, env, setter;  /* args can be a symbol, as well as a list */
      int32_t arity, opt_addr;
    } func;

    struct {                        /* optlist (internal) */
      int32_t num_exprs, num_args, len, addr, pc;
      opt_info *opts;
    } opt;

    struct {                        /* strings */
      uint32_t length;
      union {
	bool needs_free;            /* string GC */
	uint32_t accessor;          /* symbol-access */
	int32_t temp_len;           /* temp string length (sc->tmp_strs) */
      } str_ext;
      char *svalue;
      uint64_t hash;                /* string hash-index */
      s7_pointer initial_slot;      /* for built-in symbols (unlet) */
      union {
	char *documentation;        /* symbol help */
	s7_pointer ksym;            /* keyword->symbol */
      } doc;
    } string;

    struct {                       /* symbols */
      s7_pointer name, global_slot, local_slot;
      int64_t id;
      uint32_t ctr;                /* closure_is_ok (check func only defined once) */
      uint16_t tag;                /* symbol lists (quick memq) */
      uint8_t type;
    } sym;

    struct {                       /* syntax */
      s7_pointer symbol;
      opcode_t op;
      int16_t min_args, max_args;
    } syn;

    struct {                       /* slots (bindings) */
      s7_pointer sym, val, nxt, pending_value, expr;
    } slt;

    struct {                       /* environments (frames, lets) */
      s7_pointer slots, nxt;
      int64_t id;                  /* id of rootlet is -1 */
      union {
	struct {
	  s7_pointer function;     /* __func__ (code) if this is a funclet */
	  uint32_t line, file; /* __func__ location if it is known */
	} efnc;
	struct {
	  s7_pointer dox1, dox2;   /* do loop variables */
	} dox;
	struct {                   /* (catch #t ...) opts */
	  uint64_t op_stack_loc, goto_loc;
	} ctall;
      } edat;
    } envr;

    struct {                        /* special stuff like #<unspecified> */
      /* these 3 are just place-holders */
      s7_pointer unused_slots, unused_nxt;
      int64_t unused_id;
      /* these two fields are for some special case objects like #<unspecified> */
      const char *name;
      int32_t len;
    } unq;

    struct {                        /* counter (internal) */
      s7_pointer result, list, env, slots; /* env = counter_let (curlet after map/for-each frame created) */
      uint64_t cap;                 /* sc->capture_let_counter for frame reuse */
    } ctr;

    struct {                        /* random-state */
#if WITH_GMP
      gmp_randstate_t state;
#else
      uint64_t seed, carry;
#endif
    } rng;

    struct {                        /* additional object types (C) */
      int32_t type;
      void *value;                  /*  the value the caller associates with the object */
      s7_pointer e;                 /*   the method list, if any (openlet) */
      void (*mark)(void *val);
    } c_obj;

    struct {                        /* continuations */
      continuation_t *continuation;
      s7_pointer stack;
      s7_pointer *stack_start, *stack_end, *op_stack;
    } cwcc;

    struct {                        /* call-with-exit */
      uint64_t goto_loc, op_stack_loc;
      bool active;
    } rexit;

    struct {                        /* catch */
      uint64_t goto_loc, op_stack_loc;
      s7_pointer tag;
      s7_pointer handler;
    } rcatch; /* C++ reserves "catch" I guess */

    struct {                        /* dynamic-wind */
      s7_pointer in, out, body;
      uint32_t state;
    } winder;
  } object;

#if WITH_PROFILE
  int32_t file_and_line;
#endif
#if DEBUGGING
  int32_t current_alloc_line, previous_alloc_line, gc_line, alloc_line, uses;
  int64_t current_alloc_type, previous_alloc_type, debugger_bits;
  const char *current_alloc_func, *previous_alloc_func, *gc_func, *alloc_func;
#endif

} s7_cell;


typedef struct {
  s7_pointer *objs;
  int32_t size, top, ref, size2;
  bool has_hits;
  int32_t *refs;
} shared_info;


typedef struct {
  int32_t loc, curly_len, ctr;
  char *curly_str;
  s7_pointer args, orig_str, curly_arg;
  s7_pointer port, strport;
} format_data;


typedef struct gc_obj {
  s7_pointer p;
  struct gc_obj *nxt;
} gc_obj;

typedef struct {
  s7_pointer *list;
  uint32_t size, loc;
} gc_list;


static s7_pointer *small_ints, *chars;
static s7_pointer real_zero, real_NaN, real_pi, real_one, arity_not_set, max_arity, real_infinity, real_minus_infinity, minus_one, minus_two;


struct s7_scheme {
  opcode_t cur_op;                    /* making this global is much slower! */
  s7_pointer value;
  s7_pointer args;                    /* arguments of current function */
  s7_pointer code, cur_code;          /* current code */
  s7_pointer envir;                   /* curlet */
  token_t tok;

  s7_pointer stack;                   /* stack is a vector */
  uint32_t stack_size;
  s7_pointer *stack_start, *stack_end, *stack_resize_trigger;

  s7_pointer *op_stack, *op_stack_now, *op_stack_end;
  uint32_t op_stack_size, max_stack_size;

  s7_cell **heap, **free_heap, **free_heap_top, **free_heap_trigger, **previous_free_heap_top;
  int64_t heap_size, gc_freed;

#if WITH_HISTORY
  s7_pointer eval_history1, eval_history2, error_history;
  bool using_history1;
#endif

#if WITH_MULTITHREAD_CHECKS
  int32_t lock_count;
  pthread_mutex_t lock;
#endif

  gc_obj *permanent_objects;

  s7_pointer protected_objects, protected_accessors;       /* a vector of gc-protected objects */
  uint32_t *gpofl;
  uint32_t protected_objects_size, protected_accessors_size, protected_accessors_loc;
  int32_t gpofl_loc;

  s7_pointer nil;                     /* empty list */
  s7_pointer T;                       /* #t */
  s7_pointer F;                       /* #f */
  s7_pointer eof_object;              /* #<eof> */
  s7_pointer undefined;               /* #<undefined> */
  s7_pointer unspecified;             /* #<unspecified> */
  s7_pointer no_value;                /* the (values) value */
  s7_pointer gc_nil;                  /* a marker for an unoccupied slot in sc->protected_objects (and other similar stuff) */

  s7_pointer symbol_table;            /* symbol table */
  s7_pointer rootlet, shadow_rootlet; /* rootlet */
  s7_int rootlet_entries;
  s7_pointer unlet;                   /* original bindings of predefined functions */

  s7_pointer input_port;              /* current-input-port */
  s7_pointer input_port_stack;        /*   input port stack (load and read internally) */
  s7_pointer output_port;             /* current-output-port */
  s7_pointer error_port;              /* current-error-port */
  s7_pointer owlet;                   /* owlet */
  s7_pointer error_type, error_data, error_code, error_line, error_file; /* owlet slots */
  s7_pointer standard_input, standard_output, standard_error;

  s7_pointer sharp_readers;           /* the binding pair for the global *#readers* list */
  s7_pointer load_hook;               /* *load-hook* hook object */
  s7_pointer unbound_variable_hook;   /* *unbound-variable-hook* hook object */
  s7_pointer missing_close_paren_hook, rootlet_redefinition_hook;
  s7_pointer error_hook, read_error_hook; /* *error-hook* hook object, and *read-error-hook* */
  s7_pointer direct_str;

  bool gc_off;                        /* gc_off: if true, the GC won't run */
  uint32_t gc_stats;
  uint32_t gensym_counter, f_class, add_class, multiply_class, subtract_class, equal_class;
  int32_t format_column;
  uint64_t capture_let_counter;
  bool short_print, is_autoloading;
  int64_t let_number;
  double default_rationalize_error, morally_equal_float_epsilon, hash_table_float_epsilon;
  s7_int default_hash_table_length, initial_string_port_length, print_length, objstr_max_len, history_size, true_history_size;
  s7_int max_vector_length, max_string_length, max_list_length, max_vector_dimensions;
  s7_pointer stacktrace_defaults;
  vdims_t *wrap_only;

  char *typnam;
  int32_t typnam_len;
  char *help_arglist;
  int32_t print_width;
  s7_pointer *singletons;

  #define INITIAL_TMP_STR_SIZE 16
  s7_pointer *tmp_strs;

  #define INITIAL_FILE_NAMES_SIZE 8
  s7_pointer *file_names;
  int32_t file_names_size, file_names_top;

  #define INITIAL_STRBUF_SIZE 1024
  uint32_t strbuf_size;
  #define TMPBUF_SIZE 1024
  char *strbuf, *tmpbuf;

  char *read_line_buf;
  uint32_t read_line_buf_size;

  s7_pointer v, w, x, y, z;         /* evaluator local vars */
  s7_pointer temp1, temp2, temp3, temp4, temp5, temp6, temp7, temp8, temp9, temp10, temp11;
  s7_pointer temp_cell, temp_cell_1, temp_cell_2;
  s7_pointer d1, d2, d3, d4;
  s7_pointer t1_1, t2_1, t2_2, t3_1, t3_2, t3_3, z2_1, z2_2;
  s7_pointer a1_1, a2_1, a2_2, a3_1, a3_2, a3_3, a4_1, a4_2, a4_3, a4_4;
  #define T_TEMPS_SIZE 32
  s7_pointer t_temps[T_TEMPS_SIZE];              /* more eval temps */
  int32_t t_temp_ctr;

  jmp_buf goto_start;
  bool longjmp_ok;
  int32_t setjmp_loc;

  void (*begin_hook)(s7_scheme *sc, bool *val);

  int32_t no_values, current_line, s7_call_line, safety;
  const char *current_file, *s7_call_file, *s7_call_name;

  shared_info *circle_info;
  format_data **fdats;
  int32_t num_fdats;
  s7_pointer elist_1, elist_2, elist_3, elist_4, elist_5, plist_1, plist_2, plist_3, qlist_2;
  gc_list *strings, *strings1, *vectors, *input_ports, *output_ports, *continuations, *c_objects, *hash_tables, *gensyms, *optlists;
  s7_pointer *setters;
  uint32_t setters_size, setters_loc;
  char ***string_lists;
  int32_t *string_locs, *string_sizes, *string_max_sizes;

  c_object_t **c_object_types;
  int32_t c_object_types_size;
  int32_t num_c_object_types;

  uint16_t syms_tag;
  int32_t ht_iter_tag, baffle_ctr, bignum_precision;
  s7_pointer default_rng;

  /* these symbols are primarily for the generic function search */
  s7_pointer abs_symbol, acos_symbol, acosh_symbol, add_symbol, angle_symbol, append_symbol, apply_symbol, apply_values_symbol, arity_symbol,
             ash_symbol, asin_symbol, asinh_symbol, assoc_symbol, assq_symbol, assv_symbol, atan_symbol, atanh_symbol,
             autoload_symbol, autoloader_symbol,
             byte_vector_symbol, byte_vector_ref_symbol, byte_vector_set_symbol, 
             c_pointer_symbol, caaaar_symbol, caaadr_symbol, caaar_symbol, caadar_symbol, caaddr_symbol, caadr_symbol,
             caar_symbol, cadaar_symbol, cadadr_symbol, cadar_symbol, caddar_symbol, cadddr_symbol, caddr_symbol, cadr_symbol,
             call_cc_symbol, call_with_current_continuation_symbol, call_with_exit_symbol, call_with_input_file_symbol, 
             call_with_input_string_symbol, call_with_output_file_symbol, call_with_output_string_symbol, car_symbol,
             catch_symbol, cdaaar_symbol, cdaadr_symbol, cdaar_symbol, cdadar_symbol, cdaddr_symbol, cdadr_symbol, cdar_symbol,
             cddaar_symbol, cddadr_symbol, cddar_symbol, cdddar_symbol, cddddr_symbol, cdddr_symbol, cddr_symbol, cdr_symbol,
             ceiling_symbol, char_downcase_symbol, char_eq_symbol, char_geq_symbol, char_gt_symbol, char_leq_symbol, char_lt_symbol,
             char_position_symbol, char_to_integer_symbol, char_upcase_symbol, cload_directory_symbol, close_input_port_symbol, 
             close_output_port_symbol, complex_symbol, cons_symbol, copy_symbol, cos_symbol, cosh_symbol, coverlet_symbol, 
             curlet_symbol, current_error_port_symbol, current_input_port_symbol, current_output_port_symbol, cutlet_symbol, cyclic_sequences_symbol,
             denominator_symbol, dilambda_symbol, display_symbol, divide_symbol, dynamic_wind_symbol,
             eq_symbol, error_symbol, eval_string_symbol, eval_symbol, exact_to_inexact_symbol, exit_symbol, exp_symbol, expt_symbol,
             features_symbol, fill_symbol, float_vector_ref_symbol, float_vector_set_symbol, float_vector_symbol, floor_symbol, 
             flush_output_port_symbol, for_each_symbol, format_symbol, funclet_symbol,
             gc_symbol, gcd_symbol, gensym_symbol, geq_symbol, get_output_string_symbol, gt_symbol,
             hash_table_entries_symbol, hash_table_ref_symbol, hash_table_set_symbol, hash_table_star_symbol, hash_table_symbol, help_symbol, 
             imag_part_symbol, inexact_to_exact_symbol, inlet_symbol, int_vector_ref_symbol, int_vector_set_symbol, int_vector_symbol,
             integer_decode_float_symbol, integer_to_char_symbol, is_aritable_symbol, is_boolean_symbol, is_byte_vector_symbol,
             is_c_object_symbol, is_c_pointer_symbol, is_char_alphabetic_symbol, is_char_lower_case_symbol, is_char_numeric_symbol,
             is_char_symbol, is_char_upper_case_symbol, is_char_whitespace_symbol, is_complex_symbol, is_constant_symbol,
             is_continuation_symbol, is_defined_symbol, is_dilambda_symbol, is_eof_object_symbol, is_eq_symbol, is_equal_symbol,
             is_eqv_symbol, is_even_symbol, is_exact_symbol, is_float_vector_symbol, is_gensym_symbol, is_hash_table_symbol,
             is_inexact_symbol, is_infinite_symbol, is_input_port_symbol, is_int_vector_symbol, is_integer_symbol, is_iterator_symbol,
             is_keyword_symbol, is_let_symbol, is_list_symbol, is_macro_symbol, is_morally_equal_symbol, is_nan_symbol, is_negative_symbol,
             is_null_symbol, is_number_symbol, is_odd_symbol, is_openlet_symbol, is_output_port_symbol, is_pair_symbol, 
             is_port_closed_symbol, is_positive_symbol, is_procedure_symbol, is_proper_list_symbol, is_provided_symbol,
             is_random_state_symbol, is_rational_symbol, is_real_symbol, is_sequence_symbol, is_string_symbol, is_symbol_symbol, is_syntax_symbol,
             is_vector_symbol, is_zero_symbol, iterate_symbol, iterator_is_at_end_symbol, iterator_sequence_symbol,
             is_float_symbol, is_integer_or_real_at_end_symbol, is_integer_or_any_at_end_symbol, is_unspecified_symbol, is_undefined_symbol,
             keyword_to_symbol_symbol, 
             lcm_symbol, length_symbol, leq_symbol, let_ref_fallback_symbol, let_ref_symbol, let_set_fallback_symbol,
             let_set_symbol, let_temporarily_symbol, list_ref_symbol, list_set_symbol, list_symbol, list_tail_symbol, list_values_symbol, 
             load_path_symbol, load_symbol, log_symbol, logand_symbol, logbit_symbol, logior_symbol, lognot_symbol, logxor_symbol, lt_symbol,
             magnitude_symbol, make_byte_vector_symbol, make_float_vector_symbol, make_hash_table_symbol, make_int_vector_symbol,
             make_iterator_symbol, string_to_keyword_symbol, make_list_symbol, make_shared_vector_symbol, make_string_symbol,
             make_vector_symbol, map_symbol, max_symbol, member_symbol, memq_symbol, memv_symbol, min_symbol, modulo_symbol, multiply_symbol, 
             newline_symbol, not_symbol, number_to_string_symbol, numerator_symbol, 
             object_to_string_symbol, object_to_let_symbol, open_input_file_symbol, open_input_string_symbol, open_output_file_symbol, 
             open_output_string_symbol, openlet_symbol, outlet_symbol, owlet_symbol, 
             pair_filename_symbol, pair_line_number_symbol, peek_char_symbol, pi_symbol, port_filename_symbol, port_line_number_symbol,
             procedure_documentation_symbol, procedure_signature_symbol, procedure_source_symbol, provide_symbol,
             quotient_symbol, 
             random_state_symbol, random_state_to_list_symbol, random_symbol, rationalize_symbol, read_byte_symbol,
             read_char_symbol, read_line_symbol, read_string_symbol, read_symbol, real_part_symbol, remainder_symbol,
             require_symbol, reverse_symbol, reverseb_symbol, rootlet_symbol, round_symbol,
             set_car_symbol, set_cdr_symbol, sin_symbol, sinh_symbol, sort_symbol, sqrt_symbol,
             stacktrace_symbol, string_append_symbol, string_downcase_symbol, string_eq_symbol, string_fill_symbol,
             string_geq_symbol, string_gt_symbol, string_leq_symbol, string_lt_symbol, string_position_symbol, string_ref_symbol,
             string_set_symbol, string_symbol, string_to_number_symbol, string_to_symbol_symbol, string_upcase_symbol,
             sublet_symbol, substring_symbol, subtract_symbol, symbol_access_symbol, symbol_symbol, symbol_to_dynamic_value_symbol,
             symbol_to_keyword_symbol, symbol_to_string_symbol, symbol_to_value_symbol, s7_version_symbol,
             tan_symbol, tanh_symbol, throw_symbol, string_to_byte_vector_symbol, 
             tree_count_symbol, truncate_symbol, type_of_symbol,
             unlet_symbol, 
             values_symbol, varlet_symbol, vector_append_symbol, vector_dimensions_symbol, vector_fill_symbol, vector_ref_symbol,
             vector_set_symbol, vector_symbol, 
             with_input_from_file_symbol, with_input_from_string_symbol, with_output_to_file_symbol, with_output_to_string_symbol,
             write_byte_symbol, write_char_symbol, write_string_symbol, write_symbol;
         
#if (!WITH_PURE_S7)
  s7_pointer is_char_ready_symbol, char_ci_leq_symbol, char_ci_lt_symbol, char_ci_eq_symbol, char_ci_geq_symbol, char_ci_gt_symbol, 
             let_to_list_symbol, integer_length_symbol, string_ci_leq_symbol, string_ci_lt_symbol, string_ci_eq_symbol,
             string_ci_geq_symbol, string_ci_gt_symbol, string_to_list_symbol, vector_to_list_symbol, string_length_symbol, 
             string_copy_symbol, list_to_string_symbol, list_to_vector_symbol, vector_length_symbol, make_polar_symbol, 
             make_rectangular_symbol;
#endif

  /* *s7* fields */
  s7_pointer stack_top_symbol, heap_size_symbol, gc_freed_symbol, gc_protected_objects_symbol,
             free_heap_size_symbol, file_names_symbol, symbol_table_symbol, cpu_time_symbol, c_objects_symbol, float_format_precision_symbol,
             stack_size_symbol, rootlet_size_symbol, c_types_symbol, safety_symbol, max_stack_size_symbol, gc_stats_symbol, autoloading_symbol,
             strings_symbol, vectors_symbol, input_ports_symbol, output_ports_symbol, continuations_symbol, hash_tables_symbol, gensyms_symbol,
             catches_symbol, exits_symbol, stack_symbol, default_rationalize_error_symbol, max_string_length_symbol, default_random_state_symbol,
             max_list_length_symbol, max_vector_length_symbol, max_vector_dimensions_symbol, default_hash_table_length_symbol, profile_info_symbol,
             hash_table_float_epsilon_symbol, morally_equal_float_epsilon_symbol, initial_string_port_length_symbol, memory_usage_symbol,
             undefined_identifier_warnings_symbol, print_length_symbol, bignum_precision_symbol, stacktrace_defaults_symbol, history_symbol, history_size_symbol;

  /* syntax symbols et al */
  s7_pointer else_symbol, lambda_symbol, lambda_star_symbol, let_symbol, quote_symbol, unquote_symbol, macroexpand_symbol, 
             define_expansion_symbol, baffle_symbol, with_let_symbol, documentation_symbol, signature_symbol, if_symbol, 
             when_symbol, unless_symbol, begin_symbol, cond_symbol, case_symbol, and_symbol, or_symbol, do_symbol, 
             define_symbol, define_star_symbol, define_constant_symbol, with_baffle_symbol, define_macro_symbol, 
             define_macro_star_symbol, define_bacro_symbol, define_bacro_star_symbol, letrec_symbol, letrec_star_symbol, 
             let_star_symbol, key_rest_symbol, key_allow_other_keys_symbol, key_readable_symbol, value_symbol, type_symbol, 
             baffled_symbol, __func___symbol, set_symbol, body_symbol, class_name_symbol, feed_to_symbol, format_error_symbol, 
             wrong_number_of_args_symbol, read_error_symbol, string_read_error_symbol, syntax_error_symbol, division_by_zero_symbol, 
             no_catch_symbol, io_error_symbol, invalid_escape_function_symbol, wrong_type_arg_symbol, out_of_range_symbol, missing_method_symbol;

  /* optimizer symbols */
  s7_pointer and_ap_symbol, and_az_symbol, and_p_symbol, and_safe_aa_symbol, and_safe_p_symbol, 
             and_unchecked_symbol, begin1_symbol, case_a_symbol, case_unchecked_symbol,
             cond_all_x_2_symbol, cond_all_x_symbol, cond_all_x_z_symbol, cond_simple_symbol, cond_unchecked_symbol, cond_unchecked_z_symbol,
             decrement_1_symbol, define_constant_unchecked_symbol, define_funchecked_symbol, define_star_unchecked_symbol, define_unchecked_symbol,
             do_unchecked_symbol, dotimes_p_symbol, dox_symbol, cond_feed_symbol, 
             if_unchecked_symbol,

             case_a_e_s_symbol, case_a_i_s_symbol, case_a_g_s_symbol, case_a_e_g_symbol, case_a_g_g_symbol, 
             case_s_e_s_symbol, case_s_i_s_symbol, case_s_g_s_symbol, case_s_e_g_symbol, case_s_g_g_symbol, 
             case_p_e_s_symbol, case_p_i_s_symbol, case_p_g_s_symbol, case_p_e_g_symbol, case_p_g_g_symbol, 

             if_a_p_p_symbol, if_a_p_symbol, if_a_r_symbol, if_a_n_n_symbol, if_a_n_symbol, 
             if_c_p_p_symbol, if_c_p_symbol, if_c_r_symbol, if_c_n_n_symbol, if_c_n_symbol,
             if_cs_p_p_symbol, if_cs_p_symbol, if_cs_r_symbol, if_cs_n_n_symbol, if_cs_n_symbol, 
             if_csc_p_p_symbol, if_csc_p_symbol, if_csc_r_symbol, if_csc_n_n_symbol, if_csc_n_symbol,
             if_csq_p_p_symbol, if_csq_p_symbol, if_csq_r_symbol, if_csq_n_n_symbol, if_csq_n_symbol,
             if_css_p_p_symbol, if_css_p_symbol, if_css_r_symbol, if_css_n_n_symbol, if_css_n_symbol,
             if_is_type_s_p_p_symbol, if_is_type_s_p_symbol, if_is_type_s_r_symbol, if_is_type_s_n_n_symbol, if_is_type_s_n_symbol,
             if_is_type_opsq_p_p_symbol, if_is_type_opsq_p_symbol, if_is_type_opsq_r_symbol, if_is_type_opsq_n_n_symbol, if_is_type_opsq_n_symbol,
             if_opsq_p_p_symbol, if_opsq_p_symbol, if_opsq_r_symbol, if_opsq_n_n_symbol, if_opsq_n_symbol,
             if_and2_p_p_symbol, if_and2_p_symbol, if_and2_r_symbol, if_and2_n_n_symbol, if_and2_n_symbol,
             if_and3_p_p_symbol, if_and3_p_symbol, if_and3_r_symbol, if_and3_n_n_symbol, if_and3_n_symbol,
             if_andp_p_p_symbol, if_andp_p_symbol, if_andp_r_symbol, if_andp_n_n_symbol, if_andp_n_symbol,
             if_or2_p_p_symbol, if_or2_p_symbol, if_or2_r_symbol, if_or2_n_n_symbol, if_or2_n_symbol,
             if_orp_p_p_symbol, if_orp_p_symbol, if_orp_r_symbol, if_orp_n_n_symbol, if_orp_n_symbol,
             if_p_p_p_symbol, if_p_p_symbol, if_p_r_symbol, if_p_n_n_symbol, if_p_n_symbol,
             if_s_opcq_p_p_symbol, if_s_opcq_p_symbol, if_s_opcq_r_symbol, if_s_opcq_n_n_symbol, if_s_opcq_n_symbol,
             if_s_p_p_symbol, if_s_p_symbol, if_s_r_symbol, if_s_n_n_symbol, if_s_n_symbol,
             if_z_p_p_symbol, if_z_p_symbol, if_z_r_symbol, if_z_n_n_symbol, if_z_n_symbol,

             increment_1_symbol, increment_sa_symbol, increment_saa_symbol, increment_ss_symbol,
             increment_sss_symbol, increment_sz_symbol, lambda_star_unchecked_symbol, lambda_unchecked_symbol, let_all_c_symbol,
             let_all_opsq_symbol, let_all_s_symbol, let_all_x_symbol, let_c_symbol, let_no_vars_symbol, let_one_symbol,
             let_opcq_symbol, let_opsq_p_symbol, let_opsq_symbol, let_car_symbol, let_opssq_symbol, let_opssq_e_symbol, let_opassq_e_symbol, 
             let_s_symbol, let_s_z_symbol, let_star2_symbol, let_a_symbol, let_a_z_symbol,
             let_star_all_x_symbol, let_star_a2_symbol, let_star_unchecked_symbol, let_unchecked_symbol, let_z_symbol, letrec_star_unchecked_symbol,
             letrec_unchecked_symbol, named_let_no_vars_symbol, named_let_star_symbol, named_let_symbol, let_temporarily_unchecked_symbol,

             or_ap_symbol, or_p_symbol, or_safe_aa_symbol, or_safe_p_symbol,
             or_unchecked_symbol, quote_unchecked_symbol, safe_do_symbol, safe_dotimes_symbol, set_cons_symbol, set_let_all_x_symbol,
             set_let_s_symbol, set_normal_symbol, set_pair_a_symbol, set_pair_p_symbol,
             set_pair_symbol, set_dilambda_symbol, set_dilambda_z_symbol, set_pair_z_symbol, set_pair_za_symbol, 
             set_pws_symbol, set_symbol_a_symbol, set_symbol_c_symbol,
             set_symbol_opcq_symbol, set_symbol_opsq_symbol, set_symbol_opssq_symbol, set_symbol_opsssq_symbol, set_symbol_p_symbol,
             set_symbol_oplq_symbol, set_symbol_opllq_symbol, 
             set_symbol_q_symbol, set_symbol_s_symbol, set_symbol_l_symbol, set_symbol_z_symbol, set_unchecked_symbol, 
             simple_do_symbol, unless_s_symbol, unless_a_symbol, unless_unchecked_symbol, 
             when_s_symbol, when_a_symbol, when_p_symbol, when_unchecked_symbol, 
             with_baffle_unchecked_symbol, with_let_s_symbol, with_let_unchecked_symbol, 
             dox_slot_symbol, 
             do_no_vars_symbol, do_no_vars_no_opt_symbol, dotimes_one_step_symbol;

#if WITH_GMP
  s7_pointer bignum_symbol, is_bignum_symbol;
  gc_list *bigints, *bigratios, *bigreals, *bignumbers;
#endif

#if WITH_SYSTEM_EXTRAS
  s7_pointer is_directory_symbol, file_exists_symbol, delete_file_symbol, getenv_symbol, system_symbol, directory_to_list_symbol, file_mtime_symbol;
#endif

  /* setter and quasiquote functions */
  s7_pointer vector_set_function, string_set_function, list_set_function, hash_table_set_function, let_set_function, object_set_function,
             apply_values_function, multivector_function, apply_function, vector_function, last_function;

  s7_pointer wrong_type_arg_info, out_of_range_info, simple_wrong_type_arg_info, simple_out_of_range_info;
  s7_pointer err_wrap1, err_wrap2;
  s7_pointer too_many_arguments_string, not_enough_arguments_string, division_by_zero_error_string, missing_method_string;
  s7_pointer *syn_docs; /* prebuilt evaluator arg lists, syntax doc strings */
  #define NUM_SAFE_LISTS 64
  s7_pointer safe_lists[NUM_SAFE_LISTS];
  int32_t current_safe_list;

  s7_pointer autoload_table, libraries, profile_info;
  const char ***autoload_names;
  int32_t *autoload_names_sizes;
  bool **autoloaded_already;
  int32_t autoload_names_loc, autoload_names_top;
  port_t *port_heap;

  int32_t format_depth;
  int32_t slash_str_size;
  char *slash_str;

  bool undefined_identifier_warnings;
  hash_entry_t *optimizer_fixups;

  opt_info *free_opts;
  jmp_buf opt_exit;
  int32_t pc, funcalls, unwraps;
  #define OPTS_SIZE 128      /* 128 overflows twice in s7test, 64 overflows 4 times in s7test, once in tall! */
  opt_info *opts[OPTS_SIZE]; /* this form is a lot faster than opt_info**! */
};

/* (*s7* 'safety) settings */
#define NO_SAFETY 0
#define CLM_OPTIMIZATION_SAFETY 1
#define ALL_OPTIMIZATION_SAFETY 2
#define IMMUTABLE_VECTOR_SAFETY 3

typedef enum {USE_DISPLAY, USE_WRITE, USE_READABLE_WRITE, USE_WRITE_WRONG} use_write_t;

#define INITIAL_AUTOLOAD_NAMES_SIZE 4

#define INITIAL_STORED_OPTLISTS_SIZE 8
static s7_pointer *stored_optlists = NULL;
static int32_t stored_optlists_size, stofl_loc = -1;
static int32_t *stofl = NULL;

static s7_pointer prepackaged_type_names[NUM_TYPES];

static bool t_number_p[NUM_TYPES], t_real_p[NUM_TYPES], t_rational_p[NUM_TYPES], t_big_number_p[NUM_TYPES];
static bool t_simple_p[NUM_TYPES], t_structure_p[NUM_TYPES];
static bool t_any_macro_p[NUM_TYPES], t_any_closure_p[NUM_TYPES], t_has_closure_let[NUM_TYPES];
static bool t_mappable_p[NUM_TYPES], t_sequence_p[NUM_TYPES], t_vector_p[NUM_TYPES];
static bool t_procedure_p[NUM_TYPES], t_applicable_p[NUM_TYPES];

static void init_types(void)
{
  int32_t i;
  for (i = 0; i < NUM_TYPES; i++)
    {
      t_number_p[i] = false;
      t_real_p[i] = false;
      t_rational_p[i] = false;
      t_simple_p[i] = false;
      t_structure_p[i] = false;
      t_any_macro_p[i] = false;
      t_any_closure_p[i] = false;
      t_has_closure_let[i] = false;
      t_sequence_p[i] = false;
      t_mappable_p[i] = false;
      t_vector_p[i] = false;
      t_applicable_p[i] = false;
      t_procedure_p[i] = false;
    }
  t_number_p[T_INTEGER] = true;
  t_number_p[T_RATIO] = true;
  t_number_p[T_REAL] = true;
  t_number_p[T_COMPLEX] = true;

  t_rational_p[T_INTEGER] = true;
  t_rational_p[T_RATIO] = true;

  t_real_p[T_INTEGER] = true;
  t_real_p[T_RATIO] = true;
  t_real_p[T_REAL] = true;

  t_big_number_p[T_BIG_INTEGER] = true;
  t_big_number_p[T_BIG_RATIO] = true;
  t_big_number_p[T_BIG_REAL] = true;
  t_big_number_p[T_BIG_COMPLEX] = true;

  t_structure_p[T_PAIR] = true;
  t_structure_p[T_VECTOR] = true;
  t_structure_p[T_HASH_TABLE] = true;
  t_structure_p[T_SLOT] = true;
  t_structure_p[T_LET] = true;
  t_structure_p[T_ITERATOR] = true;

  t_sequence_p[T_NIL] = true;
  t_sequence_p[T_PAIR] = true;
  t_sequence_p[T_STRING] = true;
  t_sequence_p[T_VECTOR] = true;
  t_sequence_p[T_INT_VECTOR] = true;
  t_sequence_p[T_FLOAT_VECTOR] = true;
  t_sequence_p[T_HASH_TABLE] = true;
  t_sequence_p[T_LET] = true;
  t_sequence_p[T_C_OBJECT] = true;

  t_mappable_p[T_PAIR] = true;
  t_mappable_p[T_STRING] = true;
  t_mappable_p[T_VECTOR] = true;
  t_mappable_p[T_INT_VECTOR] = true;
  t_mappable_p[T_FLOAT_VECTOR] = true;
  t_mappable_p[T_HASH_TABLE] = true;
  t_mappable_p[T_LET] = true;
  t_mappable_p[T_C_OBJECT] = true;
  t_mappable_p[T_ITERATOR] = true;
  t_mappable_p[T_C_MACRO] = true;
  t_mappable_p[T_MACRO] = true;
  t_mappable_p[T_BACRO] = true;
  t_mappable_p[T_MACRO_STAR] = true;
  t_mappable_p[T_BACRO_STAR] = true;
  t_mappable_p[T_CLOSURE] = true;
  t_mappable_p[T_CLOSURE_STAR] = true;

  t_vector_p[T_VECTOR] = true;
  t_vector_p[T_INT_VECTOR] = true;
  t_vector_p[T_FLOAT_VECTOR] = true;

  t_applicable_p[T_PAIR] = true;
  t_applicable_p[T_STRING] = true;
  t_applicable_p[T_VECTOR] = true;
  t_applicable_p[T_INT_VECTOR] = true;
  t_applicable_p[T_FLOAT_VECTOR] = true;
  t_applicable_p[T_HASH_TABLE] = true;
  t_applicable_p[T_ITERATOR] = true;
  t_applicable_p[T_LET] = true;
  t_applicable_p[T_C_OBJECT] = true;
  t_applicable_p[T_C_MACRO] = true;
  t_applicable_p[T_MACRO] = true;
  t_applicable_p[T_BACRO] = true;
  t_applicable_p[T_MACRO_STAR] = true;
  t_applicable_p[T_BACRO_STAR] = true;
  t_applicable_p[T_SYNTAX] = true;
  t_applicable_p[T_C_FUNCTION] = true;
  t_applicable_p[T_C_FUNCTION_STAR] = true;
  t_applicable_p[T_C_ANY_ARGS_FUNCTION] = true;
  t_applicable_p[T_C_OPT_ARGS_FUNCTION] = true;
  t_applicable_p[T_C_RST_ARGS_FUNCTION] = true;
  t_applicable_p[T_CLOSURE] = true;
  t_applicable_p[T_CLOSURE_STAR] = true;
  t_applicable_p[T_GOTO] = true;
  t_applicable_p[T_CONTINUATION] = true;

  /* t_procedure_p[T_C_OBJECT] = true; */ 
  t_procedure_p[T_C_FUNCTION] = true;
  t_procedure_p[T_C_FUNCTION_STAR] = true;
  t_procedure_p[T_C_ANY_ARGS_FUNCTION] = true;
  t_procedure_p[T_C_OPT_ARGS_FUNCTION] = true;
  t_procedure_p[T_C_RST_ARGS_FUNCTION] = true;
  t_procedure_p[T_CLOSURE] = true;
  t_procedure_p[T_CLOSURE_STAR] = true;
  t_procedure_p[T_GOTO] = true;
  t_procedure_p[T_CONTINUATION] = true;

  t_any_macro_p[T_C_MACRO] = true;
  t_any_macro_p[T_MACRO] = true;
  t_any_macro_p[T_BACRO] = true;
  t_any_macro_p[T_MACRO_STAR] = true;
  t_any_macro_p[T_BACRO_STAR] = true;

  t_any_closure_p[T_CLOSURE] = true;
  t_any_closure_p[T_CLOSURE_STAR] = true;

  t_has_closure_let[T_MACRO] = true;
  t_has_closure_let[T_BACRO] = true;
  t_has_closure_let[T_MACRO_STAR] = true;
  t_has_closure_let[T_BACRO_STAR] = true;
  t_has_closure_let[T_CLOSURE] = true;
  t_has_closure_let[T_CLOSURE_STAR] = true;

  t_simple_p[T_NIL] = true;
  t_simple_p[T_UNDEFINED] = true;
  t_simple_p[T_EOF_OBJECT] = true;
  t_simple_p[T_BOOLEAN] = true;
  t_simple_p[T_CHARACTER] = true;
  t_simple_p[T_SYMBOL] = true;
  t_simple_p[T_SYNTAX] = true;
  t_simple_p[T_C_MACRO] = true;
  t_simple_p[T_C_FUNCTION] = true;
  t_simple_p[T_C_FUNCTION_STAR] = true;
  t_simple_p[T_C_ANY_ARGS_FUNCTION] = true;
  t_simple_p[T_C_OPT_ARGS_FUNCTION] = true;
  t_simple_p[T_C_RST_ARGS_FUNCTION] = true;
  /* not completely sure about the next ones */
  t_simple_p[T_LET] = true;
  t_simple_p[T_INPUT_PORT] = true;
  t_simple_p[T_OUTPUT_PORT] = true;
}

#if WITH_HISTORY
#define current_code(Sc)           car(Sc->cur_code)
#define set_current_code(Sc, Code) do {Sc->cur_code = cdr(Sc->cur_code); set_car(Sc->cur_code, Code);} while (0)
#define mark_current_code(Sc)      do {int32_t i; s7_pointer p; for (p = Sc->cur_code, i = 0; i < sc->history_size; i++, p = cdr(p)) S7_MARK(car(p));} while (0)
#else
#define current_code(Sc)           Sc->cur_code
#define set_current_code(Sc, Code) Sc->cur_code = Code
#define mark_current_code(Sc)      S7_MARK(Sc->cur_code)
#endif

#define typeflag(p)  ((p)->tf.flag)
#define typesflag(p) ((p)->tf.sflag)

static s7_scheme *cur_sc = NULL;

#if DEBUGGING
  static const char *check_name(int32_t typ);
  static s7_pointer check_seti(s7_scheme *sc, s7_pointer x, const char *func, int32_t line);
  static s7_pointer check_ref(s7_pointer p, uint8_t expected_type, const char *func, int32_t line, const char *func1, const char *func2);
  static s7_pointer check_ref2(s7_pointer p, uint8_t expected_type, int32_t other_type, const char *func, int32_t line, const char *func1, const char *func2);
  static s7_pointer check_ref3(s7_pointer p, const char *func, int32_t line);
  static s7_pointer check_ref4(s7_pointer p, const char *func, int32_t line);
  static s7_pointer check_ref5(s7_pointer p, const char *func, int32_t line);
  static s7_pointer check_ref6(s7_pointer p, const char *func, int32_t line);
  static s7_pointer check_ref7(s7_pointer p, const char *func, int32_t line);
  static s7_pointer check_ref8(s7_pointer p, const char *func, int32_t line);
  static s7_pointer check_ref9(s7_pointer p, const char *func, int32_t line);
  static s7_pointer check_ref10(s7_pointer p, const char *func, int32_t line);
  static s7_pointer check_ref11(s7_pointer p, const char *func, int32_t line);
  static s7_pointer check_nref(s7_pointer p, const char *func, int32_t line);
  static s7_pointer check_cell(s7_pointer p, const char *func, int32_t line);
  static void print_gc_info(s7_pointer obj, int32_t line);

  static s7_pointer opt1_1(s7_scheme *sc, s7_pointer p, uint32_t role, const char *func, int32_t line);
  static s7_pointer set_opt1_1(s7_scheme *sc, s7_pointer p, s7_pointer x, uint32_t role, const char *func, int32_t line);
  static s7_pointer opt2_1(s7_scheme *sc, s7_pointer p, uint32_t role, const char *func, int32_t line);
  static void set_opt2_1(s7_scheme *sc, s7_pointer p, s7_pointer x, uint32_t role, const char *func, int32_t line);
  static s7_pointer opt3_1(s7_scheme *sc, s7_pointer p, uint32_t role, const char *func, int32_t line);
  static void set_opt3_1(s7_scheme *sc, s7_pointer p, s7_pointer x, uint32_t role, const char *func, int32_t line);

  static uint64_t s_hash_1(s7_scheme *sc, s7_pointer p, const char *func, int32_t line);
  static void set_s_hash_1(s7_scheme *sc, s7_pointer p, uint64_t x, const char *func, int32_t line);
  static const char *s_name_1(s7_scheme *sc, s7_pointer p, const char *func, int32_t line);
  static void set_s_name_1(s7_scheme *sc, s7_pointer p, const char *str, const char *func, int32_t line);
  static uint32_t s_line_1(s7_scheme *sc, s7_pointer p, const char *func, int32_t line);
  static void set_s_line_1(s7_scheme *sc, s7_pointer p, uint32_t x, const char *func, int32_t line);
  static uint32_t s_len_1(s7_scheme *sc, s7_pointer p, const char *func, int32_t line);
  static void set_s_len_1(s7_scheme *sc, s7_pointer p, uint32_t x, const char *func, int32_t line);
  static uint32_t s_op_1(s7_scheme *sc, s7_pointer p, const char *func, int32_t line);
  static void set_s_op_1(s7_scheme *sc, s7_pointer p, uint32_t x, const char *func, int32_t line);
  static uint32_t s_syn_op_1(s7_scheme *sc, s7_pointer p, const char *func, int32_t line);
  static void set_s_syn_op_1(s7_scheme *sc, s7_pointer p, uint32_t x, const char *func, int32_t line);

  #define unchecked_type(p)           ((p)->tf.type_field)
#if WITH_GCC
  #define type(p) ({uint8_t _t_; _t_ = (p)->tf.type_field; if (((_t_ == T_FREE)) || (_t_ >= NUM_TYPES)) print_gc_info(p, __LINE__); _t_;})
#else
  #define type(p) (p)->tf.type_field
#endif

  #define set_type(p, f)						\
    do {								\
      p->previous_alloc_line = p->current_alloc_line;			\
      p->previous_alloc_func = p->current_alloc_func;			\
      p->previous_alloc_type = p->current_alloc_type;			\
      p->current_alloc_line = __LINE__;					\
      p->current_alloc_func = __func__;					\
      p->current_alloc_type = f;					\
      p->uses++; 					                \
      if ((((f) & 0xff) == T_FREE) || (((f) & 0xff) >= NUM_TYPES))	\
        fprintf(stderr, "%d: set free %p type to %" PRIx64 "\n", __LINE__, p, (int64_t)(f)); \
      else								\
	{								\
	  if (((typeflag(p) & T_IMMUTABLE) != 0) && ((typeflag(p) != (f))))						\
	    {fprintf(stderr, "%s[%d]: set immutable %p type %d to %" PRId64 "\n", __func__, __LINE__, p, unchecked_type(p), (int64_t)(f)); abort();} \
	  if (((typeflag(p) & T_LINE_NUMBER) != 0) && (((typeflag(p)) & 0xff) == T_PAIR) && (((f) & T_LINE_NUMBER) == 0)) \
            fprintf(stderr, "%d unsets line_number\n", __LINE__); \
	}								\
      typeflag(p) = f;							\
    } while (0)

  /* these check most s7cell field references (and many type bits) for consistency */
  #define _TI(P)   check_ref(P, T_INTEGER,           __func__, __LINE__, NULL, NULL)
  #define _TR(P)   check_ref(P, T_REAL,              __func__, __LINE__, NULL, NULL)
  #define _TF(P)   check_ref2(P, T_RATIO, T_INTEGER, __func__, __LINE__, NULL, NULL)
  #define _TZ(P)   check_ref(P, T_COMPLEX,           __func__, __LINE__, NULL, NULL)
  #define _TBgi(P) check_ref(P, T_BIG_INTEGER,       __func__, __LINE__, "sweep", NULL)
  #define _TBgr(P) check_ref(P, T_BIG_REAL,          __func__, __LINE__, "sweep", NULL)
  #define _TBgf(P) check_ref(P, T_BIG_RATIO,         __func__, __LINE__, "sweep", NULL)
  #define _TBgz(P) check_ref(P, T_BIG_COMPLEX,       __func__, __LINE__, "sweep", NULL)

  #define _TChr(P) check_ref(P, T_CHARACTER,         __func__, __LINE__, NULL, NULL)
  #define _TCtr(P) check_ref(P, T_COUNTER,           __func__, __LINE__, NULL, NULL)
  #define _TOpt(P) check_ref(P, T_OPTLIST,           __func__, __LINE__, "free_optlist", NULL)
  #define _TPtr(P) check_ref(P, T_C_POINTER,         __func__, __LINE__, NULL, NULL)
  #define _TBfl(P) check_ref(P, T_BAFFLE,            __func__, __LINE__, NULL, NULL)
  #define _TGot(P) check_ref(P, T_GOTO,              __func__, __LINE__, NULL, NULL)
  #define _TStk(P) check_ref(P, T_STACK,             __func__, __LINE__, NULL, NULL)
  #define _TPair(P) check_ref(P, T_PAIR,             __func__, __LINE__, NULL, NULL)
  #define _TCat(P) check_ref(P, T_CATCH,             __func__, __LINE__, NULL, NULL)
  #define _TDyn(P) check_ref(P, T_DYNAMIC_WIND,      __func__, __LINE__, NULL, NULL)
  #define _TSlt(P) check_ref(P, T_SLOT,              __func__, __LINE__, NULL, NULL)
  #define _TSlp(P) check_ref2(P, T_SLOT, T_PAIR,     __func__, __LINE__, NULL, NULL)
  #define _TSln(P) check_ref2(P, T_SLOT, T_NIL,      __func__, __LINE__, NULL, NULL)
  #define _TSld(P) check_ref2(P, T_SLOT, T_UNDEFINED,__func__, __LINE__, NULL, NULL)
  #define _TSyn(P) check_ref(P, T_SYNTAX,            __func__, __LINE__, NULL, NULL)
  #define _TMac(P) check_ref(P, T_C_MACRO,           __func__, __LINE__, NULL, NULL)
  #define _TLet(P) check_ref(P, T_LET,               __func__, __LINE__, NULL, NULL)
  #define _TLid(P) check_ref2(P, T_LET, T_NIL,       __func__, __LINE__, NULL, NULL)
  #define _TRan(P) check_ref(P, T_RANDOM_STATE,      __func__, __LINE__, NULL, NULL)
  #define _TLst(P) check_ref2(P, T_PAIR, T_NIL,      __func__, __LINE__, "gc", NULL)
  #define _TStr(P) check_ref(P, T_STRING,            __func__, __LINE__, "sweep", NULL)
  #define _TObj(P) check_ref(P, T_C_OBJECT,          __func__, __LINE__, "sweep", NULL)
  #define _THsh(P) check_ref(P, T_HASH_TABLE,        __func__, __LINE__, "sweep", "free_hash_table")
  #define _TItr(P) check_ref(P, T_ITERATOR,          __func__, __LINE__, "sweep", NULL)
  #define _TCon(P) check_ref(P, T_CONTINUATION,      __func__, __LINE__, "sweep", NULL)
  #define _TFvc(P) check_ref(P, T_FLOAT_VECTOR,      __func__, __LINE__, "sweep", NULL)
  #define _TIvc(P) check_ref(P, T_INT_VECTOR,        __func__, __LINE__, "sweep", NULL)
  #define _TSym(P) check_ref(P, T_SYMBOL,            __func__, __LINE__, "sweep", "remove_gensym_from_symbol_table")
  #define _TExp(P) check_ref2(P, T_MACRO, T_SYMBOL,  __func__, __LINE__, NULL, NULL)

  #define _TPrt(P) check_ref3(P,                     __func__, __LINE__) /* input|output_port, or free */
  #define _TVec(P) check_ref4(P,                     __func__, __LINE__) /* any vector or free */
  #define _TClo(P) check_ref5(P,                     __func__, __LINE__) /* has closure let */
  #define _TFnc(P) check_ref6(P,                     __func__, __LINE__) /* any c_function|c_macro */
  #define _TNum(P) check_ref7(P,                     __func__, __LINE__) /* any number (not bignums I think) */
  #define _TSeq(P) check_ref8(P,                     __func__, __LINE__) /* any sequence or structure */
  #define _TMet(P) check_ref9(P,                     __func__, __LINE__) /* anything that might contain a method */
  #define _TArg(P) check_ref10(P,                    __func__, __LINE__) /* closure arg (list, symbol) */
  #define _TApp(P) check_ref11(P,                    __func__, __LINE__) /* setter (any_procedure or #f) */
  #define _NFre(P) check_nref(P,                     __func__, __LINE__) /* not free */
  #define _Cell(P) check_cell(P,                     __func__, __LINE__) /* any cell */
  #define _TSet(P) check_seti(sc, P,                 __func__, __LINE__) /* set of immutable value */

#else
  #define unchecked_type(p)           ((p)->tf.type_field)
  #define type(p)                     ((p)->tf.type_field)
  #define set_type(p, f)              typeflag(p) = f
  #define _TSet(P)                    P
  #define _TI(P)                      P
  #define _TR(P)                      P
  #define _TF(P)                      P
  #define _TZ(P)                      P
  #define _TBgi(P)                    P
  #define _TBgr(P)                    P
  #define _TBgf(P)                    P
  #define _TBgz(P)                    P
  #define _TStr(P)                    P
  #define _TSyn(P)                    P
  #define _TChr(P)                    P
  #define _TObj(P)                    P
  #define _TCtr(P)                    P
  #define _TOpt(P)                    P
  #define _THsh(P)                    P
  #define _TItr(P)                    P
  #define _TPtr(P)                    P
  #define _TBfl(P)                    P
  #define _TGot(P)                    P
  #define _TCon(P)                    P
  #define _TStk(P)                    P
  #define _TPrt(P)                    P
  #define _TIvc(P)                    P
  #define _TFvc(P)                    P
  #define _TVec(P)                    P
  #define _TPair(P)                   P
  #define _TRan(P)                    P
  #define _TDyn(P)                    P
  #define _TCat(P)                    P
  #define _TClo(P)                    P
  #define _TFnc(P)                    P
  #define _TSlt(P)                    P
  #define _TSln(P)                    P
  #define _TSld(P)                    P
  #define _TSlp(P)                    P
  #define _TSym(P)                    P
  #define _TLet(P)                    P
  #define _TLid(P)                    P
  #define _TLst(P)                    P
  #define _TNum(P)                    P
  #define _TSeq(P)                    P
  #define _TMet(P)                    P
  #define _TMac(P)                    P
  #define _TArg(P)                    P
  #define _TApp(P)                    P
  #define _TExp(P)                    P
  #define _NFre(P)                    P
  #define _Cell(P)                    P
#endif

#define is_number(P)                  t_number_p[type(P)]
#define is_integer(P)                 (type(P) == T_INTEGER)
#define is_real(P)                    t_real_p[type(P)]
#define is_rational(P)                t_rational_p[type(P)]
#define is_big_number(p)              t_big_number_p[type(p)]
#define is_t_integer(p)               (type(p) == T_INTEGER)
#define is_t_ratio(p)                 (type(p) == T_RATIO)
#define is_t_real(p)                  (type(p) == T_REAL)
#define is_t_complex(p)               (type(p) == T_COMPLEX)
#define is_t_big_integer(p)           (type(p) == T_BIG_INTEGER)
#define is_t_big_ratio(p)             (type(p) == T_BIG_RATIO)
#define is_t_big_real(p)              (type(p) == T_BIG_REAL)
#define is_t_big_complex(p)           (type(p) == T_BIG_COMPLEX)
#define is_float(p)                   ((is_real(p)) && (!is_rational(p)))

#define is_free(p)                    (type(p) == T_FREE)
#define is_free_and_clear(p)          (typeflag(p) == T_FREE)
#define is_simple(P)                  t_simple_p[type(P)]
#define has_structure(P)              t_structure_p[type(P)]

#define is_any_macro(P)               t_any_macro_p[type(P)]
#define is_any_closure(P)             t_any_closure_p[type(P)]
#define is_procedure_or_macro(P)      ((t_procedure_p[type(P)]) || (t_any_macro_p[type(P)]))
#define is_any_procedure(P)           (type(P) >= T_CLOSURE)
#define has_closure_let(P)            t_has_closure_let[type(P)]

#define is_simple_sequence(P)         (t_sequence_p[type(P)])
#define is_sequence(P)                ((t_sequence_p[type(P)]) || (has_methods(P)))
#define is_mappable(P)                (t_mappable_p[type(P)])
#define is_applicable(P)              (t_applicable_p[type(P)])
/* this misses #() which actually is not applicable to anything, probably "" also, and inapplicable c-objects like random-state */
#define is_procedure(p)               ((t_procedure_p[type(p)]) || ((is_c_object(p)) && (is_safe_procedure(p))))


/* the layout of these bits does matter in several cases -- don't shadow SYNTACTIC_PAIR and OPTIMIZED_PAIR */
#define TYPE_BITS                     8

#define T_KEYWORD                     (1 << (TYPE_BITS + 0))
#define is_keyword(p)                 ((typesflag(_NFre(p)) & T_KEYWORD) != 0)
/* this bit distinguishes a symbol from a symbol that is also a keyword
 * this should be ok in the second byte because keywords are constants in s7 (never syntax)
 */

#define T_SYNTACTIC                   (1 << (TYPE_BITS + 1))
#define is_syntactic(p)               ((typesflag(_NFre(p)) & T_SYNTACTIC) != 0)
#define is_syntactic_symbol(p)        ((typesflag(_NFre(p)) & (T_SYNTACTIC | 0xff)) == (T_SYMBOL | T_SYNTACTIC))
#define SYNTACTIC_TYPE                (uint16_t)(T_SYMBOL | T_DONT_EVAL_ARGS | T_SYNTACTIC)
#define SYNTACTIC_PAIR                (uint16_t)(T_PAIR | T_SYNTACTIC)
/* this marks symbols that represent syntax objects, it should be in the second byte */
#define set_syntactic_pair(p)         typeflag(_TPair(p)) = (SYNTACTIC_PAIR | (typeflag(p) & 0xffff0000))


#define T_SIMPLE_ARG_DEFAULTS         (1 << (TYPE_BITS + 2))
#define has_simple_arg_defaults(p)    ((typeflag(_TPair(p)) & T_SIMPLE_ARG_DEFAULTS) != 0)
#define set_simple_arg_defaults(p)    typeflag(_TPair(p)) |= T_SIMPLE_ARG_DEFAULTS
/* are all lambda* default values simple? */

#define T_LIST_IN_USE                 T_SIMPLE_ARG_DEFAULTS
#define list_is_in_use(p)             ((typeflag(_TPair(p)) & T_LIST_IN_USE) != 0)
#define set_list_in_use(p)            typeflag(_TPair(p)) |= T_LIST_IN_USE
#define clear_list_in_use(p)          typeflag(_TPair(p)) &= (~T_LIST_IN_USE)
/* these could all be one permanent list, indexed from inside, and this bit is never actually protecting anything across a call */


#define T_OPTIMIZED                   (1 << (TYPE_BITS + 3))
#define set_optimized(p)              typesflag(_TPair(p)) |= T_OPTIMIZED
#define clear_optimized(p)            typeflag(_TPair(p)) &= (~(T_OPTIMIZED | T_OVERLAY | T_SYNTACTIC | T_HAS_ALL_X))
#define OPTIMIZED_PAIR                (uint16_t)(T_PAIR | T_OPTIMIZED)
#define is_optimized(p)               (typesflag(p) == OPTIMIZED_PAIR)
#define unoptimize(p)                 typesflag(_TPair(p)) &= (~T_OPTIMIZED)
/*   this is faster than the bit extraction above and the same speed as xor */
/* optimizer flag for an expression that has optimization info, it should be in the second byte
 */

#define T_SCOPE_SAFE                  T_OPTIMIZED
#define is_scope_safe(p)              ((typeflag(_TFnc(p)) & T_SCOPE_SAFE) != 0)
#define set_scope_safe(p)             typeflag(_TFnc(p)) |= T_SCOPE_SAFE


#define T_SAFE_CLOSURE                (1 << (TYPE_BITS + 4))
#define is_safe_closure(p)            ((typesflag(_NFre(p)) & T_SAFE_CLOSURE) != 0)
#define set_safe_closure(p)           typesflag(p) |= T_SAFE_CLOSURE
#define clear_safe_closure(p)         typesflag(p) &= (~T_SAFE_CLOSURE)
/* optimizer flag for a closure body that is completely simple (every expression is safe)
 *   set_safe_closure happens only in optimize_lambda, clear only in procedure_source, bits only here
 *   this has to be separate from T_SAFE_PROCEDURE, and should be in the second byte.
 *   It can be set on either the body (a pair) or the closure itself.
 * define -> optimize_lambda sets safe -> define_funchecked -> make_funclet for the frame
 *   similarly, named let -> optimize_lambda, then let creates the frame if safe
 *   thereafter, optimizer uses OP_SAFE_CLOSURE* which calls old_frame*
 */

#define T_DONT_EVAL_ARGS              (1 << (TYPE_BITS + 5))
#define dont_eval_args(p)             ((typesflag(_NFre(p)) & T_DONT_EVAL_ARGS) != 0)
/* this marks things that don't evaluate their arguments */

#define T_EXPANSION                   (1 << (TYPE_BITS + 6))
#define is_expansion(p)               ((typesflag(_TExp(p)) & T_EXPANSION) != 0)
#define clear_expansion(p)            typesflag(_TSym(p)) &= (~T_EXPANSION)
/* this marks the symbol and its run-time macro value, distinguishing it from an ordinary macro */

#define T_MULTIPLE_VALUE              (1 << (TYPE_BITS + 7))
#define is_multiple_value(p)          ((typesflag(_NFre(p)) & T_MULTIPLE_VALUE) != 0)
#define set_multiple_value(p)         typesflag(_TPair(p)) |= T_MULTIPLE_VALUE
#define clear_multiple_value(p)       typesflag(_TPair(p)) &= (~T_MULTIPLE_VALUE)
#define multiple_value(p)             p
/* this bit marks a list (from "values") that is waiting for a
 *    chance to be spliced into its caller's argument list.  It is normally
 *    on only for a very short time.
 */

#define T_MATCHED                     T_MULTIPLE_VALUE
#define is_matched_pair(p)            ((typesflag(_TPair(p)) & T_MATCHED) != 0)
#define set_match_pair(p)             typesflag(_TPair(p)) |= T_MATCHED
#define clear_match_pair(p)           typesflag(_TPair(p)) &= (~T_MATCHED)
#define is_matched_symbol(p)          ((typesflag(_TSym(p)) & T_MATCHED) != 0)
#define set_match_symbol(p)           typesflag(_TSym(p)) |= T_MATCHED
#define clear_match_symbol(p)         typesflag(_TSym(p)) &= (~T_MATCHED)

#define T_GLOBAL                      (1 << (TYPE_BITS + 8))
#define is_global(p)                  ((typeflag(_TSym(p)) & T_GLOBAL) != 0)
#define set_global(p)                 do {if (!is_local(p)) typeflag(_TSym(p)) |= T_GLOBAL;} while (0)
#if 0
  /* to find who is stomping on our symbols: */
  static char *object_to_truncated_string(s7_scheme *sc, s7_pointer p, int32_t len);

  static void set_local_1(s7_scheme *sc, s7_pointer symbol, const char *func, int32_t line)
  {
    if (is_global(symbol)) /* || (is_syntactic(symbol))) */
      fprintf(stderr, "%s[%d]: %s%s%s in %s\n", func, line, BOLD_TEXT, DISPLAY(symbol), UNBOLD_TEXT, DISPLAY_80(sc->cur_code));
    typeflag(symbol) = (typeflag(symbol) & ~(T_DONT_EVAL_ARGS | T_GLOBAL | T_SYNTACTIC));
  }
  #define set_local(Symbol) set_local_1(sc, Symbol, __func__, __LINE__)
#else
#define set_local(p)                  typeflag(_TSym(p)) = ((typeflag(p) | T_LOCAL) & ~(T_DONT_EVAL_ARGS | T_GLOBAL | T_SYNTACTIC))
#endif
/* this (T_GLOBAL) marks something defined (bound) at the top-level, and never defined locally */

#define T_UNSAFE_DO                   T_GLOBAL
#define is_unsafe_do(p)               ((typeflag(_TPair(p)) & T_UNSAFE_DO) != 0)
#define set_unsafe_do(p)              typeflag(_TPair(p)) |= T_UNSAFE_DO
/* marks do-loops that resist optimization */

#define T_COLLECTED                   (1 << (TYPE_BITS + 9))
#define is_collected(p)               ((typeflag(_TSeq(p)) & T_COLLECTED) != 0)
#define set_collected(p)              typeflag(_TSeq(p)) |= T_COLLECTED
/* #define clear_collected(p)         typeflag(_TSeq(p)) &= (~T_COLLECTED) */
/* this is a transient flag used by the printer to catch cycles.  It affects only objects that have structure.  
 *   We can't use a low bit (bit 7 for example), because collect_shared_info inspects the object's type.
 */

#define T_LINE_NUMBER                 (1 << (TYPE_BITS + 10))
#define has_line_number(p)            ((typeflag(_TPair(p)) & T_LINE_NUMBER) != 0)
#define set_has_line_number(p)        typeflag(_TPair(p)) |= T_LINE_NUMBER
/* pair in question has line/file info added during read, or the environment has function placement info 
 *   this bit should not be in the first byte -- SYNTACTIC_PAIR ignores it.
 */

#define T_LOADER_PORT                 T_LINE_NUMBER
#define is_loader_port(p)             ((typeflag(_TPrt(p)) & T_LOADER_PORT) != 0)
#define set_loader_port(p)            typeflag(_TPrt(p)) |= T_LOADER_PORT
#define clear_loader_port(p)          typeflag(_TPrt(p)) &= (~T_LOADER_PORT)
/* to block random load-time reads from screwing up the load process, this bit marks a port used by the loader */

#define T_HAS_ACCESSOR                T_LINE_NUMBER
#define symbol_has_accessor(p)        ((typeflag(_TSym(p)) & T_HAS_ACCESSOR) != 0)
#define symbol_set_has_accessor(p)    typeflag(_TSym(p)) |= T_HAS_ACCESSOR
#define slot_has_accessor(p)          ((typeflag(_TSlt(p)) & T_HAS_ACCESSOR) != 0)
#define slot_set_has_accessor(p)      typeflag(_TSlt(p)) |= T_HAS_ACCESSOR
/* marks a slot or symbol that has a setter */

#define T_WITH_LET_LET                T_LINE_NUMBER
#define is_with_let_let(p)            ((typeflag(_TLet(p)) & T_WITH_LET_LET) != 0)
#define set_with_let_let(p)           typeflag(_TLet(p)) |= T_WITH_LET_LET
/* marks a let that is the argument to with-let */

#define T_SIMPLE_DEFAULTS             T_LINE_NUMBER
#define has_simple_defaults(p)        ((typeflag(_TFnc(p)) & T_SIMPLE_DEFAULTS) != 0)
#define set_simple_defaults(p)        typeflag(_TFnc(p)) |= T_SIMPLE_DEFAULTS
#define clear_simple_defaults(p)      typeflag(_TFnc(p)) &= (~T_SIMPLE_DEFAULTS)
/* flag c_func_star arg defaults that need GC protection */


#define T_SHARED                      (1 << (TYPE_BITS + 11))
#define is_shared(p)                  ((typeflag(_TSeq(p)) & T_SHARED) != 0)
#define set_shared(p)                 typeflag(_TSeq(p)) |= T_SHARED
#define clear_collected_and_shared(p) typeflag(p) &= (~(T_COLLECTED | T_SHARED)) /* this can clear free cells = calloc */
#define is_collected_or_shared(p)     ((typeflag(p) & (T_COLLECTED | T_SHARED)) != 0)

#define T_OVERLAY                     (1 << (TYPE_BITS + 12))
#define set_overlay(p)                typeflag(_TPair(p)) |= T_OVERLAY
#define is_overlaid(p)                ((typeflag(_TPair(p)) & T_OVERLAY) != 0)
#define clear_overlay(p)              typeflag(_TPair(p)) &= (~T_OVERLAY)
/* optimizer flag that marks a cell whose opt_back [ie opt1] points to the previous cell in a list */

#define T_LOCAL                       T_OVERLAY
#define is_local(p)                   ((typeflag(_TSym(p)) & T_LOCAL) != 0)
/* marks a symbol that has been used locally (needed to protect against incorrect set_globals later) */

#define T_SAFE_PROCEDURE              (1 << (TYPE_BITS + 13))
#define is_safe_procedure(p)          ((typeflag(_NFre(p)) & T_SAFE_PROCEDURE) != 0)
#define is_safe_or_scope_safe_procedure(p) ((typeflag(_TFnc(p)) & (T_SCOPE_SAFE | T_SAFE_PROCEDURE)) != 0)
/* applicable objects that do not return or modify their arg list directly (no :rest arg in particular),
 *    and that can't call apply themselves either directly or via s7_call, and that don't mess with the stack.
 */

#define T_CHECKED                     (1 << (TYPE_BITS + 14))
#define set_checked(p)                typeflag(_TPair(p)) |= T_CHECKED
#define is_checked(p)                 ((typeflag(_TPair(p)) & T_CHECKED) != 0)
#define clear_checked(p)              typeflag(_TPair(p)) &= (~T_CHECKED)
#define set_checked_slot(p)           typeflag(_TSlt(p)) |= T_CHECKED
#define is_checked_slot(p)            ((typeflag(_TSlt(p)) & T_CHECKED) != 0)
#define is_not_checked_slot(p)        ((typeflag(_TSlt(p)) & T_CHECKED) == 0)

#define T_UNSAFE                      (1 << (TYPE_BITS + 15))
#define set_unsafe(p)                 typeflag(_TPair(p)) |= T_UNSAFE
#define set_unsafely_optimized(p)     typeflag(_TPair(p)) = (typeflag(p) & (~T_OVERLAY)) | (T_UNSAFE | T_OPTIMIZED)
#define is_unsafe(p)                  ((typeflag(_TPair(p)) & T_UNSAFE) != 0)
#define clear_unsafe(p)               typeflag(_TPair(p)) &= (~T_UNSAFE)
#define is_safely_optimized(p)        ((typeflag(p) & (T_OPTIMIZED | T_UNSAFE)) == T_OPTIMIZED)
/* optimizer flag saying "this expression is not completely self-contained.  It might involve the stack, etc" */

#define T_CLEAN_SYMBOL                T_UNSAFE
#define is_clean_symbol(p)            ((typeflag(_TSym(p)) & T_CLEAN_SYMBOL) != 0)
#define set_clean_symbol(p)           typeflag(_TSym(p)) |= T_CLEAN_SYMBOL
/* set if we know the symbol name can be printed without quotes (slashification) */

#define T_HAS_STEPPER                 T_UNSAFE
#define has_stepper(p)                ((typeflag(_TSlt(p)) & T_HAS_STEPPER) != 0)
#define set_has_stepper(p)            typeflag(_TSlt(p)) |= T_HAS_STEPPER

#define T_IMMUTABLE                   (1 << (TYPE_BITS + 16))
#define is_immutable(p)               ((typeflag(_NFre(p)) & T_IMMUTABLE) != 0)
#define set_immutable(p)              typeflag(_NFre(p)) |= T_IMMUTABLE
#define is_immutable_port(p)          ((typeflag(_TPrt(p)) & T_IMMUTABLE) != 0)
#define is_immutable_symbol(p)        ((typeflag(_TSym(p)) & T_IMMUTABLE) != 0)
#define is_immutable_pair(p)          ((typeflag(_TPair(p)) & T_IMMUTABLE) != 0)
#define is_immutable_vector(p)        ((typeflag(_TVec(p)) & T_IMMUTABLE) != 0)
#define is_immutable_string(p)        ((typeflag(_TStr(p)) & T_IMMUTABLE) != 0)
/* immutable means the value can't be changed via set! or bind -- this is separate from the symbol access stuff
 * this bit can't be in the second byte -- with-let, for example, is immutable, but we use SYNTACTIC_TYPE to 
 * recognize syntax in do loop optimizations.
 */

#define T_SETTER                      (1 << (TYPE_BITS + 17))
#define set_setter(p)                 typeflag(_TSym(p)) |= T_SETTER
#define is_setter(p)                  ((typeflag(_TSym(p)) & T_SETTER) != 0)
/* optimizer flag for a procedure that sets some variable (set-car! for example). */

#define T_ALLOW_OTHER_KEYS            T_SETTER
#define set_allow_other_keys(p)       typeflag(_TPair(p)) |= T_ALLOW_OTHER_KEYS
#define allows_other_keys(p)          ((typeflag(_TPair(p)) & T_ALLOW_OTHER_KEYS) != 0)
/* marks arglist that allows keyword args other than those in the parameter list; can't allow
 *   (define* (f :allow-other-keys)...) because there's only one nil, and besides, it does say "other".
 */

/* this bit is already in use on pairs, but the contexts never overlap, I hope... I need more bits. */
#define T_HAS_ALL_X                   T_SETTER
#define set_has_all_x(p)              typeflag(_TPair(p)) |= T_HAS_ALL_X
#define has_all_x(p)                  ((typeflag(_TPair(p)) & T_HAS_ALL_X) != 0)
#define clear_has_all_x(p)            typeflag(_TPair(p)) &= (~T_HAS_ALL_X)

/* closure stored optlists */
#define T_HAS_OPTLIST                 T_SETTER
#define has_optlist(p)                ((typeflag(_TClo(p)) & T_HAS_OPTLIST) != 0)
#define set_has_optlist(p)            typeflag(_TClo(p)) |= T_HAS_OPTLIST
#define clear_has_optlist(p)          typeflag(_TClo(p)) &= (~T_HAS_OPTLIST)

#define T_HASH_REMOVED                T_SETTER
#define hash_table_set_removed(p)     typeflag(_THsh(p)) |= T_HASH_REMOVED
#define hash_table_removed(p)         ((typeflag(_THsh(p)) & T_HASH_REMOVED) != 0)

#define T_LET_REMOVED                 T_SETTER
#define let_set_removed(p)            typeflag(_TLet(p)) |= T_LET_REMOVED
#define let_removed(p)                ((typeflag(_TLet(p)) & T_LET_REMOVED) != 0)
/* these mark objects that have been removed from the heap or checked for that possibility */


#define T_MUTABLE                     (1 << (TYPE_BITS + 18))
#define is_mutable(p)                 ((typeflag(_TNum(p)) & T_MUTABLE) != 0)
/* #define set_mutable(p)             typeflag(_TNum(p)) |= T_MUTABLE */
/* used for mutable numbers */

#define T_HAS_KEYWORD                 T_MUTABLE
#define has_keyword(p)                ((typeflag(_TSym(p)) & T_HAS_KEYWORD) != 0)
#define set_has_keyword(p)            typeflag(_TSym(p)) |= T_HAS_KEYWORD

#define T_MARK_SEQ                    T_MUTABLE
#define is_mark_seq(p)                ((typeflag(_TItr(p)) & T_MARK_SEQ) != 0)
#define set_mark_seq(p)               typeflag(_TItr(p)) |= T_MARK_SEQ
/* used in iterators for GC mark of sequence */

#define T_BYTE_VECTOR                 T_MUTABLE
#define is_byte_vector(p)             ((is_string(p)) && ((typeflag(p) & T_BYTE_VECTOR) != 0))
#define set_byte_vector(p)            typeflag(_TStr(p)) |= T_BYTE_VECTOR
/* marks a string that the caller considers a byte_vector */

#define T_STEP_END                    T_MUTABLE
#define is_step_end(p)                ((typeflag(_TSlt(p)) & T_STEP_END) != 0)
#define set_step_end(p)               typeflag(_TSlt(p)) |= T_STEP_END
/* marks a slot that holds a do-loop's step-or-end variable, numerator=current, denominator=end */


#define T_PAIR_NO_OPT                 T_MUTABLE
#define set_pair_no_opt(p)            typeflag(_TPair(p)) |= T_PAIR_NO_OPT
#define pair_no_opt(p)                ((typeflag(_TPair(p)) & T_PAIR_NO_OPT) != 0)

#define T_CLOSURE_NO_OPT              T_MUTABLE
#define set_closure_no_opt(p)         typeflag(_TClo(p)) |= T_CLOSURE_NO_OPT
#define closure_no_opt(p)             ((typeflag(_TClo(p)) & T_CLOSURE_NO_OPT) != 0)

#define T_NO_INT_OPT                  T_SETTER
#define set_no_int_opt(p)             typeflag(_TPair(p)) |= T_NO_INT_OPT
#define no_int_opt(p)                 ((typeflag(_TPair(p)) & T_NO_INT_OPT) != 0)

#define T_NO_FLOAT_OPT                T_UNSAFE
#define set_no_float_opt(p)           typeflag(_TPair(p)) |= T_NO_FLOAT_OPT
#define no_float_opt(p)               ((typeflag(_TPair(p)) & T_NO_FLOAT_OPT) != 0)

#define T_NO_BOOL_OPT                 T_SAFE_STEPPER
#define set_no_bool_opt(p)            typeflag(_TPair(p)) |= T_NO_BOOL_OPT
#define no_bool_opt(p)                ((typeflag(_TPair(p)) & T_NO_BOOL_OPT) != 0)

#define T_DIRECT_X_OPT                T_SAFE_STEPPER
#define set_direct_x_opt(p)           typeflag(_TPair(p)) |= T_DIRECT_X_OPT
#define has_direct_x_opt(p)           ((typeflag(_TPair(p)) & T_DIRECT_X_OPT) != 0)


#define T_SAFE_STEPPER                (1 << (TYPE_BITS + 19))
#define is_safe_stepper(p)            ((typeflag(_TSlp(p)) & T_SAFE_STEPPER) != 0)
#define set_safe_stepper(p)           typeflag(_TSlp(p)) |= T_SAFE_STEPPER
#define clear_safe_stepper(p)         typeflag(p) &= (~T_SAFE_STEPPER)
/* an experiment */

#define T_PRINT_NAME                  T_SAFE_STEPPER
#define has_print_name(p)             ((typeflag(_TNum(p)) & T_PRINT_NAME) != 0)
#define set_has_print_name(p)         typeflag(_TNum(p)) |= T_PRINT_NAME
/* marks numbers that have a saved version of their string representation */

#define T_MAYBE_SAFE                  T_SAFE_STEPPER
#define is_maybe_safe(p)              ((typeflag(_TFnc(p)) & T_MAYBE_SAFE) != 0)
#define set_maybe_safe(p)             typeflag(_TFnc(p)) |= T_MAYBE_SAFE

#define T_HAS_SET_FALLBACK            T_SAFE_STEPPER
#define T_HAS_REF_FALLBACK            T_MUTABLE
#define has_ref_fallback(p)           ((typeflag(_TLid(p)) & (T_HAS_REF_FALLBACK | T_HAS_METHODS)) == (T_HAS_REF_FALLBACK | T_HAS_METHODS))
#define has_set_fallback(p)           ((typeflag(_TLid(p)) & (T_HAS_SET_FALLBACK | T_HAS_METHODS)) == (T_HAS_SET_FALLBACK | T_HAS_METHODS))
#define set_has_ref_fallback(p)       typeflag(_TLet(p)) |= T_HAS_REF_FALLBACK
#define set_has_set_fallback(p)       typeflag(_TLet(p)) |= T_HAS_SET_FALLBACK
#define set_all_methods(p, e)         typeflag(_TLet(p)) |= (typeflag(e) & (T_HAS_METHODS | T_HAS_REF_FALLBACK | T_HAS_SET_FALLBACK))


#define T_COPY_ARGS                   (1 << (TYPE_BITS + 20))
#define needs_copied_args(p)          ((typeflag(_NFre(p)) & T_COPY_ARGS) != 0)
/* this marks something that might mess with its argument list, it should not be in the second byte */

#define T_LOCAL_SYMBOL                T_COPY_ARGS
#define is_local_symbol(p)            ((typeflag(_NFre(p)) & T_LOCAL_SYMBOL) != 0)
#define set_local_symbol(p)           typeflag(_TPair(p)) |= T_LOCAL_SYMBOL


#define T_GENSYM                      (1 << (TYPE_BITS + 21))
#define is_gensym(p)                  ((typeflag(_TSym(p)) & T_GENSYM) != 0)
/* symbol is from gensym (GC-able etc) */

#define T_FUNCLET                     T_GENSYM
#define is_funclet(p)                 ((typeflag(_TLet(p)) & T_FUNCLET) != 0)
#define set_funclet(p)                typeflag(_TLet(p)) |= T_FUNCLET
/* this marks a funclet */

#define T_HASH_CHOSEN                 T_GENSYM
#define hash_chosen(p)                ((typeflag(_THsh(p)) & T_HASH_CHOSEN) != 0)
#define hash_set_chosen(p)            typeflag(_THsh(p)) |= T_HASH_CHOSEN
#define hash_clear_chosen(p)          typeflag(_THsh(p)) &= (~T_HASH_CHOSEN)

#define T_DOCUMENTED                  T_GENSYM
#define is_documented(p)              ((typeflag(_TStr(p)) & T_DOCUMENTED) != 0)
#define set_documented(p)             typeflag(_TStr(p)) |= T_DOCUMENTED
/* this marks a symbol that has documentation (bit is set on name cell) */

#define T_DOTTED_PAIR                 T_GENSYM
#define is_dotted_pair(p)             ((typeflag(_TLst(p)) & T_DOTTED_PAIR) != 0)
#define pair_set_dotted(p)            typeflag(_TPair(p)) |= T_DOTTED_PAIR
/* reader indication that a list it just read was dotted */

#define T_HAS_METHODS                 (1 << (TYPE_BITS + 22))
#define has_methods(p)                ((typeflag(_NFre(p)) & T_HAS_METHODS) != 0)
#define set_has_methods(p)            typeflag(_TMet(p)) |= T_HAS_METHODS
#define clear_has_methods(p)          typeflag(_TMet(p)) &= (~T_HAS_METHODS)
/* this marks an environment or closure that is "open" for generic functions etc
 * don't reuse this bit
 */

#define T_ITER_OK                     (1LL << (TYPE_BITS + 23))
#define iter_ok(p)                    ((typeflag(_NFre(p)) & T_ITER_OK) != 0)  /* not TItr(p) here because this bit is globally unique */
#define clear_iter_ok(p)              typeflag(_TItr(p)) &= (~T_ITER_OK)


#define T_GC_MARK                     0x8000000000000000
#define is_marked(p)                  ((typeflag(p) &  T_GC_MARK) != 0)
#define set_mark(p)                   typeflag(_NFre(p)) |= T_GC_MARK
#define clear_mark(p)                 typeflag(p) &= (~T_GC_MARK)
/* using bit 23 for this makes a big difference in the GC */

static int64_t not_heap = -1;
#define heap_location(p)              (p)->hloc
#define not_in_heap(p)                ((_NFre(p))->hloc < 0)
#define in_heap(p)                    ((_NFre(p))->hloc >= 0)
#define unheap(p)                     (p)->hloc = not_heap--

#define is_eof(p)                     ((_NFre(p)) == sc->eof_object)
#define is_true(Sc, p)                ((_NFre(p)) != Sc->F)
#define is_false(Sc, p)               ((_NFre(p)) == Sc->F)

#ifdef _MSC_VER
  #define MS_WINDOWS 1
  static s7_pointer make_boolean(s7_scheme *sc, bool val) {if (val) return(sc->T); return(sc->F);}
#else
  #define MS_WINDOWS 0
  #define make_boolean(sc, Val)       ((Val) ? sc->T : sc->F)
#endif

#define is_pair(p)                    (type(p) == T_PAIR)
#define is_null(p)                    ((_NFre(p)) == sc->nil)
#define is_not_null(p)                ((_NFre(p)) != sc->nil)

#define raw_opt1(p)                   ((p)->object.cons.opt1)

#if (!DEBUGGING)

#define opt1(p, r)                    ((p)->object.cons.opt1)
#define set_opt1(p, x, r)             (p)->object.cons.opt1 = x
#define opt2(p, r)                    ((p)->object.cons.opt2)
#define set_opt2(p, x, r)             (p)->object.cons.opt2 = (s7_pointer)(x)
#define opt3(p, r)                    ((p)->object.cons.opt3)
#define set_opt3(p, x, r)             do {(p)->object.cons.opt3 = x; typeflag(p) &= ~(T_OPTIMIZED | T_LINE_NUMBER);} while (0)
/* turn off optimized because it collides with pair_syntax_op */

#define pair_line(p)                  (p)->object.sym_cons.line
#define pair_set_line(p, X)           (p)->object.sym_cons.line = X
#define pair_raw_hash(p)              (p)->object.sym_cons.hash
#define pair_set_raw_hash(p, X)       (p)->object.sym_cons.hash = X
#define pair_raw_len(p)               (p)->object.sym_cons.op
#define pair_set_raw_len(p, X)        (p)->object.sym_cons.op = X
#define pair_raw_name(p)              (p)->object.sym_cons.fstr
#define pair_set_raw_name(p, X)       (p)->object.sym_cons.fstr = X

/* opt1 == raw_hash, opt2 == raw_name, opt3 == line|ctr + op|len, but hash/name/len only apply to the symbol table so there's no collision
 */

#else

/* these 3 fields (or 8 counting sym_cons) hold most of the varigated optimizer info, so they are used in many conflicting ways.
 * the bits and funcs here try to track each such use, and report any cross-talk or collisions.
 * all of this machinery vanishes if debugging is turned off.
 */
#define S_NAME                        (1 << 25)
#define S_HASH                        (1 << 26)
#define S_OP                          (1 << 27)
#define S_LINE                        (1 << 28)
#define S_LEN                         (1 << 29)
#define S_SYNOP                       (1 << 30) /* 0x80000000 *//* (1 << 31) */

#define E_SET                         (1 << 0)
#define E_FAST                        (1 << 7)   /* fast list in member/assoc circular list check */
#define E_CFUNC                       (1 << 8)   /* c-function */
#define E_CLAUSE                      (1 << 9)   /* case clause */
#define E_BACK                        (1 << 10)  /* back pointer for doubly-linked list */
#define E_LAMBDA                      (1 << 11)  /* lambda(*) */
#define E_SYM                         (1 << 12)  /* symbol */
#define E_PAIR                        (1 << 13)  /* pair */
#define E_CON                         (1 << 14)  /* constant from eval's point of view */
#define E_GOTO                        (1 << 15)  /* call-with-exit exit func */
#define E_ANY                         (1 << 16)  /* anything -- deliberate unchecked case */
#define E_SLOT                        (1 << 17)  /* slot */
#define E_MASK                        (E_FAST | E_CFUNC | E_CLAUSE | E_BACK | E_LAMBDA | E_SYM | E_PAIR | E_CON | E_GOTO | E_ANY | E_SLOT | S_HASH)

#define opt1_is_set(p)                (((p)->debugger_bits & E_SET) != 0)
#define set_opt1_is_set(p)            (p)->debugger_bits |= E_SET
#define opt1_role_matches(p, Role)    (((p)->debugger_bits & E_MASK) == Role)
#define set_opt1_role(p, Role)        (p)->debugger_bits = (Role | ((p)->debugger_bits & ~E_MASK))
#define opt1(p, Role)                 opt1_1(cur_sc, _TPair(p), Role, __func__, __LINE__)
#define set_opt1(p, x, Role)          set_opt1_1(cur_sc, _TPair(p), x, Role, __func__, __LINE__)

#define F_SET                         (1 << 1)   
#define F_KEY                         (1 << 18)  /* case key */
#define F_SLOW                        (1 << 19)  /* slow list in member/assoc circular list check */
#define F_SYM                         (1 << 20)  /* symbol */
#define F_PAIR                        (1 << 21)  /* pair */
#define F_CON                         (1 << 22)  /* constant as above */
#define F_CALL                        (1 << 23)  /* c-func */
#define F_LAMBDA                      (1 << 24)  /* lambda form */
#define F_MASK                        (F_KEY | F_SLOW | F_SYM | F_PAIR | F_CON | F_CALL | F_LAMBDA | S_NAME)

#define opt2_is_set(p)                (((p)->debugger_bits & F_SET) != 0)
#define set_opt2_is_set(p)            (p)->debugger_bits |= F_SET
#define opt2_role_matches(p, Role)    (((p)->debugger_bits & F_MASK) == Role)
#define set_opt2_role(p, Role)        (p)->debugger_bits = (Role | ((p)->debugger_bits & ~F_MASK))
#define opt2(p, Role)                 opt2_1(cur_sc, _TPair(p), Role, __func__, __LINE__)
#define set_opt2(p, x, Role)          set_opt2_1(cur_sc, _TPair(p), (s7_pointer)x, Role, __func__, __LINE__)

/* opt3 collides with optimization and line number stuff (T_LINE_NUMBER, T_OPTIMIZED) */
#define G_SET                         (1 << 2)
#define G_ARGLEN                      (1 << 3)  /* arglist length */
#define G_SYM                         (1 << 4)  /* expression symbol access */
#define G_AND                         (1 << 5)  /* and second clause */
#define G_DIRECT                      (1 << 6)  /* direct call info */
#define G_MASK                        (G_ARGLEN | G_SYM | G_AND | S_OP | S_LINE | S_LEN | S_SYNOP | G_DIRECT)

#define opt3_is_set(p)                (((p)->debugger_bits & G_SET) != 0)
#define set_opt3_is_set(p)            (p)->debugger_bits |= G_SET
#define opt3_role_matches(p, Role)    (((p)->debugger_bits & G_MASK) == Role)
#define set_opt3_role(p, Role)        (p)->debugger_bits = (Role | ((p)->debugger_bits & ~G_MASK))
#define opt3(p, Role)                 opt3_1(cur_sc, _TPair(p), Role, __func__, __LINE__)
#define set_opt3(p, x, Role)          set_opt3_1(cur_sc, _TPair(p), x, Role, __func__, __LINE__)

/* opt1 == s_hash, opt2 == s_fstr, opt3 == s_op|len|line and op==len so they are contradictory (but only op/line|opt3 actually collide)
 * line|len|op: uint32_t set G_SET and S_* if S_LEN -> not op and vice versa
 * another collider: pair_syntax_op|optimize_op below.  Both need bits: S_SYNOP?
 */

#define pair_line(p)                  s_line_1(sc, _TPair(p), __func__, __LINE__)
#define pair_set_line(p, X)           set_s_line_1(sc, _TPair(p), X, __func__, __LINE__)
#define pair_raw_hash(p)              s_hash_1(sc, _TPair(p), __func__, __LINE__)
#define pair_set_raw_hash(p, X)       set_s_hash_1(sc, _TPair(p), X, __func__, __LINE__)
#define pair_raw_len(p)               s_len_1(sc, _TPair(p), __func__, __LINE__)
#define pair_set_raw_len(p, X)        set_s_len_1(sc, _TPair(p), X, __func__, __LINE__)
#define pair_raw_name(p)              s_name_1(sc, _TPair(p), __func__, __LINE__)
#define pair_set_raw_name(p, X)       set_s_name_1(sc, _TPair(p), X, __func__, __LINE__)

#endif

#define opt_fast(P)                   _TLst(opt1(P,              E_FAST))
#define set_opt_fast(P, X)            set_opt1(P, _TPair(X),     E_FAST)
#define opt_back(P)                   _TPair(opt1(P,             E_BACK))
#define set_opt_back(P)               set_opt1(cdr(P), _TPair(P),E_BACK)
#define has_opt_back(P)               (cdr(opt_back(P)) == P )
#define opt_cfunc(P)                  _NFre(opt1(P,              E_CFUNC))
#define set_opt_cfunc(P, X)           set_opt1(P, _NFre(X),      E_CFUNC)
#define opt_lambda_unchecked(P)       opt1(P,                    E_LAMBDA) /* can be free/null? from s7_call? */
#define opt_lambda(P)                 _TClo(opt1(P,              E_LAMBDA))
#define set_opt_lambda(P, X)          set_opt1(P, _NFre(X),      E_LAMBDA)
#define opt_goto(P)                   _TGot(opt1(P,              E_GOTO))
#define set_opt_goto(P, X)            set_opt1(P, _TGot(X),      E_GOTO)
#define opt_clause(P)                 _NFre(opt1(P,              E_CLAUSE))
#define set_opt_clause(P, X)          set_opt1(P, _NFre(X),      E_CLAUSE)
#define opt_sym1(P)                   _TSym(opt1(P,              E_SYM))
#define set_opt_sym1(P, X)            set_opt1(P, _TSym(X),      E_SYM)
#define opt_pair1(P)                  _TLst(opt1(P,              E_PAIR))
#define set_opt_pair1(P, X)           set_opt1(P, _TLst(X),      E_PAIR)
#define opt_con1(P)                   _NFre(opt1(P,              E_CON))
#define set_opt_con1(P, X)            set_opt1(P, _NFre(X),      E_CON)
#define opt_any1(P)                   opt1(P,                    E_ANY)    /* can be free in closure_is_ok */
#define opt_slot1(P)                  _TSlt(opt1(P,              E_SLOT))
#define set_opt_slot1(P, X)           set_opt1(P, _TSlt(X),      E_SLOT)

#define opt_any2(P)                    _NFre(opt2(P,             F_KEY))
#define set_opt_any2(P, X)             set_opt2(P, _NFre(X),     F_KEY)
#define opt_slow(P)                   _TLst(opt2(P,              F_SLOW))
#define set_opt_slow(P, X)            set_opt2(P, _TPair(X),     F_SLOW)
#define opt_sym2(P)                   _TSym(opt2(P,              F_SYM))
#define set_opt_sym2(P, X)            set_opt2(P, _TSym(X),      F_SYM)
#define opt_pair2(P)                  _TLst(opt2(P,              F_PAIR))
#define set_opt_pair2(P, X)           set_opt2(P, _TLst(X),      F_PAIR)
#define opt_con2(P)                   _NFre(opt2(P,              F_CON))
#define set_opt_con2(P, X)            set_opt2(P, _NFre(X),      F_CON)
#define opt_lambda2(P)                _TPair(opt2(P,             F_LAMBDA))
#define set_opt_lambda2(P, X)         set_opt2(P, _TPair(X),     F_LAMBDA)
#define opt_direct_x_call(P)          opt2(P,                    F_LAMBDA)
#define set_opt_direct_x_call(P, X)   set_opt2(P, (s7_pointer)(X), F_LAMBDA)

#define arglist_length(P)             _TI(opt3(cdr(P),           G_ARGLEN))
#define set_arglist_length(P, X)      set_opt3(cdr(P), _TI(X),   G_ARGLEN)
#define opt_sym3(P)                   _TSym(opt3(P,              G_SYM))
#define set_opt_sym3(P, X)            set_opt3(P, _TSym(X),      G_SYM)
#define opt_pair3(P)                  _TPair(opt3(P,             G_AND))
#define set_opt_pair3(P, X)           set_opt3(P, _TPair(X),     G_AND)
#define opt_any3(P)                   _NFre(opt3(P,              G_AND))
#define set_opt_any3(P, X)            set_opt3(P, _NFre(X),      G_AND)
#define opt_direct_x(P)               opt3(P,                    G_DIRECT)
#define set_opt_direct_x(P, X)        set_opt3(P, (s7_pointer)(X), G_DIRECT)

#define opt_con3(P)                   P->object.cons_ext.opt_type
#define set_opt_con3(P, X)            P->object.cons_ext.opt_type = X

#define c_callee(f)                   ((s7_function)opt2(f,      F_CALL))
#define c_call(f)                     ((s7_function)opt2(f,      F_CALL))
#if DEBUGGING
  #define set_c_call(f, _X_)          do {s7_pointer X; X = (s7_pointer)(_X_); if (!(X)) fprintf(stderr, "%s[%d] c_call null\n", __func__, __LINE__); set_opt2(f, X, F_CALL);} while (0)
  #define set_x_call_checked(f, _X_)  do {s7_pointer X; X = (s7_pointer)(_X_); if ((!(X)) && (strcmp(__func__, "check_and") != 0) && (strcmp(__func__, "check_or") != 0)) fprintf(stderr, "%s[%d] x_call null\n", __func__, __LINE__); set_opt2(f, X, F_CALL); if (X) set_has_all_x(f); else clear_has_all_x(f);} while (0)
  #define set_x_call(f, _X_)          do {s7_pointer X; X = (s7_pointer)(_X_); if (!(X)) fprintf(stderr, "%s[%d] x_call null: %s\n", __func__, __LINE__, DISPLAY(f)); set_opt2(f, X, F_CALL); if (X) set_has_all_x(f); else clear_has_all_x(f);} while (0)
#else
  #define set_c_call(f, X)            set_opt2(f, (s7_pointer)(X), F_CALL)
  #define set_x_call(f, X)            do {set_opt2(f, (s7_pointer)(X), F_CALL); set_has_all_x(f);} while (0)
  #define set_x_call_checked(f, _X_)  do {s7_pointer X; X = (s7_pointer)(_X_); set_opt2(f, X, F_CALL); if (X) set_has_all_x(f); else clear_has_all_x(f);} while (0)
#endif
#define set_x_call_direct(f, X)       do {set_opt2(f, (s7_pointer)(X), F_CALL); set_has_all_x(f);} while (0)


#define car(p)                        (_TPair(p))->object.cons.car
#define set_car(p, Val)               (_TPair(p))->object.cons.car = _NFre(Val)
#define cdr(p)                        (_TPair(p))->object.cons.cdr
#define set_cdr(p, Val)               (_TPair(p))->object.cons.cdr = _NFre(Val)
#define unchecked_car(p)              (_NFre(p))->object.cons.car
#define unchecked_cdr(p)              (_NFre(p))->object.cons.cdr

#define caar(p)                       car(car(p))
#define cadr(p)                       car(cdr(p))
#define set_cadr(p, Val)              (_TPair(p))->object.cons.cdr->object.cons.car = _NFre(Val)
#define cdar(p)                       cdr(car(p))
#define set_cdar(p, Val)              (_TPair(p))->object.cons.car->object.cons.cdr = _NFre(Val)
#define cddr(p)                       cdr(cdr(p))

#define caaar(p)                      car(car(car(p)))
#define cadar(p)                      car(cdr(car(p)))
#define cdadr(p)                      cdr(car(cdr(p)))
#define caddr(p)                      car(cdr(cdr(p)))
#define set_caddr(p, Val)             (_TPair(p))->object.cons.cdr->object.cons.cdr->object.cons.car = _NFre(Val)
#define caadr(p)                      car(car(cdr(p)))
#define cdaar(p)                      cdr(car(car(p)))
#define cdddr(p)                      cdr(cdr(cdr(p)))
#define cddar(p)                      cdr(cdr(car(p)))

#define caaadr(p)                     car(car(car(cdr(p))))
#define caadar(p)                     car(car(cdr(car(p))))
#define cadaar(p)                     car(cdr(car(car(p))))
#define cadddr(p)                     car(cdr(cdr(cdr(p))))
#define caaddr(p)                     car(car(cdr(cdr(p))))
#define cddddr(p)                     cdr(cdr(cdr(cdr(p))))
#define caddar(p)                     car(cdr(cdr(car(p))))
#define cdadar(p)                     cdr(car(cdr(car(p))))
#define cdaddr(p)                     cdr(car(cdr(cdr(p))))
#define caaaar(p)                     car(car(car(car(p))))
#define cadadr(p)                     car(cdr(car(cdr(p))))
#define cdaadr(p)                     cdr(car(car(cdr(p))))
#define cdaaar(p)                     cdr(car(car(car(p))))
#define cdddar(p)                     cdr(cdr(cdr(car(p))))
#define cddadr(p)                     cdr(cdr(car(cdr(p))))
#define cddaar(p)                     cdr(cdr(car(car(p))))

#if WITH_GCC
  /* slightly tricky because cons can be called recursively */
  #define cons(Sc, A, B)   ({s7_pointer _X_, _A_, _B_; _A_ = A; _B_ = B; new_cell(Sc, _X_, T_PAIR | T_SAFE_PROCEDURE); set_car(_X_, _A_); set_cdr(_X_, _B_); _X_;})
#else
  #define cons(Sc, A, B)              s7_cons(Sc, A, B)
#endif

#define list_1(Sc, A)                 cons(Sc, A, Sc->nil)
#define list_2(Sc, A, B)              cons_unchecked(Sc, A, cons(Sc, B, Sc->nil))
#define list_3(Sc, A, B, C)           cons_unchecked(Sc, A, cons_unchecked(Sc, B, cons(Sc, C, Sc->nil)))
#define list_4(Sc, A, B, C, D)        cons_unchecked(Sc, A, cons_unchecked(Sc, B, cons_unchecked(Sc, C, cons(Sc, D, Sc->nil))))

#define is_string(p)                  (type(p) == T_STRING)
#define string_value(p)               (_TStr(p))->object.string.svalue
#define string_length(p)              (_TStr(p))->object.string.length
#define string_hash(p)                (_TStr(p))->object.string.hash
#define string_needs_free(p)          (_TStr(p))->object.string.str_ext.needs_free
#define string_temp_true_length(p)    (_TStr(p))->object.string.str_ext.temp_len

#define tmpbuf_malloc(P, Len)         do {if ((Len) < TMPBUF_SIZE) P = sc->tmpbuf; else P = (char *)malloc((Len) * sizeof(char));} while (0)
#define tmpbuf_calloc(P, Len)         do {if ((Len) < TMPBUF_SIZE) {P = sc->tmpbuf; memset((void *)P, 0, Len);} else P = (char *)calloc(Len, sizeof(char));} while (0)
#define tmpbuf_free(P, Len)           do {if ((Len) >= TMPBUF_SIZE) free(P);} while (0)

#define character(p)                  (_TChr(p))->object.chr.c
#define upper_character(p)            (_TChr(p))->object.chr.up_c
#define is_char_alphabetic(p)         (_TChr(p))->object.chr.alpha_c
#define is_char_numeric(p)            (_TChr(p))->object.chr.digit_c
#define is_char_whitespace(p)         (_TChr(p))->object.chr.space_c
#define is_char_uppercase(p)          (_TChr(p))->object.chr.upper_c
#define is_char_lowercase(p)          (_TChr(p))->object.chr.lower_c
#define character_name(p)             (_TChr(p))->object.chr.c_name
#define character_name_length(p)      (_TChr(p))->object.chr.length

#if (!DEBUGGING)
  #define optimize_op(p)              (_TPair(p))->object.sym_cons.op
  #define set_optimize_op(P, Op)      optimize_op(P) = Op
#else
  #define optimize_op(p)              s_op_1(cur_sc, _TPair(p), __func__, __LINE__)
  #define set_optimize_op(p, Op)      set_s_op_1(cur_sc, _TPair(p), Op, __func__, __LINE__)
#endif

#define optimize_op_match(P, Q)       ((is_optimized(P)) && ((optimize_op(P) & 0xfffe) == Q))
#define op_no_hop(P)                  (optimize_op(P) & 0xfffe)
#define clear_hop(P)                  set_optimize_op(P, op_no_hop(P))
#define clear_optimize_op(P)          set_optimize_op(P, 0)
#define set_safe_optimize_op(P, Q)    do {set_optimized(P); set_optimize_op(P, Q);} while (0)
#define set_unsafe_optimize_op(P, Q)  do {set_unsafely_optimized(P); set_optimize_op(P, Q); /* if (is_unknown_op(Q)) fprintf(stderr, "%d %s %s\n", __LINE__, opt_names[Q], DISPLAY_80(P)); */ } while (0)

#define is_symbol(p)                  (type(p) == T_SYMBOL)
#define is_safe_symbol(p)             ((is_symbol(p)) && (is_slot(find_symbol(sc, p))))
#define symbol_name_cell(p)           _TStr((_TSym(p))->object.sym.name)
#define symbol_set_name_cell(p, S)    (_TSym(p))->object.sym.name = _TStr(S)
#define symbol_name(p)                string_value(symbol_name_cell(p))
#define symbol_name_length(p)         string_length(symbol_name_cell(p))
#define symbol_hmap(p)                s7_int_abs(heap_location(p))
#define symbol_global_accessor_index(p) (symbol_name_cell(p))->object.string.str_ext.accessor
#define symbol_id(p)                  (_TSym(p))->object.sym.id
#define symbol_set_id(p, X)           (_TSym(p))->object.sym.id = X
/* we need 64-bits here, since we don't want this thing to wrap around, and frames are created at a great rate
 *    callgrind says this is faster than an uint32_t!
 */
#define symbol_syntax_op(p)           syntax_opcode(slot_value(global_slot(p)))
#define symbol_type(p)                (_TSym(p))->object.sym.type

#define global_slot(p)                (_TSym(p))->object.sym.global_slot
#define set_global_slot(p, Val)       (_TSym(p))->object.sym.global_slot = _TSld(Val)
#define initial_slot(p)               (symbol_name_cell(p))->object.string.initial_slot
#define set_initial_slot(p, Val)      (symbol_name_cell(p))->object.string.initial_slot = _TSld(Val)
#define local_slot(p)                 (_TSym(p))->object.sym.local_slot
#define set_local_slot(p, Val)        (_TSym(p))->object.sym.local_slot = _TSln(Val)
#define keyword_symbol(p)             (symbol_name_cell(p))->object.string.doc.ksym
#define keyword_set_symbol(p, Val)    (symbol_name_cell(p))->object.string.doc.ksym = _TSym(Val)
#define symbol_help(p)                (symbol_name_cell(p))->object.string.doc.documentation
#define symbol_set_help(p, Doc)       (symbol_name_cell(p))->object.string.doc.documentation = Doc
#define symbol_tag(p)                 (_TSym(p))->object.sym.tag
#define symbol_set_tag(p, Val)        (_TSym(p))->object.sym.tag = Val
#define symbol_ctr(p)                 (_TSym(p))->object.sym.ctr
#define symbol_set_ctr(p, Val)        (_TSym(p))->object.sym.ctr = Val
#define symbol_increment_ctr(p)       (_TSym(p))->object.sym.ctr++
#define symbol_has_help(p)            (is_documented(symbol_name_cell(p)))
#define symbol_set_has_help(p)        set_documented(symbol_name_cell(p))

#define symbol_set_local(Symbol, Id, Slot) do {set_local_slot(Symbol, Slot); symbol_set_id(Symbol, Id); symbol_increment_ctr(Symbol);} while (0)
/* set slot before id in case Slot is an expression that tries to find the current Symbol slot (using its old Id obviously) */

#define is_slot(p)                    (type(p) == T_SLOT)
#define slot_symbol(p)                _TSym((_TSlt(p))->object.slt.sym)
#define slot_set_symbol(p, Sym)       (_TSlt(p))->object.slt.sym = _TSym(Sym)
#define slot_value(p)                 _NFre((_TSlt(p))->object.slt.val)
#define slot_set_value(p, Val)        (_TSlt(p))->object.slt.val = _NFre(Val)
#define slot_set_value_with_hook(Slot, Value) \
  do {if (hook_has_functions(sc->rootlet_redefinition_hook)) slot_set_value_with_hook_1(sc, Slot, Value); else slot_set_value(Slot, Value);} while (0)
#define next_slot(p)                  (_TSlt(p))->object.slt.nxt
#define set_next_slot(p, Val)         (_TSlt(p))->object.slt.nxt = _TSln(Val)
#define slot_pending_value(p)         (_TSlt(p))->object.slt.pending_value
#define slot_set_pending_value(p, Val) (_TSlt(p))->object.slt.pending_value = _NFre(Val)
#define slot_expression(p)            (_TSlt(p))->object.slt.expr
#define slot_set_expression(p, Val)   (_TSlt(p))->object.slt.expr = _NFre(Val)
#define slot_accessor(p)              slot_expression(p)
#define slot_set_accessor(p, Val)     slot_expression(p) = _TApp(Val)

#if DEBUGGING
  #define local_symbol_value(Sym)     check_sym(sc, _TSym(Sym))
#else
  #define local_symbol_value(Sym)     slot_value(local_slot(Sym))
#endif

#define is_syntax(p)                  (type(p) == T_SYNTAX)
#define syntax_symbol(p)              _TSym((_TSyn(p))->object.syn.symbol)
#define syntax_set_symbol(p, Sym)     (_TSyn(p))->object.syn.symbol = _TSym(Sym)
#define syntax_opcode(p)              (_TSyn(p))->object.syn.op
#define syntax_min_args(p)            (_TSyn(p))->object.syn.min_args
#define syntax_max_args(p)            (_TSyn(p))->object.syn.max_args
#define syntax_documentation(p)       sc->syn_docs[syntax_opcode(p)]

#if (!DEBUGGING)
  #define pair_syntax_op(p)           (p)->object.sym_cons.op
  #define pair_set_syntax_op(p, X)    do {unoptimize(p); (p)->object.sym_cons.op = X;} while (0)
#else
  #define pair_syntax_op(p)           s_syn_op_1(cur_sc, _TPair(p), __func__, __LINE__)
  #define pair_set_syntax_op(p, Op)   do {unoptimize(p); set_s_syn_op_1(cur_sc, _TPair(p), Op, __func__, __LINE__);} while (0)
#endif
#define pair_syntax_symbol(P)         car(opt_back(P))
#define pair_set_syntax_symbol(p, op) do {set_car(opt_back(p), op); pair_set_syntax_op(opt_back(p), symbol_syntax_op(op));} while (0)

#define ROOTLET_SIZE 512
#define let_id(p)                     (_TLid(p))->object.envr.id
#define is_let(p)                     (type(p) == T_LET)
#define let_slots(p)                  (_TLet(p))->object.envr.slots
#define let_set_slots(p, Slot)        (_TLet(p))->object.envr.slots = _TSln(Slot)
#define outlet(p)                     (_TLet(p))->object.envr.nxt
#define set_outlet(p, ol)             (_TLet(p))->object.envr.nxt = _TLid(ol)
#define funclet_function(p)           _TSym((_TLet(p))->object.envr.edat.efnc.function)
#define funclet_set_function(p, F)    (_TLet(p))->object.envr.edat.efnc.function = _TSym(F)
#define let_line(p)                   (_TLet(p))->object.envr.edat.efnc.line
#define let_set_line(p, L)            (_TLet(p))->object.envr.edat.efnc.line = L
#define let_file(p)                   (_TLet(p))->object.envr.edat.efnc.file
#define let_set_file(p, F)            (_TLet(p))->object.envr.edat.efnc.file = F
#define dox_slot1(p)                  _TSlt((_TLet(p))->object.envr.edat.dox.dox1)
#define dox_set_slot1(p, S)           (_TLet(p))->object.envr.edat.dox.dox1 = _TSlt(S)
#define dox_slot2(p)                  _TSlt((_TLet(p))->object.envr.edat.dox.dox2)
#define dox_set_slot2(p, S)           (_TLet(p))->object.envr.edat.dox.dox2 = _TSlt(S)
#define dox_slot2_unchecked(p)        _TLet(p)->object.envr.edat.dox.dox2
#define dox_set_slot2_unchecked(p, S) _TLet(p)->object.envr.edat.dox.dox2 = (S)

#define unique_name(p)                (p)->object.unq.name
#define unique_name_length(p)         (p)->object.unq.len
#define is_unspecified(p)             (type(p) == T_UNSPECIFIED)
#define unique_car(p)                 (p)->object.unq.unused_slots
#define unique_cdr(p)                 (p)->object.unq.unused_nxt

#define vector_length(p)              ((p)->object.vector.length)
#define vector_element(p, i)          ((p)->object.vector.elements.objects[i])
#define vector_elements(p)            (p)->object.vector.elements.objects
#define vector_getter(p)              (_TVec(p))->object.vector.vget
#define vector_setter(p)              (_TVec(p))->object.vector.vset
#define int_vector_element(p, i)      ((_TIvc(p))->object.vector.elements.ints[i])
#define int_vector_elements(p)        (_TIvc(p))->object.vector.elements.ints
#define float_vector_element(p, i)    ((_TFvc(p))->object.vector.elements.floats[i])
#define float_vector_elements(p)      (_TFvc(p))->object.vector.elements.floats
#define is_normal_vector(p)           (type(p) == T_VECTOR)
#define is_int_vector(p)              (type(p) == T_INT_VECTOR)
#define is_float_vector(p)            (type(p) == T_FLOAT_VECTOR)

#define vector_ndims(p)               ((_TVec(p))->object.vector.dim_info->ndims)
#define vector_dimension(p, i)        ((_TVec(p))->object.vector.dim_info->dims[i])
#define vector_dimensions(p)          ((_TVec(p))->object.vector.dim_info->dims)
#define vector_offset(p, i)           ((_TVec(p))->object.vector.dim_info->offsets[i])
#define vector_offsets(p)             ((_TVec(p))->object.vector.dim_info->offsets)
#define vector_dimension_info(p)      ((_TVec(p))->object.vector.dim_info)
#define shared_vector(p)              ((_TVec(p))->object.vector.dim_info->original)
#define vector_rank(p)                ((vector_dimension_info(p)) ? vector_ndims(p) : 1)
#define vector_has_dimensional_info(p) (vector_dimension_info(p))
#define vector_elements_allocated(p)  ((_TVec(p))->object.vector.dim_info->elements_allocated)
#define vector_dimensions_allocated(p) ((_TVec(p))->object.vector.dim_info->dimensions_allocated)

#define is_hash_table(p)              (type(p) == T_HASH_TABLE)
#define hash_table_mask(p)            (_THsh(p))->object.hasher.mask
#define hash_table_element(p, i)      ((_THsh(p))->object.hasher.elements[i])
#define hash_table_elements(p)        (_THsh(p))->object.hasher.elements
#define hash_table_entries(p)         (_THsh(p))->object.hasher.entries
#define hash_table_checker(p)         (_THsh(p))->object.hasher.hash_func
#define hash_table_mapper(p)          (_THsh(p))->object.hasher.loc
#define hash_table_checker_locked(p)  (hash_table_mapper(p) != default_hash_map)
#define hash_table_procedures(p)      _TLst((_THsh(p))->object.hasher.dproc)
#define hash_table_set_procedures(p, Lst) (_THsh(p))->object.hasher.dproc = _TLst(Lst)
#define hash_table_procedures_checker(p) car(hash_table_procedures(p))
#define hash_table_procedures_mapper(p) cdr(hash_table_procedures(p))

#define is_iterator(p)                (type(p) == T_ITERATOR)
#define iterator_sequence(p)          (_TItr(p))->object.iter.obj
#define iterator_position(p)          (_TItr(p))->object.iter.lc.loc
#define iterator_length(p)            (_TItr(p))->object.iter.lw.len
#define iterator_slow(p)              _TLst((_TItr(p))->object.iter.lw.slow)
#define iterator_set_slow(p, Val)     (_TItr(p))->object.iter.lw.slow = _TLst(Val)
#define iterator_hash_current(p)      (_TItr(p))->object.iter.lw.hcur
#define iterator_current(p)           (_TItr(p))->object.iter.cur
#define iterator_current_slot(p)      _TSln((_TItr(p))->object.iter.lc.lcur)
#define iterator_set_current_slot(p, Val) (_TItr(p))->object.iter.lc.lcur = _TSln(Val)
#define iterator_let_cons(p)          (_TItr(p))->object.iter.cur
#define iterator_next(p)              (_TItr(p))->object.iter.next
#define iterator_is_at_end(p)         ((typeflag(_TItr(p)) & T_ITER_OK) == 0)

#define ITERATOR_END eof_object
#define ITERATOR_END_NAME "#<eof>"

#define is_input_port(p)              (type(p) == T_INPUT_PORT)
#define is_output_port(p)             (type(p) == T_OUTPUT_PORT)
#define port_port(p)                  (_TPrt(p))->object.prt.port
#define port_type(p)                  (_TPrt(p))->object.prt.ptype
#define is_string_port(p)             (port_type(p) == STRING_PORT)
#define is_file_port(p)               (port_type(p) == FILE_PORT)
#define is_function_port(p)           (port_type(p) == FUNCTION_PORT)
#define port_line_number(p)           (_TPrt(p))->object.prt.line_number
#define port_file_number(p)           (_TPrt(p))->object.prt.file_number
#define port_filename(p)              port_port(p)->filename
#define port_filename_length(p)       port_port(p)->filename_length
#define port_file(p)                  port_port(p)->file
#define port_is_closed(p)             (_TPrt(p))->object.prt.is_closed
#define port_data(p)                  (_TPrt(p))->object.prt.data
#define port_data_size(p)             (_TPrt(p))->object.prt.size
#define port_position(p)              (_TPrt(p))->object.prt.point
#define port_needs_free(p)            port_port(p)->needs_free
#define port_output_function(p)       port_port(p)->output_function
#define port_input_function(p)        port_port(p)->input_function
#define port_original_input_string(p) port_port(p)->orig_str
#define port_read_character(p)        port_port(p)->read_character
#define port_read_line(p)             port_port(p)->read_line
#define port_display(p)               port_port(p)->display
#define port_write_character(p)       port_port(p)->write_character
#define port_write_string(p)          port_port(p)->write_string
#define port_read_semicolon(p)        port_port(p)->read_semicolon
#define port_read_white_space(p)      port_port(p)->read_white_space
#define port_read_name(p)             port_port(p)->read_name
#define port_read_sharp(p)            port_port(p)->read_sharp
#define port_gc_loc(p)                port_port(p)->gc_loc

#define is_c_function(f)              (type(f) >= T_C_FUNCTION)
#define is_c_function_star(f)         (type(f) == T_C_FUNCTION_STAR)
#define is_any_c_function(f)          (type(f) >= T_C_FUNCTION_STAR)
#define c_function_data(f)            (_TFnc(f))->object.fnc.c_proc
#define c_function_call(f)            (_TFnc(f))->object.fnc.ff
#define c_function_required_args(f)   (_TFnc(f))->object.fnc.required_args
#define c_function_optional_args(f)   (_TFnc(f))->object.fnc.optional_args
#define c_function_has_rest_arg(f)    (_TFnc(f))->object.fnc.rest_arg
#define c_function_all_args(f)        (_TFnc(f))->object.fnc.all_args
#define c_function_setter(f)          _TApp((_TFnc(f))->object.fnc.setter)
#define c_function_set_setter(f, Val) (_TFnc(f))->object.fnc.setter = _TApp(Val)
#define c_function_name(f)            c_function_data(f)->name
#define c_function_name_length(f)     c_function_data(f)->name_length
#define c_function_documentation(f)   c_function_data(f)->doc
#define c_function_signature(f)       c_function_data(f)->signature
#define c_function_class(f)           c_function_data(f)->id
#define c_function_chooser(f)         c_function_data(f)->chooser
#define c_function_base(f)            _TApp(c_function_data(f)->generic_ff)
#define c_function_set_base(f, Val)   c_function_data(f)->generic_ff = _TApp(Val)
#define c_function_arg_defaults(f)    c_function_data(f)->arg_defaults
#define c_function_call_args(f)       c_function_data(f)->call_args
#define c_function_arg_names(f)       c_function_data(f)->arg_names
#define set_c_function(X, f)          do {set_opt_cfunc(X, f); set_c_call(X, c_function_call(f));} while (0)
#define c_function_opt_data(f)        c_function_data(f)->opt_data

#define is_c_macro(p)                 (type(p) == T_C_MACRO)
#define c_macro_data(f)               (_TMac(f))->object.fnc.c_proc
#define c_macro_call(f)               (_TMac(f))->object.fnc.ff
#define c_macro_name(f)               c_macro_data(f)->name
#define c_macro_name_length(f)        c_macro_data(f)->name_length
#define c_macro_required_args(f)      (_TMac(f))->object.fnc.required_args
#define c_macro_all_args(f)           (_TMac(f))->object.fnc.all_args
#define c_macro_setter(f)             _TApp((_TMac(f))->object.fnc.setter)
#define c_macro_set_setter(f, Val)    (_TMac(f))->object.fnc.setter = _TApp(Val)

#define is_random_state(p)            (type(p) == T_RANDOM_STATE)
#if WITH_GMP
#define random_gmp_state(p)           (_TRan(p))->object.rng.state
#else
#define random_seed(p)                (_TRan(p))->object.rng.seed
#define random_carry(p)               (_TRan(p))->object.rng.carry
#endif

#define continuation_data(p)          (_TCon(p))->object.cwcc.continuation
#define continuation_stack(p)         (_TCon(p))->object.cwcc.stack
#define continuation_set_stack(p, Val) (_TCon(p))->object.cwcc.stack = _TStk(Val)
#define continuation_stack_end(p)     (_TCon(p))->object.cwcc.stack_end
#define continuation_stack_start(p)   (_TCon(p))->object.cwcc.stack_start
#define continuation_stack_size(p)    (_TCon(p))->object.cwcc.continuation->stack_size
#define continuation_stack_top(p)     (continuation_stack_end(p) - continuation_stack_start(p))
#define continuation_op_stack(p)      (_TCon(p))->object.cwcc.op_stack
#define continuation_op_loc(p)        (_TCon(p))->object.cwcc.continuation->op_stack_loc
#define continuation_op_size(p)       (_TCon(p))->object.cwcc.continuation->op_stack_size
#define continuation_key(p)           (_TCon(p))->object.cwcc.continuation->local_key

#define call_exit_goto_loc(p)         (_TGot(p))->object.rexit.goto_loc
#define call_exit_op_loc(p)           (_TGot(p))->object.rexit.op_stack_loc
#define call_exit_active(p)           (_TGot(p))->object.rexit.active

#define temp_stack_top(p)             (_TStk(p))->object.stk.top
#define s7_stack_top(Sc)              ((Sc)->stack_end - (Sc)->stack_start)

#define is_continuation(p)            (type(p) == T_CONTINUATION)
#define is_goto(p)                    (type(p) == T_GOTO)
#define is_macro(p)                   (type(p) == T_MACRO)
/* #define is_bacro(p)                (type(p) == T_BACRO) */
#define is_macro_star(p)              (type(p) == T_MACRO_STAR)
#define is_bacro_star(p)              (type(p) == T_BACRO_STAR)

#define is_closure(p)                 (type(p) == T_CLOSURE)
#define is_closure_star(p)            (type(p) == T_CLOSURE_STAR)
#define closure_args(p)               (_TClo(p))->object.func.args
#define closure_set_args(p, Val)      (_TClo(p))->object.func.args = _TArg(Val)
#define closure_body(p)               (_TPair((_TClo(p))->object.func.body))
#define closure_set_body(p, Val)      (_TClo(p))->object.func.body = _TPair(Val)
#define closure_let(p)                _TLid((_TClo(p))->object.func.env)
#define closure_set_let(p, L)         (_TClo(p))->object.func.env = _TLid(L)
#define closure_setter(p)             _TApp((_TClo(p))->object.func.setter)
#define closure_set_setter(p, Val)    (_TClo(p))->object.func.setter = _TApp(Val)
#define closure_arity(p)              (_TClo(p))->object.func.arity
#define closure_optlist_addr(p)       (_TClo(p))->object.func.opt_addr
#define closure_set_optlist_addr(p, addr) (_TClo(p))->object.func.opt_addr = addr
#define closure_optlist(p)            stored_optlists[closure_optlist_addr(p)]
#define CLOSURE_ARITY_NOT_SET         0x40000000
#define MAX_ARITY                     0x20000000
#define closure_arity_unknown(p)      (closure_arity(p) == CLOSURE_ARITY_NOT_SET)
#define is_thunk(Sc, Fnc)             ((type(Fnc) >= T_GOTO) && (s7_is_aritable(Sc, Fnc, 0)))

/* #define is_optlist(p)              (type(p) == T_OPTLIST) */
#define optlist_num_exprs(p)          (_TOpt(p))->object.opt.num_exprs
#define optlist_set_num_exprs(p, Val) (_TOpt(p))->object.opt.num_exprs = Val
#define optlist_addr(p)               (_TOpt(p))->object.opt.addr
#define optlist_pc(p)                 (_TOpt(p))->object.opt.pc
#define optlist_set_pc(p, Val)        (_TOpt(p))->object.opt.pc = Val
#define optlist_num_args(p)           (_TOpt(p))->object.opt.num_args
#define optlist_set_num_args(p, Val)  (_TOpt(p))->object.opt.num_args = Val
#define optlist_len(p)                (_TOpt(p))->object.opt.len
#define optlist_set_len(p, Val)       (_TOpt(p))->object.opt.len = Val
/* #define optlist_opts(p)            (_TOpt(p))->object.opt.opts */

#define hook_has_functions(p)         (is_pair(s7_hook_functions(sc, _TClo(p))))

#define catch_tag(p)                  (_TCat(p))->object.rcatch.tag
#define catch_goto_loc(p)             (_TCat(p))->object.rcatch.goto_loc
#define catch_op_loc(p)               (_TCat(p))->object.rcatch.op_stack_loc
#define catch_handler(p)              (_TCat(p))->object.rcatch.handler

#define catch_all_goto_loc(p)         (_TLet(p))->object.envr.edat.ctall.goto_loc
#define catch_all_set_goto_loc(p, L)  (_TLet(p))->object.envr.edat.ctall.goto_loc = L
#define catch_all_op_loc(p)           (_TLet(p))->object.envr.edat.ctall.op_stack_loc
#define catch_all_set_op_loc(p, L)    (_TLet(p))->object.envr.edat.ctall.op_stack_loc = L

enum {DWIND_INIT, DWIND_BODY, DWIND_FINISH};
#define dynamic_wind_state(p)         (_TDyn(p))->object.winder.state
#define dynamic_wind_in(p)            (_TDyn(p))->object.winder.in
#define dynamic_wind_out(p)           (_TDyn(p))->object.winder.out
#define dynamic_wind_body(p)          (_TDyn(p))->object.winder.body

#define is_c_object(p)                (type(p) == T_C_OBJECT)
#define c_object_value(p)             (_TObj(p))->object.c_obj.value
#define c_object_type(p)              (_TObj(p))->object.c_obj.type
#define c_object_let(p)               _TLid((_TObj(p))->object.c_obj.e)
#define c_object_set_let(p, L)        (_TObj(p))->object.c_obj.e = _TLid(L)
#define c_object_mark(p)              (_TObj(p))->object.c_obj.mark

#define c_object_info(Sc, p)          Sc->c_object_types[c_object_type(_TObj(p))]
#define c_object_free(Sc, p)          c_object_info(Sc, p)->free
#define c_object_ref(Sc, p)           c_object_info(Sc, p)->ref
#define c_object_set(Sc, p)           c_object_info(Sc, p)->set
#define c_object_print(Sc, p)         c_object_info(Sc, p)->print
#define c_object_print_readably(Sc, p) c_object_info(Sc, p)->print_readably
#define c_object_len(Sc, p)           c_object_info(Sc, p)->length
#define c_object_eql(Sc, p)           c_object_info(Sc, p)->equal
#define c_object_fill(Sc, p)          c_object_info(Sc, p)->fill
#define c_object_copy(Sc, p)          c_object_info(Sc, p)->copy
#define c_object_reverse(Sc, p)       c_object_info(Sc, p)->reverse
#define c_object_direct_ref(Sc, p)    c_object_info(Sc, p)->direct_ref
#define c_object_direct_set(Sc, p)    c_object_info(Sc, p)->direct_set
#define c_object_scheme_name(Sc, p)   _TStr(c_object_info(Sc, p)->scheme_name)
/* #define c_object_outer_type(Sc, p) c_object_info(Sc, p)->outer_type */

#define raw_pointer(p)                (_TPtr(p))->object.cptr.c_pointer
#define raw_pointer_type(p)           (_TPtr(p))->object.cptr.c_type
#define raw_pointer_info(p)           (_TPtr(p))->object.cptr.info
#define is_c_pointer(p)               (type(p) == T_C_POINTER)

#define is_counter(p)                 (type(p) == T_COUNTER)
#define counter_result(p)             (_TCtr(p))->object.ctr.result
#define counter_set_result(p, Val)    (_TCtr(p))->object.ctr.result = _NFre(Val)
#define counter_list(p)               (_TCtr(p))->object.ctr.list
#define counter_set_list(p, Val)      (_TCtr(p))->object.ctr.list = _NFre(Val)
#define counter_capture(p)            (_TCtr(p))->object.ctr.cap
#define counter_set_capture(p, Val)   (_TCtr(p))->object.ctr.cap = Val
#define counter_let(p)                _TLid((_TCtr(p))->object.ctr.env)
#define counter_set_let(p, L)         (_TCtr(p))->object.ctr.env = _TLid(L)
#define counter_slots(p)              (_TCtr(p))->object.ctr.slots
#define counter_set_slots(p, Val)     (_TCtr(p))->object.ctr.slots = _TSln(Val)

#define is_baffle(p)                  (type(p) == T_BAFFLE)
#define baffle_key(p)                 (_TBfl(p))->object.baffle_key

#if __cplusplus && HAVE_COMPLEX_NUMBERS
  using namespace std;                /* the code has to work in C as well as C++, so we can't scatter std:: all over the place */
  typedef complex<s7_double> s7_complex;
  static s7_double Real(complex<s7_double> x) {return(real(x));} /* protect the C++ name */
  static s7_double Imag(complex<s7_double> x) {return(imag(x));}
#endif

#define integer(p)                    (_TI(p))->object.number.integer_value
#define real(p)                       (_TR(p))->object.number.real_value
#define set_real(p, x)                real(p) = x
#define numerator(p)                  (_TF(p))->object.number.fraction_value.numerator
#define denominator(p)                (_TF(p))->object.number.fraction_value.denominator
#define fraction(p)                   (((long double)numerator(p)) / ((long double)denominator(p)))
#define inverted_fraction(p)          (((long double)denominator(p)) / ((long double)numerator(p)))
#define real_part(p)                  (_TZ(p))->object.number.complex_value.rl
#define set_real_part(p, x)           real_part(p) = x
#define imag_part(p)                  (_TZ(p))->object.number.complex_value.im
#define set_imag_part(p, x)           imag_part(p) = x
#if HAVE_COMPLEX_NUMBERS
  #define as_c_complex(p)             CMPLX(real_part(p), imag_part(p))
#endif

#if WITH_GMP
#define big_integer(p)                ((_TBgi(p))->object.number.big_integer)
#define big_ratio(p)                  ((_TBgf(p))->object.number.big_ratio)
#define big_real(p)                   ((_TBgr(p))->object.number.big_real)
#define big_complex(p)                ((_TBgz(p))->object.number.big_complex)
#endif

#define NUM_SMALL_INTS 2048
#define small_int(Val)                small_ints[Val]
#define is_small(n)                   ((n & ~(NUM_SMALL_INTS - 1)) == 0)

#define print_name(p)                 (char *)((_TNum(p))->object.number.pval.name + 1)
#define print_name_length(p)          (_TNum(p))->object.number.pval.name[0]

static void set_print_name(s7_pointer p, const char *name, int32_t len)
{
  if ((len < PRINT_NAME_SIZE) &&
      (!is_mutable(p)))
    {
      set_has_print_name(p);
      print_name_length(p) = (unsigned char)(len & 0xff);
      memcpy((void *)print_name(p), (void *)name, len);
    }
}

#if WITH_GCC
#define make_integer(Sc, N) ({ s7_int _N_; _N_ = (N); (is_small(_N_) ? small_int(_N_) : ({ s7_pointer _X_; new_cell(Sc, _X_, T_INTEGER); integer(_X_) = _N_; _X_;}) ); })

#define make_real(Sc, X) ({ s7_pointer _X_; s7_double _N_ = (X); new_cell(Sc, _X_, T_REAL); set_real(_X_, _N_); _X_;})

#define make_complex(Sc, R, I)						\
  ({ s7_double im; im = (I); ((im == 0.0) ? make_real(Sc, R) : ({ s7_pointer _X_; new_cell(Sc, _X_, T_COMPLEX); set_real_part(_X_, R); set_imag_part(_X_, im); _X_;}) ); })

#define real_to_double(Sc, X, Caller)   ({ s7_pointer _x_; _x_ = (X); ((type(_x_) == T_REAL) ? real(_x_) : s7_number_to_real_with_caller(Sc, _x_, Caller)); })
#define rational_to_double(Sc, X)       ({ s7_pointer _x_; _x_ = (X); ((type(_x_) == T_INTEGER) ? (s7_double)integer(_x_) : fraction(_x_)); })

#else

#define make_integer(Sc, N)           s7_make_integer(Sc, N)
#define make_real(Sc, X)              s7_make_real(Sc, X)
#define make_complex(Sc, R, I)        s7_make_complex(Sc, R, I)
#define real_to_double(Sc, X, Caller) s7_number_to_real_with_caller(Sc, X, Caller)
#define rational_to_double(Sc, X)     s7_number_to_real(Sc, X)
#endif

#define S7_LLONG_MAX 9223372036854775807LL
#define S7_LLONG_MIN (-S7_LLONG_MAX - 1LL)

#define S7_LONG_MAX 2147483647LL
#define S7_LONG_MIN (-S7_LONG_MAX - 1LL)

#define S7_SHORT_MAX 32767
#define S7_SHORT_MIN -32768

static s7_int s7_int_max = 0, s7_int_min = 0;

/* 9007199254740991LL is where a truncated double starts to skip integers (expt 2 53) = ca 1e16
 *   :(ceiling (+ 1e16 1))
 *   10000000000000000
 *   :(> 9007199254740993.0 9007199254740992.0)
 *   #f ; in non-gmp 64-bit doubles
 *
 * but we can't fix this except in the gmp case because:
 *   :(integer-decode-float (+ (expt 2.0 62) 100))
 *   (4503599627370496 10 1)
 *   :(integer-decode-float (+ (expt 2.0 62) 500))
 *   (4503599627370496 10 1)
 *   :(> (+ (expt 2.0 62) 500) (+ (expt 2.0 62) 100))
 *   #f ; non-gmp again
 *
 * i.e. the bits are identical.  We can't even detect when it has happened, so should
 *   we just give an error for any floor (or whatever) of an arg>1e16?  (sin has a similar problem)?
 *   I think in the non-gmp case I'll throw an error in these cases because the results are
 *   bogus:
 *   :(floor (+ (expt 2.0 62) 512))
 *   4611686018427387904
 *   :(floor (+ (expt 2.0 62) 513))
 *   4611686018427388928
 *
 * another case at the edge: (round 9007199254740992.51) -> 9007199254740992
 *
 * This spells trouble for normal arithmetic in this range.  If no gmp,
 *    (- (+ (expt 2.0 62) 512) (+ (expt 2.0 62) 513)) = -1024.0 (should be -1.0)
 *    but we don't currently give an error in this case -- not sure what the right thing is.
 */


/* --------------------------------------------------------------------------------
 * local versions of some standard C library functions
 * timing tests involving these are very hard to interpret -- pervasive inconsistency!
 */

static inline int32_t safe_strlen(const char *str)
{
  /* this is safer than strlen, and slightly faster */
  char *tmp = (char *)str;
  if ((!tmp) || (!(*tmp))) return(0);
  while (*tmp++) {};
  return(tmp - str - 1);
}


static char *copy_string_with_length(const char *str, int32_t len)
{
  char *newstr;
  newstr = (char *)malloc((len + 1) * sizeof(char));
  if (len != 0)
    memcpy((void *)newstr, (void *)str, len);
  newstr[len] = '\0';
  return(newstr);
}


static char *copy_string(const char *str)
{
  return(copy_string_with_length(str, safe_strlen(str)));
}


static bool local_strcmp(const char *s1, const char *s2)
{
  while (true)
    {
      if (*s1 != *s2++) return(false);
      if (*s1++ == 0) return(true);
    }
  return(true);
}

#define strings_are_equal(Str1, Str2) (local_strcmp(Str1, Str2))
/* this should only be used for internal strings -- scheme strings can have embedded nulls. */

static bool safe_strcmp(const char *s1, const char *s2)
{
  if ((!s1) || (!s2)) return(s1 == s2);
  return(local_strcmp(s1, s2));
}


static bool local_strncmp(const char *s1, const char *s2, uint32_t n)
{
#if S7_ALIGNED
  return(strncmp(s1, s2, n) == 0);
#else
#if (defined(__x86_64__) || defined(__i386__)) /* unaligned accesses are safe on i386 hardware, sez everyone */
  if (n >= 4)
    {
      int32_t *is1, *is2;
      int32_t n4 = n >> 2;
      is1 = (int32_t *)s1;
      is2 = (int32_t *)s2;
      do {if (*is1++ != *is2++) return(false);} while (--n4 > 0);
      s1 = (const char *)is1;
      s2 = (const char *)is2;
      n &= 3;
    }
#endif
  while (n > 0)
    {
      if (*s1++ != *s2++) return(false);
      n--;
    }
  return(true);
#endif
}

#define strings_are_equal_with_length(Str1, Str2, Len) (local_strncmp(Str1, Str2, Len))


static void memclr(void *s, size_t n)
{
  unsigned char *s2;
#if S7_ALIGNED
  s2 = (unsigned char *)s;
#else
#if (defined(__x86_64__) || defined(__i386__))
  if (n >= 4)
    {
      int32_t *s1 = (int32_t *)s;
      size_t n4 = n >> 2;
      do {*s1++ = 0;} while (--n4 > 0);
      n &= 3;
      s2 = (unsigned char *)s1;
    }
  else s2 = (unsigned char *)s;
#else
  s2 = (unsigned char *)s;
#endif
#endif
  while (n > 0)
    {
      *s2++ = 0;
      n--;
    }
}


/* ---------------- forward decls ---------------- */

static char *number_to_string_base_10(s7_pointer obj, int32_t width, int32_t precision, char float_choice, int32_t *nlen, use_write_t choice);
static s7_pointer iterator_finished(s7_scheme *sc, s7_pointer iterator);
static s7_pointer eval(s7_scheme *sc, opcode_t first_op);
static s7_pointer division_by_zero_error(s7_scheme *sc, s7_pointer caller, s7_pointer arg);
static s7_pointer file_error(s7_scheme *sc, const char *caller, const char *descr, const char *name);
static s7_pointer prepackaged_type_name(s7_scheme *sc, s7_pointer x);
static void s7_warn(s7_scheme *sc, int32_t len, const char *ctrl, ...);
static s7_pointer safe_reverse_in_place(s7_scheme *sc, s7_pointer list);
static s7_pointer cons_unchecked(s7_scheme *sc, s7_pointer a, s7_pointer b);
static s7_pointer permanent_cons(s7_pointer a, s7_pointer b, uint64_t type);
static s7_pointer permanent_list(s7_scheme *sc, int32_t len);
static s7_pointer make_atom(s7_scheme *sc, char *q, int32_t radix, bool want_symbol, bool with_error);
static s7_pointer apply_error(s7_scheme *sc, s7_pointer obj, s7_pointer args);
static int32_t remember_file_name(s7_scheme *sc, const char *file);
static const char *type_name(s7_scheme *sc, s7_pointer arg, int32_t article);
static s7_pointer make_vector_1(s7_scheme *sc, s7_int len, bool filled, uint64_t typ);
static s7_pointer make_string_uncopied_with_length(s7_scheme *sc, char *str, int32_t len);
static s7_pointer make_string_wrapper_with_length(s7_scheme *sc, const char *str, int32_t len);
static void check_for_substring_temp(s7_scheme *sc, s7_pointer expr);
static s7_pointer splice_in_values(s7_scheme *sc, s7_pointer args);
static void pop_input_port(s7_scheme *sc);
static char *object_to_truncated_string(s7_scheme *sc, s7_pointer p, int32_t len);
static token_t token(s7_scheme *sc);
static s7_pointer implicit_index(s7_scheme *sc, s7_pointer obj, s7_pointer indices);
static bool s7_is_morally_equal(s7_scheme *sc, s7_pointer x, s7_pointer y);
static void remove_gensym_from_symbol_table(s7_scheme *sc, s7_pointer sym);
static s7_pointer unbound_variable(s7_scheme *sc, s7_pointer sym);
static void free_hash_table(s7_pointer table);
void s7_show_let(s7_scheme *sc);
static s7_pointer g_cdr(s7_scheme *sc, s7_pointer args);

static bool is_all_x_safe(s7_scheme *sc, s7_pointer p);
static void annotate_args(s7_scheme *sc, s7_pointer args, s7_pointer e);
static void annotate_arg(s7_scheme *sc, s7_pointer arg, s7_pointer e);
static bool float_optimize(s7_scheme *sc, s7_pointer expr);
static bool int_optimize(s7_scheme *sc, s7_pointer expr);
static bool bool_optimize(s7_scheme *sc, s7_pointer expr);
static bool bool_optimize_nw(s7_scheme *sc, s7_pointer expr);
static bool cell_optimize(s7_scheme *sc, s7_pointer expr);
static void pc_fallback(s7_scheme *sc, int32_t new_pc);


#if WITH_GMP
static s7_int big_integer_to_s7_int(mpz_t n);
#else
static double next_random(s7_pointer r);
#endif

#if DEBUGGING && WITH_GCC
  static s7_pointer find_symbol_unchecked_1(s7_scheme *sc, s7_pointer symbol);
  #define find_symbol_unchecked(Sc, Sym) check_null_sym(Sc, find_symbol_unchecked_1(Sc, Sym), Sym, __LINE__, __func__)
  static s7_pointer check_null_sym(s7_scheme *sc, s7_pointer p, s7_pointer sym, int32_t line, const char *func);
  #define find_symbol_unexamined(Sc, Sym) find_symbol_unchecked_1(Sc, Sym)
  static s7_pointer check_sym(s7_scheme *sc, s7_pointer sym);
#else
  static inline s7_pointer find_symbol_unchecked(s7_scheme *sc, s7_pointer symbol);
  #define find_symbol_unexamined(Sc, Sym) find_symbol_unchecked(Sc, Sym)
#endif

#if WITH_GCC
  #if DEBUGGING
    #define find_symbol_checked(Sc, Sym) ({s7_pointer _x_; _x_ = find_symbol_unchecked_1(Sc, Sym); ((_x_) ? _x_ : unbound_variable(Sc, Sym));})
  #else
    #define find_symbol_checked(Sc, Sym) ({s7_pointer _x_; _x_ = find_symbol_unchecked(Sc, Sym); ((_x_) ? _x_ : unbound_variable(Sc, Sym));})
  #endif
#else
  #define find_symbol_checked(Sc, Sym) find_symbol_unchecked(Sc, Sym)
#endif

static s7_pointer find_method(s7_scheme *sc, s7_pointer env, s7_pointer symbol);
static s7_pointer find_let(s7_scheme *sc, s7_pointer obj);
static bool call_begin_hook(s7_scheme *sc);
static s7_pointer default_vector_setter(s7_scheme *sc, s7_pointer vec, s7_int loc, s7_pointer val);
static s7_pointer default_vector_getter(s7_scheme *sc, s7_pointer vec, s7_int loc);

static s7_pointer simple_wrong_type_arg_error_prepackaged(s7_scheme *sc, s7_pointer caller, s7_pointer arg, s7_pointer typnam, s7_pointer descr);
static s7_pointer wrong_type_arg_error_prepackaged(s7_scheme *sc, s7_pointer caller, s7_pointer arg_n, s7_pointer arg, s7_pointer typnam, s7_pointer descr);
static s7_pointer out_of_range_error_prepackaged(s7_scheme *sc, s7_pointer caller, s7_pointer arg_n, s7_pointer arg, s7_pointer descr);
static s7_pointer simple_out_of_range_error_prepackaged(s7_scheme *sc, s7_pointer caller, s7_pointer arg, s7_pointer descr);

/* putting off the type description until s7_error via the sc->gc_nil marker below makes it possible
 *    for gcc to speed up the functions that call these as tail-calls.  1-2% overall speedup!
 */
#define simple_wrong_type_argument(Sc, Caller, Arg, Desired_Type) \
  simple_wrong_type_arg_error_prepackaged(Sc, symbol_name_cell(Caller), Arg, Sc->gc_nil, prepackaged_type_names[Desired_Type])

#define wrong_type_argument(Sc, Caller, Num, Arg, Desired_Type)	\
  wrong_type_arg_error_prepackaged(Sc, symbol_name_cell(Caller), make_integer(Sc, Num), Arg, Sc->gc_nil, prepackaged_type_names[Desired_Type])

#define simple_wrong_type_argument_with_type(Sc, Caller, Arg, Type) \
  simple_wrong_type_arg_error_prepackaged(Sc, symbol_name_cell(Caller), Arg, Sc->gc_nil, Type)

#define wrong_type_argument_with_type(Sc, Caller, Num, Arg, Type)	\
  wrong_type_arg_error_prepackaged(Sc, symbol_name_cell(Caller), make_integer(Sc, Num), Arg, Sc->gc_nil, Type)


#define simple_out_of_range(Sc, Caller, Arg, Description) \
  simple_out_of_range_error_prepackaged(Sc, symbol_name_cell(Caller), Arg, Description)

#define out_of_range(Sc, Caller, Arg_Num, Arg, Description)		\
  out_of_range_error_prepackaged(Sc, symbol_name_cell(Caller), Arg_Num, Arg, Description)


static s7_pointer car_a_list_string, cdr_a_list_string, caar_a_list_string, cadr_a_list_string, cdar_a_list_string,
                  cddr_a_list_string, caaar_a_list_string, caadr_a_list_string, cadar_a_list_string, caddr_a_list_string,
                  cdaar_a_list_string, cdadr_a_list_string, cddar_a_list_string, cdddr_a_list_string, a_list_string, 
                  an_association_list_string, an_output_port_string, an_input_port_string, an_open_port_string, 
                  a_normal_real_string, a_rational_string, a_boolean_string, a_number_string, a_let_string, 
                  a_procedure_string, a_proper_list_string, a_thunk_string, something_applicable_string, a_symbol_string, 
                  a_non_negative_integer_string, a_format_port_string, an_unsigned_byte_string, value_is_missing_string,
                  a_non_constant_symbol_string, an_eq_func_string, a_sequence_string, its_too_small_string, parameter_set_twice_string,
                  a_normal_procedure_string, its_too_large_string, its_negative_string, result_is_too_large_string, 
                  its_nan_string, its_infinite_string, too_many_indices_string, a_valid_radix_string, an_input_string_port_string, 
                  an_input_file_port_string, an_output_string_port_string, an_output_file_port_string, a_random_state_object_string;

#if (!HAVE_COMPLEX_NUMBERS)
static s7_pointer no_complex_numbers_string;
#endif


/* ---------------- evaluator ops ---------------- */

enum {OP_NO_OP, OP_GC_PROTECT,
      OP_READ_INTERNAL, OP_EVAL,
      OP_EVAL_ARGS, OP_EVAL_ARGS1, OP_EVAL_ARGS2, OP_EVAL_ARGS3, OP_EVAL_ARGS4, OP_EVAL_ARGS5,
      OP_APPLY, OP_EVAL_MACRO, OP_LAMBDA, OP_QUOTE, OP_MACROEXPAND,
      OP_DEFINE, OP_DEFINE1, OP_BEGIN, OP_BEGIN1,
      OP_IF, OP_IF1, OP_WHEN, OP_WHEN1, OP_UNLESS, OP_UNLESS1, OP_SET, OP_SET1, OP_SET2,
      OP_LET, OP_LET1, OP_LET_STAR, OP_LET_STAR1, OP_LET_STAR2,
      OP_LETREC, OP_LETREC1, OP_LETREC_STAR, OP_LETREC_STAR1,
      OP_LET_TEMPORARILY, OP_LET_TEMP_UNCHECKED, OP_LET_TEMP_INIT1, OP_LET_TEMP_INIT2, OP_LET_TEMP_DONE, OP_LET_TEMP_DONE1,
      OP_COND, OP_COND1, OP_COND1_1, OP_COND_SIMPLE, OP_COND1_SIMPLE, OP_COND_UNCHECKED_Z, 
      OP_AND, OP_AND1, OP_OR, OP_OR1,
      OP_DEFINE_MACRO, OP_DEFINE_MACRO_STAR, OP_DEFINE_EXPANSION,
      OP_CASE, 
      OP_READ_LIST, OP_READ_NEXT, OP_READ_DOT, OP_READ_QUOTE,
      OP_READ_QUASIQUOTE, OP_READ_UNQUOTE, OP_READ_APPLY_VALUES,
      OP_READ_VECTOR, OP_READ_BYTE_VECTOR, OP_READ_INT_VECTOR, OP_READ_FLOAT_VECTOR, OP_READ_DONE,
      OP_LOAD_RETURN_IF_EOF, OP_LOAD_CLOSE_AND_POP_IF_EOF, OP_EVAL_DONE,
      OP_CATCH, OP_DYNAMIC_WIND, OP_DEFINE_CONSTANT, OP_DEFINE_CONSTANT1,
      OP_DO, OP_DO_END, OP_DO_END1, OP_DO_STEP, OP_DO_STEP2, OP_DO_INIT,
      OP_DEFINE_STAR, OP_LAMBDA_STAR, OP_LAMBDA_STAR_DEFAULT, OP_ERROR_QUIT, OP_UNWIND_INPUT, OP_UNWIND_OUTPUT,
      OP_ERROR_HOOK_QUIT,
      OP_WITH_LET, OP_WITH_LET1, OP_WITH_LET_UNCHECKED, OP_WITH_LET_S,
      OP_WITH_BAFFLE, OP_WITH_BAFFLE_UNCHECKED, OP_EXPANSION,
      OP_FOR_EACH, OP_FOR_EACH_1, OP_FOR_EACH_2, OP_FOR_EACH_3,
      OP_MAP, OP_MAP_1, OP_MAP_2, OP_MAP_GATHER, OP_MAP_GATHER_1, OP_MAP_GATHER_2, OP_MAP_GATHER_3, 
      OP_BARRIER, OP_DEACTIVATE_GOTO,

      OP_DEFINE_BACRO, OP_DEFINE_BACRO_STAR,
      OP_GET_OUTPUT_STRING, OP_GET_OUTPUT_STRING_1,
      OP_SORT, OP_SORT1, OP_SORT2, OP_SORT3, OP_SORT_PAIR_END, OP_SORT_VECTOR_END, OP_SORT_STRING_END,
      OP_EVAL_STRING_1, OP_EVAL_STRING_2,
      OP_MEMBER_IF, OP_ASSOC_IF, OP_MEMBER_IF1, OP_ASSOC_IF1,

      OP_QUOTE_UNCHECKED, OP_LAMBDA_UNCHECKED, OP_LET_UNCHECKED, OP_CASE_UNCHECKED, OP_WHEN_UNCHECKED, OP_UNLESS_UNCHECKED,

      OP_SET_UNCHECKED, OP_SET_SYMBOL_C, OP_SET_SYMBOL_L, OP_SET_SYMBOL_S, OP_SET_SYMBOL_Q, OP_SET_SYMBOL_P, OP_SET_SYMBOL_Z, OP_SET_SYMBOL_A,
      OP_SET_SYMBOL_opSq, OP_SET_SYMBOL_opCq, OP_SET_SYMBOL_opSSq, OP_SET_SYMBOL_opSSSq, 
      OP_SET_SYMBOL_opLq, OP_SET_SYMBOL_opLLq, 
      OP_SET_NORMAL, OP_SET_PAIR, OP_SET_DILAMBDA, OP_SET_DILAMBDA_Z, OP_SET_DILAMBDA_Z_1, OP_SET_PAIR_Z, OP_SET_PAIR_A, OP_SET_PAIR_P, OP_SET_PAIR_ZA,
      OP_SET_PAIR_P_1, OP_SET_WITH_ACCESSOR, OP_SET_PWS, OP_SET_LET_S, OP_SET_LET_ALL_X,
      OP_SET_SAFE,
      OP_INCREMENT_1, OP_DECREMENT_1, OP_SET_CONS,
      OP_INCREMENT_SS, OP_INCREMENT_SSS, OP_INCREMENT_SZ, OP_INCREMENT_SA, OP_INCREMENT_SAA,

      OP_LET_STAR_UNCHECKED, OP_LETREC_UNCHECKED, OP_LETREC_STAR_UNCHECKED, OP_COND_UNCHECKED,
      OP_LAMBDA_STAR_UNCHECKED, OP_DO_UNCHECKED, OP_DEFINE_UNCHECKED, OP_DEFINE_STAR_UNCHECKED, OP_DEFINE_FUNCHECKED, OP_DEFINE_CONSTANT_UNCHECKED,
      OP_DEFINE_WITH_ACCESSOR, OP_DEFINE_MACRO_WITH_ACCESSOR,

      OP_LET_NO_VARS, OP_NAMED_LET, OP_NAMED_LET_NO_VARS, OP_NAMED_LET_STAR,
      OP_LET_C, OP_LET_S, OP_LET_S_Z, OP_LET_ALL_C, OP_LET_ALL_S, OP_LET_ALL_X,
      OP_LET_STAR_ALL_X, OP_LET_STAR_A2, OP_LET_STAR_A, OP_LET_opCq, OP_LET_opSSq, OP_LET_opSSq_E, OP_LET_opaSSq_E,
      OP_LET_opSq, OP_LET_ALL_opSq, OP_LET_opSq_P, OP_LET_CAR, OP_LET_ONE, OP_LET_ONE_1, OP_LET_Z, OP_LET_Z_1, OP_LET_A, OP_LET_A_Z,

      OP_CASE_A_E_S, OP_CASE_A_I_S, OP_CASE_A_G_S, OP_CASE_A_E_G, OP_CASE_A_G_G, 
      OP_CASE_S_E_S, OP_CASE_S_I_S, OP_CASE_S_G_S, OP_CASE_S_E_G, OP_CASE_S_G_G, 
      OP_CASE_P_E_S, OP_CASE_P_I_S, OP_CASE_P_G_S, OP_CASE_P_E_G, OP_CASE_P_G_G, 
      OP_CASE_E_S, OP_CASE_I_S, OP_CASE_G_S, OP_CASE_E_G, OP_CASE_G_G, 

      OP_IF_UNCHECKED, OP_AND_UNCHECKED, OP_AND_P, OP_AND_P1, OP_AND_AP, OP_AND_AZ, OP_AND_SAFE_P, OP_AND_SAFE_AA, 
      OP_OR_UNCHECKED, OP_OR_P, OP_OR_P1, OP_OR_AP, OP_OR_SAFE_P, OP_OR_SAFE_AA,
      OP_COND_FEED, OP_COND_FEED_1, OP_WHEN_S, OP_WHEN_A, OP_WHEN_P, OP_UNLESS_S, OP_UNLESS_A,

      OP_IF_S_P, OP_IF_S_P_P, OP_IF_S_R, OP_IF_S_N, OP_IF_S_N_N,
      OP_IF_C_P, OP_IF_C_P_P, OP_IF_C_R, OP_IF_C_N, OP_IF_C_N_N,
      OP_IF_CS_P, OP_IF_CS_P_P, OP_IF_CS_R, OP_IF_CS_N, OP_IF_CS_N_N,
      OP_IF_CSQ_P, OP_IF_CSQ_P_P, OP_IF_CSQ_R, OP_IF_CSQ_N, OP_IF_CSQ_N_N,
      OP_IF_CSS_P, OP_IF_CSS_P_P, OP_IF_CSS_R, OP_IF_CSS_N, OP_IF_CSS_N_N,
      OP_IF_CSC_P, OP_IF_CSC_P_P, OP_IF_CSC_R, OP_IF_CSC_N, OP_IF_CSC_N_N,
      OP_IF_opSq_P, OP_IF_opSq_P_P, OP_IF_opSq_R, OP_IF_opSq_N, OP_IF_opSq_N_N, 
      OP_IF_S_opCq_P, OP_IF_S_opCq_P_P, OP_IF_S_opCq_R, OP_IF_S_opCq_N, OP_IF_S_opCq_N_N, 
      OP_IF_IS_TYPE_S_P, OP_IF_IS_TYPE_S_P_P, OP_IF_IS_TYPE_S_R, OP_IF_IS_TYPE_S_N, OP_IF_IS_TYPE_S_N_N,
      OP_IF_IS_TYPE_opSq_P, OP_IF_IS_TYPE_opSq_P_P, OP_IF_IS_TYPE_opSq_R, OP_IF_IS_TYPE_opSq_N, OP_IF_IS_TYPE_opSq_N_N,
      OP_IF_A_P, OP_IF_A_P_P, OP_IF_A_R, OP_IF_A_N, OP_IF_A_N_N,
      OP_IF_AND2_P, OP_IF_AND2_P_P, OP_IF_AND2_R, OP_IF_AND2_N, OP_IF_AND2_N_N,
      OP_IF_AND3_P, OP_IF_AND3_P_P, OP_IF_AND3_R, OP_IF_AND3_N, OP_IF_AND3_N_N,
      OP_IF_Z_P, OP_IF_Z_P_P, OP_IF_Z_R, OP_IF_Z_N, OP_IF_Z_N_N,
      OP_IF_P_P, OP_IF_P_P_P, OP_IF_P_R, OP_IF_P_N, OP_IF_P_N_N,
      OP_IF_ANDP_P, OP_IF_ANDP_P_P, OP_IF_ANDP_R, OP_IF_ANDP_N, OP_IF_ANDP_N_N,
      OP_IF_ORP_P, OP_IF_ORP_P_P, OP_IF_ORP_R, OP_IF_ORP_N, OP_IF_ORP_N_N,
      OP_IF_OR2_P, OP_IF_OR2_P_P, OP_IF_OR2_R, OP_IF_OR2_N, OP_IF_OR2_N_N,

      OP_IF_PPP, OP_IF_PP, OP_IF_PR, OP_IF_PRR,
      OP_WHEN_PP,

      OP_CATCH_1, OP_CATCH_2, OP_CATCH_ALL, OP_COND_ALL_X, OP_COND_ALL_X_2, OP_COND_ALL_X_Z,
      OP_SIMPLE_DO, OP_SIMPLE_DO_STEP, OP_SAFE_DOTIMES, OP_SAFE_DOTIMES_STEP, OP_SAFE_DOTIMES_STEP_P, OP_SAFE_DOTIMES_STEP_O,
      OP_SAFE_DO, OP_SAFE_DO_STEP, OP_DOX, OP_DOX_STEP, OP_DOX_STEP_P, 
      OP_DOTIMES_P, OP_DOTIMES_STEP_P,
      OP_DO_NO_VARS, OP_DO_NO_VARS_NO_OPT, OP_DO_NO_VARS_NO_OPT_1,
      OP_DOTIMES_ONE_STEP,

      OP_SAFE_C_P_1, OP_SAFE_C_PP_1, OP_SAFE_C_PP_3, OP_SAFE_C_PP_5, OP_SAFE_C_PP_6,
      OP_EVAL_ARGS_P_2, OP_EVAL_ARGS_P_2_MV, OP_EVAL_ARGS_P_3, OP_EVAL_ARGS_P_4, OP_EVAL_ARGS_P_3_MV,
      OP_EVAL_ARGS_AAP_1, OP_EVAL_ARGS_AAP_MV, OP_EVAL_MACRO_MV, OP_MACROEXPAND_1, OP_APPLY_LAMBDA,
      OP_SAFE_CLOSURE_P_1, OP_CLOSURE_P_1, OP_SAFE_CLOSURE_AP_1, OP_SAFE_CLOSURE_PA_1, 
      OP_SAFE_C_ZZ_1, OP_SAFE_C_ZA_1, OP_INCREMENT_SZ_1, OP_SAFE_C_SZ_SZ,
      OP_SAFE_C_ZAA_1, OP_SAFE_C_AZA_1, OP_SAFE_C_AAZ_1, 
      OP_SAFE_C_ZZA_1, OP_SAFE_C_ZZA_2, OP_SAFE_C_ZAZ_1, OP_SAFE_C_AZZ_1, OP_SAFE_C_AZZ_2,
      OP_SAFE_C_ZZZ_1, OP_SAFE_C_ZZZ_2,
      OP_C_P_1, OP_C_P_2, OP_C_AP_1, OP_NOT_P_1,
      OP_CLOSURE_AP_1, OP_CLOSURE_PA_1, 
      OP_CLOSURE_P_MV, OP_CLOSURE_AP_MV, OP_CLOSURE_PA_MV, 

      OP_SET_WITH_LET_1, OP_SET_WITH_LET_2,
      OP_MAX_DEFINED_1};

#define OP_MAX_DEFINED (OP_MAX_DEFINED_1 + 1)

typedef enum{E_C_P, E_C_PP, E_C_CP, E_C_SP, E_C_PC, E_C_PS} combine_op_t;

enum {OP_SAFE_C_C, HOP_SAFE_C_C, 
      OP_SAFE_C_AND2, HOP_SAFE_C_AND2, OP_SAFE_C_OR2, HOP_SAFE_C_OR2,
      OP_SAFE_C_S, HOP_SAFE_C_S, OP_SAFE_C_L, HOP_SAFE_C_L,
      OP_SAFE_CAR_S, HOP_SAFE_CAR_S, OP_SAFE_CDR_S, HOP_SAFE_CDR_S, OP_SAFE_CADR_S, HOP_SAFE_CADR_S,
      OP_SAFE_IS_PAIR_S, HOP_SAFE_IS_PAIR_S, OP_SAFE_IS_NULL_S, HOP_SAFE_IS_NULL_S, OP_SAFE_IS_SYMBOL_S, HOP_SAFE_IS_SYMBOL_S, /* order matters here */
      OP_SAFE_C_SS, HOP_SAFE_C_SS, OP_SAFE_C_LL, HOP_SAFE_C_LL, 
      OP_SAFE_C_SC, HOP_SAFE_C_SC, OP_SAFE_C_CS, HOP_SAFE_C_CS,
      OP_SAFE_C_SQ, HOP_SAFE_C_SQ, OP_SAFE_C_QS, HOP_SAFE_C_QS, OP_SAFE_C_QQ, HOP_SAFE_C_QQ,
      OP_SAFE_C_CQ, HOP_SAFE_C_CQ, OP_SAFE_C_QC, HOP_SAFE_C_QC,
      OP_SAFE_C_SSS, HOP_SAFE_C_SSS, OP_SAFE_C_SCS, HOP_SAFE_C_SCS, OP_SAFE_C_SSC, HOP_SAFE_C_SSC, OP_SAFE_C_CSS, HOP_SAFE_C_CSS,
      OP_SAFE_C_SCC, HOP_SAFE_C_SCC, OP_SAFE_C_CSC, HOP_SAFE_C_CSC,
      OP_SAFE_C_ALL_S, HOP_SAFE_C_ALL_S, OP_SAFE_C_ALL_X, HOP_SAFE_C_ALL_X, OP_SAFE_C_SSA, HOP_SAFE_C_SSA, OP_SAFE_C_SAS, HOP_SAFE_C_SAS,
      OP_SAFE_C_CSA, HOP_SAFE_C_CSA, OP_SAFE_C_SCA, HOP_SAFE_C_SCA, 
      OP_SAFE_C_A, HOP_SAFE_C_A, OP_SAFE_C_AA, HOP_SAFE_C_AA, OP_SAFE_C_AAA, HOP_SAFE_C_AAA, OP_SAFE_C_AAAA, HOP_SAFE_C_AAAA,
      OP_SAFE_C_SQS, HOP_SAFE_C_SQS, OP_SAFE_C_opAq, HOP_SAFE_C_opAq, OP_SAFE_C_opAAq, HOP_SAFE_C_opAAq, OP_SAFE_C_opAAAq, HOP_SAFE_C_opAAAq,
      OP_SAFE_C_S_opAq, HOP_SAFE_C_S_opAq, OP_SAFE_C_S_opAAq, HOP_SAFE_C_S_opAAq, OP_SAFE_C_S_opAAAq, HOP_SAFE_C_S_opAAAq,

      OP_SAFE_C_opCq, HOP_SAFE_C_opCq, OP_SAFE_C_opSq, HOP_SAFE_C_opSq,
      OP_SAFE_C_opCq_opSq, HOP_SAFE_C_opCq_opSq, OP_SAFE_C_opSq_opCq, HOP_SAFE_C_opSq_opCq,
      OP_SAFE_C_opSSq, HOP_SAFE_C_opSSq, OP_SAFE_C_opSCq, HOP_SAFE_C_opSCq, OP_SAFE_C_opSQq, HOP_SAFE_C_opSQq, OP_SAFE_C_opQSq, HOP_SAFE_C_opQSq,
      OP_SAFE_C_opCSq, HOP_SAFE_C_opCSq, OP_SAFE_C_S_opSq, HOP_SAFE_C_S_opSq,
      OP_SAFE_C_C_opSCq, HOP_SAFE_C_C_opSCq,
      OP_SAFE_C_S_opSCq, HOP_SAFE_C_S_opSCq, OP_SAFE_C_S_opCSq, HOP_SAFE_C_S_opCSq,
      OP_SAFE_C_opSq_S, HOP_SAFE_C_opSq_S, OP_SAFE_C_opSq_C, HOP_SAFE_C_opSq_C,
      OP_SAFE_C_opSq_opSq, HOP_SAFE_C_opSq_opSq, OP_SAFE_C_S_opSSq, HOP_SAFE_C_S_opSSq, OP_SAFE_C_C_opSq, HOP_SAFE_C_C_opSq,
      OP_SAFE_C_C_opCSq, HOP_SAFE_C_C_opCSq, OP_SAFE_C_opCSq_C, HOP_SAFE_C_opCSq_C,
      OP_SAFE_C_S_opCq, HOP_SAFE_C_S_opCq, OP_SAFE_C_opSSq_C, HOP_SAFE_C_opSSq_C, OP_SAFE_C_C_opSSq, HOP_SAFE_C_C_opSSq,
      OP_SAFE_C_C_opCq, HOP_SAFE_C_C_opCq, OP_SAFE_C_opCq_S, HOP_SAFE_C_opCq_S,
      OP_SAFE_C_opCq_opCq, HOP_SAFE_C_opCq_opCq, OP_SAFE_C_opCq_C, HOP_SAFE_C_opCq_C,
      OP_SAFE_C_opSCq_opSCq, HOP_SAFE_C_opSCq_opSCq, OP_SAFE_C_opSSq_opSSq, HOP_SAFE_C_opSSq_opSSq,
      OP_SAFE_C_opSSq_opCq, HOP_SAFE_C_opSSq_opCq, OP_SAFE_C_opSSq_opSq, HOP_SAFE_C_opSSq_opSq, OP_SAFE_C_opSq_opSSq, HOP_SAFE_C_opSq_opSSq,
      OP_SAFE_C_opSSq_S, HOP_SAFE_C_opSSq_S, OP_SAFE_C_opSCq_S, HOP_SAFE_C_opSCq_S, OP_SAFE_C_opCSq_S, HOP_SAFE_C_opCSq_S,
      OP_SAFE_C_opSCq_C, HOP_SAFE_C_opSCq_C, OP_SAFE_C_opCq_opSSq, HOP_SAFE_C_opCq_opSSq,
      OP_SAFE_C_S_op_opSSq_Sq, HOP_SAFE_C_S_op_opSSq_Sq, OP_SAFE_C_S_op_S_opSSqq, HOP_SAFE_C_S_op_S_opSSqq, OP_SAFE_C_S_op_S_opSqq, HOP_SAFE_C_S_op_S_opSqq,
      OP_SAFE_C_op_opSSq_q_C, HOP_SAFE_C_op_opSSq_q_C, OP_SAFE_C_op_opSq_q_C, HOP_SAFE_C_op_opSq_q_C, 
      OP_SAFE_C_op_opSSq_q_S, HOP_SAFE_C_op_opSSq_q_S, OP_SAFE_C_op_opSq_q_S, HOP_SAFE_C_op_opSq_q_S, 
      OP_SAFE_C_S_op_opSSq_opSSqq, HOP_SAFE_C_S_op_opSSq_opSSqq,
      OP_SAFE_C_op_opSq_q, HOP_SAFE_C_op_opSq_q, 
      OP_SAFE_C_C_op_S_opCqq, HOP_SAFE_C_C_op_S_opCqq,
      OP_SAFE_C_op_S_opSq_q, HOP_SAFE_C_op_S_opSq_q, OP_SAFE_C_op_opSq_S_q, HOP_SAFE_C_op_opSq_S_q, 
      OP_SAFE_C_opSq_Q, HOP_SAFE_C_opSq_Q, OP_SAFE_C_opSq_QS, HOP_SAFE_C_opSq_QS,

      OP_SAFE_C_STAR_A, HOP_SAFE_C_STAR_A, OP_SAFE_C_STAR_AA, HOP_SAFE_C_STAR_AA, OP_SAFE_C_STAR_ALL_X, HOP_SAFE_C_STAR_ALL_X, 

      OP_SAFE_QUOTE, HOP_SAFE_QUOTE,
      OP_SAFE_IFA_SS_A, HOP_SAFE_IFA_SS_A, OP_SAFE_C_P, HOP_SAFE_C_P,

      OP_SAFE_C_Z, HOP_SAFE_C_Z, OP_SAFE_C_ZZ, HOP_SAFE_C_ZZ, OP_SAFE_C_SZ, HOP_SAFE_C_SZ, OP_SAFE_C_ZS, HOP_SAFE_C_ZS,
      OP_SAFE_C_CZ, HOP_SAFE_C_CZ, OP_SAFE_C_ZC, HOP_SAFE_C_ZC, OP_SAFE_C_ZQ, HOP_SAFE_C_ZQ,
      OP_SAFE_C_opCq_Z, HOP_SAFE_C_opCq_Z, OP_SAFE_C_S_opSZq, HOP_SAFE_C_S_opSZq,
      OP_SAFE_C_AZ, HOP_SAFE_C_AZ, OP_SAFE_C_ZA, HOP_SAFE_C_ZA,
      OP_SAFE_C_ZAA, HOP_SAFE_C_ZAA, OP_SAFE_C_AZA, HOP_SAFE_C_AZA, OP_SAFE_C_AAZ, HOP_SAFE_C_AAZ, 
      OP_SAFE_C_ZZA, HOP_SAFE_C_ZZA, OP_SAFE_C_ZAZ, HOP_SAFE_C_ZAZ, OP_SAFE_C_AZZ, HOP_SAFE_C_AZZ,
      OP_SAFE_C_ZZZ, HOP_SAFE_C_ZZZ, 

      OP_THUNK, HOP_THUNK,
      OP_CLOSURE_S, HOP_CLOSURE_S, OP_CLOSURE_C, HOP_CLOSURE_C, OP_CLOSURE_P, HOP_CLOSURE_P, 
      OP_CLOSURE_SS, HOP_CLOSURE_SS, OP_CLOSURE_SS_P, HOP_CLOSURE_SS_P, 
      OP_CLOSURE_SC, HOP_CLOSURE_SC, OP_CLOSURE_CS, HOP_CLOSURE_CS,
      OP_CLOSURE_A, HOP_CLOSURE_A, OP_CLOSURE_AA, HOP_CLOSURE_AA, OP_CLOSURE_A_P, HOP_CLOSURE_A_P,
      OP_CLOSURE_ALL_X, HOP_CLOSURE_ALL_X, OP_CLOSURE_ALL_S, HOP_CLOSURE_ALL_S, OP_CLOSURE_ALL_S_P, HOP_CLOSURE_ALL_S_P,
      OP_CLOSURE_FA, HOP_CLOSURE_FA,
      OP_CLOSURE_AP, HOP_CLOSURE_AP, OP_CLOSURE_PA, HOP_CLOSURE_PA, 
      OP_CLOSURE_ANY_ALL_X, HOP_CLOSURE_ANY_ALL_X, 

      OP_CLOSURE_STAR_A, HOP_CLOSURE_STAR_A, OP_CLOSURE_STAR_AA, HOP_CLOSURE_STAR_AA, OP_CLOSURE_STAR_ALL_X, HOP_CLOSURE_STAR_ALL_X,

      OP_SAFE_THUNK, HOP_SAFE_THUNK, 
      OP_SAFE_THUNK_E, HOP_SAFE_THUNK_E, 
      OP_SAFE_THUNK_P, HOP_SAFE_THUNK_P, OP_SAFE_LTHUNK_P, HOP_SAFE_LTHUNK_P, 

      OP_SAFE_CLOSURE_S, HOP_SAFE_CLOSURE_S, 
      OP_SAFE_CLOSURE_C, HOP_SAFE_CLOSURE_C, OP_SAFE_CLOSURE_P, HOP_SAFE_CLOSURE_P, 
      OP_SAFE_CLOSURE_SS, HOP_SAFE_CLOSURE_SS, OP_SAFE_CLOSURE_SS_B, HOP_SAFE_CLOSURE_SS_B, 
      OP_SAFE_CLOSURE_SC, HOP_SAFE_CLOSURE_SC, OP_SAFE_CLOSURE_CS, HOP_SAFE_CLOSURE_CS,
      OP_SAFE_CLOSURE_A, HOP_SAFE_CLOSURE_A, OP_SAFE_LCLOSURE_A, HOP_SAFE_LCLOSURE_A, OP_SAFE_LCLOSURE_A_P, HOP_SAFE_LCLOSURE_A_P, 
      OP_SAFE_CLOSURE_SA, HOP_SAFE_CLOSURE_SA, 
      OP_SAFE_CLOSURE_S_P, HOP_SAFE_CLOSURE_S_P, OP_SAFE_LCLOSURE_L_P, HOP_SAFE_LCLOSURE_L_P, 
      OP_SAFE_CLOSURE_SAA, HOP_SAFE_CLOSURE_SAA, OP_SAFE_CLOSURE_S_C, HOP_SAFE_CLOSURE_S_C,
      OP_SAFE_CLOSURE_A_C, HOP_SAFE_CLOSURE_A_C, 
      OP_SAFE_CLOSURE_ALL_X, HOP_SAFE_CLOSURE_ALL_X, OP_SAFE_CLOSURE_AA, HOP_SAFE_CLOSURE_AA,
      OP_SAFE_CLOSURE_AP, HOP_SAFE_CLOSURE_AP, OP_SAFE_CLOSURE_PA, HOP_SAFE_CLOSURE_PA, 

      OP_SAFE_CLOSURE_STAR_A, HOP_SAFE_CLOSURE_STAR_A, OP_SAFE_CLOSURE_STAR_AA, HOP_SAFE_CLOSURE_STAR_AA, 
      OP_SAFE_CLOSURE_STAR_S0, HOP_SAFE_CLOSURE_STAR_S0, OP_SAFE_CLOSURE_STAR_ALL_X, HOP_SAFE_CLOSURE_STAR_ALL_X,

      /* these can't be embedded, and have to be the last thing called */
      OP_APPLY_SS, HOP_APPLY_SS, 
      OP_C_ALL_X, HOP_C_ALL_X, OP_CALL_WITH_EXIT, HOP_CALL_WITH_EXIT, 
      OP_C_CATCH, HOP_C_CATCH, OP_C_CATCH_ALL, HOP_C_CATCH_ALL, OP_C_CATCH_ALL_Z, HOP_C_CATCH_ALL_Z,
      OP_C_S_opSq, HOP_C_S_opSq, OP_C_S_opCq, HOP_C_S_opCq, OP_C_SS, HOP_C_SS,
      OP_C_S, HOP_C_S, OP_READ_S, HOP_READ_S, OP_C_P, HOP_C_P, OP_C_Z, HOP_C_Z, OP_C_AP, HOP_C_AP, OP_NOT_P, HOP_NOT_P,
      OP_C_A, HOP_C_A, OP_C_SCS, HOP_C_SCS,
      OP_C_FA, HOP_C_FA, OP_C_AA, HOP_C_AA, OP_C_FA_1, HOP_C_FA_1, 

      OP_GOTO, HOP_GOTO, OP_GOTO_C, HOP_GOTO_C, OP_GOTO_S, HOP_GOTO_S, OP_GOTO_A, HOP_GOTO_A,
      OP_ITERATE, HOP_ITERATE, 
      OP_CONTINUATION_A, HOP_CONTINUATION_A,
      OP_VECTOR_A, HOP_VECTOR_A, OP_CVECTOR_A, HOP_CVECTOR_A, 
      OP_STRING_A, HOP_STRING_A,
      OP_C_OBJECT_A, HOP_C_OBJECT_A, OP_PAIR_A, HOP_PAIR_A, OP_HASH_TABLE_A, HOP_HASH_TABLE_A,
      OP_ENVIRONMENT_Q, HOP_ENVIRONMENT_Q, OP_ENVIRONMENT_A, HOP_ENVIRONMENT_A,

      OP_UNKNOWN, HOP_UNKNOWN, OP_UNKNOWN_ALL_S, HOP_UNKNOWN_ALL_S, OP_UNKNOWN_ALL_X, HOP_UNKNOWN_ALL_X,
      OP_UNKNOWN_G, HOP_UNKNOWN_G, OP_UNKNOWN_GG, HOP_UNKNOWN_GG, OP_UNKNOWN_A, HOP_UNKNOWN_A, OP_UNKNOWN_AA, HOP_UNKNOWN_AA,

      OP_SAFE_C_PP, HOP_SAFE_C_PP,
      OP_SAFE_C_opSq_P, HOP_SAFE_C_opSq_P, 
      OP_SAFE_C_SP, HOP_SAFE_C_SP, OP_SAFE_C_CP, HOP_SAFE_C_CP, OP_SAFE_C_QP, HOP_SAFE_C_QP, OP_SAFE_C_AP, HOP_SAFE_C_AP,
      OP_SAFE_C_PS, HOP_SAFE_C_PS, OP_SAFE_C_PC, HOP_SAFE_C_PC, OP_SAFE_C_PQ, HOP_SAFE_C_PQ,
      OP_SAFE_C_AAP, HOP_SAFE_C_AAP,
      
      OP_S, HOP_S, OP_S_S, HOP_S_S, OP_S_C, HOP_S_C, OP_S_A, HOP_S_A,
      OPT_MAX_DEFINED
};

#if DEBUGGING || OP_NAMES

static const char *op_names[OP_MAX_DEFINED_1] = {
      "no_op", "gc_protect", 
      "read_internal", "eval",
      "eval_args", "eval_args1", "eval_args2", "eval_args3", "eval_args4", "eval_args5",
      "apply", "eval_macro", "lambda", "quote", "macroexpand",
      "define", "define1", "begin", "begin1",
      "if", "if1", "when", "when1", "unless", "unless1", "set", "set1", "set2",
      "let", "let1", "let*", "let*1", "let*2",
      "letrec", "letrec1", "letrec*", "letrec*1", 
      "let_temporarily", "let_temp_unchecked", "let_temp_init1", "let_temp_init2", "let_temp_done", "let_temp_done1",
      "cond", "cond1", "cond1_1", "cond_simple", "cond1_simple", "cond_unchecked_z", 
      "and", "and1", "or", "or1",
      "define_macro", "define_macro*", "define_expansion",
      "case", "read_list", "read_next", "read_dot", "read_quote",
      "read_quasiquote", "read_unquote", "read_apply_values",
      "read_vector", "read_byte_vector", "read_int_vector", "read_float_vector", "read_done",
      "load_return_if_eof", "load_close_and_pop_if_eof", "eval_done",
      "catch", "dynamic_wind", "define_constant", "define_constant1",
      "do", "do_end", "do_end1", "do_step", "do_step2", "do_init",
      "define*", "lambda*", "lambda*_default", "error_quit", "unwind_input", "unwind_output",
      "error_hook_quit",
      "with_let", "with_let1", "with_let_unchecked", "with_let_s",
      "with_baffle", "with_baffle_unchecked", "expansion",
      "for_each", "for_each_1", "for_each_2", "for_each_3", 
      "map", "map_1", "map_2", "map_gather", "map_gather_1", "map_gather_2", "map_gather_3", 
      "barrier", "deactivate_goto",

      "define_bacro", "define_bacro*",
      "get_output_string", "get_output_string_1",
      "sort", "sort1", "sort2", "sort3", "sort_pair_end", "sort_vector_end", "sort_string_end",
      "eval_string_1", "eval_string_2",
      "member_if", "assoc_if", "member_if1", "assoc_if1",

      "quote_unchecked", "lambda_unchecked", "let_unchecked", "case_unchecked", "when_unchecked", "unless_unchecked",

      "set_unchecked", "set_symbol_c", "set_symbol_l", "set_symbol_s", "set_symbol_q", "set_symbol_p", "set_symbol_z", "set_symbol_a",
      "set_symbol_opsq", "set_symbol_opcq", "set_symbol_opssq", "set_symbol_opsssq", 
      "set_symbol_oplq", "set_symbol_opllq", 
      "set_normal", "set_pair", "set_dilambda", "set_dilambda_z", "set_dilambda_z_1",
      "set_pair_z", "set_pair_a", "set_pair_p", "set_pair_za",
      "set_pair_p_1", "set_with_accessor", "set_pws", "set_let_s", "set_let_all_x",
      "set_safe",
      "increment_1", "decrement_1", "set_cons",
      "increment_ss", "increment_sss", "increment_sz", "increment_sa", "increment_saa",

      "let*_unchecked", "letrec_unchecked", "letrec*_unchecked", "cond_unchecked",
      "lambda*_unchecked", "do_unchecked", "define_unchecked", "define*_unchecked", "define_funchecked", "define_constant_unchecked",
      "define_with_accessor", "define_macro_with_accessor",

      "let_no_vars", "named_let", "named_let_no_vars", "named_let*",
      "let_c", "let_s", "let_s_z", "let_all_c", "let_all_s", "let_all_x",
      "let*_all_x", "let*_a2", "let*_a", "let_opcq", "let_opssq", "let_opssq_e", "let_opassq_e",
      "let_opsq", "let_all_opsq", "let_opsq_p", "let_car", "let_one", "let_one_1", "let_z", "let_z_1", "let_a", "let_a_z",

      "case_a_e_s", "case_a_i_s", "case_a_g_s", "case_a_e_g", "case_a_g_g", 
      "case_s_e_s", "case_s_i_s", "case_s_g_s", "case_s_e_g", "case_s_g_g", 
      "case_p_e_s", "case_p_i_s", "case_p_g_s", "case_p_e_g", "case_p_g_g", 
      "case_e_s", "case_i_s", "case_g_s", "case_e_g", "case_g_g", 

      "if_unchecked", "and_unchecked", "and_p", "and_p1", "and_ap", "and_az", "and_safe_p", "and_safe_aa", 
      "or_unchecked", "or_p", "or_p1", "or_ap", "or_safe_p", "or_safe_aa",
      "cond_feed", "cond_feed_1", "when_s", "when_a", "when_p", "unless_s", "unless_a",

      "if_s_p", "if_s_p_p", "if_s_r", "if_s_n", "if_s_n_n",
      "if_c_p", "if_c_p_p", "if_c_r", "if_c_n", "if_c_n_n",
      "if_cs_p", "if_cs_p_p", "if_cs_r", "if_cs_n", "if_cs_n_n",
      "if_csq_p", "if_csq_p_p", "if_csq_r", "if_csq_n", "if_csq_n_n",
      "if_css_p", "if_css_p_p", "if_css_r","if_css_n", "if_css_n_n",
      "if_csc_p", "if_csc_p_p", "if_csc_r", "if_csc_n", "if_csc_n_n",
      "if_opsq_p", "if_opsq_p_p", "if_opsq_r", "if_opsq_n", "if_opsq_n_n",
      "if_s_opcq_p", "if_s_opcq_p_p", "if_s_opcq_r","if_s_opcq_n", "if_s_opcq_n_n",
      "if_is_type_s_p", "if_is_type_s_p_p", "if_is_type_s_r", "if_is_type_s_n", "if_is_type_s_n_n",
      "if_is_type_opsq_p", "if_is_type_opsq_p_p", "if_is_type_opsq_r", "if_is_type_opsq_n", "if_is_type_opsq_n_n",
      "if_a_p", "if_a_p_p", "if_a_r", "if_a_n", "if_a_n_n",
      "if_and2_p", "if_and2_p_p", "if_and2_r","if_and2_n", "if_and2_n_n",
      "if_and3_p", "if_and3_p_p", "if_and3_r","if_and3_n", "if_and3_n_n",
      "if_z_p", "if_z_p_p", "if_z_r", "if_z_n", "if_z_n_n",
      "if_p_p", "if_p_p_p", "if_p_r", "if_p_n", "if_p_n_n",
      "if_andp_p", "if_andp_p_p", "if_andp_r", "if_andp_n", "if_andp_n_n",
      "if_orp_p", "if_orp_p_p", "if_orp_r","if_orp_n", "if_orp_n_n",
      "if_or2_p", "if_or2_p_p", "if_or2_r","if_or2_n", "if_or2_n_n",

      "if_ppp", "if_pp", "if_pr", "if_prr",
      "when_pp",

      "catch_1", "catch_2", "catch_all", "cond_all_x", "cond_all_x_2", "cond_all_x_z",
      "simple_do", "simple_do_step", "safe_dotimes", "safe_dotimes_step", "safe_dotimes_step_p", "safe_dotimes_step_o", 
      "safe_do", "safe_do_step", "dox", "dox_step", "dox_step_p", 
      "dotimes_p", "dotimes_step_p",
      "do_no_vars", "do_no_vars_no_opt", "do_no_vars_no_opt_1",
      "dotimes_one_step",

      "safe_c_p_1", "safe_c_pp_1", "safe_c_pp_3", "safe_c_pp_5", "safe_c_pp_6",
      "eval_args_p_2", "eval_args_p_2_mv", "eval_args_p_3", "eval_args_p_4", "eval_args_p_3_mv",
      "eval_args_aap_1", "eval_args_aap_mv", "eval_macro_mv", "macroexpand_1", "apply_lambda",
      "safe_closure_p_1", "closure_p_1", "safe_closure_ap_1", "safe_closure_pa_1", 
      "safe_c_zz_1", "safe_c_za_1", "increment_sz_1", "safe_c_sz_sz",
      "safe_c_zaa_1", "safe_c_aza_1", "safe_c_aaz_1", 
      "safe_c_zza_1", "safe_c_zza_2", "safe_c_zaz_1", "safe_c_azz_1", "safe_c_azz_2",
      "safe_c_zzz_1", "safe_c_zzz_2",
      "c_p_1", "c_p_2", "c_ap_1", "not_1",
      "closure_ap_1", "closure_pa_1", 
      "closure_p_mv", "closure_ap_mv", "closure_pa_mv",

      "set_with_let_1", "set_with_let_2",
};

static const char* opt_names[OPT_MAX_DEFINED] =
     {"safe_c_c", "h_safe_c_c", 
      "safe_c_and2", "h_safe_c_and2", "safe_c_or2", "h_safe_c_or2",
      "safe_c_s", "h_safe_c_s", "safe_c_l", "h_safe_c_l", 
      "safe_car_s", "h_safe_car_s", "safe_cdr_s", "h_safe_cdr_s", "safe_cadr_s", "h_safe_cadr_s",
      "safe_is_pair_s", "h_safe_is_pair_s", "safe_is_null_s", "h_safe_is_null_s", "safe_is_symbol_s", "h_safe_is_symbol_s", 
      "safe_c_ss", "h_safe_c_ss", "safe_c_ll", "h_safe_c_ll", 
      "safe_c_sc", "h_safe_c_sc", "safe_c_cs", "h_safe_c_cs",
      "safe_c_sq", "h_safe_c_sq", "safe_c_qs", "h_safe_c_qs", "safe_c_qq", "h_safe_c_qq",
      "safe_c_cq", "h_safe_c_cq", "safe_c_qc", "h_safe_c_qc",
      "safe_c_sss", "h_safe_c_sss", "safe_c_scs", "h_safe_c_scs", "safe_c_ssc", "h_safe_c_ssc", "safe_c_css", "h_safe_c_css",
      "safe_c_scc", "h_safe_c_scc", "safe_c_csc", "h_safe_c_csc",
      "safe_c_all_s", "h_safe_c_all_s", "safe_c_all_x", "h_safe_c_all_x", "safe_c_ssa", "h_safe_c_ssa", "safe_c_sas", "h_safe_c_sas",
      "safe_c_csa", "h_safe_c_csa", "safe_c_sca", "h_safe_c_sca", 
      "safe_c_a", "h_safe_c_a", "safe_c_aa", "h_safe_c_aa", "safe_c_aaa", "h_safe_c_aaa", "safe_c_aaaa", "h_safe_c_aaaa",
      "safe_c_sqs", "h_safe_c_sqs", "safe_c_opaq", "h_safe_c_opaq", "safe_c_opaaq", "h_safe_c_opaaq", "safe_c_opaaaq", "h_safe_c_opaaaq",
      "safe_c_s_opaq", "h_safe_c_s_opaq", "safe_c_s_opaaq", "h_safe_c_s_opaaq", "safe_c_s_opaaaq", "h_safe_c_s_opaaaq",

      "safe_c_opcq", "h_safe_c_opcq", "safe_c_opsq", "h_safe_c_opsq",
      "safe_c_opcq_opsq", "h_safe_c_opcq_opsq", "safe_c_opsq_opcq", "h_safe_c_opsq_opcq",
      "safe_c_opssq", "h_safe_c_opssq", "safe_c_opscq", "h_safe_c_opscq", "safe_c_opsqq", "h_safe_c_opsqq", "safe_c_opqsq", "h_safe_c_opqsq",
      "safe_c_opcsq", "h_safe_c_opcsq", "safe_c_s_opsq", "h_safe_c_s_opsq",
      "safe_c_c_opscq", "h_safe_c_c_opscq",
      "safe_c_s_opscq", "h_safe_c_s_opscq", "safe_c_s_opcsq", "h_safe_c_s_opcsq",
      "safe_c_opsq_s", "h_safe_c_opsq_s", "safe_c_opsq_c", "h_safe_c_opsq_c",
      "safe_c_opsq_opsq", "h_safe_c_opsq_opsq", "safe_c_s_opssq", "h_safe_c_s_opssq", "safe_c_c_opsq", "h_safe_c_c_opsq",
      "safe_c_c_opcsq", "h_safe_c_c_opcsq", "safe_c_opcsq_c", "h_safe_c_opcsq_c",
      "safe_c_s_opcq", "h_safe_c_s_opcq", "safe_c_opssq_c", "h_safe_c_opssq_c", "safe_c_c_opssq", "h_safe_c_c_opssq",
      "safe_c_c_opcq", "h_safe_c_c_opcq", "safe_c_opcq_s", "h_safe_c_opcq_s",
      "safe_c_opcq_opcq", "h_safe_c_opcq_opcq", "safe_c_opcq_c", "h_safe_c_opcq_c",
      "safe_c_opscq_opscq", "h_safe_c_opscq_opscq", "safe_c_opssq_opssq", "h_safe_c_opssq_opssq",
      "safe_c_opssq_opcq", "h_safe_c_opssq_opcq", "safe_c_opssq_opsq", "h_safe_c_opssq_opsq", "safe_c_opsq_opssq", "h_safe_c_opsq_opssq",
      "safe_c_opssq_s", "h_safe_c_opssq_s", "safe_c_opscq_s", "h_safe_c_opscq_s", "safe_c_opcsq_s", "h_safe_c_opcsq_s",
      "safe_c_opscq_c", "h_safe_c_opscq_c", "safe_c_opcq_opssq", "h_safe_c_opcq_opssq",
      "safe_c_s_op_opssq_sq", "h_safe_c_s_op_opssq_sq", "safe_c_s_op_s_opssqq", "h_safe_c_s_op_s_opssqq", "safe_c_s_op_s_opsqq", "h_safe_c_s_op_s_opsqq",
      "safe_c_op_opssq_q_c", "h_safe_c_op_opssq_q_c", "safe_c_op_opsq_q_c", "h_safe_c_op_opsq_q_c", 
      "safe_c_op_opssq_q_s", "h_safe_c_op_opssq_q_s", "safe_c_op_opsq_q_s", "h_safe_c_op_opsq_q_s", 
      "safe_c_s_op_opssq_opssqq", "h_safe_c_s_op_opssq_opssqq",
      "safe_c_op_opsq_q", "h_safe_c_op_opsq_q", 
      "safe_c_c_op_s_opcqq", "h_safe_c_c_op_s_opcqq",
      "safe_c_op_s_opsq_q", "h_safe_c_op_s_opsq_q", "safe_c_op_opsq_s_q", "h_safe_c_op_opsq_s_q",
      "safe_c_opsq_q", "h_safe_c_opsq_q", "safe_c_opsq_q_s", "h_safe_c_opsq_q_s",
      
      "safe_c*_a", "h_safe_c*_a", "safe_c*_aa", "h_safe_c*_aa", "safe_c*_all_x", "h_safe_c*_all_x", 

      "safe_quote", "h_safe_quote",
      "safe_ifa_ss_a", "h_safe_ifa_ss_a", "safe_c_p", "h_safe_c_p",

      "safe_c_z", "h_safe_c_z", "safe_c_zz", "h_safe_c_zz", "safe_c_sz", "h_safe_c_sz", "safe_c_zs", "h_safe_c_zs",
      "safe_c_cz", "h_safe_c_cz", "safe_c_zc", "h_safe_c_zc", "safe_c_zq", "h_safe_c_zq",
      "safe_c_opcq_z", "h_safe_c_opcq_z", "safe_c_s_opszq", "h_safe_c_s_opszq",
      "safe_c_az", "h_safe_c_az", "safe_c_za", "h_safe_c_za",
      "safe_c_zaa", "h_safe_c_zaa", "safe_c_aza", "h_safe_c_aza", "safe_c_aaz", "h_safe_c_aaz", 
      "safe_c_zza", "h_safe_c_zza", "safe_c_zaz", "h_safe_c_zaz", "safe_c_azz", "h_safe_c_azz",
      "safe_c_zzz", "h_safe_c_zzz", 

      "thunk", "h_thunk",
      "closure_s", "h_closure_s", "closure_c", "h_closure_c", "closure_p", "h_closure_p", 
      "closure_ss", "h_closure_ss", "closure_ss_p", "h_closure_ss_p", 
      "closure_sc", "h_closure_sc", "closure_cs", "h_closure_cs",
      "closure_a", "h_closure_a", "closure_aa", "h_closure_aa", "closure_a_p", "h_closure_a_p", 
      "closure_all_x", "h_closure_all_x", "closure_all_s", "h_closure_all_s", "closure_all_s_p", "h_closure_all_s_p",
      "closure_fa", "h_closure_fa",
      "closure_ap", "h_closure_ap", "closure_pa", "h_closure_pa", 
      "closure_any_all_x", "h_closure_any_all_x", 

      "closure*_a", "h_closure*_a", "closure*_aa", "h_closure*_aa", "closure*_all_x", "h_closure*_all_x",

      "safe_thunk", "h_safe_thunk", 
      "safe_thunk_e", "h_safe_thunk_e", 
      "safe_thunk_p", "h_safe_thunk_p", "safe_lthunk", "h_safe_lthunk", 
      "safe_closure_s", "h_safe_closure_s", 
      "safe_closure_c", "h_safe_closure_c", "safe_closure_p", "h_safe_closure_p", 
      "safe_closure_ss", "h_safe_closure_ss", "safe_closure_ss_b", "h_safe_closure_ss_b", 
      "safe_closure_sc", "h_safe_closure_sc", "safe_closure_cs", "h_safe_closure_cs",
      "safe_closure_a", "h_safe_closure_a", "safe_lclosure_a", "h_safe_lclosure_a", "safe_lclosure_a_p", "h_safe_lclosure_a_p", 
      "safe_closure_sa", "h_safe_closure_sa", 
      "safe_closure_s_p", "h_safe_closure_s_p", "safe_lclosure_l_p", "h_safe_lclosure_l_p", 
      "safe_closure_saa", "h_safe_closure_saa", "safe_closure_s_c", "h_safe_closure_s_c",
      "safe_closure_a_c", "h_safe_closure_a_c", 
      "safe_closure_all_x", "h_safe_closure_all_x", "safe_closure_aa", "h_safe_closure_aa",
      "safe_closure_ap", "h_safe_closure_ap", "safe_closure_pa", "h_safe_closure_pa", 

      "safe_closure*_a", "h_safe_closure*_a", "safe_closure*_aa", "h_safe_closure*_aa", 
      "safe_closure*_s0", "h_safe_closure*_s0", "safe_closure*_all_x", "h_safe_closure*_all_x",

      "apply_ss", "h_apply_ss",
      "c_all_x", "h_c_all_x", "call_with_exit", "h_call_with_exit", 
      "c_catch", "h_c_catch", "c_catch_all", "h_c_catch_all", "c_catch_all_z", "h_c_catch_all_z",
      "c_s_opsq", "h_c_s_opsq", "c_s_opcq", "h_c_s_opcq", "c_ss", "h_c_ss",
      "c_s", "h_c_s", "read_s", "h_read_s", "c_p", "h_c_p", "c_z", "h_c_z", "c_ap", "h_c_ap", "c_not", "h_c_not",
      "c_a", "h_c_a", "c_scs", "h_c_scs",
      "c_fa", "h_c_fa", "c_aa", "h_c_aa", "c_fa_1", "h_c_fa_1",

      "goto", "h_goto", "goto_c", "h_goto_c", "goto_s", "h_goto_s", "goto_a", "h_goto_a",
      "iterate", "h_iterate",
      "continuation_a", "h_continuation_a", 
      "vector_a", "h_vector_a", "cvector_a", "h_cvector_a", 
      "string_a", "h_string_a",
      "c_object_a", "h_c_object_a", "pair_a", "h_pair_a", "hash_table_a", "h_hash_table_a",
      "environment_q", "h_environment_q", "environment_a", "h_environment_a",

      "unknown", "h_unknown", "unknown_all_s", "h_unknown_all_s", "unknown_all_x", "h_unknown_all_x",
      "unknown_g", "h_unknown_g", "unknown_gg", "h_unknown_gg", "unknown_a", "h_unknown_a", "unknown_aa", "h_unknown_aa",

      "safe_c_pp", "h_safe_c_pp",
      "safe_c_opsq_p", "h_safe_c_opsq_p", 
      "safe_c_sp", "h_safe_c_sp", "safe_c_cp", "h_safe_c_cp", "safe_c_qp", "h_safe_c_qp", "safe_c_ap", "h_safe_c_ap",
      "safe_c_ps", "h_safe_c_ps", "safe_c_pc", "h_safe_c_pc", "safe_c_pq", "h_safe_c_pq",
      "safe_c_aap", "h_safe_c_aap",
      
      "s", "h_s", "s_s", "h_s_s", "s_c", "h_s_c", "s_a", "h_s_a",
};
#endif

#define in_reader(Sc) ((sc->cur_op >= OP_READ_LIST) && (sc->cur_op <= OP_READ_DONE) && (is_input_port(Sc->input_port)))

#define is_safe_c_op(op) (op < OP_THUNK)
#define is_unknown_op(op) ((op >= OP_UNKNOWN) && (op < OP_SAFE_C_PP))
#define is_callable_c_op(op) ((op < OP_THUNK) || (op > OP_UNKNOWN_AA)) /* used only in check_set */

static bool is_h_optimized(s7_pointer p)
{
  return((is_optimized(p)) &&
	 ((optimize_op(p) & 1) != 0) &&
	 (!is_unknown_op(optimize_op(p))));
}

#define is_h_safe_c_c(P) ((is_optimized(P)) && (optimize_op(P) >= HOP_SAFE_C_C) && (optimize_op(P) < OP_SAFE_C_S) && ((optimize_op(P) & 1) != 0)) 
#define is_h_safe_c_s(P) ((is_optimized(P)) && (optimize_op(P) >= HOP_SAFE_C_S) && (optimize_op(P) <= HOP_SAFE_IS_SYMBOL_S) && ((optimize_op(P) & 1) != 0))
#define is_safe_c_s(P)   ((is_optimized(P)) && (optimize_op(P) >= HOP_SAFE_C_S) && (optimize_op(P) <= HOP_SAFE_IS_SYMBOL_S))

static int32_t position_of(s7_pointer p, s7_pointer args)
{
  int32_t i;
  for (i = 1; p != args; i++, args = cdr(args));
  return(i);
}

s7_pointer s7_method(s7_scheme *sc, s7_pointer obj, s7_pointer method)
{
  if (has_methods(obj))
    return(find_method(sc, find_let(sc, obj), method));
  return(sc->undefined);
}


/* if a method is shadowing a built-in like abs, it should expect the same args as abs and
 *   behave the same -- no multiple values etc.
 */
#define check_method(Sc, Obj, Method, Args)     \
  {                                             \
    s7_pointer func; 	                        \
    if ((has_methods(Obj)) && ((func = find_method(Sc, find_let(Sc, Obj), Method)) != Sc->undefined)) \
      return(s7_apply_function(Sc, func, Args)); \
  }

#define apply_known_method(Sc, Let, Method, Args) return(s7_apply_function(Sc, find_method(Sc, Let, Method), Args))

#define check_two_methods(Sc, Obj, Method1, Method2, Args)	                            \
  if (has_methods(Obj))                                                                     \
    {                                                                                       \
      s7_pointer func;							                    \
      func = find_method(Sc, find_let(Sc, Obj), Method1);                           \
      if ((func == Sc->undefined) && (Method1 != Method2) && (Method2)) func = find_method(Sc, find_let(Sc, Obj), Method2); \
      if (func != Sc->undefined) return(s7_apply_function(Sc, func, Args)); \
    }

static s7_pointer check_values(s7_scheme *sc, s7_pointer obj, s7_pointer args)
{
  check_method(sc, obj, sc->values_symbol, args);
  return(sc->gc_nil);
}

/* unfortunately, in the simplest cases, where a function (like number?) accepts any argument,
 *   this costs about a factor of 1.5 in speed (we're doing the normal check like s7_is_number,
 *   but then have to check has_methods before returning #f).  We can't use the old form until
 *   openlet is seen because the prior code might use #_number? which gets the value
 *   before the switch.  These simple functions normally do not dominate timing info, so I'll
 *   go ahead. It's mostly boilerplate:
 */

static s7_pointer apply_boolean_method(s7_scheme *sc, s7_pointer obj, s7_pointer method);
static s7_pointer missing_method_error(s7_scheme *sc, s7_pointer method, s7_pointer obj);

#define check_boolean_method(Sc, Checker, Method, Args)      \
  {                                                          \
    s7_pointer p;                                            \
    p = car(Args);                                           \
    if (Checker(p)) return(Sc->T);                           \
    if (!has_methods(p)) return(Sc->F);		       \
    return(apply_boolean_method(Sc, p, Method));	       \
  }

#define check_boolean_not_method(Sc, Checker, Method, Args)  \
  {                                                          \
    s7_pointer p;                                            \
    p = find_symbol_unchecked(sc, cadar(Args));                                           \
    if (Checker(p)) return(Sc->F);                           \
    if (!has_methods(p)) return(Sc->T);		       \
    return((apply_boolean_method(Sc, p, Method) == sc->F) ? sc->T : sc->F); \
  }

#define eval_boolean_method(Sc, Checker, Method, Arg) \
    if (Checker(Arg))				      \
      Sc->value = Sc->T;			      \
    else					      \
      {						      \
	s7_pointer func;						\
	if ((has_methods(Arg)) && ((func = find_method(Sc, find_let(Sc, Arg), Method)) != Sc->undefined)) \
	  Sc->value = s7_apply_function(Sc, func, list_1(Sc, Arg));	\
	else Sc->value = Sc->F;						\
      }

static s7_pointer find_and_apply_method(s7_scheme *sc, s7_pointer lt, s7_pointer sym, s7_pointer args)
{
  s7_pointer func;
  func = find_method(sc, lt, sym);
  if (func != sc->undefined)
    return(s7_apply_function(sc, func, args));
  return(missing_method_error(sc, sym, lt));   
}

#define method_or_bust(Sc, Obj, Method, Args, Type, Num)		\
  do {									\
    if (has_methods(Obj))						\
      return(find_and_apply_method(Sc, find_let(Sc, Obj), Method, Args)); \
    return(wrong_type_argument(Sc, Method, Num, Obj, Type));	\
  } while (0)

#define method_or_bust_one_arg(Sc, Obj, Method, Args, Type)		\
  do {									\
    if (has_methods(Obj))						\
      return(find_and_apply_method(Sc, find_let(Sc, Obj), Method, Args));\
    return(simple_wrong_type_argument(Sc, Method, Obj, Type)); \
  } while (0)


#define method_or_bust_with_type(Sc, Obj, Method, Args, Type, Num)	\
  do {									\
    if (has_methods(Obj))						\
      return(find_and_apply_method(Sc, find_let(Sc, Obj), Method, Args));\
    return(wrong_type_argument_with_type(Sc, Method, Num, Obj, Type));		\
  } while (0)

#define method_or_bust_with_type_one_arg(Sc, Obj, Method, Args, Type)	\
  do {									\
    if (has_methods(Obj))						\
      return(find_and_apply_method(Sc, find_let(Sc, Obj), Method, Args));\
    return(simple_wrong_type_argument_with_type(Sc, Method, Obj, Type)); \
  } while (0)


#define eval_error_any(Sc, ErrType, ErrMsg, Obj) \
  do {static s7_pointer _Err_ = NULL; if (!_Err_) _Err_ = s7_make_permanent_string(ErrMsg); \
      return(s7_error(Sc, ErrType, set_elist_2(Sc, _Err_, Obj)));} while (0)

#define eval_error(Sc, ErrMsg, Obj)       eval_error_any(Sc, Sc->syntax_error_symbol, ErrMsg, Obj)
#define eval_type_error(Sc, ErrMsg, Obj)  eval_error_any(Sc, Sc->wrong_type_arg_symbol, ErrMsg, Obj)
#define eval_range_error(Sc, ErrMsg, Obj) eval_error_any(Sc, Sc->out_of_range_symbol, ErrMsg, Obj)

#define eval_error_no_return(Sc, ErrType, ErrMsg, Obj)	\
  do {static s7_pointer _Err_ = NULL; \
      if (!_Err_) _Err_ = s7_make_permanent_string(ErrMsg); \
      s7_error(Sc, ErrType, set_elist_2(Sc, _Err_, Obj));} while (0)

#define eval_error_with_caller(Sc, ErrMsg, Caller, Obj)	\
  do {static s7_pointer _Err_ = NULL; \
      if (!_Err_) _Err_ = s7_make_permanent_string(ErrMsg); \
      return(s7_error(Sc, Sc->syntax_error_symbol, set_elist_3(Sc, _Err_, Caller, Obj)));} while (0)

#define eval_error_with_caller_and_print_limit(Sc, ErrMsg, Caller, Obj)	\
  do {static s7_pointer _Err_ = NULL; \
      if (!_Err_) _Err_ = s7_make_permanent_string(ErrMsg); \
      return(s7_error(Sc, Sc->syntax_error_symbol, set_elist_3(Sc, _Err_, Caller, print_truncate(Sc, Obj))));} while (0)

static s7_pointer set_elist_1(s7_scheme *sc, s7_pointer x1)
{
  set_car(sc->elist_1, x1);
  return(sc->elist_1);
} 

static s7_pointer set_elist_2(s7_scheme *sc, s7_pointer x1, s7_pointer x2)
{
  set_car(sc->elist_2, x1);
  set_cadr(sc->elist_2, x2);
  return(sc->elist_2);
} 

static s7_pointer set_elist_3(s7_scheme *sc, s7_pointer x1, s7_pointer x2, s7_pointer x3)
{
  s7_pointer p;
  p = sc->elist_3;
  set_car(p, x1); p = cdr(p);
  set_car(p, x2); p = cdr(p);
  set_car(p, x3);
  return(sc->elist_3);
} 

static s7_pointer set_elist_4(s7_scheme *sc, s7_pointer x1, s7_pointer x2, s7_pointer x3, s7_pointer x4)
{
  s7_pointer p;
  p = sc->elist_4;
  set_car(p, x1); p = cdr(p);
  set_car(p, x2); p = cdr(p);
  set_car(p, x3); p = cdr(p);
  set_car(p, x4);
  return(sc->elist_4);
} 

static s7_pointer set_elist_5(s7_scheme *sc, s7_pointer x1, s7_pointer x2, s7_pointer x3, s7_pointer x4, s7_pointer x5)
{
  s7_pointer p;
  p = sc->elist_5;
  set_car(p, x1); p = cdr(p);
  set_car(p, x2); p = cdr(p);
  set_car(p, x3); p = cdr(p);
  set_car(p, x4); p = cdr(p);
  set_car(p, x5);
  return(sc->elist_5);
} 

static s7_pointer set_wlist_3(s7_scheme *sc, s7_pointer lst, s7_pointer x1, s7_pointer x2, s7_pointer x3)
{
  s7_pointer p;
  p = lst;
  set_car(p, x1); p = cdr(p);
  set_car(p, x2); p = cdr(p);
  set_car(p, x3);
  return(lst);
} 

static s7_pointer set_wlist_4(s7_scheme *sc, s7_pointer lst, s7_pointer x1, s7_pointer x2, s7_pointer x3, s7_pointer x4)
{
  s7_pointer p;
  p = lst;
  set_car(p, x1); p = cdr(p);
  set_car(p, x2); p = cdr(p);
  set_car(p, x3); p = cdr(p);
  set_car(p, x4);
  return(lst);
} 

static s7_pointer set_plist_1(s7_scheme *sc, s7_pointer x1)
{
  set_car(sc->plist_1, x1);
  return(sc->plist_1);
} 

static s7_pointer set_plist_2(s7_scheme *sc, s7_pointer x1, s7_pointer x2)
{
  set_car(sc->plist_2, x1);
  set_cadr(sc->plist_2, x2);
  return(sc->plist_2);
} 

static s7_pointer set_qlist_2(s7_scheme *sc, s7_pointer x1, s7_pointer x2)
{
  set_car(sc->qlist_2, x1);
  set_cadr(sc->qlist_2, x2);
  return(sc->qlist_2);
} 

static s7_pointer set_plist_3(s7_scheme *sc, s7_pointer x1, s7_pointer x2, s7_pointer x3)
{
  return(set_wlist_3(sc, sc->plist_3, x1, x2, x3));
} 



/* -------------------------------- constants -------------------------------- */

s7_pointer s7_f(s7_scheme *sc)
{
  return(sc->F);
}


s7_pointer s7_t(s7_scheme *sc)
{
  return(sc->T);
}


s7_pointer s7_nil(s7_scheme *sc)
{
  return(sc->nil);
}


bool s7_is_null(s7_scheme *sc, s7_pointer p)
{
  return(is_null(p));
}

static bool is_null_b(s7_pointer p) {return(type(p) == T_NIL);}


s7_pointer s7_undefined(s7_scheme *sc)
{
  return(sc->undefined);
}


s7_pointer s7_unspecified(s7_scheme *sc)
{
  return(sc->unspecified);
}


bool s7_is_unspecified(s7_scheme *sc, s7_pointer val)
{
  return(is_unspecified(val));
}


s7_pointer s7_eof_object(s7_scheme *sc)          /* returns #<eof> -- not equivalent to "eof-object?" */
{
  return(sc->eof_object);
}


static s7_pointer g_not(s7_scheme *sc, s7_pointer args)
{
  #define H_not "(not obj) returns #t if obj is #f, otherwise #t: (not ()) -> #f"
  #define Q_not pl_bt
  return(make_boolean(sc, is_false(sc, car(args))));
}

static bool not_b(s7_pointer p) {return(p == cur_sc->F);}


bool s7_boolean(s7_scheme *sc, s7_pointer x)
{
  return(x != sc->F);
}


bool s7_is_boolean(s7_pointer x)
{
  return(type(x) == T_BOOLEAN);
}


s7_pointer s7_make_boolean(s7_scheme *sc, bool x)
{
  return(make_boolean(sc, x));
}


static s7_pointer g_is_boolean(s7_scheme *sc, s7_pointer args)
{
  #define H_is_boolean "(boolean? obj) returns #t if obj is #f or #t: (boolean? ()) -> #f"
  #define Q_is_boolean pl_bt
  check_boolean_method(sc, s7_is_boolean, sc->is_boolean_symbol, args);
}


bool s7_is_constant(s7_pointer p)
{
  /* this means "always evaluates to the same thing", sort of, not "evaluates to itself":
   *   (let ((x 'x)) (and (not (constant? x)) (equal? x (eval x))))
   *   (and (constant? (list + 1)) (not (equal? (list + 1) (eval (list + 1)))))
   * but since (constant? (vector 1 2 3)) and (constant? #(1 2 3)) are both #t,
   *   how to tell in scheme that (vector-set! x 0 y) is safe?
   *   see end of this file
   */
  return((type(p) != T_SYMBOL) || (is_immutable_symbol(p)));
}


static s7_pointer g_is_constant(s7_scheme *sc, s7_pointer args)
{
  #define H_is_constant "(constant? obj) returns #t if obj is a constant (unsettable): (constant? pi) -> #t"
  #define Q_is_constant pl_bt
  check_boolean_method(sc, s7_is_constant, sc->is_constant_symbol, args);
}



/* -------------------------------- GC -------------------------------- */

/* in most code, pairs, lets, and slots dominate the heap -- each about 25% to 40% of the
 *   total cell allocations.  In snd-test, reals are 50%. slots need not be in the heap,
 *   but moving them out to their own free list was actually slower because we need (in that
 *   case) to manage them in the sweep process by tracking lets.  
 */

uint32_t s7_gc_protect(s7_scheme *sc, s7_pointer x)
{
  uint32_t loc;

  if (sc->gpofl_loc < 0)
    {
      uint32_t i, size, new_size;
      size = sc->protected_objects_size;
      new_size = 2 * size;
      vector_elements(sc->protected_objects) = (s7_pointer *)realloc(vector_elements(sc->protected_objects), new_size * sizeof(s7_pointer));
      vector_length(sc->protected_objects) = new_size;
      sc->protected_objects_size = new_size;
      sc->gpofl = (uint32_t *)realloc(sc->gpofl, new_size * sizeof(uint32_t));
      for (i = size; i < new_size; i++)
	{
	  vector_element(sc->protected_objects, i) = sc->gc_nil;
	  sc->gpofl[++sc->gpofl_loc] = i;
	}
    }

  loc = sc->gpofl[sc->gpofl_loc--];
#if DEBUGGING
  if (loc >= sc->protected_objects_size)
    fprintf(stderr, "sc->gpofl loc: %u (%d)\n", loc, sc->protected_objects_size);
  if (vector_element(sc->protected_objects, loc) != sc->gc_nil)
    fprintf(stderr, "protected object at %u about to be clobbered? %s\n", loc, DISPLAY(vector_element(sc->protected_objects, loc)));
#endif
  vector_element(sc->protected_objects, loc) = x;
  return(loc);
}

void s7_gc_unprotect(s7_scheme *sc, s7_pointer x)
{
  uint32_t i;

  for (i = 0; i < sc->protected_objects_size; i++)
    if (vector_element(sc->protected_objects, i) == x)
      {
	vector_element(sc->protected_objects, i) = sc->gc_nil;
	sc->gpofl[++sc->gpofl_loc] = i;
	return;
      }
}


void s7_gc_unprotect_at(s7_scheme *sc, uint32_t loc)
{
  if (loc < sc->protected_objects_size)
    {
      if (vector_element(sc->protected_objects, loc) != sc->gc_nil)
	sc->gpofl[++sc->gpofl_loc] = loc;
      vector_element(sc->protected_objects, loc) = sc->gc_nil;
    }
}


s7_pointer s7_gc_protected_at(s7_scheme *sc, uint32_t loc)
{
  s7_pointer obj;

  obj = sc->unspecified;
  if (loc < sc->protected_objects_size)
    obj = vector_element(sc->protected_objects, loc);

  if (obj == sc->gc_nil)
    return(sc->unspecified);

  return(obj);
}

#define gc_protected_at(Sc, Loc) vector_element(Sc->protected_objects, Loc)


static void (*mark_function[NUM_TYPES])(s7_pointer p);

#define S7_MARK(Obj) do {s7_pointer _p_; _p_ = Obj; if (!is_marked(_p_)) (*mark_function[unchecked_type(_p_)])(_p_);} while (0)

static void mark_slot(s7_pointer p)
{
  set_mark(p);
  S7_MARK(slot_value(p));
  if (slot_has_accessor(p))
    S7_MARK(slot_accessor(p));

  if (is_gensym(slot_symbol(p))) /* (let () (apply define (gensym) (list 32)) (gc) (gc) (curlet)) */
    set_mark(slot_symbol(p));
}

static void mark_symbol(s7_pointer p)
{
  if (is_gensym(p))
    set_mark(p);
  /* don't set the mark bit of a normal symbol!  It wrecks the check against SYNTACTIC_TYPE,
   *   slowing everything down by a large amount.
   * (this comment is obsolete I think)
   */
}

static void mark_noop(s7_pointer p) {}

/* ports can be alloc'd and freed at a frightening pace, so I think I'll make a special free_list for them. */

static port_t *alloc_port(s7_scheme *sc)
{
  if (sc->port_heap)
    {
      port_t *p;
      p = sc->port_heap;
      sc->port_heap = (port_t *)(p->next);
      return(p);
    }
  return((port_t *)calloc(1, sizeof(port_t)));
}


static void free_port(s7_scheme *sc, port_t *p)
{
  p->next = (void *)(sc->port_heap);
  sc->port_heap = p;
}

static void close_output_port(s7_scheme *sc, s7_pointer p);
static void free_optlist(s7_scheme *sc, s7_pointer p);

#define STRING_LISTS 256
#define STRING_LIST_INIT_SIZE 2

static void init_string_free_lists(s7_scheme *sc)
{
  int32_t i;
  sc->string_lists = (char ***)calloc(STRING_LISTS, sizeof(char **));
  sc->string_locs = (int32_t *)calloc(STRING_LISTS, sizeof(int32_t));
  sc->string_sizes = (int32_t *)malloc(STRING_LISTS * sizeof(int32_t));
  sc->string_max_sizes = (int32_t *)malloc(STRING_LISTS * sizeof(int32_t));
  for (i = 0; i < STRING_LISTS; i++)
    {
      sc->string_lists[i] = (char **)calloc(STRING_LIST_INIT_SIZE, sizeof(char *));
      sc->string_sizes[i] = STRING_LIST_INIT_SIZE;
      if (i < 16)
	sc->string_max_sizes[i] = 4096;
      else
	{
	  if (i < 32)
	    sc->string_max_sizes[i] = 1024;
	  else 
	    {
	      if (i < 64)
		sc->string_max_sizes[i] = 256;
	      else sc->string_max_sizes[i] = 32;
	    }
	}
    }
}

static char *alloc_string(s7_scheme *sc, int32_t len)
{
  if ((len < STRING_LISTS) &&
      (sc->string_locs[len] > 0))
    return(sc->string_lists[len][--sc->string_locs[len]]);
  return((char *)malloc((len + 1) * sizeof(char)));
}

static void string_to_free_list(s7_scheme *sc, char *value, int32_t len)
{
  if (len >= STRING_LISTS)
    free(value);
  else
    {
      if (sc->string_locs[len] >= sc->string_sizes[len])
	{
	  if (sc->string_sizes[len] >= sc->string_max_sizes[len])
	    {
	      free(value);
	      return;
	    }
	  sc->string_sizes[len] *= 2;
	  sc->string_lists[len] = (char **)realloc((void *)(sc->string_lists[len]), sc->string_sizes[len] * sizeof(char *));
	}
      sc->string_lists[len][sc->string_locs[len]++] = value;
    }
}


static void sweep(s7_scheme *sc)
{
  uint32_t i, j;
  s7_pointer s1;
  gc_list *gp;
  
  gp = sc->strings;
  if (gp->loc > 0)
    {
      /* unrolling this loop is not an improvement */
      for (i = 0, j = 0; i < gp->loc; i++)
	{
	  s1 = gp->list[i];
	  if (is_free_and_clear(s1))
	    {
	      if (string_needs_free(s1))
		string_to_free_list(sc, string_value(s1), string_length(s1));
	    }
	  else 
	    {
	      /* remove_from_heap can remove a string from the heap; we need to notice that removal
	       *   via in_heap, and remove it also from this cache; otherwise it just stays here
	       *   forever, slowing down the loop. 
	       */
	      if (in_heap(s1)) /* this costs more than it saves */
		gp->list[j++] = s1;
	    }
	}
      gp->loc = j;
    }

  gp = sc->strings1;
  if (gp->loc > 0)
    {
      for (i = 0, j = 0; i < gp->loc; i++)
	{
	  s1 = gp->list[i];
	  if (is_free_and_clear(s1))
	    {
	      if (string_needs_free(s1))
		free(string_value(s1));
	    }
	  else gp->list[j++] = s1;
	}
      gp->loc = j;
    }

  gp = sc->gensyms;
  if (gp->loc > 0)
    {
      for (i = 0, j = 0; i < gp->loc; i++)
	{
	  s1 = gp->list[i];
	  if (is_free_and_clear(s1))
	    {
	      remove_gensym_from_symbol_table(sc, s1); /* this uses symbol_name_cell data */
	      free(symbol_name(s1));
	      free(symbol_name_cell(s1));
	    }
	  else gp->list[j++] = s1;
	}
      gp->loc = j;
      if (j == 0) mark_function[T_SYMBOL] = mark_noop;
    }

  gp = sc->c_objects;
  if (gp->loc > 0)
    {
      for (i = 0, j = 0; i < gp->loc; i++)
	{
	  s1 = gp->list[i];
	  if (is_free_and_clear(s1))
	    (*(c_object_free(sc, s1)))(c_object_value(s1));    
	  else gp->list[j++] = s1;
	}
      gp->loc = j;
    }

  gp = sc->optlists;
  if (gp->loc > 0)
    {
      for (i = 0, j = 0; i < gp->loc; i++)
	{
	  s1 = gp->list[i];
	  if (is_free_and_clear(s1))
	    free_optlist(sc, s1);
	  else gp->list[j++] = s1;
	}
      gp->loc = j;
    }

  gp = sc->vectors;
  if (gp->loc > 0)
    {
      for (i = 0, j = 0; i < gp->loc; i++)
	{
	  s1 = gp->list[i];
	  if (is_free_and_clear(s1))
	    {
	      /* a multidimensional empty vector can have dimension info, wrapped vectors always have dimension info */
	      if (vector_dimension_info(s1))
		{
		  if (vector_dimensions_allocated(s1))
		    {
		      free(vector_dimensions(s1));
		      free(vector_offsets(s1));
		    }
		  if (vector_elements_allocated(s1))
		    free(vector_elements(s1));      /* I think this will work for any vector (int/float too) */
		  if (vector_dimension_info(s1) != sc->wrap_only)
		    free(vector_dimension_info(s1));
		}
	      else
		{
		  if (vector_length(s1) != 0)
		    free(vector_elements(s1));
		}
	    }
	  else gp->list[j++] = s1;
	}
      gp->loc = j;
    }

  gp = sc->hash_tables;
  if (gp->loc > 0)
    {
      for (i = 0, j = 0; i < gp->loc; i++)
	{
	  s1 = gp->list[i];
	  if (is_free_and_clear(s1))
	    {
	      if (hash_table_mask(s1) > 0)
		free_hash_table(s1);
	    }
	  else gp->list[j++] = s1;
	}
      gp->loc = j;
    }

  gp = sc->input_ports;
  if (gp->loc > 0)
    {
      for (i = 0, j = 0; i < gp->loc; i++)
	{
	  s1 = gp->list[i];
	  if (is_free_and_clear(s1))
	    {
	      if (port_needs_free(s1))
		{
		  if (port_data(s1))
		    {
		      free(port_data(s1));
		      port_data(s1) = NULL;
		      port_data_size(s1) = 0;
		    }
		  port_needs_free(s1) = false;
		}

	      if (port_filename(s1))
		{
		  free(port_filename(s1));
		  port_filename(s1) = NULL;
		}
	      free_port(sc, port_port(s1));
	    }
	  else gp->list[j++] = s1;
	}
      gp->loc = j;
    }

  gp = sc->output_ports;
  if (gp->loc > 0)
    {
      for (i = 0, j = 0; i < gp->loc; i++)
	{
	  s1 = gp->list[i];
	  if (is_free_and_clear(s1))
	    {
	      close_output_port(sc, s1); /* needed for free filename, etc */
	      free_port(sc, port_port(s1));
	    }
	  else gp->list[j++] = s1;
	}
      gp->loc = j;
    }

  gp = sc->continuations;
  if (gp->loc > 0)
    {
      for (i = 0, j = 0; i < gp->loc; i++)
	{
	  s1 = gp->list[i];
	  if (is_free_and_clear(s1))
	    {
	      if (continuation_op_stack(s1))
		{
		  free(continuation_op_stack(s1));
		  continuation_op_stack(s1) = NULL;
		}
	      free(continuation_data(s1));
	    }
	  else gp->list[j++] = s1;
	}
      gp->loc = j;
    }

#if WITH_GMP
  gp = sc->bigints;
  if (gp->loc > 0)
    {
      for (i = 0, j = 0; i < gp->loc; i++)
	{
	  s7_pointer s1;
	  s1 = gp->list[i];
	  if (is_free_and_clear(s1))
	    mpz_clear(big_integer(s1));
	  else gp->list[j++] = s1;
	}
      gp->loc = j;
    }

  gp = sc->bigratios;
  if (gp->loc > 0)
    {
      for (i = 0, j = 0; i < gp->loc; i++)
	{
	  s7_pointer s1;
	  s1 = gp->list[i];
	  if (is_free_and_clear(s1))
	    mpq_clear(big_ratio(s1));
	  else gp->list[j++] = s1;
	}
      gp->loc = j;
    }

  gp = sc->bigreals;
  if (gp->loc > 0)
    {
      for (i = 0, j = 0; i < gp->loc; i++)
	{
	  s7_pointer s1;
	  s1 = gp->list[i];
	  if (is_free_and_clear(s1))
	    mpfr_clear(big_real(s1));
	  else gp->list[j++] = s1;
	}
      gp->loc = j;
    }

  gp = sc->bignumbers;
  if (gp->loc > 0)
    {
      for (i = 0, j = 0; i < gp->loc; i++)
	{
	  s7_pointer s1;
	  s1 = gp->list[i];
	  if (is_free_and_clear(s1))
	    mpc_clear(big_complex(s1));
	  else gp->list[j++] = s1;
	}
      gp->loc = j;
    }
#endif
}

static inline void add_to_gc_list(gc_list *gp, s7_pointer p)
{
  if (gp->loc == gp->size)
    {
      gp->size *= 2;
      gp->list = (s7_pointer *)realloc(gp->list, gp->size * sizeof(s7_pointer));
    }
  gp->list[gp->loc++] = p;
}

static gc_list *make_gc_list(void)
{
  gc_list *gp;
  #define INIT_GC_CACHE_SIZE 4
  gp = (gc_list *)malloc(sizeof(gc_list));
  gp->size = INIT_GC_CACHE_SIZE;
  gp->loc = 0;
  gp->list = (s7_pointer *)malloc(gp->size * sizeof(s7_pointer));
  return(gp);			 
}

static void add_gensym(s7_scheme *sc, s7_pointer p)
{
  add_to_gc_list(sc->gensyms, p);
  mark_function[T_SYMBOL] = mark_symbol;
}


#define add_c_object(sc, p)     {add_to_gc_list(sc->c_objects, p);}
#define add_optlist(sc, p)      {add_to_gc_list(sc->optlists, p);}
#define add_hash_table(sc, p)   {add_to_gc_list(sc->hash_tables, p);}
#define add_string(sc, p)       {add_to_gc_list(sc->strings, p);}
#define add_string1(sc, p)      {add_to_gc_list(sc->strings1, p);}
#define add_vector(sc, p)       {add_to_gc_list(sc->vectors, p);}
#define add_input_port(sc, p)   {add_to_gc_list(sc->input_ports, p);}
#define add_output_port(sc, p)  {add_to_gc_list(sc->output_ports, p);}
#define add_continuation(sc, p) {add_to_gc_list(sc->continuations, p);}

#if WITH_GMP
#define add_bigint(sc, p)    {add_to_gc_list(sc->bigints, p);}
#define add_bigratio(sc, p)  {add_to_gc_list(sc->bigratios, p);}
#define add_bigreal(sc, p)   {add_to_gc_list(sc->bigreals, p);}
#define add_bignumber(sc, p) {add_to_gc_list(sc->bignumbers, p);}
#endif


#define INIT_GC_CACHE_SIZE 4
static void init_gc_caches(s7_scheme *sc)
{
  sc->strings = make_gc_list();
  sc->strings1 = make_gc_list();
  sc->gensyms = make_gc_list();
  sc->vectors = make_gc_list();
  sc->hash_tables = make_gc_list();
  sc->input_ports = make_gc_list();
  sc->output_ports = make_gc_list();
  sc->continuations = make_gc_list();
  sc->c_objects = make_gc_list();
  sc->optlists = make_gc_list();
#if WITH_GMP
  sc->bigints = make_gc_list();
  sc->bigratios = make_gc_list();
  sc->bigreals = make_gc_list();
  sc->bignumbers = make_gc_list();
#endif

  /* slightly unrelated... */
  sc->setters_size = 4;
  sc->setters_loc = 0;
  sc->setters = (s7_pointer *)malloc(sc->setters_size * sizeof(s7_pointer));
}


static void add_setter(s7_scheme *sc, s7_pointer p, s7_pointer setter)
{
  /* procedure-setters GC-protected. The c_function_setter field can't be used because the built-in functions
   *   are often removed from the heap and never thereafter marked.
   */
  uint32_t i;
  for (i = 0; i < sc->setters_loc; i++)
    {
      s7_pointer x;
      x = sc->setters[i];
      if (car(x) == p)
	{
	  set_cdr(x, setter);
	  return;
 	}
    }
  if (sc->setters_loc == sc->setters_size)
    {
      sc->setters_size *= 2;
      sc->setters = (s7_pointer *)realloc(sc->setters, sc->setters_size * sizeof(s7_pointer));
    }
  sc->setters[sc->setters_loc++] = permanent_cons(p, setter, T_PAIR | T_IMMUTABLE);
}


static void mark_vector_1(s7_pointer p, s7_int top)
{
  s7_pointer *tp, *tend, *tend4;

  set_mark(p);

  tp = (s7_pointer *)(vector_elements(p));
  if (!tp) return;
  tend = (s7_pointer *)(tp + top);

  tend4 = (s7_pointer *)(tend - 4);
  while (tp <= tend4)
    {
      S7_MARK(*tp++);
      S7_MARK(*tp++);
      S7_MARK(*tp++);
      S7_MARK(*tp++);
    }

  while (tp < tend)
    S7_MARK(*tp++);
}

static void mark_let(s7_pointer env)
{
  s7_pointer x;
  for (x = env; is_let(x) && (!is_marked(x)); x = outlet(x))
    {
      s7_pointer y;
      set_mark(x);
      for (y = let_slots(x); is_slot(y); y = next_slot(y))
	if (!is_marked(y)) /* slot value might be the enclosing let */
	  mark_slot(y);
    }
}

static void just_mark(s7_pointer p)
{
  set_mark(p);
}

static void mark_c_pointer(s7_pointer p)
{
  set_mark(p);
  S7_MARK(raw_pointer_type(p));
  S7_MARK(raw_pointer_info(p));
}

static void mark_c_proc_star(s7_pointer p)
{
  set_mark(p);
  if (!has_simple_defaults(p))
    {
      s7_pointer arg;
      for (arg = c_function_call_args(p); is_pair(arg); arg = cdr(arg))
	S7_MARK(car(arg));
    }
}

static void mark_pair(s7_pointer p)
{
  s7_pointer x;
  set_mark(p);
  S7_MARK(car(p));
  /* if the list is huge, recursion to cdr(p) is problematic when there are strict limits on the stack size
   *  so I'll try something else... (This form is faster according to callgrind).
   *
   * in snd-14 or so through 15.3, sc->temp_cell_2|3 were used for trailing args in eval, but that meant
   *   the !is_marked check below (which is intended to catch cyclic lists) caused cells to be missed;
   *   since sc->args could contain permanently marked cells, if these were passed to g_vector, for example, and
   *   make_vector_1 triggered a GC call, we needed to mark both the permanent (always marked) cell and its contents,
   *   and continue through the rest of the list.  But adding temp_cell_2|3 to sc->permanent_objects was not enough.
   *   Now I've already forgotten the rest of the story, and it was just an hour ago! -- the upshot is that temp_cell_2|3
   *   are not now used as arg list members.
   */
  for (x = cdr(p); (is_pair(x)) && (!is_marked(x)); x = cdr(x))
    {
      set_mark(x);
      S7_MARK(car(x));
    }
  S7_MARK(x);
}

static void mark_counter(s7_pointer p)
{
  set_mark(p);
  S7_MARK(counter_result(p));
  S7_MARK(counter_list(p));
  S7_MARK(counter_let(p));
}

static void mark_closure(s7_pointer p)
{
  set_mark(p);
  S7_MARK(closure_args(p));
  S7_MARK(closure_body(p));
  mark_let(closure_let(p));
  S7_MARK(closure_setter(p));
  if (has_optlist(p)) 
    set_mark(closure_optlist(p));
}

static void mark_stack_1(s7_pointer p, s7_int top)
{
  s7_pointer *tp, *tend;
  set_mark(p);

  tp = (s7_pointer *)(vector_elements(p));
  if (!tp) return;
  tend = (s7_pointer *)(tp + top);

  while (tp < tend)
    {
      S7_MARK(*tp++);
      S7_MARK(*tp++);
      S7_MARK(*tp++);
      tp++;
    }
}

static void mark_stack(s7_pointer p)
{
  /* we can have a bare stack awaiting a continuation to hold it if the new_cell for the continuation
   *    triggers the GC!  But we need a top-of-stack??
   */
  mark_stack_1(p, temp_stack_top(p));
}

static void mark_continuation(s7_pointer p)
{
  uint32_t i;
  set_mark(p);
  mark_stack_1(continuation_stack(p), continuation_stack_top(p));
  for (i = 0; i < continuation_op_loc(p); i++)
    S7_MARK(continuation_op_stack(p)[i]);
}

static void mark_vector(s7_pointer p)
{
  mark_vector_1(p, vector_length(p));
}

static void mark_vector_possibly_shared(s7_pointer p)
{
  /* If a subvector (an inner dimension) of a vector is the only remaining reference
   *    to the main vector, we want to make sure the main vector is not GC'd until
   *    the subvector is also GC-able.  The shared_vector field either points to the
   *    parent vector, or it is sc->F, so we need to check for a vector parent if
   *    the current is multidimensional (this will include 1-dim slices).  We need
   *    to keep the parent case separate (i.e. sc->F means the current is the original)
   *    so that we only free once (or remove_from_heap once).  
   *
   * If we have a shared-vector of a shared-vector, and the middle and original are not otherwise
   *   in use, we mark the middle one, but (since it itself is not in use anywhere else)
   *   we don't mark the original!  So we need to follow the share-vector chain marking every one.
   */
  if ((vector_has_dimensional_info(p)) &&
      (s7_is_vector(shared_vector(p))))
    mark_vector_possibly_shared(shared_vector(p));

  mark_vector_1(p, vector_length(p));
}

static void mark_int_or_float_vector(s7_pointer p)
{
  set_mark(p);
}

static void mark_int_or_float_vector_possibly_shared(s7_pointer p)
{
  if ((vector_has_dimensional_info(p)) &&
      (s7_is_vector(shared_vector(p))))
    mark_int_or_float_vector_possibly_shared(shared_vector(p));

  set_mark(p);
}

static void mark_c_object(s7_pointer p)
{
  set_mark(p);
  (*(c_object_mark(p)))(c_object_value(p));
}

static void mark_catch(s7_pointer p)
{
  set_mark(p);
  S7_MARK(catch_tag(p));
  S7_MARK(catch_handler(p));
}

static void mark_dynamic_wind(s7_pointer p)
{
  set_mark(p);
  S7_MARK(dynamic_wind_in(p));
  S7_MARK(dynamic_wind_out(p));
  S7_MARK(dynamic_wind_body(p));
}

static void mark_hash_table(s7_pointer p)
{
  set_mark(p);
  S7_MARK(hash_table_procedures(p));
  if (hash_table_entries(p) > 0)
    {
      uint32_t len;
      hash_entry_t **entries, **last;

      entries = hash_table_elements(p);
      len = hash_table_mask(p) + 1;
      last = (hash_entry_t **)(entries + len);

      while (entries < last)
	{
	  hash_entry_t *xp;
	  for (xp = *entries++; xp; xp = xp->next)
	    {
	      S7_MARK(xp->key);
	      S7_MARK(xp->value);
	    }
	  for (xp = *entries++; xp; xp = xp->next)
	    {
	      S7_MARK(xp->key);
	      S7_MARK(xp->value);
	    }
	}
    }
}

static void mark_iterator(s7_pointer p)
{
  set_mark(p);
  S7_MARK(iterator_sequence(p));
  if (is_mark_seq(p))
    S7_MARK(iterator_current(p));
}

static void mark_input_port(s7_pointer p)
{
  set_mark(p);
  set_mark(port_original_input_string(p));
}

#define clear_type(p) typeflag(p) = T_FREE

static void init_mark_functions(void)
{
  mark_function[T_FREE]                = mark_noop;
  mark_function[T_UNDEFINED]           = mark_noop;
  mark_function[T_EOF_OBJECT]          = mark_noop;
  mark_function[T_UNSPECIFIED]         = mark_noop;
  mark_function[T_NIL]                 = mark_noop;
  mark_function[T_BOOLEAN]             = mark_noop;
  mark_function[T_SYNTAX]              = mark_noop;
  mark_function[T_CHARACTER]           = mark_noop;
  mark_function[T_SYMBOL]              = mark_noop; /* this changes to mark_symbol when gensyms are in the heap */
  mark_function[T_STRING]              = just_mark;
  mark_function[T_INTEGER]             = just_mark;
  mark_function[T_RATIO]               = just_mark;
  mark_function[T_REAL]                = just_mark;
  mark_function[T_COMPLEX]             = just_mark;
  mark_function[T_BIG_INTEGER]         = just_mark;
  mark_function[T_BIG_RATIO]           = just_mark;
  mark_function[T_BIG_REAL]            = just_mark;
  mark_function[T_BIG_COMPLEX]         = just_mark;
  mark_function[T_RANDOM_STATE]        = just_mark;
  mark_function[T_GOTO]                = just_mark;
  mark_function[T_OUTPUT_PORT]         = just_mark;
  mark_function[T_OPTLIST]             = just_mark;
  mark_function[T_BAFFLE]              = just_mark;
  mark_function[T_C_MACRO]             = just_mark;
  mark_function[T_C_POINTER]           = mark_c_pointer;
  mark_function[T_C_FUNCTION]          = just_mark;
  mark_function[T_C_FUNCTION_STAR]     = just_mark;  /* changes to mark_c_proc_star if defaults involve an expression */
  mark_function[T_C_ANY_ARGS_FUNCTION] = just_mark;
  mark_function[T_C_OPT_ARGS_FUNCTION] = just_mark;
  mark_function[T_C_RST_ARGS_FUNCTION] = just_mark;
  mark_function[T_PAIR]                = mark_pair;
  mark_function[T_CLOSURE]             = mark_closure;
  mark_function[T_CLOSURE_STAR]        = mark_closure;
  mark_function[T_CONTINUATION]        = mark_continuation;
  mark_function[T_INPUT_PORT]          = mark_input_port;
  mark_function[T_VECTOR]              = mark_vector; /* this changes if shared vector created (similarly below) */
  mark_function[T_INT_VECTOR]          = mark_int_or_float_vector;
  mark_function[T_FLOAT_VECTOR]        = mark_int_or_float_vector; 
  mark_function[T_MACRO]               = mark_closure;
  mark_function[T_BACRO]               = mark_closure;
  mark_function[T_MACRO_STAR]          = mark_closure;
  mark_function[T_BACRO_STAR]          = mark_closure;
  mark_function[T_C_OBJECT]            = mark_c_object;
  mark_function[T_CATCH]               = mark_catch;
  mark_function[T_DYNAMIC_WIND]        = mark_dynamic_wind;
  mark_function[T_HASH_TABLE]          = mark_hash_table;
  mark_function[T_ITERATOR]            = mark_iterator;
  mark_function[T_LET]                 = mark_let;
  mark_function[T_STACK]               = mark_stack;
  mark_function[T_COUNTER]             = mark_counter;
  mark_function[T_SLOT]                = mark_slot;
}


static void mark_op_stack(s7_scheme *sc)
{
  s7_pointer *p, *tp;
  tp = sc->op_stack_now;
  p = sc->op_stack;
  while (p < tp)
    S7_MARK(*p++);
}

static void mark_rootlet(s7_scheme *sc)
{
  s7_pointer ge;
  s7_pointer *tmp, *top;

  ge = sc->rootlet;
  tmp = vector_elements(ge);
  top = (s7_pointer *)(tmp + sc->rootlet_entries);

  set_mark(ge);
  while (tmp < top)
    S7_MARK(slot_value(*tmp++));
}

void s7_mark_c_object(s7_pointer p)
{
  S7_MARK(p);
}

static void mark_permanent_objects(s7_scheme *sc)
{
  gc_obj *g;
  for (g = sc->permanent_objects; g; g = (gc_obj *)(g->nxt))
    S7_MARK(g->p);
}

static void unmark_permanent_objects(s7_scheme *sc)
{
  gc_obj *g;
  for (g = sc->permanent_objects; g; g = (gc_obj *)(g->nxt))
    clear_mark(g->p);
}


#if (!MS_WINDOWS)
  #include <time.h>
  #include <sys/time.h>
  static struct timeval start_time;
  static struct timezone z0;
#endif


#if DEBUGGING
static int32_t last_gc_line = 0;
static const char *last_gc_func = NULL;
#endif

#define GC_STATS 1
#define HEAP_STATS 2
#define STACK_STATS 4

#define show_gc_stats(Sc) ((Sc->gc_stats & GC_STATS) != 0)
#define show_stack_stats(Sc) ((Sc->gc_stats & STACK_STATS) != 0)
#define show_heap_stats(Sc) ((Sc->gc_stats & HEAP_STATS) != 0)


static int32_t gc(s7_scheme *sc)
{
  s7_cell **old_free_heap_top;
  /* mark all live objects (the symbol table is in permanent memory, not the heap) */
#if DEBUGGING
  #define gc_call(P, Tp) \
    p = (*tp++); \
    if (is_marked(_Cell(p))) \
       clear_mark(p); \
    else \
      { \
        if (!is_free_and_clear(p))		\
          {								\
	    p->debugger_bits = 0; p->gc_line = last_gc_line; p->gc_func = last_gc_func;	\
            clear_type(p);	\
            (*fp++) = p;\
          }}
#else
  #define gc_call(P, Tp) p = (*tp++); if (is_marked(p)) clear_mark(p); else {if (!is_free_and_clear(p)) {clear_type(p); (*fp++) = p;}}
#endif

  if (show_gc_stats(sc))
    {
      fprintf(stdout, "gc ");
#if DEBUGGING
      fprintf(stdout, "%s[%d] ", last_gc_func, last_gc_line);
#endif
#if (!MS_WINDOWS)
      /* this is apparently deprecated in favor of clock_gettime -- what compile-time switch to use here?
       *   _POSIX_TIMERS, or perhaps use CLOCK_REALTIME, but clock_gettime requires -lrt -- no thanks.
       */
      gettimeofday(&start_time, &z0);
#endif
    }

  mark_rootlet(sc);
  S7_MARK(sc->args);
  mark_let(sc->envir);

  slot_set_value(sc->error_data, sc->F); 
  /* the other choice here is to explicitly mark slot_value(sc->error_data) as we do eval_history1/2 below.
   *    in both cases, the values are permanent lists that do not mark impermanent contents.
   *    this will need circular list checks, and can't depend on marked to exit early
   */
  mark_let(sc->owlet);
#if WITH_HISTORY
  {
    s7_pointer p1, p2;
    for (p1 = sc->eval_history1, p2 = sc->eval_history2; ; p2 = cdr(p2))
      {
	S7_MARK(car(p1));
	S7_MARK(car(p2));
	p1 = cdr(p1);
	if (p1 == sc->eval_history1) break; /* these are circular lists */
      }
  }
#endif

  S7_MARK(sc->code);
  mark_current_code(sc);
  mark_stack_1(sc->stack, s7_stack_top(sc));
  S7_MARK(sc->v);
  S7_MARK(sc->w);
  S7_MARK(sc->x);
  S7_MARK(sc->y);
  S7_MARK(sc->z);
  S7_MARK(sc->value);

  S7_MARK(sc->temp1);
  S7_MARK(sc->temp2);
  S7_MARK(sc->temp3);
  S7_MARK(sc->temp4);
  S7_MARK(sc->temp5);
  S7_MARK(sc->temp6);
  S7_MARK(sc->temp7);
  S7_MARK(sc->temp8);
  S7_MARK(sc->temp9);
  S7_MARK(sc->temp10);
  S7_MARK(sc->temp11);
  {
    int32_t i;
    for (i = 0; i < T_TEMPS_SIZE; i++) {S7_MARK(sc->t_temps[i]);}
  }
  set_mark(sc->input_port);
  S7_MARK(sc->input_port_stack);
  set_mark(sc->output_port);
  set_mark(sc->error_port);
  S7_MARK(sc->stacktrace_defaults);
  S7_MARK(sc->autoload_table);
  S7_MARK(sc->default_rng);

  mark_pair(sc->temp_cell_1);
  mark_pair(sc->temp_cell_2);
  S7_MARK(car(sc->t1_1));
  S7_MARK(car(sc->t2_1));
  S7_MARK(car(sc->t2_2));
  S7_MARK(car(sc->t3_1));
  S7_MARK(car(sc->t3_2));
  S7_MARK(car(sc->t3_3));

  S7_MARK(car(sc->a4_1));
  S7_MARK(car(sc->a4_2));
  S7_MARK(car(sc->a4_3));
  S7_MARK(car(sc->a4_4));

  S7_MARK(car(sc->plist_1));
  S7_MARK(car(sc->plist_2));
  S7_MARK(cadr(sc->plist_2));
  S7_MARK(car(sc->plist_3));
  S7_MARK(cadr(sc->plist_3));
  S7_MARK(caddr(sc->plist_3));
  S7_MARK(car(sc->qlist_2));
  S7_MARK(cadr(sc->qlist_2));

  {
    uint32_t i;
    s7_pointer p;
    /* perhaps: if (sc->current_safe_list > 0) ... but this loop is down in the noise */
    for (i = 1; i < NUM_SAFE_LISTS; i++)
      if ((is_pair(sc->safe_lists[i])) &&
	  (list_is_in_use(sc->safe_lists[i])))
	for (p = sc->safe_lists[i]; is_pair(p); p = cdr(p))
	  S7_MARK(car(p));
    for (i = 0; i < sc->setters_loc; i++)
      S7_MARK(cdr(sc->setters[i]));
  }
  {
    int32_t i;
    for (i = 0; i < sc->num_fdats; i++)
      if (sc->fdats[i])
	S7_MARK(sc->fdats[i]->curly_arg);
  }
  mark_vector(sc->protected_objects);
  mark_vector(sc->protected_accessors);

  /* now protect recent allocations using the free_heap cells above the current free_heap_top (if any).
   *
   * cells above sc->free_heap_top might be malloc'd garbage (after heap reallocation), so we keep track of
   *   where the last actually freed cells were after the previous GC call.  We're trying to
   *   GC protect the previous GC_TEMPS_SIZE allocated pointers so that the caller doesn't have
   *   to gc-protect every temporary cell.
   *
   * There's one remaining possible problem.  s7_remove_from_heap frees cells outside
   *   the GC and might push free_heap_top beyond its previous_free_heap_top, then
   *   an immediate explicit gc call might not see those temp cells.
   */
  {
    s7_pointer *tmps, *tmps_top;

    tmps = sc->free_heap_top;
    tmps_top = tmps + GC_TEMPS_SIZE;
    if (tmps_top > sc->previous_free_heap_top)
      tmps_top = sc->previous_free_heap_top;

    while (tmps < tmps_top)
      S7_MARK(*tmps++);
  }
  mark_op_stack(sc);
  mark_permanent_objects(sc);

  /* free up all unmarked objects */
  old_free_heap_top = sc->free_heap_top;

  {
    s7_pointer *fp, *tp, *heap_top;
    fp = sc->free_heap_top;

    tp = sc->heap;
    heap_top = (s7_pointer *)(sc->heap + sc->heap_size);

    while (tp < heap_top)          /* != here or ^ makes no difference */
      {
	s7_pointer p;
	/* from here down is gc_call, but I wanted one case explicit for readability */
	p = (*tp++);

	if (is_marked(_Cell(p)))          /* this order is faster than checking typeflag(p) != T_FREE first */
	  clear_mark(p);
	else
	  {
	    if (!is_free_and_clear(p)) /* if T_FREE, it's an already-free object -- the free_heap is usually not empty when we call the GC */
	      {
#if DEBUGGING
		p->debugger_bits = 0;
#endif
		clear_type(p); /* (this is needed -- otherwise we try to free some objects twice) */
		(*fp++) = p;
	      }
	  }

	/* this looks crazy, but it speeds up the entire GC process by 25%!
	 *   going from 16 to 32 saves .2% so it may not matter.
	 */
	gc_call(p, tp);
	gc_call(p, tp);
	gc_call(p, tp);

	gc_call(p, tp);
	gc_call(p, tp);
	gc_call(p, tp);
	gc_call(p, tp);

	gc_call(p, tp);
	gc_call(p, tp);
	gc_call(p, tp);
	gc_call(p, tp);

	gc_call(p, tp);
	gc_call(p, tp);
	gc_call(p, tp);
	gc_call(p, tp);

	gc_call(p, tp);
	gc_call(p, tp);
	gc_call(p, tp);
	gc_call(p, tp);

	gc_call(p, tp);
	gc_call(p, tp);
	gc_call(p, tp);
	gc_call(p, tp);

	gc_call(p, tp);
	gc_call(p, tp);
	gc_call(p, tp);
	gc_call(p, tp);

	gc_call(p, tp);
	gc_call(p, tp);
	gc_call(p, tp);
	gc_call(p, tp);
      }

    sc->free_heap_top = fp;
    sweep(sc);
  }

  unmark_permanent_objects(sc);
  sc->gc_freed = (int64_t)(sc->free_heap_top - old_free_heap_top);

  if (show_gc_stats(sc))
    {
#if (!MS_WINDOWS)
      struct timeval t0;
      double secs;
      gettimeofday(&t0, &z0);
      secs = (t0.tv_sec - start_time.tv_sec) +  0.000001 * (t0.tv_usec - start_time.tv_usec);
      fprintf(stdout, "freed %" PRId64 "/%" PRId64 " (free: %" PRIdPTR "), time: %f\n", sc->gc_freed, sc->heap_size, (intptr_t)(sc->free_heap_top - sc->free_heap), secs);
#else
      fprintf(stdout, "freed %" PRId64 "/%" PRId64 "\n", sc->gc_freed, sc->heap_size);
#endif
    }
  
  /* if (sc->begin_hook) call_begin_hook(sc); */
  sc->previous_free_heap_top = sc->free_heap_top;
  return(sc->gc_freed); /* needed by cell allocator to decide when to increase heap size */
}

void s7_gc_stats(s7_scheme *sc, bool on) {sc->gc_stats = (on) ? GC_STATS : 0;}
int64_t s7_heap_size(s7_scheme *sc) {return(sc->heap_size);}
int64_t s7_gc_freed(s7_scheme *sc) {return(sc->gc_freed);}


#define GC_TRIGGER_SIZE 64

/* new_cell has to include the new cell's type.  In the free list, it is 0 (T_FREE).  If we remove it here,
 *   but then hit some error before setting the type, the GC sweep thinks it is a free cell already and
 *   does not return it to the free list: a memory leak.
 */

#if (!DEBUGGING)
#define new_cell(Sc, Obj, Type)			\
  do {						\
    if (Sc->free_heap_top <= Sc->free_heap_trigger) try_to_call_gc(Sc); \
    Obj = (*(--(Sc->free_heap_top))); \
    set_type(Obj, Type);	      \
    } while (0)

#define new_cell_no_check(Sc, Obj, Type) do {Obj = (*(--(Sc->free_heap_top))); set_type(Obj, Type);} while (0)
  /* since sc->free_heap_trigger is GC_TRIGGER_SIZE above the free heap base, we don't need
   *   to check it repeatedly after the first such check.
   */
#else
/* DEBUGGING */
static bool for_any_other_reason(s7_scheme *sc, int32_t line)
{
#if 0
  static int32_t ctr = 0;
  if ((sc->default_rng) &&
      (!sc->gc_off) &&
      (ctr > GC_TRIGGER_SIZE))
    {
      s7_double x;
      x = next_random(sc->default_rng);
      if (x > .999)
	{
	  ctr = 0;
	  return(true);
	}
    }
  ctr++;
#endif
  return(false);
}

#define new_cell(Sc, Obj, Type)			\
  do {						\
    if ((Sc->free_heap_top <= Sc->free_heap_trigger) || (for_any_other_reason(Sc, __LINE__))) {last_gc_line = __LINE__; last_gc_func = __func__; try_to_call_gc(Sc);} \
    Obj = (*(--(Sc->free_heap_top))); \
    Obj->alloc_line = __LINE__; Obj->alloc_func = __func__; Obj->debugger_bits = 0;	\
    set_type(Obj, Type);	      \
    } while (0)

#define new_cell_no_check(Sc, Obj, Type)		    \
  do {							    \
    Obj = (*(--(Sc->free_heap_top)));			    \
    Obj->alloc_line = __LINE__;	Obj->alloc_func = __func__; Obj->debugger_bits = 0; \
    set_type(Obj, Type);				    \
    } while (0)
#endif


static void resize_heap_to(s7_scheme *sc, int64_t size)
{
  /* alloc more heap */
  int64_t old_size, old_free, k;
  s7_cell *cells;
  s7_pointer p;

  old_size = sc->heap_size;
  old_free = sc->free_heap_top - sc->free_heap;

  if (size == 0)
    {
      if (sc->heap_size < 512000)
	sc->heap_size *= 2;
      else sc->heap_size += 512000;
    }
  else
    {
      if (size > sc->heap_size)
	while (sc->heap_size < size) sc->heap_size *= 2;
    }

  sc->heap = (s7_cell **)realloc(sc->heap, sc->heap_size * sizeof(s7_cell *));
  if (!(sc->heap))
    s7_warn(sc, 256, "heap reallocation failed! tried to get %" PRId64 " bytes\n", (int64_t)(sc->heap_size * sizeof(s7_cell *)));

  sc->free_heap = (s7_cell **)realloc(sc->free_heap, sc->heap_size * sizeof(s7_cell *));
  if (!(sc->free_heap))
    s7_warn(sc, 256, "free heap reallocation failed! tried to get %" PRId64 " bytes\n", (int64_t)(sc->heap_size * sizeof(s7_cell *)));

  sc->free_heap_trigger = (s7_cell **)(sc->free_heap + GC_TRIGGER_SIZE);
  sc->free_heap_top = sc->free_heap + old_free; /* incremented below, added old_free 21-Aug-12?!? */

  /* optimization suggested by K Matheussen */
  cells = (s7_cell *)calloc(sc->heap_size - old_size, sizeof(s7_cell));
  for (p = cells, k = old_size; k < sc->heap_size;)
    {
      sc->heap[k] = p;
      heap_location(p) = k++;
      (*sc->free_heap_top++) = p++;
      sc->heap[k] = p;
      heap_location(p) = k++;
      (*sc->free_heap_top++) = p++;
      sc->heap[k] = p;
      heap_location(p) = k++;
      (*sc->free_heap_top++) = p++;
      sc->heap[k] = p;
      heap_location(p) = k++;
      (*sc->free_heap_top++) = p++;
    }
  sc->previous_free_heap_top = sc->free_heap_top;

  if (show_heap_stats(sc))
    {
      fprintf(stderr, "heap grows to %" PRId64 "\n", sc->heap_size);
#if DEBUGGING
      if (sc->heap_size > 50000000) /* maybe a max-heap-size? */
	{
	  s7_show_let(sc);
	  abort();
	}
#endif
    }
}

#define resize_heap(Sc) resize_heap_to(Sc, 0)

static void try_to_call_gc(s7_scheme *sc)
{
  /* called only from new_cell and cons */
  if (sc->gc_off)
    {
      /* we can't just return here!  Someone needs a new cell, and once the heap free list is exhausted, segfault */
      resize_heap(sc);
    }
  else
    {
#if (!DEBUGGING)
      int64_t freed_heap;
      freed_heap = gc(sc);
      if ((freed_heap < sc->heap_size / 2) &&
	  (freed_heap < 1000000)) /* if huge heap */
	resize_heap(sc);
#else
      gc(sc);
      if ((int64_t)(sc->free_heap_top - sc->free_heap) < sc->heap_size / 2)
	resize_heap(sc);
#endif
    }
}

  /* originally I tried to mark each temporary value until I was done with it, but
   *   that way madness lies... By delaying GC of _every_ %$^#%@ pointer, I can dispense
   *   with hundreds of individual protections.  So the free_heap's last GC_TEMPS_SIZE
   *   allocated pointers are protected during the mark sweep.
   */


static s7_pointer g_gc(s7_scheme *sc, s7_pointer args)
{
  #define H_gc "(gc (on #t)) runs the garbage collector.  If 'on' is supplied, it turns the GC on or off. \
Evaluation produces a surprising amount of garbage, so don't leave the GC off for very long!"
  #define Q_gc s7_make_signature(sc, 2, sc->T, sc->is_boolean_symbol)

  if (is_not_null(args))
    {
      if (!s7_is_boolean(car(args)))
	method_or_bust_one_arg(sc, car(args), sc->gc_symbol, args, T_BOOLEAN);
      sc->gc_off = (car(args) == sc->F);
      if (sc->gc_off)
	return(sc->F);
    }
#if DEBUGGING
  last_gc_line = __LINE__; 
  last_gc_func = __func__;
#endif
  gc(sc);
  return(sc->unspecified);
}


s7_pointer s7_gc_on(s7_scheme *sc, bool on)
{
  sc->gc_off = !on;
  return(s7_make_boolean(sc, on));
}


static int32_t permanent_cells = 0;
#if (!WITH_THREADS)
static s7_cell *alloc_pointer(void)
{
  #define ALLOC_SIZE 256
  static uint32_t alloc_k = ALLOC_SIZE;
  static s7_cell *alloc_cells = NULL;

  if (alloc_k == ALLOC_SIZE)     /* if either no current block or the block is used up */
    {                            /*   make a new block */
      permanent_cells += ALLOC_SIZE;
      alloc_cells = (s7_cell *)calloc(ALLOC_SIZE, sizeof(s7_cell));
      alloc_k = 0;
    }
  return(&alloc_cells[alloc_k++]);
}
#else
#define alloc_pointer() (s7_cell *)calloc(1, sizeof(s7_cell))
#endif


static void add_permanent_object(s7_scheme *sc, s7_pointer obj)
{
  gc_obj *g;
  g = (gc_obj *)malloc(sizeof(gc_obj));
  g->p = obj;
  g->nxt = sc->permanent_objects;
  sc->permanent_objects = g;
}

static void free_cell(s7_scheme *sc, s7_pointer p)
{
#if DEBUGGING
  p->debugger_bits = 0;
#endif
  clear_type(p);
  (*(sc->free_heap_top++)) = p;
}

static void free_vlist(s7_scheme *sc, s7_pointer lst)
{
  if (is_pair(lst))
    {
      s7_pointer p, np;
      for (p = lst, np = cdr(lst); is_pair(p); p = np, np = unchecked_cdr(np))
	free_cell(sc, p);
    }
}

static inline s7_pointer petrify(s7_scheme *sc, s7_pointer x, int64_t loc)
{
  s7_pointer p;
  p = alloc_pointer();
  sc->heap[loc] = p;
  heap_location(p) = loc;
  free_cell(sc, p);
  unheap(x);
  return(x);
}

static void s7_remove_from_heap(s7_scheme *sc, s7_pointer x)
{
  int64_t loc;
  /* global functions are very rarely redefined, so we can remove the function body from
   *   the heap when it is defined.  If redefined, we currently lose the memory held by the
   *   old definition.  (It is not trivial to recover this memory because it is allocated
   *   in blocks, not by the pointer, I think, but s7_define is the point to try).
   *
   * There is at least one problem with this: if, for example, a function has
   *    a quoted (constant) list, then uses list-set! to change an element of it,
   *    then a GC happens, and the new element is GC'd because no one in the heap
   *    points to it, then we call the function again, and it tries to access
   *    that element.
   *
   *    (define (bad-idea)
   *      (let ((lst '(1 2 3)))
   *        (let ((result (list-ref lst 1)))
   *          (list-set! lst 1 (* 2.0 16.6))
   *          (gc)
   *          result)))
   *
   *     put that in a file, load it (to force removal), than call bad-idea a few times.
   * so... if (*s7* 'safety) is not 0, remove-from-heap is disabled.
   */
  loc = heap_location(x);
  if (not_in_heap(x)) return;

  switch (type(x))
    {
    case T_PAIR:
      petrify(sc, x, loc);
#if 0
      /* this code fixes the problem above, but at some cost (gc + mark_pair up by about 2% in the worst case (snd-test.scm)) */
      if ((car(x) == sc->quote_symbol) &&
	  (is_pair(cadr(x))))
	{
	  add_permanent_object(sc, cdr(x));
	}
      else
	{
	  s7_remove_from_heap(sc, car(x));
	  s7_remove_from_heap(sc, cdr(x));
	}
#else
      s7_remove_from_heap(sc, car(x));
      s7_remove_from_heap(sc, cdr(x));
#endif
      return;

    case T_HASH_TABLE:
    case T_LET:
    case T_VECTOR: 
      /* not int|float_vector or string because none of their elements are GC-able (so unheap below is ok)
       *   but hash-table and let seem like they need protection? And let does happen via define-class.
       */
      add_permanent_object(sc, x);
      return;

    case T_SYNTAX:
      return;

    case T_SYMBOL:
      if (is_gensym(x))
	{
	  uint32_t i;
	  gc_list *gp;
	  sc->heap[loc] = alloc_pointer();
	  free_cell(sc, sc->heap[loc]);
	  heap_location(sc->heap[loc]) = loc;

	  /* unheap(x); */
	  heap_location(x) = -heap_location(x);
	  /* if gensym is a hash-table key, then is removed from the heap, we need to be sure the hash-table map to it
	   *   continues to be valid.  symbol_hmap is abs(heap_location), and the possible overlap with other not-in-heap
	   *   ints is not problematic (they'll just hash to the same location).
	   */
	  gp = sc->gensyms;
	  for (i = 0; i < gp->loc; i++) /* sc->gensyms reaches size 512 during s7test, but this search is called 3 times and costs nothing */
	    if (gp->list[i] == x)
	      {
		uint32_t j;
		for (j = i + 1; i < gp->loc - 1; i++, j++)
		  gp->list[i] = gp->list[j];
		gp->list[i] = NULL;
		gp->loc--;
		if (gp->loc == 0) mark_function[T_SYMBOL] = mark_noop;
		break;
	      }
	}
      return;

    case T_CLOSURE: case T_CLOSURE_STAR:
    case T_MACRO:   case T_MACRO_STAR:
    case T_BACRO:   case T_BACRO_STAR:
      petrify(sc, x, loc);
      s7_remove_from_heap(sc, closure_args(x));
      s7_remove_from_heap(sc, closure_body(x));
      if (has_optlist(x))
	s7_remove_from_heap(sc, closure_optlist(x));
      return;

    default:
      break;
    }
  
  petrify(sc, x, loc);
}



/* -------------------------------- stacks -------------------------------- */

#define OP_STACK_INITIAL_SIZE 32

#if DEBUGGING
#define stop_at_error true

static void push_op_stack(s7_scheme *sc, s7_pointer op)
{
  (*sc->op_stack_now++) = _NFre(op);
  if (sc->op_stack_now > (sc->op_stack + sc->op_stack_size)) 
    {
      fprintf(stderr, "%sop_stack overflow%s\n", BOLD_TEXT, UNBOLD_TEXT);
      if (stop_at_error) abort();
    }
}

static s7_pointer pop_op_stack(s7_scheme *sc)
{
  s7_pointer op;
  op = (*(--(sc->op_stack_now)));
  if (sc->op_stack_now < sc->op_stack) 
    {
      fprintf(stderr, "%sop_stack underflow%s\n", BOLD_TEXT, UNBOLD_TEXT);
      if (stop_at_error) abort();
    }
  return(_NFre(op));
}
#else
#define push_op_stack(Sc, Op) (*Sc->op_stack_now++) = Op
#define pop_op_stack(Sc)      (*(--(Sc->op_stack_now)))
#endif

static void initialize_op_stack(s7_scheme *sc)
{
  int32_t i;
  sc->op_stack = (s7_pointer *)malloc(OP_STACK_INITIAL_SIZE * sizeof(s7_pointer));
  sc->op_stack_size = OP_STACK_INITIAL_SIZE;
  sc->op_stack_now = sc->op_stack;
  sc->op_stack_end = (s7_pointer *)(sc->op_stack + sc->op_stack_size);
  for (i = 0; i < OP_STACK_INITIAL_SIZE; i++)
    sc->op_stack[i] = sc->nil;
}


static void resize_op_stack(s7_scheme *sc)
{
  int32_t i, loc, new_size;
  loc = (int32_t)(sc->op_stack_now - sc->op_stack);
  new_size = sc->op_stack_size * 2;
  sc->op_stack = (s7_pointer *)realloc((void *)(sc->op_stack), new_size * sizeof(s7_pointer));
  for (i = sc->op_stack_size; i < new_size; i++)
    sc->op_stack[i] = sc->nil;
  sc->op_stack_size = (uint32_t)new_size;
  sc->op_stack_now = (s7_pointer *)(sc->op_stack + loc);
  sc->op_stack_end = (s7_pointer *)(sc->op_stack + sc->op_stack_size);
}


#define stack_code(Stack, Loc)  vector_element(_TStk(Stack), Loc - 3)
#define stack_let(Stack, Loc)   vector_element(_TStk(Stack), Loc - 2)
#define stack_args(Stack, Loc)  vector_element(_TStk(Stack), Loc - 1)
#define stack_op(Stack, Loc)    ((opcode_t)(vector_element(_TStk(Stack), Loc)))

#if DEBUGGING
static void pop_stack(s7_scheme *sc)
{
  sc->stack_end -= 4;
  if (sc->stack_end < sc->stack_start) 
    {
      fprintf(stderr, "%sstack underflow%s\n", BOLD_TEXT, UNBOLD_TEXT);
      if (stop_at_error) abort();
    }
  /* here and in push_stack, both code and args might be non-free only because they've been retyped
   *   inline (as in named let) -- they actually don't make sense in these cases, but are ignored,
   *   and are carried around as GC protection in other cases.
   */
  sc->code =  _NFre(sc->stack_end[0]);
  sc->envir = _TLid(sc->stack_end[1]);
  sc->args =  _NFre(sc->stack_end[2]);
  sc->cur_op = (opcode_t)(sc->stack_end[3]);
  if (sc->cur_op > OP_MAX_DEFINED) 
    {
      fprintf(stderr, "%spop_stack[%d] invalid opcode: %" PRIdPTR " %s\n", BOLD_TEXT, __LINE__, sc->cur_op, UNBOLD_TEXT);
      if (stop_at_error) abort();
    }
}

static void pop_stack_no_op(s7_scheme *sc)
{
  sc->stack_end -= 4;
  if (sc->stack_end < sc->stack_start) 
    {
      fprintf(stderr, "%sstack underflow%s\n", BOLD_TEXT, UNBOLD_TEXT);
      if (stop_at_error) abort();
    }
  sc->code =  _NFre(sc->stack_end[0]);
  sc->envir = _TLid(sc->stack_end[1]);
  sc->args =  _NFre(sc->stack_end[2]);
}

static void push_stack(s7_scheme *sc, opcode_t op, s7_pointer args, s7_pointer code)
{
  if (sc->stack_end >= sc->stack_start + sc->stack_size) 
    {
      fprintf(stderr, "%sstack overflow%s\n", BOLD_TEXT, UNBOLD_TEXT);
      if (stop_at_error) abort();
    }
  if (op > OP_MAX_DEFINED) 
    {
      fprintf(stderr, "%spush_stack[%d] invalid opcode: %" PRIdPTR " %s\n", BOLD_TEXT, __LINE__, sc->cur_op, UNBOLD_TEXT);
      if (stop_at_error) abort();
    }
  if (code) sc->stack_end[0] = _NFre(code);
  sc->stack_end[1] = _TLid(sc->envir);
  if (args) sc->stack_end[2] = _NFre(args);
  sc->stack_end[3] = (s7_pointer)op;
  sc->stack_end += 4;
}

#define push_stack_no_let_no_code(Sc, Op, Args) push_stack(Sc, Op, Args, Sc->gc_nil)
#define push_stack_no_args(Sc, Op, Code) push_stack(Sc, Op, Sc->gc_nil, Code)
#define push_stack_no_let(Sc, Op, Args, Code) push_stack(Sc, Op, Args, Code)
#define push_stack_op(Sc, Op) push_stack(Sc, Op, Sc->gc_nil, Sc->gc_nil)
#define push_stack_op_let(Sc, Op) push_stack(Sc, Op, Sc->gc_nil, Sc->gc_nil)
/* in the non-debugging case, the sc->gc_nil's here are not set, so we can (later) pop free cells */

#else
/* these macros are faster than the equivalent simple function calls.  If the s7_scheme struct is set up to reflect the
 *    stack order [code envir args op], we can use memcpy here:
 *      #define pop_stack(Sc) do {Sc->stack_end -= 4; memcpy((void *)Sc, (void *)(Sc->stack_end), 4 * sizeof(s7_pointer));} while (0)
 *    but it is not any faster (nor are other similar explicit cases)
 */

#define pop_stack(Sc) \
  do { \
      Sc->stack_end -= 4; \
      Sc->code =  Sc->stack_end[0]; \
      Sc->envir = Sc->stack_end[1]; \
      Sc->args =  Sc->stack_end[2]; \
      sc->cur_op = (opcode_t)(Sc->stack_end[3]); \
  } while (0)

#define pop_stack_no_op(Sc) \
  do { \
      Sc->stack_end -= 4; \
      Sc->code =  Sc->stack_end[0]; \
      Sc->envir = Sc->stack_end[1]; \
      Sc->args =  Sc->stack_end[2]; \
  } while (0)

#define push_stack(Sc, Op, Args, Code) \
  do { \
      Sc->stack_end[0] = Code; \
      Sc->stack_end[1] = Sc->envir; \
      Sc->stack_end[2] = Args; \
      Sc->stack_end[3] = (s7_pointer)Op; \
      Sc->stack_end += 4; \
  } while (0)

#define push_stack_no_let_no_code(Sc, Op, Args) \
  do { \
      Sc->stack_end[2] = Args; \
      Sc->stack_end[3] = (s7_pointer)Op; \
      Sc->stack_end += 4; \
  } while (0)

#define push_stack_no_args(Sc, Op, Code) \
  do { \
      Sc->stack_end[0] = Code; \
      Sc->stack_end[1] = Sc->envir; \
      Sc->stack_end[3] = (s7_pointer)Op; \
      Sc->stack_end += 4; \
  } while (0)
#define push_stack_no_let(Sc, Op, Args, Code) \
  do { \
      Sc->stack_end[0] = Code; \
      Sc->stack_end[2] = Args; \
      Sc->stack_end[3] = (s7_pointer)Op; \
      Sc->stack_end += 4; \
  } while (0)
#define push_stack_op(Sc, Op) \
  do { \
      Sc->stack_end[3] = (s7_pointer)Op; \
      Sc->stack_end += 4; \
  } while (0)
#define push_stack_op_let(Sc, Op) \
  do { \
      Sc->stack_end[1] = Sc->envir; \
      Sc->stack_end[3] = (s7_pointer)Op; \
      Sc->stack_end += 4; \
  } while (0)

#endif
/* since we don't GC mark the stack past the stack_top, push_stack_no_args and friends can cause pop_stack to set
 *   sc->code and sc->args to currently free objects.
 */

#define main_stack_op(Sc)   ((opcode_t)(Sc->stack_end[-1]))
/* #define main_stack_args(Sc) (Sc->stack_end[-2]) */
/* #define main_stack_let(Sc)  (Sc->stack_end[-3]) */
/* #define main_stack_code(Sc) (Sc->stack_end[-4]) */
/* #define pop_main_stack(Sc)  Sc->stack_end -= 4 */

/* beware of main_stack_code!  If a function has a tail-call, the main_stack_code that form sees
 *   if main_stack_op==op-begin1 can change from call to call -- the begin actually refers
 *   to the caller, which is dependent on where the current function was called, so we can't hard-wire
 *   any optimizations based on that sequence.
 */

static void stack_reset(s7_scheme *sc)
{
  sc->stack_end = sc->stack_start;
  push_stack_op(sc, OP_EVAL_DONE);
  push_stack_op(sc, OP_BARRIER);
}


static void resize_stack(s7_scheme *sc)
{
  uint32_t i, new_size, loc;  /* int64_ts?? sc->stack_size also is an uint32_t */

  loc = s7_stack_top(sc);
  new_size = sc->stack_size * 2;

  /* how can we trap infinite recursions?  Is a warning in order here?
   *   I think I'll add 'max-stack-size
   *   size currently reaches 8192 in s7test
   */
  if (new_size > sc->max_stack_size)
    s7_error(sc, s7_make_symbol(sc, "stack-too-big"), set_elist_1(sc, s7_make_string_wrapper(sc, "stack has grown past (*s7* 'max-stack-size)")));

  vector_elements(sc->stack) = (s7_pointer *)realloc(vector_elements(sc->stack), new_size * sizeof(s7_pointer));
  if (!vector_elements(sc->stack))
    s7_error(sc, s7_make_symbol(sc, "stack-too-big"), set_elist_1(sc, s7_make_string_wrapper(sc, "no room to expand stack?")));

  for (i = sc->stack_size; i < new_size; i++)
    vector_element(sc->stack, i) = sc->nil;
  vector_length(sc->stack) = new_size;
  sc->stack_size = new_size;

  sc->stack_start = vector_elements(sc->stack);
  sc->stack_end = (s7_pointer *)(sc->stack_start + loc);
  sc->stack_resize_trigger = (s7_pointer *)(sc->stack_start + sc->stack_size / 2);

  if (show_stack_stats(sc))
    {
      fprintf(stderr, "stack grows to %u, %s\n", new_size, DISPLAY_80(sc->code));
      s7_show_let(sc);
    }
}

#define check_stack_size(Sc) \
  if (Sc->stack_end >= Sc->stack_resize_trigger) \
    { \
      if ((Sc->begin_hook) && (call_begin_hook(Sc))) return(Sc->F); \
      resize_stack(Sc); \
    }



/* -------------------------------- symbols -------------------------------- */

static inline uint64_t raw_string_hash(const unsigned char *key, uint32_t len)
{
  uint64_t x;
  unsigned char *cx = (unsigned char *)&x;

  x = 0;
  if (len <= 8)
    memcpy((void *)cx, (void *)key, len);
  else
    {
      uint64_t y;
      unsigned char *cy = (unsigned char *)&y;

      memcpy((void *)cx, (void *)key, 8);
      y = 0;
      len -= 8;
      memcpy((void *)cy, (void *)(key + 8), (len > 8) ? 8 : len);
      x |= y;
    }
  return(x);
}


static inline s7_pointer make_symbol_with_length(s7_scheme *sc, const char *name, uint32_t len);

static s7_pointer new_symbol(s7_scheme *sc, const char *name, uint32_t len, uint64_t hash, uint32_t location)
{
  /* name might not be null-terminated */
  s7_pointer x, str, p;
  unsigned char *base, *val;

  base = (unsigned char *)malloc(sizeof(s7_cell) * 3 + len + 1);
#if DEBUGGING
  /* clear at least debugger_bits here and below */
  memset((void *)base, 0, sizeof(s7_cell) * 3 + len + 1);
#endif
  x = (s7_pointer)base;
  str = (s7_pointer)(base + sizeof(s7_cell));
  p = (s7_pointer)(base + 2 * sizeof(s7_cell));
  val = (unsigned char *)(base + 3 * sizeof(s7_cell));
  memcpy((void *)val, (void *)name, len);
  val[len] = '\0';

  unheap(str);
  typeflag(str) = T_STRING | T_IMMUTABLE;                  /* avoid debugging confusion involving set_type (also below) */
  string_length(str) = len;
  string_value(str) = (char *)val;
  string_hash(str) = hash;
  string_needs_free(str) = false;

  unheap(x);
  typeflag(x) = T_SYMBOL;
  symbol_set_name_cell(x, str);
  set_global_slot(x, sc->undefined);                       /* was sc->nil */
  set_initial_slot(x, sc->undefined);
  symbol_set_local(x, 0LL, sc->nil);
  symbol_set_tag(x, 0);
  symbol_set_ctr(x, 0);
  symbol_type(x) = 0;

  if (symbol_name_length(x) > 1)                           /* not 0, otherwise : is a keyword */
    {
      if (name[0] == ':')
	{
	  typeflag(x) |= (T_IMMUTABLE | T_KEYWORD);
	  keyword_set_symbol(x, make_symbol_with_length(sc, (char *)(name + 1), len - 1));
	  set_has_keyword(keyword_symbol(x));
	  set_global_slot(x, s7_make_slot(sc, sc->nil, x, x));
	}
      else
	{
	  char c;
	  c = name[symbol_name_length(x) - 1];
	  if (c == ':')
	    {
	      char *kstr;
	      uint32_t klen;
	      klen = symbol_name_length(x) - 1;
	      /* can't use tmpbuf_* here (or not safely I think) because name is already using tmpbuf */
	      kstr = (char *)malloc((klen + 1) * sizeof(char));
	      memcpy((void *)kstr, (void *)name, klen);
	      kstr[klen] = 0;
	      typeflag(x) |= (T_IMMUTABLE | T_KEYWORD);
	      keyword_set_symbol(x, make_symbol_with_length(sc, kstr, klen));
	      set_has_keyword(keyword_symbol(x));
	      set_global_slot(x, s7_make_slot(sc, sc->nil, x, x));
	      free(kstr);
	    }
	}
    }

  unheap(p);
  typeflag(p) = T_PAIR | T_IMMUTABLE;
  set_car(p, x);
  set_cdr(p, vector_element(sc->symbol_table, location));
  vector_element(sc->symbol_table, location) = p;
  pair_set_raw_hash(p, hash);
  pair_set_raw_len(p, len);
  pair_set_raw_name(p, string_value(str));
  return(x);
}

static inline s7_pointer make_symbol_with_length(s7_scheme *sc, const char *name, uint32_t len)
{
  s7_pointer x;
  uint64_t hash;
  uint32_t location;

  hash = raw_string_hash((const unsigned char *)name, len);
  location = hash % SYMBOL_TABLE_SIZE;

  if (len <= 8)
    {
      for (x = vector_element(sc->symbol_table, location); is_pair(x); x = cdr(x))
	if ((hash == pair_raw_hash(x)) &&
	    (len == pair_raw_len(x)))
	  return(car(x));
    }
  else
    {
      for (x = vector_element(sc->symbol_table, location); is_pair(x); x = cdr(x))
	if ((hash == pair_raw_hash(x)) &&
	    (len == pair_raw_len(x)) &&
	    (strings_are_equal_with_length(name, pair_raw_name(x), len))) /* length here because name might not be null-terminated */
	  return(car(x));
    }
  return(new_symbol(sc, name, len, hash, location));
}


static s7_pointer make_symbol(s7_scheme *sc, const char *name)
{
  return(make_symbol_with_length(sc, name, safe_strlen(name)));
}


s7_pointer s7_make_symbol(s7_scheme *sc, const char *name)
{
  if (!name) return(sc->F);
  return(make_symbol_with_length(sc, name, safe_strlen(name)));
}

static s7_pointer symbol_table_find_by_name(s7_scheme *sc, const char *name, uint64_t hash, uint32_t location)
{
  s7_pointer x;
  for (x = vector_element(sc->symbol_table, location); is_not_null(x); x = cdr(x))
    if ((hash == pair_raw_hash(x)) &&
	(strings_are_equal(name, pair_raw_name(x))))
      return(car(x));
  return(sc->nil);
}


s7_pointer s7_symbol_table_find_name(s7_scheme *sc, const char *name)
{
  uint64_t hash;
  uint32_t location;
  s7_pointer result;

  hash = raw_string_hash((const unsigned char *)name, safe_strlen(name));
  location = hash % SYMBOL_TABLE_SIZE;
  result = symbol_table_find_by_name(sc, name, hash, location);
  if (is_null(result))
    return(NULL);

  return(result);
}


#define FILLED true
#define NOT_FILLED false

static s7_pointer g_symbol_table(s7_scheme *sc, s7_pointer args)
{
  #define H_symbol_table "(symbol-table) returns a vector containing the current symbol-table symbols"
  #define Q_symbol_table s7_make_signature(sc, 1, sc->is_vector_symbol)

  s7_pointer lst, x;
  s7_pointer *els;
  int32_t i, j, syms = 0;

  /* this can't be optimized by returning the actual symbol-table (a vector of lists), because
   *    gensyms can cause the table's lists and symbols to change at any time.  This wreaks havoc
   *    on traversals like for-each.  So, symbol-table returns a snap-shot of the table contents
   *    at the time it is called, and we call gc before making the list.  I suppose the next step
   *    is to check that we have room, and increase the heap here if necessary!
   *
   *    (define (for-each-symbol func num) (for-each (lambda (sym) (if (> num 0) (for-each-symbol func (- num 1)) (func sym))) (symbol-table)))
   *    (for-each-symbol (lambda (sym) (gensym) 1))
   */

  for (i = 0; i < SYMBOL_TABLE_SIZE; i++)
    for (x = vector_element(sc->symbol_table, i); is_not_null(x); x = cdr(x))
      syms++;
  sc->w = make_vector_1(sc, syms, NOT_FILLED, T_VECTOR);
  els = vector_elements(sc->w);

  for (i = 0, j = 0; i < SYMBOL_TABLE_SIZE; i++)
    for (x = vector_element(sc->symbol_table, i); is_not_null(x); x = cdr(x))
      els[j++] = car(x);

  lst = sc->w;
  sc->w = sc->nil;
  return(lst);
}


bool s7_for_each_symbol_name(s7_scheme *sc, bool (*symbol_func)(const char *symbol_name, void *data), void *data)
{
  /* this includes the special constants #<unspecified> and so on for simplicity -- are there any others? */
  int32_t i;
  s7_pointer x;

  for (i = 0; i < SYMBOL_TABLE_SIZE; i++)
    for (x = vector_element(sc->symbol_table, i); is_not_null(x); x = cdr(x))
      if (symbol_func(symbol_name(car(x)), data))
	return(true);

  return((symbol_func("#t", data))             ||
	 (symbol_func("#f", data))             ||
	 (symbol_func("#<unspecified>", data)) ||
	 (symbol_func("#<undefined>", data))   ||
	 (symbol_func("#<eof>", data))         ||
	 (symbol_func("#true", data))          ||
	 (symbol_func("#false", data)));
}


bool s7_for_each_symbol(s7_scheme *sc, bool (*symbol_func)(const char *symbol_name, s7_pointer value, void *data), void *data)
{
  int32_t i;
  s7_pointer x;

  for (i = 0; i < SYMBOL_TABLE_SIZE; i++)
    for (x = vector_element(sc->symbol_table, i); is_not_null(x); x = cdr(x))
      if (symbol_func(symbol_name(car(x)), cdr(x), data))
	return(true);

  return(false);
}


static void remove_gensym_from_symbol_table(s7_scheme *sc, s7_pointer sym)
{
  /* sym is a free cell at this point (we're called after the GC), but the name_cell is still intact */
  s7_pointer x, name;
  uint32_t location;

  name = symbol_name_cell(sym);
  location = string_hash(name) % SYMBOL_TABLE_SIZE;
  x = vector_element(sc->symbol_table, location);

  if (car(x) == sym)
    {
      vector_element(sc->symbol_table, location) = cdr(x);
      free(x);
    }
  else
    {
      s7_pointer y;
      for (y = x, x = cdr(x); is_pair(x); y = x, x = cdr(x))
	{
	  if (car(x) == sym)
	    {
	      set_cdr(y, cdr(x));
	      free(x);
	      return;
	    }
	}
#if DEBUGGING
      fprintf(stderr, "could not remove %s?\n", string_value(name));
#endif
    }
}


s7_pointer s7_gensym(s7_scheme *sc, const char *prefix)
{
  char *name;
  uint32_t len, location;
  uint64_t hash;
  s7_pointer x;

  len = safe_strlen(prefix) + 32;
  tmpbuf_malloc(name, len);
  /* there's no point in heroic efforts here to avoid name collisions -- the user can screw up no matter what we do */
  len = snprintf(name, len, "{%s}-%u", prefix, sc->gensym_counter++);
  hash = raw_string_hash((const unsigned char *)name, len);
  location = hash % SYMBOL_TABLE_SIZE;
  x = new_symbol(sc, name, len, hash, location);  /* not T_GENSYM -- might be called from outside */
  tmpbuf_free(name, len);
  return(x);
}


static bool s7_is_gensym(s7_pointer g) {return((is_symbol(g)) && (is_gensym(g)));}

static s7_pointer g_is_gensym(s7_scheme *sc, s7_pointer args)
{
  #define H_is_gensym "(gensym? sym) returns #t if sym is a gensym"
  #define Q_is_gensym pl_bt

  check_boolean_method(sc, s7_is_gensym, sc->is_gensym_symbol, args);
}


static char *pos_int_to_str(s7_int num, uint32_t *len, char endc)
{
  #define INT_TO_STR_SIZE 32
  static char itos[INT_TO_STR_SIZE];
  char *p, *op;

  p = (char *)(itos + INT_TO_STR_SIZE - 1);
  op = p;
  *p-- = '\0';
  if (endc != '\0') *p-- = endc;
  do {*p-- = "0123456789"[num % 10]; num /= 10;} while (num);
  (*len) = op - p;           /* this includes the trailing #\null */
  return((char *)(p + 1));
}

static s7_pointer g_gensym(s7_scheme *sc, s7_pointer args)
{
  #define H_gensym "(gensym (prefix \"gensym\")) returns a new, unused symbol"
  #define Q_gensym s7_make_signature(sc, 2, sc->is_gensym_symbol, sc->is_string_symbol)

  const char *prefix;
  char *name, *p;
  uint32_t len, plen, nlen, location;
  uint64_t hash;
  s7_pointer x, str, stc;

  /* get symbol name */
  if (is_not_null(args))
    {
      s7_pointer name;
      name = car(args);
      if (!is_string(name))
	method_or_bust_one_arg(sc, name, sc->gensym_symbol, args, T_STRING);
      prefix = string_value(name);
    }
  else prefix = "gensym";
  plen = safe_strlen(prefix);
  len = plen + 32;
  name = (char *)malloc(len * sizeof(char));
  name[0] = '{';
  if (plen > 0) memcpy((void *)(name + 1), prefix, plen);
  name[plen + 1] = '}';
  name[plen + 2] = '-'; /* {gensym}-nnn */

  p = pos_int_to_str(sc->gensym_counter++, &len, '\0');
  memcpy((void *)(name + plen + 3), (void *)p, len);
  nlen = len + plen + 2;

  hash = raw_string_hash((const unsigned char *)name, nlen);
  location = hash % SYMBOL_TABLE_SIZE;

  /* make-string for symbol name */
#if DEBUGGING
  str = (s7_cell *)calloc(1, sizeof(s7_cell));
#else
  str = (s7_cell *)malloc(sizeof(s7_cell));
#endif
  unheap(str);
#if DEBUGGING
  typeflag(str) = 0;
#endif
  set_type(str, T_STRING | T_IMMUTABLE);
  string_length(str) = nlen;
  string_value(str) = name;
  string_needs_free(str) = false;
  string_hash(str) = hash;

  /* allocate the symbol in the heap so GC'd when inaccessible */
  new_cell(sc, x, T_SYMBOL | T_GENSYM);
  symbol_set_name_cell(x, str);
  set_global_slot(x, sc->undefined);
  set_initial_slot(x, sc->undefined);
  symbol_set_local(x, 0LL, sc->nil);
  symbol_set_ctr(x, 0);

  /* place new symbol in symbol-table, but using calloc so we can easily free it (remove it from the table) in GC sweep */
#if DEBUGGING
  stc = (s7_cell *)calloc(1, sizeof(s7_cell));
#else
  stc = (s7_cell *)malloc(sizeof(s7_cell));
#endif
#if DEBUGGING
  typeflag(stc) = 0;
#endif
  unheap(stc);
  set_type(stc, T_PAIR | T_IMMUTABLE);
  set_car(stc, x);
  set_cdr(stc, vector_element(sc->symbol_table, location));
  vector_element(sc->symbol_table, location) = stc;
  pair_set_raw_hash(stc, hash);
  pair_set_raw_len(stc, string_length(str));
  pair_set_raw_name(stc, string_value(str));

  add_gensym(sc, x);
  return(x);
}


s7_pointer s7_name_to_value(s7_scheme *sc, const char *name)
{
  return(s7_symbol_value(sc, make_symbol(sc, name)));
}


bool s7_is_syntax(s7_pointer p)
{
  return(is_syntax(p));
}

static s7_pointer g_is_syntax(s7_scheme *sc, s7_pointer args)
{
  #define H_is_syntax "(syntax? obj) returns #t if obj is a syntactic value (e.g. lambda)"
  #define Q_is_syntax pl_bt

  check_boolean_method(sc, is_syntax, sc->is_syntax_symbol, args);
}


bool s7_is_symbol(s7_pointer p)
{
  return(is_symbol(p));
}

static s7_pointer g_is_symbol(s7_scheme *sc, s7_pointer args)
{
  #define H_is_symbol "(symbol? obj) returns #t if obj is a symbol"
  #define Q_is_symbol pl_bt

  check_boolean_method(sc, is_symbol, sc->is_symbol_symbol, args);
}


const char *s7_symbol_name(s7_pointer p)
{
  return(symbol_name(p));
}


static s7_pointer g_symbol_to_string(s7_scheme *sc, s7_pointer args)
{
  #define H_symbol_to_string "(symbol->string sym) returns the symbol sym converted to a string"
  #define Q_symbol_to_string s7_make_signature(sc, 2, sc->is_string_symbol, sc->is_symbol_symbol)
  s7_pointer sym;

  sym = car(args);
  if (!is_symbol(sym))
    method_or_bust_one_arg(sc, sym, sc->symbol_to_string_symbol, args, T_SYMBOL);
  /* s7_make_string uses strlen which stops at an embedded null */
  return(s7_make_string_with_length(sc, symbol_name(sym), symbol_name_length(sym)));    /* return a copy */
}

static s7_pointer symbol_to_string_uncopied;
static s7_pointer g_symbol_to_string_uncopied(s7_scheme *sc, s7_pointer args)
{
  s7_pointer sym;

  sym = car(args);
  if (!is_symbol(sym))
    method_or_bust_one_arg(sc, sym, sc->symbol_to_string_symbol, args, T_SYMBOL);
  if (is_gensym(sym))
    return(s7_make_string_with_length(sc, symbol_name(sym), symbol_name_length(sym)));    /* return a copy of gensym name (which will be freed) */
  return(symbol_name_cell(sym));
}

static s7_pointer symbol_to_string_p(s7_pointer sym)
{
  if (!is_symbol(sym))
    simple_wrong_type_argument(cur_sc, cur_sc->symbol_to_string_symbol, sym, T_SYMBOL);
  return(s7_make_string_with_length(cur_sc, symbol_name(sym), symbol_name_length(sym))); 
}

static s7_pointer symbol_to_string_uncopied_p(s7_pointer sym)
{
  if (!is_symbol(sym))
    simple_wrong_type_argument(cur_sc, cur_sc->symbol_to_string_symbol, sym, T_SYMBOL);
  if (is_gensym(sym))
    return(s7_make_string_with_length(cur_sc, symbol_name(sym), symbol_name_length(sym)));
  return(symbol_name_cell(sym));
}


static s7_pointer g_string_to_symbol_1(s7_scheme *sc, s7_pointer str, s7_pointer caller)
{
  if (!is_string(str))
    method_or_bust_one_arg(sc, str, caller, list_1(sc, str), T_STRING);
  if (string_length(str) == 0)
    return(simple_wrong_type_argument_with_type(sc, caller, str, s7_make_string_wrapper(sc, "a non-null string")));

  /* currently if the string has an embedded null, it marks the end of the new symbol name. */
  return(make_symbol_with_length(sc, string_value(str), string_length(str)));
}


static s7_pointer g_string_to_symbol(s7_scheme *sc, s7_pointer args)
{
  #define H_string_to_symbol "(string->symbol str) returns the string str converted to a symbol"
  #define Q_string_to_symbol s7_make_signature(sc, 2, sc->is_symbol_symbol, sc->is_string_symbol)
  return(g_string_to_symbol_1(sc, car(args), sc->string_to_symbol_symbol));
}

static s7_pointer string_to_symbol_p_p(s7_pointer p)
{
  if (!is_string(p))
    simple_wrong_type_argument(cur_sc, cur_sc->string_to_symbol_symbol, p, T_STRING);
  if (string_length(p) == 0)
    simple_wrong_type_argument_with_type(cur_sc, cur_sc->string_to_symbol_symbol, p, s7_make_string_wrapper(cur_sc, "a non-null string"));
  return(make_symbol_with_length(cur_sc, string_value(p), string_length(p)));
}

static s7_pointer g_string_append(s7_scheme *sc, s7_pointer args);

static s7_pointer g_symbol(s7_scheme *sc, s7_pointer args)
{
  #define H_symbol "(symbol str ...) returns its string arguments concatenated and converted to a symbol"
  #define Q_symbol s7_make_circular_signature(sc, 1, 2, sc->is_symbol_symbol, sc->is_string_symbol)
  if (is_null(cdr(args)))
    return(g_string_to_symbol_1(sc, car(args), sc->symbol_symbol));
  return(g_string_to_symbol_1(sc, g_string_append(sc, args), sc->symbol_symbol));
}


static s7_pointer add_symbol_to_list(s7_scheme *sc, s7_pointer sym)
{
  symbol_set_tag(sym, sc->syms_tag);
  return(sym);
}

#define clear_symbol_list(Sc) Sc->syms_tag++
#define symbol_is_in_list(Sc, Sym) (symbol_tag(Sym) == Sc->syms_tag)



/* -------------------------------- environments -------------------------------- */

#define new_frame(Sc, Old_Env, New_Env)		      \
  do {						      \
    s7_pointer _x_;				      \
      new_cell(Sc, _x_, T_LET | T_SAFE_PROCEDURE);    \
      let_id(_x_) = ++sc->let_number;		      \
      let_set_slots(_x_, Sc->nil);	              \
      set_outlet(_x_, Old_Env);			      \
      New_Env = _x_;				      \
  } while (0)
/* this macro is noticeably faster than using the equivalent inlined new_frame_in_env function below */

static inline s7_pointer new_frame_in_env(s7_scheme *sc, s7_pointer old_env)
{
  /* return(cons(sc, sc->nil, old_env)); */
  s7_pointer x;
  new_cell(sc, x, T_LET | T_SAFE_PROCEDURE);
  let_id(x) = ++sc->let_number;
  let_set_slots(x, sc->nil);
  set_outlet(x, old_env);
  return(x);
}


static inline s7_pointer make_simple_let(s7_scheme *sc)
{
  s7_pointer frame;
  new_cell(sc, frame, T_LET | T_SAFE_PROCEDURE);
  let_id(frame) = sc->let_number + 1;
  let_set_slots(frame, sc->nil);
  set_outlet(frame, sc->envir);
  return(frame);
}


/* in all these macros, symbol_set_local should follow slot_set_value so that we can evaluate the
 *    slot's value in its old state.
 */
#define add_slot(Frame, Symbol, Value)			\
  do {							\
    s7_pointer _slot_, _sym_, _val_;			\
    _sym_ = Symbol; _val_ = Value;			\
    new_cell_no_check(sc, _slot_, T_SLOT);\
    slot_set_symbol(_slot_, _sym_);			\
    slot_set_value(_slot_, _val_);			\
    symbol_set_local(_sym_, let_id(Frame), _slot_);	\
    set_next_slot(_slot_, let_slots(Frame));		\
    let_set_slots(Frame, _slot_);	                \
  } while (0)

#define add_slot_checked(Frame, Symbol, Value)		\
  do {							\
    s7_pointer _slot_, _sym_, _val_;			\
    _sym_ = Symbol; _val_ = Value;			\
    new_cell(sc, _slot_, T_SLOT);		        \
    slot_set_symbol(_slot_, _sym_);			\
    slot_set_value(_slot_, _val_);			\
    symbol_set_local(_sym_, let_id(Frame), _slot_);	\
    set_next_slot(_slot_, let_slots(Frame));		\
    let_set_slots(Frame, _slot_); 	                \
  } while (0)

/* no set_local here -- presumably done earlier in check_* */

#define new_frame_with_slot(Sc, Old_Env, New_Env, Symbol, Value) \
  do {								 \
    s7_pointer _x_, _slot_, _sym_, _val_;			 \
    _sym_ = Symbol; _val_ = Value;				\
    new_cell(Sc, _x_, T_LET | T_SAFE_PROCEDURE);		\
    let_id(_x_) = ++sc->let_number;				\
    set_outlet(_x_, Old_Env);			                \
    New_Env = _x_;						\
    new_cell_no_check(Sc, _slot_, T_SLOT);	                \
    slot_set_symbol(_slot_, _sym_);				\
    slot_set_value(_slot_, _val_);	                        \
    symbol_set_local(_sym_, sc->let_number, _slot_);            \
    set_next_slot(_slot_, sc->nil);			        \
    let_set_slots(_x_, _slot_);					\
  } while (0)


#define new_frame_with_two_slots(Sc, Old_Env, New_Env, Symbol1, Value1, Symbol2, Value2) \
  do {                                   \
    s7_pointer _x_, _slot_, _sym1_, _val1_, _sym2_, _val2_;		\
    _sym1_ = Symbol1; _val1_ = Value1;					\
    _sym2_ = Symbol2; _val2_ = Value2;					\
    new_cell(Sc, _x_, T_LET | T_SAFE_PROCEDURE);			\
    let_id(_x_) = ++sc->let_number;					\
    set_outlet(_x_, Old_Env);				                \
    New_Env = _x_;							\
    new_cell_no_check(Sc, _slot_, T_SLOT);		                \
    slot_set_symbol(_slot_, _sym1_);					\
    slot_set_value(_slot_, _val1_);					\
    symbol_set_local(_sym1_, sc->let_number, _slot_);			\
    let_set_slots(_x_, _slot_);			                        \
    new_cell_no_check(Sc, _x_, T_SLOT);			                \
    slot_set_symbol(_x_, _sym2_);					\
    slot_set_value(_x_, _val2_);					\
    symbol_set_local(_sym2_, sc->let_number, _x_);			\
    set_next_slot(_x_, sc->nil);				        \
    set_next_slot(_slot_, _x_);			                        \
  } while (0)


static s7_pointer reuse_as_let(s7_scheme *sc, s7_pointer frame, s7_pointer next_frame)
{
  /* we're reusing frame here as a let -- it was probably a pair */
#if DEBUGGING
  frame->debugger_bits = 0;
#endif
  set_type(frame, T_LET | T_SAFE_PROCEDURE);
  let_set_slots(frame, sc->nil);
  set_outlet(frame, next_frame);
  let_id(frame) = ++sc->let_number;
  return(frame);
}

static s7_pointer reuse_as_slot(s7_pointer slot, s7_pointer symbol, s7_pointer value)
{
#if DEBUGGING
  slot->debugger_bits = 0;
#endif
  set_type(slot, T_SLOT);
  slot_set_symbol(slot, symbol);
  slot_set_value(slot, _NFre(value));
  return(slot);
}


static s7_pointer old_frame_with_slot(s7_scheme *sc, s7_pointer env, s7_pointer val)
{
  s7_pointer x, sym;
  uint64_t id;

  id = ++sc->let_number;
  let_id(env) = id;
  x = let_slots(env);
  slot_set_value(x, val);
  sym = slot_symbol(x);
  symbol_set_local(sym, id, x);

  return(env);
}


static s7_pointer old_frame_with_two_slots(s7_scheme *sc, s7_pointer env, s7_pointer val1, s7_pointer val2)
{
  s7_pointer x, sym;
  uint64_t id;

  id = ++sc->let_number;
  let_id(env) = id;
  x = let_slots(env);
  slot_set_value(x, val1);
  sym = slot_symbol(x);
  symbol_set_local(sym, id, x);
  x = next_slot(x);
  slot_set_value(x, val2);
  sym = slot_symbol(x);
  symbol_set_local(sym, id, x);

  return(env);
}


static s7_pointer old_frame_with_three_slots(s7_scheme *sc, s7_pointer env, s7_pointer val1, s7_pointer val2, s7_pointer val3)
{
  s7_pointer x, sym;
  uint64_t id;

  id = ++sc->let_number;
  let_id(env) = id;
  x = let_slots(env);

  slot_set_value(x, val1);
  sym = slot_symbol(x);
  symbol_set_local(sym, id, x);
  x = next_slot(x);

  slot_set_value(x, val2);
  sym = slot_symbol(x);
  symbol_set_local(sym, id, x);
  x = next_slot(x);

  slot_set_value(x, val3);
  sym = slot_symbol(x);
  symbol_set_local(sym, id, x);

  return(env);
}

static s7_pointer permanent_slot(s7_pointer symbol, s7_pointer value)
{
  s7_pointer x;
  x = alloc_pointer();
  unheap(x);
  set_type(x, T_SLOT);
  slot_set_symbol(x, symbol);
  slot_set_value(x, value);
  return(x);
}


static s7_pointer find_let(s7_scheme *sc, s7_pointer obj)
{
  if (is_let(obj)) return(obj);
  switch (type(obj))
    {
    case T_LET:
      return(obj);

    case T_MACRO:   case T_MACRO_STAR:
    case T_BACRO:   case T_BACRO_STAR:
    case T_CLOSURE: case T_CLOSURE_STAR:
      return(closure_let(obj));

    case T_C_OBJECT:
      return(c_object_let(obj));

    case T_C_POINTER:
      if (is_let(raw_pointer_info(obj)))
	return(raw_pointer_info(obj));
    }
  return(sc->nil);
}


static s7_pointer let_fill(s7_scheme *sc, s7_pointer args)
{
  s7_pointer e, val;
  e = car(args);

  if (e == sc->rootlet)
    return(out_of_range(sc, sc->fill_symbol, small_int(1), e, s7_make_string_wrapper(sc, "can't fill! rootlet")));
  if (e == sc->owlet)
    return(out_of_range(sc, sc->fill_symbol, small_int(1), e, s7_make_string_wrapper(sc, "can't fill! owlet")));
  if (is_funclet(e))
    return(out_of_range(sc, sc->fill_symbol, small_int(1), e, s7_make_string_wrapper(sc, "can't fill! a funclet")));

  val = cadr(args);
  if (val == sc->undefined)
    {
      let_set_slots(e, sc->nil);
      let_id(e) = ++sc->let_number; /* else previous symbol_id matches! */
    }
  else
    {
      s7_pointer p;
      for (p = let_slots(e); is_slot(p); p = next_slot(p))
	slot_set_value(p, val);
    }
  return(val);
}


static s7_pointer find_method(s7_scheme *sc, s7_pointer env, s7_pointer symbol)
{
  s7_pointer x;
  if (symbol_id(symbol) == 0) /* this means the symbol has never been used locally, so how can it be a method? */
    return(sc->undefined);

  /* I think the symbol_id is in sync with let_id, so the standard search should work */
  if (let_id(env) == symbol_id(symbol))
    return(slot_value(local_slot(symbol)));

  for (x = env; symbol_id(symbol) < let_id(x); x = outlet(x));

  if (let_id(x) == symbol_id(symbol))
    return(slot_value(local_slot(symbol)));

  for (; is_let(x); x = outlet(x))
    {
      s7_pointer y;
      for (y = let_slots(x); is_slot(y); y = next_slot(y))
	if (slot_symbol(y) == symbol)
	  return(slot_value(y));
    }
  return(sc->undefined);
}


static int32_t let_length(s7_scheme *sc, s7_pointer e)
{
  /* used by length, applicable_length, copy, and some length optimizations */
  int32_t i;
  s7_pointer p;

  if (e == sc->rootlet)
    return(sc->rootlet_entries);

  if (has_methods(e))
    {
      s7_pointer length_func;
      length_func = find_method(sc, e, sc->length_symbol);
      if (length_func != sc->undefined)
	{
	  p = s7_apply_function(sc, length_func, list_1(sc, e));
	  if (s7_is_integer(p))
	    return((int32_t)s7_integer(p));
	  return(-1); /* ?? */
	}
    }

  for (i = 0, p = let_slots(e); is_slot(p); i++, p = next_slot(p));
  return(i);
}


static void slot_set_value_with_hook_1(s7_scheme *sc, s7_pointer slot, s7_pointer value)
  {
    /* (set! (hook-functions *rootlet-redefinition-hook*) (list (lambda (hook) (format *stderr* "~A ~A~%" (hook 'symbol) (hook 'value))))) */
    s7_pointer symbol;
    symbol = slot_symbol(slot);
    if ((global_slot(symbol) == slot) &&      
	(value != slot_value(slot)))
      s7_call(sc, sc->rootlet_redefinition_hook, set_elist_2(sc, symbol, value));
    slot_set_value(slot, value);
  }


static s7_pointer make_slot_1(s7_scheme *sc, s7_pointer env, s7_pointer symbol, s7_pointer value)
{
  /* env is not rootlet and is a let */
  s7_pointer slot;
  new_cell(sc, slot, T_SLOT);
  slot_set_symbol(slot, symbol);
  slot_set_value(slot, value);
  set_next_slot(slot, let_slots(env));
  let_set_slots(env, slot);
  set_local(symbol);
  /* this is called by varlet so we have to be careful about the resultant let_id
   *   check for greater to ensure shadowing stays in effect, and equal to do updates (set! in effect)
   */
  if (let_id(env) >= symbol_id(symbol))
    symbol_set_local(symbol, let_id(env), slot);
  return(slot);
}

static hash_entry_t *hash_eq(s7_scheme *sc, s7_pointer table, s7_pointer key);
static s7_pointer hash_table_iterate(s7_scheme *sc, s7_pointer iterator);

static void remove_function_from_heap(s7_scheme *sc, s7_pointer value);
static void remove_let_from_heap(s7_scheme *sc, s7_pointer lt)
{
  s7_pointer p;
  for (p = let_slots(lt); is_slot(p); p = next_slot(p))
    {
      s7_pointer val;
      val = slot_value(p);
      if ((has_closure_let(val)) &&
	  (heap_location(closure_args(val)) >= 0))
	remove_function_from_heap(sc, val);
      else
	{
	  /* an experiment... */
	  if ((is_hash_table(val)) &&
	      (!hash_table_removed(val)))
	    {
	      s7_pointer iterator, p;
	      uint32_t gc_iter;
	      int32_t i, len;
	      
	      len = hash_table_entries(val);
	      iterator = s7_make_iterator(sc, val);
	      gc_iter = s7_gc_protect(sc, iterator);
	      p = cons(sc, sc->F, sc->F);
	      iterator_current(iterator) = p;
	      set_mark_seq(iterator);
	      for (i = 0; i < len; i++)
		{
		  s7_pointer key_val;
		  key_val = hash_table_iterate(sc, iterator);
		  if ((has_closure_let(cdr(key_val))) &&
		      (heap_location(closure_args(cdr(key_val))) >= 0))
		    remove_function_from_heap(sc, cdr(key_val));
		}
	      hash_table_set_removed(val);
	      s7_gc_unprotect_at(sc, gc_iter);
	      iterator_current(iterator) = sc->nil;
	      free_cell(sc, p);
	      free_cell(sc, iterator);
	    }
	}
    }
  let_set_removed(lt);
}

static void remove_function_from_heap(s7_scheme *sc, s7_pointer value)
{
  s7_pointer lt;
  s7_remove_from_heap(sc, closure_args(value));
  s7_remove_from_heap(sc, closure_body(value));
  
  /* remove closure if it's local to current func */
  lt = closure_let(value);
  if ((is_let(lt)) && (!let_removed(lt)) && (lt != sc->rootlet) && (lt != sc->shadow_rootlet))
    {
      lt = outlet(lt);
      if ((is_let(lt)) && (!let_removed(lt)) && (lt != sc->rootlet) && (lt != sc->shadow_rootlet))
	{
	  remove_let_from_heap(sc, lt);
	  lt = outlet(lt);
	  if ((is_let(lt)) && (!let_removed(lt)) && (lt != sc->rootlet) && (lt != sc->shadow_rootlet))
	    remove_let_from_heap(sc, lt);
	}
    }
}

s7_pointer s7_make_slot(s7_scheme *sc, s7_pointer env, s7_pointer symbol, s7_pointer value)
{
  if ((!is_let(env)) ||
      (env == sc->rootlet))
    {
      s7_pointer ge, slot;
      if ((sc->safety == NO_SAFETY) && 
	  (has_closure_let(value)))
	remove_function_from_heap(sc, value);

      /* first look for existing slot -- this is not always checked before calling s7_make_slot */
      if (is_slot(global_slot(symbol)))
	{
	  slot = global_slot(symbol);
	  slot_set_value_with_hook(slot, value);
	  return(slot);
	}

      ge = sc->rootlet;
      slot = permanent_slot(symbol, value);
      vector_element(ge, sc->rootlet_entries++) = slot;
      if (sc->rootlet_entries >= vector_length(ge))
	{
	  int32_t i, len;
	  vector_length(ge) *= 2;
	  len = vector_length(ge);
	  vector_elements(ge) = (s7_pointer *)realloc(vector_elements(ge), len * sizeof(s7_pointer));
	  for (i = sc->rootlet_entries; i < len; i++)
	    vector_element(ge, i) = sc->nil;
	}
      set_global_slot(symbol, slot);
      
      if (symbol_id(symbol) == 0) /* never defined locally? */
	{
	  if (initial_slot(symbol) == sc->undefined)
	    set_initial_slot(symbol, permanent_slot(symbol, value));
	  set_local_slot(symbol, slot);
	  symbol_increment_ctr(symbol);
	  set_global(symbol);
	}
      if (is_gensym(symbol))
	s7_remove_from_heap(sc, symbol);
      return(slot);
    }

  return(make_slot_1(sc, env, symbol, value));
  /* there are about the same number of frames as local variables -- this
   *   strikes me as surprising, but it holds up across a lot of code.
   */
}


static s7_pointer make_slot(s7_scheme *sc, s7_pointer variable, s7_pointer value)
{
  /* this is for a do-loop optimization -- an unattached slot */
  s7_pointer y;
  new_cell(sc, y, T_SLOT);
  slot_set_symbol(y, variable);
  slot_set_value(y, value);
  return(y);
}


/* -------------------------------- let? -------------------------------- */
bool s7_is_let(s7_pointer e)
{
  return(is_let(e));
}

static s7_pointer g_is_let(s7_scheme *sc, s7_pointer args)
{
  #define H_is_let "(let? obj) returns #t if obj is a let (an environment)."
  #define Q_is_let pl_bt

  check_boolean_method(sc, is_let, sc->is_let_symbol, args);
}


/* -------------------------------- unlet -------------------------------- */
#define UNLET_ENTRIES 512 /* 397 if not --disable-deprecated etc */

static void save_unlet(s7_scheme *sc)
{
  int32_t i, k = 0;
  s7_pointer x;
  s7_pointer *inits;

  sc->unlet = (s7_pointer)calloc(1, sizeof(s7_cell));
  set_type(sc->unlet, T_VECTOR);
  vector_length(sc->unlet) = UNLET_ENTRIES;
  vector_elements(sc->unlet) = (s7_pointer *)malloc(UNLET_ENTRIES * sizeof(s7_pointer));
  vector_getter(sc->unlet) = default_vector_getter;
  vector_setter(sc->unlet) = default_vector_setter;
  inits = vector_elements(sc->unlet);
  s7_vector_fill(sc, sc->unlet, sc->nil);
  unheap(sc->unlet);

  for (i = 0; i < SYMBOL_TABLE_SIZE; i++)
    for (x = vector_element(sc->symbol_table, i); is_not_null(x); x = cdr(x))
      {
	s7_pointer sym;
	sym = car(x);
	if (is_slot(initial_slot(sym)))
	  {
	    s7_pointer val;
	    val = slot_value(initial_slot(sym));
	    if ((is_c_function(val)) || (is_syntax(val))) /* we're assuming the initial_slots values of these guys need no GC protection */
	      inits[k++] = initial_slot(sym);

	    /* non-c_functions that are not 'set! (and therefore initial_slot GC) protected by default: 
	     *    make-hook hook-functions (and deprecated procedure-arity) 
	     * if these initial_slot values are added to unlet, they need explicit GC protection.
	     */
	    /* (let ((begin +)) (with-let (unlet) (begin 1 2))) */
#if DEBUGGING
	    if (k >= UNLET_ENTRIES)
	      fprintf(stderr, "unlet overflow\n");
#endif
	  }
      }
}

static s7_pointer g_unlet(s7_scheme *sc, s7_pointer args)
{
  /* add sc->unlet bindings to the current environment */
  #define H_unlet "(unlet) establishes the original bindings of all the predefined functions"
  #define Q_unlet s7_make_signature(sc, 1, sc->is_let_symbol)

  /* slightly confusing:
   *    :((unlet) 'abs)
   *    #<undefined>
   *    :(defined? 'abs (unlet))
   *    #t
   * this is because unlet sets up a local environment of unshadowed symbols,
   *   and s7_let_ref below only looks at the local env chain (that is, if env is not
   *   the global env, then the global env is not searched).
   *
   * Also (define hi 3) #_hi => 3, (set! hi 4), #_hi -> 3 but (with-let (unlet) hi) -> 4!
   */
  int32_t i;
  s7_pointer *inits;
  s7_pointer x;

  sc->w = new_frame_in_env(sc, sc->envir);
  inits = vector_elements(sc->unlet);

  for (i = 0; (i < UNLET_ENTRIES) && (is_slot(inits[i])); i++)
    {
      s7_pointer sym;
      x = slot_value(inits[i]);
      sym = slot_symbol(inits[i]);
      if (is_procedure(x))
	{
	  if (((!is_global(sym)) &&                  /* it might be shadowed locally */
	       (s7_symbol_local_value(sc, sym, sc->envir) != slot_value(global_slot(sym)))) ||
	      (x != slot_value(global_slot(sym))))   /* it's not shadowed, but has been changed globally */
	    make_slot_1(sc, sc->w, sym, x);
	}
      else
	{
	  if ((is_syntax(x)) &&
	      (local_slot(sym) != sc->nil))          /* this can be a freed cell, will be nil if unchanged */
	    make_slot_1(sc, sc->w, sym, x);
	}
    }
  /* if (set! + -) then + needs to be overridden, but the local bit isn't set,
   *   so we have to check the actual values in the non-local case.
   *   (define (f x) (with-let (unlet) (+ x 1)))
   */

  x = sc->w;
  sc->w = sc->nil;
  return(x);
}


/* -------------------------------- openlet? -------------------------------- */
bool s7_is_openlet(s7_pointer e)
{
  return(has_methods(e));
}

static s7_pointer g_is_openlet(s7_scheme *sc, s7_pointer args)
{
  #define H_is_openlet "(openlet? obj) returns #t is 'obj' has methods."
  #define Q_is_openlet pl_bt

  /* if car(args) is not a let, should this raise an error? -- no, easier to use this way in cond */
  check_method(sc, car(args), sc->is_openlet_symbol, args);
  return(make_boolean(sc, has_methods(car(args))));
}


/* -------------------------------- openlet -------------------------------- */
s7_pointer s7_openlet(s7_scheme *sc, s7_pointer e)
{
  set_has_methods(e);
  return(e);
}

static s7_pointer g_openlet(s7_scheme *sc, s7_pointer args)
{
  #define H_openlet "(openlet e) tells the built-in generic functions that the environment 'e might have an over-riding method."
  #define Q_openlet pcl_e
  s7_pointer e;

  e = car(args);
  check_method(sc, e, sc->openlet_symbol, args);

  if (e == sc->rootlet)
    s7_error(sc, sc->error_symbol, set_elist_1(sc, s7_make_string(sc, "can't openlet rootlet")));
    
  if ((is_let(e)) ||
      (has_closure_let(e)) ||
      ((is_c_object(e)) && (c_object_let(e) != sc->nil)) ||
      ((is_c_pointer(e)) && (is_let(raw_pointer_info(e)))))
    {
      set_has_methods(e);
      return(e);
    }

  return(simple_wrong_type_argument_with_type(sc, sc->openlet_symbol, e, a_let_string));
}



/* -------------------------------- coverlet -------------------------------- */

static s7_pointer g_coverlet(s7_scheme *sc, s7_pointer args)
{
  s7_pointer e;
  #define H_coverlet "(coverlet e) undoes an earlier openlet."
  #define Q_coverlet pcl_e

  e = car(args);
  sc->temp3 = e;
  check_method(sc, e, sc->coverlet_symbol, list_1(sc, e));
  sc->temp3 = sc->nil;
  if (e == sc->rootlet)
    s7_error(sc, sc->error_symbol, set_elist_1(sc, s7_make_string(sc, "can't coverlet rootlet")));

  if ((is_let(e)) ||
      (has_closure_let(e)) ||
      ((is_c_object(e)) && (c_object_let(e) != sc->nil)) ||
      ((is_c_pointer(e)) && (is_let(raw_pointer_info(e)))))
    {
      clear_has_methods(e);
      return(e);
    }
  return(simple_wrong_type_argument_with_type(sc, sc->coverlet_symbol, e, a_let_string));
}



/* -------------------------------- varlet -------------------------------- */
static void append_let(s7_scheme *sc, s7_pointer new_e, s7_pointer old_e)
{
  s7_pointer x;

  if (old_e == sc->rootlet)
    return;

  if (new_e != sc->rootlet)
    {
      for (x = let_slots(old_e); is_slot(x); x = next_slot(x))
	make_slot_1(sc, new_e, slot_symbol(x), slot_value(x)); /* not add_slot here because we might run off the free heap end */
    }
  else
    {
      for (x = let_slots(old_e); is_slot(x); x = next_slot(x))
	{
	  s7_pointer sym, val;
	  sym = slot_symbol(x);
	  val = slot_value(x);
	  if (is_slot(global_slot(sym)))
	    slot_set_value(global_slot(sym), val);
	  else s7_make_slot(sc, new_e, sym, val);
	}
    }
}

static s7_pointer check_c_obj_env(s7_scheme *sc, s7_pointer old_e, s7_pointer caller)
{
  if (is_c_object(old_e))
    old_e = c_object_let(old_e);
  if (!is_let(old_e))
    return(simple_wrong_type_argument_with_type(sc, caller, old_e, a_let_string));
  return(old_e);
}


s7_pointer s7_varlet(s7_scheme *sc, s7_pointer env, s7_pointer symbol, s7_pointer value)
{
  if (!is_let(env))
    return(wrong_type_argument_with_type(sc, sc->varlet_symbol, 1, env, a_let_string));

  if (!is_symbol(symbol))
    return(wrong_type_argument_with_type(sc, sc->varlet_symbol, 2, symbol, a_symbol_string));

  if ((is_slot(global_slot(symbol))) &&
      (is_syntax(slot_value(global_slot(symbol)))))
    return(wrong_type_argument_with_type(sc, sc->varlet_symbol, 2, symbol, s7_make_string_wrapper(sc, "a non-syntactic name")));

  if (env == sc->rootlet)
    {
      if (is_slot(global_slot(symbol)))
	slot_set_value(global_slot(symbol), value);
      else s7_make_slot(sc, env, symbol, value);
    }
  else make_slot_1(sc, env, symbol, value);
  return(value);
}


static s7_pointer g_varlet(s7_scheme *sc, s7_pointer args)
{
  #define H_varlet "(varlet env ...) adds its arguments (an environment, a cons: symbol . value, or a pair of arguments, the symbol and its value) \
to the environment env, and returns the environment."
  #define Q_varlet s7_make_circular_signature(sc, 2, 4, sc->is_let_symbol, \
                     s7_make_signature(sc, 2, sc->is_let_symbol, sc->is_null_symbol), \
                       s7_make_signature(sc, 3, sc->is_pair_symbol, sc->is_symbol_symbol, sc->is_let_symbol), \
                         sc->T)
  /* varlet = with-let + define */
  s7_pointer x, e, sym, val, p;

  e = car(args);
  if (is_null(e))
    e = sc->rootlet;
  else
    {
      check_method(sc, e, sc->varlet_symbol, args);
      if (!is_let(e))
	return(wrong_type_argument_with_type(sc, sc->varlet_symbol, 1, e, a_let_string));
    }

  for (x = cdr(args); is_pair(x); x = cdr(x))
    {
      p = car(x);
      switch (type(p))
	{
	case T_SYMBOL:
	  if (is_keyword(p))
	    sym = keyword_symbol(p);
	  else sym = p;
	  if (!is_pair(cdr(x)))
	    s7_error(sc, sc->error_symbol, set_elist_3(sc, value_is_missing_string, sc->varlet_symbol, car(x)));
	  x = cdr(x);
	  val = car(x);
	  break;

	case T_PAIR:
	  sym = car(p);
	  if (!is_symbol(sym))
	    return(wrong_type_argument_with_type(sc, sc->varlet_symbol, position_of(x, args), p, a_symbol_string));
	  val = cdr(p);
	  break;

	case T_LET:
	  append_let(sc, e, check_c_obj_env(sc, p, sc->varlet_symbol));
	  continue;

	default:
	  return(wrong_type_argument_with_type(sc, sc->varlet_symbol, position_of(x, args), p, a_symbol_string));
	}

      if (is_immutable_symbol(sym))
	return(wrong_type_argument_with_type(sc, sc->varlet_symbol, position_of(x, args), sym, a_non_constant_symbol_string));

      if (e == sc->rootlet)
	{
	  if (is_slot(global_slot(sym)))
	    {
	      if (is_syntax(slot_value(global_slot(sym))))
		return(wrong_type_argument_with_type(sc, sc->varlet_symbol, position_of(x, args), p, s7_make_string_wrapper(sc, "a non-syntactic keyword")));
	      /*  without this check we can end up turning our code into gibberish:
	       *   :(set! quote 1)
	       *   ;can't set! quote
	       *   :(varlet (rootlet) '(quote . 1))
	       *   :quote
	       *   1
	       * or worse set quote to a function of one arg that tries to quote something -- infinite loop
	       */
	      slot_set_value_with_hook(global_slot(sym), val);
	    }
	  else s7_make_slot(sc, e, sym, val);
	}
      else make_slot_1(sc, e, sym, val);
      /* this used to check for sym already defined, and set its value, but that greatly slows down
       *   the most common use (adding a slot), and makes it hard to shadow explicitly.  Don't use
       *   varlet as a substitute for set!/let-set!.
       */
    }
  return(e);
}


/* -------------------------------- cutlet -------------------------------- */
static s7_pointer g_cutlet(s7_scheme *sc, s7_pointer args)
{
  #define H_cutlet "(cutlet e symbol ...) removes symbols from the environment e."
  #define Q_cutlet s7_make_circular_signature(sc, 2, 3, sc->is_let_symbol, sc->is_let_symbol, sc->is_symbol_symbol)

  s7_pointer e, syms;
  #define THE_UN_ID ++sc->let_number

  e = car(args);
  if (is_null(e))
    e = sc->rootlet;
  else
    {
      check_method(sc, e, sc->cutlet_symbol, args);
      if (!is_let(e))
	return(wrong_type_argument_with_type(sc, sc->cutlet_symbol, 1, e, a_let_string));
    }
  /* besides removing the slot we have to make sure the symbol_id does not match else
   *   let-ref and others will use the old slot!  What's the un-id?  Perhaps the next one?
   *   (let ((b 1)) (let ((b 2)) (cutlet (curlet) 'b)) b)
   */
  for (syms = cdr(args); is_pair(syms); syms = cdr(syms))
    {
      s7_pointer sym, slot;
      sym = car(syms);

      if (!is_symbol(sym))
	return(wrong_type_argument_with_type(sc, sc->cutlet_symbol, position_of(syms, args), sym, a_symbol_string));

      if (is_keyword(sym))
	sym = keyword_symbol(sym);

      if (e == sc->rootlet)
	{
	  if (is_slot(global_slot(sym)))
	    {
	      symbol_set_id(sym, THE_UN_ID);
	      slot_set_value(global_slot(sym), sc->undefined);
	    }
	}
      else
	{
	  slot = let_slots(e);
	  if (is_slot(slot))
	    {
	      if (slot_symbol(slot) == sym)
		{
		  let_set_slots(e, next_slot(let_slots(e)));
		  symbol_set_id(sym, THE_UN_ID);
		}
	      else
		{
		  s7_pointer last_slot;
		  last_slot = slot;
		  for (slot = next_slot(let_slots(e)); is_slot(slot); last_slot = slot, slot = next_slot(slot))
		    {
		      if (slot_symbol(slot) == sym)
			{
			  symbol_set_id(sym, THE_UN_ID);
			  set_next_slot(last_slot, next_slot(slot));
			  break; 
			}
		    }
		}
	    }
	}
    }
  return(e);
}


/* -------------------------------- sublet -------------------------------- */
static s7_pointer sublet_1(s7_scheme *sc, s7_pointer e, s7_pointer bindings, s7_pointer caller)
{
  s7_pointer new_e;
  /* fprintf(stderr, "%s sublet: %s %s\n", DISPLAY(caller), DISPLAY(e), DISPLAY(bindings)); */

  if (e == sc->rootlet)
    new_e = new_frame_in_env(sc, sc->nil);
  else new_e = new_frame_in_env(sc, e);
  set_all_methods(new_e, e);

  if (!is_null(bindings))
    {
      s7_pointer x;
      sc->temp3 = new_e;

      for (x = bindings; is_pair(x); x = cdr(x))
	{
	  s7_pointer p, sym, val;

	  p = car(x);
	  switch (type(p))
	    {
	    case T_SYMBOL:
	      if (is_keyword(p))
		sym = keyword_symbol(p);
	      else sym = p;
	      if (!is_pair(cdr(x)))
		s7_error(sc, sc->error_symbol, set_elist_3(sc, value_is_missing_string, caller, car(x)));
	      x = cdr(x);
	      val = car(x);
	      break;

	    case T_PAIR:
	      sym = car(p);
	      if (!is_symbol(sym))
		return(wrong_type_argument_with_type(sc, caller, position_of(x, bindings), p, a_symbol_string));
	      val = cdr(p);
	      break;

	    case T_LET:
	      append_let(sc, new_e, check_c_obj_env(sc, p, caller));
	      continue;

	    default:
	      return(wrong_type_argument_with_type(sc, caller, position_of(x, bindings), p, a_symbol_string));
	    }

	  if (is_immutable_symbol(sym))
	    return(wrong_type_argument_with_type(sc, caller, position_of(x, bindings), sym, a_non_constant_symbol_string));
	  if ((is_slot(global_slot(sym))) &&
	      (is_syntax(slot_value(global_slot(sym)))))
	    return(wrong_type_argument_with_type(sc, sc->sublet_symbol, 2, sym, s7_make_string_wrapper(sc, "a non-syntactic name")));

	  /* here we know new_e is a let and is not rootlet */
	  make_slot_1(sc, new_e, sym, val);
	  if (sym == sc->let_ref_fallback_symbol)
	    set_has_ref_fallback(new_e);
	  else
	    {
	      if (sym == sc->let_set_fallback_symbol)
		set_has_set_fallback(new_e);
	    }
	}
      sc->temp3 = sc->nil;
    }
  return(new_e);
}

s7_pointer s7_sublet(s7_scheme *sc, s7_pointer e, s7_pointer bindings)
{
  return(sublet_1(sc, e, bindings, sc->sublet_symbol));
}

static s7_pointer g_sublet(s7_scheme *sc, s7_pointer args)
{
  #define H_sublet "(sublet env ...) adds its \
arguments (each an environment or a cons: symbol . value) to the environment env, and returns the \
new environment."
  #define Q_sublet Q_varlet
  s7_pointer e;

  e = car(args);
  if (is_null(e))
    e = sc->rootlet;
  else
    {
      check_method(sc, e, sc->sublet_symbol, args);
      if (!is_let(e))
	return(wrong_type_argument_with_type(sc, sc->sublet_symbol, 1, e, a_let_string));
    }
  return(sublet_1(sc, e, cdr(args), sc->sublet_symbol));
}


/* -------------------------------- inlet -------------------------------- */
s7_pointer s7_inlet(s7_scheme *sc, s7_pointer args)
{
  #define H_inlet "(inlet ...) adds its \
arguments, each an environment, a cons: '(symbol . value), or a keyword/value pair, to a new environment, and returns the \
new environment. (inlet :a 1 :b 2) or (inlet '(a . 1) '(b . 2))"
  #define Q_inlet s7_make_circular_signature(sc, 1, 2, sc->is_let_symbol, sc->T)

  return(sublet_1(sc, sc->rootlet, args, sc->inlet_symbol));
}

#define g_inlet s7_inlet


static s7_pointer simple_inlet;
static s7_pointer g_simple_inlet(s7_scheme *sc, s7_pointer args)
{
  /* here all args are paired with normal symbol/value, no fallbacks, no immutable symbols etc */
  s7_pointer new_e, x;
  int64_t id;
  new_e = new_frame_in_env(sc, sc->nil);
  sc->temp3 = new_e;
  id = let_id(new_e);
  for (x = args; is_pair(x); x = cddr(x))
    {
      s7_pointer symbol, slot;
      symbol = car(x);
      new_cell(sc, slot, T_SLOT);
      slot_set_symbol(slot, symbol);
      slot_set_value(slot, cadr(x));
      set_next_slot(slot, let_slots(new_e));
      let_set_slots(new_e, slot);
      set_local(symbol);
      symbol_set_local(symbol, id, slot);
    }
  sc->temp3 = sc->nil;
  return(new_e);
}

static bool is_proper_quote(s7_scheme *sc, s7_pointer p)
{
  return((is_pair(p)) &&
	 (car(p) == sc->quote_symbol) &&
	 (is_pair(cdr(p))));
}

static s7_pointer inlet_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
  if (!ops) return(f);
  if ((args > 0) &&
      ((args % 2) == 0))
    {
      s7_pointer p;
      for (p = cdr(expr); is_pair(p); p = cddr(p))
	{
	  s7_pointer sym;
	  if (!is_proper_quote(sc, car(p)))
	    return(f);
	  sym = cadar(p);
	  if ((!is_symbol(sym)) ||
	      (is_immutable_symbol(sym)) ||
	      (sym == sc->let_ref_fallback_symbol) ||
	      (sym == sc->let_set_fallback_symbol))
	    return(f);
	}
      return(simple_inlet);
    }
  return(f);
}



/* -------------------------------- let->list -------------------------------- */
s7_pointer s7_let_to_list(s7_scheme *sc, s7_pointer env)
{
  s7_pointer x;

  sc->temp3 = sc->w;
  sc->w = sc->nil;

  if (env == sc->rootlet)
    {
      uint32_t i, lim2;
      s7_pointer *entries;

      entries = vector_elements(env);
      lim2 = sc->rootlet_entries;
      if (lim2 & 1) lim2--;

      for (i = 0; i < lim2; )
	{
	  sc->w = cons_unchecked(sc, cons(sc, slot_symbol(entries[i]), slot_value(entries[i])), sc->w); i++;
	  sc->w = cons_unchecked(sc, cons_unchecked(sc, slot_symbol(entries[i]), slot_value(entries[i])), sc->w); i++;
	}
      if (lim2 < sc->rootlet_entries)
	sc->w = cons_unchecked(sc, cons(sc, slot_symbol(entries[i]), slot_value(entries[i])), sc->w);
    }
  else
    {
      s7_pointer iter, func;
      /* need to check make-iterator method before dropping into let->list */

      if ((has_methods(env)) && ((func = find_method(sc, env, sc->make_iterator_symbol)) != sc->undefined))
	iter = s7_apply_function(sc, func, list_1(sc, env));
      else iter = sc->nil;

      if (is_null(iter))
	{
	  for (x = let_slots(env); is_slot(x); x = next_slot(x))
	    sc->w = cons_unchecked(sc, cons(sc, slot_symbol(x), slot_value(x)), sc->w);
	}
      else
	{
	  /* (begin (load "mockery.scm") (let ((lt ((*mock-pair* 'mock-pair) 1 2 3))) (format *stderr* "~{~A ~}" lt))) */
	  while (true)
	    {
	      x = s7_iterate(sc, iter);
	      if (iterator_is_at_end(iter)) break;
	      sc->w = cons(sc, x, sc->w);
	    }
	  sc->w = safe_reverse_in_place(sc, sc->w);
	}
    }
  x = sc->w;
  sc->w = sc->temp3;
  sc->temp3 = sc->nil;
  return(x);
}

#if (!WITH_PURE_S7)
static s7_pointer g_let_to_list(s7_scheme *sc, s7_pointer args)
{
  #define H_let_to_list "(let->list env) returns env's bindings as a list of cons's: '(symbol . value)."
  #define Q_let_to_list s7_make_signature(sc, 2, sc->is_pair_symbol, sc->is_let_symbol)

  s7_pointer env;
  env = car(args);
  check_method(sc, env, sc->let_to_list_symbol, args);
  if (!is_let(env))
    {
      if (is_c_object(env))
	env = c_object_let(env);
      else
	{
	  if (is_c_pointer(env))
	    env = raw_pointer_info(env);
	}
      if (!is_let(env))
        return(simple_wrong_type_argument_with_type(sc, sc->let_to_list_symbol, env, a_let_string));
    }
  return(s7_let_to_list(sc, env));
}
#endif


/* -------------------------------- let-ref -------------------------------- */

static s7_pointer g_let_ref(s7_scheme *sc, s7_pointer args)
{
  #define H_let_ref "(let-ref env sym) returns the value of the symbol sym in the environment env"
  #define Q_let_ref s7_make_signature(sc, 3, sc->T, sc->is_let_symbol, sc->is_symbol_symbol)
  s7_pointer env, symbol, x, y;

  /* (let ((a 1)) ((curlet) 'a))
   * ((rootlet) 'abs)
   */

  env = car(args);
  if (!is_let(env))
    return(wrong_type_argument_with_type(sc, sc->let_ref_symbol, 1, env, a_let_string));

  symbol = cadr(args);
  if (is_keyword(symbol))
    symbol = keyword_symbol(symbol);

  if (!is_symbol(symbol))
    {
      check_method(sc, env, sc->let_ref_symbol, args);
      if (has_ref_fallback(env))
	check_method(sc, env, sc->let_ref_fallback_symbol, args);
      return(wrong_type_argument_with_type(sc, sc->let_ref_symbol, 2, symbol, a_symbol_string));
    }

  if (env == sc->rootlet)
    {
      y = global_slot(symbol);
      if (is_slot(y))
	return(slot_value(y));
      return(sc->undefined);
    }

  if (let_id(env) == symbol_id(symbol))
    return(slot_value(local_slot(symbol))); /* this obviously has to follow the global-env check */

  for (x = env; is_let(x); x = outlet(x))
    for (y = let_slots(x); is_slot(y); y = next_slot(y))
      if (slot_symbol(y) == symbol)
	return(slot_value(y));

  /* now for a horrible kludge.  If a let is a mock-hash-table (for example), implicit
   *   indexing of the hash-table collides with the same thing for the let (field names
   *   versus keys), and we can't just try again here because that makes it too easy to
   *   get into infinite recursion.  So, 'let-ref-fallback...
   */

  if (has_methods(env))
    {
      if (has_ref_fallback(env))
	apply_known_method(sc, env, sc->let_ref_fallback_symbol, args);

      /* why did this ignore a global value? Changed 24-May-16 to check rootlet if no methods --
       *   apparently I was using #<undefined> here (pre-rootlet-check) to indicate that an
       *   open let did not have a particular method (locally).  This seems inconsistent now,
       *   but it was far worse before.  At least (let () ((curlet) 'pi)) is pi!
       */
    }
  else
    {
      y = global_slot(symbol);
      if (is_slot(y))
	return(slot_value(y));
    }

  return(sc->undefined);
}


static s7_pointer lint_let_ref_1(s7_scheme *sc, s7_pointer lt, s7_pointer sym)
{
  s7_pointer x, y;
  for (x = lt; is_let(x); x = outlet(x))
    for (y = let_slots(x); is_slot(y); y = next_slot(y))
      if (slot_symbol(y) == sym)
	return(slot_value(y));

  if (has_methods(lt))
    {
      if (has_ref_fallback(lt))
	apply_known_method(sc, lt, sc->let_ref_fallback_symbol, set_plist_2(sc, lt, sym));
    }
  else
    {
      y = global_slot(sym);
      if (is_slot(y))
	return(slot_value(y));
    }
  return(sc->undefined);
}

s7_pointer s7_let_ref(s7_scheme *sc, s7_pointer env, s7_pointer symbol) {return(g_let_ref(sc, set_plist_2(sc, env, symbol)));}
static s7_pointer let_ref_p_pp(s7_pointer p1, s7_pointer p2) {return(g_let_ref(cur_sc, set_plist_2(cur_sc, p1, p2)));}

static s7_pointer lint_let_ref;
static s7_pointer g_lint_let_ref(s7_scheme *sc, s7_pointer args)
{
  s7_pointer lt;
  lt = find_symbol_unchecked(sc, opt_sym2(args)); /* cadar */
  if (is_pair(lt))
    {
      lt = cdr(lt);
      if (is_let(lt))
	{
	  s7_pointer y, sym;
	  sym = opt_sym3(args); /* cadadr */
	  for (y = let_slots(lt); is_slot(y); y = next_slot(y))
	    if (slot_symbol(y) == sym)
	      return(slot_value(y));
	  return(lint_let_ref_1(sc, outlet(lt), sym));
	}
      return(wrong_type_argument_with_type(sc, sc->let_ref_symbol, 1, lt, a_let_string));
    }
  return(simple_wrong_type_argument(sc, sc->cdr_symbol, lt, T_PAIR));
}

static s7_pointer local_lint_let_ref;
static s7_pointer g_local_lint_let_ref(s7_scheme *sc, s7_pointer args)
{
  s7_pointer lt;
  lt = local_symbol_value(opt_sym2(args));
  if (is_pair(lt))
    {
      lt = cdr(lt);
      if (is_let(lt))
	{
	  s7_pointer y, sym;
	  sym = opt_sym3(args);
	  for (y = let_slots(lt); is_slot(y); y = next_slot(y))
	    if (slot_symbol(y) == sym)
	      return(slot_value(y));
	  return(lint_let_ref_1(sc, outlet(lt), sym));
	}
      return(wrong_type_argument_with_type(sc, sc->let_ref_symbol, 1, lt, a_let_string));
    }
  return(simple_wrong_type_argument(sc, sc->cdr_symbol, lt, T_PAIR));
}


static s7_pointer let_ref_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
  if (!ops) return(f);
  if ((is_h_safe_c_c(expr)) &&
      ((raw_opt1(expr) == lint_let_ref) || /* perhaps check is_safe_c_op(expr) then opt_cfunc(expr) rather than using raw_opt here and in the set case below */
       (raw_opt1(expr) == local_lint_let_ref)))
    return(raw_opt1(expr));
  
  if (optimize_op(expr) == HOP_SAFE_C_opSq_Q)
    {
      s7_pointer arg1, arg2;
      arg1 = cadr(expr);
      arg2 = caddr(expr);
      if ((car(arg1) == sc->cdr_symbol) &&
	  (is_symbol(cadr(arg2))) &&
	  (!is_immutable_symbol(cadr(arg2))))
	{
	  set_optimize_op(expr, HOP_SAFE_C_C);
	  set_opt_sym2(cdr(expr), cadr(arg1));
	  set_opt_sym3(cdr(expr), cadr(arg2));
	  if (is_local_symbol(cdr(arg1)))
	    return(local_lint_let_ref);
	  return(lint_let_ref);
	}
    }
  return(f);
}


/* -------------------------------- let-set! -------------------------------- */
static s7_pointer call_accessor(s7_scheme *sc, s7_pointer slot, s7_pointer old_value)
{
  s7_pointer func, new_value;

  /* new_value = sc->error_symbol; */
  func = slot_accessor(slot);

  if (is_procedure_or_macro(func))
    {
      if (is_c_function(func))
	{
	  set_car(sc->t2_1, slot_symbol(slot));
	  set_car(sc->t2_2, old_value);
      	  new_value = c_function_call(func)(sc, sc->t2_1);
	}
      else
	{
	  bool old_off;
	  old_off = sc->gc_off;
	  sc->gc_off = true;
	  new_value = s7_apply_function(sc, func, list_2(sc, slot_symbol(slot), old_value));
	  sc->gc_off = old_off;
	}
    }
  else return(old_value);

  if (new_value == sc->error_symbol)
    return(s7_error(sc, sc->error_symbol, set_elist_3(sc, s7_make_string_wrapper(sc, "can't set! ~S to ~S"), slot_symbol(slot), old_value)));
  return(new_value);
}

static s7_pointer let_set_1(s7_scheme *sc, s7_pointer env, s7_pointer symbol, s7_pointer value)
{
  s7_pointer x, y;
  static s7_pointer err = NULL;

  if (is_keyword(symbol))
    symbol = keyword_symbol(symbol);

  if (env == sc->rootlet)
    {
      if (is_immutable_symbol(symbol))  /* (let-set! (rootlet) :rest #f) */
	return(wrong_type_argument_with_type(sc, sc->let_set_symbol, 2, symbol, a_non_constant_symbol_string));
      y = global_slot(symbol);
      if (is_slot(y))
	{
	  if (slot_has_accessor(y))
	    slot_set_value(y, call_accessor(sc, y, value));
	  else slot_set_value(y, value);
	  return(slot_value(y));
	}

      if (!err) err = s7_make_permanent_string("let-set! ~A is not defined in ~A");
      return(s7_error(sc, sc->wrong_type_arg_symbol, set_elist_3(sc, err, symbol, env)));
      /* return(sc->undefined); */
    }

  for (x = env; is_let(x); x = outlet(x))
    for (y = let_slots(x); is_slot(y); y = next_slot(y))
      if (slot_symbol(y) == symbol)
	{
	  if (slot_has_accessor(y))
	    slot_set_value(y, call_accessor(sc, y, value));
	  else slot_set_value(y, value);
	  return(slot_value(y));
	}

  if (has_methods(env))
    {
      if (has_set_fallback(env))
	apply_known_method(sc, env, sc->let_set_fallback_symbol, sc->w = list_3(sc, env, symbol, value));
    }
  else
    {
      y = global_slot(symbol);
      if (is_slot(y))
	{
	  if (slot_has_accessor(y))
	    slot_set_value(y, call_accessor(sc, y, value));
	  else slot_set_value(y, value);
	  return(slot_value(y));
	}
    }

  if (!err) err = s7_make_permanent_string("let-set! ~A is not defined in ~A");
  return(s7_error(sc, sc->wrong_type_arg_symbol, set_elist_3(sc, err, symbol, env)));
  /* return(sc->undefined); */
  /* not sure about this -- what's the most useful choice? */
}

s7_pointer s7_let_set(s7_scheme *sc, s7_pointer env, s7_pointer symbol, s7_pointer value)
{
  if (!is_let(env))
    return(wrong_type_argument_with_type(sc, sc->let_set_symbol, 1, env, a_let_string));

  if (!is_symbol(symbol))
    {
      check_method(sc, env, sc->let_set_symbol, sc->w = list_3(sc, env, symbol, value));
      if (has_set_fallback(env))
	apply_known_method(sc, env, sc->let_set_fallback_symbol, sc->w = list_3(sc, env, symbol, value));
      return(wrong_type_argument_with_type(sc, sc->let_set_symbol, 2, symbol, a_symbol_string));
    }

  return(let_set_1(sc, env, symbol, value));
}

static s7_pointer g_let_set(s7_scheme *sc, s7_pointer args)
{
  /* (let ((a 1)) (set! ((curlet) 'a) 32) a) */
  #define H_let_set "(let-set! env sym val) sets the symbol sym's value in the environment env to val"
  #define Q_let_set s7_make_signature(sc, 4, sc->T, sc->is_let_symbol, sc->is_symbol_symbol, sc->T)

  return(s7_let_set(sc, car(args), cadr(args), caddr(args)));
}

static s7_pointer let_set_p_ppp(s7_pointer p1, s7_pointer p2, s7_pointer p3) {return(s7_let_set(cur_sc, p1, p2, p3));}

static s7_pointer lint_let_set, local_lint_let_set;
static s7_pointer g_lint_let_set_1(s7_scheme *sc, s7_pointer lt1, s7_pointer sym, s7_pointer val)
{
  s7_pointer lt, x, y;
  static s7_pointer err = NULL;

  lt = (is_pair(lt1)) ? cdr(lt1) : g_cdr(sc, set_plist_1(sc, lt1));
  if (!is_let(lt))
    return(wrong_type_argument_with_type(sc, sc->let_set_symbol, 1, lt, a_let_string));

  if (lt == sc->rootlet)
    {
      y = global_slot(sym);
      if (is_slot(y))
	{
	  if (slot_has_accessor(y))
	    slot_set_value(y, call_accessor(sc, y, val));
	  else slot_set_value(y, val);
	  return(slot_value(y));
	}
      if (!err) err = s7_make_permanent_string("let-set! ~A is not defined in ~A");
      return(s7_error(sc, sc->wrong_type_arg_symbol, set_elist_3(sc, err, sym, lt)));
    }

  for (x = lt; is_let(x); x = outlet(x))
    for (y = let_slots(x); is_slot(y); y = next_slot(y))
      if (slot_symbol(y) == sym)
	{
	  if (slot_has_accessor(y))
	    slot_set_value(y, call_accessor(sc, y, val));
	  else slot_set_value(y, val);
	  return(slot_value(y));
	}

  if (has_methods(lt))
    {
      if (has_set_fallback(lt))
	apply_known_method(sc, lt, sc->let_set_fallback_symbol, sc->w = list_3(sc, lt, sym, val));
    }
  else
    {
      y = global_slot(sym);
      if (is_slot(y))
	{
	  if (slot_has_accessor(y))
	    slot_set_value(y, call_accessor(sc, y, val));
	  else slot_set_value(y, val);
	  return(slot_value(y));
	}
    }
  if (!err) err = s7_make_permanent_string("let-set! ~A is not defined in ~A");
  return(s7_error(sc, sc->wrong_type_arg_symbol, set_elist_3(sc, err, sym, lt)));
}

static s7_pointer g_lint_let_set(s7_scheme *sc, s7_pointer args)
{
  return(g_lint_let_set_1(sc, find_symbol_checked(sc, opt_sym2(args)), opt_sym3(args), find_symbol_unchecked(sc, caddr(args))));
}

static s7_pointer g_local_lint_let_set(s7_scheme *sc, s7_pointer args)
{
  return(g_lint_let_set_1(sc, local_symbol_value(opt_sym2(args)), opt_sym3(args), local_symbol_value(caddr(args))));
}


static s7_pointer let_set_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
  if (!ops) return(f);
  if ((is_h_safe_c_c(expr)) &&
      ((raw_opt1(expr) == lint_let_set) ||
       (raw_opt1(expr) == local_lint_let_set)))
    return(raw_opt1(expr));

  if (optimize_op(expr) == HOP_SAFE_C_opSq_QS)
    {
      s7_pointer arg1, arg2, arg3;
      arg1 = cadr(expr);
      arg2 = caddr(expr);
      arg3 = cadddr(expr);
      if ((car(arg1) == sc->cdr_symbol) &&
	  (is_symbol(cadr(arg2))) &&
	  (!is_immutable_symbol(cadr(arg2))) &&
	  (is_symbol(arg3)) &&
	  (!is_immutable_symbol(arg3)))
	{
	  set_optimize_op(expr, HOP_SAFE_C_C);
	  set_opt_sym2(cdr(expr), cadr(arg1));
	  set_opt_sym3(cdr(expr), cadr(arg2));
	  if ((is_local_symbol(cdr(arg1))) &&
	      (is_local_symbol(cdddr(expr))))
	    return(local_lint_let_set);
	  return(lint_let_set);
	}
    }
  return(f);
}


static s7_pointer reverse_slots(s7_scheme *sc, s7_pointer list)
{
  s7_pointer p = list, result, q;
  result = sc->nil;

  while (is_slot(p))
    {
      q = next_slot(p);
      set_next_slot(p, result);
      result = p;
      p = q;
    }
  return(result);
}


static s7_pointer let_copy(s7_scheme *sc, s7_pointer env)
{
  if (is_let(env))
    {
      s7_pointer new_e;

      if (env == sc->rootlet)   /* (copy (rootlet)) or (copy (funclet abs)) etc */
	return(sc->rootlet);

      /* we can't make copy handle environments-as-objects specially because the
       *   make-object function in define-class uses copy to make a new object!
       *   So if it is present, we get it here, and then there's almost surely trouble.
       */
      new_e = new_frame_in_env(sc, outlet(env));
      set_all_methods(new_e, env);
      sc->temp3 = new_e;
      if (is_slot(let_slots(env)))
	{
	  s7_int id;
	  s7_pointer x, y = NULL;

	  id = let_id(new_e);
	  for (x = let_slots(env); is_slot(x); x = next_slot(x))
	    {
	      s7_pointer z;
	      new_cell(sc, z, T_SLOT);
	      slot_set_symbol(z, slot_symbol(x));
	      slot_set_value(z, slot_value(x));
	      if (symbol_id(slot_symbol(z)) != id) /* keep shadowing intact */
		symbol_set_local(slot_symbol(x), id, z);
	      if (is_slot(let_slots(new_e)))
		set_next_slot(y, z);
	      else let_set_slots(new_e, z);
	      set_next_slot(z, sc->nil);              /* in case GC runs during this loop */
	      y = z;
	    }
	}
      /* We can't do a (normal) loop here then reverse the slots later because the symbol's local_slot has to
       *    match the unshadowed slot, not the last in the list:
       *    (let ((e1 (inlet 'a 1 'a 2))) (let ((e2 (copy e1))) (list (equal? e1 e2) (equal? (e1 'a) (e2 'a)))))
       */
      sc->temp3 = sc->nil;
      return(new_e);
    }
  return(sc->nil);
}


/* -------------------------------- rootlet -------------------------------- */
static s7_pointer g_rootlet(s7_scheme *sc, s7_pointer ignore)
{
  #define H_rootlet "(rootlet) returns the current top-level definitions (symbol bindings)."
  #define Q_rootlet s7_make_signature(sc, 1, sc->is_let_symbol)
  return(sc->rootlet);
}
/* as with the symbol-table, this function can lead to disaster -- user could
 *   clobber the environment etc.  But we want it to be editable and augmentable,
 *   so I guess I'll leave it alone.  (See curlet|funclet as well).
 */

s7_pointer s7_rootlet(s7_scheme *sc)
{
  return(sc->rootlet);
}

s7_pointer s7_shadow_rootlet(s7_scheme *sc)
{
  return(sc->shadow_rootlet);
}

s7_pointer s7_set_shadow_rootlet(s7_scheme *sc, s7_pointer let)
{
  sc->shadow_rootlet = let;
  return(let);
}


/* -------------------------------- curlet -------------------------------- */
static s7_pointer g_curlet(s7_scheme *sc, s7_pointer args)
{
  #define H_curlet "(curlet) returns the current definitions (symbol bindings)"
  #define Q_curlet s7_make_signature(sc, 1, sc->is_let_symbol)

  sc->capture_let_counter++;
  if (is_let(sc->envir))
    return(sc->envir);
  return(sc->rootlet);
}

s7_pointer s7_curlet(s7_scheme *sc)
{
  sc->capture_let_counter++;
  return(sc->envir);
}

s7_pointer s7_set_curlet(s7_scheme *sc, s7_pointer e)
{
  s7_pointer p, old_e;
  old_e = sc->envir;
  sc->envir = e;

  if ((is_let(e)) && (let_id(e) > 0)) /* might be () [id=-1] or rootlet [id=0] etc */
    {
      let_id(e) = ++sc->let_number;
      for (p = let_slots(e); is_slot(p); p = next_slot(p))
	{
	  s7_pointer sym;
	  sym = slot_symbol(p);
	  if (symbol_id(sym) != sc->let_number)
	    symbol_set_local(sym, sc->let_number, p);
	}
    }

  return(old_e);
}


/* -------------------------------- outlet -------------------------------- */
s7_pointer s7_outlet(s7_scheme *sc, s7_pointer e)
{
  return(outlet(e));
}

static s7_pointer g_outlet(s7_scheme *sc, s7_pointer args)
{
  #define H_outlet "(outlet env) is the environment that contains env."
  #define Q_outlet s7_make_signature(sc, 2, sc->is_let_symbol, sc->is_let_symbol)

  s7_pointer env;
  env = car(args);
  if (!is_let(env))
    method_or_bust_with_type_one_arg(sc, env, sc->outlet_symbol, args, a_let_string);

  if ((env == sc->rootlet) ||
      (is_null(outlet(env))))
    return(sc->rootlet);
  return(outlet(env));
}


static s7_pointer g_set_outlet(s7_scheme *sc, s7_pointer args)
{
  /* (let ((a 1)) (let ((b 2)) (set! (outlet (curlet)) (rootlet)) ((curlet) 'a))) */
  s7_pointer env, new_outer;

  env = car(args);
  if (!is_let(env))
    return(s7_wrong_type_arg_error(sc, "set! outlet", 1, env, "a let"));

  new_outer = cadr(args);
  if (!is_let(new_outer))
    return(s7_wrong_type_arg_error(sc, "set! outlet", 2, new_outer, "a let"));

  if (env != sc->rootlet)
    set_outlet(env, (new_outer == sc->rootlet) ? sc->nil : new_outer);
  return(new_outer);
}



static s7_pointer find_symbol(s7_scheme *sc, s7_pointer symbol)
{
  s7_pointer x;

  if (let_id(sc->envir) == symbol_id(symbol))
    return(local_slot(symbol));

  for (x = sc->envir; symbol_id(symbol) < let_id(x); x = outlet(x));

  if (let_id(x) == symbol_id(symbol))
    return(local_slot(symbol));

  for (; is_let(x); x = outlet(x))
    {
      s7_pointer y;
      for (y = let_slots(x); is_slot(y); y = next_slot(y))
	if (slot_symbol(y) == symbol)
	  return(y);
    }

  return(global_slot(symbol));
}

#if WITH_GCC && DEBUGGING
static s7_pointer find_symbol_unchecked_1(s7_scheme *sc, s7_pointer symbol)
#else
static inline s7_pointer find_symbol_unchecked(s7_scheme *sc, s7_pointer symbol) /* find_symbol_checked includes the unbound_variable call */
#endif
{
  s7_pointer x;

  if (let_id(sc->envir) == symbol_id(symbol))
    return(slot_value(local_slot(symbol)));

  for (x = sc->envir; symbol_id(symbol) < let_id(x); x = outlet(x));

  /* this looks redundant, but every attempt to improve it is much slower! */
  if (let_id(x) == symbol_id(symbol))
    return(slot_value(local_slot(symbol)));

  for (; is_let(x); x = outlet(x))
    {
      s7_pointer y;
      for (y = let_slots(x); is_slot(y); y = next_slot(y))
	if (slot_symbol(y) == symbol)
	  return(slot_value(y));
    }

  /* for a global, the loop above is not hit.  If a global is used locally (as a function parameter name for example),
   *   we run the entire search loop -- 10 times slower!  So don't use a local variable or parameter named 'car!
   */
  x = global_slot(symbol);
  if (is_slot(x))
    return(slot_value(x));

#if WITH_GCC
  return(NULL); /* much faster than various alternatives */
#else
  return(unbound_variable(sc, symbol));
#endif
}


s7_pointer s7_slot(s7_scheme *sc, s7_pointer symbol)
{
  return(find_symbol(sc, symbol));
}


s7_pointer s7_slot_value(s7_pointer slot)
{
  return(slot_value(slot));
}


s7_pointer s7_slot_set_value(s7_scheme *sc, s7_pointer slot, s7_pointer value)
{
  slot_set_value(slot, value);
  return(value);
}


void s7_slot_set_real_value(s7_scheme *sc, s7_pointer slot, s7_double value)
{
  set_real(slot_value(slot), value);
}


static s7_pointer find_local_symbol(s7_scheme *sc, s7_pointer symbol, s7_pointer e)
{
  if (!is_let(e))
    return(global_slot(symbol));

  if (symbol_id(symbol) != 0)
    {
      s7_pointer y;
      for (y = let_slots(e); is_slot(y); y = next_slot(y))
	if (slot_symbol(y) == symbol)
	  return(y);
    }
  return(sc->undefined);
}


s7_pointer s7_symbol_value(s7_scheme *sc, s7_pointer sym)
{
  s7_pointer x;

  x = find_symbol(sc, sym);
  if (is_slot(x))
    return(slot_value(x));

  return(sc->undefined);
}


s7_pointer s7_symbol_local_value(s7_scheme *sc, s7_pointer sym, s7_pointer local_env)
{
  if ((local_env == sc->rootlet) || (is_global(sym)))
    {
      if (is_slot(global_slot(sym)))
	return(slot_value(global_slot(sym)));
      return(sc->undefined);
    }

  if (is_let(local_env))
    {
      s7_pointer x;

      for (x = local_env; is_let(x); x = outlet(x))
	{
	  s7_pointer y;

	  for (y = let_slots(x); is_slot(y); y = next_slot(y))
	    if (slot_symbol(y) == sym)
	      return(slot_value(y));
	}
    }
  return(s7_symbol_value(sc, sym));
}


/* -------------------------------- symbol->value -------------------------------- */

#define find_global_symbol_checked(Sc, Sym) ((is_global(Sym)) ? slot_value(global_slot(Sym)) : find_symbol_checked(Sc, Sym))

static s7_pointer g_symbol_to_value(s7_scheme *sc, s7_pointer args)
{
  #define H_symbol_to_value "(symbol->value sym (env (curlet))) returns the binding of (the value associated with) the \
symbol sym in the given environment: (let ((x 32)) (symbol->value 'x)) -> 32"
  #define Q_symbol_to_value s7_make_signature(sc, 3, sc->T, sc->is_symbol_symbol, sc->is_let_symbol)
  /* (symbol->value 'x e) => (e 'x)? */

  s7_pointer sym;
  sym = car(args);

  if (!is_symbol(sym))
    method_or_bust(sc, sym, sc->symbol_to_value_symbol, args, T_SYMBOL, 1);

  if (is_not_null(cdr(args)))
    {
      s7_pointer local_env;

      local_env = cadr(args);
      if (local_env == sc->unlet_symbol)
	return((is_slot(initial_slot(sym))) ? slot_value(initial_slot(sym)) : sc->undefined);

      if (!is_let(local_env))
	method_or_bust_with_type(sc, local_env, sc->symbol_to_value_symbol, args, a_let_string, 2);

      return(s7_symbol_local_value(sc, sym, local_env));
    }

  if (is_global(sym))
    return(slot_value(global_slot(sym)));

  return(s7_symbol_value(sc, sym));
}


s7_pointer s7_symbol_set_value(s7_scheme *sc, s7_pointer sym, s7_pointer val)
{
  s7_pointer x;
  /* if immutable should this return an error? */
  x = find_symbol(sc, sym);
  if (is_slot(x))
    slot_set_value(x, val); /* with_hook? */
  return(val);
}


/* -------------------------------- symbol->dynamic-value -------------------------------- */

static s7_pointer find_dynamic_value(s7_scheme *sc, s7_pointer x, s7_pointer sym, int64_t *id)
{
  for (; symbol_id(sym) < let_id(x); x = outlet(x));

  if (let_id(x) == symbol_id(sym))
    {
      (*id) = let_id(x);
      return(slot_value(local_slot(sym)));
    }
  for (; (is_let(x)) && (let_id(x) > (*id)); x = outlet(x))
    {
      s7_pointer y;
      for (y = let_slots(x); is_slot(y); y = next_slot(y))
	if (slot_symbol(y) == sym)
	  {
	    (*id) = let_id(x);
	    return(slot_value(y));
	  }
    }
  return(sc->gc_nil);
}


static s7_pointer g_symbol_to_dynamic_value(s7_scheme *sc, s7_pointer args)
{
  #define H_symbol_to_dynamic_value "(symbol->dynamic-value sym) returns the dynamic binding of the symbol sym"
  #define Q_symbol_to_dynamic_value s7_make_signature(sc, 2, sc->T, sc->is_symbol_symbol)

  s7_pointer sym, val;
  int64_t i, top_id;

  sym = car(args);
  if (!is_symbol(sym))
    method_or_bust(sc, sym, sc->symbol_to_dynamic_value_symbol, args, T_SYMBOL, 1);

  if (is_global(sym))
    return(slot_value(global_slot(sym)));

  if (let_id(sc->envir) == symbol_id(sym))
    return(slot_value(local_slot(sym)));

  top_id = -1;
  val = find_dynamic_value(sc, sc->envir, sym, &top_id);
  if (top_id == symbol_id(sym))
    return(val);

  for (i = s7_stack_top(sc) - 1; i > 0; i -= 4)
    {
      s7_pointer cur_val;
      cur_val = find_dynamic_value(sc, stack_let(sc->stack, i), sym, &top_id);
      if (cur_val != sc->gc_nil)
	val = cur_val;
      if (top_id == symbol_id(sym))
	return(val);
    }

  if (val == sc->gc_nil)
    return(s7_symbol_value(sc, sym));
  return(val);
}


typedef bool (safe_sym_t)(s7_scheme *sc, s7_pointer sym, s7_pointer e);
static s7_function all_x_eval(s7_scheme *sc, s7_pointer holder, s7_pointer e, safe_sym_t *checker);

static bool direct_memq(s7_pointer symbol, s7_pointer symbols)
{
  s7_pointer x;
  for (x = symbols; is_pair(x); x = cdr(x))
    if (car(x) == symbol)
	return(true);
  return(false);
}

static bool direct_assq(s7_pointer symbol, s7_pointer symbols)
{ /* used only below in do_symbol_is_safe */
  s7_pointer x;
  for (x = symbols; is_pair(x); x = cdr(x))
    if (caar(x) == symbol)
      return(true);
  return(false);
}

static bool do_symbol_is_safe(s7_scheme *sc, s7_pointer sym, s7_pointer e)
{
  return((is_slot(global_slot(sym))) || 
	 (direct_assq(sym, e)) ||
	 (is_slot(find_symbol(sc, sym))));
}

static bool let_symbol_is_safe(s7_scheme *sc, s7_pointer sym, s7_pointer e)
{
  return((is_slot(global_slot(sym))) || 
	 ((!is_with_let_let(e)) && (is_slot(find_symbol(sc, sym)))));
}

static bool let_star_symbol_is_safe(s7_scheme *sc, s7_pointer sym, s7_pointer e)
{
  return((symbol_is_in_list(sc, sym)) || 
	 (is_slot(global_slot(sym))) || 
	 ((is_let(e)) && (!is_with_let_let(e)) && (is_slot(find_symbol(sc, sym)))));
}

static bool pair_symbol_is_safe(s7_scheme *sc, s7_pointer sym, s7_pointer e)
{
  return((is_slot(global_slot(sym))) || 
	 (direct_memq(sym, e)));
}


static inline s7_pointer collect_variables(s7_scheme *sc, s7_pointer lst, s7_pointer e)
{
  /* collect local variable names from let/do (pre-error-check) */
  s7_pointer p;
  sc->w = e;
  for (p = lst; is_pair(p); p = cdr(p))
    sc->w = cons(sc, add_symbol_to_list(sc, caar(p)), sc->w);
  return(sc->w);
}

static s7_pointer collect_parameters(s7_scheme *sc, s7_pointer lst, s7_pointer e)
{
  /* collect local variable names from lambda arglists (pre-error-check) */
  s7_pointer p;
  if (is_symbol(lst))
    return(cons(sc, add_symbol_to_list(sc, lst), e));
  sc->w = e;
  for (p = lst; is_pair(p); p = cdr(p))
    {
      s7_pointer car_p;
      car_p = car(p);
      if (is_pair(car_p))
	car_p = car(car_p);
      if ((is_symbol(car_p)) &&
	  (!is_keyword(car_p)))
	sc->w = cons(sc, add_symbol_to_list(sc, car_p), sc->w);
    }
  if (is_symbol(p)) /* rest arg */
    sc->w = cons(sc, add_symbol_to_list(sc, p), sc->w);
  return(sc->w);
}


/* make macros and closures */

typedef enum {OPT_F, OPT_T, OPT_OOPS} opt_t;
static opt_t optimize(s7_scheme *sc, s7_pointer code, int32_t hop, s7_pointer e);

static void clear_all_optimizations(s7_scheme *sc, s7_pointer p)
{
  /* I believe that we would not have been optimized to begin with if the tree were circular,
   *   and this tree is supposed to be a function call + args -- a circular list here is a bug.
   */
  if (is_pair(p))
    {
      if ((is_optimized(p)) &&
	  ((optimize_op(p) & 1) == 0)) /* protect possibly shared code?  Elsewhere we assume these aren't changed */
	{
	  clear_optimized(p);
	  clear_optimize_op(p);
	}
      clear_all_optimizations(sc, cdr(p));
      clear_all_optimizations(sc, car(p));
    }
}


static s7_pointer make_macro(s7_scheme *sc, opcode_t op)
{
  s7_pointer cx, mac;
  uint64_t typ;

  if (op == OP_DEFINE_MACRO)
    typ = T_MACRO | T_DONT_EVAL_ARGS | T_COPY_ARGS;
  else
    {
      if (op == OP_DEFINE_MACRO_STAR)
	typ = T_MACRO_STAR | T_DONT_EVAL_ARGS | T_COPY_ARGS;
      else
	{
	  if (op == OP_DEFINE_BACRO)
	    typ = T_BACRO | T_DONT_EVAL_ARGS | T_COPY_ARGS;
	  else
	    {
	      if (op == OP_DEFINE_BACRO_STAR)
		typ = T_BACRO_STAR | T_DONT_EVAL_ARGS | T_COPY_ARGS;
	      else
		{
		  if ((op == OP_DEFINE_EXPANSION) &&
		      (!is_let(sc->envir)))        /* local expansions are just normal macros */
		    typ = T_MACRO | T_EXPANSION | T_DONT_EVAL_ARGS | T_COPY_ARGS;
		  else typ = T_MACRO | T_DONT_EVAL_ARGS | T_COPY_ARGS;
		}
	    }
	}
    }

  new_cell_no_check(sc, mac, typ);
  sc->temp6 = mac;
  closure_set_args(mac, cdar(sc->code));
  closure_set_body(mac, cdr(sc->code));
  closure_set_setter(mac, sc->F);
  closure_set_let(mac, sc->envir);
  closure_arity(mac) = CLOSURE_ARITY_NOT_SET;

  sc->capture_let_counter++;
  sc->code = caar(sc->code);
  if ((op == OP_DEFINE_EXPANSION) &&
      (!is_let(sc->envir)))
    set_type(sc->code, T_EXPANSION | T_SYMBOL); /* see comment under READ_TOK */
  /* symbol? macro name has already been checked, find name in environment, and define it */
  cx = find_local_symbol(sc, sc->code, sc->envir);
  if (is_slot(cx))
    slot_set_value_with_hook(cx, mac);
  else s7_make_slot(sc, sc->envir, sc->code, mac); /* was current but we've checked immutable already */

  clear_symbol_list(sc); /* tracks names local to this macro */
  if (optimize(sc, closure_body(mac), 1, collect_parameters(sc, closure_args(mac), sc->nil)) == OPT_OOPS)
    clear_all_optimizations(sc, closure_body(mac));

  sc->temp6 = sc->nil;
  return(mac);
}


static s7_pointer make_closure(s7_scheme *sc, s7_pointer args, s7_pointer code, uint64_t type, int32_t arity)
{
  /* this is called every time a lambda form is evaluated, or during letrec, etc */
  s7_pointer x;
  new_cell(sc, x, type);
  closure_set_args(x, args);
  closure_set_body(x, code);
  closure_set_setter(x, sc->F);
  closure_arity(x) = arity;
  closure_set_let(x, sc->envir);
  sc->capture_let_counter++;
  return(x);
}

#define make_closure_with_let(Sc, X, Args, Code, Env, Arity)	\
  do {							\
    new_cell(Sc, X, T_CLOSURE | T_COPY_ARGS | ((is_safe_closure(Code)) ? T_SAFE_CLOSURE : 0)); \
    closure_set_args(X, Args);						\
    closure_set_body(X, Code);				                \
    closure_set_setter(X, sc->F);					\
    closure_arity(X) = Arity;						\
    closure_set_let(X, Env);						\
    sc->capture_let_counter++;						\
  } while (0)

static int32_t closure_length(s7_scheme *sc, s7_pointer e)
{
  /* we can't use let_length(sc, closure_let(e)) because the closure_let(closure)
   *   changes.  So the open bit is not always on.  Besides, the fallbacks need to be for closures, not environments.
   */
  s7_pointer length_func;
  length_func = find_method(sc, closure_let(e), sc->length_symbol);
  if (length_func != sc->undefined)
    return((int32_t)s7_integer(s7_apply_function(sc, length_func, list_1(sc, e))));

  /* there are cases where this should raise a wrong-type-arg error, but for now... */
  return(-1);
}

#define check_closure_for(Sc, Fnc, Sym)				    \
  if ((has_closure_let(Fnc)) && (is_let(closure_let(Fnc))))	    \
    {								    \
      s7_pointer val;						    \
      val = find_local_symbol(Sc, Sym, closure_let(Fnc));	    \
      if ((!is_slot(val)) && (is_let(outlet(closure_let(Fnc)))))    \
	val = find_local_symbol(Sc, Sym, outlet(closure_let(Fnc))); \
      if (is_slot(val))						    \
	return(slot_value(val));				    \
    }

static s7_pointer copy_tree(s7_scheme *sc, s7_pointer tree)
{
#if WITH_GCC
  #define COPY_TREE(P) ({s7_pointer _p; _p = P; cons_unchecked(sc, (is_pair(car(_p))) ? copy_tree(sc, car(_p)) : car(_p), (is_pair(cdr(_p))) ? copy_tree(sc, cdr(_p)) : cdr(_p));})
#else
  #define COPY_TREE(P) copy_tree(sc, P)
#endif

  return(cons_unchecked(sc, 
			(is_pair(car(tree))) ? COPY_TREE(car(tree)) : car(tree),
			(is_pair(cdr(tree))) ? COPY_TREE(cdr(tree)) : cdr(tree)));
}

static inline void annotate_expansion(s7_pointer p)
{
  if ((is_symbol(car(p))) &&
      (is_pair(cdr(p))))
    {
      set_opt_back(p);
      set_overlay(cdr(p));
    }
  else
    {
      if (is_pair(car(p)))
	annotate_expansion(car(p));
    }
  for (p = cdr(p); is_pair(p); p = cdr(p))
    if (is_pair(car(p)))
      annotate_expansion(car(p));
}

static s7_pointer copy_body(s7_scheme *sc, s7_pointer p)
{
  /* ideally we'd use tree_len here, but it currently does not protect against cycles */
  if (8192 >= (sc->free_heap_top - sc->free_heap))
    {
      gc(sc);
      while (8192 >= (sc->free_heap_top - sc->free_heap))
	resize_heap(sc);
    }
  sc->w = copy_tree(sc, p);
  annotate_expansion(sc->w);
  p = sc->w;
  sc->w = sc->nil;
  return(p);
}

static s7_pointer copy_closure(s7_scheme *sc, s7_pointer fnc)
{
  /* copy the source tree annotating (for eventual optimization), return a thing of the same type as fnc */
  s7_pointer x, body;

  body = copy_body(sc, closure_body(fnc));
  new_cell(sc, x, typeflag(fnc));
  closure_set_args(x, closure_args(fnc));
  closure_set_body(x, body);
  closure_set_setter(x, closure_setter(fnc));
  closure_arity(x) = closure_arity(fnc);
  closure_set_let(x, closure_let(fnc));
  return(x);
}

/* -------------------------------- defined? -------------------------------- */
static s7_pointer g_is_defined(s7_scheme *sc, s7_pointer args)
{
  #define H_is_defined "(defined? obj (env (curlet)) ignore-globals) returns #t if obj has a binding (a value) in the environment env"
  #define Q_is_defined s7_make_signature(sc, 4, sc->is_boolean_symbol, sc->is_symbol_symbol, sc->is_let_symbol, sc->is_boolean_symbol)

  s7_pointer sym;

  /* is this correct?
   *    (defined? '_x) #f (symbol->value '_x) #<undefined>
   *    (define x #<undefined>) (defined? 'x) #t
   * can't return the value here because it might be #f
   */

  sym = car(args);
  if (!is_symbol(sym))
    method_or_bust(sc, sym, sc->is_defined_symbol, args, T_SYMBOL, 1);

  if (is_pair(cdr(args)))
    {
      s7_pointer e, b, x;
      e = cadr(args);
      if (!is_let(e))
	return(wrong_type_argument_with_type(sc, sc->is_defined_symbol, 2, e, a_let_string));

      if (is_pair(cddr(args)))
	{
	  b = caddr(args);
	  if (!s7_is_boolean(b))
	    method_or_bust_with_type(sc, b, sc->is_defined_symbol, args, a_boolean_string, 3);
	}
      else b = sc->F;

      if (e == sc->rootlet) /* we checked (let? e) above */
	{
	  if (b == sc->F)
	    return(make_boolean(sc, is_slot(global_slot(sym)))); /* new_symbol and gensym initialize global_slot to #<undefined> */
	  return(sc->F);
	}

      x = find_local_symbol(sc, sym, e);
      if (is_slot(x))
	return(sc->T);

      if (b == sc->T)
	return(sc->F);

      /* here we can't fall back on find_symbol:
       *    (let ((b 2))
       *      (let ((e (curlet)))
       *        (let ((a 1))
       *          (if (defined? 'a e)
       *              (format #t "a: ~A in ~{~A ~}" (symbol->value 'a e) e))))
       *    "a: 1 in (b . 2)"
       *
       * but we also can't just return #f:
       *    (let ((b 2))
       *      (let ((e (curlet)))
       *        (let ((a 1))
       *          (format #t "~A: ~A" (defined? 'abs e) (eval '(abs -1) e)))))
       *    "#f: 1"
       */
      return(make_boolean(sc, is_slot(global_slot(sym))));
    }
  else
    {
      if (is_global(sym))
	return(sc->T);
    }
  return(make_boolean(sc, is_slot(find_symbol(sc, sym))));
}


bool s7_is_defined(s7_scheme *sc, const char *name)
{
  s7_pointer x;
  x = s7_symbol_table_find_name(sc, name);
  if (x)
    {
      x = find_symbol(sc, x);
      return(is_slot(x));
    }
  return(false);
}

static bool is_defined_b_p(s7_pointer p)
{
  if (!is_symbol(p))
    simple_wrong_type_argument(cur_sc, cur_sc->is_defined_symbol, p, T_SYMBOL);
  return(is_slot(find_symbol(cur_sc, p)));
}

static bool is_defined_b_pp(s7_pointer p, s7_pointer e) {return(g_is_defined(cur_sc, set_plist_2(cur_sc, p, e)) != cur_sc->F);}



void s7_define(s7_scheme *sc, s7_pointer envir, s7_pointer symbol, s7_pointer value)
{
  s7_pointer x;
  if ((envir == sc->nil) ||
      (envir == sc->rootlet))
    envir = sc->shadow_rootlet;
  x = find_local_symbol(sc, symbol, envir);
  if (is_slot(x))
    slot_set_value_with_hook(x, value);
  else
    {
      s7_make_slot(sc, envir, symbol, value); /* I think this means C code can override "constant" defs */
      if ((envir == sc->shadow_rootlet) &&
	  (!is_slot(global_slot(symbol))))
	{
	  set_global(symbol); /* is_global => global_slot is usable */
	  set_global_slot(symbol, local_slot(symbol));
	}
    }
}


s7_pointer s7_define_variable(s7_scheme *sc, const char *name, s7_pointer value)
{
  s7_pointer sym;
  sym = make_symbol(sc, name);
  s7_define(sc, sc->nil, sym, value);
  return(sym);
}


s7_pointer s7_define_variable_with_documentation(s7_scheme *sc, const char *name, s7_pointer value, const char *help)
{
  s7_pointer sym;
  sym = s7_define_variable(sc, name, value);
  symbol_set_has_help(sym);
  symbol_set_help(sym, copy_string(help));
  return(sym);
}


s7_pointer s7_define_constant(s7_scheme *sc, const char *name, s7_pointer value)
{
  s7_pointer sym;
  sym = make_symbol(sc, name);
  s7_define(sc, sc->nil, sym, value);
  set_immutable(sym);
  return(sym);
}

/* (define (func a) (let ((cvar (+ a 1))) cvar)) (define-constant cvar 23) (func 1) -> ;can't bind an immutable object: cvar
 * (let ((aaa 1)) (define-constant aaa 32) (set! aaa 3)) -> set!: can't alter immutable object: aaa
 */

s7_pointer s7_define_constant_with_documentation(s7_scheme *sc, const char *name, s7_pointer value, const char *help)
{
  s7_pointer sym;
  sym = s7_define_constant(sc, name, value);
  symbol_set_has_help(sym);
  symbol_set_help(sym, copy_string(help));
  return(value); /* inconsistent with variable above, but consistent with define_function? */
}


char *s7_symbol_documentation(s7_scheme *sc, s7_pointer sym)
{
  if (is_keyword(sym)) return(NULL);
  if ((is_symbol(sym)) &&
      (symbol_has_help(sym)))
    return(symbol_help(sym));
  return(NULL);
}


char *s7_symbol_set_documentation(s7_scheme *sc, s7_pointer sym, const char *new_doc)
{
  if (is_keyword(sym)) return(NULL);
  if ((is_symbol(sym)) &&
      (symbol_has_help(sym)) &&
      (symbol_help(sym)))
    free(symbol_help(sym));
  symbol_set_has_help(sym);
  symbol_set_help(sym, copy_string(new_doc));
  return(symbol_help(sym));
}


/* -------------------------------- keyword? -------------------------------- */

bool s7_is_keyword(s7_pointer obj)
{
  return(is_keyword(obj));
}


static s7_pointer g_is_keyword(s7_scheme *sc, s7_pointer args)
{
  #define H_is_keyword "(keyword? obj) returns #t if obj is a keyword, (keyword? :rest) -> #t"
  #define Q_is_keyword pl_bt
  check_boolean_method(sc, is_keyword, sc->is_keyword_symbol, args);
}


/* -------------------------------- string->keyword -------------------------------- */
s7_pointer s7_make_keyword(s7_scheme *sc, const char *key)
{
  s7_pointer sym;
  char *name;
  uint32_t slen;
  slen = safe_strlen(key);
  tmpbuf_malloc(name, slen + 2);
  name[0] = ':';                                     /* prepend ":" */
  name[1] = '\0';
  memcpy((void *)(name + 1), (void *)key, slen);
  sym = make_symbol_with_length(sc, name, slen + 1); /* keyword slot etc taken care of here (in new_symbol actually) */
  tmpbuf_free(name, slen + 2);
  return(sym);
}


static s7_pointer g_string_to_keyword(s7_scheme *sc, s7_pointer args)
{
  #define H_string_to_keyword "(string->keyword str) prepends ':' to str and defines that as a keyword"
  #define Q_string_to_keyword s7_make_signature(sc, 2, sc->is_keyword_symbol, sc->is_string_symbol)

  if (!is_string(car(args)))
    method_or_bust_one_arg(sc, car(args), sc->string_to_keyword_symbol, args, T_STRING);
  return(s7_make_keyword(sc, string_value(car(args))));
}

/* -------------------------------- keyword->symbol -------------------------------- */
static s7_pointer g_keyword_to_symbol(s7_scheme *sc, s7_pointer args)
{
  #define H_keyword_to_symbol "(keyword->symbol key) returns a symbol with the same name as key but no prepended colon"
  #define Q_keyword_to_symbol s7_make_signature(sc, 2, sc->is_symbol_symbol, sc->is_keyword_symbol)

  s7_pointer sym;
  sym = car(args);
  if (!is_keyword(sym))
    method_or_bust_with_type_one_arg(sc, sym, sc->keyword_to_symbol_symbol, args, s7_make_string_wrapper(sc, "a keyword"));
  return(keyword_symbol(sym));
}


/* -------------------------------- symbol->keyword -------------------------------- */
static s7_pointer symbol_to_keyword(s7_scheme *sc, s7_pointer sym)
{
  return(s7_make_keyword(sc, symbol_name(sym)));
}

static s7_pointer g_symbol_to_keyword(s7_scheme *sc, s7_pointer args)
{
  #define H_symbol_to_keyword "(symbol->keyword sym) returns a keyword with the same name as sym, but with a colon prepended"
  #define Q_symbol_to_keyword s7_make_signature(sc, 2, sc->is_keyword_symbol, sc->is_symbol_symbol)

  if (!is_symbol(car(args)))
    method_or_bust_one_arg(sc, car(args), sc->symbol_to_keyword_symbol, args, T_SYMBOL);
  return(symbol_to_keyword(sc, car(args)));
}


/* ---------------- uninterpreted pointers ---------------- */

bool s7_is_c_pointer(s7_pointer arg)
{
  return(is_c_pointer(arg));
}

bool s7_is_c_pointer_of_type(s7_pointer arg, s7_pointer type)
{
  return((is_c_pointer(arg)) && (raw_pointer_type(arg) == type));
}

void *s7_c_pointer(s7_pointer p)
{
  if ((is_number(p)) &&
      (s7_integer(p) == 0))
    return(NULL); /* special case where the null pointer has been cons'd up by hand */

  if (!is_c_pointer(p))
    return(NULL);

  return(raw_pointer(p));
}

s7_pointer s7_c_pointer_type(s7_pointer p)
{
  if (!is_c_pointer(p))
    return(NULL); /* as above */
  return(raw_pointer_type(p));
}

s7_pointer s7_make_c_pointer_with_type(s7_scheme *sc, void *ptr, s7_pointer type, s7_pointer info)
{
  s7_pointer x;
  new_cell(sc, x, T_C_POINTER);
  raw_pointer(x) = ptr;
  raw_pointer_type(x) = type;
  raw_pointer_info(x) = info;
  return(x);
}

s7_pointer s7_make_c_pointer(s7_scheme *sc, void *ptr) {return(s7_make_c_pointer_with_type(sc, ptr, sc->F, sc->F));}

static s7_pointer g_is_c_pointer(s7_scheme *sc, s7_pointer args)
{
  #define H_is_c_pointer "(c-pointer? obj type) returns #t if obj is a C pointer being held in s7.  If type is given, the c_pointer's type is also checked."
  #define Q_is_c_pointer s7_make_signature(sc, 3, sc->is_boolean_symbol, sc->T, sc->T)

  s7_pointer p;
  p = car(args);
  if (!is_c_pointer(p))
    {
      if (!has_methods(p)) return(sc->F);
      return(apply_boolean_method(sc, p, sc->is_c_pointer_symbol));
    }
  if (is_pair(cdr(args)))
    return(make_boolean(sc, raw_pointer_type(p) == cadr(args)));
  return(sc->T);
}

static s7_pointer g_c_pointer(s7_scheme *sc, s7_pointer args)
{
  #define H_c_pointer "(c-pointer int type info) returns a c-pointer object. The type and info args are optional, defaulting to #f."
  #define Q_c_pointer s7_make_signature(sc, 4, sc->is_c_pointer_symbol, sc->is_integer_symbol, sc->T, sc->T)

  s7_pointer arg, type, info;
  intptr_t p;

  arg = car(args);
  if (!s7_is_integer(arg))
    method_or_bust(sc, arg, sc->c_pointer_symbol, list_1(sc, arg), T_INTEGER, 1);
  p = (intptr_t)s7_integer(arg);             /* (c-pointer (bignum "1234")) */
  info = sc->F;
  if (is_pair(cdr(args)))
    {
      type = cadr(args);
      if (is_pair(cddr(args)))
	info = caddr(args);
    }
  else type = sc->F;
  return(s7_make_c_pointer_with_type(sc, (void *)p, type, info));
}


/* -------------------------------- continuations and gotos -------------------------------- */

enum {NO_JUMP, CALL_WITH_EXIT_JUMP, THROW_JUMP, CATCH_JUMP, ERROR_JUMP, ERROR_QUIT_JUMP};
enum {NO_SET_JUMP, READ_SET_JUMP, LOAD_SET_JUMP, DYNAMIC_WIND_SET_JUMP, S7_CALL_SET_JUMP, EVAL_SET_JUMP};

static s7_pointer g_is_continuation(s7_scheme *sc, s7_pointer args)
{
  #define H_is_continuation "(continuation? obj) returns #t if obj is a continuation"
  #define Q_is_continuation pl_bt

  check_boolean_method(sc, is_continuation, sc->is_continuation_symbol, args);
  /* is this the right thing?  It returns #f for call-with-exit ("goto") because
   *   that form of continuation can't continue (via a jump back to its context).
   * how to recognize the call-with-exit function?  "goto" is an internal name.
   */
}

static bool s7_is_continuation(s7_pointer p) {return(is_continuation(p));}


static s7_pointer protected_list_copy(s7_scheme *sc, s7_pointer a)
{
  s7_pointer slow, fast, p;

  sc->w = cons(sc, car(a), sc->nil);
  p = sc->w;

  slow = fast = cdr(a);
  while (true)
    {
      if (!is_pair(fast))
	{
	  if (is_null(fast))
	    return(sc->w);
	  set_cdr(p, fast);
	  return(sc->w);
	}

      set_cdr(p, cons(sc, car(fast), sc->nil));
      p = cdr(p);

      fast = cdr(fast);
      if (!is_pair(fast))
	{
	  if (is_null(fast))
	    return(sc->w);
	  set_cdr(p, fast);
	  return(sc->w);
	}
      /* if unrolled further, it's a lot slower? */
      set_cdr(p, cons(sc, car(fast), sc->nil));
      p = cdr(p);

      fast = cdr(fast);
      slow = cdr(slow);
      if (fast == slow)
	{
	  /* try to preserve the original cyclic structure */
	  s7_pointer p1, f1, p2, f2;
	  set_match_pair(a);
	  for (p1 = sc->w, f1 = a; !(is_matched_pair(cdr(f1))); f1 = cdr(f1), p1 = cdr(p1))
	    set_match_pair(f1);
	  for (p2 = sc->w, f2 = a; cdr(f1) != f2; f2 = cdr(f2), p2 = cdr(p2))
	    clear_match_pair(f2);
	  for (f1 = f2; is_pair(f1); f1 = cdr(f1), f2 = cdr(f2))
	    {
	      clear_match_pair(f1);
	      f1 = cdr(f1);
	      clear_match_pair(f1);
	      if (f1 == f2) break;
	    }
	  if (is_null(p1))
	    set_cdr(p2, p2);
	  else set_cdr(p1, p2);
	  return(sc->w);
	}
    }
  return(sc->w);
}


static s7_pointer copy_counter(s7_scheme *sc, s7_pointer obj)
{
  s7_pointer nobj;
  new_cell(sc, nobj, T_COUNTER);
  counter_set_result(nobj, counter_result(obj));
  counter_set_list(nobj, counter_list(obj));
  counter_set_capture(nobj, counter_capture(obj));
  counter_set_let(nobj, counter_let(obj));
  counter_set_slots(nobj, counter_slots(obj));
  return(nobj);
}


static s7_pointer copy_stack(s7_scheme *sc, s7_pointer old_v, int64_t top)
{
  #define CC_INITIAL_STACK_SIZE 256 /* 128 is too small here */
  int64_t i, len;
  s7_pointer new_v;
  s7_pointer *nv, *ov;

  /* stacks can grow temporarily, so sc->stack_size grows, but we don't normally need all that
   *   leftover space here, so choose the original stack size if it's smaller.
   */
  len = vector_length(old_v);
  if (len > CC_INITIAL_STACK_SIZE)
    {
      if (top < CC_INITIAL_STACK_SIZE / 4)
	len = CC_INITIAL_STACK_SIZE;
    }
  else
    {
      if (len < CC_INITIAL_STACK_SIZE)
	len = CC_INITIAL_STACK_SIZE;
    }
  if ((int64_t)(sc->free_heap_top - sc->free_heap) < (int64_t)(sc->heap_size / 4)) gc(sc);
  /* this gc call is needed if there are lots of call/cc's -- by pure bad luck
   *   we can end up hitting the end of the gc free list time after time while
   *   in successive copy_stack's below, causing s7 to core up until it runs out of memory.
   */

  new_v = make_vector_1(sc, len, NOT_FILLED, T_VECTOR);
  set_type(new_v, T_STACK);
  temp_stack_top(new_v) = top;
  nv = vector_elements(new_v);
  ov = vector_elements(old_v);
  if (len > 0)
    memcpy((void *)nv, (void *)ov, len * sizeof(s7_pointer));

  s7_gc_on(sc, false);
  for (i = 2; i < top; i += 4)
    {
      s7_pointer p;
      p = ov[i];                            /* args */
      if (is_pair(p))                       /* args need not be a list (it can be a port or #f, etc) */
	nv[i] = protected_list_copy(sc, p); /* args (copy is needed -- see s7test.scm) */
      /* lst can be dotted or circular here.  The circular list only happens in a case like:
       *    (dynamic-wind (lambda () (eq? (let ((lst (cons 1 2))) (set-cdr! lst lst) lst) (call/cc (lambda (k) k)))) (lambda () #f) (lambda () #f))
       */
      else
	{
	  if (is_counter(p))              /* these can only occur in this context */
	    nv[i] = copy_counter(sc, p);
	}
    }
  s7_gc_on(sc, true);
  return(new_v);
}


static s7_pointer make_goto(s7_scheme *sc)
{
  s7_pointer x;
  new_cell(sc, x, T_GOTO);
  call_exit_goto_loc(x) = s7_stack_top(sc);
  call_exit_op_loc(x) = (int32_t)(sc->op_stack_now - sc->op_stack);
  call_exit_active(x) = true;
  return(x);
}


static s7_pointer *copy_op_stack(s7_scheme *sc)
{
  int32_t len;
  s7_pointer *ops;
  ops = (s7_pointer *)malloc(sc->op_stack_size * sizeof(s7_pointer));
  len = (int32_t)(sc->op_stack_now - sc->op_stack);
  if (len > 0)
    memcpy((void *)ops, (void *)(sc->op_stack), len * sizeof(s7_pointer));
  return(ops);
}


/* (with-baffle . body) calls body guaranteeing that there can be no jumps into the
 *    middle of it from outside -- no outer evaluation of a continuation can jump across this
 *    barrier:  The flip-side of call-with-exit.
 *    It sets a T_BAFFLE var in a new env, that has a unique key.  Call/cc then always
 *    checks the env chain for any such variable, saving the localmost.  Apply of a continuation
 *    looks for such a saved variable, if none, go ahead, else check the current env (before the
 *    jump) for that variable.  If none, error, else go ahead.  This is different from a delimited
 *    continuation which simply delimits the extent of the continuation (why not use lambda?) -- we want to block it
 *    from coming at us from some unknown place.
 */

static s7_pointer make_baffle(s7_scheme *sc)
{
  s7_pointer x;
  new_cell(sc, x, T_BAFFLE);
  baffle_key(x) = sc->baffle_ctr++;
  return(x);
}


static bool find_baffle(s7_scheme *sc, int32_t key)
{
  /* search backwards through sc->envir for sc->baffle_symbol with key as value
   */
  s7_pointer x, y;
  for (x = sc->envir; is_let(x); x = outlet(x))
    for (y = let_slots(x); is_slot(y); y = next_slot(y))
      if ((slot_symbol(y) == sc->baffle_symbol) &&
	  (baffle_key(slot_value(y)) == key))
	return(true);

  if ((is_slot(global_slot(sc->baffle_symbol))) &&
      (is_baffle(slot_value(global_slot(sc->baffle_symbol)))))
    return(baffle_key(slot_value(global_slot(sc->baffle_symbol))) == key);

  return(false);
}


static int32_t find_any_baffle(s7_scheme *sc)
{
  /* search backwards through sc->envir for any sc->baffle_symbol
   */
  if (sc->baffle_ctr > 0)
    {
      s7_pointer x, y;
      for (x = sc->envir; is_let(x); x = outlet(x))
	for (y = let_slots(x); is_slot(y); y = next_slot(y))
	  if (slot_symbol(y) == sc->baffle_symbol)
	    return(baffle_key(slot_value(y)));

      if ((is_slot(global_slot(sc->baffle_symbol))) &&
	  (is_baffle(slot_value(global_slot(sc->baffle_symbol)))))
	return(baffle_key(slot_value(global_slot(sc->baffle_symbol))));
    }
  return(-1);
}


s7_pointer s7_make_continuation(s7_scheme *sc)
{
  s7_pointer x, stack;
  int64_t loc;

  loc = s7_stack_top(sc);
  stack = copy_stack(sc, sc->stack, loc);
  sc->temp8 = stack;

  new_cell(sc, x, T_CONTINUATION);
  continuation_data(x) = (continuation_t *)malloc(sizeof(continuation_t));
  continuation_set_stack(x, stack);
  continuation_stack_size(x) = vector_length(continuation_stack(x));   /* copy_stack can return a smaller stack than the current one */
  continuation_stack_start(x) = vector_elements(continuation_stack(x));
  continuation_stack_end(x) = (s7_pointer *)(continuation_stack_start(x) + loc);
  continuation_op_stack(x) = copy_op_stack(sc);                        /* no heap allocation here */
  continuation_op_loc(x) = (int32_t)(sc->op_stack_now - sc->op_stack);
  continuation_op_size(x) = sc->op_stack_size;
  continuation_key(x) = find_any_baffle(sc);
  sc->temp8 = sc->nil;

  add_continuation(sc, x);
  return(x);
}

static void let_temp_done(s7_scheme *sc, s7_pointer args, s7_pointer code)
{
  push_stack(sc, OP_EVAL_DONE, sc->args, sc->code);
  sc->args = args;
  sc->code = code;
  eval(sc, OP_LET_TEMP_DONE);
}


static bool check_for_dynamic_winds(s7_scheme *sc, s7_pointer c)
{
  int64_t i, s_base = 0, c_base = -1;
  opcode_t op;

  for (i = s7_stack_top(sc) - 1; i > 0; i -= 4)
    {
      op = stack_op(sc->stack, i);
      switch (op)
	{
	case OP_DYNAMIC_WIND:
	case OP_LET_TEMP_DONE:
	  {
	    s7_pointer x;
	    int64_t j;
	    x = stack_code(sc->stack, i);
	    for (j = 3; j < continuation_stack_top(c); j += 4)
	      if (((stack_op(continuation_stack(c), j) == OP_DYNAMIC_WIND) || 
		   (stack_op(continuation_stack(c), j) == OP_LET_TEMP_DONE)) &&
		  (x == stack_code(continuation_stack(c), j)))
		{
		  s_base = i;
		  c_base = j;
		  break;
		}

	    if (s_base != 0)
	      break;

	    if (op == OP_DYNAMIC_WIND)
	      {
		if (dynamic_wind_state(x) == DWIND_BODY)
		  {
		    dynamic_wind_state(x) = DWIND_FINISH;
		    if (dynamic_wind_out(x) != sc->F)
		      {
			push_stack(sc, OP_EVAL_DONE, sc->args, sc->code);
			sc->args = sc->nil;
			sc->code = dynamic_wind_out(x);
			eval(sc, OP_APPLY);
		      }
		  }
	      }
	    else let_temp_done(sc, stack_args(sc->stack, i), stack_code(sc->stack, i));
	  }
	  break;

	case OP_BARRIER:
	  if (i > continuation_stack_top(c))  /* otherwise it's some unproblematic outer eval-string? */
	    return(false);                    /*    but what if we've already evaluated a dynamic-wind closer? */
	  break;

	case OP_DEACTIVATE_GOTO:              /* here we're jumping out of an unrelated call-with-exit block */
	  if (i > continuation_stack_top(c))
	    call_exit_active(stack_args(sc->stack, i)) = false;
	  break;

	default:
	  break;
	}
    }

  for (i = c_base + 4; i < continuation_stack_top(c); i += 4)
    {
      op = stack_op(continuation_stack(c), i);

      if (op == OP_DYNAMIC_WIND)
	{
	  s7_pointer x;
	  x = stack_code(continuation_stack(c), i);
	  if (dynamic_wind_in(x) != sc->F)
	    {
	      /* this can cause an infinite loop if the call/cc is trying to jump back into
	       *   a dynamic-wind init function -- it's even possible to trick with-baffle!
	       *   I can't find any fool-proof way to catch this problem.
	       */
	      push_stack(sc, OP_EVAL_DONE, sc->args, sc->code);
	      sc->args = sc->nil;
	      sc->code = dynamic_wind_in(x);
	      eval(sc, OP_APPLY);
	    }
	  dynamic_wind_state(x) = DWIND_BODY;
	}
      else
	{
	  if (op == OP_DEACTIVATE_GOTO)
	    call_exit_active(stack_args(continuation_stack(c), i)) = true;
	  else
	    {
	      if (op == OP_LET_TEMP_DONE)
		let_temp_done(sc, stack_args(continuation_stack(c), i), stack_code(continuation_stack(c), i));
	    }
	}
    }
  return(true);
}


static bool call_with_current_continuation(s7_scheme *sc)
{
  s7_pointer c;
  c = sc->code;

  /* check for (baffle ...) blocking the current attempt to continue */
  if ((continuation_key(c) >= 0) &&
      (!(find_baffle(sc, continuation_key(c))))) /* should this raise an error? */
    return(false);

  if (!check_for_dynamic_winds(sc, c)) /* if OP_BARRIER on stack deeper than continuation top(?), but can this happen? (it doesn't in s7test) */
    return(true);

  /* we push_stack sc->code before calling an embedded eval above, so sc->code should still be c here, etc
   */
  sc->stack = copy_stack(sc, continuation_stack(c), continuation_stack_top(c));
  sc->stack_size = continuation_stack_size(c);
  sc->stack_start = vector_elements(sc->stack);
  sc->stack_end = (s7_pointer *)(sc->stack_start + continuation_stack_top(c));
  sc->stack_resize_trigger = (s7_pointer *)(sc->stack_start + sc->stack_size / 2);

  {
    int32_t i, top;
    top = continuation_op_loc(c);
    sc->op_stack_now = (s7_pointer *)(sc->op_stack + top);
    sc->op_stack_size = continuation_op_size(c);
    sc->op_stack_end = (s7_pointer *)(sc->op_stack + sc->op_stack_size);
    for (i = 0; i < top; i++)
      sc->op_stack[i] = continuation_op_stack(c)[i];
  }

  if (is_null(sc->args))
    sc->value = sc->nil;
  else
    {
      if (is_null(cdr(sc->args)))
	sc->value = car(sc->args);
      else sc->value = splice_in_values(sc, sc->args);
    }
  return(true);
}


static void call_with_exit(s7_scheme *sc)
{
  int64_t i, new_stack_top, quit = 0;

  if (!call_exit_active(sc->code))
    {
      static s7_pointer call_with_exit_error = NULL;
      if (!call_with_exit_error)
	call_with_exit_error = s7_make_permanent_string("call-with-exit escape procedure called outside its block");
      s7_error(sc, sc->invalid_escape_function_symbol, set_elist_1(sc, call_with_exit_error));
    }

  call_exit_active(sc->code) = false;
  new_stack_top = call_exit_goto_loc(sc->code);
  sc->op_stack_now = (s7_pointer *)(sc->op_stack + call_exit_op_loc(sc->code));

  /* look for dynamic-wind in the stack section that we are jumping out of */
  for (i = s7_stack_top(sc) - 1; i > new_stack_top; i -= 4)
    {
      opcode_t op;

      op = stack_op(sc->stack, i);
      switch (op)
	{
	case OP_DYNAMIC_WIND:
	  {
	    s7_pointer lx;
	    lx = stack_code(sc->stack, i);
	    if (dynamic_wind_state(lx) == DWIND_BODY)
	      {
		dynamic_wind_state(lx) = DWIND_FINISH;
		if (dynamic_wind_out(lx) != sc->F)
		  {
		    push_stack(sc, OP_EVAL_DONE, sc->args, sc->code);
		    sc->args = sc->nil;
		    sc->code = dynamic_wind_out(lx);
		    eval(sc, OP_APPLY);
		  }
	      }
	  }
	  break;

	case OP_EVAL_STRING_2:
	  s7_close_input_port(sc, sc->input_port);
	  pop_input_port(sc);
	  break;

	case OP_BARRIER:                /* oops -- we almost certainly went too far */
	  return;

	case OP_DEACTIVATE_GOTO:        /* here we're jumping into an unrelated call-with-exit block */
	  call_exit_active(stack_args(sc->stack, i)) = false;
	  break;

	case OP_LET_TEMP_DONE:
	  let_temp_done(sc, stack_args(sc->stack, i), stack_code(sc->stack, i));
	  break;

	  /* call/cc does not close files, but I think call-with-exit should */
	case OP_GET_OUTPUT_STRING_1:
	case OP_UNWIND_OUTPUT:
	  {
	    s7_pointer x;
	    x = stack_code(sc->stack, i);                /* "code" = port that we opened */
	    s7_close_output_port(sc, x);
	    x = stack_args(sc->stack, i);                /* "args" = port that we shadowed, if not #f */
	    if (x != sc->F)
	      sc->output_port = x;
	  }
	  break;

	case OP_UNWIND_INPUT:
	  s7_close_input_port(sc, stack_code(sc->stack, i)); /* "code" = port that we opened */
	  sc->input_port = stack_args(sc->stack, i);         /* "args" = port that we shadowed */
	  break;

	case OP_EVAL_DONE: /* goto called in a method -- put off the inner eval return(s) until we clean up the stack */
	  quit++;
	  break;

	default:
	  break;
	}
    }

  sc->stack_end = (s7_pointer *)(sc->stack_start + new_stack_top);

  /* the return value should have an implicit values call, just as in call/cc */
  if (is_null(sc->args))
    sc->value = sc->nil;
  else
    {
      if (is_null(cdr(sc->args)))
	sc->value = car(sc->args);
      else sc->value = splice_in_values(sc, sc->args);
    }

  if (quit > 0)
    {
      if (sc->longjmp_ok)
	{
	  pop_stack(sc);
	  longjmp(sc->goto_start, CALL_WITH_EXIT_JUMP);
	}
      for (i = 0; i < quit; i++)
	push_stack_op_let(sc, OP_EVAL_DONE);
    }
}


static s7_pointer g_call_cc(s7_scheme *sc, s7_pointer args)
{
  #define H_call_cc "(call-with-current-continuation func) is always a mistake!"
  #define Q_call_cc s7_make_signature(sc, 2, sc->values_symbol, sc->is_procedure_symbol)
  /* I think the intent is that sc->values_symbol as the proc-sig return type indicates multiple values are possible (otherwise use #t). */

  s7_pointer p;
  p = car(args);                             /* this is the procedure passed to call/cc */
  if (!is_procedure(p))                      /* this includes continuations */
    {
      check_two_methods(sc, p, sc->call_cc_symbol, sc->call_with_current_continuation_symbol, args);
      return(simple_wrong_type_argument_with_type(sc, sc->call_cc_symbol, p, a_procedure_string));
    }
  if (!s7_is_aritable(sc, p, 1))
    return(s7_error(sc, sc->wrong_type_arg_symbol, set_elist_2(sc, s7_make_string_wrapper(sc, "call/cc procedure, ~A, should take one argument"), p)));

  sc->w = s7_make_continuation(sc);
  push_stack(sc, OP_APPLY, list_1(sc, sc->w), p);
  sc->w = sc->nil;

  return(sc->nil);
}

/* we can't naively optimize call/cc to call-with-exit if the continuation is only
 *   used as a function in the call/cc body because it might (for example) be wrapped
 *   in a lambda form that is being exported.  See b-func in s7test for an example.
 */


static s7_pointer g_call_with_exit(s7_scheme *sc, s7_pointer args)
{
  #define H_call_with_exit "(call-with-exit func) is call/cc without the ability to jump back into a previous computation."
  #define Q_call_with_exit s7_make_signature(sc, 2, sc->values_symbol, sc->is_procedure_symbol)

  s7_pointer p, x;
  /* (call-with-exit (lambda (return) ...)) */

  p = car(args);
  if (!is_procedure(p))                           /* this includes continuations */
    method_or_bust_with_type_one_arg(sc, p, sc->call_with_exit_symbol, args, a_procedure_string);

  x = make_goto(sc);
  push_stack(sc, OP_DEACTIVATE_GOTO, x, p); /* this means call-with-exit is not tail-recursive */
  push_stack(sc, OP_APPLY, cons_unchecked(sc, x, sc->nil), p);

  /* if the lambda body calls the argument as a function,
   *   it is applied to its arguments, apply notices that it is a goto, and...
   *
   *      (conceptually...) sc->stack_top = call_exit_goto_loc(sc->code);
   *      s_pop(sc, (is_not_null(sc->args)) ? car(sc->args) : sc->nil);
   *
   *   which jumps to the point of the goto returning car(args).
   *
   * There is one gotcha: we can't jump back in from outside, so if the caller saves the goto
   *   and tries to invoke it outside the call-with-exit block, we have to
   *   make sure it triggers an error.  So, if the escape is called, it then
   *   deactivates itself.  Otherwise the block returns, we pop to OP_DEACTIVATE_GOTO,
   *   and it finds the goto in sc->args.
   * Even worse:
   *
       (let ((cc #f))
         (call-with-exit
           (lambda (c3)
             (call/cc (lambda (ret) (set! cc ret)))
             (c3)))
         (cc))
   *
   * where we jump back into a call-with-exit body via call/cc, the goto has to be
   * re-established.
   *
   * I think call-with-exit could be based on catch, but it's a simpler notion,
   *   and certainly at the source level it is easier to read.
   */
  return(sc->nil);
}



/* -------------------------------- numbers -------------------------------- */

#if WITH_GMP
  static char *big_number_to_string_with_radix(s7_pointer p, int32_t radix, int32_t width, int32_t *nlen, use_write_t use_write);
  static bool big_numbers_are_eqv(s7_pointer a, s7_pointer b);
  static s7_pointer string_to_either_integer(s7_scheme *sc, const char *str, int32_t radix);
  static s7_pointer string_to_either_ratio(s7_scheme *sc, const char *nstr, const char *dstr, int32_t radix);
  static s7_pointer string_to_either_real(s7_scheme *sc, const char *str, int32_t radix);
  static s7_pointer string_to_either_complex(s7_scheme *sc, char *q, char *slash1, char *ex1, bool has_dec_point1,
					     char *plus, char *slash2, char *ex2, bool has_dec_point2, int32_t radix, int32_t has_plus_or_minus);
  static s7_pointer big_add(s7_scheme *sc, s7_pointer args);
  static s7_pointer big_subtract(s7_scheme *sc, s7_pointer args);
  static s7_pointer big_multiply(s7_scheme *sc, s7_pointer args);
  static s7_pointer big_divide(s7_scheme *sc, s7_pointer args);
  static s7_pointer big_random(s7_scheme *sc, s7_pointer args);
  static s7_pointer s7_int_to_big_integer(s7_scheme *sc, s7_int val);
  static s7_pointer s7_ratio_to_big_ratio(s7_scheme *sc, s7_int num, s7_int den);
  static s7_pointer s7_number_to_big_real(s7_scheme *sc, s7_pointer p);
  static s7_pointer promote_number(s7_scheme *sc, int32_t type, s7_pointer x);
  static s7_pointer big_equal(s7_scheme *sc, s7_pointer args);
  static s7_pointer big_negate(s7_scheme *sc, s7_pointer args);
  static s7_pointer big_invert(s7_scheme *sc, s7_pointer args);
#if (!WITH_PURE_S7)
  static s7_pointer big_inexact_to_exact(s7_scheme *sc, s7_pointer args);
  static s7_pointer big_exact_to_inexact(s7_scheme *sc, s7_pointer args);
#endif
  static s7_pointer mpz_to_big_integer(s7_scheme *sc, mpz_t val);
  static s7_pointer mpq_to_big_ratio(s7_scheme *sc, mpq_t val);
  static s7_pointer mpfr_to_big_real(s7_scheme *sc, mpfr_t val);
  static s7_pointer mpc_to_big_complex(s7_scheme *sc, mpc_t val);
#endif

#if ((defined(__clang__) && (!POINTER_32) && ((__clang_major__ > 3) || (__clang_major__ == 3 && __clang_minor__ >= 4))) || (defined(__GNUC__) && __GNUC__ >= 5))
  #define HAVE_OVERFLOW_CHECKS 1
#else
  #define HAVE_OVERFLOW_CHECKS 0
  #if (!WITH_GMP)
    #warning "no arithmetic overflow checks in this version of s7"
  #endif
#endif

#if (defined(__clang__) && (!POINTER_32) && ((__clang_major__ > 3) || (__clang_major__ == 3 && __clang_minor__ >= 4))) 
  #define subtract_overflow(A, B, C)     __builtin_ssubll_overflow((long long)A, (long long)B, (long long *)C)
  #define add_overflow(A, B, C)          __builtin_saddll_overflow((long long)A, (long long)B, (long long *)C)
  #define multiply_overflow(A, B, C)     __builtin_smulll_overflow((long long)A, (long long)B, (long long *)C)
  #define int_subtract_overflow(A, B, C) __builtin_ssub_overflow(A, B, C)
  #define int_add_overflow(A, B, C)      __builtin_sadd_overflow(A, B, C)
  #define int_multiply_overflow(A, B, C) __builtin_smul_overflow(A, B, C)
#else
#if (defined(__GNUC__) && __GNUC__ >= 5)
  #define subtract_overflow(A, B, C)     __builtin_sub_overflow(A, B, C)
  #define add_overflow(A, B, C)          __builtin_add_overflow(A, B, C)
  #define multiply_overflow(A, B, C)     __builtin_mul_overflow(A, B, C)
  #define int_subtract_overflow(A, B, C) __builtin_sub_overflow(A, B, C)
  #define int_add_overflow(A, B, C)      __builtin_add_overflow(A, B, C)
  #define int_multiply_overflow(A, B, C) __builtin_mul_overflow(A, B, C)
#endif
#endif


#define s7_int_abs(x) ((x) >= 0 ? (x) : -(x))
/* can't use abs even in gcc -- it doesn't work with int64_ts! */

#if (!__NetBSD__)
  #define s7_fabsl(X) fabsl(X)
#else
  static double s7_fabsl(long double x) {if (x < 0.0) return(-x);  return(x);}
#endif


static bool is_NaN(s7_double x) {return(x != x);}
/* callgrind says this is faster than isnan, I think (very confusing data...) */


#if defined(__sun) && defined(__SVR4)
  static bool is_inf(s7_double x) {return((x == x) && (is_NaN(x - x)));} /* there's no isinf in Solaris */
#else
#if (!MS_WINDOWS)

  #if __cplusplus
    #define is_inf(x) std::isinf(x)
  #else
    #define is_inf(x) isinf(x)
  #endif

#else
  static bool is_inf(s7_double x) {return((x == x) && (is_NaN(x - x)));}  /* Another possibility: (x * 0) != 0 */

  /* in MS C, we need to provide inverse hyperbolic trig funcs and cbrt */
  static double asinh(double x) {return(log(x + sqrt(1.0 + x * x)));}
  static double acosh(double x) {return(log(x + sqrt(x * x - 1.0)));}
  /* perhaps less prone to numerical troubles (untested): 2.0 * log(sqrt(0.5 * (x + 1.0)) + sqrt(0.5 * (x - 1.0))) */
  static double atanh(double x) {return(log((1.0 + x) / (1.0 - x)) / 2.0);}
  static double cbrt(double x) {if (x >= 0.0) return(pow(x, 1.0 / 3.0)); return(-pow(-x, 1.0 / 3.0));}
#endif /* windows */
#endif /* sun */


/* for g_log, we also need round. this version is from stackoverflow, see also round_per_R5RS below */
double s7_round(double number) {return((number < 0.0) ? ceil(number - 0.5) : floor(number + 0.5));}

#if HAVE_COMPLEX_NUMBERS
#if __cplusplus
  #define _Complex_I (complex<s7_double>(0.0, 1.0))
  #define creal(x) Real(x)
  #define cimag(x) Imag(x)
  #define carg(x) arg(x)
  #define cabs(x) abs(x)
  #define csqrt(x) sqrt(x)
  #define cpow(x, y) pow(x, y)
  #define clog(x) log(x)
  #define cexp(x) exp(x)
  #define csin(x) sin(x)
  #define ccos(x) cos(x)
  #define csinh(x) sinh(x)
  #define ccosh(x) cosh(x)
#else
  typedef double complex s7_complex;
#endif


#if (!HAVE_COMPLEX_TRIG)
#if (__cplusplus)

  static s7_complex ctan(s7_complex z)   {return(csin(z) / ccos(z));}
  static s7_complex ctanh(s7_complex z)  {return(csinh(z) / ccosh(z));}
  static s7_complex casin(s7_complex z)  {return(-_Complex_I * clog(_Complex_I * z + csqrt(1.0 - z * z)));}
  static s7_complex cacos(s7_complex z)  {return(-_Complex_I * clog(z + _Complex_I * csqrt(1.0 - z * z)));}
  static s7_complex catan(s7_complex z)  {return(_Complex_I * clog((_Complex_I + z) / (_Complex_I - z)) / 2.0);}
  static s7_complex casinh(s7_complex z) {return(clog(z + csqrt(1.0 + z * z)));}
  static s7_complex cacosh(s7_complex z) {return(clog(z + csqrt(z * z - 1.0)));}
  static s7_complex catanh(s7_complex z) {return(clog((1.0 + z) / (1.0 - z)) / 2.0);}
#else

/* still not in FreeBSD! */
static s7_complex clog(s7_complex z) {return(log(fabs(cabs(z))) + carg(z) * _Complex_I);}
static s7_complex cpow(s7_complex x, s7_complex y)
{
  s7_double r = cabs(x);
  s7_double theta = carg(x);
  s7_double yre = creal(y);
  s7_double yim = cimag(y);
  s7_double nr = exp(yre * log(r) - yim * theta);
  s7_double ntheta = yre * theta + yim * log(r);
  return(nr * cos(ntheta) + (nr * sin(ntheta)) * _Complex_I); /* make-polar */
}

#if (!defined(__FreeBSD__)) || (__FreeBSD__ < 9) /* untested -- this orignally looked at __FreeBSD_version which apparently no longer exists */
  static s7_complex cexp(s7_complex z) {return(exp(creal(z)) * cos(cimag(z)) + (exp(creal(z)) * sin(cimag(z))) * _Complex_I);}
#endif

#if (!defined(__FreeBSD__)) || (__FreeBSD__ < 10)
  static s7_complex csin(s7_complex z)   {return(sin(creal(z)) * cosh(cimag(z)) + (cos(creal(z)) * sinh(cimag(z))) * _Complex_I);}
  static s7_complex ccos(s7_complex z)   {return(cos(creal(z)) * cosh(cimag(z)) + (-sin(creal(z)) * sinh(cimag(z))) * _Complex_I);}
  static s7_complex csinh(s7_complex z)  {return(sinh(creal(z)) * cos(cimag(z)) + (cosh(creal(z)) * sin(cimag(z))) * _Complex_I);}
  static s7_complex ccosh(s7_complex z)  {return(cosh(creal(z)) * cos(cimag(z)) + (sinh(creal(z)) * sin(cimag(z))) * _Complex_I);}
  static s7_complex ctan(s7_complex z)   {return(csin(z) / ccos(z));}
  static s7_complex ctanh(s7_complex z)  {return(csinh(z) / ccosh(z));}
  static s7_complex casin(s7_complex z)  {return(-_Complex_I * clog(_Complex_I * z + csqrt(1.0 - z * z)));}
  static s7_complex cacos(s7_complex z)  {return(-_Complex_I * clog(z + _Complex_I * csqrt(1.0 - z * z)));}
  static s7_complex catan(s7_complex z)  {return(_Complex_I * clog((_Complex_I + z) / (_Complex_I - z)) / 2.0);}
  static s7_complex catanh(s7_complex z) {return(clog((1.0 + z) / (1.0 - z)) / 2.0);}
  static s7_complex casinh(s7_complex z) {return(clog(z + csqrt(1.0 + z * z)));}
  static s7_complex cacosh(s7_complex z) {return(clog(z + csqrt(z * z - 1.0)));}
  /* perhaps less prone to numerical troubles (untested): 2.0 * clog(csqrt(0.5 * (z + 1.0)) + csqrt(0.5 * (z - 1.0))) */
#endif /* not FreeBSD 10 */
#endif /* not c++ */
#endif /* not HAVE_COMPLEX_TRIG */

#else  /* not HAVE_COMPLEX_NUMBERS */
  typedef double s7_complex;
  #define _Complex_I 1
  #define creal(x) x
  #define cimag(x) x
  #define csin(x) sin(x)
  #define casin(x) x
  #define ccos(x) cos(x)
  #define cacos(x) x
  #define ctan(x) x
  #define catan(x) x
  #define csinh(x) x
  #define casinh(x) x
  #define ccosh(x) x
  #define cacosh(x) x
  #define ctanh(x) x
  #define catanh(x) x
  #define cexp(x) exp(x)
  #define cpow(x, y) pow(x, y)
  #define clog(x) log(x)
  #define csqrt(x) sqrt(x)
  #define conj(x) x
#endif

#ifdef __OpenBSD__
  /* openbsd's builtin versions of these functions are not usable */
  static s7_complex catanh_1(s7_complex z) {return(clog((1.0 + z) / (1.0 - z)) / 2.0);}
  static s7_complex casinh_1(s7_complex z) {return(clog(z + csqrt(1.0 + z * z)));}
  static s7_complex cacosh_1(s7_complex z) {return(clog(z + csqrt(z * z - 1.0)));}
#endif
#ifdef __NetBSD__
  static s7_complex catanh_1(s7_complex z) {return(clog((1.0 + z) / (1.0 - z)) / 2.0);}
  static s7_complex casinh_1(s7_complex z) {return(clog(z + csqrt(1.0 + z * z)));}
#endif


bool s7_is_number(s7_pointer p)
{
#if WITH_GMP
  return((is_number(p)) || (is_big_number(p)));
#else
  return(is_number(p));
#endif
}


bool s7_is_integer(s7_pointer p)
{
#if WITH_GMP
  return((is_t_integer(p)) ||
	 (is_t_big_integer(p)));
#else
  return(is_integer(p));
#endif
}

bool s7_is_real(s7_pointer p)
{
#if WITH_GMP
  return((is_real(p)) ||
	 (is_t_big_integer(p)) ||
	 (is_t_big_ratio(p)) ||
	 (is_t_big_real(p)));
#else
  return(is_real(p)); /* in GSL, a NaN or inf is not a real, or perhaps better, finite = not (nan or inf) */
#endif
}


bool s7_is_rational(s7_pointer p)
{
#if WITH_GMP
  return((is_rational(p)) ||
	 (is_t_big_integer(p)) ||
	 (is_t_big_ratio(p)));
#else
  return(is_rational(p));
#endif
}


bool s7_is_ratio(s7_pointer p)
{
#if WITH_GMP
  return((is_t_ratio(p)) ||
	 (is_t_big_ratio(p)));
#else
  return(is_t_ratio(p));
#endif
}


bool s7_is_complex(s7_pointer p)
{
#if WITH_GMP
  return((is_number(p)) || (is_big_number(p)));
#else
  return(is_number(p));
#endif
}


static s7_int c_gcd(s7_int u, s7_int v)
{
  s7_int a, b;

  if ((u == s7_int_min) || (v == s7_int_min))
    {
      /* can't take abs of these (below) so do it by hand */
      s7_int divisor = 1;
      if (u == v) return(u);
      while (((u & 1) == 0) && ((v & 1) == 0))
	{
	  u /= 2;
	  v /= 2;
	  divisor *= 2;
	}
      return(divisor);
    }

  a = s7_int_abs(u);
  b = s7_int_abs(v);
  while (b != 0)
    {
      s7_int temp;
      temp = a % b;
      a = b;
      b = temp;
    }
  if (a < 0)
    return(-a);
  return(a);
}


static bool c_rationalize(s7_double ux, s7_double error, s7_int *numer, s7_int *denom)
{
  /*
    (define* (rat ux (err 0.0000001))
      ;; translated from CL code in Canny, Donald, Ressler, "A Rational Rotation Method for Robust Geometric Algorithms"
      (let ((x0 (- ux error))
	    (x1 (+ ux error)))
        (let ((i (ceiling x0))
	      (i0 (floor x0))
	      (i1 (ceiling x1))
	      (r 0))
          (if (>= x1 i)
	      i
	      (do ((p0 i0 (+ p1 (* r p0)))
	           (q0 1 (+ q1 (* r q0)))
	           (p1 i1 p0)
	           (q1 1 q0)
	           (e0 (- i1 x0) e1p)
	           (e1 (- x0 i0) (- e0p (* r e1p)))
	           (e0p (- i1 x1) e1)
	           (e1p (- x1 i0) (- e0 (* r e1))))
	          ((<= x0 (/ p0 q0) x1)
	           (/ p0 q0))
	        (set! r (min (floor (/ e0 e1))
			     (ceiling (/ e0p e1p)))))))))
  */

  double x0, x1;
  s7_int i, i0, i1, p0, q0, p1, q1;
  double e0, e1, e0p, e1p;
  int32_t tries = 0;
  /* don't use s7_double here;  if it is "long double", the loop below will hang */

  /* #e1e19 is a killer -- it's bigger than most-positive-fixnum, but if we ceil(ux) below
   *   it turns into most-negative-fixnum.  1e19 is trouble in many places.
   */
  if ((ux > s7_int_max) || (ux < s7_int_min))
    {
      /* can't return false here because that confuses some of the callers!
       */
      if (ux > s7_int_min) (*numer) = s7_int_max; else (*numer) = s7_int_min;
      (*denom) = 1;
      return(true);
    }

  if (error < 0.0) error = -error;
  x0 = ux - error;
  x1 = ux + error;
  i = (s7_int)ceil(x0);

  if (error >= 1.0) /* aw good grief! */
    {
      if (x0 < 0)
	{
	  if (x1 < 0)
	    (*numer) = (s7_int)floor(x1);
	  else (*numer) = 0;
	}
      else (*numer) = i;
      (*denom) = 1;
      return(true);
    }

  if (x1 >= i)
    {
      if (i >= 0)
	(*numer) = i;
      else (*numer) = (s7_int)floor(x1);
      (*denom) = 1;
      return(true);
    }

  i0 = (s7_int)floor(x0);
  i1 = (s7_int)ceil(x1);

  p0 = i0;
  q0 = 1;
  p1 = i1;
  q1 = 1;
  e0 = i1 - x0;
  e1 = x0 - i0;
  e0p = i1 - x1;
  e1p = x1 - i0;

  while (true)
    {
      s7_int old_p1, old_q1;
      double old_e0, old_e1, old_e0p, val, r, r1;
      val = (double)p0 / (double)q0;

      if (((x0 <= val) && (val <= x1)) ||
	  (e1 == 0)                    ||
	  (e1p == 0)                   ||
	  (tries > 100))
	{
	  (*numer) = p0;
	  (*denom) = q0;
	  return(true);
	}
      tries++;

      r = (s7_int)floor(e0 / e1);
      r1 = (s7_int)ceil(e0p / e1p);
      if (r1 < r) r = r1;

      /* do handles all step vars in parallel */
      old_p1 = p1;
      p1 = p0;
      old_q1 = q1;
      q1 = q0;
      old_e0 = e0;
      e0 = e1p;
      old_e0p = e0p;
      e0p = e1;
      old_e1 = e1;

      p0 = old_p1 + r * p0;
      q0 = old_q1 + r * q0;
      e1 = old_e0p - r * e1p;
      /* if the error is set too low, we can get e1 = 0 here: (rationalize (/ pi) 1e-17) */
      e1p = old_e0 - r * old_e1;
    }
  return(false);
}


s7_pointer s7_rationalize(s7_scheme *sc, s7_double x, s7_double error)
{
  s7_int numer = 0, denom = 1;
  if (c_rationalize(x, error, &numer, &denom))
    return(s7_make_ratio(sc, numer, denom));
  return(make_real(sc, x));
}


static s7_int number_to_numerator(s7_pointer n)
{
  if (is_t_ratio(n))
    return(numerator(n));
  return(integer(n));
}


static s7_int number_to_denominator(s7_pointer n)
{
  if (is_t_ratio(n))
    return(denominator(n));
  return(1);
}


s7_pointer s7_make_integer(s7_scheme *sc, s7_int n)
{
  s7_pointer x;
  if (is_small(n))              /* ((n >= 0) && (n < NUM_SMALL_INTS)) is slower */
    return(small_int(n));

  new_cell(sc, x, T_INTEGER);
  integer(x) = n;
  return(x);
}


static s7_pointer make_mutable_integer(s7_scheme *sc, s7_int n)
{
  s7_pointer x;
  new_cell(sc, x, T_INTEGER | T_MUTABLE);
  integer(x) = n;
  return(x);
}


static s7_pointer make_permanent_integer_unchecked(s7_int i)
{
  s7_pointer p;
  p = (s7_pointer)calloc(1, sizeof(s7_cell));
  typeflag(p) = T_IMMUTABLE | T_INTEGER;
  unheap(p);
  integer(p) = i;
  return(p);
}

static s7_pointer make_permanent_integer(s7_int i)
{
  if (is_small(i)) return(small_int(i));

  if (i == MAX_ARITY) return(max_arity);
  if (i == CLOSURE_ARITY_NOT_SET) return(arity_not_set);
  if (i == -1) return(minus_one);
  if (i == -2) return(minus_two);
  /* a few -3 */

  return(make_permanent_integer_unchecked(i));
}


s7_pointer s7_make_real(s7_scheme *sc, s7_double n)
{
  s7_pointer x;
  new_cell(sc, x, T_REAL);
  set_real(x, n);
  return(x);
}


s7_pointer s7_make_mutable_real(s7_scheme *sc, s7_double n)
{
  s7_pointer x;
  new_cell(sc, x, T_REAL | T_MUTABLE);
  set_real(x, n);
  return(x);
}


static s7_pointer make_permanent_real(s7_double n)
{
  s7_pointer x;
  int32_t nlen = 0;
  char *str;

  x = (s7_pointer)calloc(1, sizeof(s7_cell));
  set_type(x, T_IMMUTABLE | T_REAL);
  unheap(x);
  set_real(x, n);

  str = number_to_string_base_10(x, 0, float_format_precision, 'g', &nlen, USE_WRITE);
  set_print_name(x, str, nlen);
  return(x);
}


s7_pointer s7_make_complex(s7_scheme *sc, s7_double a, s7_double b)
{
  s7_pointer x;
  if (b == 0.0)
    {
      new_cell(sc, x, T_REAL);
      set_real(x, a);
    }
  else
    {
      new_cell(sc, x, T_COMPLEX);
      set_real_part(x, a);
      set_imag_part(x, b);
    }
  return(x);
}


s7_pointer s7_make_ratio(s7_scheme *sc, s7_int a, s7_int b)
{
  s7_pointer x;
  s7_int divisor;

  if (b == 0)
    return(division_by_zero_error(sc, s7_make_string_wrapper(sc, "make-ratio"), set_elist_2(sc, make_integer(sc, a), small_int(0))));
  if (a == 0)
    return(small_int(0));
  if (b == 1)
    return(make_integer(sc, a));

#if (!WITH_GMP)
  if (b == s7_int_min)
    {
      if (a == b)
	return(small_int(1));

      /* we've got a problem... This should not trigger an error during reading -- we might have the
       *   ratio on a switch with-bignums or whatever, so its mere occurrence is just an annoyance.
       *   We'll try to do something...
       */
      if (a & 1)
	{
	  if (a == 1)
	    return(real_NaN);
	  /* not an error here? we can't get this in the ratio reader, I think, because the denominator is negative */
	  b = b + 1;
	  /* so (/ -1 most-negative-fixnum) -> 1/9223372036854775807 -- not ideal, but ... */
	}
      else
	{
	  a /= 2;
	  b /= 2;
	}
    }
#endif

  if (b < 0)
    {
      a = -a;
      b = -b;
    }
  divisor = c_gcd(a, b);
  if (divisor != 1)
    {
      a /= divisor;
      b /= divisor;
    }
  if (b == 1)
    return(make_integer(sc, a));

  new_cell(sc, x, T_RATIO);
  numerator(x) = a;
  denominator(x) = b;

  return(x);
}
/* in fc19 as a guest running in virtualbox on OSX, the line  a /= divisor can abort with an arithmetic exception (SIGFPE)
 *    if leastfix/mostfix -- apparently this is a bug in virtualbox.
 */


#define WITH_OVERFLOW_ERROR true
#define WITHOUT_OVERFLOW_ERROR false

#if (!WITH_PURE_S7) && (!WITH_GMP)
static s7_pointer exact_to_inexact(s7_scheme *sc, s7_pointer x)
{
  /* this is tricky because a big int32_t can mess up when turned into a double:
   *   (truncate (exact->inexact most-positive-fixnum)) -> -9223372036854775808
   */
  switch (type(x))
    {
    case T_INTEGER: return(make_real(sc, (s7_double)(integer(x))));
    case T_RATIO:   return(make_real(sc, (s7_double)(fraction(x))));
    case T_REAL:                  
    case T_COMPLEX: return(x); /* apparently (exact->inexact 1+i) is not an error */
    default: 
      method_or_bust_with_type_one_arg(sc, x, sc->exact_to_inexact_symbol, list_1(sc, x), a_number_string);
    }
}

static s7_pointer inexact_to_exact(s7_scheme *sc, s7_pointer x, bool with_error)
{
  switch (type(x))
    {
    case T_INTEGER:
    case T_RATIO:
      return(x);

    case T_REAL:
      {
	s7_int numer = 0, denom = 1;
	s7_double val;

	val = s7_real(x);
	if ((is_inf(val)) || (is_NaN(val)))
	  {
	    if (with_error)
	      return(simple_wrong_type_argument_with_type(sc, sc->inexact_to_exact_symbol, x, a_normal_real_string));
	    return(sc->nil);
	  }

	if ((val > s7_int_max) ||
	    (val < s7_int_min))
	  {
	    if (with_error)
	      return(simple_out_of_range(sc, sc->inexact_to_exact_symbol, x, its_too_large_string));
	    return(sc->nil);
	  }

	if (c_rationalize(val, sc->default_rationalize_error, &numer, &denom))
	  return(s7_make_ratio(sc, numer, denom));
      }

    default:
      if (with_error)
	method_or_bust_one_arg(sc, x, sc->inexact_to_exact_symbol, list_1(sc, x), T_REAL);
      return(sc->nil);
    }
  return(x);
}
#endif

s7_double s7_number_to_real_with_caller(s7_scheme *sc, s7_pointer x, const char *caller)
{
  if (is_t_real(x))
    return(real(x));
  /* this is nearly always the case in current usage, so by avoiding the "switch" we can go twice as fast */

  switch (type(x))
    {
    case T_INTEGER:     return((s7_double)integer(x));
    case T_RATIO:       return((s7_double)numerator(x) / (s7_double)denominator(x));
    case T_REAL:        return(real(x));
#if WITH_GMP
    case T_BIG_INTEGER: return((s7_double)big_integer_to_s7_int(big_integer(x)));
    case T_BIG_RATIO:   return((s7_double)((long double)big_integer_to_s7_int(mpq_numref(big_ratio(x))) / 
					   (long double)big_integer_to_s7_int(mpq_denref(big_ratio(x)))));
    case T_BIG_REAL:    return((s7_double)mpfr_get_d(big_real(x), GMP_RNDN));
#endif
    }
  s7_wrong_type_arg_error(sc, caller, 0, x, "a real number");
  return(0.0);
}


s7_double s7_number_to_real(s7_scheme *sc, s7_pointer x)
{
  return(s7_number_to_real_with_caller(sc, x, "s7_number_to_real"));
}


s7_int s7_number_to_integer_with_caller(s7_scheme *sc, s7_pointer x, const char *caller)
{
  if (!s7_is_integer(x))
    s7_wrong_type_arg_error(sc, caller, 0, x, "an integer");
#if WITH_GMP
  if (is_t_big_integer(x))
    return(big_integer_to_s7_int(big_integer(x)));
#endif
  return(integer(x));
}

s7_int s7_number_to_integer(s7_scheme *sc, s7_pointer x) 
{
  return(s7_number_to_integer_with_caller(sc, x, "s7_number_to_integer"));
}


s7_int s7_numerator(s7_pointer x)
{
  switch (type(x))
    {
    case T_INTEGER:     return(integer(x));
    case T_RATIO:       return(numerator(x));
#if WITH_GMP
    case T_BIG_INTEGER: return(big_integer_to_s7_int(big_integer(x)));
    case T_BIG_RATIO:   return(big_integer_to_s7_int(mpq_numref(big_ratio(x))));
#endif
    }
  return(0);
}


s7_int s7_denominator(s7_pointer x)
{
  switch (type(x))
    {
    case T_RATIO:     return(denominator(x));
#if WITH_GMP
    case T_BIG_RATIO: return(big_integer_to_s7_int(mpq_denref(big_ratio(x))));
#endif
    }
  return(1);
}


s7_int s7_integer(s7_pointer p)
{
#if WITH_GMP
  if (is_t_big_integer(p))
    return(big_integer_to_s7_int(big_integer(p)));
#endif
  return(integer(p));
}


s7_double s7_real(s7_pointer p)
{
#if WITH_GMP
  if (is_t_big_real(p))
    return((s7_double)mpfr_get_d(big_real(p), GMP_RNDN));
#endif
  return(real(p));
}


#if (!WITH_GMP)
static s7_complex s7_to_c_complex(s7_pointer p)
{
#if HAVE_COMPLEX_NUMBERS
  return(CMPLX(s7_real_part(p), s7_imag_part(p)));
#else
  return(0.0);
#endif
}


static s7_pointer s7_from_c_complex(s7_scheme *sc, s7_complex z)
{
  return(s7_make_complex(sc, creal(z), cimag(z)));
}
#endif


static int32_t s7_int32_max = 0, s7_int32_min = 0, s7_int_bits = 0, s7_int_digits = 0; /* initialized later */
static int32_t s7_int_digits_by_radix[17];


#if (!WITH_GMP)
static s7_pointer s7_negate(s7_scheme *sc, s7_pointer p)     /* can't use "negate" because it confuses C++! */
{
  switch (type(p))
    {
    case T_INTEGER: return(make_integer(sc, -integer(p)));
    case T_RATIO:   return(s7_make_ratio(sc, -numerator(p), denominator(p)));
    case T_REAL:    return(make_real(sc, -real(p)));
    default:        return(s7_make_complex(sc, -real_part(p), -imag_part(p)));
    }
}
#endif


static s7_pointer s7_invert(s7_scheme *sc, s7_pointer p)      /* s7_ to be consistent... */
{
  switch (type(p))
    {
    case T_INTEGER:
      return(s7_make_ratio(sc, 1, integer(p)));      /* a already checked, not 0 */

    case T_RATIO:
      return(s7_make_ratio(sc, denominator(p), numerator(p)));

    case T_REAL:
      return(make_real(sc, 1.0 / real(p)));

    case T_COMPLEX:
      {
	s7_double r2, i2, den;
	r2 = real_part(p);
	i2 = imag_part(p);
	den = (r2 * r2 + i2 * i2);
	return(s7_make_complex(sc, r2 / den, -i2 / den));
      }

    default:
      return(wrong_type_argument_with_type(sc, sc->divide_symbol, 1, p, a_number_string));
    }
}


static s7_pointer subtract_ratios(s7_scheme *sc, s7_pointer x, s7_pointer y)
{
  s7_int d1, d2, n1, n2;
  d1 = number_to_denominator(x);
  n1 = number_to_numerator(x);
  d2 = number_to_denominator(y);
  n2 = number_to_numerator(y);

  if (d1 == d2)                                     /* the easy case -- if overflow here, it matches the int32_t case */
    return(s7_make_ratio(sc, n1 - n2, d1));

#if (!WITH_GMP) && HAVE_OVERFLOW_CHECKS
  {
    s7_int n1d2, n2d1, d1d2, dn;
    if ((multiply_overflow(d1, d2, &d1d2)) ||
	(multiply_overflow(n1, d2, &n1d2)) ||
	(multiply_overflow(n2, d1, &n2d1)) ||
	(subtract_overflow(n1d2, n2d1, &dn)))
      return(make_real(sc, ((long double)n1 / (long double)d1) - ((long double)n2 / (long double)d2)));
    return(s7_make_ratio(sc, dn, d1d2));
  }
#else
  return(s7_make_ratio(sc, n1 * d2 - n2 * d1, d1 * d2));
#endif
}


static bool s7_is_negative(s7_pointer obj)
{
  switch (type(obj))
    {
    case T_INTEGER:     return(integer(obj) < 0);
    case T_RATIO:       return(numerator(obj) < 0);
#if WITH_GMP
    case T_BIG_INTEGER: return(mpz_cmp_ui(big_integer(obj), 0) < 0);
    case T_BIG_RATIO:   return(mpq_cmp_ui(big_ratio(obj), 0, 1) < 0);
    case T_BIG_REAL:    return(mpfr_cmp_ui(big_real(obj), 0) < 0);
#endif
    default:            return(real(obj) < 0);
    }
}


static bool s7_is_positive(s7_pointer x)
{
  switch (type(x))
    {
    case T_INTEGER:     return(integer(x) > 0);
    case T_RATIO:       return(numerator(x) > 0);
#if WITH_GMP
    case T_BIG_INTEGER: return(mpz_cmp_ui(big_integer(x), 0) > 0);
    case T_BIG_RATIO:   return(mpq_cmp_ui(big_ratio(x), 0, 1) > 0);
    case T_BIG_REAL:    return(mpfr_cmp_ui(big_real(x), 0) > 0);
#endif
    default:            return(real(x) > 0.0);
    }
}


static bool s7_is_zero(s7_pointer x)
{
  switch (type(x))
    {
    case T_INTEGER:     return(integer(x) == 0);
    case T_REAL:        return(real(x) == 0.0);
#if WITH_GMP
    case T_BIG_INTEGER: return(mpz_cmp_ui(big_integer(x), 0) == 0);
    case T_BIG_REAL:    return(mpfr_zero_p(big_real(x)));
#endif
    default:            return(false); /* ratios and complex numbers here are already collapsed into integers and reals */
    }
}


static bool s7_is_one(s7_pointer x)
{
    return(((is_integer(x)) && (integer(x) == 1)) ||
	   ((is_t_real(x)) && (real(x) == 1.0)));
}


/* optimize exponents */
#define MAX_POW 32
static double pepow[17][MAX_POW], mepow[17][MAX_POW];

static void init_pows(void)
{
  int32_t i, j;
  for (i = 2; i < 17; i++)        /* radix between 2 and 16 */
    for (j = 0; j < MAX_POW; j++) /* saved exponent between 0 and +/- MAX_POW */
      {
	pepow[i][j] = pow((double)i, (double)j);
	mepow[i][j] = pow((double)i, (double)(-j));
      }
}

static double ipow(int32_t x, int32_t y)
{
  if ((y < MAX_POW) && (y > (-MAX_POW)))
    {
      if (y >= 0)
	return(pepow[x][y]);
      return(mepow[x][-y]);
    }
  return(pow((double)x, (double)y));
}


static int32_t s7_int_to_string(char *p, s7_int n, int32_t radix, int32_t width)
{
  static const char dignum[] = "0123456789abcdef";
  int32_t i, len, start, end;
  bool sign;
  s7_int pown;

  if ((radix < 2) || (radix > 16))
    return(0);

  if (n == s7_int_min) /* can't negate this, so do it by hand */
    {
      static const char *mnfs[17] = {"","",
	"-1000000000000000000000000000000000000000000000000000000000000000", "-2021110011022210012102010021220101220222",
	"-20000000000000000000000000000000", "-1104332401304422434310311213", "-1540241003031030222122212",
	"-22341010611245052052301", "-1000000000000000000000", "-67404283172107811828",	"-9223372036854775808",
	"-1728002635214590698",	"-41a792678515120368", "-10b269549075433c38", "-4340724c6c71dc7a8", "-160e2ad3246366808", "-8000000000000000"};

      len = safe_strlen(mnfs[radix]);
      if (width > len)
	{
	  start = width - len - 1;
	  memset((void *)p, (int)' ', start);
	}
      else start = 0;
      for (i = 0; i < len; i++)
	p[start + i] = mnfs[radix][i];
      p[len + start] = '\0';
      return(len + start);
    }

  sign = (n < 0);
  if (sign) n = -n;

  /* the previous version that counted up to n, rather than dividing down below n, as here,
   *   could be confused by large ints on 64 bit machines
   */
  pown = n;
  for (i = 1; i < 100; i++)
    {
      if (pown < radix)
	break;
      pown /= (s7_int)radix;
    }
  len = i - 1;
  if (sign) len++;
  end = 0;
  if (width > len)                  /* (format #f "~10B" 123) */
    {
      start = width - len - 1;
      end += start;
      memset((void *)p, (int)' ', start);
    }
  else
    {
      start = 0;
      end = 0;
    }

  if (sign)
    {
      p[start] = '-';
      end++;
    }

  for (i = start + len; i >= end; i--)
    {
      p[i] = dignum[n % radix];
      n /= radix;
    }
  p[len + start + 1] = '\0';
  return(len + start + 1);
}


static char *integer_to_string_base_10_no_width(s7_pointer obj, int32_t *nlen) /* do not free the returned string */
{
  int64_t num;
  char *p, *op;
  bool sign;
  static char int_to_str[INT_TO_STR_SIZE];

  if (has_print_name(obj))
    {
      (*nlen) = print_name_length(obj);
      return((char *)print_name(obj));
    }
  /* (*nlen) = snprintf(int_to_str, INT_TO_STR_SIZE, "%" PRId64 "", (int64_t)integer(obj));
   *  but that is very slow -- the following code is 6 times faster
   */
  num = (int64_t)integer(obj);
  if (num == s7_int_min)
    {
      (*nlen) = 20;
      return((char *)"-9223372036854775808");
    }
  p = (char *)(int_to_str + INT_TO_STR_SIZE - 1);
  op = p;
  *p-- = '\0';

  sign = (num < 0);
  if (sign) num = -num;  /* we need a positive index below */
  do {*p-- = "0123456789"[num % 10]; num /= 10;} while (num);
  if (sign)
    {
      *p = '-';
      (*nlen) = op - p;
      return(p);
    }

  (*nlen) = op - p - 1;
  return(++p);
}


#define BASE_10 10
static int32_t num_to_str_size = -1;
static char *num_to_str = NULL;
static const char *float_format_g = NULL;

static char *floatify(char *str, int32_t *nlen)
{
  if ((!strchr(str, 'e')) &&
      (!strchr(str, '.')))
    {
      /* this assumes there is room in str for 2 more chars */
      int32_t len;
      len = *nlen;
      str[len]='.';
      str[len + 1]='0';
      str[len + 2]='\0';
      (*nlen) = len + 2;
    }
  return(str);
}

static char *number_to_string_base_10(s7_pointer obj, int32_t width, int32_t precision, char float_choice, int32_t *nlen, use_write_t choice) /* don't free result */
{
  /* the rest of s7 assumes nlen is set to the correct length 
   *   a tricky case: (format #f "~f" 1e308) -- tries to print 308 digits! so 256 as default len is too small.
   *   but then even worse: (format #f "~F" 1e308+1e308i)!
   */
  int32_t len;
  len = 1024;
  if (width > len) len = 2 * width;
  if (len > num_to_str_size)
    {
      if (!num_to_str)
	num_to_str = (char *)malloc(len * sizeof(char));
      else num_to_str = (char *)realloc(num_to_str, len * sizeof(char));
      num_to_str_size = len;
    }

  /* bignums can't happen here */
  switch (type(obj))
    {
    case T_INTEGER:
      if (width == 0)
	return(integer_to_string_base_10_no_width(obj, nlen));
      (*nlen) = snprintf(num_to_str, num_to_str_size, "%*" PRId64, width, (int64_t)integer(obj));
      break;

    case T_RATIO:
      len = snprintf(num_to_str, num_to_str_size, "%" PRId64 "/%" PRId64, (int64_t)numerator(obj), (int64_t)denominator(obj));
      if (width > len)
	{
	  int32_t spaces;
	  if (width >= num_to_str_size)
	    {
	      num_to_str_size = width + 1;
	      num_to_str = (char *)realloc(num_to_str, num_to_str_size * sizeof(char));
	    }
	  spaces = width - len;
	  num_to_str[width] = '\0';
	  memmove((void *)(num_to_str + spaces), (void *)num_to_str, len);
	  memset((void *)num_to_str, (int)' ', spaces);
	  (*nlen) = width;
	}
      else (*nlen) = len;
      break;

    case T_REAL:
      {
	const char *frmt;
	if (sizeof(double) >= sizeof(s7_double))
	  frmt = (float_choice == 'g') ? "%*.*g" : ((float_choice == 'f') ? "%*.*f" : "%*.*e");
	else frmt = (float_choice == 'g') ? "%*.*Lg" : ((float_choice == 'f') ? "%*.*Lf" : "%*.*Le");

	len = snprintf(num_to_str, num_to_str_size - 4, frmt, width, precision, s7_real(obj)); /* -4 for floatify */
	(*nlen) = len;
	floatify(num_to_str, nlen);
      }
      break;

    default:
      {
	if ((choice == USE_READABLE_WRITE) &&
	    ((is_NaN(real_part(obj))) || (is_NaN(imag_part(obj))) || ((is_inf(real_part(obj))) || (is_inf(imag_part(obj))))))
	  {
	    char rbuf[128], ibuf[128];
	    char *rp, *ip;
	    if (is_NaN(real_part(obj)))
	      rp = (char *)"nan.0";
	    else
	      {
		if (is_inf(real_part(obj)))
		  {
		    if (real_part(obj) < 0.0)
		      rp = (char *)"-inf.0";
		    else rp = (char *)"inf.0";
		  }
		else
		  {
		    snprintf(rbuf, 128, float_format_g, precision, real_part(obj));
		    rp = rbuf;
		  }
	      }
	    if (is_NaN(imag_part(obj)))
	      ip = (char *)"nan.0";
	    else
	      {
		if (is_inf(imag_part(obj)))
		  {
		    if (imag_part(obj) < 0.0)
		      ip = (char *)"-inf.0";
		    else ip = (char *)"inf.0";
		  }
		else
		  {
		    snprintf(ibuf, 128, float_format_g, precision, imag_part(obj));
		    ip = ibuf;
		  }
	      }
	    len = snprintf(num_to_str, num_to_str_size, "(complex %s %s)", rp, ip);
	  }
	else
	  {
	    const char *frmt;
	    if (sizeof(double) >= sizeof(s7_double))
	      {
		if (imag_part(obj) >= 0.0)
		  frmt = (float_choice == 'g') ? "%.*g+%.*gi" : ((float_choice == 'f') ? "%.*f+%.*fi" : "%.*e+%.*ei");
		else frmt = (float_choice == 'g') ? "%.*g%.*gi" : ((float_choice == 'f') ? "%.*f%.*fi" :"%.*e%.*ei"); /* minus sign comes with the imag_part */
	      }
	    else
	      {
		if (imag_part(obj) >= 0.0)
		  frmt = (float_choice == 'g') ? "%.*Lg+%.*Lgi" : ((float_choice == 'f') ? "%.*Lf+%.*Lfi" : "%.*Le+%.*Lei");
		else frmt = (float_choice == 'g') ? "%.*Lg%.*Lgi" : ((float_choice == 'f') ? "%.*Lf%.*Lfi" : "%.*Le%.*Lei");
	      }

	    len = snprintf(num_to_str, num_to_str_size, frmt, precision, real_part(obj), precision, imag_part(obj));
	  }

	if (width > len)  /* (format #f "~20g" 1+i) */
	  {
	    int32_t spaces;
	    if (width >= num_to_str_size)
	      {
		num_to_str_size = width + 1;
		num_to_str = (char *)realloc(num_to_str, num_to_str_size * sizeof(char));
	      }
	    spaces = width - len;
	    num_to_str[width] = '\0';
	    memmove((void *)(num_to_str + spaces), (void *)num_to_str, len);
	    memset((void *)num_to_str, (int)' ', spaces);
	    (*nlen) = width;
	  }
	else (*nlen) = len;
      }
      break;
    }
  return(num_to_str);
}


static char *number_to_string_with_radix(s7_scheme *sc, s7_pointer obj, int32_t radix, int32_t width, int32_t precision, char float_choice, int32_t *nlen)
{
  /* the rest of s7 assumes nlen is set to the correct length */
  char *p;
  int32_t len, str_len;

#if WITH_GMP
  if (s7_is_bignum(obj))
    return(big_number_to_string_with_radix(obj, radix, width, nlen, USE_WRITE));
  /* this ignores precision because it's way too hard to get the mpfr string to look like
   *   C's output -- we either have to call mpfr_get_str twice (the first time just to
   *   find out what the exponent is and how long the string actually is), or we have
   *   to do messy string manipulations.  So (format #f "",3F" pi) ignores the "3" and
   *   prints the full string.
   */
#endif

  if (radix == 10)
    {
      p = number_to_string_base_10(obj, width, precision, float_choice, nlen, USE_WRITE);
      return(copy_string_with_length(p, *nlen));
    }

  switch (type(obj))
    {
    case T_INTEGER:
      p = (char *)malloc((128 + width) * sizeof(char));
      *nlen = s7_int_to_string(p, s7_integer(obj), radix, width);
      return(p);

    case T_RATIO:
      {
	char n[128], d[128];
	s7_int_to_string(n, numerator(obj), radix, 0);
	s7_int_to_string(d, denominator(obj), radix, 0);
	p = (char *)malloc(256 * sizeof(char));
	len = snprintf(p, 256, "%s/%s", n, d);
	str_len = 256;
      }
      break;

    case T_REAL:
      {
	int32_t i;
	s7_int int_part;
	s7_double x, frac_part, min_frac, base;
	bool sign = false;
	char n[128], d[256];

	x = s7_real(obj);

	if (is_NaN(x))
	  return(copy_string_with_length("nan.0", *nlen = 5));
	if (is_inf(x))
	  {
	    if (x < 0.0)
	      return(copy_string_with_length("-inf.0", *nlen = 6));
	    return(copy_string_with_length("inf.0", *nlen = 5));
	  }

	if (x < 0.0)
	  {
	    sign = true;
	    x = -x;
	  }

	if (x > 1.0e18) /* i.e. close to or greater than most-positive-fixnum (9.22e18), so the code below is unlikely to work, (format #f "~X" 1e19) */
	  {
	    int32_t ep;
	    char *p1;
	    s7_pointer r;

	    len = 0;
	    ep = (int32_t)floor(log(x) / log((double)radix));
	    r = make_real(sc, x / pow((double)radix, (double)ep)); /* divide it down to one digit, then the fractional part */
	    p1 = number_to_string_with_radix(sc, r, radix, width, precision, float_choice, &len);
	    p = (char *)malloc((len + 8) * sizeof(char));
	    (*nlen) = snprintf(p, len + 8, "%s%se%d", (sign) ? "-" : "", p1, ep);
	    free(p1);
	    return(p);
	  }

	int_part = (s7_int)floor(x);
	frac_part = x - int_part;
	s7_int_to_string(n, int_part, radix, 0);
	min_frac = (s7_double)ipow(radix, -precision);

	/* doesn't this assume precision < 128/256 and that we can fit in 256 digits (1e308)? */

	for (i = 0, base = radix; (i < precision) && (frac_part > min_frac); i++, base *= radix)
	  {
	    s7_int ipart;
	    ipart = (s7_int)(frac_part * base);
	    if (ipart >= radix)         /* rounding confusion */
	      ipart = radix - 1;
	    frac_part -= (ipart / base);
	    if (ipart < 10)
	      d[i] = (char)('0' + ipart);
	    else d[i] = (char)('a' + ipart -  10);
	  }
	if (i == 0)
	  d[i++] = '0';
	d[i] = '\0';
	p = (char *)malloc(256 * sizeof(char));
	len = snprintf(p, 256, "%s%s.%s", (sign) ? "-" : "", n, d);
	str_len = 256;
      }
      break;

    default:
      {
	char *n, *d;
	p = (char *)malloc(512 * sizeof(char));
	n = number_to_string_with_radix(sc, make_real(sc, real_part(obj)), radix, 0, precision, float_choice, &len);
	d = number_to_string_with_radix(sc, make_real(sc, imag_part(obj)), radix, 0, precision, float_choice, &len);
	len = snprintf(p, 512, "%s%s%si", n, (imag_part(obj) < 0.0) ? "" : "+", d);
	str_len = 512;
	free(n);
	free(d);
      }
      break;
    }

  if (width > len)
    {
      int32_t spaces;
      if (width >= str_len)
	{
	  str_len = width + 1;
	  p = (char *)realloc(p, str_len * sizeof(char));
	}
      spaces = width - len;
      p[width] = '\0';
      memmove((void *)(p + spaces), (void *)p, len);
      memset((void *)p, (int)' ', spaces);
      (*nlen) = width;
    }
  else (*nlen) = len;
  return(p);
}


char *s7_number_to_string(s7_scheme *sc, s7_pointer obj, int32_t radix)
{
  int32_t nlen = 0;
  return(number_to_string_with_radix(sc, obj, radix, 0, 20, 'g', &nlen));
  /* (log top 10) so we get all the digits in base 10 (??) */
}

static s7_pointer number_to_string_p(s7_pointer p)
{
  int32_t nlen = 0;
  char *res;
  if (!is_number(p))
    simple_wrong_type_argument_with_type(cur_sc, cur_sc->number_to_string_symbol, p, a_number_string);
  res = number_to_string_base_10(p, 0, 20, 'g', &nlen, USE_WRITE);
  return(s7_make_string_with_length(cur_sc, res, nlen));
}


static void prepare_temporary_string(s7_scheme *sc, int32_t len, int32_t which)
{
  s7_pointer p;
  p = sc->tmp_strs[which];
  if (len > string_temp_true_length(p))
    {
      string_value(p) = (char *)realloc(string_value(p), len * sizeof(char));
      string_temp_true_length(p) = len;
    }
}

static s7_pointer g_number_to_string_1(s7_scheme *sc, s7_pointer args, bool temporary)
{
  #define H_number_to_string "(number->string num (radix 10)) converts the number num into a string."
  #define Q_number_to_string s7_make_signature(sc, 3, sc->is_string_symbol, sc->is_number_symbol, sc->is_integer_symbol)

  s7_int radix = 10;
  int32_t size, nlen = 0;
  char *res;
  s7_pointer x;

  x = car(args);
  if (!s7_is_number(x))
    method_or_bust_with_type(sc, x, sc->number_to_string_symbol, args, a_number_string, 1);

  if (is_pair(cdr(args)))
    {
      s7_pointer y;
      y = cadr(args);
      if (s7_is_integer(y))
	radix = s7_integer(y);
      else method_or_bust(sc, y, sc->number_to_string_symbol, args, T_INTEGER, 2);
      if ((radix < 2) || (radix > 16))
	return(out_of_range(sc, sc->number_to_string_symbol, small_int(2), y, a_valid_radix_string));
    }

#if WITH_GMP
  if (s7_is_bignum(x))
    {
      res = big_number_to_string_with_radix(x, radix, 0, &nlen, USE_WRITE);
      return(make_string_uncopied_with_length(sc, res, nlen));
    }
#endif

  size = float_format_precision;
  if (!is_rational(x))
    {
      /* if size = 20, (number->string .1) gives "0.10000000000000000555", but if it's less than 20,
       *    large numbers (or very small numbers) mess up the less significant digits.
       */
      if (radix == 10)
	{
	  if (is_real(x))
	    {
	      s7_double val;
	      val = fabs(s7_real(x));
	      if ((val > (s7_int32_max / 4)) || (val < 1.0e-6))
		size += 4;
	    }
	  else
	    {
	      s7_double rl;
	      rl = fabs(s7_real_part(x));
	      if ((rl > (s7_int32_max / 4)) || (rl < 1.0e-6))
		{
		  s7_double im;
		  im = fabs(s7_imag_part(x));
		  if ((im > (s7_int32_max / 4)) || (im < 1.0e-6))
		    size += 4;
		}
	    }
	}
    }
  if (radix != 10)
    {
      res = number_to_string_with_radix(sc, x, radix, 0, size, 'g', &nlen);
      return(make_string_uncopied_with_length(sc, res, nlen));
    }
  res = number_to_string_base_10(x, 0, size, 'g', &nlen, USE_WRITE);
  if (temporary)
    {
      s7_pointer p;
      prepare_temporary_string(sc, nlen + 1, 1);
      p = sc->tmp_strs[1];
      string_length(p) = nlen;
      memcpy((void *)(string_value(p)), (void *)res, nlen);
      string_value(p)[nlen] = 0;
      return(p);
    }
  return(s7_make_string_with_length(sc, res, nlen));
}

static s7_pointer g_number_to_string(s7_scheme *sc, s7_pointer args)
{
  return(g_number_to_string_1(sc, args, false));
}

static s7_pointer number_to_string_p_p(s7_pointer p) {return(g_number_to_string_1(cur_sc, set_plist_1(cur_sc, p), false));}
static s7_pointer number_to_string_p_pp(s7_pointer p1, s7_pointer p2) {return(g_number_to_string_1(cur_sc, set_plist_2(cur_sc, p1, p2), false));}

static s7_pointer number_to_string_temp;
static s7_pointer g_number_to_string_temp(s7_scheme *sc, s7_pointer args)
{
  return(g_number_to_string_1(sc, args, true));
}



#define CTABLE_SIZE 256
static bool *exponent_table, *slashify_table, *char_ok_in_a_name, *white_space, *number_table, *symbol_slashify_table;
static int32_t *digits;

static void init_ctables(void)
{
  int32_t i;

  exponent_table = (bool *)calloc(CTABLE_SIZE, sizeof(bool));
  slashify_table = (bool *)calloc(CTABLE_SIZE, sizeof(bool));
  symbol_slashify_table = (bool *)calloc(CTABLE_SIZE, sizeof(bool));
  char_ok_in_a_name = (bool *)calloc(CTABLE_SIZE, sizeof(bool));
  white_space = (bool *)calloc(CTABLE_SIZE + 1, sizeof(bool));
  white_space++;      /* leave white_space[-1] false for white_space[EOF] */
  number_table = (bool *)calloc(CTABLE_SIZE, sizeof(bool));

  for (i = 1; i < CTABLE_SIZE; i++)
    char_ok_in_a_name[i] = true;
  char_ok_in_a_name[0] = false;
  char_ok_in_a_name[(unsigned char)'('] = false;  /* idiotic cast is for C++'s benefit */
  char_ok_in_a_name[(unsigned char)')'] = false;
  char_ok_in_a_name[(unsigned char)';'] = false;
  char_ok_in_a_name[(unsigned char)'\t'] = false;
  char_ok_in_a_name[(unsigned char)'\n'] = false;
  char_ok_in_a_name[(unsigned char)'\r'] = false;
  char_ok_in_a_name[(unsigned char)' '] = false;
  char_ok_in_a_name[(unsigned char)'"'] = false;
  /* what about stuff like vertical tab?  or comma? */

  for (i = 0; i < CTABLE_SIZE; i++)
    white_space[i] = false;
  white_space[(unsigned char)'\t'] = true;
  white_space[(unsigned char)'\n'] = true;
  white_space[(unsigned char)'\r'] = true;
  white_space[(unsigned char)'\f'] = true;
  white_space[(unsigned char)'\v'] = true;
  white_space[(unsigned char)' '] = true;
  white_space[(unsigned char)'\205'] = true; /* 133 */
  white_space[(unsigned char)'\240'] = true; /* 160 */

  /* surely only 'e' is needed... */
  exponent_table[(unsigned char)'e'] = true; exponent_table[(unsigned char)'E'] = true;
  exponent_table[(unsigned char)'@'] = true;
#if WITH_EXTRA_EXPONENT_MARKERS
  exponent_table[(unsigned char)'s'] = true; exponent_table[(unsigned char)'S'] = true;
  exponent_table[(unsigned char)'f'] = true; exponent_table[(unsigned char)'F'] = true;
  exponent_table[(unsigned char)'d'] = true; exponent_table[(unsigned char)'D'] = true;
  exponent_table[(unsigned char)'l'] = true; exponent_table[(unsigned char)'L'] = true;
#endif

  for (i = 0; i < 32; i++)
    slashify_table[i] = true;
  for (i = 127; i < 160; i++)
    slashify_table[i] = true;
  slashify_table[(unsigned char)'\\'] = true;
  slashify_table[(unsigned char)'"'] = true;
  slashify_table[(unsigned char)'\n'] = false;

  for (i = 0; i < CTABLE_SIZE; i++)
    symbol_slashify_table[i] = ((slashify_table[i]) || (!char_ok_in_a_name[i]));

  digits = (int32_t *)calloc(CTABLE_SIZE, sizeof(int32_t));
  for (i = 0; i < CTABLE_SIZE; i++)
    digits[i] = 256;

  digits[(unsigned char)'0'] = 0; digits[(unsigned char)'1'] = 1; digits[(unsigned char)'2'] = 2; digits[(unsigned char)'3'] = 3; digits[(unsigned char)'4'] = 4;
  digits[(unsigned char)'5'] = 5; digits[(unsigned char)'6'] = 6; digits[(unsigned char)'7'] = 7; digits[(unsigned char)'8'] = 8; digits[(unsigned char)'9'] = 9;
  digits[(unsigned char)'a'] = 10; digits[(unsigned char)'A'] = 10;
  digits[(unsigned char)'b'] = 11; digits[(unsigned char)'B'] = 11;
  digits[(unsigned char)'c'] = 12; digits[(unsigned char)'C'] = 12;
  digits[(unsigned char)'d'] = 13; digits[(unsigned char)'D'] = 13;
  digits[(unsigned char)'e'] = 14; digits[(unsigned char)'E'] = 14;
  digits[(unsigned char)'f'] = 15; digits[(unsigned char)'F'] = 15;

  for (i = 0; i < CTABLE_SIZE; i++)
    number_table[i] = false;
  number_table[(unsigned char)'0'] = true;
  number_table[(unsigned char)'1'] = true;
  number_table[(unsigned char)'2'] = true;
  number_table[(unsigned char)'3'] = true;
  number_table[(unsigned char)'4'] = true;
  number_table[(unsigned char)'5'] = true;
  number_table[(unsigned char)'6'] = true;
  number_table[(unsigned char)'7'] = true;
  number_table[(unsigned char)'8'] = true;
  number_table[(unsigned char)'9'] = true;
  number_table[(unsigned char)'.'] = true;
  number_table[(unsigned char)'+'] = true;
  number_table[(unsigned char)'-'] = true;
  number_table[(unsigned char)'#'] = true;
}


#define is_white_space(C) white_space[C]
  /* this is much faster than C's isspace, and does not depend on the current locale.
   * if c == EOF (-1), it indexes into the empty (0) slot we preallocated below white_space
   */


static s7_pointer check_sharp_readers(s7_scheme *sc, const char *name)
{
  s7_pointer reader, value, args;
  bool need_loader_port;
  value = sc->F;
  args = sc->F;

  /* *#reader* is assumed to be an alist of (char . proc)
   *    where each proc takes one argument, the string from just beyond the "#" to the next delimiter.
   *    The procedure can call read-char to read ahead in the current-input-port.
   *    If it returns anything other than #f, that is the value of the sharp expression.
   *    Since #f means "nothing found", it is tricky to handle #F: 
   *       (cons #\F (lambda (str) (and (string=? str "F") (list 'not #t))))
   * This search happens after #|, #t, and #f (and #nD for multivectors?). #! has a fallback.
   */

  need_loader_port = is_loader_port(sc->input_port);
  if (need_loader_port)
    clear_loader_port(sc->input_port);

  /* normally read* can't read from sc->input_port if it is in use by the loader,
   *   but here we are deliberately making that possible.
   */
  for (reader = slot_value(sc->sharp_readers); is_not_null(reader); reader = cdr(reader))
    {
      if (name[0] == s7_character(caar(reader)))
	{
	  if (args == sc->F)
	    args = list_1(sc, s7_make_string(sc, name));
	  /* args is GC protected by s7_apply_function?? (placed on the stack) */
	  value = s7_apply_function(sc, cdar(reader), args); /* this is much less error-safe than s7_call */
	  if (value != sc->F)
	    break;
	}
    }
  if (need_loader_port)
    set_loader_port(sc->input_port);
  return(value);
}


static s7_pointer g_sharp_readers_set(s7_scheme *sc, s7_pointer args)
{
  /* new value must be either () or a proper list of conses (char . func) */
  if (is_null(cadr(args))) return(cadr(args));
  if (is_pair(cadr(args)))
    {
      s7_pointer x;
      for (x = cadr(args); is_pair(x); x = cdr(x))
	{
	  if ((!is_pair(car(x))) ||
	      (!s7_is_character(caar(x))) ||
	      (!s7_is_procedure(cdar(x))))
	    return(sc->error_symbol);
	}
      if (is_null(x))
	return(cadr(args));
    }
  return(sc->error_symbol);
}


static bool is_abnormal(s7_pointer x)
{
  switch (type(x))
    {
    case T_INTEGER:
    case T_RATIO:
      return(false);

    case T_REAL:
      return(is_inf(real(x)) ||
	     is_NaN(real(x)));

    case T_COMPLEX:
      return(((is_inf(s7_real_part(x)))  ||
	      (is_inf(s7_imag_part(x)))  ||
	      (is_NaN(s7_real_part(x))) ||
	      (is_NaN(s7_imag_part(x)))));

#if WITH_GMP
    case T_BIG_INTEGER:
    case T_BIG_RATIO:
      return(false);

    case T_BIG_REAL:
      return((is_inf(s7_real_part(x))) ||
	     (is_NaN(s7_real_part(x))));

    case T_BIG_COMPLEX:
      return((is_inf(s7_real_part(x))) ||
	     (is_inf(s7_imag_part(x))) ||
	     (is_NaN(s7_real_part(x))) ||
	     (is_NaN(s7_imag_part(x))));
#endif

    default:
      return(true);
    }
}

static s7_pointer unknown_sharp_constant(s7_scheme *sc, char *name)
{
  /* check *read-error-hook* */
  if (hook_has_functions(sc->read_error_hook))
    {
      s7_pointer result;
      result = s7_call(sc, sc->read_error_hook, list_2(sc, sc->T, s7_make_string_wrapper(sc, name)));
      if (result != sc->unspecified)
	return(result);
    }
  return(sc->nil);
}

#define SYMBOL_OK true
#define NO_SYMBOLS false

static s7_pointer make_sharp_constant(s7_scheme *sc, char *name, int32_t radix, bool with_error)
{
  /* name is the stuff after the '#', return sc->nil if not a recognized #... entity */
  s7_pointer x;

  if ((name[0] == 't') &&
      ((name[1] == '\0') || (strings_are_equal(name, "true"))))
    return(sc->T);

  if ((name[0] == 'f') &&
      ((name[1] == '\0') || (strings_are_equal(name, "false"))))
    return(sc->F);

  if (is_not_null(slot_value(sc->sharp_readers)))
    {
      x = check_sharp_readers(sc, name);
      if (x != sc->F)
	return(x);
    }

  if ((name[0] == '\0') || name[1] == '\0')
    return(unknown_sharp_constant(sc, name));

  switch (name[0])
    {
      /* -------- #< ... > -------- */
    case '<':
      if (strings_are_equal(name, "<unspecified>"))
	return(sc->unspecified);

      if (strings_are_equal(name, "<undefined>"))
	return(sc->undefined);

      if (strings_are_equal(name, "<eof>"))
	return(sc->eof_object);

      return(unknown_sharp_constant(sc, name));


      /* -------- #o #d #x #b -------- */
    case 'o':   /* #o (octal) */
    case 'x':   /* #x (hex) */
    case 'b':   /* #b (binary) */
      {
	int32_t num_at = 1;
	/* the #b or whatever overrides any radix passed in earlier */
	x = make_atom(sc, (char *)(name + num_at), (name[0] == 'o') ? 8 : ((name[0] == 'x') ? 16 : ((name[0] == 'b') ? 2 : 10)), NO_SYMBOLS, with_error);
	if (is_abnormal(x))
	  return(unknown_sharp_constant(sc, name));
	return(x);
      }
      break;


      /* -------- #_... -------- */
    case '_':
      {
	s7_pointer sym;
	sym = make_symbol(sc, (char *)(name + 1));
	if (is_slot(initial_slot(sym)))
	  return(slot_value(initial_slot(sym)));
	return(s7_error(sc, sc->syntax_error_symbol, set_elist_2(sc, s7_make_string_wrapper(sc, "#~A is undefined"), s7_make_string_wrapper(sc, name))));
	/* return(sc->undefined); */
      }


      /* -------- #\... -------- */
    case '\\':
      if (name[2] == 0)                             /* the most common case: #\a */
	return(chars[(unsigned char)(name[1])]);
      /* not uint32_t here!  (uint32_t)255 (as a char) returns -1!! */
      switch (name[1])
	{
	case 'n':
	  if ((strings_are_equal(name + 1, "null")) ||
	      (strings_are_equal(name + 1, "nul")))
	    return(chars[0]);

	  if (strings_are_equal(name + 1, "newline"))
	    return(chars[(unsigned char)'\n']);
	  break;

	case 's':
	  if (strings_are_equal(name + 1, "space"))
	    return(chars[(unsigned char)' ']);
	  break;

	case 'r':
	  if (strings_are_equal(name + 1, "return"))
	    return(chars[(unsigned char)'\r']);
	  break;

	case 'l':
	  if (strings_are_equal(name + 1, "linefeed"))
	    return(chars[(unsigned char)'\n']);
	  break;

	case 't':
	  if (strings_are_equal(name + 1, "tab"))
	    return(chars[(unsigned char)'\t']);
	  break;

	case 'a':
	  /* the next 4 are for r7rs */
	  if (strings_are_equal(name + 1, "alarm"))
	    return(chars[7]);
	  break;

	case 'b':
	  if (strings_are_equal(name + 1, "backspace"))
	    return(chars[8]);
	  break;

	case 'e':
	  if (strings_are_equal(name + 1, "escape"))
	    return(chars[0x1b]);
	  break;

	case 'd':
	  if (strings_are_equal(name + 1, "delete"))
	    return(chars[0x7f]);
	  break;

	case 'x':
	  /* #\x is just x, but apparently #\x<num> is int->char? #\x65 -> #\e, and #\xcebb is lambda? */
	  {
	    /* sscanf here misses errors like #\x1.4, but make_atom misses #\x6/3,
	     *   #\x#b0, #\x#e0.0, #\x-0, #\x#e0e100 etc, so we have to do it at an even lower level.
	     * another problem: #\xbdca2cbec overflows so lval is -593310740 -> segfault unless caught
	     */
	    bool happy = true;
	    char *tmp;
	    int32_t lval = 0;

	    tmp = (char *)(name + 2);
	    while ((*tmp) && (happy) && (lval >= 0))
	      {
		int32_t dig;
		dig = digits[(int32_t)(*tmp++)];
		if (dig < 16)
		  lval = dig + (lval * 16);
		else happy = false;
	      }
	    if ((happy) &&
		(lval < 256) &&
		(lval >= 0))
	      return(chars[lval]);
	  }
	  break;
	}
    }
  return(unknown_sharp_constant(sc, name));
}


static s7_int string_to_integer(const char *str, int32_t radix, bool *overflow)
{
  bool negative = false;
  s7_int lval = 0;
  int32_t dig;
  char *tmp = (char *)str;
  char *tmp1;

  if (str[0] == '+')
    tmp++;
  else
    {
      if (str[0] == '-')
	{
	  negative = true;
	  tmp++;
	}
    }
  while (*tmp == '0') {tmp++;};
  tmp1 = tmp;

 if (radix == 10)
    {
      while (true)
	{
	  dig = digits[(unsigned char)(*tmp++)];
	  if (dig > 9) break;
#if HAVE_OVERFLOW_CHECKS
	  if (multiply_overflow(lval, (s7_int)10, &lval)) break;
	  if (add_overflow(lval, (s7_int)dig, &lval)) break;
#else
	  lval = dig + (lval * 10);
	  dig = digits[(unsigned char)(*tmp++)];
	  if (dig > 9) break;
	  lval = dig + (lval * 10);
#endif
	}
    }
  else
    {
      while (true)
	{
	  dig = digits[(unsigned char)(*tmp++)];
	  if (dig >= radix) break;
#if HAVE_OVERFLOW_CHECKS
	  if (multiply_overflow(lval, (s7_int)radix, &lval)) break;
	  if (add_overflow(lval, (s7_int)dig, &lval)) break;
#else
	  lval = dig + (lval * radix);
	  dig = digits[(unsigned char)(*tmp++)];
	  if (dig >= radix) break;
	  lval = dig + (lval * radix);
#endif
	}
    }

#if WITH_GMP
  (*overflow) = ((lval > s7_int32_max) ||
		 ((tmp - tmp1) > s7_int_digits_by_radix[radix]));
  /* this tells the string->number readers to create a bignum.  We need to be very
   *    conservative here to catch contexts such as (/ 1/524288 19073486328125)
   */
#else
  if ((tmp - tmp1 - 2) > s7_int_digits_by_radix[radix])
    {
      /* I can't decide what to do with these non-gmp overflows.  Perhaps NAN in all cases?
       *     overflow: 9223372036854775810 -> -9223372036854775806 -- this is not caught currently
       */
      (*overflow) = true;
      if (negative)
	return(s7_int_min);       /* or INFINITY? */
      return(s7_int_max);         /* 0/100000000000000000000000000000000000000000000000000000000000000000000 */
    }
#endif

  if (negative)
    return(-lval);
  return(lval);
}


/*  9223372036854775807                9223372036854775807
 * -9223372036854775808               -9223372036854775808
 * 0000000000000000000000000001.0     1.0
 * 1.0000000000000000000000000000     1.0
 * 1000000000000000000000000000.0e-40 1.0e-12
 * 0.0000000000000000000000000001e40  1.0e12
 * 1.0e00000000000000000001           10.0
 */

static s7_double string_to_double_with_radix(const char *ur_str, int32_t radix, bool *overflow)
{
  /* strtod follows LANG which is not what we want (only "." is decimal point in Scheme).
   *   To overcome LANG in strtod would require screwing around with setlocale which never works.
   *   So we use our own code -- according to valgrind, this function is much faster than strtod.
   *
   * comma as decimal point causes ambiguities: `(+ ,1 2) etc
   */

  int32_t i, sign = 1, frac_len, int_len, dig, max_len, exponent = 0;
  int64_t int_part = 0, frac_part = 0;
  char *str;
  char *ipart, *fpart;
  s7_double dval = 0.0;

  /* there's an ambiguity in number notation here if we allow "1e1" or "1.e1" in base 16 (or 15) -- is e a digit or an exponent marker?
   *   but 1e+1, for example disambiguates it -- kind of messy! -- the scheme spec says "e" can only occur in base 10.
   *   mpfr says "e" as exponent only in bases <= 10 -- else use '@' which works in any base.  This can only cause confusion
   *   in scheme, unfortunately, due to the idiotic scheme polar notation.  But we accept "s" and "l" as exponent markers
   *   so, perhaps for radix > 10, the exponent, if any, has to use one of S s L l?  Not "l"!  And "s" originally meant "short".
   *
   * '@' can now be used as the exponent marker (26-Mar-12).
   * Another slight ambiguity: 1+1/2i is parsed as 1 + 0.5i, not 1+1/(2i), or (1+1)/(2i) or (1+1/2)i etc
   */

  max_len = s7_int_digits_by_radix[radix];
  str = (char *)ur_str;

  if (*str == '+')
    str++;
  else
    {
      if (*str == '-')
	{
	  str++;
	  sign = -1;
	}
    }
  while (*str == '0') {str++;};

  ipart = str;
  while (digits[(int32_t)(*str)] < radix) str++;
  int_len = str - ipart;

  if (*str == '.') str++;
  fpart = str;
  while (digits[(int32_t)(*str)] < radix) str++;
  frac_len = str - fpart;

  if ((*str) && (exponent_table[(unsigned char)(*str)]))
    {
      int32_t exp_negative = false;
      str++;
      if (*str == '+')
	str++;
      else
	{
	  if (*str == '-')
	    {
	      str++;
	      exp_negative = true;
	    }
	}
      while ((dig = digits[(int32_t)(*str++)]) < 10) /* exponent itself is always base 10 */
	{
#if HAVE_OVERFLOW_CHECKS
	  if ((int_multiply_overflow(exponent, 10, &exponent)) ||
	      (int_add_overflow(exponent, dig, &exponent)))
	    {
	      exponent = 1000000; /* see below */
	      break;
	    }
#else
	  exponent = dig + (exponent * 10);
#endif
	}
#if (!defined(__GNUC__)) || (__GNUC__ < 5)
      if (exponent < 0)         /* we overflowed, so make sure we notice it below (need to check for 0.0e... first) (Brian Damgaard) */
	exponent = 1000000;     /*   see below for examples -- this number needs to be very big but not too big for add */
#endif
      if (exp_negative)
	exponent = -exponent;

      /*           2e12341234123123123123213123123123 -> 0.0
       * but exp len is not the decider: 2e00000000000000000000000000000000000000001 -> 20.0
       * first zero: 2e123412341231231231231
       * then:     2e12341234123123123123123123 -> inf
       * then:     2e123412341231231231231231231231231231 -> 0.0
       *           2e-123412341231231231231 -> inf
       * but:      0e123412341231231231231231231231231231
       */
    }

#if WITH_GMP
  /* 9007199254740995.0 */
  if (int_len + frac_len >= max_len)
    {
      (*overflow) = true;
      return(0.0);
    }
#endif

  str = ipart;
  if ((int_len + exponent) > max_len)
    {
      /*  12341234.56789e12                   12341234567889999872.0              1.234123456789e+19
       * -1234567890123456789.0              -1234567890123456768.0              -1.2345678901235e+18
       *  12345678901234567890.0              12345678901234567168.0              1.2345678901235e+19
       *  123.456e30                          123456000000000012741097792995328.0 1.23456e+32
       *  12345678901234567890.0e12           12345678901234569054409354903552.0  1.2345678901235e+31
       *  1.234567890123456789012e30          1234567890123456849145940148224.0   1.2345678901235e+30
       *  1e20                                100000000000000000000.0             1e+20
       *  1234567890123456789.0               1234567890123456768.0               1.2345678901235e+18
       *  123.456e16                          1234560000000000000.0               1.23456e+18
       *  98765432101234567890987654321.0e-5  987654321012345728401408.0          9.8765432101235e+23
       *  98765432101234567890987654321.0e-10 9876543210123456512.0               9.8765432101235e+18
       *  0.00000000000000001234e20           1234.0
       *  0.000000000000000000000000001234e30 1234.0
       *  0.0000000000000000000000000000000000001234e40 1234.0
       *  0.000000000012345678909876543210e15 12345.678909877
       *  0e1000                              0.0
       */

      for (i = 0; i < max_len; i++)
	{
	  dig = digits[(int32_t)(*str++)];
	  if (dig < radix)
	    int_part = dig + (int_part * radix);
	  else break;
	}

      /* if the exponent is huge, check for 0 int_part and frac_part before complaining (0e1000 or 0.0e1000)
       */
      if ((int_part == 0) &&
	  (exponent > max_len))
	{
	  /* if frac_part is also 0, return 0.0 */
	  if (frac_len == 0)
	    return(0.0);

	  str = fpart;
	  while ((dig = digits[(int32_t)(*str++)]) < radix)
	    frac_part = dig + (frac_part * radix);
	  if (frac_part == 0)
	    return(0.0);

#if WITH_GMP
	  (*overflow) = true;
#endif
	}

#if WITH_GMP
      (*overflow) = ((int_part > 0) || (exponent > 20));    /* .1e310 is a tricky case */
#endif

      if (int_part != 0) /* 0.<310 zeros here>1e310 for example --
			  *   pow (via ipow) thinks it has to be too big, returns Nan,
			  *   then Nan * 0 -> Nan and the NaN propagates
			  */
	{
	  if (int_len <= max_len)
	    dval = int_part * ipow(radix, exponent);
	  else dval = int_part * ipow(radix, exponent + int_len - max_len);
	}
      else dval = 0.0;

      /* shift by exponent, but if int_len > max_len then we assumed (see below) int_len - max_len 0's on the left */
      /*   using int_to_int or table lookups here instead of pow did not make any difference in speed */

      if (int_len < max_len)
	{
	  int32_t k, flen;
	  str = fpart;

	  for (k = 0; (frac_len > 0) && (k < exponent); k += max_len)
	    {
	      if (frac_len > max_len) flen = max_len; else flen = frac_len;
	      frac_len -= max_len;

	      frac_part = 0;
	      for (i = 0; i < flen; i++)
		frac_part = digits[(int32_t)(*str++)] + (frac_part * radix);

	      if (frac_part != 0)                                /* same pow->NaN problem as above can occur here */
		dval += frac_part * ipow(radix, exponent - flen - k);
	    }
	}
      else
	{
	  /* some of the fraction is in the integer part before the negative exponent shifts it over */
	  if (int_len > max_len)
	    {
	      int32_t ilen;
	      /* str should be at the last digit we read */
	      ilen = int_len - max_len;                          /* we read these above */
	      if (ilen > max_len)
		ilen = max_len;

	      for (i = 0; i < ilen; i++)
		frac_part = digits[(int32_t)(*str++)] + (frac_part * radix);

	      dval += frac_part * ipow(radix, exponent - ilen);
	    }
	}

      return(sign * dval);
    }

  /* int_len + exponent <= max_len */

  if (int_len <= max_len)
    {
      int32_t int_exponent;

      /* a better algorithm (since the inaccuracies are in the radix^exponent portion):
       *   strip off leading zeros and possible sign,
       *   strip off digits beyond max_len, then remove any trailing zeros.
       *     (maybe fiddle with the lowest order digit here for rounding, but I doubt it matters)
       *   read digits until end of number or max_len reached, ignoring the decimal point
       *   get exponent and use it and decimal point location to position the current result integer
       * this always combines the same integer and the same exponent no matter how the number is expressed.
       */

      int_exponent = exponent;
      if (int_len > 0)
	{
	  char *iend;
	  iend = (char *)(str + int_len - 1);
	  while ((*iend == '0') && (iend != str)) {iend--; int_exponent++;}

	  while (str <= iend)
	    int_part = digits[(int32_t)(*str++)] + (int_part * radix);
	}
      if (int_exponent != 0)
	dval = int_part * ipow(radix, int_exponent);
      else dval = (s7_double)int_part;
    }
  else
    {
      int32_t len, flen;
      int64_t frpart = 0;

      /* 98765432101234567890987654321.0e-20    987654321.012346
       * 98765432101234567890987654321.0e-29    0.98765432101235
       * 98765432101234567890987654321.0e-30    0.098765432101235
       * 98765432101234567890987654321.0e-28    9.8765432101235
       */

      len = int_len + exponent;
      for (i = 0; i < len; i++)
	int_part = digits[(int32_t)(*str++)] + (int_part * radix);

      flen = -exponent;
      if (flen > max_len)
	flen = max_len;

      for (i = 0; i < flen; i++)
	frpart = digits[(int32_t)(*str++)] + (frpart * radix);

      if (len <= 0)
	dval = int_part + frpart * ipow(radix, len - flen);
      else dval = int_part + frpart * ipow(radix, -flen);
    }

  if (frac_len > 0)
    {
      str = fpart;
      if (frac_len <= max_len)
	{
	  /* splitting out base 10 case saves very little here */
	  /* this ignores trailing zeros, so that 0.3 equals 0.300 */
	  char *fend;

	  fend = (char *)(str + frac_len - 1);
	  while ((*fend == '0') && (fend != str)) {fend--; frac_len--;} /* (= .6 0.6000) */

	  while (str <= fend)
	    frac_part = digits[(int32_t)(*str++)] + (frac_part * radix);
	  dval += frac_part * ipow(radix, exponent - frac_len);

	  /* fprintf(stderr, "frac: %" PRId64 ", exp: (%d %d) %.20f, val: %.20f\n", frac_part, exponent, frac_len, ipow(radix, exponent - frac_len), dval);
	   * 0.6:    frac:    6, exp: 0.10000000000000000555, val: 0.60000000000000008882
	   * 0.60:   frac:   60, exp: 0.01000000000000000021, val: 0.59999999999999997780
	   * 0.6000: frac: 6000, exp: 0.00010000000000000000, val: 0.59999999999999997780
	   * :(= 0.6 0.60)
	   * #f
	   * :(= #i3/5 0.6)
	   * #f
	   * so (string->number (number->string num)) == num only if both num's are the same text (or you get lucky)
	   * :(= 0.6 6e-1) ; but not 60e-2
	   * #t
	   *
	   * to fix the 0.60 case, we need to ignore trailing post-dot zeros.
	   */
	}
      else
	{
	  if (exponent <= 0)
	    {
	      for (i = 0; i < max_len; i++)
		frac_part = digits[(int32_t)(*str++)] + (frac_part * radix);

	      dval += frac_part * ipow(radix, exponent - max_len);
	    }
	  else
	    {
	      /* 1.0123456789876543210e1         10.12345678987654373771
	       * 1.0123456789876543210e10        10123456789.87654304504394531250
	       * 0.000000010000000000000000e10   100.0
	       * 0.000000010000000000000000000000000000000000000e10 100.0
	       * 0.000000012222222222222222222222222222222222222e10 122.22222222222222
	       * 0.000000012222222222222222222222222222222222222e17 1222222222.222222
	       */

	      int_part = 0;
	      for (i = 0; i < exponent; i++)
		int_part = digits[(int32_t)(*str++)] + (int_part * radix);

	      frac_len -= exponent;
	      if (frac_len > max_len)
		frac_len = max_len;

	      for (i = 0; i < frac_len; i++)
		frac_part = digits[(int32_t)(*str++)] + (frac_part * radix);

	      dval += int_part + frac_part * ipow(radix, -frac_len);
	    }
	}
    }

#if WITH_GMP
  if ((int_part == 0) &&
      (frac_part == 0))
    return(0.0);
  (*overflow) = ((frac_len - exponent) > max_len);
#endif

  return(sign * dval);
}


static s7_pointer make_atom(s7_scheme *sc, char *q, int32_t radix, bool want_symbol, bool with_error)
{
  /* make symbol or number from string */
  #define IS_DIGIT(Chr, Rad) (digits[(unsigned char)Chr] < Rad)

  char c, *p;
  bool has_dec_point1 = false;

  p = q;
  c = *p++;

  /* a number starts with + - . or digit, but so does 1+ for example */

  switch (c)
    {
    case '#':
      return(make_sharp_constant(sc, p, radix, with_error)); /* make_sharp_constant expects the '#' to be removed */

    case '+':
    case '-':
      c = *p++;
      if (c == '.')
	{
	  has_dec_point1 = true;
	  c = *p++;
	}
      if ((!c) || (!IS_DIGIT(c, radix)))
	return((want_symbol) ? make_symbol(sc, q) : sc->F);
      break;

    case '.':
      has_dec_point1 = true;
      c = *p++;

      if ((!c) || (!IS_DIGIT(c, radix)))
	return((want_symbol) ? make_symbol(sc, q) : sc->F);
      break;

    case '0':        /* these two are always digits */
    case '1':
      break;

    default:
      if (!IS_DIGIT(c, radix))
	return((want_symbol) ? make_symbol(sc, q) : sc->F);
      break;
    }

  /* now it's possibly a number -- the first character(s) could be part of a number in the current radix */
  {
    char *slash1 = NULL, *slash2 = NULL, *plus = NULL, *ex1 = NULL, *ex2 = NULL;
    bool has_i = false, has_dec_point2 = false;
    int32_t has_plus_or_minus = 0, current_radix;

#if (!WITH_GMP)
    bool overflow = false;
#endif
    current_radix = radix;  /* current_radix is 10 for the exponent portions, but radix for all the rest */

    for ( ; (c = *p) != 0; ++p)
      {
	/* what about embedded null? (string->number (string #\1 (integer->char 0) #\0))
	 *   currently we stop and return 1, but Guile returns #f
	 */
	if (!IS_DIGIT(c, current_radix))         /* moving this inside the switch statement was much slower */
	  {
	    current_radix = radix;

	    switch (c)
	      {
		/* -------- decimal point -------- */
	      case '.':
		if ((!IS_DIGIT(p[1], current_radix)) &&
		    (!IS_DIGIT(p[-1], current_radix)))
		  return((want_symbol) ? make_symbol(sc, q) : sc->F);

		if (has_plus_or_minus == 0)
		  {
		    if ((has_dec_point1) || (slash1))
		      return((want_symbol) ? make_symbol(sc, q) : sc->F);
		    has_dec_point1 = true;
		  }
		else
		  {
		    if ((has_dec_point2) || (slash2))
		      return((want_symbol) ? make_symbol(sc, q) : sc->F);
		    has_dec_point2 = true;
		  }
		continue;


		/* -------- exponent marker -------- */
#if WITH_EXTRA_EXPONENT_MARKERS
		/* 1st 3d-perspective 0.0f 128.0f 3d 1s -- in 2 million lines of public scheme code, not one actual use! */
	      case 's': case 'S':
	      case 'd': case 'D':
	      case 'f': case 'F':
	      case 'l': case 'L':
#endif
	      case 'e': case 'E':
		if (current_radix > 10)
		  return((want_symbol) ? make_symbol(sc, q) : sc->F);
		/* see note above */
		/* fall through -- if '@' used, radices>10 are ok */

	      case '@':
		current_radix = 10;

		if (((ex1) ||
		     (slash1)) &&
		    (has_plus_or_minus == 0)) /* ee */
		  return((want_symbol) ? make_symbol(sc, q) : sc->F);

		if (((ex2) ||
		     (slash2)) &&
		    (has_plus_or_minus != 0)) /* 1+1.0ee */
		  return((want_symbol) ? make_symbol(sc, q) : sc->F);

		if ((!IS_DIGIT(p[-1], radix)) && /* was current_radix but that's always 10! */
		    (p[-1] != '.'))
		  return((want_symbol) ? make_symbol(sc, q) : sc->F);

		if (has_plus_or_minus == 0)
		  {
		    ex1 = p;
		    has_dec_point1 = true; /* decimal point illegal from now on */
		  }
		else
		  {
		    ex2 = p;
		    has_dec_point2 = true;
		  }
		p++;
		if ((*p == '-') || (*p == '+')) p++;
		if (IS_DIGIT(*p, current_radix))
		  continue;
		break;


		/* -------- internal + or - -------- */
	      case '+':
	      case '-':
		if (has_plus_or_minus != 0) /* already have the separator */
		  return((want_symbol) ? make_symbol(sc, q) : sc->F);

		if (c == '+') has_plus_or_minus = 1; else has_plus_or_minus = -1;
		plus = (char *)(p + 1);
		continue;

		/* ratio marker */
	      case '/':
		if ((has_plus_or_minus == 0) &&
		    ((ex1) ||
		     (slash1) ||
		     (has_dec_point1)))
		  return((want_symbol) ? make_symbol(sc, q) : sc->F);

		if ((has_plus_or_minus != 0) &&
		    ((ex2) ||
		     (slash2) ||
		     (has_dec_point2)))
		  return((want_symbol) ? make_symbol(sc, q) : sc->F);

		if (has_plus_or_minus == 0)
		  slash1 = (char *)(p + 1);
		else slash2 = (char *)(p + 1);

		if ((!IS_DIGIT(p[1], current_radix)) ||
		    (!IS_DIGIT(p[-1], current_radix)))
		  return((want_symbol) ? make_symbol(sc, q) : sc->F);

		continue;


		/* -------- i for the imaginary part -------- */
	      case 'i':
		if ((has_plus_or_minus != 0) &&
		    (!has_i))
		  {
		    has_i = true;
		    continue;
		  }
		break;

	      default:
		break;
	      }
	    return((want_symbol) ? make_symbol(sc, q) : sc->F);
	  }
      }

    if ((has_plus_or_minus != 0) &&        /* that is, we have an internal + or - */
	(!has_i))                          /*   but no i for the imaginary part */
      return((want_symbol) ? make_symbol(sc, q) : sc->F);

    if (has_i)
      {
#if (!WITH_GMP)
	s7_double rl = 0.0, im = 0.0;
#else
	char e1 = 0, e2 = 0;
#endif
	s7_pointer result;
	int32_t len;
	char ql1, pl1;

	len = safe_strlen(q);

	if (q[len - 1] != 'i')
	  return((want_symbol) ? make_symbol(sc, q) : sc->F);

	/* save original string */
	ql1 = q[len - 1];
	pl1 = (*(plus - 1));
#if WITH_GMP
	if (ex1) {e1 = *ex1; (*ex1) = '@';} /* for mpfr */
	if (ex2) {e2 = *ex2; (*ex2) = '@';}
#endif

	/* look for cases like 1+i */
	if ((q[len - 2] == '+') || (q[len - 2] == '-'))
	  q[len - 1] = '1';
	else q[len - 1] = '\0'; /* remove 'i' */

	(*((char *)(plus - 1))) = '\0';

	/* there is a slight inconsistency here:
	   1/0      -> nan.0
           1/0+0i   -> inf.0 (0/1+0i is 0.0)
	   #i1/0+0i -> inf.0
	   0/0      -> nan.0
	   0/0+0i   -> -nan.0
	*/

#if (!WITH_GMP)
	if ((has_dec_point1) ||
	    (ex1))
	  {
	    /* (string->number "1100.1+0.11i" 2) -- need to split into 2 honest reals before passing to non-base-10 str->dbl */
	    rl = string_to_double_with_radix(q, radix, &overflow);
	  }
	else
	  {
	    if (slash1)
	      {
		/* here the overflow could be innocuous if it's in the denominator and the numerator is 0
		 *    0/100000000000000000000000000000000000000-0i
		 */
		s7_int num, den;
		num = string_to_integer(q, radix, &overflow);
		den = string_to_integer(slash1, radix, &overflow);
		if (den == 0)
		  rl = NAN;
		else
		  {
		    if (num == 0)
		      {
			rl = 0.0;
			overflow = false;
		      }
		    else rl = (s7_double)num / (s7_double)den;
		  }
	      }
	    else rl = (s7_double)string_to_integer(q, radix, &overflow);
	    if (overflow) return(real_NaN);
	  }
	if (rl == -0.0) rl = 0.0;

	if ((has_dec_point2) ||
	    (ex2))
	  im = string_to_double_with_radix(plus, radix, &overflow);
	else
	  {
	    if (slash2)
	      {
		/* same as above: 0-0/100000000000000000000000000000000000000i
		 */
		s7_int num, den;
		num = string_to_integer(plus, radix, &overflow);
		den = string_to_integer(slash2, radix, &overflow);
		if (den == 0)
		  im = NAN;
		else
		  {
		    if (num == 0)
		      {
			im = 0.0;
			overflow = false;
		      }
		    else im = (s7_double)num / (s7_double)den;
		  }
	      }
	    else im = (s7_double)string_to_integer(plus, radix, &overflow);
	    if (overflow) return(real_NaN);
	  }
	if ((has_plus_or_minus == -1) &&
	    (im != 0.0))
	  im = -im;
	result = s7_make_complex(sc, rl, im);
#else
	result = string_to_either_complex(sc, q, slash1, ex1, has_dec_point1, plus, slash2, ex2, has_dec_point2, radix, has_plus_or_minus);
#endif

	/* restore original string */
	q[len - 1] = ql1;
	(*((char *)(plus - 1))) = pl1;
#if WITH_GMP
	if (ex1) (*ex1) = e1;
	if (ex2) (*ex2) = e2;
#endif

	return(result);
      }

    /* not complex */
    if ((has_dec_point1) ||
	(ex1))
      {
	s7_pointer result;

	if (slash1)  /* not complex, so slash and "." is not a number */
	  return((want_symbol) ? make_symbol(sc, q) : sc->F);

#if (!WITH_GMP)
	result = make_real(sc, string_to_double_with_radix(q, radix, &overflow));
#else
	{
	  char old_e = 0;
	  if (ex1)
	    {
	      old_e = (*ex1);
	      (*ex1) = '@';
	    }
	  result = string_to_either_real(sc, q, radix);
	  if (ex1)
	    (*ex1) = old_e;
	}
#endif
	return(result);
      }

    /* not real */
    if (slash1)
#if (!WITH_GMP)
      {
	s7_int n, d;

	n = string_to_integer(q, radix, &overflow);
	d = string_to_integer(slash1, radix, &overflow);

	if ((n == 0) && (d != 0))                        /* 0/100000000000000000000000000000000000000 */
	  return(small_int(0));
	if ((d == 0) || (overflow))
	  return(real_NaN);
	/* it would be neat to return 1 from 10000000000000000000000000000/10000000000000000000000000000
	 *   but q is the entire number ('/' included) and slash1 is the stuff after the '/', and every
	 *   big number comes through here, so there's no clean and safe way to check that q == slash1.
	 */
	return(s7_make_ratio(sc, n, d));
      }
#else
    return(string_to_either_ratio(sc, q, slash1, radix));
#endif

    /* integer */
#if (!WITH_GMP)
    {
      s7_int x;
      x = string_to_integer(q, radix, &overflow);
      if (overflow)
	return((q[0] == '-') ? real_minus_infinity : real_infinity);
      return(make_integer(sc, x));
    }
#else
    return(string_to_either_integer(sc, q, radix));
#endif
  }
}


static s7_pointer s7_string_to_number(s7_scheme *sc, char *str, int32_t radix)
{
  s7_pointer x;
  x = make_atom(sc, str, radix, NO_SYMBOLS, WITHOUT_OVERFLOW_ERROR);
  if (s7_is_number(x))  /* only needed because str might start with '#' and not be a number (#t for example) */
    return(x);
  return(sc->F);
}


static s7_pointer g_string_to_number_1(s7_scheme *sc, s7_pointer args, s7_pointer caller)
{
  #define H_string_to_number "(string->number str (radix 10)) converts str into a number. \
If str does not represent a number, string->number returns #f.  If 'str' has an embedded radix, \
the optional 'radix' argument is ignored: (string->number \"#x11\" 2) -> 17 not 3."
  #define Q_string_to_number s7_make_signature(sc, 3, s7_make_signature(sc, 2, sc->is_number_symbol, sc->is_boolean_symbol), sc->is_string_symbol, sc->is_integer_symbol)

  s7_int radix = 0;
  char *str;

  if (!is_string(car(args)))
    method_or_bust(sc, car(args), caller, args, T_STRING, 1);

  if (is_pair(cdr(args)))
    {
      s7_pointer rad, p;
      rad = cadr(args);
      if (!s7_is_integer(rad))
	{
	  if (!s7_is_integer(p = check_values(sc, rad, cdr(args))))
	    method_or_bust(sc, rad, caller, args, T_INTEGER, 2);
	  rad = p;
	}
      radix = s7_integer(rad);
      if ((radix < 2) ||              /* what about negative int32_t as base (Knuth), reals such as phi, and some complex like -1+i */
	  (radix > 16))               /* the only problem here is printing the number; perhaps put each digit in "()" in base 10: (123)(0)(34) */
	return(out_of_range(sc, caller, small_int(2), rad, a_valid_radix_string));
    }
  else radix = 10;

  str = (char *)string_value(car(args));
  if ((!str) || (!(*str)))
    return(sc->F);

  switch (str[0])
    {
    case 'n':
      if (safe_strcmp(str, "nan.0"))
	return(real_NaN);
      break;

    case 'i':
      if (safe_strcmp(str, "inf.0"))
	return(real_infinity);
      break;

    case '-':
      if ((str[1] == 'i') && (safe_strcmp((const char *)(str + 1), "inf.0")))
	return(real_minus_infinity);
      break;

    case '+':
      if ((str[1] == 'i') && (safe_strcmp((const char *)(str + 1), "inf.0")))
	return(real_infinity);
      break;
    }
  return(s7_string_to_number(sc, str, radix));
}


static s7_pointer g_string_to_number(s7_scheme *sc, s7_pointer args)
{
  return(g_string_to_number_1(sc, args, sc->string_to_number_symbol));
}


static bool numbers_are_eqv(s7_pointer a, s7_pointer b)
{
  if (type(a) != type(b)) /* (eqv? 1 1.0) -> #f! */
    return(false);

  switch (type(a))
    {
    case T_INTEGER:
      return((integer(a) == integer(b)));

    case T_RATIO:
      return((numerator(a) == numerator(b)) &&
	     (denominator(a) == denominator(b)));

    case T_REAL:
      if (is_NaN(real(a)))
	return(false);
      return(real(a) == real(b));

    case T_COMPLEX:
      if ((is_NaN(real_part(a))) ||
	  (is_NaN(imag_part(a))))
	return(false);
      return((real_part(a) == real_part(b)) &&
	     (imag_part(a) == imag_part(b)));

    default:
#if WITH_GMP
      if ((is_big_number(a)) || (is_big_number(b))) /* this can happen if (member bignum ...) -> memv */
	return(big_numbers_are_eqv(a, b));
#endif
      break;
    }
  return(false);
}


static bool is_rational_via_method(s7_scheme *sc, s7_pointer p)
{
  if (s7_is_rational(p))
    return(true);
  if (has_methods(p))
    {
      s7_pointer f;
      f = find_method(sc, find_let(sc, p), sc->is_rational_symbol);
      if (f != sc->undefined)
	return(is_true(sc, s7_apply_function(sc, f, cons(sc, p, sc->nil))));
    }
  return(false);
}


/* -------------------------------- abs -------------------------------- */
#if (!WITH_GMP)
static s7_pointer g_abs(s7_scheme *sc, s7_pointer args)
{
  #define H_abs "(abs x) returns the absolute value of the real number x"
  #define Q_abs s7_make_signature(sc, 2, sc->is_real_symbol, sc->is_real_symbol)  

  s7_pointer x;
  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:
      if (integer(x) < 0)
	{
	  if (integer(x) == s7_int_min)
	    return(make_integer(sc, s7_int_max));
	  return(make_integer(sc, -integer(x)));
	}
      return(x);

    case T_RATIO:
      if (numerator(x) < 0)
	{
	  if (numerator(x) == s7_int_min)
	    return(s7_make_ratio(sc, s7_int_max, denominator(x)));
	  return(s7_make_ratio(sc, -numerator(x), denominator(x)));
	}
      return(x);

    case T_REAL:
      if (is_NaN(real(x)))                  /* (abs -nan.0) -> nan.0, not -nan.0 */
	return(real_NaN);
      if (real(x) < 0.0)
	return(make_real(sc, -real(x)));
      return(x);

    default:
      method_or_bust_one_arg(sc, x, sc->abs_symbol, args, T_REAL);
    }
}

static s7_double abs_d_d(s7_double x) {return((x < 0.0) ? (-x) : x);} 
static s7_int abs_i_i(s7_int x) {return((x < 0.0) ? (-x) : x);} 


/* -------------------------------- magnitude -------------------------------- */

static double my_hypot(double x, double y)
{
  /* according to callgrind, this is much faster than libc's hypot */
  if (x == 0.0) return(fabs(y));
  if (y == 0.0) return(fabs(x));
  if (x == y) return(1.414213562373095 * fabs(x));
  if ((is_NaN(x)) || (is_NaN(y))) return(NAN);
  return(sqrt(x * x + y * y));
}

static s7_pointer g_magnitude(s7_scheme *sc, s7_pointer args)
{
  #define H_magnitude "(magnitude z) returns the magnitude of z"
  #define Q_magnitude s7_make_signature(sc, 2, sc->is_real_symbol, sc->is_number_symbol)
  s7_pointer x;
  x = car(args);

  switch (type(x))
    {
    case T_INTEGER:
      if (integer(x) == s7_int_min)
	return(make_integer(sc, s7_int_max));
      /* (magnitude -9223372036854775808) -> -9223372036854775808
       *   same thing happens in abs, lcm and gcd: (gcd -9223372036854775808) -> -9223372036854775808
       */
      if (integer(x) < 0)
        return(make_integer(sc, -integer(x)));
      return(x);

    case T_RATIO:
      if (numerator(x) < 0)
	return(s7_make_ratio(sc, -numerator(x), denominator(x)));
      return(x);

    case T_REAL:
      if (is_NaN(real(x)))                 /* (magnitude -nan.0) -> nan.0, not -nan.0 */
	return(real_NaN);
      if (real(x) < 0.0)
	return(make_real(sc, -real(x)));
      return(x);

    case T_COMPLEX:
      return(make_real(sc, my_hypot(imag_part(x), real_part(x))));

    default:
      method_or_bust_with_type_one_arg(sc, x, sc->magnitude_symbol, args, a_number_string);
    }
}

static s7_double magnitude_d_p(s7_pointer p) {return(s7_number_to_real_with_caller(cur_sc, g_magnitude(cur_sc, set_plist_1(cur_sc, p)), "magnitude"));}


/* -------------------------------- rationalize -------------------------------- */
static s7_pointer g_rationalize(s7_scheme *sc, s7_pointer args)
{
  #define H_rationalize "(rationalize x err) returns the ratio with lowest denominator within err of x"
  #define Q_rationalize s7_make_signature(sc, 3, sc->is_rational_symbol, sc->is_real_symbol, sc->is_real_symbol)
  s7_double err;
  s7_pointer x;

  x = car(args);
  if (!s7_is_real(x))
    method_or_bust(sc, x, sc->rationalize_symbol, args, T_REAL, 1);

  if (is_not_null(cdr(args)))
    {
      s7_pointer ex;
      ex = cadr(args);
      if (!s7_is_real(ex))
	method_or_bust(sc, ex, sc->rationalize_symbol, args, T_REAL, 2);

      err = real_to_double(sc, ex, "rationalize");
      if (is_NaN(err))
	return(out_of_range(sc, sc->rationalize_symbol, small_int(2), cadr(args), its_nan_string));
      if (err < 0.0) err = -err;
    }
  else err = sc->default_rationalize_error;

  switch (type(x))
    {
    case T_INTEGER:
      {
	s7_int a, b, pa;
	if (err < 1.0) return(x);
	a = s7_integer(x);
	if (a < 0) pa = -a; else pa = a;
	if (err >= pa) return(small_int(0));
	b = (s7_int)err;
	pa -= b;
	if (a < 0)
	  return(make_integer(sc, -pa));
	return(make_integer(sc, pa));
      }

    case T_RATIO:
      if (err == 0.0)
	return(x);

    case T_REAL:
      {
	s7_double rat;
	s7_int numer = 0, denom = 1;

	rat = real_to_double(sc, x, "rationalize");

	if ((is_NaN(rat)) || (is_inf(rat)))
	  return(wrong_type_argument_with_type(sc, sc->rationalize_symbol, 1, x, a_normal_real_string));

	if (err >= fabs(rat))
	  return(small_int(0));

	if ((rat > 9.2233720368548e+18) || (rat < -9.2233720368548e+18))
	  return(out_of_range(sc, sc->rationalize_symbol, small_int(1), x, its_too_large_string));

	if ((fabs(rat) + fabs(err)) < 1.0e-18)
	  err = 1.0e-18;
	/* (/ 1.0 most-positive-fixnum) is 1.0842021e-19, so if we let err be less than that,
	 * (rationalize 1e-19 1e-20) hangs, but this only affects the initial ceiling, I believe.
	 */

	if (fabs(rat) < fabs(err))
	  return(small_int(0));

	if (c_rationalize(rat, err, &numer, &denom))
	  return(s7_make_ratio(sc, numer, denom));

	return(sc->F);
      }
    }
  return(sc->F); /* make compiler happy */
}


/* -------------------------------- angle -------------------------------- */
static s7_pointer g_angle(s7_scheme *sc, s7_pointer args)
{
  #define H_angle "(angle z) returns the angle of z"
  #define Q_angle s7_make_signature(sc, 2, sc->is_real_symbol, sc->is_number_symbol)
  s7_pointer x;
  /* (angle inf+infi) -> 0.78539816339745 ? 
   *   I think this should be -pi < ang <= pi
   */

  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:
      if (integer(x) < 0)
	return(real_pi);
      return(small_int(0));

    case T_RATIO:
      if (numerator(x) < 0)
	return(real_pi);
      return(small_int(0));

    case T_REAL:
      if (is_NaN(real(x))) return(x);
      if (real(x) < 0.0)
	return(real_pi);
      return(real_zero);

    case T_COMPLEX:
      return(make_real(sc, atan2(imag_part(x), real_part(x))));

    default:
      method_or_bust_with_type_one_arg(sc, x, sc->angle_symbol, args, a_number_string);
    }
}

static s7_double angle_d_p(s7_pointer x)
{
  switch (type(x))
    {
    case T_INTEGER: if (integer(x) < 0)   return(M_PI); return(0.0); break;
    case T_RATIO:   if (numerator(x) < 0) return(M_PI); return(0.0); break;
    case T_REAL:    if (is_NaN(real(x)))  return(NAN); if (real(x) < 0.0) return(M_PI); return(0.0); break;
    case T_COMPLEX: return(atan2(imag_part(x), real_part(x))); break;
    default: simple_wrong_type_argument_with_type(cur_sc, cur_sc->angle_symbol, x, a_number_string); break;
    }
  return(0.0);
}


/* -------------------------------- make-polar -------------------------------- */
#if (!WITH_PURE_S7)
static s7_pointer g_make_polar(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x, y;
  s7_double ang, mag;
  #define H_make_polar "(make-polar mag ang) returns a complex number with magnitude mag and angle ang"
  #define Q_make_polar s7_make_signature(sc, 3, sc->is_number_symbol, sc->is_real_symbol, sc->is_real_symbol)

  x = car(args);
  y = cadr(args);

  switch (type(x))
    {
    case T_INTEGER:
      switch (type(y))
	{
	case T_INTEGER:
	  if (integer(x) == 0) return(x);            /* (make-polar 0 1) -> 0 */
	  if (integer(y) == 0) return(x);            /* (make-polar 1 0) -> 1 */
	  mag = (s7_double)integer(x);
	  ang = (s7_double)integer(y);
	  break;

	case T_RATIO:
	  if (integer(x) == 0) return(x);
	  mag = (s7_double)integer(x);
	  ang = (s7_double)fraction(y);
	  break;

	case T_REAL:
	  ang = real(y);
	  if (ang == 0.0) return(x);
	  if (is_NaN(ang)) return(y);
	  if (is_inf(ang)) return(real_NaN);
	  if ((ang == M_PI) || (ang == -M_PI)) return(make_integer(sc, -integer(x)));
	  mag = (s7_double)integer(x);
	  break;

	default:
	  method_or_bust(sc, y, sc->make_polar_symbol, args, T_REAL, 2);
	}
      break;

    case T_RATIO:
      switch (type(y))
	{
	case T_INTEGER:
	  if (integer(y) == 0) return(x);
	  mag = (s7_double)fraction(x);
	  ang = (s7_double)integer(y);
	  break;

	case T_RATIO:
	  mag = (s7_double)fraction(x);
	  ang = (s7_double)fraction(y);
	  break;

	case T_REAL:
	  ang = real(y);
	  if (ang == 0.0) return(x);
	  if (is_NaN(ang)) return(y);
	  if (is_inf(ang)) return(real_NaN);
	  if ((ang == M_PI) || (ang == -M_PI)) return(s7_make_ratio(sc, -numerator(x), denominator(x)));
	  mag = (s7_double)fraction(x);
	  break;

	default:
	  method_or_bust(sc, y, sc->make_polar_symbol, args, T_REAL, 2);
	}
      break;

    case T_REAL:
      mag = real(x);
      switch (type(y))
	{
	case T_INTEGER:
	  if (is_NaN(mag)) return(x);
	  if (integer(y) == 0) return(x);
	  ang = (s7_double)integer(y);
	  break;

	case T_RATIO:
	  if (is_NaN(mag)) return(x);
	  ang = (s7_double)fraction(y);
	  break;

	case T_REAL:
	  if (is_NaN(mag)) return(x);
	  ang = real(y);
	  if (ang == 0.0) return(x);
	  if (is_NaN(ang)) return(y);
	  if (is_inf(ang)) return(real_NaN);
	  break;

	default:
	  method_or_bust(sc, y, sc->make_polar_symbol, args, T_REAL, 2);
	}
      break;

    default:
      method_or_bust(sc, x, sc->make_polar_symbol, args, T_REAL, 1);
    }

  return(s7_make_complex(sc, mag * cos(ang), mag * sin(ang)));

  /* since sin is inaccurate for large arguments, so is make-polar:
   *    (make-polar 1.0 1e40) -> -0.76267273202438+0.64678458842683i, not 8.218988919070239214448025364432557517335E-1-5.696334009536363273080341815735687231337E-1i
   */
}
#endif


/* -------------------------------- complex -------------------------------- */

static s7_pointer c_complex(s7_scheme *sc, s7_double rl, s7_double im)
{
  /* same as s7_make_complex, but assumes im is not 0.0 */
  s7_pointer x;
  new_cell(sc, x, T_COMPLEX);
  set_real_part(x, rl);
  set_imag_part(x, im);
  return(x);
}

static s7_pointer g_complex(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x, y;
  #define H_complex "(complex x1 x2) returns a complex number with real-part x1 and imaginary-part x2"
  #define Q_complex s7_make_signature(sc, 3, sc->is_number_symbol, sc->is_real_symbol, sc->is_real_symbol)

  x = car(args);
  y = cadr(args);

  switch (type(y))
    {
    case T_INTEGER:
      switch (type(x))
	{
	case T_INTEGER:
	  if (integer(y) == 0) return(x);
	  return(c_complex(sc, (s7_double)integer(x), (s7_double)integer(y)));

	case T_RATIO:
	  if (integer(y) == 0) return(x);
	  return(c_complex(sc, (s7_double)fraction(x), (s7_double)integer(y)));

	case T_REAL:
	  if (integer(y) == 0) return(x);
	  return(c_complex(sc, real(x), (s7_double)integer(y)));

	default:
	  method_or_bust(sc, x, sc->complex_symbol, args, T_REAL, 1);
	}

    case T_RATIO:
      switch (type(x))
	{
	case T_INTEGER: return(c_complex(sc, (s7_double)integer(x), (s7_double)fraction(y)));
	case T_RATIO:   return(c_complex(sc, (s7_double)fraction(x), (s7_double)fraction(y)));
	case T_REAL:    return(c_complex(sc, real(x), (s7_double)fraction(y)));
	default:
	  method_or_bust(sc, x, sc->complex_symbol, args, T_REAL, 1);
	}

    case T_REAL:
      switch (type(x))
	{
	case T_INTEGER:
	  if (real(y) == 0.0) return(x);
	  return(c_complex(sc, (s7_double)integer(x), real(y)));

	case T_RATIO:
	  if (real(y) == 0.0) return(x);
	  return(c_complex(sc, (s7_double)fraction(x), real(y)));

	case T_REAL:
	  if (real(y) == 0.0) return(x);
	  return(c_complex(sc, real(x), real(y)));

	default:
	  method_or_bust(sc, x, sc->complex_symbol, args, T_REAL, 1);
	}

    default:
      method_or_bust(sc, (is_let(x)) ? x : y, sc->complex_symbol, args, T_REAL, 2);
    }
}

static s7_pointer complex_p_ii(s7_int x, s7_int y)
{
  if (y == 0)
    return(make_real(cur_sc, (s7_double)x));
  return(c_complex(cur_sc, (s7_double)x, (s7_double)y));
}


/* -------------------------------- exp -------------------------------- */
static s7_pointer g_exp(s7_scheme *sc, s7_pointer args)
{
  #define H_exp "(exp z) returns e^z, (exp 1) is 2.718281828459"
  #define Q_exp pcl_n  

  s7_pointer x;

  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:
      if (integer(x) == 0) return(small_int(1));                       /* (exp 0) -> 1 */
      return(make_real(sc, exp((s7_double)(integer(x)))));

    case T_RATIO:
      return(make_real(sc, exp((s7_double)fraction(x))));

    case T_REAL:
      return(make_real(sc, exp(real(x))));

    case T_COMPLEX:
#if HAVE_COMPLEX_NUMBERS
      return(s7_from_c_complex(sc, cexp(as_c_complex(x))));
      /* this is inaccurate for large arguments:
       *   (exp 0+1e20i) -> -0.66491178990701-0.74692189125949i, not 7.639704044417283004001468027378811228331E-1-6.45251285265780844205811711312523007406E-1i
       */
#else
      return(out_of_range(sc, sc->exp_symbol, small_int(1), x, no_complex_numbers_string));
#endif

    default:
      method_or_bust_with_type_one_arg(sc, x, sc->exp_symbol, args, a_number_string);
    }
}

static s7_double exp_d_d(s7_double x) {return(exp(x));}


/* -------------------------------- log -------------------------------- */

#if __cplusplus
#define LOG_2 1.4426950408889634074
#else
#define LOG_2 1.4426950408889634073599246810018921L /* (/ (log 2.0)) */
#endif

static s7_pointer g_log(s7_scheme *sc, s7_pointer args)
{
  #define H_log "(log z1 (z2 e)) returns log(z1) / log(z2) where z2 (the base) defaults to e: (log 8 2) = 3"
  #define Q_log pcl_n

  s7_pointer x;
  x = car(args);
  if (!s7_is_number(x))
    method_or_bust_with_type(sc, x, sc->log_symbol, args, a_number_string, 1);

  if (is_pair(cdr(args)))
    {
      s7_pointer y;

      y = cadr(args);
      if (!(s7_is_number(y)))
	method_or_bust_with_type(sc, y, sc->log_symbol, args, a_number_string, 2);

      if (y == small_int(2))
	{
	  /* (define (2^n? x) (and (not (zero? x)) (zero? (logand x (- x 1))))) */
	  if (is_integer(x))
	    {
	      s7_int ix;
	      ix = s7_integer(x);
	      if (ix > 0)
		{
		  s7_double fx;
#if (__ANDROID__) || (MS_WINDOWS) || ((__GNUC__) && ((__GNUC__ < 4) || ((__GNUC__ == 4) && (__GNUC_MINOR__ <= 4))))
		  /* just a guess -- log2 gets a warning in gcc 4.3.2, but not in 4.4.4 */
		  fx = log((double)ix) / log(2.0);
#else
		  fx = log2((double)ix);
#endif
		  /* (s7_int)fx rounds (log 8 2) to 2 in FreeBSD! */
#if ((__GNUC__) && ((__GNUC__ < 4) || ((__GNUC__ == 4) && (__GNUC_MINOR__ < 4))))
		  return(make_real(sc, fx));
#else
		  if ((ix & (ix - 1)) == 0)
		    return(make_integer(sc, (s7_int)s7_round(fx)));
		  return(make_real(sc, fx));
#endif
		}
	    }
	  if ((s7_is_real(x)) &&
	      (s7_is_positive(x)))
	    return(make_real(sc, log(real_to_double(sc, x, "log")) * LOG_2));
	  return(s7_from_c_complex(sc, clog(s7_to_c_complex(x)) * LOG_2));
	}

      if ((x == small_int(1)) && (y == small_int(1)))  /* (log 1 1) -> 0 (this is NaN in the bignum case) */
	return(small_int(0));

      /* (log 1 0) must be 0 since everyone says (expt 0 0) is 1 */
      if (s7_is_zero(y))
	{
	  if ((y == small_int(0)) &&
	      (x == small_int(1)))
	    return(y);
	  return(out_of_range(sc, sc->log_symbol, small_int(2), y, s7_make_string_wrapper(sc, "can't be 0")));
	}

      if (s7_is_one(y))          /* this used to raise an error, but the bignum case is simpler if we return inf */
	{
	  if (s7_is_one(x))      /* but (log 1.0 1.0) -> 0.0 */
	    return(real_zero);
	  return(real_infinity); /* currently (log 1/0 1) is inf? */
	}

      if ((s7_is_real(x)) &&
	  (s7_is_real(y)) &&
	  (s7_is_positive(x)) &&
	  (s7_is_positive(y)))
	{
	  if ((s7_is_rational(x)) &&
	      (s7_is_rational(y)))
	    {
	      s7_double res;
	      s7_int ires;
	      res = log(rational_to_double(sc, x)) / log(rational_to_double(sc, y));
	      ires = (s7_int)res;
	      if (res - ires == 0.0)
		return(make_integer(sc, ires));   /* (log 8 2) -> 3 or (log 1/8 2) -> -3 */
	      return(make_real(sc, res));         /* perhaps use rationalize here? (log 2 8) -> 1/3 */
	    }
	  return(make_real(sc, log(real_to_double(sc, x, "log")) / log(real_to_double(sc, y, "log"))));
	}
      return(s7_from_c_complex(sc, clog(s7_to_c_complex(x)) / clog(s7_to_c_complex(y))));
    }

  if (s7_is_real(x))
    {
      if (s7_is_positive(x))
	return(make_real(sc, log(real_to_double(sc, x, "log"))));
      return(s7_make_complex(sc, log(-real_to_double(sc, x, "log")), M_PI));
    }
  return(s7_from_c_complex(sc, clog(s7_to_c_complex(x))));
}



/* -------------------------------- sin -------------------------------- */
static s7_pointer g_sin(s7_scheme *sc, s7_pointer args)
{
  #define H_sin "(sin z) returns sin(z)"
  #define Q_sin pcl_n  

  s7_pointer x;
  x = car(args);
  switch (type(x))
    {
    case T_REAL:
      return(make_real(sc, sin(real(x))));

    case T_INTEGER:
      if (integer(x) == 0) return(small_int(0));                      /* (sin 0) -> 0 */
      return(make_real(sc, sin((s7_double)integer(x))));

    case T_RATIO:
      return(make_real(sc, sin((s7_double)(fraction(x)))));

    case T_COMPLEX:
#if HAVE_COMPLEX_NUMBERS
      return(s7_from_c_complex(sc, csin(as_c_complex(x))));
#else
      return(out_of_range(sc, sc->sin_symbol, small_int(1), x, no_complex_numbers_string));
#endif

    default:
      method_or_bust_with_type_one_arg(sc, x, sc->sin_symbol, args, a_number_string);
    }

  /* sin is totally inaccurate over about 1e18.  There's a way to get true results,
   *   but it involves fancy "range reduction" techniques.
   *   This means that lots of things are inaccurate:
   * (sin (remainder 1e22 (* 2 pi)))
   * -0.57876806033477
   * but it should be -8.522008497671888065747423101326159661908E-1
   * ---
   * (remainder 1e22 (* 2 pi)) -> 1.0057952155665e+22 !!
   *   it should be 5.263007914620499494429139986095833592117E0
   */
}

static s7_double sin_d_d(s7_double x) {return(sin(x));}


/* -------------------------------- cos -------------------------------- */
static s7_pointer g_cos(s7_scheme *sc, s7_pointer args)
{
  #define H_cos "(cos z) returns cos(z)"
  #define Q_cos pcl_n  

  s7_pointer x;
  x = car(args);
  switch (type(x))
    {
    case T_REAL:
      return(make_real(sc, cos(real(x))));

    case T_INTEGER:
      if (integer(x) == 0) return(small_int(1));                     /* (cos 0) -> 1 */
      return(make_real(sc, cos((s7_double)integer(x))));

    case T_RATIO:
      return(make_real(sc, cos((s7_double)(fraction(x)))));

    case T_COMPLEX:
#if HAVE_COMPLEX_NUMBERS
      return(s7_from_c_complex(sc, ccos(as_c_complex(x))));
#else
      return(out_of_range(sc, sc->cos_symbol, small_int(1), x, no_complex_numbers_string));
#endif

    default:
      method_or_bust_with_type_one_arg(sc, x, sc->cos_symbol, args, a_number_string);
    }
}

static s7_double cos_d_d(s7_double x) {return(cos(x));}


/* -------------------------------- tan -------------------------------- */
static s7_pointer g_tan(s7_scheme *sc, s7_pointer args)
{
  #define H_tan "(tan z) returns tan(z)"
  #define Q_tan pcl_n  

  s7_pointer x;
  x = car(args);
  switch (type(x))
    {
    case T_REAL:
      return(make_real(sc, tan(real(x))));

    case T_INTEGER:
      if (integer(x) == 0) return(small_int(0));                      /* (tan 0) -> 0 */
      return(make_real(sc, tan((s7_double)(integer(x)))));

    case T_RATIO:
      return(make_real(sc, tan((s7_double)(fraction(x)))));

    case T_COMPLEX:
#if HAVE_COMPLEX_NUMBERS
      if (imag_part(x) > 350.0)
	return(c_complex(sc, 0.0, 1.0));
      if (imag_part(x) < -350.0)
	return(c_complex(sc, 0.0, -1.0));
      return(s7_from_c_complex(sc, ctan(as_c_complex(x))));
#else
      return(out_of_range(sc, sc->tan_symbol, small_int(1), x, no_complex_numbers_string));
#endif

    default:
      method_or_bust_with_type_one_arg(sc, x, sc->tan_symbol, args, a_number_string);
    }
}

static s7_double tan_d_d(s7_double x) {return(tan(x));}


/* -------------------------------- asin -------------------------------- */
static s7_pointer c_asin(s7_scheme *sc, s7_double x)
{
  s7_double absx, recip;
  s7_complex result;

  absx = fabs(x);
  if (absx <= 1.0)
    return(make_real(sc, asin(x)));

  /* otherwise use maxima code: */
  recip = 1.0 / absx;
  result = (M_PI / 2.0) - (_Complex_I * clog(absx * (1.0 + (sqrt(1.0 + recip) * csqrt(1.0 - recip)))));
  if (x < 0.0)
    return(s7_from_c_complex(sc, -result));
  return(s7_from_c_complex(sc, result));
}


static s7_pointer g_asin(s7_scheme *sc, s7_pointer args)
{
  #define H_asin "(asin z) returns asin(z); (sin (asin x)) = x"
  #define Q_asin pcl_n  
  s7_pointer n;

  n = car(args);
  switch (type(n))
    {
    case T_INTEGER:
      if (integer(n) == 0) return(small_int(0));                    /* (asin 0) -> 0 */
      /* in netBSD, (asin 2) returns 0.25383842987008+0.25383842987008i according to Peter Bex */
      return(c_asin(sc, (s7_double)integer(n)));

    case T_RATIO:
      return(c_asin(sc, (s7_double)numerator(n) / (s7_double)denominator(n)));

    case T_REAL:
      return(c_asin(sc, real(n)));

    case T_COMPLEX:
#if HAVE_COMPLEX_NUMBERS
      /* if either real or imag part is very large, use explicit formula, not casin */
      /*   this code taken from sbcl's src/code/irrat.lisp */
      /* break is around x+70000000i */

      if ((fabs(real_part(n)) > 1.0e7) ||
	  (fabs(imag_part(n)) > 1.0e7))
	{
	  s7_complex sq1mz, sq1pz, z;
	  z = as_c_complex(n);
	  sq1mz = csqrt(1.0 - z);
	  sq1pz = csqrt(1.0 + z);
	  return(s7_make_complex(sc, atan(real_part(n) / creal(sq1mz * sq1pz)), asinh(cimag(sq1pz * conj(sq1mz)))));
	}
      return(s7_from_c_complex(sc, casin(as_c_complex(n))));
#else
      return(out_of_range(sc, sc->asin_symbol, small_int(1), n, no_complex_numbers_string));
#endif

    default:
      method_or_bust_with_type_one_arg(sc, n, sc->asin_symbol, list_1(sc, n), a_number_string);
    }
}


/* -------------------------------- acos -------------------------------- */
static s7_pointer c_acos(s7_scheme *sc, s7_double x)
{
  s7_double absx, recip;
  s7_complex result;

  absx = fabs(x);
  if (absx <= 1.0)
    return(make_real(sc, acos(x)));

  /* else follow maxima again: */
  recip = 1.0 / absx;
  if (x > 0.0)
    result = _Complex_I * clog(absx * (1.0 + (sqrt(1.0 + recip) * csqrt(1.0 - recip))));
  else result = M_PI - _Complex_I * clog(absx * (1.0 + (sqrt(1.0 + recip) * csqrt(1.0 - recip))));
  return(s7_from_c_complex(sc, result));
}

static s7_pointer g_acos(s7_scheme *sc, s7_pointer args)
{
  #define H_acos "(acos z) returns acos(z); (cos (acos 1)) = 1"
  #define Q_acos pcl_n  
  s7_pointer n;
  
  n = car(args);
  switch (type(n))
    {
    case T_INTEGER:
      if (integer(n) == 1) return(small_int(0));
      return(c_acos(sc, (s7_double)integer(n)));

    case T_RATIO:
      return(c_acos(sc, (s7_double)numerator(n) / (s7_double)denominator(n)));

    case T_REAL:
      return(c_acos(sc, real(n)));

    case T_COMPLEX:
#if HAVE_COMPLEX_NUMBERS
      /* if either real or imag part is very large, use explicit formula, not cacos */
      /*   this code taken from sbcl's src/code/irrat.lisp */

      if ((fabs(real_part(n)) > 1.0e7) ||
	  (fabs(imag_part(n)) > 1.0e7))
	{
	  s7_complex sq1mz, sq1pz, z;
	  z = as_c_complex(n);
	  sq1mz = csqrt(1.0 - z);
	  sq1pz = csqrt(1.0 + z);
	  return(s7_make_complex(sc, 2.0 * atan(creal(sq1mz) / creal(sq1pz)), asinh(cimag(sq1mz * conj(sq1pz)))));
	}
      return(s7_from_c_complex(sc, cacos(s7_to_c_complex(n))));
#else
      return(out_of_range(sc, sc->acos_symbol, small_int(1), n, no_complex_numbers_string));
#endif

    default:
      method_or_bust_with_type_one_arg(sc, n, sc->acos_symbol, list_1(sc, n), a_number_string);
    }
}


/* -------------------------------- atan -------------------------------- */

static s7_pointer g_atan(s7_scheme *sc, s7_pointer args)
{
  #define H_atan "(atan z) returns atan(z), (atan y x) returns atan(y/x)"
  #define Q_atan s7_make_signature(sc, 3, sc->is_number_symbol, sc->is_number_symbol, sc->is_real_symbol)
  /* actually if there are two args, both should be real, but how to express that in the signature? */
  s7_pointer x, y;
  s7_double x1, x2;

  /* currently (atan inf.0 inf.0) -> 0.78539816339745, and (atan inf.0 -inf.0) -> 2.3561944901923 (etc) */

  x = car(args);
  if (!is_pair(cdr(args)))
    {
      switch (type(x))
	{
	case T_INTEGER:
	  if (integer(x) == 0) return(small_int(0));                /* (atan 0) -> 0 */

	case T_RATIO:
	case T_REAL:
	  return(make_real(sc, atan(real_to_double(sc, x, "atan"))));

	case T_COMPLEX:
#if HAVE_COMPLEX_NUMBERS
	  return(s7_from_c_complex(sc, catan(as_c_complex(x))));
#else
	  return(out_of_range(sc, sc->atan_symbol, small_int(1), x, no_complex_numbers_string));
#endif

	default:
	  method_or_bust_with_type_one_arg(sc, x, sc->atan_symbol, args, a_number_string);
	}
    }

  if (!s7_is_real(x))
    method_or_bust(sc, x, sc->atan_symbol, args, T_REAL, 1);

  y = cadr(args);
  if (!s7_is_real(y))
    method_or_bust(sc, y, sc->atan_symbol, args, T_REAL, 2);
  
  x1 = real_to_double(sc, x, "atan");
  x2 = real_to_double(sc, y, "atan");
  return(make_real(sc, atan2(x1, x2)));
}

static s7_double atan_d_dd(s7_double x, s7_double y) {return(atan2(x, y));}


/* -------------------------------- sinh -------------------------------- */
static s7_pointer g_sinh(s7_scheme *sc, s7_pointer args)
{
  #define H_sinh "(sinh z) returns sinh(z)"
  #define Q_sinh pcl_n  

  s7_pointer x;
  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:
      if (integer(x) == 0) return(small_int(0));                    /* (sinh 0) -> 0 */

    case T_REAL:
    case T_RATIO:
      return(make_real(sc, sinh(real_to_double(sc, x, "sinh"))));

    case T_COMPLEX:
#if HAVE_COMPLEX_NUMBERS
      return(s7_from_c_complex(sc, csinh(as_c_complex(x))));
#else
      return(out_of_range(sc, sc->sinh_symbol, small_int(1), x, no_complex_numbers_string));
#endif

    default:
      method_or_bust_with_type_one_arg(sc, x, sc->sinh_symbol, args, a_number_string);
    }
}

static s7_double sinh_d_d(s7_double x) {return(sinh(x));}


/* -------------------------------- cosh -------------------------------- */
static s7_pointer g_cosh(s7_scheme *sc, s7_pointer args)
{
  #define H_cosh "(cosh z) returns cosh(z)"
  #define Q_cosh pcl_n  

  s7_pointer x;
  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:
      if (integer(x) == 0) return(small_int(1));                   /* (cosh 0) -> 1 */

    case T_REAL:
    case T_RATIO:
      /* this is not completely correct when optimization kicks in.
       * :(define (hi) (do ((i 0 (+ i 1))) ((= i 1)) (display (cosh i))))
       * hi
       * :(hi)
       * 1.0()
       * :(cosh 0)
       * 1
       */
      return(make_real(sc, cosh(real_to_double(sc, x, "cosh"))));

    case T_COMPLEX:
#if HAVE_COMPLEX_NUMBERS
      return(s7_from_c_complex(sc, ccosh(as_c_complex(x))));
#else
      return(out_of_range(sc, sc->cosh_symbol, small_int(1), x, no_complex_numbers_string));
#endif

    default:
      method_or_bust_with_type_one_arg(sc, x, sc->cosh_symbol, args, a_number_string);
    }
}

static s7_double cosh_d_d(s7_double x) {return(cosh(x));}


/* -------------------------------- tanh -------------------------------- */
static s7_pointer g_tanh(s7_scheme *sc, s7_pointer args)
{
  #define H_tanh "(tanh z) returns tanh(z)"
  #define Q_tanh pcl_n  

  s7_pointer x;
  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:
      if (integer(x) == 0) return(small_int(0));  /* (tanh 0) -> 0 */

    case T_REAL:
    case T_RATIO:
      return(make_real(sc, tanh(real_to_double(sc, x, "tanh"))));

    case T_COMPLEX:
#if HAVE_COMPLEX_NUMBERS
      if (real_part(x) > 350.0)
	return(real_one);                         /* closer than 0.0 which is what ctanh is about to return! */
      if (real_part(x) < -350.0)
	return(make_real(sc, -1.0));              /* closer than ctanh's -0.0 */
      return(s7_from_c_complex(sc, ctanh(as_c_complex(x))));
#else
      return(out_of_range(sc, sc->tanh_symbol, small_int(1), x, no_complex_numbers_string));
#endif

    default:
      method_or_bust_with_type_one_arg(sc, x, sc->tanh_symbol, args, a_number_string);
    }
}

static s7_double tanh_d_d(s7_double x) {return(tanh(x));}


/* -------------------------------- asinh -------------------------------- */

static s7_pointer g_asinh(s7_scheme *sc, s7_pointer args)
{
  #define H_asinh "(asinh z) returns asinh(z)"
  #define Q_asinh pcl_n  
  s7_pointer x;
  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:
      if (integer(x) == 0) return(small_int(0));
      return(make_real(sc, asinh((s7_double)integer(x))));

    case T_RATIO:
      return(make_real(sc, asinh((s7_double)numerator(x) / (s7_double)denominator(x))));

    case T_REAL:
      return(make_real(sc, asinh(real(x))));

    case T_COMPLEX:
#if HAVE_COMPLEX_NUMBERS
  #if (defined(__OpenBSD__)) || (defined(__NetBSD__))
      return(s7_from_c_complex(sc, casinh_1(as_c_complex(x))));
  #else
      return(s7_from_c_complex(sc, casinh(as_c_complex(x))));
  #endif
#else
      return(out_of_range(sc, sc->asinh_symbol, small_int(1), x, no_complex_numbers_string));
#endif

    default:
      method_or_bust_with_type_one_arg(sc, x, sc->asinh_symbol, list_1(sc, x), a_number_string);
    }
}


/* -------------------------------- acosh -------------------------------- */

static s7_pointer g_acosh(s7_scheme *sc, s7_pointer args)
{
  #define H_acosh "(acosh z) returns acosh(z)"
  #define Q_acosh pcl_n  
  s7_pointer x;
  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:
      if (integer(x) == 1) return(small_int(0));

    case T_REAL:
    case T_RATIO:
      {
	double x1;
	x1 = real_to_double(sc, x, "acosh");
	if (x1 >= 1.0)
	  return(make_real(sc, acosh(x1)));
      }

    case T_COMPLEX:
#if HAVE_COMPLEX_NUMBERS
  #ifdef __OpenBSD__
      return(s7_from_c_complex(sc, cacosh_1(s7_to_c_complex(x))));
  #else
      return(s7_from_c_complex(sc, cacosh(s7_to_c_complex(x)))); /* not as_c_complex because x might not be complex */
  #endif
#else
      /* since we can fall through to this branch, we need a better error message than "must be a number, not 0.0" */
      return(out_of_range(sc, sc->acosh_symbol, small_int(1), x, no_complex_numbers_string));
#endif

    default:
      method_or_bust_with_type_one_arg(sc, x, sc->acosh_symbol, list_1(sc, x), a_number_string);
    }
}


/* -------------------------------- atanh -------------------------------- */

static s7_pointer g_atanh(s7_scheme *sc, s7_pointer args)
{
  #define H_atanh "(atanh z) returns atanh(z)"
  #define Q_atanh pcl_n  
  s7_pointer x;
  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:
      if (integer(x) == 0) return(small_int(0));                    /* (atanh 0) -> 0 */

    case T_REAL:
    case T_RATIO:
      {
	double x1;
	x1 = real_to_double(sc, x, "atanh");
	if (fabs(x1) < 1.0)
	  return(make_real(sc, atanh(x1)));
      }

      /* if we can't distinguish x from 1.0 even with long doubles, we'll get inf.0:
       *    (atanh 9223372036854775/9223372036854776) -> 18.714973875119
       *    (atanh 92233720368547758/92233720368547757) -> inf.0
       */
    case T_COMPLEX:
#if HAVE_COMPLEX_NUMBERS
  #if (defined(__OpenBSD__)) || (defined(__NetBSD__))
      return(s7_from_c_complex(sc, catanh_1(s7_to_c_complex(x))));
  #else
      return(s7_from_c_complex(sc, catanh(s7_to_c_complex(x))));
  #endif
#else
      return(out_of_range(sc, sc->atanh_symbol, small_int(1), x, no_complex_numbers_string));
#endif

    default:
      method_or_bust_with_type_one_arg(sc, x, sc->atanh_symbol, list_1(sc, x), a_number_string);
    }
}


/* -------------------------------- sqrt -------------------------------- */
static s7_pointer g_sqrt(s7_scheme *sc, s7_pointer args)
{
  #define H_sqrt "(sqrt z) returns the square root of z"
  #define Q_sqrt pcl_n  

  s7_pointer n;
  s7_double sqx;

  n = car(args);
  switch (type(n))
    {
    case T_INTEGER:
      if (integer(n) >= 0)
	{
	  s7_int ix;
	  sqx = sqrt((s7_double)integer(n));
	  ix = (s7_int)sqx;
	  if ((ix * ix) == integer(n))
	    return(make_integer(sc, ix));
	  return(make_real(sc, sqx));
	  /* Mark Weaver notes that
	   *     (zero? (- (sqrt 9007199136250226) 94906265.0)) -> #t
	   * but (* 94906265 94906265) -> 9007199136250225 -- oops
	   * at least we return a real here, not an incorrect integer and
	   *     (sqrt 9007199136250225) -> 94906265
	   */
	}
      sqx = (s7_double)integer(n); /* we're trying to protect against (sqrt -9223372036854775808) where we can't negate the integer argument */
      return(s7_make_complex(sc, 0.0, sqrt((s7_double)(-sqx))));

    case T_RATIO:
      sqx = (s7_double)fraction(n);
      if (sqx > 0.0) /* else it's complex, so it can't be a ratio */
	{
	  s7_int nm = 0, dn = 1;
	  if (c_rationalize(sqx, 1.0e-16, &nm, &dn)) /* 1e-16 so that (sqrt 1/1099511627776) returns 1/1048576 */
	    {
#if HAVE_OVERFLOW_CHECKS
	      s7_int nm2, dn2;
	      if ((multiply_overflow(nm, nm, &nm2)) ||
		  (multiply_overflow(dn, dn, &dn2)))
		return(make_real(sc, sqrt(sqx)));
	      if ((nm2 == numerator(n)) &&
		  (dn2 == denominator(n)))
		return(s7_make_ratio(sc, nm, dn));
#else
	      if ((nm * nm == numerator(n)) &&
		  (dn * dn == denominator(n)))
		return(s7_make_ratio(sc, nm, dn));
#endif
	    }
	  return(make_real(sc, sqrt(sqx)));
	}
      return(s7_make_complex(sc, 0.0, sqrt(-sqx)));

    case T_REAL:
      if (is_NaN(real(n)))
	return(real_NaN);
      if (real(n) >= 0.0)
	return(make_real(sc, sqrt(real(n))));
      return(s7_make_complex(sc, 0.0, sqrt(-real(n))));

    case T_COMPLEX:
      /* (* inf.0 (sqrt -1)) -> -nan+infi, but (sqrt -inf.0) -> 0+infi */
#if HAVE_COMPLEX_NUMBERS
      return(s7_from_c_complex(sc, csqrt(as_c_complex(n))));
#else
      return(out_of_range(sc, sc->sqrt_symbol, small_int(1), n, no_complex_numbers_string));
#endif

    default:
      method_or_bust_with_type_one_arg(sc, n, sc->sqrt_symbol, args, a_number_string);
    }
}


/* -------------------------------- expt -------------------------------- */

static s7_int int_to_int(s7_int x, s7_int n)
{
  /* from GSL */
  s7_int value = 1;
  do {
    if (n & 1) value *= x;
    n >>= 1;
#if HAVE_OVERFLOW_CHECKS
    if (multiply_overflow(x, x, &x))
      break;
#else
    x *= x;
#endif
  } while (n);
  return(value);
}


static const int64_t nth_roots[63] = {
  S7_LLONG_MAX, S7_LLONG_MAX, 3037000499LL, 2097151, 55108, 6208, 1448, 511, 234, 127, 78, 52, 38, 28, 22,
  18, 15, 13, 11, 9, 8, 7, 7, 6, 6, 5, 5, 5, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2,
  2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2};

static const int64_t int_nth_roots[31] = {
  S7_LONG_MAX, S7_LONG_MAX, 46340, 1290, 215, 73, 35, 21, 14, 10, 8, 7, 5, 5, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2};

static bool int_pow_ok(s7_int x, s7_int y)
{
  if (s7_int_bits > 31)
    return((y < 63) &&
	   (nth_roots[y] >= s7_int_abs(x)));
  return((y < 31) &&
	 (int_nth_roots[y] >= s7_int_abs(x)));
}


static s7_pointer g_expt(s7_scheme *sc, s7_pointer args)
{
  #define H_expt "(expt z1 z2) returns z1^z2"
  #define Q_expt pcl_n
  s7_pointer n, pw;

  n = car(args);
  if (!s7_is_number(n))
    method_or_bust_with_type(sc, n, sc->expt_symbol, args, a_number_string, 1);

  pw = cadr(args);
  if (!s7_is_number(pw))
    method_or_bust_with_type(sc, pw, sc->expt_symbol, args, a_number_string, 2);

  /* this provides more than 2 args to expt:
   *  if (is_not_null(cddr(args)))
   *    return(g_expt(sc, list_2(sc, car(args), g_expt(sc, cdr(args)))));
   *
   * but it's unusual in scheme to process args in reverse order, and the
   * syntax by itself is ambiguous (does (expt 2 2 3) = 256 or 64?)
   */

  if (s7_is_zero(n))
    {
      if (s7_is_zero(pw))
	{
	  if ((s7_is_integer(n)) && (s7_is_integer(pw)))       /* (expt 0 0) -> 1 */
	    return(small_int(1));
	  return(real_zero);                                   /* (expt 0.0 0) -> 0.0 */
	}

      if (s7_is_real(pw))
	{
	  if (s7_is_negative(pw))                              /* (expt 0 -1) */
	    return(division_by_zero_error(sc, sc->expt_symbol, args));
	  /* (Clisp gives divide-by-zero error here, Guile returns inf.0) */

	  if ((!s7_is_rational(pw)) &&                         /* (expt 0 most-positive-fixnum) */
	      (is_NaN(s7_real(pw))))                           /* (expt 0 +nan.0) */
	    return(pw);
	}
      else
	{                                                      /* (expt 0 a+bi) */
	  if (real_part(pw) < 0.0)                             /* (expt 0 -1+i) */
	    return(division_by_zero_error(sc, sc->expt_symbol, args));
	  if ((is_NaN(real_part(pw))) ||                       /* (expt 0 0+1/0i) */
	      (is_NaN(imag_part(pw))))
	    return(real_NaN);
	}

      if ((s7_is_integer(n)) && (s7_is_integer(pw)))           /* pw != 0, (expt 0 2312) */
	return(small_int(0));
      return(real_zero);                                       /* (expt 0.0 123123) */
    }

  if (s7_is_one(pw))
    {
      if (s7_is_integer(pw))
	return(n);
      if (is_rational(n))
	return(make_real(sc, rational_to_double(sc, n)));
      return(n);
    }

  if (is_t_integer(pw))
    {
      s7_int y;
      y = integer(pw);
      if (y == 0)
	{
	  if (is_rational(n))                                 /* (expt 3 0) */
	    return(small_int(1));
	  if ((is_NaN(s7_real_part(n))) ||                    /* (expt 1/0 0) -> NaN */
	      (is_NaN(s7_imag_part(n))))                      /* (expt (complex 0 1/0) 0) -> NaN */
	    return(n);
	  return(real_one);                                   /* (expt 3.0 0) */
	}

      switch (type(n))
	{
	case T_INTEGER:
	  {
	    s7_int x;
	    x = s7_integer(n);
	    if (x == 1)                                       /* (expt 1 y) */
	      return(n);

	    if (x == -1)
	      {
		if (y == s7_int_min)                        /* (expt -1 most-negative-fixnum) */
		  return(small_int(1));

		if (s7_int_abs(y) & 1)                        /* (expt -1 odd-int) */
		  return(n);
		return(small_int(1));                         /* (expt -1 even-int) */
	      }

	    if (y == s7_int_min)                            /* (expt x most-negative-fixnum) */
	      return(small_int(0));
	    if (x == s7_int_min)                            /* (expt most-negative-fixnum y) */
	      return(make_real(sc, pow((double)x, (double)y)));

	    if (int_pow_ok(x, s7_int_abs(y)))
	      {
		if (y > 0)
		  return(make_integer(sc, int_to_int(x, y)));
		return(s7_make_ratio(sc, 1, int_to_int(x, -y)));
	      }
	  }
	  break;

	case T_RATIO:
	  {
	    s7_int nm, dn;

	    nm = numerator(n);
	    dn = denominator(n);

	    if (y == s7_int_min)
	      {
		if (s7_int_abs(nm) > dn)
		  return(small_int(0));              /* (expt 4/3 most-negative-fixnum) -> 0? */
		return(real_infinity);               /* (expt 3/4 most-negative-fixnum) -> inf? */
	      }

	    if ((int_pow_ok(nm, s7_int_abs(y))) &&
		(int_pow_ok(dn, s7_int_abs(y))))
	      {
		if (y > 0)
		  return(s7_make_ratio(sc, int_to_int(nm, y), int_to_int(dn, y)));
		return(s7_make_ratio(sc, int_to_int(dn, -y), int_to_int(nm, -y)));
	      }
	  }
	  break;
	  /* occasionally int^rat can be int32_t but it happens so infrequently it's probably not worth checking
	   *  one possibly easy case: (expt 1 1/2) -> 1 (-1?) etc
	   */

	case T_REAL:
	  /* (expt -1.0 most-positive-fixnum) should be -1.0
	   * (expt -1.0 (+ (expt 2 53) 1)) -> -1.0
	   * (expt -1.0 (- 1 (expt 2 54))) -> -1.0
	   */
	  if (real(n) == -1.0)
	    {
	      if (y == s7_int_min)
		return(real_one);

	      if (s7_int_abs(y) & 1)
		return(n);
	      return(real_one);
	    }
	  break;

	case T_COMPLEX:
#if HAVE_COMPLEX_NUMBERS
	  if ((s7_real_part(n) == 0.0) &&
	      ((s7_imag_part(n) == 1.0) ||
	       (s7_imag_part(n) == -1.0)))
	    {
	      bool yp, np;
	      yp = (y > 0);
	      np = (s7_imag_part(n) > 0.0);
	      switch (s7_int_abs(y) % 4)
		{
		case 0: return(real_one);
		case 1: return(s7_make_complex(sc, 0.0, (yp == np) ? 1.0 : -1.0));
		case 2: return(make_real(sc, -1.0));
		case 3: return(s7_make_complex(sc, 0.0, (yp == np) ? -1.0 : 1.0));
		}
	    }
#else
	  return(out_of_range(sc, sc->expt_symbol, small_int(2), n, no_complex_numbers_string));
#endif
	  break;
	}
    }

  if ((s7_is_real(n)) &&
      (s7_is_real(pw)))
    {
      s7_double x, y;

      if ((is_t_ratio(pw)) &&
	  (numerator(pw) == 1))
	{
	  if (denominator(pw) == 2)
	    return(g_sqrt(sc, args));
	  if (denominator(pw) == 3)
	    return(make_real(sc, cbrt(real_to_double(sc, n, "expt")))); /* (expt 27 1/3) should be 3, not 3.0... */

	  /* but: (expt 512/729 1/3) -> 0.88888888888889
	   */
	  /* and 4 -> sqrt(sqrt...) etc? */
	}

      x = real_to_double(sc, n, "expt");
      y = real_to_double(sc, pw, "expt");

      if (is_NaN(x)) return(n);
      if (is_NaN(y)) return(pw);
      if (y == 0.0) return(real_one);

      if (x > 0.0)
	return(make_real(sc, pow(x, y)));
      /* tricky cases abound here: (expt -1 1/9223372036854775807)
       */
    }

  /* (expt 0+i 1e+16) = 0.98156860153485-0.19111012657867i ?
   * (expt 0+i 1+1/0i) = 0.0 ??
   */
  return(s7_from_c_complex(sc, cpow(s7_to_c_complex(n), s7_to_c_complex(pw))));
}


/* -------------------------------- lcm -------------------------------- */
static s7_pointer g_lcm(s7_scheme *sc, s7_pointer args)
{
  #define H_lcm "(lcm ...) returns the least common multiple of its rational arguments"
  #define Q_lcm pcl_f

  s7_int n = 1, d = 0;
  s7_pointer p;

  if (!is_pair(args))
    return(small_int(1));

  if (!is_pair(cdr(args)))
    {
      if (!is_rational(car(args)))
	method_or_bust_with_type(sc, car(args), sc->lcm_symbol, args, a_rational_string, 1);
      return(g_abs(sc, args));
    }

  for (p = args; is_pair(p); p = cdr(p))
    {
      s7_pointer x;
      s7_int b;
      x = car(p);
      switch (type(x))
	{
	case T_INTEGER:
	  if (integer(x) == 0)
	    n = 0;
	  else 
	    {
	      b = integer(x);
	      if (b < 0) b = -b;
	      n = (n / c_gcd(n, b)) * b;
	    }
	  if (d != 0) d = 1;
	  break;

	case T_RATIO:
	  b = numerator(x);
	  if (b < 0) b = -b;
	  n = (n / c_gcd(n, b)) * b;
	  if (d == 0)
	    {
	      if (p == args)
		d = s7_denominator(x);
	      else d = 1;
	    }
	  else d = c_gcd(d, s7_denominator(x));
	  break;

	default:
	  method_or_bust_with_type(sc, x, sc->lcm_symbol, cons(sc, (d <= 1) ? make_integer(sc, n) : s7_make_ratio(sc, n, d), p), a_rational_string, position_of(p, args));
	}
      if (n < 0) return(simple_out_of_range(sc, sc->lcm_symbol, args, result_is_too_large_string));
      if (n == 0)
	{
	  for (p = cdr(p); is_pair(p); p = cdr(p))
	    if (!is_rational_via_method(sc, car(p)))
	      return(wrong_type_argument_with_type(sc, sc->lcm_symbol, position_of(p, args), x, a_rational_string));
	  return(small_int(0));
	}
    }

  if (d <= 1)
    return(make_integer(sc, n));
  return(s7_make_ratio(sc, n, d));
}



/* -------------------------------- gcd -------------------------------- */
static s7_pointer g_gcd(s7_scheme *sc, s7_pointer args)
{
  #define H_gcd "(gcd ...) returns the greatest common divisor of its rational arguments"
  #define Q_gcd pcl_f
  s7_int n = 0, d = 1;
  s7_pointer p;

  if (!is_pair(args))
    return(small_int(0));

  if (!is_pair(cdr(args)))
    {
      if (!is_rational(car(args)))
	method_or_bust_with_type(sc, car(args), sc->gcd_symbol, args, a_rational_string, 1);
      return(g_abs(sc, args));
    }

  for (p = args; is_pair(p); p = cdr(p))
    {
      s7_pointer x;
      s7_int b;
      x = car(p);
      switch (type(x))
	{
	case T_INTEGER:
	  n = c_gcd(n, integer(x));
	  break;

	case T_RATIO:
	  n = c_gcd(n, s7_numerator(x));
	  b = s7_denominator(x);
	  if (b < 0) b = -b;
	  d = (d / c_gcd(d, b)) * b;
	  if (d < 0) return(simple_out_of_range(sc, sc->gcd_symbol, args, result_is_too_large_string));
	  break;

	default:
	  method_or_bust_with_type(sc, x, sc->gcd_symbol, cons(sc, (d <= 1) ? make_integer(sc, n) : s7_make_ratio(sc, n, d), p), a_rational_string, position_of(p, args));
	}
      if (n < 0) return(simple_out_of_range(sc, sc->gcd_symbol, args, result_is_too_large_string));
    }

  if (d <= 1)
    return(make_integer(sc, n));
  return(s7_make_ratio(sc, n, d));
}



static s7_pointer s7_truncate(s7_scheme *sc, s7_pointer caller, s7_double xf)   /* can't use "truncate" -- it's in unistd.h */
{
  if ((xf > s7_int_max) ||
      (xf < s7_int_min))
    return(simple_out_of_range(sc, caller, make_real(sc, xf), its_too_large_string));

  if (xf > 0.0)
    return(make_integer(sc, (s7_int)floor(xf)));
  return(make_integer(sc, (s7_int)ceil(xf)));
}

static inline s7_int c_quo_int(s7_scheme *sc, s7_int x, s7_int y)
{
  if (y == 0)
    division_by_zero_error(sc, sc->quotient_symbol, set_elist_2(sc, make_integer(sc, x), make_integer(sc, y)));
  if ((y == -1) && (x == s7_int_min))   /* (quotient most-negative-fixnum -1) */
    simple_out_of_range(sc, sc->quotient_symbol, set_elist_2(sc, make_integer(sc, x), make_integer(sc, y)), its_too_large_string);
  return(x / y);
}

static s7_double c_quo_dbl(s7_scheme *sc, s7_double x, s7_double y)
{
  s7_double xf;

  if (y == 0.0)
    division_by_zero_error(sc, sc->quotient_symbol, set_elist_2(sc, make_real(sc, x), make_real(sc, y)));
  if ((is_inf(y)) || (is_NaN(y)))
    wrong_type_argument_with_type(sc, sc->quotient_symbol, 2, make_real(sc, y), a_normal_real_string);

  xf = x / y;
  if ((xf > s7_int_max) ||
      (xf < s7_int_min))
    simple_out_of_range(sc, sc->quotient_symbol, make_real(sc, xf), its_too_large_string);

  if (xf > 0.0)
    return(floor(xf));
  return(ceil(xf));
}

static s7_int quotient_i_ii(s7_int i1, s7_int i2) {return(c_quo_int(cur_sc, i1, i2));}
static s7_int quotient_i_ii_direct(s7_int i1, s7_int i2) {return(i1 / i2);} /* i2 > 0 */

static s7_double quotient_d_dd(s7_double x1, s7_double x2) 
{
  if ((is_inf(x1)) || (is_NaN(x1)))
    wrong_type_argument_with_type(cur_sc, cur_sc->quotient_symbol, 1, make_real(cur_sc, x1), a_normal_real_string);
  return(c_quo_dbl(cur_sc, x1, x2));
}

static s7_pointer g_quotient(s7_scheme *sc, s7_pointer args)
{
  #define H_quotient "(quotient x1 x2) returns the integer quotient of x1 and x2; (quotient 4 3) = 1"
  #define Q_quotient pcl_r
  /* (define (quo x1 x2) (truncate (/ x1 x2))) ; slib
   */
  s7_pointer x, y;
  s7_int d1, d2, n1, n2;

  x = car(args);
  y = cadr(args);

  switch (type(x))
    {
    case T_INTEGER:
      switch (type(y))
	{
	case T_INTEGER:
	  return(make_integer(sc, c_quo_int(sc, integer(x), integer(y))));

	case T_RATIO:
	  n1 = integer(x);
	  d1 = 1;
	  n2 = numerator(y);
	  d2 = denominator(y);
	  /* (quotient -9223372036854775808 -1/9223372036854775807): arithmetic exception in the no-overflow-checks case */
	  goto RATIO_QUO_RATIO;

	case T_REAL:
	  if (real(y) == 0.0)
	    return(division_by_zero_error(sc, sc->quotient_symbol, args));
	  if ((is_inf(real(y))) || (is_NaN(real(y))))
	    return(wrong_type_argument_with_type(sc, sc->quotient_symbol, 2, y, a_normal_real_string));
	  return(s7_truncate(sc, sc->quotient_symbol, (s7_double)integer(x) / real(y)));

	default:
	  method_or_bust(sc, y, sc->quotient_symbol, args, T_REAL, 2);
	}

    case T_RATIO:
      switch (type(y))
	{
	case T_INTEGER:
	  if (integer(y) == 0)
	    return(division_by_zero_error(sc, sc->quotient_symbol, args));
	  n1 = numerator(x);
	  d1 = denominator(x);
	  n2 = integer(y);
	  d2 = 1;
	  goto RATIO_QUO_RATIO;
	  /* this can lose:
	   *   (quotient 1 2305843009213693952/4611686018427387903) -> 2, not 1
	   *   (quotient 21053343141/6701487259 3587785776203/1142027682075) -> 1, not 0
	   */

	case T_RATIO:
	  n1 = numerator(x);
	  d1 = denominator(x);
	  n2 = numerator(y);
	  d2 = denominator(y);
	RATIO_QUO_RATIO:
	  if (d1 == d2)
	    return(make_integer(sc, n1 / n2));              /* (quotient 3/9223372036854775807 1/9223372036854775807) */
	  if (n1 == n2)
	    return(make_integer(sc, d2 / d1));              /* (quotient 9223372036854775807/2 9223372036854775807/8) */
#if HAVE_OVERFLOW_CHECKS
	  {
	    s7_int n1d2, n2d1;
	    if ((multiply_overflow(n1, d2, &n1d2)) ||
		(multiply_overflow(n2, d1, &n2d1)))
	      return(s7_truncate(sc, sc->quotient_symbol, ((long double)n1 / (long double)n2) * ((long double)d2 / (long double)d1)));
	    return(make_integer(sc, n1d2 / n2d1));
	  }
#else
	  return(make_integer(sc, (n1 * d2) / (n2 * d1)));
#endif

	case T_REAL:
	  if (real(y) == 0.0)
	    return(division_by_zero_error(sc, sc->quotient_symbol, args));
	  if ((is_inf(real(y))) || (is_NaN(real(y))))
	    return(wrong_type_argument_with_type(sc, sc->quotient_symbol, 2, y, a_normal_real_string));
	  return(s7_truncate(sc, sc->quotient_symbol, (s7_double)fraction(x) / real(y)));

	default:
	  method_or_bust(sc, y, sc->quotient_symbol, args, T_REAL, 2);
	}

    case T_REAL:
      if ((is_inf(real(x))) || (is_NaN(real(x))))
	return(wrong_type_argument_with_type(sc, sc->quotient_symbol, 1, x, a_normal_real_string));

      /* if infs allowed we need to return infs/nans, else:
       *    (quotient inf.0 1e-309) -> -9223372036854775808
       *    (quotient inf.0 inf.0) -> -9223372036854775808
       */

      switch (type(y))
	{
	case T_INTEGER:
	  if (integer(y) == 0)
	    return(division_by_zero_error(sc, sc->quotient_symbol, args));
	  return(s7_truncate(sc, sc->quotient_symbol, real(x) / (s7_double)integer(y)));

	case T_RATIO:
	  return(s7_truncate(sc, sc->quotient_symbol, real(x) / (s7_double)fraction(y)));

	case T_REAL:
	  return(make_real(sc, c_quo_dbl(sc, real(x), real(y))));

	default:
	  method_or_bust(sc, y, sc->quotient_symbol, args, T_REAL, 2);
	}

    default:
      method_or_bust(sc, x, sc->quotient_symbol, args, T_REAL, 2);
    }
}


static inline s7_int c_rem_int(s7_scheme *sc, s7_int x, s7_int y)
{
  if (y == 0)
    division_by_zero_error(sc, sc->remainder_symbol, set_elist_2(sc, make_integer(sc, x), make_integer(sc, y)));
  if ((y == 1) || (y == -1))   /* (remainder most-negative-fixnum -1) will segfault with arithmetic exception */
    return(0);
  return(x % y);
}

static s7_double c_rem_dbl(s7_scheme *sc, s7_double x, s7_double y)
{
  s7_int quo;
  s7_double pre_quo;
  if (y == 0.0)
    division_by_zero_error(sc, sc->remainder_symbol, set_elist_2(sc, make_real(sc, x), make_real(sc, y)));
  if ((is_inf(y)) || (is_NaN(y)))
    wrong_type_argument_with_type(sc, sc->remainder_symbol, 2, set_elist_1(sc, make_real(sc, y)), a_normal_real_string);

  pre_quo = x / y;
  if ((pre_quo > s7_int_max) || (pre_quo < s7_int_min))
    simple_out_of_range(sc, sc->remainder_symbol, set_elist_2(sc, make_real(sc, x), make_real(sc, y)), its_too_large_string);
  if (pre_quo > 0.0) 
    quo = (s7_int)floor(pre_quo); 
  else quo = (s7_int)ceil(pre_quo);
  return(x - (y * quo));
}

static s7_int remainder_i_ii(s7_int i1, s7_int i2) {return(c_rem_int(cur_sc, i1, i2));}
static s7_int remainder_i_ii_direct(s7_int i1, s7_int i2) {return(i1 % i2);} /* i2 > 1 */

static s7_double remainder_d_dd(s7_double x1, s7_double x2) 
{
  if ((is_inf(x1)) || (is_NaN(x1)))
    wrong_type_argument_with_type(cur_sc, cur_sc->remainder_symbol, 1, set_elist_1(cur_sc, make_real(cur_sc, x1)), a_normal_real_string);
  return(c_rem_dbl(cur_sc, x1, x2));
}

static s7_pointer g_remainder(s7_scheme *sc, s7_pointer args)
{
  #define H_remainder "(remainder x1 x2) returns the remainder of x1/x2; (remainder 10 3) = 1"
  #define Q_remainder pcl_r
  /* (define (rem x1 x2) (- x1 (* x2 (quo x1 x2)))) ; slib, if x2 is an integer (- x1 (truncate x1 x2)), fractional part: (remainder x 1) */

  s7_pointer x, y;
  s7_int quo, d1, d2, n1, n2;
  s7_double pre_quo;

  x = car(args);
  y = cadr(args);

  switch (type(x))
    {
    case T_INTEGER:
      switch (type(y))
	{
	case T_INTEGER:
	  return(make_integer(sc, c_rem_int(sc, integer(x), integer(y))));

	case T_RATIO:
	  n1 = integer(x);
	  d1 = 1;
	  n2 = numerator(y);
	  d2 = denominator(y);
	  goto RATIO_REM_RATIO;

	case T_REAL:
	  if (real(y) == 0.0)
	    return(division_by_zero_error(sc, sc->remainder_symbol, args));
	  if ((is_inf(real(y))) || (is_NaN(real(y))))
	    return(wrong_type_argument_with_type(sc, sc->remainder_symbol, 2, y, a_normal_real_string));

	  pre_quo = (s7_double)integer(x) / real(y);
	  if ((pre_quo > s7_int_max) || (pre_quo < s7_int_min))
	    return(simple_out_of_range(sc, sc->remainder_symbol, args, its_too_large_string));
	  if (pre_quo > 0.0) quo = (s7_int)floor(pre_quo); else quo = (s7_int)ceil(pre_quo);
	  return(make_real(sc, integer(x) - real(y) * quo));

	default:
	  method_or_bust(sc, y, sc->remainder_symbol, args, T_REAL, 2);
	}

    case T_RATIO:
      switch (type(y))
	{
	case T_INTEGER:
	  n2 = integer(y);
 	  if (n2 == 0)
 	    return(division_by_zero_error(sc, sc->remainder_symbol, args));
	  n1 = numerator(x);
	  d1 = denominator(x);
	  d2 = 1;
	  goto RATIO_REM_RATIO;

	case T_RATIO:
	  n1 = numerator(x);
	  d1 = denominator(x);
	  n2 = numerator(y);
	  d2 = denominator(y);
	RATIO_REM_RATIO:
	  if (d1 == d2)
	    quo = (s7_int)(n1 / n2);
	  else
	    {
	      if (n1 == n2)
		quo = (s7_int)(d2 / d1);
	      else
		{
#if HAVE_OVERFLOW_CHECKS
		  s7_int n1d2, n2d1;
		  if ((multiply_overflow(n1, d2, &n1d2)) ||
		      (multiply_overflow(n2, d1, &n2d1)))
		    {
		      pre_quo = ((long double)n1 / (long double)n2) * ((long double)d2 / (long double)d1);
		      if ((pre_quo > s7_int_max) || (pre_quo < s7_int_min))
			return(simple_out_of_range(sc, sc->remainder_symbol, args, its_too_large_string));
		      if (pre_quo > 0.0) quo = (s7_int)floor(pre_quo); else quo = (s7_int)ceil(pre_quo);
		    }
		  else quo = n1d2 / n2d1;
#else
		  quo = (n1 * d2) / (n2 * d1);
#endif
		}
	    }
	  if (quo == 0)
	    return(x);

#if HAVE_OVERFLOW_CHECKS
	  {
	    s7_int dn, nq;
	    if (!multiply_overflow(n2, quo, &nq))
	      {
		if ((d1 == d2) &&
		    (!subtract_overflow(n1, nq, &dn)))
		  return(s7_make_ratio(sc, dn, d1));

		if ((!multiply_overflow(n1, d2, &dn)) &&
		    (!multiply_overflow(nq, d1, &nq)) &&
		    (!subtract_overflow(dn, nq, &nq)) &&
		    (!multiply_overflow(d1, d2, &d1)))
		  return(s7_make_ratio(sc, nq, d1));
	      }
	  }
#else
	  if (d1 == d2)
	    return(s7_make_ratio(sc, n1 - n2 * quo, d1));

	  return(s7_make_ratio(sc, n1 * d2 - n2 * d1 * quo, d1 * d2));
#endif
	  return(simple_out_of_range(sc, sc->remainder_symbol, args, s7_make_string_wrapper(sc, "intermediate (a/b) is too large")));

	case T_REAL:
	  {
	    s7_double frac;
	    if (real(y) == 0.0)
	      return(division_by_zero_error(sc, sc->remainder_symbol, args));
	    if ((is_inf(real(y))) || (is_NaN(real(y))))
	      return(wrong_type_argument_with_type(sc, sc->remainder_symbol, 2, y, a_normal_real_string));
	    frac = (s7_double)fraction(x);
	    pre_quo = frac / real(y);
	    if ((pre_quo > s7_int_max) || (pre_quo < s7_int_min))
	      return(simple_out_of_range(sc, sc->remainder_symbol, args, its_too_large_string));
	    if (pre_quo > 0.0) quo = (s7_int)floor(pre_quo); else quo = (s7_int)ceil(pre_quo);
	    return(make_real(sc, frac - real(y) * quo));
	  }

	default:
	  method_or_bust(sc, y, sc->remainder_symbol, args, T_REAL, 2);
	}

    case T_REAL:
      if ((is_inf(real(x))) || (is_NaN(real(x))))
	return(wrong_type_argument_with_type(sc, sc->remainder_symbol, 1, x, a_normal_real_string));

      switch (type(y))
	{
	case T_INTEGER:
	  if (integer(y) == 0)
	    return(division_by_zero_error(sc, sc->remainder_symbol, args));
	  pre_quo = real(x) / (s7_double)integer(y);
	  if ((pre_quo > s7_int_max) || (pre_quo < s7_int_min))
	    return(simple_out_of_range(sc, sc->remainder_symbol, args, its_too_large_string));
	  if (pre_quo > 0.0) quo = (s7_int)floor(pre_quo); else quo = (s7_int)ceil(pre_quo);
	  return(make_real(sc, real(x) - integer(y) * quo));
	  /* but... (remainder 1e+18 9223372036854775807) -> 1e+18 */

	case T_RATIO:
	  {
	    /* bad cases here start around 1e16: (remainder 1e15 3/13) -> 0.0 with loss of digits earlier
	     *   would long double help?
	     */
	    s7_double frac;
	    frac = (s7_double)fraction(y);
	    pre_quo = real(x) / frac;
	    if ((pre_quo > s7_int_max) || (pre_quo < s7_int_min))
	      return(simple_out_of_range(sc, sc->remainder_symbol, args, its_too_large_string));
	    if (pre_quo > 0.0) quo = (s7_int)floor(pre_quo); else quo = (s7_int)ceil(pre_quo);
	    return(make_real(sc, real(x) - frac * quo));
	  }

	case T_REAL:
	  return(make_real(sc, c_rem_dbl(sc, real(x), real(y))));

	  /* see under sin -- this calculation is completely bogus if "a" is large
	   * (quotient 1e22 (* 2 pi)) -> -9223372036854775808 -- should this return arithmetic-overflow?
	   *          but it should be 1591549430918953357688,
	   * (remainder 1e22 (* 2 pi)) -> 1.0057952155665e+22
	   * -- the "remainder" is greater than the original argument!
	   * Clisp gives 0.0 here, as does sbcl
	   * currently s7 throws an error (out-of-range).
	   */

	default:
	  method_or_bust(sc, y, sc->remainder_symbol, args, T_REAL, 2);
	}

    default:
      method_or_bust(sc, x, sc->remainder_symbol, args, T_REAL, 1);
    }
}


/* -------------------------------- floor -------------------------------- */

#define REAL_TO_INT_LIMIT 9.2233727815085e+18
/* unfortunately, this limit is only a max in a sense: (ceiling 9223372036854770.9) => 9223372036854770
 *    see s7test for more examples
 */

static s7_pointer g_floor(s7_scheme *sc, s7_pointer args)
{
  #define H_floor "(floor x) returns the integer closest to x toward -inf"
  #define Q_floor s7_make_signature(sc, 2, sc->is_integer_symbol, sc->is_real_symbol)

  s7_pointer x;

  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:
      return(x);

    case T_RATIO:
      {
	s7_int val;
	val = numerator(x) / denominator(x);
	/* C "/" truncates? -- C spec says "truncation toward 0" */
	/* we're avoiding "floor" here because the int->double conversion introduces inaccuracies for big numbers */
	if (numerator(x) < 0) /* not "val" because it might be truncated to 0 */
	  return(make_integer(sc, val - 1));
	return(make_integer(sc, val));
      }

    case T_REAL:
      {
	s7_double z;
	z = real(x);
	if (is_NaN(z))
	  return(simple_out_of_range(sc, sc->floor_symbol, x, its_nan_string));
	if (fabs(z) > REAL_TO_INT_LIMIT)
	  return(simple_out_of_range(sc, sc->floor_symbol, x, its_too_large_string));
	return(make_integer(sc, (s7_int)floor(z)));
	/* floor here rounds down, whereas a straight int<=real coercion apparently rounds towards 0 */
      }

    case T_COMPLEX:
    default:
      method_or_bust_one_arg(sc, x, sc->floor_symbol, args, T_REAL);
    }
}

static s7_int floor_i_i(s7_int i) {return(i);}
static s7_int floor_i_d(s7_double x) 
{
  if (is_NaN(x))
    simple_out_of_range(cur_sc, cur_sc->floor_symbol, make_real(cur_sc, x), its_nan_string);
  if (fabs(x) > REAL_TO_INT_LIMIT)
    simple_out_of_range(cur_sc, cur_sc->floor_symbol, make_real(cur_sc, x), its_too_large_string);
  return((s7_int)floor(x));
}

static s7_double floor_d_d(s7_double x) 
{
  if (is_NaN(x))
    simple_out_of_range(cur_sc, cur_sc->floor_symbol, make_real(cur_sc, x), its_nan_string);
  if (fabs(x) > REAL_TO_INT_LIMIT)
    simple_out_of_range(cur_sc, cur_sc->floor_symbol, make_real(cur_sc, x), its_too_large_string);
  return(floor(x));
}

static s7_int floor_i_p(s7_pointer p)
{
  if (is_t_integer(p)) return(s7_integer(p));
  if (is_t_real(p)) return((s7_int)floor(real(p)));
  if (is_t_ratio(p)) return((s7_int)(floor(fraction(p))));
  s7_wrong_type_arg_error(cur_sc, "floor", 0, p, "a real number");
  return(0);
}


/* -------------------------------- ceiling -------------------------------- */
static s7_pointer g_ceiling(s7_scheme *sc, s7_pointer args)
{
  #define H_ceiling "(ceiling x) returns the integer closest to x toward inf"
  #define Q_ceiling s7_make_signature(sc, 2, sc->is_integer_symbol, sc->is_real_symbol)

  s7_pointer x;

  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:
      return(x);

    case T_RATIO:
      {
	s7_int val;
	val = numerator(x) / denominator(x);
	if (numerator(x) < 0)
	  return(make_integer(sc, val));
	return(make_integer(sc, val + 1));
      }

    case T_REAL:
      {
	s7_double z;
	z = real(x);
	if (is_NaN(z))
	  return(simple_out_of_range(sc, sc->ceiling_symbol, x, its_nan_string));
	if ((is_inf(z)) ||
	    (z > REAL_TO_INT_LIMIT) ||
	    (z < -REAL_TO_INT_LIMIT))
	  return(simple_out_of_range(sc, sc->ceiling_symbol, x, its_too_large_string));
	return(make_integer(sc, (s7_int)ceil(real(x))));
      }

    case T_COMPLEX:
    default:
      method_or_bust_one_arg(sc, x, sc->ceiling_symbol, args, T_REAL);
    }
}

static s7_int ceiling_i_i(s7_int i) {return(i);}
static s7_int ceiling_i_d(s7_double x) 
{
  if (is_NaN(x))
    simple_out_of_range(cur_sc, cur_sc->ceiling_symbol, make_real(cur_sc, x), its_nan_string);
  if ((is_inf(x)) ||
      (x > REAL_TO_INT_LIMIT) ||
      (x < -REAL_TO_INT_LIMIT))
    simple_out_of_range(cur_sc, cur_sc->ceiling_symbol, make_real(cur_sc, x), its_too_large_string);
  return((s7_int)ceil(x));
}

static s7_double ceiling_d_d(s7_double x) 
{
  if (is_NaN(x))
    simple_out_of_range(cur_sc, cur_sc->ceiling_symbol, make_real(cur_sc, x), its_nan_string);
  if ((is_inf(x)) ||
      (x > REAL_TO_INT_LIMIT) ||
      (x < -REAL_TO_INT_LIMIT))
    simple_out_of_range(cur_sc, cur_sc->ceiling_symbol, make_real(cur_sc, x), its_too_large_string);
  return(ceil(x));
}


/* -------------------------------- truncate -------------------------------- */
static s7_pointer g_truncate(s7_scheme *sc, s7_pointer args)
{
  #define H_truncate "(truncate x) returns the integer closest to x toward 0"
  #define Q_truncate s7_make_signature(sc, 2, sc->is_integer_symbol, sc->is_real_symbol)

  s7_pointer x;
  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:
      return(x);

    case T_RATIO:
      return(make_integer(sc, (s7_int)(numerator(x) / denominator(x)))); /* C "/" already truncates */

    case T_REAL:
      {
	s7_double z;
	z = real(x);
	if (is_NaN(z))
	  return(simple_out_of_range(sc, sc->truncate_symbol, x, its_nan_string));
	if (is_inf(z))
	  return(simple_out_of_range(sc, sc->truncate_symbol, x, its_infinite_string));
	return(s7_truncate(sc, sc->truncate_symbol, real(x)));
      }

    case T_COMPLEX:
    default:
      method_or_bust_one_arg(sc, x, sc->truncate_symbol, args, T_REAL);
    }
}

static s7_int truncate_i_i(s7_int i) {return(i);}

static s7_double truncate_d_d(s7_double x) 
{
  if (is_NaN(x))
    simple_out_of_range(cur_sc, cur_sc->truncate_symbol, make_real(cur_sc, x), its_nan_string);
  if (is_inf(x))
    simple_out_of_range(cur_sc, cur_sc->truncate_symbol, make_real(cur_sc, x), its_infinite_string);
  if (x > 0.0) return(floor(x)); 
  return(ceil(x));
}

static s7_int truncate_i_d(s7_double x) {return((s7_int)truncate_d_d(x));}


/* -------------------------------- round -------------------------------- */
static s7_double round_per_R5RS(s7_double x)
{
  s7_double fl, ce, dfl, dce;

  fl = floor(x);
  ce = ceil(x);
  dfl = x - fl;
  dce = ce - x;

  if (dfl > dce) return(ce);
  if (dfl < dce) return(fl);
  if (fmod(fl, 2.0) == 0.0) return(fl);
  return(ce);
}

static s7_pointer g_round(s7_scheme *sc, s7_pointer args)
{
  #define H_round "(round x) returns the integer closest to x"
  #define Q_round s7_make_signature(sc, 2, sc->is_integer_symbol, sc->is_real_symbol)

  s7_pointer x;
  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:
      return(x);

    case T_RATIO:
      {
	s7_int truncated, remains;
	long double frac;

	truncated = numerator(x) / denominator(x);
	remains = numerator(x) % denominator(x);
	frac = s7_fabsl((long double)remains / (long double)denominator(x));

	if ((frac > 0.5) ||
	    ((frac == 0.5) &&
	     (truncated % 2 != 0)))
	  {
	    if (numerator(x) < 0)
	      return(make_integer(sc, truncated - 1));
	    return(make_integer(sc, truncated + 1));
	  }
	return(make_integer(sc, truncated));
      }

    case T_REAL:
      {
	s7_double z;
	z = real(x);
	if (is_NaN(z))
	  return(simple_out_of_range(sc, sc->round_symbol, x, its_nan_string));
	if ((is_inf(z)) ||
	    (z > REAL_TO_INT_LIMIT) ||
	    (z < -REAL_TO_INT_LIMIT))
	  return(simple_out_of_range(sc, sc->round_symbol, x, its_too_large_string));
	return(make_integer(sc, (s7_int)round_per_R5RS(z)));
      }

    case T_COMPLEX:
    default:
      method_or_bust_one_arg(sc, x, sc->round_symbol, args, T_REAL);
    }
}

static s7_int round_i_i(s7_int i) {return(i);}
static s7_int round_i_d(s7_double z) 
{
  if (is_NaN(z))
    simple_out_of_range(cur_sc, cur_sc->round_symbol, make_real(cur_sc, z), its_nan_string);
  if ((is_inf(z)) ||
      (z > REAL_TO_INT_LIMIT) ||
      (z < -REAL_TO_INT_LIMIT))
    simple_out_of_range(cur_sc, cur_sc->round_symbol, make_real(cur_sc, z), its_too_large_string);
  return((s7_int)round_per_R5RS(z));
}

static s7_double round_d_d(s7_double z) 
{
  if (is_NaN(z))
    simple_out_of_range(cur_sc, cur_sc->round_symbol, make_real(cur_sc, z), its_nan_string);
  if ((is_inf(z)) ||
      (z > REAL_TO_INT_LIMIT) ||
      (z < -REAL_TO_INT_LIMIT))
    simple_out_of_range(cur_sc, cur_sc->round_symbol, make_real(cur_sc, z), its_too_large_string);
  return(round_per_R5RS(z));
}


static s7_int c_mod(s7_int x, s7_int y)
{
  s7_int z;
  if (y == 0) return(x);     /* else arithmetic exception */
  if ((y == 1) || (y == -1)) /* else (modulo most-negative-fixnum -1) will segfault with arithmetic exception */
    return(0);
  z = x % y;
  if (((y < 0) && (z > 0)) ||
      ((y > 0) && (z < 0)))
    return(z + y);
  return(z);
}

static s7_int modulo_i_ii(s7_int i1, s7_int i2) {return(c_mod(i1, i2));}
static s7_int modulo_i_ii_direct(s7_int i1, s7_int i2)
{
  /* i2 > 1 */
  s7_int z;
  z = i1 % i2;
  if (z < 0)
    return(z + i2);
  return(z);
}
static s7_double modulo_d_dd(s7_double x1, s7_double x2) {return(x1 - x2 * (s7_int)floor(x1 / x2));}

static s7_pointer g_modulo(s7_scheme *sc, s7_pointer args)
{
  #define H_modulo "(modulo x1 x2) returns x1 mod x2; (modulo 4 3) = 1.  The arguments can be real numbers."
  #define Q_modulo pcl_r
  /* (define (mod x1 x2) (- x1 (* x2 (floor (/ x1 x2))))) from slib
   * (mod x 0) = x according to "Concrete Mathematics"
   */
  s7_pointer x, y;
  s7_double a, b;
  s7_int n1, n2, d1, d2;

  x = car(args);
  y = cadr(args);

  switch (type(x))
    {
    case T_INTEGER:
      switch (type(y))
	{
	case T_INTEGER:
	  return(make_integer(sc, c_mod(integer(x), integer(y))));

	case T_RATIO:
	  n1 = integer(x);
	  d1 = 1;
	  n2 = numerator(y);
	  d2 = denominator(y);
	  goto RATIO_MOD_RATIO;

	case T_REAL:
	  b = real(y);
	  if (b == 0.0) return(x);
	  if (is_NaN(b)) return(y);
	  if (is_inf(b)) return(real_NaN);
	  a = (s7_double)integer(x);
	  return(make_real(sc, a - b * (s7_int)floor(a / b)));

	default:
	  method_or_bust(sc, y, sc->modulo_symbol, args, T_REAL, 2);
	}

    case T_RATIO:
      switch (type(y))
	{
	case T_INTEGER:
	  if (integer(y) == 0) return(x);
	  n1 = numerator(x);
	  d1 = denominator(x);
	  n2 = integer(y);

	  if ((n2 > 0) && (n1 > 0) && (n2 > n1)) return(x);
	  if ((n2 < 0) && (n1 < 0) && (n2 < n1)) return(x);

	  if (n2 == s7_int_min)
	    return(simple_out_of_range(sc, sc->modulo_symbol, y, s7_make_string_wrapper(sc, "intermediate (a/b) is too large")));
	  /* the problem here is that (modulo 3/2 most-negative-fixnum)
	   * will segfault with signal SIGFPE, Arithmetic exception, so try to trap it.
	   */

	  d2 = 1;
	  goto RATIO_MOD_RATIO;

	case T_RATIO:
	  n1 = numerator(x);
	  d1 = denominator(x);
	  n2 = numerator(y); /* can't be 0 */
	  d2 = denominator(y);
	  if (d1 == d2)
	    return(s7_make_ratio(sc, c_mod(n1, n2), d1));

	RATIO_MOD_RATIO:

	  if ((n1 == n2) &&
	      (d1 > d2))
	    return(x);                 /* signs match so this should be ok */
#if HAVE_OVERFLOW_CHECKS
	  {
	    s7_int n2d1, n1d2, d1d2, fl;
	    if (!multiply_overflow(n2, d1, &n2d1))
	      {
		if (n2d1 == 1)
		  return(small_int(0));
		
		if (!multiply_overflow(n1, d2, &n1d2))
		  {
		    /* can't use "floor" here (int->float ruins everything) */
		    fl = (s7_int)(n1d2 / n2d1);
		    if (((n1 < 0) && (n2 > 0)) ||
			((n1 > 0) && (n2 < 0)))
		      fl -= 1;

		    if (fl == 0)
		      return(x);

		    if ((!multiply_overflow(d1, d2, &d1d2)) &&
			(!multiply_overflow(fl, n2d1, &fl)) &&
			(!subtract_overflow(n1d2, fl, &fl)))
		      return(s7_make_ratio(sc, fl, d1d2));
		  }
	      }
	  }
#else
	  {
	    s7_int n1d2, n2d1, fl;
	    n1d2 = n1 * d2;
	    n2d1 = n2 * d1;
	    
	    if (n2d1 == 1)
	      return(small_int(0));
	    
	    /* can't use "floor" here (int->float ruins everything) */
	    fl = (s7_int)(n1d2 / n2d1);
	    if (((n1 < 0) && (n2 > 0)) ||
		((n1 > 0) && (n2 < 0)))
	      fl -= 1;
	    
	    if (fl == 0)
	      return(x);
	    
	    return(s7_make_ratio(sc, n1d2 - (n2d1 * fl), d1 * d2));
	  }
#endif

	  /* there are cases here we might want to catch:
	   *    (modulo 9223372036 1/9223372036) -> error, not 0?
	   *    (modulo 1 1/9223372036854775807) -> error, not 0?
	   */
	  return(simple_out_of_range(sc, sc->modulo_symbol, x, s7_make_string_wrapper(sc, "intermediate (a/b) is too large")));

	case T_REAL:
	  b = real(y);
	  if (b == 0.0) return(x);
	  if (is_NaN(b)) return(y);
	  if (is_inf(b)) return(real_NaN);
	  a = fraction(x);
	  return(make_real(sc, a - b * (s7_int)floor(a / b)));

	default:
	  method_or_bust(sc, y, sc->modulo_symbol, args, T_REAL, 2);
	}

    case T_REAL:
      a = real(x);

      switch (type(y))
	{
	case T_INTEGER:
	  if (is_NaN(a)) return(x);
	  if (is_inf(a)) return(real_NaN);
	  if (integer(y) == 0) return(x);
	  b = (s7_double)integer(y);
	  return(make_real(sc, a - b * (s7_int)floor(a / b)));

	case T_RATIO:
	  if (is_NaN(a)) return(x);
	  if (is_inf(a)) return(real_NaN);
	  b = fraction(y);
	  return(make_real(sc, a - b * (s7_int)floor(a / b)));

	case T_REAL:
	  if (is_NaN(a)) return(x);
	  if (is_inf(a)) return(real_NaN);
	  b = real(y);
	  if (b == 0.0) return(x);
	  if (is_NaN(b)) return(y);
	  if (is_inf(b)) return(real_NaN);
	  return(make_real(sc, a - b * (s7_int)floor(a / b)));

	default:
	  method_or_bust(sc, y, sc->modulo_symbol, args, T_REAL, 2);
	}

    default:
      method_or_bust(sc, x, sc->modulo_symbol, args, T_REAL, 1);
    }
}

static s7_pointer mod_si;
static s7_pointer g_mod_si(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x;
  s7_int y;

  x = find_symbol_unchecked(sc, car(args));
  y = integer(cadr(args));

  if (is_integer(x))
    {
      s7_int z;
      /* here we know y is positive */
      z = integer(x) % y;
      if (z < 0)
	return(make_integer(sc, z + y));
      return(make_integer(sc, z));
    }

  if (is_t_real(x))
    {
      s7_double a, b;
      a = real(x);
      if (is_NaN(a)) return(x);
      if (is_inf(a)) return(real_NaN);
      b = (s7_double)y;
      return(make_real(sc, a - b * (s7_int)floor(a / b)));
    }

  if (s7_is_ratio(x))
    return(g_modulo(sc, set_plist_2(sc, x, cadr(args))));

  method_or_bust(sc, x, sc->modulo_symbol, list_2(sc, x, cadr(args)), T_REAL, 1);
}
#endif
/* !WITH_GMP */


static int32_t reduce_fraction(s7_scheme *sc, s7_int *numer, s7_int *denom)
{
  /* we're assuming in several places that we have a normal s7 rational after returning,
   *    so the denominator needs to be positive.
   */
  s7_int divisor;

  if (*numer == 0)
    {
      *denom = 1;
      return(T_INTEGER);
    }
  if (*denom < 0)
    {
      if (*denom == *numer)
	{
	  *denom = 1;
	  *numer = 1;
	  return(T_INTEGER);
	}
      if (*denom == s7_int_min)
	{
	  if (*numer & 1)
	    return(T_RATIO);
	  *denom /= 2;
	  *numer /= 2;
	}
      else
	{
	  if (*numer == s7_int_min)
	    {
	      if (*denom & 1)
		return(T_RATIO);
	      *denom /= 2;
	      *numer /= 2;
	    }
	}
      *denom = -*denom; 
      *numer = -*numer;
    }
  divisor = c_gcd(*numer, *denom);
  if (divisor != 1)
    {
      *numer /= divisor;
      *denom /= divisor;
    }
  if (*denom == 1)
    return(T_INTEGER);
  return(T_RATIO);
}



/* ---------------------------------------- add ---------------------------------------- */

static s7_pointer g_add(s7_scheme *sc, s7_pointer args)
{
  #define H_add "(+ ...) adds its arguments"
  #define Q_add pcl_n
  s7_pointer x, p;
  s7_int num_a, den_a, dn;
  s7_double rl_a, im_a;

#if (!WITH_GMP)
  if (is_null(args))
    return(small_int(0));
#endif

  x = car(args);
  p = cdr(args);
  if (is_null(p))
    {
      if (!is_number(x))
	method_or_bust_with_type_one_arg(sc, x, sc->add_symbol, args, a_number_string);
      return(x);
    }

  switch (type(x))
    {
    case T_INTEGER:
      num_a = integer(x);

    ADD_INTEGERS:
#if WITH_GMP
      if ((num_a > s7_int32_max) ||
	  (num_a < s7_int32_min))
	return(big_add(sc, cons(sc, s7_int_to_big_integer(sc, num_a), p)));
#endif
      x = car(p);
      p = cdr(p);

      switch (type(x))
	{
	case T_INTEGER:
#if HAVE_OVERFLOW_CHECKS
	  if (add_overflow(num_a, integer(x), &den_a))
	    {
	      rl_a = (s7_double)num_a + (s7_double)integer(x);
	      if (is_null(p)) return(make_real(sc, rl_a));
	      goto ADD_REALS;
	    }
#else	    
	  den_a = num_a + integer(x);
	  if (den_a < 0)
	    {
	      if ((num_a > 0) && (integer(x) > 0))
		{
		  rl_a = (s7_double)num_a + (s7_double)integer(x);
		  if (is_null(p)) return(make_real(sc, rl_a));
		  goto ADD_REALS;
		}
	    }
	  else
	    {
	      if ((num_a < 0) && (integer(x) < 0))
		{
		  rl_a = (s7_double)num_a + (s7_double)integer(x);
		  if (is_null(p)) return(make_real(sc, rl_a));

		  /* this is not ideal!  piano.scm has its own noise generator that wants integer
		   *    arithmetic to overflow as an integer.  Perhaps 'safety==0 would not check
		   *    anywhere?
		   */
		  goto ADD_REALS;
		}
	    }
#endif
	  if (is_null(p)) return(make_integer(sc, den_a));
	  num_a = den_a;
	  /* (+ 4611686018427387904 4611686018427387904) -> -9223372036854775808
	   * (+ most-positive-fixnum most-positive-fixnum) -> -2
	   * (+ most-negative-fixnum most-negative-fixnum) -> 0
	   * can't check result - arg: (- 0 most-negative-fixnum) -> most-negative-fixnum
	   */
	  goto ADD_INTEGERS;

	case T_RATIO:
	  den_a = denominator(x);
#if HAVE_OVERFLOW_CHECKS
	  if ((multiply_overflow(den_a, num_a, &dn)) || 
	      (add_overflow(dn, numerator(x), &dn)))
	    {
	      if (is_null(p))
		{
		  if (num_a == 0)                /* (+ 0 1/9223372036854775807) */
		    return(x);
		  return(make_real(sc, num_a + fraction(x)));
		}
	      rl_a = (s7_double)num_a + fraction(x);
	      goto ADD_REALS;
	    }
#else
	  dn = numerator(x) + (num_a * den_a);
#endif
	  if (is_null(p)) return(s7_make_ratio(sc, dn, den_a));
	  num_a = dn;

	  /* overflow examples:
	   *   (+ 100000 1/142857142857140) -> -832205957599110323/28571428571428
	   *   (+ 4611686018427387904 3/4) -> 3/4
	   * see s7test for more
	   */
	  goto ADD_RATIOS;

	case T_REAL:
	  if (is_null(p)) return(make_real(sc, num_a + real(x)));
	  rl_a = (s7_double)num_a + real(x);
	  goto ADD_REALS;

	case T_COMPLEX:
	  if (is_null(p)) return(s7_make_complex(sc, num_a + real_part(x), imag_part(x)));
	  rl_a = (s7_double)num_a + real_part(x);
	  im_a = imag_part(x);
	  goto ADD_COMPLEX;

	default:
	  method_or_bust_with_type(sc, x, sc->add_symbol, cons_unchecked(sc, s7_make_integer(sc, num_a), cons(sc, x, p)), a_number_string, position_of(p, args) - 1);
	}
      break;

    case T_RATIO:
      num_a = numerator(x);
      den_a = denominator(x);
    ADD_RATIOS:
#if WITH_GMP
      if ((num_a > s7_int32_max) ||
	  (den_a > s7_int32_max) ||
	  (num_a < s7_int32_min))
	return(big_add(sc, cons(sc, s7_ratio_to_big_ratio(sc, num_a, den_a), p)));
#endif
      x = car(p);
      p = cdr(p);

      switch (type(x))
	{
	case T_INTEGER:
#if HAVE_OVERFLOW_CHECKS
	  if ((multiply_overflow(den_a, integer(x), &dn)) || 
	      (add_overflow(dn, num_a, &dn)))
	    {
	      /* (+ 3/4 4611686018427387904) -> 3/4
	       * (+ 1/17179869184 1073741824) -> 1/17179869184
	       * (+ 1/8589934592 1073741824) -> -9223372036854775807/8589934592
	       */
	      if (is_null(p))
		return(make_real(sc, (s7_double)integer(x) + ((long double)num_a / (long double)den_a)));
	      rl_a = (s7_double)integer(x) + ((long double)num_a / (long double)den_a);
	      goto ADD_REALS;
	    }
#else
	  dn = num_a + (integer(x) * den_a);
#endif
	  if (is_null(p)) return(s7_make_ratio(sc, dn, den_a));
	  num_a = dn;
	  if (reduce_fraction(sc, &num_a, &den_a) == T_INTEGER)
	    goto ADD_INTEGERS;
	  goto ADD_RATIOS;

	case T_RATIO:
	  {
	    s7_int d1, d2, n1, n2;
	    d1 = den_a;
	    n1 = num_a;
	    d2 = denominator(x);
	    n2 = numerator(x);
	    if (d1 == d2)                                     /* the easy case -- if overflow here, it matches the int32_t case */
	      {
		if (is_null(p))
		  return(s7_make_ratio(sc, n1 + n2, d1));
		num_a += n2;                  /* d1 can't be zero */
	      }
	    else
	      {
#if (!WITH_GMP)
#if HAVE_OVERFLOW_CHECKS
		s7_int n1d2, n2d1;
		if ((multiply_overflow(d1, d2, &den_a)) ||
		    (multiply_overflow(n1, d2, &n1d2)) ||
		    (multiply_overflow(n2, d1, &n2d1)) ||
		    (add_overflow(n1d2, n2d1, &num_a)))
		  {
		    if (is_null(p))
		      return(make_real(sc, ((long double)n1 / (long double)d1) + ((long double)n2 / (long double)d2)));
		    rl_a = ((long double)n1 / (long double)d1) + ((long double)n2 / (long double)d2);
		    goto ADD_REALS;
		  }
#else
		num_a = n1 * d2 + n2 * d1;
		den_a = d1 * d2;
#endif
#else
		num_a = n1 * d2 + n2 * d1;
		den_a = d1 * d2;
#endif
		if (is_null(p))
		  return(s7_make_ratio(sc, num_a, den_a));
	      }
	    /* (+ 1/100 99/100 (- most-positive-fixnum 2)) should not be converted to real
	     */
	    if (reduce_fraction(sc, &num_a, &den_a) == T_INTEGER)
	    goto ADD_INTEGERS;
	  goto ADD_RATIOS;
	  }

	case T_REAL:
	  if (is_null(p)) return(make_real(sc, ((long double)num_a / (long double)den_a) + real(x)));
	  rl_a = ((long double)num_a / (long double)den_a) + real(x);
	  goto ADD_REALS;

	case T_COMPLEX:
	  if (is_null(p)) return(s7_make_complex(sc, ((long double)num_a / (long double)den_a) + real_part(x), imag_part(x)));
	  rl_a = ((long double)num_a / (long double)den_a) + real_part(x);
	  im_a = imag_part(x);
	  goto ADD_COMPLEX;

	default:
	  method_or_bust_with_type(sc, x, sc->add_symbol, cons_unchecked(sc, s7_make_ratio(sc, num_a, den_a), cons(sc, x, p)), a_number_string, position_of(p, args) - 1);
	}
      break;

    case T_REAL:
      rl_a = real(x);

    ADD_REALS:
      x = car(p);
      p = cdr(p);

      switch (type(x))
	{
	case T_INTEGER:
	  if (is_null(p)) return(make_real(sc, rl_a + integer(x)));
	  rl_a += (s7_double)integer(x);
	  goto ADD_REALS;

	case T_RATIO:
	  if (is_null(p)) return(make_real(sc, rl_a + fraction(x)));
	  rl_a += (s7_double)fraction(x);
	  goto ADD_REALS;

	case T_REAL:
	  if (is_null(p)) return(make_real(sc, rl_a + real(x)));
	  rl_a += real(x);
	  goto ADD_REALS;

	case T_COMPLEX:
	  if (is_null(p)) return(s7_make_complex(sc, rl_a + real_part(x), imag_part(x)));
	  rl_a += real_part(x);
	  im_a = imag_part(x);
	  goto ADD_COMPLEX;

	default:
	  method_or_bust_with_type(sc, x, sc->add_symbol, cons_unchecked(sc, make_real(sc, rl_a), cons(sc, x, p)), a_number_string, position_of(p, args) - 1);
	}
      break;

    case T_COMPLEX:
      rl_a = real_part(x);
      im_a = imag_part(x);

    ADD_COMPLEX:
      x = car(p);
      p = cdr(p);

      switch (type(x))
	{
	case T_INTEGER:
	  if (is_null(p)) return(s7_make_complex(sc, rl_a + integer(x), im_a));
	  rl_a += (s7_double)integer(x);
	  goto ADD_COMPLEX;

	case T_RATIO:
	  if (is_null(p)) return(s7_make_complex(sc, rl_a + fraction(x), im_a));
	  rl_a += (s7_double)fraction(x);
	  goto ADD_COMPLEX;

	case T_REAL:
	  if (is_null(p)) return(s7_make_complex(sc, rl_a + real(x), im_a));
	  rl_a += real(x);
	  goto ADD_COMPLEX;

	case T_COMPLEX:
	  if (is_null(p)) return(s7_make_complex(sc, rl_a + real_part(x), im_a + imag_part(x)));
	  rl_a += real_part(x);
	  im_a += imag_part(x);
	  if (im_a == 0.0)
	    goto ADD_REALS;
	  goto ADD_COMPLEX;

	default:
	  method_or_bust_with_type(sc, x, sc->add_symbol, cons_unchecked(sc, s7_make_complex(sc, rl_a, im_a), cons(sc, x, p)), a_number_string, position_of(p, args) - 1);
	}
      break;

    default:
      method_or_bust_with_type(sc, x, sc->add_symbol, args, a_number_string, 1);
    }
}


static s7_pointer add_2, add_1s, add_s1, add_cs1, add_cl1, add_si, add_sf, add_fs;

static s7_pointer add_ratios(s7_scheme *sc, s7_pointer x, s7_pointer y)
{
  s7_int d1, d2, n1, n2;
  d1 = number_to_denominator(x);
  n1 = number_to_numerator(x);
  d2 = number_to_denominator(y);
  n2 = number_to_numerator(y);

  if (d1 == d2)                                     /* the easy case -- if overflow here, it matches the int32_t case */
    return(s7_make_ratio(sc, n1 + n2, d1));

#if HAVE_OVERFLOW_CHECKS
  {
    s7_int n1d2, n2d1, d1d2, dn;
    if ((multiply_overflow(d1, d2, &d1d2)) ||
	(multiply_overflow(n1, d2, &n1d2)) ||
	(multiply_overflow(n2, d1, &n2d1)) ||
	(add_overflow(n1d2, n2d1, &dn)))
      return(make_real(sc, ((long double)n1 / (long double)d1) + ((long double)n2 / (long double)d2)));
    return(s7_make_ratio(sc, dn, d1d2));
  }
#else
  return(s7_make_ratio(sc, n1 * d2 + n2 * d1, d1 * d2));
#endif
}


static s7_pointer g_add_2(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x, y;
  x = car(args);
  y = cadr(args);

  if (type(x) == type(y))
    {
      if (is_t_real(x))
	return(make_real(sc, real(x) + real(y)));
      else
	{
	  switch (type(x))
	    {
#if HAVE_OVERFLOW_CHECKS
	    case T_INTEGER:
	      {
		s7_int val;
		if (add_overflow(integer(x), integer(y), &val))
		  return(make_real(sc, (double)integer(x) + (double)integer(y)));
		return(make_integer(sc, val));
	      }
#else
	    case T_INTEGER: return(make_integer(sc, integer(x) + integer(y)));
#endif
	    case T_RATIO:   return(add_ratios(sc, x, y));
	    case T_REAL:    return(make_real(sc, real(x) + real(y)));
	    case T_COMPLEX: return(make_complex(sc, real_part(x) + real_part(y), imag_part(x) + imag_part(y)));
	    default:
	      if (!is_number(x))
		method_or_bust_with_type(sc, x, sc->add_symbol, args, a_number_string, 1);
	      method_or_bust_with_type(sc, y, sc->add_symbol, args, a_number_string, 2);
	    }
	}
    }

  switch (type(x))
    {
    case T_INTEGER:
      switch (type(y))
	{
	case T_INTEGER: return(make_integer(sc, integer(x) + integer(y)));
	case T_RATIO:   return(add_ratios(sc, x, y));
	case T_REAL:    return(make_real(sc, integer(x) + real(y)));
	case T_COMPLEX: return(make_complex(sc, integer(x) + real_part(y), imag_part(y)));
	default:
	  method_or_bust_with_type(sc, y, sc->add_symbol, args, a_number_string, 2);
	}

    case T_RATIO:
      switch (type(y))
	{
	case T_INTEGER:
	case T_RATIO:   return(add_ratios(sc, x, y));
	case T_REAL:    return(make_real(sc, fraction(x) + real(y)));
	case T_COMPLEX: return(s7_make_complex(sc, fraction(x) + real_part(y), imag_part(y)));
	default:
	  method_or_bust_with_type(sc, y, sc->add_symbol, args, a_number_string, 2);
	}

    case T_REAL:
      switch (type(y))
	{
	case T_INTEGER: return(make_real(sc, real(x) + integer(y)));
	case T_RATIO:   return(make_real(sc, real(x) + fraction(y)));
	case T_REAL:    return(make_real(sc, real(x) + real(y)));
	case T_COMPLEX: return(make_complex(sc, real(x) + real_part(y), imag_part(y)));
	default:
	  method_or_bust_with_type(sc, y, sc->add_symbol, args, a_number_string, 2);
	}

    case T_COMPLEX:
      switch (type(y))
	{
	case T_INTEGER: return(s7_make_complex(sc, real_part(x) + integer(y), imag_part(x)));
	case T_RATIO:   return(s7_make_complex(sc, real_part(x) + fraction(y), imag_part(x)));
	case T_REAL:    return(s7_make_complex(sc, real_part(x) + real(y), imag_part(x)));
	case T_COMPLEX: return(make_complex(sc, real_part(x) + real_part(y), imag_part(x) + imag_part(y)));
	default:
	  method_or_bust_with_type(sc, y, sc->add_symbol, args, a_number_string, 2);
	}

    default:
      method_or_bust_with_type(sc, x, sc->add_symbol, args, a_number_string, 1);
    }
  return(x);
}

static s7_pointer g_add_s1_1(s7_scheme *sc, s7_pointer x, s7_pointer args)
{
  switch (type(x))
    {
#if HAVE_OVERFLOW_CHECKS
    case T_INTEGER:
      {
	s7_int val;
	if (add_overflow(integer(x), 1, &val))
	  return(make_real(sc, (double)integer(x) + 1.0));
	return(make_integer(sc, val));
      }
#else
    case T_INTEGER: return(make_integer(sc, integer(x) + 1));
#endif
    case T_RATIO:   return(add_ratios(sc, x, small_int(1)));
    case T_REAL:    return(make_real(sc, real(x) + 1.0));
    case T_COMPLEX: return(s7_make_complex(sc, real_part(x) + 1.0, imag_part(x)));
    default:
      method_or_bust_with_type(sc, x, sc->add_symbol, cons(sc, x, cdr(args)), a_number_string, 1);
    }
  return(x);
}

static s7_pointer g_add_s1(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x;
  x = car(args);
  if (is_t_integer(x))
    return(make_integer(sc, integer(x) + 1));
  return(g_add_s1_1(sc, x, args));
}

static s7_pointer g_add_cs1(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x;
  x = find_symbol_unchecked(sc, car(args));
  if (is_integer(x))
    return(make_integer(sc, integer(x) + 1));
  return(g_add_s1_1(sc, x, args));
}

static s7_pointer g_add_cl1(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x;
  x = local_symbol_value(car(args));
  if (is_integer(x))
    return(make_integer(sc, integer(x) + 1));
  return(g_add_s1_1(sc, x, args));
}

static s7_pointer g_add_1s(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x;

  x = cadr(args);
  if (is_integer(x))
    return(make_integer(sc, integer(x) + 1));

  switch (type(x))
    {
    case T_INTEGER: return(make_integer(sc, integer(x) + 1));
    case T_RATIO:   return(add_ratios(sc, x, small_int(1)));
    case T_REAL:    return(make_real(sc, real(x) + 1.0));
    case T_COMPLEX: return(s7_make_complex(sc, real_part(x) + 1.0, imag_part(x)));
    default:
      method_or_bust_with_type(sc, x, sc->add_symbol, args, a_number_string, 2);
    }
  return(x);
}

static s7_pointer g_add_si(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x;
  s7_int n;

  x = find_symbol_unchecked(sc, car(args));
  n = integer(cadr(args));
  if (is_integer(x))
#if HAVE_OVERFLOW_CHECKS
    {
      s7_int val;
      if (add_overflow(integer(x), n, &val))
	return(make_real(sc, (double)integer(x) + (double)n));
      return(make_integer(sc, val));
    }
#else
    return(make_integer(sc, integer(x) + n));
#endif
  switch (type(x))
    {
    case T_INTEGER: return(make_integer(sc, integer(x) + n));
    case T_RATIO:   return(add_ratios(sc, x, cadr(args)));
    case T_REAL:    return(make_real(sc, real(x) + n));
    case T_COMPLEX: return(s7_make_complex(sc, real_part(x) + n, imag_part(x)));
    default:
      method_or_bust_with_type(sc, x, sc->add_symbol, list_2(sc, x, cadr(args)), a_number_string, 1);
    }
  return(x);
}

static s7_pointer g_add_sf(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x;
  s7_double n;

  x = find_symbol_unchecked(sc, car(args));
  n = real(cadr(args));
  switch (type(x))
    {
    case T_INTEGER: return(make_real(sc, integer(x) + n));
    case T_RATIO:   return(make_real(sc, fraction(x) + n));
    case T_REAL:    return(make_real(sc, real(x) + n));
    case T_COMPLEX: return(s7_make_complex(sc, real_part(x) + n, imag_part(x)));
    default:
      method_or_bust_with_type(sc, x, sc->add_symbol, list_2(sc, x, cadr(args)), a_number_string, 1);
    }
  return(x);
}

static s7_pointer g_add_fs(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x;
  s7_double n;

  x = find_symbol_unchecked(sc, cadr(args));
  n = real(car(args));
  switch (type(x))
    {
    case T_INTEGER: return(make_real(sc, integer(x) + n));
    case T_RATIO:   return(make_real(sc, fraction(x) + n));
    case T_REAL:    return(make_real(sc, real(x) + n));
    case T_COMPLEX: return(s7_make_complex(sc, real_part(x) + n, imag_part(x)));
    default:
      method_or_bust_with_type(sc, x, sc->add_symbol, list_2(sc, x, car(args)), a_number_string, 2);
    }
  return(x);
}

static s7_pointer add_f_sf;
static s7_pointer g_add_f_sf(s7_scheme *sc, s7_pointer args)
{
  /* (+ x (* s y)) */
  s7_pointer vargs, s;
  s7_double x, y;

  x = real(car(args));
  vargs = cdadr(args);
  s = find_symbol_unchecked(sc, car(vargs));
  y = real(cadr(vargs));

  if (is_t_real(s))
    return(make_real(sc, x + (real(s) * y)));

  switch (type(s))
    {
    case T_INTEGER: return(make_real(sc, x + (integer(s) * y)));
    case T_RATIO:   return(make_real(sc, x + (fraction(s) * y)));
    case T_REAL:    return(make_real(sc, x + real(s) * y));
    case T_COMPLEX: return(s7_make_complex(sc, x + (real_part(s) * y), imag_part(s) * y));
    default:
      {
	s7_pointer func;
	if ((func = find_method(sc, find_let(sc, s), sc->multiply_symbol)) != sc->undefined)
	  return(g_add_2(sc, set_plist_2(sc, car(args), s7_apply_function(sc, func, list_2(sc, s, cadr(vargs))))));
	return(wrong_type_argument_with_type(sc, sc->multiply_symbol, 1, s, a_number_string));
      }
    }
  return(s);
}


static s7_double add_d_d(s7_double x) {return(x);}
static s7_double add_d_dd(s7_double x1, s7_double x2) {return(x1 + x2);}
static s7_double add_d_ddd(s7_double x1, s7_double x2, s7_double x3) {return(x1 + x2 + x3);}
static s7_double add_d_dddd(s7_double x1, s7_double x2, s7_double x3, s7_double x4) {return(x1 + x2 + x3 + x4);}

static s7_int add_i_ii(s7_int i1, s7_int i2) {return(i1 + i2);}
static s7_int add_i_iii(s7_int i1, s7_int i2, s7_int i3) {return(i1 + i2 + i3);}
static s7_double add_d_id(s7_int x1, s7_double x2) {return((s7_double)x1 + x2);}

#if (!WITH_GMP)
static s7_pointer add_p_dd(s7_double x1, s7_double x2) {return(make_real(cur_sc, x1 + x2));}
static s7_pointer add_p_ii(s7_int x1, s7_int x2) {return(make_integer(cur_sc, x1 + x2));}
#endif
static s7_pointer add_p_pp(s7_pointer p1, s7_pointer p2) {return(g_add_2(cur_sc, set_plist_2(cur_sc, p1, p2)));}


/* ---------------------------------------- subtract ---------------------------------------- */

static s7_pointer g_subtract(s7_scheme *sc, s7_pointer args)
{
  #define H_subtract "(- x1 ...) subtracts its trailing arguments from the first, or negates the first if only one it is given"
  #define Q_subtract pcl_n

  s7_pointer x, p;
  s7_int num_a, den_a;
  s7_double rl_a, im_a;

  x = car(args);
  p = cdr(args);

#if (!WITH_GMP)
  if (is_null(p))
    {
      if (!is_number(x))
	method_or_bust_with_type_one_arg(sc, x, sc->subtract_symbol, args, a_number_string);
      return(s7_negate(sc, x));
    }
#endif

  switch (type(x))
    {
    case T_INTEGER:
      num_a = integer(x);

    SUBTRACT_INTEGERS:
#if WITH_GMP
      if ((num_a > s7_int32_max) ||
	  (num_a < s7_int32_min))
	return(big_subtract(sc, cons(sc, s7_int_to_big_integer(sc, num_a), p)));
#endif
      x = car(p);
      p = cdr(p);

      switch (type(x))
	{
	case T_INTEGER:
#if HAVE_OVERFLOW_CHECKS
	  if (subtract_overflow(num_a, integer(x), &den_a))
	    {
	      rl_a = (s7_double)num_a - (s7_double)integer(x);
	      if (is_null(p)) return(make_real(sc, rl_a));
	      goto SUBTRACT_REALS;
	    }
#else
	  den_a = num_a - integer(x);
	  if (den_a < 0)
	    {
	      if ((num_a > 0) && (integer(x) < 0))
		{
		  rl_a = (s7_double)num_a - (s7_double)integer(x);
		  if (is_null(p)) return(make_real(sc, rl_a));
		  goto SUBTRACT_REALS;
		}
	      /* (- most-positive-fixnum most-negative-fixnum) -> -1 (1.8446744073709551615E19)
	       */
	    }
	  else
	    {
	      if ((num_a < 0) && (integer(x) > 0))
		{
		  rl_a = (s7_double)num_a - (s7_double)integer(x);
		  if (is_null(p)) return(make_real(sc, rl_a));
		  goto SUBTRACT_REALS;
		}
	      /* (- most-negative-fixnum most-positive-fixnum) -> 1 (-1.8446744073709551615E19)
	       */
	    }
#endif
	  if (is_null(p)) return(make_integer(sc, den_a));
	  num_a = den_a;
	  goto SUBTRACT_INTEGERS;

	case T_RATIO:
	  {
	    s7_int dn;
	    den_a = denominator(x);
#if HAVE_OVERFLOW_CHECKS
	    if ((multiply_overflow(num_a, den_a, &dn)) ||
		(subtract_overflow(dn, numerator(x), &dn)))
	      {
		if (is_null(p)) return(make_real(sc, num_a - fraction(x)));
		rl_a = (s7_double)num_a - fraction(x);
		goto SUBTRACT_REALS;
	      }
#else
	    dn = (num_a * den_a) - numerator(x);
#endif
	    if (is_null(p)) return(s7_make_ratio(sc, dn, den_a));
	    num_a = dn;
	    if (reduce_fraction(sc, &num_a, &den_a) == T_INTEGER)
	      goto SUBTRACT_INTEGERS;
	    goto SUBTRACT_RATIOS;
	  }

	case T_REAL:
	  if (is_null(p)) return(make_real(sc, num_a - real(x)));
	  rl_a = (s7_double)num_a - real(x);
	  goto SUBTRACT_REALS;

	case T_COMPLEX:
	  if (is_null(p)) return(s7_make_complex(sc, num_a - real_part(x), -imag_part(x)));
	  rl_a = (s7_double)num_a - real_part(x);
	  im_a = -imag_part(x);
	  goto SUBTRACT_COMPLEX;

	default:
	  method_or_bust_with_type(sc, x, sc->subtract_symbol, cons_unchecked(sc, s7_make_integer(sc, num_a), cons(sc, x, p)), a_number_string, position_of(p, args) - 1);
	}
      break;

    case T_RATIO:
      num_a = numerator(x);
      den_a = denominator(x);
    SUBTRACT_RATIOS:
#if WITH_GMP
      if ((num_a > s7_int32_max) ||
	  (den_a > s7_int32_max) ||
	  (num_a < s7_int32_min))
	return(big_subtract(sc, cons(sc, s7_ratio_to_big_ratio(sc, num_a, den_a), p)));
#endif
      x = car(p);
      p = cdr(p);

      switch (type(x))
	{
	case T_INTEGER:
#if HAVE_OVERFLOW_CHECKS
	  {
	    s7_int di;
	    if ((multiply_overflow(den_a, integer(x), &di)) ||
		(subtract_overflow(num_a, di, &di)))
	      {
		if (is_null(p)) return(make_real(sc, ((long double)num_a / (long double)den_a) - integer(x)));
		rl_a = ((long double)num_a / (long double)den_a) - integer(x);
		goto SUBTRACT_REALS;
	      }
	    if (is_null(p)) return(s7_make_ratio(sc, di, den_a));
	    num_a = di;
	  }
#else
	  if (is_null(p)) return(s7_make_ratio(sc, num_a - (den_a * integer(x)), den_a));
	  num_a -= (den_a * integer(x));
#endif
	  if (reduce_fraction(sc, &num_a, &den_a) == T_INTEGER)
	    goto SUBTRACT_INTEGERS;
	  goto SUBTRACT_RATIOS;

	case T_RATIO:
	  {
	    s7_int d1, d2, n1, n2;
	    d1 = den_a;
	    n1 = num_a;
	    d2 = denominator(x);
	    n2 = numerator(x);
	    if (d1 == d2)                                     /* the easy case -- if overflow here, it matches the int32_t case */
	      {
		if (is_null(p))
		  return(s7_make_ratio(sc, n1 - n2, d1));
		num_a -= n2;                  /* d1 can't be zero */
	      }
	    else
	      {
#if (!WITH_GMP) && HAVE_OVERFLOW_CHECKS
		s7_int n1d2, n2d1;
		if ((multiply_overflow(d1, d2, &den_a)) ||
		    (multiply_overflow(n1, d2, &n1d2)) ||
		    (multiply_overflow(n2, d1, &n2d1)) ||
		    (subtract_overflow(n1d2, n2d1, &num_a)))
		  {
		    if (is_null(p))
		      return(make_real(sc, ((long double)n1 / (long double)d1) - ((long double)n2 / (long double)d2)));
		    rl_a = ((long double)n1 / (long double)d1) - ((long double)n2 / (long double)d2);
		    goto SUBTRACT_REALS;
		  }
#else
		num_a = n1 * d2 - n2 * d1;
		den_a = d1 * d2;
#endif
		if (is_null(p))
		  return(s7_make_ratio(sc, num_a, den_a));
	      }
	    if (reduce_fraction(sc, &num_a, &den_a) == T_INTEGER)
	    goto SUBTRACT_INTEGERS;
	  goto SUBTRACT_RATIOS;
	  }

	case T_REAL:
	  if (is_null(p)) return(make_real(sc, ((long double)num_a / (long double)den_a) - real(x)));
	  rl_a = ((long double)num_a / (long double)den_a) - real(x);
	  goto SUBTRACT_REALS;

	case T_COMPLEX:
	  if (is_null(p)) return(s7_make_complex(sc, ((long double)num_a / (long double)den_a) - real_part(x), -imag_part(x)));
	  rl_a = ((long double)num_a / (long double)den_a) - real_part(x);
	  im_a = -imag_part(x);
	  goto SUBTRACT_COMPLEX;

	default:
	  method_or_bust_with_type(sc, x, sc->subtract_symbol, cons_unchecked(sc, s7_make_ratio(sc, num_a, den_a), cons(sc, x, p)), a_number_string, position_of(p, args) - 1);
	}
      break;

    case T_REAL:
      rl_a = real(x);

    SUBTRACT_REALS:
      x = car(p);
      p = cdr(p);

      switch (type(x))
	{
	case T_INTEGER:
	  if (is_null(p)) return(make_real(sc, rl_a - integer(x)));
	  rl_a -= (s7_double)integer(x);
	  goto SUBTRACT_REALS;

	case T_RATIO:
	  if (is_null(p)) return(make_real(sc, rl_a - fraction(x)));
	  rl_a -= (s7_double)fraction(x);
	  goto SUBTRACT_REALS;

	case T_REAL:
	  if (is_null(p)) return(make_real(sc, rl_a - real(x)));
	  rl_a -= real(x);
	  goto SUBTRACT_REALS;

	case T_COMPLEX:
	  if (is_null(p)) return(s7_make_complex(sc, rl_a - real_part(x), -imag_part(x)));
	  rl_a -= real_part(x);
	  im_a = -imag_part(x);
	  goto SUBTRACT_COMPLEX;

	default:
	  method_or_bust_with_type(sc, x, sc->subtract_symbol, cons_unchecked(sc, make_real(sc, rl_a), cons(sc, x, p)), a_number_string, position_of(p, args) - 1);
	}
      break;

    case T_COMPLEX:
      rl_a = real_part(x);
      im_a = imag_part(x);

    SUBTRACT_COMPLEX:
      x = car(p);
      p = cdr(p);

      switch (type(x))
	{
	case T_INTEGER:
	  if (is_null(p)) return(s7_make_complex(sc, rl_a - integer(x), im_a));
	  rl_a -= (s7_double)integer(x);
	  goto SUBTRACT_COMPLEX;

	case T_RATIO:
	  if (is_null(p)) return(s7_make_complex(sc, rl_a - fraction(x), im_a));
	  rl_a -= (s7_double)fraction(x);
	  goto SUBTRACT_COMPLEX;

	case T_REAL:
	  if (is_null(p)) return(s7_make_complex(sc, rl_a - real(x), im_a));
	  rl_a -= real(x);
	  goto SUBTRACT_COMPLEX;

	case T_COMPLEX:
	  if (is_null(p)) return(s7_make_complex(sc, rl_a - real_part(x), im_a - imag_part(x)));
	  rl_a -= real_part(x);
	  im_a -= imag_part(x);
	  if (im_a == 0.0)
	    goto SUBTRACT_REALS;
	  goto SUBTRACT_COMPLEX;

	default:
	  method_or_bust_with_type(sc, x, sc->subtract_symbol, cons_unchecked(sc, s7_make_complex(sc, rl_a, im_a), cons(sc, x, p)), a_number_string, position_of(p, args) - 1);
	}
      break;

    default:
      method_or_bust_with_type(sc, x, sc->subtract_symbol, args, a_number_string, 1);
    }
}


static s7_pointer subtract_1, subtract_s1, subtract_cs1, subtract_cl1, subtract_2, subtract_csn;
static s7_pointer g_subtract_1(s7_scheme *sc, s7_pointer args)
{
  s7_pointer p;

  p = car(args);
  switch (type(p))
    {
    case T_INTEGER:
      if (integer(p) == s7_int_min)
#if WITH_GMP
	return(big_negate(sc, set_plist_1(sc, promote_number(sc, T_BIG_INTEGER, p))));
#else
        return(make_integer(sc, s7_int_max));
#endif
      return(make_integer(sc, -integer(p)));

    case T_RATIO:
      return(s7_make_ratio(sc, -numerator(p), denominator(p)));

    case T_REAL:
      return(make_real(sc, -real(p)));

    case T_COMPLEX:
      return(s7_make_complex(sc, -real_part(p), -imag_part(p)));

    default:
      method_or_bust_with_type(sc, p, sc->subtract_symbol, args, a_number_string, 1);
    }
}

static s7_pointer g_subtract_2(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x, y;

  x = car(args);
  y = cadr(args);

  if (type(x) == type(y))
    {
      if (is_t_real(x))
	return(make_real(sc, real(x) - real(y)));
      else
	{
	  switch (type(x))
	    {
#if HAVE_OVERFLOW_CHECKS
	    case T_INTEGER:
	      {
		s7_int val;
		if (subtract_overflow(integer(x), integer(y), &val))
		  return(make_real(sc, (double)integer(x) - (double)integer(y)));
		return(make_integer(sc, val));
	      }
#else
	    case T_INTEGER: return(make_integer(sc, integer(x) - integer(y)));
#endif
	    case T_RATIO:   return(g_subtract(sc, args));
	    case T_REAL:    return(make_real(sc, real(x) - real(y)));
	    case T_COMPLEX: return(make_complex(sc, real_part(x) - real_part(y), imag_part(x) - imag_part(y)));
	    default:
	      if (!is_number(x))
		method_or_bust_with_type(sc, x, sc->subtract_symbol, args, a_number_string, 1);
	      method_or_bust_with_type(sc, y, sc->subtract_symbol, args, a_number_string, 2);
	    }
	}
    }

  switch (type(x))
    {
    case T_INTEGER:
      switch (type(y))
	{
	case T_INTEGER: return(make_integer(sc, integer(x) - integer(y)));
	case T_RATIO:   return(g_subtract(sc, args));
	case T_REAL:    return(make_real(sc, integer(x) - real(y)));
	case T_COMPLEX: return(make_complex(sc, integer(x) - real_part(y), -imag_part(y)));
	default:
	  method_or_bust_with_type(sc, y, sc->subtract_symbol, args, a_number_string, 2);
	}

    case T_RATIO:
      switch (type(y))
	{
	case T_INTEGER:
	case T_RATIO:   return(g_subtract(sc, args));
	case T_REAL:    return(make_real(sc, fraction(x) - real(y)));
	case T_COMPLEX: return(s7_make_complex(sc, fraction(x) - real_part(y), -imag_part(y)));
	default:
	  method_or_bust_with_type(sc, y, sc->subtract_symbol, args, a_number_string, 2);
	}

    case T_REAL:
      switch (type(y))
	{
	case T_INTEGER: return(make_real(sc, real(x) - integer(y)));
	case T_RATIO:   return(make_real(sc, real(x) - fraction(y)));
	case T_REAL:    return(make_real(sc, real(x) - real(y)));
	case T_COMPLEX: return(make_complex(sc, real(x) - real_part(y), -imag_part(y)));
	default:
	  method_or_bust_with_type(sc, y, sc->subtract_symbol, args, a_number_string, 2);
	}

    case T_COMPLEX:
      switch (type(y))
	{
	case T_INTEGER: return(s7_make_complex(sc, real_part(x) - integer(y), imag_part(x)));
	case T_RATIO:   return(s7_make_complex(sc, real_part(x) - fraction(y), imag_part(x)));
	case T_REAL:    return(s7_make_complex(sc, real_part(x) - real(y), imag_part(x)));
	case T_COMPLEX: return(make_complex(sc, real_part(x) - real_part(y), imag_part(x) - imag_part(y)));
	default:
	  method_or_bust_with_type(sc, y, sc->subtract_symbol, args, a_number_string, 2);
	}

    default:
      method_or_bust_with_type(sc, x, sc->subtract_symbol, args, a_number_string, 1);
    }
  return(x);
}


static s7_pointer minus_c1(s7_scheme *sc, s7_pointer x)
{
  switch (type(x))
    {
#if HAVE_OVERFLOW_CHECKS
    case T_INTEGER:
      {
	s7_int val;
	if (subtract_overflow(integer(x), 1, &val))
	  return(make_real(sc, (double)integer(x) - 1.0));
	return(make_integer(sc, val));
      }
#else
    case T_INTEGER: return(make_integer(sc, integer(x) - 1));
#endif
    case T_RATIO:   return(subtract_ratios(sc, x, small_int(1)));
    case T_REAL:    return(make_real(sc, real(x) - 1.0));
    case T_COMPLEX: return(s7_make_complex(sc, real_part(x) - 1.0, imag_part(x)));
    default:
      method_or_bust_with_type(sc, x, sc->subtract_symbol, list_2(sc, x, small_int(1)), a_number_string, 1);
    }
  return(x);
}

static s7_pointer g_subtract_cs1(s7_scheme *sc, s7_pointer args) 
{
  s7_pointer x;
  x = find_symbol_unchecked(sc, car(args));
  if (is_integer(x))
    return(make_integer(sc, integer(x) - 1));
  return(minus_c1(sc, x));
}

static s7_pointer g_subtract_cl1(s7_scheme *sc, s7_pointer args) 
{
  s7_pointer x;
  x = local_symbol_value(car(args));
  if (is_integer(x))
    return(make_integer(sc, integer(x) - 1));
  return(minus_c1(sc, x));
}

static s7_pointer g_subtract_s1(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x;
  x = car(args);
  /* this one seems to hit reals as often as integers */
  switch (type(x))
    {
#if HAVE_OVERFLOW_CHECKS
    case T_INTEGER:
      {
	s7_int val;
	if (subtract_overflow(integer(x), 1, &val))
	  return(make_real(sc, (double)integer(x) - 1.0));
	return(make_integer(sc, val));
      }
#else
    case T_INTEGER: return(make_integer(sc, integer(x) - 1));
#endif
    case T_RATIO:   return(subtract_ratios(sc, x, small_int(1)));
    case T_REAL:    return(make_real(sc, real(x) - 1.0));
    case T_COMPLEX: return(s7_make_complex(sc, real_part(x) - 1.0, imag_part(x)));
    default:
      method_or_bust_with_type(sc, x, sc->subtract_symbol, args, a_number_string, 1);
    }
  return(x);
}

static s7_pointer g_subtract_csn(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x;
  s7_int n;

  x = find_symbol_unchecked(sc, car(args));
  n = s7_integer(cadr(args));
  if (is_integer(x))
    return(make_integer(sc, integer(x) - n));

  switch (type(x))
    {
#if HAVE_OVERFLOW_CHECKS
    case T_INTEGER:
      {
	s7_int val;
	if (subtract_overflow(integer(x), n, &val))
	  return(make_real(sc, (double)integer(x) - (double)n));
	return(make_integer(sc, val));
      }
#else
    case T_INTEGER: return(make_integer(sc, integer(x) - n));
#endif
    case T_RATIO:   return(subtract_ratios(sc, x, cadr(args)));
    case T_REAL:    return(make_real(sc, real(x) - n));
    case T_COMPLEX: return(s7_make_complex(sc, real_part(x) - n, imag_part(x)));
    default:
      method_or_bust_with_type(sc, x, sc->subtract_symbol, list_2(sc, x, cadr(args)), a_number_string, 1);
    }
  return(x);
}

static s7_pointer subtract_sf;
static s7_pointer g_subtract_sf(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x;
  s7_double n;

  x = find_symbol_unchecked(sc, car(args));
  n = real(cadr(args));
  switch (type(x))
    {
    case T_INTEGER: return(make_real(sc, integer(x) - n));
    case T_RATIO:   return(make_real(sc, fraction(x) - n));
    case T_REAL:    return(make_real(sc, real(x) - n));
    case T_COMPLEX: return(s7_make_complex(sc, real_part(x) - n, imag_part(x)));
    default:
      method_or_bust_with_type(sc, x, sc->subtract_symbol, list_2(sc, x, cadr(args)), a_number_string, 1);
    }
  return(x);
}

static s7_pointer subtract_2f;
static s7_pointer g_subtract_2f(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x;
  s7_double n;

  x = car(args);
  n = real(cadr(args));
  switch (type(x))
    {
    case T_INTEGER: return(make_real(sc, integer(x) - n));
    case T_RATIO:   return(make_real(sc, fraction(x) - n));
    case T_REAL:    return(make_real(sc, real(x) - n));
    case T_COMPLEX: return(s7_make_complex(sc, real_part(x) - n, imag_part(x)));
    default:
      method_or_bust_with_type(sc, x, sc->subtract_symbol, args, a_number_string, 1);
    }
  return(x);
}

static s7_pointer subtract_fs;
static s7_pointer g_subtract_fs(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x;
  s7_double n;

  x = find_symbol_unchecked(sc, cadr(args));
  n = real(car(args));
  switch (type(x))
    {
    case T_INTEGER: return(make_real(sc, n - integer(x)));
    case T_RATIO:   return(make_real(sc, n - fraction(x)));
    case T_REAL:    return(make_real(sc, n - real(x)));
    case T_COMPLEX: return(s7_make_complex(sc, n - real_part(x), -imag_part(x)));
    default:
      method_or_bust_with_type(sc, x, sc->subtract_symbol, list_2(sc, car(args), x), a_number_string, 2);
    }
  return(x);
}

static s7_int subtract_i_ii(s7_int i1, s7_int i2) {return(i1 - i2);}
static s7_int subtract_i_i(s7_int x) {return(-x);}
static s7_int subtract_i_iii(s7_int i1, s7_int i2, s7_int i3) {return(i1 - i2 - i3);}

static s7_double subtract_d_d(s7_double x) {return(-x);}
static s7_double subtract_d_dd(s7_double x1, s7_double x2) {return(x1 - x2);}
static s7_double subtract_d_ddd(s7_double x1, s7_double x2, s7_double x3) {return(x1 - x2 - x3);}
static s7_double subtract_d_dddd(s7_double x1, s7_double x2, s7_double x3, s7_double x4) {return(x1 - x2 - x3 - x4);}

static s7_double sub_d_id(s7_int x1, s7_double x2) {return((s7_double)x1 - x2);}
#if (!WITH_GMP)
static s7_pointer sub_p_dd(s7_double x1, s7_double x2) {return(make_real(cur_sc, x1 - x2));}
static s7_pointer sub_p_ii(s7_int x1, s7_int x2) {return(make_integer(cur_sc, x1 - x2));}
#endif
static s7_pointer subtract_p_pp(s7_pointer p1, s7_pointer p2) {return(g_subtract_2(cur_sc, set_plist_2(cur_sc, p1, p2)));}


/* ---------------------------------------- multiply ---------------------------------------- */

static s7_pointer g_multiply(s7_scheme *sc, s7_pointer args)
{
  #define H_multiply "(* ...) multiplies its arguments"
  #define Q_multiply pcl_n

  s7_pointer x, p;
  s7_int num_a, den_a;
  s7_double rl_a, im_a;

#if (!WITH_GMP)
  if (is_null(args))
    return(small_int(1));
#endif

  x = car(args);
  p = cdr(args);
  if (is_null(p))
    {
      if (!is_number(x))
	method_or_bust_with_type_one_arg(sc, x, sc->multiply_symbol, args, a_number_string);
      return(x);
    }

  switch (type(x))
    {
    case T_INTEGER:
      num_a = integer(x);

    MULTIPLY_INTEGERS:
#if WITH_GMP
      if ((num_a > s7_int32_max) ||
	  (num_a < s7_int32_min))
	return(big_multiply(sc, cons(sc, s7_int_to_big_integer(sc, num_a), p)));
#endif
      x = car(p);
      p = cdr(p);
      switch (type(x))
	{
	case T_INTEGER:
#if WITH_GMP
	  if ((integer(x) > s7_int32_max) ||
	      (integer(x) < s7_int32_min))
	    return(big_multiply(sc, cons(sc, s7_int_to_big_integer(sc, num_a), cons(sc, x, p))));
#endif

#if HAVE_OVERFLOW_CHECKS
	  {
	    s7_int dn;
	    if (multiply_overflow(num_a, integer(x), &dn))
	      {
		if (is_null(p)) return(make_real(sc, (s7_double)num_a * (s7_double)integer(x)));
		rl_a = (s7_double)num_a * (s7_double)integer(x);
		goto MULTIPLY_REALS;
	      }
	    num_a = dn;
	  }
#else
	  num_a *= integer(x);
#endif
	  if (is_null(p)) return(make_integer(sc, num_a));
	  goto MULTIPLY_INTEGERS;

	case T_RATIO:
#if HAVE_OVERFLOW_CHECKS
	  {
	    s7_int dn;
	    if (multiply_overflow(numerator(x), num_a, &dn))
	      {
		if (is_null(p))
		  return(make_real(sc, (s7_double)num_a * fraction(x)));
		rl_a = (s7_double)num_a * fraction(x);
		goto MULTIPLY_REALS;
	      }
	    num_a = dn;
	  }
#else
	  num_a *= numerator(x);
#endif
	  den_a = denominator(x);
	  if (is_null(p)) return(s7_make_ratio(sc, num_a, den_a));
	  if (reduce_fraction(sc, &num_a, &den_a) == T_INTEGER)
	    goto MULTIPLY_INTEGERS;
	  goto MULTIPLY_RATIOS;

	case T_REAL:
	  if (is_null(p)) return(make_real(sc, num_a * real(x)));
	  rl_a = num_a * real(x);
	  goto MULTIPLY_REALS;

	case T_COMPLEX:
	  if (is_null(p)) return(s7_make_complex(sc, num_a * real_part(x), num_a * imag_part(x)));
	  rl_a = num_a * real_part(x);
	  im_a = num_a * imag_part(x);
	  goto MULTIPLY_COMPLEX;

	default:
	  method_or_bust_with_type(sc, x, sc->multiply_symbol, cons_unchecked(sc, s7_make_integer(sc, num_a), cons(sc, x, p)), a_number_string, position_of(p, args) - 1);
	}
      break;

    case T_RATIO:
      num_a = numerator(x);
      den_a = denominator(x);
    MULTIPLY_RATIOS:
#if WITH_GMP
      if ((num_a > s7_int32_max) ||
	  (den_a > s7_int32_max) ||
	  (num_a < s7_int32_min))
	return(big_multiply(sc, cons(sc, s7_ratio_to_big_ratio(sc, num_a, den_a), p)));
#endif
      x = car(p);
      p = cdr(p);

      switch (type(x))
	{
	case T_INTEGER:
	  /* as in +, this can overflow:
	   *   (* 8 -9223372036854775807 8) -> 64
	   *   (* 3/4 -9223372036854775807 8) -> 6
	   *   (* 8 -9223372036854775808 8) -> 0
	   *   (* -1 9223372036854775806 8) -> 16
	   *   (* -9223372036854775808 8 1e+308) -> 0.0
	   */
#if HAVE_OVERFLOW_CHECKS
	  {
	    s7_int dn;
	    if (multiply_overflow(integer(x), num_a, &dn))
	      {
		if (is_null(p))
		  return(make_real(sc, ((s7_double)integer(x) / (s7_double)den_a) * (s7_double)num_a));
		rl_a = ((s7_double)integer(x) / (s7_double)den_a) * (s7_double)num_a;
		goto MULTIPLY_REALS;
	      }
	    num_a = dn;
	  }
#else
	  num_a *= integer(x);
#endif
	  if (is_null(p)) return(s7_make_ratio(sc, num_a, den_a));
	  if (reduce_fraction(sc, &num_a, &den_a) == T_INTEGER)
	    goto MULTIPLY_INTEGERS;
	  goto MULTIPLY_RATIOS;

	case T_RATIO:
	  {
#if (!WITH_GMP)
	    s7_int d1, n1;
#endif
	    s7_int d2, n2;
	    d2 = denominator(x);
	    n2 = numerator(x);
#if (!WITH_GMP)
	    d1 = den_a;
	    n1 = num_a;
#if HAVE_OVERFLOW_CHECKS
	    if ((multiply_overflow(n1, n2, &num_a)) ||
		(multiply_overflow(d1, d2, &den_a)))
	      {
		if (is_null(p))
		  return(make_real(sc, ((long double)n1 / (long double)d1) * ((long double)n2 / (long double)d2)));
		rl_a = ((long double)n1 / (long double)d1) * ((long double)n2 / (long double)d2);
		goto MULTIPLY_REALS;
	      }
#else
	    num_a *= n2;
	    den_a *= d2;
#endif
#else
	    num_a *= n2;
	    den_a *= d2;
#endif
	    if (is_null(p)) return(s7_make_ratio(sc, num_a, den_a));
	    if (reduce_fraction(sc, &num_a, &den_a) == T_INTEGER)
	      goto MULTIPLY_INTEGERS;
	    goto MULTIPLY_RATIOS;
	  }

	case T_REAL:
	  if (is_null(p)) return(make_real(sc, ((long double)num_a / (long double)den_a) * real(x)));
	  rl_a = ((long double)num_a / (long double)den_a) * real(x);
	  goto MULTIPLY_REALS;

	case T_COMPLEX:
	  {
	    s7_double frac;
	    frac = ((long double)num_a / (long double)den_a);
	    if (is_null(p)) return(s7_make_complex(sc, frac * real_part(x), frac * imag_part(x)));
	    rl_a = frac * real_part(x);
	    im_a = frac * imag_part(x);
	    goto MULTIPLY_COMPLEX;
	  }

	default:
	  method_or_bust_with_type(sc, x, sc->multiply_symbol, cons_unchecked(sc, s7_make_ratio(sc, num_a, den_a), cons(sc, x, p)), a_number_string, position_of(p, args) - 1);
	}
      break;

    case T_REAL:
      rl_a = real(x);

    MULTIPLY_REALS:
      x = car(p);
      p = cdr(p);

      switch (type(x))
	{
	case T_INTEGER:
	  if (is_null(p)) return(make_real(sc, rl_a * integer(x)));
	  rl_a *= integer(x);
	  goto MULTIPLY_REALS;

	case T_RATIO:
	  if (is_null(p)) return(make_real(sc, rl_a * fraction(x)));
	  rl_a *= (s7_double)fraction(x);
	  goto MULTIPLY_REALS;

	case T_REAL:
	  if (is_null(p)) return(make_real(sc, rl_a * real(x)));
	  rl_a *= real(x);
	  goto MULTIPLY_REALS;

	case T_COMPLEX:
	  if (is_null(p)) return(s7_make_complex(sc, rl_a * real_part(x), rl_a * imag_part(x)));
	  im_a = rl_a * imag_part(x);
	  rl_a *= real_part(x);
	  goto MULTIPLY_COMPLEX;

	default:
	  method_or_bust_with_type(sc, x, sc->multiply_symbol, cons_unchecked(sc, make_real(sc, rl_a), cons(sc, x, p)), a_number_string, position_of(p, args) - 1);
	}
      break;

    case T_COMPLEX:
      rl_a = real_part(x);
      im_a = imag_part(x);

    MULTIPLY_COMPLEX:
      x = car(p);
      p = cdr(p);

      switch (type(x))
	{
	case T_INTEGER:
	  if (is_null(p)) return(s7_make_complex(sc, rl_a * integer(x), im_a * integer(x)));
	  rl_a *= integer(x);
	  im_a *= integer(x);
	  goto MULTIPLY_COMPLEX;

	case T_RATIO:
	  {
	    s7_double frac;
	    frac = fraction(x);
	    if (is_null(p)) return(s7_make_complex(sc, rl_a * frac, im_a * frac));
	    rl_a *= frac;
	    im_a *= frac;
	    goto MULTIPLY_COMPLEX;
	  }

	case T_REAL:
	  if (is_null(p)) return(s7_make_complex(sc, rl_a * real(x), im_a * real(x)));
	  rl_a *= real(x);
	  im_a *= real(x);
	  goto MULTIPLY_COMPLEX;

	case T_COMPLEX:
	  {
	    s7_double r1, r2, i1, i2;
	    r1 = rl_a;
	    i1 = im_a;
	    r2 = real_part(x);
	    i2 = imag_part(x);
	    if (is_null(p))
	      return(s7_make_complex(sc, r1 * r2 - i1 * i2, r1 * i2 + r2 * i1));
	    rl_a = r1 * r2 - i1 * i2;
	    im_a = r1 * i2 + r2 * i1;
	    if (im_a == 0.0)
	      goto MULTIPLY_REALS;
	    goto MULTIPLY_COMPLEX;
	  }

	default:
	  method_or_bust_with_type(sc, x, sc->multiply_symbol, cons_unchecked(sc, s7_make_complex(sc, rl_a, im_a), cons(sc, x, p)), a_number_string, position_of(p, args) - 1);
	}
      break;

    default:
      method_or_bust_with_type(sc, x, sc->multiply_symbol, args, a_number_string, 1);
    }
}

#if (!WITH_GMP)
static s7_pointer multiply_2, multiply_fs, multiply_sf, multiply_is, multiply_si;

static s7_pointer g_multiply_2(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x, y;
  x = car(args);
  y = cadr(args);

  if (type(x) == type(y))
    {
      if (is_t_real(x))
	return(make_real(sc, real(x) * real(y)));
      else
	{
	  switch (type(x))
	    {
#if HAVE_OVERFLOW_CHECKS
	    case T_INTEGER:
	      {
		s7_int n;
		if (multiply_overflow(integer(x), integer(y), &n))
		  return(make_real(sc, ((s7_double)integer(x)) * ((s7_double)integer(y))));
		return(make_integer(sc, n));
	      }
#else
	    case T_INTEGER: return(make_integer(sc, integer(x) * integer(y)));
#endif
	    case T_RATIO:   return(g_multiply(sc, args));
	    case T_REAL:    return(make_real(sc, real(x) * real(y)));
	    case T_COMPLEX:
	      {
		s7_double r1, r2, i1, i2;
		r1 = real_part(x);
		r2 = real_part(y);
		i1 = imag_part(x);
		i2 = imag_part(y);
		return(s7_make_complex(sc, r1 * r2 - i1 * i2, r1 * i2 + r2 * i1));
	      }
	    default:
	      if (!is_number(x))
		method_or_bust_with_type(sc, x, sc->multiply_symbol, args, a_number_string, 1);
	      method_or_bust_with_type(sc, y, sc->multiply_symbol, args, a_number_string, 2);
	    }
	}
    }

  switch (type(x))
    {
    case T_INTEGER:
      switch (type(y))
	{
	case T_INTEGER: return(make_integer(sc, integer(x) * integer(y)));
	case T_RATIO:   return(g_multiply(sc, args));
	case T_REAL:    return(make_real(sc, integer(x) * real(y)));
	case T_COMPLEX: return(s7_make_complex(sc, integer(x) * real_part(y), integer(x) * imag_part(y)));
	default:
	  method_or_bust_with_type(sc, y, sc->multiply_symbol, args, a_number_string, 2);
	}

    case T_RATIO:
      switch (type(y))
	{
	case T_INTEGER:
	case T_RATIO:    return(g_multiply(sc, args));
	case T_REAL:     return(make_real(sc, fraction(x) * real(y)));
	case T_COMPLEX:
	  {
	    s7_double frac;
	    frac = fraction(x);
	    return(s7_make_complex(sc, frac * real_part(y), frac * imag_part(y)));
	  }
	default:
	  method_or_bust_with_type(sc, y, sc->multiply_symbol, args, a_number_string, 2);
	}

    case T_REAL:
      switch (type(y))
	{
	case T_INTEGER: return(make_real(sc, real(x) * integer(y)));
	case T_RATIO:   return(make_real(sc, real(x) * fraction(y)));
	case T_REAL:    return(make_real(sc, real(x) * real(y)));
	case T_COMPLEX: return(s7_make_complex(sc, real(x) * real_part(y), real(x) * imag_part(y)));
	default:
	  method_or_bust_with_type(sc, y, sc->multiply_symbol, args, a_number_string, 2);
	}

    case T_COMPLEX:
      switch (type(y))
	{
	case T_INTEGER: return(s7_make_complex(sc, real_part(x) * integer(y), imag_part(x) * integer(y)));
	case T_RATIO:
	  {
	    s7_double frac;
	    frac = fraction(y);
	    return(s7_make_complex(sc, real_part(x) * frac, imag_part(x) * frac));
	  }
	case T_REAL:    return(s7_make_complex(sc, real_part(x) * real(y), imag_part(x) * real(y)));
	case T_COMPLEX:
	  {
	    s7_double r1, r2, i1, i2;
	    r1 = real_part(x);
	    r2 = real_part(y);
	    i1 = imag_part(x);
	    i2 = imag_part(y);
	    return(s7_make_complex(sc, r1 * r2 - i1 * i2, r1 * i2 + r2 * i1));
	  }
	default:
	  method_or_bust_with_type(sc, y, sc->multiply_symbol, args, a_number_string, 2);
	}

    default:
      method_or_bust_with_type(sc, x, sc->multiply_symbol, args, a_number_string, 1);
    }
  return(x);
}

/* all of these mess up if overflows occur
 *  (let () (define (f x) (* x 9223372036854775806)) (f -63)) -> -9223372036854775682, but (* -63 9223372036854775806) -> -5.810724383218509e+20
 *  how to catch this?  (affects * - +)
 */

static s7_pointer g_multiply_si(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x;
  s7_int n;

  x = find_symbol_unchecked(sc, car(args));
  n = integer(cadr(args));

  switch (type(x))
    {
#if HAVE_OVERFLOW_CHECKS
    case T_INTEGER: 
      {
	s7_int val;
	if (multiply_overflow(integer(x), n, &val))
	  return(make_real(sc, (double)integer(x) * (double)n));
	return(make_integer(sc, val));
      }
    case T_RATIO:
      {
	s7_int val;
	if (multiply_overflow(numerator(x), n, &val))
	  return(make_real(sc, fraction(x) * (double)n));
	return(s7_make_ratio(sc, val, denominator(x)));
      }
#else
    case T_INTEGER: return(make_integer(sc, integer(x) * n));
    case T_RATIO:   return(s7_make_ratio(sc, numerator(x) * n, denominator(x)));
#endif
    case T_REAL:    return(make_real(sc, real(x) * n));
    case T_COMPLEX: return(s7_make_complex(sc, real_part(x) * n, imag_part(x) * n));
    default:
      method_or_bust_with_type(sc, x, sc->multiply_symbol, list_2(sc, x, cadr(args)), a_number_string, 1);
    }
  return(x);
}

static s7_pointer g_multiply_is(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x;
  s7_int n;

  x = find_symbol_unchecked(sc, cadr(args));
  n = integer(car(args));

  switch (type(x))
    {
#if HAVE_OVERFLOW_CHECKS
    case T_INTEGER: 
      {
	s7_int val;
	if (multiply_overflow(integer(x), n, &val))
	  return(make_real(sc, (double)integer(x) * (double)n));
	return(make_integer(sc, val));
      }
    case T_RATIO:
      {
	s7_int val;
	if (multiply_overflow(numerator(x), n, &val))
	  return(make_real(sc, fraction(x) * (double)n));
	return(s7_make_ratio(sc, val, denominator(x)));
      }
#else
    case T_INTEGER: return(make_integer(sc, integer(x) * n));
    case T_RATIO:   return(s7_make_ratio(sc, numerator(x) * n, denominator(x)));
#endif
    case T_REAL:    return(make_real(sc, real(x) * n));
    case T_COMPLEX: return(s7_make_complex(sc, real_part(x) * n, imag_part(x) * n));
    default:
      method_or_bust_with_type(sc, x, sc->multiply_symbol, list_2(sc, car(args), x), a_number_string, 2);
    }
  return(x);
}

static s7_pointer g_multiply_fs(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x;
  s7_double scl;

  scl = real(car(args));
  x = find_symbol_unchecked(sc, cadr(args));

  switch (type(x))
    {
    case T_INTEGER: return(make_real(sc, integer(x) * scl));
    case T_RATIO:   return(make_real(sc, numerator(x) * scl / denominator(x)));
    case T_REAL:    return(make_real(sc, real(x) * scl));
    case T_COMPLEX: return(s7_make_complex(sc, real_part(x) * scl, imag_part(x) * scl));
    default:
      method_or_bust_with_type(sc, x, sc->multiply_symbol, list_2(sc, car(args), x), a_number_string, 1);
    }
  return(x);
}

static s7_pointer g_multiply_sf(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x;
  s7_double scl;

  scl = real(cadr(args));
  x = find_symbol_unchecked(sc, car(args));

  switch (type(x))
    {
    case T_INTEGER: return(make_real(sc, integer(x) * scl));
    case T_RATIO:   return(make_real(sc, numerator(x) * scl / denominator(x)));
    case T_REAL:    return(make_real(sc, real(x) * scl));
    case T_COMPLEX: return(s7_make_complex(sc, real_part(x) * scl, imag_part(x) * scl));
    default:
      method_or_bust_with_type(sc, x, sc->multiply_symbol, list_2(sc, x, cadr(args)), a_number_string, 2);
    }
  return(x);
}

static s7_pointer sqr_ss;
static s7_pointer g_sqr_ss(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x;
  x = find_symbol_unchecked(sc, car(args));

  switch (type(x))
    {
#if HAVE_OVERFLOW_CHECKS
    case T_INTEGER: 
      {
	s7_int val;
	if (multiply_overflow(integer(x), integer(x), &val))
	  return(make_real(sc, (double)integer(x) * (double)integer(x)));
	return(make_integer(sc, val));
      }
    case T_RATIO:
      {
	s7_int num, den;
	if ((multiply_overflow(numerator(x), numerator(x), &num)) ||
	    (multiply_overflow(denominator(x), denominator(x), &den)))
	  return(make_real(sc, fraction(x) * fraction(x)));
	return(s7_make_ratio(sc, num, den));
      }
#else
    case T_INTEGER: return(s7_make_integer(sc, integer(x) * integer(x)));
    case T_RATIO:   return(s7_make_ratio(sc, numerator(x) * numerator(x), denominator(x) * denominator(x)));
#endif
    case T_REAL:    return(make_real(sc, real(x) * real(x)));
    case T_COMPLEX: return(s7_make_complex(sc, real_part(x) * real_part(x) - imag_part(x) * imag_part(x), 2.0 * real_part(x) * imag_part(x)));
    default:
      method_or_bust_with_type(sc, x, sc->multiply_symbol, list_2(sc, x, x), a_number_string, 1);
    }
  return(x);
}
#endif /* with-gmp */

static s7_int multiply_i_ii(s7_int i1, s7_int i2) {return(i1 * i2);}
static s7_int multiply_i_iii(s7_int i1, s7_int i2, s7_int i3) {return(i1 * i2 * i3);}

static s7_double multiply_d_d(s7_double x) {return(x);}
static s7_double multiply_d_dd(s7_double x1, s7_double x2) {return(x1 * x2);}
static s7_double multiply_d_ddd(s7_double x1, s7_double x2, s7_double x3) {return(x1 * x2 * x3);}
static s7_double multiply_d_dddd(s7_double x1, s7_double x2, s7_double x3, s7_double x4) {return(x1 * x2 * x3 * x4);}
static s7_double multiply_d_id(s7_int x1, s7_double x2) {return((s7_double)x1 * x2);}
#if (!WITH_GMP)
static s7_pointer mul_p_dd(s7_double x1, s7_double x2) {return(make_real(cur_sc, x1 * x2));}
static s7_pointer mul_p_ii(s7_int x1, s7_int x2) {return(make_integer(cur_sc, x1 * x2));}

static s7_pointer multiply_p_pp(s7_pointer x1, s7_pointer x2) {return(g_multiply_2(cur_sc, set_plist_2(cur_sc, x1, x2)));}
#endif


/* ---------------------------------------- divide ---------------------------------------- */

static bool is_number_via_method(s7_scheme *sc, s7_pointer p)
{
  if (s7_is_number(p))
    return(true);
  if (has_methods(p))
    {
      s7_pointer f;
      f = find_method(sc, find_let(sc, p), sc->is_number_symbol);
      if (f != sc->undefined)
	return(is_true(sc, s7_apply_function(sc, f, cons(sc, p, sc->nil))));
    }
  return(false);
}

static s7_pointer g_divide(s7_scheme *sc, s7_pointer args)
{
  #define H_divide "(/ x1 ...) divides its first argument by the rest, or inverts the first if there is only one argument"
  #define Q_divide pcl_n

  s7_pointer x, p;
  s7_int num_a, den_a;
  s7_double rl_a, im_a;

  x = car(args);
  p = cdr(args);
  if (is_null(p))
    {
      if (!is_number(x))
	method_or_bust_with_type_one_arg(sc, x, sc->divide_symbol, args, a_number_string);
      if (s7_is_zero(x))
	return(division_by_zero_error(sc, sc->divide_symbol, args));
      return(s7_invert(sc, x));
    }

  switch (type(x))
    {
    case T_INTEGER:
      num_a = integer(x);
      if (num_a == 0)
	{
	  bool return_nan = false, return_real_zero = false;
	  for (; is_pair(p); p = cdr(p))
	    {
	      s7_pointer n;
	      n = car(p);
	      if (!s7_is_number(n))
		{
		  n = check_values(sc, n, p);
		  if (!s7_is_number(n))
		    return(wrong_type_argument_with_type(sc, sc->divide_symbol, position_of(p, args), n, a_number_string));
		}
	      if (s7_is_zero(n))
		return(division_by_zero_error(sc, sc->divide_symbol, args));
	      if (type(n) > T_RATIO)
		{
		  return_real_zero = true;
		  if (is_NaN(s7_real_part(n)))
		    return_nan = true;
		}
	    }
	  if (return_nan)
	    return(real_NaN);
	  if (return_real_zero)
	    return(real_zero);
	  return(small_int(0));
	}

    DIVIDE_INTEGERS:
#if WITH_GMP
      if ((num_a > s7_int32_max) ||
	  (num_a < s7_int32_min))
	return(big_divide(sc, cons(sc, s7_int_to_big_integer(sc, num_a), p)));
#endif
      x = car(p);
      p = cdr(p);

      switch (type(x))
	{
	case T_INTEGER:
	  if (integer(x) == 0)
	    return(division_by_zero_error(sc, sc->divide_symbol, args));

	  /* to be consistent, I suppose we should search first for NaNs in the divisor list.
	   *   (* 0 0/0) is NaN, so (/ 1 0 0/0) should equal (/ 1 0/0) = NaN.  But the whole
	   *   thing is ridiculous.
	   */
	  if (is_null(p))
	    return(s7_make_ratio(sc, num_a, integer(x)));

	  den_a = integer(x);
	  if (reduce_fraction(sc, &num_a, &den_a) == T_INTEGER)
	    goto DIVIDE_INTEGERS;
	  goto DIVIDE_RATIOS;

	case T_RATIO:
	  den_a = denominator(x);
#if HAVE_OVERFLOW_CHECKS
	  {
	    s7_int dn;
	    if (multiply_overflow(num_a, den_a, &dn))
	      {
		if (is_null(p)) return(make_real(sc, num_a * inverted_fraction(x)));
		rl_a = (s7_double)num_a * inverted_fraction(x);
		goto DIVIDE_REALS;
	      }
	    num_a = dn;
	  }
#else
	  num_a *= den_a;
#endif
	  den_a = numerator(x);
	  if (is_null(p)) return(s7_make_ratio(sc, num_a, den_a));
	  if (reduce_fraction(sc, &num_a, &den_a) == T_INTEGER)
	    goto DIVIDE_INTEGERS;
	  goto DIVIDE_RATIOS;

	case T_REAL:
	  rl_a = (s7_double)num_a;
	  if (real(x) == 0.0)
	    return(division_by_zero_error(sc, sc->divide_symbol, args));
	  if (is_null(p)) return(make_real(sc, rl_a / real(x)));
	  rl_a /= real(x);
	  goto DIVIDE_REALS;

	case T_COMPLEX:
	  {
	    s7_double i2, r2, den;
	    rl_a = (s7_double)num_a;
	    r2 = real_part(x);
	    i2 = imag_part(x);
	    den = 1.0 / (r2 * r2 + i2 * i2);
	    /* we could avoid the squaring (see Knuth II p613 16)
	     *    not a big deal: (/ 1.0e308+1.0e308i 2.0e308+2.0e308i) => nan
	     *    (gmp case is ok here)
	     */
	    if (is_null(p))
	      return(s7_make_complex(sc, rl_a * r2 * den, -(rl_a * i2 * den)));
	    im_a = -rl_a * i2 * den;
	    rl_a *= r2 * den;
	    goto DIVIDE_COMPLEX;
	  }

	default:
	  method_or_bust_with_type(sc, x, sc->divide_symbol, cons_unchecked(sc, s7_make_integer(sc, num_a), cons(sc, x, p)), a_number_string, position_of(p, args) - 1);
	}
      break;

    case T_RATIO:
      num_a = numerator(x);
      den_a = denominator(x);
    DIVIDE_RATIOS:
#if WITH_GMP
      if ((num_a > s7_int32_max) ||
	  (den_a > s7_int32_max) ||
	  (num_a < s7_int32_min))
	return(big_divide(sc, cons(sc, s7_ratio_to_big_ratio(sc, num_a, den_a), p)));
#endif
      x = car(p);
      p = cdr(p);

      switch (type(x))
	{
	case T_INTEGER:
	  if (integer(x) == 0)
	    return(division_by_zero_error(sc, sc->divide_symbol, args));
#if HAVE_OVERFLOW_CHECKS
	  {
	    s7_int dn;
	    if (multiply_overflow(den_a, integer(x), &dn))
	      {
		if (is_null(p)) return(make_real(sc, (long double)num_a / ((long double)den_a * (s7_double)integer(x))));
		rl_a = (long double)num_a / ((long double)den_a * (s7_double)integer(x));
		goto DIVIDE_REALS;
	      }
	    den_a = dn;
	  }
#else
	  den_a *= integer(x);
#endif
	  if (is_null(p)) return(s7_make_ratio(sc, num_a, den_a));
	  if (reduce_fraction(sc, &num_a, &den_a) == T_INTEGER)
	    goto DIVIDE_INTEGERS;
	  goto DIVIDE_RATIOS;

	case T_RATIO:
	  {
	    s7_int d1, d2, n1, n2;
	    d1 = den_a;
	    n1 = num_a;
	    d2 = denominator(x);
	    n2 = numerator(x);
	    if (d1 == d2)
	      {
		if (is_null(p))
		  return(s7_make_ratio(sc, n1, n2));
		den_a = n2;
	      }
	    else
	      {
#if (!WITH_GMP) && HAVE_OVERFLOW_CHECKS
		if ((multiply_overflow(n1, d2, &n1)) ||
		    (multiply_overflow(n2, d1, &d1)))
		  {
		    s7_double r1, r2;
		    r1 = ((long double)num_a / (long double)den_a);
		    r2 = inverted_fraction(x);
		    if (is_null(p)) return(make_real(sc, r1 * r2));
		    rl_a = r1 * r2;
		    goto DIVIDE_REALS;
		  }
		num_a = n1;
		den_a = d1;
#else
		num_a *= d2;
		den_a *= n2;
#endif
		if (is_null(p))
		  return(s7_make_ratio(sc, num_a, den_a));
	      }
	    if (reduce_fraction(sc, &num_a, &den_a) == T_INTEGER)
	      goto DIVIDE_INTEGERS;
	    goto DIVIDE_RATIOS;
	  }

	case T_REAL:
	  {
	    s7_double r1;
	    if (real(x) == 0.0)
	      return(division_by_zero_error(sc, sc->divide_symbol, args));
	    r1 = ((long double)num_a / (long double)den_a);
	    if (is_null(p)) return(make_real(sc, r1 / real(x)));
	    rl_a = r1 / real(x);
	    goto DIVIDE_REALS;
	  }

	case T_COMPLEX:
	  {
	    s7_double den, i2, r2;
	    rl_a = ((long double)num_a / (long double)den_a);
	    r2 = real_part(x);
	    i2 = imag_part(x);
	    den = 1.0 / (r2 * r2 + i2 * i2);
	    if (is_null(p))
	      return(s7_make_complex(sc, rl_a * r2 * den, -rl_a * i2 * den));
	    im_a = -rl_a * i2 * den;
	    rl_a *= r2 * den;
	    goto DIVIDE_COMPLEX;
	  }

	default:
	  method_or_bust_with_type(sc, x, sc->divide_symbol, cons_unchecked(sc, s7_make_ratio(sc, num_a, den_a), cons(sc, x, p)), a_number_string, position_of(p, args) - 1);
	}
      break;

    case T_REAL:
      rl_a = real(x);
      if (rl_a == 0)
	{
	  bool return_nan = false;
	  for (; is_pair(p); p = cdr(p))
	    {
	      s7_pointer n;
	      n = car(p);
	      if (!s7_is_number(n))
		{
		  n = check_values(sc, n, p);
		  if (!s7_is_number(n))
		    return(wrong_type_argument_with_type(sc, sc->divide_symbol, position_of(p, args), n, a_number_string));
		}
	      if (s7_is_zero(n))
		return(division_by_zero_error(sc, sc->divide_symbol, args));
	      if ((is_t_real(n)) &&
		  (is_NaN(real(n))))
		return_nan = true;
	    }
	  if (return_nan)
	    return(real_NaN);
	  return(real_zero);
	}

    DIVIDE_REALS:
      x = car(p);
      p = cdr(p);

      switch (type(x))
	{
	case T_INTEGER:
	  if (integer(x) == 0)
	    return(division_by_zero_error(sc, sc->divide_symbol, args));
	  if (is_null(p)) return(make_real(sc, rl_a / integer(x)));
	  rl_a /= (s7_double)integer(x);
	  goto DIVIDE_REALS;

	case T_RATIO:
	  if (is_null(p)) return(make_real(sc, rl_a * inverted_fraction(x)));
	  rl_a *= (s7_double)inverted_fraction(x);
	  goto DIVIDE_REALS;

	case T_REAL:
	  if (real(x) == 0.0)
	    return(division_by_zero_error(sc, sc->divide_symbol, args));
	  if (is_null(p)) return(make_real(sc, rl_a / real(x)));
	  rl_a /= real(x);
	  goto DIVIDE_REALS;

	case T_COMPLEX:
	  {
	    s7_double den, r2, i2;
	    r2 = real_part(x);
	    i2 = imag_part(x);
	    den = 1.0 / (r2 * r2 + i2 * i2);
	    if (is_null(p))
	      return(s7_make_complex(sc, rl_a * r2 * den, -rl_a * i2 * den));
	    im_a = -rl_a * i2 * den;
	    rl_a *= r2 * den;
	    goto DIVIDE_COMPLEX;
	  }

	default:
	  method_or_bust_with_type(sc, x, sc->divide_symbol, cons_unchecked(sc, make_real(sc, rl_a), cons(sc, x, p)), a_number_string, position_of(p, args) - 1);
	}
      break;

    case T_COMPLEX:
      rl_a = real_part(x);
      im_a = imag_part(x);

    DIVIDE_COMPLEX:
      x = car(p);
      p = cdr(p);

      switch (type(x))
	{
	case T_INTEGER:
	  {
	    s7_double r1;
	    if (integer(x) == 0)
	      return(division_by_zero_error(sc, sc->divide_symbol, args));
	    r1 = 1.0 / (s7_double)integer(x);
	    if (is_null(p)) return(s7_make_complex(sc, rl_a * r1, im_a * r1));
	    rl_a *= r1;
	    im_a *= r1;
	    goto DIVIDE_COMPLEX;
	  }

	case T_RATIO:
	  {
	    s7_double frac;
	    frac = inverted_fraction(x);
	    if (is_null(p)) return(s7_make_complex(sc, rl_a * frac, im_a * frac));
	    rl_a *= frac;
	    im_a *= frac;
	    goto DIVIDE_COMPLEX;
	  }

	case T_REAL:
	  {
	    s7_double r1;
	    if (real(x) == 0.0)
	      return(division_by_zero_error(sc, sc->divide_symbol, args));
	    r1 = 1.0 / real(x);
	    if (is_null(p)) return(s7_make_complex(sc, rl_a * r1, im_a * r1));
	    rl_a *= r1;
	    im_a *= r1;
	    goto DIVIDE_COMPLEX;
	  }

	case T_COMPLEX:
	  {
	    s7_double r1, r2, i1, i2, den;
	    r1 = rl_a;
	    i1 = im_a;
	    r2 = real_part(x);
	    i2 = imag_part(x);
	    den = 1.0 / (r2 * r2 + i2 * i2);
	    if (is_null(p))
	      return(s7_make_complex(sc, (r1 * r2 + i1 * i2) * den, (r2 * i1 - r1 * i2) * den));
	    rl_a = (r1 * r2 + i1 * i2) * den;
	    im_a = (r2 * i1 - r1 * i2) * den;
	    goto DIVIDE_COMPLEX;
	  }

	default:
	  method_or_bust_with_type(sc, x, sc->divide_symbol, cons_unchecked(sc, s7_make_complex(sc, rl_a, im_a), cons(sc, x, p)), a_number_string, position_of(p, args) - 1);
	}
      break;

    default:
      method_or_bust_with_type(sc, x, sc->divide_symbol, args, a_number_string, 1);
    }
}


#if (!WITH_GMP)
static s7_pointer invert_1;

static s7_pointer g_invert_1(s7_scheme *sc, s7_pointer args)
{
  s7_pointer p;
  p = car(args);
  switch (type(p))
    {
    case T_INTEGER:
      if (integer(p) != 0)
	return(s7_make_ratio(sc, 1, integer(p)));      /* a already checked, not 0 */
      return(division_by_zero_error(sc, sc->divide_symbol, args));

    case T_RATIO:
      return(s7_make_ratio(sc, denominator(p), numerator(p)));

    case T_REAL:
      if (real(p) != 0.0)
	return(make_real(sc, 1.0 / real(p)));
      return(division_by_zero_error(sc, sc->divide_symbol, args));

    case T_COMPLEX:
      {
	s7_double r2, i2, den;
	r2 = real_part(p);
	i2 = imag_part(p);
	den = (r2 * r2 + i2 * i2);
	return(s7_make_complex(sc, r2 / den, -i2 / den));
      }

    default:
      method_or_bust_with_type(sc, p, sc->divide_symbol, args, a_number_string, 1);
    }
}


static s7_pointer divide_1r;
static s7_pointer g_divide_1r(s7_scheme *sc, s7_pointer args)
{
  if (s7_is_real(cadr(args)))
    {
      s7_double rl;
      rl = real_to_double(sc, cadr(args), "/");
      if (rl == 0.0)
	return(division_by_zero_error(sc, sc->divide_symbol, args));
      return(make_real(sc, 1.0 / rl));
    }
  return(g_divide(sc, args));
}
#endif


static s7_double divide_d_d(s7_double x) 
{
  if (x == 0.0) division_by_zero_error(cur_sc, cur_sc->divide_symbol, set_elist_1(cur_sc, real_zero));
  return(1.0 / x);
}

static s7_double divide_d_dd(s7_double x1, s7_double x2) 
{
  if (x2 == 0.0) division_by_zero_error(cur_sc, cur_sc->divide_symbol, set_elist_1(cur_sc, real_zero));
  return(x1 / x2);
}

static s7_double divide_d_ddd(s7_double x1, s7_double x2, s7_double x3) 
{
  s7_double d;
  d = x2 * x3;
  if (d == 0.0) division_by_zero_error(cur_sc, cur_sc->divide_symbol, set_elist_1(cur_sc, real_zero));
  return(x1 / d);
}

static s7_double divide_d_dddd(s7_double x1, s7_double x2, s7_double x3, s7_double x4) 
{
  s7_double d;
  d = x2 * x3 * x4;
  if (d == 0.0) division_by_zero_error(cur_sc, cur_sc->divide_symbol, set_elist_1(cur_sc, real_zero));  
  return(x1 / d);
}

static s7_pointer divide_p_ii(s7_int x, s7_int y) {return(s7_make_ratio(cur_sc, x, y));} /* make-ratio checks for y==0 */



/* ---------------------------------------- max/min ---------------------------------------- */

static bool is_real_via_method_1(s7_scheme *sc, s7_pointer p)
{
  s7_pointer f;
  f = find_method(sc, find_let(sc, p), sc->is_real_symbol);
  if (f != sc->undefined)
    return(is_true(sc, s7_apply_function(sc, f, cons(sc, p, sc->nil))));
  return(false);
}

#define is_real_via_method(sc, p) ((s7_is_real(p)) || ((has_methods(p)) && (is_real_via_method_1(sc, p))))


static s7_pointer g_max(s7_scheme *sc, s7_pointer args)
{
  #define H_max "(max ...) returns the maximum of its arguments"
  #define Q_max pcl_r

  s7_pointer x, y, p;
  s7_int num_a, num_b, den_a, den_b;

  x = car(args);
  p = cdr(args);

  switch (type(x))
    {
    case T_INTEGER:
    MAX_INTEGERS:
      if (is_null(p)) return(x);
      y = car(p);
      p = cdr(p);
      switch (type(y))
	{
	case T_INTEGER:
	  if (integer(x) < integer(y)) x = y;
	  goto MAX_INTEGERS;

	case T_RATIO:
	  num_a = integer(x);
	  den_a = 1;
	  num_b = numerator(y);
	  den_b = denominator(y);
	  goto RATIO_MAX_RATIO;

	case T_REAL:
	  if (is_NaN(real(y)))
	    {
	      for (; is_not_null(p); p = cdr(p))
		if (!is_real_via_method(sc, car(p)))
		  return(wrong_type_argument(sc, sc->max_symbol, position_of(p, args), car(p), T_REAL));
	      return(y);
	    }
	  if (integer(x) < real(y))
	    {
	      x = y;
	      goto MAX_REALS;
	    }
	  goto MAX_INTEGERS;

	default:
	  method_or_bust(sc, y, sc->max_symbol, cons_unchecked(sc, x, cons(sc, y, p)), T_REAL, position_of(p, args) - 1);
	}


    case T_RATIO:
    MAX_RATIOS:
      if (is_null(p)) return(x);
      y = car(p);
      p = cdr(p);
      switch (type(y))
	{
	case T_INTEGER:
	  num_a = numerator(x);
	  den_a = denominator(x);
	  num_b = integer(y);
	  den_b = 1;
	  goto RATIO_MAX_RATIO;

	case T_RATIO:
	  num_a = numerator(x);
	  den_a = denominator(x);
	  num_b = numerator(y);
	  den_b = denominator(y);

	RATIO_MAX_RATIO:
	  /* there are tricky cases here where long ints outrun doubles:
	   *   (max 92233720368547758/9223372036854775807 92233720368547757/9223372036854775807)
	   * which should be 92233720368547758/9223372036854775807) but first the fraction gets reduced
	   * to 13176245766935394/1317624576693539401, so we fall into the double comparison, and
	   * there we should be comparing
	   *    9.999999999999999992410584792601468961145E-3 and
	   *    9.999999999999999883990367544051025548645E-3
	   * but if using doubles we get
	   *    0.010000000000000000208166817117 and
	   *    0.010000000000000000208166817117
	   * that is, we can't distinguish these two fractions once they're coerced to doubles.
	   *
	   * Even long doubles fail in innocuous-looking cases:
	   *     (min 21053343141/6701487259 3587785776203/1142027682075) -> 3587785776203/1142027682075
	   *     (max 21053343141/6701487259 3587785776203/1142027682075) -> 3587785776203/1142027682075
	   *
	   * Another consequence: outside gmp, we can't handle cases like
	   *    (max 9223372036854776/9223372036854775807 #i9223372036854775/9223372036854775000)
	   *    (max #i9223372036854776/9223372036854775807 9223372036854775/9223372036854775000)
	   * I guess if the user is using "inexact" numbers (#i...), he accepts their inexactness.
	   */
	  
	  if ((num_a < 0) && (num_b >= 0)) /* x < 0, y >= 0 -> y */
	    x = y;
	  else
	    {
	      if ((num_a < 0) || (num_b >= 0))
		{
		  if (den_a == den_b)
		    {
		      if (num_a < num_b)
			x = y;
		    }
		  else
		    {
		      if (num_a == num_b)
			{
			  if (((num_a >= 0) &&
			       (den_a > den_b)) ||
			      ((num_a < 0) &&
			       (den_a < den_b)))
			    x = y;
			}
		      else
			{
			  s7_int vala, valb;
			  vala = num_a / den_a;
			  valb = num_b / den_b;
			  if (!((vala > valb) ||
				((vala == valb) && (is_t_integer(y)))))
			    {
			      if ((valb > vala) ||
				  ((vala == valb) && (is_t_integer(x))) ||
				  /* sigh -- both are ratios and the int32_t parts are equal */
				  (((long double)(num_a % den_a) / (long double)den_a) <= ((long double)(num_b % den_b) / (long double)den_b)))
				x = y;
			    }
			}
		    }
		}
	    }
	  if (is_t_ratio(x))
	    goto MAX_RATIOS;
	  goto MAX_INTEGERS;

	case T_REAL:
	  /* (max 3/4 nan.0) should probably return NaN */
	  if (is_NaN(real(y)))
	    {
	      for (; is_not_null(p); p = cdr(p))
		if (!is_real_via_method(sc, car(p)))
		  return(wrong_type_argument(sc, sc->max_symbol, position_of(p, args), car(p), T_REAL));
	      return(y);
	    }

	  if (fraction(x) < real(y))
	    {
	      x = y;
	      goto MAX_REALS;
	    }
	  goto MAX_RATIOS;

	default:
	  method_or_bust(sc, y, sc->max_symbol, cons_unchecked(sc, x, cons(sc, y, p)), T_REAL, position_of(p, args) - 1);
	}


    case T_REAL:
      if (is_NaN(real(x)))
	{
	  for (; is_not_null(p); p = cdr(p))
	    if (!is_real_via_method(sc, car(p)))
	      return(wrong_type_argument(sc, sc->max_symbol, position_of(p, args), car(p), T_REAL));
	  return(x);
	}

    MAX_REALS:
      if (is_null(p)) return(x);
      y = car(p);
      p = cdr(p);

      switch (type(y))
	{
	case T_INTEGER:
	  if (real(x) < integer(y))
	    {
	      x = y;
	      goto MAX_INTEGERS;
	    }
	  goto MAX_REALS;

	case T_RATIO:
	  if (real(x) < fraction(y))
	    {
	      x = y;
	      goto MAX_RATIOS;
	    }
	  goto MAX_REALS;

	case T_REAL:
	  if (is_NaN(real(y)))
	    {
	      for (; is_not_null(p); p = cdr(p))
		if (!is_real_via_method(sc, car(p)))
		  return(wrong_type_argument(sc, sc->max_symbol, position_of(p, args), car(p), T_REAL));
	      return(y);
	    }
	  if (real(x) < real(y)) x = y;
	  goto MAX_REALS;

	default:
	  method_or_bust(sc, y, sc->max_symbol, cons_unchecked(sc, x, cons(sc, y, p)), T_REAL, position_of(p, args) - 1);
	}

    default:
      method_or_bust(sc, x, sc->max_symbol, cons(sc, x, p), T_REAL, 1);
    }
}

static s7_int max_i_ii(s7_int i1, s7_int i2) {return((i1 > i2) ? i1 : i2);}
static s7_double max_d_dd(s7_double x1, s7_double x2) {return((x1 > x2) ? x1 : x2);}

static s7_pointer g_min(s7_scheme *sc, s7_pointer args)
{
  #define H_min "(min ...) returns the minimum of its arguments"
  #define Q_min pcl_r

  s7_pointer x, y, p;
  s7_int num_a, num_b, den_a, den_b;

  x = car(args);
  p = cdr(args);

  switch (type(x))
    {
    case T_INTEGER:
    MIN_INTEGERS:
      if (is_null(p)) return(x);
      y = car(p);
      p = cdr(p);

      switch (type(y))
	{
	case T_INTEGER:
	  if (integer(x) > integer(y)) x = y;
	  goto MIN_INTEGERS;

	case T_RATIO:
	  num_a = integer(x);
	  den_a = 1;
	  num_b = numerator(y);
	  den_b = denominator(y);
	  goto RATIO_MIN_RATIO;

	case T_REAL:
	  if (is_NaN(real(y)))
	    {
	      for (; is_not_null(p); p = cdr(p))
		if (!is_real_via_method(sc, car(p)))
		  return(wrong_type_argument(sc, sc->min_symbol, position_of(p, args), car(p), T_REAL));
	      return(y);
	    }
	  if (integer(x) > real(y))
	    {
	      x = y;
	      goto MIN_REALS;
	    }
	  goto MIN_INTEGERS;

	default:
	  method_or_bust(sc, y, sc->min_symbol, cons_unchecked(sc, x, cons(sc, y, p)), T_REAL, position_of(p, args) - 1);
	}


    case T_RATIO:
    MIN_RATIOS:
      if (is_null(p)) return(x);
      y = car(p);
      p = cdr(p);

      switch (type(y))
	{
	case T_INTEGER:
	  num_a = numerator(x);
	  den_a = denominator(x);
	  num_b = integer(y);
	  den_b = 1;
	  goto RATIO_MIN_RATIO;

	case T_RATIO:
	  num_a = numerator(x);
	  den_a = denominator(x);
	  num_b = numerator(y);
	  den_b = denominator(y);

	RATIO_MIN_RATIO:
	  if ((num_a >= 0) && (num_b < 0))
	    x = y;
	  else
	    {
	      if ((num_a >= 0) || (num_b < 0))
		{
		  if (den_a == den_b)
		    {
		      if (num_a > num_b)
			x = y;
		    }
		  else
		    {
		      if (num_a == num_b)
			{
			  if (((num_a >= 0) &&
			       (den_a < den_b)) ||
			      ((num_a < 0) &&
			       (den_a > den_b)))
			    x = y;
			}
		      else
			{
			  s7_int vala, valb;
			  vala = num_a / den_a;
			  valb = num_b / den_b;
			  
			  if (!((vala < valb) ||
				((vala == valb) && (is_t_integer(x)))))
			    {
			      if ((valb < vala) ||
				  ((vala == valb) && (is_t_integer(y))) ||
				  (((long double)(num_a % den_a) / (long double)den_a) >= ((long double)(num_b % den_b) / (long double)den_b)))
				x = y;
			    }
			}
		    }
		}
	    }
	  if (is_t_ratio(x))
	    goto MIN_RATIOS;
	  goto MIN_INTEGERS;

	case T_REAL:
	  /* (min 3/4 nan.0) should probably return NaN */
	  if (is_NaN(real(y)))
	    {
	      for (; is_not_null(p); p = cdr(p))
		if (!is_real_via_method(sc, car(p)))
		  return(wrong_type_argument(sc, sc->min_symbol, position_of(p, args), car(p), T_REAL));
	      return(y);
	    }
	  if (fraction(x) > real(y))
	    {
	      x = y;
	      goto MIN_REALS;
	    }
	  goto MIN_RATIOS;

	default:
	  method_or_bust(sc, y, sc->min_symbol, cons_unchecked(sc, x, cons(sc, y, p)), T_REAL, position_of(p, args) - 1);
	}


    case T_REAL:
      if (is_NaN(real(x)))
	{
	  for (; is_not_null(p); p = cdr(p))
	    if (!is_real_via_method(sc, car(p)))
	      return(wrong_type_argument(sc, sc->min_symbol, position_of(p, args), car(p), T_REAL));
	  return(x);
	}

    MIN_REALS:
      if (is_null(p)) return(x);
      y = car(p);
      p = cdr(p);

      switch (type(y))
	{
	case T_INTEGER:
	  if (real(x) > integer(y))
	    {
	      x = y;
	      goto MIN_INTEGERS;
	    }
	  goto MIN_REALS;

	case T_RATIO:
	  if (real(x) > fraction(y))
	    {
	      x = y;
	      goto MIN_RATIOS;
	    }
	  goto MIN_REALS;

	case T_REAL:
	  if (is_NaN(real(y)))
	    {
	      for (; is_not_null(p); p = cdr(p))
		if (!is_real_via_method(sc, car(p)))
		  return(wrong_type_argument(sc, sc->min_symbol, position_of(p, args), car(p), T_REAL));
	      return(y);
	    }
	  if (real(x) > real(y)) x = y;
	  goto MIN_REALS;

	default:
	  method_or_bust(sc, y, sc->min_symbol, cons_unchecked(sc, x, cons(sc, y, p)), T_REAL, position_of(p, args) - 1);
	}

    default:
      method_or_bust(sc, x, sc->min_symbol, cons(sc, x, p), T_REAL, 1);
    }
}

static s7_int min_i_ii(s7_int i1, s7_int i2) {return((i1 < i2) ? i1 : i2);}
static s7_double min_d_dd(s7_double x1, s7_double x2) {return((x1 < x2) ? x1 : x2);}


/* ---------------------------------------- = > < >= <= ---------------------------------------- */

static s7_pointer g_equal(s7_scheme *sc, s7_pointer args)
{
  #define H_equal "(= z1 ...) returns #t if all its arguments are equal"
  #define Q_equal s7_make_circular_signature(sc, 1, 2, sc->is_boolean_symbol, sc->is_number_symbol)
  s7_pointer x, p;
  s7_int num_a, den_a;
  s7_double rl_a, im_a;

  x = car(args);
  p = cdr(args);

  switch (type(x))
    {
    case T_INTEGER:
      num_a = integer(x);
      while (true)
	{
	  x = car(p);
	  p = cdr(p);
	  switch (type(x))
	    {
	    case T_INTEGER:
	      if (num_a != integer(x)) goto NOT_EQUAL;
	      break;

	    case T_RATIO:
	    case T_COMPLEX:
	      goto NOT_EQUAL;

	    case T_REAL:
	      if (num_a != real(x)) goto NOT_EQUAL;
	      break;

	    default:
	      method_or_bust_with_type(sc, x, sc->eq_symbol, cons_unchecked(sc, make_integer(sc, num_a), cons(sc, x, p)), a_number_string, position_of(p, args) - 1);
	    }
	  if (is_null(p))
	    return(sc->T);
	}

    case T_RATIO:
      num_a = numerator(x);
      den_a = denominator(x);
      rl_a = 0.0;
      while (true)
	{
	  x = car(p);
	  p = cdr(p);
	  switch (type(x))
	    {
	    case T_INTEGER:
	    case T_COMPLEX:
	      goto NOT_EQUAL;

	    case T_RATIO:
	      if ((num_a != numerator(x)) || (den_a != denominator(x)))	goto NOT_EQUAL; /* hidden cast here */
	      break;

	    case T_REAL:
	      if (rl_a == 0.0)
		rl_a = ((long double)num_a) / ((long double)den_a);
	      if (rl_a != real(x)) goto NOT_EQUAL;
	      break;

	    default:
	      method_or_bust_with_type(sc, x, sc->eq_symbol, cons_unchecked(sc, s7_make_ratio(sc, num_a, den_a), cons(sc, x, p)), a_number_string, position_of(p, args) - 1);
	    }
	  if (is_null(p))
	    return(sc->T);
	}

    case T_REAL:
      rl_a = real(x);
      while (true)
	{
	  x = car(p);
	  p = cdr(p);
	  switch (type(x))
	    {
	    case T_INTEGER:
	      if (rl_a != integer(x)) goto NOT_EQUAL;
	      break;

	    case T_RATIO:
	      if (rl_a != (double)fraction(x)) goto NOT_EQUAL;
	      /* the cast to double is needed because rl_a is s7_double and we want (= ratio real) to be the same as (= real ratio):
	       *   (= 1.0 9223372036854775807/9223372036854775806)
	       *   (= 9223372036854775807/9223372036854775806 1.0)
	       */
	      break;

	    case T_REAL:
	      if (rl_a != real(x)) goto NOT_EQUAL;
	      break;

	    case T_COMPLEX:
	      goto NOT_EQUAL;

	    default:
	      method_or_bust_with_type(sc, x, sc->eq_symbol, cons_unchecked(sc, make_real(sc, rl_a), cons(sc, x, p)), a_number_string, position_of(p, args) - 1);
	    }
	  if (is_null(p))
	    return(sc->T);
	}

    case T_COMPLEX:
      rl_a = real_part(x);
      im_a = imag_part(x);
      while (true)
	{
	  x = car(p);
	  p = cdr(p);
	  switch (type(x))
	    {
	    case T_INTEGER:
	    case T_RATIO:
	    case T_REAL:
	      goto NOT_EQUAL;
	      break;

	    case T_COMPLEX:
	      if ((rl_a != real_part(x)) || (im_a != imag_part(x)))
		goto NOT_EQUAL;
	      break;

	    default:
	      method_or_bust_with_type(sc, x, sc->eq_symbol, cons_unchecked(sc, s7_make_complex(sc, rl_a, im_a), cons(sc, x, p)), a_number_string, position_of(p, args) - 1);
	    }
	  if (is_null(p))
	    return(sc->T);
	}

    default:
      method_or_bust_with_type(sc, x, sc->eq_symbol, args, a_number_string, 1);
    }

 NOT_EQUAL:
  for (; is_pair(p); p = cdr(p))
    if (!is_number_via_method(sc, car(p)))
      return(wrong_type_argument_with_type(sc, sc->eq_symbol, position_of(p, args), car(p), a_number_string));

  return(sc->F);
}


static s7_int c_object_length_to_int(s7_scheme *sc, s7_pointer obj);

#if (!WITH_GMP)
static s7_pointer equal_s_ic;
static s7_pointer g_equal_s_ic(s7_scheme *sc, s7_pointer args)
{
  s7_int y;
  s7_pointer val;

  val = find_symbol_unchecked(sc, car(args));
  y = s7_integer(cadr(args));
  if (is_integer(val))
    return(make_boolean(sc, integer(val) == y));

  switch (type(val))
    {
    case T_INTEGER: return(make_boolean(sc, integer(val) == y));
    case T_RATIO:   return(sc->F);
    case T_REAL:    return(make_boolean(sc, real(val) == y));
    case T_COMPLEX: return(sc->F);
    default:
      method_or_bust_with_type(sc, val, sc->eq_symbol, list_2(sc, val, cadr(args)), a_number_string, 1);
    }
  return(sc->T);
}

static s7_pointer equal_length_ic;
static s7_pointer g_equal_length_ic(s7_scheme *sc, s7_pointer args)
{
  /* avoid make_integer (and telescope opts), we get here with car=length expr, cadr=int */
  s7_int ilen;
  s7_pointer val;

  val = find_symbol_unchecked(sc, cadar(args));
  ilen = s7_integer(cadr(args));

  switch (type(val))
    {
    case T_PAIR:         return(make_boolean(sc, s7_list_length(sc, val) == ilen));
    case T_NIL:          return(make_boolean(sc, ilen == 0));
    case T_STRING:       return(make_boolean(sc, string_length(val) == ilen));
    case T_HASH_TABLE:   return(make_boolean(sc, (hash_table_mask(val) + 1) == ilen));
    case T_ITERATOR:     return(make_boolean(sc, iterator_length(val) == ilen));
    case T_C_OBJECT:     return(make_boolean(sc, c_object_length_to_int(sc, val) == ilen));
    case T_LET:          return(make_boolean(sc, let_length(sc, val) == ilen));
    case T_INT_VECTOR:
    case T_FLOAT_VECTOR:
    case T_VECTOR:       return(make_boolean(sc, vector_length(val) == ilen));
    case T_CLOSURE:
    case T_CLOSURE_STAR: if (has_methods(val)) return(make_boolean(sc, closure_length(sc, val) == ilen));
    default:             return(simple_wrong_type_argument_with_type(sc, sc->length_symbol, val, a_sequence_string));
      /* here we already lost because we checked for the length above */
    }
  return(sc->F);
}


static bool equal_b_pi(s7_scheme *sc, s7_pointer x, s7_int y)
{
  if (is_integer(x))
    return(integer(x) == y);

  switch (type(x))
    {
    case T_INTEGER: return(integer(x) == y);
    case T_RATIO:   return(false);
    case T_REAL:    return(real(x) == y);
    case T_COMPLEX: return(false);
    default:
      simple_wrong_type_argument_with_type(cur_sc, sc->eq_symbol, x, a_number_string);
    }
  return(false);
}
#endif

static s7_pointer c_equal_2(s7_scheme *sc, s7_pointer x, s7_pointer y)
{
#if (!MS_WINDOWS)
  if (type(x) == type(y))
    {
      if (type(x) == T_INTEGER)
	return(make_boolean(sc, integer(x) == integer(y)));
      if (type(x) == T_REAL)
	return(make_boolean(sc, real(x) == real(y)));
      if (type(x) == T_COMPLEX)
	return(make_boolean(sc, (real_part(x) == real_part(y)) && (imag_part(x) == imag_part(y))));
      if (type(x) == T_RATIO)
	return(make_boolean(sc, (numerator(x) == numerator(y)) && (denominator(x) == denominator(y))));
    }
#endif

  switch (type(x))
    {
    case T_INTEGER:
      switch (type(y))
	{
	case T_INTEGER: return(make_boolean(sc, integer(x) == integer(y)));
	case T_RATIO:   return(sc->F);
	case T_REAL:    return(make_boolean(sc, integer(x) == real(y)));
	case T_COMPLEX: return(sc->F);
	default:
	  method_or_bust_with_type(sc, y, sc->eq_symbol, list_2(sc, x, y), a_number_string, 2);
	}
      break;

    case T_RATIO:
      switch (type(y))
	{
	case T_INTEGER: return(sc->F);
	case T_RATIO:   return(make_boolean(sc, (numerator(x) == numerator(y)) && (denominator(x) == denominator(y))));
	case T_REAL:    return(make_boolean(sc, fraction(x) == real(y)));            /* this could avoid the divide via numerator == denominator * x */
	case T_COMPLEX: return(sc->F);
	default:
	  method_or_bust_with_type(sc, y, sc->eq_symbol, list_2(sc, x, y), a_number_string, 2);
	}
      break;

    case T_REAL:
      switch (type(y))
	{
	case T_INTEGER: return(make_boolean(sc, real(x) == integer(y)));
	case T_RATIO:   return(make_boolean(sc, real(x) == fraction(y)));
	case T_REAL:    return(make_boolean(sc, real(x) == real(y)));
	case T_COMPLEX: return(sc->F);
	default:
	  method_or_bust_with_type(sc, y, sc->eq_symbol, list_2(sc, x, y), a_number_string, 2);
	}
      break;

    case T_COMPLEX:
      switch (type(y))
	{
	case T_INTEGER:
	case T_RATIO:
	case T_REAL:
	  return(sc->F);

#if (!MS_WINDOWS)
	case T_COMPLEX:
	  return(make_boolean(sc, (real_part(x) == real_part(y)) && (imag_part(x) == imag_part(y))));
#else
	case T_COMPLEX:
	  if ((real_part(x) == real_part(y)) && (imag_part(x) == imag_part(y))) return(sc->T); else return(sc->F);
#endif
	default:
	  method_or_bust_with_type(sc, y, sc->eq_symbol, list_2(sc, x, y), a_number_string, 2);
	}
      break;

    default:
      method_or_bust_with_type(sc, x, sc->eq_symbol, list_2(sc, x, y), a_number_string, 1);
    }
  return(sc->F);
}

#if (!WITH_GMP)
static s7_pointer equal_p_pp(s7_pointer p1, s7_pointer p2) {return(c_equal_2(cur_sc, p1, p2));}
static s7_pointer equal_p_pi(s7_pointer p1, s7_int p2) 
{
  if (is_integer(p1)) 
    return((integer(p1) == p2) ? cur_sc->T : cur_sc->F);
  if (is_t_real(p1))
    return((real(p1) == p2) ? cur_sc->T : cur_sc->F);
  if (is_number(p1))
    return(cur_sc->F);
  return(wrong_type_argument_with_type(cur_sc, cur_sc->eq_symbol, 1, p1, a_number_string));  
}

static s7_pointer equal_p_dd(s7_double x1, s7_double x2) {return(make_boolean(cur_sc, x1 == x2));}
static s7_pointer gt_p_dd(s7_double x1, s7_double x2) {return(make_boolean(cur_sc, x1 > x2));}
static s7_pointer geq_p_dd(s7_double x1, s7_double x2) {return(make_boolean(cur_sc, x1 >= x2));}
static s7_pointer lt_p_dd(s7_double x1, s7_double x2) {return(make_boolean(cur_sc, x1 < x2));}
static s7_pointer leq_p_dd(s7_double x1, s7_double x2) {return(make_boolean(cur_sc, x1 <= x2));}
static s7_pointer equal_p_ii(s7_int x1, s7_int x2) {return(make_boolean(cur_sc, x1 == x2));}
static s7_pointer gt_p_ii(s7_int x1, s7_int x2) {return(make_boolean(cur_sc, x1 > x2));}
static s7_pointer geq_p_ii(s7_int x1, s7_int x2) {return(make_boolean(cur_sc, x1 >= x2));}
static s7_pointer lt_p_ii(s7_int x1, s7_int x2) {return(make_boolean(cur_sc, x1 < x2));}
static s7_pointer leq_p_ii(s7_int x1, s7_int x2) {return(make_boolean(cur_sc, x1 <= x2));}

static s7_pointer equal_2;
#endif

static s7_pointer g_equal_2(s7_scheme *sc, s7_pointer args) {return(c_equal_2(sc, car(args), cadr(args)));}

#if (!WITH_GMP)
static s7_pointer g_less(s7_scheme *sc, s7_pointer args)
{
  #define H_less "(< x1 ...) returns #t if its arguments are in increasing order"
  #define Q_less s7_make_circular_signature(sc, 1, 2, sc->is_boolean_symbol, sc->is_real_symbol)

  s7_pointer x, y, p;

  x = car(args);
  p = cdr(args);

  switch (type(x))
    {
    case T_INTEGER:
    INTEGER_LESS:
      y = car(p);
      p = cdr(p);
      switch (type(y))
	{
	case T_INTEGER:
	  if (integer(x) >= integer(y)) goto NOT_LESS;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto INTEGER_LESS;

	case T_RATIO:
	  /* no gmp here, but this can overflow: (< 9223372036 1/9223372036), but conversion to real is also problematic
	   */
	  if ((integer(x) >= 0) && (numerator(y) < 0)) goto NOT_LESS;  /* (< 1 -1/2), ratio numerator can't be 0 */
	  if ((integer(x) <= 0) && (numerator(y) > 0))                 /* (< 0 1/2) */
	    {
	      if (is_null(p)) return(sc->T);
	      x = y;
	      goto RATIO_LESS;
	    }
	  if ((integer(x) < s7_int32_max) &&
	      (integer(x) > s7_int32_min) &&
	      (denominator(y) < s7_int32_max))
	    {
	      if ((integer(x) * denominator(y)) >= numerator(y)) goto NOT_LESS;
	    }
	  else
	    {
	      if (integer(x) >= fraction(y)) goto NOT_LESS;
	    }
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto RATIO_LESS;

	case T_REAL:
	  if (is_NaN(real(y))) goto NOT_LESS;
	  if (integer(x) >= real(y)) goto NOT_LESS;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto REAL_LESS;

	default:
	  method_or_bust(sc, y, sc->lt_symbol, cons_unchecked(sc, x, cons(sc, y, p)), T_REAL, position_of(p, args) - 1);
	}


    case T_RATIO:
    RATIO_LESS:
      y = car(p);
      p = cdr(p);
      switch (type(y))
	{
	case T_INTEGER:
	  if ((numerator(x) > 0) && (integer(y) <= 0)) goto NOT_LESS;
	  if ((numerator(x) < 0) && (integer(y) >= 0))
	    {
	      if (is_null(p)) return(sc->T);
	      x = y;
	      goto INTEGER_LESS;
	    }
	  if ((integer(y) < s7_int32_max) &&
	      (integer(y) > s7_int32_min) &&
	      (denominator(x) < s7_int32_max))
	    {
	      if (numerator(x) >= (integer(y) * denominator(x))) goto NOT_LESS;
	    }
	  else
	    {
	      if (fraction(x) >= integer(y)) goto NOT_LESS;
	    }
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto INTEGER_LESS;

	case T_RATIO:
	  /* conversion to real and >= is not safe here (see comment under g_greater) */
	  {
	    s7_int d1, d2, n1, n2;
	    d1 = denominator(x);
	    n1 = numerator(x);
	    d2 = denominator(y);
	    n2 = numerator(y);
	    if (d1 == d2)
	      {
		if (n1 >= n2) goto NOT_LESS;
	      }
	    else
	      {
#if HAVE_OVERFLOW_CHECKS
		if ((multiply_overflow(n1, d2, &n1)) ||
		    (multiply_overflow(n2, d1, &n2)))
		  {
		    if (fraction(x) >= fraction(y)) goto NOT_LESS;
		  }
		else
		  {
		    if (n1 >= n2) goto NOT_LESS;
		  }
#else
		if ((n1 * d2) >=  (n2 * d1)) goto NOT_LESS;
#endif
	      }
	  }
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto RATIO_LESS;

	case T_REAL:
	  if (is_NaN(real(y))) goto NOT_LESS;
	  if (fraction(x) >= real(y)) goto NOT_LESS;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto REAL_LESS;

	default:
	  method_or_bust(sc, y, sc->lt_symbol, cons_unchecked(sc, x, cons(sc, y, p)), T_REAL, position_of(p, args) - 1);
	}


    case T_REAL:
      if (is_NaN(real(x))) goto NOT_LESS;

    REAL_LESS:
      y = car(p);
      p = cdr(p);
      switch (type(y))
	{
	case T_INTEGER:
	  if (real(x) >= integer(y)) goto NOT_LESS;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto INTEGER_LESS;

	case T_RATIO:
	  if (real(x) >= fraction(y)) goto NOT_LESS;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto RATIO_LESS;

	case T_REAL:
	  if (is_NaN(real(y))) goto NOT_LESS;
	  if (real(x) >= real(y)) goto NOT_LESS;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto REAL_LESS;

	default:
	  method_or_bust(sc, y, sc->lt_symbol, cons_unchecked(sc, x, cons(sc, y, p)), T_REAL, position_of(p, args) - 1);
	}

    default:
      method_or_bust(sc, x, sc->lt_symbol, args, T_REAL, 1);
    }

 NOT_LESS:
  for (; is_pair(p); p = cdr(p))
    if (!is_real_via_method(sc, car(p)))
      return(wrong_type_argument(sc, sc->lt_symbol, position_of(p, args), car(p), T_REAL));

  return(sc->F);
}


static s7_pointer g_less_or_equal(s7_scheme *sc, s7_pointer args)
{
  #define H_less_or_equal "(<= x1 ...) returns #t if its arguments are in increasing order"
  #define Q_less_or_equal s7_make_circular_signature(sc, 1, 2, sc->is_boolean_symbol, sc->is_real_symbol)

  s7_pointer x, y, p;

  x = car(args);
  p = cdr(args);

  switch (type(x))
    {
    case T_INTEGER:
    INTEGER_LEQ:
      y = car(p);
      p = cdr(p);
      switch (type(y))
	{
	case T_INTEGER:
	  if (integer(x) > integer(y)) goto NOT_LEQ;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto INTEGER_LEQ;

	case T_RATIO:
	  /* no gmp here, but this can overflow: (< 9223372036 1/9223372036), but conversion to real is also problematic
	   */
	  if ((integer(x) >= 0) && (numerator(y) < 0)) goto NOT_LEQ;  /* (< 1 -1/2), ratio numerator can't be 0 */
	  if ((integer(x) <= 0) && (numerator(y) > 0))                 /* (< 0 1/2) */
	    {
	      if (is_null(p)) return(sc->T);
	      x = y;
	      goto RATIO_LEQ;
	    }
	  if ((integer(x) < s7_int32_max) &&
	      (integer(x) > s7_int32_min) &&
	      (denominator(y) < s7_int32_max))
	    {
	      if ((integer(x) * denominator(y)) > numerator(y)) goto NOT_LEQ;
	    }
	  else
	    {
	      if (integer(x) > fraction(y)) goto NOT_LEQ;
	    }
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto RATIO_LEQ;

	case T_REAL:
	  if (is_NaN(real(y))) goto NOT_LEQ;
	  if (integer(x) > real(y)) goto NOT_LEQ;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto REAL_LEQ;

	default:
	  method_or_bust(sc, y, sc->leq_symbol, cons_unchecked(sc, x, cons(sc, y, p)), T_REAL, position_of(p, args) - 1);
	}


    case T_RATIO:
    RATIO_LEQ:
      y = car(p);
      p = cdr(p);
      switch (type(y))
	{
	case T_INTEGER:
	  if ((numerator(x) > 0) && (integer(y) <= 0)) goto NOT_LEQ;
	  if ((numerator(x) < 0) && (integer(y) >= 0))
	    {
	      if (is_null(p)) return(sc->T);
	      x = y;
	      goto INTEGER_LEQ;
	    }
	  if ((integer(y) < s7_int32_max) &&
	      (integer(y) > s7_int32_min) &&
	      (denominator(x) < s7_int32_max))
	    {
	      if (numerator(x) > (integer(y) * denominator(x))) goto NOT_LEQ;
	    }
	  else
	    {
	      if (fraction(x) > integer(y)) goto NOT_LEQ;
	    }
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto INTEGER_LEQ;

	case T_RATIO:
	  {
	    s7_int d1, d2, n1, n2;
	    d1 = denominator(x);
	    n1 = numerator(x);
	    d2 = denominator(y);
	    n2 = numerator(y);
	    if (d1 == d2)
	      {
		if (n1 > n2) goto NOT_LEQ;
	      }
	    else
	      {
#if HAVE_OVERFLOW_CHECKS
		if ((multiply_overflow(n1, d2, &n1)) ||
		    (multiply_overflow(n2, d1, &n2)))
		  {
		    if (fraction(x) > fraction(y)) goto NOT_LEQ;
		  }
		else
		  {
		    if (n1 > n2) goto NOT_LEQ;
		  }
#else
		if ((n1 * d2) >  (n2 * d1)) goto NOT_LEQ;
#endif
	      }
	  }
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto RATIO_LEQ;

	case T_REAL:
	  if (is_NaN(real(y))) goto NOT_LEQ;
	  if (fraction(x) > real(y)) goto NOT_LEQ;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto REAL_LEQ;

	default:
	  method_or_bust(sc, y, sc->leq_symbol, cons_unchecked(sc, x, cons(sc, y, p)), T_REAL, position_of(p, args) - 1);
	}


    case T_REAL:
      if (is_NaN(real(x))) goto NOT_LEQ;

    REAL_LEQ:
      y = car(p);
      p = cdr(p);
      switch (type(y))
	{
	case T_INTEGER:
	  if (real(x) > integer(y)) goto NOT_LEQ;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto INTEGER_LEQ;

	case T_RATIO:
	  if (real(x) > fraction(y)) goto NOT_LEQ;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto RATIO_LEQ;

	case T_REAL:
	  if (is_NaN(real(y))) goto NOT_LEQ;
	  if (real(x) > real(y)) goto NOT_LEQ;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto REAL_LEQ;

	default:
	  method_or_bust(sc, y, sc->leq_symbol, cons_unchecked(sc, x, cons(sc, y, p)), T_REAL, position_of(p, args) - 1);
	}

    default:
      method_or_bust(sc, x, sc->leq_symbol, args, T_REAL, 1);
    }

 NOT_LEQ:
  for (; is_pair(p); p = cdr(p))
    if (!is_real_via_method(sc, car(p)))
      return(wrong_type_argument(sc, sc->leq_symbol, position_of(p, args), car(p), T_REAL));

  return(sc->F);
}


static s7_pointer g_greater(s7_scheme *sc, s7_pointer args)
{
  #define H_greater "(> x1 ...) returns #t if its arguments are in decreasing order"
  #define Q_greater s7_make_circular_signature(sc, 1, 2, sc->is_boolean_symbol, sc->is_real_symbol)

  s7_pointer x, y, p;
  x = car(args);
  p = cdr(args);

  switch (type(x))
    {
    case T_INTEGER:
    INTEGER_GREATER:
      y = car(p);
      p = cdr(p);
      switch (type(y))
	{
	case T_INTEGER:
	  if (integer(x) <= integer(y)) goto NOT_GREATER;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto INTEGER_GREATER;

	case T_RATIO:
	  /* no gmp here, but this can overflow: (< 9223372036 1/9223372036), but conversion to real is also problematic
	   */
	  if ((integer(x) <= 0) && (numerator(y) > 0)) goto NOT_GREATER;
	  if ((integer(x) >= 0) && (numerator(y) < 0))
	    {
	      if (is_null(p)) return(sc->T);
	      x = y;
	      goto RATIO_GREATER;
	    }
	  if ((integer(x) < s7_int32_max) &&
	      (integer(x) > s7_int32_min) &&
	      (denominator(y) < s7_int32_max))
	    {
	      if ((integer(x) * denominator(y)) <= numerator(y)) goto NOT_GREATER;
	    }
	  else
	    {
	      if (integer(x) <= fraction(y)) goto NOT_GREATER;
	    }
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto RATIO_GREATER;

	case T_REAL:
	  if (is_NaN(real(y))) goto NOT_GREATER;
	  if (integer(x) <= real(y)) goto NOT_GREATER;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto REAL_GREATER;

	default:
	  method_or_bust(sc, y, sc->gt_symbol, cons_unchecked(sc, x, cons(sc, y, p)), T_REAL, position_of(p, args) - 1);
	}


    case T_RATIO:
    RATIO_GREATER:
      y = car(p);
      p = cdr(p);
      switch (type(y))
	{
	case T_INTEGER:
	  if ((numerator(x) < 0) && (integer(y) >= 0)) goto NOT_GREATER;
	  if ((numerator(x) > 0) && (integer(y) <= 0))
	    {
	      if (is_null(p)) return(sc->T);
	      x = y;
	      goto INTEGER_GREATER;
	    }
	  if ((integer(y) < s7_int32_max) &&
	      (integer(y) > s7_int32_min) &&
	      (denominator(x) < s7_int32_max))
	    {
	      if (numerator(x) <= (integer(y) * denominator(x))) goto NOT_GREATER;
	    }
	  else
	    {
	      if (fraction(x) <= integer(y)) goto NOT_GREATER;
	    }
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto INTEGER_GREATER;

	case T_RATIO:
	  {
	    s7_int d1, d2, n1, n2;
	    d1 = denominator(x);
	    n1 = numerator(x);
	    d2 = denominator(y);
	    n2 = numerator(y);
	    if (d1 == d2)
	      {
		if (n1 <= n2) goto NOT_GREATER;
	      }
	    else
	      {
#if HAVE_OVERFLOW_CHECKS
		if ((multiply_overflow(n1, d2, &n1)) ||
		    (multiply_overflow(n2, d1, &n2)))
		  {
		    if (fraction(x) <= fraction(y)) goto NOT_GREATER;
		  }
		else
		  {
		    if (n1 <= n2) goto NOT_GREATER;
		  }
#else
		if ((n1 * d2) <=  (n2 * d1)) goto NOT_GREATER;
#endif
	      }
	  }
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto RATIO_GREATER;

	case T_REAL:
	  if (is_NaN(real(y))) goto NOT_GREATER;
	  if (fraction(x) <= real(y)) goto NOT_GREATER;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto REAL_GREATER;

	default:
	  method_or_bust(sc, y, sc->gt_symbol, cons_unchecked(sc, x, cons(sc, y, p)), T_REAL, position_of(p, args) - 1);
	}


    case T_REAL:
      if (is_NaN(real(x))) goto NOT_GREATER;

    REAL_GREATER:
      y = car(p);
      p = cdr(p);
      switch (type(y))
	{
	case T_INTEGER:
	  if (real(x) <= integer(y)) goto NOT_GREATER;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto INTEGER_GREATER;

	case T_RATIO:
	  if (real(x) <= fraction(y)) goto NOT_GREATER;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto RATIO_GREATER;

	case T_REAL:
	  if (is_NaN(real(y))) goto NOT_GREATER;
	  if (real(x) <= real(y)) goto NOT_GREATER;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto REAL_GREATER;

	default:
	  method_or_bust(sc, y, sc->gt_symbol, cons_unchecked(sc, x, cons(sc, y, p)), T_REAL, position_of(p, args) - 1);
	}

    default:
      method_or_bust(sc, x, sc->gt_symbol, args, T_REAL, 1);
    }

 NOT_GREATER:
  for (; is_pair(p); p = cdr(p))
    if (!is_real_via_method(sc, car(p)))
      return(wrong_type_argument(sc, sc->gt_symbol, position_of(p, args), car(p), T_REAL));

  return(sc->F);
}


static s7_pointer g_greater_or_equal(s7_scheme *sc, s7_pointer args)
{
  #define H_greater_or_equal "(>= x1 ...) returns #t if its arguments are in decreasing order"
  #define Q_greater_or_equal s7_make_circular_signature(sc, 1, 2, sc->is_boolean_symbol, sc->is_real_symbol)
  /* (>= 1+i 1+i) is an error which seems unfortunate */
  s7_pointer x, y, p;

  x = car(args);
  p = cdr(args);

  switch (type(x))
    {
    case T_INTEGER:
    INTEGER_GEQ:
      y = car(p);
      p = cdr(p);
      switch (type(y))
	{
	case T_INTEGER:
	  if (integer(x) < integer(y)) goto NOT_GEQ;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto INTEGER_GEQ;

	case T_RATIO:
	  /* no gmp here, but this can overflow: (< 9223372036 1/9223372036), but conversion to real is also problematic
	   */
	  if ((integer(x) <= 0) && (numerator(y) > 0)) goto NOT_GEQ;
	  if ((integer(x) >= 0) && (numerator(y) < 0))
	    {
	      if (is_null(p)) return(sc->T);
	      x = y;
	      goto RATIO_GEQ;
	    }
	  if ((integer(x) < s7_int32_max) &&
	      (integer(x) > s7_int32_min) &&
	      (denominator(y) < s7_int32_max))
	    {
	      if ((integer(x) * denominator(y)) < numerator(y)) goto NOT_GEQ;
	    }
	  else
	    {
	      if (integer(x) < fraction(y)) goto NOT_GEQ;
	    }
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto RATIO_GEQ;

	case T_REAL:
	  if (is_NaN(real(y))) goto NOT_GEQ;
	  if (integer(x) < real(y)) goto NOT_GEQ;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto REAL_GEQ;

	default:
	  method_or_bust(sc, y, sc->geq_symbol, cons_unchecked(sc, x, cons(sc, y, p)), T_REAL, position_of(p, args) - 1);
	}


    case T_RATIO:
    RATIO_GEQ:
      y = car(p);
      p = cdr(p);
      switch (type(y))
	{
	case T_INTEGER:
	  if ((numerator(x) < 0) && (integer(y) >= 0)) goto NOT_GEQ;
	  if ((numerator(x) > 0) && (integer(y) <= 0))
	    {
	      if (is_null(p)) return(sc->T);
	      x = y;
	      goto INTEGER_GEQ;
	    }
	  if ((integer(y) < s7_int32_max) &&
	      (integer(y) > s7_int32_min) &&
	      (denominator(x) < s7_int32_max))
	    {
	      if (numerator(x) < (integer(y) * denominator(x))) goto NOT_GEQ;
	    }
	  else
	    {
	      if (fraction(x) < integer(y)) goto NOT_GEQ;
	    }
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto INTEGER_GEQ;

	case T_RATIO:
	  {
	    s7_int d1, d2, n1, n2;
	    d1 = denominator(x);
	    n1 = numerator(x);
	    d2 = denominator(y);
	    n2 = numerator(y);
	    if (d1 == d2)
	      {
		if (n1 < n2) goto NOT_GEQ;
	      }
	    else
	      {
#if HAVE_OVERFLOW_CHECKS
		if ((multiply_overflow(n1, d2, &n1)) ||
		    (multiply_overflow(n2, d1, &n2)))
		  {
		    if (fraction(x) < fraction(y)) goto NOT_GEQ;
		  }
		else
		  {
		    if (n1 < n2) goto NOT_GEQ;
		  }
#else
		if ((n1 * d2) <  (n2 * d1)) goto NOT_GEQ;
#endif
	      }
	  }
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto RATIO_GEQ;

	case T_REAL:
	  if (is_NaN(real(y))) goto NOT_GEQ;
	  if (fraction(x) < real(y)) goto NOT_GEQ;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto REAL_GEQ;

	default:
	  method_or_bust(sc, y, sc->geq_symbol, cons_unchecked(sc, x, cons(sc, y, p)), T_REAL, position_of(p, args) - 1);
	}


    case T_REAL:
      if (is_NaN(real(x))) goto NOT_GEQ;

    REAL_GEQ:
      y = car(p);
      p = cdr(p);
      switch (type(y))
	{
	case T_INTEGER:
	  if (real(x) < integer(y)) goto NOT_GEQ;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto INTEGER_GEQ;

	case T_RATIO:
	  if (real(x) < fraction(y)) goto NOT_GEQ;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto RATIO_GEQ;

	case T_REAL:
	  if (is_NaN(real(y))) goto NOT_GEQ;
	  if (real(x) < real(y)) goto NOT_GEQ;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto REAL_GEQ;

	default:
	  method_or_bust(sc, y, sc->geq_symbol, cons_unchecked(sc, x, cons(sc, y, p)), T_REAL, position_of(p, args) - 1);
	}

    default:
      method_or_bust(sc, x, sc->geq_symbol, args, T_REAL, 1);
    }

 NOT_GEQ:
  for (; is_pair(p); p = cdr(p))
    if (!is_real_via_method(sc, car(p)))
      return(wrong_type_argument(sc, sc->geq_symbol, position_of(p, args), car(p), T_REAL));

  return(sc->F);

}


static s7_pointer less_s_ic, less_s0;
static s7_pointer g_less_s0(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x;
  x = car(args);
  if (is_integer(x))
    return(make_boolean(sc, integer(x) < 0));
  if (is_real(x))
    return(make_boolean(sc, s7_is_negative(x)));
  method_or_bust(sc, x, sc->lt_symbol, args, T_REAL, 1);
}

static bool ratio_lt_pi(s7_pointer x, s7_int y)
{
  if ((y >= 0) && (numerator(x) < 0))
    return(true);
  if ((y <= 0) && (numerator(x) > 0))
    return(false);
  if (denominator(x) < s7_int32_max)
    return(numerator(x) < (y * denominator(x)));
  return(fraction(x) < y);
}

static s7_pointer g_less_s_ic(s7_scheme *sc, s7_pointer args)
{
  s7_int y;
  s7_pointer x;

  x = car(args);
  y = integer(cadr(args));

  if (type(x) == T_INTEGER)
    return(make_boolean(sc, integer(x) < y));
  if (type(x) == T_REAL)
    return(make_boolean(sc, real(x) < y));
  if (type(x) == T_RATIO)
    return(make_boolean(sc, ratio_lt_pi(x, y)));
  method_or_bust(sc, x, sc->lt_symbol, args, T_REAL, 1);
}

static s7_pointer less_length_ic;
static s7_pointer g_less_length_ic(s7_scheme *sc, s7_pointer args)
{
  s7_int ilen;
  s7_pointer val;

  val = find_symbol_unchecked(sc, cadar(args));
  ilen = s7_integer(cadr(args));

  switch (type(val))
    {
    case T_PAIR:         return(make_boolean(sc, s7_list_length(sc, val) < ilen));
    case T_NIL:          return(make_boolean(sc, ilen > 0));
    case T_STRING:       return(make_boolean(sc, string_length(val) < ilen));
    case T_HASH_TABLE:   return(make_boolean(sc, (hash_table_mask(val) + 1) < ilen)); /* was <=? -- changed 15-Dec-15, then again 6-Jan-17: mask is len-1 */
    case T_ITERATOR:     return(make_boolean(sc, iterator_length(val) < ilen));
    case T_C_OBJECT:     return(make_boolean(sc, c_object_length_to_int(sc, val) < ilen));
    case T_LET:          return(make_boolean(sc, let_length(sc, val) < ilen));  /* this works because let_length handles the length method itself! */
    case T_INT_VECTOR:
    case T_FLOAT_VECTOR:
    case T_VECTOR:       return(make_boolean(sc, vector_length(val) < ilen));
    case T_CLOSURE:
    case T_CLOSURE_STAR: if (has_methods(val)) return(make_boolean(sc, closure_length(sc, val) < ilen));
    default:             return(simple_wrong_type_argument_with_type(sc, sc->length_symbol, val, a_sequence_string)); /* no check method here because we checked above */
    }
  return(sc->F);
}

static s7_pointer c_less_2(s7_scheme *sc, s7_pointer x, s7_pointer y)
{
#if (!MS_WINDOWS)
  if (type(x) == type(y))
    {
      if (type(x) == T_INTEGER)
	return(make_boolean(sc, integer(x) < integer(y)));
      if (type(x) == T_REAL)
	return(make_boolean(sc, real(x) < real(y)));
      if (type(x) == T_RATIO)
	return(make_boolean(sc, fraction(x) < fraction(y)));
    }
#endif

  switch (type(x))
    {
    case T_INTEGER:
      switch (type(y))
	{
	case T_INTEGER:
	  return(make_boolean(sc, integer(x) < integer(y)));

	case T_RATIO:
	  return(g_less(sc, set_plist_2(sc, x, y)));

	case T_REAL:
	  if (is_NaN(real(y))) return(sc->F);
	  return(make_boolean(sc, integer(x) < real(y)));

	default:
	  method_or_bust(sc, y, sc->lt_symbol, list_2(sc, x, y), T_REAL, 2);
	}
      break;

    case T_RATIO:
      return(g_less(sc, set_plist_2(sc, x, y)));

    case T_REAL:
      switch (type(y))
	{
	case T_INTEGER:
	  if (is_NaN(real(x))) return(sc->F);
	  return(make_boolean(sc, real(x) < integer(y)));

	case T_RATIO:
	  if (is_NaN(real(x))) return(sc->F);
	  return(make_boolean(sc, real(x) < fraction(y)));

	case T_REAL:
	  if (is_NaN(real(x))) return(sc->F);
	  /* if (is_NaN(real(y))) return(sc->F); */
	  return(make_boolean(sc, real(x) < real(y)));

	default:
	  method_or_bust(sc, y, sc->lt_symbol, list_2(sc, x, y), T_REAL, 2);
	}
      break;

    default:
      method_or_bust(sc, x, sc->lt_symbol, list_2(sc, x, y), T_REAL, 1);
    }
  return(sc->T);
}

static s7_pointer lt_p_pp(s7_pointer p1, s7_pointer p2) {return(c_less_2(cur_sc, p1, p2));}

static s7_pointer less_2;
static s7_pointer g_less_2(s7_scheme *sc, s7_pointer args) {return(c_less_2(sc, car(args), cadr(args)));}

static bool ratio_leq_pi(s7_pointer x, s7_int y)
{
  if ((y >= 0) && (numerator(x) <= 0))
    return(true);
  if ((y <= 0) && (numerator(x) > 0))
    return(false);
  if (denominator(x) < s7_int32_max)
    return(numerator(x) <= (y * denominator(x)));
  return(fraction(x) <= y);
}

static s7_pointer leq_s_ic;
static s7_pointer g_leq_s_ic(s7_scheme *sc, s7_pointer args)
{
  s7_int y;
  s7_pointer x;

  x = car(args);
  y = s7_integer(cadr(args));

  if (type(x) == T_INTEGER)
    return(make_boolean(sc, integer(x) <= y));
  if (type(x) == T_REAL)
    return(make_boolean(sc, real(x) <= y));
  if (type(x) == T_RATIO)
    return(make_boolean(sc, ratio_leq_pi(x, y)));
  method_or_bust(sc, x, sc->leq_symbol, args, T_REAL, 1);
}


static s7_pointer c_leq_2(s7_scheme *sc, s7_pointer x, s7_pointer y)
{

#if (!MS_WINDOWS)
  if (type(x) == type(y))
    {
      if (type(x) == T_INTEGER)
	return(make_boolean(sc, integer(x) <= integer(y)));
      if (type(x) == T_REAL)
	return(make_boolean(sc, real(x) <= real(y)));
      if (type(x) == T_RATIO)
	return(make_boolean(sc, fraction(x) <= fraction(y)));
    }
#endif

  switch (type(x))
    {
    case T_INTEGER:
      switch (type(y))
	{
	case T_INTEGER:
	  return(make_boolean(sc, integer(x) <= integer(y)));

	case T_RATIO:
	  return(g_less_or_equal(sc, set_plist_2(sc, x, y)));

	case T_REAL:
	  if (is_NaN(real(y))) return(sc->F);
	  return(make_boolean(sc, integer(x) <= real(y)));

	default:
	  method_or_bust(sc, y, sc->leq_symbol, list_2(sc, x, y), T_REAL, 2);
	}
      break;

    case T_RATIO:
      return(g_less_or_equal(sc, set_plist_2(sc, x, y)));

    case T_REAL:
      switch (type(y))
	{
	case T_INTEGER:
	  if (is_NaN(real(x))) return(sc->F);
	  return(make_boolean(sc, real(x) <= integer(y)));

	case T_RATIO:
	  if (is_NaN(real(x))) return(sc->F);
	  return(make_boolean(sc, real(x) <= fraction(y)));

	case T_REAL:
	  if (is_NaN(real(x))) return(sc->F);
	  /* if (is_NaN(real(y))) return(sc->F); */
	  return(make_boolean(sc, real(x) <= real(y)));

	default:
	  method_or_bust(sc, y, sc->leq_symbol, list_2(sc, x, y), T_REAL, 2);
	}
      break;

    default:
      method_or_bust(sc, x, sc->leq_symbol, list_2(sc, x, y), T_REAL, 1);
    }
  return(sc->T);
}

static s7_pointer leq_p_pp(s7_pointer p1, s7_pointer p2) {return(c_leq_2(cur_sc, p1, p2));}

static s7_pointer leq_2;
static s7_pointer g_leq_2(s7_scheme *sc, s7_pointer args) {return(c_leq_2(sc, car(args), cadr(args)));}

static s7_pointer greater_s_ic, greater_s_fc;
static s7_pointer g_greater_s_ic(s7_scheme *sc, s7_pointer args)
{
  s7_int y;
  s7_pointer x;

  x = car(args);
  y = integer(cadr(args));

  if (type(x) == T_INTEGER)
    return(make_boolean(sc, integer(x) > y));
  if (type(x) == T_REAL)
    return(make_boolean(sc, real(x) > y));
  if (type(x) == T_RATIO)
    return(make_boolean(sc, !ratio_leq_pi(x, y)));
  method_or_bust_with_type(sc, x, sc->gt_symbol, args, a_number_string, 1);
}

static s7_pointer g_greater_s_fc(s7_scheme *sc, s7_pointer args)
{
  s7_double y;
  s7_pointer x;

  x = car(args);
  y = real(cadr(args));

  if (is_t_real(x))
    return(make_boolean(sc, real(x) > y));

  switch (type(x))
    {
    case T_INTEGER:
      return(make_boolean(sc, integer(x) > y));

    case T_RATIO:
      /* (> 9223372036854775807/9223372036854775806 1.0) */
      if (denominator(x) < s7_int32_max) /* y range check was handled in greater_chooser */
	return(make_boolean(sc, (numerator(x) > (y * denominator(x)))));
      return(make_boolean(sc, fraction(x) > y));

    case T_REAL:
      return(make_boolean(sc, real(x) > y));

    default:
      method_or_bust_with_type(sc, x, sc->gt_symbol, args, a_number_string, 1);
    }
  return(sc->T);
}


static s7_pointer c_greater_2(s7_scheme *sc, s7_pointer x, s7_pointer y)
{
#if (!MS_WINDOWS)
  if (type(x) == type(y))
    {
      if (type(x) == T_INTEGER)
	return(make_boolean(sc, integer(x) > integer(y)));
      if (type(x) == T_REAL)
	return(make_boolean(sc, real(x) > real(y)));
      if (type(x) == T_RATIO)
	return(make_boolean(sc, fraction(x) > fraction(y)));
    }
#endif

  switch (type(x))
    {
    case T_INTEGER:
      switch (type(y))
	{
	case T_INTEGER:
	  return(make_boolean(sc, integer(x) > integer(y)));

	case T_RATIO:
	  return(g_greater(sc, set_plist_2(sc, x, y)));

	case T_REAL:
	  if (is_NaN(real(y))) return(sc->F);
	  return(make_boolean(sc, integer(x) > real(y)));

	default:
	  method_or_bust(sc, y, sc->gt_symbol, list_2(sc, x, y), T_REAL, 2);
	}
      break;

    case T_RATIO:
      return(g_greater(sc, set_plist_2(sc, x, y)));

    case T_REAL:
      switch (type(y))
	{
	case T_INTEGER:
	  if (is_NaN(real(x))) return(sc->F);
	  return(make_boolean(sc, real(x) > integer(y)));

	case T_RATIO:
	  if (is_NaN(real(x))) return(sc->F);
	  return(make_boolean(sc, real(x) > fraction(y)));

	case T_REAL:
	  if (is_NaN(real(x))) return(sc->F);
	  /* if (is_NaN(real(y))) return(sc->F); */
	  return(make_boolean(sc, real(x) > real(y)));

	default:
	  method_or_bust(sc, y, sc->gt_symbol, list_2(sc, x, y), T_REAL, 2);
	}
      break;

    default:
      method_or_bust(sc, x, sc->gt_symbol, list_2(sc, x, y), T_REAL, 1);
    }
  return(sc->T);
}

static s7_pointer gt_p_pp(s7_pointer p1, s7_pointer p2) {return(c_greater_2(cur_sc, p1, p2));}

static s7_pointer greater_2;
static s7_pointer g_greater_2(s7_scheme *sc, s7_pointer args) {return(c_greater_2(sc, car(args), cadr(args)));}

static s7_pointer c_geq_2(s7_scheme *sc, s7_pointer x, s7_pointer y)
{
#if (!MS_WINDOWS)
  if (type(x) == type(y))
    {
      if (type(x) == T_INTEGER)
	return(make_boolean(sc, integer(x) >= integer(y)));
      if (type(x) == T_REAL)
	return(make_boolean(sc, real(x) >= real(y)));
      if (type(x) == T_RATIO)
	return(make_boolean(sc, fraction(x) >= fraction(y)));
    }
#endif

  switch (type(x))
    {
    case T_INTEGER:
      switch (type(y))
	{
	case T_INTEGER:
	  return(make_boolean(sc, integer(x) >= integer(y)));

	case T_RATIO:
	  return(g_greater_or_equal(sc, set_plist_2(sc, x, y)));

	case T_REAL:
	  if (is_NaN(real(y))) return(sc->F);
	  return(make_boolean(sc, integer(x) >= real(y)));

	default:
	  method_or_bust(sc, y, sc->geq_symbol, list_2(sc, x, y), T_REAL, 2);
	}
      break;

    case T_RATIO:
      return(g_greater_or_equal(sc, set_plist_2(sc, x, y)));

    case T_REAL:
      switch (type(y))
	{
	case T_INTEGER:
	  if (is_NaN(real(x))) return(sc->F);
	  return(make_boolean(sc, real(x) >= integer(y)));

	case T_RATIO:
	  if (is_NaN(real(x))) return(sc->F);
	  return(make_boolean(sc, real(x) >= fraction(y)));

	case T_REAL:
	  if (is_NaN(real(x))) return(sc->F);
	  /* if (is_NaN(real(y))) return(sc->F); */
	  return(make_boolean(sc, real(x) >= real(y)));

	default:
	  method_or_bust(sc, y, sc->geq_symbol, list_2(sc, x, y), T_REAL, 2);
	}
      break;

    default:
      method_or_bust(sc, x, sc->geq_symbol, list_2(sc, x, y), T_REAL, 1);
    }
  return(sc->T);
}
static s7_pointer geq_p_pp(s7_pointer p1, s7_pointer p2) {return(c_geq_2(cur_sc, p1, p2));}

#endif

static s7_pointer geq_2 = NULL;

#if (!WITH_GMP)
static s7_pointer g_geq_2(s7_scheme *sc, s7_pointer args) {return(c_geq_2(sc, car(args), cadr(args)));}

static s7_pointer geq_s_fc;
static s7_pointer g_geq_s_fc(s7_scheme *sc, s7_pointer args)
{
  s7_double y;
  s7_pointer x;

  x = car(args);
  y = real(cadr(args));

  if (is_t_real(x))
    return(make_boolean(sc, real(x) >= y));
  return(g_geq_2(sc, args));
}

static s7_pointer geq_s_ic;
static s7_pointer g_geq_s_ic(s7_scheme *sc, s7_pointer args)
{
  s7_int y;
  s7_pointer x;

  x = car(args);
  y = s7_integer(cadr(args));

  if (type(x) == T_INTEGER)
    return(make_boolean(sc, integer(x) >= y));
  if (type(x) == T_REAL)
    return(make_boolean(sc, real(x) >= y));
  if (type(x) == T_RATIO)
    return(make_boolean(sc, !ratio_lt_pi(x, y)));
  method_or_bust(sc, x, sc->geq_symbol, args, T_REAL, 1);
}

static bool lt_b_pp(s7_pointer x, s7_pointer y) 
{
#if (!MS_WINDOWS)
  if (type(x) == type(y))
    {
      if (type(x) == T_INTEGER)
	return(integer(x) < integer(y));
      if (type(x) == T_REAL)
	return(real(x) < real(y));
      if (type(x) == T_RATIO)
	return(fraction(x) < fraction(y));
    }
#endif
  return(c_less_2(cur_sc, x, y) != cur_sc->F);
}

static bool leq_b_pp(s7_pointer x, s7_pointer y) 
{
#if (!MS_WINDOWS)
  if (type(x) == type(y))
    {
      if (type(x) == T_INTEGER)
	return(integer(x) <= integer(y));
      if (type(x) == T_REAL)
	return(real(x) <= real(y));
      if (type(x) == T_RATIO)
	return(fraction(x) <= fraction(y));
    }
#endif
  return(c_leq_2(cur_sc, x, y) != cur_sc->F);
}

static bool gt_b_pp(s7_pointer x, s7_pointer y) 
{
#if (!MS_WINDOWS)
  if (type(x) == type(y))
    {
      if (type(x) == T_INTEGER)
	return(integer(x) > integer(y));
      if (type(x) == T_REAL)
	return(real(x) > real(y));
      if (type(x) == T_RATIO)
	return(fraction(x) > fraction(y));
    }
#endif
  return(c_greater_2(cur_sc, x, y) != cur_sc->F);
}

static bool geq_b_pp(s7_pointer x, s7_pointer y) 
{
#if (!MS_WINDOWS)
  if (type(x) == type(y))
    {
      if (type(x) == T_INTEGER)
	return(integer(x) >= integer(y));
      if (type(x) == T_REAL)
	return(real(x) >= real(y));
      if (type(x) == T_RATIO)
	return(fraction(x) >= fraction(y));
    }
#endif
  return(c_geq_2(cur_sc, x, y) != cur_sc->F);
}

static bool req_b_pp(s7_pointer x, s7_pointer y) 
{
#if (!MS_WINDOWS)
  if (type(x) == type(y))
    {
      if (type(x) == T_INTEGER)
	return(integer(x) == integer(y));
      if (type(x) == T_REAL)
	return(real(x) == real(y));
      if (type(x) == T_RATIO)
	return(fraction(x) == fraction(y));
    }
#endif
  return(c_equal_2(cur_sc, x, y) != cur_sc->F);
}

static bool req_b_pi(s7_pointer i1, s7_int i2) {return(equal_b_pi(cur_sc, i1, i2));}

static bool lt_b_pi(s7_pointer p1, s7_int p2)
{
  if (is_integer(p1)) return(integer(p1) < p2);
  if (is_t_real(p1))  return(real(p1) < p2);
  if (is_t_ratio(p1)) return(ratio_lt_pi(p1, p2));
  simple_wrong_type_argument(cur_sc, cur_sc->lt_symbol, p1, T_REAL);
  return(false);
}

static s7_pointer lt_p_pi(s7_pointer p1, s7_int p2) {return(make_boolean(cur_sc, lt_b_pi(p1, p2)));}

static bool leq_b_pi(s7_pointer p1, s7_int p2) 
{
  if (is_integer(p1)) return(integer(p1) <= p2);
  if (is_t_real(p1))  return(real(p1) <= p2);
  if (is_t_ratio(p1)) return(ratio_leq_pi(p1, p2));
  simple_wrong_type_argument(cur_sc, cur_sc->leq_symbol, p1, T_REAL);
  return(false);
}

static s7_pointer leq_p_pi(s7_pointer p1, s7_int p2) {return(make_boolean(cur_sc, leq_b_pi(p1, p2)));}

static bool gt_b_pi(s7_pointer p1, s7_int p2) 
{
  if (is_integer(p1)) return(integer(p1) > p2);
  if (is_t_real(p1))  return(real(p1) > p2);
  if (is_t_ratio(p1)) return(!ratio_leq_pi(p1, p2));
  simple_wrong_type_argument(cur_sc, cur_sc->gt_symbol, p1, T_REAL);
  return(false);
}

static s7_pointer gt_p_pi(s7_pointer p1, s7_int p2) {return(make_boolean(cur_sc, gt_b_pi(p1, p2)));}

static bool geq_b_pi(s7_pointer p1, s7_int p2) 
{
  if (is_integer(p1)) return(integer(p1) >= p2);
  if (is_t_real(p1))  return(real(p1) >= p2);
  if (is_t_ratio(p1)) return(!ratio_lt_pi(p1, p2));
  simple_wrong_type_argument(cur_sc, cur_sc->geq_symbol, p1, T_REAL);
  return(false);
}

static s7_pointer geq_p_pi(s7_pointer p1, s7_int p2) {return(make_boolean(cur_sc, geq_b_pi(p1, p2)));}
#endif
/* end (!WITH_GMP) */

static bool req_b_ii(s7_int i1, s7_int i2) {return(i1 == i2);}
static bool lt_b_ii(s7_int i1, s7_int i2) {return(i1 < i2);}
static bool leq_b_ii(s7_int i1, s7_int i2) {return(i1 <= i2);}
static bool gt_b_ii(s7_int i1, s7_int i2) {return(i1 > i2);}
static bool geq_b_ii(s7_int i1, s7_int i2) {return(i1 >= i2);}
static bool req_b_dd(s7_double i1, s7_double i2) {return(i1 == i2);}
static bool lt_b_dd(s7_double i1, s7_double i2) {return(i1 < i2);}
static bool leq_b_dd(s7_double i1, s7_double i2) {return(i1 <= i2);}
static bool gt_b_dd(s7_double i1, s7_double i2) {return(i1 > i2);}
static bool geq_b_dd(s7_double i1, s7_double i2) {return(i1 >= i2);}


/* ---------------------------------------- real-part imag-part ---------------------------------------- */

s7_double s7_real_part(s7_pointer x)
{
  switch(type(x))
    {
    case T_INTEGER:     return((s7_double)integer(x));
    case T_RATIO:       return(fraction(x));
    case T_REAL:        return(real(x));
    case T_COMPLEX:     return(real_part(x));
#if WITH_GMP
    case T_BIG_INTEGER: return((s7_double)big_integer_to_s7_int(big_integer(x)));
    case T_BIG_RATIO:   return((s7_double)((long double)big_integer_to_s7_int(mpq_numref(big_ratio(x))) / (long double)big_integer_to_s7_int(mpq_denref(big_ratio(x)))));
    case T_BIG_REAL:    return((s7_double)mpfr_get_d(big_real(x), GMP_RNDN));
    case T_BIG_COMPLEX: return((s7_double)mpfr_get_d(mpc_realref(big_complex(x)), GMP_RNDN));
#endif
    }
  return(0.0);
}


s7_double s7_imag_part(s7_pointer x)
{
  switch (type(x))
    {
    case T_COMPLEX:     return(imag_part(x));
#if WITH_GMP
    case T_BIG_COMPLEX: return((s7_double)mpfr_get_d(mpc_imagref(big_complex(x)), GMP_RNDN));
#endif
    }
  return(0.0);
}

static s7_pointer g_real_part(s7_scheme *sc, s7_pointer args)
{
  #define H_real_part "(real-part num) returns the real part of num"
  #define Q_real_part s7_make_signature(sc, 2, sc->is_real_symbol, sc->is_number_symbol)  

  s7_pointer p;
  p = car(args);
  switch (type(p))
    {
    case T_INTEGER:
    case T_RATIO:
    case T_REAL:
      return(p);

    case T_COMPLEX:
      return(make_real(sc, real_part(p)));

#if WITH_GMP
    case T_BIG_INTEGER:
    case T_BIG_RATIO:
    case T_BIG_REAL:
      return(p);

    case T_BIG_COMPLEX:
      {
	s7_pointer x;

	new_cell(sc, x, T_BIG_REAL);
	add_bigreal(sc, x);
	mpfr_init(big_real(x));
	mpc_real(big_real(x), big_complex(p), GMP_RNDN);

	return(x);
      }
#endif

    default:
      method_or_bust_with_type_one_arg(sc, p, sc->real_part_symbol, args, a_number_string);
    }
}


static s7_pointer g_imag_part(s7_scheme *sc, s7_pointer args)
{
  #define H_imag_part "(imag-part num) returns the imaginary part of num"
  #define Q_imag_part s7_make_signature(sc, 2, sc->is_real_symbol, sc->is_number_symbol)  
  s7_pointer p;
  /* currently (imag-part nan.0) -> 0.0 ? it's true but maybe confusing */

  p = car(args);
  switch (type(p))
    {
    case T_INTEGER:
    case T_RATIO:
      return(small_int(0));

    case T_REAL:
      return(real_zero);

    case T_COMPLEX:
      return(make_real(sc, imag_part(p)));

#if WITH_GMP
    case T_BIG_INTEGER:
    case T_BIG_RATIO:
      return(small_int(0));

    case T_BIG_REAL:
      return(real_zero);

    case T_BIG_COMPLEX:
      {
	s7_pointer x;
	new_cell(sc, x, T_BIG_REAL);
	add_bigreal(sc, x);
	mpfr_init(big_real(x));
	mpc_imag(big_real(x), big_complex(p), GMP_RNDN);

	return(x);
      }
#endif

    default:
      method_or_bust_with_type_one_arg(sc, p, sc->imag_part_symbol, args, a_number_string);
    }
}

#if (!WITH_GMP)
static s7_double real_part_d_p(s7_pointer p)
{
  if (!is_number(p))
    simple_wrong_type_argument_with_type(cur_sc, cur_sc->real_part_symbol, p, a_number_string);
  return(s7_real_part(p));
}

static s7_double imag_part_d_p(s7_pointer p)
{
  if (!is_number(p))
    simple_wrong_type_argument_with_type(cur_sc, cur_sc->imag_part_symbol, p, a_number_string);
  return(s7_imag_part(p));
}
#endif


/* ---------------------------------------- numerator denominator ---------------------------------------- */

static s7_pointer g_numerator(s7_scheme *sc, s7_pointer args)
{
  #define H_numerator "(numerator rat) returns the numerator of the rational number rat"
  #define Q_numerator s7_make_signature(sc, 2, sc->is_integer_symbol, sc->is_rational_symbol)

  s7_pointer x;
  x = car(args);
  switch (type(x))
    {
    case T_RATIO:       return(make_integer(sc, numerator(x)));
    case T_INTEGER:     return(x);
#if WITH_GMP
    case T_BIG_INTEGER: return(x);
    case T_BIG_RATIO:   return(mpz_to_big_integer(sc, mpq_numref(big_ratio(x))));
#endif
    default:            method_or_bust_with_type_one_arg(sc, x, sc->numerator_symbol, args, a_rational_string);
    }
}

static s7_int numerator_i(s7_pointer p) 
{
  if (!is_rational(p)) 
    simple_wrong_type_argument(cur_sc, cur_sc->numerator_symbol, p, T_RATIO);
  return(numerator(p));
}


static s7_pointer g_denominator(s7_scheme *sc, s7_pointer args)
{
  #define H_denominator "(denominator rat) returns the denominator of the rational number rat"
  #define Q_denominator s7_make_signature(sc, 2, sc->is_integer_symbol, sc->is_rational_symbol)

  s7_pointer x;
  x = car(args);
  switch (type(x))
    {
    case T_RATIO:       return(make_integer(sc, denominator(x)));
    case T_INTEGER:     return(small_int(1));
#if WITH_GMP
    case T_BIG_INTEGER: return(small_int(1));
    case T_BIG_RATIO:   return(mpz_to_big_integer(sc, mpq_denref(big_ratio(x))));
#endif
    default:            method_or_bust_with_type_one_arg(sc, x, sc->denominator_symbol, args, a_rational_string);
    }
}

static s7_int denominator_i(s7_pointer p) 
{
  if (!is_rational(p)) 
    simple_wrong_type_argument(cur_sc, cur_sc->denominator_symbol, p, T_RATIO);
  if (is_integer(p))
    return(1);
  return(denominator(p));
}


/* ---------------------------------------- nan? infinite? ---------------------------------------- */

static s7_pointer g_is_nan(s7_scheme *sc, s7_pointer args)
{
  #define H_is_nan "(nan? obj) returns #t if obj is a NaN"
  #define Q_is_nan pl_bn  

  s7_pointer x;
  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:
    case T_RATIO:
      return(sc->F);

    case T_REAL:
      return(make_boolean(sc, is_NaN(real(x))));

    case T_COMPLEX:
      return(make_boolean(sc, (is_NaN(real_part(x))) || (is_NaN(imag_part(x)))));

#if WITH_GMP
    case T_BIG_INTEGER:
    case T_BIG_RATIO:
      return(sc->F);

    case T_BIG_REAL:
      return(make_boolean(sc, is_NaN(s7_real_part(x))));

    case T_BIG_COMPLEX:
      return(make_boolean(sc, (is_NaN(s7_real_part(x))) || (is_NaN(s7_imag_part(x)))));
#endif

    default:
      method_or_bust_with_type_one_arg(sc, x, sc->is_nan_symbol, list_1(sc, x), a_number_string);
    }
}

static bool is_nan_b(s7_pointer p) {return(g_is_nan(cur_sc, set_plist_1(cur_sc, p)) != cur_sc->F);}


static s7_pointer g_is_infinite(s7_scheme *sc, s7_pointer args)
{
  #define H_is_infinite "(infinite? obj) returns #t if obj is an infinite real"
  #define Q_is_infinite pl_bn  

  s7_pointer x;
  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:
    case T_RATIO:
      return(sc->F);

    case T_REAL:
      return(make_boolean(sc, is_inf(real(x))));

    case T_COMPLEX:
      return(make_boolean(sc, (is_inf(real_part(x))) || (is_inf(imag_part(x)))));

#if WITH_GMP
    case T_BIG_INTEGER:
    case T_BIG_RATIO:
      return(sc->F);

    case T_BIG_REAL:
      return(make_boolean(sc, mpfr_inf_p(big_real(x)) != 0));

    case T_BIG_COMPLEX:
      return(make_boolean(sc,
			  (mpfr_inf_p(big_real(g_real_part(sc, list_1(sc, x)))) != 0) ||
			  (mpfr_inf_p(big_real(g_imag_part(sc, list_1(sc, x)))) != 0)));
#endif

    default:
      method_or_bust_with_type_one_arg(sc, x, sc->is_infinite_symbol, list_1(sc, x), a_number_string);
    }
}

static bool is_infinite_b(s7_pointer p) {return(g_is_infinite(cur_sc, set_plist_1(cur_sc, p)) != cur_sc->F);}


/* ---------------------------------------- number? complex? integer? rational? real?  ---------------------------------------- */

static s7_pointer g_is_number(s7_scheme *sc, s7_pointer args)
{
  #define H_is_number "(number? obj) returns #t if obj is a number"
  #define Q_is_number pl_bt  
  check_boolean_method(sc, s7_is_number, sc->is_number_symbol, args); /* we need the s7_* versions here for the GMP case */
}


static s7_pointer g_is_integer(s7_scheme *sc, s7_pointer args)
{
  #define H_is_integer "(integer? obj) returns #t if obj is an integer"
  #define Q_is_integer pl_bt  
  check_boolean_method(sc, s7_is_integer, sc->is_integer_symbol, args);
}


static s7_pointer g_is_real(s7_scheme *sc, s7_pointer args)
{
  #define H_is_real "(real? obj) returns #t if obj is a real number"
  #define Q_is_real pl_bt  
  check_boolean_method(sc, s7_is_real, sc->is_real_symbol, args);
}


static s7_pointer g_is_complex(s7_scheme *sc, s7_pointer args)
{
  #define H_is_complex "(complex? obj) returns #t if obj is a number"
  #define Q_is_complex pl_bt  
  check_boolean_method(sc, s7_is_number, sc->is_complex_symbol, args);
}


static s7_pointer g_is_rational(s7_scheme *sc, s7_pointer args)
{
  #define H_is_rational "(rational? obj) returns #t if obj is a rational number (either an integer or a ratio)"
  #define Q_is_rational pl_bt  
  check_boolean_method(sc, s7_is_rational, sc->is_rational_symbol, args);
  /* in the non-gmp case, (rational? 455702434782048082459/86885567283849955830) -> #f, not #t
   *  and similarly for exact? etc.
   */
}


/* ---------------------------------------- even? odd?---------------------------------------- */

static s7_pointer g_is_even(s7_scheme *sc, s7_pointer args)
{
  #define H_is_even "(even? int) returns #t if the integer int32_t is even"
  #define Q_is_even s7_make_signature(sc, 2, sc->is_boolean_symbol, sc->is_integer_symbol)

  s7_pointer p;
  p = car(args);
  switch (type(p))
    {
    case T_INTEGER:     return(make_boolean(sc, ((integer(p) & 1) == 0)));
#if WITH_GMP
    case T_BIG_INTEGER: return(make_boolean(sc, mpz_even_p(big_integer(p))));
#endif
    default:            method_or_bust_one_arg(sc, p, sc->is_even_symbol, list_1(sc, p), T_INTEGER);
    }
}

static bool is_even_b(s7_pointer p)
{
  if (!s7_is_integer(p))
    simple_wrong_type_argument(cur_sc, cur_sc->is_even_symbol, p, T_INTEGER);
  return((integer(p) & 1) == 0);
}

static bool is_even_i(s7_int i1) {return((i1 & 1) == 0);}


static s7_pointer g_is_odd(s7_scheme *sc, s7_pointer args)
{
  #define H_is_odd "(odd? int) returns #t if the integer int32_t is odd"
  #define Q_is_odd s7_make_signature(sc, 2, sc->is_boolean_symbol, sc->is_integer_symbol)

  s7_pointer p;
  p = car(args);
  switch (type(p))
    {
    case T_INTEGER:     return(make_boolean(sc, ((integer(p) & 1) == 1)));
#if WITH_GMP
    case T_BIG_INTEGER: return(make_boolean(sc, mpz_odd_p(big_integer(p))));
#endif
    default:            method_or_bust_one_arg(sc, p, sc->is_odd_symbol, list_1(sc, p), T_INTEGER);
    }
}

static bool is_odd_b(s7_pointer p)
{
  if (!s7_is_integer(p))
    simple_wrong_type_argument(cur_sc, cur_sc->is_odd_symbol, p, T_INTEGER);
  return((integer(p) & 1) == 1);
}

static bool is_odd_i(s7_int i1) {return((i1 & 1) == 1);}



/* ---------------------------------------- zero? ---------------------------------------- */

static s7_pointer g_is_zero(s7_scheme *sc, s7_pointer args)
{
  #define H_is_zero "(zero? num) returns #t if the number num is zero"
  #define Q_is_zero pl_bn
  s7_pointer x;
  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:     return(make_boolean(sc, integer(x) == 0));
    case T_REAL:        return(make_boolean(sc, real(x) == 0.0));
    case T_RATIO:
    case T_COMPLEX:     return(sc->F);      /* ratios and complex numbers are already collapsed into integers and reals */
#if WITH_GMP
    case T_BIG_INTEGER: return(make_boolean(sc, mpz_cmp_ui(big_integer(x), 0) == 0));
    case T_BIG_REAL:    return(make_boolean(sc, mpfr_zero_p(big_real(x))));
    case T_BIG_RATIO:
    case T_BIG_COMPLEX: return(sc->F);
#endif
    default:
      method_or_bust_with_type_one_arg(sc, x, sc->is_zero_symbol, list_1(sc, x), a_number_string);
    }
}

static bool is_zero_b(s7_pointer p)
{
  if (!is_number(p))
    simple_wrong_type_argument_with_type(cur_sc, cur_sc->is_zero_symbol, p, a_number_string);
  if (is_t_integer(p))
    return(integer(p) == 0);
  if (is_t_real(p))
    return(real(p) == 0.0);
  return(false);
}

static bool is_zero_i(s7_int p) {return(p == 0);}
static bool is_zero_d(s7_double p) {return(p == 0.0);}


/* -------------------------------- positive? -------------------------------- */

static s7_pointer g_is_positive(s7_scheme *sc, s7_pointer args)
{
  #define H_is_positive "(positive? num) returns #t if the real number num is positive (greater than 0)"
  #define Q_is_positive s7_make_signature(sc, 2, sc->is_boolean_symbol, sc->is_real_symbol)
  s7_pointer x;
  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:     return(make_boolean(sc, integer(x) > 0));
    case T_RATIO:       return(make_boolean(sc, numerator(x) > 0));
    case T_REAL:        return(make_boolean(sc, real(x) > 0.0));
#if WITH_GMP
    case T_BIG_INTEGER: return(make_boolean(sc, (mpz_cmp_ui(big_integer(x), 0) > 0)));
    case T_BIG_RATIO:   return(make_boolean(sc, (mpq_cmp_ui(big_ratio(x), 0, 1) > 0)));
    case T_BIG_REAL:    return(make_boolean(sc, (mpfr_cmp_ui(big_real(x), 0) > 0)));
#endif
    default:
      method_or_bust_one_arg(sc, x, sc->is_positive_symbol, list_1(sc, x), T_REAL);
    }
}

static bool is_positive_b(s7_pointer p)
{
  if (!is_real(p))
    simple_wrong_type_argument(cur_sc, cur_sc->is_positive_symbol, p, T_REAL);
  if (is_t_integer(p))
    return(integer(p) > 0);
  if (is_t_real(p))
    return(real(p) > 0.0);
  return(numerator(p) > 0);
}

static bool is_positive_i(s7_int p) {return(p > 0);}
static bool is_positive_d(s7_double p) {return(p > 0.0);}


/* -------------------------------- negative? -------------------------------- */

static s7_pointer g_is_negative(s7_scheme *sc, s7_pointer args)
{
  #define H_is_negative "(negative? num) returns #t if the real number num is negative (less than 0)"
  #define Q_is_negative s7_make_signature(sc, 2, sc->is_boolean_symbol, sc->is_real_symbol)
  s7_pointer x;
  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:     return(make_boolean(sc, integer(x) < 0));
    case T_RATIO:       return(make_boolean(sc, numerator(x) < 0));
    case T_REAL:        return(make_boolean(sc, real(x) < 0.0));
#if WITH_GMP
    case T_BIG_INTEGER: return(make_boolean(sc, (mpz_cmp_ui(big_integer(x), 0) < 0)));
    case T_BIG_RATIO:   return(make_boolean(sc, (mpq_cmp_ui(big_ratio(x), 0, 1) < 0)));
    case T_BIG_REAL:    return(make_boolean(sc, (mpfr_cmp_ui(big_real(x), 0) < 0)));
#endif
    default:
      method_or_bust_one_arg(sc, x, sc->is_negative_symbol, list_1(sc, x), T_REAL);
    }
}

static bool is_negative_b(s7_pointer p)
{
  if (!is_real(p))
    simple_wrong_type_argument(cur_sc, cur_sc->is_negative_symbol, p, T_REAL);
  if (is_t_integer(p))
    return(integer(p) < 0);
  if (is_t_real(p))
    return(real(p) < 0.0);
  return(numerator(p) < 0);
}

static bool is_negative_i(s7_int p) {return(p < 0);}
static bool is_negative_d(s7_double p) {return(p < 0.0);}


bool s7_is_ulong(s7_pointer arg) {return(is_integer(arg));}
unsigned long s7_ulong(s7_pointer p)  {return((_NFre(p))->object.number.ul_value);}

s7_pointer s7_make_ulong(s7_scheme *sc, unsigned long n)
{
  s7_pointer x;
  new_cell(sc, x, T_INTEGER);
  x->object.number.ul_value = n;
  return(x);
}

bool s7_is_ulong_long(s7_pointer arg) {return(is_integer(arg));}
uint64_t s7_ulong_long(s7_pointer p)  {return((_NFre(p))->object.number.ull_value);}

s7_pointer s7_make_ulong_long(s7_scheme *sc, uint64_t n)
{
  s7_pointer x;
  new_cell(sc, x, T_INTEGER);
  x->object.number.ull_value = n;
  return(x);
}


#if (!WITH_PURE_S7)
#if (!WITH_GMP)
/* ---------------------------------------- exact<->inexact exact? inexact? ---------------------------------------- */

static s7_pointer g_exact_to_inexact(s7_scheme *sc, s7_pointer args)
{
  #define H_exact_to_inexact "(exact->inexact num) converts num to an inexact number; (exact->inexact 3/2) = 1.5"
  #define Q_exact_to_inexact pcl_r
  return(exact_to_inexact(sc, car(args)));
}

static s7_pointer g_inexact_to_exact(s7_scheme *sc, s7_pointer args)
{
  #define H_inexact_to_exact "(inexact->exact num) converts num to an exact number; (inexact->exact 1.5) = 3/2"
  #define Q_inexact_to_exact s7_make_signature(sc, 2, sc->is_rational_symbol, sc->is_real_symbol)
  return(inexact_to_exact(sc, car(args), WITH_OVERFLOW_ERROR));
}
#endif
/* (!WITH_GMP) */


static s7_pointer g_is_exact(s7_scheme *sc, s7_pointer args)
{
  #define H_is_exact "(exact? num) returns #t if num is exact (an integer or a ratio)"
  #define Q_is_exact pl_bn

  s7_pointer x;
  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:
    case T_RATIO:       return(sc->T);
    case T_REAL:
    case T_COMPLEX:     return(sc->F);
#if WITH_GMP
    case T_BIG_INTEGER:
    case T_BIG_RATIO:   return(sc->T);
    case T_BIG_REAL:
    case T_BIG_COMPLEX: return(sc->F);
#endif
    default:
      method_or_bust_with_type_one_arg(sc, x, sc->is_exact_symbol, args, a_number_string);
    }
}

static bool is_exact_b(s7_pointer p) 
{
  if (!is_number(p))
    simple_wrong_type_argument_with_type(cur_sc, cur_sc->is_exact_symbol, p, a_number_string);
  return(is_rational(p));
}


static s7_pointer g_is_inexact(s7_scheme *sc, s7_pointer args)
{
  #define H_is_inexact "(inexact? num) returns #t if num is inexact (neither an integer nor a ratio)"
  #define Q_is_inexact pl_bn

  s7_pointer x;
  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:
    case T_RATIO:       return(sc->F);
    case T_REAL:
    case T_COMPLEX:     return(sc->T);
#if WITH_GMP
    case T_BIG_INTEGER:
    case T_BIG_RATIO:   return(sc->F);
    case T_BIG_REAL:
    case T_BIG_COMPLEX: return(sc->T);
#endif
    default:
      method_or_bust_with_type_one_arg(sc, x, sc->is_inexact_symbol, args, a_number_string);
    }
}

static bool is_inexact_b(s7_pointer p) 
{
  if (!is_number(p))
    simple_wrong_type_argument_with_type(cur_sc, cur_sc->is_inexact_symbol, p, a_number_string);
  return(!is_rational(p));
}


/* ---------------------------------------- integer-length, integer-decode-float ---------------------------------------- */

static int32_t integer_length(s7_int a)
{
  static const int32_t bits[256] =
    {0, 1, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8};

  #define I_8 256LL
  #define I_16 65536LL
  #define I_24 16777216LL
  #define I_32 4294967296LL
  #define I_40 1099511627776LL
  #define I_48 281474976710656LL
  #define I_56 72057594037927936LL

  /* a might be most-negative-fixnum! in Clisp: (integer-length -9223372036854775808) -> 63
   */
  if (a < 0)
    {
      if (a == s7_int_min) return(63);
      a = -a;
    }
  if (a < I_8) return(bits[a]);
  if (a < I_16) return(8 + bits[a >> 8]);
  if (a < I_24) return(16 + bits[a >> 16]);
  if (a < I_32) return(24 + bits[a >> 24]);
  if (a < I_40) return(32 + bits[a >> 32]);
  if (a < I_48) return(40 + bits[a >> 40]);
  if (a < I_56) return(48 + bits[a >> 48]);
  return(56 + bits[a >> 56]);
}

static s7_pointer g_integer_length(s7_scheme *sc, s7_pointer args)
{
  #define H_integer_length "(integer-length arg) returns the number of bits required to represent the integer 'arg': (ceiling (log (abs arg) 2))"
  #define Q_integer_length pcl_i

  s7_int x;
  s7_pointer p;

  p = car(args);
  if (!s7_is_integer(p))
    method_or_bust_one_arg(sc, p, sc->integer_length_symbol, args, T_INTEGER);


  x = s7_integer(p);
  if (x < 0)
    return(make_integer(sc, integer_length(-(x + 1))));
  return(make_integer(sc, integer_length(x)));
}

static s7_int integer_length_i_i(s7_int x)
{
  if (x < 0)
    return(integer_length(-(x + 1)));
  return(integer_length(x));
}
#endif /* !pure s7 */


static s7_pointer g_integer_decode_float(s7_scheme *sc, s7_pointer args)
{
  #define H_integer_decode_float "(integer-decode-float x) returns a list containing the significand, exponent, and \
sign of 'x' (1 = positive, -1 = negative).  (integer-decode-float 0.0): (0 0 1)"
  #define Q_integer_decode_float s7_make_signature(sc, 2, sc->is_pair_symbol, sc->is_float_symbol)

  /* no matter what s7_double is, integer-decode-float acts as if x is a C double */

  typedef struct decode_float_t {
    union {
      int64_t ix;
      double fx;
    } value;
  } decode_float_t;

 decode_float_t num;
  s7_pointer x;
  x = car(args);

  switch (type(x))
    {
    case T_REAL:
      num.value.fx = (double)real(x);
      break;

#if WITH_GMP
    case T_BIG_REAL:
      num.value.fx = (double)real_to_double(sc, x, "integer-decode-float");
      break;
#endif

    default:
      method_or_bust_with_type_one_arg(sc, x, sc->integer_decode_float_symbol, args, s7_make_string_wrapper(sc, "a non-rational real"));
    }

  if (num.value.fx == 0.0)
    return(list_3(sc, small_int(0), small_int(0), small_int(1)));

  return(list_3(sc,
		make_integer(sc, (s7_int)((num.value.ix & 0xfffffffffffffLL) | 0x10000000000000LL)),
		make_integer(sc, (s7_int)(((num.value.ix & 0x7fffffffffffffffLL) >> 52) - 1023 - 52)),
		make_integer(sc, ((num.value.ix & 0x8000000000000000LL) != 0) ? -1 : 1)));
}


/* -------------------------------- logior -------------------------------- */
static s7_pointer g_logior(s7_scheme *sc, s7_pointer args)
{
  #define H_logior "(logior int32_t ...) returns the bitwise OR of its integer arguments (the bits that are on in any of the arguments)"
  #define Q_logior pcl_i
  s7_int result = 0;
  s7_pointer x;

  for (x = args; is_not_null(x); x = cdr(x))
    {
      if (!s7_is_integer(car(x)))
	method_or_bust(sc, car(x), sc->logior_symbol, cons(sc, make_integer(sc, result), x), T_INTEGER, position_of(x, args));
      result |= s7_integer(car(x));
    }
  return(make_integer(sc, result));
}

static s7_int logior_i_ii(s7_int i1, s7_int i2) {return(i1 | i2);}


/* -------------------------------- logxor -------------------------------- */
static s7_pointer g_logxor(s7_scheme *sc, s7_pointer args)
{
  #define H_logxor "(logxor int32_t ...) returns the bitwise XOR of its integer arguments (the bits that are on in an odd number of the arguments)"
  #define Q_logxor pcl_i
  s7_int result = 0;
  s7_pointer x;

  for (x = args; is_not_null(x); x = cdr(x))
    {
      if (!s7_is_integer(car(x)))
	method_or_bust(sc, car(x), sc->logxor_symbol, cons(sc, make_integer(sc, result), x), T_INTEGER, position_of(x, args));
      result ^= s7_integer(car(x));
    }
  return(make_integer(sc, result));
}

static s7_int logxor_i_ii(s7_int i1, s7_int i2) {return(i1 ^ i2);}


/* -------------------------------- logand -------------------------------- */
static s7_pointer g_logand(s7_scheme *sc, s7_pointer args)
{
  #define H_logand "(logand int32_t ...) returns the bitwise AND of its integer arguments (the bits that are on in every argument)"
  #define Q_logand pcl_i
  s7_int result = -1;
  s7_pointer x;

  for (x = args; is_not_null(x); x = cdr(x))
    {
      if (!s7_is_integer(car(x)))
	method_or_bust(sc, car(x), sc->logand_symbol, cons(sc, make_integer(sc, result), x), T_INTEGER, position_of(x, args));
      result &= s7_integer(car(x));
    }
  return(make_integer(sc, result));
}

static s7_int logand_i_ii(s7_int i1, s7_int i2) {return(i1 & i2);}


/* -------------------------------- lognot -------------------------------- */

static s7_pointer g_lognot(s7_scheme *sc, s7_pointer args)
{
  #define H_lognot "(lognot num) returns the bitwise negation (the complement, the bits that are not on) in num: (lognot 0) -> -1"
  #define Q_lognot pcl_i
  if (!s7_is_integer(car(args)))
    method_or_bust_one_arg(sc, car(args), sc->lognot_symbol, args, T_INTEGER);
  return(make_integer(sc, ~s7_integer(car(args))));
}

static s7_int lognot_i_i(s7_int i1) {return(~i1);}


/* -------------------------------- logbit? -------------------------------- */
/* logbit?  CL is (logbitp index int) using 2^index, but that order strikes me as backwards
 *   at least gmp got the arg order right!
 */

static s7_pointer g_logbit(s7_scheme *sc, s7_pointer args)
{
  #define H_logbit "(logbit? int32_t index) returns #t if the index-th bit is on in int, otherwise #f. The argument \
order here follows gmp, and is the opposite of the CL convention.  (logbit? int32_t bit) is the same as (not (zero? (logand int32_t (ash 1 bit))))."
  #define Q_logbit s7_make_circular_signature(sc, 1, 2, sc->is_boolean_symbol, sc->is_integer_symbol)

  s7_pointer x, y;
  s7_int index;      /* index in gmp is mp_bitcnt which is an unsigned long int32_t */

  x = car(args);
  y = cadr(args);

  if (!s7_is_integer(x))
    method_or_bust(sc, x, sc->logbit_symbol, args, T_INTEGER, 1);
  if (!s7_is_integer(y))
    method_or_bust(sc, y, sc->logbit_symbol, args, T_INTEGER, 2);

  index = s7_integer(y);
  if (index < 0)
    return(out_of_range(sc, sc->logbit_symbol, small_int(2), y, its_negative_string));

#if WITH_GMP
  if (is_t_big_integer(x))
    return(make_boolean(sc, (mpz_tstbit(big_integer(x), index) != 0)));
#endif

  if (index >= s7_int_bits)           /* not sure about the >: (logbit? -1 64) ?? */
    return(make_boolean(sc, integer(x) < 0));

  /* :(zero? (logand most-positive-fixnum (ash 1 63)))
   *   -> ash argument 2, 63, is out of range (shift is too large)
   *   so logbit? has a wider range than the logand/ash shuffle above.
   */

  /* all these int64_ts are necessary, else C turns it into an int, gets confused about signs etc */
  return(make_boolean(sc, ((((int64_t)(1LL << (int64_t)index)) & (int64_t)integer(x)) != 0)));
}

static bool logbit_b_ii(s7_int i1, s7_int i2) 
{
  if (i2 < 0)
    simple_out_of_range(cur_sc, cur_sc->logbit_symbol, make_integer(cur_sc, i2), its_negative_string);
  if (i2 >= s7_int_bits)
    return(i1 < 0);
  return((((int64_t)(1LL << (int64_t)i2)) & (int64_t)i1) != 0);
}

#if (!WITH_GMP)
static bool logbit_b_pp(s7_pointer p1, s7_pointer p2)
{
  if (is_integer(p1))
    {
      if (is_integer(p2))
	return(logbit_b_ii(integer(p1), integer(p2)));
      simple_wrong_type_argument(cur_sc, cur_sc->logbit_symbol, p2, T_INTEGER);
    }
  simple_wrong_type_argument(cur_sc, cur_sc->logbit_symbol, p1, T_INTEGER);
  return(false);
}
#endif


/* -------------------------------- ash -------------------------------- */
static s7_int c_ash(s7_scheme *sc, s7_int arg1, s7_int arg2)
{
  if (arg1 == 0) return(0);

  if (arg2 >= s7_int_bits)
    out_of_range(sc, sc->ash_symbol, small_int(2), make_integer(sc, arg2), its_too_large_string);

  if (arg2 < -s7_int_bits)
    {
      if (arg1 < 0)                      /* (ash -31 -100) */
	return(-1);
      return(0);
    }

  /* I can't see any point in protecting this: (ash 9223372036854775807 1) -> -2, but anyone using ash must know something about bits */
  if (arg2 >= 0)
    {
      if (arg1 < 0)
	{
	  uint64_t z;
	  z = (uint64_t)arg1;
	  return((s7_int)(z << arg2));
	}
      return(arg1 << arg2);
    }
  return(arg1 >> -arg2);
}

static s7_pointer g_ash(s7_scheme *sc, s7_pointer args)
{
  #define H_ash "(ash i1 i2) returns i1 shifted right or left i2 times, i1 << i2, (ash 1 3) -> 8, (ash 8 -3) -> 1"
  #define Q_ash pcl_i
  s7_pointer x, y;

  x = car(args);
  if (!s7_is_integer(x))
    method_or_bust(sc, x, sc->ash_symbol, args, T_INTEGER, 1);

  y = cadr(args);
  if (!s7_is_integer(y))
    method_or_bust(sc, y, sc->ash_symbol, args, T_INTEGER, 2);

  return(make_integer(sc, c_ash(sc, s7_integer(x), s7_integer(y))));
}

static s7_int ash_i_ii(s7_int i1, s7_int i2) {return(c_ash(cur_sc, i1, i2));}
#if (!WITH_GMP)
static s7_int rsh_i_ii_direct(s7_int i1, s7_int i2) {return(i1 >> (-i2));}
static s7_int lsh_i_ii_direct(s7_int i1, s7_int i2) {return(i1 << i2);}
static s7_int rsh_i_i2_direct(s7_int i1, s7_int i2) {return(i1 >> 1);}
#endif


/* ---------------------------------------- random ---------------------------------------- */

/* random numbers.  The simple version used in clm.c is probably adequate,
 *   but here I'll use Marsaglia's MWC algorithm.
 *     (random num) -> a number (0..num), if num == 0 return 0, use global default state
 *     (random num state) -> same but use this state
 *     (random-state seed) -> make a new state
 *   to save the current seed, use copy
 *   to save it across load, random-state->list and list->random-state.
 *   random-state? returns #t if its arg is one of these guys
 */

#if (!WITH_GMP)
s7_pointer s7_random_state(s7_scheme *sc, s7_pointer args)
{
  #define H_random_state "(random-state seed (carry plausible-default)) returns a new random number state initialized with 'seed'. \
Pass this as the second argument to 'random' to get a repeatable random number sequence:\n\
    (let ((seed (random-state 1234))) (random 1.0 seed))"
  #define Q_random_state s7_make_circular_signature(sc, 1, 2, sc->is_random_state_symbol, sc->is_integer_symbol)

  s7_pointer r1, r2, p;
  s7_int i1, i2;

  r1 = car(args);
  if (!s7_is_integer(r1))
    method_or_bust(sc, r1, sc->random_state_symbol, args, T_INTEGER, 1);
  i1 = s7_integer(r1);
  if (i1 < 0)
    return(out_of_range(sc, sc->random_state_symbol, small_int(1), r1, its_negative_string));

  if (is_null(cdr(args)))
    {
      new_cell(sc, p, T_RANDOM_STATE);
      random_seed(p) = (uint64_t)i1;
      random_carry(p) = 1675393560;                          /* should this be dependent on the seed? */
      return(p);
    }

  r2 = cadr(args);
  if (!s7_is_integer(r2))
    method_or_bust(sc, r2, sc->random_state_symbol, args, T_INTEGER, 2);
  i2 = s7_integer(r2);
  if (i2 < 0)
    return(out_of_range(sc, sc->random_state_symbol, small_int(2), r2, its_negative_string));

  new_cell(sc, p, T_RANDOM_STATE);
  random_seed(p) = (uint64_t)i1;
  random_carry(p) = (uint64_t)i2;
  return(p);
}

#define g_random_state s7_random_state
#endif

static s7_pointer rng_copy(s7_scheme *sc, s7_pointer args)
{
#if WITH_GMP
  return(sc->F); /* I can't find a way to copy a gmp random generator */
#else
  s7_pointer obj;
  obj = car(args);
  if (is_random_state(obj))
    {
      s7_pointer new_r;
      new_cell(sc, new_r, T_RANDOM_STATE);
      random_seed(new_r) = random_seed(obj);
      random_carry(new_r) = random_carry(obj);
      return(new_r);
    }
  return(sc->F);
#endif
}


static s7_pointer g_is_random_state(s7_scheme *sc, s7_pointer args)
{
  #define H_is_random_state "(random-state? obj) returns #t if obj is a random-state object (from random-state)."
  #define Q_is_random_state pl_bt
  check_boolean_method(sc, is_random_state, sc->is_random_state_symbol, args);
}

static bool is_random_state_b(s7_pointer p) {return(type(p) == T_RANDOM_STATE);}

s7_pointer s7_random_state_to_list(s7_scheme *sc, s7_pointer args)
{
  #define H_random_state_to_list "(random-state->list r) returns the random state object as a list.\
You can later apply random-state to this list to continue a random number sequence from any point."
  #define Q_random_state_to_list s7_make_signature(sc, 2, sc->is_pair_symbol, sc->is_random_state_symbol)

#if WITH_GMP
  if ((is_pair(args)) &&
      (!is_random_state(car(args))))
    method_or_bust_with_type(sc, car(args), sc->random_state_to_list_symbol, args, a_random_state_object_string, 1);
  return(sc->nil);
#else
  s7_pointer r;
  if (is_null(args))
    r = sc->default_rng;
  else
    {
      r = car(args);
      if (!is_random_state(r))
	method_or_bust_with_type(sc, r, sc->random_state_to_list_symbol, args, a_random_state_object_string, 1);
    }
  return(list_2(sc, make_integer(sc, random_seed(r)), make_integer(sc, random_carry(r))));
#endif
}

#define g_random_state_to_list s7_random_state_to_list

void s7_set_default_random_state(s7_scheme *sc, s7_int seed, s7_int carry)
{
#if (!WITH_GMP)
  s7_pointer p;
  new_cell(sc, p, T_RANDOM_STATE);
  random_seed(p) = (uint64_t)seed;
  random_carry(p) = (uint64_t)carry;
  sc->default_rng = p;
#endif
}

#if (!WITH_GMP)
/* -------------------------------- random -------------------------------- */

static double next_random(s7_pointer r)
{
  /* The multiply-with-carry generator for 32-bit integers:
   *        x(n)=a*x(n-1) + carry mod 2^32
   * Choose multiplier a from this list:
   *   1791398085 1929682203 1683268614 1965537969 1675393560
   *   1967773755 1517746329 1447497129 1655692410 1606218150
   *   2051013963 1075433238 1557985959 1781943330 1893513180
   *   1631296680 2131995753 2083801278 1873196400 1554115554
   * ( or any 'a' for which both a*2^32-1 and a*2^31-1 are prime)
   */
  double result;
  uint64_t temp;
  #define RAN_MULT 2131995753UL

  temp = random_seed(r) * RAN_MULT + random_carry(r);
  random_seed(r) = (temp & 0xffffffffUL);
  random_carry(r) = (temp >> 32);
  result = (double)((uint32_t)(random_seed(r))) / 4294967295.5;
  /* divisor was 2^32-1 = 4294967295.0, but somehow this can round up once in a billion tries?
   *   do we want the double just less than 2^32?
   */

  /* (let ((mx 0) (mn 1000)) (do ((i 0 (+ i 1))) ((= i 10000)) (let ((val (random 123))) (set! mx (max mx val)) (set! mn (min mn val)))) (list mn mx)) */
  return(result);
}


s7_double s7_random(s7_scheme *sc, s7_pointer state)
{
  if (!state)
    return(next_random(sc->default_rng));
  return(next_random(state));
}


static s7_pointer g_random(s7_scheme *sc, s7_pointer args)
{
  #define H_random "(random num (state #f)) returns a random number between 0 and num (0 if num=0)."
  #define Q_random s7_make_signature(sc, 3, sc->is_number_symbol, sc->is_number_symbol, sc->is_random_state_symbol)
  s7_pointer r, num;

  num = car(args);
  if (!s7_is_number(num))
    method_or_bust_with_type(sc, num, sc->random_symbol, args, a_number_string, 1);

  if (is_not_null(cdr(args)))
    {
      r = cadr(args);
      if (!is_random_state(r))
	method_or_bust_with_type(sc, r, sc->random_symbol, args, a_random_state_object_string, 2);
    }
  else r = sc->default_rng;

  switch (type(num))
    {
    case T_INTEGER:
      return(make_integer(sc, (s7_int)(integer(num) * next_random(r))));

    case T_RATIO:
      {
	s7_double x, error;
	s7_int numer = 0, denom = 1;
	/* the error here needs to take the size of the fraction into account.  Otherwise, if
	 *    error is (say) 1e-6 and the fraction is (say) 9000000/9223372036854775807,
	 *    c_rationalize will always return 0.  But even that isn't foolproof:
	 *    (random 1/562949953421312) -> 1/376367230475000
	 */
	x = fraction(num);
	if ((x < 1.0e-10) && (x > -1.0e-10))
	  {
	    /* 1e-12 is not tight enough:
	     *    (random 1/2251799813685248) -> 1/2250240579436280
	     *    (random -1/4503599627370496) -> -1/4492889778435526
	     *    (random 1/140737488355328) -> 1/140730223985746
	     *    (random -1/35184372088832) -> -1/35183145492420
	     *    (random -1/70368744177664) -> -1/70366866392738
	     *    (random 1/4398046511104) -> 1/4398033095756
	     *    (random 1/137438953472) -> 1/137438941127
	     */
	    if (numerator(num) < -10)
	      numer = -(s7_int)(floor(-numerator(num) * next_random(r)));
	    else
	      {
		if (numerator(num) > 10)
		  numer = (s7_int)floor(numerator(num) * next_random(r));
		else
		  {
		    int64_t diff;
		    numer = numerator(num);
		    diff = s7_int_max - denominator(num);
		    if (diff < 100)
		      return(s7_make_ratio(sc, numer, denominator(num)));
		    denom = denominator(num) + (s7_int)floor(diff * next_random(r));
		    return(s7_make_ratio(sc, numer, denom));
		  }
	      }
	    return(s7_make_ratio(sc, numer, denominator(num)));
	  }
	if ((x < 1e-6) && (x > -1e-6))
	  error = 1e-18;
	else error = 1e-12;
	c_rationalize(x * next_random(r), error, &numer, &denom);
	return(s7_make_ratio(sc, numer, denom));
      }

    case T_REAL:
      return(make_real(sc, real(num) * next_random(r)));

    case T_COMPLEX:
      return(s7_make_complex(sc, real_part(num) * next_random(r), imag_part(num) * next_random(r)));
    }
  return(sc->F);
}

static s7_double random_d_d(s7_double x)
{
  return(x * next_random(cur_sc->default_rng));
}

static s7_int random_i_i(s7_int i)
{
  return((s7_int)(i * next_random(cur_sc->default_rng)));
}

static s7_pointer random_p_p(s7_pointer p)
{
  return(g_random(cur_sc, set_plist_1(cur_sc, p)));
}

static s7_pointer random_ic, random_rc;

static s7_pointer g_random_ic(s7_scheme *sc, s7_pointer args)
{
  return(make_integer(sc, (s7_int)(integer(car(args)) * next_random(sc->default_rng))));
}

static s7_pointer g_random_rc(s7_scheme *sc, s7_pointer args)
{
  return(make_real(sc, real(car(args)) * next_random(sc->default_rng)));
}

static s7_pointer random_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
  if (!ops) return(f);
  if (args == 1)
    {
      s7_pointer arg1;
      arg1 = cadr(expr);
      if (s7_is_integer(arg1))
	{
	  set_optimize_op(expr, HOP_SAFE_C_C);
	  return(random_ic);
	}
      if (is_float(arg1))
	{
	  set_optimize_op(expr, HOP_SAFE_C_C);
	  return(random_rc);
	}
    }
  return(f);
}
#endif /* gmp */



/* -------------------------------- characters -------------------------------- */

#define NUM_CHARS 256

static s7_pointer g_char_to_integer(s7_scheme *sc, s7_pointer args)
{
  #define H_char_to_integer "(char->integer c) converts the character c to an integer"
  #define Q_char_to_integer s7_make_signature(sc, 2, sc->is_integer_symbol, sc->is_char_symbol)

  if (!s7_is_character(car(args)))
    method_or_bust_one_arg(sc, car(args), sc->char_to_integer_symbol, args, T_CHARACTER);
  return(small_int(character(car(args))));
}

static s7_int char_to_integer_i(s7_pointer p) 
{
  if (!s7_is_character(p)) 
    simple_wrong_type_argument(cur_sc, cur_sc->char_to_integer_symbol, p, T_CHARACTER);
  return(character(p));
}


static s7_pointer g_integer_to_char(s7_scheme *sc, s7_pointer args)
{
  #define H_integer_to_char "(integer->char i) converts the non-negative integer i to a character"
  #define Q_integer_to_char s7_make_signature(sc, 2, sc->is_char_symbol, sc->is_integer_symbol)

  s7_pointer x;
  s7_int ind;

  x = car(args);
  if (!s7_is_integer(x))
    method_or_bust_one_arg(sc, x, sc->integer_to_char_symbol, list_1(sc, x), T_INTEGER);
  ind = s7_integer(x);
  if ((ind < 0) || (ind >= NUM_CHARS))
    return(simple_wrong_type_argument_with_type(sc, sc->integer_to_char_symbol, x, s7_make_string_wrapper(sc, "an integer that can represent a character")));
  return(s7_make_character(sc, (unsigned char)ind));
}

static s7_pointer integer_to_char_p_p(s7_pointer x)
{
  s7_int ind;
  if (!s7_is_integer(x))
    simple_wrong_type_argument(cur_sc, cur_sc->integer_to_char_symbol, x, T_INTEGER);
  ind = s7_integer(x);
  if ((ind < 0) || (ind >= NUM_CHARS))
    return(simple_wrong_type_argument_with_type(cur_sc, cur_sc->integer_to_char_symbol, x, s7_make_string_wrapper(cur_sc, "an integer that can represent a character")));
  return(s7_make_character(cur_sc, (unsigned char)ind));
}


static unsigned char uppers[256], lowers[256];
static void init_uppers(void)
{
  int32_t i;
  for (i = 0; i < 256; i++)
    {
      uppers[i] = (unsigned char)toupper(i);
      lowers[i] = (unsigned char)tolower(i);
    }
}

static s7_pointer g_char_upcase(s7_scheme *sc, s7_pointer args)
{
  #define H_char_upcase "(char-upcase c) converts the character c to upper case"
  #define Q_char_upcase pcl_c
  if (!s7_is_character(car(args)))
    method_or_bust_one_arg(sc, car(args), sc->char_upcase_symbol, args, T_CHARACTER);
  return(s7_make_character(sc, upper_character(car(args))));
}

static s7_pointer g_char_downcase(s7_scheme *sc, s7_pointer args)
{
  #define H_char_downcase "(char-downcase c) converts the character c to lower case"
  #define Q_char_downcase pcl_c
  if (!s7_is_character(car(args)))
    method_or_bust_one_arg(sc, car(args), sc->char_downcase_symbol, args, T_CHARACTER);
  return(s7_make_character(sc, lowers[character(car(args))]));
}

static s7_pointer g_is_char_alphabetic(s7_scheme *sc, s7_pointer args)
{
  #define H_is_char_alphabetic "(char-alphabetic? c) returns #t if the character c is alphabetic"
  #define Q_is_char_alphabetic pl_bc
  if (!s7_is_character(car(args)))
    method_or_bust_one_arg(sc, car(args), sc->is_char_alphabetic_symbol, args, T_CHARACTER);
  return(make_boolean(sc, is_char_alphabetic(car(args))));

  /* isalpha returns #t for (integer->char 226) and others in that range */
}

static bool is_char_alphabetic_b(s7_pointer c) {return((s7_is_character(c)) && (is_char_alphabetic(c)));}


static s7_pointer g_is_char_numeric(s7_scheme *sc, s7_pointer args)
{
  s7_pointer arg;
  #define H_is_char_numeric "(char-numeric? c) returns #t if the character c is a digit"
  #define Q_is_char_numeric pl_bc

  arg = car(args);
  if (!s7_is_character(arg))
    method_or_bust_one_arg(sc, arg, sc->is_char_numeric_symbol, set_plist_1(sc, arg), T_CHARACTER);
  return(make_boolean(sc, is_char_numeric(arg)));
}

static bool is_char_numeric_b(s7_pointer c) {return((s7_is_character(c)) && (is_char_numeric(c)));}


static s7_pointer g_is_char_whitespace(s7_scheme *sc, s7_pointer args)
{
  s7_pointer arg;
  #define H_is_char_whitespace "(char-whitespace? c) returns #t if the character c is non-printing character"
  #define Q_is_char_whitespace pl_bc

  arg = car(args);
  if (!s7_is_character(arg))
    method_or_bust_one_arg(sc, arg, sc->is_char_whitespace_symbol, set_plist_1(sc, arg), T_CHARACTER);
  return(make_boolean(sc, is_char_whitespace(arg)));
}

static bool is_char_whitespace_b(s7_pointer c) {return((s7_is_character(c)) && (is_char_whitespace(c)));}


static s7_pointer g_is_char_upper_case(s7_scheme *sc, s7_pointer args)
{
  s7_pointer arg;
  #define H_is_char_upper_case "(char-upper-case? c) returns #t if the character c is in upper case"
  #define Q_is_char_upper_case pl_bc
  
  arg = car(args);
  if (!s7_is_character(arg))
    method_or_bust_one_arg(sc, arg, sc->is_char_upper_case_symbol, set_plist_1(sc, arg), T_CHARACTER);
  return(make_boolean(sc, is_char_uppercase(arg)));
}

static bool is_char_upper_case_b(s7_pointer c) 
{
  if (!s7_is_character(c))
    simple_wrong_type_argument(cur_sc, cur_sc->is_char_upper_case_symbol, c, T_CHARACTER);
  return(is_char_uppercase(c));
}


static s7_pointer g_is_char_lower_case(s7_scheme *sc, s7_pointer args)
{
  s7_pointer arg;
  #define H_is_char_lower_case "(char-lower-case? c) returns #t if the character c is in lower case"
  #define Q_is_char_lower_case pl_bc
  
  arg = car(args);
  if (!s7_is_character(arg))
    method_or_bust_one_arg(sc, arg, sc->is_char_lower_case_symbol, set_plist_1(sc, arg), T_CHARACTER);
  return(make_boolean(sc, is_char_lowercase(arg)));
}

static bool is_char_lower_case_b(s7_pointer c)
{
  if (!s7_is_character(c))
    simple_wrong_type_argument(cur_sc, cur_sc->is_char_lower_case_symbol, c, T_CHARACTER);
  return(is_char_lowercase(c));
}


static s7_pointer g_is_char(s7_scheme *sc, s7_pointer args)
{
  #define H_is_char "(char? obj) returns #t if obj is a character"
  #define Q_is_char pl_bt
  check_boolean_method(sc, s7_is_character, sc->is_char_symbol, args);
}


s7_pointer s7_make_character(s7_scheme *sc, uint32_t c)
{
  return(chars[c]);
}


bool s7_is_character(s7_pointer p)
{
  return(type(p) == T_CHARACTER);
}


char s7_character(s7_pointer p)
{
  return(character(p));
}


static int32_t charcmp(unsigned char c1, unsigned char c2)
{
  return((c1 == c2) ? 0 : (c1 < c2) ? -1 : 1);
  /* not tolower here -- the single case is apparently supposed to be upper case
   *   this matters in a case like (char-ci<? #\_ #\e) which Guile and Gauche say is #f
   *   although (char<? #\_ #\e) is #t -- the spec does not say how to interpret this!
   */
}


static bool is_character_via_method(s7_scheme *sc, s7_pointer p)
{
  if (s7_is_character(p))
    return(true);
  if (has_methods(p))
    {
      s7_pointer f;
      f = find_method(sc, find_let(sc, p), sc->is_char_symbol);
      if (f != sc->undefined)
	return(is_true(sc, s7_apply_function(sc, f, cons(sc, p, sc->nil))));
    }
  return(false);
}


static s7_pointer g_char_cmp(s7_scheme *sc, s7_pointer args, int32_t val, s7_pointer sym)
{
  s7_pointer x, y;

  y = car(args);
  if (!s7_is_character(y))
    method_or_bust(sc, y, sym, args, T_CHARACTER, 1);

  for (x = cdr(args); is_pair(x); x = cdr(x))
    {
      if (!s7_is_character(car(x)))
	method_or_bust(sc, car(x), sym, cons(sc, y, x), T_CHARACTER, position_of(x, args));

      if (charcmp(character(y), character(car(x))) != val)
	{
	  for (y = cdr(x); is_pair(y); y = cdr(y))
	    if (!is_character_via_method(sc, car(y)))
	      return(wrong_type_argument(sc, sym, position_of(y, args), car(y), T_CHARACTER));
	  return(sc->F);
	}
      y = car(x);
    }
  return(sc->T);
}


static s7_pointer g_char_cmp_not(s7_scheme *sc, s7_pointer args, int32_t val, s7_pointer sym)
{
  s7_pointer x, y;

  y = car(args);
  if (!s7_is_character(y))
    method_or_bust(sc, y, sym, args, T_CHARACTER, 1);

  for (x = cdr(args); is_pair(x); x = cdr(x))
    {
      if (!s7_is_character(car(x)))
	method_or_bust(sc, car(x), sym, cons(sc, y, x), T_CHARACTER, position_of(x, args));

      if (charcmp(character(y), character(car(x))) == val)
	{
	  for (y = cdr(x); is_pair(y); y = cdr(y))
	    if (!is_character_via_method(sc, car(y)))
	      return(wrong_type_argument(sc, sym, position_of(y, args), car(y), T_CHARACTER));
	  return(sc->F);
	}
      y = car(x);
    }
  return(sc->T);
}


static s7_pointer g_chars_are_equal(s7_scheme *sc, s7_pointer args)
{
  #define H_chars_are_equal "(char=? char ...) returns #t if all the character arguments are equal"
  #define Q_chars_are_equal pcl_bc

  s7_pointer x, y;

  y = car(args);
  if (!s7_is_character(y))
    method_or_bust(sc, y, sc->char_eq_symbol, args, T_CHARACTER, 1);

  for (x = cdr(args); is_pair(x); x = cdr(x))
    {
      if (!s7_is_character(car(x)))
	method_or_bust(sc, car(x), sc->char_eq_symbol, cons(sc, y, x), T_CHARACTER, position_of(x, args));

      if (car(x) != y)
	{
	  for (y = cdr(x); is_pair(y); y = cdr(y))
	    if (!is_character_via_method(sc, car(y)))
	      return(wrong_type_argument(sc, sc->char_eq_symbol, position_of(y, args), car(y), T_CHARACTER));
	  return(sc->F);
	}
    }
  return(sc->T);
}


static s7_pointer g_chars_are_less(s7_scheme *sc, s7_pointer args)
{
  #define H_chars_are_less "(char<? char ...) returns #t if all the character arguments are increasing"
  #define Q_chars_are_less pcl_bc

  return(g_char_cmp(sc, args, -1, sc->char_lt_symbol));
}


static s7_pointer g_chars_are_greater(s7_scheme *sc, s7_pointer args)
{
  #define H_chars_are_greater "(char>? char ...) returns #t if all the character arguments are decreasing"
  #define Q_chars_are_greater pcl_bc

  return(g_char_cmp(sc, args, 1, sc->char_gt_symbol));
}


static s7_pointer g_chars_are_geq(s7_scheme *sc, s7_pointer args)
{
  #define H_chars_are_geq "(char>=? char ...) returns #t if all the character arguments are equal or decreasing"
  #define Q_chars_are_geq pcl_bc

  return(g_char_cmp_not(sc, args, -1, sc->char_geq_symbol));
}


static s7_pointer g_chars_are_leq(s7_scheme *sc, s7_pointer args)
{
  #define H_chars_are_leq "(char<=? char ...) returns #t if all the character arguments are equal or increasing"
  #define Q_chars_are_leq pcl_bc

  return(g_char_cmp_not(sc, args, 1, sc->char_leq_symbol));
}

static s7_pointer simple_char_eq;
static s7_pointer g_simple_char_eq(s7_scheme *sc, s7_pointer args)
{
  return(make_boolean(sc, character(car(args)) == character(cadr(args))));
}


static void check_char2_args(s7_scheme *sc, s7_pointer caller, s7_pointer p1, s7_pointer p2)
{
  if (!s7_is_character(p1))
    simple_wrong_type_argument(sc, caller, p1, T_CHARACTER);
  if (!s7_is_character(p2))
    simple_wrong_type_argument(sc, caller, p2, T_CHARACTER);
}

static bool char_lt_b_direct(s7_pointer p1, s7_pointer p2) {return(character(p1) < character(p2));}
static bool char_lt_b(s7_pointer p1, s7_pointer p2)
{
  check_char2_args(cur_sc, cur_sc->char_lt_symbol, p1, p2);
  return(character(p1) < character(p2));
}

static bool char_leq_b_direct(s7_pointer p1, s7_pointer p2) {return(character(p1) <= character(p2));}
static bool char_leq_b(s7_pointer p1, s7_pointer p2)
{
  check_char2_args(cur_sc, cur_sc->char_leq_symbol, p1, p2);
  return(character(p1) <= character(p2));
}

static bool char_gt_b_direct(s7_pointer p1, s7_pointer p2) {return(character(p1) > character(p2));}
static bool char_gt_b(s7_pointer p1, s7_pointer p2)
{
  check_char2_args(cur_sc, cur_sc->char_gt_symbol, p1, p2);
  return(character(p1) > character(p2));
}

static bool char_geq_b_direct(s7_pointer p1, s7_pointer p2) {return(character(p1) >= character(p2));}
static bool char_geq_b(s7_pointer p1, s7_pointer p2)
{
  check_char2_args(cur_sc, cur_sc->char_geq_symbol, p1, p2);
  return(character(p1) >= character(p2));
}

static bool char_eq_b_direct(s7_pointer p1, s7_pointer p2) {return(character(p1) == character(p2));}
static bool char_eq_b(s7_pointer p1, s7_pointer p2)
{
  check_char2_args(cur_sc, cur_sc->char_eq_symbol, p1, p2);
  return(character(p1) == character(p2));
}


static s7_pointer char_equal_s_ic, char_equal_2;
static s7_pointer g_char_equal_s_ic(s7_scheme *sc, s7_pointer args)
{
  s7_pointer c;
  c = find_symbol_unchecked(sc, car(args));
  if (c == cadr(args))
    return(sc->T);
  if (s7_is_character(c))
    return(sc->F);
  method_or_bust(sc, c, sc->char_eq_symbol, list_2(sc, c, cadr(args)), T_CHARACTER, 1);
}

static s7_pointer g_char_equal_2(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_character(car(args)))
    method_or_bust(sc, car(args), sc->char_eq_symbol, args, T_CHARACTER, 1);
  if (car(args) == cadr(args))
    return(sc->T);
  if (!s7_is_character(cadr(args)))
    method_or_bust(sc, cadr(args), sc->char_eq_symbol, args, T_CHARACTER, 2);
  return(sc->F);
}

static s7_pointer char_less_2;
static s7_pointer g_char_less_2(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_character(car(args)))
    method_or_bust(sc, car(args), sc->char_lt_symbol, args, T_CHARACTER, 1);
  if (!s7_is_character(cadr(args)))
    method_or_bust(sc, cadr(args), sc->char_lt_symbol, args, T_CHARACTER, 2);
  return(make_boolean(sc, character(car(args)) < character(cadr(args))));
}


static s7_pointer char_greater_2;
static s7_pointer g_char_greater_2(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_character(car(args)))
    method_or_bust(sc, car(args), sc->char_gt_symbol, args, T_CHARACTER, 1);
  if (!s7_is_character(cadr(args)))
    method_or_bust(sc, cadr(args), sc->char_gt_symbol, args, T_CHARACTER, 2);
  return(make_boolean(sc, character(car(args)) > character(cadr(args))));
}

#if (!WITH_PURE_S7)
static s7_pointer g_char_cmp_ci(s7_scheme *sc, s7_pointer args, int32_t val, s7_pointer sym)
{
  s7_pointer x, y;

  y = car(args);
  if (!s7_is_character(y))
    method_or_bust(sc, y, sym, args, T_CHARACTER, 1);

  for (x = cdr(args); is_pair(x); x = cdr(x))
    {
      if (!s7_is_character(car(x)))
	method_or_bust(sc, car(x), sym, cons(sc, y, x), T_CHARACTER, position_of(x, args));
      if (charcmp(upper_character(y), upper_character(car(x))) != val)
	{
	  for (y = cdr(x); is_pair(y); y = cdr(y))
	    if (!is_character_via_method(sc, car(y)))
	      return(wrong_type_argument(sc, sym, position_of(y, args), car(y), T_CHARACTER));
	  return(sc->F);
	}
      y = car(x);
    }
  return(sc->T);
}

static s7_pointer g_char_cmp_ci_not(s7_scheme *sc, s7_pointer args, int32_t val, s7_pointer sym)
{
  s7_pointer x, y;

  y = car(args);
  if (!s7_is_character(y))
    method_or_bust(sc, y, sym, args, T_CHARACTER, 1);
  for (x = cdr(args); is_pair(x); x = cdr(x))
    {
      if (!s7_is_character(car(x)))
	method_or_bust(sc, car(x), sym, cons(sc, y, x), T_CHARACTER, position_of(x, args));
      if (charcmp(upper_character(y), upper_character(car(x))) == val)
	{
	  for (y = cdr(x); is_pair(y); y = cdr(y))
	    if (!is_character_via_method(sc, car(y)))
	      return(wrong_type_argument(sc, sym, position_of(y, args), car(y), T_CHARACTER));
	  return(sc->F);
	}
      y = car(x);
    }
  return(sc->T);
}


static s7_pointer g_chars_are_ci_equal(s7_scheme *sc, s7_pointer args)
{
  #define H_chars_are_ci_equal "(char-ci=? char ...) returns #t if all the character arguments are equal, ignoring case"
  #define Q_chars_are_ci_equal pcl_bc

  return(g_char_cmp_ci(sc, args, 0, sc->char_ci_eq_symbol));
}

static s7_pointer g_chars_are_ci_less(s7_scheme *sc, s7_pointer args)
{
  #define H_chars_are_ci_less "(char-ci<? char ...) returns #t if all the character arguments are increasing, ignoring case"
  #define Q_chars_are_ci_less pcl_bc

  return(g_char_cmp_ci(sc, args, -1, sc->char_ci_lt_symbol));
}

static s7_pointer g_chars_are_ci_greater(s7_scheme *sc, s7_pointer args)
{
  #define H_chars_are_ci_greater "(char-ci>? char ...) returns #t if all the character arguments are decreasing, ignoring case"
  #define Q_chars_are_ci_greater pcl_bc

  return(g_char_cmp_ci(sc, args, 1, sc->char_ci_gt_symbol));
}

static s7_pointer g_chars_are_ci_geq(s7_scheme *sc, s7_pointer args)
{
  #define H_chars_are_ci_geq "(char-ci>=? char ...) returns #t if all the character arguments are equal or decreasing, ignoring case"
  #define Q_chars_are_ci_geq pcl_bc

  return(g_char_cmp_ci_not(sc, args, -1, sc->char_ci_geq_symbol));
}

static s7_pointer g_chars_are_ci_leq(s7_scheme *sc, s7_pointer args)
{
  #define H_chars_are_ci_leq "(char-ci<=? char ...) returns #t if all the character arguments are equal or increasing, ignoring case"
  #define Q_chars_are_ci_leq pcl_bc

  return(g_char_cmp_ci_not(sc, args, 1, sc->char_ci_leq_symbol));
}


static bool char_ci_lt_b_direct(s7_pointer p1, s7_pointer p2) {return(upper_character(p1) < upper_character(p2));}
static bool char_ci_lt_b(s7_pointer p1, s7_pointer p2)
{
  check_char2_args(cur_sc, cur_sc->char_ci_lt_symbol, p1, p2);
  return(upper_character(p1) < upper_character(p2));
}

static bool char_ci_leq_b_direct(s7_pointer p1, s7_pointer p2) {return(upper_character(p1) <= upper_character(p2));}
static bool char_ci_leq_b(s7_pointer p1, s7_pointer p2)
{
  check_char2_args(cur_sc, cur_sc->char_ci_leq_symbol, p1, p2);
  return(upper_character(p1) <= upper_character(p2));
}

static bool char_ci_gt_b_direct(s7_pointer p1, s7_pointer p2) {return(upper_character(p1) > upper_character(p2));}
static bool char_ci_gt_b(s7_pointer p1, s7_pointer p2)
{
  check_char2_args(cur_sc, cur_sc->char_ci_gt_symbol, p1, p2);
  return(upper_character(p1) > upper_character(p2));
}

static bool char_ci_geq_b_direct(s7_pointer p1, s7_pointer p2) {return(upper_character(p1) >= upper_character(p2));}
static bool char_ci_geq_b(s7_pointer p1, s7_pointer p2)
{
  check_char2_args(cur_sc, cur_sc->char_ci_geq_symbol, p1, p2);
  return(upper_character(p1) >= upper_character(p2));
}

static bool char_ci_eq_b_direct(s7_pointer p1, s7_pointer p2) {return(upper_character(p1) == upper_character(p2));}
static bool char_ci_eq_b(s7_pointer p1, s7_pointer p2)
{
  check_char2_args(cur_sc, cur_sc->char_ci_eq_symbol, p1, p2);
  return(upper_character(p1) == upper_character(p2));
}

#endif /* not pure s7 */



static s7_pointer g_char_position(s7_scheme *sc, s7_pointer args)
{
  #define H_char_position "(char-position char-or-str str (start 0)) returns the position of the first occurrence of char in str, or #f"
  #define Q_char_position s7_make_signature(sc, 4, s7_make_signature(sc, 2, sc->is_integer_symbol, sc->is_boolean_symbol), s7_make_signature(sc, 2, sc->is_char_symbol, sc->is_string_symbol), sc->is_string_symbol, sc->is_integer_symbol)

  const char *porig, *pset;
  s7_int start, pos, len; /* not "int" because start arg might be most-negative-fixnum */
  s7_pointer arg1, arg2;

  arg1 = car(args);
  if ((!s7_is_character(arg1)) &&
      (!is_string(arg1)))
    method_or_bust(sc, arg1, sc->char_position_symbol, args, T_CHARACTER, 1);

  arg2 = cadr(args);
  if (!is_string(arg2))
    method_or_bust(sc, arg2, sc->char_position_symbol, args, T_STRING, 2);

  porig = string_value(arg2);
  len = string_length(arg2);

  if (is_pair(cddr(args)))
    {
      s7_pointer arg3;
      arg3 = caddr(args);
      if (!s7_is_integer(arg3))
	{
	  s7_pointer p;
	  if (!s7_is_integer(p = check_values(sc, arg3, cddr(args))))
	    method_or_bust(sc, arg3, sc->char_position_symbol, args, T_INTEGER, 3);
	  arg3 = p;
	}
      start = s7_integer(arg3);
      if (start < 0)
	return(wrong_type_argument_with_type(sc, sc->char_position_symbol, 3, arg3, a_non_negative_integer_string));
    }
  else start = 0;
  if (start >= len) return(sc->F);

  if (s7_is_character(arg1))
    {
      char c;
      const char *p;
      c = character(arg1);
      p = strchr((const char *)(porig + start), (int)c); /* use strchrnul in Gnu C to catch embedded null case */
      if (p)
	return(make_integer(sc, p - porig));
      return(sc->F);
    }

  if (string_length(arg1) == 0)
    return(sc->F);
  pset = string_value(arg1);

  pos = strcspn((const char *)(porig + start), (const char *)pset);
  if ((pos + start) < len)
    return(make_integer(sc, pos + start));

  /* but if the string has an embedded null, we can get erroneous results here --
   *   perhaps check for null at pos+start?  What about a searched-for string that
   *   also has embedded nulls?
   *
   * The embedded nulls are for byte-vector usages, where presumably you're not talking
   *   about chars and strings, so I think I'll ignore these cases.  In unicode, you'd
   *   want to use unicode-aware searchers, so that also is irrelevant.
   */
  return(sc->F);
}

static s7_pointer char_position_p_ppi(s7_pointer p1, s7_pointer p2, s7_int start)
{
  /* p1 is char, p2 is string, p3 is int32_t */
  const char *porig, *p;
  char c;
  s7_int len;
  c = character(p1);
  if (!is_string(p2))
    simple_wrong_type_argument(cur_sc, cur_sc->char_position_symbol, p2, T_STRING);
  len = string_length(p2);
  porig = string_value(p2);
  if (start >= len) return(cur_sc->F);
  p = strchr((const char *)(porig + start), (int)c);
  if (p) return(make_integer(cur_sc, p - porig));
  return(cur_sc->F);  
}


static s7_pointer char_position_csi;
static s7_pointer g_char_position_csi(s7_scheme *sc, s7_pointer args)
{
  /* assume char arg1, no end */
  const char *porig, *p;
  char c;
  s7_pointer arg2;
  s7_int start, len;

  c = character(car(args));
  arg2 = cadr(args);

  if (!is_string(arg2))
    return(g_char_position(sc, args));

  len = string_length(arg2); /* can't return #f here if len==0 -- need start error check first */
  porig = string_value(arg2);

  if (is_pair(cddr(args)))
    {
      s7_pointer arg3;
      arg3 = caddr(args);
      if (!s7_is_integer(arg3))
	return(g_char_position(sc, args));
      start = s7_integer(arg3);
      if (start < 0)
	return(wrong_type_argument_with_type(sc, sc->char_position_symbol, 3, arg3, a_non_negative_integer_string));
      if (start >= len) return(sc->F);
    }
  else start = 0;

  if (len == 0) return(sc->F);
  p = strchr((const char *)(porig + start), (int)c);
  if (p)
    return(make_integer(sc, p - porig));
  return(sc->F);
}


static s7_pointer g_string_position(s7_scheme *sc, s7_pointer args)
{
  #define H_string_position "(string-position str1 str2 (start 0)) returns the starting position of str1 in str2 or #f"
  #define Q_string_position s7_make_signature(sc, 4, s7_make_signature(sc, 2, sc->is_integer_symbol, sc->is_boolean_symbol), sc->is_string_symbol, sc->is_string_symbol, sc->is_integer_symbol)
  const char *s1, *s2, *p2;
  s7_int start = 0;
  s7_pointer s1p, s2p;

  s1p = car(args);
  if (!is_string(s1p))
    method_or_bust(sc, s1p, sc->string_position_symbol, args, T_STRING, 1);

  s2p = cadr(args);
  if (!is_string(s2p))
    method_or_bust(sc, s2p, sc->string_position_symbol, args, T_STRING, 2);

  if (is_pair(cddr(args)))
    {
      s7_pointer arg3;
      arg3 = caddr(args);
      if (!s7_is_integer(arg3))
	{
	  s7_pointer p;
	  if (!s7_is_integer(p = check_values(sc, arg3, cddr(args))))
	    method_or_bust(sc, arg3, sc->string_position_symbol, args, T_INTEGER, 3);
	  arg3 = p;
	}
      start = s7_integer(arg3);
      if (start < 0)
	return(wrong_type_argument_with_type(sc, sc->string_position_symbol, 3, arg3, a_non_negative_integer_string));
    }

  if (string_length(s1p) == 0)
    return(sc->F);
  s1 = string_value(s1p);
  s2 = string_value(s2p);
  if (start >= string_length(s2p))
    return(sc->F);

  p2 = strstr((const char *)(s2 + start), s1);
  if (!p2) return(sc->F);
  return(make_integer(sc, p2 - s2));
}


/* -------------------------------- strings -------------------------------- */

s7_pointer s7_make_string_with_length(s7_scheme *sc, const char *str, int32_t len)
{
  s7_pointer x;
  new_cell(sc, x, T_STRING | T_SAFE_PROCEDURE);
  string_value(x) = alloc_string(sc, len); 
  if (len != 0)                                  /* memcpy can segfault if string_value(x) is NULL */
    memcpy((void *)string_value(x), (void *)str, len);
  string_value(x)[len] = 0;
  string_length(x) = len;
  string_hash(x) = 0;
  string_needs_free(x) = true;
  add_string(sc, x);
  return(x);
}


static s7_pointer make_string_uncopied_with_length(s7_scheme *sc, char *str, int32_t len)
{
  s7_pointer x;
  new_cell(sc, x, T_STRING | T_SAFE_PROCEDURE);
  string_value(x) = str;
  string_length(x) = len;
  string_hash(x) = 0;
  string_needs_free(x) = true;
  add_string1(sc, x);
  return(x);
}


static s7_pointer make_string_wrapper_with_length(s7_scheme *sc, const char *str, int32_t len)
{
  s7_pointer x;
  new_cell(sc, x, T_STRING | T_IMMUTABLE | T_SAFE_PROCEDURE);
  string_value(x) = (char *)str;
  string_length(x) = len;
  string_hash(x) = 0;
  string_needs_free(x) = false;
  return(x);
}


s7_pointer s7_make_string_wrapper(s7_scheme *sc, const char *str)
{
  return(make_string_wrapper_with_length(sc, str, safe_strlen(str)));
}


static s7_pointer make_empty_string(s7_scheme *sc, int32_t len, char fill)
{
  s7_pointer x;
  new_cell(sc, x, T_STRING);
  string_value(x) = alloc_string(sc, len);
  if (fill != 0)
    memset((void *)(string_value(x)), fill, len);
  string_value(x)[len] = 0;
  string_hash(x) = 0;
  string_length(x) = len;
  string_needs_free(x) = true;
  add_string(sc, x);
  return(x);
}


s7_pointer s7_make_string(s7_scheme *sc, const char *str)
{
  if (str)
    return(s7_make_string_with_length(sc, str, safe_strlen(str)));
  return(make_empty_string(sc, 0, 0));
}


static char *make_permanent_c_string(const char *str)
{
  char *x;
  int32_t len;
  len = safe_strlen(str);
  x = (char *)malloc((len + 1) * sizeof(char));
  memcpy((void *)x, (void *)str, len);
  x[len] = 0;
  return(x);
}


s7_pointer s7_make_permanent_string(const char *str)
{
  /* for the symbol table which is never GC'd */
  s7_pointer x;
  x = alloc_pointer();
  unheap(x);
  set_type(x, T_STRING | T_IMMUTABLE);
  if (str)
    {
      uint32_t len;
      len = safe_strlen(str);
      string_length(x) = len;
      string_value(x) = (char *)malloc((len + 1) * sizeof(char));
      memcpy((void *)string_value(x), (void *)str, len);
      string_value(x)[len] = 0;
    }
  else
    {
      string_value(x) = NULL;
      string_length(x) = 0;
    }
  string_hash(x) = 0;
  string_needs_free(x) = false;
  return(x);
}

static s7_pointer make_permanent_string_wrapper(void)
{
  s7_pointer x;
  x = alloc_pointer();
  unheap(x);
  set_type(x, T_STRING);
  string_value(x) = NULL;
  string_length(x) = 0;
  string_hash(x) = 0;
  string_needs_free(x) = false;
  return(x);
}


static s7_pointer make_temporary_string(s7_scheme *sc, const char *str, int32_t len)
{
  s7_pointer p;
  p = sc->tmp_strs[0];
  prepare_temporary_string(sc, len + 1, 0);
  string_length(p) = len;
  if (len > 0)
    memmove((void *)(string_value(p)), (void *)str, len); /* not memcpy because str might be a temp string (i.e. sc->tmp_str_chars -> itself) */
  string_value(p)[len] = 0;
  return(p);
}


bool s7_is_string(s7_pointer p)
{
  return(is_string(p));
}


const char *s7_string(s7_pointer p)
{
  return(string_value(p));
}


static s7_pointer g_is_string(s7_scheme *sc, s7_pointer args)
{
  #define H_is_string "(string? obj) returns #t if obj is a string"
  #define Q_is_string pl_bt

  check_boolean_method(sc, is_string, sc->is_string_symbol, args);
}


/* -------------------------------- make-string -------------------------------- */
static s7_pointer g_make_string(s7_scheme *sc, s7_pointer args)
{
  #define H_make_string "(make-string len (val #\\space)) makes a string of length len filled with the character val (default: space)"
  #define Q_make_string s7_make_signature(sc, 3, sc->is_string_symbol, sc->is_integer_symbol, sc->is_char_symbol)

  s7_pointer n;
  s7_int len;
  char fill = ' ';

  n = car(args);
  if (!s7_is_integer(n))
    {
      check_two_methods(sc, n, sc->make_string_symbol, sc->make_byte_vector_symbol, args);
      return(wrong_type_argument(sc, sc->make_string_symbol, 1, n, T_INTEGER));
    }

  len = s7_integer(n);
  if ((len < 0) || (len > sc->max_string_length))
    return(out_of_range(sc, sc->make_string_symbol, small_int(1), n, (len < 0) ? its_negative_string : its_too_large_string));

  if (is_not_null(cdr(args)))
    {
      if (!s7_is_character(cadr(args)))
	method_or_bust(sc, cadr(args), sc->make_string_symbol, args, T_CHARACTER, 2);
      fill = s7_character(cadr(args));
    }
  n = make_empty_string(sc, (int32_t)len, fill);
  if (fill == '\0')
    memset((void *)string_value(n), 0, (int)len);
  return(n);
}


#if (!WITH_PURE_S7)
static s7_pointer g_string_length(s7_scheme *sc, s7_pointer args)
{
  #define H_string_length "(string-length str) returns the length of the string str"
  #define Q_string_length s7_make_signature(sc, 2, sc->is_integer_symbol, sc->is_string_symbol)
  s7_pointer p;
  p = car(args);
  if (!is_string(p))
    method_or_bust_one_arg(sc, p, sc->string_length_symbol, args, T_STRING);
  return(make_integer(sc, string_length(p)));
}

static s7_int string_length_i(s7_pointer p) 
{
  if (!is_string(p)) 
    simple_wrong_type_argument(cur_sc, cur_sc->string_length_symbol, p, T_STRING);
  return(string_length(p));
}
#endif


/* -------------------------------- string-up|downcase -------------------------------- */
static s7_pointer g_string_downcase(s7_scheme *sc, s7_pointer args)
{
  #define H_string_downcase "(string-downcase str) returns the lower case version of str."
  #define Q_string_downcase pcl_s

  s7_pointer p, newstr;
  int32_t i, len;
  unsigned char *nstr, *ostr;

  p = car(args);
  if (!is_string(p))
    {
      sc->temp3 = p;
      method_or_bust_one_arg(sc, p, sc->string_downcase_symbol, list_1(sc, p), T_STRING);
      sc->temp3 = sc->nil;
    }
  len = string_length(p);
  newstr = make_empty_string(sc, len, 0);

  ostr = (unsigned char *)string_value(p);
  nstr = (unsigned char *)string_value(newstr);
  for (i = 0; i < len; i++)
    nstr[i] = lowers[(int32_t)ostr[i]];

  return(newstr);
}

static s7_pointer g_string_upcase(s7_scheme *sc, s7_pointer args)
{
  #define H_string_upcase "(string-upcase str) returns the upper case version of str."
  #define Q_string_upcase pcl_s

  s7_pointer p, newstr;
  int32_t i, len;
  unsigned char *nstr, *ostr;

  p = car(args);
  if (!is_string(p))
    {
      sc->temp3 = p;
      method_or_bust_one_arg(sc, p, sc->string_upcase_symbol, list_1(sc, p), T_STRING);
      sc->temp3 = sc->nil;
    }
  len = string_length(p);
  newstr = make_empty_string(sc, len, 0);

  ostr = (unsigned char *)string_value(p);
  nstr = (unsigned char *)string_value(newstr);
  for (i = 0; i < len; i++)
    nstr[i] = uppers[(int32_t)ostr[i]];

  return(newstr);
}

uint32_t s7_string_length(s7_pointer str)
{
  return(string_length(str));
}


/* -------------------------------- string-ref -------------------------------- */
static s7_pointer string_ref_1(s7_scheme *sc, s7_pointer strng, s7_pointer index)
{
  /* every use of this has already checked for the byte-vector case */
  char *str;
  s7_int ind;

  if (!s7_is_integer(index))
    {
      s7_pointer p;
      if (!s7_is_integer(p = check_values(sc, index, cons(sc, index, sc->nil))))
	method_or_bust(sc, index, sc->string_ref_symbol, list_2(sc, strng, index), T_INTEGER, 2);
      index = p;
    }
  ind = s7_integer(index);
  if (ind < 0)
    return(wrong_type_argument_with_type(sc, sc->string_ref_symbol, 2, index, a_non_negative_integer_string));
  if (ind >= string_length(strng))
    return(out_of_range(sc, sc->string_ref_symbol, small_int(2), index, its_too_large_string));

  str = string_value(strng);
  return(s7_make_character(sc, ((unsigned char *)str)[ind]));
}


static s7_pointer g_string_ref_2(s7_scheme *sc, s7_pointer strng, s7_pointer args, s7_pointer caller)
{
  s7_pointer index, p;
  char *str;
  s7_int ind;

  index = cadr(args);
  if (!s7_is_integer(index))
    {
      if (!s7_is_integer(p = check_values(sc, index, cdr(args))))
	method_or_bust(sc, index, caller, args, T_INTEGER, 2);
      index = p;
    }
  ind = s7_integer(index);
  if (ind < 0)
    return(wrong_type_argument_with_type(sc, caller, 2, index, a_non_negative_integer_string));
  if (ind >= string_length(strng))
    return(out_of_range(sc, caller, small_int(2), index, its_too_large_string));

  str = string_value(strng);
  if (is_byte_vector(strng))
    return(small_int((unsigned char)(str[ind])));
  return(s7_make_character(sc, ((unsigned char *)str)[ind]));
}

static s7_pointer g_string_ref(s7_scheme *sc, s7_pointer args)
{
  #define H_string_ref "(string-ref str index) returns the character at the index-th element of the string str"
  #define Q_string_ref s7_make_signature(sc, 3, sc->is_char_symbol, sc->is_string_symbol, sc->is_integer_symbol)

  s7_pointer strng;
  strng = car(args);
  if (!is_string(strng))
    method_or_bust(sc, strng, sc->string_ref_symbol, args, T_STRING, 1);
  return(g_string_ref_2(sc, strng, args, sc->string_ref_symbol));
}

static s7_pointer g_byte_vector_ref(s7_scheme *sc, s7_pointer args)
{
  #define H_byte_vector_ref "(byte-vector-ref vect index) returns the byte at the index-th element of the byte-vector vect"
  #define Q_byte_vector_ref s7_make_signature(sc, 3, sc->is_integer_symbol, sc->is_byte_vector_symbol, sc->is_integer_symbol)
  
  s7_pointer v;
  v = car(args);
  if (!is_byte_vector(v))
    method_or_bust(sc, v, sc->byte_vector_ref_symbol, args, T_STRING, 1);
  return(g_string_ref_2(sc, v, args, sc->byte_vector_ref_symbol));
}



/* -------------------------------- string-set! -------------------------------- */
static s7_pointer g_string_set_2(s7_scheme *sc, s7_pointer x, s7_pointer args, s7_pointer caller)
{
  s7_pointer c, index;
  char *str;
  s7_int ind;

  index = cadr(args);
  if (!s7_is_integer(index))
    {
      s7_pointer p;
      if (!s7_is_integer(p = check_values(sc, index, cdr(args))))
	method_or_bust(sc, index, caller, args, T_INTEGER, 2);
      index = p;
    }
  ind = s7_integer(index);
  if (ind < 0)
    return(wrong_type_argument_with_type(sc, caller, 2, index, a_non_negative_integer_string));
  if (ind >= string_length(x))
    return(out_of_range(sc, caller, small_int(2), index, its_too_large_string));
  str = string_value(_TSet(x));

  c = caddr(args);
  if ((typeflag(x) & T_BYTE_VECTOR) == 0)
    {
      if (!s7_is_character(c))
	method_or_bust(sc, c, caller, list_3(sc, x, index, c), T_CHARACTER, 3);
      str[ind] = (char)s7_character(c);
    }
  else
    {
      s7_int ic;  /* not int32_t here! */
      if (!(s7_is_integer(c)))
	method_or_bust(sc, c, caller, list_3(sc, x, index, c), T_INTEGER, 3);
      ic = s7_integer(c);
      if ((ic < 0) || (ic > 255))
	return(wrong_type_argument_with_type(sc, caller, 3, c, an_unsigned_byte_string));
      str[ind] = (char)ic;
    }
  return(c);
}

static s7_pointer g_string_set(s7_scheme *sc, s7_pointer args)
{
  #define H_string_set "(string-set! str index chr) sets the index-th element of the string str to the character chr"
  #define Q_string_set s7_make_signature(sc, 4, sc->is_char_symbol, sc->is_string_symbol, sc->is_integer_symbol, sc->is_char_symbol)

  s7_pointer strng;
  strng = car(args);
  if (!is_string(strng))
    method_or_bust(sc, strng, sc->string_set_symbol, args, T_STRING, 1);
  return(g_string_set_2(sc, strng, args, sc->string_set_symbol));
}

static s7_pointer g_byte_vector_set(s7_scheme *sc, s7_pointer args)
{
  #define H_byte_vector_set "(byte-vector-set! vect index byte) sets the index-th element of the byte-vector vect to the integer byte"
  #define Q_byte_vector_set s7_make_signature(sc, 4, sc->is_integer_symbol, sc->is_byte_vector_symbol, sc->is_integer_symbol, sc->is_integer_symbol)

  s7_pointer v;
  v = car(args);
  if (!is_byte_vector(v))
    method_or_bust(sc, v, sc->byte_vector_set_symbol, args, T_STRING, 1);
  return(g_string_set_2(sc, v, args, sc->byte_vector_set_symbol));
}

static s7_pointer string_ref_p_pi(s7_pointer p1, s7_int i1)
{
  if (!is_string(p1))
    simple_wrong_type_argument(cur_sc, cur_sc->string_ref_symbol, p1, T_STRING);
  if ((i1 < 0) || (i1 >= string_length(p1)))
    out_of_range(cur_sc, cur_sc->string_ref_symbol, small_int(2), make_integer(cur_sc, i1), (i1 < 0) ? its_negative_string : its_too_large_string);
  return(chars[((unsigned char *)string_value(p1))[i1]]);
}

static s7_pointer string_ref_p_pi_direct(s7_pointer p1, s7_int i1)
{
  if ((i1 < 0) || (i1 >= string_length(p1)))
    out_of_range(cur_sc, cur_sc->string_ref_symbol, small_int(2), make_integer(cur_sc, i1), (i1 < 0) ? its_negative_string : its_too_large_string);
  return(chars[((unsigned char *)string_value(p1))[i1]]);
}

static s7_pointer string_ref_unchecked(s7_pointer p1, s7_int i1) {return(chars[((unsigned char *)string_value(p1))[i1]]);}

static s7_pointer string_set_p_pip(s7_pointer p1, s7_int i1, s7_pointer p2)
{
  if (!is_string(p1))
    simple_wrong_type_argument(cur_sc, cur_sc->string_set_symbol, p1, T_STRING);
  if (!s7_is_character(p2))
    simple_wrong_type_argument(cur_sc, cur_sc->string_set_symbol, p2, T_CHARACTER);
  if ((i1 < 0) || (i1 >= string_length(p1)))
    out_of_range(cur_sc, cur_sc->string_set_symbol, small_int(2), make_integer(cur_sc, i1), (i1 < 0) ? its_negative_string : its_too_large_string);
  string_value(p1)[i1] = s7_character(p2);
  return(p2);
}

static s7_pointer string_set_p_pip_direct(s7_pointer p1, s7_int i1, s7_pointer p2)
{
  if ((i1 < 0) || (i1 >= string_length(p1)))
    out_of_range(cur_sc, cur_sc->string_set_symbol, small_int(2),make_integer(cur_sc, i1), (i1 < 0) ? its_negative_string : its_too_large_string);
  string_value(p1)[i1] = s7_character(p2);
  return(p2);
}

static s7_pointer string_set_unchecked(s7_pointer p1, s7_int i1, s7_pointer p2) {string_value(p1)[i1] = s7_character(p2); return(p2);}

static s7_int byte_vector_ref_i(s7_pointer p1, s7_int i1)
{
  if (!is_byte_vector(p1))
    simple_wrong_type_argument_with_type(cur_sc, cur_sc->byte_vector_ref_symbol, p1, s7_make_string(cur_sc, "a byte-vector"));
  if ((i1 < 0) || (i1 >= string_length(p1)))
    out_of_range(cur_sc, cur_sc->byte_vector_ref_symbol, small_int(2), make_integer(cur_sc, i1), (i1 < 0) ? its_negative_string : its_too_large_string);
  return((s7_int)((unsigned char)(string_value(p1)[i1])));
}

static s7_int byte_vector_set_i(s7_pointer p1, s7_int i1, s7_int i2)
{
  if (!is_byte_vector(p1))
    simple_wrong_type_argument_with_type(cur_sc, cur_sc->byte_vector_set_symbol, p1, s7_make_string(cur_sc, "a byte-vector"));
  if ((i2 < 0) || (i2 > 255))
    simple_wrong_type_argument_with_type(cur_sc, cur_sc->byte_vector_set_symbol, make_integer(cur_sc, i2), an_unsigned_byte_string);
  if ((i1 < 0) || (i1 >= string_length(p1)))
    simple_out_of_range(cur_sc, cur_sc->byte_vector_set_symbol, make_integer(cur_sc, i1), (i1 < 0) ? its_negative_string : its_too_large_string);
  string_value(p1)[i1] = (char)i2;
  return(i2);
}



/* -------------------------------- string-append -------------------------------- */
static s7_pointer g_string_append_1(s7_scheme *sc, s7_pointer args, bool use_temp)
{
  s7_int len = 0;
  s7_pointer x, newstr;
  char *pos;

  if (is_null(args))
    return(s7_make_string_with_length(sc, "", 0));

  /* get length for new string */
  for (x = args; is_not_null(x); x = cdr(x))
    {
      s7_pointer p;
      p = car(x);
      if (!is_string(p))
	{
	  /* look for string-append and if found, cobble up a plausible intermediate call */
	  if (has_methods(p))
	    {
	      s7_pointer func;
	      func = find_method(sc, find_let(sc, p), sc->string_append_symbol);
	      if (func != sc->undefined)
		{
		  s7_pointer y;
		  if (len == 0)
		    return(s7_apply_function(sc, func, args));
		  newstr = make_empty_string(sc, len, 0);
		  for (pos = string_value(newstr), y = args; y != x; pos += string_length(car(y)), y = cdr(y))
		    memcpy(pos, string_value(car(y)), string_length(car(y)));
		  return(s7_apply_function(sc, func, cons(sc, newstr, x)));
		}
	    }
	  return(wrong_type_argument(sc, sc->string_append_symbol, position_of(x, args), p, T_STRING));
	}
      len += string_length(p);
    }

  if (len > sc->max_string_length)
    return(s7_error(sc, sc->out_of_range_symbol, 
		    set_elist_3(sc, s7_make_string_wrapper(sc, "string-append new string length, ~D, is larger than (*s7* 'max-string-length): ~D"),
				make_integer(sc, len), 
				make_integer(sc, sc->max_string_length))));
  if (use_temp)
    {
      newstr = sc->tmp_strs[0];
      prepare_temporary_string(sc, len + 1, 0);
      string_length(newstr) = len;
      string_value(newstr)[len] = 0;
    }
  else
    {
      /* store the contents of the argument strings into the new string */
      newstr = make_empty_string(sc, len, 0);
    }
  for (pos = string_value(newstr), x = args; is_not_null(x); pos += string_length(car(x)), x = cdr(x))
    memcpy(pos, string_value(car(x)), string_length(car(x)));

  if (is_byte_vector(car(args)))
    set_byte_vector(newstr);

  return(newstr);
}

static s7_pointer g_string_append(s7_scheme *sc, s7_pointer args)
{
  #define H_string_append "(string-append str1 ...) appends all its string arguments into one string"
  #define Q_string_append pcl_s
  return(g_string_append_1(sc, args, false));
}

static s7_pointer string_append_to_temp;
static s7_pointer g_string_append_to_temp(s7_scheme *sc, s7_pointer args)
{
  return(g_string_append_1(sc, args, true));
}


#if (!WITH_PURE_S7)
static s7_pointer g_string_copy(s7_scheme *sc, s7_pointer args)
{
  #define H_string_copy "(string-copy str) returns a copy of its string argument"
  #define Q_string_copy s7_make_signature(sc, 2, sc->is_string_symbol, sc->is_string_symbol)
  s7_pointer p;
  p = car(args);
  if (!is_string(p))
    method_or_bust(sc, p, sc->string_copy_symbol, args, T_STRING, 1);
  return(s7_make_string_with_length(sc, string_value(p), string_length(p)));
}
#endif


/* -------------------------------- substring -------------------------------- */
static s7_pointer start_and_end(s7_scheme *sc, s7_pointer caller, s7_pointer fallback,
				s7_pointer start_and_end_args, s7_pointer args, int32_t position, s7_int *start, s7_int *end)
{
  /* we assume that *start=0 and *end=length, that end is "exclusive"
   *   return true if the start/end points are not changed.
   */
  s7_pointer pstart, pend, p;
  s7_int index;

#if DEBUGGING
  if (is_null(start_and_end_args))
    {
      fprintf(stderr, "start_and_end_args is null\n");
      return(sc->gc_nil);
    }
#endif

  pstart = car(start_and_end_args);
  if (!s7_is_integer(pstart))
    {
      if (!s7_is_integer(p = check_values(sc, pstart, start_and_end_args)))
	{
	  check_two_methods(sc, pstart, caller, fallback, args);
	  return(wrong_type_argument(sc, caller, position, pstart, T_INTEGER));
	}
      else pstart = p;
    }

  index = s7_integer(pstart);
  if ((index < 0) ||
      (index > *end)) /* *end == length here */
    return(out_of_range(sc, caller, small_int(position), pstart, (index < 0) ? its_negative_string : its_too_large_string));
  *start = index;

  if (is_null(cdr(start_and_end_args)))
    return(sc->gc_nil);

  pend = cadr(start_and_end_args);
  if (!s7_is_integer(pend))
    {
      if (!s7_is_integer(p = check_values(sc, pend, cdr(start_and_end_args))))
	{
	  check_two_methods(sc, pend, caller, fallback,
			    (position == 2) ? list_3(sc, car(args), pstart, pend) : list_4(sc, car(args), cadr(args), pstart, pend));
	  return(wrong_type_argument(sc, caller, position + 1, pend, T_INTEGER));
	}
      else pend = p;
    }
  index = s7_integer(pend);
  if ((index < *start) ||
      (index > *end))
    return(out_of_range(sc, caller, small_int(position + 1), pend, (index < *start) ? its_too_small_string : its_too_large_string));
  *end = index;
  return(sc->gc_nil);
}


static s7_pointer g_substring(s7_scheme *sc, s7_pointer args)
{
  #define H_substring "(substring str start (end (length str))) returns the portion of the string str between start and \
end: (substring \"01234\" 1 2) -> \"1\""
  #define Q_substring s7_make_circular_signature(sc, 2, 3, sc->is_string_symbol, sc->is_string_symbol, sc->is_integer_symbol)

  s7_pointer x, str;
  s7_int start = 0, end;
  int32_t len;
  char *s;

  str = car(args);
  if (!is_string(str))
    method_or_bust(sc, str, sc->substring_symbol, args, T_STRING, 1);

  end = string_length(str);
  if (!is_null(cdr(args)))
    {
      x = start_and_end(sc, sc->substring_symbol, NULL, cdr(args), args, 2, &start, &end);
      if (x != sc->gc_nil) return(x);
    }
  s = string_value(str);
  len = (int32_t)(end - start);
  x = s7_make_string_with_length(sc, (char *)(s + start), len);
  string_value(x)[len] = 0;
  return(x);
}


static s7_pointer substring_to_temp;
static s7_pointer g_substring_to_temp(s7_scheme *sc, s7_pointer args)
{
  s7_pointer str;
  s7_int start = 0, end;

  str = car(args);
  if (!is_string(str))
    method_or_bust(sc, str, sc->substring_symbol, args, T_STRING, 1);

  end = string_length(str);
  if (!is_null(cdr(args)))
    {
      s7_pointer x;
      x = start_and_end(sc, sc->substring_symbol, NULL, cdr(args), args, 2, &start, &end);
      if (x != sc->gc_nil) return(x);
    }
  return(make_temporary_string(sc, (const char *)(string_value(str) + start), (int32_t)(end - start)));
}


/* -------------------------------- object->string -------------------------------- */
static use_write_t write_choice(s7_scheme *sc, s7_pointer arg)
{
  if (arg == sc->F) return(USE_DISPLAY);
  if (arg == sc->T) return(USE_WRITE);
  if (arg == sc->key_readable_symbol) return(USE_READABLE_WRITE);
  return(USE_WRITE_WRONG);
}

#define DONT_USE_DISPLAY(Choice) ((Choice == USE_DISPLAY) ? USE_WRITE : Choice)

static char *s7_object_to_c_string_1(s7_scheme *sc, s7_pointer obj, use_write_t use_write, int32_t *nlen);

static s7_pointer g_object_to_string(s7_scheme *sc, s7_pointer args)
{
  #define H_object_to_string "(object->string obj (write #t) (max-len most-positive-fixnum)) returns a string representation of obj."
  #define Q_object_to_string s7_make_signature(sc, 4, sc->is_string_symbol, sc->T, s7_make_signature(sc, 2, sc->is_boolean_symbol, sc->is_keyword_symbol), sc->is_integer_symbol)

  use_write_t choice;
  char *str;
  s7_pointer obj;
  int32_t out_len = 0;
  sc->objstr_max_len = s7_int_max;

  if (is_not_null(cdr(args)))
    {
      choice = write_choice(sc, cadr(args));
      if (choice == USE_WRITE_WRONG)
	method_or_bust(sc, cadr(args), sc->object_to_string_symbol, args, T_BOOLEAN, 2);

      if (is_not_null(cddr(args)))
	{
	  if (!is_integer(caddr(args)))
	    return(wrong_type_argument(sc, sc->object_to_string_symbol, 3, caddr(args), T_INTEGER));
	  sc->objstr_max_len = integer(caddr(args));
	}
    }
  else choice = USE_WRITE;
  /* can't use s7_object_to_string here anymore because it assumes use_write arg is a boolean */

  obj = car(args);
  check_method(sc, obj, sc->object_to_string_symbol, args);
  str = s7_object_to_c_string_1(sc, obj, choice, &out_len);
  sc->objstr_max_len = s7_int_max;
  if (str)
    return(make_string_uncopied_with_length(sc, str, out_len));
  return(s7_make_string_with_length(sc, "", 0));
}


/* -------------------------------- string comparisons -------------------------------- */
static int32_t scheme_strcmp(s7_pointer s1, s7_pointer s2)
{
  /* tricky here because str[i] must be treated as unsigned
   *   (string<? (string (integer->char #xf0)) (string (integer->char #x70)))
   * also null or lack thereof does not say anything about the string end
   *   so we have to go by its length.
   */
  int32_t i, len, len1, len2;
  char *str1, *str2;

  len1 = string_length(s1);
  len2 = string_length(s2);
  if (len1 > len2)
    len = len2;
  else len = len1;

  str1 = string_value(s1);
  str2 = string_value(s2);

  for (i = 0; i < len; i++)
    if ((unsigned char)(str1[i]) < (unsigned char )(str2[i]))
      return(-1);
    else
      {
	if ((unsigned char)(str1[i]) > (unsigned char)(str2[i]))
	  return(1);
      }

  if (len1 < len2)
    return(-1);
  if (len1 > len2)
    return(1);
  return(0);
}


static bool is_string_via_method(s7_scheme *sc, s7_pointer p)
{
  if (s7_is_string(p))
    return(true);
  if (has_methods(p))
    {
      s7_pointer f;
      f = find_method(sc, find_let(sc, p), sc->is_string_symbol);
      if (f != sc->undefined)
	return(is_true(sc, s7_apply_function(sc, f, cons(sc, p, sc->nil))));
    }
  return(false);
}

static s7_pointer g_string_cmp(s7_scheme *sc, s7_pointer args, int32_t val, s7_pointer sym)
{
  s7_pointer x, y;

  y = car(args);
  if (!is_string(y))
    method_or_bust(sc, y, sym, args, T_STRING, 1);

  for (x = cdr(args); is_not_null(x); x = cdr(x))
    {
      if (!is_string(car(x)))
	method_or_bust(sc, car(x), sym, cons(sc, y, x), T_STRING, position_of(x, args));
      if (scheme_strcmp(y, car(x)) != val)
	{
	  for (y = cdr(x); is_pair(y); y = cdr(y))
	    if (!is_string_via_method(sc, car(y)))
	      return(wrong_type_argument(sc, sym, position_of(y, args), car(y), T_STRING));
	  return(sc->F);
	}
      y = car(x);
    }
  return(sc->T);
}


static s7_pointer g_string_cmp_not(s7_scheme *sc, s7_pointer args, int32_t val, s7_pointer sym)
{
  s7_pointer x, y;

  y = car(args);
  if (!is_string(y))
    method_or_bust(sc, y, sym, args, T_STRING, 1);

  for (x = cdr(args); is_not_null(x); x = cdr(x))
    {
      if (!is_string(car(x)))
	method_or_bust(sc, car(x), sym, cons(sc, y, x), T_STRING, position_of(x, args));
      if (scheme_strcmp(y, car(x)) == val)
	{
	  for (y = cdr(x); is_pair(y); y = cdr(y))
	    if (!is_string_via_method(sc, car(y)))
	      return(wrong_type_argument(sc, sym, position_of(y, args), car(y), T_STRING));
	  return(sc->F);
	}
      y = car(x);
    }
  return(sc->T);
}


static bool scheme_strings_are_equal(s7_pointer x, s7_pointer y)
{
  return((string_length(x) == string_length(y)) &&
	 (strings_are_equal_with_length(string_value(x), string_value(y), string_length(x))));
}


static s7_pointer g_strings_are_equal(s7_scheme *sc, s7_pointer args)
{
  #define H_strings_are_equal "(string=? str ...) returns #t if all the string arguments are equal"
  #define Q_strings_are_equal pcl_bs

  /* C-based check stops at null, but we can have embedded nulls.
   *   (let ((s1 "1234") (s2 "1245")) (string-set! s1 1 #\null) (string-set! s2 1 #\null) (string=? s1 s2))
   */
  s7_pointer x, y;
  bool happy = true;

  y = car(args);
  if (!is_string(y))
    method_or_bust(sc, y, sc->string_eq_symbol, args, T_STRING, 1);

  for (x = cdr(args); is_pair(x); x = cdr(x))
    {
      s7_pointer p;
      p = car(x);
      if (y != p)
	{
	  if (!is_string(p))
	    method_or_bust(sc, p, sc->string_eq_symbol, cons(sc, y, x), T_STRING, position_of(x, args));
	  if (happy)
	    happy = scheme_strings_are_equal(p, y);
	}
    }
  if (!happy)
    return(sc->F);
  return(sc->T);
}


static s7_pointer g_strings_are_less(s7_scheme *sc, s7_pointer args)
{
  #define H_strings_are_less "(string<? str ...) returns #t if all the string arguments are increasing"
  #define Q_strings_are_less pcl_bs

  return(g_string_cmp(sc, args, -1, sc->string_lt_symbol));
}


static s7_pointer g_strings_are_greater(s7_scheme *sc, s7_pointer args)
{
  #define H_strings_are_greater "(string>? str ...) returns #t if all the string arguments are decreasing"
  #define Q_strings_are_greater pcl_bs

  return(g_string_cmp(sc, args, 1, sc->string_gt_symbol));
}


static s7_pointer g_strings_are_geq(s7_scheme *sc, s7_pointer args)
{
  #define H_strings_are_geq "(string>=? str ...) returns #t if all the string arguments are equal or decreasing"
  #define Q_strings_are_geq pcl_bs

  return(g_string_cmp_not(sc, args, -1, sc->string_geq_symbol));
}


static s7_pointer g_strings_are_leq(s7_scheme *sc, s7_pointer args)
{
  #define H_strings_are_leq "(string<=? str ...) returns #t if all the string arguments are equal or increasing"
  #define Q_strings_are_leq pcl_bs

  return(g_string_cmp_not(sc, args, 1, sc->string_leq_symbol));
}

static s7_pointer string_equal_2;
static s7_pointer g_string_equal_2(s7_scheme *sc, s7_pointer args)
{
  if (!is_string(car(args)))
    method_or_bust(sc, car(args), sc->string_eq_symbol, args, T_STRING, 1);
  if (!is_string(cadr(args)))
    method_or_bust(sc, cadr(args), sc->string_eq_symbol, args, T_STRING, 2);
  return(make_boolean(sc, scheme_strings_are_equal(car(args), cadr(args))));
}


static s7_pointer string_less_2;
static s7_pointer g_string_less_2(s7_scheme *sc, s7_pointer args)
{
  if (!is_string(car(args)))
    method_or_bust(sc, car(args), sc->string_lt_symbol, args, T_STRING, 1);
  if (!is_string(cadr(args)))
    method_or_bust(sc, cadr(args), sc->string_lt_symbol, args, T_STRING, 2);
  return(make_boolean(sc, scheme_strcmp(car(args), cadr(args)) == -1));
}


static s7_pointer string_greater_2;
static s7_pointer g_string_greater_2(s7_scheme *sc, s7_pointer args)
{
  if (!is_string(car(args)))
    method_or_bust(sc, car(args), sc->string_gt_symbol, args, T_STRING, 1);
  if (!is_string(cadr(args)))
    method_or_bust(sc, cadr(args), sc->string_gt_symbol, args, T_STRING, 2);
  return(make_boolean(sc, scheme_strcmp(car(args), cadr(args)) == 1));
}

static inline void check_string2_args(s7_scheme *sc, s7_pointer caller, s7_pointer p1, s7_pointer p2)
{
  if (!is_string(p1))
    simple_wrong_type_argument(sc, caller, p1, T_STRING);
  if (!s7_is_string(p2))
    simple_wrong_type_argument(sc, caller, p2, T_STRING);
}

static bool string_lt_b_direct(s7_pointer p1, s7_pointer p2) {return(scheme_strcmp(p1, p2) == -1);}
static bool string_lt_b(s7_pointer p1, s7_pointer p2)
{
  check_string2_args(cur_sc, cur_sc->string_lt_symbol, p1, p2);
  return(scheme_strcmp(p1, p2) == -1);
}

static bool string_leq_b_direct(s7_pointer p1, s7_pointer p2) {return(scheme_strcmp(p1, p2) != 1);}
static bool string_leq_b(s7_pointer p1, s7_pointer p2)
{
  check_string2_args(cur_sc, cur_sc->string_leq_symbol, p1, p2);
  return(scheme_strcmp(p1, p2) != 1);
}

static bool string_gt_b_direct(s7_pointer p1, s7_pointer p2) {return(scheme_strcmp(p1, p2) == 1);}
static bool string_gt_b(s7_pointer p1, s7_pointer p2)
{
  check_string2_args(cur_sc, cur_sc->string_gt_symbol, p1, p2);
  return(scheme_strcmp(p1, p2) == 1);
}

static bool string_geq_b_direct(s7_pointer p1, s7_pointer p2) {return(scheme_strcmp(p1, p2) != -1);}
static bool string_geq_b(s7_pointer p1, s7_pointer p2)
{
  check_string2_args(cur_sc, cur_sc->string_geq_symbol, p1, p2);
  return(scheme_strcmp(p1, p2) != -1);
}

static bool string_eq_b_direct(s7_pointer p1, s7_pointer p2) {return(scheme_strings_are_equal(p1, p2));}
static bool string_eq_b(s7_pointer p1, s7_pointer p2)
{
  check_string2_args(cur_sc, cur_sc->string_eq_symbol, p1, p2);
  return(scheme_strings_are_equal(p1, p2));
}


#if (!WITH_PURE_S7)

static int32_t scheme_strcasecmp(s7_pointer s1, s7_pointer s2)
{
  /* same as scheme_strcmp -- watch out for unwanted sign! and lack of trailing null (length sets string end).
   */
  int32_t i, len, len1, len2;
  unsigned char *str1, *str2;

  len1 = string_length(s1);
  len2 = string_length(s2);
  if (len1 > len2)
    len = len2;
  else len = len1;

  str1 = (unsigned char *)string_value(s1);
  str2 = (unsigned char *)string_value(s2);

  for (i = 0; i < len; i++)
    if (uppers[(int32_t)str1[i]] < uppers[(int32_t)str2[i]])
      return(-1);
    else
      {
	if (uppers[(int32_t)str1[i]] > uppers[(int32_t)str2[i]])
	  return(1);
      }

  if (len1 < len2)
    return(-1);
  if (len1 > len2)
    return(1);
  return(0);
}


static bool scheme_strequal_ci(s7_pointer s1, s7_pointer s2)
{
  /* same as scheme_strcmp -- watch out for unwanted sign! */
  int32_t i, len, len2;
  unsigned char *str1, *str2;

  len = string_length(s1);
  len2 = string_length(s2);
  if (len != len2)
    return(false);

  str1 = (unsigned char *)string_value(s1);
  str2 = (unsigned char *)string_value(s2);

  for (i = 0; i < len; i++)
    if (uppers[(int32_t)str1[i]] != uppers[(int32_t)str2[i]])
      return(false);
  return(true);
}


static s7_pointer g_string_ci_cmp(s7_scheme *sc, s7_pointer args, int32_t val, s7_pointer sym)
{
  s7_pointer x, y;

  y = car(args);
  if (!is_string(y))
    method_or_bust(sc, y, sym, args, T_STRING, 1);

  for (x = cdr(args); is_not_null(x); x = cdr(x))
    {
      if (!is_string(car(x)))
	method_or_bust(sc, car(x), sym, cons(sc, y, x), T_STRING, position_of(x, args));
      if (val == 0)
	{
	  if (!scheme_strequal_ci(y, car(x)))
	    {
	      for (y = cdr(x); is_pair(y); y = cdr(y))
		if (!is_string_via_method(sc, car(y)))
		  return(wrong_type_argument(sc, sym, position_of(y, args), car(y), T_STRING));
	      return(sc->F);
	    }
	}
      else
	{
	  if (scheme_strcasecmp(y, car(x)) != val)
	    {
	      for (y = cdr(x); is_pair(y); y = cdr(y))
		if (!is_string_via_method(sc, car(y)))
		  return(wrong_type_argument(sc, sym, position_of(y, args), car(y), T_STRING));
	      return(sc->F);
	    }
	}
      y = car(x);
    }
  return(sc->T);
}


static s7_pointer g_string_ci_cmp_not(s7_scheme *sc, s7_pointer args, int32_t val, s7_pointer sym)
{
  s7_pointer x, y;

  y = car(args);
  if (!is_string(y))
    method_or_bust(sc, y, sym, args, T_STRING, 1);

  for (x = cdr(args); is_not_null(x); x = cdr(x))
    {
      if (!is_string(car(x)))
	method_or_bust(sc, car(x), sym, cons(sc, y, x), T_STRING, position_of(x, args));
      if (scheme_strcasecmp(y, car(x)) == val)
	{
	  for (y = cdr(x); is_pair(y); y = cdr(y))
	    if (!is_string_via_method(sc, car(y)))
	      return(wrong_type_argument(sc, sym, position_of(y, args), car(y), T_STRING));
	  return(sc->F);
	}
      y = car(x);
    }
  return(sc->T);
}


static s7_pointer g_strings_are_ci_equal(s7_scheme *sc, s7_pointer args)
{
  #define H_strings_are_ci_equal "(string-ci=? str ...) returns #t if all the string arguments are equal, ignoring case"
  #define Q_strings_are_ci_equal pcl_bs
  return(g_string_ci_cmp(sc, args, 0, sc->string_ci_eq_symbol));
}


static s7_pointer g_strings_are_ci_less(s7_scheme *sc, s7_pointer args)
{
  #define H_strings_are_ci_less "(string-ci<? str ...) returns #t if all the string arguments are increasing, ignoring case"
  #define Q_strings_are_ci_less pcl_bs
  return(g_string_ci_cmp(sc, args, -1, sc->string_ci_lt_symbol));
}


static s7_pointer g_strings_are_ci_greater(s7_scheme *sc, s7_pointer args)
{
  #define H_strings_are_ci_greater "(string-ci>? str ...) returns #t if all the string arguments are decreasing, ignoring case"
  #define Q_strings_are_ci_greater pcl_bs
  return(g_string_ci_cmp(sc, args, 1, sc->string_ci_gt_symbol));
}


static s7_pointer g_strings_are_ci_geq(s7_scheme *sc, s7_pointer args)
{
  #define H_strings_are_ci_geq "(string-ci>=? str ...) returns #t if all the string arguments are equal or decreasing, ignoring case"
  #define Q_strings_are_ci_geq pcl_bs
  return(g_string_ci_cmp_not(sc, args, -1, sc->string_ci_geq_symbol));
}


static s7_pointer g_strings_are_ci_leq(s7_scheme *sc, s7_pointer args)
{
  #define H_strings_are_ci_leq "(string-ci<=? str ...) returns #t if all the string arguments are equal or increasing, ignoring case"
  #define Q_strings_are_ci_leq pcl_bs
  return(g_string_ci_cmp_not(sc, args, 1, sc->string_ci_leq_symbol));
}


static bool string_ci_lt_b_direct(s7_pointer p1, s7_pointer p2) {return(scheme_strcasecmp(p1, p2) == -1);}
static bool string_ci_lt_b(s7_pointer p1, s7_pointer p2)
{
  check_string2_args(cur_sc, cur_sc->string_ci_lt_symbol, p1, p2);
  return(scheme_strcasecmp(p1, p2) == -1);
}

static bool string_ci_leq_b_direct(s7_pointer p1, s7_pointer p2) {return(scheme_strcasecmp(p1, p2) != 1);}
static bool string_ci_leq_b(s7_pointer p1, s7_pointer p2)
{
  check_string2_args(cur_sc, cur_sc->string_ci_leq_symbol, p1, p2);
  return(scheme_strcasecmp(p1, p2) != 1);
}

static bool string_ci_gt_b_direct(s7_pointer p1, s7_pointer p2) {return(scheme_strcasecmp(p1, p2) == 1);}
static bool string_ci_gt_b(s7_pointer p1, s7_pointer p2)
{
  check_string2_args(cur_sc, cur_sc->string_ci_gt_symbol, p1, p2);
  return(scheme_strcasecmp(p1, p2) == 1);
}

static bool string_ci_geq_b_direct(s7_pointer p1, s7_pointer p2) {return(scheme_strcasecmp(p1, p2) != -1);}
static bool string_ci_geq_b(s7_pointer p1, s7_pointer p2)
{
  check_string2_args(cur_sc, cur_sc->string_ci_geq_symbol, p1, p2);
  return(scheme_strcasecmp(p1, p2) != -1);
}

static bool string_ci_eq_b_direct(s7_pointer p1, s7_pointer p2) {return(scheme_strcasecmp(p1, p2) == 0);}
static bool string_ci_eq_b(s7_pointer p1, s7_pointer p2)
{
  check_string2_args(cur_sc, cur_sc->string_ci_eq_symbol, p1, p2);
  return(scheme_strcasecmp(p1, p2) == 0);
}

#endif /* pure s7 */


static s7_pointer g_string_fill(s7_scheme *sc, s7_pointer args)
{
  #define H_string_fill "(string-fill! str chr start end) fills the string str with the character chr"
  #define Q_string_fill s7_make_circular_signature(sc, 3, 4, s7_make_signature(sc, 2, sc->is_char_symbol, sc->is_integer_symbol), sc->is_string_symbol, sc->is_char_symbol, sc->is_integer_symbol)

  s7_pointer x, chr;
  s7_int start = 0, end, byte = 0;
  x = car(args);

  if (!is_string(x))
    method_or_bust(sc, x, sc->string_fill_symbol, args, T_STRING, 1); /* not two methods here */

  if ((sc->safety > NO_SAFETY) &&
      (is_immutable_string(x)))
    return(s7_error(sc, sc->error_symbol, set_elist_2(sc, s7_make_string_wrapper(sc, "can't fill! ~S (it is immutable)"), x)));

  chr = cadr(args);
  if (!is_byte_vector(x))
    {
      if (!s7_is_character(chr))
	{
	  check_two_methods(sc, chr, sc->string_fill_symbol, sc->fill_symbol, args);
	  return(wrong_type_argument(sc, sc->string_fill_symbol, 2, chr, T_CHARACTER));
	}
    }
  else
    {
      if (!is_integer(chr))
	{
	  check_two_methods(sc, chr, sc->string_fill_symbol, sc->fill_symbol, args);
	  return(wrong_type_argument(sc, sc->fill_symbol, 2, chr, T_INTEGER));
	}
      byte = integer(chr);
      if ((byte < 0) || (byte > 255))
	return(simple_wrong_type_argument_with_type(sc, sc->string_fill_symbol, chr, an_unsigned_byte_string));
    }

  end = string_length(x);
  if (!is_null(cddr(args)))
    {
      s7_pointer p;
      p = start_and_end(sc, sc->string_fill_symbol, sc->fill_symbol, cddr(args), args, 3, &start, &end);
      if (p != sc->gc_nil) return(p);
      if (start == end) return(chr);
    }
  if (end == 0) return(chr);

  if (!is_byte_vector(x))
    memset((void *)(string_value(x) + start), (int)character(chr), end - start);
  else memset((void *)(string_value(x) + start), (int)byte, end - start);

  return(chr);
}


static s7_pointer g_string_1(s7_scheme *sc, s7_pointer args, s7_pointer sym)
{
  int32_t i, len;
  s7_pointer x, newstr;
  char *str;

  /* get length for new string and check arg types */
  for (len = 0, x = args; is_not_null(x); len++, x = cdr(x))
    {
      s7_pointer p;
      p = car(x);
      if (!s7_is_character(p))
	{
	  if (has_methods(p))
	    {
	      s7_pointer func;
	      func = find_method(sc, find_let(sc, p), sym);
	      if (func != sc->undefined)
		{
		  s7_pointer y;
		  if (len == 0)
		    return(s7_apply_function(sc, func, args));
		  newstr = make_empty_string(sc, len, 0);
		  str = string_value(newstr);
		  for (i = 0, y = args; y != x; i++, y = cdr(y))
		    str[i] = character(car(y));
		  return(g_string_append(sc, set_plist_2(sc, newstr, s7_apply_function(sc, func, x))));
		}
	    }
	  return(wrong_type_argument(sc, sym, len + 1, car(x), T_CHARACTER));
	}
    }
  newstr = make_empty_string(sc, len, 0);
  str = string_value(newstr);
  for (i = 0, x = args; is_not_null(x); i++, x = cdr(x))
    str[i] = character(car(x));

  return(newstr);
}


static s7_pointer g_string(s7_scheme *sc, s7_pointer args)
{
  #define H_string "(string chr...) appends all its character arguments into one string"
  #define Q_string s7_make_circular_signature(sc, 1, 2, sc->is_string_symbol, sc->is_char_symbol)

  if (is_null(args))                                /* (string) but not (string ()) */
    return(s7_make_string_with_length(sc, "", 0));
  return(g_string_1(sc, args, sc->string_symbol));
}


#if (!WITH_PURE_S7)
static s7_pointer g_list_to_string(s7_scheme *sc, s7_pointer args)
{
  #define H_list_to_string "(list->string lst) appends all the list's characters into one string; (apply string lst)"
  #define Q_list_to_string s7_make_signature(sc, 2, sc->is_string_symbol, sc->is_proper_list_symbol)

  if (is_null(car(args)))
    return(s7_make_string_with_length(sc, "", 0));

  if (!s7_is_proper_list(sc, car(args))) 
    method_or_bust_with_type_one_arg(sc, car(args), sc->list_to_string_symbol, args, s7_make_string_wrapper(sc, "a (proper, non-circular) list of characters"));
  return(g_string_1(sc, car(args), sc->list_to_string_symbol));
}
#endif

static s7_pointer s7_string_to_list(s7_scheme *sc, const char *str, int32_t len)
{
  int32_t i;
  s7_pointer result;

  if (len == 0) 
    return(sc->nil);
  if (len >= (sc->free_heap_top - sc->free_heap))
    {
      gc(sc);
      while (len >= (sc->free_heap_top - sc->free_heap))
	resize_heap(sc);
    }

  sc->v = sc->nil;
  for (i = len - 1; i >= 0; i--)
    sc->v = cons_unchecked(sc, s7_make_character(sc, ((unsigned char)str[i])), sc->v);
  result = sc->v;
  sc->v = sc->nil;
  return(result);
}

#if (!WITH_PURE_S7)
static s7_pointer g_string_to_list(s7_scheme *sc, s7_pointer args)
{
  #define H_string_to_list "(string->list str start end) returns the elements of the string str in a list; (map values str)"
  #define Q_string_to_list s7_make_circular_signature(sc, 2, 3, sc->is_proper_list_symbol, sc->is_string_symbol, sc->is_integer_symbol)

  s7_int i, start = 0, end;
  s7_pointer p, str;

  str = car(args);
  if (!is_string(str))
    method_or_bust_one_arg(sc, str, sc->string_to_list_symbol, args, T_STRING);

  end = string_length(str);
  if (!is_null(cdr(args)))
    {
      p = start_and_end(sc, sc->string_to_list_symbol, NULL, cdr(args), args, 2, &start, &end);
      if (p != sc->gc_nil) return(p);
      if (start == end) return(sc->nil);
    }
  else
    {
      if (end == 0) return(sc->nil);
    }
  if ((start == 0) && (end == string_length(str)))
    return(s7_string_to_list(sc, string_value(str), string_length(str)));

  sc->w = sc->nil;
  for (i = end - 1; i >= start; i--)
    sc->w = cons(sc, s7_make_character(sc, ((unsigned char)string_value(str)[i])), sc->w);

  p = sc->w;
  sc->w = sc->nil;
  return(p);
}
#endif


/* -------------------------------- byte-vectors --------------------------------
 *
 * these are just strings with the T_BYTE_VECTOR bit set.
 */

static bool s7_is_byte_vector(s7_pointer b) {return((is_string(b)) && (is_byte_vector(b)));}

static s7_pointer g_is_byte_vector(s7_scheme *sc, s7_pointer args)
{
  #define H_is_byte_vector "(byte-vector? obj) returns #t if obj is a byte-vector"
  #define Q_is_byte_vector pl_bt

  check_boolean_method(sc, s7_is_byte_vector, sc->is_byte_vector_symbol, args);
}

static s7_pointer g_string_to_byte_vector(s7_scheme *sc, s7_pointer args)
{
  #define H_string_to_byte_vector "(string->byte-vector obj) turns a string into a byte-vector."
  #define Q_string_to_byte_vector s7_make_signature(sc, 2, sc->is_byte_vector_symbol, sc->is_string_symbol)
  s7_pointer str;
  str = car(args);
  if (is_integer(str))
    str = s7_make_string_with_length(sc, (const char *)(&(integer(str))), sizeof(s7_int));
  else
    {
      if (!is_string(str))
	method_or_bust(sc, str, sc->string_to_byte_vector_symbol, set_plist_1(sc, str), T_STRING, 1);
    }
  set_byte_vector(str);
  return(str);
}

static s7_pointer g_make_byte_vector(s7_scheme *sc, s7_pointer args)
{
  #define H_make_byte_vector "(make-byte-vector len (byte 0)) makes a byte-vector of length len filled with byte."
  #define Q_make_byte_vector s7_make_circular_signature(sc, 1, 2, sc->is_byte_vector_symbol, sc->is_integer_symbol)

  s7_pointer str;
  if (is_null(cdr(args)))
    {
      str = g_make_string(sc, args);
      if (is_string(str))
	memclr((void *)(string_value(str)), string_length(str));
    }
  else
    {
      s7_pointer len, byte;
      s7_int b;
      len = car(args);
      if (!is_integer(len))
	method_or_bust(sc, len, sc->make_byte_vector_symbol, args, T_INTEGER, 1);

      byte = cadr(args);
      if (!s7_is_integer(byte))
	method_or_bust(sc, byte, sc->make_byte_vector_symbol, args, T_INTEGER, 2);

      b = s7_integer(byte);
      if ((b < 0) || (b > 255))
	return(simple_wrong_type_argument_with_type(sc, sc->make_byte_vector_symbol, byte, an_unsigned_byte_string));
      str = g_make_string(sc, set_plist_2(sc, len, chars[b]));
    }
  set_byte_vector(str);
  return(str);
}


static s7_pointer g_byte_vector(s7_scheme *sc, s7_pointer args)
{
  #define H_byte_vector "(byte-vector ...) returns a byte-vector whose elements are the arguments"
  #define Q_byte_vector s7_make_circular_signature(sc, 1, 2, sc->is_byte_vector_symbol, sc->is_integer_symbol)

  s7_int i, len;
  s7_pointer vec, x;
  char *str;

  len = s7_list_length(sc, args);
  vec = make_empty_string(sc, len, 0);
  str = string_value(vec);

  for (i = 0, x = args; is_pair(x); i++, x = cdr(x))
    {
      s7_pointer byte;
      s7_int b;
      byte = car(x);
      if (!s7_is_integer(byte))
	{
	  if (has_methods(byte))
	    {
	      s7_pointer func;
	      func = find_method(sc, find_let(sc, byte), sc->byte_vector_symbol);
	      if (func != sc->undefined)
		{
		  if (i == 0)
		    return(s7_apply_function(sc, func, args));
		  string_length(vec) = i;
		  vec = g_string_append(sc, set_plist_2(sc, vec, s7_apply_function(sc, func, x)));
		  set_byte_vector(vec);
		  return(vec);
		}
	    }
	  return(wrong_type_argument(sc, sc->byte_vector_symbol, i + 1, byte, T_INTEGER));
	}
      b = s7_integer(byte);
      if ((b < 0) || (b > 255))
	return(simple_wrong_type_argument_with_type(sc, sc->byte_vector_symbol, byte, an_unsigned_byte_string));
      str[i] = (unsigned char)b;
    }
  set_byte_vector(vec);
  return(vec);
}

static s7_pointer byte_vector_to_list(s7_scheme *sc, const char *str, int32_t len)
{
  int32_t i;
  s7_pointer p;
  if (len == 0) return(sc->nil);
  sc->w = sc->nil;
  for (i = len - 1; i >= 0; i--)
    sc->w = cons(sc, small_int((uint32_t)((unsigned char)(str[i]))), sc->w); /* extra cast is not redundant! */
  p = sc->w;
  sc->w = sc->nil;
  return(p);
}



/* -------------------------------- ports --------------------------------
 *
 * originally nil served as stdin and friends, but that made it impossible to catch an error
 *   like (read-line (current-output-port)) when the latter was stdout.  So we now have
 *   the built-in constant ports *stdin*, *stdout*, and *stderr*.  Some way is needed to
 *   refer to these directly so that (read-line *stdin*) for example can insist on reading
 *   from the terminal, or whatever stdin is.
 */

static s7_pointer g_is_port_closed(s7_scheme *sc, s7_pointer args)
{
  #define H_is_port_closed "(port-closed? p) returns #t if the port p is closed."
  #define Q_is_port_closed s7_make_signature(sc, 2, sc->is_boolean_symbol, s7_make_signature(sc, 2, sc->is_input_port_symbol, sc->is_output_port_symbol))
  s7_pointer x;

  x = car(args);
  if ((is_input_port(x)) || (is_output_port(x)))
    return(make_boolean(sc, port_is_closed(x)));

  method_or_bust_with_type_one_arg(sc, x, sc->is_port_closed_symbol, args, s7_make_string_wrapper(sc, "a port"));
}

static bool is_port_closed_b(s7_pointer x)
{
  if ((!is_input_port(x)) && (!is_output_port(x)))
    simple_wrong_type_argument_with_type(cur_sc, cur_sc->is_port_closed_symbol, x, s7_make_string_wrapper(cur_sc, "a port"));
  return(port_is_closed(x));
}


static s7_pointer c_port_line_number(s7_scheme *sc, s7_pointer x)
{
  if ((!(is_input_port(x))) ||
      (port_is_closed(x)))
    method_or_bust_with_type_one_arg(sc, x, sc->port_line_number_symbol, list_1(sc, x), an_input_port_string);
  return(make_integer(sc, port_line_number(x)));
}

static s7_pointer g_port_line_number(s7_scheme *sc, s7_pointer args)
{
  #define H_port_line_number "(port-line-number input-file-port) returns the current read line number of port"
  #define Q_port_line_number s7_make_signature(sc, 2, sc->is_integer_symbol, s7_make_signature(sc, 2, sc->is_input_port_symbol, sc->is_null_symbol))

  if ((is_null(args)) || (is_null(car(args))))
    return(c_port_line_number(sc, sc->input_port));
  return(c_port_line_number(sc, car(args)));
}

int32_t s7_port_line_number(s7_pointer p)
{
  if (!(is_input_port(p)))
    simple_wrong_type_argument(cur_sc, cur_sc->port_line_number_symbol, p, T_INPUT_PORT);
  return(port_line_number(p));
}

static s7_int port_line_number_i_p(s7_pointer p) 
{
  return(s7_port_line_number(p));
}

static s7_pointer g_set_port_line_number(s7_scheme *sc, s7_pointer args)
{
  s7_pointer p, line;

  if ((is_null(car(args))) || 
      ((is_null(cdr(args))) && (is_integer(car(args)))))
    p = sc->input_port;
  else 
    {
      p = car(args);
      if (!(is_input_port(p)))
	return(s7_wrong_type_arg_error(sc, "set! port-line-number", 1, p, "an input port"));
    }

  line = (is_null(cdr(args)) ? car(args) : cadr(args));
  if (!is_integer(line))
    return(s7_wrong_type_arg_error(sc, "set! port-line-number", 2, line, "an integer"));
  port_line_number(p) = integer(line);
  return(line);
}


const char *s7_port_filename(s7_pointer x)
{
  if (((is_input_port(x)) ||
       (is_output_port(x))) &&
      (!port_is_closed(x)))
    return(port_filename(x));
  return(NULL);
}


static s7_pointer c_port_filename(s7_scheme *sc, s7_pointer x)
{
  if (((is_input_port(x)) ||
       (is_output_port(x))) &&
      (!port_is_closed(x)))
    {
      if (port_filename(x))
	return(make_string_wrapper_with_length(sc, port_filename(x), port_filename_length(x)));
      return(s7_make_string_with_length(sc, "", 0));
      /* otherwise (eval-string (port-filename)) and (string->symbol (port-filename)) segfault */
    }
  method_or_bust_with_type_one_arg(sc, x, sc->port_filename_symbol, list_1(sc, x), an_open_port_string);
}

static s7_pointer g_port_filename(s7_scheme *sc, s7_pointer args)
{
  #define H_port_filename "(port-filename file-port) returns the filename associated with port"
  #define Q_port_filename s7_make_signature(sc, 2, sc->is_string_symbol, s7_make_signature(sc, 2, sc->is_input_port_symbol, sc->is_output_port_symbol))

  if (is_null(args))
    return(c_port_filename(sc, sc->input_port));
  return(c_port_filename(sc, car(args)));
}


bool s7_is_input_port(s7_scheme *sc, s7_pointer p)
{
  return(is_input_port(p));
}

static bool is_input_port_b(s7_pointer p) {return(is_input_port(p));}


static s7_pointer g_is_input_port(s7_scheme *sc, s7_pointer args)
{
  #define H_is_input_port "(input-port? p) returns #t if p is an input port"
  #define Q_is_input_port pl_bt
  check_boolean_method(sc, is_input_port, sc->is_input_port_symbol, args);
}


bool s7_is_output_port(s7_scheme *sc, s7_pointer p)
{
  return(is_output_port(p));
}

static bool is_output_port_b(s7_pointer p) {return(is_output_port(p));}


static s7_pointer g_is_output_port(s7_scheme *sc, s7_pointer args)
{
  #define H_is_output_port "(output-port? p) returns #t if p is an output port"
  #define Q_is_output_port pl_bt
  check_boolean_method(sc, is_output_port, sc->is_output_port_symbol, args);
}


s7_pointer s7_current_input_port(s7_scheme *sc)
{
  return(sc->input_port);
}


static s7_pointer g_current_input_port(s7_scheme *sc, s7_pointer args)
{
  #define H_current_input_port "(current-input-port) returns the current input port"
  #define Q_current_input_port s7_make_signature(sc, 1, sc->is_input_port_symbol)
  return(sc->input_port);
}

#if (!WITH_PURE_S7)
static s7_pointer g_set_current_input_port(s7_scheme *sc, s7_pointer args)
{
  #define H_set_current_input_port "(set-current-input-port port) sets the current-input port to port and returns the previous value of the input port"
  #define Q_set_current_input_port s7_make_signature(sc, 2, sc->is_input_port_symbol, sc->is_input_port_symbol)

  s7_pointer old_port, port;
  old_port = sc->input_port;
  port = car(args);
  if ((is_input_port(port)) &&
      (!port_is_closed(port)))
    sc->input_port = port;
  else
    {
      check_method(sc, port, s7_make_symbol(sc, "set-current-input-port"), args);
      return(s7_wrong_type_arg_error(sc, "set-current-input-port", 0, port, "an open input port"));
    }
  return(old_port);
}
#endif

s7_pointer s7_set_current_input_port(s7_scheme *sc, s7_pointer port)
{
  s7_pointer old_port;
  old_port = sc->input_port;
  sc->input_port = port;
  return(old_port);
}


s7_pointer s7_current_output_port(s7_scheme *sc)
{
  return(sc->output_port);
}


s7_pointer s7_set_current_output_port(s7_scheme *sc, s7_pointer port)
{
  s7_pointer old_port;
  old_port = sc->output_port;
  sc->output_port = port;
  return(old_port);
}


static s7_pointer g_current_output_port(s7_scheme *sc, s7_pointer args)
{
  #define H_current_output_port "(current-output-port) returns the current output port"
  #define Q_current_output_port s7_make_signature(sc, 1, sc->is_output_port_symbol)
  return(sc->output_port);
}

#if (!WITH_PURE_S7)
static s7_pointer g_set_current_output_port(s7_scheme *sc, s7_pointer args)
{
  #define H_set_current_output_port "(set-current-output-port port) sets the current-output port to port and returns the previous value of the output port"
  #define Q_set_current_output_port s7_make_signature(sc, 2, sc->is_output_port_symbol, sc->is_output_port_symbol)

  s7_pointer old_port, port;
  old_port = sc->output_port;
  port = car(args);
  if (((is_output_port(port)) &&
       (!port_is_closed(port))) ||
      (port == sc->F))
    sc->output_port = port;
  else
    {
      check_method(sc, port, s7_make_symbol(sc, "set-current-output-port"), args);
      return(s7_wrong_type_arg_error(sc, "set-current-output-port", 0, port, "an open output port"));
    }
  return(old_port);
}
#endif

s7_pointer s7_current_error_port(s7_scheme *sc)
{
  return(sc->error_port);
}


s7_pointer s7_set_current_error_port(s7_scheme *sc, s7_pointer port)
{
  s7_pointer old_port;
  old_port = sc->error_port;
  sc->error_port = port;
  return(old_port);
}


static s7_pointer g_current_error_port(s7_scheme *sc, s7_pointer args)
{
  #define H_current_error_port "(current-error-port) returns the current error port"
  #define Q_current_error_port s7_make_signature(sc, 1, sc->is_output_port_symbol)
  return(sc->error_port);
}


static s7_pointer g_set_current_error_port(s7_scheme *sc, s7_pointer args)
{
  #define H_set_current_error_port "(set-current-error-port port) sets the current-error port to port and returns the previous value of the error port"
  #define Q_set_current_error_port s7_make_signature(sc, 2, sc->is_output_port_symbol, sc->is_output_port_symbol)
  s7_pointer old_port, port;

  old_port = sc->error_port;
  port = car(args);
  if (((is_output_port(port)) &&
       (!port_is_closed(port))) ||
      (port == sc->F))
    sc->error_port = port;
  else
    {
      check_method(sc, port, s7_make_symbol(sc, "set-current-error-port"), args);
      return(s7_wrong_type_arg_error(sc, "set-current-error-port", 0, port, "an open output port"));
    }
  return(old_port);
}


#if (!WITH_PURE_S7)
static s7_pointer g_is_char_ready(s7_scheme *sc, s7_pointer args)
{
  #define H_is_char_ready "(char-ready? (port (current-input-port))) returns #t if a character is ready for input on the given port"
  #define Q_is_char_ready s7_make_signature(sc, 2, sc->is_boolean_symbol, sc->is_input_port_symbol)
  if (is_not_null(args))
    {
      s7_pointer pt = car(args);
      if (!is_input_port(pt))
	method_or_bust_with_type_one_arg(sc, pt, sc->is_char_ready_symbol, args, an_input_port_string);
      if (port_is_closed(pt))
	return(simple_wrong_type_argument_with_type(sc, sc->is_char_ready_symbol, pt, an_open_port_string));

      if (is_function_port(pt))
	return((*(port_input_function(pt)))(sc, S7_IS_CHAR_READY, pt));
      return(make_boolean(sc, is_string_port(pt)));
    }
  return(make_boolean(sc, (is_input_port(sc->input_port)) && (is_string_port(sc->input_port))));
}

static bool is_char_ready_b_p(s7_pointer p) {return(g_is_char_ready(cur_sc, set_plist_1(cur_sc, p)) != cur_sc->F);}
#endif


static s7_pointer g_is_eof_object(s7_scheme *sc, s7_pointer args)
{
  #define H_is_eof_object "(eof-object? val) returns #t if val is the end-of-file object"
  #define Q_is_eof_object pl_bt
  check_boolean_method(sc, is_eof, sc->is_eof_object_symbol, args);
}

static bool s7_is_eof_object(s7_pointer p) {return(p == cur_sc->eof_object);}


static int32_t closed_port_read_char(s7_scheme *sc, s7_pointer port);
static s7_pointer closed_port_read_line(s7_scheme *sc, s7_pointer port, bool with_eol, bool copied);
static void closed_port_write_char(s7_scheme *sc, int32_t c, s7_pointer port);
static void closed_port_write_string(s7_scheme *sc, const char *str, int32_t len, s7_pointer port);
static void closed_port_display(s7_scheme *sc, const char *s, s7_pointer port);

void s7_close_input_port(s7_scheme *sc, s7_pointer p)
{
#if DEBUGGING
  if (!is_input_port(p))
    fprintf(stderr, "s7_close_input_port: %s\n", DISPLAY(p));
#endif
  if ((is_immutable_port(p)) ||
      ((is_input_port(p)) && (port_is_closed(p))))
    return;

  if (port_filename(p))
    {
      free(port_filename(p));
      port_filename(p) = NULL;
    }

  if (is_file_port(p))
    {
      if (port_file(p))
	{
	  fclose(port_file(p));
	  port_file(p) = NULL;
	}
    }
  else
    {
      if ((is_string_port(p)) &&
	  (port_gc_loc(p) != -1))
	s7_gc_unprotect_at(sc, port_gc_loc(p));
    }
  if (port_needs_free(p))
    {
      if (port_data(p))
	{
	  free(port_data(p));
	  port_data(p) = NULL;
	  port_data_size(p) = 0;
	}
      port_needs_free(p) = false;
    }

  port_read_character(p) = closed_port_read_char;
  port_read_line(p) = closed_port_read_line;
  port_write_character(p) = closed_port_write_char;
  port_write_string(p) = closed_port_write_string;
  port_display(p) = closed_port_display;
  port_is_closed(p) = true;
}


static s7_pointer g_close_input_port(s7_scheme *sc, s7_pointer args)
{
  s7_pointer pt;
  #define H_close_input_port "(close-input-port port) closes the port"
  #define Q_close_input_port s7_make_signature(sc, 2, sc->is_unspecified_symbol, sc->is_input_port_symbol)

  pt = car(args);
  if (!is_input_port(pt))
    method_or_bust_with_type_one_arg(sc, pt, sc->close_input_port_symbol, set_plist_1(sc, pt), an_input_port_string);
  if (!is_immutable_port(pt))
    s7_close_input_port(sc, pt);
  return(sc->unspecified);
}


void s7_flush_output_port(s7_scheme *sc, s7_pointer p)
{
  if ((!is_output_port(p)) ||
      (!is_file_port(p)) ||
      (port_is_closed(p)) ||
      (p == sc->F))
    return;

  if (port_file(p))
    {
      if (port_position(p) > 0)
	{
	  if (fwrite((void *)(port_data(p)), 1, port_position(p), port_file(p)) != port_position(p))
	    s7_warn(sc, 64, "fwrite trouble in flush-output-port\n");
	  port_position(p) = 0;
	}
      fflush(port_file(p));
    }
}

static s7_pointer g_flush_output_port(s7_scheme *sc, s7_pointer args)
{
  #define H_flush_output_port "(flush-output-port port) flushes the port"
  #define Q_flush_output_port s7_make_signature(sc, 2, sc->T, sc->is_output_port_symbol)
  s7_pointer pt;

  if (is_null(args))
    pt = sc->output_port;
  else pt = car(args);

  if (!is_output_port(pt))
    {
      if (pt == sc->F) return(pt);
      method_or_bust_with_type_one_arg(sc, pt, sc->flush_output_port_symbol, args, an_output_port_string);
    }
  s7_flush_output_port(sc, pt);
  return(pt);
}


static void close_output_port(s7_scheme *sc, s7_pointer p)
{
  if (is_file_port(p))
    {
      if (port_filename(p)) /* only a file (output) port has a filename */
	{
	  free(port_filename(p));
	  port_filename(p) = NULL;
	  port_filename_length(p) = 0;
	}

      if (port_file(p))
	{
	  if (port_position(p) > 0)
	    {
	      if (fwrite((void *)(port_data(p)), 1, port_position(p), port_file(p)) != port_position(p))
		s7_warn(sc, 64, "fwrite trouble in close-output-port\n");
	      port_position(p) = 0;
	    }
	  free(port_data(p));
	  fflush(port_file(p));
	  fclose(port_file(p));
	  port_file(p) = NULL;
	}
    }
  else
    {
      if ((is_string_port(p)) &&
	  (port_data(p)))
	{
	  free(port_data(p));
	  port_data(p) = NULL;
	  port_data_size(p) = 0;
	  port_needs_free(p) = false;
	}
    }
  port_read_character(p) = closed_port_read_char;
  port_read_line(p) = closed_port_read_line;
  port_write_character(p) = closed_port_write_char;
  port_write_string(p) = closed_port_write_string;
  port_display(p) = closed_port_display;
  port_is_closed(p) = true;
}

void s7_close_output_port(s7_scheme *sc, s7_pointer p)
{
  if ((is_immutable_port(p)) ||
      ((is_output_port(p)) && (port_is_closed(p))) ||
      (p == sc->F))
    return;
  close_output_port(sc, p);
}


static s7_pointer g_close_output_port(s7_scheme *sc, s7_pointer args)
{
  s7_pointer pt;
  #define H_close_output_port "(close-output-port port) closes the port"
  #define Q_close_output_port s7_make_signature(sc, 2, sc->is_unspecified_symbol, sc->is_output_port_symbol)

  pt = car(args);
  if (!is_output_port(pt))
    {
      if (pt == sc->F) return(sc->unspecified);
      method_or_bust_with_type_one_arg(sc, pt, sc->close_output_port_symbol, set_plist_1(sc, pt), an_output_port_string);
    }
  if (!(is_immutable_port(pt)))
    s7_close_output_port(sc, pt);
  return(sc->unspecified);
}


/* -------- read character functions -------- */

static int32_t file_read_char(s7_scheme *sc, s7_pointer port)
{
  return(fgetc(port_file(port)));
}


static int32_t function_read_char(s7_scheme *sc, s7_pointer port)
{
  return(character((*(port_input_function(port)))(sc, S7_READ_CHAR, port)));
}


static int32_t string_read_char(s7_scheme *sc, s7_pointer port)
{
  if (port_data_size(port) <= port_position(port)) /* port_string_length is 0 if no port string */
    return(EOF);
  return((unsigned char)port_data(port)[port_position(port)++]);
}


static int32_t output_read_char(s7_scheme *sc, s7_pointer port)
{
  simple_wrong_type_argument_with_type(sc, sc->read_char_symbol, port, an_input_port_string);
  return(0);
}


static int32_t closed_port_read_char(s7_scheme *sc, s7_pointer port)
{
  simple_wrong_type_argument_with_type(sc, sc->read_char_symbol, port, an_open_port_string);
  return(0);
}



/* -------- read line functions -------- */

static s7_pointer output_read_line(s7_scheme *sc, s7_pointer port, bool with_eol, bool copied)
{
  return(simple_wrong_type_argument_with_type(sc, sc->read_line_symbol, port, an_input_port_string));
}


static s7_pointer closed_port_read_line(s7_scheme *sc, s7_pointer port, bool with_eol, bool copied)
{
  return(simple_wrong_type_argument_with_type(sc, sc->read_line_symbol, port, an_open_port_string));
}


static s7_pointer function_read_line(s7_scheme *sc, s7_pointer port, bool with_eol, bool copied)
{
  return((*(port_input_function(port)))(sc, S7_READ_LINE, port));
}


static s7_pointer stdin_read_line(s7_scheme *sc, s7_pointer port, bool with_eol, bool copied)
{
  if (!sc->read_line_buf)
    {
      sc->read_line_buf_size = 1024;
      sc->read_line_buf = (char *)malloc(sc->read_line_buf_size * sizeof(char));
    }

  if (fgets(sc->read_line_buf, sc->read_line_buf_size, stdin))
    return(s7_make_string(sc, sc->read_line_buf)); /* fgets adds the trailing '\0' */
  return(s7_make_string_with_length(sc, NULL, 0));
}


static s7_pointer file_read_line(s7_scheme *sc, s7_pointer port, bool with_eol, bool copied)
{
  char *buf;
  int32_t read_size, previous_size = 0;

  if (!sc->read_line_buf)
    {
      sc->read_line_buf_size = 1024;
      sc->read_line_buf = (char *)malloc(sc->read_line_buf_size * sizeof(char));
    }

  buf = sc->read_line_buf;
  read_size = sc->read_line_buf_size;

  while (true)
    {
      char *p, *rtn;
      size_t len;

      p = fgets(buf, read_size, port_file(port));
      if (!p)
	return(sc->eof_object);

      rtn = strchr(buf, (int)'\n');
      if (rtn)
	{
	  port_line_number(port)++;
	  return(s7_make_string_with_length(sc, sc->read_line_buf, (with_eol) ? (previous_size + rtn - p + 1) : (previous_size + rtn - p)));
	}
      /* if no newline, then either at eof or need bigger buffer */
      len = strlen(sc->read_line_buf);

      if ((len + 1) < sc->read_line_buf_size)
	return(s7_make_string_with_length(sc, sc->read_line_buf, len));

      previous_size = sc->read_line_buf_size;
      sc->read_line_buf_size *= 2;
      sc->read_line_buf = (char *)realloc(sc->read_line_buf, sc->read_line_buf_size * sizeof(char));
      read_size = previous_size;
      previous_size -= 1;
      buf = (char *)(sc->read_line_buf + previous_size);
    }
  return(sc->eof_object);
}


static s7_pointer string_read_line(s7_scheme *sc, s7_pointer port, bool with_eol, bool copied)
{
  uint32_t i, port_start;
  unsigned char *port_str, *cur, *start;

  port_start = port_position(port);
  port_str = port_data(port);
  start = (unsigned char *)(port_str + port_start);

  cur = (unsigned char *)strchr((const char *)start, (int)'\n'); /* this can run off the end making valgrind unhappy, but I think it's innocuous */
  if (cur)
      {
	port_line_number(port)++;
	i = cur - port_str;
	port_position(port) = i + 1;
	if (copied)
	  return(s7_make_string_with_length(sc, (const char *)start, ((with_eol) ? i + 1 : i) - port_start));
	return(make_string_wrapper_with_length(sc, (char *)start, ((with_eol) ? i + 1 : i) - port_start));
      }
  i = port_data_size(port);
  port_position(port) = i;
  if (i <= port_start)         /* the < part can happen -- if not caught we try to create a string of length -1 -> segfault */
    return(sc->eof_object);

  if (copied)
    return(s7_make_string_with_length(sc, (const char *)start, i - port_start));
  return(make_string_wrapper_with_length(sc, (char *)start, i - port_start));
}


/* -------- write character functions -------- */

static void resize_port_data(s7_pointer pt, uint32_t new_size)
{
  uint32_t loc;
  loc = port_data_size(pt);
  if (new_size < loc) return;
  port_data_size(pt) = new_size;
  port_data(pt) = (unsigned char *)realloc(port_data(pt), new_size * sizeof(unsigned char));
  memclr((void *)(port_data(pt) + loc), new_size - loc);
}

static void string_write_char(s7_scheme *sc, int32_t c, s7_pointer pt)
{
  if (port_position(pt) >= port_data_size(pt))
    resize_port_data(pt, port_data_size(pt) * 2);
  port_data(pt)[port_position(pt)++] = c;
}

static void stdout_write_char(s7_scheme *sc, int32_t c, s7_pointer port)
{
  fputc(c, stdout);
}

static void stderr_write_char(s7_scheme *sc, int32_t c, s7_pointer port)
{
  fputc(c, stderr);
}

static void function_write_char(s7_scheme *sc, int32_t c, s7_pointer port)
{
  (*(port_output_function(port)))(sc, c, port);
}


#define PORT_DATA_SIZE 256
static void file_write_char(s7_scheme *sc, int32_t c, s7_pointer port)
{
  if (port_position(port) == PORT_DATA_SIZE)
    {
      if (fwrite((void *)(port_data(port)), 1, PORT_DATA_SIZE, port_file(port)) != PORT_DATA_SIZE)
	s7_warn(sc, 64, "fwrite trouble during write-char\n");
      port_position(port) = 0;
    }
  port_data(port)[port_position(port)++] = (unsigned char)c;
}


static void input_write_char(s7_scheme *sc, int32_t c, s7_pointer port)
{
  simple_wrong_type_argument_with_type(sc, sc->write_char_symbol, port, an_output_port_string);
}


static void closed_port_write_char(s7_scheme *sc, int32_t c, s7_pointer port)
{
  simple_wrong_type_argument_with_type(sc, sc->write_char_symbol, port, an_open_port_string);
}



/* -------- write string functions -------- */

static void input_write_string(s7_scheme *sc, const char *str, int32_t len, s7_pointer port)
{
  simple_wrong_type_argument_with_type(sc, sc->write_symbol, port, an_output_port_string);
}


static void closed_port_write_string(s7_scheme *sc, const char *str, int32_t len, s7_pointer port)
{
  simple_wrong_type_argument_with_type(sc, sc->write_symbol, port, an_open_port_string);
}


static void input_display(s7_scheme *sc, const char *s, s7_pointer port)
{
  simple_wrong_type_argument_with_type(sc, sc->write_symbol, port, an_output_port_string);
}

static void closed_port_display(s7_scheme *sc, const char *s, s7_pointer port)
{
  simple_wrong_type_argument_with_type(sc, sc->write_symbol, port, an_open_port_string);
}

static void stdout_write_string(s7_scheme *sc, const char *str, int32_t len, s7_pointer port)
{
  if (str[len] == '\0')
    fputs(str, stdout);
  else
    {
      int32_t i;
      for (i = 0; i < len; i++)
	fputc(str[i], stdout);
    }
}

static void stderr_write_string(s7_scheme *sc, const char *str, int32_t len, s7_pointer port)
{
  if (str[len] == '\0')
    fputs(str, stderr);
  else
    {
      int32_t i;
      for (i = 0; i < len; i++)
	fputc(str[i], stderr);
    }
}

static void string_write_string(s7_scheme *sc, const char *str, int32_t len, s7_pointer pt)
{
  uint32_t new_len;  /* len is known to be non-zero, str may not be 0-terminated */
  new_len = port_position(pt) + (uint32_t)len;
  if (new_len >= port_data_size(pt))
    resize_port_data(pt, new_len * 2);
  memcpy((void *)(port_data(pt) + port_position(pt)), (void *)str, len);
  /* memcpy is much faster than the equivalent while loop, and faster than using the 4-bytes-at-a-time shuffle */
  port_position(pt) = new_len;
}


static s7_pointer write_string_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
  check_for_substring_temp(sc, expr);
  return(f);
}


static void file_display(s7_scheme *sc, const char *s, s7_pointer port)
{
  if (s)
    {
      if (port_position(port) > 0)
	{
	  if (fwrite((void *)(port_data(port)), 1, port_position(port), port_file(port)) != port_position(port))
	    s7_warn(sc, 64, "fwrite trouble in display\n");
	  port_position(port) = 0;
	}
      if (fputs(s, port_file(port)) == EOF)
	s7_warn(sc, 64, "write to %s: %s\n", port_filename(port), strerror(errno));
    }
}

static void file_write_string(s7_scheme *sc, const char *str, int32_t len, s7_pointer pt)
{
  int32_t new_len;
  new_len = port_position(pt) + len;
  if (new_len >= PORT_DATA_SIZE)
    {
      if (port_position(pt) > 0)
	{
	  if (fwrite((void *)(port_data(pt)), 1, port_position(pt), port_file(pt)) != port_position(pt))
	    s7_warn(sc, 64, "fwrite trouble in write-string\n");
	  port_position(pt) = 0;
	}
      if (fwrite((void *)str, 1, len, port_file(pt)) != (size_t)len)
	s7_warn(sc, 64, "fwrite trouble in write-string\n");
    }
  else
    {
      memcpy((void *)(port_data(pt) + port_position(pt)), (void *)str, len);
      port_position(pt) = new_len;
    }
}

static void string_display(s7_scheme *sc, const char *s, s7_pointer port)
{
  if (s)
    string_write_string(sc, s, safe_strlen(s), port);
}


static void function_display(s7_scheme *sc, const char *s, s7_pointer port)
{
  if (s)
    {
      for (; *s; s++)
	(*(port_output_function(port)))(sc, *s, port);
    }
}

static void function_write_string(s7_scheme *sc, const char *str, int32_t len, s7_pointer pt)
{
  int32_t i;
  for (i = 0; i < len; i++)
    (*(port_output_function(pt)))(sc, str[i], pt);
}

static void stdout_display(s7_scheme *sc, const char *s, s7_pointer port)
{
  if (s) fputs(s, stdout);
}


static void stderr_display(s7_scheme *sc, const char *s, s7_pointer port)
{
  if (s) fputs(s, stderr);
}


static s7_pointer g_write_string(s7_scheme *sc, s7_pointer args)
{
  #define H_write_string "(write-string str port start end) writes str to port."
  #define Q_write_string s7_make_circular_signature(sc, 3, 4, sc->is_string_symbol, sc->is_string_symbol, sc->is_output_port_symbol, sc->is_integer_symbol)
  s7_pointer str, port;
  s7_int start = 0, end;

  str = car(args);
  if (!is_string(str))
    method_or_bust(sc, str, sc->write_string_symbol, args, T_STRING, 1);

  end = string_length(str);
  if (!is_null(cdr(args)))
    {
      s7_pointer inds;
      port = cadr(args);
      inds = cddr(args);
      if (!is_null(inds))
	{
	  s7_pointer p;
	  p = start_and_end(sc, sc->write_string_symbol, NULL, inds, args, 3, &start, &end);
	  if (p != sc->gc_nil) return(p);
	}
    }
  else port = sc->output_port;
  if (!is_output_port(port))
    {
      if (port == sc->F)
	{
	  s7_pointer x;
	  int32_t len;
	  if ((start == 0) && (end == string_length(str)))
	    return(str);
	  len = (int32_t)(end - start);
	  x = s7_make_string_with_length(sc, (char *)(string_value(str) + start), len);
	  string_value(x)[len] = 0;
	  return(x);
	}
      method_or_bust_with_type(sc, port, sc->write_string_symbol, args, an_output_port_string, 2);
    }

  if (start == 0)
    port_write_string(port)(sc, string_value(str), end, port);
  else port_write_string(port)(sc, (char *)(string_value(str) + start), (end - start), port);
  return(str);
}


/* -------- skip to newline readers -------- */

static token_t file_read_semicolon(s7_scheme *sc, s7_pointer pt)
{
  int32_t c;
  do (c = fgetc(port_file(pt))); while ((c != '\n') && (c != EOF));
  port_line_number(pt)++;
  if (c == EOF)
    return(TOKEN_EOF);
  return(token(sc));
}


static token_t string_read_semicolon(s7_scheme *sc, s7_pointer pt)
{
  const char *orig_str, *str;
  str = (const char *)(port_data(pt) + port_position(pt));
  orig_str = strchr(str, (int)'\n');
  if (!orig_str)
    {
      port_position(pt) = port_data_size(pt);
      return(TOKEN_EOF);
    }
  port_position(pt) += (orig_str - str + 1); /* + 1 because strchr leaves orig_str pointing at the newline */
  port_line_number(pt)++;
  return(token(sc));
}


/* -------- white space readers -------- */

static int32_t file_read_white_space(s7_scheme *sc, s7_pointer port)
{
  int32_t c;
  while (is_white_space(c = fgetc(port_file(port))))
    if (c == '\n')
      port_line_number(port)++;
  return(c);
}


static int32_t terminated_string_read_white_space(s7_scheme *sc, s7_pointer pt)
{
  const unsigned char *str;
  unsigned char c;
  /* here we know we have null termination and white_space[#\null] is false. */
  str = (const unsigned char *)(port_data(pt) + port_position(pt));

  while (white_space[c = *str++]) /* (let ((a 1)) a) -- 255 is not -1 = EOF */
    if (c == '\n')
      port_line_number(pt)++;
  if (c)
    port_position(pt) = str - port_data(pt);
  else port_position(pt) = port_data_size(pt);
  return((int32_t)c);
}


/* name (alphanumeric token) readers */

static void resize_strbuf(s7_scheme *sc, uint32_t needed_size)
{
  uint32_t i, old_size;
  old_size = sc->strbuf_size;
  while (sc->strbuf_size <= needed_size) sc->strbuf_size *= 2;
  sc->strbuf = (char *)realloc(sc->strbuf, sc->strbuf_size * sizeof(char));
  for (i = old_size; i < sc->strbuf_size; i++) sc->strbuf[i] = '\0';
}


static s7_pointer file_read_name_or_sharp(s7_scheme *sc, s7_pointer pt, bool atom_case)
{
  int32_t c;
  uint32_t i = 1;
  /* sc->strbuf[0] has the first char of the string we're reading */

  do {
    c = fgetc(port_file(pt)); /* might return EOF */
    if (c == '\n')
      port_line_number(pt)++;

    sc->strbuf[i++] = c;
    if (i >= sc->strbuf_size)
      resize_strbuf(sc, i);
  } while ((c != EOF) && (char_ok_in_a_name[c]));

  if ((i == 2) &&
      (sc->strbuf[0] == '\\'))
    sc->strbuf[2] = '\0';
  else
    {
      if (c != EOF)
	{
	  if (c == '\n')
	    port_line_number(pt)--;
	  ungetc(c, port_file(pt));
	}
      sc->strbuf[i - 1] = '\0';
    }

  if (atom_case)
    return(make_atom(sc, sc->strbuf, BASE_10, SYMBOL_OK, WITH_OVERFLOW_ERROR));

  return(make_sharp_constant(sc, sc->strbuf, BASE_10, WITH_OVERFLOW_ERROR));
}

static s7_pointer file_read_name(s7_scheme *sc, s7_pointer pt)
{
  return(file_read_name_or_sharp(sc, pt, true));
}

static s7_pointer file_read_sharp(s7_scheme *sc, s7_pointer pt)
{
  return(file_read_name_or_sharp(sc, pt, false));
}


static s7_pointer string_read_name_no_free(s7_scheme *sc, s7_pointer pt)
{
  /* sc->strbuf[0] has the first char of the string we're reading */
  uint32_t k;
  char *str, *orig_str;

  str = (char *)(port_data(pt) + port_position(pt));

  if (!char_ok_in_a_name[(unsigned char)*str])
    {
      s7_pointer result;
      result = sc->singletons[(unsigned char)(sc->strbuf[0])];
      if (!result)
	{
	  sc->strbuf[1] = '\0';
	  result = make_symbol_with_length(sc, sc->strbuf, 1);
	  sc->singletons[(unsigned char)(sc->strbuf[0])] = result;
	}
      return(result);
    }

  orig_str = (char *)(str - 1);
  str++;
  while (char_ok_in_a_name[(unsigned char)(*str)]) {str++;}
  k = str - orig_str;
  if (*str != 0)
    port_position(pt) += (k - 1);
  else port_position(pt) = port_data_size(pt);

  /* this is equivalent to:
   *    str = strpbrk(str, "(); \"\t\r\n");
   *    if (!str)
   *      {
   *        k = strlen(orig_str);
   *        str = (char *)(orig_str + k);
   *      }
   *    else k = str - orig_str;
   * but slightly faster.
   */

  if (!number_table[(unsigned char)(*orig_str)])
    return(make_symbol_with_length(sc, orig_str, k));

  /* eval_c_string string is a constant so we can't set and unset the token's end char */
  if ((k + 1) >= sc->strbuf_size)
    resize_strbuf(sc, k + 1);

  memcpy((void *)(sc->strbuf), (void *)orig_str, k);
  sc->strbuf[k] = '\0';
  return(make_atom(sc, sc->strbuf, BASE_10, SYMBOL_OK, WITH_OVERFLOW_ERROR));
}


static s7_pointer string_read_sharp(s7_scheme *sc, s7_pointer pt)
{
  /* sc->strbuf[0] has the first char of the string we're reading.
   *   since a *#readers* function might want to get further input, we can't mess with the input even when it is otherwise safe
   */
  uint32_t k;
  char *orig_str, *str;

  str = (char *)(port_data(pt) + port_position(pt));

  if (!char_ok_in_a_name[(unsigned char)*str])
    {
      if (sc->strbuf[0] == 'f')
	return(sc->F);
      if (sc->strbuf[0] == 't')
	return(sc->T);
      if (sc->strbuf[0] == '\\')
	{
	  /* must be from #\( and friends -- a character that happens to be not ok-in-a-name */
	  sc->strbuf[1] = str[0];
	  sc->strbuf[2] = '\0';
	  port_position(pt)++;
	}
      else sc->strbuf[1] = '\0';
      return(make_sharp_constant(sc, sc->strbuf, BASE_10, WITH_OVERFLOW_ERROR));
    }

  orig_str = (char *)(str - 1);
  str++;
  while (char_ok_in_a_name[(unsigned char)(*str)]) {str++;}
  k = str - orig_str;
  if (*str != 0)
    port_position(pt) += (k - 1);
  else port_position(pt) += k;

  if ((k + 1) >= sc->strbuf_size)
    resize_strbuf(sc, k + 1);

  memcpy((void *)(sc->strbuf), (void *)orig_str, k);
  sc->strbuf[k] = '\0';
  return(make_sharp_constant(sc, sc->strbuf, BASE_10, WITH_OVERFLOW_ERROR));
}


static s7_pointer string_read_name(s7_scheme *sc, s7_pointer pt)
{
  /* port_string was allocated (and read from a file) so we can mess with it directly */
  s7_pointer result;
  uint32_t k;
  char *orig_str, *str;
  char endc;

  str = (char *)(port_data(pt) + port_position(pt));
  if (!char_ok_in_a_name[(unsigned char)*str])
    {
      s7_pointer result;
      result = sc->singletons[(unsigned char)(sc->strbuf[0])];
      if (!result)
	{
	  sc->strbuf[1] = '\0';
	  result = make_symbol_with_length(sc, sc->strbuf, 1);
	  sc->singletons[(unsigned char)(sc->strbuf[0])] = result;
	}
      return(result);
    }

  orig_str = (char *)(str - 1);
  str++;
  while (char_ok_in_a_name[(unsigned char)(*str)]) {str++;}
  k = str - orig_str;
  if (*str != 0)
    port_position(pt) += (k - 1);
  else port_position(pt) = port_data_size(pt);

  if (!number_table[(unsigned char)(*orig_str)])
    return(make_symbol_with_length(sc, orig_str, k));

  endc = (*str);
  (*str) = '\0';
  result = make_atom(sc, orig_str, BASE_10, SYMBOL_OK, WITH_OVERFLOW_ERROR);
  (*str) = endc;
  return(result);
}


static s7_pointer read_file(s7_scheme *sc, FILE *fp, const char *name, int64_t max_size, const char *caller)
{
  s7_pointer port;
#if (!MS_WINDOWS)
  int64_t size;
#endif
  uint32_t port_loc;

  new_cell(sc, port, T_INPUT_PORT);
  port_loc = s7_gc_protect(sc, port);
  port_port(port) = alloc_port(sc);
  port_is_closed(port) = false;
  port_original_input_string(port) = sc->nil;
  port_write_character(port) = input_write_char;
  port_write_string(port) = input_write_string;

  /* if we're constantly opening files, and each open saves the file name in permanent
   *   memory, we gradually core-up.
   */
  port_filename_length(port) = safe_strlen(name);
  port_filename(port) = copy_string_with_length(name, port_filename_length(port));
  port_line_number(port) = 1;  /* first line is numbered 1 */
  add_input_port(sc, port);

#if (!MS_WINDOWS)
  /* this doesn't work in MS C */
  fseek(fp, 0, SEEK_END);
  size = ftell(fp);
  rewind(fp);

  /* pseudo files (under /proc for example) have size=0, but we can read them, so don't assume a 0 length file is empty
   */

  if ((size > 0) &&                          /* if (size != 0) we get (open-input-file "/dev/tty") -> (open "/dev/tty") read 0 bytes of an expected -1? */
      ((max_size < 0) || (size < max_size))) /* load uses max_size = -1 */
    {
      size_t bytes;
      unsigned char *content;

      content = (unsigned char *)malloc((size + 2) * sizeof(unsigned char));
      bytes = fread(content, sizeof(unsigned char), size, fp);
      if (bytes != (size_t)size)
	{
	  char tmp[256];
	  int32_t len;
	  len = snprintf(tmp, 256, "(%s \"%s\") read %ld bytes of an expected %" PRId64 "?", caller, name, (long)bytes, size);
	  port_write_string(sc->output_port)(sc, tmp, len, sc->output_port);
	  size = bytes;
	}
      content[size] = '\0';
      content[size + 1] = '\0';
      fclose(fp);

      port_type(port) = STRING_PORT;
      port_data(port) = content;
      port_data_size(port) = size;
      port_position(port) = 0;
      port_needs_free(port) = true;
      port_gc_loc(port) = -1;
      port_read_character(port) = string_read_char;
      port_read_line(port) = string_read_line;
      port_display(port) = input_display;
      port_read_semicolon(port) = string_read_semicolon;
      port_read_white_space(port) = terminated_string_read_white_space;
      port_read_name(port) = string_read_name;
      port_read_sharp(port) = string_read_sharp;
    }
  else
    {
      port_file(port) = fp;
      port_type(port) = FILE_PORT;
      port_needs_free(port) = false;
      port_read_character(port) = file_read_char;
      port_read_line(port) = file_read_line;
      port_display(port) = input_display;
      port_read_semicolon(port) = file_read_semicolon;
      port_read_white_space(port) = file_read_white_space;
      port_read_name(port) = file_read_name;
      port_read_sharp(port) = file_read_sharp; /* was string_read_sharp?? */
    }
#else
  /* _stat64 is no better than the fseek/ftell route, and
   *    GetFileSizeEx and friends requires Windows.h which makes hash of everything else.
   *    fread until done takes too long on big files, so use a file port
   */
  port_file(port) = fp;
  port_type(port) = FILE_PORT;
  port_needs_free(port) = false;
  port_read_character(port) = file_read_char;
  port_read_line(port) = file_read_line;
  port_display(port) = input_display;
  port_read_semicolon(port) = file_read_semicolon;
  port_read_white_space(port) = file_read_white_space;
  port_read_name(port) = file_read_name;
  port_read_sharp(port) = file_read_sharp;
#endif

  s7_gc_unprotect_at(sc, port_loc);
  return(port);
}


static s7_pointer make_input_file(s7_scheme *sc, const char *name, FILE *fp)
{
  #define MAX_SIZE_FOR_STRING_PORT 5000000
  return(read_file(sc, fp, name, MAX_SIZE_FOR_STRING_PORT, "open"));
}

#if (!MS_WINDOWS)
#include <sys/stat.h>
#endif

static bool is_directory(const char *filename)
{
#if (!MS_WINDOWS)
  #ifdef S_ISDIR
    struct stat statbuf;
    return((stat(filename, &statbuf) >= 0) &&
	   (S_ISDIR(statbuf.st_mode)));
  #endif
#endif
  return(false);
}


static s7_pointer open_input_file_1(s7_scheme *sc, const char *name, const char *mode, const char *caller)
{
  FILE *fp;
  /* see if we can open this file before allocating a port */

  if (is_directory(name))
    return(file_error(sc, caller, "is a directory", name));

  errno = 0;
  fp = fopen(name, mode);
  if (!fp)
    {
#if (!MS_WINDOWS)
      if (errno == EINVAL)
	return(file_error(sc, caller, "invalid mode", mode));
  #if WITH_GCC
      /* catch one special case, "~/..." */
      if ((name[0] == '~') &&
	  (name[1] == '/'))
	{
	  char *home;
	  home = getenv("HOME");
	  if (home)
	    {
	      char *filename;
	      int32_t len;
	      len = safe_strlen(name) + safe_strlen(home) + 1;
	      tmpbuf_malloc(filename, len);
	      snprintf(filename, len, "%s%s", home, (char *)(name + 1));
	      fp = fopen(filename, "r");
	      tmpbuf_free(filename, len);
	      if (fp)
		return(make_input_file(sc, name, fp));
	    }
	}
  #endif
#endif
      return(file_error(sc, caller, strerror(errno), name));
    }
  return(make_input_file(sc, name, fp));
}


s7_pointer s7_open_input_file(s7_scheme *sc, const char *name, const char *mode)
{
  return(open_input_file_1(sc, name, mode, "open-input-file"));
}


static s7_pointer g_open_input_file(s7_scheme *sc, s7_pointer args)
{
  #define H_open_input_file "(open-input-file filename (mode \"r\")) opens filename for reading"
  #define Q_open_input_file s7_make_signature(sc, 3, sc->is_input_port_symbol, sc->is_string_symbol, sc->is_string_symbol)
  s7_pointer name = car(args);
  /* open-input-file can create a new output file if the file to be opened does not exist, and the "a" mode is given */

  if (!is_string(name))
    method_or_bust(sc, name, sc->open_input_file_symbol, args, T_STRING, 1);
  /* what if the file name is a byte-vector? currently we accept it */

  if (is_pair(cdr(args)))
    {
      s7_pointer mode;
      mode = cadr(args);
      if (!is_string(mode))
	method_or_bust_with_type(sc, mode, sc->open_input_file_symbol, args, s7_make_string_wrapper(sc, "a string (a mode such as \"r\")"), 2);
      /* since scheme allows embedded nulls, dumb stuff is accepted here: (open-input-file file "a\x00b") -- should this be an error? */
      return(open_input_file_1(sc, string_value(name), string_value(mode), "open-input-file"));
    }
  return(open_input_file_1(sc, string_value(name), "r", "open-input-file"));
}

static void make_standard_ports(s7_scheme *sc)
{
  s7_pointer x;

  /* standard output */
  x = alloc_pointer();
  unheap(x);
  set_type(x, T_OUTPUT_PORT | T_IMMUTABLE);
  port_port(x) = (port_t *)calloc(1, sizeof(port_t));
  port_type(x) = FILE_PORT;
  port_data(x) = NULL;
  port_is_closed(x) = false;
  port_filename_length(x) = 8;
  port_filename(x) = copy_string_with_length("*stdout*", 8);
  port_file_number(x) = remember_file_name(sc, port_filename(x)); /* these numbers need to be correct for the evaluator (__FUNC__ data) */
  port_line_number(x) = 0;
  port_file(x) = stdout;
  port_needs_free(x) = false;
  port_read_character(x) = output_read_char;
  port_read_line(x) = output_read_line;
  port_display(x) = stdout_display;
  port_write_character(x) = stdout_write_char;
  port_write_string(x) = stdout_write_string;
  sc->standard_output = x;

  /* standard error */
  x = alloc_pointer();
  unheap(x);
  set_type(x, T_OUTPUT_PORT | T_IMMUTABLE);
  port_port(x) = (port_t *)calloc(1, sizeof(port_t));
  port_type(x) = FILE_PORT;
  port_data(x) = NULL;
  port_is_closed(x) = false;
  port_filename_length(x) = 8;
  port_filename(x) = copy_string_with_length("*stderr*", 8);
  port_file_number(x) = remember_file_name(sc, port_filename(x));
  port_line_number(x) = 0;
  port_file(x) = stderr;
  port_needs_free(x) = false;
  port_read_character(x) = output_read_char;
  port_read_line(x) = output_read_line;
  port_display(x) = stderr_display;
  port_write_character(x) = stderr_write_char;
  port_write_string(x) = stderr_write_string;
  sc->standard_error = x;

  /* standard input */
  x = alloc_pointer();
  unheap(x);
  set_type(x, T_INPUT_PORT | T_IMMUTABLE);
  port_port(x) = (port_t *)calloc(1, sizeof(port_t));
  port_type(x) = FILE_PORT;
  port_is_closed(x) = false;
  port_original_input_string(x) = sc->nil;
  port_filename_length(x) = 7;
  port_filename(x) = copy_string_with_length("*stdin*", 7);
  port_file_number(x) = remember_file_name(sc, port_filename(x));
  port_line_number(x) = 0;
  port_file(x) = stdin;
  port_needs_free(x) = false;
  port_read_character(x) = file_read_char;
  port_read_line(x) = stdin_read_line;
  port_display(x) = input_display;
  port_read_semicolon(x) = file_read_semicolon;
  port_read_white_space(x) = file_read_white_space;
  port_read_name(x) = file_read_name;
  port_read_sharp(x) = file_read_sharp;
  port_write_character(x) = input_write_char;
  port_write_string(x) = input_write_string;
  sc->standard_input = x;

  s7_define_constant(sc, "*stdin*", sc->standard_input);
  s7_define_constant(sc, "*stdout*", sc->standard_output);
  s7_define_constant(sc, "*stderr*", sc->standard_error);

  sc->input_port = sc->standard_input;
  sc->output_port = sc->standard_output;
  sc->error_port = sc->standard_error;
  sc->current_file = NULL;
  sc->current_line = -1;
}


s7_pointer s7_open_output_file(s7_scheme *sc, const char *name, const char *mode)
{
  FILE *fp;
  s7_pointer x;
  /* see if we can open this file before allocating a port */

  errno = 0;
  fp = fopen(name, mode);
  if (!fp)
    {
#if (!MS_WINDOWS)
      if (errno == EINVAL)
	return(file_error(sc, "open-output-file", "invalid mode", mode));
#endif
      return(file_error(sc, "open-output-file", strerror(errno), name));
    }

  new_cell(sc, x, T_OUTPUT_PORT);
  port_port(x) = alloc_port(sc);
  port_type(x) = FILE_PORT;
  port_is_closed(x) = false;
  port_filename_length(x) = safe_strlen(name);
  port_filename(x) = copy_string_with_length(name, port_filename_length(x));
  port_line_number(x) = 1;
  port_file(x) = fp;
  port_needs_free(x) = false;
  port_read_character(x) = output_read_char;
  port_read_line(x) = output_read_line;
  port_display(x) = file_display;
  port_write_character(x) = file_write_char;
  port_write_string(x) = file_write_string;
  port_position(x) = 0;
  port_data_size(x) = PORT_DATA_SIZE;
  port_data(x) = (unsigned char *)malloc(PORT_DATA_SIZE); /* was +8? */
  add_output_port(sc, x);
  return(x);
}


static s7_pointer g_open_output_file(s7_scheme *sc, s7_pointer args)
{
  #define H_open_output_file "(open-output-file filename (mode \"w\")) opens filename for writing"
  #define Q_open_output_file s7_make_signature(sc, 3, sc->is_output_port_symbol, sc->is_string_symbol, sc->is_string_symbol)
  s7_pointer name = car(args);

  if (!is_string(name))
    method_or_bust(sc, name, sc->open_output_file_symbol, args, T_STRING, 1);

  if (is_pair(cdr(args)))
    {
      if (!is_string(cadr(args)))
	method_or_bust_with_type(sc, cadr(args), sc->open_output_file_symbol, args, s7_make_string_wrapper(sc, "a string (a mode such as \"w\")"), 2);
      return(s7_open_output_file(sc, string_value(name), string_value(cadr(args))));
    }
  return(s7_open_output_file(sc, string_value(name), "w"));
}

static s7_pointer open_input_string(s7_scheme *sc, const char *input_string, int32_t len)
{
  s7_pointer x;
  new_cell(sc, x, T_INPUT_PORT);
  port_port(x) = alloc_port(sc);
  port_type(x) = STRING_PORT;
  port_is_closed(x) = false;
  port_original_input_string(x) = sc->nil;
  port_data(x) = (unsigned char *)input_string;
  port_data_size(x) = len;
  port_position(x) = 0;
  port_filename_length(x) = 0;
  port_filename(x) = NULL;
  port_file_number(x) = 0; /* uint32_t */
  port_line_number(x) = 0;
  port_needs_free(x) = false;
  port_gc_loc(x) = -1;
  port_read_character(x) = string_read_char;
  port_read_line(x) = string_read_line;
  port_display(x) = input_display;
  port_read_semicolon(x) = string_read_semicolon;
#if DEBUGGING
  if (input_string[len] != '\0')
    fprintf(stderr, "read_white_space string is not terminated: %s", input_string);
#endif
  port_read_white_space(x) = terminated_string_read_white_space;
  port_read_name(x) = string_read_name_no_free;
  port_read_sharp(x) = string_read_sharp;
  port_write_character(x) = input_write_char;
  port_write_string(x) = input_write_string;
  add_input_port(sc, x);
  return(x);
}


static s7_pointer open_and_protect_input_string(s7_scheme *sc, s7_pointer str)
{
  s7_pointer p;
  p = open_input_string(sc, string_value(str), string_length(str));
  port_gc_loc(p) = s7_gc_protect(sc, str);
  return(p);
}


s7_pointer s7_open_input_string(s7_scheme *sc, const char *input_string)
{
  return(open_input_string(sc, input_string, safe_strlen(input_string)));
}


static s7_pointer g_open_input_string(s7_scheme *sc, s7_pointer args)
{
  #define H_open_input_string "(open-input-string str) opens an input port reading str"
  #define Q_open_input_string s7_make_signature(sc, 2, sc->is_input_port_symbol, sc->is_string_symbol)
  s7_pointer input_string, port;

  input_string = car(args);
  if (!is_string(input_string))
    method_or_bust_one_arg(sc, input_string, sc->open_input_string_symbol, args, T_STRING);
  port = open_and_protect_input_string(sc, input_string);
  return(port);
}

#define FORMAT_PORT_LENGTH 128
/* the large majority (> 99% in my tests) of the output strings have less than 128 chars when the port is finally closed
 *   256 is slightly slower (the calloc time below dominates the realloc time in string_write_string)
 *   64 is much slower (realloc dominates)
 */

static s7_pointer open_output_string(s7_scheme *sc, int32_t len)
{
  s7_pointer x;
  new_cell(sc, x, T_OUTPUT_PORT);
  port_port(x) = alloc_port(sc);
  port_type(x) = STRING_PORT;
  port_is_closed(x) = false;
  port_data_size(x) = len;
  port_data(x) = (unsigned char *)malloc(len * sizeof(unsigned char)); /* was +8? */
  port_data(x)[0] = '\0';   /* in case s7_get_output_string before any output */
  port_position(x) = 0;
  port_needs_free(x) = true;
  port_read_character(x) = output_read_char;
  port_read_line(x) = output_read_line;
  port_display(x) = string_display;
  port_write_character(x) = string_write_char;
  port_write_string(x) = string_write_string;
  add_output_port(sc, x);
  return(x);
}

s7_pointer s7_open_output_string(s7_scheme *sc) {return(open_output_string(sc, sc->initial_string_port_length));}

static s7_pointer open_output_string_p(void) {return(s7_open_output_string(cur_sc));}

static s7_pointer g_open_output_string(s7_scheme *sc, s7_pointer args)
{
  #define H_open_output_string "(open-output-string) opens an output string port"
  #define Q_open_output_string s7_make_signature(sc, 1, sc->is_output_port_symbol)
  return(s7_open_output_string(sc));
}


const char *s7_get_output_string(s7_scheme *sc, s7_pointer p)
{
  port_data(p)[port_position(p)] = '\0';
  return((const char *)port_data(p));
}


static s7_pointer g_get_output_string(s7_scheme *sc, s7_pointer args)
{
  #define H_get_output_string "(get-output-string port clear-port) returns the output accumulated in port.  \
If the optional 'clear-port' is #t, the current string is flushed."
  #define Q_get_output_string s7_make_signature(sc, 3, sc->is_string_symbol, sc->is_output_port_symbol, sc->is_boolean_symbol)

  s7_pointer p, result;
  bool clear_port = false;

  if (is_pair(cdr(args)))
    {
      p = cadr(args);
      if (!s7_is_boolean(p))
	return(wrong_type_argument(sc, sc->get_output_string_symbol, 2, p, T_BOOLEAN));
      clear_port = (p == sc->T);
    }
  p = car(args);
  if ((!is_output_port(p)) ||
      (!is_string_port(p)))
    {
      if (p == sc->F) return(make_empty_string(sc, 0, 0));
      method_or_bust_with_type_one_arg(sc, p, sc->get_output_string_symbol, args, s7_make_string_wrapper(sc, "an output string port"));
    }
  if (port_is_closed(p))
    return(simple_wrong_type_argument_with_type(sc, sc->get_output_string_symbol, p, s7_make_string_wrapper(sc, "an active (open) string port")));

  result = s7_make_string_with_length(sc, (const char *)port_data(p), port_position(p));
  if (clear_port)
    {
      port_position(p) = 0;
      port_data(p)[0] = '\0';
    }
  return(result);
}

s7_pointer s7_open_input_function(s7_scheme *sc, s7_pointer (*function)(s7_scheme *sc, s7_read_t read_choice, s7_pointer port))
{
  s7_pointer x;
  new_cell(sc, x, T_INPUT_PORT);
  port_port(x) = alloc_port(sc);
  port_type(x) = FUNCTION_PORT;
  port_is_closed(x) = false;
  port_original_input_string(x) = sc->nil;
  port_needs_free(x) = false;
  port_input_function(x) = function;
  port_read_character(x) = function_read_char;
  port_read_line(x) = function_read_line;
  port_display(x) = input_display;
  port_write_character(x) = input_write_char;
  port_write_string(x) = input_write_string;
  add_input_port(sc, x);
  return(x);
}


s7_pointer s7_open_output_function(s7_scheme *sc, void (*function)(s7_scheme *sc, unsigned char c, s7_pointer port))
{
  s7_pointer x;
  new_cell(sc, x, T_OUTPUT_PORT);
  port_port(x) = alloc_port(sc);
  port_type(x) = FUNCTION_PORT;
  port_data(x) = NULL;
  port_is_closed(x) = false;
  port_needs_free(x) = false;
  port_output_function(x) = function;
  port_read_character(x) = output_read_char;
  port_read_line(x) = output_read_line;
  port_display(x) = function_display;
  port_write_character(x) = function_write_char;
  port_write_string(x) = function_write_string;
  add_output_port(sc, x);
  return(x);
}


static void push_input_port(s7_scheme *sc, s7_pointer new_port)
{
  sc->input_port_stack = cons(sc, sc->input_port, sc->input_port_stack);
  sc->input_port = new_port;
}


static void pop_input_port(s7_scheme *sc)
{
  if (is_pair(sc->input_port_stack))
    {
      s7_pointer nxt;
      sc->input_port = car(sc->input_port_stack);
      nxt = cdr(sc->input_port_stack);
      /* is this safe? */
      free_cell(sc, sc->input_port_stack); 
      sc->input_port_stack = nxt;
    }
  else sc->input_port = sc->standard_input;
}

#if DEBUGGING && (0)
static void print_input_stack(s7_scheme *sc)
{
  if (is_null(sc->input_port_stack))
    fprintf(stderr, "no input stack\n");
  else
    {
      s7_pointer p;
      for (p = sc->input_port_stack; is_pair(p); p = cdr(p))
	fprintf(stderr, "%s\n", s7_object_to_c_string(sc, car(p)));
    }
}
#endif


static int32_t inchar(s7_pointer pt)
{
  int32_t c;
  if (is_file_port(pt))
    c = fgetc(port_file(pt)); /* not unsigned char! -- could be EOF */
  else
    {
      if (port_data_size(pt) <= port_position(pt))
	return(EOF);
      c = (unsigned char)port_data(pt)[port_position(pt)++];
    }

  if (c == '\n')
    port_line_number(pt)++;

  return(c);
}


static void backchar(char c, s7_pointer pt)
{
  if (c == '\n')
    port_line_number(pt)--;

  if (is_file_port(pt))
    ungetc(c, port_file(pt));
  else
    {
      if (port_position(pt) > 0)
	port_position(pt)--;
    }
}


int s7_read_char(s7_scheme *sc, s7_pointer port)
{
  /* needs to be int32_t return value so EOF=-1, but not 255 */
  return(port_read_character(port)(sc, port));
}


int s7_peek_char(s7_scheme *sc, s7_pointer port)
{
  int32_t c;              /* needs to be an int32_t so EOF=-1, but not 255 */
  c = port_read_character(port)(sc, port);
  if (c != EOF)
    backchar(c, port);
  return(c);
}


void s7_write_char(s7_scheme *sc, int32_t c, s7_pointer pt)
{
  if (pt != sc->F)
    port_write_character(pt)(sc, c, pt);
}


static s7_pointer input_port_if_not_loading(s7_scheme *sc)
{
  s7_pointer port;
  port = sc->input_port;
  if (is_loader_port(port)) /* this flag is turned off by the reader macros, so we aren't in that context */
    {
      int32_t c;
      c = port_read_white_space(port)(sc, port);
      if (c > 0)            /* we can get either EOF or NULL at the end */
	{
	  backchar(c, port);
	  return(NULL);
	}
      return(sc->standard_input);
    }
  return(port);
}

static s7_pointer g_read_char(s7_scheme *sc, s7_pointer args)
{
  #define H_read_char "(read-char (port (current-input-port))) returns the next character in the input port"
  #define Q_read_char s7_make_signature(sc, 2, s7_make_signature(sc, 2, sc->is_char_symbol, sc->is_eof_object_symbol), sc->is_input_port_symbol)
  s7_pointer port;

  if (is_not_null(args))
    port = car(args);
  else
    {
      port = input_port_if_not_loading(sc);
      if (!port) return(sc->eof_object);
    }
  if (!is_input_port(port))
    method_or_bust_with_type_one_arg(sc, port, sc->read_char_symbol, args, an_input_port_string);
  return(chars[port_read_character(port)(sc, port)]);
}


static s7_pointer g_write_char(s7_scheme *sc, s7_pointer args)
{
  #define H_write_char "(write-char char (port (current-output-port))) writes char to the output port"
  #define Q_write_char s7_make_signature(sc, 3, sc->is_char_symbol, sc->is_char_symbol, sc->is_output_port_symbol)
  s7_pointer port, chr;

  chr = car(args);
  if (!s7_is_character(chr))
    method_or_bust(sc, chr, sc->write_char_symbol, args, T_CHARACTER, 1);

  if (is_pair(cdr(args)))
    port = cadr(args);
  else port = sc->output_port;
  if (port == sc->F) return(chr);
  if (!is_output_port(port))
    method_or_bust_with_type(sc, port, sc->write_char_symbol, args, an_output_port_string, 2);

  port_write_character(port)(sc, s7_character(chr), port);
  return(chr);
}

static s7_pointer write_char_p_p(s7_pointer c)
{
  if (!s7_is_character(c))
    simple_wrong_type_argument(cur_sc, cur_sc->write_char_symbol, c, T_CHARACTER);
  if (cur_sc->output_port == cur_sc->F) return(c);
  port_write_character(cur_sc->output_port)(cur_sc, s7_character(c), cur_sc->output_port);
  return(c);
}

static s7_pointer write_char_p_pp(s7_pointer c, s7_pointer port)
{
  if (!s7_is_character(c))
    simple_wrong_type_argument(cur_sc, cur_sc->write_char_symbol, c, T_CHARACTER);
  if (port == cur_sc->F) return(c);
  if (!is_output_port(port))
    simple_wrong_type_argument_with_type(cur_sc, cur_sc->write_char_symbol, port, an_output_port_string);
  port_write_character(port)(cur_sc, s7_character(c), port);
  return(c);
}

/* (with-output-to-string (lambda () (write-char #\space))) -> " "
 * (with-output-to-string (lambda () (write #\space))) -> "#\\space"
 * (with-output-to-string (lambda () (display #\space))) -> " "
 * is this correct?  It's what Guile does.  write-char is actually display-char.
 */


static s7_pointer g_peek_char(s7_scheme *sc, s7_pointer args)
{
  #define H_peek_char "(peek-char (port (current-input-port))) returns the next character in the input port, but does not remove it from the input stream"
  #define Q_peek_char s7_make_signature(sc, 2, s7_make_signature(sc, 2, sc->is_char_symbol, sc->is_eof_object_symbol), sc->is_input_port_symbol)
  s7_pointer port;

  if (is_not_null(args))
    port = car(args);
  else port = sc->input_port;

  if (!is_input_port(port))
    method_or_bust_with_type_one_arg(sc, port, sc->peek_char_symbol, args, an_input_port_string);
  if (port_is_closed(port))
    return(simple_wrong_type_argument_with_type(sc, sc->peek_char_symbol, port, an_open_port_string));

  if (is_function_port(port))
    return((*(port_input_function(port)))(sc, S7_PEEK_CHAR, port));
  return(chars[s7_peek_char(sc, port)]);
}


static s7_pointer g_read_byte(s7_scheme *sc, s7_pointer args)
{
  #define H_read_byte "(read-byte (port (current-input-port))): reads a byte from the input port"
  #define Q_read_byte s7_make_signature(sc, 2, s7_make_signature(sc, 2, sc->is_integer_symbol, sc->is_eof_object_symbol), sc->is_input_port_symbol)
  s7_pointer port;
  int32_t c;

  if (is_not_null(args))
    port = car(args);
  else
    {
      port = input_port_if_not_loading(sc);
      if (!port) return(sc->eof_object);
    }
  if (!is_input_port(port))
    method_or_bust_with_type_one_arg(sc, port, sc->read_byte_symbol, args, an_input_port_string);

  c = port_read_character(port)(sc, port);
  if (c == EOF)
    return(sc->eof_object);
  return(small_int(c));
}


static s7_pointer g_write_byte(s7_scheme *sc, s7_pointer args)
{
  #define H_write_byte "(write-byte byte (port (current-output-port))): writes byte to the output port"
  #define Q_write_byte s7_make_signature(sc, 3, sc->is_integer_symbol, sc->is_integer_symbol, sc->is_output_port_symbol)
  s7_pointer port, b;
  s7_int val;

  b = car(args);
  if (!s7_is_integer(b))
    method_or_bust(sc, car(args), sc->write_byte_symbol, args, T_INTEGER, 1);

  val = s7_integer(b);
  if ((val < 0) || (val > 255)) /* need to check this before port==#f, else (write-byte most-positive-fixnum #f) is not an error */
    return(wrong_type_argument_with_type(sc, sc->write_byte_symbol, 1, b, an_unsigned_byte_string));

  if (is_pair(cdr(args)))
    port = cadr(args);
  else port = sc->output_port;

  if (!is_output_port(port))
    {
      if (port == sc->F) return(car(args));
      method_or_bust_with_type_one_arg(sc, port, sc->write_byte_symbol, args, an_output_port_string);
    }

  s7_write_char(sc, (int32_t)val, port);
  return(b);
}


static s7_pointer g_read_line(s7_scheme *sc, s7_pointer args)
{
  #define H_read_line "(read-line port (with-eol #f)) returns the next line from port, or #<eof>.\
If 'with-eol' is not #f, read-line includes the trailing end-of-line character."
  #define Q_read_line s7_make_signature(sc, 3, s7_make_signature(sc, 2, sc->is_string_symbol, sc->is_eof_object_symbol), sc->is_input_port_symbol, sc->is_boolean_symbol)

  s7_pointer port;
  bool with_eol = false;

  if (is_not_null(args))
    {
      port = car(args);
      if (!is_input_port(port))
	method_or_bust_with_type(sc, port, sc->read_line_symbol, args, an_input_port_string, 1);

      if (is_not_null(cdr(args)))
	with_eol = (cadr(args) != sc->F);
    }
  else
    {
      port = input_port_if_not_loading(sc);
      if (!port) return(sc->eof_object);
    }
  return(port_read_line(port)(sc, port, with_eol, true));
}

static s7_pointer read_line_uncopied;
static s7_pointer g_read_line_uncopied(s7_scheme *sc, s7_pointer args)
{
  s7_pointer port;
  bool with_eol = false;
  port = car(args);
  if (!is_input_port(port))
    return(g_read_line(sc, args));
  if (is_not_null(cdr(args)))
    with_eol = (cadr(args) != sc->F);
  return(port_read_line(port)(sc, port, with_eol, false));
}


static s7_pointer g_read_string(s7_scheme *sc, s7_pointer args)
{
  /* read-chars would be a better name -- read-string could mean CL-style read-from-string (like eval-string) 
   *   similarly read-bytes could return a byte-vector (rather than r7rs's read-bytevector)
   *   and write-string -> write-chars, write-bytevector -> write-bytes
   */
  #define H_read_string "(read-string k port) reads k characters from port into a new string and returns it."
  #define Q_read_string s7_make_signature(sc, 3, s7_make_signature(sc, 2, sc->is_string_symbol, sc->is_eof_object_symbol), sc->is_integer_symbol, sc->is_input_port_symbol)
  s7_pointer k, port, s;
  s7_int i, chars;
  unsigned char *str;

  k = car(args);
  if (!s7_is_integer(k))
    method_or_bust(sc, k, sc->read_string_symbol, args, T_INTEGER, 1);
  chars = s7_integer(k);

  if (!is_null(cdr(args)))
    port = cadr(args);
  else port = input_port_if_not_loading(sc);

  if (chars < 0)
    return(wrong_type_argument_with_type(sc, sc->read_string_symbol, 1, make_integer(sc, chars), a_non_negative_integer_string));
  if (chars > sc->max_string_length)
    return(out_of_range(sc, sc->read_string_symbol, small_int(1), make_integer(sc, chars), its_too_large_string));

  if (!port) return(sc->eof_object);
  if (!is_input_port(port))
    method_or_bust_with_type(sc, port, sc->read_string_symbol, list_2(sc, make_integer(sc, chars), port), an_input_port_string, 2);

  if (chars == 0)
    return(make_empty_string(sc, 0, 0));

  s = make_empty_string(sc, chars, 0);
  str = (unsigned char *)string_value(s);
  for (i = 0; i < chars; i++)
    {
      int32_t c;
      c = port_read_character(port)(sc, port);
      if (c == EOF)
	{
	  if (i == 0)
	    return(sc->eof_object);
	  string_length(s) = i;
	  return(s);
	}
      str[i] = (unsigned char)c;
    }
  return(s);
}


#define declare_jump_info() bool old_longjmp; int32_t old_jump_loc, jump_loc; jmp_buf old_goto_start

#define store_jump_info(Sc)						\
  do {									\
      old_longjmp = Sc->longjmp_ok;					\
      old_jump_loc = Sc->setjmp_loc;					\
      memcpy((void *)old_goto_start, (void *)(Sc->goto_start), sizeof(jmp_buf));\
  } while (0)

#define restore_jump_info(Sc)						\
  do {									\
      Sc->longjmp_ok = old_longjmp;					\
      Sc->setjmp_loc = old_jump_loc;					\
      memcpy((void *)(Sc->goto_start), (void *)old_goto_start, sizeof(jmp_buf));\
      if ((jump_loc == ERROR_JUMP) &&\
          (sc->longjmp_ok))\
        longjmp(sc->goto_start, ERROR_JUMP);\
  } while (0)

#define set_jump_info(Sc, Tag)			\
  do {						\
      sc->longjmp_ok = true;			\
      sc->setjmp_loc = Tag;				\
      jump_loc = setjmp(sc->goto_start);		\
  } while (0)


s7_pointer s7_read(s7_scheme *sc, s7_pointer port)
{
  if (is_input_port(port))
    {
      s7_pointer old_envir;
      declare_jump_info();

      old_envir = sc->envir;
      sc->envir = sc->nil;
      push_input_port(sc, port);

      store_jump_info(sc);
      set_jump_info(sc, READ_SET_JUMP);
      if (jump_loc != NO_JUMP)
	{
	  if (jump_loc != ERROR_JUMP)
	    eval(sc, sc->cur_op);
	}
      else 
	{
	  push_stack_no_let_no_code(sc, OP_BARRIER, port);
	  push_stack(sc, OP_EVAL_DONE, sc->args, sc->code);

	  eval(sc, OP_READ_INTERNAL);

	  if (sc->tok == TOKEN_EOF)
	    sc->value = sc->eof_object;

	  if ((sc->cur_op == OP_EVAL_DONE) &&
	      (stack_op(sc->stack, s7_stack_top(sc) - 1) == OP_BARRIER))
	    pop_stack(sc); 
	}
      pop_input_port(sc);
      sc->envir = old_envir;

      restore_jump_info(sc);
      return(sc->value);
    }
  return(simple_wrong_type_argument_with_type(sc, sc->read_symbol, port, an_input_port_string));
}


static s7_pointer g_read(s7_scheme *sc, s7_pointer args)
{
  /* would it be useful to add an environment arg here?  (just set sc->envir at the end?)
   *    except for expansions, nothing is evaluated at read time, unless...
   *    say we set up a dot reader:
   *        (set! *#readers* (cons (cons #\. (lambda (str) (if (string=? str ".") (eval (read)) #f))) *#readers*))
   *    then
   *        (call-with-input-string "(+ 1 #.(+ 1 hiho))" (lambda (p) (read p)))
   *    evaluates hiho in the rootlet, but how to pass the env to the inner eval or read?
   * (eval, eval-string and load already have an env arg)
   */
  #define H_read "(read (port (current-input-port))) returns the next object in the input port, or #<eof> at the end"
  #define Q_read s7_make_signature(sc, 2, sc->T, sc->is_input_port_symbol)
  s7_pointer port;

  if (is_not_null(args))
    port = car(args);
  else
    {
      port = input_port_if_not_loading(sc);
      if (!port) return(sc->eof_object);
    }

  if (!is_input_port(port))
    method_or_bust_with_type_one_arg(sc, port, sc->read_symbol, args, an_input_port_string);

  if (is_function_port(port))
    return((*(port_input_function(port)))(sc, S7_READ, port));

  if ((is_string_port(port)) &&
      (port_data_size(port) <= port_position(port)))
    return(sc->eof_object);

  push_input_port(sc, port);
  push_stack_op_let(sc, OP_READ_DONE); /* this stops the internal read process so we only get one form */
  push_stack_op_let(sc, OP_READ_INTERNAL);

  return(port);
}


/* -------------------------------- load -------------------------------- */

#if WITH_MULTITHREAD_CHECKS
typedef struct {
  s7_scheme* sc;
  const int32_t lock_count; /* Remember lock count in case we have skipped calls to leave_track_scope by longjmp-ing. */
} lock_scope_t;

static lock_scope_t enter_lock_scope(s7_scheme *sc) 
{
  if (pthread_mutex_trylock(&sc->lock) != 0) abort();
  sc->lock_count++;
  {
    lock_scope_t st = {.sc = sc, .lock_count = sc->lock_count};
    return(st);
  }
}

static void leave_lock_scope(lock_scope_t *st)
{
  while (st->sc->lock_count > st->lock_count)
    {
      st->sc->lock_count--;
      pthread_mutex_unlock(&st->sc->lock);
    }
}
#define TRACK(Sc) lock_scope_t lock_scope __attribute__ ((__cleanup__(leave_lock_scope))) = enter_lock_scope(Sc)
#else
#define TRACK(Sc)
#endif


static FILE *search_load_path(s7_scheme *sc, const char *name)
{
  int32_t i, len;
  s7_pointer lst;

  lst = s7_load_path(sc);
  len = s7_list_length(sc, lst);
  for (i = 0; i < len; i++)
    {
      const char *new_dir;
      new_dir = string_value(s7_list_ref(sc, lst, i));
      if (new_dir)
	{
	  FILE *fp;
	  snprintf(sc->tmpbuf, TMPBUF_SIZE, "%s/%s", new_dir, name);
	  fp = fopen(sc->tmpbuf, "r");
	  if (fp) return(fp);
	}
    }
  return(NULL);
}


s7_pointer s7_load_with_environment(s7_scheme *sc, const char *filename, s7_pointer e)
{
  s7_pointer port;
  FILE *fp;
  char *new_filename = NULL;
  declare_jump_info();
  TRACK(sc);

  fp = fopen(filename, "r");
  if (!fp)
    {
      fp = search_load_path(sc, filename);
      if (fp) 
	new_filename = copy_string(sc->tmpbuf); /* (require libc.scm) for example needs the directory for cload in some cases */
    }
  if (!fp)
    return(file_error(sc, "load", "can't open", filename));

  if (hook_has_functions(sc->load_hook))
    s7_call(sc, sc->load_hook, list_1(sc, sc->temp4 = s7_make_string(sc, filename)));

  port = read_file(sc, fp, (new_filename) ? (const char *)new_filename : filename, -1, "load");   /* -1 means always read its contents into a local string */
  port_file_number(port) = remember_file_name(sc, filename);
  if (new_filename) free(new_filename);
  set_loader_port(port);
  sc->temp6 = port;
  push_input_port(sc, port);
  sc->temp6 = sc->nil;

  /* it's possible to call this recursively (s7_load is Xen_load_file which can be invoked via s7_call)
   *   but in that case, we actually want it to behave like g_load and continue the evaluation upon completion
   */
  sc->envir = e;
  push_stack(sc, OP_LOAD_RETURN_IF_EOF, port, sc->code);

  store_jump_info(sc);
  set_jump_info(sc, LOAD_SET_JUMP);
  if (jump_loc != NO_JUMP)
    {
      if (jump_loc != ERROR_JUMP)
	eval(sc, sc->cur_op);
    }
  else eval(sc, OP_READ_INTERNAL);

  pop_input_port(sc);
  if (is_input_port(port))
    s7_close_input_port(sc, port);

  restore_jump_info(sc);
  if (is_multiple_value(sc->value))
    sc->value = splice_in_values(sc, multiple_value(sc->value));
 return(sc->value);
}


s7_pointer s7_load(s7_scheme *sc, const char *filename)
{
  return(s7_load_with_environment(sc, filename, sc->nil));
}


#if WITH_C_LOADER
#include <dlfcn.h>

static char *full_filename(const char *filename)
{
  int32_t len;
  char *pwd, *rtn;
  pwd = getcwd(NULL, 0); /* docs say this means it will return a new string of the right size */
  len = safe_strlen(pwd) + safe_strlen(filename) + 8;
  rtn = (char *)malloc(len * sizeof(char));
  if (pwd)
    {
      snprintf(rtn, len, "%s/%s", pwd, filename);
      free(pwd);
    }
  else snprintf(rtn, len, "%s", filename);
  return(rtn);
}
#endif


static s7_pointer g_load(s7_scheme *sc, s7_pointer args)
{
  #define H_load "(load file (env (rootlet))) loads the scheme file 'file'. The 'env' argument \
defaults to the rootlet.  To load into the current environment instead, pass (curlet)."
  #define Q_load s7_make_signature(sc, 3, sc->values_symbol, sc->is_string_symbol, sc->is_let_symbol)

  FILE *fp = NULL;
  s7_pointer name, port;
  const char *fname;

  name = car(args);
  if (!is_string(name))
    method_or_bust(sc, name, sc->load_symbol, args, T_STRING, 1);

  if (is_not_null(cdr(args)))
    {
      s7_pointer e;
      e = cadr(args);
      if (!is_let(e))
	return(wrong_type_argument_with_type(sc, sc->load_symbol, 2, e, a_let_string));
      if (e == sc->rootlet)
	sc->envir = sc->nil;
      else sc->envir = e;
    }
  else sc->envir = sc->nil;

  fname = string_value(name);
  if ((!fname) || (!(*fname)))                 /* fopen("", "r") returns a file pointer?? */
    return(s7_error(sc, sc->out_of_range_symbol, set_elist_2(sc, s7_make_string_wrapper(sc, "load's first argument, ~S, should be a filename"), name)));

  if (is_directory(fname))
    return(s7_error(sc, sc->wrong_type_arg_symbol, set_elist_2(sc, s7_make_string_wrapper(sc, "load argument, ~S, is a directory"), name)));

#if WITH_C_LOADER
  /* if fname ends in .so, try loading it as a c shared object
   *   (load "/home/bil/cl/m_j0.so" (inlet (cons 'init_func 'init_m_j0)))
   */
  {
    int32_t fname_len;

    fname_len = safe_strlen(fname);
    if ((fname_len > 3) &&
	(is_pair(cdr(args))) &&
	(local_strcmp((const char *)(fname + (fname_len - 3)), ".so")))
      {
	s7_pointer init;

	init = g_let_ref(sc, set_plist_2(sc, (is_null(sc->envir)) ? sc->rootlet : sc->envir, s7_make_symbol(sc, "init_func")));
	if (is_symbol(init))
	  {
	    void *library;
	    char *pwd_name = NULL;

	    if (fname[0] != '/')
	      pwd_name = full_filename(fname); /* this is necessary, at least in Linux -- we can't blithely dlopen whatever is passed to us */
	    library = dlopen((pwd_name) ? pwd_name : fname, RTLD_NOW);
	    if (library)
	      {
		const char *init_name = NULL;
		void *init_func;

		init_name = symbol_name(init);
		init_func = dlsym(library, init_name);
		if (init_func)
		  {
		    typedef void *(*dl_func)(s7_scheme *sc);
		    ((dl_func)init_func)(sc);
		    if (pwd_name) free(pwd_name);
		    return(sc->T);
		  }
		else
		  {
		    s7_warn(sc, 512, "loaded %s, but can't find %s (%s)?\n", fname, init_name, dlerror());
		    dlclose(library);
		  }
	      }
	    else s7_warn(sc, 512, "load %s failed: %s\n", (pwd_name) ? pwd_name : fname, dlerror());
	    if (pwd_name) free(pwd_name);
	  }
	else s7_warn(sc, 512, "can't load %s: no init function\n", fname);
	return(sc->F);
      }
  }
#endif

  fp = fopen(fname, "r");

#if WITH_GCC
  if (!fp)
    {
      /* catch one special case, "~/..." since it causes 99.9% of the "can't load ..." errors */
      if ((fname[0] == '~') &&
	  (fname[1] == '/'))
	{
	  char *home;
	  home = getenv("HOME");
	  if (home)
	    {
	      char *filename;
	      int32_t len;
	      len = safe_strlen(fname) + safe_strlen(home) + 1;
	      tmpbuf_malloc(filename, len);
	      snprintf(filename, len, "%s%s", home, (char *)(fname + 1));
	      fp = fopen(filename, "r");
	      tmpbuf_free(filename, len);
	    }
	}
    }
#endif

  if (!fp)
    {
      fp = search_load_path(sc, fname);
      if (!fp)
	return(file_error(sc, "load", "can't open", fname));
    }

  port = read_file(sc, fp, fname, -1, "load");
  port_file_number(port) = remember_file_name(sc, fname);
  set_loader_port(port);
  sc->temp6 = port;
  push_input_port(sc, port);
  sc->temp6 = sc->nil;

  push_stack_op_let(sc, OP_LOAD_CLOSE_AND_POP_IF_EOF);  /* was pushing args and code, but I don't think they're used later */
  push_stack_op_let(sc, OP_READ_INTERNAL);

  /* now we've opened and moved to the file to be loaded, and set up the stack to return
   *   to where we were.  Call *load-hook* if it is a procedure.
   */

  if (hook_has_functions(sc->load_hook))
    s7_apply_function(sc, sc->load_hook, list_1(sc, sc->temp4 = s7_make_string(sc, fname)));

  return(sc->unspecified);
}


s7_pointer s7_load_path(s7_scheme *sc)
{
  return(s7_symbol_value(sc, sc->load_path_symbol));
}


s7_pointer s7_add_to_load_path(s7_scheme *sc, const char *dir)
{
  s7_symbol_set_value(sc,
		      sc->load_path_symbol,
		      cons(sc,
			   s7_make_string(sc, dir),
			   s7_symbol_value(sc, sc->load_path_symbol)));
  return(s7_symbol_value(sc, sc->load_path_symbol));
}


static s7_pointer g_load_path_set(s7_scheme *sc, s7_pointer args)
{
  /* new value must be either () or a proper list of strings */
  if (is_null(cadr(args))) return(cadr(args));
  if (is_pair(cadr(args)))
    {
      s7_pointer x;
      for (x = cadr(args); is_pair(x); x = cdr(x))
	if (!is_string(car(x)))
	  return(sc->error_symbol);
      if (is_null(x))
	return(cadr(args));
    }
  return(sc->error_symbol);
}

static s7_pointer g_cload_directory_set(s7_scheme *sc, s7_pointer args)
{
  s7_pointer cl_dir;
  cl_dir = cadr(args);
  if (!is_string(cl_dir))
    return(sc->error_symbol);
  s7_symbol_set_value(sc, sc->cload_directory_symbol, cl_dir);
  if (safe_strlen(string_value(cl_dir)) > 0)
    s7_add_to_load_path(sc, (const char *)(string_value(cl_dir)));
  return(cl_dir);
}


/* ---------------- autoload ---------------- */

void s7_autoload_set_names(s7_scheme *sc, const char **names, int32_t size)
{
  /* the idea here is that by sticking to string constants we can handle 90% of the work at compile-time,
   *   with less start-up memory.  Then eventually we'll add C libraries a la xg (gtk) as environments
   *   and every name in that library will come as an import once dlopen has picked up the library.
   *   So, hopefully, we can pre-declare as many names as we want from as many libraries as we want,
   *   without a bloated mess of a run-time image.  And new libraries are easy to accommodate --
   *   add the names to be auto-exported to this list with the name of the scheme file that cloads
   *   the library and exports the given name. So, we'll need a separate such file for each library?
   *
   * the environment variable could use the library base name in *: *libm* or *libgtk*
   *   (*libm* 'j0)
   * why not just predeclare these libraries?  The caller could import what he wants via require.
   * So the autoloader need only know which libraries, but this doesn't fit the current use of gtk in xg
   * In fact, we only need to see *libm* -> libm.so etc, but we still need the arg/return types of each function, etc
   * And libgtk is enormous -- seems too bad to tie-in everything via the FFI when we need less than 1% of it.
   * Perhaps each module as an environment within the main one: ((*libgtk* *gtkwidget*) 'gtk_widget_new)?
   * But that requires inside knowlege of the library, and changes without notice.
   *
   * Also we need to decide how to handle name collisions (by order of autoload lib setup)
   * And (lastly?) how to handle different library versions?
   *
   *
   * so autoload known libs here in s7 so we're indepentdent of snd
   *   (currently these are included in make-index.scm[line 575] -> snd-xref.c)
   * for each module, include an env in the lib env (*libgtk* 'gtkwidget.h) or whatever that has the names in that header
   * in autoload below, don't sort! -- just build a list of autoload tables and check each in order at autoload time (we want startup to be fast)
   * for versions, include wrapper macro at end of each c-define choice
   * in the xg case, there's no savings in delaying the defines
   *
   */

  if (!sc->autoload_names)
    {
      sc->autoload_names = (const char ***)calloc(INITIAL_AUTOLOAD_NAMES_SIZE, sizeof(const char **));
      sc->autoload_names_sizes = (int32_t *)calloc(INITIAL_AUTOLOAD_NAMES_SIZE, sizeof(int32_t));
      sc->autoloaded_already = (bool **)calloc(INITIAL_AUTOLOAD_NAMES_SIZE, sizeof(bool *));
      sc->autoload_names_top = INITIAL_AUTOLOAD_NAMES_SIZE;
      sc->autoload_names_loc = 0;
    }
  else
    {
      if (sc->autoload_names_loc >= sc->autoload_names_top)
	{
	  int32_t i;
	  sc->autoload_names_top *= 2;
	  sc->autoload_names = (const char ***)realloc(sc->autoload_names, sc->autoload_names_top * sizeof(const char **));
	  sc->autoload_names_sizes = (int32_t *)realloc(sc->autoload_names_sizes, sc->autoload_names_top * sizeof(int32_t));
	  sc->autoloaded_already = (bool **)realloc(sc->autoloaded_already, sc->autoload_names_top * sizeof(bool *));
	  for (i = sc->autoload_names_loc; i < sc->autoload_names_top; i++)
	    {
	      sc->autoload_names[i] = NULL;
	      sc->autoload_names_sizes[i] = 0;
	      sc->autoloaded_already[i] = NULL;
	    }
	}
    }

  sc->autoload_names[sc->autoload_names_loc] = names;
  sc->autoload_names_sizes[sc->autoload_names_loc] = size;
  sc->autoloaded_already[sc->autoload_names_loc] = (bool *)calloc(size, sizeof(bool));
  sc->autoload_names_loc++;
}


static const char *find_autoload_name(s7_scheme *sc, s7_pointer symbol, bool *already_loaded, bool loading)
{
  int32_t l = 0, pos = -1, lib, libs;
  const char *name, *this_name;

  name = symbol_name(symbol);
  libs = sc->autoload_names_loc;

  for (lib = 0; lib < libs; lib++)
    {
      const char **names;
      int32_t u;
      u = sc->autoload_names_sizes[lib] - 1;
      names = sc->autoload_names[lib];

      while (true)
	{
	  int32_t comp;
	  if (u < l) break;
	  pos = (l + u) / 2;
	  this_name = names[pos * 2];
	  comp = strcmp(this_name, name);
	  if (comp == 0)
	    {
	      *already_loaded = sc->autoloaded_already[lib][pos];
	      if (loading) sc->autoloaded_already[lib][pos] = true;
	      return(names[pos * 2 + 1]);             /* file name given func name */
	    }
	  if (comp < 0)
	    l = pos + 1;
	  else u = pos - 1;
	}
    }
  return(NULL);
}


s7_pointer s7_autoload(s7_scheme *sc, s7_pointer symbol, s7_pointer file_or_function)
{
  /* add '(symbol . file) to s7's autoload table */
  if (is_null(sc->autoload_table))
    sc->autoload_table = s7_make_hash_table(sc, sc->default_hash_table_length);
  s7_hash_table_set(sc, sc->autoload_table, symbol, file_or_function);
  return(file_or_function);
}


static s7_pointer g_autoload(s7_scheme *sc, s7_pointer args)
{
  #define H_autoload "(autoload symbol file-or-function) adds the symbol to its table of autoloadable symbols. \
If that symbol is encountered as an unbound variable, s7 either loads the file (following *load-path*), or calls \
the function.  The function takes one argument, the calling environment.  Presumably the symbol is defined \
in the file, or by the function."
  #define Q_autoload s7_make_signature(sc, 3, sc->T, sc->is_symbol_symbol, sc->T)

  s7_pointer sym, value;

  sym = car(args);
  if (is_string(sym))
    {
      if (string_length(sym) == 0)                   /* (autoload "" ...) */
	return(s7_wrong_type_arg_error(sc, "autoload", 1, sym, "a symbol-name or a symbol"));
      sym = make_symbol_with_length(sc, string_value(sym), string_length(sym));
    }
  if (!is_symbol(sym))
    {
      check_method(sc, sym, sc->autoload_symbol, args);
      return(s7_wrong_type_arg_error(sc, "autoload", 1, sym, "a string (symbol-name) or a symbol"));
    }
  if (is_keyword(sym))
    return(s7_wrong_type_arg_error(sc, "autoload", 1, sym, "a normal symbol (a keyword is never unbound)"));

  value = cadr(args);
  if (is_string(value))
    return(s7_autoload(sc, sym, value));
  if (((is_closure(value)) || (is_closure_star(value))) &&
      (s7_is_aritable(sc, value, 1)))
    return(s7_autoload(sc, sym, value));

  check_method(sc, value, sc->autoload_symbol, args);
  return(s7_wrong_type_arg_error(sc, "autoload", 2, value, "a string (file-name) or a thunk"));
}

static s7_pointer g_autoloader(s7_scheme *sc, s7_pointer args)
{
  #define H_autoloader "(*autoload* sym) returns the autoload info for the symbol sym, or #f."
  #define Q_autoloader s7_make_signature(sc, 2, sc->T, sc->is_symbol_symbol)
  s7_pointer sym;

  sym = car(args);
  if (!is_symbol(sym))
    {
      check_method(sc, sym, sc->autoloader_symbol, args);
      return(s7_wrong_type_arg_error(sc, "*autoload*", 1, sym, "a symbol"));
    }
  if (sc->autoload_names)
    {
      const char *file;
      bool loaded = false;
      file = find_autoload_name(sc, sym, &loaded, false);
      if (file)
	return(s7_make_string(sc, file));
    }
  if (is_hash_table(sc->autoload_table))
    return(s7_hash_table_ref(sc, sc->autoload_table, sym));

  return(sc->F);
}


static s7_pointer g_require(s7_scheme *sc, s7_pointer args)
{
  #define H_require "(require . symbols) loads each file associated with each symbol if it has not been loaded already.\
The symbols refer to the argument to \"provide\".  (require lint.scm)"
  #define Q_require s7_make_circular_signature(sc, 1, 2, sc->T, sc->is_symbol_symbol)

  s7_pointer p;
  push_stack_no_let_no_code(sc, OP_GC_PROTECT, args);
  for (p = args; is_pair(p); p = cdr(p))
    {
      s7_pointer sym;
      if (is_symbol(car(p)))
	sym = car(p);
      else
	{
	  if ((is_proper_quote(sc, car(p))) && 
	      (is_symbol(cadar(p))))
	    sym = cadar(p);
	  else return(s7_error(sc, sc->wrong_type_arg_symbol, set_elist_2(sc, s7_make_string_wrapper(sc, "require: ~S is not a symbol"), car(p))));
	}

      if ((!is_slot(find_symbol(sc, sym))) &&
	  (sc->is_autoloading))
	{
	  s7_pointer f;
	  f = g_autoloader(sc, list_1(sc, sym));
	  if (is_string(f))
	    s7_load_with_environment(sc, string_value(f), sc->envir);
	  else
	    {
	      sc->temp5 = sc->nil; 
	      return(s7_error(sc, make_symbol(sc, "autoload-error"), 
			      set_elist_2(sc, s7_make_string_wrapper(sc, "require: no autoload info for ~S"), sym)));
	    }
	}
    }
  sc->stack_end -= 4;  
  return(sc->T);
}


/* -------------------------------- eval-string -------------------------------- */

s7_pointer s7_eval_c_string_with_environment(s7_scheme *sc, const char *str, s7_pointer e)
{
  s7_pointer code, port;
  TRACK(sc);
  port = s7_open_input_string(sc, str);
  code = s7_read(sc, port);
  s7_close_input_port(sc, port);
  return(s7_eval(sc, _NFre(code), e));
}


s7_pointer s7_eval_c_string(s7_scheme *sc, const char *str)
{
  return(s7_eval_c_string_with_environment(sc, str, sc->nil));
}

static s7_pointer g_eval_string(s7_scheme *sc, s7_pointer args)
{
  #define H_eval_string "(eval-string str (env (curlet))) returns the result of evaluating the string str as Scheme code"
  #define Q_eval_string s7_make_signature(sc, 3, sc->values_symbol, sc->is_string_symbol, sc->is_let_symbol)
  s7_pointer port, str;
  
  str = car(args);
  if (!is_string(str))
    method_or_bust(sc, str, sc->eval_string_symbol, args, T_STRING, 1);

  if (is_not_null(cdr(args)))
    {
      s7_pointer e;
      e = cadr(args);
      if (!is_let(e))
 	return(wrong_type_argument_with_type(sc, sc->eval_string_symbol, 2, e, a_let_string));
      if (e == sc->rootlet)
	sc->envir = sc->nil;
      else sc->envir = e;
    }

  port = open_and_protect_input_string(sc, str);
  push_input_port(sc, port);

  sc->temp3 = sc->args;
  push_stack(sc, OP_EVAL_STRING_1, args, sc->code); 
  push_stack_op_let(sc, OP_READ_INTERNAL);

  return(sc->F);
}

static s7_pointer eval_string_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
  check_for_substring_temp(sc, expr);
  return(f);
}


static s7_pointer call_with_input(s7_scheme *sc, s7_pointer port, s7_pointer args)
{
  s7_pointer p;
  p = cadr(args);
  port_original_input_string(port) = car(args);
  push_stack(sc, OP_UNWIND_INPUT, sc->input_port, port);
  push_stack(sc, OP_APPLY, list_1(sc, port), p);
  return(sc->F);
}


/* -------------------------------- call-with-input-string -------------------------------- */

static s7_pointer g_call_with_input_string(s7_scheme *sc, s7_pointer args)
{
  s7_pointer str, proc;
  #define H_call_with_input_string "(call-with-input-string str proc) opens a string port for str and applies proc to it"
  #define Q_call_with_input_string pl_sf
  /* (call-with-input-string "44" (lambda (p) (+ 1 (read p)))) -> 45 */

  str = car(args);
  if (!is_string(str))
    method_or_bust(sc, str, sc->call_with_input_string_symbol, args, T_STRING, 1);

  proc = cadr(args);
  if (is_let(proc))
    check_method(sc, proc, sc->call_with_input_string_symbol, args);

  if (!s7_is_aritable(sc, proc, 1))
    return(wrong_type_argument_with_type(sc, sc->call_with_input_string_symbol, 2, proc,
					 s7_make_string_wrapper(sc, "a procedure of one argument (the port)")));

  if ((is_continuation(proc)) || (is_goto(proc)))
    return(wrong_type_argument_with_type(sc, sc->call_with_input_string_symbol, 2, proc, a_normal_procedure_string));

  return(call_with_input(sc, open_and_protect_input_string(sc, str), args));
}



/* -------------------------------- call-with-input-file -------------------------------- */

static s7_pointer g_call_with_input_file(s7_scheme *sc, s7_pointer args)
{
  #define H_call_with_input_file "(call-with-input-file filename proc) opens filename and calls proc with the input port as its argument"
  #define Q_call_with_input_file pl_sf
  s7_pointer str, proc;

  str = car(args);
  if (!is_string(str))
    method_or_bust(sc, str, sc->call_with_input_file_symbol, args, T_STRING, 1);

  proc = cadr(args);
  if (!s7_is_aritable(sc, proc, 1))
    return(wrong_type_argument_with_type(sc, sc->call_with_input_file_symbol, 2, proc,
					 s7_make_string_wrapper(sc, "a procedure of one argument (the port)")));
  if ((is_continuation(proc)) || (is_goto(proc)))
    return(wrong_type_argument_with_type(sc, sc->call_with_input_file_symbol, 2, proc, a_normal_procedure_string));

  return(call_with_input(sc, open_input_file_1(sc, string_value(str), "r", "call-with-input-file"), args));
}


static s7_pointer with_input(s7_scheme *sc, s7_pointer port, s7_pointer args)
{
  s7_pointer old_input_port, p;
  old_input_port = sc->input_port;
  sc->input_port = port;
  port_original_input_string(port) = car(args);
  push_stack(sc, OP_UNWIND_INPUT, old_input_port, port);
  p = cadr(args);
  push_stack(sc, OP_APPLY, sc->nil, p);
  return(sc->F);
}


/* -------------------------------- with-input-from-string -------------------------------- */

static s7_pointer g_with_input_from_string(s7_scheme *sc, s7_pointer args)
{
  #define H_with_input_from_string "(with-input-from-string str thunk) opens str as the temporary current-input-port and calls thunk"
  #define Q_with_input_from_string pl_sf
  s7_pointer str;

  str = car(args);
  if (!is_string(str))
    method_or_bust(sc, str, sc->with_input_from_string_symbol, args, T_STRING, 1);

  if (!is_thunk(sc, cadr(args)))
    method_or_bust_with_type(sc, cadr(args), sc->with_input_from_string_symbol, args, a_thunk_string, 2);

  /* since the arguments are evaluated before we get here, we can get some confusing situations:
   *   (with-input-from-string "#x2.1" (read))
   *   (read) -> whatever it can get from the current input port!
   *   ";with-input-from-string argument 2, #<eof>, is untyped but should be a thunk"
   */
  return(with_input(sc, open_and_protect_input_string(sc, str), args));
}



/* -------------------------------- with-input-from-file -------------------------------- */

static s7_pointer g_with_input_from_file(s7_scheme *sc, s7_pointer args)
{
  #define H_with_input_from_file "(with-input-from-file filename thunk) opens filename as the temporary current-input-port and calls thunk"
  #define Q_with_input_from_file pl_sf

  if (!is_string(car(args)))
    method_or_bust(sc, car(args), sc->with_input_from_file_symbol, args, T_STRING, 1);

  if (!is_thunk(sc, cadr(args)))
    method_or_bust_with_type(sc, cadr(args), sc->with_input_from_file_symbol, args, a_thunk_string, 2);

  return(with_input(sc, open_input_file_1(sc, string_value(car(args)), "r", "with-input-from-file"), args));
}



/* -------------------------------- iterators -------------------------------- */

static s7_pointer g_is_iterator(s7_scheme *sc, s7_pointer args)
{
  #define H_is_iterator "(iterator? obj) returns #t if obj is an iterator."
  #define Q_is_iterator pl_bt
  s7_pointer x;

  x = car(args);
  if (is_iterator(x)) return(sc->T);
  check_closure_for(sc, x, sc->is_iterator_symbol);
  check_boolean_method(sc, is_iterator, sc->is_iterator_symbol, args);
  return(sc->F);
}


static s7_pointer iterator_copy(s7_scheme *sc, s7_pointer p)
{
  /* fields are obj cur [loc|lcur] [len|slow|hcur] next */
  s7_pointer iter;
  new_cell(sc, iter, T_ITERATOR | T_SAFE_PROCEDURE | (iter_ok(p) ? T_ITER_OK : 0));
  iterator_sequence(iter) = iterator_sequence(p);   /* obj */
  iterator_position(iter) = iterator_position(p);   /* loc|lcur (loc is s7_int) */
  iterator_length(iter) = iterator_length(p);       /* len|slow|hcur (len is s7_int) */
  iterator_current(iter) = iterator_current(p);     /* cur */
  iterator_next(iter) = iterator_next(p);           /* next */
  return(iter);
}


static s7_pointer iterator_finished(s7_scheme *sc, s7_pointer iterator)
{
  return(sc->ITERATOR_END);
}

static s7_pointer let_iterate(s7_scheme *sc, s7_pointer iterator)
{
  s7_pointer slot;
  slot = iterator_current_slot(iterator);
  if (is_slot(slot))
    {
      iterator_set_current_slot(iterator, next_slot(slot));
      if (iterator_let_cons(iterator))
	{
	  s7_pointer p;
	  p = iterator_let_cons(iterator);
	  set_car(p, slot_symbol(slot));
	  set_cdr(p, slot_value(slot));
	  return(p);
	}
      return(cons(sc, slot_symbol(slot), slot_value(slot)));
    }
  iterator_next(iterator) = iterator_finished;
  clear_iter_ok(iterator);
  return(sc->ITERATOR_END);
}

static s7_pointer rootlet_iterate(s7_scheme *sc, s7_pointer iterator)
{
  s7_pointer slot;
  slot = iterator_current(iterator);
  if (is_slot(slot))
    {
      if (iterator_position(iterator) < sc->rootlet_entries)
	{
	  iterator_position(iterator)++;
	  iterator_current(iterator) = vector_element(sc->rootlet, iterator_position(iterator));
	}
      else iterator_current(iterator) = sc->nil;
      return(cons(sc, slot_symbol(slot), slot_value(slot)));
    }
  iterator_next(iterator) = iterator_finished;
  clear_iter_ok(iterator);
  return(sc->ITERATOR_END);
}

static s7_pointer hash_table_iterate(s7_scheme *sc, s7_pointer iterator)
{
  s7_pointer table;
  int32_t loc, len;
  hash_entry_t **elements;
  hash_entry_t *lst;

  lst = iterator_hash_current(iterator);
  if (lst)
    {
      iterator_hash_current(iterator) = lst->next;
      if (iterator_current(iterator))
	{
	  s7_pointer p;
	  p = iterator_current(iterator);
	  set_car(p, lst->key);
	  set_cdr(p, lst->value);
	  return(p);
	}
      return(cons(sc, lst->key, lst->value));
    }

  table = iterator_sequence(iterator); /* using iterator_length and hash_table_entries here was slightly slower */
  len = hash_table_mask(table) + 1;
  elements = hash_table_elements(table);

  for (loc = iterator_position(iterator) + 1; loc < len;  loc++)
    {
      hash_entry_t *x;
      x = elements[loc];
      if (x)
	{
	  iterator_position(iterator) = loc;
	  iterator_hash_current(iterator) = x->next;
	  if (iterator_current(iterator))
	    {
	      s7_pointer p;
	      p = iterator_current(iterator);
	      set_car(p, x->key);
	      set_cdr(p, x->value);
	      return(p);
	    }
	  return(cons(sc, x->key, x->value));
	}
    }
  iterator_next(iterator) = iterator_finished;
  clear_iter_ok(iterator);
  return(sc->ITERATOR_END);
}

static s7_pointer string_iterate(s7_scheme *sc, s7_pointer obj)
{
  if (iterator_position(obj) < iterator_length(obj))
    return(s7_make_character(sc, (unsigned char)(string_value(iterator_sequence(obj))[iterator_position(obj)++])));
  iterator_next(obj) = iterator_finished;
  clear_iter_ok(obj);
  return(sc->ITERATOR_END);
}

static s7_pointer byte_vector_iterate(s7_scheme *sc, s7_pointer obj)
{
  if (iterator_position(obj) < iterator_length(obj))
    return(small_int((unsigned char)(string_value(iterator_sequence(obj))[iterator_position(obj)++])));
  iterator_next(obj) = iterator_finished;
  clear_iter_ok(obj);
  return(sc->ITERATOR_END);
}

static s7_pointer float_vector_iterate(s7_scheme *sc, s7_pointer obj)
{
  if (iterator_position(obj) < iterator_length(obj))
    return(make_real(sc, float_vector_element(iterator_sequence(obj), iterator_position(obj)++)));
  iterator_next(obj) = iterator_finished;
  clear_iter_ok(obj);
  return(sc->ITERATOR_END);
}

static s7_pointer int_vector_iterate(s7_scheme *sc, s7_pointer obj)
{
  if (iterator_position(obj) < iterator_length(obj))
    return(make_integer(sc, int_vector_element(iterator_sequence(obj), iterator_position(obj)++)));
  iterator_next(obj) = iterator_finished;
  clear_iter_ok(obj);
  return(sc->ITERATOR_END);
}

static s7_pointer vector_iterate(s7_scheme *sc, s7_pointer obj)
{
  if (iterator_position(obj) < iterator_length(obj))
    return(vector_element(iterator_sequence(obj), iterator_position(obj)++));
  iterator_next(obj) = iterator_finished;
  clear_iter_ok(obj);
  return(sc->ITERATOR_END);
}

static s7_pointer closure_iterate(s7_scheme *sc, s7_pointer obj)
{
  s7_pointer result;
  result = s7_apply_function(sc, iterator_sequence(obj), sc->nil);
  if (result == sc->ITERATOR_END)
    {
      iterator_next(obj) = iterator_finished;
      clear_iter_ok(obj);
    }
  return(result);
}

static s7_pointer c_object_direct_iterate(s7_scheme *sc, s7_pointer obj)
{
  if (iterator_position(obj) < iterator_length(obj))
    {
      s7_pointer result, p;
      p = iterator_sequence(obj);
      result = c_object_direct_ref(sc, p)(sc, p, iterator_position(obj));
      iterator_position(obj)++;
      if (result == sc->ITERATOR_END)
	{
	  iterator_next(obj) = iterator_finished;
	  clear_iter_ok(obj);
	}
      return(result);
    }
  iterator_next(obj) = iterator_finished;
  clear_iter_ok(obj);
  return(sc->ITERATOR_END);
}

static s7_pointer c_object_iterate(s7_scheme *sc, s7_pointer obj)
{
  if (iterator_position(obj) < iterator_length(obj))
    {
      s7_pointer result, p, cur;
      p = iterator_sequence(obj);
      cur = iterator_current(obj);
      set_car(sc->z2_1, sc->x);
      set_car(sc->z2_2, sc->z); /* is this necessary? */
      set_car(cur, make_integer(sc, iterator_position(obj)));
      result = (*(c_object_ref(sc, p)))(sc, p, cur);
      sc->x = car(sc->z2_1);
      sc->z = car(sc->z2_2);
      iterator_position(obj)++;
      if (result == sc->ITERATOR_END)
	{
	  iterator_next(obj) = iterator_finished;
	  clear_iter_ok(obj);
	}
      return(result);
    }
  iterator_next(obj) = iterator_finished;
  clear_iter_ok(obj);
  return(sc->ITERATOR_END);
}


static s7_pointer pair_iterate_1(s7_scheme *sc, s7_pointer obj);
static s7_pointer pair_iterate(s7_scheme *sc, s7_pointer obj)
{
  if (is_pair(iterator_current(obj)))
    {
      s7_pointer result;
      result = car(iterator_current(obj));
      iterator_current(obj) = cdr(iterator_current(obj));
      if (iterator_current(obj) == iterator_slow(obj))
	{
	  iterator_next(obj) = iterator_finished;
	  clear_iter_ok(obj);
	  return(result);
	}
      iterator_next(obj) = pair_iterate_1;
      return(result);
    }
  iterator_next(obj) = iterator_finished;
  clear_iter_ok(obj);
  return(sc->ITERATOR_END);
}

static s7_pointer pair_iterate_1(s7_scheme *sc, s7_pointer obj)
{
  if (is_pair(iterator_current(obj)))
    {
      s7_pointer result;
      result = car(iterator_current(obj));
      iterator_current(obj) = cdr(iterator_current(obj));
      if (iterator_current(obj) == iterator_slow(obj))
	{
	  iterator_next(obj) = iterator_finished;
	  clear_iter_ok(obj);
	  return(result);
	}
      iterator_set_slow(obj, cdr(iterator_slow(obj)));
      iterator_next(obj) = pair_iterate;
      return(result);
    }
  iterator_next(obj) = iterator_finished;
  clear_iter_ok(obj);
  return(sc->ITERATOR_END);
}

static s7_pointer iterator_method(s7_scheme *sc, s7_pointer e)
{
  s7_pointer func;
  if ((has_methods(e)) && 
      ((func = find_method(sc, find_let(sc, e), sc->make_iterator_symbol)) != sc->undefined))
    {
      s7_pointer it;
      it = s7_apply_function(sc, func, list_1(sc, e));
      if (!is_iterator(it))
	return(s7_error(sc, sc->error_symbol, set_elist_2(sc, s7_make_string_wrapper(sc, "make-iterator method must return an interator: ~S"), it)));
      return(it);
    }
  return(NULL);
}

s7_pointer s7_make_iterator(s7_scheme *sc, s7_pointer e)
{
  s7_pointer iter;

  new_cell(sc, iter, T_ITERATOR | T_SAFE_PROCEDURE | T_ITER_OK);
  iterator_sequence(iter) = e;
  iterator_position(iter) = 0;

  switch (type(e))
    {
    case T_LET:
      if (e == sc->rootlet)
	{
	  iterator_current(iter) = vector_element(e, 0); /* unfortunately tricky -- let_iterate uses different fields */
	  iterator_next(iter) = rootlet_iterate;
	}
      else
	{
	  s7_pointer f;
	  sc->temp6 = iter;
	  f = iterator_method(sc, e);
	  sc->temp6 = sc->nil;
	  if (f) {free_cell(sc, iter); return(f);}
	  iterator_set_current_slot(iter, let_slots(e));
	  iterator_next(iter) = let_iterate;
	  iterator_let_cons(iter) = NULL;
	}
      break;

    case T_HASH_TABLE:
      iterator_hash_current(iter) = NULL;
      iterator_current(iter) = NULL;
      iterator_position(iter) = -1;
      iterator_next(iter) = hash_table_iterate;
      break;

    case T_STRING:
      iterator_length(iter) = string_length(e);
      if (is_byte_vector(e))
	iterator_next(iter) = byte_vector_iterate;
      else iterator_next(iter) = string_iterate;
      break;

    case T_VECTOR:
      iterator_length(iter) = vector_length(e);
      iterator_next(iter) = vector_iterate;
      break;

    case T_INT_VECTOR:
      iterator_length(iter) = vector_length(e);
      iterator_next(iter) = int_vector_iterate;
      break;

    case T_FLOAT_VECTOR:
      iterator_length(iter) = vector_length(e);
      iterator_next(iter) = float_vector_iterate;
      break;

    case T_PAIR:
      iterator_current(iter) = e;
      iterator_next(iter) = pair_iterate;
      iterator_set_slow(iter, e);
      break;

    case T_MACRO:   case T_MACRO_STAR:
    case T_BACRO:   case T_BACRO_STAR:
    case T_CLOSURE: case T_CLOSURE_STAR:
      {
	s7_pointer p;
	p = cons(sc, e, sc->nil);
	if (g_is_iterator(sc, p) != sc->F)
	  {
	    set_car(p, small_int(0));
	    iterator_current(iter) = p;
	    set_mark_seq(iter);
	    iterator_next(iter) = closure_iterate;
	    if (has_methods(e))
	      iterator_length(iter) = closure_length(sc, e);
	    else iterator_length(iter) = s7_int_max;
	  }
	else 
	  {
	    free_cell(sc, iter);
	    return(simple_wrong_type_argument_with_type(sc, sc->make_iterator_symbol, e, s7_make_string_wrapper(sc, "a closure/macro with an 'iterator local that is not #f")));
	  }
      }
      break;

    case T_C_OBJECT:
      iterator_length(iter) = c_object_length_to_int(sc, e);
      if (c_object_direct_ref(sc, e))
	iterator_next(iter) = c_object_direct_iterate;
      else
	{
	  s7_pointer f;
	  sc->temp6 = iter;
	  f = iterator_method(sc, e);
	  sc->temp6 = sc->nil;
	  if (f) {free_cell(sc, iter); return(f);}
	  iterator_current(iter) = cons(sc, small_int(0), sc->nil);
	  set_mark_seq(iter);
	  iterator_next(iter) = c_object_iterate;
	}
      break;

    default:
      return(simple_wrong_type_argument_with_type(sc, sc->make_iterator_symbol, e, a_sequence_string));
    }
  return(iter);
}


static s7_pointer g_make_iterator(s7_scheme *sc, s7_pointer args)
{
  #define H_make_iterator "(make-iterator sequence) returns an iterator object that returns the next value \
in the sequence each time it is called.  When it reaches the end, it returns " ITERATOR_END_NAME "."
  #define Q_make_iterator s7_make_signature(sc, 3, sc->is_iterator_symbol, sc->is_sequence_symbol, sc->is_pair_symbol)
  
  s7_pointer seq;
  seq = car(args);

  if (is_pair(cdr(args)))
    {
      if (is_pair(cadr(args)))
	{
	  if (is_hash_table(seq))
	    {
	      s7_pointer iter;
	      iter = s7_make_iterator(sc, seq);
	      iterator_current(iter) = cadr(args);
	      set_mark_seq(iter);
	      return(iter);
	    }
	  if ((is_let(seq)) && (seq != sc->rootlet))
	    {
	      s7_pointer iter;
	      iter = s7_make_iterator(sc, seq);
	      iterator_let_cons(iter) = cadr(args);
	      set_mark_seq(iter);
	      return(iter);
	    }
	}
      else return(simple_wrong_type_argument(sc, sc->make_iterator_symbol, cadr(args), T_PAIR));
    }
  return(s7_make_iterator(sc, seq));
}

static s7_pointer g_iterate(s7_scheme *sc, s7_pointer args)
{
  #define H_iterate "(iterate obj) returns the next element from the iterator obj, or " ITERATOR_END_NAME "."
  #define Q_iterate s7_make_signature(sc, 2, sc->T, sc->is_iterator_symbol)

  s7_pointer iter;
  iter = car(args);
  if (!is_iterator(iter))
    method_or_bust_one_arg(sc, iter, sc->iterate_symbol, args, T_ITERATOR);
  return((iterator_next(iter))(sc, iter));
}

s7_pointer s7_iterate(s7_scheme *sc, s7_pointer obj)
{
  return((iterator_next(obj))(sc, obj));
}

bool s7_is_iterator(s7_pointer obj)
{
  return(is_iterator(obj));
}

static bool is_iterator_b(s7_pointer obj) {return(g_is_iterator(cur_sc, set_plist_1(cur_sc, obj)) != cur_sc->F);}

bool s7_iterator_is_at_end(s7_scheme *sc, s7_pointer obj)
{
  if (iter_ok(obj))
    return(false);
  if (!is_iterator(obj))
    simple_wrong_type_argument(sc, sc->iterator_is_at_end_symbol, obj, T_ITERATOR);
  return(true);
}

bool iterator_is_at_end_b(s7_pointer obj)
{
  if (iter_ok(obj))
    return(false);
  if (!is_iterator(obj))
    simple_wrong_type_argument(cur_sc, cur_sc->iterator_is_at_end_symbol, obj, T_ITERATOR);
  return(true);
}

static s7_pointer g_iterator_is_at_end(s7_scheme *sc, s7_pointer args)
{
  #define H_iterator_is_at_end "(iterator-at-end? iter) returns #t if the iterator has reached the end of its sequence."
  #define Q_iterator_is_at_end s7_make_signature(sc, 2, sc->is_boolean_symbol, sc->is_iterator_symbol)
  s7_pointer iter;

  iter = car(args);
  if (iter_ok(iter))
    return(sc->F);
  if (!is_iterator(iter))
    return(simple_wrong_type_argument(sc, sc->iterator_is_at_end_symbol, iter, T_ITERATOR));
  return(sc->T);
}


static s7_pointer g_iterator_sequence(s7_scheme *sc, s7_pointer args)
{
  #define H_iterator_sequence "(iterator-sequence iterator) returns the sequence that iterator is traversing."
  #define Q_iterator_sequence s7_make_signature(sc, 2, sc->is_sequence_symbol, sc->is_iterator_symbol)

  s7_pointer iter;

  iter = car(args);
  if (!is_iterator(iter))
    return(simple_wrong_type_argument(sc, sc->iterator_sequence_symbol, iter, T_ITERATOR));
  return(iterator_sequence(iter));
}



/* -------------------------------------------------------------------------------- */

#define INITIAL_SHARED_INFO_SIZE 8

static int32_t shared_ref(shared_info *ci, s7_pointer p)
{
  /* from print after collecting refs, not called by equality check */
  int32_t i;
  s7_pointer *objs;

  if (!is_collected(p)) return(0);

  objs = ci->objs;
  for (i = 0; i < ci->top; i++)
    if (objs[i] == p)
      {
	int32_t val;
	val = ci->refs[i];
	if (val > 0)
	  ci->refs[i] = -ci->refs[i];
	return(val);
      }
  return(0);
}


static int32_t peek_shared_ref(shared_info *ci, s7_pointer p)
{
  /* returns 0 if not found, otherwise the ref value for p */
  int32_t i;
  s7_pointer *objs;

  objs = ci->objs;
  for (i = 0; i < ci->top; i++)
    if (objs[i] == p) return(ci->refs[i]);

  return(0);
}


static void enlarge_shared_info(shared_info *ci)
{
  int32_t i;
  ci->size *= 2;
  ci->size2 = ci->size - 2;
  ci->objs = (s7_pointer *)realloc(ci->objs, ci->size * sizeof(s7_pointer));
  ci->refs = (int32_t *)realloc(ci->refs, ci->size * sizeof(int32_t));
  for (i = ci->top; i < ci->size; i++)
    {
      ci->refs[i] = 0;
      ci->objs[i] = NULL;
    }
}

static void add_shared_ref(shared_info *ci, s7_pointer x, int32_t ref_x)
{
  /* called only in equality check, not printer */

  if (ci->top == ci->size)
    enlarge_shared_info(ci);

  set_collected(x);

  ci->objs[ci->top] = x;
  ci->refs[ci->top++] = ref_x;
}

static bool collect_shared_info(s7_scheme *sc, shared_info *ci, s7_pointer top, bool stop_at_print_length);
static hash_entry_t *hash_equal(s7_scheme *sc, s7_pointer table, s7_pointer key);

static bool collect_vector_info(s7_scheme *sc, shared_info *ci, s7_pointer top, bool stop_at_print_length)
{
  s7_int i, plen;
  bool cyclic = false;

  if (stop_at_print_length)
    {
      plen = sc->print_length;
      if (plen > vector_length(top))
	plen = vector_length(top);
    }
  else plen = vector_length(top);

  for (i = 0; i < plen; i++)
    if ((has_structure(vector_element(top, i))) &&
	(collect_shared_info(sc, ci, vector_element(top, i), stop_at_print_length)))
      cyclic = true;
  return(cyclic);
}

static bool check_collected(s7_pointer top, shared_info *ci)
{
  s7_pointer *p, *objs_end;
  int32_t i;
  objs_end = (s7_pointer *)(ci->objs + ci->top);
  for (p = ci->objs; p < objs_end; p++)
    if ((*p) == top)
      {
	i = (int32_t)(p - ci->objs);
	if (ci->refs[i] == 0)
	  {
	    ci->has_hits = true;
	    ci->refs[i] = ++ci->ref;  /* if found, set the ref number */
	  }
	break;
      }
  return(true);
}

static bool collect_shared_info(s7_scheme *sc, shared_info *ci, s7_pointer top, bool stop_at_print_length)
{
  /* look for top in current list.
   *
   * As we collect objects (guaranteed to have structure) we set the collected bit.  If we ever
   *   encounter an object with that bit on, we've seen it before so we have a possible cycle.
   *   Once the collection pass is done, we run through our list, and clear all these bits.
   */
  bool top_cyclic;
  if (is_collected_or_shared(top))
    {
      if (is_shared(top))
	return(false);
      return(check_collected(top, ci));
    }

  /* top not seen before -- add it to the list */
  set_collected(top);
  
  if (ci->top == ci->size)
    enlarge_shared_info(ci);
  ci->objs[ci->top++] = top;
  
  top_cyclic = false;
  /* now search the rest of this structure */
  switch (type(top))
    {
    case T_PAIR:
#if 0
      /* old form: simple understandable slow */
      if ((has_structure(car(top))) &&
	  (collect_shared_info(sc, ci, car(top), stop_at_print_length)))
	top_cyclic = true;
      if ((has_structure(cdr(top))) &&
	  (collect_shared_info(sc, ci, cdr(top), stop_at_print_length)))
	top_cyclic = true;
#endif
      {
	s7_pointer p, cp;
	if ((has_structure(car(top))) &&
	    (collect_shared_info(sc, ci, car(top), stop_at_print_length)))
	  top_cyclic = true;
	for (p = cdr(top); is_pair(p); p = cdr(p))
	  {
	    if (is_collected_or_shared(p))
	      {
		if (is_shared(p))
		  {
		    if (!top_cyclic)
		      for (cp = top; cp != p; cp = cdr(cp)) set_shared(cp);
		    return(top_cyclic);
		  }
		return(check_collected(p, ci));
	      }
	    set_collected(p);
	    if (ci->top == ci->size)
	      enlarge_shared_info(ci);
	    ci->objs[ci->top++] = p;
	    if ((has_structure(car(p))) &&
		(collect_shared_info(sc, ci, car(p), stop_at_print_length)))
	      top_cyclic = true;
	  }
	if ((has_structure(p)) &&
	    (collect_shared_info(sc, ci, p, stop_at_print_length)))
	  return(true);

	if (!top_cyclic)
	  for (cp = top; is_pair(cp); cp = cdr(cp)) set_shared(cp);
	return(top_cyclic);
      }
      break;
      
    case T_VECTOR:
      if (collect_vector_info(sc, ci, top, stop_at_print_length))
	top_cyclic = true;
      break;
      
    case T_ITERATOR:
      if (collect_shared_info(sc, ci, iterator_sequence(top), stop_at_print_length))
	top_cyclic = true;
      break;
      
    case T_HASH_TABLE:
      if (hash_table_entries(top) > 0)
	{
	  uint32_t i, len;
	  hash_entry_t **entries;
	  bool keys_safe;
	  
	  keys_safe = ((hash_table_checker(top) != hash_equal) && (!hash_table_checker_locked(top)));
	  entries = hash_table_elements(top);
	  len = hash_table_mask(top) + 1;
	  for (i = 0; i < len; i++)
	    {
	      hash_entry_t *p;
	      for (p = entries[i]; p; p = p->next)
		{
		  if ((!keys_safe) &&
		      (has_structure(p->key)) &&
		      (collect_shared_info(sc, ci, p->key, stop_at_print_length)))
		    top_cyclic = true;
		  if ((has_structure(p->value)) &&
		      (collect_shared_info(sc, ci, p->value, stop_at_print_length)))
		    top_cyclic = true;
		}
	    }
	}
      break;
      
    case T_SLOT:
      if ((has_structure(slot_value(top))) &&
	  (collect_shared_info(sc, ci, slot_value(top), stop_at_print_length)))
	top_cyclic = true;
      break;
      
    case T_LET:
      if (top == sc->rootlet)
	{
	  if (collect_vector_info(sc, ci, top, stop_at_print_length))
	    top_cyclic = true;
	}
      else
	{
	  s7_pointer p;
	  for (p = let_slots(top); is_slot(p); p = next_slot(p))
	    if ((has_structure(slot_value(p))) &&
		(collect_shared_info(sc, ci, slot_value(p), stop_at_print_length)))
	      top_cyclic = true;
	}
      break;
    }
  if (!top_cyclic)
    set_shared(top);
  return(top_cyclic);
}

static shared_info *init_circle_info(void)
{
  shared_info *ci;
  ci = (shared_info *)calloc(1, sizeof(shared_info));
  ci->size = INITIAL_SHARED_INFO_SIZE;
  ci->size2 = ci->size - 2;
  ci->objs = (s7_pointer *)malloc(ci->size * sizeof(s7_pointer));
  ci->refs = (int32_t *)calloc(ci->size, sizeof(int32_t));   /* finder expects 0 = unseen previously */
  return(ci);
}

static shared_info *new_shared_info(s7_scheme *sc)
{
  shared_info *ci;
  int32_t i;
  ci = sc->circle_info;
  memclr((void *)(ci->refs), ci->top * sizeof(int32_t));
  for (i = 0; i < ci->top; i++)
    clear_collected_and_shared(ci->objs[i]);
  ci->top = 0;
  ci->ref = 0;
  ci->has_hits = false;
  return(ci);
}

static shared_info *make_shared_info(s7_scheme *sc, s7_pointer top, bool stop_at_print_length)
{
  /* for the printer */
  shared_info *ci;
  int32_t i, refs;
  s7_pointer *ci_objs;
  int32_t *ci_refs;
  bool no_problem = true, cyclic = false;
  s7_int k, stop_len;

  /* check for simple cases first */
  if (is_pair(top))
    {
      s7_pointer x;
      x = top;
      if (stop_at_print_length)
	{
	  s7_pointer slow;
	  stop_len = sc->print_length;
	  slow = top;
	  for (k = 0; k < stop_len; k += 2)
	    {
	      if (!is_pair(x))
		break;
	      if (has_structure(car(x)))
		{
		  no_problem = false;
		  break;
		}
	      x = cdr(x);
	      if (!is_pair(x))
		break;
	      if (has_structure(car(x)))
		{
		  no_problem = false;
		  break;
		}
	      x = cdr(x);
	      slow = cdr(slow);
	      if (x == slow)
		{
		  no_problem = false;
		  break;
		}
	    }
	}
      else
	{
	  if (s7_list_length(sc, top) == 0) /* it is circular at the top level (following cdr) */
	    no_problem = false;
	  else
	    {
	      for (; is_pair(x); x = cdr(x))
		if (has_structure(car(x)))
		  {
		    /* it can help a little in some cases to scan vectors here (and slots):
		     *   if no element has structure, it's ok (maybe also hash_table_entries == 0)
		     */
		    no_problem = false;
		    break;
		  }
	    }
	}
      if ((no_problem) &&
	  (!is_null(x)) &&
	  (has_structure(x)))
	no_problem = false;
	      
      if (no_problem)
	return(NULL);
    }
  else
    {
      if (s7_is_vector(top))
	{
	  if (!is_normal_vector(top))
	    return(NULL);

	  stop_len = vector_length(top);
	  if ((stop_at_print_length) &&
	      (stop_len > sc->print_length))
	    stop_len = sc->print_length;

	  for (k = 0; k < stop_len; k++)
	    if (has_structure(vector_element(top, k)))
	      {
		no_problem = false;
		break;
	      }
	  if (no_problem)
	    return(NULL);
	}
    }

  ci = new_shared_info(sc);

  /* collect all pointers associated with top */
  cyclic = collect_shared_info(sc, ci, top, stop_at_print_length);

  for (i = 0; i < ci->top; i++)
    {
      s7_pointer p;
      p = ci->objs[i];
      clear_collected_and_shared(p);
    }
  if (!cyclic)
    return(NULL);

  if (!(ci->has_hits))
    return(NULL);

  ci_objs = ci->objs;
  ci_refs = ci->refs;

  /* find if any were referenced twice (once for just being there, so twice=shared)
   *   we know there's at least one such reference because has_hits is true.
   */
  for (i = 0, refs = 0; i < ci->top; i++)
    if (ci_refs[i] > 0)
      {
	set_collected(ci_objs[i]);
	if (i == refs)
	  refs++;
	else
	  {
	    ci_objs[refs] = ci_objs[i];
	    ci_refs[refs++] = ci_refs[i];
	    ci_refs[i] = 0;
	    ci_objs[i] = NULL;
	  }
      }
  ci->top = refs;
  return(ci);
}


/* -------------------------------- cyclic-sequences -------------------------------- */

static s7_pointer cyclic_sequences(s7_scheme *sc, s7_pointer obj, bool return_list)
{
  if (has_structure(obj))
    {
      shared_info *ci;
      ci = make_shared_info(sc, obj, false); /* false=don't stop at print length (vectors etc) */
      if (ci)
	{
	  if (return_list)
	    {
	      int32_t i;
	      s7_pointer lst;
	      sc->w = sc->nil;
	      for (i = 0; i < ci->top; i++)
		sc->w = cons(sc, ci->objs[i], sc->w);
	      lst = sc->w;
	      sc->w = sc->nil;
	      return(lst);
	    }
	  else return(sc->T);
	}
    }
  return(sc->nil);
}

static s7_pointer g_cyclic_sequences(s7_scheme *sc, s7_pointer args)
{
  #define H_cyclic_sequences "(cyclic-sequences obj) returns a list of elements that are cyclic."
  #define Q_cyclic_sequences s7_make_signature(sc, 2, sc->is_proper_list_symbol, sc->T)
  return(cyclic_sequences(sc, car(args), true));
}

static int32_t circular_list_entries(s7_pointer lst)
{
  int32_t i;
  s7_pointer x;
  for (i = 1, x = cdr(lst); ; i++, x = cdr(x))
    {
      int32_t j;
      s7_pointer y;
      for (y = lst, j = 0; j < i; y = cdr(y), j++)
	if (x == y)
	  return(i);
    }
}

static void object_to_port_with_circle_check(s7_scheme *sc, s7_pointer vr, s7_pointer port, use_write_t use_write, shared_info *ci);
/* static void object_to_port(s7_scheme *sc, s7_pointer obj, s7_pointer port, use_write_t use_write, shared_info *ci); */

static void (*display_functions[256])(s7_scheme *sc, s7_pointer obj, s7_pointer port, use_write_t use_write, shared_info *ci); 
#define object_to_port(Sc, Obj, Port, Use_Write, Ci) (*display_functions[unchecked_type(Obj)])(Sc, Obj, Port, Use_Write, Ci)

static s7_pointer object_out(s7_scheme *sc, s7_pointer obj, s7_pointer strport, use_write_t choice);

static bool string_needs_slashification(const char *str, int32_t len)
{
  /* we have to go by len (str len) not *s==0 because s7 strings can have embedded nulls */
  unsigned char *p, *pend;
  pend = (unsigned char *)(str + len);
  for (p = (unsigned char *)str; p < pend; p++)
    if (slashify_table[*p])
      return(true);
  return(false);
}

#define IN_QUOTES true
#define NOT_IN_QUOTES false

static char *slashify_string(s7_scheme *sc, const char *p, int32_t len, bool quoted, int32_t *nlen) /* do not free result */
{
  int32_t j = 0, cur_size, size;
  char *s;
  unsigned char *pcur, *pend;

  pend = (unsigned char *)(p + len);
  size = len * 4 + 256;
  if (size > sc->slash_str_size)
    {
      if (sc->slash_str) free(sc->slash_str);
      sc->slash_str_size = size;
      sc->slash_str = (char *)malloc(size);
    }
  else size = sc->slash_str_size;
  cur_size = size - 2;

  /* memset((void *)sc->slash_str, 0, size); */
  s = sc->slash_str;

  if (quoted) s[j++] = '"';

  /* what about the trailing nulls? Guile writes them out (as does s7 currently)
   *    but that is not ideal.  I'd like to use ~S for error messages, so that
   *    strings are clearly identified via the double-quotes, but this way of
   *    writing them is ugly:
   *
   *    :(let ((str (make-string 8 #\null))) (set! (str 0) #\a) str)
   *    "a\x00\x00\x00\x00\x00\x00\x00"
   *
   * but it would be misleading to omit them because:
   *
   *    :(let ((str (make-string 8 #\null))) (set! (str 0) #\a) (string-append str "bc"))
   *    "a\x00\x00\x00\x00\x00\x00\x00bc"
   */

  for (pcur = (unsigned char *)p; pcur < pend; pcur++)
    {
      if (slashify_table[*pcur])
	{
	  s[j++] = '\\';
	  switch (*pcur)
	    {
	    case '"':   s[j++] = '"';  break;
	    case '\\':  s[j++] = '\\'; break;
	    case '\'':  s[j++] = '\'';  break;
	    case '\t':  s[j++] = 't';  break;
	    case '\r':  s[j++] = 'r';  break;
	    case '\b':  s[j++] = 'b';  break;
	    /* case '\v':  s[j++] = 'v';  break; */
	    case '\f':  s[j++] = 'f';  break;
	    case '\?':  s[j++] = '?';  break;
	    case 'x':   s[j++] = 'x'; break;
	    default:
	      {
		uint32_t n;
		static char dignum[] = "0123456789abcdef";
		s[j++] = 'x';
		n = (uint32_t)(*pcur);
		if (n < 16)
		  s[j++] = '0';
		else s[j++] = dignum[(n / 16) % 16];
		s[j++] = dignum[n % 16];
		s[j++] = ';';
	      }
	      break;
	    }
	}
      else s[j++] = *pcur;
      if (j >= cur_size)
	{
	  size *= 2;
	  sc->slash_str = (char *)realloc(sc->slash_str, size * sizeof(char));    
	  sc->slash_str_size = size;
	  cur_size = size - 4;
	  s = sc->slash_str;
	}
    }
  if (quoted) s[j++] = '"';
  s[j] = '\0';
  (*nlen) = j;
  return(s);
}

static void output_port_to_port(s7_scheme *sc, s7_pointer obj, s7_pointer port, use_write_t use_write, shared_info *ci)
{
  if ((obj == sc->standard_output) ||
      (obj == sc->standard_error))
    port_write_string(port)(sc, port_filename(obj), port_filename_length(obj), port);
  else
    {
      int32_t nlen;
      if (use_write == USE_READABLE_WRITE)
	{
	  if (port_is_closed(obj))
	    port_write_string(port)(sc, "(let ((p (open-output-string))) (close-output-port p) p)", 56, port);
	  else 
	    {
	      char *str;
	      if (is_string_port(obj))
		{
		  port_write_string(port)(sc, "(let ((p (open-output-string)))", 31, port);
		  if (port_position(obj) > 0)
		    {
		      port_write_string(port)(sc, " (display ", 10, port);
		      str = slashify_string(sc, (const char *)port_data(obj), port_position(obj), IN_QUOTES, &nlen);
		      port_write_string(port)(sc, str, nlen, port);
		      port_write_string(port)(sc, " p)", 3, port);
		    }
		  port_write_string(port)(sc, " p)", 3, port);
		}
	      else 
		{
		  str = (char *)malloc(256 * sizeof(char));
		  nlen = snprintf(str, 256, "(open-output-file \"%s\" \"a\")", port_filename(obj));
		  port_write_string(port)(sc, str, nlen, port);
		  free(str);
		}
	    }
	}
      else
	{
	  if (is_string_port(obj))
	    port_write_string(port)(sc, "<output-string-port", 19, port);
	  else
	    {
	      if (is_file_port(obj))
		port_write_string(port)(sc, "<output-file-port", 17, port);
	      else port_write_string(port)(sc, "<output-function-port", 21, port);
	    }
	  if (port_is_closed(obj)) 
	    port_write_string(port)(sc, " (closed)>", 10, port);
	  else port_write_character(port)(sc, '>', port);
	}
    }
}

static void input_port_to_port(s7_scheme *sc, s7_pointer obj, s7_pointer port, use_write_t use_write, shared_info *ci)
{
  if (obj == sc->standard_input)
    port_write_string(port)(sc, port_filename(obj), port_filename_length(obj), port);
  else
    {
      int32_t nlen = 0;
      if (use_write == USE_READABLE_WRITE)
	{
	  if (port_is_closed(obj))
	    port_write_string(port)(sc, "(call-with-input-string \"\" (lambda (p) p))", 42, port);
	  else
	    {
	      if (is_function_port(obj))
		port_write_string(port)(sc, "#<function input port>", 22, port);
	      else
		{
		  char *str;
		  if (is_file_port(obj))
		    {
		      str = (char *)malloc(256 * sizeof(char));
		      nlen = snprintf(str, 256, "(open-input-file \"%s\")", port_filename(obj));
		      port_write_string(port)(sc, str, nlen, port);
		      free(str);
		    }
		  else
		    {
		      /* if the string is large, slashify_string is a problem. Usually this is actually
		       *   a file port where the contents were read in one (up to 5MB) gulp, so the
		       *   readable version could be: open file, read-char to the current port_position.
		       *   s7_port_filename(port) has the char* name if any.
		       */
		      int32_t data_len;
		      data_len = port_data_size(obj) - port_position(obj);
		      if (data_len > 100)
			{
			  const char *filename;
			  filename = (const char *)s7_port_filename(obj);
			  if (filename)
			    {
			      #define DO_STR_LEN 1024
			      char *do_str;
			      int32_t len;
			      do_str = (char *)malloc(DO_STR_LEN * sizeof(char));
			      if (port_position(obj) > 0)
				{
				  len = snprintf(do_str, DO_STR_LEN, "(let ((port (open-input-file \"%s\")))", filename);
				  port_write_string(port)(sc, do_str, len, port);
				  len = snprintf(do_str, DO_STR_LEN, " (do ((i 0 (+ i 1)) (c (read-char port) (read-char port))) ((= i %u) port)))", 
						 port_position(obj) - 1);
				}
			      else len = snprintf(do_str, DO_STR_LEN, "(open-input-file \"%s\")", filename);
			      port_write_string(port)(sc, do_str, len, port);
			      free(do_str);
			      return;
			    }
			}
		      port_write_string(port)(sc, "(open-input-string ", 19, port);
		      /* not port_write_string here because there might be embedded double-quotes */
		      str = slashify_string(sc, (const char *)(port_data(obj) + port_position(obj)), port_data_size(obj) - port_position(obj), IN_QUOTES, &nlen);
		      port_write_string(port)(sc, str, nlen, port);
		      port_write_character(port)(sc, ')', port);
		    }
		}
	    }
	}
      else
	{
	  if (is_string_port(obj))
	    port_write_string(port)(sc, "<input-string-port", 18, port);
	  else
	    {
	      if (is_file_port(obj))
		port_write_string(port)(sc, "<input-file-port", 16, port);
	      else port_write_string(port)(sc, "<input-function-port", 20, port);
	    }
	  if (port_is_closed(obj)) 
	    port_write_string(port)(sc, " (closed)>", 10, port);
	  else port_write_character(port)(sc, '>', port);
	}
    }
}

static bool symbol_needs_slashification(s7_pointer obj)
{
  unsigned char *p, *pend;
  const char *str;
  int32_t len;
  str = symbol_name(obj);
  if (str[0] == '#')
    return(true);
  len = symbol_name_length(obj);
  pend = (unsigned char *)(str + len);
  for (p = (unsigned char *)str; p < pend; p++)
    if (symbol_slashify_table[*p])
      return(true);
  set_clean_symbol(obj);
  return(false);
}

static void symbol_to_port(s7_scheme *sc, s7_pointer obj, s7_pointer port, use_write_t use_write, shared_info *ci)
{
  /* I think this is the only place we print a symbol's name
   *   but in the readable case, what about (symbol "1;3")? it actually seems ok!
   */
  if ((!is_clean_symbol(obj)) &&
      (symbol_needs_slashification(obj)))
    {
      int32_t nlen = 0;
      char *str, *symstr;
      str = slashify_string(sc, symbol_name(obj), symbol_name_length(obj), NOT_IN_QUOTES, &nlen);
      nlen += 16;
      tmpbuf_malloc(symstr, nlen);
      nlen = snprintf(symstr, nlen, "(symbol \"%s\")", str);
      port_write_string(port)(sc, symstr, nlen, port);
      tmpbuf_free(symstr, nlen);
    }
  else
    {
      if ((use_write == USE_READABLE_WRITE) &&
	  (!is_keyword(obj)))
	port_write_character(port)(sc, '\'', port);
      if (is_string_port(port))
	{
	  uint32_t new_len;
	  new_len = port_position(port) + symbol_name_length(obj);
	  if (new_len >= port_data_size(port))
	    resize_port_data(port, new_len * 2);
	  memcpy((void *)(port_data(port) + port_position(port)), (void *)symbol_name(obj), symbol_name_length(obj));
	  port_position(port) = new_len;
	}
      else port_write_string(port)(sc, symbol_name(obj), symbol_name_length(obj), port);
    }
}

static void string_to_port(s7_scheme *sc, s7_pointer obj, s7_pointer port, use_write_t use_write)
{
  if (string_length(obj) > 0)
    {
      /* since string_length is a scheme length, not C, this write can embed nulls from C's point of view */
      if (string_length(obj) > 10000)
	{
	  size_t size;
	  char buf[128];
	  buf[0] = string_value(obj)[0];
	  buf[1] = '\0';
	  size = strspn((const char *)(string_value(obj) + 1), buf); /* if all #\null, this won't work */
	  if (size == string_length(obj) - 1)
	    {
	      int32_t nlen;
	      s7_pointer c;
	      c = chars[(int32_t)buf[0]];
	      nlen = snprintf(buf, 128, "(make-string %u ", string_length(obj));
	      port_write_string(port)(sc, buf, nlen, port);
	      port_write_string(port)(sc, character_name(c), character_name_length(c), port);
	      port_write_character(port)(sc, ')', port);
	      return;
	    }
	}
      if (use_write == USE_DISPLAY)
	port_write_string(port)(sc, string_value(obj), string_length(obj), port);
      else
	{
	  if (!string_needs_slashification(string_value(obj), string_length(obj)))
	    {
	      port_write_character(port)(sc, '"', port);
	      port_write_string(port)(sc, string_value(obj), string_length(obj), port);
	      port_write_character(port)(sc, '"', port);
	    }
	  else
	    {
	      char *str;
	      int32_t nlen = 0;
	      str = slashify_string(sc, string_value(obj), string_length(obj), IN_QUOTES, &nlen);
	      port_write_string(port)(sc, str, nlen, port);
	    }
	}
    }
  else
    {
      if (use_write != USE_DISPLAY)
	port_write_string(port)(sc, "\"\"", 2, port);
    }
}


static char *multivector_indices_to_string(s7_scheme *sc, s7_int index, s7_pointer vect, char *str, int32_t cur_dim)
{
  s7_int size, ind;
  char buf[64];

  size = vector_dimension(vect, cur_dim);
  ind = index % size;
  if (cur_dim > 0)
    multivector_indices_to_string(sc, (index - ind) / size, vect, str, cur_dim - 1);

  snprintf(buf, 64, " %" PRId64, ind);
#ifdef __OpenBSD__
  strlcat(str, buf, 128); /* 128=length of str */
#else
  strcat(str, buf);
#endif
  return(str);
}


static int32_t multivector_to_port(s7_scheme *sc, s7_pointer vec, s7_pointer port,
			       int32_t out_len, int32_t flat_ref, int32_t dimension, int32_t dimensions, bool *last,
			       use_write_t use_write, shared_info *ci)
{
  int32_t i;

  if (use_write != USE_READABLE_WRITE)
    {
      if (*last)
	port_write_string(port)(sc, " (", 2, port);
      else port_write_character(port)(sc, '(', port);
      (*last) = false;
    }

  for (i = 0; i < vector_dimension(vec, dimension); i++)
    {
      if (dimension == (dimensions - 1))
	{
	  if (flat_ref < out_len)
	    {
	      if (use_write == USE_READABLE_WRITE)
		{
		  int32_t plen;
		  char buf[128];
		  char *indices;
		  /* need to translate flat_ref into a set of indices
		   */
		  tmpbuf_calloc(indices, 128);
		  plen = snprintf(buf, 128, "(set! ({v}%s) ", multivector_indices_to_string(sc, flat_ref, vec, indices, dimension));
		  port_write_string(port)(sc, buf, plen, port);
		  tmpbuf_free(indices, 128);
		}
	      object_to_port_with_circle_check(sc, vector_getter(vec)(sc, vec, flat_ref), port, DONT_USE_DISPLAY(use_write), ci);

	      if (use_write == USE_READABLE_WRITE)
		port_write_string(port)(sc, ") ", 2, port);
	      flat_ref++;
	    }
	  else
	    {
	      port_write_string(port)(sc, "...)", 4, port);
	      return(flat_ref);
	    }
	  if ((use_write != USE_READABLE_WRITE) &&
	      (i < (vector_dimension(vec, dimension) - 1)))
	    port_write_character(port)(sc, ' ', port);
	}
      else
	{
	  if (flat_ref < out_len)
	    flat_ref = multivector_to_port(sc, vec, port, out_len, flat_ref, dimension + 1, dimensions, last, DONT_USE_DISPLAY(use_write), ci);
	  else
	    {
	      port_write_string(port)(sc, "...)", 4, port);
	      return(flat_ref);
	    }
	}
    }
  if (use_write != USE_READABLE_WRITE)
    port_write_character(port)(sc, ')', port);
  (*last) = true;
  return(flat_ref);
}


static void make_vector_to_port(s7_scheme *sc, s7_pointer vect, s7_pointer port)
{
  s7_int vlen;
  int32_t plen;
  char buf[128];
  const char* vtyp = "";

  if (is_float_vector(vect))
    vtyp = "float-";
  else
    {
      if (is_int_vector(vect))
	vtyp = "int-";
    }

  vlen = vector_length(vect);
  if (vector_rank(vect) == 1)
    {
      plen = snprintf(buf, 128, "(make-%svector %" PRId64 " ", vtyp, vlen);
      port_write_string(port)(sc, buf, plen, port);
    }
  else
    {
      uint32_t dim;
      plen = snprintf(buf, 128, "(make-%svector '(", vtyp);
      port_write_string(port)(sc, buf, plen, port);
      for (dim = 0; dim < vector_ndims(vect) - 1; dim++)
	{
	  plen = snprintf(buf, 128, "%" PRId64 " ", vector_dimension(vect, dim));
	  port_write_string(port)(sc, buf, plen, port);
	}
      plen = snprintf(buf, 128, "%" PRId64 ") ", vector_dimension(vect, dim));
      port_write_string(port)(sc, buf, plen, port);
    }
}
  
static void vector_to_port(s7_scheme *sc, s7_pointer vect, s7_pointer port, use_write_t use_write, shared_info *ci)
{
  s7_int i, len;
  int32_t plen;
  bool too_long = false;
  char buf[128];

  len = vector_length(vect);
  if (len == 0)
    {
      if (vector_rank(vect) > 1)
	{
	  plen = snprintf(buf, 32, "#%uD()", vector_ndims(vect));
	  port_write_string(port)(sc, buf, plen, port);
	}
      else port_write_string(port)(sc, "#()", 3, port);
      return;
    }

  if (use_write != USE_READABLE_WRITE)
    {
      if (sc->print_length <= 0)
	{
	  if (vector_rank(vect) > 1)
	    {
	      plen = snprintf(buf, 32, "#%uD(...)", vector_ndims(vect));
	      port_write_string(port)(sc, buf, plen, port);
	    }
	  else port_write_string(port)(sc, "#(...)", 6, port);
	  return;
	}

      if (len > sc->print_length)
	{
	  too_long = true;
	  len = sc->print_length;
	}
    }
  if ((!ci) &&
      (len > 1000))
    {
      s7_int vlen;
      s7_pointer p0;
      s7_pointer *els;
      vlen = vector_length(vect);
      els = vector_elements(vect);
      p0 = els[0];
      for (i = 1; i < vlen; i++)
	if (els[i] != p0)
	  break;
      if (i == vlen)
	{
	  make_vector_to_port(sc, vect, port);
	  object_to_port(sc, p0, port, use_write, NULL);
	  port_write_character(port)(sc, ')', port);
	}
      return;
    }

  if (use_write == USE_READABLE_WRITE)
    {
      if ((ci) &&
	  (is_collected(vect)) &&
	  (peek_shared_ref(ci, vect) != 0))
	{
	  port_write_string(port)(sc, "(let (({v} (make-vector ", 24, port);
	  if (vector_rank(vect) > 1)
	    {
	      uint32_t dim;
	      port_write_string(port)(sc, "'(", 2, port);
	      for (dim = 0; dim < vector_ndims(vect); dim++)
		{
		  plen = snprintf(buf, 128, "%" PRId64 " ", vector_dimension(vect, dim));
		  port_write_string(port)(sc, buf, plen, port);
		}
	      port_write_string(port)(sc, ")))) ", 5, port);
	    }
	  else 
	    {
	      plen = snprintf(buf, 128, "%" PRId64 "))) ", vector_length(vect));
	      port_write_string(port)(sc, buf, plen, port);
	    }
	  if (shared_ref(ci, vect) < 0)
	    {
	      plen = snprintf(buf, 128, "(set! {%d} {v}) ", -shared_ref(ci, vect));
	      port_write_string(port)(sc, buf, plen, port);
	    }
	  
	  if (vector_rank(vect) > 1)
	    {
	      bool last = false;
	      multivector_to_port(sc, vect, port, len, 0, 0, vector_ndims(vect), &last, use_write, ci);
	    }
	  else
	    {
	      for (i = 0; i < len; i++)
		{
		  port_write_string(port)(sc, "(set! ({v} ", 11, port);
		  plen = snprintf(buf, 128, "%" PRId64 ") ", i);
		  port_write_string(port)(sc, buf, plen, port);
		  object_to_port_with_circle_check(sc, vector_element(vect, i), port, use_write, ci);
		  port_write_string(port)(sc, ") ", 2, port);
		}
	    }
	  port_write_string(port)(sc, "{v})", 4, port);
	}
      else /* simple readable case */
	{
	  if (vector_rank(vect) > 1)
	    port_write_string(port)(sc, "(make-shared-vector (vector", 27, port);
	  else port_write_string(port)(sc, "(vector", 7, port);

	  for (i = 0; i < len; i++)
	    {
	      port_write_character(port)(sc, ' ', port);
	      object_to_port_with_circle_check(sc, vector_element(vect, i), port, use_write, ci);
	    }
	  port_write_character(port)(sc, ')', port);

	  if (vector_rank(vect) > 1)
	    {
	      uint32_t dim;
	      port_write_string(port)(sc, " '(", 3, port);
	      for (dim = 0; dim < vector_ndims(vect) - 1; dim++)
		{
		  plen = snprintf(buf, 128, "%" PRId64 " ", vector_dimension(vect, dim));
		  port_write_string(port)(sc, buf, plen, port);
		}
	      plen = snprintf(buf, 128, "%" PRId64, vector_dimension(vect, dim));
	      port_write_string(port)(sc, buf, plen, port);
	      port_write_string(port)(sc, "))", 2, port);
	    }
	}
    }
  else
    {
      if (vector_rank(vect) > 1)
	{
	  bool last = false;
	  if (vector_ndims(vect) > 1)
	    {
	      plen = snprintf(buf, 32, "#%uD", vector_ndims(vect));
	      port_write_string(port)(sc, buf, plen, port);
	    }
	  else port_write_character(port)(sc, '#', port);
	  multivector_to_port(sc, vect, port, len, 0, 0, vector_ndims(vect), &last, use_write, ci);
	}
      else
	{
	  port_write_string(port)(sc, "#(", 2, port);
	  for (i = 0; i < len - 1; i++)
	    {
	      object_to_port_with_circle_check(sc, vector_element(vect, i), port, DONT_USE_DISPLAY(use_write), ci);
	      port_write_character(port)(sc, ' ', port);
	    }
	  object_to_port_with_circle_check(sc, vector_element(vect, i), port, DONT_USE_DISPLAY(use_write), ci);

	  if (too_long)
	    port_write_string(port)(sc, " ...)", 5, port);
	  else port_write_character(port)(sc, ')', port);
	}
    }
}


static void int_or_float_vector_to_port(s7_scheme *sc, s7_pointer vect, s7_pointer port, use_write_t use_write, shared_info *ignored)
{
  s7_int i, len;
  int32_t plen;
  bool too_long = false;
  char buf[128];

  len = vector_length(vect);
  if (len == 0)
    {
      if (vector_rank(vect) > 1)
	{
	  plen = snprintf(buf, 32, "#%c%uD()", (is_int_vector(vect)) ? 'i' : 'r', vector_ndims(vect));
	  port_write_string(port)(sc, buf, plen, port);
	}
      else port_write_string(port)(sc, "#()", 3, port);
      return;
    }

  if (use_write == USE_READABLE_WRITE)
    plen = len;
  else plen = sc->print_length;

  if (plen <= 0)
    {
      if (vector_rank(vect) > 1)
	{
	  plen = snprintf(buf, 32, "#%c%uD(...)", (is_int_vector(vect)) ? 'i' : 'r', vector_ndims(vect));
	  port_write_string(port)(sc, buf, plen, port);
	}
      else 
	{
	  if (is_int_vector(vect))
	    port_write_string(port)(sc, "#i(...)", 7, port);
	  else port_write_string(port)(sc, "#r(...)", 7, port);
	}
      return;
    }

  if (len > plen)
    {
      too_long = true;
      len = plen;
    }

  if (len > 1000)
    {
      s7_int vlen;
      vlen = vector_length(vect);
      if (is_float_vector(vect))
	{
	  s7_double first;
	  s7_double *els;
	  els = float_vector_elements(vect);
	  first = els[0];
	  for (i = 1; i < vlen; i++)
	    if (els[i] != first)
	      break;
	}
      else
	{
	  s7_int first;
	  s7_int *els;
	  els = int_vector_elements(vect);
	  first = els[0];
	  for (i = 1; i < vlen; i++)
	    if (els[i] != first)
	      break;
	}
      if (i == vlen)
	{
	  make_vector_to_port(sc, vect, port);
	  if (is_float_vector(vect))
	    plen = snprintf(buf, 128, float_format_g, float_format_precision, float_vector_element(vect, 0));
	  else plen = snprintf(buf, 128, "%" PRId64, int_vector_element(vect, 0));
	  port_write_string(port)(sc, buf, plen, port);
	  port_write_character(port)(sc, ')', port);
	}
      return;
    }

  if (vector_rank(vect) == 1)
    {
      if (is_int_vector(vect))
	{
	  port_write_string(port)(sc, "#i(", 3, port);
	  if (!is_string_port(port))
	    {
	      plen = snprintf(buf, 128, "%" PRId64, int_vector_element(vect, 0));
	      port_write_string(port)(sc, buf, plen, port);
	      for (i = 1; i < len; i++)
		{
		  plen = snprintf(buf, 128, " %" PRId64, int_vector_element(vect, i));
		  port_write_string(port)(sc, buf, plen, port);
		}
	    }
	  else
	    {
	      /* an experiment */
	      uint32_t new_len, next_len;
	      unsigned char *dbuf;
	      new_len = port_position(port);
	      next_len = port_data_size(port) - 128;
	      dbuf = port_data(port);
	      
	      if (new_len >= next_len)
		{
		  resize_port_data(port, port_data_size(port) * 2);
		  next_len = port_data_size(port) - 128;
		  dbuf = port_data(port);
		}
	      plen = snprintf((char *)(dbuf + new_len), 128, "%" PRId64, int_vector_element(vect, 0));
	      new_len += plen;
	      for (i = 1; i < len; i++)
		{
		  if (new_len >= next_len)
		    {
		      resize_port_data(port, port_data_size(port) * 2);
		      next_len = port_data_size(port) - 128;
		      dbuf = port_data(port);
		    }
		  plen = snprintf((char *)(dbuf + new_len), 128, " %" PRId64, int_vector_element(vect, i));
		  new_len += plen;
		}
	      port_position(port) = new_len;
	    }
	}
      else
	{
	  port_write_string(port)(sc, "#r(", 3, port);
	  plen = snprintf(buf, 124, float_format_g, float_format_precision, float_vector_element(vect, 0)); /* 124 so floatify has room */
	  floatify(buf, &plen);
	  port_write_string(port)(sc, buf, plen, port);
	  for (i = 1; i < len; i++)
	    {
	      port_write_character(port)(sc, ' ', port);
	      plen = snprintf(buf, 124, float_format_g, float_format_precision, float_vector_element(vect, i));
	      floatify(buf, &plen);
	      port_write_string(port)(sc, buf, plen, port);
	    }
	}
      
      if (too_long)
	port_write_string(port)(sc, " ...)", 5, port);
      else port_write_character(port)(sc, ')', port);
      return;
    }

  /* multidimensional case */
  {
    bool last = false;
    plen = snprintf(buf, 32, "#%c%uD", (is_int_vector(vect)) ? 'i' : 'r', vector_ndims(vect));
    port_write_string(port)(sc, buf, plen, port);
    multivector_to_port(sc, vect, port, len, 0, 0, vector_ndims(vect), &last, USE_DISPLAY, NULL);
  }
}


static void byte_vector_to_port(s7_scheme *sc, s7_pointer vect, s7_pointer port, use_write_t use_write)
{
  s7_int i, len;
  int32_t plen;
  bool too_long = false;

  len = string_length(vect);
  if (use_write == USE_READABLE_WRITE)
    plen = len;
  else plen = sc->print_length;

  if (len == 0)
    port_write_string(port)(sc, "#u8()", 5, port);
  else
    {
      if (plen <= 0)
	port_write_string(port)(sc, "#u8(...)", 8, port);
      else
	{
	  uint32_t nlen;
	  char *p;
	  if (len > plen)
	    {
	      too_long = true;
	      len = plen;
	    }

	  if (len > 1000)
	    {
	      uint32_t vlen;
	      char c;
	      char *data;
	      vlen = string_length(vect);
	      data = string_value(vect);
	      c = data[0];
	      for (i = 1; i < vlen; i++)
		if (data[i] != c)
		  break;
	      if (i == vlen)
		{
		  char buf[128];
		  plen = snprintf(buf, 128, "(make-byte-vector %u ", vlen);
		  port_write_string(port)(sc, buf, plen, port);
		  plen = snprintf(buf, 128, "%u)", (uint32_t)c);
		  port_write_string(port)(sc, buf, plen, port);
		  return;
		}
	    }

	  port_write_string(port)(sc, "#u8(", 4, port);
	  for (i = 0; i < len - 1; i++)
	    {
	      p = pos_int_to_str((int32_t)((unsigned char)string_value(vect)[i]), &nlen, ' ');
	      port_write_string(port)(sc, p, nlen - 1, port);
	    }
	  p = pos_int_to_str((int32_t)((unsigned char)string_value(vect)[i]), &nlen, (too_long) ? '\0' : ')');
	  port_write_string(port)(sc, p, nlen - 1, port);

	  if (too_long)
	    port_write_string(port)(sc, " ...)", 5, port);
	}
    }
}


static void pair_to_port(s7_scheme *sc, s7_pointer lst, s7_pointer port, use_write_t use_write, shared_info *ci)
{
  /* we need list_to_starboard... */
  s7_pointer x;
  s7_int i, len, true_len;

  true_len = s7_list_length(sc, lst);
  if (true_len < 0)                    /* a dotted list -- handle cars, then final cdr */
    len = (-true_len + 1);
  else
    {
      if (true_len == 0)               /* either () or a circular list */
	{
	  if (is_not_null(lst))
	    len = circular_list_entries(lst);
	  else
	    {
	      port_write_string(port)(sc, "()", 2, port);
	      return;
	    }
	}
      else len = true_len;
    }

  if (((car(lst) == sc->quote_symbol) ||
       (car(lst) == sc->quote_unchecked_symbol)) && /* this can happen (see lint.scm) */
      (true_len == 2))
    {
      /* len == 1 is important, otherwise (list 'quote 1 2) -> '1 2 which looks weird
       *   or (object->string (apply . `''1)) -> "'quote 1"
       * so (quote x) = 'x but (quote x y z) should be left alone (if evaluated, it's an error)
       */
      port_write_character(port)(sc, '\'', port);
      object_to_port_with_circle_check(sc, cadr(lst), port, USE_WRITE, ci);
      return;
    }
  else port_write_character(port)(sc, '(', port);

  if (is_multiple_value(lst))
    port_write_string(port)(sc, "values ", 7, port);

  if (use_write == USE_READABLE_WRITE)
    {
      if (ci)
	{
	  int32_t plen;
	  char buf[128];

	  port_write_string(port)(sc, "let (({lst} (make-list ", 23, port);
	  plen = snprintf(buf, 128, "%" PRId64 "))) ", len);
	  port_write_string(port)(sc, buf, plen, port);

	  if ((shared_ref(ci, lst) < 0))
	    {
	      plen = snprintf(buf, 128, "(set! {%d} {lst}) ", -shared_ref(ci, lst));
	      port_write_string(port)(sc, buf, plen, port);
	    }

	  port_write_string(port)(sc, "(let (({x} {lst})) ", 19, port);
	  for (i = 0, x = lst; (i < len) && (is_pair(x)); i++, x = cdr(x))
	    {
	      port_write_string(port)(sc, "(set-car! {x} ", 14, port);
	      object_to_port_with_circle_check(sc, car(x), port, use_write, ci);
	      port_write_string(port)(sc, ") ", 2, port);
	      if (i < len - 1)
		port_write_string(port)(sc, "(set! {x} (cdr {x})) ", 21, port);
	    }
	  if (!is_null(x))
	    {
	      port_write_string(port)(sc, "(set-cdr! {x} ", 14, port);
	      object_to_port_with_circle_check(sc, x, port, use_write, ci);
	      port_write_string(port)(sc, ") ", 2, port);
	    }
	  port_write_string(port)(sc, ") {lst})", 8, port);
	}
      else
	{
	  /* the easier cases: no circles or shared refs to patch up */
	  if (true_len > 0)
	    {
	      port_write_string(port)(sc, "list", 4, port);
	      for (x = lst; is_pair(x); x = cdr(x))
		{
		  port_write_character(port)(sc, ' ', port);
		  object_to_port_with_circle_check(sc, car(x), port, use_write, ci);
		}
	      port_write_character(port)(sc, ')', port);
	    }
	  else
	    {
	      port_write_string(port)(sc, "cons ", 5, port);
	      object_to_port_with_circle_check(sc, car(lst), port, use_write, ci);
	      for (x = cdr(lst); is_pair(x); x = cdr(x))
		{
		  port_write_character(port)(sc, ' ', port);
		  port_write_string(port)(sc, "(cons ", 6, port);
		  object_to_port_with_circle_check(sc, car(x), port, use_write, ci);
		}
	      port_write_character(port)(sc, ' ', port);
	      object_to_port_with_circle_check(sc, x, port, use_write, ci);
	      for (i = 1; i < len; i++)
		port_write_character(port)(sc, ')', port);
	    }
	}
    }
  else
    {
      if (ci)
	{
	  for (x = lst, i = 0; (is_pair(x)) && (i < len) && ((!ci) || (i == 0) || (!is_collected(x)) || (peek_shared_ref(ci, x) == 0)); i++, x = cdr(x))
	    {
	      object_to_port_with_circle_check(sc, car(x), port, DONT_USE_DISPLAY(use_write), ci);
	      if (i < (len - 1))
		port_write_character(port)(sc, ' ', port);
	    }
	  if (is_not_null(x))
	    {
	      if ((true_len == 0) &&
		  (i == len))
		port_write_string(port)(sc, " . ", 3, port);
	      else port_write_string(port)(sc, ". ", 2, port);
	      object_to_port_with_circle_check(sc, x, port, DONT_USE_DISPLAY(use_write), ci);
	    }
	  port_write_character(port)(sc, ')', port);
	}
      else
	{
	  s7_int len1;
	  len1 = len - 1;
	  if (is_string_port(port))
	    {
	      for (x = lst, i = 0; (is_pair(x)) && (i < len1); i++, x = cdr(x))
		{
		  object_to_port(sc, car(x), port, DONT_USE_DISPLAY(use_write), ci);
		  if (port_position(port) >= sc->objstr_max_len)
		    return;
		  if (port_position(port) >= port_data_size(port))
		    resize_port_data(port, port_data_size(port) * 2);
		  port_data(port)[port_position(port)++] = (unsigned char)' ';
		}
	    }
	  else
	    {
	      for (x = lst, i = 0; (is_pair(x)) && (i < len1); i++, x = cdr(x))
		{
		  object_to_port(sc, car(x), port, DONT_USE_DISPLAY(use_write), ci);
		  port_write_character(port)(sc, ' ', port);
		}
	    }
	  if (is_pair(x))
	    {
	      object_to_port(sc, car(x), port, DONT_USE_DISPLAY(use_write), ci);
	      x = cdr(x);
	    }
	  if (is_not_null(x))
	    {
	      port_write_string(port)(sc, ". ", 2, port);
	      object_to_port(sc, x, port, DONT_USE_DISPLAY(use_write), ci);
	    }
	  port_write_character(port)(sc, ')', port);
	}
    }
}


static void hash_table_to_port(s7_scheme *sc, s7_pointer hash, s7_pointer port, use_write_t use_write, shared_info *ci)
{
  int32_t i, len;
  uint32_t gc_iter;
  bool too_long = false;
  s7_pointer iterator, p;

  /* if hash is a member of ci, just print its number
   * (let ((ht (hash-table '(a . 1)))) (hash-table-set! ht 'b ht))
   *
   * since equal? does not care about the hash-table lengths, we can ignore that complication in the :readable case
   */
  len = hash_table_entries(hash);
  if (len == 0)
    {
      port_write_string(port)(sc, "(hash-table)", 12, port);
      return;
    }

  if (use_write != USE_READABLE_WRITE)
    {
      s7_int plen;
      plen = sc->print_length;
      if (plen <= 0)
	{
	  port_write_string(port)(sc, "(hash-table ...)", 16, port);
	  return;
	}
      if (len > plen)
	{
	  too_long = true;
	  len = plen;
	}
    }

  iterator = s7_make_iterator(sc, hash);
  gc_iter = s7_gc_protect(sc, iterator);
  p = cons(sc, sc->F, sc->F);
  iterator_current(iterator) = p;
  set_mark_seq(iterator);

  if ((use_write == USE_READABLE_WRITE) &&
      (ci) &&
      (is_collected(hash)) &&
      (peek_shared_ref(ci, hash) != 0))
    {
      port_write_string(port)(sc, "(let (({ht} (make-hash-table)))", 31, port);
      if (shared_ref(ci, hash) < 0)
	{
	  int32_t plen;
	  char buf[64];
	  plen = snprintf(buf, 64, "(set! {%d} {ht}) ", -shared_ref(ci, hash));
	  port_write_string(port)(sc, buf, plen, port);
	}
      for (i = 0; i < len; i++)
	{
	  s7_pointer key_val, key, val;

	  key_val = hash_table_iterate(sc, iterator);
	  key = car(key_val);
	  val = cdr(key_val);

	  port_write_string(port)(sc, " (set! ({ht} ", 13, port);
	  if (key == hash)
	    port_write_string(port)(sc, "{ht}", 4, port);
	  else object_to_port_with_circle_check(sc, key, port, USE_READABLE_WRITE, ci);
	  port_write_string(port)(sc, ") ", 2, port);
	  if (val == hash)
	    port_write_string(port)(sc, "{ht}", 4, port);
	  else object_to_port_with_circle_check(sc, val, port, USE_READABLE_WRITE, ci);
	  port_write_character(port)(sc, ')', port);
	}
      port_write_string(port)(sc, " {ht})", 6, port);
    }
  else
    {
      port_write_string(port)(sc, "(hash-table", 11, port);
      for (i = 0; i < len; i++)
	{
	  s7_pointer key_val;
	  if (use_write == USE_READABLE_WRITE)
	    port_write_character(port)(sc, ' ', port);
	  else port_write_string(port)(sc, " '", 2, port);
	  key_val = hash_table_iterate(sc, iterator);
	  object_to_port_with_circle_check(sc, key_val, port, DONT_USE_DISPLAY(use_write), ci);
	}

      if (too_long)
	port_write_string(port)(sc, " ...)", 5, port);
      else port_write_character(port)(sc, ')', port);
    }

  s7_gc_unprotect_at(sc, gc_iter);
  iterator_current(iterator) = sc->nil;
  free_cell(sc, p);
  free_cell(sc, iterator);
}


static int32_t slot_to_port_1(s7_scheme *sc, s7_pointer x, s7_pointer port, use_write_t use_write, shared_info *ci, int32_t n)
{
  if (is_slot(x))
    {
      n = slot_to_port_1(sc, next_slot(x), port, use_write, ci, n);
      if (n <= sc->print_length)
	{
	  port_write_character(port)(sc, ' ', port);
	  object_to_port_with_circle_check(sc, x, port, use_write, ci);
	}
      if (n == (sc->print_length + 1))
	port_write_string(port)(sc, " ...", 4, port);
    }
  return(n + 1);
}

static void let_to_port(s7_scheme *sc, s7_pointer obj, s7_pointer port, use_write_t use_write, shared_info *ci)
{
  /* if outer env points to (say) method list, the object needs to specialize object->string itself */
  if (has_methods(obj))
    {
      s7_pointer print_func;
      print_func = find_method(sc, obj, sc->object_to_string_symbol);
      if (print_func != sc->undefined)
	{
	  s7_pointer p;
	  /* what needs to be protected here? for one, the function might not return a string! */

	  clear_has_methods(obj);
	  if (use_write == USE_WRITE)
	    p = s7_apply_function(sc, print_func, list_1(sc, obj));
	  else p = s7_apply_function(sc, print_func, list_2(sc, obj, (use_write == USE_DISPLAY) ? sc->F : sc->key_readable_symbol));
	  set_has_methods(obj);

	  if ((is_string(p)) && 
	      (string_length(p) > 0))
	    port_write_string(port)(sc, string_value(p), string_length(p), port);
	  return;
	}
    }
  if (obj == sc->rootlet)
    port_write_string(port)(sc, "(rootlet)", 9, port);
  else
    {
      if (sc->short_print)
	port_write_string(port)(sc, "#<let>", 6, port);
      else
	{
	  /* circles can happen here:
	   *    (let () (let ((b (curlet))) (curlet)))
	   *    #<let 'b #<let>>
	   * or (let ((b #f)) (set! b (curlet)) (curlet))
	   *    #1=#<let 'b #1#>
	   */
	  if ((use_write == USE_READABLE_WRITE) &&
	      (ci) &&
	      (is_collected(obj)) &&
	      (peek_shared_ref(ci, obj) != 0))
	    {
	      s7_pointer x;
	      port_write_string(port)(sc, "(let (({e} (inlet))) ", 21, port);
	      if ((ci) &&
		  (shared_ref(ci, obj) < 0))
		{
		  int32_t plen;
		  char buf[64];
		  plen = snprintf(buf, 64, "(set! {%d} {e}) ", -shared_ref(ci, obj));
		  port_write_string(port)(sc, buf, plen, port);
		}
	      
	      port_write_string(port)(sc, "(apply varlet {e} (reverse (list ", 33, port);
	      for (x = let_slots(obj); is_slot(x); x = next_slot(x))
		{
		  port_write_string(port)(sc, "(cons ", 6, port);
		  symbol_to_port(sc, slot_symbol(x), port, use_write, NULL);
		  port_write_character(port)(sc, ' ', port);
		  object_to_port_with_circle_check(sc, slot_value(x), port, use_write, ci);
		  port_write_character(port)(sc, ')', port);
		}
	      port_write_string(port)(sc, "))) {e})", 8, port);
	    }
	  else
	    {
	      port_write_string(port)(sc, "(inlet", 6, port);
	      slot_to_port_1(sc, let_slots(obj), port, use_write, ci, 0);
	      port_write_character(port)(sc, ')', port);
	    }
	}
    }
}


static void write_macro_readably(s7_scheme *sc, s7_pointer obj, s7_pointer port)
{
  s7_pointer arglist, body, expr;

  body = closure_body(obj);
  arglist = closure_args(obj);

  port_write_string(port)(sc, "(define-", 8, port);
  port_write_string(port)(sc, ((is_macro(obj)) || (is_macro_star(obj))) ? "macro" : "bacro", 5, port);
  if ((is_macro_star(obj)) || (is_bacro_star(obj)))
    port_write_character(port)(sc, '*', port);
  port_write_string(port)(sc, " (_m_", 5, port);
  if (is_symbol(arglist))
    {
      port_write_string(port)(sc, " . ", 3, port);
      port_write_string(port)(sc, symbol_name(arglist), symbol_name_length(arglist), port);
    }
  else
    {
      if (is_pair(arglist))
	{
	  for (expr = arglist; is_pair(expr); expr = cdr(expr))
	    {
	      port_write_character(port)(sc, ' ', port);
	      object_to_port(sc, car(expr), port, USE_WRITE, NULL);
	    }
	  if (!is_null(expr))
	    {
	      port_write_string(port)(sc, " . ", 3, port);
	      object_to_port(sc, expr, port, USE_WRITE, NULL);
	    }
	}
    }
  port_write_string(port)(sc, ") ", 2, port);
  for (expr = body; is_pair(expr); expr = cdr(expr))
    object_to_port(sc, car(expr), port, USE_WRITE, NULL);
  port_write_character(port)(sc, ')', port);
}


static s7_pointer match_symbol(s7_scheme *sc, s7_pointer symbol, s7_pointer e)
{
  s7_pointer y, le;
  for (le = e; is_let(le) && (le != sc->rootlet); le = outlet(le))
    for (y = let_slots(le); is_slot(y); y = next_slot(y))
      if (slot_symbol(y) == symbol)
	return(y);
  return(NULL);
}

static bool slot_memq(s7_pointer symbol, s7_pointer symbols)
{
  s7_pointer x;
  for (x = symbols; is_pair(x); x = cdr(x))
    if (slot_symbol(car(x)) == symbol)
      return(true);
  return(false);
}

static bool arg_memq(s7_pointer symbol, s7_pointer args)
{
  s7_pointer x;
  for (x = args; is_pair(x); x = cdr(x))
    if ((car(x) == symbol) ||
	((is_pair(car(x))) &&
	 (caar(x) == symbol)))
      return(true);
  return(false);
}


static void collect_locals(s7_scheme *sc, s7_pointer body, s7_pointer e, s7_pointer args, uint32_t gc_loc)
{
  if (is_pair(body))
    {
      collect_locals(sc, car(body), e, args, gc_loc);
      collect_locals(sc, cdr(body), e, args, gc_loc);
    }
  else
    {
      if ((is_symbol(body)) &&
	  (!arg_memq(body, args)) &&
	  (!slot_memq(body, gc_protected_at(sc, gc_loc))))
	{
	  s7_pointer slot;
	  slot = match_symbol(sc, body, e);
	  if (slot)
	    gc_protected_at(sc, gc_loc) = cons(sc, slot, gc_protected_at(sc, gc_loc));
	}
    }
}


static s7_pointer find_closure(s7_scheme *sc, s7_pointer closure, s7_pointer cur_env)
{
  s7_pointer e, y;
  for (e = cur_env; is_let(e); e = outlet(e))
    {
      if ((is_funclet(e)) &&
	  (is_global(funclet_function(e))) &&         /* (define (f1) (lambda () 1)) shouldn't say the returned closure is named f1 */
	  (slot_value(global_slot(funclet_function(e))) == closure))
	return(funclet_function(e));

      for (y = let_slots(e); is_slot(y); y = next_slot(y))
	if (slot_value(y) == closure)
	  return(slot_symbol(y));
    }
  return(sc->nil);
}

static void write_closure_name(s7_scheme *sc, s7_pointer closure, s7_pointer port)
{
  s7_pointer x;
  x = find_closure(sc, closure, closure_let(closure));
  /* this can be confusing!  In some cases, the function is in its environment, and in other very similar-looking cases it isn't:
   * (let ((a (lambda () 1))) a)
   * #<lambda ()>
   * (letrec ((a (lambda () 1))) a)
   * a
   * (let () (define (a) 1) a)
   * a
   */
  if (is_symbol(x)) /* after find_closure */
    {
      port_write_string(port)(sc, symbol_name(x), symbol_name_length(x), port);
      return;
    }
  
  /* names like #<closure> and #<macro> are useless -- try to be a bit more informative */
  switch (type(closure))
    {
    case T_CLOSURE:      
      port_write_string(port)(sc, "#<lambda ", 9, port);  
      break;

    case T_CLOSURE_STAR: 
      port_write_string(port)(sc, "#<lambda* ", 10, port);  
      break;

    case T_MACRO:        
      if (is_expansion(closure)) 
	port_write_string(port)(sc, "#<expansion ", 12, port); 
      else port_write_string(port)(sc, "#<macro ", 8, port); 
      break;

    case T_MACRO_STAR:   
      port_write_string(port)(sc, "#<macro* ", 9, port);  
      break;
						   
    case T_BACRO:        
      port_write_string(port)(sc, "#<bacro ", 8, port);   
      break;

    case T_BACRO_STAR:   
      port_write_string(port)(sc, "#<bacro* ", 9, port); 
      break;
    }

  if (is_null(closure_args(closure)))
    port_write_string(port)(sc, "()>", 3, port);
  else
    {
      s7_pointer args;
      args = closure_args(closure);
      if (is_symbol(args))
	{
	  port_write_string(port)(sc, symbol_name(args), symbol_name_length(args), port);
	  port_write_character(port)(sc, '>', port);    /* (lambda a a) -> #<lambda a> */
	}
      else
	{
	  port_write_character(port)(sc, '(', port);
	  x = car(args);
	  if (is_pair(x)) x = car(x);
	  port_write_string(port)(sc, symbol_name(x), symbol_name_length(x), port);
	  if (!is_null(cdr(args)))
	    {
	      s7_pointer y;
	      port_write_character(port)(sc, ' ', port);
	      if (is_pair(cdr(args)))
		{
		  y = cadr(args);
		  if (is_pair(y)) 
		    y = car(y);
		  else
		    {
		      if (y == sc->key_rest_symbol)
			{
			  port_write_string(port)(sc, ":rest ", 6, port);
			  args = cdr(args);
			  y = cadr(args);
			  if (is_pair(y)) y = car(y);
			}
		    }
		}
	      else 
		{
		  port_write_string(port)(sc, ". ", 2, port);
		  y = cdr(args);
		}
	      port_write_string(port)(sc, symbol_name(y), symbol_name_length(y), port);
	      if ((is_pair(cdr(args))) &&
		  (!is_null(cddr(args))))
		port_write_string(port)(sc, " ...", 4, port);
	    }
	  port_write_string(port)(sc, ")>", 2, port);
	}
    }
}

static s7_pointer closure_name(s7_scheme *sc, s7_pointer closure)
{
  /* this is used by the error handlers to get the current function name */
  s7_pointer x;

  x = find_closure(sc, closure, sc->envir);
  if (is_symbol(x))
    return(x);

  if (is_pair(current_code(sc)))
    return(current_code(sc));

  return(closure); /* desperation -- the parameter list (caar here) will cause endless confusion in OP_APPLY errors! */
}


static void write_closure_readably_1(s7_scheme *sc, s7_pointer obj, s7_pointer arglist, s7_pointer body, s7_pointer port)
{
  s7_int old_print_length;
  s7_pointer p;

  if (type(obj) == T_CLOSURE_STAR)  
    port_write_string(port)(sc, "(lambda* ", 9, port);
  else port_write_string(port)(sc, "(lambda ", 8, port);

  if ((is_pair(arglist)) &&
      (allows_other_keys(arglist)))
    {
      sc->temp9 = s7_append(sc, arglist, cons(sc, sc->key_allow_other_keys_symbol, sc->nil));
      object_out(sc, sc->temp9, port, USE_WRITE);
      sc->temp9 = sc->nil;
    }
  else object_out(sc, arglist, port, USE_WRITE); /* here we just want the straight output (a b) not (list 'a 'b) */

  old_print_length = sc->print_length;
  sc->print_length = 1048576;
  for (p = body; is_pair(p); p = cdr(p))
    {
      port_write_character(port)(sc, ' ', port);
      object_out(sc, car(p), port, USE_WRITE);
    }
  port_write_character(port)(sc, ')', port);
  sc->print_length = old_print_length;
}

static void write_closure_readably(s7_scheme *sc, s7_pointer obj, s7_pointer port)
{
  s7_pointer body, arglist, pe, local_slots, setter = NULL;
  uint32_t gc_loc;
  
  body = closure_body(obj);
  arglist = closure_args(obj);
  pe = closure_let(obj);               /* perhaps check for documentation? */

  gc_loc = s7_gc_protect(sc, sc->nil);
  collect_locals(sc, body, pe, arglist, gc_loc);   /* collect locals used only here */
  if (s7_is_dilambda(obj))
    {
      setter = closure_setter(obj);
      if ((!(has_closure_let(setter))) ||
	  (closure_let(setter) != pe))
	setter = NULL;
    }
  if (setter)
    collect_locals(sc, closure_body(setter), pe, closure_args(setter), gc_loc);
  local_slots = _TLst(gc_protected_at(sc, gc_loc)); /* possibly a list of slots */

  if (!is_null(local_slots))
    {
      s7_pointer x;
      port_write_string(port)(sc, "(let (", 6, port);
      for (x = local_slots; is_pair(x); x = cdr(x))
	{
	  s7_pointer slot;
	  slot = car(x);
	  port_write_character(port)(sc, '(', port);
	  port_write_string(port)(sc, symbol_name(slot_symbol(slot)), symbol_name_length(slot_symbol(slot)), port);
	  port_write_character(port)(sc, ' ', port);
	  object_out(sc, slot_value(slot), port, USE_WRITE);
	  if (is_null(cdr(x)))
	    port_write_character(port)(sc, ')', port);
	  else port_write_string(port)(sc, ") ", 2, port);
	}
      port_write_string(port)(sc, ") ", 2, port);
    }

  if (setter)
    port_write_string(port)(sc, "(dilambda ", 10, port);

  write_closure_readably_1(sc, obj, arglist, body, port);

  if (setter)
    {
      port_write_character(port)(sc, ' ', port);
      write_closure_readably_1(sc, setter, closure_args(setter), closure_body(setter), port);
      port_write_character(port)(sc, ')', port);
    }

  if (!is_null(local_slots))
    port_write_character(port)(sc, ')', port);
  s7_gc_unprotect_at(sc, gc_loc);
}


#if TRAP_SEGFAULT
#include <signal.h>
static sigjmp_buf senv; /* global here is not a problem -- it is used only to protect s7_is_valid */
static volatile sig_atomic_t can_jump = 0;
static void segv(int32_t ignored) {if (can_jump) siglongjmp(senv, 1);}
#endif

bool s7_is_valid(s7_scheme *sc, s7_pointer arg)
{
  bool result = false;
  if (!arg) return(false);

#if TRAP_SEGFAULT
  if (sigsetjmp(senv, 1) == 0)
    {
      void (*old_segv)(int32_t sig);
      can_jump = 1;
      old_segv = signal(SIGSEGV, segv);
#endif
      result = ((!is_free(arg)) &&
		(type(arg) < NUM_TYPES) &&
		(arg->hloc >= not_heap) &&
		((arg->hloc < 0) ||
		 ((arg->hloc < sc->heap_size) && (sc->heap[arg->hloc] == arg))));

#if TRAP_SEGFAULT
      signal(SIGSEGV, old_segv);
    }
  else result = false;
  can_jump = 0;
#endif

  return(result);
}

enum {NO_ARTICLE, INDEFINITE_ARTICLE};

static char *describe_type_bits(s7_scheme *sc, s7_pointer obj)
{
  uint64_t full_typ;
  uint8_t typ;
  char *buf;

  buf = (char *)malloc(512 * sizeof(char));
  typ = unchecked_type(obj);
  full_typ = typeflag(obj);

  /* if debugging all of these bits are being watched, so we need to access them directly */
  snprintf(buf, 512, "type: %d (%s), flags: #x%" PRIx64 "%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s",
	   typ,
	   type_name(sc, obj, NO_ARTICLE),
	   full_typ,
	   /* bit 0 (the first 8 bits are easy...) */
	   ((full_typ & T_KEYWORD) != 0) ?        " keyword" : "",
	   /* bit 1 */
	   ((full_typ & T_SYNTACTIC) != 0) ?      " syntactic" : "",
	   /* bit 2 */
	   ((full_typ & T_SIMPLE_ARG_DEFAULTS) != 0) ? ((is_pair(obj)) ? " simple-args|in-use" : " ?2?") : "",

	   /* bit 3 */
	   ((full_typ & T_OPTIMIZED) != 0) ?      ((is_c_function(obj)) ? " scope-safe" : 
						   ((is_pair(obj)) ? " optimized" :
						    " ?3?")) : "",
	   /* bit 4 */
	   ((full_typ & T_SAFE_CLOSURE) != 0) ?   " safe-closure" : "",
	   /* bit 5 */
	   ((full_typ & T_DONT_EVAL_ARGS) != 0) ? " dont-eval-args" : "",
	   /* bit 6 */
	   ((full_typ & T_EXPANSION) != 0) ?      (((is_symbol(obj)) || (is_macro(obj))) ? " expansion" : 
						   " ?6?") : "",
	   /* bit 7 */
	   ((full_typ & T_MULTIPLE_VALUE) != 0) ? ((is_symbol(obj)) ? " matched" :
						   ((is_pair(obj)) ? " values|matched" : 
						    " ?7?")) : "",
	   /* bit 8 */
	   ((full_typ & T_GLOBAL) != 0) ?         ((is_pair(obj)) ? " unsafe-do" : 
						   ((is_symbol(obj)) ? " global" :
						    " ?8?")) : "",
	   /* bit 9 */
	   ((full_typ & T_COLLECTED) != 0) ?      " collected" : "",
	   /* bit 10 */
	   ((full_typ & T_LINE_NUMBER) != 0) ?    ((is_pair(obj)) ? " line-number" : 
						   ((is_input_port(obj)) ? " loader-port" : 
						    ((is_let(obj)) ? " with-let" : 
						     ((is_c_function(obj)) ? " simple-defaults" : 
						      (((is_symbol(obj)) || (is_slot(obj))) ? " has-accessor" :
						       " ?10?"))))) : "",
	   /* bit 11 */
	   ((full_typ & T_SHARED) != 0) ?         " shared" : "",
	   /* bit 12 */
	   ((full_typ & T_OVERLAY) != 0) ?        ((is_symbol(obj)) ? " local" : 
						   ((is_pair(obj)) ? " overlay" :
						    " ?12?")) : "",
	   /* bit 13 */
	   ((full_typ & T_SAFE_PROCEDURE) != 0) ? " safe-procedure" : "",
	   /* bit 14 */
	   ((full_typ & T_CHECKED) != 0) ?        " checked" : "",
	   /* bit 15 */
	   ((full_typ & T_UNSAFE) != 0) ?         ((is_symbol(obj)) ? " clean-symbol" : 
						   ((is_slot(obj)) ? " has-stepper" : 
						    ((is_pair(obj)) ? " unsafe|no-float-opt" :
						     " ?15?"))) : "",
	   /* bit 16 */
	   ((full_typ & T_IMMUTABLE) != 0) ?      " immutable" : "",
	   /* bit 17 */
	   ((full_typ & T_SETTER) != 0) ?         ((is_symbol(obj)) ? " setter" :
						   ((is_pair(obj)) ? " allow-other-keys|has-all-x|no-int-opt" :
						    ((is_closure(obj)) ? " has-optlist" :
						     (((is_hash_table(obj)) || (is_let(obj))) ? " removed" :
						      " ?17?")))) : "",
	   /* bit 18 */
	   ((full_typ & T_MUTABLE) != 0) ?        ((is_number(obj)) ? " mutable" :
						   ((is_string(obj)) ? " byte-vector" : 
						    ((is_symbol(obj)) ? " has-keyword" :
						     ((is_let(obj)) ? " let-ref-fallback" : 
						      ((is_iterator(obj)) ? " mark-sequence" : 
						       ((is_slot(obj)) ? " step-end" : 
							((is_let(obj)) ? " ref-fallback" :
							 (((is_pair(obj)) || (is_any_closure(obj))) ? " no-opt" :
							  " ?18?")))))))) : "",
	   /* bit 19 */
	   ((full_typ & T_SAFE_STEPPER) != 0) ?   ((is_let(obj)) ? " set-fallback" : 
						   ((is_slot(obj)) ? " safe-stepper" : 
						    ((is_c_function(obj)) ? " maybe-safe" :
						     ((is_number(obj)) ? " print-name" :
						      " ?19?")))) : "",
	   /* bit 20 */
	   ((full_typ & T_COPY_ARGS) != 0) ?      ((is_pair(obj))? " local-symbol" : 
						   (((is_any_macro(obj)) || (is_any_closure(obj))) ? " copy-args" :
						    "?20?")) : "",
	   /* bit 21 */
	   ((full_typ & T_GENSYM) != 0) ?         ((is_let(obj)) ? " funclet" : 
						   ((is_symbol(obj)) ? " gensym" :
						    ((is_string(obj)) ? " documented-symbol" :
						     ((is_hash_table(obj)) ? " hash-chosen" :
						      ((is_pair(obj)) ? " dotted" :
						       " ?21?"))))) : "",
	   /* bit 22 */
	   ((full_typ & T_HAS_METHODS) != 0) ?    " has-methods" : "",
	   /* bit 23 */
	   ((full_typ & T_ITER_OK) != 0) ?        " iter-ok" : "",
	   /* bit 55 */
	   (((full_typ & T_GC_MARK) != 0) && (in_heap(obj))) ? " gc-marked" : "");
  return(buf);
}


void s7_show_let(s7_scheme *sc) /* debugging convenience */
{
  s7_pointer olet; 
  for (olet = sc->envir; (is_let(olet)) && (olet != sc->rootlet); olet = outlet(olet))
    {
      if (olet == sc->owlet)
	fprintf(stderr, "(owlet): ");
      else
	{
	  if (is_funclet(olet))
	    fprintf(stderr, "(%s funclet): ", DISPLAY(funclet_function(olet)));
	  else
	    {
	      if (olet == sc->shadow_rootlet)
		fprintf(stderr, "(shadow rootlet): ");
	    }
	}
      fprintf(stderr, "%s\n", DISPLAY(olet));
    }
}

#if DEBUGGING
static const char *check_name(int32_t typ)
{
  if ((typ >= 0) && (typ < NUM_TYPES))
    {
      s7_pointer p;
      p = prepackaged_type_names[typ];
      if (is_string(p)) return(string_value(p));
      
      switch (typ)
	{
	case T_C_OBJECT:    return("a c-object");
	case T_INPUT_PORT:  return("an input port");
	case T_OUTPUT_PORT: return("an output port");
	}
    }
  return("unknown type!");
}

static s7_pointer check_seti(s7_scheme *sc, s7_pointer x, const char *func, int32_t line)
{
  if (is_immutable(x)) /* can be vector|pair|string -- incomplete constant arg check I think, TODO: handle immutable vectors */
    {
      fprintf(stderr, "%s%s[%d]: set! immutable %s: %s%s\n", BOLD_TEXT, func, line, type_name(sc, x, NO_ARTICLE), DISPLAY(x), UNBOLD_TEXT);
      if (stop_at_error) abort();
    }
  return(x);
}

static char *safe_object_to_string(s7_pointer p)
{
  uint8_t typ;
  char *buf;
  typ = unchecked_type(p);
  if ((typ > T_FREE) && (typ < NUM_TYPES))
    return(s7_object_to_c_string(cur_sc, p));
  buf = (char *)calloc(128, sizeof(char));
  snprintf(buf, 128, "type=%d", typ);
  return(buf);
}

static s7_pointer check_ref(s7_pointer p, uint8_t expected_type, const char *func, int32_t line, const char *func1, const char *func2)
{
  if (!p)
    fprintf(stderr, "%s[%d]: null pointer passed to check_ref\n", func, line);
  else
    {
      uint8_t typ;
      typ = unchecked_type(p);
      if (typ != expected_type)
	{
	  if ((!func1) || (typ != T_FREE))
	    {
	      fprintf(stderr, "%s%s[%d]: not %s, but %s (%s)%s\n", 
		      BOLD_TEXT, 
		      func, line, check_name(expected_type), check_name(typ), safe_object_to_string(p),
		      UNBOLD_TEXT);
	      if (stop_at_error) abort();
	    }
	  else
	    {
	      if ((strcmp(func, func1) != 0) &&
		  ((!func2) || (strcmp(func, func2) != 0)))
		{
		  fprintf(stderr, "%s%s[%d]: free cell, not %s%s\n", BOLD_TEXT, func, line, check_name(expected_type), UNBOLD_TEXT);
		  if (stop_at_error) abort();
		}
	    }
	}
    }
  return(p);
}

static s7_pointer check_ref2(s7_pointer p, uint8_t expected_type, int32_t other_type, const char *func, int32_t line, const char *func1, const char *func2)
{
  if (!p)
    fprintf(stderr, "%s[%d]: null pointer passed to check_ref2\n", func, line);
  else
    {
      uint8_t typ;
      typ = unchecked_type(p);
      if ((typ != expected_type) && (typ != other_type))
	return(check_ref(p, expected_type, func, line, func1, func2));
    }
  return(p);
}

static s7_pointer check_ref3(s7_pointer p, const char *func, int32_t line)
{
  uint8_t typ;
  typ = unchecked_type(p);
  if ((typ != T_INPUT_PORT) && (typ != T_OUTPUT_PORT) && (typ != T_FREE))
    {
      fprintf(stderr, "%s%s[%d]: not a port, but %s (%s)%s\n", 
	      BOLD_TEXT, 
	      func, line, check_name(typ), safe_object_to_string(p),
	      UNBOLD_TEXT);
      if (stop_at_error) abort();
    }
  return(p);
}

static s7_pointer check_ref4(s7_pointer p, const char *func, int32_t line)
{
  uint8_t typ;
  typ = unchecked_type(p);
  if ((typ != T_VECTOR) && (typ != T_FLOAT_VECTOR) && (typ != T_INT_VECTOR) && (typ != T_FREE))
    {
      fprintf(stderr, "%s%s[%d]: not a vector, but %s (%s)%s\n", 
	      BOLD_TEXT, 
	      func, line, check_name(typ), safe_object_to_string(p),
	      UNBOLD_TEXT);
      if (stop_at_error) abort();
    }
  return(p);
}

static s7_pointer check_ref5(s7_pointer p, const char *func, int32_t line)
{
  uint8_t typ;
  typ = unchecked_type(p);
  if (!t_has_closure_let[typ])
    {
      fprintf(stderr, "%s%s[%d]: not a closure, but %s (%s)%s\n", 
	      BOLD_TEXT, 
	      func, line, check_name(typ), safe_object_to_string(p),
	      UNBOLD_TEXT);
      if (stop_at_error) abort();
    }
  return(p);
}

static s7_pointer check_ref6(s7_pointer p, const char *func, int32_t line)
{
  uint8_t typ;
  typ = unchecked_type(p);
  if ((typ < T_C_FUNCTION_STAR) && (typ != T_C_MACRO))
    {
      fprintf(stderr, "%s%s[%d]: not a c function, but %s (%s)%s\n", 
	      BOLD_TEXT, 
	      func, line, check_name(typ), safe_object_to_string(p),
	      UNBOLD_TEXT);
      if (stop_at_error) abort();
    }
  return(p);
}

static s7_pointer check_ref7(s7_pointer p, const char *func, int32_t line)
{
  if ((!func) || (strcmp(func, "decribe_type_bits") != 0))
    {
      uint8_t typ;
      typ = unchecked_type(p);
      if ((typ < T_INTEGER) || (typ > T_COMPLEX))
	{
	  fprintf(stderr, "%s%s[%d]: not a number, but %s (%s)%s\n", 
		  BOLD_TEXT, 
		  func, line, check_name(typ), safe_object_to_string(p),
		  UNBOLD_TEXT);
	  if (stop_at_error) abort();
	}
    }
  return(p);
}

static s7_pointer check_ref8(s7_pointer p, const char *func, int32_t line)
{
  uint8_t typ;
  typ = unchecked_type(p);
  if ((!t_sequence_p[typ]) && (!t_structure_p[typ]) && (!is_any_closure(p))) /* closure calling itself an iterator?? */
    {
      fprintf(stderr, "%s%s[%d]: not a sequence or structure, but %s (%s)%s\n", 
	      BOLD_TEXT, 
	      func, line, check_name(typ), safe_object_to_string(p),
	      UNBOLD_TEXT);
      if (stop_at_error) abort();
    }
  return(p);
}

static s7_pointer check_ref9(s7_pointer p, const char *func, int32_t line)
{
  uint8_t typ;
  typ = unchecked_type(p);
  if ((typ != T_LET) && (typ != T_C_OBJECT) && (!is_any_closure(p)) && (!is_any_macro(p)) && (typ != T_C_POINTER))
    {
      fprintf(stderr, "%s%s[%d]: not a possible method holder, but %s (%s)%s\n", 
	      BOLD_TEXT, 
	      func, line, check_name(typ), safe_object_to_string(p),
	      UNBOLD_TEXT);
      if (stop_at_error) abort();
    }
  return(p);
}

static s7_pointer check_ref10(s7_pointer p, const char *func, int32_t line)
{
  uint8_t typ;
  typ = unchecked_type(p);
  if ((typ != T_PAIR) && (typ != T_NIL) && (typ != T_SYMBOL))
    {
      fprintf(stderr, "%s%s[%d]: arglist is %s (%s)%s?\n", 
	      BOLD_TEXT, 
	      func, line, check_name(typ), safe_object_to_string(p),
	      UNBOLD_TEXT);
      if (stop_at_error) abort();
    }
  return(p);
}

static s7_pointer check_ref11(s7_pointer p, const char *func, int32_t line)
{
  uint8_t typ;
  typ = unchecked_type(p);
  if ((typ < T_CLOSURE) && (typ != T_BOOLEAN) && (typ != T_PAIR)) /* actually #t is an error here */
    {
      fprintf(stderr, "%s%s[%d]: setter is %s (%s)%s?\n", 
	      BOLD_TEXT, 
	      func, line, check_name(typ), safe_object_to_string(p),
	      UNBOLD_TEXT);
      if (stop_at_error) abort();
    }
  return(p);
}

static s7_pointer check_sym(s7_scheme *sc, s7_pointer sym)
{
  if (!is_slot(local_slot(sym)))
    fprintf(stderr, "%s local_slot: %s\n", DISPLAY(sym), DISPLAY(local_slot(sym)));
  else
    {
      s7_pointer local_val, search_val;
      local_val = slot_value(local_slot(sym));
      search_val = find_symbol_checked(sc, sym);
      if (local_val != search_val)
	{
	  fprintf(stderr, "local %s: %p %p ", symbol_name(sym), local_val, search_val);
	  fprintf(stderr, "%s ", DISPLAY_80(local_val));
	  fprintf(stderr, "%s", DISPLAY_80(search_val));
	  fprintf(stderr, "\n");
	}
      return(local_val);
    }
  return(sc->nil);
}

static s7_pointer check_cell(s7_pointer p, const char *func, int32_t line)
{
  uint8_t typ;
  if (!p)
    {
      fprintf(stderr, "%s%s[%d]: null pointer!%s\n", BOLD_TEXT, func, line, UNBOLD_TEXT);
      if (stop_at_error) abort();
    }    
  typ = unchecked_type(p);
  if ((typ < 0) || (typ >= NUM_TYPES))
    {
      fprintf(stderr, "%s%s[%d]: attempt to use messed up cell (type: %d)%s\n", BOLD_TEXT, func, line, typ, UNBOLD_TEXT);
      if (stop_at_error) abort();
    }
  return(p);
}

static s7_pointer check_nref(s7_pointer p, const char *func, int32_t line)
{
  uint8_t typ;
  check_cell(p, func, line);
  typ = unchecked_type(p);
  if (typ == T_FREE)
    {
      fprintf(stderr, "%s%s[%d]: attempt to use free cell%s\n", BOLD_TEXT, func, line, UNBOLD_TEXT);
      if (stop_at_error) abort();
    }
  return(p);
}

static void print_gc_info(s7_pointer obj, int32_t line)
{
  if (!obj)
    fprintf(stderr, "[%d]: obj is %p\n", line, obj);
  else
    {
      if (unchecked_type(obj) != T_FREE)
	fprintf(stderr, "[%d]: %p type is %d?\n", line, obj, type(obj));
      else
	{
	  fprintf(stderr, "%s%p is free (line %d), current: %s[%d], previous: %s[%d],  gc call: %s[%d], alloc: %s[%d]%s\n",
		  BOLD_TEXT, 
		  obj, line,
		  obj->current_alloc_func, obj->current_alloc_line,
		  obj->previous_alloc_func, obj->previous_alloc_line,
		  obj->gc_func, obj->gc_line, obj->alloc_func, obj->alloc_line,
		  UNBOLD_TEXT);
	}
    }
  abort();
}

static const char *opt1_role_name(int32_t role)
{
  if (role == E_FAST) return("opt_fast");
  if (role == E_BACK) return("opt_back");
  if (role == E_CFUNC) return("opt_cfunc");
  if (role == E_LAMBDA) return("opt_lambda");
  if (role == E_CLAUSE) return("opt_clause");
  if (role == E_GOTO) return("opt_goto");
  if (role == E_SYM) return("opt_sym1");
  if (role == E_PAIR) return("opt_pair1");
  if (role == E_CON) return("opt_con1");
  if (role == E_ANY) return("opt_any1");
  if (role == E_SLOT) return("opt_slot1");
  return("unknown");
}

static const char *opt2_role_name(int32_t role)
{
  if (role == F_CALL) return("c_call(ee)");
  if (role == F_KEY) return("opt_any2");
  if (role == F_SLOW) return("opt_slow");
  if (role == F_SYM) return("opt_sym2");
  if (role == F_PAIR) return("opt_pair2");
  if (role == F_CON) return("opt_con2");
  if (role == F_LAMBDA) return("opt_lambda2");
  return("unknown");
}

static const char *opt3_role_name(int32_t role)
{
  if (role == G_ARGLEN) return("arglist_length");
  if (role == G_SYM) return("opt_sym3");
  if (role == G_AND) return("opt_pair3 or opt_any3");
  if (role == G_DIRECT) return("direct_opt3");
  if (role == S_OP) return("s_op");
  if (role == S_SYNOP) return("s_synop");
  if (role == S_LEN) return("s_len");
  if (role == S_LINE) return("s_line");
  if (role == S_HASH) return("s_hash");
  return("unknown");
}

static char* show_debugger_bits(uint32_t bits)
{
  char *bits_str;
  bits_str = (char *)calloc(512, sizeof(char));
  snprintf(bits_str, 512, " %s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s",
	  ((bits & E_SET) != 0) ? " e-set" : "",
	  ((bits & E_FAST) != 0) ? " opt_fast" : "",
	  ((bits & E_CFUNC) != 0) ? " opt_cfunc" : "",
	  ((bits & E_CLAUSE) != 0) ? " opt_clause" : "",
	  ((bits & E_BACK) != 0) ? " opt_back" : "",
	  ((bits & E_LAMBDA) != 0) ? " opt_lambda" : "",
	  ((bits & E_SYM) != 0) ? " opt_sym1" : "",
	  ((bits & E_PAIR) != 0) ? " opt_pair1" : "",
	  ((bits & E_CON) != 0) ? " opt_con1" : "",
	  ((bits & E_GOTO) != 0) ? " opt_goto" : "",
	  ((bits & E_ANY) != 0) ? " opt_any1" : "",
	  ((bits & E_SLOT) != 0) ? " opt_slot1" : "",
	  ((bits & F_SET) != 0) ? " f-set" : "",
	  ((bits & F_KEY) != 0) ? " opt_any2" : "",
	  ((bits & F_SLOW) != 0) ? " opt_slow" : "",
	  ((bits & F_SYM) != 0) ? " opt_sym2" : "",
	  ((bits & F_PAIR) != 0) ? " opt_pair2" : "",
	  ((bits & F_CON) != 0) ? " opt_con2" : "",
	  ((bits & F_CALL) != 0) ? " c_call(ee)" : "",
	  ((bits & F_LAMBDA) != 0) ? " opt_lambda2" : "",
	  ((bits & G_SET) != 0) ? " g-set" : "",
	  ((bits & G_ARGLEN) != 0) ? " arglist_length" : "",
	  ((bits & G_SYM) != 0) ? " opt_sym3" : "",
	  ((bits & G_AND) != 0) ? " opt_pair3 or opt_any3 " : "",
	  ((bits & G_DIRECT) != 0) ? " opt_direct_x3" : "",
	  ((bits & S_NAME) != 0) ? " raw-name" : "",
	  ((bits & S_HASH) != 0) ? " raw-hash" : "",	
	  ((bits & S_LINE) != 0) ? " line" : "",
	  ((bits & S_LEN) != 0) ? " len" : "",
	  ((bits & S_OP) != 0) ? " op" : "",
	  ((bits & S_SYNOP) != 0) ? " syn-op" : "");
  return(bits_str);
}

static void show_opt1_bits(s7_scheme *sc, s7_pointer p, const char *func, int32_t line, uint32_t role)
{
  char *bits;
  bits = show_debugger_bits(p->debugger_bits);
  fprintf(stderr, "%s%s[%d]: opt1: %p->%p wants %s, debugger bits are %" PRIx64 "%s but expects %x%s\n", 
	  BOLD_TEXT, 
	  func, line, p, p->object.cons.opt1, 
	  opt1_role_name(role),
	  p->debugger_bits, bits, role,
	  UNBOLD_TEXT);
  free(bits);
}

static s7_pointer opt1_1(s7_scheme *sc, s7_pointer p, uint32_t role, const char *func, int32_t line)
{
  if ((!opt1_is_set(p)) ||
      ((!opt1_role_matches(p, role)) &&
       (role != E_ANY)))
    {
      show_opt1_bits(sc, p, func, line, role);
      if (stop_at_error) abort();
    }
  return(p->object.cons.opt1);
}

static s7_pointer set_opt1_1(s7_scheme *sc, s7_pointer p, s7_pointer x, uint32_t role, const char *func, int32_t line)
{
  /* fprintf(stderr, "%s[%d]:%d set %p %p\n", func, line, role, p, x); */
  p->object.cons.opt1 = x;
  set_opt1_role(p, role);
  set_opt1_is_set(p);
  return(x);
}

static uint64_t s_hash_1(s7_scheme *sc, s7_pointer p, const char *func, int32_t line)
{
  if ((!opt1_is_set(p)) ||
      (!opt1_role_matches(p, S_HASH)))
    {
      show_opt1_bits(sc, p, func, line, (uint32_t)S_HASH);
      if (stop_at_error) abort();
    }
  return(p->object.sym_cons.hash);
}

static void set_s_hash_1(s7_scheme *sc, s7_pointer p, uint64_t x, const char *func, int32_t line)
{
  p->object.sym_cons.hash = x;
  set_opt1_role(p, S_HASH);
  set_opt1_is_set(p);
}

static void show_opt2_bits(s7_scheme *sc, s7_pointer p, const char *func, int32_t line, uint32_t role)
{
  char *bits;
  bits = show_debugger_bits(p->debugger_bits);
  fprintf(stderr, "%s%s[%d]: opt2: %p->%p wants %s, debugger bits are %" PRIx64 "%s but expects %x%s%s%s%s%s%s%s%s%s%s\n", 
	  BOLD_TEXT, 
	  func, line, p, p->object.cons.opt2, 
	  opt2_role_name(role),
	  p->debugger_bits, bits, role,
	  ((role & F_SET) != 0) ? " f-set" : "",
	  ((role & F_KEY) != 0) ? " key" : "",
	  ((role & F_SLOW) != 0) ? " slow" : "",
	  ((role & F_SYM) != 0) ? " sym" : "",
	  ((role & F_PAIR) != 0) ? " pair" : "",
	  ((role & F_CON) != 0) ? " con" : "",
	  ((role & F_CALL) != 0) ? " call" : "",
	  ((role & F_LAMBDA) != 0) ? " lambda" : "",
	  ((role & S_NAME) != 0) ? " raw-name" : "",
	  UNBOLD_TEXT);
  free(bits);
}

static s7_pointer opt2_1(s7_scheme *sc, s7_pointer p, uint32_t role, const char *func, int32_t line)
{
  if ((!opt2_is_set(p)) ||
      (!opt2_role_matches(p, role)))
    {
      show_opt2_bits(sc, p, func, line, role);
      fprintf(stderr, "p: %s\n", DISPLAY(p));
      if (stop_at_error) abort();
    }
  return(p->object.cons.opt2);
}

static void set_opt2_1(s7_scheme *sc, s7_pointer p, s7_pointer x, uint32_t role, const char *func, int32_t line)
{
  if ((role == F_CALL) &&
      (x == NULL))  /* this happens apparently innocuously in check_and|or */
    {
      if ((safe_strcmp(func, "check_and") != 0) &&
	  (safe_strcmp(func, "check_or") != 0))
	fprintf(stderr, "%s[%d]: set c_call for %s to null\n", func, line, DISPLAY_80(p));
    }
  p->object.cons.opt2 = x;
  set_opt2_role(p, role);
  set_opt2_is_set(p);
}

static const char *s_name_1(s7_scheme *sc, s7_pointer p, const char *func, int32_t line)
{
  if ((!opt2_is_set(p)) ||
      (!opt2_role_matches(p, S_NAME)))
    {
      show_opt2_bits(sc, p, func, line, (uint32_t)S_NAME);
      if (stop_at_error) abort();
    }
  return(p->object.sym_cons.fstr);
}

static void set_s_name_1(s7_scheme *sc, s7_pointer p, const char *str, const char *func, int32_t line)
{
  p->object.sym_cons.fstr = str;
  set_opt2_role(p, S_NAME);
  set_opt2_is_set(p);
}

static void show_opt3_bits(s7_scheme *sc, s7_pointer p, const char *func, int32_t line, int32_t role)
{
  char *bits;
  bits = show_debugger_bits(p->debugger_bits);
  fprintf(stderr, "%s%s[%d]: opt3: %s %" PRIx64 "%s%s\n", 
	  BOLD_TEXT,
	  func, line, opt3_role_name(role), p->debugger_bits, bits, 
	  UNBOLD_TEXT);
  free(bits);
}

static s7_pointer opt3_1(s7_scheme *sc, s7_pointer p, uint32_t role, const char *func, int32_t line)
{
  if ((!opt3_is_set(p)) || 
      (!opt3_role_matches(p, role)))
    {
      show_opt3_bits(sc, p, func, line, role);
      if (stop_at_error) abort();
    }
  return(p->object.cons.opt3);
}

static void set_opt3_1(s7_scheme *sc, s7_pointer p, s7_pointer x, uint32_t role, const char *func, int32_t line)
{
  typeflag(p) &= ~(T_OPTIMIZED | T_LINE_NUMBER);
  p->object.cons.opt3 = x;
  set_opt3_is_set(p);
  set_opt3_role(p, role);
}

/* S_LINE */
static uint32_t s_line_1(s7_scheme *sc, s7_pointer p, const char *func, int32_t line)
{
  if ((!opt3_is_set(p)) || 
      ((p->debugger_bits & S_LINE) == 0) ||
      (!has_line_number(p)))
    {
      show_opt3_bits(sc, p, func, line, (uint32_t)S_LINE);
      if (stop_at_error) abort();
    }
  return(p->object.sym_cons.line);
}

static void set_s_line_1(s7_scheme *sc, s7_pointer p, uint32_t x, const char *func, int32_t line)
{
  p->object.sym_cons.line = x;
  (p)->debugger_bits = (S_LINE | (p->debugger_bits & ~S_LEN)); /* turn on line, cancel len */
  set_opt3_is_set(p);
}

/* S_LEN (collides with S_LINE) */
static uint32_t s_len_1(s7_scheme *sc, s7_pointer p, const char *func, int32_t line)
{
  if ((!opt3_is_set(p)) || 
      ((p->debugger_bits & S_LEN) == 0) ||
      (has_line_number(p)))
    {
      show_opt3_bits(sc, p, func, line, (uint32_t)S_LEN);
      if (stop_at_error) abort();
    }
  return(p->object.sym_cons.line);
}

static void set_s_len_1(s7_scheme *sc, s7_pointer p, uint32_t x, const char *func, int32_t line)
{
  typeflag(p) &= ~(T_LINE_NUMBER);
  p->object.sym_cons.line = x;
  (p)->debugger_bits = (S_LEN | (p->debugger_bits & ~(S_LINE)));
  set_opt3_is_set(p);
}

/* S_OP */
static uint32_t s_op_1(s7_scheme *sc, s7_pointer p, const char *func, int32_t line)
{
  if ((!opt3_is_set(p)) || 
      ((p->debugger_bits & (S_OP | S_SYNOP)) == 0))
    {
      show_opt3_bits(sc, p, func, line, (uint32_t)S_SYNOP);
      if (stop_at_error) abort();
    }
  return(p->object.sym_cons.op);
}

static void set_s_op_1(s7_scheme *sc, s7_pointer p, uint32_t x, const char *func, int32_t line)
{
  p->object.sym_cons.op = x;
  (p)->debugger_bits = (S_OP | (p->debugger_bits & ~(S_SYNOP)));
  set_opt3_is_set(p);
}

/* S_SYNOP (collides with S_OP, but the optimize bit needs to stay on) */
static uint32_t s_syn_op_1(s7_scheme *sc, s7_pointer p, const char *func, int32_t line)
{
  if ((!opt3_is_set(p)) || 
      ((p->debugger_bits & (S_SYNOP | S_OP)) == 0))
    {
      show_opt3_bits(sc, p, func, line, (uint32_t)S_OP);
      if (stop_at_error) abort();
    }
  return(p->object.sym_cons.op);
}

static void set_s_syn_op_1(s7_scheme *sc, s7_pointer p, uint32_t x, const char *func, int32_t line)
{
  p->object.sym_cons.op = x;
  (p)->debugger_bits = (S_SYNOP | (p->debugger_bits & ~(S_OP)));
  set_opt3_is_set(p);
}

static void print_debugging_state(s7_scheme *sc, s7_pointer obj, s7_pointer port)
{
  /* show current state, current allocated state, and previous allocated state.
   */
  char *current_bits, *allocated_bits, *previous_bits, *str;
  int64_t save_typeflag;
  int32_t len, nlen;
  const char *excl_name;

  if (is_free(obj))
    excl_name = "free cell!";
  else excl_name = "unknown object!";

  current_bits = describe_type_bits(sc, obj);
  save_typeflag = typeflag(obj);
  typeflag(obj) = obj->current_alloc_type;
  allocated_bits = describe_type_bits(sc, obj);
  typeflag(obj) = obj->previous_alloc_type;
  previous_bits = describe_type_bits(sc, obj);
  typeflag(obj) = save_typeflag;

  len = safe_strlen(excl_name) +
    safe_strlen(current_bits) + safe_strlen(allocated_bits) + safe_strlen(previous_bits) +
    safe_strlen(obj->previous_alloc_func) + safe_strlen(obj->current_alloc_func) + 512;
  tmpbuf_malloc(str, len);

  nlen = snprintf(str, len,
		  "\n<%s %s,\n  current: %s[%d] %s,\n  previous: %s[%d] %s\n  hloc: %" PRId64 " (%d uses), free: %s[%d], alloc: %s[%d]>",
		  excl_name, current_bits,
		  obj->current_alloc_func, obj->current_alloc_line, allocated_bits,
		  obj->previous_alloc_func, obj->previous_alloc_line, previous_bits,
		  heap_location(obj), obj->uses,
		  obj->gc_func, obj->gc_line, obj->alloc_func, obj->alloc_line);

  free(current_bits);
  free(allocated_bits);
  free(previous_bits);
  if (is_null(port))
    fprintf(stderr, "%p: %s\n", obj, str);
  else port_write_string(port)(sc, str, nlen, port);
  tmpbuf_free(str, len);
}

#if DEBUGGING
static s7_pointer g_is_local_symbol(s7_scheme *sc, s7_pointer p) {return(make_boolean(sc, is_local_symbol(car(p))));}
#endif

static s7_pointer check_null_sym(s7_scheme *sc, s7_pointer p, s7_pointer sym, int32_t line, const char *func)
{
  if (!p)
    {
      fprintf(stderr, "%s%s[%d]: %s unbound%s\n", BOLD_TEXT, func, line, symbol_name(sym), UNBOLD_TEXT);
      if (stop_at_error) abort();
    }
  return(p);
}

#if 0
/* these are bits usable on (say) let code without colliding with other pair-wise uses */
/*    we need has_all_x = T_SETTER currently
 *    T_MUTABLE and T_SAFE_STEPPER are let_ref|set fallback bits 
 *    T_IMMUTABLE is hard to predict, 
 *    maybe T_SAFE_STEPPER for unsafe_locals
 */
static void check_pair_bits(s7_scheme *sc, s7_pointer p)
{
  /* if ((typeflag(p) & T_MUTABLE) != 0) fprintf(stderr, "mutable: %s\n", DISPLAY_80(p)); */ /* now no_opt flag */
  if ((typeflag(p) & T_IMMUTABLE) != 0) fprintf(stderr, "immutable: %s\n", DISPLAY_80(p));
  if ((typeflag(p) & T_GENSYM) != 0) fprintf(stderr, "gensym: %s\n", DISPLAY_80(p));
  if ((typeflag(p) & T_SETTER) != 0) fprintf(stderr, "setter: %s\n", DISPLAY_80(p));
  if ((typeflag(p) & T_COPY_ARGS) != 0) fprintf(stderr, "copy_args: %s\n", DISPLAY_80(p));
  if ((typeflag(p) & T_SAFE_STEPPER) != 0) fprintf(stderr, "safe_stepper: %s\n", DISPLAY_80(p));
}
#endif
#endif

static void iterator_to_port(s7_scheme *sc, s7_pointer obj, s7_pointer port, use_write_t use_write, shared_info *ci)
{
  if (use_write == USE_READABLE_WRITE)
    {
      if (iterator_is_at_end(obj))
	port_write_string(port)(sc, "(make-iterator #())", 19, port);
      else
	{
	  s7_pointer seq;
	  seq = iterator_sequence(obj);
	  if ((is_string(seq)) && (!is_byte_vector(seq)))
	    {
	      port_write_string(port)(sc, "(make-iterator \"", 16, port);
	      port_write_string(port)(sc, (char *)(string_value(seq) + iterator_position(obj)), string_length(seq) - iterator_position(obj), port);
	      port_write_string(port)(sc, "\")", 2, port);
	    }
	  else
	    {
	      if (iterator_position(obj) > 0)
		port_write_string(port)(sc, "(let ((iter (make-iterator ", 27, port);
	      else port_write_string(port)(sc, "(make-iterator ", 15, port);
	      object_to_port_with_circle_check(sc, iterator_sequence(obj), port, use_write, ci);
	      if (iterator_position(obj) > 0)
		{
		  int32_t nlen;
		  char *str;
		  str = (char *)malloc(128 * sizeof(char));
		  nlen = snprintf(str, 128, "))) (do ((i 0 (+ i 1))) ((= i %" PRId64 ") iter) (iterate iter)))", iterator_position(obj));
		  port_write_string(port)(sc, str, nlen, port);
		  free(str);
		}
	      else port_write_character(port)(sc, ')', port);
	    }
	}
    }
  else
    {
      const char *str;
      str = type_name(sc, iterator_sequence(obj), NO_ARTICLE);
      port_write_string(port)(sc, "#<iterator: ", 12, port);
      port_write_string(port)(sc, str, safe_strlen(str), port);
      port_write_character(port)(sc, '>', port);
    }
}

static void baffle_to_port(s7_scheme *sc, s7_pointer obj, s7_pointer port, use_write_t use_write, shared_info *ci)
{
  int32_t nlen;
  char buf[64];
  nlen = snprintf(buf, 64, "#<baffle: %d>", baffle_key(obj));
  port_write_string(port)(sc, buf, nlen, port);
}

static void c_pointer_to_port(s7_scheme *sc, s7_pointer obj, s7_pointer port, use_write_t use_write, shared_info *ci)
{
  int32_t nlen;
  char buf[128];

  if (use_write == USE_READABLE_WRITE)
    {
      nlen = snprintf(buf, 128, "(c-pointer %" PRIdPTR, (intptr_t)raw_pointer(obj));
      port_write_string(port)(sc, buf, nlen, port);
      if ((raw_pointer_type(obj) != sc->F) ||
	  (raw_pointer_info(obj) != sc->F))
	{
	  port_write_character(port)(sc, ' ', port);
	  object_to_port_with_circle_check(sc, raw_pointer_type(obj), port, use_write, ci);
	  port_write_character(port)(sc, ' ', port);
	  object_to_port_with_circle_check(sc, raw_pointer_info(obj), port, use_write, ci);
	}
      port_write_character(port)(sc, ')', port);
    }
  else 
    {
      if (is_symbol(raw_pointer_type(obj)))
	nlen = snprintf(buf, 128, "#<%s %p>", symbol_name(raw_pointer_type(obj)), raw_pointer(obj));
      else nlen = snprintf(buf, 128, "#<c_pointer %p>", raw_pointer(obj));
      port_write_string(port)(sc, buf, nlen, port);
    }
}

static void rng_to_port(s7_scheme *sc, s7_pointer obj, s7_pointer port, use_write_t use_write, shared_info *ci)
{
  int32_t nlen;
  char buf[128];
#if WITH_GMP
  if (use_write == USE_READABLE_WRITE)
    nlen = snprintf(buf, 128, "#<unprint-readable object>");
  else nlen = snprintf(buf, 128, "#<rng %p>", obj);
#else
  if (use_write == USE_READABLE_WRITE)
    nlen = snprintf(buf, 128, "(random-state %" PRIu64 " %" PRIu64 ")", random_seed(obj), random_carry(obj));
  else nlen = snprintf(buf, 128, "#<rng %" PRIu64 " %" PRIu64 ">", random_seed(obj), random_carry(obj));
#endif
  port_write_string(port)(sc, buf, nlen, port);
}

static void display_any(s7_scheme *sc, s7_pointer obj, s7_pointer port, use_write_t use_write, shared_info *ci)
{
#if DEBUGGING
  print_debugging_state(sc, obj, port);
#else
  {
    char *str, *tmp;
    int32_t nlen, len;
    tmp = describe_type_bits(sc, obj);
    len = 32 + safe_strlen(tmp);
    tmpbuf_malloc(str, len);
    if (is_free(obj))
      nlen = snprintf(str, len, "<free cell! %s>", tmp);
    else nlen = snprintf(str, len, "<unknown object! %s>", tmp);
    free(tmp);
    port_write_string(port)(sc, str, nlen, port);
    tmpbuf_free(str, len);
  }
#endif
}

static void unique_to_port(s7_scheme *sc, s7_pointer obj, s7_pointer port, use_write_t use_write, shared_info *ci)
{
  port_write_string(port)(sc, unique_name(obj), unique_name_length(obj), port);
}

static void eof_to_port(s7_scheme *sc, s7_pointer obj, s7_pointer port, use_write_t use_write, shared_info *ci)
{
  /* if file has #<eof> it causes read to return #<eof> -> end of read! what is readable version? '#<eof> or (begin #<eof>) as below
   * but this is silly -- to fool read, the #<eof> has to be all by itself at the top-level!
   * and the read of #<eof> does not affect the port, so if you know it's there, just ignore #<eof> and continue reading.
   */
  if (use_write == USE_READABLE_WRITE)
    port_write_string(port)(sc, "(begin #<eof>)", 14, port);
  else port_write_string(port)(sc, unique_name(obj), unique_name_length(obj), port);
}

static void counter_to_port(s7_scheme *sc, s7_pointer obj, s7_pointer port, use_write_t use_write, shared_info *ci)
{
  port_write_string(port)(sc, "#<counter>", 10, port);
}

static void optlist_to_port(s7_scheme *sc, s7_pointer obj, s7_pointer port, use_write_t use_write, shared_info *ci)
{
  port_write_string(port)(sc, "#<optlist>", 10, port);
}

static void integer_to_port(s7_scheme *sc, s7_pointer obj, s7_pointer port, use_write_t use_write, shared_info *ci)
{
  if (has_print_name(obj))
    port_write_string(port)(sc, print_name(obj), print_name_length(obj), port);
  else
    {
      int32_t nlen;
      char *str;
      nlen = 0;
      str = integer_to_string_base_10_no_width(obj, &nlen);
      if (nlen > 0)
	{
	  set_print_name(obj, str, nlen);
	  port_write_string(port)(sc, str, nlen, port);
	}
      else port_display(port)(sc, str, port);
    }
}

static void number_to_port(s7_scheme *sc, s7_pointer obj, s7_pointer port, use_write_t use_write, shared_info *ci)
{
  if (has_print_name(obj))
    port_write_string(port)(sc, print_name(obj), print_name_length(obj), port);
  else
    {
      int32_t nlen;
      char *str;
      nlen = 0;
      str = number_to_string_base_10(obj, 0, float_format_precision, 'g', &nlen, use_write); /* was 14 */
      set_print_name(obj, str, nlen);
      port_write_string(port)(sc, str, nlen, port);
    }
}

#if WITH_GMP
static void big_number_to_port(s7_scheme *sc, s7_pointer obj, s7_pointer port, use_write_t use_write, shared_info *ci)
{
  int32_t nlen;
  char *str;
  nlen = 0;
  str = big_number_to_string_with_radix(obj, BASE_10, 0, &nlen, use_write);
  port_write_string(port)(sc, str, nlen, port);
  free(str);
}
#endif

static void syntax_to_port(s7_scheme *sc, s7_pointer obj, s7_pointer port, use_write_t use_write, shared_info *ci)
{
  port_display(port)(sc, symbol_name(syntax_symbol(obj)), port);
}

static void string_to_port_1(s7_scheme *sc, s7_pointer obj, s7_pointer port, use_write_t use_write, shared_info *ci)
{
  if (is_byte_vector(obj))
    byte_vector_to_port(sc, obj, port, use_write);
  else string_to_port(sc, obj, port, use_write);
}

static void character_to_port(s7_scheme *sc, s7_pointer obj, s7_pointer port, use_write_t use_write, shared_info *ci)
{
  if (use_write == USE_DISPLAY)
    port_write_character(port)(sc, character(obj), port);
  else port_write_string(port)(sc, character_name(obj), character_name_length(obj), port);
}

static void closure_to_port(s7_scheme *sc, s7_pointer obj, s7_pointer port, use_write_t use_write, shared_info *ci)
{
  if (has_methods(obj))
    {
      /* look for object->string method else fallback on ordinary case.
       * can't use recursion on closure_let here because then the fallback name is #<let>.
       */
      s7_pointer print_func;
      print_func = find_method(sc, closure_let(obj), sc->object_to_string_symbol);
      if (print_func != sc->undefined)
	{
	  s7_pointer p;
	  p = s7_apply_function(sc, print_func, list_1(sc, obj));
	  if (string_length(p) > 0)
	    port_write_string(port)(sc, string_value(p), string_length(p), port);
	  return;
	}
    }
  if (use_write == USE_READABLE_WRITE)
    write_closure_readably(sc, obj, port);
  else write_closure_name(sc, obj, port);
}

static void macro_to_port(s7_scheme *sc, s7_pointer obj, s7_pointer port, use_write_t use_write, shared_info *ci)
{
  if (use_write == USE_READABLE_WRITE)
    write_macro_readably(sc, obj, port);
  else write_closure_name(sc, obj, port);
}

static void c_function_to_port(s7_scheme *sc, s7_pointer obj, s7_pointer port, use_write_t use_write, shared_info *ci)
{
  port_write_string(port)(sc, c_function_name(obj), c_function_name_length(obj), port);
}

static void c_macro_to_port(s7_scheme *sc, s7_pointer obj, s7_pointer port, use_write_t use_write, shared_info *ci)
{
  port_write_string(port)(sc, c_macro_name(obj), c_macro_name_length(obj), port);
}

static void continuation_to_port(s7_scheme *sc, s7_pointer obj, s7_pointer port, use_write_t use_write, shared_info *ci)
{
  if (use_write == USE_READABLE_WRITE)
    port_write_string(port)(sc, "continuation", 12, port);
  else port_write_string(port)(sc, "#<continuation>", 15, port);
}

static void goto_to_port(s7_scheme *sc, s7_pointer obj, s7_pointer port, use_write_t use_write, shared_info *ci)
{
  if (use_write == USE_READABLE_WRITE)
    port_write_string(port)(sc, "goto", 4, port);
  else port_write_string(port)(sc, "#<goto>", 7, port);
}

static void catch_to_port(s7_scheme *sc, s7_pointer obj, s7_pointer port, use_write_t use_write, shared_info *ci)
{
  port_write_string(port)(sc, "#<catch>", 8, port);
}

static void dynamic_wind_to_port(s7_scheme *sc, s7_pointer obj, s7_pointer port, use_write_t use_write, shared_info *ci)
{
  /* this can happen because (*s7* 'stack) can involve dynamic-wind markers */
  port_write_string(port)(sc, "#<dynamic-wind>", 15, port);
}

static void c_object_to_port(s7_scheme *sc, s7_pointer obj, s7_pointer port, use_write_t use_write, shared_info *ci)
{
  char *str;
  if (use_write == USE_READABLE_WRITE)
    str = ((*(c_object_print_readably(sc, obj)))(sc, c_object_value(obj)));
  else str = ((*(c_object_print(sc, obj)))(sc, c_object_value(obj)));
  port_display(port)(sc, str, port);
  free(str);
}

static void slot_to_port(s7_scheme *sc, s7_pointer obj, s7_pointer port, use_write_t use_write, shared_info *ci)
{
  if (use_write != USE_READABLE_WRITE)
    port_write_character(port)(sc, '\'', port);
  symbol_to_port(sc, slot_symbol(obj), port, use_write, ci);
  port_write_character(port)(sc, ' ', port);
  object_to_port_with_circle_check(sc, slot_value(obj), port, use_write, ci);
}


static void init_display_functions(void)
{
  int32_t i;
  for (i = 0; i < 256; i++) display_functions[i] = display_any;
  display_functions[T_FLOAT_VECTOR] = int_or_float_vector_to_port;
  display_functions[T_INT_VECTOR] =   int_or_float_vector_to_port;
  display_functions[T_VECTOR] =       vector_to_port;
  display_functions[T_PAIR] =         pair_to_port;
  display_functions[T_HASH_TABLE] =   hash_table_to_port;
  display_functions[T_ITERATOR] =     iterator_to_port;
  display_functions[T_LET] =          let_to_port;
  display_functions[T_BOOLEAN] =      unique_to_port;
  display_functions[T_NIL] =          unique_to_port;
  display_functions[T_UNSPECIFIED] =  unique_to_port;
  display_functions[T_UNDEFINED] =    unique_to_port;
  display_functions[T_EOF_OBJECT] =   eof_to_port;
  display_functions[T_INPUT_PORT] =   input_port_to_port;
  display_functions[T_OUTPUT_PORT] =  output_port_to_port;
  display_functions[T_COUNTER] =      counter_to_port;
  display_functions[T_OPTLIST] =      optlist_to_port;
  display_functions[T_BAFFLE] =       baffle_to_port;
  display_functions[T_INTEGER] =      integer_to_port;
  display_functions[T_RATIO] =        number_to_port;
  display_functions[T_REAL] =         number_to_port;
  display_functions[T_COMPLEX] =      number_to_port;
#if WITH_GMP
  display_functions[T_BIG_INTEGER] =  big_number_to_port;
  display_functions[T_BIG_RATIO] =    big_number_to_port;
  display_functions[T_BIG_REAL] =     big_number_to_port;
  display_functions[T_BIG_COMPLEX] =  big_number_to_port;
#endif
  display_functions[T_SYMBOL] =       symbol_to_port;
  display_functions[T_SYNTAX] =       syntax_to_port;
  display_functions[T_STRING] =       string_to_port_1;
  display_functions[T_CHARACTER] =    character_to_port;
  display_functions[T_CLOSURE] =      closure_to_port;
  display_functions[T_CLOSURE_STAR] = closure_to_port;
  display_functions[T_MACRO] =        macro_to_port;
  display_functions[T_MACRO_STAR] =   macro_to_port;
  display_functions[T_BACRO] =        macro_to_port;
  display_functions[T_BACRO_STAR] =   macro_to_port;
  display_functions[T_C_OPT_ARGS_FUNCTION] = c_function_to_port;
  display_functions[T_C_RST_ARGS_FUNCTION] = c_function_to_port;
  display_functions[T_C_ANY_ARGS_FUNCTION] = c_function_to_port;
  display_functions[T_C_FUNCTION] =   c_function_to_port;
  display_functions[T_C_FUNCTION_STAR] = c_function_to_port;
  display_functions[T_C_MACRO] =      c_macro_to_port;
  display_functions[T_C_POINTER] =    c_pointer_to_port;
  display_functions[T_RANDOM_STATE] = rng_to_port;
  display_functions[T_CONTINUATION] = continuation_to_port;
  display_functions[T_GOTO] =         goto_to_port;
  display_functions[T_CATCH] =        catch_to_port;
  display_functions[T_DYNAMIC_WIND] = dynamic_wind_to_port;
  display_functions[T_C_OBJECT] =     c_object_to_port;
  display_functions[T_SLOT] =         slot_to_port;
}

static void object_to_port_with_circle_check(s7_scheme *sc, s7_pointer vr, s7_pointer port, use_write_t use_write, shared_info *ci)
{
  if ((ci) &&
      (has_structure(vr)))
    {
      int32_t ref;
      ref = shared_ref(ci, vr);
      if (ref != 0)
	{
	  char buf[32];
	  int32_t nlen;
	  char *p;
	  uint32_t len;
	  if (ref > 0)
	    {
	      if (use_write == USE_READABLE_WRITE)
		{
		  nlen = snprintf(buf, 32, "(set! {%d} ", ref);
		  port_write_string(port)(sc, buf, nlen, port);
		  object_to_port(sc, vr, port, USE_READABLE_WRITE, ci);
		  port_write_character(port)(sc, ')', port);
		}
	      else
		{
		  p = pos_int_to_str((s7_int)ref, &len, '=');
		  *--p = '#';
		  port_write_string(port)(sc, p, len, port);
		  object_to_port(sc, vr, port, DONT_USE_DISPLAY(use_write), ci);
		}
	    }
	  else
	    {
	      if (use_write == USE_READABLE_WRITE)
		{
		  nlen = snprintf(buf, 32, "{%d}", -ref);
		  port_write_string(port)(sc, buf, nlen, port);
		}
	      else 
		{
		  p = pos_int_to_str((s7_int)(-ref), &len, '#');
		  *--p = '#';
		  port_write_string(port)(sc, p, len, port);
		}
	    }
	  return;
	}
    }
  object_to_port(sc, vr, port, use_write, ci);
}


static void setup_shared_reads(s7_scheme *sc, s7_pointer port, shared_info *ci)
{
  int32_t i;
  char buf[64];

  port_write_string(port)(sc, "(let (", 6, port);
  for (i = 1; i <= ci->top; i++)
    {
      int32_t len;
      len = snprintf(buf, 64, "({%d} #f)", i);
      port_write_string(port)(sc, buf, len, port);
    }
  port_write_string(port)(sc, ") ", 2, port);
}

static void finish_shared_reads(s7_scheme *sc, s7_pointer port, shared_info *ci)
{
  port_write_character(port)(sc, ')', port);
}

static s7_pointer object_out(s7_scheme *sc, s7_pointer obj, s7_pointer strport, use_write_t choice)
{
  if ((has_structure(obj)) &&
      (obj != sc->rootlet))
    {
      shared_info *ci;
      ci = make_shared_info(sc, obj, choice != USE_READABLE_WRITE);
      if (ci)
	{
	  if (choice == USE_READABLE_WRITE)
	    {
	      setup_shared_reads(sc, strport, ci);
	      object_to_port_with_circle_check(sc, obj, strport, choice, ci);
	      finish_shared_reads(sc, strport, ci);
	    }
	  else object_to_port_with_circle_check(sc, obj, strport, choice, ci);
	  return(obj);
	}
    }
  object_to_port(sc, obj, strport, choice, NULL);
  return(obj);
}
  

static s7_pointer format_ports = NULL;

static s7_pointer open_format_port(s7_scheme *sc)
{
  s7_pointer x;
  int32_t len;

  if (format_ports)
    {
      x = format_ports;
      format_ports = (s7_pointer)(port_port(x)->next);
      port_position(x) = 0;
      port_data(x)[0] = '\0';
      return(x);
    }

  len = FORMAT_PORT_LENGTH;
  x = alloc_pointer();
  set_type(x, T_OUTPUT_PORT);
  port_port(x) = (port_t *)calloc(1, sizeof(port_t));
  port_type(x) = STRING_PORT;
  port_is_closed(x) = false;
  port_data_size(x) = len;
  port_data(x) = (unsigned char *)malloc(len * sizeof(unsigned char)); /* was +8 */
  port_data(x)[0] = '\0';
  port_position(x) = 0;
  port_needs_free(x) = false;
  port_read_character(x) = output_read_char;
  port_read_line(x) = output_read_line;
  port_display(x) = string_display;
  port_write_character(x) = string_write_char;
  port_write_string(x) = string_write_string;
  return(x);
}

static void close_format_port(s7_scheme *sc, s7_pointer port)
{
  port_port(port)->next = (void *)format_ports;
  format_ports = port;
}


static char *s7_object_to_c_string_1(s7_scheme *sc, s7_pointer obj, use_write_t use_write, int32_t *nlen)
{
  char *str;
  s7_pointer strport;

  strport = open_format_port(sc);
  object_out(sc, obj, strport, use_write);
  if (nlen) (*nlen) = port_position(strport);

  str = (char *)malloc((port_position(strport) + 1) * sizeof(char));
  memcpy((void *)str, (void *)port_data(strport), port_position(strport));
  str[port_position(strport)] = '\0';
  close_format_port(sc, strport);

  return(str); 
}


char *s7_object_to_c_string(s7_scheme *sc, s7_pointer obj)
{
  TRACK(sc);
  if ((sc->safety > NO_SAFETY) &&
      (!s7_is_valid(sc, obj)))
    fprintf(stderr, "bad arg to %s: %p\n", __func__, obj);

  return(s7_object_to_c_string_1(sc, obj, USE_WRITE, NULL));
}


s7_pointer s7_object_to_string(s7_scheme *sc, s7_pointer obj, bool use_write) /* unavoidable backwards compatibility rigidity here */
{
  char *str;
  int32_t len = 0;

  if ((sc->safety > NO_SAFETY) &&
      (!s7_is_valid(sc, obj)))
    fprintf(stderr, "bad arg to %s: %p\n", __func__, obj);

  str = s7_object_to_c_string_1(sc, obj, (use_write) ? USE_WRITE : USE_DISPLAY, &len);
  if (str)
    return(make_string_uncopied_with_length(sc, str, len));
  return(s7_make_string_with_length(sc, "", 0));
}


/* -------------------------------- newline -------------------------------- */
void s7_newline(s7_scheme *sc, s7_pointer port)
{
  s7_write_char(sc, '\n', port);
}

static s7_pointer g_newline(s7_scheme *sc, s7_pointer args)
{
  #define H_newline "(newline (port (current-output-port))) writes a carriage return to the port"
  #define Q_newline s7_make_signature(sc, 2, sc->is_unspecified_symbol, sc->is_output_port_symbol)
  s7_pointer port;

  if (is_not_null(args))
    port = car(args);
  else port = sc->output_port;
  if (!is_output_port(port))
    {
      if (port == sc->F) return(sc->unspecified);
      method_or_bust_with_type_one_arg(sc, port, sc->newline_symbol, args, an_output_port_string);
    }
  s7_newline(sc, port);
  return(sc->unspecified);
}

static s7_pointer newline_p(void) {s7_write_char(cur_sc, '\n', cur_sc->output_port); return(cur_sc->unspecified);}
static s7_pointer newline_p_p(s7_pointer port) {s7_write_char(cur_sc, '\n', port); return(cur_sc->unspecified);}


/* -------------------------------- write -------------------------------- */
void s7_write(s7_scheme *sc, s7_pointer obj, s7_pointer port)
{
  if (port != sc->F)
    {
      if (port_is_closed(port))
	s7_wrong_type_arg_error(sc, "write", 2, port, "an open output port");
      object_out(sc, obj, port, USE_WRITE);
    }
}


static s7_pointer g_write(s7_scheme *sc, s7_pointer args)
{
  #define H_write "(write obj (port (current-output-port))) writes (object->string obj) to the output port"
  #define Q_write s7_make_signature(sc, 3, sc->T, sc->T, sc->is_output_port_symbol)
  s7_pointer port;

  if (is_pair(cdr(args)))
    port = cadr(args);
  else port = sc->output_port;
  if (!is_output_port(port))
    {
      if (port == sc->F) return(car(args));
      method_or_bust_with_type(sc, port, sc->write_symbol, args, an_output_port_string, 2);
    }
  if (port_is_closed(port))
    return(s7_wrong_type_arg_error(sc, "write", 2, port, "an open output port"));
  return(object_out(sc, car(args), port, USE_WRITE));
}

static s7_pointer write_p_p(s7_pointer x) 
{
  if (cur_sc->output_port == cur_sc->F) return(x);
  return(object_out(cur_sc, x, cur_sc->output_port, USE_WRITE));
}
static s7_pointer write_p_pp(s7_pointer x, s7_pointer port) 
{
  if (port == cur_sc->F)
    return(x);
  if ((!is_output_port(port)) ||
      (port_is_closed(port)))
    s7_wrong_type_arg_error(cur_sc, "write", 2, port, "an open output port");
  if (port == cur_sc->F) return(x);
  return(object_out(cur_sc, x, port, USE_WRITE));
}


/* -------------------------------- display -------------------------------- */
void s7_display(s7_scheme *sc, s7_pointer obj, s7_pointer port)
{
  if (port != sc->F)
    {
      if (port_is_closed(port))
	s7_wrong_type_arg_error(sc, "display", 2, port, "an open output port");
      object_out(sc, obj, port, USE_DISPLAY);
    }
}

static s7_pointer g_display(s7_scheme *sc, s7_pointer args)
{
  #define H_display "(display obj (port (current-output-port))) prints obj"
  #define Q_display s7_make_signature(sc, 3, sc->T, sc->T, sc->is_output_port_symbol)
  s7_pointer port;

  if (is_pair(cdr(args)))
    port = cadr(args);
  else port = sc->output_port;
  if (!is_output_port(port))
    {
      if (port == sc->F) return(car(args));
      method_or_bust_with_type(sc, port, sc->display_symbol, args, an_output_port_string, 2);
    }
  if (port_is_closed(port))
    return(s7_wrong_type_arg_error(sc, "display", 2, port, "an open output port"));
  return(object_out(sc, car(args), port, USE_DISPLAY));
}

static s7_pointer display_p_p(s7_pointer x) 
{
  if (cur_sc->output_port == cur_sc->F) return(x);
  return(object_out(cur_sc, x, cur_sc->output_port, USE_DISPLAY));
}
static s7_pointer display_p_pp(s7_pointer x, s7_pointer port) 
{
  if (port == cur_sc->F)
    return(x);
  if ((!is_output_port(port)) ||
      (port_is_closed(port)))
    s7_wrong_type_arg_error(cur_sc, "display", 2, port, "an open output port");
  if (port == cur_sc->F) return(x);
  return(object_out(cur_sc, x, port, USE_DISPLAY));
}


/* -------------------------------- call-with-output-string -------------------------------- */
static s7_pointer g_call_with_output_string(s7_scheme *sc, s7_pointer args)
{
  #define H_call_with_output_string "(call-with-output-string proc) opens a string port applies proc to it, then returns the collected output"
  #define Q_call_with_output_string s7_make_signature(sc, 2, sc->is_string_symbol, sc->is_procedure_symbol)
  s7_pointer port, proc;

  proc = car(args);
  if (is_let(proc))
    check_method(sc, proc, sc->call_with_output_string_symbol, args);
  if (!s7_is_aritable(sc, proc, 1))
    method_or_bust_with_type(sc, proc, sc->call_with_output_string_symbol, args, s7_make_string_wrapper(sc, "a procedure of one argument (the port)"), 1);

  if ((is_continuation(proc)) || (is_goto(proc)))
    return(wrong_type_argument_with_type(sc, sc->call_with_output_string_symbol, 1, proc, a_normal_procedure_string));

  port = s7_open_output_string(sc);
  push_stack(sc, OP_GET_OUTPUT_STRING_1, sc->F, port); /* args checked in call_with_exit */
  push_stack(sc, OP_APPLY, list_1(sc, port), proc);
  return(sc->F);
}



/* -------------------------------- call-with-output-file -------------------------------- */
static s7_pointer g_call_with_output_file(s7_scheme *sc, s7_pointer args)
{
  #define H_call_with_output_file "(call-with-output-file filename proc) opens filename and calls proc with the output port as its argument"
  #define Q_call_with_output_file pl_sf
  s7_pointer port, file, proc;

  file = car(args);
  if (!is_string(file))
    method_or_bust(sc, file, sc->call_with_output_file_symbol, args, T_STRING, 1);

  proc = cadr(args);
  if (!s7_is_aritable(sc, proc, 1))
    method_or_bust_with_type(sc, proc, sc->call_with_output_file_symbol, args, s7_make_string_wrapper(sc, "a procedure of one argument (the port)"), 2);

  if ((is_continuation(proc)) || is_goto(proc))
    return(wrong_type_argument_with_type(sc, sc->call_with_output_file_symbol, 2, proc, a_normal_procedure_string));

  port = s7_open_output_file(sc, string_value(file), "w");
  push_stack(sc, OP_UNWIND_OUTPUT, sc->F, port);
  push_stack(sc, OP_APPLY, list_1(sc, port), proc);
  return(sc->F);
}



/* -------------------------------- with-output-to-string -------------------------------- */
static s7_pointer g_with_output_to_string(s7_scheme *sc, s7_pointer args)
{
  #define H_with_output_to_string "(with-output-to-string thunk) opens a string as a temporary current-output-port, calls thunk, then returns the collected output"
  #define Q_with_output_to_string s7_make_signature(sc, 2, sc->is_string_symbol, sc->is_procedure_symbol)
  s7_pointer old_output_port, p;

  p = car(args);
  if (!is_thunk(sc, p))
    method_or_bust_with_type(sc, p, sc->with_output_to_string_symbol, args, a_thunk_string, 1);

  old_output_port = sc->output_port;
  sc->output_port = s7_open_output_string(sc);
  push_stack(sc, OP_GET_OUTPUT_STRING_1, old_output_port, sc->output_port);
  push_stack(sc, OP_APPLY, sc->nil, p);
  return(sc->F);
}

/* (let () (define-macro (mac) (write "123")) (with-output-to-string mac))
 * (string-ref (with-output-to-string (lambda () (write "1234") (values (get-output-string) 1))))
 */


/* -------------------------------- with-output-to-file -------------------------------- */
static s7_pointer g_with_output_to_file(s7_scheme *sc, s7_pointer args)
{
  #define H_with_output_to_file "(with-output-to-file filename thunk) opens filename as the temporary current-output-port and calls thunk"
  #define Q_with_output_to_file pl_sf
  s7_pointer old_output_port, file, proc;

  file = car(args);
  if (!is_string(file))
    method_or_bust(sc, file, sc->with_output_to_file_symbol, args, T_STRING, 1);

  proc = cadr(args);
  if (!is_thunk(sc, proc))
    method_or_bust_with_type(sc, proc, sc->with_output_to_file_symbol, args, a_thunk_string, 2);

  old_output_port = sc->output_port;
  sc->output_port = s7_open_output_file(sc, string_value(file), "w");
  push_stack(sc, OP_UNWIND_OUTPUT, old_output_port, sc->output_port);
  push_stack(sc, OP_APPLY, sc->nil, proc);
  return(sc->F);
}



/* -------------------------------- format -------------------------------- */

static s7_pointer format_error_1(s7_scheme *sc, s7_pointer msg, const char *str, s7_pointer args, format_data *fdat)
{
  s7_pointer x = NULL, ctrl_str;
  static s7_pointer format_string_1 = NULL, format_string_2, format_string_3, format_string_4;

  if (!format_string_1)
    {
      format_string_1 = s7_make_permanent_string("format: ~S ~{~S~^ ~}: ~A");
      format_string_2 = s7_make_permanent_string("format: ~S: ~A");
      format_string_3 = s7_make_permanent_string("format: ~S ~{~S~^ ~}~&~NT^: ~A");
      format_string_4 = s7_make_permanent_string("format: ~S~&~NT^: ~A");
    }

  if (fdat->orig_str)
    ctrl_str = fdat->orig_str;
  else ctrl_str = s7_make_string_wrapper(sc, str);

  if (fdat->loc == 0)
    {
      if (is_pair(args))
	x = set_elist_4(sc, format_string_1, ctrl_str, args, msg);
      else x = set_elist_3(sc, format_string_2, ctrl_str, msg);
    }
  else
    {
      if (is_pair(args))
	x = set_elist_5(sc, format_string_3, ctrl_str, args, make_integer(sc, fdat->loc + 20), msg);
      else x = set_elist_4(sc, format_string_4, ctrl_str, make_integer(sc, fdat->loc + 20), msg);
    }
  if (fdat->port)
    {
      close_format_port(sc, fdat->port);
      fdat->port = NULL;
    }
  return(s7_error(sc, sc->format_error_symbol, x));
}

#define format_error(Sc, Msg, Str, Args, Fdat) \
  do {static s7_pointer _Err_ = NULL; if (!_Err_) _Err_ = s7_make_permanent_string(Msg);  return(format_error_1(Sc, _Err_, Str, Args, Fdat));} while (0)

#define just_format_error(Sc, Msg, Str, Args, Fdat) \
  do {static s7_pointer _Err_ = NULL; if (!_Err_) _Err_ = s7_make_permanent_string(Msg);  format_error_1(Sc, _Err_, Str, Args, Fdat);} while (0)

static void format_append_char(s7_scheme *sc, format_data *fdat, char c, s7_pointer port)
{
  port_write_character(port)(sc, c, port);
  sc->format_column++;

  /* if c is #\null, is this the right thing to do?
   * We used to return "1 2 3 4" because ~C was first turned into a string (empty in this case)
   *   (format #f "1 2~C3 4" #\null)
   *   "1 2"
   * Clisp does this:
   *   (format nil "1 2~C3 4" (int-char 0))
   *   "1 23 4"
   * whereas sbcl says int-char is undefined, and
   * Guile returns "1 2\x003 4"
   */
}

static void format_append_newline(s7_scheme *sc, format_data *fdat, s7_pointer port)
{
  port_write_character(port)(sc, '\n', port);
  sc->format_column = 0;
}


static void format_append_string(s7_scheme *sc, format_data *fdat, const char *str, int32_t len, s7_pointer port)
{
  port_write_string(port)(sc, str, len, port);
  fdat->loc += len;
  sc->format_column += len;
}

static void format_append_chars(s7_scheme *sc, format_data *fdat, char pad, int32_t chars, s7_pointer port)
{
  if (chars > 0)
    {
      if (chars < TMPBUF_SIZE)
	{
	  int32_t j;
	  for (j = 0; j < chars; j++)
	    sc->tmpbuf[j] = pad;
	  sc->tmpbuf[chars] = '\0';
	  format_append_string(sc, fdat, sc->tmpbuf, chars, port);
	}
      else
	{
	  int32_t j;
	  for (j = 0; j < chars; j++)
	    format_append_char(sc, fdat, pad, port);
	}
    }
}


static int32_t format_read_integer(s7_scheme *sc, int32_t *cur_i, int32_t str_len, const char *str, s7_pointer args, format_data *fdat)
{
  /* we know that str[*cur_i] is a digit */
  int32_t i, lval = 0;
  for (i = *cur_i; i < str_len - 1; i++)
    {
      int32_t dig;
      dig = digits[(unsigned char)str[i]];
      if (dig < 10)
	{
#if HAVE_OVERFLOW_CHECKS
	  if ((int_multiply_overflow(lval, 10, &lval)) ||
	      (int_add_overflow(lval, dig, &lval)))
	    break;
#else        
	  lval = dig + (lval * 10);
#endif
	}
      else break;
    }

  if (i >= str_len)
    just_format_error(sc, "numeric argument, but no directive!", str, args, fdat);
  *cur_i = i;
  return(lval);
}


static void format_number(s7_scheme *sc, format_data *fdat, int32_t radix, int32_t width, int32_t precision, char float_choice, char pad, s7_pointer port)
{
  char *tmp;
  int32_t nlen = 0;
  if (width < 0) width = 0;

  /* precision choice depends on float_choice if it's -1 */
  if (precision < 0)
    {
      if ((float_choice == 'e') ||
	  (float_choice == 'f') ||
	  (float_choice == 'g'))
	precision = 6;
      else
	{
	  /* in the "int" cases, precision depends on the arg type */
	  switch (type(car(fdat->args)))
	    {
	    case T_INTEGER:
	    case T_RATIO:
	      precision = 0;
	      break;

	    default:
	      precision = 6;
	      break;
	    }
	}
    }
  /* should (format #f "~F" 1/3) return "1/3"?? in CL it's "0.33333334" */

  tmp = number_to_string_with_radix(sc, car(fdat->args), radix, width, precision, float_choice, &nlen);
  if (pad != ' ')
    {
      char *padtmp;
      padtmp = tmp;
      while (*padtmp == ' ') (*(padtmp++)) = pad;
    }
  format_append_string(sc, fdat, tmp, nlen, port);

  free(tmp);
  fdat->args = cdr(fdat->args);
  fdat->ctr++;
}


static int32_t format_nesting(const char *str, char opener, char closer, int32_t start, int32_t end) /* start=i, end=str_len-1 */
{
  int32_t k, nesting = 1;
  for (k = start + 2; k < end; k++)
    if (str[k] == '~')
      {
	if (str[k + 1] == closer)
	  {
	    nesting--;
	    if (nesting == 0)
	      return(k - start - 1);
	  }
	else
	  {
	    if (str[k + 1] == opener)
	      nesting++;
	  }
      }
  return(-1);
}

static bool format_method(s7_scheme *sc, const char *str, format_data *fdat, s7_pointer port)
{
  s7_pointer obj, func;

  obj = car(fdat->args);
  if ((has_methods(obj)) &&
      ((func = find_method(sc, find_let(sc, obj), sc->format_symbol)) != sc->undefined))
    {
      s7_pointer ctrl_str;
      if (fdat->orig_str)
	ctrl_str = fdat->orig_str;
      else ctrl_str = s7_make_string_wrapper(sc, str);

      obj = s7_apply_function(sc, func, cons(sc, ctrl_str, fdat->args));
      if (is_string(obj))
	{
	  format_append_string(sc, fdat, string_value(obj), string_length(obj), port);
	  fdat->args = cdr(fdat->args);
	  fdat->ctr++;
	  return(true);
	}
    }
  return(false);
}


#define MAX_FORMAT_NUMERIC_ARG 10000
static int32_t format_n_arg(s7_scheme *sc, const char *str, int32_t str_len, format_data *fdat, s7_pointer args)
{
  int32_t n;

  if (is_null(fdat->args))          /* (format #f "~nT") */
    just_format_error(sc, "~~N: missing argument", str, args, fdat);
  if (!s7_is_integer(car(fdat->args)))
    just_format_error(sc, "~~N: integer argument required", str, args, fdat);
  n = (int32_t)s7_integer(car(fdat->args));

  if (n < 0)
    just_format_error(sc, "~~N value is negative?", str, args, fdat);
  else
    {
      if (n > MAX_FORMAT_NUMERIC_ARG)
	just_format_error(sc, "~~N value is too big", str, args, fdat);
    }

  fdat->args = cdr(fdat->args);    /* I don't think fdat->ctr should be incremented here -- it's for *vector-print-length* etc */
  return(n);
}


static int32_t format_numeric_arg(s7_scheme *sc, const char *str, int32_t str_len, format_data *fdat, s7_pointer args, int32_t *i)
{
  int32_t width;
  width = format_read_integer(sc, i, str_len, str, args, fdat);
  if (width < 0)
    just_format_error(sc, "width value is negative?", str, fdat->args, fdat);
  else
    {
      if (width > MAX_FORMAT_NUMERIC_ARG)
	just_format_error(sc, "width value is too big", str, fdat->args, fdat);
    }
  return(width);
}


#if WITH_GMP
static bool s7_is_one_or_big_one(s7_pointer p);
#else
#define s7_is_one_or_big_one(Num) s7_is_one(Num)
#endif

static s7_pointer object_to_list(s7_scheme *sc, s7_pointer obj);

static s7_pointer format_to_port_1(s7_scheme *sc, s7_pointer port, const char *str, s7_pointer args,
				   s7_pointer *next_arg, bool with_result, bool columnized, int32_t len, s7_pointer orig_str)
{
  int32_t i, str_len;
  format_data *fdat;
  s7_pointer deferred_port;

  if ((!with_result) &&
      (port == sc->F))
    return(sc->F);

  if (len <= 0)
    {
      str_len = safe_strlen(str);
      if (str_len == 0)
	{
	  if (is_not_null(args))
	    {
	      static s7_pointer null_err = NULL;
	      if (!null_err)
		null_err = s7_make_permanent_string("format control string is null, but there are arguments: ~S");
	      return(s7_error(sc, sc->format_error_symbol, set_elist_2(sc, null_err, args)));
	    }
	  if (with_result)
	    return(make_string_wrapper_with_length(sc, "", 0));
	  return(sc->F);
	}
    }
  else str_len = len;

  sc->format_depth++;
  if (sc->format_depth >= sc->num_fdats)
    {
      int32_t k, new_num_fdats;
      new_num_fdats = sc->format_depth * 2;
      sc->fdats = (format_data **)realloc(sc->fdats, sizeof(format_data *) * new_num_fdats);
      for (k = sc->num_fdats; k < new_num_fdats; k++) sc->fdats[k] = NULL;
      sc->num_fdats = new_num_fdats;
    }

  fdat = sc->fdats[sc->format_depth];
  if (!fdat)
    {
      fdat = (format_data *)malloc(sizeof(format_data));
      sc->fdats[sc->format_depth] = fdat;
      fdat->curly_len = 0;
      fdat->curly_str = NULL;
      fdat->ctr = 0;
    }
  else
    {
      if (fdat->port)
	close_format_port(sc, fdat->port);
      if (fdat->strport)
	close_format_port(sc, fdat->strport);
    }
  fdat->port = NULL;
  fdat->strport = NULL;
  fdat->loc = 0;
  fdat->args = args;
  fdat->orig_str = orig_str;
  fdat->curly_arg = sc->nil;

  /* choose whether to write to a temporary string port, or simply use the in-coming port
   *   if with_result, returned string is wanted.
   *   if port is sc->F, no non-string result is wanted.
   *   if port is not boolean, it better be a port.
   *   if we are about to goto START in eval, and main_stack_op(Sc) == OP_BEGIN1, no return string is wanted -- yow, this is not true
   */

  if (with_result)
    {
      deferred_port = port;
      port = open_format_port(sc); 
      fdat->port = port;
    }
  else deferred_port = sc->F;

  for (i = 0; i < str_len - 1; i++)
    {
      if ((unsigned char)(str[i]) == (unsigned char)'~') /* what does MS C want? */
	{
	  use_write_t use_write;
	  switch (str[i + 1])
	    {
	    case '%':                           /* -------- newline -------- */
	      /* sbcl apparently accepts numeric args here (including 0) */

	      if ((port_data(port)) &&
		  (port_position(port) < port_data_size(port)))
		{
		  port_data(port)[port_position(port)++] = '\n';
		  /* which is actually a bad idea, but as a desperate stopgap, I simply padded
		   *  the string port string with 8 chars that are not in the length.
		   */
		  sc->format_column = 0;
		}
	      else format_append_newline(sc, fdat, port);
	      i++;
	      break;

	    case '&':                           /* -------- conditional newline -------- */
	      /* this only works if all output goes through format -- display/write for example do not update format_column */
	      if (sc->format_column > 0)
		format_append_newline(sc, fdat, port);
	      i++;
	      break;

	    case '~':                           /* -------- tilde -------- */
	      format_append_char(sc, fdat, '~', port);
	      i++;
	      break;

	    case '\n':                          /* -------- trim white-space -------- */
	      for (i = i + 2; i <str_len - 1; i++)
		if (!(white_space[(unsigned char)(str[i])]))
		  {
		    i--;
		    break;
		  }
	      break;

	    case '*':                           /* -------- ignore arg -------- */
	      i++;
	      if (is_null(fdat->args))          /* (format #f "~*~A") */
		format_error(sc, "can't skip argument!", str, args, fdat);
	      fdat->args = cdr(fdat->args);
	      break;

	    case '|':                           /* -------- exit if args nil or ctr > (*s7* 'print-length) -------- */
	      if ((is_pair(fdat->args)) &&
		  (fdat->ctr >= sc->print_length))
		{
		  format_append_string(sc, fdat, " ...", 4, port);
		  fdat->args = sc->nil;
		}
	      /* fall through */

	    case '^':                           /* -------- exit -------- */
	      if (is_null(fdat->args))
		{
		  i = str_len;
		  goto ALL_DONE;
		}
	      i++;
	      break;

	    case '@':                           /* -------- plural, 'y' or 'ies' -------- */
	      i += 2;
	      if ((str[i] != 'P') && (str[i] != 'p'))
		format_error(sc, "unknown '@' directive", str, args, fdat);
	      if (!is_pair(fdat->args))
		format_error(sc, "'@' directive argument missing", str, args, fdat);
	      if (!s7_is_real(car(fdat->args)))        /* CL accepts non numbers here */
		format_error(sc, "'@P' directive argument is not a real number", str, args, fdat);

	      if (!s7_is_one_or_big_one(car(fdat->args)))
		format_append_string(sc, fdat, "ies", 3, port);
	      else format_append_char(sc, fdat, 'y', port);

	      fdat->args = cdr(fdat->args);
	      break;

	    case 'P': case 'p':                 /* -------- plural in 's' -------- */
	      if (!is_pair(fdat->args))
		format_error(sc, "'P' directive argument missing", str, args, fdat);
	      if (!s7_is_real(car(fdat->args)))
		format_error(sc, "'P' directive argument is not a real number", str, args, fdat);
	      if (!s7_is_one_or_big_one(car(fdat->args)))
		format_append_char(sc, fdat, 's', port);
	      i++;
	      fdat->args = cdr(fdat->args);
	      break;

	    case '{':                           /* -------- iteration -------- */
	      {
		int32_t curly_len;

		if (is_null(fdat->args))
		  format_error(sc, "missing argument", str, args, fdat);

		curly_len = format_nesting(str, '{', '}', i, str_len - 1);

		if (curly_len == -1)
		  format_error(sc, "'{' directive, but no matching '}'", str, args, fdat);
		if (curly_len == 1)
		  format_error(sc, "~{~}' doesn't consume any arguments!", str, args, fdat);

		/* what about cons's here?  I can't see any way in CL either to specify the car or cdr of a cons within the format string
		 *   (cons 1 2) is applicable: ((cons 1 2) 0) -> 1
		 *   also there can be applicable objects that won't work in the map context (arg not integer etc)
		 */
		if (is_not_null(car(fdat->args)))               /* (format #f "~{~A ~}" ()) -> "" */
		  {
		    s7_pointer curly_arg;
		    /* perhaps use an iterator here -- rootlet->list is expensive! */
		    curly_arg = object_to_list(sc, car(fdat->args)); /* if a pair, this simply returns the original */
		    if (is_not_null(curly_arg))                /* (format #f "~{~A ~}" #()) -> "" */
		      {
			char *curly_str = NULL;                /* this is the local (nested) format control string */
			s7_pointer orig_arg;

			if (!s7_is_proper_list(sc, curly_arg))
			  format_error(sc, "'{' directive argument should be a proper list or something we can turn into a list", str, args, fdat);
			
			fdat->curly_arg = curly_arg;
			if (curly_arg != car(fdat->args))
			  orig_arg = curly_arg;
			else orig_arg = sc->nil;

			if (curly_len > fdat->curly_len)
			  {
			    if (fdat->curly_str) free (fdat->curly_str);
			    fdat->curly_len = curly_len;
			    fdat->curly_str = (char *)malloc(curly_len * sizeof(char));
			  }
			curly_str = fdat->curly_str;
			memcpy((void *)curly_str, (void *)(str + i + 2), curly_len - 1);
			curly_str[curly_len - 1] = '\0';

			if ((sc->format_depth < sc->num_fdats - 1) &&
			    (sc->fdats[sc->format_depth + 1]))
			  sc->fdats[sc->format_depth + 1]->ctr = 0;
			
			/* it's not easy to use an iterator here instead of a list (so object->list isn't needed above),
			 *   because the curly brackets may enclose multiple arguments -- we would need to use
			 *   iterators throughout this function.
			 */
			while (is_not_null(curly_arg))
			  {
			    s7_pointer new_arg = sc->nil;
			    format_to_port_1(sc, port, curly_str, curly_arg, &new_arg, false, columnized, curly_len - 1, NULL);
			    if (curly_arg == new_arg)
			      {
				fdat->curly_arg = sc->nil;
				format_error(sc, "'{...}' doesn't consume any arguments!", str, args, fdat);
			      }
			    curly_arg = new_arg;
			  }
			fdat->curly_arg = sc->nil;
			while (is_pair(orig_arg))
 			  {
 			    s7_pointer p;
 			    p = orig_arg;
 			    orig_arg = cdr(orig_arg);
 			    free_cell(sc, p);   /* if car(fdar->args) is a hash-table, we could also free_cell(car(p)), but not in any other case */
			  }
		      }
		  }

		i += (curly_len + 2); /* jump past the ending '}' too */
		fdat->args = cdr(fdat->args);
		fdat->ctr++;
	      }
	      break;

	    case '}':
	      format_error(sc, "unmatched '}'", str, args, fdat);

	    case 'W': case 'w':
	      use_write = USE_READABLE_WRITE;
	      goto OBJSTR;

	    case 'S': case 's':
	      use_write = USE_WRITE;
	      goto OBJSTR;

	    case 'A': case 'a':
	      use_write = USE_DISPLAY;
	    OBJSTR:
	      /* object->string */
	      {
		s7_pointer obj, strport;
		if (is_null(fdat->args))
		  format_error(sc, "missing argument", str, args, fdat);

		i++;
		obj = car(fdat->args);
		/* for the column check, we need to know the length of the object->string output */
		if (columnized)
		  {
		    strport = open_format_port(sc);
		    fdat->strport = strport;
		  }
		else strport = port;
		object_out(sc, obj, strport, use_write);
		if (columnized)
		  {
		    if (port_position(strport) >= port_data_size(strport))
		      resize_port_data(strport, port_data_size(strport) * 2);
		    
		    port_data(strport)[port_position(strport)] = '\0';		    
		    if (port_position(strport) > 0)
		      format_append_string(sc, fdat, (const char *)port_data(strport), port_position(strport), port);
		    close_format_port(sc, strport);
		    fdat->strport = NULL;
		  }

		fdat->args = cdr(fdat->args);
		fdat->ctr++;
	      }
	      break;


	      /* -------- numeric args -------- */
	    case '0': case '1': case '2': case '3': case '4': case '5':
	    case '6': case '7': case '8': case '9': case ',':
	    case 'N': case 'n':

	    case 'B': case 'b':
	    case 'D': case 'd':
	    case 'E': case 'e':
	    case 'F': case 'f':
	    case 'G': case 'g':
	    case 'O': case 'o':
	    case 'X': case 'x':

	    case 'T': case 't':
	    case 'C': case 'c':
	      {
		int32_t width = -1, precision = -1;
		char pad = ' ';
		i++;                                      /* str[i] == '~' */

		if (isdigit((int32_t)(str[i])))
		  width = format_numeric_arg(sc, str, str_len, fdat, args, &i); 
		else
		  {
		    if ((str[i] == 'N') || (str[i] == 'n'))
		      {
			i++;
			width = format_n_arg(sc, str, str_len, fdat, args); 
		      }
		  }
		if (str[i] == ',')
		  {
		    i++;                                  /* is (format #f "~12,12D" 1) an error?  The precision has no use here. */
		    if (isdigit((int32_t)(str[i])))
		      precision = format_numeric_arg(sc, str, str_len, fdat, args, &i); 
		    else
		      {
			if ((str[i] == 'N') || (str[i] == 'n'))
			  {
			    i++;
			    precision = format_n_arg(sc, str, str_len, fdat, args); 
			  }
			else
			  {
			    if (str[i] == '\'')              /* (format #f "~12,'xD" 1) -> "xxxxxxxxxxx1" */
			      {
				pad = str[i + 1];
				i += 2;
				if (i >= str_len)            /* (format #f "~,'") */
				  format_error(sc, "incomplete numeric argument", str, args, fdat);
			      }
			    /* is (let ((str "~12,'xD")) (set! (str 5) #\null) (format #f str 1)) an error? */
			  }
		      }
		  }

		switch (str[i])
		  {
		    /* -------- pad to column --------
		     *   are columns numbered from 1 or 0?  there seems to be disagreement about this directive
		     *   does "space over to" mean including?
		     */

		  case 'T': case 't':
		    if (width == -1) width = 0;
		    if (precision == -1) precision = 0;
		    if ((width > 0) || (precision > 0))         /* (format #f "a~8Tb") */
		      {
			/* (length (substring (format #f "~%~10T.") 1)) == (length (format #f "~10T."))
			 * (length (substring (format #f "~%-~10T.~%") 1)) == (length (format #f "-~10T.~%"))
			 */
			if (precision > 0)
			  {
			    int32_t mult;
			    mult = (int32_t)(ceil((s7_double)(sc->format_column + 1 - width) / (s7_double)precision)); /* CLtL2 ("least positive int") */
			    if (mult < 1) mult = 1;
			    width += (precision * mult);
			  }
			format_append_chars(sc, fdat, pad, width - sc->format_column - 1, port);
		      }
		    break;

		  case 'C': case 'c':
		    {
		      s7_pointer obj;

		      if (is_null(fdat->args))
			format_error(sc, "~~C: missing argument", str, args, fdat);
		      /* the "~~" here and below protects against "~C" being treated as a directive */
		      /* i++; */
		      obj = car(fdat->args);

		      if (!s7_is_character(obj))
			{
			  if (!format_method(sc, str, fdat, port))
			    format_error(sc, "'C' directive requires a character argument", str, args, fdat);
			}
		      else
			{
			  /* here use_write is false, so we just add the char, not its name */
			  if (width == -1)
			    format_append_char(sc, fdat, character(obj), port);
			  else format_append_chars(sc, fdat, character(obj), width, port);
			  fdat->args = cdr(fdat->args);
			  fdat->ctr++;
			}
		    }
		    break;

		    /* -------- numbers -------- */
		  case 'F': case 'f':
		    if (is_null(fdat->args))
		      format_error(sc, "~~F: missing argument", str, args, fdat);
		    if (!(s7_is_number(car(fdat->args))))
		      {
			if (!format_method(sc, str, fdat, port))
			  format_error(sc, "~~F: numeric argument required", str, args, fdat);
		      }
		    else format_number(sc, fdat, 10, width, precision, 'f', pad, port);
		    break;

		  case 'G': case 'g':
		    if (is_null(fdat->args))
		      format_error(sc, "~~G: missing argument", str, args, fdat);
		    if (!(s7_is_number(car(fdat->args))))
		      {
			if (!format_method(sc, str, fdat, port))
			  format_error(sc, "~~G: numeric argument required", str, args, fdat);
		      }
		    else format_number(sc, fdat, 10, width, precision, 'g', pad, port);
		    break;

		  case 'E': case 'e':
		    if (is_null(fdat->args))
		      format_error(sc, "~~E: missing argument", str, args, fdat);
		    if (!(s7_is_number(car(fdat->args))))
		      {
			if (!format_method(sc, str, fdat, port))
			  format_error(sc, "~~E: numeric argument required", str, args, fdat);
		      }
		    else format_number(sc, fdat, 10, width, precision, 'e', pad, port);
		    break;

		    /* how to handle non-integer arguments in the next 4 cases?  clisp just returns
		     *   the argument: (format nil "~X" 1.25) -> "1.25" which is perverse (ClTl2 p 581:
		     *   "if arg is not an integer, it is printed in ~A format and decimal base")!!
		     *   I think I'll use the type of the number to choose the output format.
		     */
		  case 'D': case 'd':
		    if (is_null(fdat->args))
		      format_error(sc, "~~D: missing argument", str, args, fdat);
		    if (!(s7_is_number(car(fdat->args))))
		      {
			/* (let () (require mockery.scm) (format #f "~D" ((*mock-number* 'mock-number) 123)))
			 *    port here is a string-port, str has the width/precision data if the caller wants it,
			 *    args is the current arg.  But format_number handles fdat->args and so on, so
			 *    I think I'll pass the format method the current control string (str), the
			 *    current object (car(fdat->args)), and the arglist (args), and assume it will
			 *    return a (scheme) string.
			 */
			if (!format_method(sc, str, fdat, port))
			  format_error(sc, "~~D: numeric argument required", str, args, fdat);
		      }
		    else format_number(sc, fdat, 10, width, precision, 'd', pad, port);
		    break;

		  case 'O': case 'o':
		    if (is_null(fdat->args))
		      format_error(sc, "~~O: missing argument", str, args, fdat);
		    if (!(s7_is_number(car(fdat->args))))
		      {
			if (!format_method(sc, str, fdat, port))
			  format_error(sc, "~~O: numeric argument required", str, args, fdat);
		      }
		    else format_number(sc, fdat, 8, width, precision, 'o', pad, port);
		    break;

		  case 'X': case 'x':
		    if (is_null(fdat->args))
		      format_error(sc, "~~X: missing argument", str, args, fdat);
		    if (!(s7_is_number(car(fdat->args))))
		      {
			if (!format_method(sc, str, fdat, port))
			  format_error(sc, "~~X: numeric argument required", str, args, fdat);
		      }
		    else format_number(sc, fdat, 16, width, precision, 'x', pad, port);
		    break;

		  case 'B': case 'b':
		    if (is_null(fdat->args))
		      format_error(sc, "~~B: missing argument", str, args, fdat);
		    if (!(s7_is_number(car(fdat->args))))
		      {
			if (!format_method(sc, str, fdat, port))
			  format_error(sc, "~~B: numeric argument required", str, args, fdat);
		      }
		    else format_number(sc, fdat, 2, width, precision, 'b', pad, port);
		    break;

		  default:
		    if (width > 0) 
		      format_error(sc, "unused numeric argument", str, args, fdat);
		    format_error(sc, "unimplemented format directive", str, args, fdat);
		  }
	      }
	      break;

	    default:
	      format_error(sc, "unimplemented format directive", str, args, fdat);
	    }
	}
      else /* str[i] is not #\~ */
	{
	  int32_t j, new_len;
	  const char *p;

	  p = (char *)strchr((const char *)(str + i + 1), (int)'~');
	  if (!p)
	    j = str_len;
	  else j = (int32_t)(p - str);
	  new_len = j - i;

	  if ((port_data(port)) &&
	      ((port_position(port) + new_len) < port_data_size(port)))
	    {
	      memcpy((void *)(port_data(port) + port_position(port)), (void *)(str + i), new_len);
	      port_position(port) += new_len;
	    }
	  else port_write_string(port)(sc, (char *)(str + i), new_len, port);
	  fdat->loc += new_len;
	  sc->format_column += new_len;
	  i = j - 1;
	}
    }

 ALL_DONE:
  if (next_arg)
    (*next_arg) = fdat->args;
  else
    {
      if (is_not_null(fdat->args))
	format_error(sc, "too many arguments", str, args, fdat);
    }
  if (i < str_len)
    {
      if (str[i] == '~')
	format_error(sc, "control string ends in tilde", str, args, fdat);
      format_append_char(sc, fdat, str[i], port);
    }

  sc->format_depth--;

  if (with_result)
    {
      s7_pointer result;

      if ((is_output_port(deferred_port)) &&
	  (port_position(port) > 0))
	{
	  port_data(port)[port_position(port)] = '\0';
	  port_write_string(deferred_port)(sc, (const char *)port_data(port), port_position(port), deferred_port);
	}
      result = s7_make_string_with_length(sc, (char *)port_data(port), port_position(port));
      close_format_port(sc, port);
      fdat->port = NULL;
      return(result);
    }
  return(sc->F);
}


static bool is_columnizing(const char *str)
{
  /* look for ~t ~,<int>T ~<int>,<int>t */
  char *p;

  for (p = (char *)str; (*p);)
    if (*p++ == '~') /* this is faster than strchr */
      {
	char c;
	c = *p++;
	if ((c == 't') || (c == 'T')) return(true);
	if (!c) return(false);
	if ((c == ',') || ((c >= '0') && (c <= '9')) || (c == 'n') || (c == 'N'))
	  {
	    while (((c >= '0') && (c <= '9')) || (c == 'n') || (c == 'N')) c = *p++;
	    if ((c == 't') || (c == 'T')) return(true);
	    if (!c) return(false);                       /* ~,1 for example */
	    if (c == ',')
	      {
		c = *p++;
		while (((c >= '0') && (c <= '9')) || (c == 'n') || (c == 'N')) c = *p++;
		if ((c == 't') || (c == 'T')) return(true);
		if (!c) return(false);
	      }
	  }
      }
  return(false);
}


static s7_pointer format_to_port(s7_scheme *sc, s7_pointer port, const char *str, s7_pointer args, s7_pointer *next_arg, bool with_result, int32_t len)
{
  return(format_to_port_1(sc, port, str, args, next_arg, with_result, true /* is_columnizing(str) */, len, NULL));
  /* is_columnizing on every call is much slower than ignoring the issue */
}


static s7_pointer g_format_1(s7_scheme *sc, s7_pointer args)
{
  s7_pointer pt, str;
  sc->format_column = 0;
  pt = car(args);

  if (is_string(pt))
    return(format_to_port_1(sc, sc->F, string_value(pt), cdr(args), NULL, true, true, string_length(pt), pt));
  if (is_null(pt)) pt = sc->output_port;     /* () -> (current-output-port) */

  if (!((s7_is_boolean(pt)) ||               /* #f or #t */
	((is_output_port(pt)) &&             /* (current-output-port) or call-with-open-file arg, etc */
	 (!port_is_closed(pt)))))
    method_or_bust_with_type(sc, pt, sc->format_symbol, args, an_output_port_string, 1);

  if (!is_pair(cdr(args)))                   /* (format #f) */
    return(s7_error(sc, sc->format_error_symbol, set_elist_2(sc, s7_make_string_with_length(sc, "format has no control string: ~S", 32), args)));
  str = cadr(args);
  if (!is_string(str))
    method_or_bust(sc, str, sc->format_symbol, args, T_STRING, 2);

  return(format_to_port_1(sc, (pt == sc->T) ? sc->output_port : pt, 
			  string_value(str), cddr(args), NULL, !is_output_port(pt), true, string_length(str), str));
}


static s7_pointer g_format(s7_scheme *sc, s7_pointer args)
{
  #define H_format "(format out str . args) substitutes args into str sending the result to out. Most of \
s7's format directives are taken from CL: ~% = newline, ~& = newline if the preceding output character was \
no a newline, ~~ = ~, ~<newline> trims white space, ~* skips an argument, ~^ exits {} iteration if the arg list is exhausted, \
~nT spaces over to column n, ~A prints a representation of any object, ~S is the same, but puts strings in double quotes, \
~C prints a character, numbers are handled by ~F, ~E, ~G, ~B, ~O, ~D, and ~X with preceding numbers giving \
spacing (and spacing character) and precision.  ~{ starts an embedded format directive which is ended by ~}: \n\
\n\
  >(format #f \"dashed: ~{~A~^-~}\" '(1 2 3))\n\
  \"dashed: 1-2-3\"\n\
\n\
~P inserts \"s\" if the current it is not 1 or 1.0 (use ~@P for \"ies\" or \"y\").\n\
~B is number->string in base 2, ~O in base 8, ~D base 10, ~X base 16,\n\
~E: (format #f \"~E\" 100.1) -&gt; \"1.001000e+02\" (%e in C)\n\
~F: (format #f \"~F\" 100.1) -&gt; \"100.100000\"   (%f in C)\n\
~G: (format #f \"~G\" 100.1) -&gt; \"100.1\"        (%g in C)\n\
\n\
If the 'out' it is not an output port, the resultant string is returned.  If it \
is #t, the string is also sent to the current-output-port."

  #define Q_format s7_make_circular_signature(sc, 1, 2, s7_make_signature(sc, 2, sc->is_string_symbol, sc->is_boolean_symbol), sc->T)
  return(g_format_1(sc, args));
}


const char *s7_format(s7_scheme *sc, s7_pointer args)
{
  s7_pointer result;
  result = g_format_1(sc, args);
  if (is_string(result))
    return(string_value(result));
  return(NULL);
}


/* -------------------------------- system extras -------------------------------- */

#if WITH_SYSTEM_EXTRAS
#include <fcntl.h>

static s7_pointer g_is_directory(s7_scheme *sc, s7_pointer args)
{
  #define H_is_directory "(directory? str) returns #t if str is the name of a directory"
  #define Q_is_directory s7_make_signature(sc, 2, sc->is_boolean_symbol, sc->is_string_symbol)
  s7_pointer name;
  name = car(args);

  if (!is_string(name))
    method_or_bust_one_arg(sc, name, sc->is_directory_symbol, args, T_STRING);
  return(s7_make_boolean(sc, is_directory(string_value(name))));
}

static bool is_directory_b(s7_pointer p)
{
  if (!is_string(p))
    simple_wrong_type_argument(cur_sc, cur_sc->is_directory_symbol, p, T_STRING);
  return(is_directory(string_value(p)));
}


static bool file_probe(const char *arg)
{
#if (!MS_WINDOWS)
  return(access(arg, F_OK) == 0);
#else
  int32_t fd;
  fd = open(arg, O_RDONLY, 0);
  if (fd == -1) return(false);
  close(fd);
  return(true);
#endif
}


static s7_pointer g_file_exists(s7_scheme *sc, s7_pointer args)
{
  #define H_file_exists "(file-exists? filename) returns #t if the file exists"
  #define Q_file_exists s7_make_signature(sc, 2, sc->is_boolean_symbol, sc->is_string_symbol)

  s7_pointer name;
  name = car(args);

  if (!is_string(name))
    method_or_bust_one_arg(sc, name, sc->file_exists_symbol, args, T_STRING);
  return(s7_make_boolean(sc, file_probe(string_value(name))));
}

static bool file_exists_b(s7_pointer p)
{
  if (!is_string(p))
    simple_wrong_type_argument(cur_sc, cur_sc->file_exists_symbol, p, T_STRING);
  return(file_probe(string_value(p)));
}


static s7_pointer g_delete_file(s7_scheme *sc, s7_pointer args)
{
  #define H_delete_file "(delete-file filename) deletes the file filename."
  #define Q_delete_file s7_make_signature(sc, 2, sc->is_integer_symbol, sc->is_string_symbol)

  s7_pointer name;
  name = car(args);

  if (!is_string(name))
    method_or_bust_one_arg(sc, name, sc->delete_file_symbol, args, T_STRING);
  return(make_integer(sc, unlink(string_value(name))));
}


static s7_pointer g_getenv(s7_scheme *sc, s7_pointer args)
{
  #define H_getenv "(getenv var) returns the value of an environment variable."
  #define Q_getenv pcl_s

  s7_pointer name;
  name = car(args);

  if (!is_string(name))
    method_or_bust_one_arg(sc, name, sc->getenv_symbol, args, T_STRING);
  return(s7_make_string(sc, getenv(string_value(name))));
}


static s7_pointer g_system(s7_scheme *sc, s7_pointer args)
{
  #define H_system "(system command) executes the command.  If the optional second argument is #t, \
system captures the output as a string and returns it."
  #define Q_system s7_make_signature(sc, 3, s7_make_signature(sc, 2, sc->is_integer_symbol, sc->is_string_symbol), sc->is_string_symbol, sc->is_boolean_symbol)

  s7_pointer name;
  name = car(args);

  if (!is_string(name))
    method_or_bust_one_arg(sc, name, sc->system_symbol, args, T_STRING);

  if ((is_pair(cdr(args))) &&
      (cadr(args) == sc->T))
    {
      #define BUF_SIZE 256
      char buf[BUF_SIZE];
      char *str = NULL;
      int32_t cur_len = 0, full_len = 0;
      FILE *fd;
      s7_pointer res;

      fd = popen(string_value(name), "r");
      while (fgets(buf, BUF_SIZE, fd))
	{
	  int32_t buf_len;
	  buf_len = safe_strlen(buf);
	  if (cur_len + buf_len >= full_len)
	    {
	      full_len += BUF_SIZE * 2;
	      if (str)
		str = (char *)realloc(str, full_len * sizeof(char));
	      else str = (char *)malloc(full_len * sizeof(char));
	    }
	  memcpy((void *)(str + cur_len), (void *)buf, buf_len);
	  cur_len += buf_len;
	}
      pclose(fd);

      res = s7_make_string_with_length(sc, str, cur_len);
      if (str) free(str);
      return(res);
    }
  return(make_integer(sc, system(string_value(name))));
}


#if (!MS_WINDOWS)
#include <dirent.h>

static s7_pointer g_directory_to_list(s7_scheme *sc, s7_pointer args)
{
  s7_pointer name;
  DIR *dpos;
  s7_pointer result;

  #define H_directory_to_list "(directory->list directory) returns the contents of the directory as a list of strings (filenames)."
  #define Q_directory_to_list s7_make_signature(sc, 2, sc->is_pair_symbol, sc->is_string_symbol)

  name = car(args);
  if (!is_string(name))
    method_or_bust_one_arg(sc, name, sc->directory_to_list_symbol, list_1(sc, name), T_STRING);

  sc->w = sc->nil;
  if ((dpos = opendir(string_value(name))))
    {
      struct dirent *dirp;
      while ((dirp = readdir(dpos)))
	sc->w = cons(sc, s7_make_string(sc, dirp->d_name), sc->w);
      closedir(dpos);
    }

  result = sc->w;
  sc->w = sc->nil;
  return(result);
}

static s7_pointer g_file_mtime(s7_scheme *sc, s7_pointer args)
{
  #define H_file_mtime "(file-mtime file): return the write date of file"
  #define Q_file_mtime s7_make_signature(sc, 2, sc->is_integer_symbol, sc->is_string_symbol)

  struct stat statbuf;
  int32_t err;
  s7_pointer name;

  name = car(args);
  if (!is_string(name))
    method_or_bust_one_arg(sc, name, sc->file_mtime_symbol, args, T_STRING);

  err = stat(string_value(name), &statbuf);
  if (err < 0)
    return(file_error(sc, "file-mtime", strerror(errno), string_value(name)));

  return(s7_make_integer(sc, (s7_int)(statbuf.st_mtime)));
}
#endif
#endif



/* -------------------------------- lists -------------------------------- */

s7_pointer s7_cons(s7_scheme *sc, s7_pointer a, s7_pointer b)
{
  s7_pointer x;
  new_cell(sc, x, T_PAIR | T_SAFE_PROCEDURE);
  set_car(x, a);
  set_cdr(x, b);
  return(x);
}


static s7_pointer cons_unchecked(s7_scheme *sc, s7_pointer a, s7_pointer b)
{
  /* apparently slightly faster as a function? */
  s7_pointer x;
  new_cell_no_check(sc, x, T_PAIR | T_SAFE_PROCEDURE);
  set_car(x, a);
  set_cdr(x, b);
  return(x);
}


static s7_pointer permanent_cons(s7_pointer a, s7_pointer b, uint64_t type)
{
  /* for the symbol table which is never GC'd (and its contents aren't marked) */
  s7_pointer x;
  x = alloc_pointer();
  set_type(x, type);
  unheap(x);
  set_car(x, a);
  set_cdr(x, b);
  return(x);
}

static s7_pointer permanent_list(s7_scheme *sc, int32_t len)
{
  int32_t j;
  s7_pointer p;
  p = sc->nil;
  for (j = 0; j < len; j++)
    p = permanent_cons(sc->nil, p, T_PAIR | T_IMMUTABLE);
  return(p);
}


static void check_sig_entry(s7_scheme *sc, s7_pointer p, s7_pointer res, bool circle)
{
  if ((!is_symbol(car(p))) && 
      (!s7_is_boolean(car(p))) &&
      (!is_pair(car(p))))
    {
      s7_pointer np;
      int32_t i;
      for (np = res, i = 0; np != p; np = cdr(np), i++);
      fprintf(stderr, "s7_make_%ssignature got an invalid entry at position %d: (", (circle) ? "circular_" : "", i);
      for (np = res; np != p; np = cdr(np))
	fprintf(stderr, "%s ", DISPLAY(car(np)));
      fprintf(stderr, "...");
      set_car(p, sc->nil);
    }
}

s7_pointer s7_make_signature(s7_scheme *sc, int32_t len, ...)
{
  va_list ap;
  s7_pointer p, res;

  res = permanent_list(sc, len);
  va_start(ap, len);
  for (p = res; is_pair(p); p = cdr(p))
    {
      set_car(p, va_arg(ap, s7_pointer));
      check_sig_entry(sc, p, res, false);
    }
  va_end(ap);

  return((s7_pointer)res);
}

s7_pointer s7_make_circular_signature(s7_scheme *sc, int32_t cycle_point, int32_t len, ...)
{
  va_list ap;
  int32_t i;
  s7_pointer p, res, back = NULL, end = NULL;

  res = permanent_list(sc, len);
  va_start(ap, len);
  for (p = res, i = 0; is_pair(p); p = cdr(p), i++)
    {
      set_car(p, va_arg(ap, s7_pointer));
      check_sig_entry(sc, p, res, true);
      if (i == cycle_point) back = p;
      if (i == (len - 1)) end = p;
    }
  va_end(ap);
  if (end) set_cdr(end, back);
  if (i < len) 
    fprintf(stderr, "s7_make_circular_signature got too few entries: %s\n", DISPLAY(res));
  return((s7_pointer)res);
}


bool s7_is_pair(s7_pointer p) {return(is_pair(p));}
static s7_pointer is_pair_p_p(s7_pointer p) {return((is_pair(p)) ? cur_sc->T : cur_sc->F);}


s7_pointer s7_car(s7_pointer p) {return(car(p));}
s7_pointer s7_cdr(s7_pointer p) {return(cdr(p));}

s7_pointer s7_cadr(s7_pointer p) {return(cadr(p));}
s7_pointer s7_cddr(s7_pointer p) {return(cddr(p));}
s7_pointer s7_cdar(s7_pointer p) {return(cdar(p));}
s7_pointer s7_caar(s7_pointer p) {return(caar(p));}

s7_pointer s7_caadr(s7_pointer p) {return(caadr(p));}
s7_pointer s7_caddr(s7_pointer p) {return(caddr(p));}
s7_pointer s7_cadar(s7_pointer p) {return(cadar(p));}
s7_pointer s7_caaar(s7_pointer p) {return(caaar(p));}
s7_pointer s7_cdadr(s7_pointer p) {return(cdadr(p));}
s7_pointer s7_cdddr(s7_pointer p) {return(cdddr(p));}
s7_pointer s7_cddar(s7_pointer p) {return(cddar(p));}
s7_pointer s7_cdaar(s7_pointer p) {return(cdaar(p));}

s7_pointer s7_caaadr(s7_pointer p) {return(caaadr(p));}
s7_pointer s7_caaddr(s7_pointer p) {return(caaddr(p));}
s7_pointer s7_caadar(s7_pointer p) {return(caadar(p));}
s7_pointer s7_caaaar(s7_pointer p) {return(caaaar(p));}
s7_pointer s7_cadadr(s7_pointer p) {return(cadadr(p));}
s7_pointer s7_cadddr(s7_pointer p) {return(cadddr(p));}
s7_pointer s7_caddar(s7_pointer p) {return(caddar(p));}
s7_pointer s7_cadaar(s7_pointer p) {return(cadaar(p));}

s7_pointer s7_cdaadr(s7_pointer p) {return(cdaadr(p));}
s7_pointer s7_cdaddr(s7_pointer p) {return(cdaddr(p));}
s7_pointer s7_cdadar(s7_pointer p) {return(cdadar(p));}
s7_pointer s7_cdaaar(s7_pointer p) {return(cdaaar(p));}
s7_pointer s7_cddadr(s7_pointer p) {return(cddadr(p));}
s7_pointer s7_cddddr(s7_pointer p) {return(cddddr(p));}
s7_pointer s7_cdddar(s7_pointer p) {return(cdddar(p));}
s7_pointer s7_cddaar(s7_pointer p) {return(cddaar(p));}


s7_pointer s7_set_car(s7_pointer p, s7_pointer q)
{
  set_car(p, q);
  return(q);  /* was p? 5-Aug-17 */
}


s7_pointer s7_set_cdr(s7_pointer p, s7_pointer q)
{
  set_cdr(p, q);
  return(q);  /* was p? 5-Aug-17 */
}

/* -------------------------------------------------------------------------------- */

s7_pointer s7_apply_1(s7_scheme *sc, s7_pointer args, s7_pointer (*f1)(s7_pointer a1))
{
  /* not currently used */
  return(f1(car(args)));
}

s7_pointer s7_apply_2(s7_scheme *sc, s7_pointer args, s7_pointer (*f2)(s7_pointer a1, s7_pointer a2))
{
  return(f2(car(args), cadr(args)));
}

s7_pointer s7_apply_3(s7_scheme *sc, s7_pointer args, s7_pointer (*f3)(s7_pointer a1, s7_pointer a2, s7_pointer a3))
{
  s7_pointer a1;
  a1 = car(args);  args = cdr(args);
  return(f3(a1, car(args), cadr(args)));
}

s7_pointer s7_apply_4(s7_scheme *sc, s7_pointer args, s7_pointer (*f4)(s7_pointer a1, s7_pointer a2, s7_pointer a3, s7_pointer a4))
{
  s7_pointer a1, a2;
  a1 = car(args);  a2 = cadr(args);  args = cddr(args);
  return(f4(a1, a2, car(args), cadr(args)));
}

s7_pointer s7_apply_5(s7_scheme *sc, s7_pointer args, s7_pointer (*f5)(s7_pointer a1, s7_pointer a2, s7_pointer a3, s7_pointer a4, s7_pointer a5))
{
  s7_pointer a1, a2, a3, a4;
  a1 = car(args);  a2 = cadr(args);  args = cddr(args);
  a3 = car(args);  a4 = cadr(args);  args = cddr(args);
  return(f5(a1, a2, a3, a4, car(args)));
}

s7_pointer s7_apply_6(s7_scheme *sc, s7_pointer args, s7_pointer (*f6)(s7_pointer a1, s7_pointer a2, s7_pointer a3, s7_pointer a4, s7_pointer a5, s7_pointer a6))
{
  s7_pointer a1, a2, a3, a4;
  a1 = car(args);  a2 = cadr(args);  args = cddr(args);
  a3 = car(args);  a4 = cadr(args);  args = cddr(args);
  return(f6(a1, a2, a3, a4, car(args), cadr(args)));
}

s7_pointer s7_apply_7(s7_scheme *sc, s7_pointer args,
		      s7_pointer (*f7)(s7_pointer a1, s7_pointer a2, s7_pointer a3, s7_pointer a4, s7_pointer a5, s7_pointer a6, s7_pointer a7))
{
  s7_pointer a1, a2, a3, a4, a5, a6;
  a1 = car(args);  a2 = cadr(args);  args = cddr(args);
  a3 = car(args);  a4 = cadr(args);  args = cddr(args);
  a5 = car(args);  a6 = cadr(args);  args = cddr(args);
  return(f7(a1, a2, a3, a4, a5, a6, car(args)));
}

s7_pointer s7_apply_8(s7_scheme *sc, s7_pointer args,
		      s7_pointer (*f8)(s7_pointer a1, s7_pointer a2, s7_pointer a3, s7_pointer a4, s7_pointer a5, s7_pointer a6, s7_pointer a7, s7_pointer a8))
{
  s7_pointer a1, a2, a3, a4, a5, a6;
  a1 = car(args);  a2 = cadr(args);  args = cddr(args);
  a3 = car(args);  a4 = cadr(args);  args = cddr(args);
  a5 = car(args);  a6 = cadr(args);  args = cddr(args);
  return(f8(a1, a2, a3, a4, a5, a6, car(args), cadr(args)));
}

s7_pointer s7_apply_9(s7_scheme *sc, s7_pointer args, s7_pointer (*f9)(s7_pointer a1, s7_pointer a2, s7_pointer a3, s7_pointer a4,
								       s7_pointer a5, s7_pointer a6, s7_pointer a7, s7_pointer a8, s7_pointer a9))
{
  s7_pointer a1, a2, a3, a4, a5, a6;
  a1 = car(args);  a2 = cadr(args);  args = cddr(args);
  a3 = car(args);  a4 = cadr(args);  args = cddr(args);
  a5 = car(args);  a6 = cadr(args);  args = cddr(args);
  return(f9(a1, a2, a3, a4, a5, a6, car(args), cadr(args), caddr(args)));
}

s7_pointer s7_apply_n_1(s7_scheme *sc, s7_pointer args, s7_pointer (*f1)(s7_pointer a1))
{
  if (is_pair(args))
    return(f1(car(args)));
  return(f1(sc->undefined));
}

s7_pointer s7_apply_n_2(s7_scheme *sc, s7_pointer args, s7_pointer (*f2)(s7_pointer a1, s7_pointer a2))
{
  if (is_pair(args))
    {
      if (is_pair(cdr(args)))
	return(f2(car(args), cadr(args)));
      return(f2(car(args), sc->undefined));
    }
  return(f2(sc->undefined, sc->undefined));
}

s7_pointer s7_apply_n_3(s7_scheme *sc, s7_pointer args, s7_pointer (*f3)(s7_pointer a1, s7_pointer a2, s7_pointer a3))
{
  if (is_pair(args))
    {
      s7_pointer a1;
      a1 = car(args); args = cdr(args);
      if (is_pair(args))
	{
	  s7_pointer a2;
	  a2 = car(args);
	  if (is_pair(cdr(args)))
	    return(f3(a1, a2, cadr(args)));
	  return(f3(a1, a2, sc->undefined));
	}
      return(f3(a1, sc->undefined, sc->undefined));
    }
  return(f3(sc->undefined, sc->undefined, sc->undefined));
}

s7_pointer s7_apply_n_4(s7_scheme *sc, s7_pointer args, s7_pointer (*f4)(s7_pointer a1, s7_pointer a2, s7_pointer a3, s7_pointer a4))
{
  if (is_pair(args))
    {
      s7_pointer a1;
      a1 = car(args); args = cdr(args);
      if (is_pair(args))
	{
	  s7_pointer a2;
	  a2 = car(args); args = cdr(args);
	  if (is_pair(args))
	    {
	      s7_pointer a3;
	      a3 = car(args);
	      if (is_pair(cdr(args)))
		return(f4(a1, a2, a3, cadr(args)));
	      return(f4(a1, a2, a3, sc->undefined));
	    }
	  return(f4(a1, a2, sc->undefined, sc->undefined));
	}
      return(f4(a1, sc->undefined, sc->undefined, sc->undefined));
    }
  return(f4(sc->undefined, sc->undefined, sc->undefined, sc->undefined));
}

s7_pointer s7_apply_n_5(s7_scheme *sc, s7_pointer args,
			s7_pointer (*f5)(s7_pointer a1, s7_pointer a2, s7_pointer a3, s7_pointer a4, s7_pointer a5))
{
  if (is_pair(args))
    {
      s7_pointer a1;
      a1 = car(args); args = cdr(args);
      if (is_pair(args))
	{
	  s7_pointer a2;
	  a2 = car(args); args = cdr(args);
	  if (is_pair(args))
	    {
	      s7_pointer a3;
	      a3 = car(args); args = cdr(args);
	      if (is_pair(args))
		{
		  s7_pointer a4;
		  a4 = car(args);
		  if (is_pair(cdr(args)))
		    return(f5(a1, a2, a3, a4, cadr(args)));
		  return(f5(a1, a2, a3, a4, sc->undefined));
		}
	      return(f5(a1, a2, a3, sc->undefined, sc->undefined));
	    }
	  return(f5(a1, a2, sc->undefined, sc->undefined, sc->undefined));
	}
      return(f5(a1, sc->undefined, sc->undefined, sc->undefined, sc->undefined));
    }
  return(f5(sc->undefined, sc->undefined, sc->undefined, sc->undefined, sc->undefined));
}

s7_pointer s7_apply_n_6(s7_scheme *sc, s7_pointer args,
			s7_pointer (*f6)(s7_pointer a1, s7_pointer a2, s7_pointer a3, s7_pointer a4, s7_pointer a5, s7_pointer a6))
{
  s7_pointer a1, a2, a3, a4, a5, a6;
  a1 = sc->undefined; a2 = sc->undefined; a3 = sc->undefined; a4 = sc->undefined; a5 = sc->undefined; a6 = sc->undefined;
  if (is_pair(args))
    {
      a1 = car(args); args = cdr(args);
      if (is_pair(args))
	{
	  a2 = car(args); args = cdr(args);
	  if (is_pair(args))
	    {
	      a3 = car(args); args = cdr(args);
	      if (is_pair(args))
		{
		  a4 = car(args); args = cdr(args);
		  if (is_pair(args))
		    {
		      a5 = car(args);
		      if (is_pair(cdr(args))) a6 = cadr(args);
		    }}}}}
  return(f6(a1, a2, a3, a4, a5, a6));
}

s7_pointer s7_apply_n_7(s7_scheme *sc, s7_pointer args,
			s7_pointer (*f7)(s7_pointer a1, s7_pointer a2, s7_pointer a3, s7_pointer a4,
					 s7_pointer a5, s7_pointer a6, s7_pointer a7))
{
  s7_pointer a1, a2, a3, a4, a5, a6, a7;
  a1 = sc->undefined; a2 = sc->undefined; a3 = sc->undefined; a4 = sc->undefined; a5 = sc->undefined;
  a6 = sc->undefined, a7 = sc->undefined;
  if (is_pair(args))
    {
      a1 = car(args); args = cdr(args);
      if (is_pair(args))
	{
	  a2 = car(args); args = cdr(args);
	  if (is_pair(args))
	    {
	      a3 = car(args); args = cdr(args);
	      if (is_pair(args))
		{
		  a4 = car(args); args = cdr(args);
		  if (is_pair(args))
		    {
		      a5 = car(args); args = cdr(args);
		      if (is_pair(args))
			{
			  a6 = car(args);
			  if (is_pair(cdr(args))) a7 = cadr(args);
			}}}}}}
  return(f7(a1, a2, a3, a4, a5, a6, a7));
}

s7_pointer s7_apply_n_8(s7_scheme *sc, s7_pointer args,
			s7_pointer (*f8)(s7_pointer a1, s7_pointer a2, s7_pointer a3, s7_pointer a4,
					 s7_pointer a5, s7_pointer a6, s7_pointer a7, s7_pointer a8))
{
  s7_pointer a1, a2, a3, a4, a5, a6, a7, a8;
  a1 = sc->undefined; a2 = sc->undefined; a3 = sc->undefined; a4 = sc->undefined; a5 = sc->undefined;
  a6 = sc->undefined, a7 = sc->undefined; a8 = sc->undefined;
  if (is_pair(args))
    {
      a1 = car(args); args = cdr(args);
      if (is_pair(args))
	{
	  a2 = car(args); args = cdr(args);
	  if (is_pair(args))
	    {
	      a3 = car(args); args = cdr(args);
	      if (is_pair(args))
		{
		  a4 = car(args); args = cdr(args);
		  if (is_pair(args))
		    {
		      a5 = car(args); args = cdr(args);
		      if (is_pair(args))
			{
			  a6 = car(args); args = cdr(args);
			  if (is_pair(args))
			    {
			      a7 = car(args);
			      if (is_pair(cdr(args))) a8 = cadr(args);
			    }}}}}}}
  return(f8(a1, a2, a3, a4, a5, a6, a7, a8));
}

s7_pointer s7_apply_n_9(s7_scheme *sc, s7_pointer args,
			s7_pointer (*f9)(s7_pointer a1, s7_pointer a2, s7_pointer a3, s7_pointer a4,
					 s7_pointer a5, s7_pointer a6, s7_pointer a7, s7_pointer a8,
					 s7_pointer a9))
{
  s7_pointer a1, a2, a3, a4, a5, a6, a7, a8, a9;
  a1 = sc->undefined; a2 = sc->undefined; a3 = sc->undefined; a4 = sc->undefined; a5 = sc->undefined;
  a6 = sc->undefined, a7 = sc->undefined; a8 = sc->undefined; a9 = sc->undefined;
  if (is_pair(args))
    {
      a1 = car(args); args = cdr(args);
      if (is_pair(args))
	{
	  a2 = car(args); args = cdr(args);
	  if (is_pair(args))
	    {
	      a3 = car(args); args = cdr(args);
	      if (is_pair(args))
		{
		  a4 = car(args); args = cdr(args);
		  if (is_pair(args))
		    {
		      a5 = car(args); args = cdr(args);
		      if (is_pair(args))
			{
			  a6 = car(args); args = cdr(args);
			  if (is_pair(args))
			    {
			      a7 = car(args); args = cdr(args);
			      if (is_pair(args))
				{
				  a8 = car(args);
				  if (is_pair(cdr(args))) a9 = cadr(args);
				}}}}}}}}
  return(f9(a1, a2, a3, a4, a5, a6, a7, a8, a9));
}

/* -------------------------------------------------------------------------------- */



s7_pointer s7_list_ref(s7_scheme *sc, s7_pointer lst, int32_t num)
{
  int32_t i;
  s7_pointer x;
  for (x = lst, i = 0; (i < num) && (is_pair(x)); i++, x = cdr(x)) {}
  if ((i == num) && (is_pair(x)))
    return(car(x));
  return(sc->nil);
}


s7_pointer s7_list_set(s7_scheme *sc, s7_pointer lst, int32_t num, s7_pointer val)
{
  int32_t i;
  s7_pointer x;
  for (x = lst, i = 0; (i < num) && (is_pair(x)); i++, x = cdr(x)) {}
  if ((i == num) &&
      (is_pair(x)))
    set_car(x, _NFre(val));
  return(val);
}


s7_pointer s7_member(s7_scheme *sc, s7_pointer sym, s7_pointer lst)
{
  s7_pointer x;
  for (x = lst; is_pair(x); x = cdr(x))
    if (s7_is_equal(sc, sym, car(x)))
      return(x);
  return(sc->F);
}


static bool symbol_is_in_arg_list(s7_pointer sym, s7_pointer lst)
{
  s7_pointer x;
  for (x = lst; is_pair(x); x = cdr(x))
    if ((sym == car(x)) ||
	((is_pair(car(x))) &&
	 (sym == caar(x))))
	return(true);
  return(sym == x);
}

static s7_int tree_len_1(s7_scheme *sc, s7_pointer p)
{
  s7_int sum;
  for (sum = 0; is_pair(p); p = cdr(p))
    {
      if ((!is_pair(car(p))) ||
	  (caar(p) == sc->quote_symbol))
	sum++;
      else sum += tree_len_1(sc, car(p));
    }
  if (!is_null(p)) return(sum + 1);
  return(sum);
}

static s7_int tree_len(s7_scheme *sc, s7_pointer p)
{
  if (is_null(p))
    return(0);
  if ((!is_pair(p)) ||
      (car(p) == sc->quote_symbol))
    return(1);
  return(tree_len_1(sc, p));
}

static s7_int tree_leaves_i(s7_pointer p) {return(tree_len(cur_sc, p));}

static s7_pointer g_tree_leaves(s7_scheme *sc, s7_pointer args)
{
  #define H_tree_leaves "(tree-leaves tree) returns the number of leaves in the tree"
  #define Q_tree_leaves s7_make_signature(sc, 2, sc->is_integer_symbol, sc->is_list_symbol)
  return(s7_make_integer(sc, tree_len(sc, car(args))));
}


bool s7_tree_memq(s7_scheme *sc, s7_pointer sym, s7_pointer tree)
{
  if (sym == tree) return(true);
  if (!is_pair(tree)) return(false);
  if (car(tree) == sc->quote_symbol)
    {
      if ((is_symbol(sym)) || (is_pair(sym)) || (!is_pair(cdr(tree))))
	return(false);
      return(sym == cadr(tree));
    }
  do {
    if ((sym == cdr(tree)) ||    /* "sym" need not be a symbol */
	(s7_tree_memq(sc, sym, car(tree))))
      return(true);
    tree = cdr(tree);
  } while (is_pair(tree));
  return(false);
}

static s7_pointer g_tree_memq(s7_scheme *sc, s7_pointer args)
{
  #define H_tree_memq "(tree-memq obj tree) is a tree-oriented version of memq, but returning #t if the object is in the tree."
  #define Q_tree_memq s7_make_signature(sc, 3, sc->is_boolean_symbol, sc->T, sc->is_list_symbol)
  return(make_boolean(sc, s7_tree_memq(sc, car(args), cadr(args))));
}

static bool tree_memq_b_pp(s7_pointer sym, s7_pointer tree) {return(s7_tree_memq(cur_sc, sym, tree));}


static bool tree_set_memq(s7_scheme *sc, s7_pointer tree)
{
  if (is_symbol(tree))
    return(symbol_is_in_list(sc, tree));
  if ((!is_pair(tree)) ||
      (car(tree) == sc->quote_symbol))
    return(false);
  do {
    if (tree_set_memq(sc, car(tree)))
      return(true);
    tree = cdr(tree);
  } while (is_pair(tree));
  return((is_symbol(tree)) && (symbol_is_in_list(sc, tree)));
}

static s7_pointer g_tree_set_memq(s7_scheme *sc, s7_pointer args)
{
  #define H_tree_set_memq "(tree-set-memq symbols tree) returns #t if any of the list of symbols is in the tree"
  #define Q_tree_set_memq s7_make_signature(sc, 3, sc->is_boolean_symbol, sc->is_list_symbol, sc->is_list_symbol)
  s7_pointer syms, p, tree;
  syms = car(args);
  if (!is_pair(syms)) return(sc->F);
  tree = cadr(args);
  clear_symbol_list(sc);
  for (p = syms; is_pair(p); p = cdr(p))
    if (is_symbol(car(p)))
      add_symbol_to_list(sc, car(p));
  return(make_boolean(sc, tree_set_memq(sc, tree)));
}

static bool tree_set_memq_b_pp(s7_pointer syms, s7_pointer tree) {return(g_tree_set_memq(cur_sc, set_plist_2(cur_sc, syms, tree)) != cur_sc->F);}

static s7_int tree_count(s7_scheme *sc, s7_pointer x, s7_pointer p, s7_int count)
{
  if (p == x)
    return(count + 1);
  if ((!is_pair(p)) || (car(p) == sc->quote_symbol))
    return(count);
  return(tree_count(sc, x, cdr(p), tree_count(sc, x, car(p), count)));
}

static inline s7_int tree_count_at_least(s7_scheme *sc, s7_pointer x, s7_pointer p, s7_int count, s7_int top)
{
  if (p == x)
    return(count + 1);
  if ((!is_pair(p)) || (car(p) == sc->quote_symbol) || (count >= top))
    return(count);
  return(tree_count_at_least(sc, x, cdr(p), tree_count_at_least(sc, x, car(p), count, top), top));
}

static s7_pointer g_tree_count(s7_scheme *sc, s7_pointer args)
{
  #define H_tree_count "(tree-count obj tree max-count) returns how many times obj is in tree, stopping at max-count (if specified)"
  #define Q_tree_count s7_make_signature(sc, 4, sc->is_integer_symbol, sc->T, sc->is_list_symbol, sc->is_integer_symbol)
  s7_pointer obj, tree, count;

  obj = car(args);
  tree = cadr(args);
  if (!is_pair(tree))
    return((obj == tree) ? small_int(1) : small_int(0));
  if (is_null(cddr(args)))
    return(s7_make_integer(sc, tree_count(sc, obj, tree, 0)));
  count = caddr(args);
  if (!s7_is_integer(count))
    return(simple_wrong_type_argument(sc, sc->tree_count_symbol, count, T_INTEGER));
  return(s7_make_integer(sc, tree_count_at_least(sc, obj, tree, 0, s7_integer(count))));
}


static s7_pointer print_truncate(s7_scheme *sc, s7_pointer code)
{
  if (tree_len(sc, code) > sc->print_length)
    {
      char *str;
      str = object_to_truncated_string(sc, code, sc->print_length * 10);
      return(make_string_uncopied_with_length(sc, str, safe_strlen(str)));
    }
  return(code);
}  
 

s7_pointer s7_assoc(s7_scheme *sc, s7_pointer sym, s7_pointer lst)
{
  s7_pointer x, y;

  if (!is_pair(lst))
    return(sc->F);

  x = lst;
  y = lst;
  while (true)
    {
      if ((is_pair(car(x))) && (s7_is_equal(sc, sym, caar(x)))) return(car(x));
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      if ((is_pair(car(x))) && (s7_is_equal(sc, sym, caar(x)))) return(car(x));
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      y = cdr(y);
      if (x == y) return(sc->F);
    }
  return(sc->F);
}


s7_pointer s7_reverse(s7_scheme *sc, s7_pointer a)
{
  /* reverse list -- produce new list (other code assumes this function does not return the original!) */
  s7_pointer x, p;

  if (is_null(a)) return(a);

  if (!is_pair(cdr(a)))
    {
      if (is_not_null(cdr(a)))
	return(cons(sc, cdr(a), car(a)));
      return(cons(sc, car(a), sc->nil)); /* don't return 'a' itself */
    }

  sc->w = list_1(sc, car(a));
  for (x = cdr(a), p = a; is_pair(x); x = cdr(x), p = cdr(p))
    {
      sc->w = cons(sc, car(x), sc->w);
      if (is_pair(cdr(x)))
	{
	  x = cdr(x);
	  sc->w = cons(sc, car(x), sc->w);
	}
      if (x == p) /* this can take awhile to notice there's a cycle, but what does the caller expect? */
	break;
    }

  if (is_not_null(x))
    p = cons(sc, x, sc->w);    /* ?? this means that (reverse '(1 2 . 3)) returns '(3 2 1) -- we used to return () here */
  else p = sc->w;

  sc->w = sc->nil;
  return(p);
}

/* s7_reverse sometimes tacks extra nodes on the end of a reversed circular list (it detects the cycle too late)
 *  (let ((lst (list 0))) (set! (cdr lst) lst) (reverse lst)) -> (#1=(0 . #1#) 0 0 0)
 */


static s7_pointer reverse_in_place(s7_scheme *sc, s7_pointer term, s7_pointer list)
{
  s7_pointer p = list, result = term, q;

  while (is_not_null(p))
    {
      q = cdr(p);
      if ((!is_pair(q)) &&
	  (is_not_null(q)))
	return(sc->nil); /* improper list? */
      set_cdr(p, result);
      result = p;
      p = q;
    }
  return(result);
}


static s7_pointer reverse_in_place_unchecked(s7_scheme *sc, s7_pointer term, s7_pointer list)
{
  s7_pointer p = list, result = term, q;

  while (is_not_null(p))
    {
      q = cdr(p);
      set_cdr(p, result);
      result = p;
      p = q;

      if (is_null(p)) break;
      q = cdr(p);
      set_cdr(p, result);
      result = p;
      p = q;
    }
  return(result);
}


static s7_pointer safe_reverse_in_place(s7_scheme *sc, s7_pointer list) /* "safe" here means we guarantee this list is unproblematic */
{
  s7_pointer p = list, result, q;
  result = sc->nil;

  while (is_not_null(p))
    {
      q = cdr(p);
      /*   also if (is_null(list)) || (is_null(cdr(list))) return(list) */
      set_cdr(p, result);
      result = p;
      p = q;

      /* unroll the loop for speed */
      if (is_null(p)) break;
      q = cdr(p);
      set_cdr(p, result);
      result = p;
      p = q;

      if (is_null(p)) break;
      q = cdr(p);
      set_cdr(p, result);
      result = p;
      p = q;

      if (is_null(p)) break;
      q = cdr(p);
      set_cdr(p, result);
      result = p;
      p = q;
    }
  return(result);
}


s7_pointer s7_append(s7_scheme *sc, s7_pointer a, s7_pointer b)
{
  s7_pointer p, tp, np;
  if (is_null(a)) return(b);

  tp = cons(sc, car(a), sc->nil);
  sc->y = tp;
  for (p = cdr(a), np = tp; is_pair(p); p = cdr(p), np = cdr(np))
    set_cdr(np, cons(sc, car(p), sc->nil));
  set_cdr(np, b);
  sc->y = sc->nil;

  return(tp);
}


static s7_pointer copy_list(s7_scheme *sc, s7_pointer lst)
{
  s7_pointer p, tp, np;
  if (!is_pair(lst)) return(sc->nil);
  tp = cons(sc, car(lst), sc->nil);
  sc->y = tp;
  for (p = cdr(lst), np = tp; is_pair(p); p = cdr(p), np = cdr(np))
    set_cdr(np, cons(sc, car(p), sc->nil));
  sc->y = sc->nil;
  return(tp);
}


static s7_pointer copy_list_with_arglist_error(s7_scheme *sc, s7_pointer lst)
{
  s7_pointer p, tp, np;
  if (is_null(lst)) return(sc->nil);
  if (!is_pair(lst))
    s7_error(sc, sc->syntax_error_symbol, set_elist_2(sc, s7_make_string_wrapper(sc, "stray dot?: ~S"), lst));
  tp = cons(sc, car(lst), sc->nil);
  sc->y = tp;
  for (p = cdr(lst), np = tp; is_pair(p); p = cdr(p), np = cdr(np))
    set_cdr(np, cons(sc, car(p), sc->nil));
  sc->y = sc->nil;
  if (!is_null(p))
    s7_error(sc, sc->syntax_error_symbol, set_elist_2(sc, s7_make_string_wrapper(sc, "improper list of arguments: ~S"), lst));
  return(tp);
}


static s7_pointer revappend(s7_scheme *sc, s7_pointer a, s7_pointer b)
{
  /* (map (lambda (x) (if (odd? x) (apply values '(1 2 3)) (values))) (list 1 2 3 4))
   *   is a bad case -- we have to copy the incoming list.
   */
  s7_pointer p = b, q;

  if (is_not_null(a))
    {
      a = copy_list(sc, a);
      while (is_not_null(a))
	{
	  q = cdr(a);
	  set_cdr(a, p);
	  p = a;
	  a = q;
	}
    }
  return(p);
}

static int32_t safe_list_length(s7_scheme *sc, s7_pointer a)
{
  /* assume that "a" is a proper list */
  int32_t i = 0;
  s7_pointer b;
  for (b = a; is_pair(b); i++, b = cdr(b)) {};
  return(i);
}

int s7_list_length(s7_scheme *sc, s7_pointer a)
{
  /* returns -len if list is dotted, 0 if it's (directly) circular */
  int32_t i;
  s7_pointer slow, fast;

  slow = fast = a;
  for (i = 0; ; i += 2)
    {
      if (!is_pair(fast))
	{
	  if (is_null(fast))
	    return(i);
	  return(-i);
	}

      fast = cdr(fast);
      if (!is_pair(fast))
	{
	  if (is_null(fast))
	    return(i + 1);
	  return(-i - 1);
	}
      /* if unrolled further, it's a lot slower? */

      fast = cdr(fast);
      slow = cdr(slow);
      if (fast == slow)
	return(0);
    }
  return(0);
}


/* -------------------------------- null? pair? -------------------------------- */
static s7_pointer g_is_null(s7_scheme *sc, s7_pointer args)
{
  #define H_is_null "(null? obj) returns #t if obj is the empty list"
  #define Q_is_null pl_bt
  check_boolean_method(sc, is_null, sc->is_null_symbol, args);
  /* as a generic this could be: has_structure and length == 0 */
}

static s7_pointer g_is_pair(s7_scheme *sc, s7_pointer args)
{
  #define H_is_pair "(pair? obj) returns #t if obj is a pair (a non-empty list)"
  #define Q_is_pair pl_bt
  check_boolean_method(sc, is_pair, sc->is_pair_symbol, args);
}


/* -------------------------------- list? proper-list? -------------------------------- */
bool s7_is_list(s7_scheme *sc, s7_pointer p)
{
  return((is_pair(p)) ||
	 (is_null(p)));
}

static bool is_list_b(s7_pointer p) {return((is_pair(p)) || (type(p) == T_NIL));}


bool s7_is_proper_list(s7_scheme *sc, s7_pointer lst)
{
  /* #t if () or undotted/non-circular pair */
  s7_pointer slow, fast;

  fast = lst;
  slow = lst;
  while (true)
    {
      if (!is_pair(fast))
	return(is_null(fast)); /* else it's an improper list */

      fast = cdr(fast);
      if (!is_pair(fast)) return(is_null(fast));

      fast = cdr(fast);
      if (!is_pair(fast)) return(is_null(fast));

      fast = cdr(fast);
      if (!is_pair(fast)) return(is_null(fast));

      fast = cdr(fast);
      slow = cdr(slow);
      if (fast == slow) return(false);
    }
  return(true);
}


static s7_pointer g_is_list(s7_scheme *sc, s7_pointer args)
{
  #define H_is_list "(list? obj) returns #t if obj is a pair or null"
  #define Q_is_list pl_bt
  #define is_a_list(p) s7_is_list(sc, p)
  check_boolean_method(sc, is_a_list, sc->is_list_symbol, args);
}


/* -------------------------------- make-list -------------------------------- */
static s7_pointer make_list(s7_scheme *sc, int32_t len, s7_pointer init)
{
  switch (len)
    {
    case 0: return(sc->nil);
    case 1: return(cons(sc, init, sc->nil));
    case 2: return(cons_unchecked(sc, init, cons(sc, init, sc->nil)));
    case 3: return(cons_unchecked(sc, init, cons_unchecked(sc, init, cons(sc, init, sc->nil))));
    case 4: return(cons_unchecked(sc, init, cons_unchecked(sc, init, cons_unchecked(sc, init, cons(sc, init, sc->nil)))));
    case 5: return(cons_unchecked(sc, init, cons_unchecked(sc, init, cons_unchecked(sc, init,
		    cons_unchecked(sc, init, cons(sc, init, sc->nil))))));
    case 6: return(cons_unchecked(sc, init, cons_unchecked(sc, init, cons_unchecked(sc, init,
		    cons_unchecked(sc, init, cons_unchecked(sc, init, cons(sc, init, sc->nil)))))));
    case 7: return(cons_unchecked(sc, init, cons_unchecked(sc, init, cons_unchecked(sc, init,
		    cons_unchecked(sc, init, cons_unchecked(sc, init, cons_unchecked(sc, init, cons(sc, init, sc->nil))))))));
    default:
      {
	s7_pointer result;
	int32_t i;

	if (len >= (sc->free_heap_top - sc->free_heap))
	  {
	    gc(sc);
	    while (len >= (sc->free_heap_top - sc->free_heap))
	      resize_heap(sc);
	  }

	sc->v = sc->nil;
	for (i = 0; i < len; i++)
	  sc->v = cons_unchecked(sc, init, sc->v);
	result = sc->v;
	sc->v = sc->nil;
	return(result);
	}
    }
  return(sc->nil); /* never happens, I hope */
}


static s7_pointer g_make_list(s7_scheme *sc, s7_pointer args)
{
  #define H_make_list "(make-list length (initial-element #f)) returns a list of 'length' elements whose value is 'initial-element'."
  #define Q_make_list s7_make_signature(sc, 3, sc->is_proper_list_symbol, sc->is_integer_symbol, sc->T)

  s7_pointer init;
  s7_int len;

  if (!s7_is_integer(car(args)))
    method_or_bust(sc, car(args), sc->make_list_symbol, args, T_INTEGER, 1);

  len = s7_integer(car(args));            /* needs to be s7_int here so that (make-list most-negative-fixnum) is handled correctly */
  if (len < 0)
    return(out_of_range(sc, sc->make_list_symbol, small_int(1), car(args), its_negative_string));
  if (len == 0) return(sc->nil);          /* what about (make-list 0 123)? */
  if (len > sc->max_list_length)
    return(out_of_range(sc, sc->make_list_symbol, small_int(1), car(args), its_too_large_string));

  if (is_pair(cdr(args)))
    init = cadr(args);
  else init = sc->F;
  return(make_list(sc, (int32_t)len, init));
}


/* -------------------------------- list-ref -------------------------------- */

static s7_pointer list_ref_1(s7_scheme *sc, s7_pointer lst, s7_pointer ind)
{
  s7_int i, index;
  s7_pointer p;

  if (!s7_is_integer(ind))
    {
      if (!s7_is_integer(p = check_values(sc, ind, cons(sc, ind, sc->nil))))
	method_or_bust(sc, ind, sc->list_ref_symbol, list_2(sc, lst, ind), T_INTEGER, 2);
      ind = p;
    }
  index = s7_integer(ind);
  if ((index < 0) || (index > sc->max_list_length))
    return(out_of_range(sc, sc->list_ref_symbol, small_int(2), ind, (index < 0) ? its_negative_string : its_too_large_string));

  for (i = 0, p = lst; (i < index) && is_pair(p); i++, p = cdr(p)) {}

  if (!is_pair(p))
    {
      if (is_null(p))
	return(out_of_range(sc, sc->list_ref_symbol, small_int(2), ind, its_too_large_string));
      return(wrong_type_argument_with_type(sc, sc->list_ref_symbol, 1, lst, a_proper_list_string));
    }
  return(car(p));
}


static s7_pointer g_list_ref(s7_scheme *sc, s7_pointer args)
{
  #define H_list_ref "(list-ref lst i ...) returns the i-th element (0-based) of the list"
  #define Q_list_ref s7_make_circular_signature(sc, 2, 3, sc->T, sc->is_pair_symbol, sc->is_integer_symbol)

  /* (let ((L '((1 2 3) (4 5 6)))) (list-ref L 1 2))

    (define (lref L . args)
      (if (null? (cdr args))
          (list-ref L (car args))
          (apply lref (list-ref L (car args)) (cdr args))))
  */
  s7_pointer lst, inds;

  lst = car(args);
  if (!is_pair(lst))
    method_or_bust(sc, lst, sc->list_ref_symbol, args, T_PAIR, 1);

  inds = cdr(args);
  while (true)
    {
      lst = list_ref_1(sc, lst, car(inds));
      if (is_null(cdr(inds)))
	return(lst);
      inds = cdr(inds);
      if (!is_pair(lst)) /* trying to avoid a cons here at the cost of one extra type check */
	return(implicit_index(sc, lst, inds));
    }
}



/* -------------------------------- list-set! -------------------------------- */
static s7_pointer g_list_set_1(s7_scheme *sc, s7_pointer lst, s7_pointer args, int32_t arg_num)
{
  #define H_list_set "(list-set! lst i ... val) sets the i-th element (0-based) of the list to val"
  #define Q_list_set s7_make_circular_signature(sc, 3, 4, sc->T, sc->is_pair_symbol, sc->is_integer_symbol, sc->is_integer_or_any_at_end_symbol)

  int32_t i;
  s7_int index;
  s7_pointer p, ind;

  /* (let ((L '((1 2 3) (4 5 6)))) (list-set! L 1 2 32) L) */

  if (!is_pair(lst))
    method_or_bust(sc, lst, sc->list_set_symbol, cons(sc, lst, args), T_PAIR, 1);

  ind = car(args);
  if (!s7_is_integer(ind))
    {
      if (!s7_is_integer(p = check_values(sc, ind, args)))
	method_or_bust(sc, ind, sc->list_set_symbol, cons(sc, lst, args), T_INTEGER, arg_num);
      ind = p;
    }
  index = s7_integer(ind);
  if ((index < 0) || (index > sc->max_list_length))
    return(out_of_range(sc, sc->list_set_symbol, small_int(arg_num), ind, (index < 0) ? its_negative_string : its_too_large_string));

  for (i = 0, p = _TSet(lst); (i < index) && is_pair(p); i++, p = cdr(p)) {}

  if (!is_pair(p))
    {
      if (is_null(p))
	return(out_of_range(sc, sc->list_set_symbol, small_int(arg_num), ind, its_too_large_string));
      return(wrong_type_argument_with_type(sc, sc->list_set_symbol, 1, lst, a_proper_list_string));
    }
  if (is_null(cddr(args)))
    set_car(p, cadr(args));
  else return(g_list_set_1(sc, car(p), cdr(args), arg_num + 1));

  return(cadr(args));
}

static s7_pointer g_list_set(s7_scheme *sc, s7_pointer args)
{
  return(g_list_set_1(sc, car(args), cdr(args), 2));
}

static s7_pointer list_ref_p_pi_direct(s7_pointer p1, s7_int i1)
{
  s7_pointer p;
  s7_int i;
  if ((i1 < 0) || (i1 > cur_sc->max_list_length))
    out_of_range(cur_sc, cur_sc->list_ref_symbol, small_int(2), make_integer(cur_sc, i1), (i1 < 0) ? its_negative_string : its_too_large_string);
  for (i = 0, p = p1; ((is_pair(p)) && (i < i1)); i++, p = cdr(p));
  if (!is_pair(p))
    {
      if (type(p) == T_NIL)
	out_of_range(cur_sc, cur_sc->list_ref_symbol, small_int(2), make_integer(cur_sc, i1), its_too_large_string);
      else simple_wrong_type_argument_with_type(cur_sc, cur_sc->list_ref_symbol, p1, a_proper_list_string);
    }
  return(car(p));
}

static s7_pointer list_ref_p_pi(s7_pointer p1, s7_int i1)
{
  if (!is_pair(p1))
    simple_wrong_type_argument(cur_sc, cur_sc->list_ref_symbol, p1, T_PAIR);
  return(list_ref_p_pi_direct(p1, i1));
}

static s7_pointer list_set_p_pip_direct(s7_pointer p1, s7_int i1, s7_pointer p2)
{
  s7_pointer p;
  s7_int i;
  if ((i1 < 0) || (i1 > cur_sc->max_list_length))
    out_of_range(cur_sc, cur_sc->list_set_symbol, small_int(2), make_integer(cur_sc, i1), (i1 < 0) ? its_negative_string : its_too_large_string);
  for (i = 0, p = p1; ((is_pair(p)) && (i < i1)); i++, p = cdr(p));
  if (!is_pair(p))
    {
      if (type(p) == T_NIL)
	out_of_range(cur_sc, cur_sc->list_set_symbol, small_int(2), make_integer(cur_sc, i1), its_too_large_string);
      else simple_wrong_type_argument_with_type(cur_sc, cur_sc->list_set_symbol, p1, a_proper_list_string);
    }
  set_car(p, p2);
  return(p2);
}

static s7_pointer list_set_p_pip(s7_pointer p1, s7_int i1, s7_pointer p2)
{
  if (!is_pair(p1))
    simple_wrong_type_argument(cur_sc, cur_sc->list_set_symbol, p1, T_PAIR);
  return(list_set_p_pip_direct(p1, i1, p2));
}

static s7_pointer list_set_ic;
static s7_pointer g_list_set_ic(s7_scheme *sc, s7_pointer args)
{
  s7_pointer p, lst, val;
  s7_int i, index;
  lst = car(args);
  if (!is_pair(lst))
    method_or_bust(sc, lst, sc->list_set_symbol, args, T_PAIR, 1);

  index = s7_integer(cadr(args));
  if ((index < 0) || (index > sc->max_list_length))
    return(out_of_range(sc, sc->list_set_symbol, small_int(2), make_integer(sc, index), (index < 0) ? its_negative_string : its_too_large_string));

  for (i = 0, p = lst; (i < index) && is_pair(p); i++, p = cdr(p)) {}
  if (!is_pair(p))
    {
      if (is_null(p))
	return(out_of_range(sc, sc->list_set_symbol, small_int(2), make_integer(sc, index), its_too_large_string));
      return(wrong_type_argument_with_type(sc, sc->list_set_symbol, 1, lst, a_proper_list_string));
    }
  
  val = caddr(args);
  set_car(p, val);
  return(val);
}
 


/* -------------------------------- list-tail -------------------------------- */

static s7_pointer g_list_tail(s7_scheme *sc, s7_pointer args)
{
  #define H_list_tail "(list-tail lst i) returns the list from the i-th element on"
  #define Q_list_tail s7_make_signature(sc, 3, sc->is_list_symbol, sc->is_pair_symbol, sc->is_integer_symbol)
  s7_pointer lst, p;
  s7_int i, index;

  lst = car(args);
  p = cadr(args);
  if (!s7_is_integer(p))
    {
      s7_pointer p1;
      if (!s7_is_integer(p1 = check_values(sc, p, cdr(args))))
	method_or_bust(sc, p, sc->list_tail_symbol, args, T_INTEGER, 2);
      p = p1;
    }
  index = s7_integer(p);

  if (!s7_is_list(sc, lst))
    method_or_bust_with_type(sc, lst, sc->list_tail_symbol, list_2(sc, lst, make_integer(sc, index)), a_list_string, 1);

  if ((index < 0) || (index > sc->max_list_length))
    return(out_of_range(sc, sc->list_tail_symbol, small_int(2), make_integer(sc, index), (index < 0) ? its_negative_string : its_too_large_string));

  for (i = 0, p = lst; (i < index) && (is_pair(p)); i++, p = cdr(p)) {}
  if (i < index)
    return(out_of_range(sc, sc->list_tail_symbol, small_int(2), make_integer(sc, index), its_too_large_string));
  return(p);
}


/* -------------------------------- cons -------------------------------- */
static s7_pointer g_cons(s7_scheme *sc, s7_pointer args)
{
  /* n-ary cons could be the equivalent of CL's list*? */
  /*   it would be neater to have a single cons cell able to contain (directly) any number of elements */
  /*   (set! (cadr (cons 1 2 3)) 4) -> (1 4 . 3) */

  #define H_cons "(cons a b) returns a pair containing a and b"
  #define Q_cons s7_make_signature(sc, 3, sc->is_pair_symbol, sc->T, sc->T)

  /* set_cdr(args, cadr(args));
   * this is not safe -- it changes a variable's value directly:
   *   (let ((lst (list 1 2))) (list (apply cons lst) lst)) -> '((1 . 2) (1 . 2))
   */
  s7_pointer x;
  new_cell(sc, x, T_PAIR | T_SAFE_PROCEDURE);
  set_car(x, car(args));
  set_cdr(x, cadr(args));
  return(x);
}

static s7_pointer cons_p_pp(s7_pointer p1, s7_pointer p2)
{
  s7_pointer x;
  new_cell(cur_sc, x, T_PAIR | T_SAFE_PROCEDURE);
  set_car(x, p1);
  set_cdr(x, p2);
  return(x);
}


static void init_car_a_list(void)
{
  car_a_list_string = s7_make_permanent_string("a list whose car is also a list");
  cdr_a_list_string = s7_make_permanent_string("a list whose cdr is also a list");

  caar_a_list_string = s7_make_permanent_string("a list whose caar is also a list");
  cadr_a_list_string = s7_make_permanent_string("a list whose cadr is also a list");
  cdar_a_list_string = s7_make_permanent_string("a list whose cdar is also a list");
  cddr_a_list_string = s7_make_permanent_string("a list whose cddr is also a list");

  caaar_a_list_string = s7_make_permanent_string("a list whose caaar is also a list");
  caadr_a_list_string = s7_make_permanent_string("a list whose caadr is also a list");
  cadar_a_list_string = s7_make_permanent_string("a list whose cadar is also a list");
  caddr_a_list_string = s7_make_permanent_string("a list whose caddr is also a list");
  cdaar_a_list_string = s7_make_permanent_string("a list whose cdaar is also a list");
  cdadr_a_list_string = s7_make_permanent_string("a list whose cdadr is also a list");
  cddar_a_list_string = s7_make_permanent_string("a list whose cddar is also a list");
  cdddr_a_list_string = s7_make_permanent_string("a list whose cdddr is also a list");

  a_list_string =                 s7_make_permanent_string("a list");
  an_eq_func_string =             s7_make_permanent_string("a procedure that can take 2 arguments");
  an_association_list_string =    s7_make_permanent_string("an association list");
  a_normal_real_string =          s7_make_permanent_string("a normal real");
  a_rational_string =             s7_make_permanent_string("an integer or a ratio");
  a_number_string =               s7_make_permanent_string("a number");
  a_procedure_string =            s7_make_permanent_string("a procedure");
  a_normal_procedure_string =     s7_make_permanent_string("a normal procedure (not a continuation)");
  a_let_string =                  s7_make_permanent_string("a let (environment)");
  a_proper_list_string =          s7_make_permanent_string("a proper list");
  a_boolean_string =              s7_make_permanent_string("a boolean");
  an_input_port_string =          s7_make_permanent_string("an input port");
  an_open_port_string =           s7_make_permanent_string("an open port");
  an_output_port_string =         s7_make_permanent_string("an output port");
  an_input_string_port_string =   s7_make_permanent_string("an input string port");
  an_input_file_port_string =     s7_make_permanent_string("an input file port");
  an_output_string_port_string =  s7_make_permanent_string("an output string port");
  an_output_file_port_string =    s7_make_permanent_string("an output file port");
  a_thunk_string =                s7_make_permanent_string("a thunk");
  a_symbol_string =               s7_make_permanent_string("a symbol");
  a_non_negative_integer_string = s7_make_permanent_string("a non-negative integer");
  an_unsigned_byte_string =       s7_make_permanent_string("an unsigned byte");
  something_applicable_string =   s7_make_permanent_string("a procedure or something applicable");
  a_random_state_object_string =  s7_make_permanent_string("a random-state object");
  a_format_port_string =          s7_make_permanent_string("#f, #t, (), or an open output port");
  a_non_constant_symbol_string =  s7_make_permanent_string("a non-constant symbol");
  a_sequence_string =             s7_make_permanent_string("a sequence");
  a_valid_radix_string =          s7_make_permanent_string("should be between 2 and 16");
  result_is_too_large_string =    s7_make_permanent_string("result is too large");
  its_too_large_string =          s7_make_permanent_string("it is too large");
  its_too_small_string =          s7_make_permanent_string("it is less than the start position");
  its_negative_string =           s7_make_permanent_string("it is negative");
  its_nan_string =                s7_make_permanent_string("NaN usually indicates a numerical error");
  its_infinite_string =           s7_make_permanent_string("it is infinite");
  too_many_indices_string =       s7_make_permanent_string("too many indices");
  value_is_missing_string =       s7_make_permanent_string("~A argument '~A's value is missing");
  parameter_set_twice_string =    s7_make_permanent_string("parameter set twice, ~S in ~S");
#if (!HAVE_COMPLEX_NUMBERS)
  no_complex_numbers_string =     s7_make_permanent_string("this version of s7 does not support complex numbers");
#endif
}


/* -------- car -------- */

static s7_pointer g_car(s7_scheme *sc, s7_pointer args)
{
  #define H_car "(car pair) returns the first element of the pair"
  #define Q_car pl_p  

  s7_pointer lst;
  lst = car(args);
  if (is_pair(lst))
    return(car(lst));
  method_or_bust_one_arg(sc, lst, sc->car_symbol, args, T_PAIR);
}

static s7_pointer car_p_p(s7_pointer p)
{
  if (is_pair(p))
    return(car(p));
  return(simple_wrong_type_argument(cur_sc, cur_sc->car_symbol, p, T_PAIR));
}


static s7_pointer g_set_car(s7_scheme *sc, s7_pointer args)
{
  #define H_set_car "(set-car! pair val) sets the pair's first element to val"
  #define Q_set_car s7_make_signature(sc, 3, sc->T, sc->is_pair_symbol, sc->T)
  s7_pointer p;

  p = car(args);
  if (is_pair(p))
    {
      set_car(p, cadr(args));
      return(car(p));
    }
  method_or_bust(sc, p, sc->set_car_symbol, args, T_PAIR, 1);
}

static s7_pointer set_car_p_pp(s7_pointer p1, s7_pointer p2)
{
  if (!is_pair(p1))
    simple_wrong_type_argument(cur_sc, cur_sc->set_car_symbol, p1, T_PAIR);
  set_car(p1, p2);
  return(p2);
}


/* -------- cdr -------- */
static s7_pointer g_cdr(s7_scheme *sc, s7_pointer args)
{
  #define H_cdr "(cdr pair) returns the second element of the pair"
  #define Q_cdr pl_p  

  s7_pointer lst;
  lst = car(args);
  if (is_pair(lst))
    return(cdr(lst));
  method_or_bust_one_arg(sc, lst, sc->cdr_symbol, args, T_PAIR);
}

static s7_pointer cdr_p_p(s7_pointer p)
{
  if (is_pair(p))
    return(cdr(p));
  return(simple_wrong_type_argument(cur_sc, cur_sc->cdr_symbol, p, T_PAIR));
}


static s7_pointer g_set_cdr(s7_scheme *sc, s7_pointer args)
{
  #define H_set_cdr "(set-cdr! pair val) sets the pair's second element to val"
  #define Q_set_cdr s7_make_signature(sc, 3, sc->T, sc->is_pair_symbol, sc->T)
  s7_pointer p;

  p = car(args);
  if (!is_pair(p))
    method_or_bust(sc, p, sc->set_cdr_symbol, args, T_PAIR, 1);

  set_cdr(p, cadr(args));
  return(cdr(p));
}

static s7_pointer set_cdr_p_pp(s7_pointer p1, s7_pointer p2)
{
  if (!is_pair(p1))
    simple_wrong_type_argument(cur_sc, cur_sc->set_cdr_symbol, p1, T_PAIR);
  set_cdr(p1, p2);
  return(p2);
}


/* -------- caar --------*/
static s7_pointer g_caar(s7_scheme *sc, s7_pointer args)
{
  #define H_caar "(caar lst) returns (car (car lst)): (caar '((1 2))) -> 1"
  #define Q_caar pl_p  

  s7_pointer lst;
  lst = car(args);
  if (!is_pair(lst)) method_or_bust_one_arg(sc, lst, sc->caar_symbol, args, T_PAIR);
  if (!is_pair(car(lst))) return(simple_wrong_type_argument_with_type(sc, sc->caar_symbol, lst, car_a_list_string));
  /* it makes no difference in timing to move lst here or below (i.e. lst=car(lst) then return(car(lst)) and so on) */
  return(caar(lst));
}

static s7_pointer caar_p_p(s7_pointer p)
{
  if ((is_pair(p)) &&
      (is_pair(car(p))))
    return(caar(p));
  return(simple_wrong_type_argument(cur_sc, cur_sc->caar_symbol, p, T_PAIR));
}


/* -------- cadr --------*/
static s7_pointer g_cadr(s7_scheme *sc, s7_pointer args)
{
  #define H_cadr "(cadr lst) returns (car (cdr lst)): (cadr '(1 2 3)) -> 2"
  #define Q_cadr pl_p  

  s7_pointer lst;
  lst = car(args);
  if (!is_pair(lst)) method_or_bust_one_arg(sc, lst, sc->cadr_symbol, args, T_PAIR);
  if (!is_pair(cdr(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cadr_symbol, lst, cdr_a_list_string));
  return(cadr(lst));
}

static s7_pointer cadr_p_p(s7_pointer p)
{
  if ((is_pair(p)) &&
      (is_pair(cdr(p))))
    return(cadr(p));
  return(simple_wrong_type_argument(cur_sc, cur_sc->cadr_symbol, p, T_PAIR));
}


/* -------- cdar -------- */
static s7_pointer g_cdar(s7_scheme *sc, s7_pointer args)
{
  #define H_cdar "(cdar lst) returns (cdr (car lst)): (cdar '((1 2 3))) -> '(2 3)"
  #define Q_cdar pl_p  

  s7_pointer lst;
  lst = car(args);
  if (!is_pair(lst)) method_or_bust_one_arg(sc, lst, sc->cdar_symbol, args, T_PAIR);
  if (!is_pair(car(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cdar_symbol, lst, car_a_list_string));
  return(cdar(lst));
}

static s7_pointer cdar_p_p(s7_pointer p)
{
  if ((is_pair(p)) &&
      (is_pair(car(p))))
    return(cdar(p));
  return(simple_wrong_type_argument(cur_sc, cur_sc->cdar_symbol, p, T_PAIR));
}


/* -------- cddr -------- */
static s7_pointer g_cddr(s7_scheme *sc, s7_pointer args)
{
  #define H_cddr "(cddr lst) returns (cdr (cdr lst)): (cddr '(1 2 3 4)) -> '(3 4)"
  #define Q_cddr pl_p  

  s7_pointer lst;
  lst = car(args);
  if (!is_pair(lst)) method_or_bust_one_arg(sc, lst, sc->cddr_symbol, args, T_PAIR);
  if (!is_pair(cdr(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cddr_symbol, lst, cdr_a_list_string));
  return(cddr(lst));
}

static s7_pointer cddr_p_p(s7_pointer p)
{
  if ((is_pair(p)) &&
      (is_pair(cdr(p))))
    return(cddr(p));
  return(simple_wrong_type_argument(cur_sc, cur_sc->cddr_symbol, p, T_PAIR));
}


/* -------- caaar -------- */
static s7_pointer g_caaar(s7_scheme *sc, s7_pointer args)
{
  s7_pointer lst;
  #define H_caaar "(caaar lst) returns (car (car (car lst))): (caaar '(((1 2)))) -> 1"
  #define Q_caaar pl_p  
  
  lst = car(args);
  if (!is_pair(lst)) method_or_bust_one_arg(sc, lst, sc->caaar_symbol, set_plist_1(sc, lst), T_PAIR);
  if (!is_pair(car(lst))) return(simple_wrong_type_argument_with_type(sc, sc->caaar_symbol, lst, car_a_list_string));
  if (!is_pair(caar(lst))) return(simple_wrong_type_argument_with_type(sc, sc->caaar_symbol, lst, caar_a_list_string));
  return(caaar(lst));
}


/* -------- caadr -------- */
static s7_pointer g_caadr(s7_scheme *sc, s7_pointer args)
{
  s7_pointer lst;
  #define H_caadr "(caadr lst) returns (car (car (cdr lst))): (caadr '(1 (2 3))) -> 2"
  #define Q_caadr pl_p  
  
  lst = car(args);
  if (!is_pair(lst)) method_or_bust_one_arg(sc, lst, sc->caadr_symbol, set_plist_1(sc, lst), T_PAIR);
  if (!is_pair(cdr(lst))) return(simple_wrong_type_argument_with_type(sc, sc->caadr_symbol, lst, cdr_a_list_string));
  if (!is_pair(cadr(lst))) return(simple_wrong_type_argument_with_type(sc, sc->caadr_symbol, lst, cadr_a_list_string));
  return(caadr(lst));
}


/* -------- cadar -------- */
static s7_pointer g_cadar(s7_scheme *sc, s7_pointer args)
{
  s7_pointer lst;
  #define H_cadar "(cadar lst) returns (car (cdr (car lst))): (cadar '((1 2 3))) -> 2"
  #define Q_cadar pl_p  
  
  lst = car(args);
  if (!is_pair(lst)) method_or_bust_one_arg(sc, lst, sc->cadar_symbol, set_plist_1(sc, lst), T_PAIR);
  if (!is_pair(car(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cadar_symbol, lst, car_a_list_string));
  if (!is_pair(cdar(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cadar_symbol, lst, cdar_a_list_string));
  return(cadar(lst));
}


/* -------- cdaar -------- */
static s7_pointer g_cdaar(s7_scheme *sc, s7_pointer args)
{
  s7_pointer lst;
  #define H_cdaar "(cdaar lst) returns (cdr (car (car lst))): (cdaar '(((1 2 3)))) -> '(2 3)"
  #define Q_cdaar pl_p  
  
  lst = car(args);
  if (!is_pair(lst)) method_or_bust_one_arg(sc, lst, sc->cdaar_symbol, set_plist_1(sc, lst), T_PAIR);
  if (!is_pair(car(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cdaar_symbol, lst, car_a_list_string));
  if (!is_pair(caar(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cdaar_symbol, lst, caar_a_list_string));
  return(cdaar(lst));
}


/* -------- caddr -------- */
static s7_pointer g_caddr(s7_scheme *sc, s7_pointer args)
{
  s7_pointer lst;
  #define H_caddr "(caddr lst) returns (car (cdr (cdr lst))): (caddr '(1 2 3 4)) -> 3"
  #define Q_caddr pl_p  
  
  lst = car(args);
  if (!is_pair(lst)) method_or_bust_one_arg(sc, lst, sc->caddr_symbol, set_plist_1(sc, lst), T_PAIR);
  if (!is_pair(cdr(lst))) return(simple_wrong_type_argument_with_type(sc, sc->caddr_symbol, lst, cdr_a_list_string));
  if (!is_pair(cddr(lst))) return(simple_wrong_type_argument_with_type(sc, sc->caddr_symbol, lst, cddr_a_list_string));
  return(caddr(lst));
}


/* -------- cdddr -------- */
static s7_pointer g_cdddr(s7_scheme *sc, s7_pointer args)
{
  s7_pointer lst;
  #define H_cdddr "(cdddr lst) returns (cdr (cdr (cdr lst))): (cdddr '(1 2 3 4)) -> '(4)"
  #define Q_cdddr pl_p  
  
  lst = car(args);
  if (!is_pair(lst)) method_or_bust_one_arg(sc, lst, sc->cdddr_symbol, set_plist_1(sc, lst), T_PAIR);
  if (!is_pair(cdr(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cdddr_symbol, lst, cdr_a_list_string));
  if (!is_pair(cddr(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cdddr_symbol, lst, cddr_a_list_string));
  return(cdddr(lst));
}


/* -------- cdadr -------- */
static s7_pointer g_cdadr(s7_scheme *sc, s7_pointer args)
{
  s7_pointer lst;
  #define H_cdadr "(cdadr lst) returns (cdr (car (cdr lst))): (cdadr '(1 (2 3 4))) -> '(3 4)"
  #define Q_cdadr pl_p  
  
  lst = car(args);
  if (!is_pair(lst)) method_or_bust_one_arg(sc, lst, sc->cdadr_symbol, set_plist_1(sc, lst), T_PAIR);
  if (!is_pair(cdr(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cdadr_symbol, lst, cdr_a_list_string));
  if (!is_pair(cadr(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cdadr_symbol, lst, cadr_a_list_string));
  return(cdadr(lst));
}


/* -------- cddar -------- */
static s7_pointer g_cddar(s7_scheme *sc, s7_pointer args)
{
  s7_pointer lst;
  #define H_cddar "(cddar lst) returns (cdr (cdr (car lst))): (cddar '((1 2 3 4))) -> '(3 4)"
  #define Q_cddar pl_p  
  
  lst = car(args);
  if (!is_pair(lst)) method_or_bust_one_arg(sc, lst, sc->cddar_symbol, set_plist_1(sc, lst), T_PAIR);
  if (!is_pair(car(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cddar_symbol, lst, car_a_list_string));
  if (!is_pair(cdar(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cddar_symbol, lst, cdar_a_list_string));
  return(cddar(lst));
}


/* -------- caaaar -------- */
static s7_pointer g_caaaar(s7_scheme *sc, s7_pointer args)
{
  s7_pointer lst;
  #define H_caaaar "(caaaar lst) returns (car (car (car (car lst)))): (caaaar '((((1 2))))) -> 1"
  #define Q_caaaar pl_p  
  
  lst = car(args);
  if (!is_pair(lst)) method_or_bust_one_arg(sc, lst, sc->caaaar_symbol, set_plist_1(sc, lst), T_PAIR);
  if (!is_pair(car(lst))) return(simple_wrong_type_argument_with_type(sc, sc->caaaar_symbol, lst, car_a_list_string));
  if (!is_pair(caar(lst))) return(simple_wrong_type_argument_with_type(sc, sc->caaaar_symbol, lst, caar_a_list_string));
  if (!is_pair(caaar(lst))) return(simple_wrong_type_argument_with_type(sc, sc->caaaar_symbol, lst, caaar_a_list_string));
  return(caaaar(lst));
}


/* -------- caaadr -------- */
static s7_pointer g_caaadr(s7_scheme *sc, s7_pointer args)
{
  s7_pointer lst;
  #define H_caaadr "(caaadr lst) returns (car (car (car (cdr lst)))): (caaadr '(1 ((2 3)))) -> 2"
  #define Q_caaadr pl_p  
  
  lst = car(args);
  if (!is_pair(lst)) method_or_bust_one_arg(sc, lst, sc->caaadr_symbol, set_plist_1(sc, lst), T_PAIR);
  if (!is_pair(cdr(lst))) return(simple_wrong_type_argument_with_type(sc, sc->caaadr_symbol, lst, cdr_a_list_string));
  if (!is_pair(cadr(lst))) return(simple_wrong_type_argument_with_type(sc, sc->caaadr_symbol, lst, cadr_a_list_string));
  if (!is_pair(caadr(lst))) return(simple_wrong_type_argument_with_type(sc, sc->caaadr_symbol, lst, caadr_a_list_string));
  return(caaadr(lst));
}


/* -------- caadar -------- */
static s7_pointer g_caadar(s7_scheme *sc, s7_pointer args)
{
  s7_pointer lst;
  #define H_caadar "(caadar lst) returns (car (car (cdr (car lst)))): (caadar '((1 (2 3)))) -> 2"
  #define Q_caadar pl_p  
  
  lst = car(args);
  if (!is_pair(lst)) method_or_bust_one_arg(sc, lst, sc->caadar_symbol, set_plist_1(sc, lst), T_PAIR);
  if (!is_pair(car(lst))) return(simple_wrong_type_argument_with_type(sc, sc->caadar_symbol, lst, car_a_list_string));
  if (!is_pair(cdar(lst))) return(simple_wrong_type_argument_with_type(sc, sc->caadar_symbol, lst, cdar_a_list_string));
  if (!is_pair(cadar(lst))) return(simple_wrong_type_argument_with_type(sc, sc->caadar_symbol, lst, cadar_a_list_string));
  return(caadar(lst));
}


/* -------- cadaar -------- */
static s7_pointer g_cadaar(s7_scheme *sc, s7_pointer args)
{
  s7_pointer lst;
  #define H_cadaar "(cadaar lst) returns (car (cdr (car (car lst)))): (cadaar '(((1 2 3)))) -> 2"
  #define Q_cadaar pl_p  
  
  lst = car(args);
  if (!is_pair(lst)) method_or_bust_one_arg(sc, lst, sc->cadaar_symbol, set_plist_1(sc, lst), T_PAIR);
  if (!is_pair(car(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cadaar_symbol, lst, car_a_list_string));
  if (!is_pair(caar(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cadaar_symbol, lst, caar_a_list_string));
  if (!is_pair(cdaar(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cadaar_symbol, lst, cdaar_a_list_string));
  return(cadaar(lst));
}


/* -------- caaddr -------- */
static s7_pointer g_caaddr(s7_scheme *sc, s7_pointer args)
{
  s7_pointer lst;
  #define H_caaddr "(caaddr lst) returns (car (car (cdr (cdr lst)))): (caaddr '(1 2 (3 4))) -> 3"
  #define Q_caaddr pl_p  
  
  lst = car(args);
  if (!is_pair(lst)) method_or_bust_one_arg(sc, lst, sc->caaddr_symbol, set_plist_1(sc, lst), T_PAIR);
  if (!is_pair(cdr(lst))) return(simple_wrong_type_argument_with_type(sc, sc->caaddr_symbol, lst, cdr_a_list_string));
  if (!is_pair(cddr(lst))) return(simple_wrong_type_argument_with_type(sc, sc->caaddr_symbol, lst, cddr_a_list_string));
  if (!is_pair(caddr(lst))) return(simple_wrong_type_argument_with_type(sc, sc->caaddr_symbol, lst, caddr_a_list_string));
  return(caaddr(lst));
}


/* -------- cadddr -------- */
static s7_pointer g_cadddr(s7_scheme *sc, s7_pointer args)
{
  s7_pointer lst;
  #define H_cadddr "(cadddr lst) returns (car (cdr (cdr (cdr lst)))): (cadddr '(1 2 3 4 5)) -> 4"
  #define Q_cadddr pl_p  
  
  lst = car(args);
  if (!is_pair(lst)) method_or_bust_one_arg(sc, lst, sc->cadddr_symbol, set_plist_1(sc, lst), T_PAIR);
  if (!is_pair(cdr(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cadddr_symbol, lst, cdr_a_list_string));
  if (!is_pair(cddr(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cadddr_symbol, lst, cddr_a_list_string));
  if (!is_pair(cdddr(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cadddr_symbol, lst, cdddr_a_list_string));
  return(cadddr(lst));
}


/* -------- cadadr -------- */
static s7_pointer g_cadadr(s7_scheme *sc, s7_pointer args)
{
  s7_pointer lst;
  #define H_cadadr "(cadadr lst) returns (car (cdr (car (cdr lst)))): (cadadr '(1 (2 3 4))) -> 3"
  #define Q_cadadr pl_p  
  
  lst = car(args);
  if (!is_pair(lst)) method_or_bust_one_arg(sc, lst, sc->cadadr_symbol, set_plist_1(sc, lst), T_PAIR);
  if (!is_pair(cdr(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cadadr_symbol, lst, cdr_a_list_string));
  if (!is_pair(cadr(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cadadr_symbol, lst, cadr_a_list_string));
  if (!is_pair(cdadr(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cadadr_symbol, lst, cdadr_a_list_string));
  return(cadadr(lst));
}


/* -------- caddar -------- */
static s7_pointer g_caddar(s7_scheme *sc, s7_pointer args)
{
  s7_pointer lst;
  #define H_caddar "(caddar lst) returns (car (cdr (cdr (car lst)))): (caddar '((1 2 3 4))) -> 3"
  #define Q_caddar pl_p  
  
  lst = car(args);
  if (!is_pair(lst)) method_or_bust_one_arg(sc, lst, sc->caddar_symbol, set_plist_1(sc, lst), T_PAIR);
  if (!is_pair(car(lst))) return(simple_wrong_type_argument_with_type(sc, sc->caddar_symbol, lst, car_a_list_string));
  if (!is_pair(cdar(lst))) return(simple_wrong_type_argument_with_type(sc, sc->caddar_symbol, lst, cdar_a_list_string));
  if (!is_pair(cddar(lst))) return(simple_wrong_type_argument_with_type(sc, sc->caddar_symbol, lst, cddar_a_list_string));
  return(caddar(lst));
}


/* -------- cdaaar -------- */
static s7_pointer g_cdaaar(s7_scheme *sc, s7_pointer args)
{
  s7_pointer lst;
  #define H_cdaaar "(cdaaar lst) returns (cdr (car (car (car lst)))): (cdaaar '((((1 2 3))))) -> '(2 3)"
  #define Q_cdaaar pl_p  
  
  lst = car(args);
  if (!is_pair(lst)) method_or_bust_one_arg(sc, lst, sc->cdaaar_symbol, set_plist_1(sc, lst), T_PAIR);
  if (!is_pair(car(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cdaaar_symbol, lst, car_a_list_string));
  if (!is_pair(caar(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cdaaar_symbol, lst, caar_a_list_string));
  if (!is_pair(caaar(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cdaaar_symbol, lst, caaar_a_list_string));
  return(cdaaar(lst));
}


/* -------- cdaadr -------- */
static s7_pointer g_cdaadr(s7_scheme *sc, s7_pointer args)
{
  s7_pointer lst;
  #define H_cdaadr "(cdaadr lst) returns (cdr (car (car (cdr lst)))): (cdaadr '(1 ((2 3 4)))) -> '(3 4)"
  #define Q_cdaadr pl_p  
  
  lst = car(args);
  if (!is_pair(lst)) method_or_bust_one_arg(sc, lst, sc->cdaadr_symbol, set_plist_1(sc, lst), T_PAIR);
  if (!is_pair(cdr(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cdaadr_symbol, lst, cdr_a_list_string));
  if (!is_pair(cadr(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cdaadr_symbol, lst, cadr_a_list_string));
  if (!is_pair(caadr(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cdaadr_symbol, lst, caadr_a_list_string));
  return(cdaadr(lst));
}


/* -------- cdadar -------- */
static s7_pointer g_cdadar(s7_scheme *sc, s7_pointer args)
{
  s7_pointer lst;
  #define H_cdadar "(cdadar lst) returns (cdr (car (cdr (car lst)))): (cdadar '((1 (2 3 4)))) -> '(3 4)"
  #define Q_cdadar pl_p  
  
  lst = car(args);
  if (!is_pair(lst)) method_or_bust_one_arg(sc, lst, sc->cdadar_symbol, set_plist_1(sc, lst), T_PAIR);
  if (!is_pair(car(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cdadar_symbol, lst, car_a_list_string));
  if (!is_pair(cdar(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cdadar_symbol, lst, cdar_a_list_string));
  if (!is_pair(cadar(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cdadar_symbol, lst, cadar_a_list_string));
  return(cdadar(lst));
}


/* -------- cddaar -------- */
static s7_pointer g_cddaar(s7_scheme *sc, s7_pointer args)
{
  s7_pointer lst;
  #define H_cddaar "(cddaar lst) returns (cdr (cdr (car (car lst)))): (cddaar '(((1 2 3 4)))) -> '(3 4)"
  #define Q_cddaar pl_p  
  
  lst = car(args);
  if (!is_pair(lst)) method_or_bust_one_arg(sc, lst, sc->cddaar_symbol, set_plist_1(sc, lst), T_PAIR);
  if (!is_pair(car(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cddaar_symbol, lst, car_a_list_string));
  if (!is_pair(caar(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cddaar_symbol, lst, caar_a_list_string));
  if (!is_pair(cdaar(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cddaar_symbol, lst, cdaar_a_list_string));
  return(cddaar(lst));
}


/* -------- cdaddr -------- */
static s7_pointer g_cdaddr(s7_scheme *sc, s7_pointer args)
{
  s7_pointer lst;
  #define H_cdaddr "(cdaddr lst) returns (cdr (car (cdr (cdr lst)))): (cdaddr '(1 2 (3 4 5))) -> '(4 5)"
  #define Q_cdaddr pl_p  
  
  lst = car(args);
  if (!is_pair(lst)) method_or_bust_one_arg(sc, lst, sc->cdaddr_symbol, set_plist_1(sc, lst), T_PAIR);
  if (!is_pair(cdr(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cdaddr_symbol, lst, cdr_a_list_string));
  if (!is_pair(cddr(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cdaddr_symbol, lst, cddr_a_list_string));
  if (!is_pair(caddr(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cdaddr_symbol, lst, caddr_a_list_string));
  return(cdaddr(lst));
}


/* -------- cddddr -------- */
static s7_pointer g_cddddr(s7_scheme *sc, s7_pointer args)
{
  s7_pointer lst;
  #define H_cddddr "(cddddr lst) returns (cdr (cdr (cdr (cdr lst)))): (cddddr '(1 2 3 4 5)) -> '(5)"
  #define Q_cddddr pl_p  
  
  lst = car(args);
  if (!is_pair(lst)) method_or_bust_one_arg(sc, lst, sc->cddddr_symbol, set_plist_1(sc, lst), T_PAIR);
  if (!is_pair(cdr(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cddddr_symbol, lst, cdr_a_list_string));
  if (!is_pair(cddr(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cddddr_symbol, lst, cddr_a_list_string));
  if (!is_pair(cdddr(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cddddr_symbol, lst, cdddr_a_list_string));
  return(cddddr(lst));
}


/* -------- cddadr -------- */
static s7_pointer g_cddadr(s7_scheme *sc, s7_pointer args)
{
  s7_pointer lst;
  #define H_cddadr "(cddadr lst) returns (cdr (cdr (car (cdr lst)))): (cddadr '(1 (2 3 4 5))) -> '(4 5)"
  #define Q_cddadr pl_p  
  
  lst = car(args);
  if (!is_pair(lst)) method_or_bust_one_arg(sc, lst, sc->cddadr_symbol, set_plist_1(sc, lst), T_PAIR);
  if (!is_pair(cdr(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cddadr_symbol, lst, cdr_a_list_string));
  if (!is_pair(cadr(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cddadr_symbol, lst, cadr_a_list_string));
  if (!is_pair(cdadr(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cddadr_symbol, lst, cdadr_a_list_string));
  return(cddadr(lst));
}


/* -------- cdddar -------- */
static s7_pointer g_cdddar(s7_scheme *sc, s7_pointer args)
{
  s7_pointer lst;
  #define H_cdddar "(cdddar lst) returns (cdr (cdr (cdr (car lst)))): (cdddar '((1 2 3 4 5))) -> '(4 5)"
  #define Q_cdddar pl_p  
  
  lst = car(args);
  if (!is_pair(lst)) method_or_bust_one_arg(sc, lst, sc->cdddar_symbol, set_plist_1(sc, lst), T_PAIR);
  if (!is_pair(car(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cdddar_symbol, lst, car_a_list_string));
  if (!is_pair(cdar(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cdddar_symbol, lst, cdar_a_list_string));
  if (!is_pair(cddar(lst))) return(simple_wrong_type_argument_with_type(sc, sc->cdddar_symbol, lst, cddar_a_list_string));
  return(cdddar(lst));
}


s7_pointer s7_assq(s7_scheme *sc, s7_pointer obj, s7_pointer x)
{
  s7_pointer y;
  y = x;
  while (true)
    {
      /* we can blithely take the car of anything, since we're not treating it as an object,
       *   then if we get a bogus match, the following check that caar made sense ought to catch it.
       *
       * if car(#<unspecified>) = #<unspecified> (initialization time), then cdr(nil)->unspec
       *   and subsequent caar(unspc)->unspec so we could forgo half the is_pair checks below.
       *   This breaks if "x" is a dotted list -- the last cdr is not nil, so we lose.
       */
      if ((obj == unchecked_car(car(x))) && (is_pair(car(x)))) return(car(x));
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      if ((obj == unchecked_car(car(x))) && (is_pair(car(x)))) return(car(x));
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      if ((obj == unchecked_car(car(x))) && (is_pair(car(x)))) return(car(x));
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      if ((obj == unchecked_car(car(x))) && (is_pair(car(x)))) return(car(x));
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      if ((obj == unchecked_car(car(x))) && (is_pair(car(x)))) return(car(x));
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      if ((obj == unchecked_car(car(x))) && (is_pair(car(x)))) return(car(x));
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      y = cdr(y);
      if (x == y) return(sc->F);
    }
  return(sc->F); /* not reached */
}


static s7_pointer g_assq(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x, y;
  #define H_assq "(assq obj alist) returns the key-value pair associated (via eq?) with the key obj in the association list alist"
  #define Q_assq s7_make_signature(sc, 3, s7_make_signature(sc, 2, sc->is_pair_symbol, sc->is_boolean_symbol), sc->T, sc->is_list_symbol)

  x = car(args);
  y = cadr(args);
  if (is_pair(y))
    return(s7_assq(sc, x, y));
  if (is_null(y)) 
    return(sc->F);
  method_or_bust_with_type(sc, y, sc->assq_symbol, list_2(sc, x, y), an_association_list_string, 2);
  /* we don't check for (pair? (car x)) here (or in assv) so we get some inconsistency with assoc:
   *  (assq #f '(#f 2 . 3)) -> #f
   *  (assoc #f '(#f 2 . 3)) -> 'error 
   */
}


static s7_pointer g_assv(s7_scheme *sc, s7_pointer args)        /* g_assv is called by g_assoc below */
{ 
  s7_pointer x, y, z;
  #define H_assv "(assv obj alist) returns the key-value pair associated (via eqv?) with the key obj in the association list alist"
  #define Q_assv Q_assq

  x = car(args);
  y = cadr(args);
  if (!is_pair(y))
    {
      if (is_null(y)) return(sc->F);
      method_or_bust_with_type(sc, y, sc->assv_symbol, list_2(sc, x, y), an_association_list_string, 2);
    }

  if (is_simple(x))
    return(s7_assq(sc, x, y));

  z = y;
  while (true)
    {
      /* here we can't play the assq == game because s7_is_eqv thinks it's getting a legit s7 object */
      if ((is_pair(car(y))) && (s7_is_eqv(x, caar(y)))) return(car(y));
      y = cdr(y);
      if (!is_pair(y)) return(sc->F);

      if ((is_pair(car(y))) && (s7_is_eqv(x, caar(y)))) return(car(y));
      y = cdr(y);
      if (!is_pair(y)) return(sc->F);

      z = cdr(z);
      if (z == y) return(sc->F);
    }
  return(sc->F); /* not reached */
}


static s7_pointer all_x_c_ss(s7_scheme *sc, s7_pointer arg);
static s7_pointer all_x_c_c(s7_scheme *sc, s7_pointer arg);
static s7_pointer g_is_eq(s7_scheme *sc, s7_pointer args);
static s7_pointer g_is_eqv(s7_scheme *sc, s7_pointer args);
static s7_function s7_bool_optimize(s7_scheme *sc, s7_pointer expr);
static s7_pointer opt_bool_any(s7_scheme *sc, s7_pointer expr);

static s7_pointer g_assoc(s7_scheme *sc, s7_pointer args)
{
  #define H_assoc "(assoc obj alist (func #f)) returns the key-value pair associated (via equal?) with the key obj in the association list alist.\
If 'func' is a function of 2 arguments, it is used for the comparison instead of 'equal?"
  #define Q_assoc s7_make_signature(sc, 4, s7_make_signature(sc, 2, sc->is_pair_symbol, sc->is_boolean_symbol), sc->T, sc->is_list_symbol, sc->is_procedure_symbol)

  s7_pointer x, y, obj, eq_func = NULL;

  x = cadr(args);
  if (!is_null(x))
    {
      if (!is_pair(x))
	method_or_bust_with_type(sc, x, sc->assoc_symbol, args, an_association_list_string, 2);

      if ((is_pair(x)) && (!is_pair(car(x))))
	return(wrong_type_argument_with_type(sc, sc->assoc_symbol, 2, x, an_association_list_string)); /* we're assuming caar below so it better exist */
    }

  if (is_not_null(cddr(args)))
    {
      /* check third arg before second (trailing arg error check) */
      eq_func = caddr(args);

      if (type(eq_func) < T_GOTO)
	method_or_bust_with_type_one_arg(sc, eq_func, sc->assoc_symbol, args, a_procedure_string);

      if (!s7_is_aritable(sc, eq_func, 2))
	return(wrong_type_argument_with_type(sc, sc->assoc_symbol, 3, eq_func, an_eq_func_string));
    }
  if (is_null(x)) return(sc->F);

  if (eq_func)
    {
      /* now maybe there's a simple case */
      if (s7_list_length(sc, x) > 0)
	{
	  if ((is_safe_procedure(eq_func)) &&
	      (is_c_function(eq_func)))
	    {
	      s7_function func;

	      func = c_function_call(eq_func);
	      if (func == g_is_eq) return(s7_assq(sc, car(args), x));
	      if (func == g_is_eqv) return(g_assv(sc, args));
	      set_car(sc->t2_1, car(args));

	      for (; is_pair(x); x = cdr(x))
		{
		  if (is_pair(car(x)))
		    {
		      set_car(sc->t2_2, caar(x));
		      if (is_true(sc, func(sc, sc->t2_1)))
			return(car(x));
		      /* I wonder if the assoc equality function should get the cons, not just caar?
		       */
		    }
		  else return(wrong_type_argument_with_type(sc, sc->assoc_symbol, 2, cadr(args), an_association_list_string));
		}
	      return(sc->F);
	    }

	  if ((is_closure(eq_func)) &&
	      (is_pair(closure_args(eq_func))) &&
	      (is_pair(cdr(closure_args(eq_func))))) /* not dotted arg list */
	    {
	      s7_pointer body;
	      body = closure_body(eq_func);
	      if (is_null(cdr(body)))
		{
		  s7_function func;

		  new_frame_with_two_slots(sc, sc->envir, sc->envir, car(closure_args(eq_func)), car(args), cadr(closure_args(eq_func)), sc->F);
		  func = s7_bool_optimize(sc, body);
		  if (func)
		    {
		      s7_pointer b;
		      b = next_slot(let_slots(sc->envir));

		      if (func == opt_bool_any)
			{
			  opt_info *o;
			  o = sc->opts[0];
			  for (; is_pair(x); x = cdr(x))
			    {
			      slot_set_value(b, caar(x));
			      sc->pc = 0;
			      if (o->v7.fb(o))
				return(car(x));
			    }
			}
		      else
			{
			  for (; is_pair(x); x = cdr(x))
			    {
			      slot_set_value(b, caar(x));
			      if (is_true(sc, func(sc, car(body))))
				return(car(x));
			    }
			}
		      return(sc->F);
		    }
		}
	    }
	}

      /* sc->value = sc->F; */
      y = cons(sc, args, sc->nil);
      set_opt_fast(y, x);
      set_opt_slow(y, x);
      push_stack(sc, OP_ASSOC_IF, y, eq_func);
      push_stack(sc, OP_APPLY, list_2(sc, car(args), caar(x)), eq_func);
      return(sc->unspecified);
    }

  x = cadr(args);
  obj = car(args);
  if (is_simple(obj))
    return(s7_assq(sc, obj, x));

  y = x;
  if (is_string(obj))
    {
      s7_pointer val;
      while (true)
	{
	  if (is_pair(car(x)))
	    {
	      val = caar(x);
	      if ((val == obj) ||
		  ((is_string(val)) &&
		   (scheme_strings_are_equal(obj, val))))
		return(car(x));
	    }
	  x = cdr(x);
	  if (!is_pair(x)) return(sc->F);

	  if (is_pair(car(x)))
	    {
	      val = caar(x);
	      if ((val == obj) ||
		  ((is_string(val)) &&
		   (scheme_strings_are_equal(obj, val))))
		return(car(x));
	    }
	  x = cdr(x);
	  if (!is_pair(x)) return(sc->F);

	  y = cdr(y);
	  if (x == y) return(sc->F);
	}
      return(sc->F);
    }

  while (true)
    {
      if ((is_pair(car(x))) && (s7_is_equal(sc, obj, caar(x)))) return(car(x));
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      if ((is_pair(car(x))) && (s7_is_equal(sc, obj, caar(x)))) return(car(x));
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      y = cdr(y);
      if (x == y) return(sc->F);
    }
  return(sc->F); /* not reached */
}

static s7_pointer assoc_p_pp(s7_pointer p1, s7_pointer p2) {return(g_assoc(cur_sc, set_plist_2(cur_sc, p1, p2)));}



/* ---------------- member, memv, memq ---------------- */

s7_pointer s7_memq(s7_scheme *sc, s7_pointer obj, s7_pointer x)
{
  s7_pointer y;
  y = x;
  while (true)
    {
      if (obj == car(x)) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      if (obj == car(x)) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      if (obj == car(x)) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      if (obj == car(x)) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      y = cdr(y);
      if (x == y) return(sc->F);
    }
  return(sc->F);
}

static s7_pointer g_memq(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x, y;
  #define H_memq "(memq obj list) looks for obj in list and returns the list from that point if it is found, otherwise #f. memq uses eq?"
  #define Q_memq pl_tl

  x = car(args);
  y = cadr(args);
  if (is_pair(y))
    return(s7_memq(sc, x, y));
  if (is_null(y)) 
    return(sc->F);
  method_or_bust_with_type(sc, y, sc->memq_symbol, list_2(sc, x, y), a_list_string, 2);
}

/* I think (memq 'c '(a b . c)) should return #f because otherwise
 *   (memq () ...) would return the () at the end.
 */

/* if memq's list is a quoted list, it won't be changing, so we can tell ahead of time that it is
 *   a proper list, and what its length is.
 */
static s7_pointer memq_2, memq_3, memq_4, memq_any;

static s7_pointer g_memq_2(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x, obj;
  x = cadr(args);
  obj = car(args);
  if (obj == car(x)) return(x);
  if (obj == cadr(x)) return(cdr(x));
  return(sc->F);
}

static s7_pointer g_memq_3(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x, obj;
  x = cadr(args);
  obj = car(args);
  while (true)
    {
      if (obj == car(x)) return(x);
      x = cdr(x);
      if (obj == car(x)) return(x);
      x = cdr(x);
      if (obj == car(x)) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);
    }
  return(sc->F);
}

static s7_pointer g_memq_4(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x, obj;
  x = cadr(args);
  obj = car(args);
  while (true)
    {
      if (obj == car(x)) return(x);
      x = cdr(x);
      if (obj == car(x)) return(x);
      x = cdr(x);
      if (obj == car(x)) return(x);
      x = cdr(x);
      if (obj == car(x)) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);
    }
  return(sc->F);
}

static s7_pointer g_memq_any(s7_scheme *sc, s7_pointer args)
{
  /* no circular list check needed in this case */
  s7_pointer x, obj;
  x = cadr(args);
  obj = car(args);
  while (true)
    {
      if (obj == car(x)) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F); /* every other pair check could be omitted */

      if (obj == car(x)) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      if (obj == car(x)) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      if (obj == car(x)) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);
    }
  return(sc->F);
}


static s7_pointer memq_car, memq_car_2;
static s7_pointer g_memq_car(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x, obj;

  obj = find_symbol_unchecked(sc, opt_sym2(args));
  if (is_pair(obj))
    obj = car(obj);
  else obj = g_car(sc, set_plist_1(sc, obj));
  x = opt_pair3(args);

  while (true)
    {
      if (obj == car(x)) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      if (obj == car(x)) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);
    }
  return(sc->F);
}

static s7_pointer g_memq_car_2(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x, obj;

  obj = find_symbol_unchecked(sc, opt_sym2(args));
  if (is_pair(obj))
    obj = car(obj);
  else obj = g_car(sc, set_plist_1(sc, obj));
  x = opt_pair3(args);
  if (obj == car(x)) return(x);
  if (obj == cadr(x)) return(cdr(x));
  return(sc->F);
}

static s7_pointer memq_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
  if ((is_proper_quote(sc, caddr(expr))) &&
      (is_pair(cadr(caddr(expr)))))
    {
      int32_t len;
      len = s7_list_length(sc, cadr(caddr(expr)));

      if ((ops) && (is_h_safe_c_s(cadr(expr))) &&
	  (c_callee(cadr(expr)) == g_car))
	{
	  set_optimize_op(expr, HOP_SAFE_C_C);
	  set_opt_sym2(cdr(expr), cadadr(expr));
	  set_opt_pair3(cdr(expr), cadr(caddr(expr)));
	  if (len == 2)
	    return(memq_car_2);
	  return(memq_car);
	}

      if (len > 0)
	{
	  if (len == 2)
	    return(memq_2);
	  if ((len % 4) == 0)
	    return(memq_4);
	  if ((len % 3) == 0)
	    return(memq_3);
	  return(memq_any);
	}
    }
  return(f);
}


static s7_pointer memv_number(s7_scheme *sc, s7_pointer obj, s7_pointer x)
{
  s7_pointer y;
  y = x;
  while (true)
    {
      if ((s7_is_number(car(x))) && (numbers_are_eqv(obj, car(x)))) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);
      if ((s7_is_number(car(x))) && (numbers_are_eqv(obj, car(x)))) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);
      if ((s7_is_number(car(x))) && (numbers_are_eqv(obj, car(x)))) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);
      if ((s7_is_number(car(x))) && (numbers_are_eqv(obj, car(x)))) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);
      y = cdr(y);
      if (x == y) return(sc->F);
    }
  return(sc->F);
}


static s7_pointer g_memv(s7_scheme *sc, s7_pointer args)
{
  #define H_memv "(memv obj list) looks for obj in list and returns the list from that point if it is found, otherwise #f. memv uses eqv?"
  #define Q_memv pl_tl
  s7_pointer x, y, z;

  x = car(args);
  y = cadr(args);
  if (!is_pair(y))
    {
      if (is_null(y)) return(sc->F);
      method_or_bust_with_type(sc, y, sc->memv_symbol, list_2(sc, x, y), a_list_string, 2);
    }

  if (is_simple(x)) return(s7_memq(sc, x, y));
  if (s7_is_number(x)) return(memv_number(sc, x, y));

  z = y;
  while (true)
    {
      if (s7_is_eqv(x, car(y))) return(y);
      y = cdr(y);
      if (!is_pair(y)) return(sc->F);

      if (s7_is_eqv(x, car(y))) return(y);
      y = cdr(y);
      if (!is_pair(y)) return(sc->F);

      z = cdr(z);
      if (z == y) return(sc->F);
    }
  return(sc->F); /* not reached */
}


static s7_pointer member(s7_scheme *sc, s7_pointer obj, s7_pointer x)
{
  s7_pointer y;

  y = x;
  if (is_string(obj))
    {
      while (true)
	{
	  if ((obj == car(x)) ||
	      ((is_string(car(x))) &&
	       (scheme_strings_are_equal(obj, car(x)))))
	    return(x);
	  x = cdr(x);
	  if (!is_pair(x)) return(sc->F);

	  if ((obj == car(x)) ||
	      ((is_string(car(x))) &&
	       (scheme_strings_are_equal(obj, car(x)))))
	    return(x);
	  x = cdr(x);
	  if (!is_pair(x)) return(sc->F);

	  y = cdr(y);
	  if (x == y) return(sc->F);
	}
      return(sc->F);
    }

  while (true)
    {
      if (s7_is_equal(sc, obj, car(x))) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      if (s7_is_equal(sc, obj, car(x))) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      if (s7_is_equal(sc, obj, car(x))) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      if (s7_is_equal(sc, obj, car(x))) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      y = cdr(y);
      if (x == y) return(sc->F);
    }
  return(sc->F); /* not reached */
}


static s7_pointer g_member(s7_scheme *sc, s7_pointer args)
{
  #define H_member "(member obj list (func #f)) looks for obj in list and returns the list from that point if it is found, otherwise #f. \
member uses equal?  If 'func' is a function of 2 arguments, it is used for the comparison instead of 'equal?"
  #define Q_member s7_make_signature(sc, 4, s7_make_signature(sc, 2, sc->is_pair_symbol, sc->is_boolean_symbol), sc->T, sc->is_list_symbol, sc->is_procedure_symbol)

  /* this could be extended to accept sequences:
   *    (member #\a "123123abnfc" char=?) -> "abnfc"
   *    (member "abc" "123abc321" string=?) -> "abc321" but there's the string length complication
   *    (member 1 #(0 1 2) =) -> #(1 2) etc but what would it do for a hash-table?
   * the third arg can be weird: (member #f (list #t) cons) -> (#t) -- cons returns '(#f . #t) which is true, so we get '(#t)
   * should this be an error: (member '(1 2 3) () '(1 . 2)) -- the third arg is bogus, but the second is nil
   *
   * here as in assoc, sort, and make-hash-table we accept macros, but I can't think of a good reason to do so.
   */

  s7_pointer x, y, obj, eq_func = NULL;
  x = cadr(args);

  if ((!is_pair(x)) && (!is_null(x)))
    method_or_bust_with_type(sc, x, sc->member_symbol, args, a_list_string, 2);

  if (is_not_null(cddr(args)))
    {
      /* check third arg before second (trailing arg error check) */
      eq_func = caddr(args);

      if (type(eq_func) < T_GOTO)
	method_or_bust_with_type(sc, eq_func, sc->member_symbol, args, a_procedure_string, 3);

      if (!s7_is_aritable(sc, eq_func, 2))
	return(wrong_type_argument_with_type(sc, sc->member_symbol, 3, eq_func, an_eq_func_string));
    }

  if (is_null(x)) return(sc->F);
  if (eq_func)
    {
      s7_pointer slow;
	  
      /* now maybe there's a simple case */
      if ((is_safe_procedure(eq_func)) &&
	  (is_c_function(eq_func)))
	{
	  s7_function func;
	  func = c_function_call(eq_func);
	  if (func == g_is_eq) return(s7_memq(sc, car(args), x));
	  if (func == g_is_eqv) return(g_memv(sc, args));
#if (!WITH_GMP)
	  if (func == g_less) func = g_less_2;
	  if (func == g_greater) func = g_greater_2;
#endif
	  set_car(sc->t2_1, car(args));
	  
	  for (slow = x; is_pair(x); x = cdr(x), slow = cdr(slow))
	    {
	      set_car(sc->t2_2, car(x));
	      if (is_true(sc, func(sc, sc->t2_1)))
		return(x);
	      
	      if (!is_pair(cdr(x)))
		return(sc->F);
	      x = cdr(x);
	      if (x == slow)
		return(sc->F);

	      set_car(sc->t2_2, car(x));
	      if (is_true(sc, func(sc, sc->t2_1)))
		return(x);
	    }
	  return(sc->F);
	}

      if ((is_closure(eq_func)) &&
	  (is_pair(closure_args(eq_func))) &&
	  (is_pair(cdr(closure_args(eq_func))))) /* not dotted arg list */
	{
	  s7_pointer body;
	  body = closure_body(eq_func);
	  if (is_null(cdr(body)))
	    {
	      s7_function func;
	      
	      new_frame_with_two_slots(sc, sc->envir, sc->envir, car(closure_args(eq_func)), car(args), cadr(closure_args(eq_func)), sc->F);
	      func = s7_bool_optimize(sc, body);

	      if (func)
		{
		  s7_pointer b;
		  b = next_slot(let_slots(sc->envir));
		  
		  if (func == opt_bool_any)
		    {
		      opt_info *o;
		      o = sc->opts[0];
		      for (slow = x; is_pair(x); x = cdr(x), slow = cdr(slow))
			{
			  slot_set_value(b, car(x));
			  sc->pc = 0;
			  if (o->v7.fb(o)) return(x);

			  if (!is_pair(cdr(x))) return(sc->F);
			  x = cdr(x);
			  if (x == slow) return(sc->F);

			  slot_set_value(b, car(x));
			  sc->pc = 0;
			  if (o->v7.fb(o)) return(x);
			}
		    }
		  else
		    {
		      for (slow = x; is_pair(x); x = cdr(x), slow = cdr(slow))
			{
			  slot_set_value(b, car(x));
			  if (is_true(sc, func(sc, car(body)))) return(x);
			  
			  if (!is_pair(cdr(x))) return(sc->F);
			  x = cdr(x);
			  if (x == slow) return(sc->F);
			  
			  slot_set_value(b, car(x));
			  if (is_true(sc, func(sc, sc->t2_1))) return(x);
			}
		    }
		  return(sc->F);
		}
	    }
	}

      y = cons(sc, args, sc->nil); /* this could probably be handled with a counter cell (cdr here is unused) */
      set_opt_fast(y, x);
      set_opt_slow(y, x);
      push_stack(sc, OP_MEMBER_IF, y, eq_func);
      if (needs_copied_args(eq_func))
	push_stack(sc, OP_APPLY, list_2(sc, car(args), car(x)), eq_func);
      else
	{
	  set_car(sc->t2_1, car(args));
	  set_car(sc->t2_2, car(x));
	  push_stack(sc, OP_APPLY, sc->t2_1, eq_func);
	}
      return(sc->unspecified);
    }

  obj = car(args);
  if (is_simple(obj))
    return(s7_memq(sc, obj, x));

  /* the only things that aren't simply == here are c_object, string, number, vector, hash-table, pair, and c_pointer
   *   but all the other cases are unlikely.
   */
  if (s7_is_number(obj))
    return(memv_number(sc, obj, x));

  return(member(sc, obj, x));
}

static s7_pointer member_sq;
static s7_pointer g_member_sq(s7_scheme *sc, s7_pointer args)
{
  s7_pointer obj, lst;
  lst = cadadr(args);
  obj = find_symbol_unchecked(sc, car(args));

  if (is_simple(obj))
    return(s7_memq(sc, obj, lst));

  if (s7_is_number(obj))
    return(memv_number(sc, obj, lst));

  return(member(sc, obj, lst));
}

static s7_pointer member_ss;
static s7_pointer g_member_ss(s7_scheme *sc, s7_pointer args)
{
  s7_pointer obj, x;

  obj = find_symbol_unchecked(sc, car(args));
  x = find_symbol_unchecked(sc, cadr(args));
  if (!is_pair(x))
    {
      if (is_null(x)) return(sc->F);
      method_or_bust_with_type(sc, x, sc->member_symbol, list_2(sc, obj, x), a_list_string, 2);
    }

  if (is_simple(obj))
    return(s7_memq(sc, obj, x));

  if (s7_is_number(obj))
    return(memv_number(sc, obj, x));

  return(member(sc, obj, x));
}

static s7_pointer member_p_pp(s7_pointer p1, s7_pointer p2) {return(g_member(cur_sc, set_plist_2(cur_sc, p1, p2)));}

static s7_pointer member_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
  if (!ops) return(f);
  if (args == 2)
    {
      if (is_symbol(caddr(expr)))
	{
	  if ((optimize_op(expr) == HOP_SAFE_C_SS) || (optimize_op(expr) == HOP_SAFE_C_LL) ||
	      ((is_h_safe_c_c(expr)) &&
	       (is_symbol(cadr(expr)))))
	    {
	      set_optimize_op(expr, HOP_SAFE_C_C);
	      return(member_ss);                    /* (member obj lst) */
	    }
	}
      else
	{
	  if ((optimize_op(expr) == HOP_SAFE_C_SQ) ||
	      ((is_h_safe_c_c(expr)) &&
	       (is_symbol(cadr(expr))) &&
	       (is_proper_quote(sc, caddr(expr))) &&
	       (is_pair(cadr(caddr(expr))))))
	    {
	      set_optimize_op(expr, HOP_SAFE_C_C);
	      return(member_sq);                    /* (member q '(quote lambda case)) */
	    }
	}
    }

  if ((args == 3) &&
      (is_symbol(cadddr(expr))) &&
      (cadddr(expr) == sc->is_eq_symbol))
    return(memq_chooser(sc, f, 2, expr, ops));

  return(f);
}


static bool is_memq(s7_pointer sym, s7_pointer lst)
{
  s7_pointer x;
  for (x = lst; is_pair(x); x = cdr(x))
    if (sym == car(x))
      return(true);
  return(false);
}


static s7_pointer g_is_provided(s7_scheme *sc, s7_pointer args)
{
  #define H_is_provided "(provided? symbol) returns #t if symbol is a member of the *features* list"
  #define Q_is_provided s7_make_signature(sc, 2, sc->is_boolean_symbol, sc->is_symbol_symbol)
  s7_pointer sym, topf, x;

  sym = car(args);
  if (!is_symbol(sym))
    method_or_bust_one_arg(sc, sym, sc->is_provided_symbol, list_1(sc, sym), T_SYMBOL);

  /* here the *features* list is spread out (or can be anyway) along the curlet chain,
   *   so we need to travel back all the way to the top level checking each *features* list in turn.
   *   Since *features* grows via cons (newest first), we can stop the scan if we hit the shared
   *   top-level at least.
   */
  topf = slot_value(global_slot(sc->features_symbol));
  if (is_memq(sym, topf))
    return(sc->T);

  if (is_global(sc->features_symbol))
    return(sc->F);
  for (x = sc->envir; symbol_id(sc->features_symbol) < let_id(x); x = outlet(x));
  for (; is_let(x); x = outlet(x))
    {
      s7_pointer y;
      for (y = let_slots(x); is_slot(y); y = next_slot(y))
	if (slot_symbol(y) == sc->features_symbol)
	  {
	    if ((slot_value(y) != topf) &&
		(is_memq(sym, slot_value(y))))
	      return(sc->T);
	  }
    }
  return(sc->F);
}


bool s7_is_provided(s7_scheme *sc, const char *feature)
{
  return(is_memq(s7_make_symbol(sc, feature), s7_symbol_value(sc, sc->features_symbol))); /* this goes from local outward */
}

bool is_provided_b(s7_pointer sym) 
{
  if (!is_symbol(sym))
    simple_wrong_type_argument(cur_sc, cur_sc->is_provided_symbol, sym, T_SYMBOL);
  return(is_memq(sym, s7_symbol_value(cur_sc, cur_sc->features_symbol)));
}


static s7_pointer c_provide(s7_scheme *sc, s7_pointer sym)
{
  /* this has to be relative to the curlet: (load file env)
   *   the things loaded are only present in env, and go away with it, so should not be in the global *features* list
   */
  s7_pointer p, lst;
  if (!is_symbol(sym))
    method_or_bust_one_arg(sc, sym, sc->provide_symbol, list_1(sc, sym), T_SYMBOL);

  p = find_local_symbol(sc, sc->features_symbol, sc->envir); /* if sc->envir is nil, this returns the global slot, else local slot */
  lst = slot_value(find_symbol(sc, sc->features_symbol));    /* in either case, we want the current *features* list */

  if (p == sc->undefined)
    make_slot_1(sc, sc->envir, sc->features_symbol, cons(sc, sym, lst));
  else
    {
      if (!is_memq(sym, lst))
	slot_set_value(p, cons(sc, sym, lst));
    }

  /* require looks up its symbol argument to see if the associated code (dsp.scm etc) has been autoloaded,
   *   so here we're defining the provided symbol for that possibility.  Perhaps better would be to forgo
   *   the definition here, and in require check *features* -- it is local to the current env.
   */
  if (!is_slot(find_symbol(sc, sym))) /* *features* name might be the same as an existing function */
    s7_define(sc, sc->envir, sym, sym);

  return(sym);
}

static s7_pointer g_provide(s7_scheme *sc, s7_pointer args)
{
  #define H_provide "(provide symbol) adds symbol to the *features* list"
  #define Q_provide s7_make_signature(sc, 2, sc->is_symbol_symbol, sc->is_symbol_symbol)
  return(c_provide(sc, car(args)));
}

void s7_provide(s7_scheme *sc, const char *feature) {c_provide(sc, s7_make_symbol(sc, feature));}


static s7_pointer g_features_set(s7_scheme *sc, s7_pointer args)
{
  /* symbol_access for set/let of *features* which can only be changed via provide */
  if (s7_is_list(sc, cadr(args)))
    return(cadr(args));
  return(sc->error_symbol);
}

static s7_pointer g_list(s7_scheme *sc, s7_pointer args)
{
  #define H_list "(list ...) returns its arguments in a list"
  #define Q_list s7_make_circular_signature(sc, 1, 2, sc->is_proper_list_symbol, sc->T)
  
  if (is_pair(args))
    {
      if (is_null(cdr(args)))
	return(cons(sc, car(args), sc->nil));
      if (is_null(cddr(args)))
	return(list_2(sc, car(args), cadr(args)));

      return(copy_list(sc, args));
    }
  return(sc->nil);
}

static void check_list_validity(s7_scheme *sc, const char *caller, s7_pointer lst)
{
  s7_pointer p;
  int32_t i;
  for (i = 1, p = lst; is_pair(p); p = cdr(p), i++)
    if (!s7_is_valid(sc, car(p)))
      fprintf(stderr, "bad arg (#%d) to %s: %p\n", i, caller, car(p));
}

s7_pointer s7_list(s7_scheme *sc, int32_t num_values, ...)
{
  int32_t i;
  va_list ap;
  s7_pointer p;

  if (num_values == 0)
    return(sc->nil);

  sc->w = sc->nil;
  va_start(ap, num_values);
  for (i = 0; i < num_values; i++)
    sc->w = cons(sc, va_arg(ap, s7_pointer), sc->w);
  va_end(ap);

  if (sc->safety > NO_SAFETY)
    check_list_validity(sc, "s7_list", sc->w);

  p = sc->w;
  sc->w = sc->nil;
  return(safe_reverse_in_place(sc, p));
}

static s7_int sequence_length(s7_scheme *sc, s7_pointer lst);

static s7_pointer g_list_append(s7_scheme *sc, s7_pointer args)
{
  s7_pointer y, tp, np = NULL, pp;
  bool args_are_lists = true;

  /* we know here that args is a pair and cdr(args) is a pair */
  tp = sc->nil;
  for (y = args; is_pair(y); y = cdr(y)) /* arglist so not dotted */
    {
      s7_pointer p;
      p = car(y);

      check_method(sc, p, sc->append_symbol, (is_null(tp)) ? args : cons(sc, tp, y));

      if (is_null(cdr(y)))
	{
	  if (is_null(tp))
	    return(p);
	  /* (append (list 1) "hi") should return '(1 . "hi") not '(1 #\h #\i) 
	   *   but this is inconsistent with (append (list 1) "hi" "hi") -> '(1 #\h #\i . "hi") ?
	   *   Perhaps if all args but last are lists, returned dotted list?
	   */
#if DEBUGGING
	  if (!np) fprintf(stderr, "%s[%d]: np is null\n", __func__, __LINE__);
#endif
	  if (args_are_lists || (is_null(p)))
	    set_cdr(np, p);
	  else 
	    {
	      s7_int len;
	      len = sequence_length(sc, p);
	      if (len > 0)
		set_cdr(np, s7_copy(sc, set_plist_2(sc, p, make_list(sc, len, sc->F))));
	      else 
		{
		  if (len < 0)
		    set_cdr(np, p);
		}
	    }
	  sc->y = sc->nil;
	  return(tp);
	}

      if (!is_sequence(p))
	return(wrong_type_argument_with_type(sc, sc->append_symbol, position_of(y, args), p, a_sequence_string));

      if (!is_null(p))
	{
	  if (is_pair(p))
	    {
	      if (!s7_is_proper_list(sc, p))
		{
		  sc->y = sc->nil;
		  return(wrong_type_argument_with_type(sc, sc->append_symbol, position_of(y, args), p, a_proper_list_string));
		}
	      /* is this error correct?
	       *     (append '(3) '(1 . 2)) -> '(3 1 . 2) ; (old) guile also returns this
	       * but (append '(1 . 2) '(3)) -> this error
	       */
	      
	      if (is_null(tp))
		{
		  tp = cons(sc, car(p), sc->nil);
		  np = tp;
		  sc->y = tp; /* GC protect? */
		  pp = cdr(p);
		}
	      else pp = p;
	      for (; is_pair(pp); pp = cdr(pp), np = cdr(np))
		set_cdr(np, cons(sc, car(pp), sc->nil));
	    }
	  else
	    {
	      s7_int len;
	      args_are_lists = false;
	      len = sequence_length(sc, p);
	      if (len > 0)
		{
		  if (is_null(tp))
		    {
		      tp = s7_copy(sc, set_plist_2(sc, p, make_list(sc, len, sc->F)));
		      np = tp;
		      sc->y = tp;
		    }
		  else set_cdr(np, s7_copy(sc, set_plist_2(sc, p, make_list(sc, len, sc->F))));
		  for (; is_pair(cdr(np)); np = cdr(np));
		}
	      else 
		{
		  if (len < 0)
		    return(wrong_type_argument_with_type(sc, sc->append_symbol, position_of(y, args), p, a_sequence_string));
		}
	    }
	}
    }
  return(tp);
}


static s7_pointer append_in_place(s7_scheme *sc, s7_pointer a, s7_pointer b)
{
  /* tack b onto the end of a without copying either -- 'a' is changed! */
  s7_pointer p;
  if (is_null(a))
    return(b);
  p = a;
  while (is_not_null(cdr(p))) p = cdr(p);
  set_cdr(p, b);
  return(a);
}


/* -------------------------------- vectors -------------------------------- */

bool s7_is_vector(s7_pointer p)
{
  return(t_vector_p[type(p)]);
}


bool s7_is_float_vector(s7_pointer p)
{
  return(type(p) == T_FLOAT_VECTOR);
}


bool s7_is_int_vector(s7_pointer p)
{
  return(type(p) == T_INT_VECTOR);
}


static s7_pointer default_vector_setter(s7_scheme *sc, s7_pointer vec, s7_int loc, s7_pointer val)
{
  vector_element(vec, loc) = val;
  return(val);
}

static s7_pointer default_vector_getter(s7_scheme *sc, s7_pointer vec, s7_int loc)
{
  return(vector_element(vec, loc));
}

static s7_pointer int_vector_setter(s7_scheme *sc, s7_pointer vec, s7_int loc, s7_pointer val)
{
  if (!s7_is_integer(val)) 
    s7_wrong_type_arg_error(sc, "int_vector_set!", 3, val, "an integer");
  int_vector_element(vec, loc) = s7_integer(val);
  return(val);
}

static s7_pointer int_vector_getter(s7_scheme *sc, s7_pointer vec, s7_int loc)
{
  return(make_integer(sc, int_vector_element(vec, loc)));
}

static s7_pointer float_vector_setter(s7_scheme *sc, s7_pointer vec, s7_int loc, s7_pointer val)
{
  float_vector_element(vec, loc) = real_to_double(sc, val, "float-vector-set!");
  return(val);
}

static s7_pointer float_vector_getter(s7_scheme *sc, s7_pointer vec, s7_int loc)
{
  return(make_real(sc, float_vector_element(vec, loc)));
}

static s7_pointer make_simple_vector(s7_scheme *sc, s7_int len) /* len >= 0 and < max */
{
  s7_pointer x;
  new_cell(sc, x, T_VECTOR | T_SAFE_PROCEDURE);
  vector_length(x) = len;
  if (len > 0)
    vector_elements(x) = (s7_pointer *)malloc(len * sizeof(s7_pointer));
  else vector_elements(x) = NULL;
  vector_dimension_info(x) = NULL;
  vector_getter(x) = default_vector_getter;
  vector_setter(x) = default_vector_setter;
  add_vector(sc, x);
  return(x);
}

static s7_pointer make_simple_float_vector(s7_scheme *sc, s7_int len) /* len >= 0 and < max */
{
  s7_pointer x;
  new_cell(sc, x, T_FLOAT_VECTOR | T_SAFE_PROCEDURE);
  vector_length(x) = len;
  if (len > 0)
    float_vector_elements(x) = (s7_double *)malloc(len * sizeof(s7_double));
  else float_vector_elements(x) = NULL;
  vector_dimension_info(x) = NULL;
  vector_getter(x) = float_vector_getter;
  vector_setter(x) = float_vector_setter;
  add_vector(sc, x);
  return(x);
}

static s7_pointer make_vector_1(s7_scheme *sc, s7_int len, bool filled, uint64_t typ)
{
  s7_pointer x;
  if (len < 0)
    return(wrong_type_argument_with_type(sc, sc->make_vector_symbol, 1, make_integer(sc, len), a_non_negative_integer_string));
  if (len > sc->max_vector_length)
    return(out_of_range(sc, sc->make_vector_symbol, small_int(1), make_integer(sc, len), its_too_large_string));

  /* this has to follow the error checks! (else garbage in free_heap temps portion confuses GC when "vector" is finalized) */
  new_cell(sc, x, typ | T_SAFE_PROCEDURE); /* (v 0) as vector-ref is safe */
  vector_length(x) = 0;
  vector_elements(x) = NULL;
  vector_dimension_info(x) = NULL;

  if (len > 0)
    {
      vector_length(x) = len;
      if (typ == T_VECTOR)
	{
	  vector_elements(x) = (s7_pointer *)malloc(len * sizeof(s7_pointer));
	  if (!vector_elements(x))
	    return(s7_error(sc, make_symbol(sc, "out-of-memory"), set_elist_1(sc, s7_make_string_wrapper(sc, "make-vector allocation failed!"))));
	  vector_getter(x) = default_vector_getter;
	  vector_setter(x) = default_vector_setter;
	  if (filled) s7_vector_fill(sc, x, sc->nil);  /* make_hash_table assumes nil as the default value */
	}
      else
	{
	  if (typ == T_FLOAT_VECTOR)
	    {
	      if (filled)
		float_vector_elements(x) = (s7_double *)calloc(len, sizeof(s7_double));
	      else float_vector_elements(x) = (s7_double *)malloc(len * sizeof(s7_double));
	      if (!float_vector_elements(x))
		return(s7_error(sc, make_symbol(sc, "out-of-memory"), set_elist_1(sc, s7_make_string_wrapper(sc, "make-float-vector allocation failed!"))));
	      vector_getter(x) = float_vector_getter;
	      vector_setter(x) = float_vector_setter;
	    }
	  else
	    {
	      if (filled)
		int_vector_elements(x) = (s7_int *)calloc(len, sizeof(s7_int));
	      else int_vector_elements(x) = (s7_int *)malloc(len * sizeof(s7_int));
	      if (!int_vector_elements(x))
		return(s7_error(sc, make_symbol(sc, "out-of-memory"), set_elist_1(sc, s7_make_string_wrapper(sc, "make-int-vector allocation failed!"))));
	      vector_getter(x) = int_vector_getter;
	      vector_setter(x) = int_vector_setter;
	    }
	}
    }

  add_vector(sc, x);
  return(x);
}


s7_pointer s7_make_vector(s7_scheme *sc, s7_int len)
{
  return(make_vector_1(sc, len, FILLED, T_VECTOR));
}

static vdims_t *make_wrap_only(s7_scheme *sc)
{
  vdims_t *v;
  v = (vdims_t *)malloc(sizeof(vdims_t));
  v->original = sc->F;
  v->elements_allocated = false;
  v->ndims = 1;
  v->dimensions_allocated = false;
  v->dims = NULL;
  v->offsets = NULL;
  return(v);
}

#define make_vdims(Sc, Alloc, Dims, Info) ((((Dims) == 1) && (!(Alloc))) ? sc->wrap_only : make_vdims_1(Sc, Alloc, Dims, Info))

static vdims_t *make_vdims_1(s7_scheme *sc, bool elements_allocated, int32_t dims, s7_int *dim_info)
{
  vdims_t *v;

  v = (vdims_t *)malloc(sizeof(vdims_t));
  v->original = sc->F;
  v->elements_allocated = elements_allocated;
  v->ndims = dims;
  if (dims > 1)
    {
      int32_t i;
      s7_int offset = 1;
      v->dimensions_allocated = true;
      v->dims = (s7_int *)malloc(v->ndims * sizeof(s7_int));
      v->offsets = (s7_int *)malloc(v->ndims * sizeof(s7_int));

      for (i = 0; i < dims; i++)
	v->dims[i] = dim_info[i];
      for (i = v->ndims - 1; i >= 0; i--)
	{
	  v->offsets[i] = offset;
	  offset *= v->dims[i];
	}
    }
  else
    {
      v->dimensions_allocated = false;
      v->dims = NULL;
      v->offsets = NULL;
    }
  return(v);
}


s7_pointer s7_make_int_vector(s7_scheme *sc, s7_int len, int32_t dims, s7_int *dim_info)
{
  s7_pointer p;
  p = make_vector_1(sc, len, FILLED, T_INT_VECTOR);
  if (dim_info)
    vector_dimension_info(p) = make_vdims(sc, true, dims, dim_info);
  return(p);
}


s7_pointer s7_make_float_vector(s7_scheme *sc, s7_int len, int32_t dims, s7_int *dim_info)
{
  s7_pointer p;
  p = make_vector_1(sc, len, FILLED, T_FLOAT_VECTOR);
  if (dim_info)
    vector_dimension_info(p) = make_vdims(sc, true, dims, dim_info);
  return(p);
}


s7_pointer s7_make_float_vector_wrapper(s7_scheme *sc, s7_int len, s7_double *data, int32_t dims, s7_int *dim_info, bool free_data)
{
  /* this wraps up a C-allocated/freed double array as an s7 vector.
   */
  s7_pointer x;

  new_cell(sc, x, T_FLOAT_VECTOR | T_SAFE_PROCEDURE);
  float_vector_elements(x) = data;
  vector_getter(x) = float_vector_getter;
  vector_setter(x) = float_vector_setter;
  vector_length(x) = len;
  if (!dim_info)
    {
      if (!free_data)    /* here we need the dim info to tell the GC to leave the data alone */
	{
	  s7_int di[1];
	  di[0] = len;
	  vector_dimension_info(x) = make_vdims(sc, free_data, 1, di);
	}
      else vector_dimension_info(x) = NULL;
    }
  else vector_dimension_info(x) = make_vdims(sc, free_data, dims, dim_info);
  add_vector(sc, x);
  return(x);
}


s7_int s7_print_length(s7_scheme *sc) {return(sc->print_length);}
s7_int s7_set_print_length(s7_scheme *sc, s7_int new_len)
{
  s7_int old_len;
  old_len = sc->print_length;
  sc->print_length = new_len;
 return(old_len);
}

#if (defined(__GNUC__) && __GNUC__ >= 5)
  #define WITH_VECTORIZE 1
#else
  #define WITH_VECTORIZE 0
#endif


#if WITH_VECTORIZE
static void vector_fill(s7_scheme *sc, s7_pointer vec, s7_pointer obj) __attribute__((optimize("tree-vectorize")));
#endif

static void vector_fill(s7_scheme *sc, s7_pointer vec, s7_pointer obj)
{
  s7_int len, i, left;

  len = vector_length(vec);
  if (len == 0) return;
  left = len - 8;
  i = 0;

  switch (type(vec))
    {
    case T_FLOAT_VECTOR:
      if (!s7_is_real(obj))
	s7_wrong_type_arg_error(sc, "(float) vector-fill!", 2, obj, "a real");
      else
	{
	  s7_double x;
	  x = real_to_double(sc, obj, "vector-fill!");
	  if (x == 0.0)
	    memclr((void *)float_vector_elements(vec), len * sizeof(s7_double));
	  else
	    {
	      s7_double *orig;
	      orig = float_vector_elements(vec);
	      while (i <= left)
		{
		  orig[i++] = x;
		  orig[i++] = x;
		  orig[i++] = x;
		  orig[i++] = x;
		  orig[i++] = x;
		  orig[i++] = x;
		  orig[i++] = x;
		  orig[i++] = x;
		}
	      for (; i < len; i++)
		orig[i] = x;
	    }
	}
      break;

    case T_INT_VECTOR:
      if (!s7_is_integer(obj)) /* possibly a bignum */
	s7_wrong_type_arg_error(sc, "int-vector fill!", 2, obj, "an integer");
      else
	{
	  s7_int k;
	  k = s7_integer(obj);
	  if (k == 0)
	    memclr((void *)int_vector_elements(vec), len * sizeof(s7_int));
	  else
	    {
	      s7_int* orig;
	      orig = int_vector_elements(vec);
	      while (i <= left)
		{
		  orig[i++] = k;
		  orig[i++] = k;
		  orig[i++] = k;
		  orig[i++] = k;
		  orig[i++] = k;
		  orig[i++] = k;
		  orig[i++] = k;
		  orig[i++] = k;
		}
	      for (; i < len; i++)
		orig[i] = k;
	    }
	}
      break;

    default:
      {
	s7_pointer *orig;
	orig = vector_elements(vec);
	while (i <= left)
	  {
	    orig[i++] = obj;
	    orig[i++] = obj;
	    orig[i++] = obj;
	    orig[i++] = obj;
	    orig[i++] = obj;
	    orig[i++] = obj;
	    orig[i++] = obj;
	    orig[i++] = obj;
	  }
	for (; i < len; i++)
	  orig[i] = obj;
      }
    }
}

#if (!WITH_GMP)
void s7_vector_fill(s7_scheme *sc, s7_pointer vec, s7_pointer obj) {vector_fill(sc, vec, obj);}
#endif

static s7_pointer g_vector_fill(s7_scheme *sc, s7_pointer args)
{
  #define H_vector_fill "(vector-fill! v val start end) sets all elements of the vector v between start and end to val"
  #define Q_vector_fill s7_make_circular_signature(sc, 3, 4, sc->T, sc->is_vector_symbol, sc->T, sc->is_integer_symbol)

  s7_pointer x, fill;
  s7_int start = 0, end;

  x = car(args);
  if (!s7_is_vector(x))
    {
      check_method(sc, x, sc->vector_fill_symbol, args);
      /* not two_methods (and fill!) here else we get stuff like:
       *   (let ((e (openlet (inlet 'fill! (lambda (obj val) (string-fill! (obj 'value) val)) 'value "01234")))) (vector-fill! e #\a) (e 'value)) -> "aaaaa"
       */
      return(wrong_type_argument(sc, sc->vector_fill_symbol, 1, x, T_VECTOR));
    }

  if ((sc->safety > NO_SAFETY) &&
      (is_immutable_vector(x)))
    return(s7_error(sc, sc->error_symbol, set_elist_2(sc, s7_make_string_wrapper(sc, "can't fill! ~S (it is immutable)"), x)));

  fill = cadr(args);
  if (is_float_vector(x))
    {
      if (!s7_is_real(fill)) /* possibly a bignum */
	{
	  check_two_methods(sc, fill, sc->vector_fill_symbol, sc->fill_symbol, args);
	  s7_wrong_type_arg_error(sc, "(float) vector-fill!", 2, fill, "a real");
	}
    }
  else
    {
      if (is_int_vector(x))
	{
	  if (!s7_is_integer(fill))
	    {
	      check_two_methods(sc, fill, sc->vector_fill_symbol, sc->fill_symbol, args);
	      s7_wrong_type_arg_error(sc, "int-vector fill!", 2, fill, "an integer");
	    }
	}
    }

  end = vector_length(x);
  if (!is_null(cddr(args)))
    {
      s7_pointer p;
      p = start_and_end(sc, sc->vector_fill_symbol, sc->fill_symbol, cddr(args), args, 3, &start, &end);
      if (p != sc->gc_nil) return(p);
      if (start == end) return(fill);
    }
  if (end == 0) return(fill);

  if ((start == 0) && (end == vector_length(x)))
    s7_vector_fill(sc, x, fill);
  else
    {
      s7_int i;
      if (is_normal_vector(x))
	{
	  for (i = start; i < end; i++)
	    vector_element(x, i) = fill;
	}
      else
	{
	  if (is_int_vector(x))
	    {
	      s7_int k;
	      k = s7_integer(fill);
	      if (k == 0)
		memclr((void *)(int_vector_elements(x) + start), (end - start) * sizeof(s7_int));
	      else
		{
		  for (i = start; i < end; i++)
		    int_vector_element(x, i) = k;
		}
	    }
	  else
	    {
	      if (is_float_vector(x))
		{
		  s7_double y;
		  y = real_to_double(sc, fill, "vector-fill!");
		  if (y == 0.0)
		    memclr((void *)(float_vector_elements(x) + start), (end - start) * sizeof(s7_double));
		  else
		    {
		      s7_double *orig;
		      s7_int left;
		      orig = float_vector_elements(x);
		      left = end - 8;
		      i = start;
		      while (i <= left)
			{
			  orig[i++] = y;
			  orig[i++] = y;
			  orig[i++] = y;
			  orig[i++] = y;
			  orig[i++] = y;
			  orig[i++] = y;
			  orig[i++] = y;
			  orig[i++] = y;
			}
		      for (; i < end; i++)
			orig[i] = y;
		    }
		}
	    }
	}
    }
  return(fill);
}


s7_pointer s7_vector_ref(s7_scheme *sc, s7_pointer vec, s7_int index)
{
  if (index >= vector_length(vec))
    return(out_of_range(sc, sc->vector_ref_symbol, small_int(2), make_integer(sc, index), its_too_large_string));

  return(vector_getter(vec)(sc, vec, index));
}


s7_pointer s7_vector_set(s7_scheme *sc, s7_pointer vec, s7_int index, s7_pointer a)
{
  if (index >= vector_length(vec))
    return(out_of_range(sc, sc->vector_set_symbol, small_int(2), make_integer(sc, index), its_too_large_string));

  vector_setter(vec)(sc, vec, index, _NFre(a));
  return(a);
}


s7_pointer *s7_vector_elements(s7_pointer vec)      {return(vector_elements(vec));}
s7_int *s7_int_vector_elements(s7_pointer vec)      {return(int_vector_elements(vec));}
s7_double *s7_float_vector_elements(s7_pointer vec) {return(float_vector_elements(vec));}


s7_int *s7_vector_dimensions(s7_pointer vec)
{
  static s7_int *dims = NULL;
  if (vector_dimension_info(vec))
    return(vector_dimensions(vec));
  if (!dims) dims = (s7_int *)malloc(sizeof(s7_int));
  dims[0] = vector_length(vec);
  return(dims);
}


s7_int *s7_vector_offsets(s7_pointer vec)
{
  static s7_int *offs = NULL;
  if (vector_dimension_info(vec))
    return(vector_offsets(vec));
  if (!offs) offs = (s7_int *)malloc(sizeof(s7_int));
  offs[0] = 1;
  return(offs);
}


#if (!WITH_PURE_S7)
static s7_pointer vector_append(s7_scheme *sc, s7_pointer args, uint8_t typ);

static s7_pointer g_vector_append(s7_scheme *sc, s7_pointer args)
{
  /* returns a one-dimensional vector.  To handle multidimensional vectors, we'd need to
   *   ensure all the dimensional data matches (rank, size of each dimension except the last etc),
   *   which is too much trouble.
   */
  #define H_vector_append "(vector-append . vectors) returns a new (1-dimensional) vector containing the elements of its vector arguments."
  #define Q_vector_append pcl_v

  s7_pointer p;
  int32_t i;

  if (is_null(args))
    return(make_vector_1(sc, 0, NOT_FILLED, T_VECTOR));

  for (i = 0, p = args; is_pair(p); p = cdr(p), i++)
    {
      s7_pointer x;
      x = car(p);
      if (!s7_is_vector(x))
	{
	  if (is_byte_vector(x))
	    return(wrong_type_argument_with_type(sc, sc->vector_append_symbol, i + 1, x,
						 s7_make_string_wrapper(sc, "a byte-vector is actually a string: use append or string-append")));
	  if (has_methods(x))
	    {
	      s7_pointer func;
	      func = find_method(sc, find_let(sc, x), sc->vector_append_symbol);
	      if (func != sc->undefined)
		{
		  int32_t k;
		  s7_pointer v, y;
		  if (i == 0)
		    return(s7_apply_function(sc, func, args));
		  /* we have to copy the arglist here */
		  sc->temp9 = make_list(sc, i, sc->F);
		  for (k = 0, y = args, v = sc->temp9; k < i; k++, y = cdr(y), v = cdr(v))
		    set_car(v, car(y));
		  v = g_vector_append(sc, sc->temp9);
		  y = s7_apply_function(sc, func, cons(sc, v, p));
		  sc->temp9 = sc->nil;
		  return(y);
		}
	    }
	  return(wrong_type_argument(sc, sc->vector_append_symbol, i + 1, x, T_VECTOR));
	}
    }
  return(vector_append(sc, args, type(car(args))));
}

static s7_pointer vector_append_p_pp(s7_pointer p1, s7_pointer p2) 
{
  s7_pointer val;
  cur_sc->temp7 = list_2(cur_sc, p1, p2);
  val = g_vector_append(cur_sc, cur_sc->temp7);
  cur_sc->temp7 = cur_sc->nil;
  return(val);
}

static s7_pointer vector_append_p_ppp(s7_pointer p1, s7_pointer p2, s7_pointer p3) 
{
  s7_pointer val;
  cur_sc->temp7 = list_3(cur_sc, p1, p2, p3);
  val = g_vector_append(cur_sc, cur_sc->temp7);
  cur_sc->temp7 = cur_sc->nil;
  return(val);
}
#endif

s7_pointer s7_vector_ref_n(s7_scheme *sc, s7_pointer vector, int32_t indices, ...)
{
  /* from s7.html */
  int32_t ndims;

  ndims = s7_vector_rank(vector);
  if (ndims == indices)
    {
      va_list ap;
      s7_int index = 0;
      va_start(ap, indices);

      if (ndims == 1)
	{
	  index = va_arg(ap, s7_int);
	  va_end(ap);
	  return(s7_vector_ref(sc, vector, index));
	}
      else
	{
	  int32_t i;
	  s7_int *offsets, *dimensions;

	  dimensions = s7_vector_dimensions(vector);
	  offsets = s7_vector_offsets(vector);

	  for (i = 0; i < indices; i++)
	    {
	      int32_t ind;
	      ind = va_arg(ap, int);
	      if ((ind < 0) ||
		  (ind >= dimensions[i]))
		{
		  va_end(ap);
		  return(out_of_range(sc, sc->vector_ref_symbol, small_int(i), make_integer(sc, ind), (ind < 0) ? its_negative_string : its_too_large_string));
		}
	      index += (ind * offsets[i]);
	    }
	  va_end(ap);
	  return(vector_getter(vector)(sc, vector, index));
	}
    }
  return(s7_wrong_number_of_args_error(sc, "s7_vector_ref_n: wrong number of indices: ~A", s7_make_integer(sc, indices)));
}


s7_pointer s7_vector_set_n(s7_scheme *sc, s7_pointer vector, s7_pointer value, int32_t indices, ...)
{
  int32_t ndims;

  ndims = s7_vector_rank(vector);
  if (ndims == indices)
    {
      va_list ap;
      s7_int index = 0;
      va_start(ap, indices);

      if (ndims == 1)
	{
	  index = va_arg(ap, s7_int);
	  va_end(ap);
	  s7_vector_set(sc, vector, index, value);
	  return(value);
	}
      else
	{
	  int32_t i;
	  s7_int *offsets, *dimensions;

	  dimensions = s7_vector_dimensions(vector);
	  offsets = s7_vector_offsets(vector);

	  for (i = 0; i < indices; i++)
	    {
	      int32_t ind;
	      ind = va_arg(ap, int);
	      if ((ind < 0) ||
		  (ind >= dimensions[i]))
		{
		  va_end(ap);
		  return(s7_out_of_range_error(sc, "s7_vector_set_n", i, s7_make_integer(sc, ind), "should be a valid index"));
		}
	      index += (ind * offsets[i]);
	    }
	  va_end(ap);
	  vector_setter(vector)(sc, vector, index, value);
	  return(value);
	}
    }
  return(s7_wrong_number_of_args_error(sc, "s7_vector_set_n: wrong number of indices: ~A", s7_make_integer(sc, indices)));
}


s7_pointer s7_vector_to_list(s7_scheme *sc, s7_pointer vect)
{
  s7_int i, len;
  s7_pointer result;

  len = vector_length(vect);
  if (len == 0) 
    return(sc->nil);
  if (len >= (sc->free_heap_top - sc->free_heap))
    {
      gc(sc);
      while (len >= (sc->free_heap_top - sc->free_heap))
	resize_heap(sc);
    }

  sc->v = sc->nil;
  for (i = len - 1; i >= 0; i--)
    sc->v = cons_unchecked(sc, vector_getter(vect)(sc, vect, i), sc->v);
  result = sc->v;
  sc->v = sc->nil;
  return(result);
}

#if (!WITH_PURE_S7)
static s7_pointer g_vector_to_list(s7_scheme *sc, s7_pointer args)
{
  s7_int i, start = 0, end;
  s7_pointer p, vec;
  #define H_vector_to_list "(vector->list v start end) returns the elements of the vector v as a list; (map values v)"
  #define Q_vector_to_list s7_make_circular_signature(sc, 2, 3, sc->is_proper_list_symbol, sc->is_vector_symbol, sc->is_integer_symbol)

  vec = car(args);
  if (!s7_is_vector(vec))
    method_or_bust_one_arg(sc, vec, sc->vector_to_list_symbol, args, T_VECTOR);

  end = vector_length(vec);
  if (!is_null(cdr(args)))
    {
      p = start_and_end(sc, sc->vector_to_list_symbol, NULL, cdr(args), args, 2, &start, &end);
      if (p != sc->gc_nil) return(p);
      if (start == end) return(sc->nil);
    }
  if ((start == 0) && (end == vector_length(vec)))
    return(s7_vector_to_list(sc, vec));

  sc->w = sc->nil;
  for (i = end - 1; i >= start; i--)
    sc->w = cons(sc, vector_getter(vec)(sc, vec, i), sc->w);
  p = sc->w;
  sc->w = sc->nil;
  return(p);
}

static s7_pointer vector_to_list_p_p(s7_pointer p) 
{
  s7_pointer val;
  cur_sc->temp7 = list_1(cur_sc, p);
  val = g_vector_to_list(cur_sc, cur_sc->temp7);
  cur_sc->temp7 = cur_sc->nil;
  return(val);
}
#endif


s7_pointer s7_make_and_fill_vector(s7_scheme *sc, s7_int len, s7_pointer fill)
{
  s7_pointer vect;
  vect = make_vector_1(sc, len, NOT_FILLED, T_VECTOR);
  s7_vector_fill(sc, vect, fill);
  return(vect);
}


static s7_pointer g_vector(s7_scheme *sc, s7_pointer args)
{
  #define H_vector "(vector ...) returns a vector whose elements are the arguments"
  #define Q_vector s7_make_circular_signature(sc, 1, 2, sc->is_vector_symbol, sc->T)

  s7_int len;
  s7_pointer vec;

  len = safe_list_length(sc, args); /* was s7_list_length but don't we ensure that arglists are proper? */
  vec = make_simple_vector(sc, len);
  if (len > 0)
    {
      s7_int i;
      s7_pointer x;
      for (x = args, i = 0; is_pair(x); x = cdr(x), i++)
	vector_element(vec, i) = car(x);
    }
  return(vec);
}

static s7_pointer g_is_float_vector(s7_scheme *sc, s7_pointer args)
{
  #define H_is_float_vector "(float-vector? obj) returns #t if obj is an homogeneous float vector"
  #define Q_is_float_vector pl_bt
  check_boolean_method(sc, s7_is_float_vector, sc->is_float_vector_symbol, args);
}

static s7_pointer g_float_vector(s7_scheme *sc, s7_pointer args)
{
  #define H_float_vector "(float-vector ...) returns an homogeneous float vector whose elements are the arguments"
  #define Q_float_vector s7_make_circular_signature(sc, 1, 2, sc->is_float_vector_symbol, sc->is_real_symbol)

  s7_int len;
  s7_pointer vec;

  len = safe_list_length(sc, args);
  vec = make_simple_float_vector(sc, len);
  sc->w = vec;
  if (len > 0)
    {
      s7_int i;
      s7_pointer x;
      for (x = args, i = 0; is_pair(x); x = cdr(x), i++)
	{
	  if (s7_is_real(car(x))) /* bignum is ok here */
	    float_vector_element(vec, i) = real_to_double(sc, car(x), "float-vector");
	  else return(simple_wrong_type_argument(sc, sc->float_vector_symbol, car(x), T_REAL));
	}
    }
  sc->w = sc->nil;
  return(vec);
}

static s7_pointer g_is_int_vector(s7_scheme *sc, s7_pointer args)
{
  #define H_is_int_vector "(int-vector? obj) returns #t if obj is an homogeneous int32_t vector"
  #define Q_is_int_vector pl_bt
  check_boolean_method(sc, is_int_vector, sc->is_int_vector_symbol, args);
}

static s7_pointer g_int_vector(s7_scheme *sc, s7_pointer args)
{
  #define H_int_vector "(int-vector ...) returns an homogeneous int32_t vector whose elements are the arguments"
  #define Q_int_vector s7_make_circular_signature(sc, 1, 2, sc->is_int_vector_symbol, sc->is_integer_symbol)

  s7_int len;
  s7_pointer vec;

  len = safe_list_length(sc, args);
  vec = make_vector_1(sc, len, NOT_FILLED, T_INT_VECTOR);
  if (len > 0)
    {
      s7_int i;
      s7_pointer x;
      for (x = args, i = 0; is_pair(x); x = cdr(x), i++)
	int_vector_element(vec, i) = s7_number_to_integer_with_caller(sc, car(x), "int-vector");
    }
  return(vec);
}

s7_int s7_vector_length(s7_pointer vec)
{
  return(vector_length(vec));
}

#if (!WITH_PURE_S7)
static s7_pointer g_list_to_vector(s7_scheme *sc, s7_pointer args)
{
  s7_pointer p;
  #define H_list_to_vector "(list->vector lst) returns a vector containing the elements of lst; (apply vector lst)"
  #define Q_list_to_vector s7_make_signature(sc, 2, sc->is_vector_symbol, sc->is_proper_list_symbol)

  p = car(args);
  sc->temp3 = p;
  if (is_null(p))
    return(s7_make_vector(sc, 0));

  if (!s7_is_proper_list(sc, p))
    method_or_bust_with_type_one_arg(sc, p, sc->list_to_vector_symbol, list_1(sc, p), a_proper_list_string);

  return(g_vector(sc, p));  
}

static s7_pointer g_vector_length(s7_scheme *sc, s7_pointer args)
{
  s7_pointer vec;
  #define H_vector_length "(vector-length v) returns the length of vector v"
  #define Q_vector_length s7_make_signature(sc, 2, sc->is_integer_symbol, sc->is_vector_symbol)

  vec = car(args);
  if (!s7_is_vector(vec))
    method_or_bust_one_arg(sc, vec, sc->vector_length_symbol, args, T_VECTOR);

  return(make_integer(sc, vector_length(vec)));
}

static s7_int vector_length_i(s7_pointer p) 
{
  if (!s7_is_vector(p)) 
    simple_wrong_type_argument(cur_sc, cur_sc->vector_length_symbol, p, T_VECTOR);
  return(vector_length(p));
}
#endif


static s7_pointer make_shared_vector(s7_scheme *sc, s7_pointer vect, int32_t skip_dims, s7_int index)
{
  s7_pointer x;
  vdims_t *v;

  /* (let ((v #2d((1 2) (3 4)))) (v 1))
   * (let ((v (make-vector '(2 3 4) 0))) (v 1 2))
   * (let ((v #3d(((0 1 2 3) (4 5 6 7) (8 9 10 11)) ((12 13 14 15) (16 17 18 19) (20 21 22 23))))) (v 0 1))
   */

  new_cell(sc, x, typeflag(vect) | T_SAFE_PROCEDURE);
  vector_length(x) = 0;
  vector_elements(x) = NULL;
  vector_getter(x) = vector_getter(vect);
  vector_setter(x) = vector_setter(vect);

  v = (vdims_t *)malloc(sizeof(vdims_t));
  v->ndims = vector_ndims(vect) - skip_dims;
  v->dims = (s7_int *)(vector_dimensions(vect) + skip_dims);
  v->offsets = (s7_int *)(vector_offsets(vect) + skip_dims);
  v->original = vect; /* shared_vector */
  if (is_normal_vector(vect))
    mark_function[T_VECTOR] = mark_vector_possibly_shared; 
  else mark_function[type(vect)] = mark_int_or_float_vector_possibly_shared; 
  v->elements_allocated = false;
  v->dimensions_allocated = false;
  vector_dimension_info(x) = v;

  if (skip_dims > 0)
    vector_length(x) = vector_offset(vect, skip_dims - 1);
  else vector_length(x) = vector_length(vect);

  if (is_int_vector(vect))
    int_vector_elements(x) = (s7_int *)(int_vector_elements(vect) + index);
  else
    {
      if (is_float_vector(vect))
	float_vector_elements(x) = (s7_double *)(float_vector_elements(vect) + index);
      else vector_elements(x) = (s7_pointer *)(vector_elements(vect) + index);
    }
  add_vector(sc, x);
  return(x);
}


static s7_pointer g_make_shared_vector(s7_scheme *sc, s7_pointer args)
{
  #define H_make_shared_vector "(make-shared-vector original-vector new-dimensions (offset 0)) returns \
a vector that points to the same elements as the original-vector but with different dimensional info."
  #define Q_make_shared_vector s7_make_signature(sc, 4, sc->is_vector_symbol, sc->is_vector_symbol, s7_make_signature(sc, 2, sc->is_pair_symbol, sc->is_integer_symbol), sc->is_integer_symbol)

  /* (let ((v1 #2d((1 2 3) (4 5 6)))) (let ((v2 (make-shared-vector v1 '(6)))) v2)) -> #(1 2 3 4 5 6)
   * (let ((v1 #(1 2 3 4 5 6))) (let ((v2 (make-shared-vector v1 '(3 2)))) v2)) -> #2D((1 2) (3 4) (5 6))
   * this is most useful in generic functions -- they can still use (v n) as the accessor.
   */
  s7_pointer orig, dims, y, x;
  vdims_t *v;
  int32_t i;
  s7_int new_len = 1, orig_len, offset = 0;

  orig = car(args);
  if (!s7_is_vector(orig))
    method_or_bust(sc, orig, sc->make_shared_vector_symbol, args, T_VECTOR, 1);

  orig_len = vector_length(orig);

  if (!is_null(cddr(args)))
    {
      s7_pointer off;
      off = caddr(args);
      if (s7_is_integer(off))
	{
	  offset = s7_integer(off);
	  if ((offset < 0) ||
	      (offset >= orig_len))  /* we need this if, for example, offset == 9223372036854775807 */
	    return(out_of_range(sc, sc->make_shared_vector_symbol, small_int(3), off, (offset < 0) ? its_negative_string : its_too_large_string));
	}
      else method_or_bust(sc, off, sc->make_shared_vector_symbol, args, T_INTEGER, 3);
    }

  dims = cadr(args);
  if (is_integer(dims))
    {
      if ((s7_integer(dims) < 0) ||
	  (s7_integer(dims) >= orig_len))
	return(out_of_range(sc, sc->make_shared_vector_symbol, small_int(2), dims, (s7_integer(dims) < 0) ? its_negative_string : its_too_large_string));
      dims = list_1(sc, dims);
    }
  else
    {
      if ((is_null(dims)) ||
	  (!s7_is_proper_list(sc, dims)))
	method_or_bust(sc, dims, sc->make_shared_vector_symbol, args, T_PAIR, 2);

      for (y = dims; is_pair(y); y = cdr(y))
	if ((!s7_is_integer(car(y)))        ||       /* (make-shared-vector v '((1 2) (3 4))) */
	    (s7_integer(car(y)) > orig_len) ||
	    (s7_integer(car(y)) < 0))
	  return(s7_error(sc, sc->wrong_type_arg_symbol, set_elist_1(sc, s7_make_string_wrapper(sc, "a list of integers that fits the original vector"))));
    }

  v = (vdims_t *)malloc(sizeof(vdims_t));
  v->ndims = safe_list_length(sc, dims);
  v->dims = (s7_int *)malloc(v->ndims * sizeof(s7_int));
  v->offsets = (s7_int *)malloc(v->ndims * sizeof(s7_int));
  v->dimensions_allocated = true;
  v->elements_allocated = false;
  v->original = orig; /* shared_vector */
  if (is_normal_vector(orig))
    mark_function[T_VECTOR] = mark_vector_possibly_shared; 
  else mark_function[type(orig)] = mark_int_or_float_vector_possibly_shared; 

  for (i = 0, y = dims; is_pair(y); i++, y = cdr(y))
    v->dims[i] = s7_integer(car(y));

  for (i = v->ndims - 1; i >= 0; i--)
    {
      v->offsets[i] = new_len;
      new_len *= v->dims[i];
    }

  if ((new_len < 0) || ((new_len + offset) > vector_length(orig)))
    {
      free(v->dims);
      free(v->offsets);
      free(v);
      return(out_of_range(sc, sc->make_shared_vector_symbol, small_int(2), dims, s7_make_string_wrapper(sc, "a shared vector has to fit in the original vector")));
    }

  new_cell(sc, x, typeflag(orig) | T_SAFE_PROCEDURE);
  vector_dimension_info(x) = v;
  vector_length(x) = new_len;                 /* might be less than original length */
  vector_getter(x) = vector_getter(orig);
  vector_setter(x) = vector_setter(orig);

  if (is_int_vector(orig))
    int_vector_elements(x) = (s7_int *)(int_vector_elements(orig) + offset);
  else
    {
      if (is_float_vector(orig))
	float_vector_elements(x) = (s7_double *)(float_vector_elements(orig) + offset);
      else vector_elements(x) = (s7_pointer *)(vector_elements(orig) + offset);
    }

  add_vector(sc, x);
  return(x);
}


static s7_pointer make_vector_wrapper(s7_scheme *sc, s7_int size, s7_pointer *elements)
{
  s7_pointer x;
  new_cell(sc, x, T_VECTOR | T_SAFE_PROCEDURE);
  vector_length(x) = size;
  vector_elements(x) = elements;
  vector_getter(x) = default_vector_getter;
  vector_setter(x) = default_vector_setter;
  vector_dimension_info(x) = NULL;
  /* don't add_vector -- no need for sweep to see this */
  return(x);
}

static s7_pointer make_subvector(s7_scheme *sc, s7_pointer v)
{
  s7_pointer x;
  new_cell(sc, x, type(v));
  vector_length(x) = vector_length(v);
  if (is_normal_vector(v))
    vector_elements(x) = vector_elements(v);
  else
    {
      if (is_float_vector(v))
	float_vector_elements(x) = float_vector_elements(v);
      else int_vector_elements(x) = int_vector_elements(v);
    }
  vector_getter(x) = vector_getter(v);
  vector_setter(x) = vector_setter(v);
  vector_dimension_info(x) = NULL;
  return(x);
}


static s7_pointer vector_ref_1(s7_scheme *sc, s7_pointer vect, s7_pointer indices)
{
  s7_int index = 0;
  if (vector_length(vect) == 0)
    return(out_of_range(sc, sc->vector_ref_symbol, small_int(1), vect, its_too_large_string));

  if (vector_rank(vect) > 1)
    {
      uint32_t i;
      s7_pointer x;
      for (x = indices, i = 0; (is_not_null(x)) && (i < vector_ndims(vect)); x = cdr(x), i++)
	{
	  s7_int n;
	  s7_pointer p, p1;
	  p = car(x);
	  if (!s7_is_integer(p))
	    {
	      if (!s7_is_integer(p1 = check_values(sc, p, x)))
		method_or_bust(sc, p, sc->vector_ref_symbol, cons(sc, vect, indices), T_INTEGER, i + 2);
	      p = p1;
	    }
	  n = s7_integer(p);
	  if ((n < 0) ||
	      (n >= vector_dimension(vect, i)))
	    return(out_of_range(sc, sc->vector_ref_symbol, make_integer(sc, i + 2), p, (n < 0) ? its_negative_string : its_too_large_string));

	  index += n * vector_offset(vect, i);
	}
      if (is_not_null(x))
	{
	  if (!is_normal_vector(vect))
	    return(out_of_range(sc, sc->vector_ref_symbol, small_int(2), indices, too_many_indices_string));
	  return(implicit_index(sc, vector_element(vect, index), x));
	}

      /* if not enough indices, return a shared vector covering whatever is left */
      if (i < vector_ndims(vect))
	return(make_shared_vector(sc, vect, i, index));
    }
  else
    {
      s7_pointer p, p1;
      /* (let ((hi (make-vector 3 0.0)) (sum 0.0)) (do ((i 0 (+ i 1))) ((= i 3)) (set! sum (+ sum (hi i)))) sum) */
      p = car(indices);

      if (!s7_is_integer(p))
	{
	  if (!s7_is_integer(p1 = check_values(sc, p, indices)))
	    method_or_bust(sc, p, sc->vector_ref_symbol, cons(sc, vect, indices), T_INTEGER, 2);
	  p = p1;
	}
      index = s7_integer(p);
      if ((index < 0) ||
	  (index >= vector_length(vect)))
	return(out_of_range(sc, sc->vector_ref_symbol, small_int(2), p, (index < 0) ? its_negative_string : its_too_large_string));

      if (is_not_null(cdr(indices)))                /* (let ((L #(#(1 2 3) #(4 5 6)))) (vector-ref L 1 2)) */
	{
	  if (!is_normal_vector(vect))
	    return(out_of_range(sc, sc->vector_ref_symbol, small_int(2), indices, too_many_indices_string));
	  return(implicit_index(sc, vector_element(vect, index), cdr(indices)));
	}
    }
  return((vector_getter(vect))(sc, vect, index));
}


static s7_pointer g_vector_ref(s7_scheme *sc, s7_pointer args)
{
  #define H_vector_ref "(vector-ref v ... i) returns the i-th element of vector v."
  #define Q_vector_ref s7_make_circular_signature(sc, 2, 3, sc->T, sc->is_vector_symbol, sc->is_integer_symbol)

  s7_pointer vec;

  vec = car(args);
  if (!s7_is_vector(vec))
    method_or_bust(sc, vec, sc->vector_ref_symbol, args, T_VECTOR, 1);
  return(vector_ref_1(sc, vec, cdr(args)));
}

static s7_pointer g_vector_ref_ic_n(s7_scheme *sc, s7_pointer args, s7_int index)
{
  s7_pointer vec;
  vec = find_symbol_unchecked(sc, car(args));
  if (!s7_is_vector(vec))
    method_or_bust(sc, vec, sc->vector_ref_symbol, list_2(sc, vec, cadr(args)), T_VECTOR, 1);

  if (index >= vector_length(vec))
    return(out_of_range(sc, sc->vector_ref_symbol, small_int(2), cadr(args), its_too_large_string));
  if (vector_rank(vec) > 1)
    {
      if (index >= vector_dimension(vec, 0))
	return(out_of_range(sc, sc->vector_ref_symbol, small_int(2), cadr(args), its_too_large_string));
      return(make_shared_vector(sc, vec, 1, index * vector_offset(vec, 0)));
    }
  return(vector_getter(vec)(sc, vec, index));
}

static s7_pointer vector_ref_p_pi(s7_pointer v, s7_int i) 
{
  if ((!s7_is_vector(v)) ||
      (vector_rank(v) > 1) ||
      (i < 0) ||
      (i >= vector_length(v)))
    return(g_vector_ref(cur_sc, set_plist_2(cur_sc, v, make_integer(cur_sc, i))));
  return(vector_getter(v)(cur_sc, v, i));
}

static s7_pointer vector_ref_p_pi_direct(s7_pointer v, s7_int i) 
{
  if ((i < 0) || (i >= vector_length(v)))
    out_of_range(cur_sc, cur_sc->vector_ref_symbol, small_int(2), make_integer(cur_sc, i), (i < 0) ? its_negative_string : its_too_large_string);
  return(vector_getter(v)(cur_sc, v, i));
}

static s7_pointer vector_ref_unchecked(s7_pointer v, s7_int i) {return(vector_getter(v)(cur_sc, v, i));}

static s7_pointer vector_ref_ic;
static s7_pointer g_vector_ref_ic(s7_scheme *sc, s7_pointer args) {return(g_vector_ref_ic_n(sc, args, s7_integer(cadr(args))));}
static s7_pointer vector_ref_ic_0;
static s7_pointer g_vector_ref_ic_0(s7_scheme *sc, s7_pointer args) {return(g_vector_ref_ic_n(sc, args, 0));}
static s7_pointer vector_ref_ic_1;
static s7_pointer g_vector_ref_ic_1(s7_scheme *sc, s7_pointer args) {return(g_vector_ref_ic_n(sc, args, 1));}
static s7_pointer vector_ref_ic_2;
static s7_pointer g_vector_ref_ic_2(s7_scheme *sc, s7_pointer args) {return(g_vector_ref_ic_n(sc, args, 2));}
static s7_pointer vector_ref_ic_3;
static s7_pointer g_vector_ref_ic_3(s7_scheme *sc, s7_pointer args) {return(g_vector_ref_ic_n(sc, args, 3));}

static s7_pointer vector_ref_2;
static s7_pointer g_vector_ref_2(s7_scheme *sc, s7_pointer args)
{
  s7_pointer vec, ind;
  s7_int index;

  vec = car(args);
  if (!s7_is_vector(vec))
    method_or_bust(sc, vec, sc->vector_ref_symbol, args, T_VECTOR, 1);       /* should be ok because we go to g_vector_ref below */

  if (vector_rank(vec) > 1)
    return(g_vector_ref(sc, args));

  ind = cadr(args);
  if (!s7_is_integer(ind))
    method_or_bust(sc, ind, sc->vector_ref_symbol, args, T_INTEGER, 2);

  index = s7_integer(ind);
  if ((index < 0) || (index >= vector_length(vec)))
    return(out_of_range(sc, sc->vector_ref_symbol, small_int(2), ind, (index < 0) ? its_negative_string : its_too_large_string));

  return(vector_getter(vec)(sc, vec, index));
}

static s7_pointer vector_ref_2_direct;
static s7_pointer g_vector_ref_2_direct(s7_scheme *sc, s7_pointer args)
{
  s7_pointer vec, ind;
  s7_int index;
  vec = car(args);
  ind = cadr(args);
  if (!s7_is_integer(ind))
    method_or_bust(sc, ind, sc->vector_ref_symbol, args, T_INTEGER, 2);
  index = s7_integer(ind);
  if ((index < 0) || (index >= vector_length(vec)))
    return(out_of_range(sc, sc->vector_ref_symbol, small_int(2), ind, (index < 0) ? its_negative_string : its_too_large_string));
  return(vector_element(vec, index));
}



static s7_pointer g_vector_set(s7_scheme *sc, s7_pointer args)
{
  #define H_vector_set "(vector-set! v i ... value) sets the i-th element of vector v to value."
  #define Q_vector_set s7_make_circular_signature(sc, 3, 4, sc->T, sc->is_vector_symbol, sc->is_integer_symbol, sc->is_integer_or_any_at_end_symbol)

  s7_pointer vec, val;
  s7_int index;

  vec = car(args);
  if (!s7_is_vector(vec))
    method_or_bust(sc, vec, sc->vector_set_symbol, args, T_VECTOR, 1);

  if (vector_length(_TSet(vec)) == 0)
    return(out_of_range(sc, sc->vector_set_symbol, small_int(1), vec, its_too_large_string));

  if (vector_rank(vec) > 1)
    {
      uint32_t i;
      s7_pointer x;
      index = 0;
      for (x = cdr(args), i = 0; (is_not_null(cdr(x))) && (i < vector_ndims(vec)); x = cdr(x), i++)
	{
	  s7_int n;
	  s7_pointer p, p1;
	  p = car(x);
	  if (!s7_is_integer(p))
	    {
	      if (!s7_is_integer(p1 = check_values(sc, p, x)))
		method_or_bust(sc, p, sc->vector_set_symbol, args, T_INTEGER, i + 2);
	      p = p1;
	    }
	  n = s7_integer(p);
	  if ((n < 0) ||
	      (n >= vector_dimension(vec, i)))
	    return(out_of_range(sc, sc->vector_set_symbol, make_integer(sc, i + 2), p, (n < 0) ? its_negative_string : its_too_large_string));

	  index += n * vector_offset(vec, i);
	}

      if (is_not_null(cdr(x)))
	return(s7_wrong_number_of_args_error(sc, "too many args for vector-set!: ~S", args));
      if (i != vector_ndims(vec))
	return(s7_wrong_number_of_args_error(sc, "not enough args for vector-set!: ~S", args));

      val = car(x);
    }
  else
    {
      s7_pointer p, p1;
      p = cadr(args);
      if (!s7_is_integer(p))
	{
	  if (!s7_is_integer(p1 = check_values(sc, p, cdr(args))))
	    method_or_bust(sc, p, sc->vector_set_symbol, args, T_INTEGER, 2);
	  p = p1;
	}
      index = s7_integer(p);
      if ((index < 0) ||
	  (index >= vector_length(vec)))
	return(out_of_range(sc, sc->vector_set_symbol, small_int(2), p, (index < 0) ? its_negative_string : its_too_large_string));

      if (is_not_null(cdddr(args)))
	{
	  set_car(sc->temp_cell_2, vector_getter(vec)(sc, vec, index));
	  set_cdr(sc->temp_cell_2, cddr(args));
	  return(g_vector_set(sc, sc->temp_cell_2));
	}
      val = caddr(args);
    }

  vector_setter(vec)(sc, vec, index, val);
  return(val);
}

static s7_pointer vector_set_p_pip(s7_pointer v, s7_int i, s7_pointer p) 
{
  if ((!s7_is_vector(v)) ||
      (vector_rank(v) > 1) ||
      (i < 0) ||
      (i >= vector_length(v)))
    return(g_vector_set(cur_sc, set_plist_3(cur_sc, v, make_integer(cur_sc, i), p)));
  vector_setter(v)(cur_sc, v, i, p);
  return(p);
}

static s7_pointer vector_set_p_pip_direct(s7_pointer v, s7_int i, s7_pointer p) 
{
#if DEBUGGING
  if (!is_normal_vector(v)) abort();
#endif
  if ((i < 0) || (i >= vector_length(v)))
    out_of_range(cur_sc, cur_sc->vector_set_symbol, small_int(2), make_integer(cur_sc, i), (i < 0) ? its_negative_string : its_too_large_string);
  vector_element(v, i) = p;
  return(p);
}

static s7_pointer vector_set_unchecked(s7_pointer v, s7_int i, s7_pointer p) 
{  
#if DEBUGGING
  if (!is_normal_vector(v)) abort();
#endif
  vector_element(v, i) = p;
  return(p);
}

static s7_pointer vector_set_ic;
static s7_pointer g_vector_set_ic(s7_scheme *sc, s7_pointer args)
{
  /* (vector-set! vec 0 x) */
  s7_pointer vec, val;
  s7_int index;

  vec = find_symbol_unchecked(sc, car(args));
  if (!s7_is_vector(vec))
    method_or_bust(sc, vec, sc->vector_set_symbol, list_3(sc, vec, cadr(args), find_symbol_unchecked(sc, caddr(args))), T_VECTOR, 1);
  /* the list_3 happens only if we find the method */

  if (vector_rank(vec) > 1)
    return(g_vector_set(sc, set_plist_3(sc, vec, cadr(args), find_symbol_unchecked(sc, caddr(args)))));

  index = s7_integer(cadr(args));
  if (index >= vector_length(vec))
    return(out_of_range(sc, sc->vector_set_symbol, small_int(2), cadr(args), its_too_large_string));

  val = find_symbol_unchecked(sc, caddr(args));
  vector_setter(vec)(sc, vec, index, val);
  return(val);
}


static s7_pointer c_vector_set_3(s7_scheme *sc, s7_pointer vec, s7_int index, s7_pointer val)
{
  /* (vector-set! vec ind val) where are all predigested */

  if (!s7_is_vector(vec))
    method_or_bust(sc, vec, sc->vector_set_symbol, list_3(sc, vec, make_integer(sc, index), val), T_VECTOR, 1);

  if (vector_rank(vec) > 1)
    return(g_vector_set(sc, set_plist_3(sc, vec, make_integer(sc, index), val)));

  if ((index < 0) ||
      (index >= vector_length(vec)))
    return(out_of_range(sc, sc->vector_set_symbol, small_int(2), make_integer(sc, index), (index < 0) ? its_negative_string : its_too_large_string));

  vector_setter(vec)(sc, vec, index, val);
  return(val);
}


static s7_pointer vector_set_3;
static s7_pointer g_vector_set_3(s7_scheme *sc, s7_pointer args)
{
  s7_pointer ind;
  ind = cadr(args);
  if (!s7_is_integer(ind))
    {
      s7_pointer p;
      if (!s7_is_integer(p = check_values(sc, ind, cdr(args))))
	return(wrong_type_argument(sc, sc->vector_set_symbol, 2, ind, T_INTEGER));
      else ind = p;
    }
  return(c_vector_set_3(sc, car(args), s7_integer(ind), caddr(args)));
}


static s7_pointer g_make_vector_1(s7_scheme *sc, s7_pointer args, s7_pointer err_sym)
{
  s7_int len;
  s7_pointer x, fill, vec;
  int32_t result_type = T_VECTOR;

  fill = sc->unspecified;
  x = car(args);
  if (s7_is_integer(x))
    {
      len = s7_integer(x);
      if (len < 0)
	return(wrong_type_argument_with_type(sc, err_sym, 1, x, a_non_negative_integer_string));
    }
  else
    {
      if (!(is_pair(x)))
	method_or_bust_with_type(sc, x, err_sym, args, s7_make_string_wrapper(sc, "an integer or a list of integers"), 1);

      if (!s7_is_integer(car(x)))
	return(wrong_type_argument_with_type(sc, err_sym, 1, car(x),
					     s7_make_string_wrapper(sc, "each dimension should be an integer")));
      if (is_null(cdr(x)))
	len = s7_integer(car(x));
      else
	{
	  int32_t dims;
	  s7_pointer y;

	  dims = s7_list_length(sc, x);
	  if (dims <= 0)                /* 0 if circular, negative if dotted */
	    return(wrong_type_argument_with_type(sc, err_sym, 1, x, a_proper_list_string));
	  if (dims > sc->max_vector_dimensions)
	    return(out_of_range(sc, err_sym, small_int(1), x, its_too_large_string));

	  for (len = 1, y = x; is_not_null(y); y = cdr(y))
	    {
	      if (!s7_is_integer(car(y)))
		return(wrong_type_argument(sc, err_sym, position_of(y, x), car(y), T_INTEGER));
	      len *= s7_integer(car(y));
	      if (len < 0)
		return(wrong_type_argument_with_type(sc, err_sym, position_of(y, x), car(y), a_non_negative_integer_string));
	    }
	}
    }

  if (is_not_null(cdr(args)))
    {
      fill = cadr(args);
      if (is_not_null(cddr(args)))
	{
	  if (caddr(args) == sc->T)
	    {
	      /* here bignums can cause confusion, so use is_integer not s7_is_integer etc */
	      if (is_integer(fill))
		result_type = T_INT_VECTOR;
	      else
		{
		  if (s7_is_real(fill)) /* might be gmp with big_real by accident (? see above) */
		    result_type = T_FLOAT_VECTOR;
		  else method_or_bust_with_type(sc, fill, err_sym, args, s7_make_string_wrapper(sc, "an integer or a real since 'homogeneous' is #t"), 2);
		}
	    }
	  else
	    {
	      if (caddr(args) != sc->F)
		method_or_bust_with_type(sc, caddr(args), err_sym, args, a_boolean_string, 3);
	    }
	}
    }

  vec = make_vector_1(sc, len, NOT_FILLED, result_type);
  if (len > 0) s7_vector_fill(sc, vec, fill);

  if ((is_pair(x)) &&
      (is_pair(cdr(x))))
    {
      int32_t i;
      s7_int offset = 1;
      s7_pointer y;
      vdims_t *v;

      v = (vdims_t *)malloc(sizeof(vdims_t));
      v->ndims = safe_list_length(sc, x);
      v->dims = (s7_int *)malloc(v->ndims * sizeof(s7_int));
      v->offsets = (s7_int *)malloc(v->ndims * sizeof(s7_int));
      v->original = sc->F;
      v->dimensions_allocated = true;
      v->elements_allocated = (len > 0);

      for (i = 0, y = x; is_not_null(y); i++, y = cdr(y))
	v->dims[i] = s7_integer(car(y));

      for (i = v->ndims - 1; i >= 0; i--)
	{
	  v->offsets[i] = offset;
	  offset *= v->dims[i];
	}
      vector_dimension_info(vec) = v;
    }
  return(vec);
}

static s7_pointer g_make_vector(s7_scheme *sc, s7_pointer args)
{
  #define H_make_vector "(make-vector len (value #<unspecified>)) returns a vector of len elements initialized to value. \
To create a multidimensional vector, put the dimension bounds in a list (this is to avoid ambiguities such as \
(make-vector 1 2) where it's not clear whether the '2' is an initial value or a dimension size).  (make-vector '(2 3) 1.0) \
returns a 2 dimensional vector of 6 total elements, all initialized to 1.0."
  #define Q_make_vector s7_make_signature(sc, 3, sc->is_vector_symbol, s7_make_signature(sc, 2, sc->is_integer_symbol, sc->is_pair_symbol), sc->T)
  return(g_make_vector_1(sc, args, sc->make_vector_symbol));
}

static s7_pointer g_make_float_vector(s7_scheme *sc, s7_pointer args)
{
  #define H_make_float_vector "(make-float-vector len (init 0.0)) returns a float-vector."
  #define Q_make_float_vector s7_make_signature(sc, 3, sc->is_float_vector_symbol, s7_make_signature(sc, 2, sc->is_integer_symbol, sc->is_pair_symbol), sc->is_real_symbol)
  s7_int len;
  s7_pointer x, p;
  s7_double *arr;

  p = car(args);
  if ((is_pair(cdr(args))) ||
      (!is_integer(p)))
    {
      s7_pointer init;
      if (is_pair(cdr(args)))
	{
	  init = cadr(args);
	  if (!s7_is_real(init))
	    method_or_bust(sc, init, sc->make_float_vector_symbol, args, T_REAL, 2);
#if WITH_GMP
	  if (s7_is_bignum(init))
	    return(g_make_vector_1(sc, set_plist_3(sc, p, make_real(sc, real_to_double(sc, init, "make-float-vector")), sc->T), sc->make_float_vector_symbol));
#endif
	  if (is_rational(init))
	    return(g_make_vector_1(sc, set_plist_3(sc, p, make_real(sc, rational_to_double(sc, init)), sc->T), sc->make_float_vector_symbol));
	}
      else init = real_zero;
      return(g_make_vector_1(sc, set_plist_3(sc, p, init, sc->T), sc->make_float_vector_symbol));
    }

  len = s7_integer(p);
  if (len < 0)
    return(wrong_type_argument_with_type(sc, sc->make_float_vector_symbol, 1, p, a_non_negative_integer_string));
  if (len > sc->max_vector_length)
    return(out_of_range(sc, sc->make_float_vector_symbol, small_int(1), p, its_too_large_string));

  if (len > 0)
    arr = (s7_double *)calloc(len, sizeof(s7_double));
  else arr = NULL;

  new_cell(sc, x, T_FLOAT_VECTOR | T_SAFE_PROCEDURE);
  vector_length(x) = len;
  float_vector_elements(x) = arr;
  vector_dimension_info(x) = NULL;
  vector_getter(x) = float_vector_getter;
  vector_setter(x) = float_vector_setter;

  add_vector(sc, x);
  return(x);
}


static s7_pointer g_make_int_vector(s7_scheme *sc, s7_pointer args)
{
  #define H_make_int_vector "(make-int-vector len (init 0)) returns an int-vector."
  #define Q_make_int_vector s7_make_signature(sc, 3, sc->is_int_vector_symbol, s7_make_signature(sc, 2, sc->is_integer_symbol, sc->is_pair_symbol), sc->is_integer_symbol)

  s7_int len;
  s7_pointer x, p;
  s7_int *arr;

  p = car(args);
  if ((is_pair(cdr(args))) ||
      (!is_integer(p)))
    {
      s7_pointer init;
      if (is_pair(cdr(args)))
	{
	  init = cadr(args);
	  if (!is_integer(init))
	    method_or_bust(sc, init, sc->make_int_vector_symbol, args, T_INTEGER, 2);
	}
      else init = small_int(0);
      return(g_make_vector_1(sc, set_plist_3(sc, p, init, sc->T), sc->make_int_vector_symbol));
    }

  len = s7_integer(p);
  if (len < 0)
    return(wrong_type_argument_with_type(sc, sc->make_int_vector_symbol, 1, p, a_non_negative_integer_string));
  if (len > sc->max_vector_length)
    return(out_of_range(sc, sc->make_int_vector_symbol, small_int(1), p, its_too_large_string));

  if (len > 0)
    arr = (s7_int *)calloc(len, sizeof(s7_int));
  else arr = NULL;

  new_cell(sc, x, T_INT_VECTOR | T_SAFE_PROCEDURE);
  vector_length(x) = len;
  int_vector_elements(x) = arr;
  vector_dimension_info(x) = NULL;
  vector_getter(x) = int_vector_getter;
  vector_setter(x) = int_vector_setter;

  add_vector(sc, x);
  return(x);
}

static s7_pointer g_is_vector(s7_scheme *sc, s7_pointer args)
{
  #define H_is_vector "(vector? obj) returns #t if obj is a vector"
  #define Q_is_vector pl_bt
  check_boolean_method(sc, s7_is_vector, sc->is_vector_symbol, args);
}


int s7_vector_rank(s7_pointer vect)
{
  return(vector_rank(vect));
}


static s7_pointer g_vector_dimensions(s7_scheme *sc, s7_pointer args)
{
  #define H_vector_dimensions "(vector-dimensions vect) returns a list of vect's dimensions.  In srfi-63 terms:\n\
    (define array-dimensions vector-dimensions)\n\
    (define (array-rank v) (length (vector-dimensions v)))"
  #define Q_vector_dimensions s7_make_signature(sc, 2, sc->is_pair_symbol, sc->is_vector_symbol)

  s7_pointer x;
  x = car(args);
  if (!s7_is_vector(x))
    method_or_bust_one_arg(sc, x, sc->vector_dimensions_symbol, args, T_VECTOR);

  if (vector_rank(x) > 1)
    {
      int32_t i;
      sc->w = sc->nil;
      for (i = vector_ndims(x) - 1; i >= 0; i--)
	sc->w = cons(sc, make_integer(sc, vector_dimension(x, i)), sc->w);
      x = sc->w;
      sc->w = sc->nil;
      return(x);
    }
  return(list_1(sc, make_integer(sc, vector_length(x))));
}


#define MULTIVECTOR_TOO_MANY_ELEMENTS -1
#define MULTIVECTOR_NOT_ENOUGH_ELEMENTS -2

static int32_t traverse_vector_data(s7_scheme *sc, s7_pointer vec, int32_t flat_ref, int32_t dimension, int32_t dimensions, int32_t *sizes, s7_pointer lst)
{
  /* we're filling vec, we're currently looking for element (flat-wise) flat_ref,
   *   we're at ref in dimension of dimensions, where sizes gives the bounds, and lst is our data
   *   #3D(((1 2 3) (4 5 6)) ((7 8 9) (10 11 12)))
   */
  int32_t i;
  s7_pointer x;

  for (i = 0, x = lst; i < sizes[dimension]; i++, x = cdr(x))
    {
      if (!is_pair(x))
	return(MULTIVECTOR_NOT_ENOUGH_ELEMENTS);

      if (dimension == (dimensions - 1))
	vector_setter(vec)(sc, vec, flat_ref++, car(x));
      else
	{
	  flat_ref = traverse_vector_data(sc, vec, flat_ref, dimension + 1, dimensions, sizes, car(x));
	  if (flat_ref < 0) return(flat_ref);
	}
    }
  if (is_not_null(x))
    return(MULTIVECTOR_TOO_MANY_ELEMENTS);
  return(flat_ref);
}


static s7_pointer s7_multivector_error(s7_scheme *sc, const char *message, s7_pointer data)
{
  return(s7_error(sc, sc->read_error_symbol, 
		  set_elist_3(sc, s7_make_string_wrapper(sc, "reading constant vector, ~A: ~A"), s7_make_string_wrapper(sc, message), data)));
}


static s7_pointer g_multivector(s7_scheme *sc, s7_int dims, s7_pointer data)
{
  /* get the dimension bounds from data, make the new vector, fill it from data
   *
   * dims needs to be s7_int so we can at least give correct error messages.
   * also should we let an empty vector have any number of dimensions? currently ndims is an int.
   */
  s7_pointer vec, x;
  int32_t i, err;
  uint32_t vec_loc;
  int32_t *sizes;

  /* (#2d((1 2 3) (4 5 6)) 0 0) -> 1
   * (#2d((1 2 3) (4 5 6)) 0 1) -> 2
   * (#2d((1 2 3) (4 5 6)) 1 1) -> 5
   * (#3D(((1 2) (3 4)) ((5 6) (7 8))) 0 0 0) -> 1
   * (#3D(((1 2) (3 4)) ((5 6) (7 8))) 1 1 0) -> 7
   * #3D(((1 2) (3 4)) ((5 6) (7))) -> error, #3D(((1 2) (3 4)) ((5 6) (7 8 9))), #3D(((1 2) (3 4)) (5 (7 8 9))) etc
   *
   * but a special case: #nD() is an n-dimensional empty vector
   */

  if (dims <= 0)      /* #0d(...) #2147483649D() [if dims is int32_t this is negative] */
    return(s7_out_of_range_error(sc, "#nD(...) dimensions", 1, make_integer(sc, dims), "must be 1 or more"));
  if (dims > sc->max_vector_dimensions)
    return(s7_out_of_range_error(sc, "#nD(...) dimensions", 1, make_integer(sc, dims), "must be < 512")); /* sc->max_vector_dimensions=512 currently */

  sc->w = sc->nil;
  if (is_null(data))  /* dims are already 0 (calloc above) */
    return(g_make_vector(sc, set_plist_1(sc, make_list(sc, dims, small_int(0)))));

  sizes = (int32_t *)calloc(dims, sizeof(int32_t));
  for (x = data, i = 0; i < dims; i++)
    {
      sizes[i] = safe_list_length(sc, x);
      sc->w = cons(sc, make_integer(sc, sizes[i]), sc->w);
      x = car(x);
      if ((i < (dims - 1)) &&
	  (!is_pair(x)))
	{
	  free(sizes);
	  return(s7_multivector_error(sc, "we need a list that fully specifies the vector's elements", data));
	}
    }

  vec = g_make_vector(sc, set_plist_1(sc, sc->w = safe_reverse_in_place(sc, sc->w)));
  vec_loc = s7_gc_protect(sc, vec);
  sc->w = sc->nil;

  /* now fill the vector checking that all the lists match */
  err = traverse_vector_data(sc, vec, 0, 0, dims, sizes, data);

  free(sizes);
  s7_gc_unprotect_at(sc, vec_loc);
  if (err < 0)
    return(s7_multivector_error(sc, (err == MULTIVECTOR_TOO_MANY_ELEMENTS) ? "found too many elements" : "not enough elements found", data));

  return(vec);
}

static s7_pointer g_int_multivector(s7_scheme *sc, s7_int dims, s7_pointer data)
{
  /* dims > 1, sc->value is a pair (not null) */
  s7_pointer *src;
  s7_int i, len;
  sc->value = g_multivector(sc, dims, sc->value);
  src = (s7_pointer *)vector_elements(sc->value);
  len = vector_length(sc->value);
  for (i = 0; i < len; i++)
    if (!is_t_integer(src[i]))
      return(s7_wrong_type_arg_error(sc, "#i(...)", i + 1, src[i], "an integer"));
  sc->args = g_make_vector_1(sc, set_plist_3(sc, g_vector_dimensions(sc, set_plist_1(sc, sc->value)), small_int(0), sc->T), sc->make_int_vector_symbol);
  return(s7_copy(sc, set_plist_2(sc, sc->value, sc->args)));
}

static s7_pointer g_float_multivector(s7_scheme *sc, s7_int dims, s7_pointer data)
{
  /* dims > 1, sc->value is a pair (not null) */
  s7_pointer *src;
  s7_int i, len;
  sc->value = g_multivector(sc, dims, sc->value);
  src = (s7_pointer *)vector_elements(sc->value);
  len = vector_length(sc->value);
  for (i = 0; i < len; i++)
    if (!s7_is_real(src[i]))
      return(s7_wrong_type_arg_error(sc, "#r(...)", i + 1, src[i], "a real"));
  sc->args = g_make_vector_1(sc, set_plist_3(sc, g_vector_dimensions(sc, set_plist_1(sc, sc->value)), real_zero, sc->T), sc->make_float_vector_symbol);
  return(s7_copy(sc, set_plist_2(sc, sc->value, sc->args)));
}

#if WITH_VECTORIZE
static s7_pointer s7_vector_copy_1(s7_scheme *sc, s7_pointer old_vect) __attribute__((optimize("tree-vectorize")));
static s7_pointer s7_vector_copy_1(s7_scheme *sc, s7_pointer old_vect)
#else
s7_pointer s7_vector_copy(s7_scheme *sc, s7_pointer old_vect)
#endif
{
  s7_int i, len;
  s7_pointer new_vect;

  len = vector_length(old_vect);
  if (is_float_vector(old_vect))
    {
      s7_double *src, *dst;
      if (vector_rank(old_vect) > 1)
	new_vect = g_make_vector_1(sc, set_plist_3(sc, g_vector_dimensions(sc, set_plist_1(sc, old_vect)), real_zero, sc->T), sc->make_float_vector_symbol);
      else new_vect = make_vector_1(sc, len, NOT_FILLED, T_FLOAT_VECTOR);
      /* if (len > 0) memcpy((void *)(float_vector_elements(new_vect)), (void *)(float_vector_elements(old_vect)), len * sizeof(s7_double)); */
      src = (s7_double *)float_vector_elements(old_vect);
      dst = (s7_double *)float_vector_elements(new_vect);
      for (i = len; i > 0; i--) *dst++ = *src++;
    }
  else
    {
      if (is_int_vector(old_vect))
	{
	  s7_int *src, *dst;
	  if (vector_rank(old_vect) > 1)
	    new_vect = g_make_vector_1(sc, set_plist_3(sc, g_vector_dimensions(sc, set_plist_1(sc, old_vect)), small_int(0), sc->T), sc->make_int_vector_symbol);
	  else new_vect = make_vector_1(sc, len, NOT_FILLED, T_INT_VECTOR);
	  /* if (len > 0) memcpy((void *)(int_vector_elements(new_vect)), (void *)(int_vector_elements(old_vect)), len * sizeof(s7_int)); */
	  src = (s7_int *)int_vector_elements(old_vect);
	  dst = (s7_int *)int_vector_elements(new_vect);
	  for (i = len; i > 0; i--) *dst++ = *src++;
	}
      else
	{
	  s7_pointer *src, *dst;
	  if (vector_rank(old_vect) > 1)
	    new_vect = g_make_vector(sc, set_plist_1(sc, g_vector_dimensions(sc, list_1(sc, old_vect))));
	  else new_vect = make_vector_1(sc, len, NOT_FILLED, T_VECTOR);

	  /* here and in vector-fill! we have a problem with bignums -- should new bignums be allocated? (copy_list also) */
	  /* if (len > 0) memcpy((void *)(vector_elements(new_vect)), (void *)(vector_elements(old_vect)), len * sizeof(s7_pointer)); */
	  src = (s7_pointer *)vector_elements(old_vect);
	  dst = (s7_pointer *)vector_elements(new_vect);
	  for (i = len; i > 0; i--) *dst++ = *src++;
	}
      }
  return(new_vect);
}

#if WITH_VECTORIZE
s7_pointer s7_vector_copy(s7_scheme *sc, s7_pointer old_vect) {return(s7_vector_copy_1(sc, old_vect));}
#endif

static s7_pointer univect_ref(s7_scheme *sc, s7_pointer args, bool flt)
{
  s7_pointer v, caller;
  s7_int ind;
  int32_t typ;

  caller = (flt) ? sc->float_vector_ref_symbol : sc->int_vector_ref_symbol;
  typ = (flt) ? T_FLOAT_VECTOR : T_INT_VECTOR;

  v = car(args);
  if (type(v) != typ)
    method_or_bust(sc, v, caller, args, typ, 1);

  if (vector_rank(v) == 1)
    {
      s7_pointer index;
      index = cadr(args);
      if (!s7_is_integer(index))
	{
	  s7_pointer p;
	  if (!s7_is_integer(p = check_values(sc, index, cdr(args))))
	    return(wrong_type_argument(sc, caller, 2, index, T_INTEGER));
	  else index = p;
	}
      ind = s7_integer(index);
      if ((ind < 0) || (ind >= vector_length(v)))
	return(simple_out_of_range(sc, caller, index, (ind < 0) ? its_negative_string : its_too_large_string));
      if (!is_null(cddr(args)))
	return(out_of_range(sc, caller, small_int(2), cdr(args), too_many_indices_string));
    }
  else
    {
      uint32_t i;
      s7_pointer x;
      ind = 0;
      for (x = cdr(args), i = 0; (is_not_null(x)) && (i < vector_ndims(v)); x = cdr(x), i++)
	{
	  s7_int n;
	  if (!s7_is_integer(car(x)))
	    {
	      s7_pointer p;
	      if (!s7_is_integer(p = check_values(sc, car(x), x)))
		return(wrong_type_argument(sc, caller, i + 2, car(x), T_INTEGER));
	      n = s7_integer(p);
	    }
	  else n = s7_integer(car(x));
	  if ((n < 0) ||
	      (n >= vector_dimension(v, i)))
	    return(out_of_range(sc, caller, make_integer(sc, i + 2), car(x), (n < 0) ? its_negative_string : its_too_large_string));

	  ind += n * vector_offset(v, i);
	}
      if (is_not_null(x))
	return(out_of_range(sc, caller, small_int(2), cdr(args), too_many_indices_string));

      /* if not enough indices, return a shared vector covering whatever is left */
      if (i < vector_ndims(v))
	return(make_shared_vector(sc, v, i, ind));
    }
  if (flt)
    return(make_real(sc, float_vector_element(v, ind)));
  return(make_integer(sc, int_vector_element(v, ind)));
}


static s7_pointer univect_set(s7_scheme *sc, s7_pointer args, bool flt)
{
  s7_pointer vec, val, caller;
  s7_int index;
  int32_t typ;

  caller = (flt) ? sc->float_vector_set_symbol : sc->int_vector_set_symbol;
  typ = (flt) ? T_FLOAT_VECTOR : T_INT_VECTOR;

  vec = car(args);
  if (type(vec) != typ)
    method_or_bust(sc, vec, caller, args, typ, 1);

  if (vector_rank(vec) > 1)
    {
      uint32_t i;
      s7_pointer x;
      index = 0;
      for (x = cdr(args), i = 0; (is_not_null(cdr(x))) && (i < vector_ndims(vec)); x = cdr(x), i++)
	{
	  s7_int n;
	  if (!s7_is_integer(car(x)))
	    {
	      s7_pointer p;
	      if (!s7_is_integer(p = check_values(sc, car(x), x)))
		method_or_bust(sc, car(x), caller, args, T_INTEGER, i + 2);
	      n = s7_integer(p);
	    }
	  else n = s7_integer(car(x));
	  if ((n < 0) ||
	      (n >= vector_dimension(vec, i)))
	    return(out_of_range(sc, caller, make_integer(sc, i + 2), car(x), (n < 0) ? its_negative_string : its_too_large_string));

	  index += n * vector_offset(vec, i);
	}

      if (is_not_null(cdr(x)))
	return(s7_wrong_number_of_args_error(sc, "too many args: ~S", args));
      if (i != vector_ndims(vec))
	return(s7_wrong_number_of_args_error(sc, "not enough args: ~S", args));

      val = car(x);
    }
  else
    {
      s7_pointer p;
      p = cdr(args);
      if (!s7_is_integer(car(p)))
	{
	  s7_pointer z;
	  if (!s7_is_integer(z = check_values(sc, car(p), p)))
	    method_or_bust(sc, car(p), caller, args, T_INTEGER, 2);
	  index = s7_integer(z);
	}
      else index = s7_integer(car(p));
      if ((index < 0) ||
	  (index >= vector_length(vec)))
	return(out_of_range(sc, caller, small_int(2), car(p), (index < 0) ? its_negative_string : its_too_large_string));

      if (is_not_null(cddr(p)))
	return(s7_wrong_number_of_args_error(sc, "too many args: ~S", args));
      val = cadr(p);
    }

  if (flt)
    {
      if (!s7_is_real(val))
	method_or_bust(sc, val, caller, args, T_REAL, 3);
      float_vector_element(vec, index) = real_to_double(sc, val, "float-vector-set!");
    }
  else
    {
      if (!s7_is_integer(val))
	method_or_bust(sc, val, caller, args, T_INTEGER, 3);
      int_vector_element(vec, index) = s7_integer(val);
    }
  return(val);
}


static s7_pointer g_float_vector_ref(s7_scheme *sc, s7_pointer args)
{
  #define H_float_vector_ref "(float-vector-ref v ...) returns an element of the float-vector v."
  #define Q_float_vector_ref s7_make_circular_signature(sc, 2, 3, s7_make_signature(sc, 2, sc->is_float_symbol, sc->is_float_vector_symbol), sc->is_float_vector_symbol, sc->is_integer_symbol)
  return(univect_ref(sc, args, true));
}

static s7_double float_vector_ref_unchecked(s7_pointer v, s7_int i) {return(float_vector_element(v, i));}
static s7_int ref_check_index(s7_pointer v, s7_int i)
{
  /* according to valgrind, it is faster to split out the bounds check */
  if ((i < 0) || (i >= vector_length(v)))
    out_of_range(cur_sc, cur_sc->float_vector_ref_symbol, small_int(2), make_integer(cur_sc, i), (i < 0) ? its_negative_string : its_too_large_string);
  return(i);
}
static s7_double float_vector_ref_d(s7_pointer v, s7_int i) {return(float_vector_element(v, ref_check_index(v, i)));}
static s7_pointer float_vector_ref_unchecked_p(s7_pointer v, s7_int i) {return(float_vector_getter(cur_sc, v, i));}


static s7_pointer g_float_vector_set(s7_scheme *sc, s7_pointer args)
{
  #define H_float_vector_set "(float-vector-set! v i ... value) sets the i-th element of the float-vector v to value."
  #define Q_float_vector_set s7_make_circular_signature(sc, 3, 4, sc->is_real_symbol, sc->is_float_vector_symbol, sc->is_integer_symbol, sc->is_integer_or_real_at_end_symbol)
  return(univect_set(sc, args, true));
}


static s7_double float_vector_set_unchecked(s7_pointer v, s7_int i, s7_double x) {float_vector_element(v, i) = x; return(x);}
static s7_int set_check_index(s7_pointer v, s7_int i)
{
  if ((i < 0) || (i >= vector_length(v)))
    out_of_range(cur_sc, cur_sc->float_vector_set_symbol, small_int(2), make_integer(cur_sc, i), (i < 0) ? its_negative_string : its_too_large_string);
  return(i);
}
static s7_double float_vector_set_d(s7_pointer v, s7_int i, s7_double x) {float_vector_element(v, (set_check_index(v, i))) = x; return(x);}
static s7_pointer float_vector_set_unchecked_p(s7_pointer v, s7_int i, s7_pointer p) {return(float_vector_setter(cur_sc, v, i, p));}


static s7_pointer g_int_vector_ref(s7_scheme *sc, s7_pointer args)
{
  #define H_int_vector_ref "(int-vector-ref v ...) returns an element of the int-vector v."
  #define Q_int_vector_ref s7_make_circular_signature(sc, 2, 3, s7_make_signature(sc, 2, sc->is_integer_symbol, sc->is_int_vector_symbol), sc->is_int_vector_symbol, sc->is_integer_symbol)
  /* actually here and in float-vector-ref, we could return a vector, not an integer: (int-vector-ref #i2d((1 2) (3 4)) 0) -> #i(1 2) */
  return(univect_ref(sc, args, false));
}

static s7_int int_vector_ref_unchecked(s7_pointer v, s7_int i) {return(int_vector_element(v, i));}
static s7_int int_vector_ref_i(s7_pointer v, s7_int i) 
{
  if ((i < 0) || (i >= vector_length(v)))
    out_of_range(cur_sc, cur_sc->int_vector_ref_symbol, small_int(2), make_integer(cur_sc, i), (i < 0) ? its_negative_string : its_too_large_string);
  return(int_vector_element(v, i));
} 
static s7_pointer int_vector_ref_unchecked_p(s7_pointer v, s7_int i) {return(int_vector_getter(cur_sc, v, i));}


static s7_pointer g_int_vector_set(s7_scheme *sc, s7_pointer args)
{
  #define H_int_vector_set "(int-vector-set! v i ... value) sets the i-th element of the int-vector v to value."
  #define Q_int_vector_set s7_make_circular_signature(sc, 2, 3, sc->is_integer_symbol, sc->is_int_vector_symbol, sc->is_integer_symbol)
  return(univect_set(sc, args, false));
}

static s7_int int_vector_set_unchecked(s7_pointer v, s7_int i, s7_int x) {int_vector_element(v, i) = x; return(x);}
static s7_int int_vector_set_i(s7_pointer v, s7_int i, s7_int x) 
{
  if ((i < 0) || (i >= vector_length(v)))
    out_of_range(cur_sc, cur_sc->int_vector_set_symbol, small_int(2), make_integer(cur_sc, i), (i < 0) ? its_negative_string : its_too_large_string);
  int_vector_element(v, i) = x;
  return(x);
} 
static s7_pointer int_vector_set_unchecked_p(s7_pointer v, s7_int i, s7_pointer p) {return(int_vector_setter(cur_sc, v, i, p));}


/* -------------------------------------------------------------------------------- */

static bool c_function_is_ok(s7_scheme *sc, s7_pointer x)
{
  /* macro version of this (below) is much slower! Since this is almost never false, 
   *   I tried __builtin_expect throughout eval below.  The result was not faster.
   */
  s7_pointer p;

  p = car(x); /* function name (symbol) */
  if (is_global(p)) 
    p = slot_value(global_slot(p)); 
  else p = find_symbol_unchecked(sc, p);

  /* this is nearly always global and p == opt_cfunc(x) 
   * p can be null if we evaluate some code, optimizing it, then eval it again in a context
   *   where the incoming p was undefined(!) -- explicit use of eval and so on.
   *   I guess ideally eval would ignore optimization info -- copy :readable or something.
   */
  return((p == opt_any1(x)) ||
	 ((is_any_c_function(p)) &&
	  (c_function_class(p) == c_function_class(opt_cfunc(x)))));
}

static bool arglist_has_rest(s7_scheme *sc, s7_pointer args)
{
  s7_pointer p;
  for (p = args; is_pair(p); p = cdr(p))
    if (car(p) == sc->key_rest_symbol)
      return(true);
  return(!is_null(p));
}


/* -------- sort! -------- */

#if (!WITH_GMP)
static int32_t dbl_less(const void *f1, const void *f2)
{
  if ((*((s7_double *)f1)) < (*((s7_double *)f2))) return(-1);
  if ((*((s7_double *)f1)) > (*((s7_double *)f2))) return(1);
  return(0);
}

static int32_t int_less(const void *f1, const void *f2)
{
  if ((*((s7_int *)f1)) < (*((s7_int *)f2))) return(-1);
  if ((*((s7_int *)f1)) > (*((s7_int *)f2))) return(1);
  return(0);
}

static int32_t dbl_greater(const void *f1, const void *f2) {return(-dbl_less(f1, f2));}
static int32_t int_greater(const void *f1, const void *f2) {return(-int_less(f1, f2));}

static int32_t byte_less(const void *f1, const void *f2)
{
  if ((*((unsigned char *)f1)) < (*((unsigned char *)f2))) return(-1);
  if ((*((unsigned char *)f1)) > (*((unsigned char *)f2))) return(1);
  return(0);
}

static int32_t byte_greater(const void *f1, const void *f2) {return(-byte_less(f1, f2));}

static int32_t dbl_less_2(const void *f1, const void *f2)
{
  s7_pointer p1, p2;
  p1 = (*((s7_pointer *)f1));
  p2 = (*((s7_pointer *)f2));
  if (real(p1) < real(p2)) return(-1);
  if (real(p1) > real(p2)) return(1);
  return(0);
}

static int32_t int_less_2(const void *f1, const void *f2)
{
  s7_pointer p1, p2;
  p1 = (*((s7_pointer *)f1));
  p2 = (*((s7_pointer *)f2));
  if (integer(p1) < integer(p2)) return(-1);
  if (integer(p1) > integer(p2)) return(1);
  return(0);
}

static int32_t dbl_greater_2(const void *f1, const void *f2) {return(-dbl_less_2(f1, f2));}
static int32_t int_greater_2(const void *f1, const void *f2) {return(-int_less_2(f1, f2));}

static int32_t str_less_2(const void *f1, const void *f2)
{
  s7_pointer p1, p2;
  p1 = (*((s7_pointer *)f1));
  p2 = (*((s7_pointer *)f2));
  return(scheme_strcmp(p1, p2));
}

static int32_t str_greater_2(const void *f1, const void *f2) {return(-str_less_2(f1, f2));}

static int32_t chr_less_2(const void *f1, const void *f2)
{
  s7_pointer p1, p2;
  p1 = (*((s7_pointer *)f1));
  p2 = (*((s7_pointer *)f2));
  if (character(p1) < character(p2)) return(-1);
  if (character(p1) > character(p2)) return(1);
  return(0);
}

static int32_t chr_greater_2(const void *f1, const void *f2) {return(-chr_less_2(f1, f2));}
#endif

static s7_scheme *compare_sc;
static s7_function compare_func;
static s7_pointer compare_args, compare_begin, compare_v1, compare_v2;
static opcode_t compare_op;
static int32_t compare_body_len = 0;
static bool p_to_b(void *p);
static int32_t vector_compare(const void *v1, const void *v2)
{
  set_car(compare_args, (*(s7_pointer *)v1));
  set_cadr(compare_args, (*(s7_pointer *)v2));
  return(((*(compare_func))(compare_sc, compare_args) != compare_sc->F) ? -1 : 1);
}

static int32_t vector_car_compare(const void *v1, const void *v2)
{
  s7_pointer a, b;
  a = (*(s7_pointer *)v1);
  b = (*(s7_pointer *)v2);
  set_car(compare_args, (is_pair(a)) ? car(a) : g_car(compare_sc, set_plist_1(compare_sc, a)));
  set_cadr(compare_args, (is_pair(b)) ? car(b) : g_car(compare_sc, set_plist_1(compare_sc, b)));
  return(((*(compare_func))(compare_sc, compare_args) != compare_sc->F) ? -1 : 1);
}

static int32_t vector_cdr_compare(const void *v1, const void *v2)
{
  s7_pointer a, b;
  a = (*(s7_pointer *)v1);
  b = (*(s7_pointer *)v2);
  set_car(compare_args, (is_pair(a)) ? cdr(a) : g_cdr(compare_sc, set_plist_1(compare_sc, a)));
  set_cadr(compare_args, (is_pair(b)) ? cdr(b) : g_cdr(compare_sc, set_plist_1(compare_sc, b)));
  return(((*(compare_func))(compare_sc, compare_args) != compare_sc->F) ? -1 : 1);
}

static int32_t all_x_compare(const void *v1, const void *v2)
{
  slot_set_value(compare_v1, (*(s7_pointer *)v1));
  slot_set_value(compare_v2, (*(s7_pointer *)v2));
  return((compare_func(compare_sc, compare_args) != compare_sc->F) ? -1 : 1);
}

static int32_t opt_bool_compare(const void *v1, const void *v2)
{
  slot_set_value(compare_v1, (*(s7_pointer *)v1));
  slot_set_value(compare_v2, (*(s7_pointer *)v2));
  compare_sc->pc = 0; /* always opt_bool_call here, so insert it */
  return((compare_sc->opts[0]->v7.fb(compare_sc->opts[0])) ? -1 : 1);
}

static int32_t opt_bool_compare_p(const void *v1, const void *v2)
{
  slot_set_value(compare_v1, (*(s7_pointer *)v1));
  slot_set_value(compare_v2, (*(s7_pointer *)v2));
  compare_sc->pc = 0;
  return((compare_sc->opts[0]->v8.fp(compare_sc->opts[0]) == compare_sc->F) ? 1 : -1);
}

static int32_t opt_begin_bool_compare_b(const void *v1, const void *v2)
{
  int32_t i;
  opt_info *o;
  slot_set_value(compare_v1, (*(s7_pointer *)v1));
  slot_set_value(compare_v2, (*(s7_pointer *)v2));
  compare_sc->pc = -1;
  for (i = 0; i < compare_body_len - 1; i++)
    {
      o = compare_sc->opts[++compare_sc->pc];
      o->v7.fp(o);
    }
  o = compare_sc->opts[++compare_sc->pc];
  return((o->v7.fb(o)) ? -1 : 1);
}

static int32_t opt_begin_bool_compare_p(const void *v1, const void *v2)
{
  int32_t i;
  opt_info *o;
  s7_pointer val;
  slot_set_value(compare_v1, (*(s7_pointer *)v1));
  slot_set_value(compare_v2, (*(s7_pointer *)v2));
  compare_sc->pc = -1;
  for (i = 0; i < compare_body_len - 1; i++)
    {
      o = compare_sc->opts[++compare_sc->pc];
      o->v7.fp(o);
    }
  o = compare_sc->opts[++compare_sc->pc];
  val = o->v7.fp(o);
  return((val != compare_sc->F) ? -1 : 1);
}

static int32_t closure_compare(const void *v1, const void *v2)
{
  slot_set_value(compare_v1, (*(s7_pointer *)v1));
  slot_set_value(compare_v2, (*(s7_pointer *)v2));
  push_stack(compare_sc, OP_EVAL_DONE, compare_sc->args, compare_sc->code);
  compare_sc->code = compare_args; /* this should be ok because we checked in advance that it is a safe closure (no sort! for example) */
  eval(compare_sc, compare_op);
  return((compare_sc->value != compare_sc->F) ? -1 : 1);
}

static int32_t closure_compare_begin(const void *v1, const void *v2)
{
  slot_set_value(compare_v1, (*(s7_pointer *)v1));
  slot_set_value(compare_v2, (*(s7_pointer *)v2));
  push_stack(compare_sc, OP_EVAL_DONE, compare_sc->args, compare_sc->code);
  push_stack_no_args(compare_sc, OP_BEGIN1, compare_begin);
  compare_sc->code = compare_args;
  eval(compare_sc, compare_op);
  return((compare_sc->value != compare_sc->F) ? -1 : 1);
}

static s7_pointer g_sort(s7_scheme *sc, s7_pointer args)
{
  #define H_sort "(sort! sequence less?) sorts a sequence using the function 'less?' to compare elements."
  #define Q_sort s7_make_signature(sc, 3, sc->is_sequence_symbol, sc->is_sequence_symbol, sc->is_procedure_symbol)

  s7_pointer data, lessp, lx;
  s7_int len = 0, n, k;
  int32_t (*sort_func)(const void *v1, const void *v2);
  s7_pointer *elements;
  uint32_t gc_loc = 0;

  /* both the intermediate vector (if any) and the current args pointer need GC protection,
   *   but it is a real bother to unprotect args at every return statement, so I'll use temp3
   */
  sc->temp3 = args; /* this is needed! */
  data = car(args);
  if (is_null(data))
    {
      /* (apply sort! () #f) should be an error I think */
      lessp = cadr(args);
      if (type(lessp) < T_GOTO)
	method_or_bust_with_type(sc, lessp, sc->sort_symbol, args, a_procedure_string, 2);
      if (!s7_is_aritable(sc, lessp, 2))
	return(wrong_type_argument_with_type(sc, sc->sort_symbol, 2, lessp, an_eq_func_string));
      return(sc->nil);
    }

  if ((sc->safety > NO_SAFETY) &&
      (is_immutable(data)))
    return(s7_error(sc, sc->error_symbol, set_elist_2(sc, s7_make_string_wrapper(sc, "can't sort! ~S (it is immutable)"), data)));

  lessp = cadr(args);
  if (type(lessp) < T_GOTO)
    method_or_bust_with_type(sc, lessp, sc->sort_symbol, args, a_procedure_string, 2);
  if (!s7_is_aritable(sc, lessp, 2))
    return(wrong_type_argument_with_type(sc, sc->sort_symbol, 2, lessp, an_eq_func_string));

  if ((is_continuation(lessp)) || is_goto(lessp))
    return(wrong_type_argument_with_type(sc, sc->sort_symbol, 2, lessp, a_normal_procedure_string));

  sort_func = vector_compare;
  compare_func = NULL;
  compare_args = sc->t2_1;
  compare_sc = sc;

  if ((is_safe_procedure(lessp)) &&     /* (sort! a <) */
      (is_c_function(lessp)))
    {
      s7_pointer sig;
      sig = c_function_signature(lessp);
      if ((sig) &&
	  (is_pair(sig)) &&
	  (car(sig) != sc->is_boolean_symbol))
	return(wrong_type_argument_with_type(sc, sc->sort_symbol, 2, lessp, s7_make_string_wrapper(sc, "sort! function should return a boolean")));
      compare_func = c_function_call(lessp);
    }
  else
    {
      if (is_closure(lessp))
	{
	  s7_pointer expr, largs;
	  expr = car(closure_body(lessp));
	  largs = closure_args(lessp);

	  if ((is_pair(largs)) &&       /* closure args not a symbol, etc */
	      (!arglist_has_rest(sc, largs)))
	    {
	      if (is_null(cdr(closure_body(lessp))))
		{
		  if ((is_optimized(expr)) &&
		      /* since (sort seq (lambda (a b) ...)) can't return a "safe closure" (the hop bit is off in
		       *   optimize in this case, for some arcane reason), the optimized expression won't be hop_safe,
		       *   but that is irrelevant at this point -- if c_function_is_ok, we're good to go.
		       */
		      (((optimize_op(expr) & 1) != 0) ||
		       ((is_global(car(expr))) &&            /* (sort! x (lambda (car y) (car x)...))! */
			(c_function_is_ok(sc, expr)))))
		    {
		      int32_t orig_data;
		      orig_data = optimize_op(expr);
		      set_optimize_op(expr, optimize_op(expr) | 1);
		      if (((optimize_op(expr) == HOP_SAFE_C_SS) || (optimize_op(expr) == HOP_SAFE_C_LL)) &&
			  (car(largs) == cadr(expr)) &&
			  (cadr(largs) == caddr(expr)))
			{
			  lessp = find_symbol_unchecked(sc, car(expr));
			  compare_func = c_function_call(lessp);
			}
		      else
			{
			  if ((optimize_op(expr) == HOP_SAFE_C_opSq_opSq) &&
			      ((caadr(expr) == sc->car_symbol) || (caadr(expr) == sc->cdr_symbol)) &&
			      (caadr(expr) == caaddr(expr)) &&
			      (car(largs) == cadadr(expr)) &&
			      (cadr(largs) == cadr(caddr(expr))))
			    {
			      lessp = find_symbol_unchecked(sc, car(expr));
			      compare_func = c_function_call(lessp);
			      sort_func = ((caadr(expr) == sc->car_symbol) ? vector_car_compare : vector_cdr_compare);
			    }
			}
		      set_optimize_op(expr, orig_data);
		    }
		}
		  
	      if (!compare_func)
		{
		  s7_pointer init_val, old_e;
		  if (is_float_vector(data))
		    init_val = real_zero;
		  else
		    {
		      if (is_int_vector(data))
			init_val = small_int(0);
		      else init_val = sc->F;
		    }
		  old_e = sc->envir;
		  new_frame_with_two_slots(sc, closure_let(lessp), sc->envir, car(largs), init_val, cadr(largs), init_val);
		  compare_args = expr;
		  compare_v1 = let_slots(sc->envir);
		  compare_v2 = next_slot(let_slots(sc->envir));
		  if (is_null(cdr(closure_body(lessp))))
		    {
		      compare_func = s7_bool_optimize(sc, closure_body(lessp));
		      if (compare_func)
			{
			  if (compare_func == opt_bool_any)
			    {
			      if (sc->opts[0]->v7.fb == p_to_b)
				sort_func = opt_bool_compare_p;
			      else sort_func = opt_bool_compare;
			    }
			  else sort_func = all_x_compare;
			}
		    }
		  else
		    {
		      if (setjmp(sc->opt_exit) == 0)
			{
			  s7_pointer p;
			  compare_body_len = s7_list_length(sc, closure_body(lessp));
			  sc->pc = 0;
			  for (p = closure_body(lessp); is_pair(cdr(p)); p = cdr(p))
			    if (!cell_optimize(sc, p))
			      break;
			  if (is_null(cdr(p)))
			    {
			      int32_t start;
			      start = sc->pc;
			      if (bool_optimize_nw(sc, p))
				{
				  compare_func = opt_bool_any;
				  sort_func = opt_begin_bool_compare_b;
				}
			      else
				{
				  pc_fallback(sc, start);
				  if (cell_optimize(sc, p))
				    {
				      compare_func = opt_bool_any;
				      sort_func = opt_begin_bool_compare_p;
				    }
				}
			    }
			}
		    }
		  if (!compare_func)
		    sc->envir = old_e;
		}
	      
	      if ((!compare_func) &&
		  (is_safe_closure(lessp))) /* no embedded sort! or call/cc, etc */
		{
		  new_frame_with_two_slots(sc, closure_let(lessp), sc->envir, car(largs), sc->F, cadr(largs), sc->F);
		  compare_func = (s7_function)lessp;       /* not used -- just a flag */
		  compare_args = car(closure_body(lessp));
		  compare_begin = cdr(closure_body(lessp));
		  if (is_null(compare_begin))
		    sort_func = closure_compare;
		  else sort_func = closure_compare_begin;
		  if (typesflag(compare_args) == SYNTACTIC_PAIR)
		    {
		      compare_op = (opcode_t)pair_syntax_op(compare_args);
		      compare_args = cdr(compare_args);
		    }
		  else compare_op = OP_EVAL;
		  compare_v1 = let_slots(sc->envir);
		  compare_v2 = next_slot(let_slots(sc->envir));
		}
	    }
	}
    }

  if (compare_func == g_strings_are_less) 
    compare_func = g_string_less_2;
  else
    {
      if (compare_func == g_strings_are_greater) 
	compare_func = g_string_greater_2;
      else
	{
	  if (compare_func == g_chars_are_less)
	    compare_func = g_char_less_2;
	  else
	    {
	      if (compare_func == g_chars_are_greater)
		compare_func = g_char_greater_2;
	    }
	}
    }
#if (!WITH_GMP)
  if (compare_func == g_less)
    compare_func = g_less_2;
  else
    {
      if (compare_func == g_greater)
	compare_func = g_greater_2;
    }
#endif

  switch (type(data))
    {
    case T_PAIR:
      len = s7_list_length(sc, data);            /* 0 here == infinite */
      if (len <= 0)
	return(s7_error(sc, sc->wrong_type_arg_symbol, set_elist_2(sc, s7_make_string_wrapper(sc, "sort! argument 1 should be a proper list: ~S"), data)));
      if (len < 2)
	return(data);
      if (compare_func)
	{
	  s7_int i;
	  s7_pointer vec, p;

	  vec = g_vector(sc, data);
	  gc_loc = s7_gc_protect(sc, vec);
	  elements = s7_vector_elements(vec);

	  sc->v = vec;
	  qsort((void *)elements, len, sizeof(s7_pointer), sort_func);
	  for (p = data, i = 0; i < len; i++, p = cdr(p))
	    set_car(p, elements[i]);

	  s7_gc_unprotect_at(sc, gc_loc);
	  return(data);
	}

      push_stack(sc, OP_SORT_PAIR_END, cons(sc, data, lessp), sc->code); /* save and gc protect the original list and func */
      set_car(args, g_vector(sc, data));
      break;

    case T_STRING:
      {
	/* byte-vectors here also, so this isn't completely silly */
	s7_int i;
	s7_pointer vec;
	unsigned char *chrs;

	len = string_length(data);
	if (len < 2) 
	  return(data);

#if (!WITH_GMP)
	if (is_c_function(lessp))
	  {
	    if (((!is_byte_vector(data)) && (compare_func == g_char_less_2)) ||
		((is_byte_vector(data)) && (compare_func == g_less_2)))
	      {
		qsort((void *)vector_elements(data), len, sizeof(unsigned char), byte_less);
		return(data);
	      }
	    if (((!is_byte_vector(data)) && (compare_func == g_char_greater_2)) ||
		((is_byte_vector(data)) && (compare_func == g_greater_2)))
	      {
		qsort((void *)vector_elements(data), len, sizeof(unsigned char), byte_greater);
		return(data);
	      }
	  }
#endif

	vec = make_vector_1(sc, len, NOT_FILLED, T_VECTOR);
	gc_loc = s7_gc_protect(sc, vec);
	elements = s7_vector_elements(vec);
	chrs = (unsigned char *)string_value(data);

	if (is_byte_vector(data))
	  {
	    for (i = 0; i < len; i++)
	      elements[i] = small_int(chrs[i]);
	  }
	else
	  {
	    for (i = 0; i < len; i++)
	      elements[i] = chars[chrs[i]];
	  }

	if (compare_func)
	  {
	    sc->v = vec;
	    qsort((void *)elements, len, sizeof(s7_pointer), sort_func);

	    if (is_byte_vector(data))
	      {
		for (i = 0; i < len; i++)
		  chrs[i] = (char)integer(elements[i]);
	      }
	    else
	      {
		for (i = 0; i < len; i++)
		  chrs[i] = character(elements[i]);
	      }
	    s7_gc_unprotect_at(sc, gc_loc);
	    return(data);
	  }

	push_stack(sc, OP_SORT_STRING_END, cons(sc, data, lessp), sc->code);
	set_car(args, vec);
	s7_gc_unprotect_at(sc, gc_loc);
      }
      break;

    case T_INT_VECTOR:
    case T_FLOAT_VECTOR:
      {
	s7_int i;
	s7_pointer vec;

	len = vector_length(data);
	if (len < 2) 
	  return(data);
#if (!WITH_GMP)
	if (is_c_function(lessp))
	  {
	    if (compare_func == g_less_2)
	      {
		if (is_float_vector(data))
		  qsort((void *)vector_elements(data), len, sizeof(s7_double), dbl_less);
		else qsort((void *)vector_elements(data), len, sizeof(s7_int), int_less);
		return(data);
	      }
	    if (compare_func == g_greater_2)
	      {
		if (is_float_vector(data))
		  qsort((void *)vector_elements(data), len, sizeof(s7_double), dbl_greater);
		else qsort((void *)vector_elements(data), len, sizeof(s7_int), int_greater);
		return(data);
	      }
	  }
#endif

	/* currently we have to make the ordinary vector here even if not compare_func
	 *   because the sorter uses vector_element to access sort args (see SORT_DATA in eval).
	 *   This is probably better than passing down getter/setter (fewer allocations).
	 *   get/set macro in eval is SORT_DATA(k) then s7_vector_to_list if pair at start (sort_*_end)
	 */
	vec = make_vector_1(sc, len, FILLED, T_VECTOR);
	/* we need this vector prefilled because vector_getter below makes reals/int, causing possible GC
	 *   at any time during that loop, and the GC mark process expects the vector to have an s7_pointer
	 *   at every element.
	 */
	gc_loc = s7_gc_protect(sc, vec);
	elements = s7_vector_elements(vec);

	for (i = 0; i < len; i++)
	  elements[i] = vector_getter(data)(sc, data, i);

	if (compare_func)
	  {
	    sc->v = vec;
	    qsort((void *)elements, len, sizeof(s7_pointer), sort_func);

	    for (i = 0; i < len; i++)
	      vector_setter(data)(sc, data, i, elements[i]);

	    s7_gc_unprotect_at(sc, gc_loc);
	    return(data);
	  }

	push_stack(sc, OP_SORT_VECTOR_END, cons(sc, data, lessp), sc->code); /* save and gc protect the original homogeneous vector and func */
	set_car(args, vec);
	s7_gc_unprotect_at(sc, gc_loc);
      }
      break;

    case T_VECTOR:
      len = vector_length(data);
      if (len < 2) 
	return(data);
      if (compare_func)
	{
	  /* here if, for example, compare_func == string<?, we could precheck for strings,
	   *   then qsort without the type checks.  Also common is (lambda (a b) (f (car a) (car b))).
	   */
#if (!WITH_GMP)
	  if ((compare_func == g_less_2) || (compare_func == g_greater_2) ||
	      (compare_func == g_string_less_2) || (compare_func == g_string_greater_2) ||
	      (compare_func == g_char_less_2) || (compare_func == g_char_greater_2))
	    {
	      int32_t typ;
	      s7_pointer *els;
	      els = s7_vector_elements(data);
	      typ = type(els[0]);
	      if ((typ == T_INTEGER) || (typ == T_REAL) || (typ == T_STRING) || (typ == T_CHARACTER))
		{
		  s7_int i;
		  for (i = 1; i < len; i++)
		    if (type(els[i]) != typ)
		      {
			typ = T_FREE;
			break;
		      }
		}
	      if ((compare_func == g_less_2) || (compare_func == g_greater_2))
		{
		  if (typ == T_INTEGER)
		    {
		      qsort((void *)els, len, sizeof(s7_pointer), ((compare_func == g_less_2) ? int_less_2 : int_greater_2));
		      return(data);
		    }
		  if (typ == T_REAL)
		    {
		      qsort((void *)els, len, sizeof(s7_pointer), ((compare_func == g_less_2) ? dbl_less_2 : dbl_greater_2));
		      return(data);
		    }
		}
	      if ((typ == T_STRING) &&
		  ((compare_func == g_string_less_2) || (compare_func == g_string_greater_2)))
		{
		  qsort((void *)els, len, sizeof(s7_pointer), ((compare_func == g_string_less_2) ? str_less_2 : str_greater_2));
		  return(data);
		}
	      if ((typ == T_CHARACTER) &&
		  ((compare_func == g_char_less_2) || (compare_func == g_char_greater_2)))
		{
		  qsort((void *)els, len, sizeof(s7_pointer), ((compare_func == g_char_less_2) ? chr_less_2 : chr_greater_2));
		  return(data);
		}
	    }
#endif
	  qsort((void *)s7_vector_elements(data), len, sizeof(s7_pointer), sort_func);
	  return(data);
	}
      break;

    default:
      method_or_bust_with_type(sc, data, sc->sort_symbol, args, a_sequence_string, 1);
    }

  n = len - 1;
  k = ((int32_t)(n / 2)) + 1;

  lx = s7_make_vector(sc, (sc->safety == NO_SAFETY) ? 4 : 6);
  gc_loc = s7_gc_protect(sc, lx);
  sc->v = lx;

  vector_element(lx, 0) = make_mutable_integer(sc, n);
  vector_element(lx, 1) = make_mutable_integer(sc, k);
  vector_element(lx, 2) = make_mutable_integer(sc, 0);
  vector_element(lx, 3) = make_mutable_integer(sc, 0);
  if (sc->safety > NO_SAFETY)
    {
      vector_element(lx, 4) = make_mutable_integer(sc, 0);
      vector_element(lx, 5) = make_integer(sc, n * n);
    }
  push_stack(sc, OP_SORT, args, lx);
  s7_gc_unprotect_at(sc, gc_loc);

  return(sc->F);
  /* if the comparison function waffles, sort! can hang: (sort! '(1 2 3) (lambda (a b) (= a b)))
   * set 'safety to 1 to add a check for this loop, but the "safe" procedures are direct, so unchecked.
   */
}


/* these are for the eval sort -- sort a vector, then if necessary put that data into the original sequence */
static s7_pointer vector_into_list(s7_pointer vect, s7_pointer lst)
{
  s7_pointer p;
  s7_pointer *elements;
  int32_t i, len;

  elements = s7_vector_elements(vect);
  len = vector_length(vect);
  for (i = 0, p = lst; i < len; i++, p = cdr(p))
    set_car(p, elements[i]);
  return(lst);
}

static s7_pointer vector_into_fi_vector(s7_pointer source, s7_pointer dest)
{
  s7_pointer *elements;
  int32_t i, len;

  elements = s7_vector_elements(source);
  len = vector_length(source);
  
  if (is_float_vector(dest))
    {
      s7_double *flts;
      flts = float_vector_elements(dest);
      for (i = 0; i < len; i++)
	flts[i] = real(elements[i]);
    }
  else
    {
      s7_int *ints;
      ints = int_vector_elements(dest);
      for (i = 0; i < len; i++)
	ints[i] = integer(elements[i]);
    }
  return(dest);
}

static s7_pointer vector_into_string(s7_pointer vect, s7_pointer dest)
{
  s7_pointer *elements;
  int32_t i, len;
  unsigned char *str;

  elements = s7_vector_elements(vect);
  len = vector_length(vect);
  str = (unsigned char *)string_value(dest);
  
  if (is_byte_vector(dest))
    {
      for (i = 0; i < len; i++)
	str[i] = (unsigned char)integer(elements[i]);
    }
  else
    {
      for (i = 0; i < len; i++)
	str[i] = character(elements[i]);
    }
  return(dest);
}



/* -------- hash tables -------- */

static hash_entry_t *hash_free_list = NULL;

static void free_hash_table(s7_pointer table)
{
  hash_entry_t **entries;
  entries = hash_table_elements(table);

  if (hash_table_entries(table) > 0)
    {
      uint32_t i, len;
      len = hash_table_mask(table) + 1;
      for (i = 0; i < len; i++)
	{
	  hash_entry_t *p, *n;
	  for (p = entries[i++]; p; p = n)
	    {
	      n = p->next;
	      p->next = hash_free_list;
	      hash_free_list = p;
	    }
	  for (p = entries[i]; p; p = n)
	    {
	      n = p->next;
	      p->next = hash_free_list;
	      hash_free_list = p;
	    }
	}
    }
  free(entries);
}

static hash_entry_t *make_hash_entry(s7_pointer key, s7_pointer value, uint32_t raw_hash)
{
  hash_entry_t *p;
  if (hash_free_list)
    {
      p = hash_free_list;
      hash_free_list = p->next;
    }
  else p = (hash_entry_t *)malloc(sizeof(hash_entry_t));
  p->key = key;
  p->value = value;
  p->raw_hash = raw_hash;
  return(p);
}


/* -------------------------------- hash-table? -------------------------------- */
bool s7_is_hash_table(s7_pointer p)
{
  return(is_hash_table(p));
}

static s7_pointer g_is_hash_table(s7_scheme *sc, s7_pointer args)
{
  #define H_is_hash_table "(hash-table? obj) returns #t if obj is a hash-table"
  #define Q_is_hash_table pl_bt
  check_boolean_method(sc, is_hash_table, sc->is_hash_table_symbol, args);
}


/* -------------------------------- hash-table-entries -------------------------------- */
static s7_pointer g_hash_table_entries(s7_scheme *sc, s7_pointer args)
{
  #define H_hash_table_entries "(hash-table-entries obj) returns the number of entries in the hash-table obj"
  #define Q_hash_table_entries s7_make_signature(sc, 2, sc->is_integer_symbol, sc->is_hash_table_symbol)

  if (!is_hash_table(car(args)))
    method_or_bust_one_arg(sc, car(args), sc->hash_table_entries_symbol, args, T_HASH_TABLE);
  return(make_integer(sc, hash_table_entries(car(args))));
}

static s7_int hash_table_entries_i(s7_pointer p) 
{
  if (!is_hash_table(p)) 
    simple_wrong_type_argument(cur_sc, cur_sc->hash_table_entries_symbol, p, T_HASH_TABLE);
  return(hash_table_entries(p));
}


/* ---------------- mappers ---------------- */
static uint32_t hash_float_location(s7_double x)
{
  int32_t loc;
#if defined(__clang__)
  if ((is_inf(x)) || (is_NaN(x))) return(0);
#endif
  x = fabs(x);
  if (x < 100.0)
    loc = 1000.0 * x;     /* this means hash_table_float_epsilon only works if it is less than about .001 */
  else loc = x;

  if (loc < 0)
    return(0);
  return(loc);
}

/* built in hash loc tables for eq? eqv? equal? morally-equal? = string=? string-ci=? char=? char-ci=? (default=equal?) */

#define hash_loc(Sc, Table, Key) (*(hash_table_mapper(Table)[type(Key)]))(Sc, Table, Key)

static hash_map_t *eq_hash_map, *eqv_hash_map, *string_eq_hash_map, *number_eq_hash_map, *char_eq_hash_map, *closure_hash_map;
static hash_map_t *morally_equal_hash_map, *c_function_hash_map;
#if (!WITH_PURE_S7)
static hash_map_t *string_ci_eq_hash_map, *char_ci_eq_hash_map;
#endif

static uint32_t hash_map_nil(s7_scheme *sc, s7_pointer table, s7_pointer key)     {return(type(key));}
static uint32_t hash_map_int(s7_scheme *sc, s7_pointer table, s7_pointer key)     {return((uint32_t)(s7_int_abs(integer(key))));}
static uint32_t hash_map_char(s7_scheme *sc, s7_pointer table, s7_pointer key)    {return(character(key));}
static uint32_t hash_map_ratio(s7_scheme *sc, s7_pointer table, s7_pointer key)   {return((uint32_t)denominator(key));} /* overflow possible as elsewhere */
static uint32_t hash_map_complex(s7_scheme *sc, s7_pointer table, s7_pointer key) {return(hash_float_location(real_part(key)));}
static uint32_t hash_map_symbol(s7_scheme *sc, s7_pointer table, s7_pointer key)  {return(symbol_hmap(key));}
static uint32_t hash_map_syntax(s7_scheme *sc, s7_pointer table, s7_pointer key)  {return(symbol_hmap(syntax_symbol(key)));}

#if WITH_GMP
static uint32_t hash_map_big_int(s7_scheme *sc, s7_pointer table, s7_pointer key)     
{
  return((uint32_t)(big_integer_to_s7_int(big_integer(key))));
}

static uint32_t hash_map_big_ratio(s7_scheme *sc, s7_pointer table, s7_pointer key)   
{
  return((uint32_t)(big_integer_to_s7_int(mpq_denref(big_ratio(key)))));
}

static uint32_t hash_map_big_real(s7_scheme *sc, s7_pointer table, s7_pointer key)    
{
  return((uint32_t)mpfr_get_d(big_real(key), GMP_RNDN));
}

static uint32_t hash_map_big_complex(s7_scheme *sc, s7_pointer table, s7_pointer key) 
{
  return((uint32_t)mpfr_get_d(mpc_realref(big_complex(key)), GMP_RNDN));
}
#endif

static uint32_t hash_map_string(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  if (string_hash(key) == 0)
    string_hash(key) = raw_string_hash((const unsigned char *)string_value(key), string_length(key));
  return(string_hash(key));
}

#if (!WITH_PURE_S7)
static uint32_t hash_map_ci_char(s7_scheme *sc, s7_pointer table, s7_pointer key) {return(upper_character(key));}

static uint32_t hash_map_ci_string(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  int32_t len;
  len = string_length(key);
  if (len == 0) return(0);
  return(len + (uppers[(int32_t)(string_value(key)[0])] << 4));
}
#endif

static uint32_t hash_map_real(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  return(hash_float_location(real(key)));
  /* currently 1e300 goes to most-negative-fixnum! -> 0 after logand size, I hope
   *
   * we need round, not floor for the location calculation in the real/complex cases else
   *    1-eps doesn't match 1.0, but 1+eps does.  And what if round(val) is too big for int?
   *    lrint is complex and requires special compiler flags to get any speed (-fno-math-errno).
   *    all we need is (int32_t)(val+0.5) -- all the other stuff is pointless in this context
   */
}

static uint32_t hash_map_real_eq(s7_scheme *sc, s7_pointer table, s7_pointer x)
{
  if (real(x) < 0.0)
    return((uint32_t)(s7_round(-real(x))));
  return((uint32_t)s7_round(real(x)));
}

static uint32_t hash_map_ratio_eq(s7_scheme *sc, s7_pointer table, s7_pointer y)
{
  s7_double x;
  x = fraction(y);
  if (x < 0.0)
    return((uint32_t)s7_round(-x));
  return((uint32_t)s7_round(x));
}

static uint32_t hash_map_hash_table(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  /* hash-tables are equal if key/values match independent of table size and entry order.
   * if not using morally-equal?, hash_table_checker|mapper must also be the same.
   * Keys are supposed to be constant while keys, so a hash-table shouldn't be a key of itself.
   */
  return(hash_table_entries(key));
}

static uint32_t hash_map_int_vector(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  if (vector_length(key) == 0)
    return(0);
  if (vector_length(key) == 1)
    return((uint32_t)(s7_int_abs(int_vector_element(key, 0))));
  return((uint32_t)(vector_length(key) + s7_int_abs(int_vector_element(key, 0)) + s7_int_abs(int_vector_element(key, 1))));
}

static uint32_t hash_map_float_vector(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  if (vector_length(key) == 0)
    return(0);
  if (vector_length(key) == 1)
    return(hash_float_location(float_vector_element(key, 0)));
  return((uint32_t)(vector_length(key) + hash_float_location(float_vector_element(key, 0)) + hash_float_location(float_vector_element(key, 1))));
}

static uint32_t hash_map_vector(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  if ((vector_length(key) == 0) ||
      (is_sequence(vector_element(key, 0))))
    return(vector_length(key));
  if ((vector_length(key) == 1) ||
      (is_sequence(vector_element(key, 1))))
    return(hash_loc(sc, table, vector_element(key, 0)));
  return(vector_length(key) + hash_loc(sc, table, vector_element(key, 0)) + hash_loc(sc, table, vector_element(key, 1)));
}

static uint32_t hash_map_eq(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  int32_t x;
  x = heap_location(key);
  if (x < 0) return(-x);
  return(x);
}

static uint32_t hash_map_closure(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  s7_pointer f, old_e, args, body;

  f = hash_table_procedures_mapper(table);
  old_e = sc->envir;
  args = closure_args(f);
  body = closure_body(f);
  new_frame_with_slot(sc, closure_let(f), sc->envir, (is_symbol(car(args))) ? car(args) : caar(args), key);
  push_stack(sc, OP_EVAL_DONE, sc->args, sc->code);
  if (is_pair(cdr(body)))
    push_stack_no_args(sc, OP_BEGIN1, cdr(body));
  sc->code = car(body);
  eval(sc, OP_EVAL);
  sc->envir = old_e;
  return(integer(sc->value));
}

static uint32_t hash_map_c_function(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  s7_function f;
  f = c_function_call(hash_table_procedures_mapper(table));
  set_car(sc->t1_1, key);
  return(integer(f(sc, sc->t1_1)));
}

static uint32_t hash_map_let(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  /* lets are equal if same symbol/value pairs, independent of order, taking into account shadowing
   *   (length (inlet 'a 1 'a 2)) = 2
   * but this counts as just one entry from equal?'s point of view, so if more than one entry, we have a problem.
   *   (equal? (inlet 'a 1) (inlet 'a 3 'a 2 'a 1)) = #t
   * also currently equal? follows outlet, but that is ridiculous here, so in this case hash equal?
   *   is not the same as equal?  Surely anyone using lets as keys wants eq?
   */
  s7_pointer slot;
  int32_t slots;

  if ((key == sc->rootlet) ||
      (!is_slot(let_slots(key))))
    return(0);
  slot = let_slots(key);
  if (!is_slot(next_slot(slot)))
    {
      if (is_sequence(slot_value(slot))) /* avoid loop if cycles */
	return(symbol_hmap(slot_symbol(slot)));
      return(symbol_hmap(slot_symbol(slot)) + hash_loc(sc, table, slot_value(slot)));
    }
  slots = 0;
  for (; is_slot(slot); slot = next_slot(slot))
    if (!is_matched_symbol(slot_symbol(slot)))
      {
	set_match_symbol(slot_symbol(slot));
	slots++;
      }
  for (slot = let_slots(key); is_slot(slot); slot = next_slot(slot))
    clear_match_symbol(slot_symbol(slot));

  if (slots == 1)
    {
      slot = let_slots(key);
      if (is_sequence(slot_value(slot))) /* avoid loop if cycles */
	return(symbol_hmap(slot_symbol(slot)));
      return(symbol_hmap(slot_symbol(slot)) + hash_loc(sc, table, slot_value(slot)));
    }
  
  return(slots);
}

static uint32_t hash_map_pair(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  /* len+loc(car) is not horrible, but it means (for example) every list '(set! ...) is hashed to the same location,
   *   so at least we need to take cadr into account if possible.  Better would combine the list_length
   *   with stats like symbols/pairs/constants at top level, then use those to spread it out over all the locs.
   */
  s7_pointer p1;
  uint32_t loc = 0;

  if (!is_sequence(car(key)))
    loc = hash_loc(sc, table, car(key)) + 1;
  else 
    {
      if ((is_pair(car(key))) &&
	  (!is_sequence(caar(key))))
	loc = hash_loc(sc, table, caar(key)) + 1;
    }
  p1 = cdr(key);
  if (is_pair(p1))
    {
      if (!is_sequence(car(p1)))
	loc += hash_loc(sc, table, car(p1)) + 1;
      else
	{
	  if ((is_pair(car(p1))) &&
	      (!is_sequence(caar(p1))))
	    loc += hash_loc(sc, table, caar(p1)) + 1;
	}
    }
  return(loc);
}


/* ---------------- checkers ---------------- */
static hash_entry_t *hash_empty(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  return(NULL);
}


static hash_entry_t *hash_int(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  if (is_integer(key))
    {
      s7_int keyval;
      hash_entry_t *x;
      uint32_t loc, hash_len;

      hash_len = hash_table_mask(table);
      keyval = integer(key);
      if (keyval < 0)
	loc = (uint32_t)((-keyval) & hash_len);
      else loc = (uint32_t)(keyval & hash_len); 
      /* I think this assumes hash_map_int is using s7_int_abs (and high order bits are ignored) */

      for (x = hash_table_element(table, loc); x; x = x->next)
	if (integer(x->key) == keyval)
	  return(x);
    }
  return(NULL);
}


static hash_entry_t *hash_string(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  if (is_string(key))
    {
      hash_entry_t *x;
      uint32_t hash_len, key_len;
      uint64_t hash;
      const char *key_str;

      key_len = string_length(key);
      key_str = string_value(key);

      hash_len = hash_table_mask(table);
      if (string_hash(key) == 0)
	string_hash(key) = raw_string_hash((const unsigned char *)string_value(key), string_length(key));
      hash = string_hash(key);

      if (key_len <= 8)
	{
	  for (x = hash_table_element(table, hash & hash_len); x; x = x->next)
	    if ((hash == string_hash(x->key)) &&
		(key_len == string_length(x->key)))
	      return(x);
	}
      else
	{
	  for (x = hash_table_element(table, hash & hash_len); x; x = x->next)
	    if ((hash == string_hash(x->key)) &&
		(key_len == string_length(x->key)) &&        /* these are scheme strings, so we can't assume 0=end of string */
		(strings_are_equal_with_length(key_str, string_value(x->key), key_len)))
	      return(x);
	}
    }
  return(NULL);
}

#if (!WITH_PURE_S7)
static hash_entry_t *hash_ci_string(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  if (is_string(key))
    {
      hash_entry_t *x;
      uint32_t hash, hash_len;

      hash_len = hash_table_mask(table);
      hash = hash_map_ci_string(sc, table, key);

      for (x = hash_table_element(table, hash & hash_len); x; x = x->next)
	if (scheme_strequal_ci(key, x->key))
	  return(x);
    }
  return(NULL);
}

static hash_entry_t *hash_ci_char(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  if (s7_is_character(key))
    {
      hash_entry_t *x;
      uint32_t hash_len, loc;

      hash_len = hash_table_mask(table);
      loc = hash_loc(sc, table, key) & hash_len;

      for (x = hash_table_element(table, loc); x; x = x->next)
	if (upper_character(key) == upper_character(x->key))
	  return(x);
    }
  return(NULL);
}
#endif

static hash_entry_t *hash_float_1(s7_scheme *sc, s7_pointer table, uint32_t loc, s7_double keyval)
{
  hash_entry_t *x;
  bool look_for_nan;
  look_for_nan = is_NaN(keyval);

  for (x = hash_table_element(table, loc); x; x = x->next)
    {
      if (is_t_real(x->key)) /* we're possibly called from hash_equal, so keys might not be T_REAL */
	{
	  s7_double val;
	  val = real(x->key);
	  if (look_for_nan)
	    {
	      if (is_NaN(val))
		return(x);
	    }
	  else
	    {
	      if ((val == keyval) ||   /* inf case */
		  (fabs(val - keyval) < sc->hash_table_float_epsilon))
		return(x);
	    }
	}
    }
  return(NULL);
}


static hash_entry_t *hash_float(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  /* give the equality check some room. also inf == inf and nan == nan
   */
  if (type(key) == T_REAL)
    {
      s7_double keyval;
      uint32_t hash_len, loc;
      
      hash_len = hash_table_mask(table);
      keyval = real(key);
      loc = hash_float_location(keyval) & hash_len;
      
      return(hash_float_1(sc, table, loc, keyval));
    }
  return(NULL);
}


static hash_entry_t *hash_complex_1(s7_scheme *sc, s7_pointer table, uint32_t loc, s7_pointer key)
{
  hash_entry_t *x;
  for (x = hash_table_element(table, loc); x; x = x->next)
    if ((is_t_complex(x->key)) &&
	(s7_is_morally_equal(sc, x->key, key)))
      return(x);
  return(NULL);
}


static hash_entry_t *hash_equal_real(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  return(hash_float_1(sc, table, hash_loc(sc, table, key) & hash_table_mask(table), real(key)));
}


static hash_entry_t *hash_equal_complex(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  return(hash_complex_1(sc, table, hash_loc(sc, table, key) & hash_table_mask(table), key));
}


static hash_entry_t *hash_equal_syntax(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  hash_entry_t *x;
  uint32_t loc;
  loc = hash_loc(sc, table, key) & hash_table_mask(table);
  for (x = hash_table_element(table, loc); x; x = x->next)
    if ((is_syntax(x->key)) &&
	(syntax_symbol(x->key) == syntax_symbol(key))) /* the opcodes might differ, but the symbols should not */
      return(x);
  return(NULL);
}


static hash_entry_t *hash_equal_eq(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  hash_entry_t *x;
  uint32_t loc;
  loc = hash_loc(sc, table, key) & hash_table_mask(table);
  for (x = hash_table_element(table, loc); x; x = x->next)
    if (x->key == key)
      return(x);
  return(NULL);
}


static hash_entry_t *hash_equal_any(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  hash_entry_t *x;
  uint32_t loc;
  loc = hash_loc(sc, table, key) & hash_table_mask(table);

  /* we can get into an infinite loop here, but it requires 2 hash tables that are members of each other
   *    and key is one of them, so I changed the equality check above to use eq? -- not sure this is right.
   */
#if 0
  /* hope for an easy case... (apparently this never happens) */
  for (x = hash_table_element(table, loc); x; x = x->next)
    if (x->key == key)
      return(x);
#endif
  for (x = hash_table_element(table, loc); x; x = x->next)
    if (s7_is_equal(sc, x->key, key))
      return(x);
  return(NULL);
}


static hash_entry_t *(*default_hash_checks[NUM_TYPES])(s7_scheme *sc, s7_pointer table, s7_pointer key);
static hash_entry_t *(*equal_hash_checks[NUM_TYPES])(s7_scheme *sc, s7_pointer table, s7_pointer key);
static hash_entry_t *(*morally_equal_hash_checks[NUM_TYPES])(s7_scheme *sc, s7_pointer table, s7_pointer key);

static hash_entry_t *hash_equal(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  return((*(equal_hash_checks[type(key)]))(sc, table, key));
}

static hash_entry_t *hash_morally_equal(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  hash_entry_t *x;
  uint32_t loc;
  loc = hash_loc(sc, table, key) & hash_table_mask(table);

  for (x = hash_table_element(table, loc); x; x = x->next)
    if (x->key == key)
      return(x);

  for (x = hash_table_element(table, loc); x; x = x->next)
    if (s7_is_morally_equal(sc, x->key, key))
      return(x);
  return(NULL);
}

static hash_entry_t *hash_c_function(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  hash_entry_t *x;
  uint32_t hash_len, loc;
  s7_function f;

  f = c_function_call(hash_table_procedures_checker(table));
  hash_len = hash_table_mask(table);
  loc = hash_loc(sc, table, key) & hash_len;
  
  set_car(sc->t2_1, key);
  for (x = hash_table_element(table, loc); x; x = x->next)
    {
      set_car(sc->t2_2, x->key);
      if (is_true(sc, f(sc, sc->t2_1)))
	return(x);
    }
  return(NULL);
}


static hash_entry_t *hash_eq(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  /* explicit eq? as hash equality func or (for example) symbols as keys */
  hash_entry_t *x;
  uint32_t hash_len, loc;

  hash_len = hash_table_mask(table);
  loc = hash_loc(sc, table, key) & hash_len;

  for (x = hash_table_element(table, loc); x; x = x->next)
    if (key == x->key)
      return(x);

  return(NULL);
}

static hash_entry_t *hash_eqv(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  hash_entry_t *x;
  uint32_t hash_len, loc;

  hash_len = hash_table_mask(table);
  loc = hash_loc(sc, table, key) & hash_len;

  for (x = hash_table_element(table, loc); x; x = x->next)
    if (s7_is_eqv(key, x->key))
      return(x);

  return(NULL);
}


static hash_entry_t *hash_number(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  if (is_number(key))
    {
      hash_entry_t *x;
      uint32_t hash_len, loc;

      hash_len = hash_table_mask(table);
      loc = hash_loc(sc, table, key) & hash_len;

#if (!WITH_GMP)
      for (x = hash_table_element(table, loc); x; x = x->next)
	if ((is_number(x->key)) &&
	    (is_true(sc, c_equal_2(sc, key, x->key))))
	  return(x);
#else
      for (x = hash_table_element(table, loc); x; x = x->next)
	if ((is_number(x->key)) &&
	    (is_true(sc, big_equal(sc, set_plist_2(sc, key, x->key)))))
	  return(x);
#endif
    }
  return(NULL);
}

static hash_entry_t *hash_symbol(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  if (is_symbol(key))
    {
      hash_entry_t *x;
      for (x = hash_table_element(table, symbol_hmap(key) & hash_table_mask(table)); x; x = x->next)
	if (key == x->key)
	  return(x);
    }
  return(NULL);
}


static hash_entry_t *hash_char(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  if (s7_is_character(key))
    return(hash_eq(sc, table, key));
  return(NULL);
}

static hash_entry_t *hash_closure(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  hash_entry_t *x;
  uint32_t hash_len, loc;
  s7_pointer f, args, body, old_e;

  f = hash_table_procedures_checker(table);
  hash_len = hash_table_mask(table);
  loc = hash_loc(sc, table, key) & hash_len;

  old_e = sc->envir;
  args = closure_args(f);    /* in lambda* case, car/cadr(args) can be lists */
  body = closure_body(f);
  new_frame_with_two_slots(sc, closure_let(f), sc->envir,
			   (is_symbol(car(args))) ? car(args) : caar(args), key,
			   (is_symbol(cadr(args))) ? cadr(args) : caadr(args), sc->F);

  for (x = hash_table_element(table, loc); x; x = x->next)
    {
      slot_set_value(next_slot(let_slots(sc->envir)), x->key);
      push_stack(sc, OP_EVAL_DONE, sc->args, sc->code);
      if (is_pair(cdr(body)))
	push_stack_no_args(sc, OP_BEGIN1, cdr(body));
      sc->code = car(body);
      eval(sc, OP_EVAL);
      if (is_true(sc, sc->value))
	{
	  sc->envir = old_e;
	  return(x);
	}
    }
  sc->envir = old_e;
  return(NULL);
}


static s7_pointer remove_from_hash_table(s7_scheme *sc, s7_pointer table, s7_pointer key, hash_entry_t *p)
{
  hash_entry_t *x;
  uint32_t hash_len, loc;

  hash_len = hash_table_mask(table);
#if DEBUGGING
  if (p->raw_hash != hash_loc(sc, table, key)) 
    fprintf(stderr, "%s[%d]: %s raw: %u, loc: %u\n", __func__, __LINE__, DISPLAY(key), p->raw_hash, hash_loc(sc, table, key));
#endif
  loc = p->raw_hash & hash_len;


  x = hash_table_element(table, loc);
  if (x == p)
    hash_table_element(table, loc) = x->next;
  else
    {
      hash_entry_t *y;
      for (y = x, x = x->next; x; y = x, x = x->next)
	if (x == p)
	  {
	    y->next = x->next;
	    break;
	  }
    }
  hash_table_entries(table)--;
  if ((hash_table_entries(table) == 0) &&
      (!hash_table_checker_locked(table)))
    {
      hash_table_checker(table) = hash_empty;
      hash_clear_chosen(table);
    }
  x->next = hash_free_list;
  hash_free_list = x;
  return(sc->F);
}

/* -------------------------------- make-hash-table -------------------------------- */

s7_pointer s7_make_hash_table(s7_scheme *sc, s7_int size)
{
  s7_pointer table;
  hash_entry_t **els;
  /* size is rounded up to the next power of 2 */

  if (size < 2)
    size = 2;
  else
    {
      if ((size & (size - 1)) != 0)          /* already 2^n ? */
	{
	  if ((size & (size + 1)) != 0)      /* already 2^n - 1 ? */
	    {
	      size--;
	      size |= (size >> 1);
	      size |= (size >> 2);
	      size |= (size >> 4);
	      size |= (size >> 8);
	      size |= (size >> 16);
	      if (s7_int_bits > 31)          /* this is either 31 or 63 */
		size |= (size >> 32);
	    }
	  size++;
	}
    }

  els = (hash_entry_t **)calloc(size, sizeof(hash_entry_t *));
  if (!els) return(s7_error(sc, make_symbol(sc, "out-of-memory"), set_elist_1(sc, s7_make_string_wrapper(sc, "make-hash-table allocation failed!"))));

  new_cell(sc, table, T_HASH_TABLE | T_SAFE_PROCEDURE);
  hash_table_mask(table) = size - 1;
  hash_table_elements(table) = els;
  hash_table_checker(table) = hash_empty;
  hash_table_mapper(table) = default_hash_map;
  hash_table_entries(table) = 0;
  hash_table_set_procedures(table, sc->nil);
  add_hash_table(sc, table);

  return(table);
}

static s7_pointer g_is_equal(s7_scheme *sc, s7_pointer args);
static s7_pointer g_is_morally_equal(s7_scheme *sc, s7_pointer args);

static s7_pointer g_make_hash_table(s7_scheme *sc, s7_pointer args)
{
  #define H_make_hash_table "(make-hash-table (size 8) eq-func) returns a new hash table"
  #define Q_make_hash_table s7_make_signature(sc, 3, sc->is_hash_table_symbol, sc->is_integer_symbol, s7_make_signature(sc, 2, sc->is_procedure_symbol, sc->is_pair_symbol))

  s7_int size;
  size = sc->default_hash_table_length;

  if (is_not_null(args))
    {
      s7_pointer p;
      p = car(args);
      if (!s7_is_integer(p))
	{
	  s7_pointer p1;
	  if (!s7_is_integer(p1 = check_values(sc, p, args)))
	    method_or_bust(sc, p, sc->make_hash_table_symbol, args, T_INTEGER, 1);
	  p = p1;
	}
      size = s7_integer(p);
      if (size <= 0)                      /* we need s7_int here to catch (make-hash-table most-negative-fixnum) etc */
	return(simple_out_of_range(sc, sc->make_hash_table_symbol, p, s7_make_string_wrapper(sc, "should be a positive integer")));
      if (size > sc->max_vector_length)
	return(simple_out_of_range(sc, sc->make_hash_table_symbol, p, its_too_large_string));

      if (is_not_null(cdr(args)))
	{
	  s7_pointer ht, proc;
	  proc = cadr(args);

	  if (is_c_function(proc))
	    {
	      if (!s7_is_aritable(sc, proc, 2))
		return(wrong_type_argument_with_type(sc, sc->make_hash_table_symbol, 3, proc, an_eq_func_string));

	      ht = s7_make_hash_table(sc, size);
	      hash_set_chosen(ht);
	      if (c_function_call(proc) == g_is_equal)
		{
		  hash_table_checker(ht) = hash_equal;
		  return(ht);
		}
	      if (c_function_call(proc) == g_is_morally_equal)
		{
		  hash_table_checker(ht) = hash_morally_equal;
		  hash_table_mapper(ht) = morally_equal_hash_map;
		  return(ht);
		}
	      if (c_function_call(proc) == g_is_eq)
		{
		  hash_table_checker(ht) = hash_eq;
		  hash_table_mapper(ht) = eq_hash_map;
		  return(ht);
		}
	      if (c_function_call(proc) == g_strings_are_equal)
		{
		  hash_table_checker(ht) = hash_string; 
		  hash_table_mapper(ht) = string_eq_hash_map;
		  return(ht);
		}
#if (!WITH_PURE_S7)
	      if (c_function_call(proc) == g_strings_are_ci_equal)
		{
		  hash_table_checker(ht) = hash_ci_string; 
		  hash_table_mapper(ht) = string_ci_eq_hash_map;
		  return(ht);
		}
	      if (c_function_call(proc) == g_chars_are_ci_equal)
		{
		  hash_table_checker(ht) = hash_ci_char; 
		  hash_table_mapper(ht) = char_ci_eq_hash_map;
		  return(ht);
		}
#endif
	      if (c_function_call(proc) == g_chars_are_equal)
		{
		  hash_table_checker(ht) = hash_char; 
		  hash_table_mapper(ht) = char_eq_hash_map;
		  return(ht);
		}
#if (!WITH_GMP)
	      if (c_function_call(proc) == g_equal)
#else
	      if ((c_function_call(proc) == g_equal) ||
		  (c_function_call(proc) == big_equal))
#endif
		{
		  hash_table_checker(ht) = hash_number; 
		  hash_table_mapper(ht) = number_eq_hash_map;
		  return(ht);
		}
	      if (c_function_call(proc) == g_is_eqv)
		{
		  hash_table_checker(ht) = hash_eqv; 
		  hash_table_mapper(ht) = eqv_hash_map;
		  return(ht);
		}
	      return(wrong_type_argument_with_type(sc, sc->make_hash_table_symbol, 3, proc, s7_make_string_wrapper(sc, "a hash function")));
	    }
	  /* proc not c_function */
	  else
	    {
	      if (is_pair(proc))
		{
		  s7_pointer checker, mapper;
		  checker = car(proc);
		  mapper = cdr(proc);

		  if (((is_any_c_function(checker)) || (is_any_closure(checker))) &&
		      ((is_any_c_function(mapper)) || (is_any_closure(mapper))) &&
		      (s7_is_aritable(sc, checker, 2)) &&
		      (s7_is_aritable(sc, mapper, 1)))
		    {
		      s7_pointer sig;
		      ht = s7_make_hash_table(sc, size);
		      hash_set_chosen(ht);
		      if (is_any_c_function(checker))
			{
			  sig = c_function_signature(checker);
			  if ((sig) &&
			      (is_pair(sig)) &&
			      (car(sig) != sc->is_boolean_symbol))
			    return(wrong_type_argument_with_type(sc, sc->make_hash_table_symbol, 3, proc, 
								 s7_make_string_wrapper(sc, "equality function should return a boolean")));
			  hash_table_checker(ht) = hash_c_function;
			}
		      else hash_table_checker(ht) = hash_closure;
		      if (is_any_c_function(mapper))
			{
			  sig = c_function_signature(mapper);
			  if ((sig) &&
			      (is_pair(sig)) &&
			      (car(sig) != sc->is_integer_symbol))
			    return(wrong_type_argument_with_type(sc, sc->make_hash_table_symbol, 3, proc, 
								 s7_make_string_wrapper(sc, "mapping function should return an integer")));
			  hash_table_mapper(ht) = c_function_hash_map;
			}
		      else hash_table_mapper(ht) = closure_hash_map;
		      hash_table_set_procedures(ht, proc);
		      return(ht);
		    }
		}
	      return(wrong_type_argument_with_type(sc, sc->make_hash_table_symbol, 3, proc, 
						   s7_make_string_wrapper(sc, "a cons of two functions")));
	    }
	}
    }
  return(s7_make_hash_table(sc, size));
}


void init_hash_maps(void)
{
  int32_t i;
  
  default_hash_map = (hash_map_t *)malloc(NUM_TYPES * sizeof(hash_map_t));
  eq_hash_map = (hash_map_t *)malloc(NUM_TYPES * sizeof(hash_map_t));
  eqv_hash_map = (hash_map_t *)malloc(NUM_TYPES * sizeof(hash_map_t));
  string_eq_hash_map = (hash_map_t *)malloc(NUM_TYPES * sizeof(hash_map_t));
  number_eq_hash_map = (hash_map_t *)malloc(NUM_TYPES * sizeof(hash_map_t));
  char_eq_hash_map = (hash_map_t *)malloc(NUM_TYPES * sizeof(hash_map_t));
#if (!WITH_PURE_S7)
  string_ci_eq_hash_map = (hash_map_t *)malloc(NUM_TYPES * sizeof(hash_map_t));
  char_ci_eq_hash_map = (hash_map_t *)malloc(NUM_TYPES * sizeof(hash_map_t));
#endif
  closure_hash_map = (hash_map_t *)malloc(NUM_TYPES * sizeof(hash_map_t));
  c_function_hash_map = (hash_map_t *)malloc(NUM_TYPES * sizeof(hash_map_t));
  morally_equal_hash_map = (hash_map_t *)malloc(NUM_TYPES * sizeof(hash_map_t));

  for (i = 0; i < NUM_TYPES; i++) 
    {
      default_hash_map[i] = hash_map_nil;
      string_eq_hash_map[i] = hash_map_nil;
      char_eq_hash_map[i] = hash_map_nil;
#if (!WITH_PURE_S7)
      string_ci_eq_hash_map[i] = hash_map_nil;
      char_ci_eq_hash_map[i] = hash_map_nil;
#endif
      number_eq_hash_map[i] = hash_map_nil;
      closure_hash_map[i] = hash_map_closure;
      c_function_hash_map[i] = hash_map_c_function;
      eq_hash_map[i] = hash_map_eq;
      eqv_hash_map[i] = hash_map_eq;

      equal_hash_checks[i] = hash_equal_any;
      morally_equal_hash_checks[i] = hash_equal_any;
      default_hash_checks[i] = hash_equal;
    }
  default_hash_map[T_INTEGER] =       hash_map_int;
  default_hash_map[T_RATIO] =         hash_map_ratio;
  default_hash_map[T_REAL] =          hash_map_real;
  default_hash_map[T_COMPLEX] =       hash_map_complex;
  default_hash_map[T_CHARACTER] =     hash_map_char;
  default_hash_map[T_SYMBOL] =        hash_map_symbol;
  default_hash_map[T_SYNTAX] =        hash_map_syntax;
  default_hash_map[T_STRING] =        hash_map_string;
  default_hash_map[T_HASH_TABLE] =    hash_map_hash_table;
  default_hash_map[T_VECTOR] =        hash_map_vector;
  default_hash_map[T_INT_VECTOR] =    hash_map_int_vector;
  default_hash_map[T_FLOAT_VECTOR] =  hash_map_float_vector;
  default_hash_map[T_LET] =           hash_map_let;
  default_hash_map[T_PAIR] =          hash_map_pair;
#if WITH_GMP
  default_hash_map[T_BIG_INTEGER] =   hash_map_big_int;
  default_hash_map[T_BIG_RATIO] =     hash_map_big_ratio;
  default_hash_map[T_BIG_REAL] =      hash_map_big_real;
  default_hash_map[T_BIG_COMPLEX] =   hash_map_big_complex;
#endif
  
  for (i = 0; i < NUM_TYPES; i++) morally_equal_hash_map[i] = default_hash_map[i];

  string_eq_hash_map[T_STRING] =      hash_map_string;
  char_eq_hash_map[T_CHARACTER] =     hash_map_char;
#if (!WITH_PURE_S7)
  string_ci_eq_hash_map[T_STRING] =   hash_map_ci_string;
  char_ci_eq_hash_map[T_CHARACTER] =  hash_map_ci_char;
#endif

  number_eq_hash_map[T_INTEGER] =     hash_map_int;
  number_eq_hash_map[T_RATIO] =       hash_map_ratio_eq;
  number_eq_hash_map[T_REAL] =        hash_map_real_eq;
  number_eq_hash_map[T_COMPLEX] =     hash_map_complex;
#if (WITH_GMP)
  number_eq_hash_map[T_BIG_INTEGER] = hash_map_big_int;
  number_eq_hash_map[T_BIG_RATIO] =   hash_map_big_ratio;
  number_eq_hash_map[T_BIG_REAL] =    hash_map_big_real;
  number_eq_hash_map[T_BIG_COMPLEX] = hash_map_big_complex;
#endif

  eqv_hash_map[T_INTEGER] =           hash_map_int;
  eqv_hash_map[T_RATIO] =             hash_map_ratio_eq;
  eqv_hash_map[T_REAL] =              hash_map_real_eq;
  eqv_hash_map[T_COMPLEX] =           hash_map_complex;

  morally_equal_hash_map[T_INTEGER] = hash_map_int;
  morally_equal_hash_map[T_RATIO] =   hash_map_ratio_eq;
  morally_equal_hash_map[T_REAL] =    hash_map_real_eq;
  morally_equal_hash_map[T_COMPLEX] = hash_map_complex;

  equal_hash_checks[T_REAL] =         hash_equal_real;
  equal_hash_checks[T_COMPLEX] =      hash_equal_complex;
  equal_hash_checks[T_SYNTAX] =       hash_equal_syntax;
  equal_hash_checks[T_SYMBOL] =       hash_equal_eq;
  equal_hash_checks[T_CHARACTER] =    hash_equal_eq;

  default_hash_checks[T_STRING] =     hash_string;
  default_hash_checks[T_INTEGER] =    hash_int;
  default_hash_checks[T_REAL] =       hash_float;
  default_hash_checks[T_SYMBOL] =     hash_symbol;
  default_hash_checks[T_CHARACTER] =  hash_char;
}


static uint32_t resize_hash_table(s7_scheme *sc, s7_pointer table)
{
  /* resize the table */
  uint32_t hash_len, loc, i, old_size, new_size;
  hash_entry_t **new_els, **old_els;
  
  old_size = hash_table_mask(table) + 1;
  new_size = old_size * 4;
  hash_len = new_size - 1;
  new_els = (hash_entry_t **)calloc(new_size, sizeof(hash_entry_t *));
  old_els = hash_table_elements(table);
  
  for (i = 0; i < old_size; i++)
    {
      hash_entry_t *x, *n;
      for (x = old_els[i]; x; x = n)
	{
	  n = x->next;
	  loc = x->raw_hash & hash_len;
	  x->next = new_els[loc];
	  new_els[loc] = x;
	}
    }
  hash_table_elements(table) = new_els;
  free(old_els);
  hash_table_mask(table) = new_size - 1;
  return(hash_len);
}


/* -------------------------------- hash-table-ref -------------------------------- */

s7_pointer s7_hash_table_ref(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  hash_entry_t *x;
  x = (*hash_table_checker(table))(sc, table, key);
  if (x) return(x->value);
  return(sc->F);
}


static s7_pointer g_hash_table_ref(s7_scheme *sc, s7_pointer args)
{
  #define H_hash_table_ref "(hash-table-ref table key) returns the value associated with key in the hash table"
  #define Q_hash_table_ref s7_make_circular_signature(sc, 2, 3, sc->T, sc->is_hash_table_symbol, sc->T)

  s7_pointer table;
  table = car(args);
  if (!is_hash_table(table))
    method_or_bust(sc, table, sc->hash_table_ref_symbol, args, T_HASH_TABLE, 1);
  /*
    (define (href H . args)
      (if (null? (cdr args))
          (hash-table-ref H (car args))
          (apply href (hash-table-ref H (car args)) (cdr args))))
  */
  if (is_null(cddr(args)))
    return(s7_hash_table_ref(sc, table, cadr(args)));
  return(implicit_index(sc, s7_hash_table_ref(sc, table, cadr(args)), cddr(args)));
}


static s7_pointer hash_table_ref_2;
static s7_pointer g_hash_table_ref_2(s7_scheme *sc, s7_pointer args)
{
  s7_pointer table;
  hash_entry_t *x;

  table = car(args);
  if (!is_hash_table(table))
    method_or_bust(sc, table, sc->hash_table_ref_symbol, args, T_HASH_TABLE, 1);

  x = (*hash_table_checker(table))(sc, table, cadr(args));
  if (x) return(x->value);
  return(sc->F);
}

static s7_pointer hash_table_ref_ss;
static s7_pointer g_hash_table_ref_ss(s7_scheme *sc, s7_pointer args)
{
  s7_pointer table, key;
  hash_entry_t *x;

  table = find_symbol_unchecked(sc, car(args));
  key = find_symbol_unchecked(sc, cadr(args));
  if (!is_hash_table(table))
    method_or_bust(sc, table, sc->hash_table_ref_symbol, list_2(sc, table, key), T_HASH_TABLE, 1);
  
  x = (*hash_table_checker(table))(sc, table, key);
  if (x) return(x->value);
  return(sc->F);
}

static s7_pointer hash_table_ref_car;
static s7_pointer g_hash_table_ref_car(s7_scheme *sc, s7_pointer args)
{
  s7_pointer y, table;
  hash_entry_t *x;

  table = find_symbol_unchecked(sc, car(args));
  y = find_symbol_unchecked(sc, opt_sym3(args));
  if (!is_pair(y))
    return(simple_wrong_type_argument(sc, sc->car_symbol, y, T_PAIR));

  if (!is_hash_table(table))
    method_or_bust(sc, table, sc->hash_table_ref_symbol, list_2(sc, table, car(y)), T_HASH_TABLE, 1);

  x = (*hash_table_checker(table))(sc, table, car(y));
  if (x) return(x->value);
  return(sc->F);
}

static s7_pointer hash_table_ref_p_pp(s7_pointer p1, s7_pointer p2)
{
  if (!is_hash_table(p1))
    simple_wrong_type_argument(cur_sc, cur_sc->hash_table_ref_symbol, p1, T_HASH_TABLE);
  return(s7_hash_table_ref(cur_sc, p1, p2));
}

static s7_pointer hash_table_ref_p_pp_direct(s7_pointer p1, s7_pointer p2)
{
  return(s7_hash_table_ref(cur_sc, p1, p2));
}


/* -------------------------------- hash-table-set! -------------------------------- */

static void hash_table_set_checker(s7_pointer table, uint8_t typ)
{
  if (/* (hash_table_checker(table) != hash_equal) && */
      (hash_table_checker(table) != default_hash_checks[typ]))
    {
      if (hash_table_checker(table) == hash_empty)
	hash_table_checker(table) = default_hash_checks[typ];
      else 
	{
	  hash_table_checker(table) = hash_equal;
	  hash_set_chosen(table);
	}
    }
}

static void extend_hash_free_list(void)
{
  int32_t i;
  hash_entry_t *p;
  hash_free_list = (hash_entry_t *)malloc(16 * sizeof(hash_entry_t));
  for (p = hash_free_list, i = 0; i < 15; i++) {p->next = p + 1; p++;}
  p->next = NULL;
}

s7_pointer s7_hash_table_set(s7_scheme *sc, s7_pointer table, s7_pointer key, s7_pointer value)
{
  hash_entry_t *x;
  x = (*hash_table_checker(table))(sc, table, key);

  if (x)
    {
      if (value == sc->F)
	return(remove_from_hash_table(sc, table, key, x));
      x->value = _NFre(value);
    }
  else 
    {
      uint32_t hash_len, raw_hash, loc;
      hash_entry_t *p;
      if (value == sc->F) return(sc->F);
      
      if (!hash_chosen(table))
	hash_table_set_checker(table, type(key));

      hash_len = hash_table_mask(table);
      if (hash_table_entries(table) > hash_len)
	hash_len = resize_hash_table(sc, table);
      raw_hash = hash_loc(sc, table, key);

      if (!hash_free_list)
	extend_hash_free_list();

      p = hash_free_list;
      hash_free_list = p->next;
      p->key = key;
      p->value = _NFre(value);
      p->raw_hash = raw_hash;

      loc = raw_hash & hash_len;
      p->next = hash_table_element(table, loc);
      hash_table_element(table, loc) = p;
      hash_table_entries(table)++;
    }
  return(value);
}


static s7_pointer g_hash_table_set(s7_scheme *sc, s7_pointer args)
{
  #define H_hash_table_set "(hash-table-set! table key value) sets the value associated with key in the hash table to value"
  #define Q_hash_table_set s7_make_signature(sc, 4, sc->T, sc->is_hash_table_symbol, sc->T, sc->T)

  s7_pointer table;
  table = car(args);
  if (!is_hash_table(table))
    method_or_bust(sc, table, sc->hash_table_set_symbol, args, T_HASH_TABLE, 1);
  return(s7_hash_table_set(sc, table, cadr(args), caddr(args)));
}

static s7_pointer hash_table_set_p_ppp(s7_pointer p1, s7_pointer p2, s7_pointer p3)
{
  if (!is_hash_table(p1))
    simple_wrong_type_argument(cur_sc, cur_sc->hash_table_set_symbol, p1, T_HASH_TABLE);
  return(s7_hash_table_set(cur_sc, p1, p2, p3));
}

static s7_pointer hash_table_set_p_ppp_direct(s7_pointer p1, s7_pointer p2, s7_pointer p3)
{
  return(s7_hash_table_set(cur_sc, p1, p2, p3));
}


/* -------------------------------- hash-table -------------------------------- */
static s7_pointer g_hash_table(s7_scheme *sc, s7_pointer args)
{
  #define H_hash_table "(hash-table ...) returns a hash-table containing the cons's passed as its arguments. \
That is, (hash-table '(\"hi\" . 3) (\"ho\" . 32)) returns a new hash-table with the two key/value pairs preinstalled."
  #define Q_hash_table s7_make_circular_signature(sc, 1, 2, sc->is_hash_table_symbol, sc->is_list_symbol)

  int32_t len;
  s7_pointer x, ht;

  /* this accepts repeated keys: (hash-table '(a . 1) '(a . 1)) */
  for (len = 0, x = args; is_pair(x); x = cdr(x), len++)
    if ((!is_pair(car(x))) &&
	(!is_null(car(x))))
      return(wrong_type_argument(sc, sc->hash_table_symbol, position_of(x, args), car(x), T_PAIR));

  ht = s7_make_hash_table(sc, (len > sc->default_hash_table_length) ? len : sc->default_hash_table_length);
  if (len > 0)
    {
      uint32_t ht_loc;
      ht_loc = s7_gc_protect(sc, ht); /* hash_table_set can cons, so we need to protect this */
      for (x = args; is_pair(x); x = cdr(x))
	if (is_pair(car(x)))
	  s7_hash_table_set(sc, ht, caar(x), cdar(x));
      s7_gc_unprotect_at(sc, ht_loc);
    }
  return(ht);
}


/* -------------------------------- hash-table* -------------------------------- */
static s7_pointer g_hash_table_star(s7_scheme *sc, s7_pointer args)
{
  #define H_hash_table_star "(hash-table* ...) returns a hash-table containing the symbol/value pairs passed as its arguments. \
That is, (hash-table* 'a 1 'b 2) returns a new hash-table with the two key/value pairs preinstalled."
  #define Q_hash_table_star s7_make_circular_signature(sc, 1, 2, sc->is_hash_table_symbol, sc->T)

  int32_t len;
  s7_pointer ht;

  len = safe_list_length(sc, args);
  if (len & 1) 
    return(s7_error(sc, sc->wrong_number_of_args_symbol, set_elist_2(sc, s7_make_string_wrapper(sc, "hash-table* got an odd number of arguments: ~S"), args)));
  len /= 2;

  ht = s7_make_hash_table(sc, (len > sc->default_hash_table_length) ? len : sc->default_hash_table_length);
  if (len > 0)
    {
      uint32_t ht_loc;
      s7_pointer x, y;
      ht_loc = s7_gc_protect(sc, ht); /* hash_table_set can cons, so we need to protect this */

      for (x = args, y = cdr(args); is_pair(x); x = cddr(x), y = unchecked_cdr(cdr(y)))
	s7_hash_table_set(sc, ht, car(x), car(y));

      s7_gc_unprotect_at(sc, ht_loc);
    }
  return(ht);
}

static s7_pointer hash_table_copy(s7_scheme *sc, s7_pointer old_hash, s7_pointer new_hash, uint32_t start, uint32_t end)
{
  uint32_t i, old_len, new_len, count = 0;
  hash_entry_t **old_lists, **new_lists;
  hash_entry_t *x, *p;

  old_len = hash_table_mask(old_hash) + 1;
  new_len = hash_table_mask(new_hash);
  old_lists = hash_table_elements(old_hash);
  new_lists = hash_table_elements(new_hash);
  
  if (hash_table_entries(new_hash) == 0)
    {
      hash_table_checker(new_hash) = hash_table_checker(old_hash);
      if (hash_chosen(old_hash)) hash_set_chosen(new_hash);
      if ((start == 0) && 
	  (end >= hash_table_entries(old_hash)))
	{
	  for (i = 0; i < old_len; i++)
	    for (x = old_lists[i]; x; x = x->next)
	      {
		uint32_t loc;
		loc = x->raw_hash & new_len;
		p = make_hash_entry(x->key, x->value, x->raw_hash);
		p->next = new_lists[loc];
		new_lists[loc] = p;
	      }
	  hash_table_entries(new_hash) = hash_table_entries(old_hash);
	  return(new_hash);
	}
      for (i = 0; i < old_len; i++)
	for (x = old_lists[i]; x; x = x->next)
	  {
	    if (count >= end)
	      {
		hash_table_entries(new_hash) = end - start;
		return(new_hash);
	      }
	    if (count >= start)
	      {
		uint32_t loc;
		loc = x->raw_hash & new_len;
		p = make_hash_entry(x->key, x->value, x->raw_hash);
		p->next = new_lists[loc];
		new_lists[loc] = p;
	      }
	    count++;
	  }
      hash_table_entries(new_hash) = count - start;
      return(new_hash);
    }
      
  /* this can't be optimized much because we have to look for key matches (we're copying old_hash into the exisiting, non-empty new_hash) */
  for (i = 0; i < old_len; i++)
    for (x = old_lists[i]; x; x = x->next)
      {
	if (count >= end)
	  return(new_hash);
	if (count >= start)
	  {
	    hash_entry_t *y;
	    y = (*hash_table_checker(new_hash))(sc, new_hash, x->key);
	    if (y)
	      y->value = x->value;
	    else
	      {
		uint32_t loc;
		loc = x->raw_hash & new_len;
		p = make_hash_entry(x->key, x->value, x->raw_hash);
		p->next = new_lists[loc];
		new_lists[loc] = p;
		hash_table_entries(new_hash)++;
		if (!hash_chosen(new_hash))
		  hash_table_set_checker(new_hash, type(x->key));
	      }
	  }
	count++;
      }
  return(new_hash);
}

static s7_pointer hash_table_fill(s7_scheme *sc, s7_pointer args)
{
  s7_pointer val, table;
  table = car(args);
  val = cadr(args);
  if (hash_table_entries(table) > 0)
    {
      int32_t len;
      hash_entry_t **entries;
      entries = hash_table_elements(table);
      len = hash_table_mask(table) + 1;
      /* hash-table-ref returns #f if it can't find a key, so val == #f here means empty the table */
      if (val == sc->F)
	{
	  hash_entry_t **hp;
	  hash_entry_t *p;
	  hp = entries;
	  if (len == 1)
	    {
	      if (*hp)
		{
		  p = *hp;
		  while (p->next) p = p->next;
		  p->next = hash_free_list;
		  hash_free_list = *hp;
		}
	    }
	  else
	    {
	      hash_entry_t **hn;
	      /* here we assume we can go by 2's */
	      hn = (hash_entry_t **)(hp + len);
	      for (; hp < hn; hp++)
		{
		  if (*hp)
		    {
		      p = *hp;
		      while (p->next) p = p->next;
		      p->next = hash_free_list;
		      hash_free_list = *hp;
		    }
		  hp++;
		  if (*hp)
		    {
		      p = *hp;
		      while (p->next) p = p->next;
		      p->next = hash_free_list;
		      hash_free_list = *hp;
		    }
		}
	    }
	  memset(entries, 0, len * sizeof(hash_entry_t *));
	  if (!hash_table_checker_locked(table))
	    {
	      hash_table_checker(table) = hash_empty;
	      hash_clear_chosen(table);
	    }
	  hash_table_entries(table) = 0;
	}
      else
	{
	  int32_t i;
	  hash_entry_t *x;
	  for (i = 0; i < len; i++)
	    for (x = entries[i]; x; x = x->next)
	      x->value = val;
	  /* keys haven't changed, so no need to mess with hash_table_checker */
	}
    }
  return(val);
}


static s7_pointer hash_table_reverse(s7_scheme *sc, s7_pointer old_hash)
{
  int32_t i, len;
  s7_pointer new_hash;
  hash_entry_t **old_lists;
  uint32_t gc_loc;

  len = hash_table_mask(old_hash) + 1;
  new_hash = s7_make_hash_table(sc, len);
  gc_loc = s7_gc_protect(sc, new_hash);

  /* I don't think the original hash functions can make any sense in general, so ignore them */
  old_lists = hash_table_elements(old_hash);
  for (i = 0; i < len; i++)
    {
      hash_entry_t *x;
      for (x = old_lists[i]; x; x = x->next)
	s7_hash_table_set(sc, new_hash, x->value, x->key);
    }
  s7_gc_unprotect_at(sc, gc_loc);
  return(new_hash);
}



/* -------------------------------- functions -------------------------------- */

bool s7_is_function(s7_pointer p)
{
  return(is_c_function(p));
}


static s7_pointer fallback_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
  return(f);
}

static void s7_function_set_class(s7_pointer f, s7_pointer base_f)
{
  c_function_class(f) = c_function_class(base_f);
  c_function_set_base(f, base_f);
}

static int32_t c_functions = 0;

s7_pointer s7_make_function(s7_scheme *sc, const char *name, s7_function f, int32_t required_args, int32_t optional_args, bool rest_arg, const char *doc)
{
  c_proc_t *ptr;
  uint32_t ftype = T_C_FUNCTION;
  s7_pointer x;

  x = alloc_pointer();
  unheap(x);

  ptr = (c_proc_t *)malloc(sizeof(c_proc_t));
  c_functions++;
  if (required_args == 0)
    {
      if (rest_arg)
	ftype = T_C_ANY_ARGS_FUNCTION;
      else
	{
	  if (optional_args != 0)
	    ftype = T_C_OPT_ARGS_FUNCTION;
	  /* a thunk needs to check for no args passed */
	}
    }
  else
    {
      if (rest_arg)
	ftype = T_C_RST_ARGS_FUNCTION;
    }

  set_type(x, ftype);

  c_function_data(x) = ptr;
  c_function_call(x) = f;
  /* f is _TApp but needs cast */
  c_function_set_base(x, x);
  c_function_set_setter(x, sc->F);
  c_function_name(x) = name;   /* (procedure-name proc) => (format #f "~A" proc) */
  c_function_name_length(x) = safe_strlen(name);
  if (doc)
    c_function_documentation(x) = make_permanent_c_string(doc);
  else c_function_documentation(x) = NULL;
  c_function_signature(x) = sc->F;

  c_function_required_args(x) = required_args;
  c_function_optional_args(x) = optional_args;
  c_function_has_rest_arg(x) = rest_arg;
  if (rest_arg)
    c_function_all_args(x) = MAX_ARITY;
  else c_function_all_args(x) = required_args + optional_args;

  c_function_class(x) = ++sc->f_class;
  c_function_chooser(x) = fallback_chooser;
  c_function_opt_data(x) = NULL;

  return(x);
}

s7_pointer s7_make_safe_function(s7_scheme *sc, const char *name, s7_function f, 
				 int32_t required_args, int32_t optional_args, bool rest_arg, const char *doc)
{
  s7_pointer p;
  p = s7_make_function(sc, name, f, required_args, optional_args, rest_arg, doc);
  typeflag(p) |= T_SAFE_PROCEDURE; 
  return(p);
}


s7_pointer s7_make_typed_function(s7_scheme *sc, const char *name, s7_function f, 
				  int32_t required_args, int32_t optional_args, bool rest_arg, const char *doc, s7_pointer signature)
{
  s7_pointer func;
  func = s7_make_function(sc, name, f, required_args, optional_args, rest_arg, doc);
  typeflag(func) |= T_SAFE_PROCEDURE;
  if (signature) c_function_signature(func) = signature;
  return(func);
}


bool s7_is_procedure(s7_pointer x)
{
  return(is_procedure(x));
}


static s7_pointer g_is_procedure(s7_scheme *sc, s7_pointer args)
{
  #define H_is_procedure "(procedure? obj) returns #t if obj is a procedure"
  #define Q_is_procedure pl_bt

  return(make_boolean(sc, is_procedure(car(args))));
}


static void s7_function_set_setter(s7_scheme *sc, const char *getter, const char *setter)
{
  /* this is internal, used only with c_function setters, so we don't need to worry about the GC mark choice
   */
  c_function_set_setter(s7_name_to_value(sc, getter), s7_name_to_value(sc, setter));
}


s7_pointer s7_closure_body(s7_scheme *sc, s7_pointer p)
{
  if (has_closure_let(p))
    return(closure_body(p));
  return(sc->nil);
}


s7_pointer s7_closure_let(s7_scheme *sc, s7_pointer p)
{
  if (has_closure_let(p))
    return(closure_let(p));
  return(sc->nil);
}


s7_pointer s7_closure_args(s7_scheme *sc, s7_pointer p)
{
  if (has_closure_let(p))
    return(closure_args(p));
  return(sc->nil);
}

static s7_pointer g_procedure_source(s7_scheme *sc, s7_pointer args)
{
  /* make it look like a scheme-level lambda */
  s7_pointer p;
  #define H_procedure_source "(procedure-source func) tries to return the definition of func"
  #define Q_procedure_source s7_make_signature(sc, 2, sc->is_list_symbol, sc->is_procedure_symbol)

  p = car(args);
  if (is_symbol(p))
    {
      p = s7_symbol_value(sc, p);
      if (p == sc->undefined)
	return(s7_error(sc, sc->wrong_type_arg_symbol, set_elist_2(sc, s7_make_string_wrapper(sc, "procedure-source arg, '~S, is unbound"), p)));
    }

  if ((is_c_function(p)) || (is_c_macro(p)))
    return(sc->nil);

  check_method(sc, p, sc->procedure_source_symbol, list_1(sc, p));
  if (has_closure_let(p))
    {
      s7_pointer body;
      body = closure_body(p);
      if (is_safe_closure(body))
	clear_safe_closure(body);
      return(append_in_place(sc, list_2(sc, ((is_closure_star(p)) || 
					     (is_macro_star(p)) || 
					     (is_bacro_star(p))) ? sc->lambda_star_symbol : sc->lambda_symbol, 
					closure_args(p)), body));
    }

  if (!is_procedure(p))
    return(simple_wrong_type_argument_with_type(sc, sc->procedure_source_symbol, p, s7_make_string_wrapper(sc, "a procedure or a macro")));
  return(sc->nil);
}

s7_pointer s7_funclet(s7_scheme *sc, s7_pointer p)
{
  if (has_closure_let(p))
    return(closure_let(p));
  return(sc->rootlet);
}


static s7_pointer g_funclet(s7_scheme *sc, s7_pointer args)
{
  s7_pointer p, e;
  #define H_funclet "(funclet func) tries to return an object's environment"
  #define Q_funclet s7_make_signature(sc, 2, sc->is_let_symbol, sc->is_procedure_symbol)

  /* this procedure gives direct access to a function's closure -- see s7test.scm
   *   for some wild examples.  At least it provides a not-too-kludgey way for several functions
   *   to share a closure.
   */

  p = car(args);
  if (is_symbol(p))
    {
      p = s7_symbol_value(sc, p);
      if (p == sc->undefined)
	return(s7_error(sc, sc->wrong_type_arg_symbol, set_elist_2(sc, s7_make_string_wrapper(sc, "funclet arg, '~S, is unbound"), car(args)))); /* not p here */
    }
  check_method(sc, p, sc->funclet_symbol, args);

  if (!((is_procedure_or_macro(p)) || (is_c_object(p))))
    return(simple_wrong_type_argument_with_type(sc, sc->funclet_symbol, p, s7_make_string_wrapper(sc, "a procedure or a macro")));

  e = find_let(sc, p);
  if ((is_null(e)) &&
      (!is_c_object(p)))
    return(sc->rootlet);

  return(e);
}


s7_pointer s7_define_function(s7_scheme *sc, const char *name, s7_function fnc,
			      int32_t required_args, int32_t optional_args, bool rest_arg, const char *doc)
{
  s7_pointer func, sym;
  func = s7_make_function(sc, name, fnc, required_args, optional_args, rest_arg, doc);
  sym = make_symbol(sc, name);
  s7_define(sc, sc->nil, sym, func);
  return(sym);
}


s7_pointer s7_define_safe_function(s7_scheme *sc, const char *name, s7_function fnc,
				   int32_t required_args, int32_t optional_args, bool rest_arg, const char *doc)
{
  /* returns (string->symbol name), not the c_proc_t func */
  s7_pointer func, sym;
  func = s7_make_safe_function(sc, name, fnc, required_args, optional_args, rest_arg, doc);
  sym = make_symbol(sc, name);
  s7_define(sc, sc->nil, sym, func);
  return(sym);
}


s7_pointer s7_define_typed_function(s7_scheme *sc, const char *name, s7_function fnc,
				    int32_t required_args, int32_t optional_args, bool rest_arg, 
				    const char *doc, s7_pointer signature)
{
  /* returns (string->symbol name), not the c_proc_t func */
  s7_pointer func, sym;
  func = s7_make_typed_function(sc, name, fnc, required_args, optional_args, rest_arg, doc, signature);
  sym = make_symbol(sc, name);
  s7_define(sc, sc->nil, sym, func);
  return(sym);
}


s7_pointer s7_define_unsafe_typed_function(s7_scheme *sc, const char *name, s7_function fnc,
					   int32_t required_args, int32_t optional_args, bool rest_arg, 
					   const char *doc, s7_pointer signature)
{
  /* returns (string->symbol name), not the c_proc_t func */
  s7_pointer func, sym;
  func = s7_make_function(sc, name, fnc, required_args, optional_args, rest_arg, doc);
  if (signature) c_function_signature(func) = signature;
  sym = make_symbol(sc, name);
  s7_define(sc, sc->nil, sym, func);
  return(sym);
}


s7_pointer s7_define_macro(s7_scheme *sc, const char *name, s7_function fnc,
			   int32_t required_args, int32_t optional_args, bool rest_arg, const char *doc)
{
  s7_pointer func, sym;
  func = s7_make_function(sc, name, fnc, required_args, optional_args, rest_arg, doc);
  set_type(func, T_C_MACRO | T_DONT_EVAL_ARGS);
  sym = make_symbol(sc, name);
  s7_define(sc, sc->nil, sym, func);
  return(sym);
}


bool s7_is_macro(s7_scheme *sc, s7_pointer x)
{
  return(is_any_macro(x));
}

static bool is_macro_b(s7_pointer x) {return(is_any_macro(x));}


static s7_pointer g_is_macro(s7_scheme *sc, s7_pointer args)
{
  #define H_is_macro "(macro? arg) returns #t if 'arg' is a macro or a bacro"
  #define Q_is_macro pl_bt
  check_boolean_method(sc, is_any_macro, sc->is_macro_symbol, args);
}


static s7_pointer s7_macroexpand(s7_scheme *sc, s7_pointer mac, s7_pointer args)
{
  push_stack(sc, OP_EVAL_DONE, sc->args, sc->code);
  sc->code = mac;
  sc->args = copy_list(sc, args);
  new_frame(sc, closure_let(sc->code), sc->envir);
  eval(sc, OP_APPLY_LAMBDA);
  return(sc->value);
}


s7_pointer s7_make_function_star(s7_scheme *sc, const char *name, s7_function fnc, const char *arglist, const char *doc)
{
  s7_pointer func, local_args, p;
  char *internal_arglist;
  int32_t i, len, n_args;
  uint32_t gc_loc;
  s7_pointer *names, *defaults;

  len = safe_strlen(arglist) + 8;
  tmpbuf_malloc(internal_arglist, len);
  snprintf(internal_arglist, len, "'(%s)", arglist);
  local_args = s7_eval_c_string(sc, internal_arglist);
  gc_loc = s7_gc_protect(sc, local_args);
  tmpbuf_free(internal_arglist, len);
  n_args = safe_list_length(sc, local_args);  /* currently rest arg not supported, and we don't notice :allow-other-keys etc */

  func = s7_make_function(sc, name, fnc, 0, n_args, false, doc);
  set_type(func, T_C_FUNCTION_STAR);
  c_function_call_args(func) = NULL;

  names = (s7_pointer *)malloc(n_args * sizeof(s7_pointer));
  c_function_arg_names(func) = names;
  defaults = (s7_pointer *)malloc(n_args * sizeof(s7_pointer));
  c_function_arg_defaults(func) = defaults;
  set_simple_defaults(func);

  for (p = local_args, i = 0; i < n_args; p = cdr(p), i++)
    {
      s7_pointer arg;
      arg = car(p);
      if (is_pair(arg))
	{
	  names[i] = symbol_to_keyword(sc, car(arg));
	  defaults[i] = cadr(arg);
	  s7_remove_from_heap(sc, cadr(arg));
	  if ((is_symbol(defaults[i])) ||
	      (is_pair(defaults[i])))
	    {
	      clear_simple_defaults(func);
	      mark_function[T_C_FUNCTION_STAR] = mark_c_proc_star;
	    }
	}
      else
	{
	  names[i] = symbol_to_keyword(sc, arg);
	  defaults[i] = sc->F;
	}
    }
  s7_gc_unprotect_at(sc, gc_loc);
  return(func);
}

s7_pointer s7_make_safe_function_star(s7_scheme *sc, const char *name, s7_function fnc, const char *arglist, const char *doc)
{
  s7_pointer func;
  func = s7_make_function_star(sc, name, fnc, arglist, doc);
  set_type(func, T_C_FUNCTION_STAR | T_SAFE_PROCEDURE);
  c_function_call_args(func) = make_list(sc, c_function_optional_args(func), sc->F);
  s7_remove_from_heap(sc, c_function_call_args(func));
  return(func);
}

static void define_function_star_1(s7_scheme *sc, const char *name, s7_function fnc, const char *arglist, const char *doc, bool safe, s7_pointer signature)
{
  s7_pointer func, sym;
  if (safe)
    func = s7_make_safe_function_star(sc, name, fnc, arglist, doc);
  else func = s7_make_function_star(sc, name, fnc, arglist, doc);
  sym = make_symbol(sc, name);
  s7_define(sc, sc->nil, sym, func);
  if (signature) c_function_signature(func) = signature;
}

void s7_define_function_star(s7_scheme *sc, const char *name, s7_function fnc, const char *arglist, const char *doc)
{
  define_function_star_1(sc, name, fnc, arglist, doc, false, NULL);
}

void s7_define_safe_function_star(s7_scheme *sc, const char *name, s7_function fnc, const char *arglist, const char *doc)
{
  define_function_star_1(sc, name, fnc, arglist, doc, true, NULL);
}

void s7_define_typed_function_star(s7_scheme *sc, const char *name, s7_function fnc, const char *arglist, const char *doc, s7_pointer signature)
{
  define_function_star_1(sc, name, fnc, arglist, doc, true, signature);
}


/* -------------------------------- procedure-documentation -------------------------------- */
static s7_pointer get_doc(s7_scheme *sc, s7_pointer x)
{
  check_closure_for(sc, x, sc->documentation_symbol);
  return(NULL);
}

const char *s7_procedure_documentation(s7_scheme *sc, s7_pointer x)
{
  s7_pointer val;
  if (is_symbol(x))
    {
      if ((symbol_has_help(x)) &&
	  (is_global(x)))
	return(symbol_help(x));
      x = s7_symbol_value(sc, x); /* this is needed by Snd */
    }

  if ((is_any_c_function(x)) ||
      (is_c_macro(x)))
    return((char *)c_function_documentation(x));

  val = get_doc(sc, x);
  if ((val) && (is_string(val)))
    return(string_value(val));

  return(NULL);
}

static s7_pointer g_procedure_documentation(s7_scheme *sc, s7_pointer args)
{
  s7_pointer p;
  #define H_procedure_documentation "(procedure-documentation func) returns func's documentation string"
  #define Q_procedure_documentation s7_make_signature(sc, 2, sc->is_string_symbol, sc->is_procedure_symbol)

  p = car(args);
  if (is_symbol(p))
    {
      if ((symbol_has_help(p)) &&
	  (is_global(p)))
	return(s7_make_string(sc, symbol_help(p)));
      p = s7_symbol_value(sc, p);
    }

  check_method(sc, p, sc->procedure_documentation_symbol, list_1(sc, p));
  if ((!is_procedure(p)) &&
      (!is_any_macro(p)))
    return(simple_wrong_type_argument_with_type(sc, sc->procedure_documentation_symbol, p, a_procedure_string));

  return(s7_make_string(sc, s7_procedure_documentation(sc, p)));
}


/* -------------------------------- help -------------------------------- */
const char *s7_help(s7_scheme *sc, s7_pointer obj)
{
  if (is_syntax(obj))
    return(string_value(syntax_documentation(obj)));

  if (is_symbol(obj))
    {
      /* here look for name */
      if (s7_symbol_documentation(sc, obj))
	return(s7_symbol_documentation(sc, obj));
      obj = s7_symbol_value(sc, obj);
    }

  if (is_procedure_or_macro(obj))
    return(s7_procedure_documentation(sc, obj));

  /* if is string, apropos? (can scan symbol table) */
  return(NULL);
}


static s7_pointer g_help(s7_scheme *sc, s7_pointer args)
{
  #define H_help "(help obj) returns obj's documentation"
  #define Q_help s7_make_signature(sc, 2, s7_make_signature(sc, 2, sc->is_string_symbol, sc->is_boolean_symbol), sc->T)
  const char *doc;

  check_method(sc, car(args), sc->help_symbol, args);
  doc = s7_help(sc, car(args));
  if (!doc)
    return(sc->F);
  return(s7_make_string(sc, doc));
}


/* -------------------------------- procedure-signature -------------------------------- */
static s7_pointer get_signature(s7_scheme *sc, s7_pointer x)
{
  check_closure_for(sc, x, sc->signature_symbol);
  return(sc->F);
}

s7_pointer s7_procedure_signature(s7_scheme *sc, s7_pointer func)
{
  if ((is_any_c_function(func)) ||
      (is_c_macro(func)))
    return((s7_pointer)c_function_signature(func));
  return(get_signature(sc, func));
}

static s7_pointer g_procedure_signature(s7_scheme *sc, s7_pointer args)
{
  s7_pointer p;
  #define H_procedure_signature "(procedure-signature func) returns func's signature"
  #define Q_procedure_signature s7_make_signature(sc, 2, s7_make_signature(sc, 2, sc->is_pair_symbol, sc->is_boolean_symbol), sc->is_procedure_symbol)

  p = car(args);
  if (is_symbol(p)) 
    {
      p = s7_symbol_value(sc, p);
      if (is_procedure(p))
	return(s7_procedure_signature(sc, p));
      if (p == sc->undefined)
	return(sc->F);
    }
  else
    {
      if (is_procedure(p))
	return(s7_procedure_signature(sc, p));
    }
  check_method(sc, p, sc->procedure_signature_symbol, list_1(sc, p));
  return(sc->F);
}

/* -------------------------------- new types (c_objects) -------------------------------- */

static void fallback_free(void *value) {}
static void fallback_mark(void *value) {}

static char *fallback_print(s7_scheme *sc, void *val)
{
  return(copy_string("#<unprintable c_object>"));
}

static char *fallback_print_readably(s7_scheme *sc, void *val)
{
  return(copy_string("#<unprint-readable c_object>"));
}

static bool fallback_equal(void *val1, void *val2)
{
  return(val1 == val2);
}

static s7_pointer fallback_ref(s7_scheme *sc, s7_pointer obj, s7_pointer args)
{
  return(apply_error(sc, obj, args));
}

static s7_pointer fallback_set(s7_scheme *sc, s7_pointer obj, s7_pointer args)
{
  eval_error(sc, "attempt to set ~S?", obj);
}

static s7_pointer fallback_length(s7_scheme *sc, s7_pointer obj)
{
  return(sc->F);
}


bool s7_is_c_object(s7_pointer p)
{
  return(is_c_object(p));
}

static s7_pointer g_is_c_object(s7_scheme *sc, s7_pointer args)
{
  #define H_is_c_object "(c-object? obj) returns the c_object's type tag."
  #define Q_is_c_object pl_bt

  s7_pointer p;
  p = car(args);
  if (is_c_object(p)) 
    return(make_integer(sc, c_object_type(p))); /* this is the c_object_types table index = tag */
  check_method(sc, p, sc->is_c_object_symbol, args);
  return(sc->F);
}


static s7_pointer g_internal_c_object_set(s7_scheme *sc, s7_pointer args)
{
  return((*(c_object_set(sc, car(args))))(sc, car(args), cdr(args)));
}

int32_t s7_make_c_type(s7_scheme *sc, const char *name)
{
  int32_t tag;
  tag = sc->num_c_object_types++;
  if (tag >= sc->c_object_types_size)
    {
      if (sc->c_object_types_size == 0)
	{
	  sc->c_object_types_size = 8;
	  sc->c_object_types = (c_object_t **)calloc(sc->c_object_types_size, sizeof(c_object_t *));
	}
      else
	{
	  sc->c_object_types_size = tag + 8;
	  sc->c_object_types = (c_object_t **)realloc((void *)(sc->c_object_types), sc->c_object_types_size * sizeof(c_object_t *));
	}
    }
  sc->c_object_types[tag] = (c_object_t *)calloc(1, sizeof(c_object_t));
  sc->c_object_types[tag]->type = tag;
  sc->c_object_types[tag]->name = copy_string(name);
  sc->c_object_types[tag]->scheme_name = s7_make_permanent_string(name);

  sc->c_object_types[tag]->free = fallback_free;
  sc->c_object_types[tag]->print = fallback_print;
  sc->c_object_types[tag]->equal = fallback_equal;
  sc->c_object_types[tag]->gc_mark = fallback_mark;
  sc->c_object_types[tag]->ref = fallback_ref;
  sc->c_object_types[tag]->set = fallback_set;
  sc->c_object_types[tag]->outer_type = T_C_OBJECT;
  sc->c_object_types[tag]->length = fallback_length;
  sc->c_object_types[tag]->copy = NULL;
  sc->c_object_types[tag]->reverse = NULL;
  sc->c_object_types[tag]->fill = NULL;
  sc->c_object_types[tag]->print_readably = fallback_print_readably;
  
  return(tag);
}

void s7_c_type_set_print(s7_scheme *sc, int32_t tag, char *(*print)(s7_scheme *sc, void *value)) 
{ 
  sc->c_object_types[tag]->print = print;
}

void s7_c_type_set_print_readably(s7_scheme *sc, int32_t type, char *(*printer)(s7_scheme *sc, void *val))
{
  sc->c_object_types[type]->print_readably = printer;
}

void s7_c_type_set_free(s7_scheme *sc, int32_t tag, void (*gc_free)(void *value))
{
  sc->c_object_types[tag]->free = gc_free;
}

void s7_c_type_set_equal(s7_scheme *sc, int32_t tag, bool (*equal)(void *val1, void *val2))
{
  sc->c_object_types[tag]->equal = equal;
}

void s7_c_type_set_mark(s7_scheme *sc, int32_t tag, void (*gc_mark)(void *val))
{
  sc->c_object_types[tag]->gc_mark = gc_mark;
}

void s7_c_type_set_apply(s7_scheme *sc, int32_t tag, s7_pointer (*ref)(s7_scheme *sc, s7_pointer obj, s7_pointer args))
{
  sc->c_object_types[tag]->ref = ref;
  if (sc->c_object_types[tag]->ref != fallback_ref)
    sc->c_object_types[tag]->outer_type = (T_C_OBJECT | T_SAFE_PROCEDURE);
}

void s7_c_type_set_apply_direct(s7_scheme *sc, int32_t tag, s7_pointer (*dref)(s7_scheme *sc, s7_pointer obj, s7_int index))
{
  sc->c_object_types[tag]->direct_ref = dref;  
}

void s7_c_type_set_set(s7_scheme *sc, int32_t tag, s7_pointer (*set)(s7_scheme *sc, s7_pointer obj, s7_pointer args))
{
  sc->c_object_types[tag]->set = set;
}

void s7_c_type_set_set_direct(s7_scheme *sc, int32_t tag, s7_pointer (*dset)(s7_scheme *sc, s7_pointer obj, s7_int index, s7_pointer val))
{
  sc->c_object_types[tag]->direct_set = dset;
}

void s7_c_type_set_length(s7_scheme *sc, int32_t tag, s7_pointer (*length)(s7_scheme *sc, s7_pointer obj))
{
  sc->c_object_types[tag]->length = length;
}

void s7_c_type_set_copy(s7_scheme *sc, int32_t tag, s7_pointer (*copy)(s7_scheme *sc, s7_pointer args))
{
  sc->c_object_types[tag]->copy = copy;
}

void s7_c_type_set_fill(s7_scheme *sc, int32_t tag, s7_pointer (*fill)(s7_scheme *sc, s7_pointer args))
{
  sc->c_object_types[tag]->fill = fill;
}

void s7_c_type_set_reverse(s7_scheme *sc, int32_t tag, s7_pointer (*reverse)(s7_scheme *sc, s7_pointer args))
{
  sc->c_object_types[tag]->reverse = reverse;
}


/* #if (!DISABLE_DEPRECATED) */
void s7_object_type_set_direct(int32_t tag, 
			       s7_pointer (*dref)(s7_scheme *sc, s7_pointer obj, s7_int index), 
			       s7_pointer (*dset)(s7_scheme *sc, s7_pointer obj, s7_int index, s7_pointer val))
{
  cur_sc->c_object_types[tag]->direct_ref = dref;
  cur_sc->c_object_types[tag]->direct_set = dset;
}

int s7_new_type(const char *name,
		char *(*print)(s7_scheme *sc, void *value),
		void (*gc_free)(void *value),
		bool (*equal)(void *val1, void *val2),
		void (*gc_mark)(void *val),
                s7_pointer (*ref)(s7_scheme *sc, s7_pointer obj, s7_pointer args),
                s7_pointer (*set)(s7_scheme *sc, s7_pointer obj, s7_pointer args))
{
  int32_t tag;
  tag = s7_make_c_type(cur_sc, name);

  if (gc_free) cur_sc->c_object_types[tag]->free = gc_free;
  if (print) cur_sc->c_object_types[tag]->print = print;
  if (equal) cur_sc->c_object_types[tag]->equal = equal;
  if (gc_mark) cur_sc->c_object_types[tag]->gc_mark = gc_mark;
  if (set) cur_sc->c_object_types[tag]->set = set;

  if (ref) cur_sc->c_object_types[tag]->ref = ref;
  if (cur_sc->c_object_types[tag]->ref != fallback_ref)
    cur_sc->c_object_types[tag]->outer_type = (T_C_OBJECT | T_SAFE_PROCEDURE);

  return(tag);
}

int s7_new_type_x(s7_scheme *sc,
		  const char *name,
		  char *(*print)(s7_scheme *sc, void *value),
		  void (*free)(void *value),
		  bool (*equal)(void *val1, void *val2),
		  void (*gc_mark)(void *val),
		  s7_pointer (*apply)(s7_scheme *sc, s7_pointer obj, s7_pointer args),
		  s7_pointer (*set)(s7_scheme *sc, s7_pointer obj, s7_pointer args),
		  s7_pointer (*length)(s7_scheme *sc, s7_pointer obj),
		  s7_pointer (*copy)(s7_scheme *sc, s7_pointer args),
		  s7_pointer (*reverse)(s7_scheme *sc, s7_pointer args),
		  s7_pointer (*fill)(s7_scheme *sc, s7_pointer args))
{
  int32_t tag;
  tag = s7_new_type(name, print, free, equal, gc_mark, apply, set);
  if (length) sc->c_object_types[tag]->length = length;
  sc->c_object_types[tag]->copy = copy;
  sc->c_object_types[tag]->reverse = reverse;
  sc->c_object_types[tag]->fill = fill;
  return(tag);
}
/* #endif */


static bool c_objects_are_equal(s7_scheme *sc, s7_pointer a, s7_pointer b)
{
  return((c_object_type(a) == c_object_type(b)) &&
	 ((*(c_object_eql(sc, a)))(c_object_value(a), c_object_value(b))));
}


void *s7_c_object_value(s7_pointer obj)
{
  return(c_object_value(obj));
}


void *s7_c_object_value_checked(s7_pointer obj, int32_t type)
{
  if ((is_c_object(obj)) &&
      (c_object_type(obj) == type))
    return(c_object_value(obj));
  return(NULL);
}


int s7_c_object_type(s7_pointer obj)
{
  if (is_c_object(obj))
    return(c_object_type(obj));
  return(-1);
}


s7_pointer s7_make_c_object_with_let(s7_scheme *sc, int32_t type, void *value, s7_pointer let)
{
  s7_pointer x;
  new_cell(sc, x, sc->c_object_types[type]->outer_type);

  /* c_object_info(x) = &(sc->c_object_types[type]); */
  /* that won't work because c_object_types can move when it is realloc'd and the old stuff is freed by realloc
   *   and since we're checking (for example) ref_2 existence as not null, we can't use a table of c_object_t's!
   */
  c_object_type(x) = type;
  c_object_value(x) = value;
  c_object_set_let(x, let);
  c_object_mark(x) = sc->c_object_types[type]->gc_mark;
  add_c_object(sc, x);
  return(x);
}

s7_pointer s7_make_c_object(s7_scheme *sc, int32_t type, void *value) 
{
  return(s7_make_c_object_with_let(sc, type, value, sc->nil));
}


s7_pointer s7_c_object_let(s7_pointer obj)
{
  return(c_object_let(obj));
}


s7_pointer s7_c_object_set_let(s7_pointer obj, s7_pointer e)
{
  c_object_set_let(obj, e);
  return(e);
}


static s7_pointer c_object_pi_direct(s7_pointer obj, s7_int i) {return((c_object_direct_ref(cur_sc, obj))(cur_sc, obj, i));}

static s7_pointer c_object_length(s7_scheme *sc, s7_pointer obj)
{
  if (c_object_len(sc, obj))
    return((*(c_object_len(sc, obj)))(sc, obj));
  eval_error(sc, "attempt to get length of ~S?", obj);
}


static s7_int c_object_length_to_int(s7_scheme *sc, s7_pointer obj)
{
  if (c_object_len(sc, obj))
    {
      s7_pointer res;
      res = (*(c_object_len(sc, obj)))(sc, obj);
      if (s7_is_integer(res))
	return(s7_integer(res));
    }
  return(-1);
}


static s7_pointer object_copy(s7_scheme *sc, s7_pointer args)
{
  s7_pointer obj;
  obj = car(args);
  check_method(sc, obj, sc->copy_symbol, args);
  if (c_object_copy(sc, obj))
    return((*(c_object_copy(sc, obj)))(sc, args));
  eval_error(sc, "attempt to copy ~S?", obj);
}




/* -------- dilambda -------- */

s7_pointer s7_dilambda(s7_scheme *sc,
		       const char *name,
		       s7_pointer (*getter)(s7_scheme *sc, s7_pointer args),
		       int32_t get_req_args, int32_t get_opt_args,
		       s7_pointer (*setter)(s7_scheme *sc, s7_pointer args),
		       int32_t set_req_args, int32_t set_opt_args,
		       const char *documentation)
{
  s7_pointer get_func, set_func;
  char *internal_set_name;
  int32_t len;

  len = 16 + safe_strlen(name);
  internal_set_name = (char *)malloc(len * sizeof(char));
  snprintf(internal_set_name, len, "[set-%s]", name);

  get_func = s7_make_safe_function(sc, name, getter, get_req_args, get_opt_args, false, documentation);
  s7_define(sc, sc->nil, make_symbol(sc, name), get_func);
  set_func = s7_make_function(sc, internal_set_name, setter, set_req_args, set_opt_args, false, documentation);
  c_function_set_setter(get_func, set_func);

  return(get_func);
}

s7_pointer s7_typed_dilambda(s7_scheme *sc,
			     const char *name,
			     s7_pointer (*getter)(s7_scheme *sc, s7_pointer args),
			     int32_t get_req_args, int32_t get_opt_args,
			     s7_pointer (*setter)(s7_scheme *sc, s7_pointer args),
			     int32_t set_req_args, int32_t set_opt_args,
			     const char *documentation,
			     s7_pointer get_sig, s7_pointer set_sig)
{
  s7_pointer get_func, set_func;
  get_func = s7_dilambda(sc, name, getter, get_req_args, get_opt_args, setter, set_req_args, set_opt_args, documentation);
  set_func = c_function_setter(get_func);
  if (get_sig) c_function_signature(get_func) = get_sig;
  if (set_sig) c_function_signature(set_func) = set_sig;
  return(get_func);
}


bool s7_is_dilambda(s7_pointer obj)
{
  return(((is_c_function(obj)) &&
	  (is_c_function(c_function_setter(obj)))) ||
	 ((is_any_closure(obj)) &&
	  (is_procedure(closure_setter(obj)))));
}

static s7_pointer g_is_dilambda(s7_scheme *sc, s7_pointer args)
{
  #define H_is_dilambda "(dilambda? obj) returns #t if obj is a procedure with setter."
  #define Q_is_dilambda pl_bt
  check_boolean_method(sc, s7_is_dilambda, sc->is_dilambda_symbol, args);
}

static s7_pointer c_set_setter(s7_scheme *sc, s7_pointer p, s7_pointer setter)
{
  switch (type(p))
    {
    case T_MACRO:   case T_MACRO_STAR:
    case T_BACRO:   case T_BACRO_STAR:
    case T_CLOSURE: case T_CLOSURE_STAR:
      closure_set_setter(p, setter);
      break;

    case T_C_FUNCTION:
    case T_C_ANY_ARGS_FUNCTION:
    case T_C_OPT_ARGS_FUNCTION:
    case T_C_RST_ARGS_FUNCTION:
    case T_C_FUNCTION_STAR:
      c_function_set_setter(p, setter);
      if ((is_any_closure(setter)) ||
	  (is_any_macro(setter)))
	add_setter(sc, p, setter);
      break;

    case T_C_MACRO:
      if ((is_any_closure(setter)) ||
	  (is_any_macro(setter)))
	add_setter(sc, p, setter);
      c_macro_set_setter(p, setter);
      break;
    }
  return(setter);
}

static s7_pointer g_dilambda(s7_scheme *sc, s7_pointer args)
{
  #define H_dilambda "(dilambda getter setter) sets getter's procedure-setter to be setter."
  #define Q_dilambda s7_make_signature(sc, 3, sc->is_procedure_symbol, sc->is_procedure_symbol, sc->is_procedure_symbol)
  s7_pointer getter, setter;

  getter = car(args);
  if (!is_any_procedure(getter))
    return(wrong_type_argument_with_type(sc, sc->dilambda_symbol, 1, getter, s7_make_string_wrapper(sc, "a procedure or macro")));

  setter = cadr(args);
  if (!is_any_procedure(setter))
    return(wrong_type_argument_with_type(sc, sc->dilambda_symbol, 2, setter, s7_make_string_wrapper(sc, "a procedure or macro")));
  
  c_set_setter(sc, getter, setter);
  return(getter);
}


s7_pointer s7_procedure_setter(s7_scheme *sc, s7_pointer obj)
{
  if (is_c_function(obj))
    return(c_function_setter(obj));

  return(closure_setter(obj));
}

static s7_pointer g_procedure_setter(s7_scheme *sc, s7_pointer args)
{
  #define H_procedure_setter "(procedure-setter obj) returns the setter associated with obj, or #f"
  #define Q_procedure_setter s7_make_signature(sc, 2, sc->T, sc->is_procedure_symbol)
  s7_pointer p;

  p = car(args);
  switch (type(p))
    {
    case T_MACRO:   case T_MACRO_STAR:
    case T_BACRO:   case T_BACRO_STAR:
    case T_CLOSURE: case T_CLOSURE_STAR:
      return(closure_setter(p));

    case T_C_FUNCTION:
    case T_C_FUNCTION_STAR:
    case T_C_ANY_ARGS_FUNCTION:
    case T_C_OPT_ARGS_FUNCTION:
    case T_C_RST_ARGS_FUNCTION:
      return(c_function_setter(p));

    case T_C_MACRO:
      return(c_macro_setter(p));

    case T_GOTO:
    case T_CONTINUATION:
      return(sc->F);

    case T_C_OBJECT: 
      /* this can satisfy procedure? if T_SAFE_PROCEDURE bit is set -- has apply method, see s7_c_type_set_apply */
      check_method(sc, p, s7_make_symbol(sc, "procedure-setter"), args);
      return(make_boolean(sc, (c_object_set(sc, p) != fallback_set)));
      /* unfortunately ref/set are not s7_functions (they have an extra object arg), so we can't return c_object_set */

    case T_LET:
      check_method(sc, p, s7_make_symbol(sc, "procedure-setter"), args);
      break;

    case T_ITERATOR:
      if (is_any_closure(iterator_sequence(p)))
	return(closure_setter(iterator_sequence(p)));
      return(sc->F);
    }
  return(s7_wrong_type_arg_error(sc, "procedure-setter", 0, p, "a procedure or a reasonable facsimile thereof"));
}

static s7_pointer g_procedure_set_setter(s7_scheme *sc, s7_pointer args)
{
  s7_pointer p, setter;

  p = car(args);
  if (!is_any_procedure(p))
    return(s7_wrong_type_arg_error(sc, "set! procedure-setter procedure", 1, p, "a procedure"));

  setter = cadr(args);
  if ((setter != sc->F) &&
      (!is_any_procedure(setter)))
    return(s7_wrong_type_arg_error(sc, "set! procedure-setter setter", 2, setter, "a procedure or #f"));

  /* should we check that p != setter?
   *   :(set! (procedure-setter <) <)
   *   <
   *   :(set! (< 3 2) 3)
   *   #f
   *   :(set! (< 1) 2)
   *   #t
   * can this make sense?
   */
  return(c_set_setter(sc, p, setter));
}


void s7_define_function_with_setter(s7_scheme *sc, const char *name, s7_function get_fnc, s7_function set_fnc, int32_t req_args, int32_t opt_args, const char *doc)
{
  s7_dilambda(sc, name, get_fnc, req_args, opt_args, set_fnc, req_args + 1, opt_args, doc);
}


/* -------------------------------- arity -------------------------------- */

static s7_pointer closure_arity_to_cons(s7_scheme *sc, s7_pointer x, s7_pointer x_args)
{
  /* x_args is unprocessed -- it is exactly the list as used in the closure[*] definition
   */
  int32_t len;

  if (is_symbol(x_args))                    /* any number of args is ok */
    return(s7_cons(sc, small_int(0), max_arity));

  if (closure_arity_unknown(x))
    closure_arity(x) = s7_list_length(sc, x_args);
  len = closure_arity(x);
  if (len < 0)                               /* dotted list => rest arg, (length '(a b . c)) is -2 */
    return(s7_cons(sc, s7_make_integer(sc, -len), max_arity));
  return(s7_cons(sc, s7_make_integer(sc, len), s7_make_integer(sc, len)));
}

static void closure_star_arity_1(s7_scheme *sc, s7_pointer x, s7_pointer args)
{
  if (closure_arity_unknown(x))
    {
      if (is_null(args))
	closure_arity(x) = 0;
      else
	{
	  if (allows_other_keys(args))
	    closure_arity(x) = -1;
	  else
	    {
	      s7_pointer p;
	      int32_t i;
	      for (i = 0, p = args; is_pair(p); p = cdr(p))
		{
		  s7_pointer arg;
		  arg = car(p);
		  if (arg == sc->key_rest_symbol)
		    break;
		  i++;
		}
	      if (is_null(p))
		closure_arity(x) = i;
	      else closure_arity(x) = -1; /* see below */
	    }
	}
    }
}

static s7_pointer closure_star_arity_to_cons(s7_scheme *sc, s7_pointer x, s7_pointer x_args)
{
  if (is_symbol(x_args))
    return(s7_cons(sc, small_int(0), max_arity));

  closure_star_arity_1(sc, x, x_args);

  if (closure_arity(x) == -1)
    return(s7_cons(sc, small_int(0), max_arity));
  return(s7_cons(sc, small_int(0), s7_make_integer(sc, closure_arity(x))));
}


static int32_t closure_arity_to_int(s7_scheme *sc, s7_pointer x)
{
  /* not lambda* here */
  if (closure_arity_unknown(x))
    {
      int32_t i;
      s7_pointer b;
      for (i = 0, b = closure_args(x); is_pair(b); i++, b = cdr(b)) {};
      if (is_null(b))
	closure_arity(x) = i;
      else
	{
	  if (i == 0)
	    return(-1);
	  closure_arity(x) = -i;
	}
    }
  return(closure_arity(x));
}


static int32_t closure_star_arity_to_int(s7_scheme *sc, s7_pointer x)
{
  /* not lambda here */
  closure_star_arity_1(sc, x, closure_args(x));
  return(closure_arity(x));
}


s7_pointer s7_arity(s7_scheme *sc, s7_pointer x)
{
  switch (type(x))
    {
    case T_C_OPT_ARGS_FUNCTION:
    case T_C_RST_ARGS_FUNCTION:
    case T_C_FUNCTION:
      return(s7_cons(sc, s7_make_integer(sc, c_function_required_args(x)), s7_make_integer(sc, c_function_all_args(x))));

    case T_C_ANY_ARGS_FUNCTION:
    case T_C_FUNCTION_STAR:
      return(s7_cons(sc, small_int(0), s7_make_integer(sc, c_function_all_args(x)))); /* should this be *2? */

    case T_MACRO:
    case T_BACRO:
    case T_CLOSURE:
      return(closure_arity_to_cons(sc, x, closure_args(x)));

    case T_MACRO_STAR:
    case T_BACRO_STAR:
    case T_CLOSURE_STAR:
      return(closure_star_arity_to_cons(sc, x, closure_args(x)));

    case T_C_MACRO:
      return(s7_cons(sc, s7_make_integer(sc, c_macro_required_args(x)), s7_make_integer(sc, c_macro_all_args(x))));

    case T_GOTO:
    case T_CONTINUATION:
      return(s7_cons(sc, small_int(0), max_arity));

    case T_STRING:
      if (string_length(x) == 0)
	return(sc->F);

    case T_LET:
      /* check_method(sc, x, sc->arity_symbol, args); */
      return(s7_cons(sc, small_int(1), small_int(1)));

    case T_C_OBJECT:
      /* check_method(sc, x, sc->arity_symbol, args); */
      if (is_procedure(x))
	return(s7_cons(sc, small_int(0), max_arity));
      return(sc->F);

    case T_INT_VECTOR:
    case T_FLOAT_VECTOR:
    case T_VECTOR:
      if (vector_length(x) == 0)
	return(sc->F);

    case T_PAIR:
    case T_HASH_TABLE:
      return(s7_cons(sc, small_int(1), max_arity));

    case T_ITERATOR:
      return(s7_cons(sc, small_int(0), small_int(0)));

    case T_SYNTAX:
      return(s7_cons(sc, small_int(syntax_min_args(x)), (syntax_max_args(x) == -1) ? max_arity : small_int(syntax_max_args(x))));
    }
  return(sc->F);
}


static s7_pointer g_arity(s7_scheme *sc, s7_pointer args)
{
  #define H_arity "(arity obj) the min and max acceptable args for obj if it is applicable, otherwise #f."
  #define Q_arity s7_make_signature(sc, 2, s7_make_signature(sc, 2, sc->is_pair_symbol, sc->is_boolean_symbol), sc->T)
  /* check_method(sc, p, sc->arity_symbol, args); */
  return(s7_arity(sc, car(args)));
}


static bool closure_is_aritable(s7_scheme *sc, s7_pointer x, s7_pointer x_args, int32_t args)
{
  /* x_args is unprocessed -- it is exactly the list as used in the closure definition
   */
  int32_t len;

  if (args == 0)
    return(!is_pair(x_args));

  if (is_symbol(x_args))                    /* any number of args is ok */
    return(true);

  len = closure_arity(x);
  if (len == CLOSURE_ARITY_NOT_SET)
    {
      len = s7_list_length(sc, x_args);
      closure_arity(x) = len;
    }
  if (len < 0)                               /* dotted list => rest arg, (length '(a b . c)) is -2 */
    return((-len) <= args);                  /*   so we have enough to take care of the required args */
  return(args == len);                       /* in a normal lambda list, there are no other possibilities */
}


static bool closure_star_is_aritable(s7_scheme *sc, s7_pointer x, s7_pointer x_args, int32_t args)
{
  if (is_symbol(x_args))
    return(true);

  closure_star_arity_1(sc, x, x_args);
  return((closure_arity(x) == -1) ||
	 (args <= closure_arity(x)));
}


bool s7_is_aritable(s7_scheme *sc, s7_pointer x, int32_t args)
{
  switch (type(x))
    {
    case T_C_RST_ARGS_FUNCTION:
    case T_C_FUNCTION:
      return(((int32_t)c_function_required_args(x) <= args) &&
	     ((int32_t)c_function_all_args(x) >= args));

    case T_C_OPT_ARGS_FUNCTION: /* any/opt req args == 0 */
    case T_C_ANY_ARGS_FUNCTION:
    case T_C_FUNCTION_STAR:
      return((int32_t)c_function_all_args(x) >= args);

    case T_MACRO:
    case T_BACRO:
    case T_CLOSURE:
      return(closure_is_aritable(sc, x, closure_args(x), args));

    case T_MACRO_STAR:
    case T_BACRO_STAR:
    case T_CLOSURE_STAR:
      return(closure_star_is_aritable(sc, x, closure_args(x), args));

    case T_C_MACRO:
      return(((int32_t)c_macro_required_args(x) <= args) &&
	     ((int32_t)c_macro_all_args(x) >= args));

    case T_GOTO:
    case T_CONTINUATION:
      return(true);

    case T_STRING:
      return((args == 1) &&
	     (string_length(x) > 0)); /* ("" 0) -> error */

    case T_C_OBJECT:
      /* check_method(sc, x, sc->is_aritable_symbol, list_2(sc, x, s7_make_integer(sc, args))); -- see below */
      return(is_procedure(x)); /* i.e. is_applicable */

    case T_INT_VECTOR:
    case T_FLOAT_VECTOR:
    case T_VECTOR:
      return((args > 0) &&
	     (vector_length(x) > 0) &&   /* (#() 0) -> error */
	     ((uint32_t)args <= vector_rank(x)));

    case T_LET:
      /* check_method(sc, x, sc->is_aritable_symbol, list_2(sc, x, s7_make_integer(sc, args))); */
      /* this slows us down a lot */
    case T_HASH_TABLE:
    case T_PAIR:
      return(args == 1);

    case T_ITERATOR:
      return(args == 0);

    case T_SYNTAX:
      return((args >= syntax_min_args(x)) && ((args <= syntax_max_args(x)) || (syntax_max_args(x) == -1)));
    }
  return(false);
}

static s7_pointer g_is_aritable(s7_scheme *sc, s7_pointer args)
{
  #define H_is_aritable "(aritable? obj num-args) returns #t if 'obj can be applied to 'num-args arguments."
  #define Q_is_aritable s7_make_signature(sc, 3, sc->is_boolean_symbol, sc->T, sc->is_integer_symbol)

  s7_pointer n;
  s7_int num;

  n = cadr(args);
  if (!s7_is_integer(n)) /* remember gmp case! */
    method_or_bust(sc, n, sc->is_aritable_symbol, args, T_INTEGER, 2);

  num = s7_integer(n);
  if (num < 0)
    return(out_of_range(sc, sc->is_aritable_symbol, small_int(2), n, its_negative_string));
  if (num > MAX_ARITY) num = MAX_ARITY;

  return(make_boolean(sc, s7_is_aritable(sc, car(args), (int32_t)num)));
}

static bool is_aritable_b_pp(s7_pointer f, s7_pointer i) {return(g_is_aritable(cur_sc, set_plist_2(cur_sc, f, i)) != cur_sc->F);}


/* -------- sequence? -------- */
static s7_pointer g_is_sequence(s7_scheme *sc, s7_pointer args)
{
  #define H_is_sequence "(sequence? obj) returns #t if obj is a sequence (vector, string, pair, etc)"
  #define Q_is_sequence pl_bt
  check_boolean_method(sc, is_simple_sequence, sc->is_sequence_symbol, args);
}

static bool is_sequence_b(s7_pointer p) {return(is_simple_sequence(p));}



/* -------------------------------- symbol-access ------------------------------------------------ */

static uint32_t protect_accessor(s7_scheme *sc, s7_pointer acc)
{
  uint32_t loc;
  if (sc->protected_accessors_size == sc->protected_accessors_loc)
    {
      int32_t i, new_size, size;
      size = sc->protected_accessors_size;
      new_size = 2 * size;
      vector_elements(sc->protected_accessors) = (s7_pointer *)realloc(vector_elements(sc->protected_accessors), new_size * sizeof(s7_pointer));
      vector_length(sc->protected_accessors) = new_size;
      for (i = size; i < new_size; i++)
	vector_element(sc->protected_accessors, i) = sc->gc_nil;
      sc->protected_accessors_size = new_size;
    }
  loc = sc->protected_accessors_loc++;
  vector_element(sc->protected_accessors, loc) = acc;
  return(loc);
}

s7_pointer s7_symbol_access(s7_scheme *sc, s7_pointer sym)
{
  /* these refer to the rootlet */
  if ((is_slot(global_slot(sym))) &&
      (slot_has_accessor(global_slot(sym))))
    return(slot_accessor(global_slot(sym)));
  return(sc->F);
}

s7_pointer s7_symbol_set_access(s7_scheme *sc, s7_pointer symbol, s7_pointer func)
{
  if (slot_has_accessor(global_slot(symbol)))
    {
      uint32_t index;
      index = symbol_global_accessor_index(symbol);
      if (index < sc->protected_accessors_size)
	{
	  if (is_immutable(vector_element(sc->protected_accessors, index))) /* a function */
	    return(func);
	  vector_element(sc->protected_accessors, index) = func;
	  slot_set_accessor(global_slot(symbol), func);
	  return(func);
	}
    }
  if (func != sc->F)
    {
      slot_set_has_accessor(global_slot(symbol));
      symbol_set_has_accessor(symbol);
      symbol_global_accessor_index(symbol) = protect_accessor(sc, func);  
    }
  slot_set_accessor(global_slot(symbol), func);
  return(func);
}

/* (let () (define xxx 23) (define (hix) (set! xxx 24)) (hix) (set! (symbol-access 'xxx) (lambda (sym val) (format *stderr* "val: ~A~%" val) val)) (hix))
 *    so set symbol-access before use!
 */

static s7_pointer g_symbol_access(s7_scheme *sc, s7_pointer args)
{
  #define H_symbol_access "(symbol-access sym (env (curlet))) is the function called when the symbol is set!."
  #define Q_symbol_access s7_make_signature(sc, 3, s7_make_signature(sc, 2, sc->is_boolean_symbol, sc->is_procedure_symbol), sc->is_symbol_symbol, sc->is_let_symbol)
  s7_pointer sym, p;

  sym = car(args);
  if (!is_symbol(sym))
    method_or_bust_one_arg(sc, sym, sc->symbol_access_symbol, args, T_SYMBOL);
  if (is_keyword(sym))
    return(sc->F);

  if (is_pair(cdr(args)))
    {
      s7_pointer e, old_e;
      e = cadr(args);
      if ((e == sc->rootlet) || (e == sc->nil))
	p = global_slot(sym);
      else
	{
	  if (!is_let(e))
	    return(wrong_type_argument(sc, sc->symbol_access_symbol, 2, e, T_LET));
	  old_e = sc->envir;
	  sc->envir = e;
	  p = find_symbol(sc, sym);
	  sc->envir = old_e;
	}
    }
  else p = find_symbol(sc, sym);

  if (!is_slot(p))
    return(sc->F);

  if (slot_has_accessor(p))
    return(slot_accessor(p));

  return(sc->F);
}


static s7_pointer g_symbol_set_access(s7_scheme *sc, s7_pointer args)
{
  s7_pointer sym, func, p;

  sym = car(args);
  if (!is_symbol(sym))                 /* no check method because no method name? */
    return(s7_wrong_type_arg_error(sc, "set! symbol-access", 1, sym, "a symbol"));
  if (is_keyword(sym))
    return(s7_wrong_type_arg_error(sc, "set! symbol-access", 1, sym, "a normal symbol (a keyword can't be set)"));

  /* (set! (symbol-access sym) f) or (set! (symbol-access sym env) f) */
  if (is_pair(cddr(args)))
    {
      s7_pointer e, old_e;
      e = cadr(args);
      func = caddr(args);
      if ((e == sc->rootlet) || (e == sc->nil))
	p = global_slot(sym);
      else
	{
	  if (!is_let(e))
	    return(s7_wrong_type_arg_error(sc, "set! symbol-access", 2, e, "a let"));
	  old_e = sc->envir;
	  sc->envir = e;
	  p = find_symbol(sc, sym);
	  sc->envir = old_e;
	}
    }
  else
    {
      p = find_symbol(sc, sym);
      func = cadr(args);
    }

  if ((!is_procedure_or_macro(func)) &&
      (func != sc->F))
    return(s7_wrong_type_arg_error(sc, "set! symbol-access", 3, func, "a function or #f"));

  if (!is_slot(p))
    return(sc->F);

  if (p == global_slot(sym))
    {
      s7_symbol_set_access(sc, sym, func); /* special GC protection for global vars */
      return(func);
    }

  slot_set_accessor(p, func);
  if (func != sc->F)
    {
      slot_set_has_accessor(p);
      symbol_set_has_accessor(sym);
    }
  return(func);
}


static s7_pointer bind_accessed_symbol(s7_scheme *sc, opcode_t op, s7_pointer symbol, s7_pointer new_value)
{
  /* this refers to (define (sym ...)) and friends -- define cases
   *    see call_accessor for the set! cases
   */
  s7_pointer func;

  func = g_symbol_access(sc, set_plist_2(sc, symbol, sc->envir));
  if (is_procedure_or_macro(func))
    {
      if (is_c_function(func))
	{
	  s7_pointer old_value;
	  old_value = new_value;
	  set_car(sc->t2_1, symbol);
	  set_car(sc->t2_2, new_value);
	  new_value = c_function_call(func)(sc, sc->t2_1);
	  if (new_value == sc->error_symbol)
 	    return(s7_error(sc, sc->error_symbol, set_elist_3(sc, s7_make_string_wrapper(sc, "can't bind ~S to ~S"), symbol, old_value)));
	}
      else
	{
	  sc->args = list_2(sc, symbol, new_value);
	  push_stack(sc, op, sc->args, sc->code);
	  sc->code = func;
	  return(sc->no_value); /* this means the accessor in set! needs to goto APPLY to get the new value */
	}
    }
  return(new_value);
}



/* -------------------------------- hooks -------------------------------- */

s7_pointer s7_hook_functions(s7_scheme *sc, s7_pointer hook)
{
  return(s7_symbol_local_value(sc, sc->body_symbol, closure_let(hook)));
}


s7_pointer s7_hook_set_functions(s7_scheme *sc, s7_pointer hook, s7_pointer functions)
{
  if (s7_is_list(sc, functions))
    s7_let_set(sc, closure_let(hook), sc->body_symbol, functions);
  return(functions);
}



/* -------------------------------- eq etc -------------------------------- */

bool s7_is_eq(s7_pointer obj1, s7_pointer obj2)
{
  return(obj1 == obj2); /* so floats and NaNs might be eq? but not eqv? */
}


static s7_pointer g_is_eq(s7_scheme *sc, s7_pointer args)
{
  #define H_is_eq "(eq? obj1 obj2) returns #t if obj1 is eq to (the same object as) obj2"
  #define Q_is_eq pcl_bt
  return(make_boolean(sc, ((car(args) == cadr(args)) ||
			   ((is_unspecified(car(args))) && (is_unspecified(cadr(args)))))));
  /* (eq? (apply apply apply values '(())) #<unspecified>) should return #t
   */
}


bool s7_is_eqv(s7_pointer a, s7_pointer b)
{
#if WITH_GMP
  if ((is_big_number(a)) || (is_big_number(b)))
    return(big_numbers_are_eqv(a, b));
#endif

  if (type(a) != type(b))
    return(false);

  if ((a == b) && (!is_number(a)))
    return(true);

  if (is_string(a))
    return(string_value(a) == string_value(b));

  if (s7_is_number(a))
    return(numbers_are_eqv(a, b));

  if (is_unspecified(a))                             /* types are the same so we know b is also unspecified */
    return(true);

  return(false);
}


static s7_pointer g_is_eqv(s7_scheme *sc, s7_pointer args)
{
  #define H_is_eqv "(eqv? obj1 obj2) returns #t if obj1 is equivalent to obj2"
  #define Q_is_eqv pcl_bt
  return(make_boolean(sc, s7_is_eqv(car(args), cadr(args))));
}



static bool floats_are_morally_equal(s7_scheme *sc, s7_double x, s7_double y)
{
  if (x == y) return(true);

  if ((is_NaN(x)) || (is_NaN(y)))
    return((is_NaN(x)) && (is_NaN(y)));

  return(fabs(x - y) <= sc->morally_equal_float_epsilon);
}

static bool eq_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
  return(x == y);
}

static bool symbol_morally_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
  if (x == y) return(true);
  if (!is_symbol(y)) return(false);                   /* (morally-equal? ''(1) '(1)) */
  return((is_slot(global_slot(x))) &&                 /* the optimizer can replace the original symbol with its own */
	 (is_syntax(slot_value(global_slot(x)))) &&
	 (is_slot(global_slot(y))) &&
	 (is_syntax(slot_value(global_slot(y)))) &&
	 (syntax_symbol(slot_value(global_slot(x))) == syntax_symbol(slot_value(global_slot(y)))));
}

static bool unspecified_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
  return(is_unspecified(y));
}

static bool c_pointer_morally_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
  return((s7_is_c_pointer(y)) && 
	 (raw_pointer(x) == raw_pointer(y)));
}

static bool c_pointer_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
  return((s7_is_c_pointer(y)) && 
	 (raw_pointer(x) == raw_pointer(y)) &&
	 (raw_pointer_type(x) == raw_pointer_type(y)) &&
	 (raw_pointer_info(x) == raw_pointer_info(y)));    /* should these use s7_is_equal? */
}

static bool string_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
  return((is_string(y)) && (scheme_strings_are_equal(x, y)));
}

static bool syntax_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
  return((is_syntax(y)) && (syntax_symbol(x) == syntax_symbol(y)));
}

static bool c_object_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
  return((is_c_object(y)) && (c_objects_are_equal(sc, x, y)));
}

static bool port_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
  return(x == y);
}

static bool port_morally_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
  if (x == y) return(true);
  if ((type(x) != type(y)) || (port_type(x) != port_type(y))) return(false);
  if ((port_is_closed(x)) && (port_is_closed(y))) return(true);
  return((is_string_port(x)) &&
	 (port_position(x) == port_position(y)) &&
	 (port_data_size(x) == port_data_size(y)) &&
	 (local_strncmp((const char *)port_data(x), (const char *)port_data(y), (is_input_port(x)) ? port_data_size(x) : port_position(x))));
}

#define equal_ref(Sc, X, Y, Ci) \
  do {   \
    /* here we know x and y are pointers to the same type of structure */ \
    int32_t ref_y;							\
    ref_y = (is_collected(Y)) ? peek_shared_ref(Ci, Y) : 0;		\
    if (is_collected(X))						\
      {									\
	int32_t ref_x;							\
	ref_x = peek_shared_ref(Ci, X);					\
	if (ref_y != 0) return(ref_x == ref_y);				\
	/* try to harmonize the new guy -- there can be more than one structure equal to the current one */ \
	if (ref_x != 0) add_shared_ref(Ci, Y, ref_x);			\
      }									\
    else								\
      {									\
	if (ref_y != 0)							\
	  add_shared_ref(Ci, X, ref_y);					\
	else								\
	  {								\
	    /* assume neither x nor y is in the table, and that they should share a ref value, \
	     *   called only in equality check, not printer.		\
	     */								\
	    if (Ci->top >= Ci->size2) enlarge_shared_info(Ci);		\
	    set_collected(X);						\
	    set_collected(Y);						\
	    Ci->objs[Ci->top] = X;					\
	    Ci->refs[Ci->top++] = ++Ci->ref;				\
	    Ci->objs[Ci->top] = Y;					\
	    Ci->refs[Ci->top++] = Ci->ref;				\
	  }								\
      }									\
  } while (0)

static bool s7_is_equal_1(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci);
static bool s7_is_morally_equal_1(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci);

static bool hash_table_morally_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
  hash_entry_t **lists;
  int32_t i, len;
  shared_info *nci = ci;

  if (x == y)
    return(true);
  if (!is_hash_table(y))
    {
      if (has_methods(y))
	{
	  s7_pointer equal_func;
	  equal_func = find_method(sc, find_let(sc, y), sc->is_morally_equal_symbol);
	  if (equal_func != sc->undefined)
	    return(s7_boolean(sc, s7_apply_function(sc, equal_func, list_2(sc, y, x))));
	}
      return(false);
    }
  if (ci)
    equal_ref(sc, x, y, ci);

  if (hash_table_entries(x) != hash_table_entries(y))
    return(false);
  if (hash_table_entries(x) == 0)
    return(true);
  len = hash_table_mask(x) + 1;
  lists = hash_table_elements(x);
  if (!nci) nci = new_shared_info(sc);

  for (i = 0; i < len; i++)
    {
      hash_entry_t *p;
      for (p = lists[i]; p; p = p->next)
	{
	  hash_entry_t *y_val;
	  y_val = (*hash_table_checker(y))(sc, y, p->key);

	  if ((!y_val) ||
	      (!s7_is_morally_equal_1(sc, p->value, y_val->value, nci)))
	    return(false);
	}
    }
  return(true);
}

static bool hash_table_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
  hash_entry_t **lists;
  int32_t i, len;
  shared_info *nci = ci;

  if (x == y)
    return(true);
  if (!is_hash_table(y)) return(false);
  if (ci)
    equal_ref(sc, x, y, ci);

  if (hash_table_entries(x) != hash_table_entries(y))
    return(false);
  if (hash_table_entries(x) == 0)
    return(true);
  if ((hash_table_checker_locked(x)) || (hash_table_checker_locked(y)))
    {
      if (hash_table_checker(x) != hash_table_checker(y))
	return(false);
      if (hash_table_mapper(x) != hash_table_mapper(y))
	return(false);
    }

  len = hash_table_mask(x) + 1;
  lists = hash_table_elements(x);
  if (!nci) nci = new_shared_info(sc);

  for (i = 0; i < len; i++)
    {
      hash_entry_t *p;
      for (p = lists[i]; p; p = p->next)
	{
	  hash_entry_t *y_val;
	  y_val = (*hash_table_checker(y))(sc, y, p->key);

	  if ((!y_val) ||
	      (!s7_is_equal_1(sc, p->value, y_val->value, nci)))
	    return(false);
	}
    }
  /* if we get here, every key/value in x has a corresponding key/value in y, and the number of entries match,
   *   so surely the tables are equal??
   */
  return(true);
}


static bool slots_match(s7_scheme *sc, s7_pointer px, s7_pointer y, shared_info *nci)
{
  s7_pointer ey, py;
  for (ey = y; (is_let(ey)) && (ey != sc->rootlet); ey = outlet(ey))
    for (py = let_slots(ey); is_slot(py); py = next_slot(py))
      if (slot_symbol(px) == slot_symbol(py)) /* we know something will match */
	return(s7_is_equal_1(sc, slot_value(px), slot_value(py), nci));
  return(false);
}

static bool slots_morally_match(s7_scheme *sc, s7_pointer px, s7_pointer y, shared_info *nci)
{
  s7_pointer ey, py;
  for (ey = y; (is_let(ey)) && (ey != sc->rootlet); ey = outlet(ey))
    for (py = let_slots(ey); is_slot(py); py = next_slot(py))
      if (slot_symbol(px) == slot_symbol(py)) /* we know something will match */
	return(s7_is_morally_equal_1(sc, slot_value(px), slot_value(py), nci));
  return(false);
}

static bool let_equal_1(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci, bool morally)
{
  s7_pointer ex, ey, px, py;
  shared_info *nci = ci;
  int32_t x_len, y_len;

  if (!is_let(y))
    return(false);
  if ((x == sc->rootlet) || (y == sc->rootlet))
    return(false);

  if (ci)
    equal_ref(sc, x, y, ci);

  clear_symbol_list(sc);
  for (x_len = 0, ex = x; (is_let(ex)) && (ex != sc->rootlet); ex = outlet(ex))
    for (px = let_slots(ex); is_slot(px); px = next_slot(px))
      if (!symbol_is_in_list(sc, slot_symbol(px)))
	{
	  add_symbol_to_list(sc, slot_symbol(px));
	  x_len++;
	}

  for (ey = y; (is_let(ey)) && (ey != sc->rootlet); ey = outlet(ey))
    for (py = let_slots(ey); is_slot(py); py = next_slot(py))
      if (!symbol_is_in_list(sc, slot_symbol(py)))          /* symbol in y, not in x */
	return(false);

  for (y_len = 0, ey = y; (is_let(ey)) && (ey != sc->rootlet); ey = outlet(ey))
    for (py = let_slots(ey); is_slot(py); py = next_slot(py))
      if (symbol_tag(slot_symbol(py)) != 0)
	{
	  y_len ++;
	  symbol_set_tag(slot_symbol(py), 0);
	}
  
  if (x_len != y_len)                                        /* symbol in x, not in y */
    return(false);

  if (!nci) nci = new_shared_info(sc);

  for (ex = x; (is_let(ex)) && (ex != sc->rootlet); ex = outlet(ex))
    for (px = let_slots(ex); is_slot(px); px = next_slot(px))
      if (symbol_tag(slot_symbol(px)) == 0)                /* unshadowed */
	{
	  symbol_set_tag(slot_symbol(px), sc->syms_tag);      /* values don't match */
	  if (((!morally) && (!slots_match(sc, px, y, nci))) ||
	      ((morally) && (!slots_morally_match(sc, px, y, nci))))
	    return(false);
	}
  return(true);
}


static bool let_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
  /* x == y if all unshadowed vars match, leaving aside the rootlet, so that for any local variable,
   *   we get the same value in either x or y.
   */
  if (x == y)
    return(true);
  return(let_equal_1(sc, x, y, ci, false));
}

static bool let_morally_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
  s7_pointer equal_func;

  if (x == y)
    return(true);

  if (has_methods(x))
    {
      equal_func = find_method(sc, find_let(sc, x), sc->is_morally_equal_symbol);
      if (equal_func != sc->undefined)
	return(s7_boolean(sc, s7_apply_function(sc, equal_func, list_2(sc, x, y))));
    }
  if (has_methods(y))
    {
      equal_func = find_method(sc, find_let(sc, y), sc->is_morally_equal_symbol);
      if (equal_func != sc->undefined)
	return(s7_boolean(sc, s7_apply_function(sc, equal_func, list_2(sc, y, x))));
    }
  return(let_equal_1(sc, x, y, ci, true));
}

static bool closure_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
  if (x == y)
    return(true);
  if (type(x) != type(y))
    return(false);
  if ((has_methods(x)) &&
      (has_methods(y)))
    {
      s7_pointer equal_func;
      equal_func = find_method(sc, closure_let(x), sc->is_equal_symbol);
      if (equal_func != sc->undefined)
	return(s7_boolean(sc, s7_apply_function(sc, equal_func, list_2(sc, x, y))));
    }
  return(false);
}

static bool closure_morally_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
  if (x == y)
    return(true);
  if (type(x) != type(y))
    return(false);
  if ((has_methods(x)) &&
      (has_methods(y)))
    {
      s7_pointer equal_func;
      equal_func = find_method(sc, closure_let(x), sc->is_morally_equal_symbol);
      if (equal_func != sc->undefined)
	return(s7_boolean(sc, s7_apply_function(sc, equal_func, list_2(sc, x, y))));
    }
  /* not sure about this -- we can't simply check let_equal(closure_let(x), closure_let(y))
   *   because locally defined constant functions on the second pass find the outer let.
   */
  return((s7_is_morally_equal_1(sc, closure_args(x), closure_args(y), ci)) &&
	 (s7_is_morally_equal_1(sc, closure_body(x), closure_body(y), ci)));
}

static bool pair_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
  shared_info *nci;
  s7_pointer px, py;

  if (x == y)
    return(true);
  if (!is_pair(y))
    return(false);

  if (ci)
    {
      nci = ci;
      equal_ref(sc, x, y, ci);
    }
  else nci = new_shared_info(sc);
  
  if (!s7_is_equal_1(sc, car(x), car(y), nci)) return(false);
  for (px = cdr(x), py = cdr(y); (is_pair(px)) && (is_pair(py)); px = cdr(px), py = cdr(py))
    {
      if (!s7_is_equal_1(sc, car(px), car(py), nci)) return(false);
      equal_ref(sc, px, py, nci);
    }
  if (px == py) /* normally nil? */
    return(true);
  return(s7_is_equal_1(sc, px, py, nci));
}

static bool pair_morally_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
  shared_info *nci;
  s7_pointer px, py;

  if (x == y)
    return(true);
  if (!is_pair(y))
    {
      if (has_methods(y))
	{
	  s7_pointer equal_func;
	  equal_func = find_method(sc, find_let(sc, y), sc->is_morally_equal_symbol);
	  if (equal_func != sc->undefined)
	    return(s7_boolean(sc, s7_apply_function(sc, equal_func, list_2(sc, y, x))));
	}
      return(false);
    }

  if (ci)
    {
      nci = ci;
      equal_ref(sc, x, y, ci);
    }
  else nci = new_shared_info(sc);
  
  if (!s7_is_morally_equal_1(sc, car(x), car(y), nci)) return(false);
  for (px = cdr(x), py = cdr(y); (is_pair(px)) && (is_pair(py)); px = cdr(px), py = cdr(py))
    {
      if (!s7_is_morally_equal_1(sc, car(px), car(py), nci)) return(false);
      equal_ref(sc, px, py, nci);
    }
  if (px == py) /* normally nil? */
    return(true);
  return(s7_is_morally_equal_1(sc, px, py, nci));
}

static bool vector_rank_match(s7_scheme *sc, s7_pointer x, s7_pointer y)
{
  int32_t x_dims, y_dims;

  if (vector_has_dimensional_info(x))
    x_dims = vector_ndims(x);
  else x_dims = 1;
  if (vector_has_dimensional_info(y))
    y_dims = vector_ndims(y);
  else y_dims = 1;

  if (x_dims != y_dims)
    return(false);

  if (x_dims > 1)
    {
      int32_t j;
      for (j = 0; j < x_dims; j++)
	if (vector_dimension(x, j) != vector_dimension(y, j))
	  return(false);
    }
  return(true);
}

static bool vector_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
  s7_int i, len;
  shared_info *nci = ci;

  if (x == y)
    return(true);
  if (!s7_is_vector(y)) return(false);

  len = vector_length(x);
  if (len != vector_length(y)) return(false);
  if (len == 0)
    return(vector_rank_match(sc, x, y));

  if (type(x) != type(y))
    return(false);
  if (!vector_rank_match(sc, x, y))
    return(false);

  if (is_float_vector(x))
    {
      for (i = 0; i < len; i++)
	{
	  s7_double z;
	  z = float_vector_element(x, i);
	  if ((is_NaN(z)) ||
	      (z != float_vector_element(y, i)))
	    return(false);
	}
      return(true);
    }

  if (is_int_vector(x))
    {
      for (i = 0; i < len; i++)
	if (int_vector_element(x, i) != int_vector_element(y, i))
	  return(false);
      return(true);
    }

  if (ci)
    equal_ref(sc, x, y, ci);
  else nci = new_shared_info(sc);

  for (i = 0; i < len; i++)
    if (!(s7_is_equal_1(sc, vector_element(x, i), vector_element(y, i), nci)))
      return(false);
  return(true);
}

static bool vector_morally_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
  s7_int i, len;
  shared_info *nci = ci;

  if (x == y)
    return(true);
  if (!s7_is_vector(y))
    {
      if (has_methods(y))
	{
	  s7_pointer equal_func;
	  equal_func = find_method(sc, find_let(sc, y), sc->is_morally_equal_symbol);
	  if (equal_func != sc->undefined)
	    return(s7_boolean(sc, s7_apply_function(sc, equal_func, list_2(sc, y, x))));
	}
      return(false);
    }
  len = vector_length(x);
  if (len != vector_length(y)) return(false);
  if (len == 0) return(true);
  if (!vector_rank_match(sc, x, y))
    return(false);

  if (type(x) != type(y))
    {
      /* (morally-equal? (make-int-vector 3 0) (make-vector 3 0)) -> #t
       * (morally-equal? (make-float-vector 3 1.0) (vector 1 1 1)) -> #t
       */
      for (i = 0; i < len; i++)
	if (!s7_is_morally_equal_1(sc, vector_getter(x)(sc, x, i), vector_getter(y)(sc, y, i), NULL)) /* this could be greatly optimized */
	  return(false);
      return(true);
    }

  if (is_float_vector(x))
    {
      s7_double *arr1, *arr2;
      s7_double fudge;
      arr1 = float_vector_elements(x);
      arr2 = float_vector_elements(y);
      fudge = sc->morally_equal_float_epsilon;
      if (fudge == 0.0)
	{
	  for (i = 0; i < len; i++)
	    if ((arr1[i] != arr2[i]) &&
		((!is_NaN(arr1[i])) || (!is_NaN(arr2[i]))))
	      return(false);
	}
      else
	{
	  for (i = 0; i < len; i++)
	    {
	      s7_double diff;
	      diff = fabs(arr1[i] - arr2[i]);
	      if (diff > fudge) return(false);
	      if ((is_NaN(diff)) &&
		  ((!is_NaN(arr1[i])) || (!is_NaN(arr2[i]))))
		return(false);
	    }
	}
      return(true);
    }

  if (is_int_vector(x))
    {
      for (i = 0; i < len; i++)
	if (int_vector_element(x, i) != int_vector_element(y, i))
	  return(false);
      return(true);
    }

  if (ci)
    equal_ref(sc, x, y, ci);
  else nci = new_shared_info(sc);

  for (i = 0; i < len; i++)
    if (!(s7_is_morally_equal_1(sc, vector_element(x, i), vector_element(y, i), nci)))
      return(false);
  return(true);
}

static bool iterator_equal_1(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci, bool morally)
{
  if (x == y) return(true);
  if (!is_iterator(y)) return(false);

  switch (type(iterator_sequence(x)))
    {
    case T_STRING:
      return((is_string(iterator_sequence(y))) &&
	     (iterator_position(x) == iterator_position(y)) &&
	     (string_equal(sc, iterator_sequence(x), iterator_sequence(y), ci)));

    case T_VECTOR:
    case T_INT_VECTOR:
    case T_FLOAT_VECTOR:
      return((s7_is_vector(iterator_sequence(y))) &&
	     (iterator_position(x) == iterator_position(y)) &&
	     ((morally) ? (vector_morally_equal(sc, iterator_sequence(x), iterator_sequence(y), ci)) : 
	                  (vector_equal(sc, iterator_sequence(x), iterator_sequence(y), ci))));

    case T_PAIR:
      return((iterator_sequence(x) == iterator_sequence(y)) &&
	     (iterator_next(x) == iterator_next(y)) &&           /* even if seqs are equal, one might be at end */
	     (iterator_current(x) == iterator_current(y)));      /* current pointer into the sequence */

    case T_HASH_TABLE:
      return((iterator_sequence(x) == iterator_sequence(y)) &&
	     (iterator_next(x) == iterator_next(y)) &&
	     (iterator_current(x) == iterator_current(y)) &&
	     (iterator_hash_current(x) == iterator_hash_current(y)) &&
	     (iterator_position(x) == iterator_position(y)));

    default:
      break;
    }
  return(false);
}

static bool iterator_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
  return(iterator_equal_1(sc, x, y, ci, false));
}

static bool iterator_morally_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
  return(iterator_equal_1(sc, x, y, ci, true));
}

static bool bignum_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
  if (!s7_is_number(y)) return(false);
#if WITH_GMP
  return(big_numbers_are_eqv(x, y));
#else
  return(false);
#endif
}

static bool bignum_morally_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
  if (!s7_is_number(y)) return(false);
#if WITH_GMP
  return(big_equal(sc, set_plist_2(sc, x, y)) != sc->F);
#else
  return(false);
#endif
}

static bool integer_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
#if WITH_GMP
  if (is_big_number(y))
    return(big_numbers_are_eqv(x, y));
#endif
  if (is_integer(y))
    return(integer(x) == integer(y));
  return(false);
}

static bool integer_morally_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
#if WITH_GMP
  if (is_big_number(y))
    return(big_equal(sc, set_plist_2(sc, x, y)) != sc->F);
#endif
  if (is_integer(y))
    return(integer(x) == integer(y));
  if (!is_number(y))
    return(false);

  if (is_t_real(y))
    return((!is_NaN(real(y))) &&
	   (fabs(integer(x) - real(y)) <= sc->morally_equal_float_epsilon));

  if (is_t_ratio(y))
    return(s7_fabsl(integer(x) - fraction(y)) <= sc->morally_equal_float_epsilon);

  return((!is_NaN(real_part(y))) &&
	 (!is_NaN(imag_part(y))) &&
	 (fabs(integer(x) - real_part(y)) <= sc->morally_equal_float_epsilon) &&
	 (fabs(imag_part(y)) <= sc->morally_equal_float_epsilon));
}

/* apparently ratio_equal is predefined in g++ -- name collision on mac */
static bool fraction_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
#if WITH_GMP
  if (is_big_number(y))
    return(big_numbers_are_eqv(x, y));
#endif
  return((s7_is_ratio(y)) &&
	 (numerator(x) == numerator(y)) &&
	 (denominator(x) == denominator(y)));
}

static bool fraction_morally_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
#if WITH_GMP
  if (is_big_number(y))
    return(big_equal(sc, set_plist_2(sc, x, y)) != sc->F);
#endif
  if (is_t_ratio(y))
    return(s7_fabsl(fraction(x) - fraction(y)) <= sc->morally_equal_float_epsilon);

  if (is_t_real(y))
    return(floats_are_morally_equal(sc, fraction(x), real(y)));

  if (is_integer(y))
    return(s7_fabsl(fraction(x) - integer(y)) <= sc->morally_equal_float_epsilon);

  if (is_t_complex(y))
    return((!is_NaN(real_part(y))) &&
	   (!is_NaN(imag_part(y))) &&
	   (s7_fabsl(fraction(x) - real_part(y)) <= sc->morally_equal_float_epsilon) &&
	   (fabs(imag_part(y)) <= sc->morally_equal_float_epsilon));
  return(false);
}

static bool real_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
#if WITH_GMP
  if (is_big_number(y))
    return(big_numbers_are_eqv(x, y));
#endif
  return((is_t_real(y)) &&
	 (real(x) == real(y)));
}

static bool real_morally_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
#if WITH_GMP
  if (is_big_number(y))
    return(big_equal(sc, set_plist_2(sc, x, y)) != sc->F);
#endif
  if (!is_number(y)) return(false);

  if (is_t_real(y))
    return(floats_are_morally_equal(sc, real(x), real(y)));

  if (is_integer(y))
    return((!is_NaN(real(x))) &&
	   (fabs(real(x) - integer(y)) <= sc->morally_equal_float_epsilon));

  if (is_t_ratio(y))
    return(floats_are_morally_equal(sc, real(x), fraction(y)));

  if (is_NaN(real(x)))
    return((is_NaN(real_part(y))) &&
	   (fabs(imag_part(y)) <= sc->morally_equal_float_epsilon));

  return((!is_NaN(real(x))) &&
	 (!is_NaN(real_part(y))) &&
	 (!is_NaN(imag_part(y))) &&
	 ((real(x) == real_part(y)) ||
	  (fabs(real(x) - real_part(y)) <= sc->morally_equal_float_epsilon)) &&
	 (fabs(imag_part(y)) <= sc->morally_equal_float_epsilon));
}

static bool complex_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
#if WITH_GMP
  if (is_big_number(y))
    return(big_numbers_are_eqv(x, y));
#endif
  return((is_t_complex(y)) &&
	 (!is_NaN(real_part(x))) &&
	 (!is_NaN(imag_part(x))) &&
	 (real_part(x) == real_part(y)) &&
	 (imag_part(x) == imag_part(y)));
}

static bool complex_morally_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
#if WITH_GMP
  if (is_big_number(y))
    return(big_equal(sc, set_plist_2(sc, x, y)) != sc->F);
#endif
  if (!is_number(y)) return(false);

  if (is_integer(y))
    return((!is_NaN(real_part(x))) &&
	   (!is_NaN(imag_part(x))) &&
	   (fabs(real_part(x) - integer(y)) <= sc->morally_equal_float_epsilon) &&
	   (fabs(imag_part(x)) <= sc->morally_equal_float_epsilon));

  if (s7_is_ratio(y))
    return((!is_NaN(real_part(x))) &&
	   (!is_NaN(imag_part(x))) &&
	   (s7_fabsl(real_part(x) - fraction(y)) <= sc->morally_equal_float_epsilon) &&
	   (fabs(imag_part(x)) <= sc->morally_equal_float_epsilon));

  if (is_real(y))
    {
      if (is_NaN(imag_part(x)))
	return(false);
      if (is_NaN(real(y)))
	return((is_NaN(real_part(x))) &&
	       (fabs(imag_part(x)) <= sc->morally_equal_float_epsilon));
      return(((real_part(x) == real(y)) ||
	      (fabs(real_part(x) - real(y)) <= sc->morally_equal_float_epsilon)) &&
	     (fabs(imag_part(x)) <= sc->morally_equal_float_epsilon));
    }

  /* should (morally-equal? nan.0 (complex nan.0 nan.0)) be #t (it's #f above)? */
  if (is_NaN(real_part(x)))
    return((is_NaN(real_part(y))) &&
	   (((is_NaN(imag_part(x))) && (is_NaN(imag_part(y)))) ||
	    (imag_part(x) == imag_part(y)) ||
	    (fabs(imag_part(x) - imag_part(y)) <= sc->morally_equal_float_epsilon)));

  if (is_NaN(imag_part(x)))
    return((is_NaN(imag_part(y))) &&
	   ((real_part(x) == real_part(y)) ||
	    (fabs(real_part(x) - real_part(y)) <= sc->morally_equal_float_epsilon)));

  if ((is_NaN(real_part(y))) ||
      (is_NaN(imag_part(y))))
    return(false);

  return(((real_part(x) == real_part(y)) ||
	  (fabs(real_part(x) - real_part(y)) <= sc->morally_equal_float_epsilon)) &&
	 ((imag_part(x) == imag_part(y)) ||
	  (fabs(imag_part(x) - imag_part(y)) <= sc->morally_equal_float_epsilon)));
}

static bool rng_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
#if WITH_GMP
  return(x == y);
#else
  return((x == y) ||
	 ((is_random_state(y)) &&
	  (random_seed(x) == random_seed(y)) &&
	  (random_carry(x) == random_carry(y))));
#endif
}

static bool (*equals[NUM_TYPES])(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci);
static bool (*morally_equals[NUM_TYPES])(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci);

static void init_equals(void)
{
  int32_t i;
  for (i = 0; i < NUM_TYPES; i++) {equals[i] = eq_equal; morally_equals[i] = eq_equal;}
  equals[T_SYMBOL] =       eq_equal;
  equals[T_C_POINTER] =    c_pointer_equal;
  equals[T_UNSPECIFIED] =  unspecified_equal;
  equals[T_STRING] =       string_equal;
  equals[T_SYNTAX] =       syntax_equal;
  equals[T_C_OBJECT] =     c_object_equal;
  equals[T_RANDOM_STATE] = rng_equal;
  equals[T_ITERATOR] =     iterator_equal;
  equals[T_INPUT_PORT] =   port_equal;
  equals[T_OUTPUT_PORT] =  port_equal;
  equals[T_MACRO] =        closure_equal;
  equals[T_MACRO_STAR] =   closure_equal;
  equals[T_BACRO] =        closure_equal;
  equals[T_BACRO_STAR] =   closure_equal;
  equals[T_CLOSURE] =      closure_equal;
  equals[T_CLOSURE_STAR] = closure_equal;
  equals[T_HASH_TABLE] =   hash_table_equal;
  equals[T_LET] =          let_equal;
  equals[T_PAIR] =         pair_equal;
  equals[T_VECTOR] =       vector_equal;
  equals[T_INT_VECTOR] =   vector_equal;
  equals[T_FLOAT_VECTOR] = vector_equal;
  equals[T_INTEGER] =      integer_equal;
  equals[T_RATIO] =        fraction_equal;
  equals[T_REAL] =         real_equal;
  equals[T_COMPLEX] =      complex_equal;
  equals[T_BIG_INTEGER] =  bignum_equal;
  equals[T_BIG_RATIO] =    bignum_equal;
  equals[T_BIG_REAL] =     bignum_equal;
  equals[T_BIG_COMPLEX] =  bignum_equal;

  morally_equals[T_SYMBOL] =       symbol_morally_equal;
  morally_equals[T_C_POINTER] =    c_pointer_morally_equal;
  morally_equals[T_UNSPECIFIED] =  unspecified_equal;
  morally_equals[T_STRING] =       string_equal;
  morally_equals[T_SYNTAX] =       syntax_equal;
  morally_equals[T_C_OBJECT] =     c_object_equal;
  morally_equals[T_RANDOM_STATE] = rng_equal;
  morally_equals[T_ITERATOR] =     iterator_morally_equal;
  morally_equals[T_INPUT_PORT] =   port_morally_equal;
  morally_equals[T_OUTPUT_PORT] =  port_morally_equal;
  morally_equals[T_MACRO] =        closure_morally_equal;
  morally_equals[T_MACRO_STAR] =   closure_morally_equal;
  morally_equals[T_BACRO] =        closure_morally_equal;
  morally_equals[T_BACRO_STAR] =   closure_morally_equal;
  morally_equals[T_CLOSURE] =      closure_morally_equal;
  morally_equals[T_CLOSURE_STAR] = closure_morally_equal;
  morally_equals[T_HASH_TABLE] =   hash_table_morally_equal;
  morally_equals[T_LET] =          let_morally_equal;
  morally_equals[T_PAIR] =         pair_morally_equal;
  morally_equals[T_VECTOR] =       vector_morally_equal;
  morally_equals[T_INT_VECTOR] =   vector_morally_equal;
  morally_equals[T_FLOAT_VECTOR] = vector_morally_equal;
  morally_equals[T_INTEGER] =      integer_morally_equal;
  morally_equals[T_RATIO] =        fraction_morally_equal;
  morally_equals[T_REAL] =         real_morally_equal;
  morally_equals[T_COMPLEX] =      complex_morally_equal;
  morally_equals[T_BIG_INTEGER] =  bignum_morally_equal;
  morally_equals[T_BIG_RATIO] =    bignum_morally_equal;
  morally_equals[T_BIG_REAL] =     bignum_morally_equal;
  morally_equals[T_BIG_COMPLEX] =  bignum_morally_equal;
}

static bool s7_is_equal_1(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
  return((*(equals[type(x)]))(sc, x, y, ci));
}

bool s7_is_equal(s7_scheme *sc, s7_pointer x, s7_pointer y)
{
  return(s7_is_equal_1(sc, x, y, NULL));
}

static bool s7_is_morally_equal_1(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
  return((*(morally_equals[type(x)]))(sc, x, y, ci));
}

bool s7_is_morally_equal(s7_scheme *sc, s7_pointer x, s7_pointer y)
{
  return(s7_is_morally_equal_1(sc, x, y, NULL));
}

static s7_pointer g_is_equal(s7_scheme *sc, s7_pointer args)
{
  #define H_is_equal "(equal? obj1 obj2) returns #t if obj1 is equal to obj2"
  #define Q_is_equal pcl_bt
  return(make_boolean(sc, s7_is_equal(sc, car(args), cadr(args))));
}

static s7_pointer g_is_morally_equal(s7_scheme *sc, s7_pointer args)
{
  #define H_is_morally_equal "(morally-equal? obj1 obj2) returns #t if obj1 is close enough to obj2."
  #define Q_is_morally_equal pcl_bt
  return(make_boolean(sc, s7_is_morally_equal(sc, car(args), cadr(args))));
}

static bool is_equal_b_pp(s7_pointer a, s7_pointer b) {return(s7_is_equal(cur_sc, a, b));}
static bool is_morally_equal_b_pp(s7_pointer a, s7_pointer b) {return(s7_is_morally_equal(cur_sc, a, b));}


static s7_pointer is_equal_p_pp(s7_pointer a, s7_pointer b) {return((s7_is_equal(cur_sc, a, b)) ? cur_sc->T : cur_sc->F);}
static s7_pointer is_morally_equal_p_pp(s7_pointer a, s7_pointer b) {return((s7_is_morally_equal(cur_sc, a, b)) ? cur_sc->T : cur_sc->F);}


/* ---------------------------------------- length, copy, fill ---------------------------------------- */

static s7_pointer (*length_functions[256])(s7_scheme *sc, s7_pointer obj);
static s7_pointer any_length(s7_scheme *sc, s7_pointer obj) {return(sc->F);}

static s7_pointer pair_length(s7_scheme *sc, s7_pointer a)
{
  int32_t i;
  s7_pointer slow, fast;
  slow = a;
  fast = cdr(a); /* we know a is a pair */
  i = 1;
  while (true)
    {
      if (!is_pair(fast)) return(make_integer(sc, (is_null(fast)) ? i : -i));
      fast = cdr(fast);
      i++;
      if (!is_pair(fast)) return(make_integer(sc, (is_null(fast)) ? i : -i));
      fast = cdr(fast);
      i++;
      slow = cdr(slow);
      if (fast == slow)	return(real_infinity);
    }
  return(real_infinity);
}

static s7_pointer nil_length(s7_scheme *sc, s7_pointer lst) {return(small_int(0));}
static s7_pointer v_length(s7_scheme *sc, s7_pointer v) {return(make_integer(sc, vector_length(v)));}
static s7_pointer str_length(s7_scheme *sc, s7_pointer v) {return(make_integer(sc, string_length(v)));}
static s7_pointer iter_length(s7_scheme *sc, s7_pointer lst) {return(make_integer(sc, iterator_length(lst)));} /* in several cases, this is incorrect */
static s7_pointer h_length(s7_scheme *sc, s7_pointer lst) {return(make_integer(sc, hash_table_mask(lst) + 1));}

static s7_pointer c_obj_length(s7_scheme *sc, s7_pointer lst)
{
  check_method(sc, lst, sc->length_symbol, list_1(sc, lst));
  return(c_object_length(sc, lst));
}

static s7_pointer lt_length(s7_scheme *sc, s7_pointer lst)
{
  check_method(sc, lst, sc->length_symbol, list_1(sc, lst));
  return(make_integer(sc, let_length(sc, lst)));
}

static s7_pointer fnc_length(s7_scheme *sc, s7_pointer lst)
{
  if (has_methods(lst))
    return(make_integer(sc, closure_length(sc, lst)));
  return(sc->F);
}

static s7_pointer io_length(s7_scheme *sc, s7_pointer lst)
{
  if (is_string_port(lst))
    return(make_integer(sc, port_data_size(lst)));
  return(sc->F);
}

static void init_length_functions(void)
{
  int i;
  for (i = 0; i < 256; i++) length_functions[i] = any_length;
  length_functions[T_NIL]          = nil_length;
  length_functions[T_PAIR]         = pair_length;
  length_functions[T_VECTOR]       = v_length;
  length_functions[T_FLOAT_VECTOR] = v_length;
  length_functions[T_INT_VECTOR]   = v_length;
  length_functions[T_STRING]       = str_length;
  length_functions[T_ITERATOR]     = iter_length;
  length_functions[T_HASH_TABLE]   = h_length;
  length_functions[T_C_OBJECT]     = c_obj_length;
  length_functions[T_LET]          = lt_length;
  length_functions[T_CLOSURE]      = fnc_length;
  length_functions[T_CLOSURE_STAR] = fnc_length;
  length_functions[T_INPUT_PORT]   = io_length;
}


static s7_pointer s7_length(s7_scheme *sc, s7_pointer lst)
{
  return((*length_functions[unchecked_type(lst)])(sc, lst));
}

static s7_pointer g_length(s7_scheme *sc, s7_pointer args)
{
  #define H_length "(length obj) returns the length of obj, which can be a list, vector, string, or hash-table. \
The length of a dotted list does not include the final cdr, and is returned as a negative number.  A circular \
list has infinite length.  Length of anything else returns #f."
  #define Q_length s7_make_signature(sc, 2, s7_make_signature(sc, 2, sc->is_real_symbol, sc->is_boolean_symbol), sc->T)
  return((*length_functions[unchecked_type(car(args))])(sc, car(args)));
}

/* what about (length file)?  input port, read_file gets the file length, so perhaps save it
 *   but we're actually looking at the port, so its length is what remains to be read? (if input port)
 */


/* -------------------------------- copy -------------------------------- */

static s7_pointer copy_to_string_error = NULL, copy_to_byte_vector_error = NULL;

static void set_string_error_source(s7_scheme *sc, s7_pointer source)
{
  if (!copy_to_string_error)
    copy_to_string_error = s7_make_permanent_string("copy ~A to string, ~S is not a character");
  if (!copy_to_byte_vector_error)
    copy_to_byte_vector_error = s7_make_permanent_string("copy ~A to byte-vector, ~S is not a byte");
  set_cadr(sc->elist_3, prepackaged_type_name(sc, source));
}

static s7_pointer string_setter(s7_scheme *sc, s7_pointer str, s7_int loc, s7_pointer val)
{
  if (s7_is_character(val))
    {
      string_value(str)[loc] = s7_character(val);
      return(val);
    }
  /* (copy #(3) "123"): wrong type arg because not a char, but it's very confusing to report
   *   error: copy argument 3, 3, is an integer but should be a character
   * perhaps better, copy #(3) to string, 3 is not a character
   */

  if (!copy_to_string_error)
    copy_to_string_error = s7_make_permanent_string("copy ~A to string, ~S is not a character");

  set_car(sc->elist_3, copy_to_string_error);
  set_caddr(sc->elist_3, val);
  return(s7_error(sc, sc->wrong_type_arg_symbol, sc->elist_3));
}

static s7_pointer byte_vector_setter(s7_scheme *sc, s7_pointer str, s7_int loc, s7_pointer val)
{
  if (s7_is_integer(val))
    {
      s7_int byte;
      byte = s7_integer(val);
      if ((byte >= 0) && (byte < 256))
	string_value(str)[loc] = (unsigned char)byte;
      else return(simple_wrong_type_argument_with_type(sc, sc->copy_symbol, val, an_unsigned_byte_string));
      return(val);
    }

  if (!copy_to_byte_vector_error)
    copy_to_byte_vector_error = s7_make_permanent_string("copy ~A to byte-vector, ~S is not a byte");

  set_car(sc->elist_3, copy_to_byte_vector_error);
  set_caddr(sc->elist_3, val);
  return(s7_error(sc, sc->wrong_type_arg_symbol, sc->elist_3));
}

static s7_pointer string_getter(s7_scheme *sc, s7_pointer str, s7_int loc)
{
  return(s7_make_character(sc, (unsigned char)(string_value(str)[loc]))); /* cast needed else (copy (string (integer->char 255))...) is trouble */
}

static s7_pointer byte_vector_getter(s7_scheme *sc, s7_pointer str, s7_int loc)
{
  return(make_integer(sc, (unsigned char)(string_value(str)[loc])));
}

static s7_pointer c_object_setter(s7_scheme *sc, s7_pointer obj, s7_int loc, s7_pointer val)
{
  set_car(sc->t2_1, make_integer(sc, loc));
  set_car(sc->t2_2, val);
  return((*(c_object_set(sc, obj)))(sc, obj, sc->t2_1));
}

static s7_pointer c_object_getter(s7_scheme *sc, s7_pointer obj, s7_int loc)
{
  set_car(sc->t1_1, make_integer(sc, loc));
  return((*(c_object_ref(sc, obj)))(sc, obj, sc->t1_1));
}

static s7_pointer let_setter(s7_scheme *sc, s7_pointer e, s7_int loc, s7_pointer val)
{
  /* loc is irrelevant here
   * val has to be of the form (cons symbol value)
   * if symbol is already in e, its value is changed, otherwise a new slot is added to e
   */
  static s7_pointer ls_err = NULL;
  s7_pointer sym;
  if (!is_pair(val))
    {
      if (!ls_err) ls_err = s7_make_permanent_string("(cons symbol value)");
      return(wrong_type_argument_with_type(sc, sc->copy_symbol, 3, e, ls_err));
    }
  sym = car(val);
  if (!is_symbol(sym))
    {
      if (!ls_err) ls_err = s7_make_permanent_string("(cons symbol value)");
      return(wrong_type_argument_with_type(sc, sc->copy_symbol, 3, e, ls_err));
    }
  if ((symbol_id(sym) < let_id(e)) ||
      (s7_let_set(sc, e, sym, cdr(val)) != cdr(val)))
    make_slot_1(sc, e, sym, cdr(val));
  return(val);
}

static s7_pointer hash_table_setter(s7_scheme *sc, s7_pointer e, s7_int loc, s7_pointer val)
{
  /* loc is irrelevant here
   * val has to be of the form (cons key value)
   * if key is already in e, its value is changed, otherwise a new slot is added to e
   */
  if (!is_pair(val))
    return(wrong_type_argument_with_type(sc, sc->copy_symbol, 1, e, a_list_string));
  return(s7_hash_table_set(sc, e, car(val), cdr(val)));
}


s7_pointer s7_copy(s7_scheme *sc, s7_pointer args)
{
  #define H_copy "(copy obj) returns a copy of obj, (copy src dest) copies src into dest, (copy src dest start end) copies src from start to end."
  /* #define Q_copy s7_make_circular_signature(sc, 3, 4, sc->T, sc->is_sequence_symbol, sc->is_sequence_symbol, sc->is_integer_symbol) */
  /* this is not right when c-object types are handled in lint -- a generator or Snd object need not consider itself a sequence,
   *   but it can provide a copy method.  So, I think I'll just use #t
   */
  #define Q_copy s7_make_circular_signature(sc, 3, 4, sc->T, sc->T, sc->T, sc->is_integer_symbol)

  s7_pointer source, dest;
  s7_int i, j, dest_len, start, end, source_len;
  s7_pointer (*set)(s7_scheme *sc, s7_pointer obj, s7_int loc, s7_pointer val) = NULL;
  s7_pointer (*get)(s7_scheme *sc, s7_pointer obj, s7_int loc) = NULL;
  bool have_indices;

  source = car(args);
  if (is_null(cdr(args)))                  /* (copy obj) */
    {
      switch (type(source))
	{
	case T_STRING:
	  {
	    s7_pointer ns;
	    ns = s7_make_string_with_length(sc, string_value(source), string_length(source));
	    if (is_byte_vector(source))
	      set_byte_vector(ns);
	    return(ns);
	  }
	  
	case T_C_OBJECT:
	  return(object_copy(sc, args));

	case T_RANDOM_STATE:
	  return(rng_copy(sc, args));
	  
	case T_HASH_TABLE:              /* this has to copy nearly everything */
	  {
	    uint32_t gc_loc;
	    s7_pointer new_hash;
	    new_hash = s7_make_hash_table(sc, hash_table_mask(source) + 1);
	    gc_loc = s7_gc_protect(sc, new_hash);
	    hash_table_checker(new_hash) = hash_table_checker(source);
	    if (hash_chosen(source)) hash_set_chosen(new_hash);
	    hash_table_mapper(new_hash) = hash_table_mapper(source);
	    hash_table_set_procedures(new_hash, hash_table_procedures(source));
	    hash_table_copy(sc, source, new_hash, 0, hash_table_entries(source));
	    s7_gc_unprotect_at(sc, gc_loc);
	    return(new_hash);
	  }
	  
	case T_ITERATOR:
	  return(iterator_copy(sc, source));
	  
	case T_LET:
	  check_method(sc, source, sc->copy_symbol, args);
	  return(let_copy(sc, source));   /* this copies only the local env and points to outer envs */
	  
	case T_CLOSURE: case T_CLOSURE_STAR:
	case T_MACRO: case T_MACRO_STAR:
	case T_BACRO: case T_BACRO_STAR:
	  check_method(sc, source, sc->copy_symbol, args);
	  return(copy_closure(sc, source));
	  
	case T_INT_VECTOR:
	case T_FLOAT_VECTOR:
	case T_VECTOR:
	  return(s7_vector_copy(sc, source)); /* "shallow" copy */
	  
	case T_PAIR:                    /* top level only, as in the other cases, last arg checks for circles */
	  return(protected_list_copy(sc, source));
	  
	case T_INTEGER:
	  new_cell(sc, dest, T_INTEGER);
	  integer(dest) = integer(source);
	  return(dest);

	case T_RATIO:
	  new_cell(sc, dest, T_RATIO);
	  numerator(dest) = numerator(source);
	  denominator(dest) = denominator(source);
	  return(dest);

	case T_REAL:
	  new_cell(sc, dest, T_REAL);
	  set_real(dest, real(source));
	  return(dest);

	case T_COMPLEX:
	  new_cell(sc, dest, T_COMPLEX);
	  set_real_part(dest, real_part(source));
	  set_imag_part(dest, imag_part(source));
	  return(dest);

#if WITH_GMP
	case T_BIG_INTEGER: return(mpz_to_big_integer(sc, big_integer(source)));
	case T_BIG_RATIO:   return(mpq_to_big_ratio(sc, big_ratio(source)));
	case T_BIG_REAL:    return(mpfr_to_big_real(sc, big_real(source)));
	case T_BIG_COMPLEX: return(mpc_to_big_complex(sc, big_complex(source)));
#endif
	  
	case T_C_POINTER:
	  return(s7_make_c_pointer_with_type(sc, raw_pointer(source), raw_pointer_type(source), raw_pointer_info(source)));
	}
      return(source);
    }

  have_indices = (is_pair(cddr(args)));
  dest = cadr(args);
  if ((source == dest) && (!have_indices))
    return(dest);
  
  switch (type(source))
    {
    case T_PAIR:
      if (dest == sc->key_readable_symbol)  /* a kludge, but I can't think of anything less stupid */
	return(copy_body(sc, source));

      end = s7_list_length(sc, source);
      if (end == 0) 
	end = circular_list_entries(source);
      else
	{
	  if (end < 0) end = -end;
	}
      break;

    case T_INT_VECTOR:
    case T_FLOAT_VECTOR:
    case T_VECTOR:
      get = vector_getter(source);
      end = vector_length(source);
      break;

    case T_STRING:
      if (is_byte_vector(source))
	get = byte_vector_getter;
      else get = string_getter;
      end = string_length(source);
      break;

    case T_HASH_TABLE:
      end = hash_table_entries(source);
      break;

    case T_C_OBJECT:
      check_method(sc, source, sc->copy_symbol, args);
      {
	s7_pointer x;
	x = object_copy(sc, args);
	if (x == dest)
	  return(dest);
	/* if object_copy can't handle args for some reason, it should return #f (not dest), and we'll soldier on... */
      }
      get = c_object_direct_ref(sc, source);
      if (!get) get = c_object_getter;
      end = c_object_length_to_int(sc, source);
      break;

    case T_LET:
      check_method(sc, source, sc->copy_symbol, args);
      if (source == sc->rootlet)
	return(wrong_type_argument_with_type(sc, sc->copy_symbol, 1, source, s7_make_string_wrapper(sc, "a sequence other than the rootlet")));
      end = let_length(sc, source);
      break;

    case T_NIL:
      end = 0;
      if (is_sequence(dest))
	break;

    default:
      return(wrong_type_argument_with_type(sc, sc->copy_symbol, 1, source, a_sequence_string));
      /* copy doesn't have to duplicate fill!, so (copy 1 #(...)) need not be supported */
    }

  start = 0;
  if (have_indices)
    {
      s7_pointer p;
      p = start_and_end(sc, sc->copy_symbol, NULL, cddr(args), args, 3, &start, &end);
      if (p != sc->gc_nil) return(p);
    }
  if ((start == 0) && (source == dest))
    return(dest);
  source_len = end - start;
  if (source_len == 0)
    {
      if (!is_sequence(dest))
	return(wrong_type_argument_with_type(sc, sc->copy_symbol, 2, dest, a_sequence_string));	
      return(dest);
    }

  switch (type(dest))
    {
    case T_PAIR:
      dest_len = source_len;
      break;

    case T_INT_VECTOR:
    case T_FLOAT_VECTOR:
    case T_VECTOR:
      set = vector_setter(dest);
      dest_len = vector_length(dest);
      break;

    case T_STRING:
      if (is_byte_vector(dest))
	set = byte_vector_setter;
      else set = string_setter;
      dest_len = string_length(dest);
      break;

    case T_HASH_TABLE:
      set = hash_table_setter;
      dest_len = source_len;
      break;

    case T_C_OBJECT:
      set = c_object_direct_set(sc, dest);
      if (!set) set = c_object_setter;
      dest_len = c_object_length_to_int(sc, dest);
      break;

    case T_LET:
      if (dest == sc->rootlet)
	return(wrong_type_argument_with_type(sc, sc->copy_symbol, 2, dest, s7_make_string_wrapper(sc, "a sequence other than the rootlet")));
      set = let_setter;
      dest_len = source_len;                     /* grows via set, so dest_len isn't relevant */
      break;

    case T_NIL:
      return(sc->nil);

    default:
      return(wrong_type_argument_with_type(sc, sc->copy_symbol, 2, dest, a_sequence_string));
    }

  if (dest_len == 0)
    return(dest);

  /* end is source_len if not set explicitly */
  if (dest_len < source_len)
    {
      end = dest_len + start;
      source_len = dest_len;
    }

  if ((source != dest) &&
      (type(source) == type(dest)))
    {
      switch (type(source))
	{
	case T_PAIR:
	  {
	    s7_pointer ps, pd;

	    ps = source;
	    for (i = 0; i < start; i++)
	      ps = cdr(ps);
	    for (pd = dest; (i < end) && is_pair(ps) && is_pair(pd); i++, ps = cdr(ps), pd = cdr(pd))
	      set_car(pd, car(ps));
	    return(dest);
	  }

	case T_VECTOR:
	  memcpy((void *)(vector_elements(dest)), (void *)((vector_elements(source)) + start), source_len * sizeof(s7_pointer));
	  return(dest);

	case T_INT_VECTOR:
	  memcpy((void *)(int_vector_elements(dest)), (void *)((int_vector_elements(source)) + start), source_len * sizeof(s7_int));
	  return(dest);

	case T_FLOAT_VECTOR:
	  memcpy((void *)(float_vector_elements(dest)), (void *)((float_vector_elements(source)) + start), source_len * sizeof(s7_double));
	  return(dest);

	case T_STRING: /* this is 4 cases (string/byte-vector) */
	  memcpy((void *)string_value(dest), (void *)((string_value(source)) + start), source_len * sizeof(char));
	  return(dest);

	case T_C_OBJECT:
	  {
	    s7_pointer mi, mj;
	    uint32_t gc_loc1, gc_loc2;
	    s7_pointer (*ref)(s7_scheme *sc, s7_pointer obj, s7_pointer args);
	    s7_pointer (*set)(s7_scheme *sc, s7_pointer obj, s7_pointer args);

	    mi = make_mutable_integer(sc, start);
	    mj = make_mutable_integer(sc, end);
	    gc_loc1 = s7_gc_protect(sc, mi);
	    gc_loc2 = s7_gc_protect(sc, mj);
	    ref = c_object_ref(sc, source);
	    set = c_object_set(sc, dest);

	    for (i = start, j = 0; i < end; i++, j++)
	      {
		integer(mi) = i;
		integer(mj) = j;
		set_car(sc->t1_1, mi);
		set_car(sc->t2_2, ref(sc, source, sc->t1_1));
		set_car(sc->t2_1, mj);
		set(sc, dest, sc->t2_1);
	      }
	    s7_gc_unprotect_at(sc, gc_loc1);
	    s7_gc_unprotect_at(sc, gc_loc2);
	    free_cell(sc, mi);
	    free_cell(sc, mj);
	    return(dest);
	  }

	case T_LET:
	  break;

	case T_HASH_TABLE:
	  {
	    s7_pointer p;
	    p = hash_table_copy(sc, source, dest, start, end);
	    if ((hash_table_checker(source) != hash_table_checker(dest)) &&
		(!hash_table_checker_locked(dest)))
	      {
		if (hash_table_checker(dest) == hash_empty)
		  hash_table_checker(dest) = hash_table_checker(source);
		else 
		  {
		    hash_table_checker(dest) = hash_equal;
		    hash_set_chosen(dest);
		  }
	      }
	    return(p);
	  }
	  break;

	default:
	  return(dest);
	}
    }

  switch (type(source))
    {
    case T_PAIR:
      {
	s7_pointer p;
	p = source;
	if (start > 0)
	  for (i = 0; i < start; i++)
	    p = cdr(p);
	/* dest won't be a pair here -- the pair->pair case was caught above */
	if (is_string(dest)) set_string_error_source(sc, source);
	for (i = start, j = 0; i < end; i++, j++, p = cdr(p))
	  set(sc, dest, j, car(p));
	return(dest);
      }

    case T_LET:
      /* implicit index can give n-way reality check (ht growth by new entries) 
       * if shadowed entries are they unshadowed by reversal?
       */
      {
	/* source and dest can't be rootlet (checked above) */
	s7_pointer slot;
	slot = let_slots(source);
	for (i = 0; i < start; i++) slot = next_slot(slot);
	if (is_pair(dest))
	  {
	    s7_pointer p;
	    for (i = start, p = dest; (i < end) && (is_pair(p)); i++, p = cdr(p), slot = next_slot(slot))
	      set_car(p, cons(sc, slot_symbol(slot), slot_value(slot)));
	  }
	else
	  {
	    if (is_let(dest))
	      {
		for (i = start; i < end; i++, slot = next_slot(slot))
		  make_slot_1(sc, dest, slot_symbol(slot), slot_value(slot));
	      }
	    else
	      {
		if (is_hash_table(dest))
		  {
		    for (i = start; i < end; i++, slot = next_slot(slot))
		      s7_hash_table_set(sc, dest, slot_symbol(slot), slot_value(slot));
		  }
		else
		  {
		    for (i = start, j = 0; i < end; i++, j++, slot = next_slot(slot))
		      set(sc, dest, j, cons(sc, slot_symbol(slot), slot_value(slot)));
		  }
	      }
	  }
	return(dest);
      }

    case T_HASH_TABLE:
      {
	int32_t loc, skip;
	hash_entry_t **elements;
	hash_entry_t *x = NULL;
	elements = hash_table_elements(source);
	loc = -1;

	skip = start;
	while (skip > 0)
	  {
	    while (!x) x = elements[++loc];
	    skip--;
	    x = x->next;
	  }

      if (is_pair(dest))
	{
	  s7_pointer p;
	  for (i = start, p = dest; (i < end) && (is_pair(p)); i++, p = cdr(p))
	    {
	      while (!x) x = elements[++loc];
	      set_car(p, cons(sc, x->key, x->value));
	      x = x->next;
	    }
	}
      else
	{
	  if (is_let(dest))
	    {
	      for (i = start; i < end; i++)
		{
		  while (!x) x = elements[++loc];
		  if (!is_symbol(x->key))
		    return(simple_wrong_type_argument(sc, sc->copy_symbol, x->key, T_SYMBOL));
		  make_slot_1(sc, dest, x->key, x->value);
		  x = x->next;
		}
	    }
	  else
	    {
	      for (i = start, j = 0; i < end; i++, j++)
		{
		  while (!x) x = elements[++loc];
		  set(sc, dest, j, cons(sc, x->key, x->value));
		  x = x->next;
		}
	    }
	}
      return(dest);
      }

    case T_VECTOR:
      if (is_float_vector(dest))
	{
	  for (i = start, j = 0; i < end; i++, j++)
	    float_vector_element(dest, j) = real_to_double(sc, vector_element(source, i), "copy");
	  return(dest);
	}
      if (is_int_vector(dest))
	{
	  for (i = start, j = 0; i < end; i++, j++)
	    {
	      s7_pointer val;
	      val = vector_element(source, i);
	      if (!s7_is_integer(val)) 
		s7_wrong_type_arg_error(sc, "copy", 3, val, "an integer");
	      int_vector_element(dest, j) = s7_integer(val);
	    }
	  return(dest);
	}
      break;

    case T_FLOAT_VECTOR:
      if (is_int_vector(dest))
	{
	  for (i = start, j = 0; i < end; i++, j++)
	    int_vector_element(dest, j) = (s7_int)(float_vector_element(source, i));
	  return(dest);
	}
      if (is_normal_vector(dest))
	{
	  for (i = start, j = 0; i < end; i++, j++)
	    vector_element(dest, j) = make_real(sc, float_vector_element(source, i));
	  return(dest);
	}
      break;

    case T_INT_VECTOR:
      if (is_float_vector(dest))
	{
	  for (i = start, j = 0; i < end; i++, j++)
	    float_vector_element(dest, j) = (s7_double)(int_vector_element(source, i));
	  return(dest);
	}
      if (is_string(dest)) /* includes byte-vector, as below */
	{
	  for (i = start, j = 0; i < end; i++, j++)
	    string_value(dest)[j] = (unsigned char)int_vector_element(source, i);
	  return(dest);
	}
      if (is_normal_vector(dest))
	{
	  for (i = start, j = 0; i < end; i++, j++)
	    vector_element(dest, j) = s7_make_integer(sc, int_vector_element(source, i));
	  return(dest);
	}
      break;

    case T_STRING:
      if (is_normal_vector(dest))
	{
	  if (is_byte_vector(source))
	    {
	      for (i = start, j = 0; i < end; i++, j++)
		vector_element(dest, j) = make_integer(sc, (s7_int)((unsigned char)string_value(source)[i]));
	    }
	  else
	    {
	      for (i = start, j = 0; i < end; i++, j++)
		vector_element(dest, j) = s7_make_character(sc, (unsigned char)string_value(source)[i]);
	    }
	  return(dest);
	}
      if (is_int_vector(dest))
	{
	  for (i = start, j = 0; i < end; i++, j++)
	    int_vector_element(dest, j) = (s7_int)((unsigned char)(string_value(source)[i]));
	  return(dest);
	}
      if (is_float_vector(dest))
	{
	  for (i = start, j = 0; i < end; i++, j++)
	    float_vector_element(dest, j) = (s7_double)((unsigned char)(string_value(source)[i]));
	  return(dest);
	}
    }

  if (is_pair(dest))
    {
      s7_pointer p;
      for (i = start, p = dest; (i < end) && (is_pair(p)); i++, p = cdr(p))
	set_car(p, get(sc, source, i));
    }
  else
    {
      /* if source == dest here, we're moving data backwards, so this is safe in either case */
      if (is_string(dest)) set_string_error_source(sc, source);
      for (i = start, j = 0; i < end; i++, j++)
	set(sc, dest, j, get(sc, source, i));
    }
  /* some choices probably should raise an error, but don't:
   *   (copy (make-hash-table) "1") ; nothing to copy (empty hash table), so no error
   */
  return(dest);
}

#define g_copy s7_copy


/* -------------------------------- reverse -------------------------------- */
 
static s7_pointer g_reverse(s7_scheme *sc, s7_pointer args)
{
  #define H_reverse "(reverse lst) returns a list with the elements of lst in reverse order.  reverse \
also accepts a string or vector argument."
  #define Q_reverse s7_make_signature(sc, 2, sc->is_sequence_symbol, sc->is_sequence_symbol)

  s7_pointer p, np;

  p = car(args);
  sc->temp3 = p;
  np = sc->nil;

  switch (type(p))
    {
    case T_NIL:
      return(sc->nil);

    case T_PAIR:
      return(s7_reverse(sc, p));

    case T_STRING:
      {
	char *source, *dest, *end;
	int32_t len;
	len = string_length(p);
	source = string_value(p);
	end = (char *)(source + len);
	dest = (char *)malloc((len + 1) * sizeof(char));
	dest[len] = 0;
	np = make_string_uncopied_with_length(sc, dest, len);
	dest += len;
	while (source < end) *(--dest) = *source++;
	if (is_byte_vector(p))
	  set_byte_vector(np);
      }
      break;

    case T_INT_VECTOR:
      {
	s7_int *source, *dest, *end;
	s7_int len;
	len = vector_length(p);
	if (vector_rank(p) > 1)
	  np = g_make_vector_1(sc, set_plist_3(sc, g_vector_dimensions(sc, set_plist_1(sc, p)), small_int(0), sc->T), sc->make_int_vector_symbol);
	else np = make_vector_1(sc, len, NOT_FILLED, T_INT_VECTOR);
	source = int_vector_elements(p);
	end = (s7_int *)(source + len);
	dest = (s7_int *)(int_vector_elements(np) + len);
	while (source < end) *(--dest) = *source++;
      }
      break;

    case T_FLOAT_VECTOR:
      {
	s7_double *source, *dest, *end;
	s7_int len;
	len = vector_length(p);
	if (vector_rank(p) > 1)
	  np = g_make_vector_1(sc, set_plist_3(sc, g_vector_dimensions(sc, set_plist_1(sc, p)), real_zero, sc->T), sc->make_float_vector_symbol);
	else np = make_vector_1(sc, len, NOT_FILLED, T_FLOAT_VECTOR);
	source = float_vector_elements(p);
	end = (s7_double *)(source + len);
	dest = (s7_double *)(float_vector_elements(np) + len);
	while (source < end) *(--dest) = *source++;
      }
      break;

    case T_VECTOR:
      {
	s7_pointer *source, *dest, *end;
	s7_int len;
	len = vector_length(p);
	if (vector_rank(p) > 1)
	  np = g_make_vector(sc, set_plist_1(sc, g_vector_dimensions(sc, set_plist_1(sc, p))));
	else np = make_vector_1(sc, len, NOT_FILLED, T_VECTOR);
	source = vector_elements(p);
	end = (s7_pointer *)(source + len);
	dest = (s7_pointer *)(vector_elements(np) + len);
	while (source < end) *(--dest) = *source++;
      }
      break;

    case T_HASH_TABLE:
      return(hash_table_reverse(sc, p));

    case T_C_OBJECT:
      check_method(sc, p, sc->reverse_symbol, args);
      if (c_object_reverse(sc, p))
	return((*(c_object_reverse(sc, p)))(sc, args));
      eval_error(sc, "attempt to reverse ~S?", p);

    case T_LET:
      {
	s7_pointer new_e, x;
	s7_int id;
	check_method(sc, p, sc->reverse_symbol, args);
	if ((p == sc->rootlet) ||
	    (!is_slot(let_slots(p))) ||
	    (!is_slot(next_slot(let_slots(p)))))
	  return(p);
	new_e = new_frame_in_env(sc, outlet(p));
	set_all_methods(new_e, p);
	sc->temp3 = new_e;
	id = let_id(new_e);
	for (x = let_slots(p); is_slot(x); x = next_slot(x))
	  {
	    s7_pointer z;
	    new_cell(sc, z, T_SLOT);
	    slot_set_symbol(z, slot_symbol(x));
	    slot_set_value(z, slot_value(x));
	    if (symbol_id(slot_symbol(z)) != id) /* keep shadowing intact */
	      symbol_set_local(slot_symbol(x), id, z);
	    next_slot(z) = let_slots(new_e);
	    let_slots(new_e) = z;
	  }
	sc->temp3 = sc->nil;
	return(new_e);
      }

    default:
      method_or_bust_with_type_one_arg(sc, p, sc->reverse_symbol, args, a_sequence_string);
    }
  return(np);
}

static s7_pointer g_reverse_in_place(s7_scheme *sc, s7_pointer args)
{
  s7_pointer p;
  #define H_reverse_in_place "(reverse! lst) reverses lst in place"
  #define Q_reverse_in_place s7_make_signature(sc, 2, sc->is_sequence_symbol, sc->is_sequence_symbol)

  p = car(args);

  if ((sc->safety > NO_SAFETY) &&
      (is_immutable(p)))
    return(s7_error(sc, sc->error_symbol, set_elist_2(sc, s7_make_string_wrapper(sc, "can't reverse! ~S (it is immutable)"), p)));

  switch (type(p))
    {
    case T_NIL:
      return(sc->nil);

    case T_PAIR:
      {
	s7_pointer np;
	np = reverse_in_place(sc, sc->nil, p);
	if (is_null(np))
	  return(simple_wrong_type_argument_with_type(sc, sc->reverseb_symbol, p, a_proper_list_string));
	return(np);
      }
      break;
      /* (reverse! p) is supposed to change p directly and lisp programmers expect reverse! to be fast
       * so in a sense this is different from the other cases: it assumes (set! p (reverse! p))
       * To make (reverse! p) direct:
       *    for (l = p, r = cdr(p); is_pair(r); l = r, r = cdr(r)) opt1(r) = l;
       *    if (!is_null(r)) return(simple_wrong_type_argument_with_type(sc, sc->reverseb_symbol, p, a_proper_list_string));
       *    for (r = l, l = p; l != r; l = cdr(l)) {t = car(l); set_car(l, car(r)); set_car(r, t); if (cdr(l) != r) r = opt1(r);}
       */

    case T_STRING:
      {
	int32_t len;
	char *s1, *s2;
	len = string_length(p);
	if (len < 2) return(p);
	s1 = string_value(p);
	s2 = (char *)(s1 + len - 1);
	while (s1 < s2) {char c; c = *s1; *s1++ = *s2; *s2-- = c;}
      }
      break;

    case T_INT_VECTOR:
      {
	s7_int len;
	s7_int *s1, *s2;
	len = vector_length(p);
	if (len < 2) return(p);
	s1 = int_vector_elements(p);
	s2 = (s7_int *)(s1 + len - 1);
	while (s1 < s2) {s7_int c; c = *s1; *s1++ = *s2; *s2-- = c;}
      }
      break;

    case T_FLOAT_VECTOR:
      {
	s7_int len;
	s7_double *s1, *s2;
	len = vector_length(p);
	if (len < 2) return(p);
	s1 = float_vector_elements(p);
	s2 = (s7_double *)(s1 + len - 1);
	while (s1 < s2) {s7_double c; c = *s1; *s1++ = *s2; *s2-- = c;}
      }
      break;

    case T_VECTOR:
      {
	s7_int len;
	s7_pointer *s1, *s2;
	len = vector_length(p);
	if (len < 2) return(p);
	s1 = vector_elements(p);
	s2 = (s7_pointer *)(s1 + len - 1);
	while (s1 < s2) {s7_pointer c; c = *s1; *s1++ = *s2; *s2-- = c;}
      }
      break;

    default:
      if ((is_simple_sequence(p)) &&
	  (!has_methods(p)))
	return(simple_wrong_type_argument_with_type(sc, sc->reverseb_symbol, p, s7_make_string_wrapper(sc, "a vector, string, or list")));
      method_or_bust_with_type_one_arg(sc, p, sc->reverseb_symbol, list_1(sc, p), a_sequence_string);
    }
  return(p);
}


/* -------------------------------- fill! -------------------------------- */

static s7_pointer pair_fill(s7_scheme *sc, s7_pointer args)
{
  /* ambiguous ("tree-fill"?) but if it's like vector-fill, we just stomp on the top level */
  s7_pointer x, y, obj, val, p;
  s7_int i, start = 0, end, len;

  obj = car(args);
  if ((sc->safety > NO_SAFETY) &&
      (is_immutable_pair(obj)))
    return(s7_error(sc, sc->error_symbol, set_elist_2(sc, s7_make_string_wrapper(sc, "can't fill! ~S (it is immutable)"), obj)));

  val = cadr(args);
  len = s7_list_length(sc, obj);
  end = len;
  if (end < 0) end = -end; else {if (end == 0) end = 123123123;}
  if (!is_null(cddr(args)))
    {
      p = start_and_end(sc, sc->fill_symbol, sc->fill_symbol, cddr(args), args, 3, &start, &end);
      if (p != sc->gc_nil) return(p);
      if (start == end) return(val);
    }
  if (len > 0)
    {
      s7_int i;
      if (end < len) len = end;
      for (i = 0, p = obj; i < start; p = cdr(p), i++);
      for (; i < len; p = cdr(p), i++) set_car(p, val);
      return(val);
    }

  for (x = obj, y = obj, i = 0; ;i++)
    {
      if ((end > 0) && (i >= end))
	return(val);
      if (i >= start) set_car(x, val);
      if (!is_pair(cdr(x)))
	{
	  if (!is_null(cdr(x)))
	    set_cdr(x, val);
	  return(val);
	}
      x = cdr(x);
      if ((i & 1) != 0) y = cdr(y);
      if (x == y) 
	return(val);
    }
  return(val);
}


s7_pointer s7_fill(s7_scheme *sc, s7_pointer args)
{
  #define H_fill "(fill! obj val (start 0) end) fills obj with val"
  #define Q_fill s7_make_circular_signature(sc, 3, 4, sc->T, sc->is_sequence_symbol, sc->T, sc->is_integer_symbol)
  s7_pointer p;

  p = car(args);
  switch (type(p))
    {
    case T_STRING:
      return(g_string_fill(sc, args)); /* redundant type check here and below */

    case T_INT_VECTOR:
    case T_FLOAT_VECTOR:
    case T_VECTOR:
      return(g_vector_fill(sc, args));

    case T_PAIR:
      return(pair_fill(sc, args));

    case T_NIL:
      return(cadr(args));        /* this parallels the empty vector case */

    case T_HASH_TABLE:
      return(hash_table_fill(sc, args));

    case T_LET:
      check_method(sc, p, sc->fill_symbol, args);
      return(let_fill(sc, args));

    case T_C_OBJECT:
      check_method(sc, p, sc->fill_symbol, args);
      if (c_object_fill(sc, p))
	return((*(c_object_fill(sc, p)))(sc, args));
      eval_error(sc, "attempt to fill ~S?", p);

    default:
      check_method(sc, p, sc->fill_symbol, args);
    }
  return(wrong_type_argument_with_type(sc, sc->fill_symbol, 1, p, a_sequence_string)); /* (fill! 1 0) */
}

#define g_fill s7_fill
/* perhaps (fill iterator obj) could fill the underlying sequence (if any) -- not let/closure
 *   similarly for length, reverse etc
 */


/* -------------------------------- append -------------------------------- */

static s7_int sequence_length(s7_scheme *sc, s7_pointer lst)
{
  switch (type(lst))
    {
    case T_PAIR:
      {
	int32_t len;
	len = s7_list_length(sc, lst);
	if (len == 0) return(-1);
	return(len);
      }
    case T_NIL:        return(0);
    case T_INT_VECTOR: 
    case T_FLOAT_VECTOR: 
    case T_VECTOR:     return(vector_length(lst));
    case T_STRING:     return(string_length(lst));
    case T_HASH_TABLE: return(hash_table_entries(lst));
    case T_LET:        return(let_length(sc, lst));
    case T_C_OBJECT:   
      {
	s7_pointer x;
	x = c_object_length(sc, lst);
	if (s7_is_integer(x))
	  return(s7_integer(x));
      }
    }
  return(-1);
}

static s7_int total_sequence_length(s7_scheme *sc, s7_pointer args, s7_pointer caller, uint8_t typ)
{
  s7_pointer p;
  int32_t i;
  s7_int len = 0;

  for (i = 1, p = args; is_pair(p); p = cdr(p), i++)
    {
      s7_pointer seq;
      s7_int n;
      seq = car(p);
      n = sequence_length(sc, seq);
      if ((n > 0) && 
	  (typ != T_FREE) && 
	  ((type(seq) == T_HASH_TABLE) ||  /* can't append hash-tables (no obvious meaning to the operation) */
	   ((type(seq) == T_LET) &&        /*   similarly for lets, unless this is a mock-string or something similar */
	    ((!has_methods(seq)) || (find_method(sc, seq, caller) == sc->undefined)))))
	{
	  wrong_type_argument(sc, caller, i, seq, typ);
	  return(0);
	}
      if (n < 0)
	{
	  wrong_type_argument_with_type(sc, caller, i, seq, (is_pair(seq)) ? a_proper_list_string : a_sequence_string);
	  return(0);
	}
      len += n;
    }
  return(len);
}

static s7_pointer vector_append(s7_scheme *sc, s7_pointer args, uint8_t typ)
{
  s7_pointer new_vec;
  s7_int len;

  len = total_sequence_length(sc, args, sc->vector_append_symbol, (typ == T_VECTOR) ? T_FREE : ((typ == T_FLOAT_VECTOR) ? T_REAL : T_INTEGER));
  new_vec = make_vector_1(sc, len, (typ == T_VECTOR) ? FILLED : NOT_FILLED, typ);  /* might hit GC in loop below so we can't use NOT_FILLED here */

  if (len > 0)
    {
      s7_pointer p, sv;
      int32_t i;

      sc->temp9 = new_vec; /* s7_copy below can call s7_error so s7_gc_protect here is tricky -- use a preset position perhaps? */
      sv = make_subvector(sc, new_vec);
      sc->temp10 = sv;

      for (i = 0, p = args; is_pair(p); p = cdr(p))
	{
	  s7_int n;
	  s7_pointer x;
	  x = car(p);
	  n = sequence_length(sc, x);
	  if (n > 0)
	    {
	      vector_length(sv) = n;
	      s7_copy(sc, set_plist_2(sc, x, sv));
	      vector_length(sv) = 0; /* so GC doesn't march off the end */
	      i += n;
	      if (typ == T_VECTOR)
		vector_elements(sv) = (s7_pointer *)(vector_elements(new_vec) + i);
	      else
		{
		  if (typ == T_FLOAT_VECTOR)
		    float_vector_elements(sv) = (s7_double *)(float_vector_elements(new_vec) + i);
		  else int_vector_elements(sv) = (s7_int *)(int_vector_elements(new_vec) + i);
		}
	    }
	}
      set_plist_2(sc, sc->nil, sc->nil);
      sc->temp9 = sc->nil;
      sc->temp10 = sc->nil;
      free_cell(sc, sv);
      vector_length(sv) = 0;
    }
  return(new_vec);
}

static s7_pointer string_append(s7_scheme *sc, s7_pointer args)
{
  s7_pointer new_str;
  s7_int len;

  len = total_sequence_length(sc, args, sc->string_append_symbol, (is_byte_vector(car(args))) ? T_INTEGER : T_CHARACTER);
  new_str = make_empty_string(sc, len, 0);
  if (is_byte_vector(car(args)))
    set_byte_vector(new_str);

  if (len > 0)
    {
      s7_pointer p, sv;
      int32_t i;

      sc->temp9 = new_str;
      sv = make_string_wrapper_with_length(sc, (const char *)string_value(new_str), len);
      if (is_byte_vector(new_str))
	set_byte_vector(sv);
      sc->temp10 = sv;

      for (i = 0, p = args; is_pair(p); p = cdr(p))
	{
	  s7_pointer x;
	  s7_int n;
	  x = car(p);
	  n = sequence_length(sc, x);
	  if (n > 0)
	    {
	      string_length(sv) = n;
	      s7_copy(sc, set_plist_2(sc, x, sv));
	      i += n;
	      string_value(sv) = (char *)(string_value(new_str) + i);
	    }
	}
      set_plist_2(sc, sc->nil, sc->nil);
      sc->temp9 = sc->nil;
      sc->temp10 = sc->nil;
      string_length(sv) = 0;
    }

  return(new_str);
}

static s7_pointer hash_table_append(s7_scheme *sc, s7_pointer args)
{
  s7_pointer new_hash, p;
  uint32_t gc_loc;
  new_hash = s7_make_hash_table(sc, sc->default_hash_table_length);
  gc_loc = s7_gc_protect(sc, new_hash);
  for (p = args; is_pair(p); p = cdr(p))
    s7_copy(sc, set_plist_2(sc, car(p), new_hash));
  set_plist_2(sc, sc->nil, sc->nil);
  s7_gc_unprotect_at(sc, gc_loc);
  return(new_hash);
}

static s7_pointer let_append(s7_scheme *sc, s7_pointer args)
{
  s7_pointer new_let, p, e;
  
  e = car(args);
  check_method(sc, e, sc->append_symbol, args);
  new_let = new_frame_in_env(sc, sc->nil);
  for (p = args; is_pair(p); p = cdr(p))
    s7_copy(sc, set_plist_2(sc, car(p), new_let)); 
  set_plist_2(sc, sc->nil, sc->nil);
  return(new_let);
}

static s7_pointer g_append(s7_scheme *sc, s7_pointer args)
{
  #define H_append "(append ...) returns its argument sequences appended into one sequence"
  #define Q_append s7_make_circular_signature(sc, 0, 1, sc->T)
  s7_pointer a1;

  /* it's possible to see brand-new lists at optimization time and set them to be uncopied here,
   *   but the various overheads swamp the gain.
   */
  if (is_null(args)) return(sc->nil);  /* (append) -> () */
  a1 = car(args);                      /* first arg determines result type unless all args but last are empty (sigh) */
  if (is_null(cdr(args))) return(a1);  /* (append <anything>) -> <anything> */

  switch (type(a1))
    {
    case T_NIL:
    case T_PAIR:
      return(g_list_append(sc, args)); /* only list case accepts any trailing arg because dotted lists are special */

    case T_VECTOR:
    case T_INT_VECTOR:
    case T_FLOAT_VECTOR:
      return(vector_append(sc, args, type(a1)));

    case T_STRING:
      return(string_append(sc, args));

    case T_HASH_TABLE:
      return(hash_table_append(sc, args));

    case T_LET:
      return(let_append(sc, args));

     default:
      check_method(sc, a1, sc->append_symbol, args);
    }
  return(wrong_type_argument_with_type(sc, sc->append_symbol, 1, a1, a_sequence_string)); /* (append 1 0) */
}

static s7_pointer append_p_pp(s7_pointer p1, s7_pointer p2) 
{
  /* plist in use above */
  s7_pointer val;
  cur_sc->temp7 = list_2(cur_sc, p1, p2);
  val = g_append(cur_sc, cur_sc->temp7);
  cur_sc->temp7 = cur_sc->nil;
  return(val);
}

static s7_pointer append_p_ppp(s7_pointer p1, s7_pointer p2, s7_pointer p3)
{
  s7_pointer val;
  cur_sc->temp7 = list_3(cur_sc, p1, p2, p3);
  val = g_append(cur_sc, cur_sc->temp7);
  cur_sc->temp7 = cur_sc->nil;
  return(val);
}

static s7_pointer object_to_list(s7_scheme *sc, s7_pointer obj)
{
  /* used only in format_to_port_1 and (map values ...) */
  switch (type(obj))
    {
    case T_INT_VECTOR:
    case T_FLOAT_VECTOR:
    case T_VECTOR:
      return(s7_vector_to_list(sc, obj));

    case T_STRING:
      if (is_byte_vector(obj))
	return(byte_vector_to_list(sc, string_value(obj), string_length(obj)));
      return(s7_string_to_list(sc, string_value(obj), string_length(obj)));

    case T_HASH_TABLE:
      if (hash_table_entries(obj) > 0)
	{
	  s7_pointer x, iterator;
	  iterator = s7_make_iterator(sc, obj);
	  sc->temp8 = iterator;
	  sc->w = sc->nil;
	  while (true)
	    {
	      x = s7_iterate(sc, iterator);
	      if (iterator_is_at_end(iterator)) break;
	      sc->w = cons(sc, x, sc->w);
	    }
	  x = sc->w;
	  sc->w = sc->nil;
	  sc->temp8 = sc->nil;
	  return(x);
	}
      return(sc->nil);

    case T_LET:
#if (!WITH_PURE_S7)
      check_method(sc, obj, sc->let_to_list_symbol, list_1(sc, obj));
#endif
      return(s7_let_to_list(sc, obj));

    case T_ITERATOR:
      {
	s7_pointer result, p = NULL;
	int32_t results = 0;
	result = sc->nil;
	while (true)
	  {
	    s7_pointer val;
	    val = s7_iterate(sc, obj);
	    if ((val == sc->ITERATOR_END) &&
		(iterator_is_at_end(obj)))
	      {
		sc->temp8 = sc->nil;
		return(result);
	      }
	    if (sc->safety > NO_SAFETY)
	      {
		results++;
		if (results > 10000)
		  {
		    fprintf(stderr, "iterator in object->list is creating a very long list!\n");
		    results = S7_LONG_MIN;
		  }
	      }
	    if (val != sc->no_value)
	      {
		if (is_null(result))
		  {
		    if (is_multiple_value(val))
		      {
			result = multiple_value(val);
			clear_multiple_value(val);
			for (p = result; is_pair(cdr(p)); p = cdr(p));
		      }
		    else 
		      {
			result = cons(sc, val, sc->nil);
			p = result;
		      }
		    sc->temp8 = result;
		  }
		else
		  {
		    if (is_multiple_value(val))
		      {
			set_cdr(p, multiple_value(val));
			clear_multiple_value(val);
			for (; is_pair(cdr(p)); p = cdr(p));
		      }
		    else 
		      {
			set_cdr(p, cons(sc, val, sc->nil));
			p = cdr(p);
		      }
		  }
	      }
	  }
      }

    case T_C_OBJECT:
      {
	int64_t i, len;
	s7_pointer x, z, result;
	uint32_t gc_z;

	x = c_object_length(sc, obj);
	if (s7_is_integer(x))
	  len = s7_integer(x);
	else return(sc->F);

	if (len < 0)
	  return(sc->F);
	if (len == 0)
	  return(sc->nil);

	result = make_list(sc, len, sc->nil);
	sc->temp8 = result;
	z = list_1(sc, sc->F);
	gc_z = s7_gc_protect(sc, z);

	set_car(sc->z2_1, sc->x);
	set_car(sc->z2_2, sc->z);
	for (i = 0, x = result; i < len; i++, x = cdr(x))
	  {
	    set_car(z, make_integer(sc, i));
	    set_car(x, (*(c_object_ref(sc, obj)))(sc, obj, z));
	  }
	sc->x = car(sc->z2_1);
	sc->z = car(sc->z2_2);
	s7_gc_unprotect_at(sc, gc_z);
	sc->temp8 = sc->nil;
	return(result);
      }
    }
  return(obj);
}



/* -------------------------------- object->let -------------------------------- */

static s7_pointer stack_entries(s7_scheme *sc, s7_pointer stack, int64_t top);

static s7_pointer g_object_to_let(s7_scheme *sc, s7_pointer args)
{
  #define H_object_to_let "(object->let obj) returns a let (namespace) describing obj."
  #define Q_object_to_let s7_make_signature(sc, 2, sc->is_let_symbol, sc->T)

  s7_pointer obj;
  obj = car(args);

  switch (type(obj))
    {
    case T_NIL:
      return(s7_inlet(sc, s7_list(sc, 4, sc->value_symbol, obj, sc->type_symbol, sc->is_null_symbol)));

    case T_UNSPECIFIED:
      return(s7_inlet(sc, s7_list(sc, 4, sc->value_symbol, obj, sc->type_symbol, sc->is_unspecified_symbol)));

    case T_UNDEFINED:
      return(s7_inlet(sc, s7_list(sc, 4, sc->value_symbol, obj, sc->type_symbol, obj)));

    case T_SYNTAX:
      return(s7_inlet(sc, s7_list(sc, 4, sc->value_symbol, obj, sc->type_symbol, sc->is_syntax_symbol)));

    case T_EOF_OBJECT:
      return(s7_inlet(sc, s7_list(sc, 4, sc->value_symbol, obj, sc->type_symbol, sc->is_eof_object_symbol)));

    case T_BOOLEAN:
      return(s7_inlet(sc, s7_list(sc, 4, sc->value_symbol, obj, sc->type_symbol, sc->is_boolean_symbol)));
      
    case T_SYMBOL:
      return(s7_inlet(sc, s7_list(sc, 4, sc->value_symbol, obj, sc->type_symbol, (is_keyword(obj)) ? sc->is_keyword_symbol : sc->is_symbol_symbol)));
      
    case T_CHARACTER:
      return(s7_inlet(sc, s7_list(sc, 4, sc->value_symbol, obj, sc->type_symbol, sc->is_char_symbol)));
      
    case T_INTEGER:
    case T_BIG_INTEGER:
      return(s7_inlet(sc, s7_list(sc, 4, sc->value_symbol, obj, sc->type_symbol, sc->is_integer_symbol)));
      
    case T_RATIO:
    case T_BIG_RATIO:
      return(s7_inlet(sc, s7_list(sc, 4, sc->value_symbol, obj, sc->type_symbol, sc->is_rational_symbol)));
      
    case T_REAL:
    case T_BIG_REAL:
      return(s7_inlet(sc, s7_list(sc, 4, sc->value_symbol, obj, sc->type_symbol, sc->is_real_symbol)));
      
    case T_COMPLEX:
    case T_BIG_COMPLEX:
      return(s7_inlet(sc, s7_list(sc, 4, sc->value_symbol, obj, sc->type_symbol, sc->is_complex_symbol)));

    case T_STRING:
      return(s7_inlet(sc, s7_list(sc, 6, sc->value_symbol, obj, 
				  sc->type_symbol, (is_byte_vector(obj)) ? sc->is_byte_vector_symbol : sc->is_string_symbol,
				  sc->length_symbol, s7_length(sc, obj))));

    case T_PAIR:
      return(s7_inlet(sc, s7_list(sc, 6, sc->value_symbol, obj, 
				  sc->type_symbol, sc->is_pair_symbol,
				  sc->length_symbol, s7_length(sc, obj))));

    case T_RANDOM_STATE:
#if WITH_GMP
      return(s7_inlet(sc, s7_list(sc, 4, sc->value_symbol, obj, sc->type_symbol, sc->is_random_state_symbol)));
#else
      return(s7_inlet(sc, s7_list(sc, 8, sc->value_symbol, obj,
				  sc->type_symbol, sc->is_random_state_symbol,
				  s7_make_symbol(sc, "seed"), s7_make_integer(sc, random_seed(obj)),
				  s7_make_symbol(sc, "carry"), s7_make_integer(sc, random_carry(obj)))));
#endif

    case T_GOTO:
      return(s7_inlet(sc, s7_list(sc, 6, sc->value_symbol, obj, 
				  sc->type_symbol, s7_make_symbol(sc, "goto?"),
				  s7_make_symbol(sc, "active"), s7_make_boolean(sc, call_exit_active(obj)))));

    case T_INT_VECTOR:
    case T_FLOAT_VECTOR:
    case T_VECTOR:
      return(s7_inlet(sc, s7_list(sc, 10, sc->value_symbol, obj, 
				  sc->type_symbol, 
				    (is_int_vector(obj)) ? sc->is_int_vector_symbol : ((is_float_vector(obj)) ? sc->is_float_vector_symbol : sc->is_vector_symbol),
				  sc->length_symbol, s7_length(sc, obj),
				  s7_make_symbol(sc, "dimensions"), g_vector_dimensions(sc, set_plist_1(sc, obj)),
				  s7_make_symbol(sc, "shared"), 
				    ((vector_has_dimensional_info(obj)) && (is_normal_vector(shared_vector(obj)))) ? shared_vector(obj) : sc->F)));

    case T_C_POINTER:
      /* raw_pointer_info can be a let and might have an object->let method (see c_object below) */
      return(s7_inlet(sc, s7_list(sc, 10, sc->value_symbol, obj, 
				  sc->type_symbol, sc->is_c_pointer_symbol,
				  s7_make_symbol(sc, "c-pointer"), s7_make_integer(sc, (s7_int)raw_pointer(obj)),
				  s7_make_symbol(sc, "c-type"), raw_pointer_type(obj),
				  s7_make_symbol(sc, "info"), raw_pointer_info(obj))));

    case T_CONTINUATION:
      {
	s7_pointer let;
	uint32_t gc_loc;
	let = s7_inlet(sc, s7_list(sc, 4, sc->value_symbol, obj, sc->type_symbol, sc->is_continuation_symbol));
	gc_loc = s7_gc_protect(sc, let);
	s7_varlet(sc, let, s7_make_symbol(sc, "stack"), stack_entries(sc, continuation_stack(obj), continuation_stack_top(obj)));
	s7_gc_unprotect_at(sc, gc_loc);
	return(let);
      }

    case T_ITERATOR:
      {
	s7_pointer let, seq;
	seq = iterator_sequence(obj);
	let = s7_inlet(sc, s7_list(sc, 8, sc->value_symbol, obj,
				   sc->type_symbol, sc->is_iterator_symbol,
				   s7_make_symbol(sc, "at-end"), s7_make_boolean(sc, iterator_is_at_end(obj)),
				   s7_make_symbol(sc, "sequence"), iterator_sequence(obj)));
	if (is_pair(seq))
	  s7_varlet(sc, let, sc->length_symbol, s7_length(sc, seq));
	else
	  {
	    if (is_hash_table(seq))
	      s7_varlet(sc, let, sc->length_symbol, s7_make_integer(sc, hash_table_entries(seq)));
	    else s7_varlet(sc, let, sc->length_symbol, s7_length(sc, obj));
	  }
	if ((is_string(seq)) || 
	    (is_normal_vector(seq)) ||
	    (is_int_vector(seq)) ||
	    (is_float_vector(seq)) ||
	    (seq == sc->rootlet) ||
	    (is_c_object(seq)) ||
	    (is_hash_table(seq)))
	  s7_varlet(sc, let, s7_make_symbol(sc, "position"), s7_make_integer(sc, iterator_position(obj)));
	else
	  {
	    if (is_pair(seq))
	      s7_varlet(sc, let, s7_make_symbol(sc, "position"), iterator_current(obj));
	  }
	return(let);
      }

    case T_HASH_TABLE:
      {
	s7_pointer let;
	let = s7_inlet(sc, s7_list(sc, 10, sc->value_symbol, obj,
				   sc->type_symbol, sc->is_hash_table_symbol,
				   sc->length_symbol, s7_length(sc, obj),
				   s7_make_symbol(sc, "entries"), s7_make_integer(sc, hash_table_entries(obj)),
				   s7_make_symbol(sc, "locked"), s7_make_boolean(sc, hash_table_checker_locked(obj))));

	if ((hash_table_checker(obj) == hash_eq) ||
	    (hash_table_checker(obj) == hash_c_function) ||
	    (hash_table_checker(obj) == hash_closure) ||
	    (hash_table_checker(obj) == hash_equal_eq) ||
	    (hash_table_checker(obj) == hash_equal_syntax) ||
	    (hash_table_checker(obj) == hash_symbol))
	  s7_varlet(sc, let, s7_make_symbol(sc, "function"), sc->is_eq_symbol);
	else
	  {
	    if (hash_table_checker(obj) == hash_eqv) 
	      s7_varlet(sc, let, s7_make_symbol(sc, "function"), sc->is_eqv_symbol);
	    else
	      {
		if ((hash_table_checker(obj) == hash_equal) || 
		    (hash_table_checker(obj) == hash_empty))
		  s7_varlet(sc, let, s7_make_symbol(sc, "function"), sc->is_equal_symbol);
		else
		  {
		    if (hash_table_checker(obj) == hash_morally_equal) 
		      s7_varlet(sc, let, s7_make_symbol(sc, "function"), sc->is_morally_equal_symbol);
		    else
		      {
			if ((hash_table_checker(obj) == hash_number) ||
			    (hash_table_checker(obj) == hash_int) ||
			    (hash_table_checker(obj) == hash_float) ||
			    (hash_table_checker(obj) == hash_equal_real) ||
			    (hash_table_checker(obj) == hash_equal_complex))
			  s7_varlet(sc, let, s7_make_symbol(sc, "function"), sc->eq_symbol);
			else
			  {
			    if (hash_table_checker(obj) == hash_string) 
			      s7_varlet(sc, let, s7_make_symbol(sc, "function"), sc->string_eq_symbol);
			    else
			      {
				if (hash_table_checker(obj) == hash_char) 
				  s7_varlet(sc, let, s7_make_symbol(sc, "function"), sc->char_eq_symbol);
#if (!WITH_PURE_S7)
				else
				  {
				    if (hash_table_checker(obj) == hash_ci_char) 
				      s7_varlet(sc, let, s7_make_symbol(sc, "function"), sc->char_ci_eq_symbol);
				    else
				      {
					if (hash_table_checker(obj) == hash_ci_string) 
					  s7_varlet(sc, let, s7_make_symbol(sc, "function"), sc->string_ci_eq_symbol);
				      }}
#endif
			      }}}}}}
	return(let);
      }

    case T_LET:
      {
	s7_pointer let;
	let = s7_inlet(sc, s7_list(sc, 10, sc->value_symbol, obj,
				   sc->type_symbol, sc->is_let_symbol,
				   sc->length_symbol, s7_length(sc, obj),
				   s7_make_symbol(sc, "open"), s7_make_boolean(sc, has_methods(obj)),
				   sc->outlet_symbol, (obj == sc->rootlet) ? sc->nil : outlet(obj)));
	if (obj == sc->rootlet)
	  s7_varlet(sc, let, s7_make_symbol(sc, "alias"), sc->rootlet_symbol);
	else
	  {
	    if (obj == sc->owlet)
	      s7_varlet(sc, let, s7_make_symbol(sc, "alias"), sc->owlet_symbol);
	    else
	      {
		if (is_funclet(obj))
		  {
		    s7_varlet(sc, let, s7_make_symbol(sc, "function"), funclet_function(obj));
		    if ((let_file(obj) > 0) && 
			(let_file(obj) < (s7_int)sc->file_names_top) &&
			(let_line(obj) > 0))
		      {
			s7_varlet(sc, let, s7_make_symbol(sc, "file"), sc->file_names[let_file(obj)]);
			s7_varlet(sc, let, s7_make_symbol(sc, "line"), make_integer(sc, let_line(obj)));
		      }
		  }
	      }
	  }
	if (has_methods(obj))
	  {
	    s7_pointer func;
	    func = find_method(sc, obj, sc->object_to_let_symbol);
	    if (func != sc->undefined)
	      {
		uint32_t gc_loc;
		gc_loc = s7_gc_protect(sc, let);
		s7_apply_function(sc, func, list_2(sc, obj, let));
		s7_gc_unprotect_at(sc, gc_loc);
	      }
	  }
	return(let);
      }

    case T_C_OBJECT:
      {
	s7_pointer let, clet;
	clet = c_object_let(obj);
	let = s7_inlet(sc, s7_list(sc, 10,
				   sc->value_symbol, obj,
				   sc->type_symbol, sc->is_c_object_symbol,
				   s7_make_symbol(sc, "c-object-type"), s7_make_integer(sc, c_object_type(obj)),
				   s7_make_symbol(sc, "c-object-let"), clet,
				   s7_make_symbol(sc, "class"), c_object_scheme_name(sc, obj)));
	/* TODO: local 'let entry causes trouble -- it's an error now in s7_varlet 
	 *       also how to examine the c_object_type table?
	 */
	if ((is_let(clet)) &&
	    ((has_methods(clet)) || (has_methods(obj))))
	  {
	    s7_pointer func;
	    func = find_method(sc, clet, sc->object_to_let_symbol);
	    if (func != sc->undefined)
	      {
		uint32_t gc_loc;
		gc_loc = s7_gc_protect(sc, let);
		s7_apply_function(sc, func, list_2(sc, obj, let));
		s7_gc_unprotect_at(sc, gc_loc);
	      }
	  }
	return(let);
      }

    case T_INPUT_PORT:
    case T_OUTPUT_PORT:
      {
	s7_pointer let;
	uint32_t gc_loc;
	let = s7_inlet(sc, s7_list(sc, 8, sc->value_symbol, obj,
				   sc->type_symbol, (is_input_port(obj)) ? sc->is_input_port_symbol : sc->is_output_port_symbol,
				   s7_make_symbol(sc, "port-type"), 
				     (is_string_port(obj)) ? sc->string_symbol : 
				       ((is_file_port(obj)) ? s7_make_symbol(sc, "file") : s7_make_symbol(sc, "function")),
				   s7_make_symbol(sc, "closed"), s7_make_boolean(sc, port_is_closed(obj))));
	gc_loc = s7_gc_protect(sc, let);
	if (is_file_port(obj))
	  {
	    s7_varlet(sc, let, s7_make_symbol(sc, "file"), g_port_filename(sc, set_plist_1(sc, obj)));
	    if (is_input_port(obj))
	      s7_varlet(sc, let, s7_make_symbol(sc, "line"), g_port_line_number(sc, set_plist_1(sc, obj)));
	  }
	if (port_data_size(obj) > 0)
	  {
	    s7_varlet(sc, let, sc->length_symbol, s7_make_integer(sc, port_data_size(obj)));
	    s7_varlet(sc, let, s7_make_symbol(sc, "position"), s7_make_integer(sc, port_position(obj)));
	    /* I think port_data need not be null-terminated, but s7_make_string assumes it is:
	     *   both valgrind and lib*san complain about the uninitialized data during strlen.
	     */
	    s7_varlet(sc, let, s7_make_symbol(sc, "data"), s7_make_string_with_length(sc, (const char *)port_data(obj), port_position(obj)));
	  }
	s7_gc_unprotect_at(sc, gc_loc);
	return(let);
      }

    case T_CLOSURE:
    case T_CLOSURE_STAR:
    case T_MACRO:
    case T_MACRO_STAR:
    case T_BACRO:
    case T_BACRO_STAR:
      {
	s7_pointer let, sig;
	const char* doc;
	uint32_t gc_loc;
	let = s7_inlet(sc, s7_list(sc, 6, sc->value_symbol, obj,
				   sc->type_symbol, (is_procedure(obj)) ? sc->is_procedure_symbol : sc->is_macro_symbol,
				   s7_make_symbol(sc, "arity"), s7_arity(sc, obj)));
	gc_loc = s7_gc_protect(sc, let);

	sig = s7_procedure_signature(sc, obj);
	if (is_pair(sig))
	  s7_varlet(sc, let, sc->signature_symbol, sig);
	
	doc = s7_procedure_documentation(sc, obj);
	if (doc)
	  s7_varlet(sc, let, sc->documentation_symbol, s7_make_string(sc, doc));
	
	if (is_let(closure_let(obj)))
	  {
	    s7_pointer flet;
	    flet = closure_let(obj);
	    if ((let_file(flet) > 0) && 
		(let_file(flet) < (s7_int)sc->file_names_top) &&
		(let_line(flet) > 0))
	      {
		s7_varlet(sc, let, s7_make_symbol(sc, "file"), sc->file_names[let_file(flet)]);
		s7_varlet(sc, let, s7_make_symbol(sc, "line"), make_integer(sc, let_line(flet)));
	      }
	  }
	
	if (closure_setter(obj) != sc->F)
	  s7_varlet(sc, let, s7_make_symbol(sc, "setter"), closure_setter(obj));

	s7_varlet(sc, let, s7_make_symbol(sc, "source"), 
			    append_in_place(sc, list_2(sc, (is_closure_star(obj)) ? sc->lambda_star_symbol : sc->lambda_symbol, 
						       closure_args(obj)), 
					    closure_body(obj)));
	s7_gc_unprotect_at(sc, gc_loc);
	return(let);
      }

    case T_C_MACRO:
    case T_C_FUNCTION_STAR:
    case T_C_FUNCTION:
    case T_C_ANY_ARGS_FUNCTION:
    case T_C_OPT_ARGS_FUNCTION:
    case T_C_RST_ARGS_FUNCTION:
      {
	s7_pointer let, sig;
	const char* doc;
	let = s7_inlet(sc, s7_list(sc, 6, sc->value_symbol, obj,
				   sc->type_symbol, (is_procedure(obj)) ? sc->is_procedure_symbol : sc->is_macro_symbol,
				   s7_make_symbol(sc, "arity"), s7_arity(sc, obj)));

	sig = s7_procedure_signature(sc, obj);
	if (is_pair(sig))
	  s7_varlet(sc, let, sc->signature_symbol, sig);
	
	doc = s7_procedure_documentation(sc, obj);
	if (doc)
	  s7_varlet(sc, let, sc->documentation_symbol, s7_make_string(sc, doc));

	if (c_function_setter(obj) != sc->F)
	  s7_varlet(sc, let, s7_make_symbol(sc, "setter"), c_function_setter(obj));
	
	return(let);
      }

    default:
#if DEBUGGING
      fprintf(stderr, "object->let: %s, type: %d\n", DISPLAY(obj), type(obj));
#endif
      return(sc->F);
    }

  return(sc->F);
}


/* ---------------- stacktrace ---------------- */

static s7_pointer stacktrace_find_caller(s7_scheme *sc, s7_pointer e)
{
  if ((is_let(e)) && (e != sc->rootlet))
    {
      if (is_funclet(e))
	return(funclet_function(e));
      return(stacktrace_find_caller(sc, outlet(e)));
    }
  return(sc->F);
}

static bool stacktrace_find_let(s7_scheme *sc, int64_t loc, s7_pointer e)
{
  return((loc > 0) &&
	 ((stack_let(sc->stack, loc) == e) ||
	  (stacktrace_find_let(sc, loc - 4, e))));
}

static int32_t stacktrace_find_error_hook_quit(s7_scheme *sc)
{
  int64_t i;
  for (i = s7_stack_top(sc) - 1; i >= 3; i -= 4)
    if (stack_op(sc->stack, i) == OP_ERROR_HOOK_QUIT)
      return(i);
  return(-1);
}

static bool stacktrace_in_error_handler(s7_scheme *sc, int64_t loc)
{
  return((outlet(sc->owlet) == sc->envir) ||
	 (stacktrace_find_let(sc, loc * 4, outlet(sc->owlet))) ||
	 (stacktrace_find_error_hook_quit(sc) > 0));
}


static bool stacktrace_error_hook_function(s7_scheme *sc, s7_pointer sym)
{
  if (is_symbol(sym))
    {
      s7_pointer f;
      f = s7_symbol_value(sc, sym);
      return((is_procedure(f)) &&
	     (is_procedure(sc->error_hook)) &&
	     (hook_has_functions(sc->error_hook)) &&
	     (direct_memq(f, s7_hook_functions(sc, sc->error_hook))));
    }
  return(false);
}

static char *stacktrace_walker(s7_scheme *sc, s7_pointer code, s7_pointer e,
			       char *notes, uint32_t gc_syms,
			       int32_t code_cols, int32_t total_cols, int32_t notes_start_col,
			       bool as_comment)
{
  s7_pointer syms;
  syms = gc_protected_at(sc, gc_syms);

  if (is_symbol(code))
    {
      if ((!direct_memq(code, syms)) &&
	  (!is_slot(global_slot(code))))
	{
	  s7_pointer val;

	  syms = cons(sc, code, syms);
	  gc_protected_at(sc, gc_syms) = syms;

	  val = s7_symbol_local_value(sc, code, e);
	  if ((val) && (val != sc->undefined) &&
	      (!is_any_macro(val)))
	    {
	      int32_t typ;

	      typ = type(val);
	      if (typ < T_GOTO)
		{
		  char *objstr, *str;
		  const char *spaces;
		  int32_t objlen, new_note_len, notes_max, cur_line_len = 0, spaces_len;
		  bool new_notes_line = false, old_short_print;
		  s7_int old_len;

		  spaces = "                                                                                ";
		  spaces_len = strlen(spaces);

		  if (notes_start_col < 0) notes_start_col = 50;
		  if (notes_start_col > total_cols) notes_start_col = 0;
		  notes_max = total_cols - notes_start_col;

		  old_short_print = sc->short_print;
		  sc->short_print = true;
		  old_len = sc->print_length;
		  if (sc->print_length > 4) sc->print_length = 4;
		  objstr = s7_object_to_c_string(sc, val);
		  objlen = safe_strlen(objstr);
		  if ((objlen > notes_max) && 
		      (notes_max > 5))
		    {
		      objstr[notes_max - 4] = '.';
		      objstr[notes_max - 3] = '.';
		      objstr[notes_max - 2] = '.';
		      objstr[notes_max - 1] = '\0';
		      objlen = notes_max;
		    }
		  sc->short_print = old_short_print;
		  sc->print_length = old_len;

		  new_note_len = symbol_name_length(code) + 3 + objlen;
		  /* we want to append this much info to the notes, but does it need a new line? */
		  if (notes_start_col < code_cols)
		    new_notes_line = true;
		  else
		    {
		      if (notes)
			{
			  char *last_newline;
			  last_newline = strrchr(notes, (int)'\n'); /* returns ptr to end if none = nil if not found? */
			  if (last_newline)
			    cur_line_len = strlen(notes) - strlen(last_newline);
			  else cur_line_len = strlen(notes);
			  new_notes_line = ((cur_line_len + new_note_len) > notes_max);
			}
		    }

		  if (new_notes_line)
		    {
		      new_note_len += (4 + notes_start_col + ((notes) ? strlen(notes) : 0));
		      str = (char *)malloc(new_note_len * sizeof(char));
		      snprintf(str, new_note_len, "%s\n%s%s%s%s: %s",
			       (notes) ? notes : "",
			       (as_comment) ? "; " : "",
			       (spaces_len >= notes_start_col) ? (char *)(spaces + spaces_len - notes_start_col) : "",
			       (as_comment) ? "" : " ; ",
			       symbol_name(code),
			       objstr);
		    }
		  else
		    {
		      new_note_len += ((notes) ? strlen(notes) : 0) + 4;
		      str = (char *)malloc(new_note_len * sizeof(char));
		      snprintf(str, new_note_len, "%s%s%s: %s",
			       (notes) ? notes : "",
			       (notes) ? ", " : " ; ",
			       symbol_name(code),
			       objstr);
		    }
		  free(objstr);
		  if (notes) free(notes);
		  return(str);
		}
	    }
	}
      return(notes);
    }
  if (is_pair(code))
    {
      notes = stacktrace_walker(sc, car(code), e, notes, gc_syms, code_cols, total_cols, notes_start_col, as_comment);
      return(stacktrace_walker(sc, cdr(code), e, notes, gc_syms, code_cols, total_cols, notes_start_col, as_comment));
    }
  return(notes);
}

static char *stacktrace_add_func(s7_scheme *sc, s7_pointer f, s7_pointer code, char *errstr, char *notes, int32_t code_max, bool as_comment)
{
  int32_t newlen, errlen;
  char *newstr, *str;

  errlen = strlen(errstr);
  if ((is_symbol(f)) &&
      (f != car(code)))
    {
      newlen = symbol_name_length(f) + errlen + 10;
      newstr = (char *)malloc(newlen * sizeof(char));
      errlen = snprintf(newstr, newlen, "%s: %s", symbol_name(f), errstr);
    }
  else
    {
      newlen = errlen + 8;
      newstr = (char *)malloc(newlen * sizeof(char));
      if ((errlen > 2) && (errstr[2] == '('))
	errlen = snprintf(newstr, newlen, "  %s", errstr);
      else errlen = snprintf(newstr, newlen, "%s", errstr);
    }

  newlen = code_max + 8 + ((notes) ? strlen(notes) : 0);
  str = (char *)malloc(newlen * sizeof(char));

  if (errlen >= code_max)
    {
      newstr[code_max - 4] = '.';
      newstr[code_max - 3] = '.';
      newstr[code_max - 2] = '.';
      newstr[code_max - 1] = '\0';
      snprintf(str, newlen, "%s%s%s\n", (as_comment) ? "; " : "", newstr, (notes) ? notes : "");
    }
  else
    {
      /* send out newstr, pad with spaces to code_max, then notes */
      int32_t len;
      len = snprintf(str, newlen, "%s%s", (as_comment) ? "; " : "", newstr);
      if (notes)
	{
	  int32_t i;
	  for (i = len; i < code_max - 1; i++)
	    str[i] = ' ';
	  str[i] = '\0';
#ifdef __OpenBSD__
	  strlcat(str, notes, newlen);
	  strlcat(str, "\n", newlen);
#else
	  strcat(str, notes);
	  strcat(str, "\n");
#endif
	}
    }
  free(newstr);
  return(str);
}


static char *stacktrace_1(s7_scheme *sc, int32_t frames_max, int32_t code_cols, int32_t total_cols, int32_t notes_start_col, bool as_comment)
{
  char *str;
  int64_t loc, top, frames = 0;
  uint32_t gc_syms;

  gc_syms = s7_gc_protect(sc, sc->nil);
  str = NULL;
  top = (sc->stack_end - sc->stack_start) / 4; /* (*s7* 'stack_top), not s7_stack_top! */

  if (stacktrace_in_error_handler(sc, top))
    {
      s7_pointer err_code;
      err_code = slot_value(sc->error_code);
      if (is_pair(err_code))
	{
	  char *errstr, *notes = NULL;
	  s7_pointer cur_env, f;

	  errstr = s7_object_to_c_string(sc, err_code);
	  cur_env = outlet(sc->owlet);
	  f = stacktrace_find_caller(sc, cur_env); /* this is a symbol */
	  if ((is_let(cur_env)) &&
	      (cur_env != sc->rootlet))
	    notes = stacktrace_walker(sc, err_code, cur_env, NULL, gc_syms, code_cols, total_cols, notes_start_col, as_comment);
	  str = stacktrace_add_func(sc, f, err_code, errstr, notes, code_cols, as_comment);
	  free(errstr);
	}

      /* now if OP_ERROR_HOOK_QUIT is in the stack, jump past it!
       */
      loc = stacktrace_find_error_hook_quit(sc);
      if (loc > 0) top = (loc + 1) / 4;
    }

  for (loc = top - 1; loc > 0; loc--)
    {
      s7_pointer code;
      int32_t true_loc;

      true_loc = (int32_t)(loc + 1) * 4 - 1;
      code = stack_code(sc->stack, true_loc); /* can code be free here? [hit this once, could not repeat it] */

      if (is_pair(code))
	{
	  char *codestr;
	  codestr = s7_object_to_c_string(sc, code);
	  if (codestr)
	    {
	      if ((!local_strcmp(codestr, "(result)")) &&
		  (!local_strcmp(codestr, "(#f)")) &&
		  (!strstr(codestr, "(stacktrace)")) &&
		  (!strstr(codestr, "(stacktrace ")))
		{
		  s7_pointer e, f;

		  e = stack_let(sc->stack, true_loc);
		  f = stacktrace_find_caller(sc, e);
		  if (!stacktrace_error_hook_function(sc, f))
		    {
		      char *notes = NULL, *newstr;
		      int32_t newlen;

		      frames++;
		      if (frames > frames_max)
			{
			  free(codestr);
			  s7_gc_unprotect_at(sc, gc_syms);
			  return(str);
			}

		      if ((is_let(e)) && (e != sc->rootlet))
			notes = stacktrace_walker(sc, code, e, NULL, gc_syms, code_cols, total_cols, notes_start_col, as_comment);
		      newstr = stacktrace_add_func(sc, f, code, codestr, notes, code_cols, as_comment);
		      free(codestr);
		      if ((notes) && (notes != newstr) && (is_let(e)) && (e != sc->rootlet)) free(notes); /* double free somehow?? */

		      newlen = strlen(newstr) + 1 + ((str) ? strlen(str) : 0);
		      codestr = (char *)malloc(newlen * sizeof(char));
		      snprintf(codestr, newlen, "%s%s", (str) ? str : "", newstr);
		      if (str) free(str);
		      free(newstr);
		      str = codestr;
		      codestr = NULL;
		    }
		  else free(codestr);
		}
	      else free(codestr);
	    }
	}
    }

  s7_gc_unprotect_at(sc, gc_syms);
  return(str);
}


s7_pointer s7_stacktrace(s7_scheme *sc)
{
  char *str;
  str = stacktrace_1(sc, 30, 45, 80, 45, false);
  return(make_string_uncopied_with_length(sc, str, safe_strlen(str)));
}


static s7_pointer g_stacktrace(s7_scheme *sc, s7_pointer args)
{
  #define H_stacktrace "(stacktrace (max-frames 30) (code-cols 50) (total-cols 80) (note-col 50) as-comment) returns \
a stacktrace as a string.  Each line has two portions, the code being evaluated and a note giving \
the value of local variables in that code.  The first argument sets how many lines are displayed. \
The next three arguments set the length and layout of those lines.  'as-comment' if #t causes each \
line to be preceded by a semicolon."
  #define Q_stacktrace s7_make_signature(sc, 6, sc->is_string_symbol, sc->is_integer_symbol, sc->is_integer_symbol, sc->is_integer_symbol, sc->is_integer_symbol, sc->is_boolean_symbol)

  s7_int max_frames = 30, code_cols = 50, total_cols = 80, notes_start_col = 50;
  bool as_comment = false;
  char *str;

  if (!is_null(args))
    {
      if (s7_is_integer(car(args)))
	{
	  max_frames = s7_integer(car(args));
	  if ((max_frames <= 0) || (max_frames > s7_int32_max))
	    max_frames = 30;
	  args = cdr(args);
	  if (!is_null(args))
	    {
	      if (s7_is_integer(car(args)))
		{
		  code_cols = s7_integer(car(args));
		  if ((code_cols <= 8) || (code_cols > s7_int32_max))
		    code_cols = 50;
		  args = cdr(args);
		  if (!is_null(args))
		    {
		      if (s7_is_integer(car(args)))
			{
			  total_cols = s7_integer(car(args));
			  if ((total_cols <= code_cols) || (total_cols > s7_int32_max))
			    total_cols = 80;
			  args = cdr(args);
			  if (!is_null(args))
			    {
			      if (s7_is_integer(car(args)))
				{
				  notes_start_col = s7_integer(car(args));
				  if ((notes_start_col <= 0) || (notes_start_col > s7_int32_max))
				    notes_start_col = 50;
				  args = cdr(args);
				  if (!is_null(args))
				    {
				      if (s7_is_boolean(car(args)))
					as_comment = s7_boolean(sc, car(args));
				      else return(wrong_type_argument(sc, sc->stacktrace_symbol, 5, car(args), T_BOOLEAN));
				    }
				}
			      else return(wrong_type_argument(sc, sc->stacktrace_symbol, 4, car(args), T_INTEGER));
			    }
			}
		      else return(wrong_type_argument(sc, sc->stacktrace_symbol, 3, car(args), T_INTEGER));
		    }
		}
	      else return(wrong_type_argument(sc, sc->stacktrace_symbol, 2, car(args), T_INTEGER));
	    }
	}
      else method_or_bust(sc, car(args), sc->stacktrace_symbol, args, T_INTEGER, 1);
    }
  str = stacktrace_1(sc, (int32_t)max_frames, (int32_t)code_cols, (int32_t)total_cols, (int32_t)notes_start_col, as_comment);
  return(make_string_uncopied_with_length(sc, str, safe_strlen(str)));
}


/* -------- s7_history, s7_add_to_history -------- */

s7_pointer s7_add_to_history(s7_scheme *sc, s7_pointer entry)
{
#if WITH_HISTORY
  set_current_code(sc, entry);
#endif
  return(entry);
}

s7_pointer s7_history(s7_scheme *sc)
{
  return(sc->cur_code);
}



/* -------- error handlers -------- */

static const char *make_type_name(s7_scheme *sc, const char *name, int32_t article)
{
  int32_t i, slen, len;

  slen = safe_strlen(name);
  len = slen + 8;
  if (len > sc->typnam_len)
    {
      if (sc->typnam) free(sc->typnam);
      sc->typnam = (char *)malloc(len * sizeof(char));
      sc->typnam_len = len;
    }
  if (article == INDEFINITE_ARTICLE)
    {
      i = 1;
      sc->typnam[0] = 'a';
      if ((name[0] == 'a') || (name[0] == 'e') || (name[0] == 'i') || (name[0] == 'o') || (name[0] == 'u'))
	sc->typnam[i++] = 'n';
      sc->typnam[i++] = ' ';
    }
  else i = 0;
  memcpy((void *)(sc->typnam + i), (void *)name, slen);
  sc->typnam[i + slen] = '\0';
  return(sc->typnam);
}


static const char *type_name_from_type(s7_scheme *sc, int32_t typ, int32_t article)
{
  static const char *frees[2] =          {"free cell",          "a free cell"};
  static const char *nils[2] =           {"nil",                "nil"};
  static const char *eofs[2] =           {"eof",                "eof"};
  static const char *unspecs[2] =        {"unspecified",        "unspecified"};
  static const char *undefs[2] =         {"undefined",          "undefined"};
  static const char *booleans[2] =       {"boolean",            "boolean"};
  static const char *strings[2] =        {"string",             "a string"};
  static const char *symbols[2] =        {"symbol",             "a symbol"};
  static const char *syntaxes[2] =       {"syntax",             "syntactic"};
  static const char *pairs[2] =          {"pair",               "a pair"};
  static const char *gotos[2] =          {"goto",               "a goto (from call-with-exit)"};
  static const char *continuations[2] =  {"continuation",       "a continuation"};
  static const char *c_functions[2] =    {"c-function",         "a c-function"};
  static const char *c_function_s[2] =   {"c-function*",        "a c-function*"};
  static const char *macros[2] =         {"macro",              "a macro"};
  static const char *c_macros[2] =       {"c-macro",            "a c-macro"};
  static const char *bacros[2] =         {"bacro",              "a bacro"};
  static const char *vectors[2] =        {"vector",             "a vector"};
  static const char *int_vectors[2] =    {"int-vector",         "an int-vector"};
  static const char *float_vectors[2] =  {"float-vector",       "a float-vector"};
  static const char *c_pointers[2] =     {"C pointer",          "a raw C pointer"};
  static const char *counters[2] =       {"internal counter",   "an internal counter"};
  static const char *optlists[2] =       {"internal optlist",   "an internal optlist"};
  static const char *baffles[2] =        {"baffle",             "a baffle"};
  static const char *slots[2] =          {"slot",               "a slot (variable binding)"};
  static const char *characters[2] =     {"character",          "a character"};
  static const char *catches[2] =        {"catch",              "a catch"};
  static const char *dynamic_winds[2] =  {"dynamic-wind",       "a dynamic-wind"};
  static const char *hash_tables[2] =    {"hash-table",         "a hash-table"};
  static const char *iterators[2] =      {"iterator",           "an iterator"};
  static const char *environments[2] =   {"environment",        "an environment"};
  static const char *integers[2] =       {"integer",            "an integer"};
  static const char *big_integers[2] =   {"big integer",        "a big integer"};
  static const char *ratios[2] =         {"ratio",              "a ratio"};
  static const char *big_ratios[2] =     {"big ratio",          "a big ratio"};
  static const char *reals[2] =          {"real",               "a real"};
  static const char *big_reals[2] =      {"big real",           "a big real"};
  static const char *complexes[2] =      {"complex number",     "a complex number"};
  static const char *big_complexes[2] =  {"big complex number", "a big complex number"};
  static const char *functions[2] =      {"function",           "a function"};
  static const char *function_stars[2] = {"function*",          "a function*"};
  static const char *rngs[2] =           {"random-state",       "a random-state"};
  static const char *inputs[2] =         {"input port",         "an input port"};
  static const char *outputs[2] =        {"output port",        "an output port"};
  static const char *c_objects[2] =      {"c_object",           "a c_object"};

  switch (typ)
    {
    case T_FREE:            return(frees[article]);
    case T_NIL:             return(nils[article]);
    case T_EOF_OBJECT:      return(eofs[article]);
    case T_UNSPECIFIED:     return(unspecs[article]);
    case T_UNDEFINED:       return(undefs[article]);
    case T_BOOLEAN:         return(booleans[article]);
    case T_STRING:          return(strings[article]);
    case T_SYMBOL:          return(symbols[article]);
    case T_SYNTAX:          return(syntaxes[article]);
    case T_PAIR:            return(pairs[article]);
    case T_GOTO:            return(gotos[article]);
    case T_CONTINUATION:    return(continuations[article]);
    case T_C_OPT_ARGS_FUNCTION:
    case T_C_RST_ARGS_FUNCTION:
    case T_C_ANY_ARGS_FUNCTION:
    case T_C_FUNCTION:      return(c_functions[article]);
    case T_C_FUNCTION_STAR: return(c_function_s[article]);
    case T_CLOSURE:         return(functions[article]);
    case T_CLOSURE_STAR:    return(function_stars[article]);
    case T_C_MACRO:         return(c_macros[article]);
    case T_C_POINTER:       return(c_pointers[article]);
    case T_CHARACTER:       return(characters[article]);
    case T_VECTOR:          return(vectors[article]);
    case T_INT_VECTOR:      return(int_vectors[article]);
    case T_FLOAT_VECTOR:    return(float_vectors[article]);
    case T_MACRO_STAR:
    case T_MACRO:           return(macros[article]);
    case T_BACRO_STAR:
    case T_BACRO:           return(bacros[article]);
    case T_CATCH:           return(catches[article]); /* are these 2 possible? */
    case T_DYNAMIC_WIND:    return(dynamic_winds[article]);
    case T_HASH_TABLE:      return(hash_tables[article]);
    case T_ITERATOR:        return(iterators[article]);
    case T_LET:             return(environments[article]);
    case T_COUNTER:         return(counters[article]);
    case T_OPTLIST:         return(optlists[article]);
    case T_BAFFLE:          return(baffles[article]);
    case T_RANDOM_STATE:    return(rngs[article]);
    case T_SLOT:            return(slots[article]);
    case T_INTEGER:         return(integers[article]);
    case T_RATIO:           return(ratios[article]);
    case T_REAL:            return(reals[article]);
    case T_COMPLEX:         return(complexes[article]);
    case T_BIG_INTEGER:     return(big_integers[article]);
    case T_BIG_RATIO:       return(big_ratios[article]);
    case T_BIG_REAL:        return(big_reals[article]);
    case T_BIG_COMPLEX:     return(big_complexes[article]);
    case T_INPUT_PORT:      return(inputs[article]);
    case T_OUTPUT_PORT:     return(outputs[article]);
    case T_C_OBJECT:        return(c_objects[article]);
    }
  return(NULL);
}


static const char *type_name(s7_scheme *sc, s7_pointer arg, int32_t article)
{
  switch (unchecked_type(arg))
    {
    case T_C_OBJECT:
      return(make_type_name(sc, sc->c_object_types[c_object_type(arg)]->name, article));

    case T_INPUT_PORT:
      return(make_type_name(sc, (is_file_port(arg)) ? "input file port" : ((is_string_port(arg)) ? "input string port" : "input port"), article));

    case T_OUTPUT_PORT:
      return(make_type_name(sc, (is_file_port(arg)) ? "output file port" : ((is_string_port(arg)) ? "output string port" : "output port"), article));

    case T_LET:
      if (has_methods(arg))
	{
	  s7_pointer class_name;
	  class_name = find_method(sc, arg, sc->class_name_symbol);
	  if (is_symbol(class_name))
	    return(make_type_name(sc, symbol_name(class_name), article));
	}

    default:
      {
	const char *str;
	str = type_name_from_type(sc, unchecked_type(arg), article);
	if (str) return(str);
      }
    }
  return("messed up object");
}


static s7_pointer prepackaged_type_name(s7_scheme *sc, s7_pointer x)
{
  s7_pointer p;

  if (has_methods(x))
    {
      p = find_method(sc, find_let(sc, x), sc->class_name_symbol);
      if (is_symbol(p))
	return(symbol_name_cell(p));
    }

  p = prepackaged_type_names[type(x)];
  if (is_string(p)) return(p);

  switch (type(x))
    {
    case T_C_OBJECT:    return(c_object_scheme_name(sc, x));
    case T_INPUT_PORT:  return((is_file_port(x)) ? an_input_file_port_string : ((is_string_port(x)) ? an_input_string_port_string : an_input_port_string));
    case T_OUTPUT_PORT: return((is_file_port(x)) ? an_output_file_port_string : ((is_string_port(x)) ? an_output_string_port_string : an_output_port_string));
    }
  return(s7_make_string_wrapper(sc, "unknown type!"));
}

static s7_pointer type_name_string(s7_scheme *sc, s7_pointer arg)
{
  if (type(arg) < NUM_TYPES)
    {
      s7_pointer p;
      p = prepackaged_type_names[type(arg)]; /* these use INDEFINITE_ARTICLE */
      if (is_string(p)) return(p);
    }
  return(s7_make_string_wrapper(sc, type_name(sc, arg, INDEFINITE_ARTICLE)));
}


static s7_pointer wrong_type_arg_error_prepackaged(s7_scheme *sc, s7_pointer caller, s7_pointer arg_n, s7_pointer arg, s7_pointer typnam, s7_pointer descr)
{
  /* info list is '(format_string caller arg_n arg type_name descr) */
  s7_pointer p;
  p = cdr(sc->wrong_type_arg_info);  /* info list is '(format_string caller arg_n arg type_name descr) */
  set_car(p, caller);  p = cdr(p);
  set_car(p, arg_n);   p = cdr(p);
  set_car(p, arg);     p = cdr(p);
  set_car(p, (typnam == sc->gc_nil) ? prepackaged_type_name(sc, arg) : typnam);
  p = cdr(p);
  set_car(p, descr);
  return(s7_error(sc, sc->wrong_type_arg_symbol, sc->wrong_type_arg_info));
}


static s7_pointer simple_wrong_type_arg_error_prepackaged(s7_scheme *sc, s7_pointer caller, s7_pointer arg, s7_pointer typnam, s7_pointer descr)
{
  set_wlist_4(sc, cdr(sc->simple_wrong_type_arg_info), caller, arg, (typnam == sc->gc_nil) ? prepackaged_type_name(sc, arg) : typnam, descr);
  return(s7_error(sc, sc->wrong_type_arg_symbol, sc->simple_wrong_type_arg_info));
}


s7_pointer s7_wrong_type_arg_error(s7_scheme *sc, const char *caller, int32_t arg_n, s7_pointer arg, const char *descr)
{
  /* info list is '(format_string caller arg_n arg type_name descr) */
  string_value(sc->err_wrap1) = (char *)caller;
  string_length(sc->err_wrap1) = safe_strlen(caller);
  string_value(sc->err_wrap2) = (char *)descr;
  string_length(sc->err_wrap2) = safe_strlen(descr);
  if (arg_n < 0) arg_n = 0;

  if (arg_n > 0)
    return(wrong_type_arg_error_prepackaged(sc, sc->err_wrap1, make_integer(sc, arg_n), arg, type_name_string(sc, arg), sc->err_wrap2));
  return(simple_wrong_type_arg_error_prepackaged(sc, sc->err_wrap1, arg, type_name_string(sc, arg), sc->err_wrap2));
}


static s7_pointer out_of_range_error_prepackaged(s7_scheme *sc, s7_pointer caller, s7_pointer arg_n, s7_pointer arg, s7_pointer descr)
{
  /* info list is '(format_string caller arg_n arg descr) */
  set_wlist_4(sc, cdr(sc->out_of_range_info), caller, arg_n, arg, descr);
  return(s7_error(sc, sc->out_of_range_symbol, sc->out_of_range_info));
}


static s7_pointer simple_out_of_range_error_prepackaged(s7_scheme *sc, s7_pointer caller, s7_pointer arg, s7_pointer descr)
{
  set_wlist_3(sc, cdr(sc->simple_out_of_range_info), caller, arg, descr);
  return(s7_error(sc, sc->out_of_range_symbol, sc->simple_out_of_range_info));
}


s7_pointer s7_out_of_range_error(s7_scheme *sc, const char *caller, int32_t arg_n, s7_pointer arg, const char *descr)
{
  /* info list is '(format_string caller arg_n arg descr) */
  string_value(sc->err_wrap1) = (char *)caller;
  string_length(sc->err_wrap1) = safe_strlen(caller);
  string_value(sc->err_wrap2) = (char *)descr;
  string_length(sc->err_wrap2) = safe_strlen(descr);
  if (arg_n < 0) arg_n = 0;

  if (arg_n > 0)
    return(out_of_range_error_prepackaged(sc, sc->err_wrap1, make_integer(sc, arg_n), arg, sc->err_wrap2));
  return(simple_out_of_range_error_prepackaged(sc, sc->err_wrap1, arg, sc->err_wrap2));
}


s7_pointer s7_wrong_number_of_args_error(s7_scheme *sc, const char *caller, s7_pointer args)
{
  return(s7_error(sc, sc->wrong_number_of_args_symbol,  set_elist_2(sc, s7_make_string_wrapper(sc, caller), args))); /* "caller" includes the format directives */
}


static s7_pointer division_by_zero_error(s7_scheme *sc, s7_pointer caller, s7_pointer arg)
{
  return(s7_error(sc, sc->division_by_zero_symbol, set_elist_3(sc, sc->division_by_zero_error_string, caller, arg)));
}


static s7_pointer file_error(s7_scheme *sc, const char *caller, const char *descr, const char *name)
{
  return(s7_error(sc, sc->io_error_symbol, 
		  set_elist_4(sc, s7_make_string_wrapper(sc, "~A: ~A ~S"), 
				  s7_make_string_wrapper(sc, caller), 
				  s7_make_string_wrapper(sc, descr), 
				  s7_make_string_wrapper(sc, name))));
}


static s7_pointer missing_method_error(s7_scheme *sc, s7_pointer method, s7_pointer obj)
{
  return(s7_error(sc, sc->missing_method_symbol, set_elist_3(sc, sc->missing_method_string, method, obj)));
}


static s7_pointer closure_or_f(s7_scheme *sc, s7_pointer p)
{
  s7_pointer body;
  if (!is_closure(p)) return(p);
  body = closure_body(p);
  if (is_pair(cdr(body))) return(p);
  if (!is_pair(car(body))) return(sc->F);
  if (caar(body) == sc->quote_symbol) return(sc->F);
  return(p);
}


static s7_pointer g_dynamic_wind(s7_scheme *sc, s7_pointer args)
{
  #define H_dynamic_wind "(dynamic-wind init body finish) calls init, then body, then finish, \
each a function of no arguments, guaranteeing that finish is called even if body is exited"
  #define Q_dynamic_wind s7_make_circular_signature(sc, 1, 2, sc->values_symbol, sc->is_procedure_symbol)

  s7_pointer p;

  if (!is_thunk(sc, car(args)))
    method_or_bust_with_type(sc, car(args), sc->dynamic_wind_symbol, args, a_thunk_string, 1);
  if (!is_thunk(sc, cadr(args)))
    method_or_bust_with_type(sc, cadr(args), sc->dynamic_wind_symbol, args, a_thunk_string, 2);
  if (!is_thunk(sc, caddr(args)))
    method_or_bust_with_type(sc, caddr(args), sc->dynamic_wind_symbol, args, a_thunk_string, 3);

  /* this won't work:

       (let ((final (lambda (a b c) (list a b c))))
         (dynamic-wind
           (lambda () #f)
           (lambda () (set! final (lambda () (display "in final"))))
           final))

   * but why not?  'final' is a thunk by the time it is evaluated.
   *   catch (the error handler) is similar.
   *
   * It can't work here because we set up the dynamic_wind_out slot below and
   *   even if the thunk check was removed, we'd still be trying to apply the original function.
   */

  new_cell(sc, p, T_DYNAMIC_WIND);                          /* don't mark car/cdr, don't copy */
  dynamic_wind_in(p) = closure_or_f(sc, car(args));
  dynamic_wind_body(p) = cadr(args);
  dynamic_wind_out(p) = closure_or_f(sc, caddr(args));

  /* since we don't care about the in and out results, and they are thunks, if the body is not a pair,
   *   or is a quoted thing, we just ignore that function.
   */

  push_stack(sc, OP_DYNAMIC_WIND, sc->nil, p);          /* args will be the saved result, code = s7_dynwind_t obj */
  if (dynamic_wind_in(p) != sc->F)
    {
      dynamic_wind_state(p) = DWIND_INIT;
      push_stack(sc, OP_APPLY, sc->nil, dynamic_wind_in(p));
    }
  else
    {
      dynamic_wind_state(p) = DWIND_BODY;
      push_stack(sc, OP_APPLY, sc->nil, dynamic_wind_body(p));
    }
  return(sc->F);
}


s7_pointer s7_dynamic_wind(s7_scheme *sc, s7_pointer init, s7_pointer body, s7_pointer finish)
{
  /* this is essentially s7_call with a dynamic-wind wrapper around "body" */
  s7_pointer p;
  declare_jump_info();

  sc->temp1 = ((init == sc->F) ? finish : init);
  sc->temp2 = body;

  store_jump_info(sc);
  set_jump_info(sc, DYNAMIC_WIND_SET_JUMP); 
  if (jump_loc != NO_JUMP)
    {
      if (jump_loc != ERROR_JUMP)
	eval(sc, sc->cur_op);
    }
  else
    {
      push_stack(sc, OP_EVAL_DONE, sc->args, sc->code);
      sc->args = sc->nil;
      
      new_cell(sc, p, T_DYNAMIC_WIND);
      dynamic_wind_in(p) = _NFre(init);
      dynamic_wind_body(p) = _NFre(body);
      dynamic_wind_out(p) = _NFre(finish);
      push_stack(sc, OP_DYNAMIC_WIND, sc->nil, p);
      if (init != sc->F)
	{
	  dynamic_wind_state(p) = DWIND_INIT;
	  sc->code = init;
	}
      else
	{
	  dynamic_wind_state(p) = DWIND_BODY;
	  sc->code = body;
	}
      eval(sc, OP_APPLY);
    }
  restore_jump_info(sc);
  sc->temp1 = sc->nil;
  sc->temp2 = sc->nil;

  if (is_multiple_value(sc->value))
    sc->value = splice_in_values(sc, multiple_value(sc->value));
 return(sc->value);
}


static s7_pointer g_catch(s7_scheme *sc, s7_pointer args)
{
  #define H_catch "(catch tag thunk handler) evaluates thunk; if an error occurs that matches the tag (#t matches all), the handler is called"
  #define Q_catch s7_make_signature(sc, 4, sc->values_symbol, s7_make_signature(sc, 2, sc->is_symbol_symbol, sc->is_boolean_symbol), sc->is_procedure_symbol, sc->is_procedure_symbol)

  s7_pointer p, proc, err;

  /* Guile sets up the catch before looking for arg errors:
   *   (catch #t log (lambda args "hiho")) -> "hiho"
   * which is consistent in that (catch #t (lambda () (log))...) should probably be the same as (catch #t log ...)
   * but what if the error handler arg is messed up?  Weird to handle args in reverse order with an intervening frame etc.
   */

  proc = cadr(args);
  err = caddr(args);
  /* if (is_let(err)) check_method(sc, err, sc->catch_symbol, args); */ /* causes exit from s7! */

  new_cell(sc, p, T_CATCH);
  catch_tag(p) = car(args);
  catch_goto_loc(p) = s7_stack_top(sc);
  catch_op_loc(p) = (int32_t)(sc->op_stack_now - sc->op_stack);
  catch_handler(p) = err;

  if (is_any_macro(err))
    push_stack(sc, OP_CATCH_2, args, p);
  else push_stack(sc, OP_CATCH, args, p);      /* args ignored but maybe safer for GC? */

  /* not sure about these error checks -- they can be omitted */
  if (!is_thunk(sc, proc))
    return(wrong_type_argument_with_type(sc, sc->catch_symbol, 2, proc, a_thunk_string));

  if (!is_applicable(err))
    return(wrong_type_argument_with_type(sc, sc->catch_symbol, 3, err, something_applicable_string));

  /* should we check here for (aritable? err 2)? -- right now:
   *  (catch #t (lambda () 1) "hiho") -> 1
   * currently this is checked only if the error handler is called
   */

  if (is_closure(proc))                        /* not also lambda* here because we need to handle the arg defaults */
    {
      /* is_thunk above checks is_aritable(proc, 0), but if it's (lambda args ...) we have to set up the frame with args=() 
       *    the case that caught this: (catch #t make-hook ...)
       */
      sc->code = closure_body(proc);
      if (is_symbol(closure_args(proc)))
	new_frame_with_slot(sc, closure_let(proc), sc->envir, closure_args(proc), sc->nil);
      else new_frame(sc, closure_let(proc), sc->envir);
      push_stack_no_args(sc, OP_BEGIN1, _TPair(sc->code));
    }
  else push_stack(sc, OP_APPLY, sc->nil, proc);

  return(sc->F);
}

/* s7_catch(sc, tag, body, error): return(g_catch(sc, list(sc, 3, tag, body, error))) */

/* error reporting info -- save filename and line number */

#define remember_location(Line, File) (((File) << 20) | (Line))
#define remembered_line_number(Line) ((Line) & 0xfffff)
#define remembered_file_name(Line)   ((((Line) >> 20) <= sc->file_names_top) ? sc->file_names[Line >> 20] : sc->F)
/* this gives room for 4000 files each of 1000000 lines */


static int32_t remember_file_name(s7_scheme *sc, const char *file)
{
  int32_t i;

  for (i = 0; i <= sc->file_names_top; i++)
    if (safe_strcmp(file, string_value(sc->file_names[i])))
      return(i);

  sc->file_names_top++;
  if (sc->file_names_top >= sc->file_names_size)
    {
      int32_t old_size = 0;
      if (sc->file_names_size == 0)
	{
	  sc->file_names_size = INITIAL_FILE_NAMES_SIZE;
	  sc->file_names = (s7_pointer *)calloc(sc->file_names_size, sizeof(s7_pointer));
	}
      else
	{
	  old_size = sc->file_names_size;
	  sc->file_names_size *= 2;
	  sc->file_names = (s7_pointer *)realloc(sc->file_names, sc->file_names_size * sizeof(s7_pointer));
	}
      for (i = old_size; i < sc->file_names_size; i++)
	sc->file_names[i] = sc->F;
    }
  sc->file_names[sc->file_names_top] = s7_make_permanent_string(file);

  return(sc->file_names_top);
}


static s7_pointer init_owlet(s7_scheme *sc)
{
  s7_pointer e;
  e = new_frame_in_env(sc, sc->rootlet);
  sc->temp3 = e;
  sc->error_type = make_slot_1(sc, e, make_symbol(sc, "error-type"), sc->F);  /* the error type or tag ('division-by-zero) */
  sc->error_data = make_slot_1(sc, e, make_symbol(sc, "error-data"), sc->F);  /* the message or information passed by the error function */
  sc->error_code = make_slot_1(sc, e, make_symbol(sc, "error-code"), sc->F);  /* the code that s7 thinks triggered the error */
  sc->error_line = make_slot_1(sc, e, make_symbol(sc, "error-line"), sc->F);  /* the line number of that code */
  sc->error_file = make_slot_1(sc, e, make_symbol(sc, "error-file"), sc->F);  /* the file name of that code */
#if WITH_HISTORY
  sc->error_history = make_slot_1(sc, e, make_symbol(sc, "error-history"), sc->F); /* buffer of previous evaluations */
#endif
  sc->temp3 = sc->nil;
  return(e);
}


static s7_pointer g_owlet(s7_scheme *sc, s7_pointer args)
{
#if WITH_HISTORY
  #define H_owlet "(owlet) returns the environment at the point of the last error. \
It has the additional local variables: error-type, error-data, error-code, error-line, error-file, and error-history."
#else
  #define H_owlet "(owlet) returns the environment at the point of the last error. \
It has the additional local variables: error-type, error-data, error-code, error-line, and error-file."
#endif
  #define Q_owlet s7_make_signature(sc, 1, sc->is_let_symbol)
  /* if owlet is not copied, (define e (owlet)), e changes as owlet does!
   */
  s7_pointer e, x;
  uint32_t gc_loc;

  e = let_copy(sc, sc->owlet);
  gc_loc = s7_gc_protect(sc, e);

  /* also make sure the pairs are copied: should be error-data, error-code, and possibly error-history */
  for (x = let_slots(e); is_slot(x); x = next_slot(x))
    if (is_pair(slot_value(x)))
      slot_set_value(x, protected_list_copy(sc, slot_value(x)));

  s7_gc_unprotect_at(sc, gc_loc);
  return(e);
}


static s7_pointer active_catches(s7_scheme *sc)
{
  int64_t i;
  s7_pointer x, lst;
  lst = sc->nil;
  for (i = s7_stack_top(sc) - 1; i >= 3; i -= 4)
    switch (stack_op(sc->stack, i))
      {
      case OP_CATCH_ALL:
	lst = cons(sc, sc->T, lst);
	break;

      case OP_CATCH_2:
      case OP_CATCH_1:
      case OP_CATCH:
	x = stack_code(sc->stack, i);
	lst = cons(sc, catch_tag(x), lst);
	break;
      }
  return(reverse_in_place_unchecked(sc, sc->nil, lst));
}

static s7_pointer active_exits(s7_scheme *sc)
{
  /* (call-with-exit (lambda (exiter) (*s7* 'exits))) */
  int64_t i;
  s7_pointer lst;
  lst = sc->nil;
  for (i = s7_stack_top(sc) - 1; i >= 3; i -= 4)
    if (stack_op(sc->stack, i) == OP_DEACTIVATE_GOTO)
      {
	s7_pointer func, jump;
	func = stack_code(sc->stack, i);  /* presumably this has the goto name */
	jump = stack_args(sc->stack, i);  /* call this to jump */

	if (is_any_closure(func))
	  lst = cons_unchecked(sc, cons(sc, car(closure_args(func)), jump), lst);
	else
	  {
	    if ((is_pair(func)) && (car(func) == sc->call_with_exit_symbol))
	      lst = cons_unchecked(sc, cons(sc, car(cadadr(func)), jump), lst); /* (call-with-exit (lambda (three) ...)) */
	    else lst = cons_unchecked(sc, cons(sc, sc->unspecified, jump), lst);
	  }
	sc->w = lst;
      }
  return(reverse_in_place_unchecked(sc, sc->nil, lst));
}

static s7_pointer stack_entries(s7_scheme *sc, s7_pointer stack, int64_t top)
{
  int64_t i;
  s7_pointer lst;
  lst = sc->nil;
  for (i = top - 1; i >= 3; i -= 4)
    {
      s7_pointer func, args, e;
      opcode_t op;
      func = stack_code(stack, i);
      args = stack_args(stack, i);
      e = stack_let(stack, i);
      op = stack_op(stack, i);
      if ((s7_is_valid(sc, func)) &&
	  (s7_is_valid(sc, args)) &&
	  (s7_is_valid(sc, e)) &&
	  (op < OP_MAX_DEFINED))
	{
#if DEBUGGING
	  if (op < OP_MAX_DEFINED_1)
	    lst = cons(sc, list_4(sc, func, args, e, s7_make_string_wrapper(sc, op_names[op])), lst);
	  else lst = cons(sc, list_4(sc, func, args, e, make_integer(sc, op)), lst);
#else
	  lst = cons(sc, list_4(sc, func, args, e, make_integer(sc, op)), lst);
#endif
	  sc->w = lst;
	}
    }
  return(reverse_in_place_unchecked(sc, sc->nil, lst));
}


/* catch handlers */

typedef bool (*catch_function)(s7_scheme *sc, int32_t i, s7_pointer type, s7_pointer info, bool *reset_hook);
static catch_function catchers[OP_MAX_DEFINED + 1];

/* here and below, don't free the catcher */

static bool catch_all_function(s7_scheme *sc, int32_t i, s7_pointer type, s7_pointer info, bool *reset_hook)
{
  s7_pointer catcher;
  catcher = stack_let(sc->stack, i);
  sc->value = stack_args(sc->stack, i);
  sc->op_stack_now = (s7_pointer *)(sc->op_stack + catch_all_op_loc(catcher));
  sc->stack_end = (s7_pointer *)(sc->stack_start + catch_all_goto_loc(catcher));
  pop_stack(sc);
  if (is_pair(sc->value))
    {
      if (car(sc->value) == sc->quote_symbol)
	sc->value = cadr(sc->value);
      else sc->value = type;
    }
  else
    {
      if (is_symbol(sc->value))
	sc->value = type;
    }
  return(true);
}

static bool catch_2_function(s7_scheme *sc, int32_t i, s7_pointer type, s7_pointer info, bool *reset_hook)
{
  /* this is the macro-error-handler case from g_catch
   *    (let () (define-macro (m . args) (apply (car args) (cadr args))) (catch #t (lambda () (error abs -1)) m))
   */
  s7_pointer x;
  x = stack_code(sc->stack, i);
  if ((catch_tag(x) == sc->T) ||
      (catch_tag(x) == type) ||
      (type == sc->T))
    {
      int64_t loc;
      loc = catch_goto_loc(x);
      sc->op_stack_now = (s7_pointer *)(sc->op_stack + catch_op_loc(x));
      sc->stack_end = (s7_pointer *)(sc->stack_start + loc);
      sc->code = catch_handler(x);

      if (needs_copied_args(sc->code))
	sc->args = list_2(sc, type, info);
      else
	{
	  set_car(sc->t2_1, type);
	  set_car(sc->t2_2, info);
	  sc->args = sc->t2_1;
	}
      sc->cur_op = OP_APPLY;
      return(true);
    }
  return(false);
}

static bool catch_1_function(s7_scheme *sc, int32_t i, s7_pointer type, s7_pointer info, bool *reset_hook)
{
  s7_pointer x;
  x = stack_code(sc->stack, i);
  if ((catch_tag(x) == sc->T) ||
      (catch_tag(x) == type) ||
      (type == sc->T))
    {
      uint64_t loc;
      opcode_t op;
      s7_pointer catcher, error_func, error_body, error_args;

      op = stack_op(sc->stack, i);
      sc->temp4 = stack_let(sc->stack, i); /* GC protect this, since we're moving the stack top below */
      catcher = x;
      loc = catch_goto_loc(catcher);
      sc->op_stack_now = (s7_pointer *)(sc->op_stack + catch_op_loc(catcher));
      sc->stack_end = (s7_pointer *)(sc->stack_start + loc);
      error_func = catch_handler(catcher);

      /* very often the error handler just returns either a constant ('error or #f), or
       *   the args passed to it, so there's no need to laboriously make a closure,
       *   and apply it -- just set sc->value to the closure body (or the args) and
       *   return.
       *
       * so first examine closure_body(error_func)
       *   if it is a constant, or quoted symbol, return that,
       *   if it is the args symbol, return (list type info)
       */

      /* if OP_CATCH_1, we deferred making the error handler until it is actually needed */
      if (op == OP_CATCH_1)
	{
	  error_body = cdr(error_func);
	  error_args = car(error_func);
	}
      else
	{
	  if (is_closure(error_func))
	    {
	      error_body = closure_body(error_func);
	      error_args = closure_args(error_func);
	    }
	  else 
	    {
	      error_body = NULL;
	      error_args = NULL;
	    }
	}

      if ((error_body) && (is_null(cdr(error_body))))
	{
	  s7_pointer y = NULL;
	  error_body = car(error_body);
	  if (is_pair(error_body))
	    {
	      if (car(error_body) == sc->quote_symbol)
		y = cadr(error_body);
	      else
		{
		  if ((car(error_body) == sc->car_symbol) &&
		      (cadr(error_body) == error_args))
		    y = type;
		}
	    }
	  else
	    {
	      if (is_symbol(error_body))
		{
		  if (error_body == error_args)
		    y = list_2(sc, type, info);
		  else
		    {
		      if ((is_pair(error_args)) &&
			  (error_body == car(error_args)))
			y = type;
		    }
		}
	      else y = error_body; /* not pair or symbol */
	    }
	  if (y)
	    {
	      if (loc > 4)
		pop_stack(sc);
	      /* we're at OP_CATCH, normally we want to pop that away, but (handwaving...) if we're coming
	       *   from s7_eval (indirectly perhaps through s7_eval_c_string), we might push the OP_EVAL_DONE
	       *   to end that call, but it's pushed at the precatch stack end (far beyond the catch loc).
	       *   If we catch an error, catch unwinds to its starting point, and the pop_stack above
	       *   puts us at the bottom of the stack (i.e. stack_end == stack_start), OP_EVAL_DONE.
	       *   Now we return true, ending up back in eval, because the error handler jumped out of eval,
	       *   back to wherever we were in eval when we hit the error.  eval jumps back to the start
	       *   of its loop, and pops the stack to see what to do next!  So the (loc > 4) at least
	       *   protects against stack underflow, but ideally we'd know we came from OP_CATCH+s7_eval.
	       *   We can't do anything fancy here because we have to unwind the C stack as well as s7's stack.
	       *   s7_eval doesn't know anything about the catches on the stack.  We can't look back for
	       *   OP_EVAL_DONE -- segfault in OP_BEGIN.  Hmmmm.  Perhaps catch should not unwind until the
	       *   end?  But we want the error handler to run as a part of the calling expression, and
	       *   in any case the OP_EVAL_DONE is not useful (it marks the end of the no-error case).
	       */
	      sc->value = y;
	      sc->temp4 = sc->nil;

	      if (loc == 4)
		sc->code = cons(sc, sc->value, sc->nil); /* if we end up at op_begin, give it something it can handle */

	      return(true);
	    }
	}
      if (op == OP_CATCH_1)
	{
	  s7_pointer x;
	  new_cell(sc, x, T_CLOSURE | T_COPY_ARGS); /* never a safe_closure, apparently */
	  closure_set_args(x, car(error_func));
	  closure_set_body(x, cdr(error_func));
	  closure_set_setter(x, sc->F);
	  closure_arity(x) = CLOSURE_ARITY_NOT_SET;
	  closure_set_let(x, sc->temp4);
	  sc->code = x;
	}
      else sc->code = error_func;
      sc->temp4 = sc->nil;

      /* if user (i.e. yers truly!) copies/pastes the preceding lambda () into the
       *   error handler portion of the catch, he gets the inexplicable message:
       *       ;(): too many arguments: (a1 ())
       *   when this apply tries to call the handler.  So, we need a special case
       *   error check here!
       */

      if (!s7_is_aritable(sc, sc->code, 2))
	{
	  s7_wrong_number_of_args_error(sc, "catch error handler should accept 2 args: ~S", sc->code);
	  return(false);
	}
      
      sc->args = list_2(sc, type, info); /* almost never able to skip this -- costs more to check! */
      sc->cur_op = OP_APPLY;

      /* explicit eval needed if s7_call called into scheme where a caught error occurred (ex6 in exs7.c)
       *  but putting it here (via eval(sc, OP_APPLY)) means the C stack is not cleared correctly in non-s7-call cases,
       *  so defer it until s7_call
       */
      return(true);
    }
  return(false);
}

static bool catch_dw_function(s7_scheme *sc, int32_t i, s7_pointer type, s7_pointer info, bool *reset_hook)
{
  s7_pointer x;
  x = stack_code(sc->stack, i);
  if (dynamic_wind_state(x) == DWIND_BODY)
    {
      dynamic_wind_state(x) = DWIND_FINISH;    /* make sure an uncaught error in the exit thunk doesn't cause us to loop */
      if (dynamic_wind_out(x) != sc->F)
	{
	  push_stack(sc, OP_EVAL_DONE, sc->args, sc->code);
	  sc->code = dynamic_wind_out(x);
	  sc->args = sc->nil;
	  eval(sc, OP_APPLY);                  /* I guess this means no call/cc out of the exit thunk in an error-catching context */
	}
    }
  return(false);
}

static bool catch_out_function(s7_scheme *sc, int32_t i, s7_pointer type, s7_pointer info, bool *reset_hook)
{
  s7_pointer x;
  x = stack_code(sc->stack, i);                /* "code" = port that we opened */
  s7_close_output_port(sc, x);
  x = stack_args(sc->stack, i);                /* "args" = port that we shadowed, if not #f */
  if (x != sc->F)
    sc->output_port = x;
  return(false);
}

static bool catch_in_function(s7_scheme *sc, int32_t i, s7_pointer type, s7_pointer info, bool *reset_hook)
{
  s7_close_input_port(sc, stack_code(sc->stack, i)); /* "code" = port that we opened */
  sc->input_port = stack_args(sc->stack, i);         /* "args" = port that we shadowed */
  return(false);
}

static bool catch_read_function(s7_scheme *sc, int32_t i, s7_pointer type, s7_pointer info, bool *reset_hook)
{
  pop_input_port(sc);
  return(false);
}

static bool catch_eval_function(s7_scheme *sc, int32_t i, s7_pointer type, s7_pointer info, bool *reset_hook)
{
  s7_close_input_port(sc, sc->input_port);
  pop_input_port(sc);
  return(false);
}

static bool catch_barrier_function(s7_scheme *sc, int32_t i, s7_pointer type, s7_pointer info, bool *reset_hook)
{
  if (is_input_port(stack_args(sc->stack, i)))      /* (eval-string "'(1 .)") */
    {
      if (sc->input_port == stack_args(sc->stack, i))
	pop_input_port(sc);
      s7_close_input_port(sc, stack_args(sc->stack, i));
    }
  return(false);
}

static bool catch_hook_function(s7_scheme *sc, int32_t i, s7_pointer type, s7_pointer info, bool *reset_hook)
{
  sc->error_hook = stack_code(sc->stack, i);
  /* apparently there was an error during *error-hook* evaluation, but Rick wants the hook re-established anyway */
  (*reset_hook) = true;
  /* avoid infinite loop -- don't try to (re-)evaluate (buggy) *error-hook*! */
  return(false);
}

static bool catch_goto_function(s7_scheme *sc, int32_t i, s7_pointer type, s7_pointer info, bool *reset_hook)
{
  call_exit_active(stack_args(sc->stack, i)) = false;
  return(false);
}

static void init_catchers(void)
{
  int32_t i;
  for (i = 0; i <= OP_MAX_DEFINED; i++) catchers[i] = NULL;
  catchers[OP_CATCH_ALL] =           catch_all_function;
  catchers[OP_CATCH_2] =             catch_2_function;
  catchers[OP_CATCH_1] =             catch_1_function;
  catchers[OP_CATCH] =               catch_1_function;
  catchers[OP_DYNAMIC_WIND] =        catch_dw_function;
  catchers[OP_GET_OUTPUT_STRING_1] = catch_out_function;
  catchers[OP_UNWIND_OUTPUT] =       catch_out_function;
  catchers[OP_UNWIND_INPUT] =        catch_in_function;
  catchers[OP_READ_DONE] =           catch_read_function;      /* perhaps an error during (read) */
  catchers[OP_EVAL_STRING_1] =       catch_eval_function;      /* perhaps an error happened before we could push the OP_EVAL_STRING_2 */
  catchers[OP_EVAL_STRING_2] =       catch_eval_function;
  catchers[OP_BARRIER] =             catch_barrier_function;
  catchers[OP_DEACTIVATE_GOTO] =     catch_goto_function;
  catchers[OP_ERROR_HOOK_QUIT] =     catch_hook_function;
}

static s7_pointer g_throw(s7_scheme *sc, s7_pointer args)
{
  #define H_throw "(throw tag . info) is like (error ...) but it does not affect the owlet. \
It looks for an existing catch with a matching tag, and jumps to it if found.  Otherwise it raises an error."
  #define Q_throw s7_make_circular_signature(sc, 2, 3, sc->values_symbol, sc->is_symbol_symbol, sc->T)

  bool ignored_flag = false;
  int64_t i;
  s7_pointer type, info;

  type = car(args);
  info = cdr(args);
  /* look for a catcher */

  for (i = s7_stack_top(sc) - 1; i >= 3; i -= 4)
    {
      catch_function catcher;
      catcher = catchers[stack_op(sc->stack, i)];
      if ((catcher) &&
	  (catcher(sc, i, type, info, &ignored_flag)))
	{
	  if (sc->longjmp_ok) longjmp(sc->goto_start, THROW_JUMP);
	  return(sc->value);
	}
    }
  if (is_let(car(args))) 
    check_method(sc, car(args), sc->throw_symbol, args);
  return(s7_error(sc, make_symbol(sc, "uncaught-throw"), 
		  set_elist_3(sc, s7_make_string_wrapper(sc, "no catch found for (throw ~W~{~^ ~S~})"), type, info)));
}


static void s7_warn(s7_scheme *sc, int32_t len, const char *ctrl, ...)
{
  va_list ap;
  char *str;

  str = (char *)malloc(len * sizeof(char));
  va_start(ap, ctrl);
  len = vsnprintf(str, len, ctrl, ap);
  va_end(ap);

  if (port_is_closed(sc->error_port))
    sc->error_port = sc->standard_error;
  s7_display(sc, make_string_uncopied_with_length(sc, str, len), sc->error_port);
}


s7_pointer s7_error(s7_scheme *sc, s7_pointer type, s7_pointer info)
{
  bool reset_error_hook = false;
  s7_pointer cur_code;

  /* type is a symbol normally, and info is compatible with format: (apply format #f info) --
   *    car(info) is the control string, cdr(info) its args
   *    type/range errors have cadr(info)=caller, caddr(info)=offending arg number
   *    null info can mean symbol table is locked so make-symbol uses s7_error to get out
   *
   * set up (owlet), look for a catch that matches 'type', if found
   *   call its error-handler, else if *error-hook* is bound, call it,
   *   else send out the error info ourselves.
   */
  sc->no_values = 0;
  sc->format_depth = -1;
  sc->gc_off = false;            /* this is in case we were triggered from the sort function -- clumsy! */

  if (sc->current_safe_list > 0) 
    {
      clear_list_in_use(sc->safe_lists[sc->current_safe_list]);
      sc->current_safe_list = 0;
    }
  slot_set_value(sc->error_type, type);
  slot_set_value(sc->error_data, info);

  if ((unchecked_type(sc->envir) != T_LET) &&
      (sc->envir != sc->nil))
    sc->envir = sc->nil;          /* in reader, the envir frame is mostly ignored so it can be (and usually is) garbage */

  set_outlet(sc->owlet, sc->envir);

  cur_code = current_code(sc);
  slot_set_value(sc->error_code, cur_code);
#if WITH_HISTORY
  slot_set_value(sc->error_history, sc->cur_code);
  if (sc->using_history1)
    sc->cur_code = sc->eval_history2;
  else sc->cur_code = sc->eval_history1;
  sc->using_history1 = (!sc->using_history1);
#endif

  if ((is_pair(cur_code)) &&     /* can be () if unexpected close paren read error */
      (has_line_number(cur_code)))
    {
      static int32_t last_line = -1;
      int32_t line;
      line = (int32_t)pair_line(cur_code); /* cast to int32_t (from uint32_t) for last_line */
      if (line != last_line)
	{
	  last_line = line;
	  if (line > 0)
	    {
	      slot_set_value(sc->error_line, make_integer(sc, remembered_line_number(line)));
	      slot_set_value(sc->error_file, remembered_file_name(line));
	    }
	  else
	    {
	      if (in_reader(sc))
		{
		  slot_set_value(sc->error_line, make_integer(sc, port_line_number(sc->input_port)));
		  slot_set_value(sc->error_file, make_string_wrapper_with_length(sc, port_filename(sc->input_port), port_filename_length(sc->input_port)));
		}
	      else
		{
		  slot_set_value(sc->error_line, sc->F);
		  slot_set_value(sc->error_file, sc->F);
		}
	    }
	}
    }
  else
    {
      if (in_reader(sc))
	{
	  slot_set_value(sc->error_line, make_integer(sc, port_line_number(sc->input_port)));
	  slot_set_value(sc->error_file, make_string_wrapper_with_length(sc, port_filename(sc->input_port), port_filename_length(sc->input_port)));
	}
      else 
	{
	  slot_set_value(sc->error_line, sc->F);
	  slot_set_value(sc->error_file, sc->F);
	}
    }

  { /* look for a catcher */
    int64_t i;
    /* top is 1 past actual top, top - 1 is op, if op = OP_CATCH, top - 4 is the cell containing the catch struct */
    for (i = s7_stack_top(sc) - 1; i >= 3; i -= 4)
      {
	catch_function catcher;
	catcher = catchers[stack_op(sc->stack, i)];
	if ((catcher) &&
	    (catcher(sc, i, type, info, &reset_error_hook)))
	  {
	    if (sc->longjmp_ok) longjmp(sc->goto_start, CATCH_JUMP);
	    /* all the rest of the code expects s7_error to jump, not return, 
	     *   so presumably if we get here, we're in trouble -- try to send out an error message
	     */
	    /* return(type); */
	  }
      }
  }

  /* error not caught */
  /* (set! *error-hook* (list (lambda (hook) (apply format #t (hook 'args))))) */

  if ((!reset_error_hook) &&
      (is_procedure(sc->error_hook)) &&
      (hook_has_functions(sc->error_hook)))
    {
      s7_pointer error_hook_func;
      /* (set! (hook-functions *error-hook*) (list (lambda (h) (format *stderr* "got error ~A~%" (h 'args))))) */

      error_hook_func = sc->error_hook;
      sc->error_hook = sc->F;
      /* if the *error-hook* functions trigger an error, we had better not have *error-hook* still set! */

      push_stack(sc, OP_ERROR_HOOK_QUIT, sc->nil, error_hook_func); /* restore *error-hook* upon successful (or any!) evaluation */
      sc->code = error_hook_func;
      sc->args = list_2(sc, type, info);

      /* if we drop into the longjmp below, the hook functions are not called!
       *   OP_ERROR_HOOK_QUIT performs the longjmp, so it should be safe to go to eval.
       */
      eval(sc, OP_APPLY);
    }
  else
    {
      if (port_is_closed(sc->error_port))
	sc->error_port = sc->standard_error;
      /* if info is not a list, send object->string to current error port,
       *   else assume car(info) is a format control string, and cdr(info) are its args
       *
       * if at all possible, get some indication of where we are!
       */
      if ((!s7_is_list(sc, info)) ||
	  (!is_string(car(info))))
	format_to_port(sc, sc->error_port, "\n;~S ~S", set_plist_2(sc, type, info), NULL, false, 7);
      else
	{
	  /* it's possible that the error string is just a string -- not intended for format */
	  if ((type != sc->format_error_symbol) &&      /* avoid an infinite loop of format errors */
	      (strchr(string_value(car(info)), '~')))
	    {
	      char *errstr;
	      int32_t len, str_len;
	      len = string_length(car(info)) + 8;
	      tmpbuf_malloc(errstr, len);
	      str_len = snprintf(errstr, len, "\n;%s", string_value(car(info)));
	      format_to_port(sc, sc->error_port, errstr, cdr(info), NULL, false, str_len);
	      tmpbuf_free(errstr, len);
	    }
	  else format_to_port(sc, sc->error_port, "\n;~S ~S", set_plist_2(sc, type, info), NULL, false, 7);
	}

      /* now display location at end */

      if ((is_input_port(sc->input_port)) &&
	  (port_file(sc->input_port) != stdin) &&
	  (!port_is_closed(sc->input_port)))
	{
	  const char *filename = NULL;
	  int32_t line;

	  filename = port_filename(sc->input_port);
	  line = port_line_number(sc->input_port);

	  if (filename)
	    format_to_port(sc, sc->error_port, "\n;  ~A[~D]", set_plist_2(sc, s7_make_string_wrapper(sc, filename), make_integer(sc, line)), NULL, false, 10);
	  else
	    {
	      if ((line > 0) &&
		  (slot_value(sc->error_line) != sc->F))
		format_to_port(sc, sc->error_port, "\n;  line ~D", set_plist_1(sc, make_integer(sc, line)), NULL, false, 11);
	      else
		{
		  if (is_pair(sc->input_port_stack))
		    {
		      s7_pointer p;
		      p = car(sc->input_port_stack);
		      if ((is_input_port(p)) &&
			  (port_file(p) != stdin) &&
			  (!port_is_closed(p)))
			{
			  filename = port_filename(p);
			  line = port_line_number(p);
			  if (filename)
			    format_to_port(sc, sc->error_port, "\n;  ~A[~D]", 
					   set_plist_2(sc, s7_make_string_wrapper(sc, filename), make_integer(sc, line)), NULL, false, 10);
			}
		    }
		}
	    }
	}
      else
	{
	  const char *call_name;
	  call_name = sc->s7_call_name;

	  /* sc->s7_call_name = NULL; */
	  if (call_name)
	    {
	      sc->s7_call_name = NULL;
	      if ((sc->s7_call_file) &&
		  (sc->s7_call_line >= 0))
		{
		  format_to_port(sc, sc->error_port, "\n;  ~A ~A[~D]",
				 set_plist_3(sc,
					     s7_make_string_wrapper(sc, call_name),
					     s7_make_string_wrapper(sc, sc->s7_call_file),
					     make_integer(sc, sc->s7_call_line)),
				 NULL, false, 13);
		}
	    }
	}
      s7_newline(sc, sc->error_port);

      if (is_string(slot_value(sc->error_file)))
	{
	  format_to_port(sc, sc->error_port, ";    ~S, line ~D",
			 set_plist_2(sc, slot_value(sc->error_file),	slot_value(sc->error_line)),
			 NULL, false, 16);
	  s7_newline(sc, sc->error_port);
	}

      /* look for __func__ in the error environment etc */
      if (sc->error_port != sc->F)
	{
	  char *errstr;
	  errstr = stacktrace_1(sc,
				s7_integer(car(sc->stacktrace_defaults)),
				s7_integer(cadr(sc->stacktrace_defaults)),
				s7_integer(caddr(sc->stacktrace_defaults)),
				s7_integer(cadddr(sc->stacktrace_defaults)),
				s7_boolean(sc, s7_list_ref(sc, sc->stacktrace_defaults, 4)));
	  if (errstr)
	    {
	      port_write_string(sc->error_port)(sc, ";\n", 2, sc->error_port);
	      port_write_string(sc->error_port)(sc, errstr, strlen(errstr), sc->error_port);
	      free(errstr);
	      port_write_character(sc->error_port)(sc, '\n', sc->error_port);
	    }
	}
      else
	{
	  if (is_pair(slot_value(sc->error_code)))
	    {
	      format_to_port(sc, sc->error_port, ";    ~S", set_plist_1(sc, slot_value(sc->error_code)), NULL, false, 7);
	      s7_newline(sc, sc->error_port);
	    }
	}

      /* if (is_continuation(type))
       *   go into repl here with access to continuation?  Or expect *error-handler* to deal with it?
       */
      sc->value = type;
      /* stack_reset(sc); */
      sc->cur_op = OP_ERROR_QUIT;
    }

  if (sc->longjmp_ok) longjmp(sc->goto_start, ERROR_JUMP);
  return(type);
}


static s7_pointer apply_error(s7_scheme *sc, s7_pointer obj, s7_pointer args)
{
  /* the operator type is needed here else the error message is confusing:
   *    (apply '+ (list 1 2))) -> ;attempt to apply + to (1 2)?
   */
  static s7_pointer errstr = NULL;
  if (is_null(obj))
    return(s7_error(sc, sc->syntax_error_symbol, set_elist_2(sc, make_string_wrapper_with_length(sc, "attempt to apply nil to ~S?", 27), args)));
  if (!errstr)
    errstr = s7_make_permanent_string("attempt to apply ~A ~S to ~S?");
  return(s7_error(sc, sc->syntax_error_symbol, set_elist_4(sc, errstr, type_name_string(sc, obj), obj, args)));
}


static s7_pointer read_error_1(s7_scheme *sc, const char *errmsg, bool string_error)
{
  /* reader errors happen before the evaluator gets involved, so forms such as:
   *   (catch #t (lambda () (car '( . ))) (lambda arg 'error))
   * do not catch the error if we simply signal an error when we encounter it.
   */
  char *msg;
  int32_t len;
  s7_pointer pt;

  pt = sc->input_port;
  if (!string_error)
    {
      /* make an heroic effort to find where we slid off the tracks */

      if (is_string_port(sc->input_port))
	{
          #define QUOTE_SIZE 40
	  uint32_t i, j, start = 0, end, slen, size;
	  char *recent_input = NULL;

	  /* we can run off the end in cases like (eval-string "(. . ,.)") or (eval-string " (@ . ,.)") */
	  if (port_position(pt) >= port_data_size(pt))
	    port_position(pt) = port_data_size(pt) - 1;

	  /* start at current position and look back a few chars */
	  for (i = port_position(pt), j = 0; (i > 0) && (j < QUOTE_SIZE); i--, j++)
	    if ((port_data(pt)[i] == '\0') ||
		(port_data(pt)[i] == '\n') ||
		(port_data(pt)[i] == '\r'))
	      break;
	  start = i;

	  /* start at current position and look ahead a few chars */
	  size = port_data_size(pt);
	  for (i = port_position(pt), j = 0; (i < size) && (j < QUOTE_SIZE); i++, j++)
	    if ((port_data(pt)[i] == '\0') ||
		(port_data(pt)[i] == '\n') ||
		(port_data(pt)[i] == '\r'))
	      break;

	  end = i;
	  slen = end - start;
	  /* hopefully this is more or less the current line where the read error happened */

	  if (slen > 0)
	    {
	      recent_input = (char *)calloc((slen + 9), sizeof(char));
	      for (i = 0; i < (slen + 8); i++) recent_input[i] = '.';
	      recent_input[3] = ' ';
	      recent_input[slen + 4] = ' ';
	      for (i = 0; i < slen; i++) recent_input[i + 4] = port_data(pt)[start + i];
	    }

	  if ((port_line_number(pt) > 0) &&
	      (port_filename(pt)))
	    {
	      len = safe_strlen(recent_input) + safe_strlen(errmsg) + port_filename_length(pt) + safe_strlen(sc->current_file) + 64;
	      msg = (char *)malloc(len * sizeof(char));
	      len = snprintf(msg, len, "%s: %s %s[%u], last top-level form at: %s[%d]",
			     errmsg, (recent_input) ? recent_input : "", port_filename(pt), port_line_number(pt),
			     sc->current_file, sc->current_line);
	    }
	  else
	    {
	      len = safe_strlen(recent_input) + safe_strlen(errmsg) + safe_strlen(sc->current_file) + 64;
	      msg = (char *)malloc(len * sizeof(char));

	      if ((sc->current_file) &&
		  (sc->current_line >= 0))
		len = snprintf(msg, len, "%s: %s, last top-level form at %s[%d]",
			       errmsg, (recent_input) ? recent_input : "",
			       sc->current_file, sc->current_line);
	      else len = snprintf(msg, len, "%s: %s", errmsg, (recent_input) ? recent_input : "");
	    }

	  if (recent_input) free(recent_input);
	  return(s7_error(sc, sc->read_error_symbol, set_elist_1(sc, make_string_uncopied_with_length(sc, msg, len))));
	}
    }

  if ((port_line_number(pt) > 0) &&
      (port_filename(pt)))
    {
      len = safe_strlen(errmsg) + port_filename_length(pt) + safe_strlen(sc->current_file) + 128;
      msg = (char *)malloc(len * sizeof(char));

      if (string_error)
	len = snprintf(msg, len, "%s %s[%u],\n;  possible culprit: \"%s...\"\n;  last top-level form at %s[%d]",
		       errmsg, port_filename(pt), port_line_number(pt),
		       sc->strbuf, sc->current_file, sc->current_line);
      else len = snprintf(msg, len, "%s %s[%u], last top-level form at %s[%d]",
			  errmsg, port_filename(pt), port_line_number(pt),
			  sc->current_file, sc->current_line);
      return(s7_error(sc, sc->read_error_symbol, set_elist_1(sc, make_string_uncopied_with_length(sc, msg, len))));
    }

  return(s7_error(sc, (string_error) ? sc->string_read_error_symbol : sc->read_error_symbol, set_elist_1(sc, s7_make_string_wrapper(sc, (char *)errmsg))));
}

static s7_pointer read_error(s7_scheme *sc, const char *errmsg)
{
  return(read_error_1(sc, errmsg, false));
}

static s7_pointer string_read_error(s7_scheme *sc, const char *errmsg)
{
  return(read_error_1(sc, errmsg, true));
}


static s7_pointer g_error(s7_scheme *sc, s7_pointer args)
{
  #define H_error "(error type ...) signals an error.  The 'type' can be used with catch to trap \
particular errors.  If the error is not caught, s7 treats the second argument as a format control string, \
and applies it to the rest of the arguments."
  #define Q_error s7_make_circular_signature(sc, 2, 3, sc->values_symbol, sc->is_symbol_symbol, sc->T)

  if (is_not_null(args))
    {
      if (is_string(car(args)))                     /* CL-style error? -- use tag = 'no-catch */
	{
	  s7_error(sc, sc->no_catch_symbol, args);  /* this can have trailing args (implicit format) */
	  return(sc->unspecified);
	}
      return(s7_error(sc, car(args), cdr(args)));
    }
  return(s7_error(sc, sc->nil, sc->nil));
}


static char *truncate_string(char *form, int32_t len, use_write_t use_write, int32_t *form_len)
{
  unsigned char *f;
  f = (unsigned char *)form;

  if (use_write != USE_DISPLAY)
    {
      /* I guess we need to protect the outer double quotes in this case */
      int32_t i;
      for (i = len - 5; i >= (len / 2); i--)
	if (is_white_space((int32_t)f[i]))
	  {
	    form[i] = '.';
	    form[i + 1] = '.';
	    form[i + 2] = '.';
	    form[i + 3] = '"';
	    form[i + 4] = '\0';
	    (*form_len) = i + 4;
	    return(form);
	  }
      i = len - 5;
      if (i > 0)
	{
	  form[i] = '.';
	  form[i + 1] = '.';
	  form[i + 2] = '.';
	  form[i + 3] = '"';
	  form[i + 4] = '\0';
	}
      else
	{
	  if (len >= 2)
	    {
	      form[len - 1] = '"';
	      form[len] = '\0';
	    }
	}
    }
  else
    {
      int32_t i;
      for (i = len - 4; i >= (len / 2); i--)
	if (is_white_space((int32_t)f[i]))
	  {
	    form[i] = '.';
	    form[i + 1] = '.';
	    form[i + 2] = '.';
	    form[i + 3] = '\0';
	    (*form_len) = i + 3;
	    return(form);
	  }
      i = len - 4;
      if (i >= 0)
	{
	  form[i] = '.';
	  form[i + 1] = '.';
	  form[i + 2] = '.';
	  form[i + 3] = '\0';
	}
      else form[len] = '\0';
    }
  return(form);
}


static char *object_to_truncated_string(s7_scheme *sc, s7_pointer p, int32_t len)
{
  char *s;
  int32_t s_len;
  sc->objstr_max_len = len + 2;
  s = s7_object_to_c_string(sc, p);
  sc->objstr_max_len = s7_int_max;
  s_len = safe_strlen(s);
  if (s_len > len)
    return(truncate_string(s, len, USE_DISPLAY, &s_len));
  return(s);
}


static s7_pointer tree_descend(s7_scheme *sc, s7_pointer p, uint32_t line)
{
  s7_pointer tp;
  if (!is_pair(p)) return(NULL);
  if (has_line_number(p))
    {
      uint32_t x;
      x = (uint32_t)remembered_line_number(pair_line(p));
      if (x > 0)
	{
	  if (line == 0) /* first line number we encounter will be the current reader location (i.e. the end of the form) */
	    line = x;
	  else
	    {
	      if (x < line)
		return(p);
	    }
	}
    }
  tp = tree_descend(sc, car(p), line);
  if (tp) return(tp);
  return(tree_descend(sc, cdr(p), line));
}

static char *current_input_string(s7_scheme *sc, s7_pointer pt)
{
  /* try to show the current input */
  if ((is_input_port(pt)) &&
      (!port_is_closed(pt)) &&
      (port_data(pt)) &&
      (port_position(pt) > 0))
    {
      const unsigned char *str;
      char *msg;
      int32_t i, j, start, pos;
      pos = (int32_t)port_position(pt);
      start = pos - 40;
      if (start < 0) start = 0;
      msg = (char *)malloc(64 * sizeof(char));
      str = (const unsigned char *)port_data(pt);
      for (i = start, j = 0; i < pos; i++, j++)
	msg[j] = str[i];
      msg[j] = '\0';
      return(msg);
    }
  return(NULL);
}


static s7_pointer missing_close_paren_error(s7_scheme *sc)
{
  int32_t len;
  char *msg, *syntax_msg = NULL;
  s7_pointer pt;

  if ((unchecked_type(sc->envir) != T_LET) &&
      (sc->envir != sc->nil))
    sc->envir = sc->nil;

  pt = sc->input_port;

  /* check *missing-close-paren-hook* */
  if (hook_has_functions(sc->missing_close_paren_hook))
    {
      s7_pointer result;
      if ((port_line_number(pt) > 0) &&
	  (port_filename(pt)))
	{
	  slot_set_value(sc->error_line, make_integer(sc, port_line_number(pt)));
	  slot_set_value(sc->error_file, s7_make_string_wrapper(sc, port_filename(pt)));
	}
      result = s7_call(sc, sc->missing_close_paren_hook, sc->nil);
      if (result != sc->unspecified)
	return(g_throw(sc, list_1(sc, result)));
    }

  if (is_pair(sc->args))
    {
      s7_pointer p;
      p = tree_descend(sc, sc->args, 0);
      if ((p) && (is_pair(p)) &&
	  (has_line_number(p)))
	{
	  int32_t msg_len, form_len;
	  char *form;
	  form = object_to_truncated_string(sc, p, 40);
	  form_len = safe_strlen(form);
	  msg_len = form_len + 128;
	  syntax_msg = (char *)malloc(msg_len * sizeof(char));
	  snprintf(syntax_msg, msg_len, ";  current form awaiting a close paren starts around line %u: %s", remembered_line_number(pair_line(p)), form);
	  free(form);
	}
    }

  if ((port_line_number(pt) > 0) &&
      (port_filename(pt)))
    {
      len = port_filename_length(pt) + safe_strlen(sc->current_file) + safe_strlen(syntax_msg) + 128;
      msg = (char *)malloc(len * sizeof(char));
      if (syntax_msg)
	{
	  len = snprintf(msg, len, "missing close paren, %s[%u], last top-level form at %s[%d]\n%s",
			 port_filename(pt), port_line_number(pt),
			 sc->current_file, sc->current_line, syntax_msg);
	  free(syntax_msg);
	}
      else len = snprintf(msg, len, "missing close paren, %s[%u], last top-level form at %s[%d]",
			  port_filename(pt), port_line_number(pt),
			  sc->current_file, sc->current_line);
      return(s7_error(sc, sc->read_error_symbol, set_elist_1(sc, make_string_uncopied_with_length(sc, msg, len))));
    }

  if (syntax_msg)
    {
      len = safe_strlen(syntax_msg) + 128;
      msg = (char *)malloc(len * sizeof(char));
      len = snprintf(msg, len, "missing close paren\n%s\n", syntax_msg);
      free(syntax_msg);
      return(s7_error(sc, sc->read_error_symbol, set_elist_1(sc, make_string_uncopied_with_length(sc, msg, len))));
    }

  {
    char *str;
    msg = (char *)malloc(128 * sizeof(char));
    str = current_input_string(sc, pt);
    len = snprintf(msg, 128, "missing close paren: %s", str);
    free(str);
    return(s7_error(sc, sc->read_error_symbol, set_elist_1(sc, make_string_uncopied_with_length(sc, msg, len))));
  }

  return(s7_error(sc, sc->read_error_symbol, set_elist_1(sc, s7_make_string_wrapper(sc, "missing close paren"))));
}


static void improper_arglist_error(s7_scheme *sc)
{
  /* sc->code is the last (dotted) arg, sc->args is the arglist reversed not including sc->code
   *   the original was `(,@(reverse args) . ,code) essentially
   */
  if (sc->args == sc->nil)               /* (abs . 1) */
    s7_error(sc, sc->syntax_error_symbol, set_elist_1(sc, s7_make_string_wrapper(sc, "function call is a dotted list?")));
  else s7_error(sc, sc->syntax_error_symbol,
		set_elist_2(sc, s7_make_string_wrapper(sc, "improper list of arguments: ~S"),
				append_in_place(sc, sc->args = safe_reverse_in_place(sc, sc->args), sc->code)));
}



/* -------------------------------- leftovers -------------------------------- */


void (*s7_begin_hook(s7_scheme *sc))(s7_scheme *sc, bool *val)
{
  return(sc->begin_hook);
}


void s7_set_begin_hook(s7_scheme *sc, void (*hook)(s7_scheme *sc, bool *val))
{
  sc->begin_hook = hook;
}


static bool call_begin_hook(s7_scheme *sc)
{
  bool result = false;
  /* originally begin_hook was bool (*hook)(s7_scheme *sc): the value was returned directly,
   *   rather than going through a *bool arg (&result below).  That works in gcc (Linux/OSX),
   *   but does not work in MS Visual C++.  In the latter, the compiler apparently completely
   *   eliminates any local, returning (for example) a thread-relative stack-allocated value
   *   directly, but then by the time we get here, that variable has vanished, and we get
   *   garbage.  We had to thwart the optimization by adding if ((flag) && (!flag)) fprintf(...);
   *   So, in the new form (26-Jun-13), the value is passed directly into an s7 variable
   *   that I hope can't be optimized out of existence.
   */
  opcode_t op;
  op = sc->cur_op;

  push_stack(sc, OP_BARRIER, sc->args, sc->code);
  sc->begin_hook(sc, &result);
  if (result)
    {
      /* set (owlet) in case we were interrupted and need to see why something was hung */
      slot_set_value(sc->error_type, sc->F);
      slot_set_value(sc->error_data, sc->value); /* was sc->F but we now clobber this below */
      slot_set_value(sc->error_code, current_code(sc));
      slot_set_value(sc->error_line, sc->F);
      slot_set_value(sc->error_file, sc->F);
#if WITH_HISTORY
      slot_set_value(sc->error_history, sc->F);
#endif
      set_outlet(sc->owlet, sc->envir);

      sc->value = s7_make_symbol(sc, "begin-hook-interrupt");
      /* otherwise the evaluator returns whatever random thing is in sc->value (normally #<closure>)
       *   which makes debugging unnecessarily difficult.
       */
      s7_quit(sc);     /* don't call gc here -- perhaps at restart somehow? */
      return(true);
    }
  pop_stack_no_op(sc);
  sc->cur_op = op;         /* for better error handling.  otherwise we get "barrier" as the offending function name in eval_error */
  return(false);
}

static s7_pointer apply_list_star(s7_scheme *sc, s7_pointer d) 
{
  s7_pointer p, q;
  /* we check this ahead of time: if (is_null(cdr(d))) return(car(d)); */
  p = cons(sc, car(d), cdr(d));
  q = p;
  while (is_not_null(cdr(cdr(p))))
    {
      d = cdr(d);
      set_cdr(p, cons(sc, car(d), cdr(d)));
      if (is_not_null(cdr(d)))
	p = cdr(p);
    }
  set_cdr(p, cadr(p));
  return(q);
}

static s7_pointer apply_list_error(s7_scheme *sc, s7_pointer lst)
{
  return(s7_error(sc, sc->wrong_type_arg_symbol, set_elist_2(sc, s7_make_string_wrapper(sc, "apply's last argument should be a proper list: ~S"), lst)));
}


static s7_pointer g_apply(s7_scheme *sc, s7_pointer args)
{
  #define H_apply "(apply func ...) applies func to the rest of the arguments"
  #define Q_apply s7_make_circular_signature(sc, 2, 3, sc->values_symbol, s7_make_signature(sc, 2, sc->is_procedure_symbol, sc->is_sequence_symbol), sc->T)

  /* can apply always be replaced with apply values?
   *   (apply + '(1 2 3)) is the same as (+ (apply values '(1 2 3)))
   * not if apply* in disguise, I think:
   *   (apply + 1 2 ()) -> 3
   *   (apply + 1 2 (apply values ())) -> error
   */
  sc->code = car(args);
  if (is_null(cdr(args)))
    {
      sc->args = sc->nil;
      push_stack(sc, OP_APPLY, sc->args, sc->code);
      return(sc->nil);
    }

  if (is_safe_procedure(sc->code))
    {
      s7_pointer p, q;
      
      for (q = args, p = cdr(args); is_not_null(cdr(p)); q = p, p = cdr(p));
      /* the last arg is supposed to be a list, it will be spliced onto the end of the previous arg list (if any) below */
      
      if (!s7_is_proper_list(sc, car(p)))        /* (apply + #f) etc */
	return(apply_list_error(sc, args));
      set_cdr(q, car(p));
      /* this would work: if (is_c_function(sc->code)) return(c_function_call(sc->code)(sc, cdr(args)));
       *   but it omits the arg number check, but if we copy the APPLY table here (returning sc->value)
       *   the overhead from the now non-inline function calls is greater than the fewer-eval-jumps savings.
       */
      push_stack(sc, OP_APPLY, cdr(args), sc->code);
      return(sc->nil);
    }

  /* here we may have to copy the arg list */
  if (is_null(cddr(args)))
    sc->args = cadr(args);
  else sc->args = apply_list_star(sc, cdr(args));
  if (!s7_is_proper_list(sc, sc->args)) 
    return(apply_list_error(sc, args));
  
  push_stack(sc, OP_APPLY, (needs_copied_args(sc->code)) ? copy_list(sc, sc->args) : sc->args, sc->code);
  return(sc->nil);
}


s7_pointer s7_apply_function(s7_scheme *sc, s7_pointer fnc, s7_pointer args)
{
  TRACK(sc);
#if DEBUGGING
  {
    s7_pointer p;
    int32_t argnum;
    _NFre(fnc);
    for (argnum = 0, p = _NFre(args); is_pair(p); argnum++, p = _NFre(cdr(p)))
      _NFre(car(p));
  }
#endif
  if (sc->safety > NO_SAFETY)
    set_current_code(sc, cons(sc, fnc, args));
  else set_current_code(sc, fnc);

  if (is_c_function(fnc))
    return(c_function_call(fnc)(sc, args));

  push_stack(sc, OP_EVAL_DONE, sc->args, sc->code);
  sc->code = fnc;
  sc->args = (needs_copied_args(sc->code)) ? copy_list(sc, args) : args;
  eval(sc, OP_APPLY);
  /* we're limited in choices here -- the caller might be (say) car(sc->t1_1) = c_call(...) where the c_call
   *   happens to fallback on a method -- we can't just push OP_APPLY and drop back into the evaluator normally.
   */
  return(sc->value);
}


static int32_t next_tx(s7_scheme *sc)
{
  sc->t_temp_ctr++;
  if (sc->t_temp_ctr >= T_TEMPS_SIZE) sc->t_temp_ctr = 0;
  return(sc->t_temp_ctr);
}

static s7_pointer set_c_function_star_args(s7_scheme *sc)
{
  int32_t i, j, n_args;
  s7_pointer arg, par, call_args, func;
  s7_pointer *df;

  func = sc->code;
  n_args = c_function_all_args(func);     /* not counting keywords, I think */
  if (is_safe_procedure(func))
    call_args = c_function_call_args(func);
  else 
    {
      int32_t tx;
      tx = next_tx(sc);
      call_args = make_list(sc, c_function_optional_args(func), sc->F);
      sc->t_temps[tx] = call_args;
    }

  /* assume at the start that there are no keywords */
  for (i = 0, arg = sc->args, par = call_args; (i < n_args) && (is_pair(arg)); i++, arg = cdr(arg), par = cdr(par))
    {
      if (!is_keyword(car(arg)))
	set_car(par, car(arg));
      else
	{
	  s7_pointer kpar, karg;
	  int32_t ki;
	  /* oops -- there are keywords, change scanners (much duplicated code...) */
	  for (kpar = call_args; kpar != par; kpar = cdr(kpar))
	    set_checked(kpar);
	  for (; is_pair(kpar); kpar = cdr(kpar))
	    clear_checked(kpar);
	  df = c_function_arg_names(func);
	  for (ki = i, karg = arg, kpar = par; (ki < n_args) && (is_pair(karg)); ki++, karg = cdr(karg), kpar = cdr(kpar))
	    {
	      if (!is_keyword(car(karg)))
		{
		  if (is_checked(kpar))
		    return(s7_error(sc, sc->wrong_type_arg_symbol, set_elist_3(sc, parameter_set_twice_string, car(kpar), sc->args)));
		  set_checked(kpar);
		  set_car(kpar, car(karg));
		}
	      else
		{
		  s7_pointer p;
		  for (j = 0, p = call_args; j < n_args; j++, p = cdr(p))
		    if (df[j] == car(karg))
		      break;
		  if (j == n_args)
		    return(s7_error(sc, sc->wrong_type_arg_symbol, set_elist_2(sc, s7_make_string_wrapper(sc, "~A: not a parameter name?"), car(karg))));
		  if (is_checked(p))
		    return(s7_error(sc, sc->wrong_type_arg_symbol, set_elist_3(sc, parameter_set_twice_string, car(p), sc->args)));
		  if (!is_pair(cdr(karg)))
		    return(s7_error(sc, sc->error_symbol, set_elist_3(sc, value_is_missing_string, func, car(karg))));
		  set_checked(p);
		  karg = cdr(karg);
		  set_car(p, car(karg));
		}
	    }
	  if (!is_null(karg))
	    return(s7_error(sc, sc->wrong_number_of_args_symbol, set_elist_3(sc, sc->too_many_arguments_string, func, sc->args)));
	  if (ki < n_args)
	    {
	      df = c_function_arg_defaults(func);
	      if (has_simple_defaults(func))
		{
		  for (ki = i, kpar = par; ki < n_args; ki++, kpar = cdr(kpar))
		    if (!is_checked(kpar))
		      set_car(kpar, df[ki]);
		}
	      else
		{
		  for (ki = i, kpar = par; ki < n_args; ki++, kpar = cdr(kpar))
		    if (!is_checked(kpar))
		      {
			s7_pointer defval;
			defval = df[ki];
			if (is_symbol(defval))
			  set_car(kpar, find_symbol_checked(sc, defval));
			else
			  {
			    if (is_pair(defval))
			      set_car(kpar, s7_eval(sc, defval, sc->nil));
			    else set_car(kpar, defval);
			  }
		      }
		}
	    }
	  return(call_args);
	}
    }
  if (!is_null(arg))
    return(s7_error(sc, sc->wrong_number_of_args_symbol, set_elist_3(sc, sc->too_many_arguments_string, func, sc->args)));
  if (i < n_args)
    {
      df = c_function_arg_defaults(func);
      if (has_simple_defaults(func))
	{
	  for (; i < n_args; i++, par = cdr(par))
	    set_car(par, df[i]);
	}
      else
	{
	  for (; i < n_args; i++, par = cdr(par))
	    {
	      s7_pointer defval;
	      defval = df[i];
	      if (is_symbol(defval))
		set_car(par, find_symbol_checked(sc, defval));
	      else
		{
		  if (is_pair(defval))
		    set_car(par, s7_eval(sc, defval, sc->nil));
		  else set_car(par, defval);
		}
	    }
	}
    }
  return(call_args);
}

#define apply_c_function_star(Sc) Sc->value = c_function_call(Sc->code)(Sc, set_c_function_star_args(Sc))

s7_pointer s7_apply_function_star(s7_scheme *sc, s7_pointer fnc, s7_pointer args)
{
  TRACK(sc);
  if (sc->safety > NO_SAFETY)
    set_current_code(sc, cons(sc, fnc, args));
  else set_current_code(sc, fnc);

  if (is_c_function_star(fnc))
    {
      sc->w = sc->args;
      sc->z = sc->code;
      sc->args = args;
      sc->code = fnc;
      apply_c_function_star(sc);
      sc->args = sc->w;
      sc->code = sc->z;
      return(sc->value);
    }
  push_stack(sc, OP_EVAL_DONE, sc->args, sc->code);
  sc->code = fnc;
  sc->args = (needs_copied_args(sc->code)) ? copy_list(sc, args) : args;
  eval(sc, OP_APPLY);
  return(sc->value);
}


s7_pointer s7_eval(s7_scheme *sc, s7_pointer code, s7_pointer e)
{
  declare_jump_info();
  TRACK(sc);

  if (sc->safety > NO_SAFETY)
    {
      if (!s7_is_valid(sc, code))
	fprintf(stderr, "bad code arg to %s: %p\n", __func__, code);
      if (!s7_is_valid(sc, e))
	fprintf(stderr, "bad environment arg to %s: %p\n", __func__, e);
    }

#if DEBUGGING
  _NFre(code);
#endif

  store_jump_info(sc);
  set_jump_info(sc, EVAL_SET_JUMP);
  if (jump_loc != NO_JUMP)
    {
      if (jump_loc != ERROR_JUMP)
	eval(sc, sc->cur_op);
    }
  else 
    {
      push_stack(sc, OP_EVAL_DONE, sc->args, sc->code);
      sc->code = code;
      if ((e != sc->rootlet) &&
	  (is_let(e)))
	sc->envir = e;
      else sc->envir = sc->nil; 
      eval(sc, OP_EVAL);
    }
  restore_jump_info(sc);

  if (is_multiple_value(sc->value))
    sc->value = splice_in_values(sc, multiple_value(sc->value));
  return(sc->value);
}


static s7_pointer g_eval(s7_scheme *sc, s7_pointer args)
{
  #define H_eval "(eval code (env (curlet))) evaluates code in the environment env. 'env' \
defaults to the curlet; to evaluate something in the top-level environment instead, \
pass (rootlet):\n\
\n\
  (define x 32) \n\
  (let ((x 3))\n\
    (eval 'x (rootlet)))\n\
\n\
  returns 32"
  #define Q_eval s7_make_signature(sc, 3, sc->values_symbol, sc->T, sc->is_let_symbol)

  if (is_not_null(cdr(args)))
    {
      s7_pointer e;
      e = cadr(args);
      if (!is_let(e))
	return(wrong_type_argument_with_type(sc, sc->eval_symbol, 2, e, a_let_string));
      if (e == sc->rootlet)
	sc->envir = sc->nil;
      else sc->envir = e;
    }
  sc->code = car(args);
  if (sc->safety > NO_SAFETY)
    {
      if (cyclic_sequences(sc, sc->code, false) == sc->T)
	return(wrong_type_argument_with_type(sc, sc->eval_symbol, 1, sc->code, a_proper_list_string));
      sc->code = copy_body(sc, sc->code);
    }
  else
   {
     if (is_optimized(sc->code))
       clear_all_optimizations(sc, sc->code);
   }
  if (s7_stack_top(sc) < 12)
    push_stack_op(sc, OP_BARRIER);
  push_stack(sc, OP_EVAL, sc->args, sc->code);

  return(sc->nil);
}


s7_pointer s7_call(s7_scheme *sc, s7_pointer func, s7_pointer args)
{
  declare_jump_info();
  TRACK(sc);

  if (sc->safety > NO_SAFETY)
    set_current_code(sc, cons(sc, func, args));
  else set_current_code(sc, func); 

  if (is_c_function(func))
    return(c_function_call(func)(sc, _NFre(args)));  /* no check for wrong-number-of-args -- is that reasonable? */

  sc->temp1 = _NFre(func);                           /* this is feeble GC protection */
  sc->temp2 = _NFre(args);

  store_jump_info(sc);
  set_jump_info(sc, S7_CALL_SET_JUMP);
  if (jump_loc != NO_JUMP)
    {
      if (jump_loc != ERROR_JUMP)
	eval(sc, sc->cur_op);

      if ((jump_loc == CATCH_JUMP) &&        /* we're returning (back to eval) from an error in catch */
	  (sc->stack_end == sc->stack_start))
	push_stack_op(sc, OP_ERROR_QUIT);
    }
  else
    {
      if (sc->safety > NO_SAFETY)
	check_list_validity(sc, "s7_call", args);

      push_stack(sc, OP_EVAL_DONE, sc->args, sc->code); /* this saves the current evaluation and will eventually finish this (possibly) nested call */
      sc->code = func;
      sc->args = (needs_copied_args(func)) ? copy_list(sc, args) : args;
      /* besides a closure, "func" can also be an object (T_C_OBJECT) -- in Snd, a generator for example  */
      eval(sc, OP_APPLY);
    }
  restore_jump_info(sc);
  /* don't clear temp1 or temp2 here -- lots of (Snd) code calls s7_call repeatedly and assumes the
   *   "func" arg is protected between calls.
   */
  return(sc->value);
}


s7_pointer s7_call_with_location(s7_scheme *sc, s7_pointer func, s7_pointer args, const char *caller, const char *file, int32_t line)
{
  s7_pointer result;

  if (caller)
    {
      sc->s7_call_name = caller;
      sc->s7_call_file = file;
      sc->s7_call_line = line;
    }

  result = s7_call(sc, func, args);

  if (caller)
    {
      sc->s7_call_name = NULL;
      sc->s7_call_file = NULL;
      sc->s7_call_line = -1;
    }
  return(result);
}


static s7_pointer implicit_index(s7_scheme *sc, s7_pointer obj, s7_pointer indices)
{
  /* (let ((lst '("12" "34"))) (lst 0 1)) -> #\2
   * (let ((lst (list #(1 2) #(3 4)))) (lst 0 1)) -> 2
   *
   * this can get tricky:
   *   ((list (lambda (a) (+ a 1)) (lambda (b) (* b 2))) 1 2) -> 4
   * but what if func takes rest/optional args, etc?
   *   ((list (lambda args (car args))) 0 "hi" 0)
   *   should this return #\h or "hi"??
   *   currently it is "hi" which is consistent with
   *  ((lambda args (car args)) "hi" 0)
   * but...
   *   ((lambda (arg) arg) "hi" 0)
   * is currently an error (too many arguments)
   * it should be (((lambda (arg) arg) "hi") 0) -> #\h
   *
   * this applies to non-homogeneous cases, so float|int-vectors don't get here
   */

  switch (type(obj))
    {
    case T_VECTOR:                       /* (#(#(1 2) #(3 4)) 1 1) -> 4 */
      return(vector_ref_1(sc, obj, indices));

    case T_STRING:                       /* (#("12" "34") 0 1) -> #\2 */
      if (is_null(cdr(indices)))
	{
	  if (is_byte_vector(obj))       /* ((vector (byte-vector 1)) 0 0) */
	    return(small_int((uint32_t)(character(string_ref_1(sc, obj, car(indices))))));
	  return(string_ref_1(sc, obj, car(indices)));
	}
      return(s7_error(sc, sc->wrong_number_of_args_symbol, set_elist_3(sc, sc->too_many_arguments_string, obj, indices)));

    case T_PAIR:                         /* (#((1 2) (3 4)) 1 0) -> 3, (#((1 (2 3))) 0 1 0) -> 2 */
      obj = list_ref_1(sc, obj, car(indices));
      if (is_pair(cdr(indices)))
	return(implicit_index(sc, obj, cdr(indices)));
      return(obj);

    case T_HASH_TABLE:                   /* ((vector (hash-table '(a . 1) '(b . 2))) 0 'a) -> 1 */
      obj = s7_hash_table_ref(sc, obj, car(indices));
      if (is_pair(cdr(indices)))
	return(implicit_index(sc, obj, cdr(indices)));
      return(obj);

    case T_C_OBJECT:
      return((*(c_object_ref(sc, obj)))(sc, obj, indices));

    case T_LET:
      obj = s7_let_ref(sc, obj, car(indices));
      if (is_pair(cdr(indices)))
	return(implicit_index(sc, obj, cdr(indices)));
      return(obj);

    default:                             /* (#(a b c) 0 1) -> error, but ((list (lambda (x) x)) 0 "hi") -> "hi" */
      if (is_applicable(obj))
	return(g_apply(sc, list_2(sc, obj, indices)));
      return(s7_error(sc, sc->wrong_number_of_args_symbol, set_elist_3(sc, sc->too_many_arguments_string, obj, indices)));
    }
}


/* -------------------------------- type-of -------------------------------- */
static s7_pointer type_to_typers[NUM_TYPES];
static void init_typers(s7_scheme *sc)
{
  type_to_typers[T_FREE] =                sc->F;
  type_to_typers[T_PAIR] =                sc->is_pair_symbol;
  type_to_typers[T_NIL] =                 sc->is_null_symbol;
  type_to_typers[T_EOF_OBJECT] =          sc->is_eof_object_symbol;
  type_to_typers[T_UNDEFINED] =           sc->is_undefined_symbol;
  type_to_typers[T_UNSPECIFIED] =         sc->is_unspecified_symbol;
  type_to_typers[T_BOOLEAN] =             sc->is_boolean_symbol;
  type_to_typers[T_CHARACTER] =           sc->is_char_symbol;
  type_to_typers[T_SYMBOL] =              sc->is_symbol_symbol;      /* and keyword? */
  type_to_typers[T_SYNTAX] =              sc->is_syntax_symbol;
  type_to_typers[T_INTEGER] =             sc->is_integer_symbol;
  type_to_typers[T_RATIO] =               sc->is_rational_symbol;
  type_to_typers[T_REAL] =                sc->is_float_symbol;
  type_to_typers[T_COMPLEX] =             sc->is_complex_symbol;
  type_to_typers[T_BIG_INTEGER] =         sc->is_integer_symbol;
  type_to_typers[T_BIG_RATIO] =           sc->is_rational_symbol;
  type_to_typers[T_BIG_REAL] =            sc->is_float_symbol;
  type_to_typers[T_BIG_COMPLEX] =         sc->is_complex_symbol;
  type_to_typers[T_STRING] =              sc->is_string_symbol;      /* and byte-vector? */
  type_to_typers[T_C_OBJECT] =            sc->is_c_object_symbol;
  type_to_typers[T_VECTOR] =              sc->is_vector_symbol;
  type_to_typers[T_INT_VECTOR] =          sc->is_int_vector_symbol;
  type_to_typers[T_FLOAT_VECTOR] =        sc->is_float_vector_symbol;
  type_to_typers[T_CATCH] =               sc->F;
  type_to_typers[T_DYNAMIC_WIND] =        sc->F;
  type_to_typers[T_HASH_TABLE] =          sc->is_hash_table_symbol;
  type_to_typers[T_LET] =                 sc->is_let_symbol;
  type_to_typers[T_ITERATOR] =            sc->is_iterator_symbol;
  type_to_typers[T_STACK] =               sc->F;
  type_to_typers[T_COUNTER] =             sc->F;
  type_to_typers[T_OPTLIST] =             sc->F;
  type_to_typers[T_SLOT] =                sc->F;
  type_to_typers[T_C_POINTER] =           sc->is_c_pointer_symbol;
  type_to_typers[T_OUTPUT_PORT] =         sc->is_output_port_symbol;
  type_to_typers[T_INPUT_PORT] =          sc->is_input_port_symbol;
  type_to_typers[T_BAFFLE] =              sc->F;
  type_to_typers[T_RANDOM_STATE] =        sc->is_random_state_symbol;
  type_to_typers[T_GOTO] =                sc->F;                       /* (continuation? goto) -> #f -- we need a type indicator for this */
  type_to_typers[T_CONTINUATION] =        sc->is_continuation_symbol;
  type_to_typers[T_CLOSURE] =             sc->is_procedure_symbol;
  type_to_typers[T_CLOSURE_STAR] =        sc->is_procedure_symbol;
  type_to_typers[T_C_MACRO] =             sc->is_macro_symbol;
  type_to_typers[T_MACRO] =               sc->is_macro_symbol;
  type_to_typers[T_MACRO_STAR] =          sc->is_macro_symbol;
  type_to_typers[T_BACRO] =               sc->is_macro_symbol;
  type_to_typers[T_BACRO_STAR] =          sc->is_macro_symbol;
  type_to_typers[T_C_FUNCTION] =          sc->is_procedure_symbol;
  type_to_typers[T_C_FUNCTION_STAR] =     sc->is_procedure_symbol;
  type_to_typers[T_C_ANY_ARGS_FUNCTION] = sc->is_procedure_symbol;
  type_to_typers[T_C_OPT_ARGS_FUNCTION] = sc->is_procedure_symbol; 
  type_to_typers[T_C_RST_ARGS_FUNCTION] = sc->is_procedure_symbol;
}

s7_pointer s7_type_of(s7_pointer arg) {return(type_to_typers[type(arg)]);}

static s7_pointer g_type_of(s7_scheme *sc, s7_pointer args) 
{
  #define H_type_of "(type-of obj) returns a symbol describing obj's type"
  #define Q_type_of s7_make_signature(sc, 2, s7_make_signature(sc, 2, sc->is_symbol_symbol, sc->is_boolean_symbol), sc->T)
  int32_t tp;

  tp = type(car(args));
  if ((tp >= 0) && (tp < NUM_TYPES))
    return(type_to_typers[type(car(args))]);
  return(s7_make_symbol(sc, "unknown!"));
}


/* -------------------------------- s7-version -------------------------------- */

static s7_pointer g_s7_version(s7_scheme *sc, s7_pointer args)
{
  #define H_s7_version "(s7-version) returns some string describing the current s7"
  #define Q_s7_version pcl_s

  return(s7_make_string(sc, "s7 " S7_VERSION ", " S7_DATE));
}

static s7_pointer s7_version_p(void) {return(s7_make_string(cur_sc, "s7 " S7_VERSION ", " S7_DATE));}


void s7_quit(s7_scheme *sc)
{
  sc->longjmp_ok = false;

  pop_input_port(sc);
  stack_reset(sc);
  push_stack_op_let(sc, OP_EVAL_DONE);
}

/* -------------------------------- exit -------------------------------- */
static s7_pointer g_emergency_exit(s7_scheme *sc, s7_pointer args)
{
  #define H_emergency_exit "(emergency-exit obj) exits s7 immediately"
  #define Q_emergency_exit s7_make_signature(sc, 2, sc->T, sc->T)

  s7_pointer obj;
#ifndef EXIT_SUCCESS
  #define EXIT_SUCCESS 0
  #define EXIT_FAILURE 1
#endif
  if (is_null(args))
    _exit(EXIT_SUCCESS);          /* r7rs spec says use _exit here */
  obj = car(args);
  if (obj == sc->F)
    _exit(EXIT_FAILURE);
  if ((obj == sc->T) || (!s7_is_integer(obj)))
    _exit(EXIT_SUCCESS);
  _exit((int)s7_integer(obj));
  return(sc->F);
}


static s7_pointer g_exit(s7_scheme *sc, s7_pointer args)
{
  #define H_exit "(exit obj) exits s7"
  #define Q_exit s7_make_signature(sc, 2, sc->T, sc->T)

  s7_quit(sc);
  return(g_emergency_exit(sc, args));
}


#if DEBUGGING
static s7_pointer g_abort(s7_scheme *sc, s7_pointer args) {abort();}
#endif


/* -------------------------------- optimizer stuff -------------------------------- */

static s7_function all_x_function[OPT_MAX_DEFINED];
#define is_all_x_op(Op) (all_x_function[Op])

static bool is_all_x_safe(s7_scheme *sc, s7_pointer p)
{
  return((!is_pair(p)) ||
	 ((is_optimized(p)) && (is_all_x_op(optimize_op(p)))) ||
	 ((car(p) == sc->quote_symbol) && (is_pair(cdr(p))) && (is_null(cddr(p))))) ; /* (if #t (quote . -1)) */
}

static int32_t all_x_count(s7_scheme *sc, s7_pointer x)
{
  int32_t count = 0;
  s7_pointer p;
  for (p = cdr(x); is_pair(p); p = cdr(p))
    if (is_all_x_safe(sc, car(p)))
      count++;
  return(count);
}


static s7_pointer apply_boolean_method(s7_scheme *sc, s7_pointer obj, s7_pointer method)
{
  s7_pointer func;
  func = find_method(sc, find_let(sc, obj), method);
  if (func == sc->undefined) return(sc->F);
  return(s7_apply_function(sc, func, list_1(sc, obj)));
}

#define all_x_bool(Sc, Checker, Method, Arg1)		       \
  {							       \
    s7_pointer Arg = Arg1;				       \
    if (Checker(Arg)) return(Sc->T);                           \
    if (!has_methods(Arg)) return(sc->F);		       \
    return(apply_boolean_method(sc, Arg, Method));			       \
  }

#define all_x_not_bool(Sc, Checker, Method, Arg1)		       \
  {							       \
    s7_pointer Arg = Arg1;				       \
    if (Checker(Arg)) return(Sc->F);                           \
    if (!has_methods(Arg)) return(sc->T);		       \
    return((apply_boolean_method(sc, Arg, Method) == sc->F) ? sc->T : sc->F);			       \
  }


/* arg here is the full expression */
static s7_pointer all_x_c(s7_scheme *sc, s7_pointer arg)       {return(arg);}
static s7_pointer all_x_q(s7_scheme *sc, s7_pointer arg)       {return(cadr(arg));}
static s7_pointer all_x_unsafe_s(s7_scheme *sc, s7_pointer arg){return(find_symbol_checked(sc, arg));}
static s7_pointer all_x_s(s7_scheme *sc, s7_pointer arg)       {return(find_symbol_unchecked(sc, arg));}
static s7_pointer local_x_s(s7_scheme *sc, s7_pointer arg)     {return(local_symbol_value(arg));}
static s7_pointer all_x_k(s7_scheme *sc, s7_pointer arg)       {return(arg);}
static s7_pointer all_x_c_c(s7_scheme *sc, s7_pointer arg)     {return(c_call(arg)(sc, cdr(arg)));}
static s7_pointer all_x_not_c_c(s7_scheme *sc, s7_pointer arg) {return(make_boolean(sc, is_false(sc, c_call(cadr(arg))(sc, cdadr(arg)))));}

static s7_pointer all_x_c_add1(s7_scheme *sc, s7_pointer arg)  
{
  s7_pointer x;
  x = find_symbol_unchecked(sc, cadr(arg));
  if (is_integer(x))
    return(make_integer(sc, integer(x) + 1));
  return(g_add_s1_1(sc, x, cdr(arg))); /* arg=(+ x 1) */
}

static s7_pointer local_x_c_add1(s7_scheme *sc, s7_pointer arg)  
{
  s7_pointer x;
  x = local_symbol_value(cadr(arg));
  if (is_integer(x))
    return(make_integer(sc, integer(x) + 1));
  return(g_add_s1_1(sc, x, cdr(arg))); /* arg=(+ x 1) */
}

static s7_pointer local_x_c_sub1(s7_scheme *sc, s7_pointer arg)  
{
  s7_pointer x;
  x = local_symbol_value(cadr(arg));
  if (is_integer(x))
    return(make_integer(sc, integer(x) - 1));
  return(minus_c1(sc, x));
}

static s7_pointer all_x_c_addi(s7_scheme *sc, s7_pointer arg)  
{
  s7_pointer x;
  x = find_symbol_unchecked(sc, cadr(arg));
  if (is_integer(x))
    return(make_integer(sc, integer(x) + integer(caddr(arg))));
  return(g_add_2(sc, set_plist_2(sc, x, caddr(arg))));
}

static s7_pointer local_x_c_addi(s7_scheme *sc, s7_pointer arg)  
{
  s7_pointer x;
  x = local_symbol_value(cadr(arg));
  if (is_integer(x))
    return(make_integer(sc, integer(x) + integer(caddr(arg))));
  return(g_add_2(sc, set_plist_2(sc, x, caddr(arg))));
}

static s7_pointer all_x_c_char_eq(s7_scheme *sc, s7_pointer arg)  
{
  s7_pointer c;
  c = find_symbol_unchecked(sc, cadr(arg));
  if (c == caddr(arg))
    return(sc->T);
  if (s7_is_character(c))
    return(sc->F);
  method_or_bust(sc, c, sc->char_eq_symbol, set_plist_2(sc, c, caddr(arg)), T_CHARACTER, 1);
}

static s7_pointer all_x_not_is_eq_car_q(s7_scheme *sc, s7_pointer arg)  
{
  s7_pointer lst, a;
  a = opt_pair2(cdr(arg));
  lst = find_symbol_unchecked(sc, opt_sym2(a));
  if (!is_pair(lst))
    return(make_boolean(sc, is_false(sc, g_is_eq(sc, set_plist_2(sc, g_car(sc, set_plist_1(sc, lst)), opt_sym3(a))))));
  return(make_boolean(sc, car(lst) != opt_sym3(a)));
}

static s7_pointer local_x_not_is_eq_car_q(s7_scheme *sc, s7_pointer arg)  
{
  s7_pointer lst, a;
  a = opt_pair2(cdr(arg));
  lst = local_symbol_value(opt_sym2(a));
  if (!is_pair(lst))
    return(make_boolean(sc, is_false(sc, g_is_eq(sc, set_plist_2(sc, g_car(sc, set_plist_1(sc, lst)), opt_sym3(a))))));
  return(make_boolean(sc, car(lst) != opt_sym3(a)));
}

static s7_pointer all_x_is_pair_cdr_s(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer p;
  p = find_symbol_unchecked(sc, opt_sym2(cdr(arg)));
  if (is_pair(p))
    return(make_boolean(sc, is_pair(cdr(p))));
  return(g_is_pair(sc, set_plist_1(sc, g_cdr(sc, set_plist_1(sc, p)))));
}

static s7_pointer local_x_is_pair_cdr_s(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer p;
  p = local_symbol_value(opt_sym2(cdr(arg)));
  if (is_pair(p))
    return(make_boolean(sc, is_pair(cdr(p))));
  return(g_is_pair(sc, set_plist_1(sc, g_cdr(sc, set_plist_1(sc, p)))));
}

static s7_pointer all_x_c_s(s7_scheme *sc, s7_pointer arg)
{
  set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(arg)));
  return(c_call(arg)(sc, sc->t1_1));
}

static s7_pointer local_x_c_s(s7_scheme *sc, s7_pointer arg)
{
  set_car(sc->t1_1, local_symbol_value(cadr(arg)));
  return(c_call(arg)(sc, sc->t1_1));
}

static s7_pointer all_x_cdr_s(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer val;
  val = find_symbol_unchecked(sc, cadr(arg));
  return((is_pair(val)) ? cdr(val) : g_cdr(sc, set_plist_1(sc, val)));
}

static s7_pointer local_x_cdr_s(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer val;
  val = local_symbol_value(cadr(arg));
  return((is_pair(val)) ? cdr(val) : g_cdr(sc, set_plist_1(sc, val)));
}

static s7_pointer all_x_car_s(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer val;
  val = find_symbol_unchecked(sc, cadr(arg));
  return((is_pair(val)) ? car(val) : g_car(sc, set_plist_1(sc, val)));
}

static s7_pointer local_x_car_s(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer val;
  val = local_symbol_value(cadr(arg));
  return((is_pair(val)) ? car(val) : g_car(sc, set_plist_1(sc, val)));
}

static s7_pointer all_x_cadr_s(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer val;
  val = find_symbol_unchecked(sc, cadr(arg));
  return(((is_pair(val)) && (is_pair(cdr(val)))) ? cadr(val) : g_cadr(sc, set_plist_1(sc, val)));
}

static s7_pointer local_x_cadr_s(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer val;
  val = local_symbol_value(cadr(arg));
  return(((is_pair(val)) && (is_pair(cdr(val)))) ? cadr(val) : g_cadr(sc, set_plist_1(sc, val)));
}

static s7_pointer all_x_is_null_s(s7_scheme *sc, s7_pointer arg)
{
  all_x_bool(sc, is_null, sc->is_null_symbol, find_symbol_unchecked(sc, cadr(arg)));
}

static s7_pointer local_x_is_null_s(s7_scheme *sc, s7_pointer arg)
{
  all_x_bool(sc, is_null, sc->is_null_symbol, local_symbol_value(cadr(arg)));
}

static s7_pointer all_x_is_symbol_s(s7_scheme *sc, s7_pointer arg)
{
  all_x_bool(sc, is_symbol, sc->is_symbol_symbol, find_symbol_unchecked(sc, cadr(arg)));
}

static s7_pointer local_x_is_symbol_s(s7_scheme *sc, s7_pointer arg)
{
  all_x_bool(sc, is_symbol, sc->is_symbol_symbol, local_symbol_value(cadr(arg)));
}

static s7_pointer all_x_is_pair_s(s7_scheme *sc, s7_pointer arg)
{
  all_x_bool(sc, is_pair, sc->is_pair_symbol, find_symbol_unchecked(sc, cadr(arg)));
}

static s7_pointer local_x_is_pair_s(s7_scheme *sc, s7_pointer arg)
{
  all_x_bool(sc, is_pair, sc->is_pair_symbol, local_symbol_value(cadr(arg)));
}

static s7_pointer all_x_is_keyword_s(s7_scheme *sc, s7_pointer arg)
{
  all_x_bool(sc, is_keyword, sc->is_keyword_symbol, find_symbol_unchecked(sc, cadr(arg)));
}

static s7_pointer all_x_is_integer_s(s7_scheme *sc, s7_pointer arg)
{
  all_x_bool(sc, is_integer, sc->is_integer_symbol, find_symbol_unchecked(sc, cadr(arg)));
}

static s7_pointer all_x_is_procedure_s(s7_scheme *sc, s7_pointer arg)
{
  all_x_bool(sc, is_procedure, sc->is_procedure_symbol, find_symbol_unchecked(sc, cadr(arg)));
}

static s7_pointer all_x_is_string_s(s7_scheme *sc, s7_pointer arg)
{
  all_x_bool(sc, is_string, sc->is_string_symbol, find_symbol_unchecked(sc, cadr(arg)));
}

static s7_pointer all_x_is_vector_s(s7_scheme *sc, s7_pointer arg)
{
  all_x_bool(sc, s7_is_vector, sc->is_vector_symbol, find_symbol_unchecked(sc, cadr(arg)));
}

static s7_pointer all_x_is_proper_list_s(s7_scheme *sc, s7_pointer arg)
{
    s7_pointer lst;
    lst = find_symbol_unchecked(sc, cadr(arg));
    if (s7_is_proper_list(sc, lst)) return(sc->T);
    if (!has_methods(lst)) return(sc->F);
    return(apply_boolean_method(sc, lst, sc->is_proper_list_symbol));
}

static s7_pointer all_x_not_s(s7_scheme *sc, s7_pointer arg)
{
  return(make_boolean(sc, is_false(sc, find_symbol_unchecked(sc, cadr(arg)))));
}

static s7_pointer all_x_not_is_pair_s(s7_scheme *sc, s7_pointer arg)
{
  all_x_not_bool(sc, is_pair, sc->is_pair_symbol, find_symbol_unchecked(sc, opt_sym3(arg)));
}

static s7_pointer local_x_not_is_pair_s(s7_scheme *sc, s7_pointer arg)
{
  all_x_not_bool(sc, is_pair, sc->is_pair_symbol, local_symbol_value(opt_sym3(arg)));
}

static s7_pointer all_x_c_sc(s7_scheme *sc, s7_pointer arg)
{
  set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(arg)));
  set_car(sc->t2_2, caddr(arg));
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer local_x_c_sc(s7_scheme *sc, s7_pointer arg)
{
  set_car(sc->t2_1, local_symbol_value(cadr(arg)));
  set_car(sc->t2_2, caddr(arg));
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_c_cs(s7_scheme *sc, s7_pointer arg)
{
  set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(arg)));
  set_car(sc->t2_1, cadr(arg));
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_c_ss(s7_scheme *sc, s7_pointer arg)
{
  set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(arg)));
  set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(arg)));
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_c_ls(s7_scheme *sc, s7_pointer arg)
{
  set_car(sc->t2_1, local_symbol_value(cadr(arg)));
  set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(arg)));
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer x_c_hash_table_ref_ss(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  hash_entry_t *x;
  if (!is_hash_table(table))
    return(g_hash_table(sc, set_plist_2(sc, table, key)));
  x = (*hash_table_checker(table))(sc, table, key);
  if (x) return(x->value);
  return(sc->F);
}

static s7_pointer all_x_c_hash_table_ref_ss(s7_scheme *sc, s7_pointer arg)
{
  return(x_c_hash_table_ref_ss(sc, find_symbol_unchecked(sc, cadr(arg)), find_symbol_unchecked(sc, caddr(arg))));
}

static s7_pointer local_x_c_hash_table_ref_ss(s7_scheme *sc, s7_pointer arg)
{
  return(x_c_hash_table_ref_ss(sc, local_symbol_value(cadr(arg)), local_symbol_value(caddr(arg))));
}

static s7_pointer all_x_c_hash_table_ref_car(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer table, lst;
  hash_entry_t *x;

  table = find_symbol_unchecked(sc, cadr(arg));
  lst = find_symbol_unchecked(sc, opt_sym3(cdr(arg)));
  if (!is_pair(lst))
    return(simple_wrong_type_argument(sc, sc->car_symbol, lst, T_PAIR));

  if (!is_hash_table(table))
    return(g_hash_table(sc, set_plist_2(sc, table, car(lst))));

  x = (*hash_table_checker(table))(sc, table, car(lst));
  if (x) return(x->value);
  return(sc->F);
}

static s7_pointer local_x_c_ss(s7_scheme *sc, s7_pointer arg)
{
  set_car(sc->t2_1, local_symbol_value(cadr(arg)));
  set_car(sc->t2_2, local_symbol_value(caddr(arg)));
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_c_qs(s7_scheme *sc, s7_pointer arg)
{
  set_car(sc->t2_1, cadadr(arg));
  set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(arg)));
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_c_sq(s7_scheme *sc, s7_pointer arg)
{
  set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(arg)));
  set_car(sc->t2_2, opt_con2(cdr(arg)));
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer local_x_c_sq(s7_scheme *sc, s7_pointer arg)
{
  set_car(sc->t2_1, local_symbol_value(cadr(arg)));
  set_car(sc->t2_2, opt_con2(cdr(arg)));
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_c_cq(s7_scheme *sc, s7_pointer arg)
{
  set_car(sc->t2_1, cadr(arg));
  set_car(sc->t2_2, opt_con2(cdr(arg)));
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_c_sss(s7_scheme *sc, s7_pointer arg)
{
  set_car(sc->t3_1, find_symbol_unchecked(sc, cadr(arg)));
  set_car(sc->t3_2, find_symbol_unchecked(sc, caddr(arg)));
  set_car(sc->t3_3, find_symbol_unchecked(sc, cadddr(arg)));
  return(c_call(arg)(sc, sc->t3_1));
}

static s7_pointer local_x_c_sss(s7_scheme *sc, s7_pointer arg)
{
  set_car(sc->t3_1, local_symbol_value(cadr(arg)));
  set_car(sc->t3_2, local_symbol_value(caddr(arg)));
  set_car(sc->t3_3, local_symbol_value(cadddr(arg)));
  return(c_call(arg)(sc, sc->t3_1));
}

static s7_pointer all_x_c_scs(s7_scheme *sc, s7_pointer arg)
{
  set_car(sc->t3_1, find_symbol_unchecked(sc, cadr(arg)));
  set_car(sc->t3_3, find_symbol_unchecked(sc, cadddr(arg)));
  set_car(sc->t3_2, caddr(arg));
  return(c_call(arg)(sc, sc->t3_1));
}

static s7_pointer all_x_c_css(s7_scheme *sc, s7_pointer arg)
{
  set_car(sc->t3_2, find_symbol_unchecked(sc, caddr(arg)));
  set_car(sc->t3_3, find_symbol_unchecked(sc, cadddr(arg)));
  set_car(sc->t3_1, cadr(arg));
  return(c_call(arg)(sc, sc->t3_1));
}

static s7_pointer all_x_c_csc(s7_scheme *sc, s7_pointer arg)
{
  set_car(sc->t3_2, find_symbol_unchecked(sc, caddr(arg)));
  set_car(sc->t3_1, cadr(arg));
  set_car(sc->t3_3, cadddr(arg));
  return(c_call(arg)(sc, sc->t3_1));
}

static s7_pointer all_x_c_ssc(s7_scheme *sc, s7_pointer arg)
{
  set_car(sc->t3_1, find_symbol_unchecked(sc, cadr(arg)));
  set_car(sc->t3_2, find_symbol_unchecked(sc, caddr(arg)));
  set_car(sc->t3_3, cadddr(arg));
  return(c_call(arg)(sc, sc->t3_1));
}

static s7_pointer all_x_c_opcq(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  largs = cadr(arg);
  set_car(sc->t1_1, c_call(largs)(sc, cdr(largs)));
  return(c_call(arg)(sc, sc->t1_1));
}

static s7_pointer all_x_c_s_opcq(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  largs = caddr(arg);
  set_car(sc->t2_2, c_call(largs)(sc, cdr(largs)));
  set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(arg)));
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_c_c_opcq(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  largs = caddr(arg);
  set_car(sc->t2_2, c_call(largs)(sc, cdr(largs)));
  set_car(sc->t2_1, cadr(arg));
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_c_opcq_s(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  largs = cadr(arg);
  set_car(sc->t2_1, c_call(largs)(sc, cdr(largs)));
  set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(arg)));
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_c_opcq_c(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  largs = cadr(arg);
  set_car(sc->t2_1, c_call(largs)(sc, cdr(largs)));
  set_car(sc->t2_2, caddr(arg));
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer safe_list_if_possible(s7_scheme *sc, int32_t num_args)
{
  if ((num_args != 0) &&
      (num_args < NUM_SAFE_LISTS))
    {
      sc->current_safe_list = num_args;
      if (!is_pair(sc->safe_lists[num_args]))
	sc->safe_lists[num_args] = permanent_list(sc, num_args);
      if (!list_is_in_use(sc->safe_lists[num_args]))
	{
	  set_list_in_use(sc->safe_lists[num_args]);
	  return(sc->safe_lists[num_args]);
	}
    }
  return(make_list(sc, num_args, sc->nil));
}

static s7_pointer all_x_c_all_s(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer args, p;
  int32_t tx;
  tx = next_tx(sc);
  sc->t_temps[tx] = safe_list_if_possible(sc, integer(arglist_length(arg)));
  for (args = cdr(arg), p = sc->t_temps[tx]; is_pair(args); args = cdr(args), p = cdr(p))
    set_car(p, find_symbol_unchecked(sc, car(args)));
  clear_list_in_use(sc->t_temps[tx]);
  sc->current_safe_list = 0;
  return(c_call(arg)(sc, sc->t_temps[tx]));
}

/* h_safe_c_all_x (etc), h_safe_c_s_opaq, h_safe_c_c_opscq, h_safe_c_scc */
 
static s7_pointer all_x_c_opcq_opcq(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  int32_t tx;
  tx = next_tx(sc);
  largs = cadr(arg);
  sc->t_temps[tx] = c_call(largs)(sc, cdr(largs));
  largs = caddr(arg);
  set_car(sc->t2_2, c_call(largs)(sc, cdr(largs)));
  set_car(sc->t2_1, sc->t_temps[tx]);
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_c_opsq(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  largs = cadr(arg);
  set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(largs)));
  set_car(sc->t1_1, c_call(largs)(sc, sc->t1_1));
  return(c_call(arg)(sc, sc->t1_1));
}

static s7_pointer local_x_c_opsq(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  largs = cadr(arg);
  set_car(sc->t1_1, local_symbol_value(cadr(largs)));
  set_car(sc->t1_1, c_call(largs)(sc, sc->t1_1));
  return(c_call(arg)(sc, sc->t1_1));
}

static s7_pointer all_x_c_car_s(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer val;
  val = find_symbol_unchecked(sc, opt_sym2(cdr(arg)));
  set_car(sc->t1_1, (is_pair(val)) ? car(val) : g_car(sc, set_plist_1(sc, val)));
  return(c_call(arg)(sc, sc->t1_1));
}

static s7_pointer local_x_c_car_s(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer val;
  val = local_symbol_value(opt_sym2(cdr(arg)));
  set_car(sc->t1_1, (is_pair(val)) ? car(val) : g_car(sc, set_plist_1(sc, val)));
  return(c_call(arg)(sc, sc->t1_1));
}

static s7_pointer all_x_c_cdr_s(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer val;
  val = find_symbol_unchecked(sc, opt_sym2(cdr(arg)));
  set_car(sc->t1_1, (is_pair(val)) ? cdr(val) : g_cdr(sc, set_plist_1(sc, val)));
  return(c_call(arg)(sc, sc->t1_1));
}

static s7_pointer local_x_c_cdr_s(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer val;
  val = local_symbol_value(opt_sym2(cdr(arg)));
  set_car(sc->t1_1, (is_pair(val)) ? cdr(val) : g_cdr(sc, set_plist_1(sc, val)));
  return(c_call(arg)(sc, sc->t1_1));
}

static s7_pointer all_x_c_not_opsq(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  largs = cadr(arg);
  set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(largs)));
  if (c_call(largs)(sc, sc->t1_1) == sc->F)
    return(sc->T);
  return(sc->F);
}

static s7_pointer local_x_c_not_opsq(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  largs = cadr(arg);
  set_car(sc->t1_1, local_symbol_value(cadr(largs)));
  if (c_call(largs)(sc, sc->t1_1) == sc->F)
    return(sc->T);
  return(sc->F);
}

static s7_pointer all_x_c_opssq(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  largs = cadr(arg);
  set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(largs)));
  set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(largs)));
  set_car(sc->t1_1, c_call(largs)(sc, sc->t2_1));
  return(c_call(arg)(sc, sc->t1_1));
}

static s7_pointer all_x_c_not_opssq(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  largs = cadr(arg);
  set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(largs)));
  set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(largs)));
  if (c_call(largs)(sc, sc->t2_1) == sc->F)
    return(sc->T);
  return(sc->F);
}

static s7_pointer local_x_c_opssq(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  largs = cadr(arg);
  set_car(sc->t2_1, local_symbol_value(cadr(largs)));
  set_car(sc->t2_2, local_symbol_value(caddr(largs)));
  set_car(sc->t1_1, c_call(largs)(sc, sc->t2_1));
  return(c_call(arg)(sc, sc->t1_1));
}

static s7_pointer all_x_c_opscq(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  largs = cadr(arg);
  set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(largs)));
  set_car(sc->t2_2, caddr(largs));
  set_car(sc->t1_1, c_call(largs)(sc, sc->t2_1));
  return(c_call(arg)(sc, sc->t1_1));
}

static s7_pointer all_x_c_opcsq(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  largs = cadr(arg);
  set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(largs)));
  set_car(sc->t2_1, cadr(largs));
  set_car(sc->t1_1, c_call(largs)(sc, sc->t2_1));
  return(c_call(arg)(sc, sc->t1_1));
}

static s7_pointer all_x_c_opsqq(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  largs = cadr(arg);
  set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(largs)));
  set_car(sc->t2_2, cadr(caddr(largs)));
  set_car(sc->t1_1, c_call(largs)(sc, sc->t2_1));
  return(c_call(arg)(sc, sc->t1_1));
}

static s7_pointer all_x_c_opqsq(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  largs = cadr(arg);
  set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(largs)));
  set_car(sc->t2_1, cadadr(largs));
  set_car(sc->t1_1, c_call(largs)(sc, sc->t2_1));
  return(c_call(arg)(sc, sc->t1_1));
}

static s7_pointer all_x_c_opcsq_c(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  largs = cadr(arg);
  set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(largs)));
  set_car(sc->t2_1, cadr(largs));
  set_car(sc->t2_1, c_call(largs)(sc, sc->t2_1));
  set_car(sc->t2_2, caddr(arg));
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_c_opssq_s(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  largs = cadr(arg);
  set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(largs)));
  set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(largs)));
  set_car(sc->t2_1, c_call(largs)(sc, sc->t2_1));
  set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(arg)));
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_c_opscq_s(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  largs = cadr(arg);
  set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(largs)));
  set_car(sc->t2_2, caddr(largs));
  set_car(sc->t2_1, c_call(largs)(sc, sc->t2_1));
  set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(arg)));
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_c_opscq_c(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  largs = cadr(arg);
  set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(largs)));
  set_car(sc->t2_2, caddr(largs));
  set_car(sc->t2_1, c_call(largs)(sc, sc->t2_1));
  set_car(sc->t2_2, caddr(arg));
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer local_x_c_opssq_s(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  largs = cadr(arg);
  set_car(sc->t2_1, local_symbol_value(cadr(largs)));
  set_car(sc->t2_2, local_symbol_value(caddr(largs)));
  set_car(sc->t2_1, c_call(largs)(sc, sc->t2_1));
  set_car(sc->t2_2, local_symbol_value(caddr(arg)));
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_c_opssq_c(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  largs = cadr(arg);
  set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(largs)));
  set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(largs)));
  set_car(sc->t2_1, c_call(largs)(sc, sc->t2_1));
  set_car(sc->t2_2, caddr(arg));
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_c_opsq_s(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  largs = cadr(arg);
  set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(largs)));
  set_car(sc->t2_1, c_call(largs)(sc, sc->t1_1));
  set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(arg)));
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_c_opsq_q(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  largs = cadr(arg);
  set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(largs)));
  set_car(sc->t2_1, c_call(largs)(sc, sc->t1_1));
  set_car(sc->t2_2, cadr(caddr(arg)));
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_c_opsq_qs(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  largs = cadr(arg);
  set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(largs)));
  set_car(sc->t3_1, c_call(largs)(sc, sc->t1_1));
  set_car(sc->t3_2, cadr(caddr(arg)));
  set_car(sc->t3_3, find_symbol_unchecked(sc, cadddr(arg)));
  return(c_call(arg)(sc, sc->t3_1));
}

static s7_pointer local_x_c_opsq_c(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  largs = cadr(arg);
  set_car(sc->t1_1, local_symbol_value(cadr(largs)));
  set_car(sc->t2_1, c_call(largs)(sc, sc->t1_1));
  set_car(sc->t2_2, caddr(arg));
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_c_opsq_c(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  largs = cadr(arg);
  set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(largs)));
  set_car(sc->t2_1, c_call(largs)(sc, sc->t1_1));
  set_car(sc->t2_2, caddr(arg));
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_c_s_opssq(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  largs = caddr(arg);
  set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(largs)));
  set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(largs)));
  set_car(sc->t2_2, c_call(largs)(sc, sc->t2_1));
  set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(arg)));
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_c_c_opssq(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  largs = caddr(arg);
  set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(largs)));
  set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(largs)));
  set_car(sc->t2_2, c_call(largs)(sc, sc->t2_1));
  set_car(sc->t2_1, cadr(arg));
  return(c_call(arg)(sc, sc->t2_1));
}



/* p_pd (oscil o (oscil o1 x)) s_opssq
 * d_pd for (oscil o1 x)?     
 * p_dd for (* .1 (tap d x))   c_opssq
 * s_opsq s_op_s_opsqq
 * all_x_c_s handles a lot of g_env_w
 */

static s7_pointer direct_x_c_c_opssq(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  s7_double x2;
  largs = caddr(arg);
  x2 = ((s7_d_pd_t)opt_direct_x(cddr(arg)))(find_symbol_unchecked(sc, cadr(largs)), 
					    real_to_double(sc, find_symbol_unchecked(sc, caddr(largs)), "number_to_double"));
  return(((s7_p_dd_t)opt_direct_x_call(cdr(arg)))(real_to_double(sc, cadr(arg), "*"), x2)); 
}


static s7_pointer local_x_c_s_opssq(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  largs = caddr(arg);
  set_car(sc->t2_1, local_symbol_value(cadr(largs)));
  set_car(sc->t2_2, local_symbol_value(caddr(largs)));
  set_car(sc->t2_2, c_call(largs)(sc, sc->t2_1));
  set_car(sc->t2_1, local_symbol_value(cadr(arg)));
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_c_s_opscq(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  largs = caddr(arg);
  set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(largs)));
  set_car(sc->t2_2, caddr(largs));
  set_car(sc->t2_2, c_call(largs)(sc, sc->t2_1));
  set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(arg)));
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_c_s_opsq(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  largs = caddr(arg);
  set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(largs)));
  set_car(sc->t2_2, c_call(largs)(sc, sc->t1_1));
  set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(arg)));
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer local_x_c_s_opsq(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  largs = caddr(arg);
  set_car(sc->t1_1, local_symbol_value(cadr(largs)));
  set_car(sc->t2_2, c_call(largs)(sc, sc->t1_1));
  set_car(sc->t2_1, local_symbol_value(cadr(arg)));
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_c_op_s_opsq_q(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer outer, args;
  outer = cadr(arg);
  args = caddr(outer);
  set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(args)));
  set_car(sc->t2_2, c_call(args)(sc, sc->t1_1));
  set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(outer)));
  set_car(sc->t1_1, c_call(outer)(sc, sc->t2_1));
  return(c_call(arg)(sc, sc->t1_1));
}

static s7_pointer all_x_c_op_opsq_s_q(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer outer, args;
  outer = cadr(arg);
  args = cadr(outer);
  set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(args)));
  set_car(sc->t2_1, c_call(args)(sc, sc->t1_1));
  set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(outer)));
  set_car(sc->t1_1, c_call(outer)(sc, sc->t2_1));
  return(c_call(arg)(sc, sc->t1_1));
}

static s7_pointer all_x_c_c_opsq(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  largs = caddr(arg);
  set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(largs)));
  set_car(sc->t2_2, c_call(largs)(sc, sc->t1_1));
  set_car(sc->t2_1, cadr(arg));
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer direct_x_c_opsq_opsq(s7_scheme *sc, s7_pointer arg)
{
  s7_double x1, x2;
  cur_sc = sc;
  x1 = ((s7_d_p_t)opt_direct_x(cdr(arg)))(find_symbol_unchecked(sc, cadr(cadr(arg))));
  x2 = ((s7_d_p_t)opt_direct_x(cddr(arg)))(find_symbol_unchecked(sc, cadr(caddr(arg))));
  return(((s7_p_dd_t)opt_direct_x_call(cdr(arg)))(x1, x2)); 
}


static s7_pointer all_x_c_opsq_opsq(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  int32_t tx;
  tx = next_tx(sc);
  largs = cdr(arg);
  set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(car(largs))));
  sc->t_temps[tx] = c_call(car(largs))(sc, sc->t1_1);
  largs = cadr(largs);
  set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(largs)));
  set_car(sc->t2_2, c_call(largs)(sc, sc->t1_1));
  set_car(sc->t2_1, sc->t_temps[tx]);
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_c_opcq_opsq(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  int32_t tx;
  tx = next_tx(sc);
  largs = cdr(arg);
  sc->t_temps[tx] = c_call(car(largs))(sc, cdar(largs));
  largs = cadr(largs);
  set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(largs)));
  set_car(sc->t2_2, c_call(largs)(sc, sc->t1_1));
  set_car(sc->t2_1, sc->t_temps[tx]);
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_c_opsq_opcq(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  int32_t tx;
  tx = next_tx(sc);
  largs = cdr(arg);
  set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(car(largs))));
  sc->t_temps[tx] = c_call(car(largs))(sc, sc->t1_1);
  largs = cadr(largs);
  set_car(sc->t2_2, c_call(largs)(sc, cdr(largs)));
  set_car(sc->t2_1, sc->t_temps[tx]);
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_c_opsq_opssq(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  int32_t tx;
  tx = next_tx(sc);
  largs = cdr(arg);
  set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(car(largs))));
  sc->t_temps[tx] = c_call(car(largs))(sc, sc->t1_1);
  largs = cadr(largs);
  set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(largs)));
  set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(largs)));
  set_car(sc->t2_2, c_call(largs)(sc, sc->t2_1));
  set_car(sc->t2_1, sc->t_temps[tx]);
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_c_opssq_opsq(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  int32_t tx;
  tx = next_tx(sc);
  largs = cdr(arg);
  set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(car(largs))));
  set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(car(largs))));
  sc->t_temps[tx] = c_call(car(largs))(sc, sc->t2_1);
  largs = cadr(largs);
  set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(largs)));
  set_car(sc->t2_2, c_call(largs)(sc, sc->t1_1));
  set_car(sc->t2_1, sc->t_temps[tx]);
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_c_opssq_opcq(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  int32_t tx;
  tx = next_tx(sc);
  largs = cdr(arg);
  set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(car(largs))));
  set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(car(largs))));
  sc->t_temps[tx] = c_call(car(largs))(sc, sc->t2_1);
  largs = cadr(largs);
  set_car(sc->t2_2, c_call(largs)(sc, cdr(largs)));
  set_car(sc->t2_1, sc->t_temps[tx]);
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_c_opssq_opssq(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer largs;
  int32_t tx;
  tx = next_tx(sc);
  largs = cdr(arg);
  set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(car(largs))));
  set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(car(largs))));
  sc->t_temps[tx] = c_call(car(largs))(sc, sc->t2_1);
  largs = cadr(largs);
  set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(largs)));
  set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(largs)));
  set_car(sc->t2_2, c_call(largs)(sc, sc->t2_1));
  set_car(sc->t2_1, sc->t_temps[tx]);
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_c_op_opssq_q_c(s7_scheme *sc, s7_pointer code)
{
  s7_pointer arg;
  arg = cadadr(code);
  set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(arg)));
  set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(arg)));
  set_car(sc->t1_1, c_call(arg)(sc, sc->t2_1));
  set_car(sc->t2_1, c_call(cadr(code))(sc, sc->t1_1));
  set_car(sc->t2_2, caddr(code));
  return(c_call(code)(sc, sc->t2_1));
}

static s7_pointer all_x_c_op_opsq_q(s7_scheme *sc, s7_pointer code)
{
  s7_pointer arg;
  arg = cadadr(code);
  set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(arg)));
  set_car(sc->t1_1, c_call(arg)(sc, sc->t1_1));
  set_car(sc->t1_1, c_call(cadr(code))(sc, sc->t1_1));
  return(c_call(code)(sc, sc->t1_1));
}

static s7_pointer all_x_c_s_op_s_opsqq(s7_scheme *sc, s7_pointer code)
{
  s7_pointer args, val, val1;
  args = caddr(code); 
  val1 = caddr(args);
  val = find_symbol_unchecked(sc, cadr(args));
  set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(val1))); 
  set_car(sc->t2_2, c_call(val1)(sc, sc->t1_1));
  set_car(sc->t2_1, val);
  set_car(sc->t2_2, c_call(args)(sc, sc->t2_1));
  set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(code))); 
  return(c_call(code)(sc, sc->t2_1)); 
}

static s7_pointer all_x_c_op_opsq_q_s(s7_scheme *sc, s7_pointer code)
{
  s7_pointer arg;
  arg = cadadr(code);
  set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(arg)));
  set_car(sc->t1_1, c_call(arg)(sc, sc->t1_1));
  set_car(sc->t2_1, c_call(cadr(code))(sc, sc->t1_1));
  set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(code)));
  return(c_call(code)(sc, sc->t2_1));
}

static s7_pointer all_x_c_op_opsq_q_c(s7_scheme *sc, s7_pointer code)
{
  s7_pointer arg;
  arg = cadadr(code);
  set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(arg)));
  set_car(sc->t1_1, c_call(arg)(sc, sc->t1_1));
  set_car(sc->t2_1, c_call(cadr(code))(sc, sc->t1_1));
  set_car(sc->t2_2, caddr(code));
  return(c_call(code)(sc, sc->t2_1));
}

static s7_pointer all_x_c_a(s7_scheme *sc, s7_pointer arg)
{
  set_car(sc->t1_1, c_call(cdr(arg))(sc, cadr(arg))); 
  return(c_call(arg)(sc, sc->t1_1));
}

static s7_pointer all_x_c_ssa(s7_scheme *sc, s7_pointer arg)
{
  int32_t tx;
  tx = next_tx(sc);
  sc->t_temps[tx] = c_call(cdddr(arg))(sc, cadddr(arg));
  set_car(sc->t3_1, find_symbol_unchecked(sc, cadr(arg)));
  set_car(sc->t3_2, find_symbol_unchecked(sc, caddr(arg)));
  set_car(sc->t3_3, sc->t_temps[tx]);
  return(c_call(arg)(sc, sc->t3_1));
}

static s7_pointer all_x_c_sas(s7_scheme *sc, s7_pointer arg)
{
  int32_t tx;
  tx = next_tx(sc);
  sc->t_temps[tx] = c_call(cddr(arg))(sc, caddr(arg));
  set_car(sc->t3_1, find_symbol_unchecked(sc, cadr(arg)));
  set_car(sc->t3_3, find_symbol_unchecked(sc, cadddr(arg)));
  set_car(sc->t3_2, sc->t_temps[tx]);
  return(c_call(arg)(sc, sc->t3_1));
}

static s7_pointer all_x_c_sca(s7_scheme *sc, s7_pointer arg)
{
  int32_t tx;
  tx = next_tx(sc);
  sc->t_temps[tx] = c_call(cdddr(arg))(sc, cadddr(arg));
  set_car(sc->t3_1, find_symbol_unchecked(sc, cadr(arg)));
  set_car(sc->t3_2, caddr(arg));
  set_car(sc->t3_3, sc->t_temps[tx]);
  return(c_call(arg)(sc, sc->t3_1));
}

static s7_pointer all_x_c_csa(s7_scheme *sc, s7_pointer arg)
{
  int32_t tx;
  tx = next_tx(sc);
  sc->t_temps[tx] = c_call(cdddr(arg))(sc, cadddr(arg));
  set_car(sc->t3_1, cadr(arg));
  set_car(sc->t3_2, find_symbol_unchecked(sc, caddr(arg)));
  set_car(sc->t3_3, sc->t_temps[tx]);
  return(c_call(arg)(sc, sc->t3_1));
}

static s7_pointer all_x_c_ca(s7_scheme *sc, s7_pointer arg)
{
  int32_t tx;
  tx = next_tx(sc);
  sc->t_temps[tx] = c_call(cddr(arg))(sc, caddr(arg));
  set_car(sc->t2_1, cadr(arg));
  set_car(sc->t2_2, sc->t_temps[tx]);
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_c_ac(s7_scheme *sc, s7_pointer arg)
{
  int32_t tx;
  tx = next_tx(sc);
  sc->t_temps[tx] = c_call(cdr(arg))(sc, cadr(arg));
  set_car(sc->t2_2, caddr(arg));
  set_car(sc->t2_1, sc->t_temps[tx]);
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_c_sa(s7_scheme *sc, s7_pointer arg)
{
  int32_t tx;
  tx = next_tx(sc);
  sc->t_temps[tx] = c_call(cddr(arg))(sc, caddr(arg));
  set_car(sc->t2_1, find_symbol_checked(sc, cadr(arg)));
  set_car(sc->t2_2, sc->t_temps[tx]);
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_c_as(s7_scheme *sc, s7_pointer arg)
{
  int32_t tx;
  tx = next_tx(sc);
  sc->t_temps[tx] = c_call(cdr(arg))(sc, cadr(arg));
  set_car(sc->t2_2, find_symbol_checked(sc, caddr(arg)));
  set_car(sc->t2_1, sc->t_temps[tx]);
  return(c_call(arg)(sc, sc->t2_1));
}

static s7_pointer all_x_if_x2(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer p;
  p = cdr(arg);
  if (is_true(sc, c_call(p)(sc, car(p))))
    p = cdr(p);
  else p = cddr(p);
  return(c_call(p)(sc, car(p)));
}

static s7_pointer all_x_and2(s7_scheme *sc, s7_pointer arg)
{
  /* arg is the full expr: (and ...) */
  s7_pointer p, val;
  p = cdr(arg);
  val = c_call(p)(sc, car(p));
  if (val == sc->F) return(val);
  p = cdr(p);
  return(c_call(p)(sc, car(p)));
}

static s7_pointer all_x_and3(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer p, val;
  p = cdr(arg);
  val = c_call(p)(sc, car(p));
  if (val == sc->F) return(val);
  p = cdr(p);
  val = c_call(p)(sc, car(p));
  if (val == sc->F) return(val);
  p = cdr(p);
  return(c_call(p)(sc, car(p)));
}

static s7_pointer all_x_or2(s7_scheme *sc, s7_pointer arg)
{
  s7_pointer p, val;
  p = cdr(arg);
  val = c_call(p)(sc, car(p));
  if (val != sc->F) return(val);
  p = cdr(p);
  return(c_call(p)(sc, car(p)));
}

static s7_pointer all_x_closure_s(s7_scheme *sc, s7_pointer code)
{
  /* no gain from all_x for the body rather than safe_c_c */
  s7_pointer result, old_e;
  old_e = sc->envir;
  sc->envir = old_frame_with_slot(sc, closure_let(opt_lambda(code)), find_symbol_unchecked(sc, opt_sym2(code)));
  code = car(closure_body(opt_lambda(code)));
  result = c_call(code)(sc, cdr(code));
  sc->envir = old_e;
  return(result);
}

static s7_pointer all_x_and_2_closure_s(s7_scheme *sc, s7_pointer code)
{
  s7_pointer result, old_e;

  old_e = sc->envir;
  sc->envir = old_frame_with_slot(sc, closure_let(opt_lambda(code)), find_symbol_unchecked(sc, opt_sym2(code)));

  code = cdar(closure_body(opt_lambda(code)));
  result = c_call(code)(sc, car(code));
  if (result != sc->F)
    result = c_call(cdr(code))(sc, cadr(code));

  sc->envir = old_e;
  return(result);
}

static s7_pointer all_x_closure_a(s7_scheme *sc, s7_pointer code)
{
  s7_pointer result, old_e;
  old_e = sc->envir;
  sc->envir = old_frame_with_slot(sc, closure_let(opt_lambda(code)), c_call(cdr(code))(sc, cadr(code)));
  code = car(closure_body(opt_lambda(code)));
  result = c_call(code)(sc, cdr(code));
  sc->envir = old_e;
  return(result);
}

static void all_x_function_init(void)
{
  int32_t i;
  for (i = 0; i < OPT_MAX_DEFINED; i++)
    all_x_function[i] = NULL;

  all_x_function[HOP_SAFE_C_C] = all_x_c_c;
  all_x_function[HOP_SAFE_C_OR2] = all_x_or2;
  all_x_function[HOP_SAFE_C_AND2] = all_x_and2;
  all_x_function[HOP_SAFE_C_A] = all_x_c_a;
  all_x_function[HOP_SAFE_C_S] = all_x_c_s;
  all_x_function[HOP_SAFE_C_L] = local_x_c_s;
  all_x_function[HOP_SAFE_CAR_S] = all_x_car_s;
  all_x_function[HOP_SAFE_CDR_S] = all_x_cdr_s;
  all_x_function[HOP_SAFE_CADR_S] = all_x_cadr_s;
  all_x_function[HOP_SAFE_IS_PAIR_S] = all_x_is_pair_s;
  all_x_function[HOP_SAFE_IS_NULL_S] = all_x_is_null_s;
  all_x_function[HOP_SAFE_IS_SYMBOL_S] = all_x_is_symbol_s;

  all_x_function[HOP_SAFE_C_opCq] = all_x_c_opcq;
  all_x_function[HOP_SAFE_C_opSq] = all_x_c_opsq;
  all_x_function[HOP_SAFE_C_opSSq] = all_x_c_opssq;
  all_x_function[HOP_SAFE_C_opSCq] = all_x_c_opscq;
  all_x_function[HOP_SAFE_C_opCSq] = all_x_c_opcsq;
  all_x_function[HOP_SAFE_C_opSQq] = all_x_c_opsqq;
  all_x_function[HOP_SAFE_C_opQSq] = all_x_c_opqsq;

  all_x_function[HOP_SAFE_C_SC] = all_x_c_sc;
  all_x_function[HOP_SAFE_C_CS] = all_x_c_cs;
  all_x_function[HOP_SAFE_C_CQ] = all_x_c_cq;
  all_x_function[HOP_SAFE_C_SQ] = all_x_c_sq;
  all_x_function[HOP_SAFE_C_QS] = all_x_c_qs;
  all_x_function[HOP_SAFE_C_SS] = all_x_c_ss;
  all_x_function[HOP_SAFE_C_LL] = local_x_c_ss;

  all_x_function[HOP_SAFE_C_opSq_S] = all_x_c_opsq_s;
  all_x_function[HOP_SAFE_C_opSq_C] = all_x_c_opsq_c;
  all_x_function[HOP_SAFE_C_opSq_Q] = all_x_c_opsq_q;
  all_x_function[HOP_SAFE_C_opSq_QS] = all_x_c_opsq_qs;
  all_x_function[HOP_SAFE_C_S_opSq] = all_x_c_s_opsq;
  all_x_function[HOP_SAFE_C_S_opCq] = all_x_c_s_opcq;
  all_x_function[HOP_SAFE_C_opCq_S] = all_x_c_opcq_s;
  all_x_function[HOP_SAFE_C_opCq_C] = all_x_c_opcq_c;
  all_x_function[HOP_SAFE_C_C_opSq] = all_x_c_c_opsq;
  all_x_function[HOP_SAFE_C_C_opCq] = all_x_c_c_opcq;
  all_x_function[HOP_SAFE_C_opCSq_C] = all_x_c_opcsq_c;
  all_x_function[HOP_SAFE_C_opSSq_C] = all_x_c_opssq_c;
  all_x_function[HOP_SAFE_C_opSCq_S] = all_x_c_opscq_s;
  all_x_function[HOP_SAFE_C_opSCq_C] = all_x_c_opscq_c;
  all_x_function[HOP_SAFE_C_opSSq_S] = all_x_c_opssq_s;
  all_x_function[HOP_SAFE_C_S_opSSq] = all_x_c_s_opssq;
  all_x_function[HOP_SAFE_C_C_opSSq] = all_x_c_c_opssq;
  all_x_function[HOP_SAFE_C_S_opSCq] = all_x_c_s_opscq;
  all_x_function[HOP_SAFE_C_opSq_opSq] = all_x_c_opsq_opsq;
  all_x_function[HOP_SAFE_C_opSq_opCq] = all_x_c_opsq_opcq;
  all_x_function[HOP_SAFE_C_opCq_opSq] = all_x_c_opcq_opsq;
  all_x_function[HOP_SAFE_C_opSq_opSSq] = all_x_c_opsq_opssq;
  all_x_function[HOP_SAFE_C_opCq_opCq] = all_x_c_opcq_opcq;
  all_x_function[HOP_SAFE_C_opSSq_opSq] = all_x_c_opssq_opsq;
  all_x_function[HOP_SAFE_C_opSSq_opCq] = all_x_c_opssq_opcq;
  all_x_function[HOP_SAFE_C_opSSq_opSSq] = all_x_c_opssq_opssq;
  all_x_function[HOP_SAFE_C_op_opSSq_q_C] = all_x_c_op_opssq_q_c;
  all_x_function[HOP_SAFE_C_op_opSq_q] = all_x_c_op_opsq_q;
  all_x_function[HOP_SAFE_C_op_opSq_q_S] = all_x_c_op_opsq_q_s;
  all_x_function[HOP_SAFE_C_op_opSq_q_C] = all_x_c_op_opsq_q_c;
  all_x_function[HOP_SAFE_C_op_S_opSq_q] = all_x_c_op_s_opsq_q;
  all_x_function[HOP_SAFE_C_op_opSq_S_q] = all_x_c_op_opsq_s_q;
  all_x_function[HOP_SAFE_C_S_op_S_opSqq] = all_x_c_s_op_s_opsqq;
  all_x_function[HOP_SAFE_C_CSA] = all_x_c_csa;
  all_x_function[HOP_SAFE_C_SCA] = all_x_c_sca;
  all_x_function[HOP_SAFE_C_SAS] = all_x_c_sas;
  all_x_function[HOP_SAFE_C_SSA] = all_x_c_ssa;
  all_x_function[HOP_SAFE_C_SSC] = all_x_c_ssc;
  all_x_function[HOP_SAFE_C_SSS] = all_x_c_sss;
  all_x_function[HOP_SAFE_C_SCS] = all_x_c_scs;
  all_x_function[HOP_SAFE_C_CSS] = all_x_c_css;
  all_x_function[HOP_SAFE_C_CSC] = all_x_c_csc;
  all_x_function[HOP_SAFE_CLOSURE_S_C] = all_x_closure_s;
  all_x_function[HOP_SAFE_CLOSURE_A_C] = all_x_closure_a;
  all_x_function[HOP_SAFE_C_ALL_S] = all_x_c_all_s;
  all_x_function[HOP_SAFE_QUOTE] = all_x_q;
}

static s7_pointer g_not_c_c(s7_scheme *sc, s7_pointer args);
static s7_pointer g_not_is_pair_s(s7_scheme *sc, s7_pointer args);
static s7_pointer g_is_pair_cdr_s(s7_scheme *sc, s7_pointer args);
static s7_pointer g_is_eq_car(s7_scheme *sc, s7_pointer args);
static s7_pointer g_is_eq_car_q(s7_scheme *sc, s7_pointer args);
static s7_pointer g_if_x2(s7_scheme *sc, s7_pointer args);
static s7_pointer g_and_2(s7_scheme *sc, s7_pointer args);
static s7_pointer g_and_3(s7_scheme *sc, s7_pointer args);
static s7_pointer g_or_2(s7_scheme *sc, s7_pointer args);
static inline s7_pointer check_quote(s7_scheme *sc, s7_pointer code);

static s7_function all_x_eval(s7_scheme *sc, s7_pointer holder, s7_pointer e, safe_sym_t *checker)
{
  s7_pointer arg;
  arg = car(holder);
  if (is_pair(arg))
    {
      if (is_optimized(arg))
	{
	  switch (optimize_op(arg))
	    {
	    case HOP_SAFE_C_C:
	      if (c_call(arg) == g_not_is_pair_s)
		{
		  set_opt_sym3(arg, cadadr(arg));
		  return((is_local_symbol(cdadr(arg))) ? local_x_not_is_pair_s : all_x_not_is_pair_s);
		}
	      if (c_call(arg) == g_is_pair_cdr_s)
		{
		  set_opt_sym2(cdr(arg), cadadr(arg));
		  return((is_local_symbol(cdadr(arg))) ? local_x_is_pair_cdr_s : all_x_is_pair_cdr_s);
		}
	      if (c_call(arg) == g_add_cs1) return(all_x_c_add1);
	      if (c_call(arg) == g_add_cl1) return(local_x_c_add1);
	      if (c_call(arg) == g_subtract_cl1) return(local_x_c_sub1);
	      if (c_call(arg) == g_if_x2) return(all_x_if_x2);	      /* g_if_x1 doesn't happen much */
	      if (c_call(arg) == g_and_2) return(all_x_and2);
	      if (c_call(arg) == g_or_2) return(all_x_or2);
	      if (c_call(arg) == g_and_3) return(all_x_and3);
	      if ((c_call(arg) == g_add_si) &&
		  (checker(sc, cadr(arg), e)))
		return((is_local_symbol(cdr(arg))) ? local_x_c_addi : all_x_c_addi);
	      if ((c_call(arg) == g_char_equal_s_ic) &&
		  (checker(sc, cadr(arg), e)))
		return(all_x_c_char_eq);
	      if (c_call(arg) == g_not_c_c)
		{
		  if (c_call(cadr(arg)) == g_is_eq_car_q)
		    {
		      set_opt_pair2(cdr(arg), cdadr(arg));
		      return((is_local_symbol(cdr(cadadr(arg)))) ? local_x_not_is_eq_car_q : all_x_not_is_eq_car_q);
		    }
		  return(all_x_not_c_c);
		}
	      if (c_call(arg) == g_hash_table_ref_ss)
		return((is_local_symbol(cdr(arg)) && is_local_symbol(cddr(arg))) ? local_x_c_hash_table_ref_ss : all_x_c_hash_table_ref_ss);
	      if (c_call(arg) == g_hash_table_ref_car)
		return(all_x_c_hash_table_ref_car);
	      return(all_x_c_c);
	      
	    case HOP_SAFE_C_S:
	      {
		bool is_local;
		is_local = (is_local_symbol(cdr(arg)) || (is_immutable(cadr(arg))));
		if (car(arg) == sc->cdr_symbol) return((is_local) ? local_x_cdr_s : all_x_cdr_s);
		if (car(arg) == sc->car_symbol) return((is_local) ? local_x_car_s : all_x_car_s);
		if (car(arg) == sc->cadr_symbol) return((is_local) ? local_x_cadr_s : all_x_cadr_s);
		if (car(arg) == sc->is_null_symbol) return((is_local) ? local_x_is_null_s : all_x_is_null_s);
		if (car(arg) == sc->is_pair_symbol) return((is_local) ? local_x_is_pair_s : all_x_is_pair_s);
		if (car(arg) == sc->is_symbol_symbol) return((is_local) ? local_x_is_symbol_s : all_x_is_symbol_s);
		if (car(arg) == sc->is_keyword_symbol) return(all_x_is_keyword_s);
		if (car(arg) == sc->is_integer_symbol) return(all_x_is_integer_s);
		if (car(arg) == sc->is_procedure_symbol) return(all_x_is_procedure_s);
		if (car(arg) == sc->is_string_symbol) return(all_x_is_string_s);
		if (car(arg) == sc->is_vector_symbol) return(all_x_is_vector_s);
		if (car(arg) == sc->is_proper_list_symbol) return(all_x_is_proper_list_s);
		if (car(arg) == sc->not_symbol) return(all_x_not_s);
		return((is_local) ? local_x_c_s : all_x_c_s);
	      }

	    case HOP_SAFE_C_L:
	      if (car(arg) == sc->cdr_symbol) return(local_x_cdr_s);
	      if (car(arg) == sc->car_symbol) return(local_x_car_s);
	      if (car(arg) == sc->cadr_symbol) return(local_x_cadr_s);
	      if (car(arg) == sc->is_null_symbol) return(local_x_is_null_s);
	      if (car(arg) == sc->is_pair_symbol) return(local_x_is_pair_s);
	      if (car(arg) == sc->is_symbol_symbol) return(local_x_is_symbol_s);
	      return(local_x_c_s);
	      
	    case HOP_SAFE_C_opSq:
	      if (car(arg) == sc->not_symbol)   return((is_local_symbol(cdadr(arg))) ? local_x_c_not_opsq : all_x_c_not_opsq);
	      if (caadr(arg) == sc->car_symbol) 
		{
		  set_opt_sym2(cdr(arg), cadadr(arg));
		  return((is_local_symbol(cdr(cadr(arg)))) ? local_x_c_car_s : all_x_c_car_s);
		}
	      if (caadr(arg) == sc->cdr_symbol) 
		{
		  set_opt_sym2(cdr(arg), cadadr(arg));
		  return((is_local_symbol(cdadr(arg))) ? local_x_c_cdr_s : all_x_c_cdr_s);
		}
	      return((is_local_symbol(cdr(cadr(arg)))) ? local_x_c_opsq : all_x_c_opsq);

	    case HOP_SAFE_C_opSq_C:
	      return((is_local_symbol(cdr(cadr(arg)))) ? local_x_c_opsq_c : all_x_c_opsq_c);

	    case HOP_SAFE_C_SC:
	      return((is_local_symbol(cdr(arg))) ? local_x_c_sc : all_x_c_sc);
	    case HOP_SAFE_C_SQ:
	      set_opt_con2(cdr(arg), cadr(caddr(arg)));
	      return((is_local_symbol(cdr(arg))) ? local_x_c_sq : all_x_c_sq);
	    case HOP_SAFE_C_CQ:
	      set_opt_con2(cdr(arg), cadr(caddr(arg)));
	      return(all_x_c_cq);
	      
	    case HOP_SAFE_C_SS:
	      if ((is_local_symbol(cdr(arg))) &&
		  (is_local_symbol(cddr(arg))))
		return(local_x_c_ss);
	      if (is_local_symbol(cdr(arg)))
		return(all_x_c_ls);
	      return(all_x_c_ss);
	      
	    case HOP_SAFE_C_SSS:
	      if ((is_local_symbol(cdr(arg))) &&
		  (is_local_symbol(cddr(arg))) &&
		  (is_local_symbol(cdddr(arg))))
		return(local_x_c_sss);
	      return(all_x_c_sss);
	      
	    case HOP_SAFE_C_opSSq:
	      if (car(arg) == sc->not_symbol)
		return(all_x_c_not_opssq);
	      if ((is_local_symbol(cdr(cadr(arg)))) &&
		  (is_local_symbol(cddr(cadr(arg)))))
		return(local_x_c_opssq);
	      return(all_x_c_opssq);

	    case HOP_SAFE_C_opSSq_S:
	      if ((is_local_symbol(cdr(cadr(arg)))) &&
		  (is_local_symbol(cddr(cadr(arg)))) &&
		  (is_local_symbol(cddr(arg))))
		return(local_x_c_opssq_s);
	      return(all_x_c_opssq_s);
		  
	    case HOP_SAFE_C_S_opSSq:
	      if ((is_local_symbol(cdr(caddr(arg)))) &&
		  (is_local_symbol(cddr(caddr(arg)))) &&
		  (is_local_symbol(cdr(arg))))
		return(local_x_c_s_opssq);
	      return(all_x_c_s_opssq);
		  
	    case HOP_SAFE_C_S_opSq:
	      if ((is_local_symbol(cdr(caddr(arg)))) &&
		  (is_local_symbol(cdr(arg))))
		return(local_x_c_s_opsq);
	      return(all_x_c_s_opsq);

	    case HOP_SAFE_C_C_opSSq:
	      if (has_direct_x_opt(arg))
		return(direct_x_c_c_opssq);
	      return(all_x_c_c_opssq);

	    case HOP_SAFE_C_opSq_opSq:
	      if (has_direct_x_opt(arg))
		return(direct_x_c_opsq_opsq);
	      return(all_x_c_opsq_opsq);

	    case HOP_SAFE_C_AA:
	      if (is_symbol(cadr(arg)))
		return(all_x_c_sa);
	      if (is_symbol(caddr(arg)))
		return(all_x_c_as);
	      if (!is_pair(cadr(arg)))
		return(all_x_c_ca);
	      if (!is_pair(caddr(arg)))
		return(all_x_c_ac);
	      return(NULL);
	      
	    case HOP_SAFE_CLOSURE_S_C:
	      if (c_call(car(closure_body(opt_lambda(arg)))) == g_and_2)
		return(all_x_and_2_closure_s);
	      return(all_x_closure_s);

	    default:
	      /* if (!all_x_function[optimize_op(arg)]) fprintf(stderr, "all_x_eval %s %s\n", DISPLAY(arg), (is_optimized(arg)) ? opt_names[optimize_op(arg)] : "unopt"); */
	      return(all_x_function[optimize_op(arg)]);
	    }
	} /* is_optimized */
      if (car(arg) == sc->quote_symbol)
	{
	  check_quote(sc, cdr(arg));
	  return(all_x_q);
	}
      /* fprintf(stderr, "all_x_eval %s %d %d %s\n", DISPLAY(arg), car(arg) == sc->quote_symbol, is_optimized(arg), (is_optimized(arg)) ? opt_names[optimize_op(arg)] : "unopt"); */
      return(NULL);
    }
  if (is_symbol(arg))
    {
      if ((is_keyword(arg)) ||
	  ((arg == sc->else_symbol) &&
	   (is_global(arg))))
	return(all_x_k);
      if (is_local_symbol(holder))
	return(local_x_s);
      if (checker(sc, arg, e))
	return(all_x_s);
      return(all_x_unsafe_s);
    }
  return(all_x_c);
}


/* -------------------------------------------------------------------------------- */

enum {o_d_v, o_d_vd, o_d_vdd, o_d_vid, o_d_id, o_d_pi, o_d_ip, o_d_pd, o_d_pid, o_d, o_d_d, o_d_dd, o_d_ddd, o_d_dddd, 
      o_i_d, o_i_i, o_i_ii, o_i_iii, o_i_p, o_i_pi, o_i_pii, o_d_p, o_b_p, o_b_pp, o_b_pp_direct, o_b_pi, o_b_ii, o_b_dd, 
      o_p, o_p_p, o_p_ii, o_p_dd,
      o_p_pp, o_p_pp_direct, o_p_ppp, o_p_ppp_direct, o_p_pi, o_p_pi_direct, o_p_ppi, o_p_pip, o_p_pip_direct, o_b_i, o_b_d};

static void add_opt_func(s7_pointer f, int32_t typ, void *func)
{
  if (is_c_function(f))
    {
      opt_funcs *op;
      op = (opt_funcs *)malloc(sizeof(opt_funcs));
      op->typ = typ;
      op->func = func;
      op->next = c_function_opt_data(f);
      c_function_opt_data(f) = op;
    }
}

static void *opt_func(s7_pointer f, int32_t typ)
{
  if (is_c_function(f))
    {
      opt_funcs *p;
      for (p = c_function_opt_data(f); p; p = p->next)
	if (p->typ == typ)
	  return(p->func);
    }
  return(NULL);
}

/* clm2xen.c */
void s7_set_d_function(s7_pointer f, s7_d_t df) {add_opt_func(f, o_d, (void *)df);}
s7_d_t s7_d_function(s7_pointer f) {return((s7_d_t)opt_func(f, o_d));}

void s7_set_d_d_function(s7_pointer f, s7_d_d_t df) {add_opt_func(f, o_d_d, (void *)df);}
s7_d_d_t s7_d_d_function(s7_pointer f) {return((s7_d_d_t)opt_func(f, o_d_d));}

void s7_set_d_dd_function(s7_pointer f, s7_d_dd_t df) {add_opt_func(f, o_d_dd, (void *)df);}
s7_d_dd_t s7_d_dd_function(s7_pointer f) {return((s7_d_dd_t)opt_func(f, o_d_dd));}

void s7_set_d_v_function(s7_pointer f, s7_d_v_t df) {add_opt_func(f, o_d_v, (void *)df);}
s7_d_v_t s7_d_v_function(s7_pointer f) {return((s7_d_v_t)opt_func(f, o_d_v));}

void s7_set_d_vd_function(s7_pointer f, s7_d_vd_t df) {add_opt_func(f, o_d_vd, (void *)df);}
s7_d_vd_t s7_d_vd_function(s7_pointer f) {return((s7_d_vd_t)opt_func(f, o_d_vd));}

void s7_set_d_vdd_function(s7_pointer f, s7_d_vdd_t df) {add_opt_func(f, o_d_vdd, (void *)df);}
s7_d_vdd_t s7_d_vdd_function(s7_pointer f) {return((s7_d_vdd_t)opt_func(f, o_d_vdd));}

void s7_set_d_vid_function(s7_pointer f, s7_d_vid_t df) {add_opt_func(f, o_d_vid, (void *)df);}
s7_d_vid_t s7_d_vid_function(s7_pointer f) {return((s7_d_vid_t)opt_func(f, o_d_vid));}

void s7_set_d_id_function(s7_pointer f, s7_d_id_t df) {add_opt_func(f, o_d_id, (void *)df);}
s7_d_id_t s7_d_id_function(s7_pointer f) {return((s7_d_id_t)opt_func(f, o_d_id));}

void s7_set_d_pid_function(s7_pointer f, s7_d_pid_t df) {add_opt_func(f, o_d_pid, (void *)df);}
s7_d_pid_t s7_d_pid_function(s7_pointer f) {return((s7_d_pid_t)opt_func(f, o_d_pid));}

void s7_set_d_ip_function(s7_pointer f, s7_d_ip_t df) {add_opt_func(f, o_d_ip, (void *)df);}
s7_d_ip_t s7_d_ip_function(s7_pointer f) {return((s7_d_ip_t)opt_func(f, o_d_ip));}

void s7_set_d_pd_function(s7_pointer f, s7_d_pd_t df) {add_opt_func(f, o_d_pd, (void *)df);}
s7_d_pd_t s7_d_pd_function(s7_pointer f) {return((s7_d_pd_t)opt_func(f, o_d_pd));}

void s7_set_i_p_function(s7_pointer f, s7_i_p_t df) {add_opt_func(f, o_i_p, (void *)df);}
s7_i_p_t s7_i_p_function(s7_pointer f) {return((s7_i_p_t)opt_func(f, o_i_p));}

void s7_set_d_p_function(s7_pointer f, s7_d_p_t df) {add_opt_func(f, o_d_p, (void *)df);}
s7_d_p_t s7_d_p_function(s7_pointer f) {return((s7_d_p_t)opt_func(f, o_d_p));}

void s7_set_b_p_function(s7_pointer f, s7_b_p_t df) {add_opt_func(f, o_b_p, (void *)df);}
s7_b_p_t s7_b_p_function(s7_pointer f) {return((s7_b_p_t)opt_func(f, o_b_p));}

void s7_set_d_pi_function(s7_pointer f, s7_d_pi_t df) {add_opt_func(f, o_d_pi, (void *)df);}
s7_d_pi_t s7_d_pi_function(s7_pointer f) {return((s7_d_pi_t)opt_func(f, o_d_pi));}

/* cload.scm */
void s7_set_d_ddd_function(s7_pointer f, s7_d_ddd_t df) {add_opt_func(f, o_d_ddd, (void *)df);}
s7_d_ddd_t s7_d_ddd_function(s7_pointer f) {return((s7_d_ddd_t)opt_func(f, o_d_ddd));}

void s7_set_d_dddd_function(s7_pointer f, s7_d_dddd_t df) {add_opt_func(f, o_d_dddd, (void *)df);}
s7_d_dddd_t s7_d_dddd_function(s7_pointer f) {return((s7_d_dddd_t)opt_func(f, o_d_dddd));}

void s7_set_i_i_function(s7_pointer f, s7_i_i_t df) {add_opt_func(f, o_i_i, (void *)df);}
s7_i_i_t s7_i_i_function(s7_pointer f) {return((s7_i_i_t)opt_func(f, o_i_i));}

void s7_set_i_ii_function(s7_pointer f, s7_i_ii_t df) {add_opt_func(f, o_i_ii, (void *)df);}
s7_i_ii_t s7_i_ii_function(s7_pointer f) {return((s7_i_ii_t)opt_func(f, o_i_ii));}

void s7_set_i_d_function(s7_pointer f, s7_i_d_t df) {add_opt_func(f, o_i_d, (void *)df);}
s7_i_d_t s7_i_d_function(s7_pointer f) {return((s7_i_d_t)opt_func(f, o_i_d));}

static void s7_set_i_iii_function(s7_pointer f, s7_i_iii_t df) {add_opt_func(f, o_i_iii, (void *)df);}
s7_i_iii_t s7_i_iii_function(s7_pointer f) {return((s7_i_iii_t)opt_func(f, o_i_iii));}

static void s7_set_p_pi_function(s7_pointer f, s7_p_pi_t df) {add_opt_func(f, o_p_pi, (void *)df);}
static s7_p_pi_t s7_p_pi_function(s7_pointer f) {return((s7_p_pi_t)opt_func(f, o_p_pi));}

static void s7_set_p_ppi_function(s7_pointer f, s7_p_ppi_t df) {add_opt_func(f, o_p_ppi, (void *)df);}
static s7_p_ppi_t s7_p_ppi_function(s7_pointer f) {return((s7_p_ppi_t)opt_func(f, o_p_ppi));}

static void s7_set_i_pi_function(s7_pointer f, s7_i_pi_t df) {add_opt_func(f, o_i_pi, (void *)df);}
static s7_i_pi_t s7_i_pi_function(s7_pointer f) {return((s7_i_pi_t)opt_func(f, o_i_pi));}

static void s7_set_i_pii_function(s7_pointer f, s7_i_pii_t df) {add_opt_func(f, o_i_pii, (void *)df);}
static s7_i_pii_t s7_i_pii_function(s7_pointer f) {return((s7_i_pii_t)opt_func(f, o_i_pii));}

static void s7_set_b_d_function(s7_pointer f, s7_b_d_t df) {add_opt_func(f, o_b_d, (void *)df);}
static s7_b_d_t s7_b_d_function(s7_pointer f) {return((s7_b_d_t)opt_func(f, o_b_d));}

static void s7_set_b_i_function(s7_pointer f, s7_b_i_t df) {add_opt_func(f, o_b_i, (void *)df);}
static s7_b_i_t s7_b_i_function(s7_pointer f) {return((s7_b_i_t)opt_func(f, o_b_i));}

static void s7_set_b_pp_function(s7_pointer f, s7_b_pp_t df) {add_opt_func(f, o_b_pp, (void *)df);}
static s7_b_pp_t s7_b_pp_function(s7_pointer f) {return((s7_b_pp_t)opt_func(f, o_b_pp));}

#if (!WITH_GMP)
static void s7_set_b_pi_function(s7_pointer f, s7_b_pi_t df) {add_opt_func(f, o_b_pi, (void *)df);}
#endif
static s7_b_pi_t s7_b_pi_function(s7_pointer f) {return((s7_b_pi_t)opt_func(f, o_b_pi));}

static void s7_set_b_ii_function(s7_pointer f, s7_b_ii_t df) {add_opt_func(f, o_b_ii, (void *)df);}
static s7_b_ii_t s7_b_ii_function(s7_pointer f) {return((s7_b_ii_t)opt_func(f, o_b_ii));}

static void s7_set_b_dd_function(s7_pointer f, s7_b_dd_t df) {add_opt_func(f, o_b_dd, (void *)df);}
static s7_b_dd_t s7_b_dd_function(s7_pointer f) {return((s7_b_dd_t)opt_func(f, o_b_dd));}

static void s7_set_p_p_function(s7_pointer f, s7_p_p_t df) {add_opt_func(f, o_p_p, (void *)df);}
static s7_p_p_t s7_p_p_function(s7_pointer f) {return((s7_p_p_t)opt_func(f, o_p_p));}

static void s7_set_p_function(s7_pointer f, s7_p_t df) {add_opt_func(f, o_p, (void *)df);}
static s7_p_t s7_p_function(s7_pointer f) {return((s7_p_t)opt_func(f, o_p));}

static void s7_set_p_pp_function(s7_pointer f, s7_p_pp_t df) {add_opt_func(f, o_p_pp, (void *)df);}
static s7_p_pp_t s7_p_pp_function(s7_pointer f) {return((s7_p_pp_t)opt_func(f, o_p_pp));}

static void s7_set_p_ppp_function(s7_pointer f, s7_p_ppp_t df) {add_opt_func(f, o_p_ppp, (void *)df);}
static s7_p_ppp_t s7_p_ppp_function(s7_pointer f) {return((s7_p_ppp_t)opt_func(f, o_p_ppp));}

static void s7_set_p_pip_function(s7_pointer f, s7_p_pip_t df) {add_opt_func(f, o_p_pip, (void *)df);}
static s7_p_pip_t s7_p_pip_function(s7_pointer f) {return((s7_p_pip_t)opt_func(f, o_p_pip));}

static void s7_set_p_pi_direct_function(s7_pointer f, s7_p_pi_t df) {add_opt_func(f, o_p_pi_direct, (void *)df);}
static s7_p_pi_t s7_p_pi_direct_function(s7_pointer f) {return((s7_p_pi_t)opt_func(f, o_p_pi_direct));}

static void s7_set_p_pip_direct_function(s7_pointer f, s7_p_pip_t df) {add_opt_func(f, o_p_pip_direct, (void *)df);}
static s7_p_pip_t s7_p_pip_direct_function(s7_pointer f) {return((s7_p_pip_t)opt_func(f, o_p_pip_direct));}

static void s7_set_p_pp_direct_function(s7_pointer f, s7_p_pp_t df) {add_opt_func(f, o_p_pp_direct, (void *)df);}
static s7_p_pp_t s7_p_pp_direct_function(s7_pointer f) {return((s7_p_pp_t)opt_func(f, o_p_pp_direct));}

static void s7_set_p_ppp_direct_function(s7_pointer f, s7_p_ppp_t df) {add_opt_func(f, o_p_ppp_direct, (void *)df);}
static s7_p_ppp_t s7_p_ppp_direct_function(s7_pointer f) {return((s7_p_ppp_t)opt_func(f, o_p_ppp_direct));}

static void s7_set_b_pp_direct_function(s7_pointer f, s7_b_pp_t df) {add_opt_func(f, o_b_pp_direct, (void *)df);}
static s7_b_pp_t s7_b_pp_direct_function(s7_pointer f) {return((s7_b_pp_t)opt_func(f, o_b_pp_direct));}

static void s7_set_p_ii_function(s7_pointer f, s7_p_ii_t df) {add_opt_func(f, o_p_ii, (void *)df);}
static s7_p_ii_t s7_p_ii_function(s7_pointer f) {return((s7_p_ii_t)opt_func(f, o_p_ii));}

#if (!WITH_GMP)
static void s7_set_p_dd_function(s7_pointer f, s7_p_dd_t df) {add_opt_func(f, o_p_dd, (void *)df);}
#endif
static s7_p_dd_t s7_p_dd_function(s7_pointer f) {return((s7_p_dd_t)opt_func(f, o_p_dd));}


#if DEBUGGING
static opt_info *alloc_opo(s7_scheme *sc, s7_pointer expr)
#else
#define alloc_opo(Sc, Expr) alloc_opo_1(Sc)
static opt_info *alloc_opo_1(s7_scheme *sc)
#endif
{
  opt_info *o;
  if (sc->pc >= OPTS_SIZE)
    {
#if DEBUGGING
      fprintf(stderr, "opts overflow: %s\n", DISPLAY(expr));
#endif
      longjmp(sc->opt_exit, 1);
    }
  o = sc->opts[sc->pc++];
  o->v8.fd = NULL;
#if DEBUGGING
  o->expr = expr;
#endif
  return(o);
}

#define OPT_PRINT 0

/* t600 has tests, t593 runs t*.scm */

static bool return_false(s7_scheme *sc, s7_pointer expr, const char *func, int32_t line)
{
#if OPT_PRINT
  /* if (strcmp(func, "cell_optimize") == 0) */
  fprintf(stderr, "   %s%s[%d]%s: %s\n", BOLD_TEXT, func, line, UNBOLD_TEXT, DISPLAY_80(expr));
#endif
  return(false);
}

#define is_opt_int(p) is_t_integer(p)
#define is_opt_real(p) is_real(p)


/* all_x fallback for all optimizers */
static s7_function all_x_optimize(s7_scheme *sc, s7_pointer expr)
{
  if ((is_optimized(car(expr))) &&
      (is_all_x_safe(sc, car(expr))))
    {
      sc->unwraps++;
      /* fprintf(stderr, "%s\n", DISPLAY_80(expr)); */
      /* lt: (if (eq? (var-name v) named-let) (values) (var-initial-value v))
       * s7test (eq? (car a) 'car) (list (cdr slot) '(gensym))
       * tgen: (out-any i x 0)
       */
      return(all_x_eval(sc, expr, sc->envir, let_symbol_is_safe));
    }
  return(NULL);
}

static s7_double opt_float_any(s7_scheme *sc, s7_pointer expr)
{
  /* caller for s7_float_optimize */
  cur_sc = sc;
  sc->pc = 0;
  return(sc->opts[0]->v7.fd(sc->opts[0]));
}

static s7_pointer opt_bool_any(s7_scheme *sc, s7_pointer expr)
{
  /* caller for s7_bool_optimize */
  cur_sc = sc;
  sc->pc = 0;
  return((sc->opts[0]->v7.fb(sc->opts[0])) ? sc->T : sc->F);
}

static s7_pointer opt_float_any_nr(s7_scheme *sc, s7_pointer expr)
{
  cur_sc = sc;
  sc->pc = 0;
  sc->opts[0]->v7.fd(sc->opts[0]);
  return(NULL);
}

static s7_pointer opt_int_any_nr(s7_scheme *sc, s7_pointer expr)
{
  cur_sc = sc;
  sc->pc = 0;
  sc->opts[0]->v7.fi(sc->opts[0]);
  return(NULL);
}

static s7_pointer opt_cell_any_nr(s7_scheme *sc, s7_pointer expr)
{
  cur_sc = sc;
  sc->pc = 0;
  return(sc->opts[0]->v7.fp(sc->opts[0])); /* faster than returning NULL */
}

static s7_pointer opt_bool_any_nr(s7_scheme *sc, s7_pointer expr)
{
  cur_sc = sc;
  sc->pc = 0;
  sc->opts[0]->v7.fb(sc->opts[0]);
  return(NULL);
}


/* callers for s7_optimize */
static s7_pointer opt_wrap_float(s7_scheme *sc, s7_pointer expr) {cur_sc = sc; sc->pc = 0; return(make_real(sc, sc->opts[0]->v7.fd(sc->opts[0])));}
static s7_pointer opt_wrap_int(s7_scheme *sc, s7_pointer expr)   {cur_sc = sc; sc->pc = 0; return(make_integer(sc, sc->opts[0]->v7.fi(sc->opts[0])));}
static s7_pointer opt_wrap_cell(s7_scheme *sc, s7_pointer expr)  {cur_sc = sc; sc->pc = 0; return(sc->opts[0]->v7.fp(sc->opts[0]));}
static s7_pointer opt_wrap_bool(s7_scheme *sc, s7_pointer expr)  {cur_sc = sc; sc->pc = 0; return((sc->opts[0]->v7.fb(sc->opts[0])) ? sc->T : sc->F);}

static s7_pointer b_to_p(void *p)    {opt_info *o = (opt_info *)p; return((o->v8.fb(o)) ? cur_sc->T : cur_sc->F);}
static bool p_to_b(void *p)          {opt_info *o = (opt_info *)p; return(o->v8.fp(o) != cur_sc->F);}
static s7_pointer d_to_p(void *p)    {opt_info *o = (opt_info *)p; return(make_real(cur_sc, o->v8.fd(o)));}
static s7_pointer d_to_p_nr(void *p) {opt_info *o = (opt_info *)p; o->v8.fd(o); return(NULL);}
static s7_pointer i_to_p(void *p)    {opt_info *o = (opt_info *)p; return(make_integer(cur_sc, o->v8.fi(o)));}
static s7_pointer i_to_p_nr(void *p) {opt_info *o = (opt_info *)p; o->v8.fi(o); return(NULL);}


/* -------------------------------- int32_t opts -------------------------------- */

static s7_int opt_unwrap_int(void *p)
{
  opt_info *o = (opt_info *)p;
  return(integer(o->v2.all_f(cur_sc, car(o->v1.p))));
}

static s7_int opt_i_c(void *p) {opt_info *o = (opt_info *)p; return(o->v1.i);}
static s7_int opt_i_s(void *p) {opt_info *o = (opt_info *)p; return(integer(slot_value(o->v1.p)));}

static bool opt_int_not_pair(s7_scheme *sc, s7_pointer car_x)
{
  opt_info *opc;
  if (is_opt_int(car_x))
    {
      opc = alloc_opo(sc, car_x);
      opc->v1.i = integer(car_x);
      opc->v7.fi = opt_i_c;
      return(true);
    }
  if (is_symbol(car_x))
    {
      s7_pointer p;
      p = find_symbol(sc, car_x);
      if ((is_slot(p)) &&
	  (is_opt_int(slot_value(p))))
	{
	  opc = alloc_opo(sc, car_x);
	  opc->v1.p = p;
	  opc->v7.fi = opt_i_s;
	  return(true);
	}
    }
  return(return_false(sc, car_x, __func__, __LINE__));
}

/* -------- i_idp -------- */
static s7_int opt_i_i_c(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v2.i_i_f(o->v1.i));
}

static s7_int opt_i_i_s(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v2.i_i_f(integer(slot_value(o->v1.p))));
}

static s7_int opt_i_i_f(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v2.i_i_f(o1->v7.fi(o1)));
}


static s7_int opt_i_d_c(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v2.i_d_f(o->v1.x));
}

static s7_int opt_i_d_s(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v2.i_d_f(real(slot_value(o->v1.p))));
}

static s7_int opt_i_d_f(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v2.i_d_f(o1->v7.fd(o1)));
}

static s7_int opt_i_p_f(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v2.i_p_f(o1->v7.fp(o1)));
}

static bool i_idp_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x)
{
  s7_i_i_t func;
  s7_i_d_t idf;
  s7_i_p_t ipf;
  int32_t start;
  start = sc->pc;
  
  func = s7_i_i_function(s_func);
  if (func)
    {
      opc->v2.i_i_f = func;
      if (is_opt_int(cadr(car_x)))
	{
	  opc->v1.i = integer(cadr(car_x));
	  opc->v7.fi = opt_i_i_c;
	  return(true);
	}
      if (is_symbol(cadr(car_x)))
	{
	  opc->v1.p = find_symbol(sc, cadr(car_x));
	  if ((is_slot(opc->v1.p)) &&
	      (is_integer(slot_value(opc->v1.p))))
	    {
	      opc->v7.fi = opt_i_i_s;
	      return(true);
	    }
	  return(return_false(sc, car_x, __func__, __LINE__));
	}
      else /* is pair arg */
	{
	  if (int_optimize(sc, cdr(car_x)))
	    {
	      opc->v7.fi = opt_i_i_f;
	      return(true);
	    }
	  pc_fallback(sc, start);
	}
    }
  idf = s7_i_d_function(s_func);
  if (idf)
    {
      opc->v2.i_d_f = idf;
      if (is_real(cadr(car_x)))
	{
	  opc->v1.x = s7_number_to_real(sc, cadr(car_x));
	  opc->v7.fi = opt_i_d_c;
	  return(true);
	}
      if (is_symbol(cadr(car_x)))
	{
	  opc->v1.p = find_symbol(sc, cadr(car_x));
	  if (is_slot(opc->v1.p))
	    {
	      if (is_float(slot_value(opc->v1.p)))
		opc->v7.fi = opt_i_d_s;
	      else
		{
		  if (float_optimize(sc, cdr(car_x)))
		    opc->v7.fi = opt_i_d_f;
		  else return(return_false(sc, car_x, __func__, __LINE__));
		}
	      return(true);
	    }
	}
      else /* is pair arg */
	{
	  if (float_optimize(sc, cdr(car_x)))
	    {
	      opc->v7.fi = opt_i_d_f;
	      return(true);
	    }
	  pc_fallback(sc, start);
	}
    }
  ipf = s7_i_p_function(s_func);
  if (ipf)
    {
      opc->v2.i_p_f = ipf;
      if (cell_optimize(sc, cdr(car_x)))
	{
	  opc->v7.fi = opt_i_p_f;
	  return(true);
	}
      pc_fallback(sc, start);
    }
  return(false);
}


/* -------- i_pi -------- */

static s7_int opt_i_pi_ss(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.i_pi_f(slot_value(o->v1.p), integer(slot_value(o->v2.p))));
}

static s7_int opt_i_pi_sf(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.i_pi_f(slot_value(o->v1.p), o1->v7.fi(o1)));
}

static bool i_pi_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x)
{
  s7_i_pi_t pfunc;
  pfunc = s7_i_pi_function(s_func);
  if (pfunc)
    {
      s7_pointer sig;
      sig = s7_procedure_signature(sc, s_func);
      if (is_pair(sig))
	{
	  s7_pointer arg1, arg2;
	  int32_t start;
	  start = sc->pc;
	  arg1 = cadr(car_x);
	  arg2 = caddr(car_x);

	  if ((is_symbol(cadr(sig))) &&
	      (is_symbol(arg1)))
	    {
	      s7_pointer obj, checker;
	      checker = s7_symbol_value(sc, cadr(sig));
	      obj = s7_symbol_value(sc, arg1);
	      if (s7_apply_function(sc, checker, set_plist_1(sc, obj)) == sc->T)
		{
		  opc->v1.p = find_symbol(sc, arg1);
		  if (is_slot(opc->v1.p))
		    {
		      if ((car(car_x) == sc->int_vector_ref_symbol) &&
			  ((!is_int_vector(slot_value(opc->v1.p))) ||
			   (vector_rank(slot_value(opc->v1.p)) > 1)))
			return(return_false(sc, car_x, __func__, __LINE__));
		      
		      opc->v3.i_pi_f = pfunc;
		      if (is_symbol(arg2))
			{
			  opc->v2.p = find_symbol(sc, arg2);
			  if ((is_slot(opc->v2.p)) &&
			      (is_opt_int(slot_value(opc->v2.p))))
			    {
			      opc->v7.fi = opt_i_pi_ss;
			      if ((car(car_x) == sc->int_vector_ref_symbol) &&
				  (is_step_end(opc->v2.p)) &&
				  (denominator(slot_value(opc->v2.p)) <= vector_length(slot_value(opc->v1.p))))
				opc->v3.i_pi_f = int_vector_ref_unchecked;
			      return(true);
			    }
			}
		      if (int_optimize(sc, cddr(car_x)))
			{
			  opc->v7.fi = opt_i_pi_sf;
			  return(true);
			}
		      pc_fallback(sc, start);
		    }
		}
	      else return(return_false(sc, car_x, __func__, __LINE__));
	    }
	}
    }
  return(false);
}

/* -------- i_ii -------- */
static s7_int opt_i_ii_cc(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.i_ii_f(o->v1.i, o->v2.i));
}

static s7_int opt_i_ii_cs(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.i_ii_f(o->v1.i, integer(slot_value(o->v2.p))));
}

static s7_int opt_i_ii_sc(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.i_ii_f(integer(slot_value(o->v1.p)), o->v2.i));
}

static s7_int opt_i_ii_ss(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.i_ii_f(integer(slot_value(o->v1.p)), integer(slot_value(o->v2.p))));
}

static s7_int opt_i_ii_cf(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.i_ii_f(o->v1.i, o1->v7.fi(o1)));
}

static s7_int opt_i_ii_sf(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.i_ii_f(integer(slot_value(o->v1.p)), o1->v7.fi(o1)));
}

static s7_int opt_i_ii_ff(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  s7_int i1;
  o1 = cur_sc->opts[++cur_sc->pc];
  i1 = o1->v7.fi(o1);
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.i_ii_f(i1, o1->v7.fi(o1)));
}

#if (!WITH_GMP)
static s7_int opt_i_ii_fc(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.i_ii_f(o1->v7.fi(o1), o->v2.i));
}

static s7_int opt_i_ii_fco(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.i_ii_f(o->v4.i_pi_f(slot_value(o->v1.p), integer(slot_value(o->v2.p))), o->v5.i));
}

static bool i_ii_fc_combinable(s7_scheme *sc, opt_info *opc)
{
  if ((sc->pc > 1) &&
      (opc == sc->opts[sc->pc - 2]))
    {
      opt_info *o1;
      o1 = sc->opts[sc->pc - 1];
      if (o1->v7.fi == opt_i_pi_ss)
	{
	  opc->v5.i = opc->v2.i; /* move v2.i ("c" in fc = arg2) out of the symbols' way */
	  opc->v4.i_pi_f = o1->v3.i_pi_f;
	  opc->v1.p = o1->v1.p;
	  opc->v2.p = o1->v2.p;	
	  opc->v7.fi = opt_i_ii_fco;
	  sc->pc--;
	  return(true);
	}
    }
  return(false);
}
#endif

static bool i_ii_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x)
{
  s7_i_ii_t ifunc;
  ifunc = s7_i_ii_function(s_func);
  if (ifunc)
    {
      s7_pointer sig;
      sig = s7_procedure_signature(sc, s_func);
      if (is_pair(sig))
	{
	  s7_pointer arg1, arg2;
	  int32_t start;
	  start = sc->pc;
	  arg1 = cadr(car_x);
	  arg2 = caddr(car_x);
	  
	  opc->v3.i_ii_f = ifunc;
	  if (is_opt_int(arg1))
	    {
	      opc->v1.i = integer(arg1);
	      if (is_opt_int(arg2))
		{
		  opc->v2.i = integer(arg2);
		  opc->v7.fi = opt_i_ii_cc;
		  return(true);
		}
	      if (is_symbol(arg2))
		{
		  opc->v2.p = find_symbol(sc, arg2);
		  if (is_slot(opc->v2.p))
		    {
		      if (is_integer(slot_value(opc->v2.p)))
			{
			  opc->v7.fi = opt_i_ii_cs;
			  return(true);
			}
		    }
		}
	      else
		{
		  if (int_optimize(sc, cddr(car_x)))
		    {
		      opc->v7.fi = opt_i_ii_cf;
		      return(true);
		    }
		  pc_fallback(sc, start);
		}
	    }
	  else
	    {
	      if (is_symbol(arg1))
		{
		  opc->v1.p = find_symbol(sc, arg1);
		  if (is_slot(opc->v1.p))
		    {
		      if (is_opt_int(slot_value(opc->v1.p)))
			{
			  if (is_opt_int(arg2))
			    {
			      opc->v2.i = integer(arg2);
			      opc->v7.fi = opt_i_ii_sc;
#if (!WITH_GMP)
			      if ((car(car_x) == sc->modulo_symbol) &&
				  (integer(arg2) > 1))
				opc->v3.i_ii_f = modulo_i_ii_direct;
			      else
				{
				  if (car(car_x) == sc->ash_symbol)
				    {
				      if (opc->v2.i < 0)
					{
					  if (opc->v2.i == -1)
					    opc->v3.i_ii_f = rsh_i_i2_direct;
					  else opc->v3.i_ii_f = rsh_i_ii_direct;
					}
				      else opc->v3.i_ii_f = lsh_i_ii_direct;
				    }
				  else
				    {
				      if (opc->v2.i > 0)
					{
					  if (opc->v3.i_ii_f == quotient_i_ii)
					    opc->v3.i_ii_f = quotient_i_ii_direct;
					  else
					    {
					      if (opc->v2.i > 1)
						{
						  if (opc->v3.i_ii_f == remainder_i_ii)
						    opc->v3.i_ii_f = remainder_i_ii_direct;
						}
					    }
					}
				      }
				}
#endif
			      return(true);
			    }
			  if (is_symbol(arg2))
			    {
			      opc->v2.p = find_symbol(sc, arg2);
			      if ((is_slot(opc->v2.p)) &&
				  (is_opt_int(slot_value(opc->v2.p))))
				{
				  opc->v7.fi = opt_i_ii_ss;
				  return(true);
				}
			    }
			  else
			    {
			      if (int_optimize(sc, cddr(car_x)))
				{
				  opc->v7.fi = opt_i_ii_sf;
				  return(true);
				}
			      pc_fallback(sc, start);
			    }
			}
		    }
		}
	      else
		{
#if (!WITH_GMP)
		  if (is_opt_int(arg2))
		    {
		      opc->v2.i = integer(arg2);
		      if (int_optimize(sc, cdr(car_x)))
			{
			  if (!i_ii_fc_combinable(sc, opc))
			    {
			      if (opc->v2.i > 0)
				{
				  if (opc->v3.i_ii_f == quotient_i_ii)
				    opc->v3.i_ii_f = quotient_i_ii_direct;
				  else
				    {
				      if (opc->v2.i > 1)
					{
					  if (opc->v3.i_ii_f == remainder_i_ii)
					    opc->v3.i_ii_f = remainder_i_ii_direct;
					}
				    }
				}
			      opc->v7.fi = opt_i_ii_fc;
			    }
			  return(true);
			}
		      pc_fallback(sc, start);
		    }
		  else
#endif
		    {
		      if ((int_optimize(sc, cdr(car_x))) &&
			  (int_optimize(sc, cddr(car_x))))
			{
			  opc->v7.fi = opt_i_ii_ff;
			  return(true);
			}
		      pc_fallback(sc, start);
		    }
		}
	    }
	}
    }
  return(false);
}

/* -------- i_iii -------- */
static s7_int opt_i_iii_fff(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  s7_int i1, i2;
  o1 = cur_sc->opts[++cur_sc->pc];
  i1 = o1->v7.fi(o1);
  o1 = cur_sc->opts[++cur_sc->pc];
  i2 = o1->v7.fi(o1);
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.i_iii_f(i1, i2, o1->v7.fi(o1)));
}

static bool i_iii_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x)
{
  s7_i_iii_t ifunc;
  ifunc = s7_i_iii_function(s_func);
  if (ifunc)
    {
      int32_t start;
      start = sc->pc;
      if ((int_optimize(sc, cdr(car_x))) &&
	  (int_optimize(sc, cddr(car_x))) &&
	  (int_optimize(sc, cdddr(car_x))))
	{
	  opc->v3.i_iii_f = ifunc;
	  opc->v7.fi = opt_i_iii_fff;
	  return(true);
	}
      pc_fallback(sc, start);
      return(return_false(sc, car_x, __func__, __LINE__));
    }
  return(false);
}

/* -------- i_pii -------- */
static s7_int opt_i_pii_ssf(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.i_pii_f(slot_value(o->v1.p), integer(slot_value(o->v2.p)), o1->v7.fi(o1)));
}

static s7_int opt_i_pii_sff(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1, *o2;
  s7_int i1;
  o1 = cur_sc->opts[++cur_sc->pc];
  i1 = o1->v7.fi(o1);
  o2 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.i_pii_f(slot_value(o->v1.p), i1, o2->v7.fi(o2)));
}

static bool opt_int_vector_set(s7_scheme *sc, opt_info *opc, s7_pointer v, s7_pointer indexp, s7_pointer valp)
{
  s7_pointer settee;
  /* fprintf(stderr, "opt intv: %s %s %s\n", DISPLAY(v), DISPLAY(indexp), DISPLAY(valp)); */
  settee = find_symbol(sc, v);
  if (is_slot(settee))
    {
      opc->v1.p = settee;
      if ((is_int_vector(slot_value(settee))) &&
	  (vector_rank(slot_value(settee)) == 1))
	{
	  opc->v3.i_pii_f = int_vector_set_i;
	  if (is_symbol(car(indexp)))
	    {
	      s7_pointer slot;
	      slot = find_symbol(sc, car(indexp));
	      if ((is_slot(slot)) &&
		  (is_integer(slot_value(slot))) &&
		  (int_optimize(sc, valp)))
		{
		  opc->v7.fi = opt_i_pii_ssf;
		  opc->v2.p = slot;
		  if ((is_step_end(opc->v2.p)) &&
		      (denominator(slot_value(opc->v2.p)) <= vector_length(slot_value(settee))))
		    opc->v3.i_pii_f = int_vector_set_unchecked;
		  return(true);
		}
	    }
	  else
	    {
	      if ((int_optimize(sc, indexp)) &&
		  (int_optimize(sc, valp)))
		{
		  opc->v7.fi = opt_i_pii_sff;
		  return(true);
		}
	    }
	}
    }
  return(return_false(sc, v, __func__, __LINE__));
}

static bool i_pii_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x)
{
  s7_i_pii_t pfunc;
  pfunc = s7_i_pii_function(s_func);
  if (pfunc)
    {
      s7_pointer sig;
      sig = s7_procedure_signature(sc, s_func);
      if ((is_pair(sig)) &&
	  (is_symbol(cadr(car_x))))
	{
	  s7_pointer obj, checker;
	  
	  if (car(car_x) == sc->int_vector_set_symbol)
	    return(opt_int_vector_set(sc, opc, cadr(car_x), cddr(car_x), cdddr(car_x)));
	  
	  checker = s7_symbol_value(sc, cadr(sig));
	  obj = s7_symbol_value(sc, cadr(car_x));
	  if (s7_apply_function(sc, checker, set_plist_1(sc, obj)) == sc->T)
	    {
	      int32_t start;
	      start = sc->pc;
	      opc->v3.i_pii_f = pfunc;
	      opc->v1.p = find_symbol(sc, cadr(car_x));
	      if (is_slot(opc->v1.p))
		{
		  s7_pointer arg2;
		  arg2 = caddr(car_x);
		  if (is_symbol(arg2))
		    {
		      opc->v2.p = find_symbol(sc, arg2);
		      if ((is_slot(opc->v2.p)) &&
			  (is_opt_int(slot_value(opc->v2.p))) &&
			  (int_optimize(sc, cdddr(car_x))))
			{
			  opc->v7.fi = opt_i_pii_ssf;
			  return(true);
			}
		    }
		  else
		    {
		      if ((int_optimize(sc, cddr(car_x))) &&
			  (int_optimize(sc, cdddr(car_x))))
			{
			  opc->v7.fi = opt_i_pii_sff;
			  return(true);
			}
		    }
		  pc_fallback(sc, start);
		}
	    }
	  else return(return_false(sc, car_x, __func__, __LINE__));
	}
    }
  return(false);
}

/* -------- i_add|multiply_any -------- */
static s7_int opt_i_add_any_f(void *p)
{
  opt_info *o = (opt_info *)p;
  s7_int sum = 0;
  int32_t i;
  for (i = 0; i < o->v1.i; i++)
    {
      opt_info *o1;
      o1 = cur_sc->opts[++cur_sc->pc];
      sum += o1->v7.fi(o1);
    }
  return(sum);
}

static s7_int opt_i_add2(void *p)
{
  s7_int sum;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  sum = o1->v7.fi(o1);
  o1 = cur_sc->opts[++cur_sc->pc];
  return(sum + o1->v7.fi(o1));
}

static s7_int opt_i_mul2(void *p)
{
  s7_int sum;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  sum = o1->v7.fi(o1);
  o1 = cur_sc->opts[++cur_sc->pc];
  return(sum * o1->v7.fi(o1));
}

static s7_int opt_i_multiply_any_f(void *p)
{
  opt_info *o = (opt_info *)p;
  s7_int sum = 1;
  int32_t i;
  for (i = 0; i < o->v1.i; i++)
    {
      opt_info *o1;
      o1 = cur_sc->opts[++cur_sc->pc];
      sum *= o1->v7.fi(o1);
    }
  return(sum);
}
		
static bool i_add_any_ok(s7_scheme *sc, opt_info *opc, s7_pointer car_x)
{
  s7_pointer p, head;
  int32_t cur_len = 0, start;
  start = sc->pc;
  head = car(car_x);
  for (p = cdr(car_x); is_pair(p); p = cdr(p))
    {
      if (is_pair(cdr(p)))
	{
	  /* TODO: 3-way case if it were able to handle symbols and constants */
	  if (!int_optimize(sc, set_plist_1(sc, set_elist_3(sc, head, car(p), cadr(p)))))
	    break;
	  cur_len++;
	  p = cdr(p);
	}
      else
	{
	  if (!int_optimize(sc, p))
	    break;
	  cur_len++;
	}
    }
  if (is_null(p))
    {
      opc->v1.i = cur_len;
      if (cur_len == 2)
	opc->v7.fi = (head == sc->add_symbol) ? opt_i_add2 :opt_i_mul2;
      else opc->v7.fi = (head == sc->add_symbol) ? opt_i_add_any_f : opt_i_multiply_any_f;
      return(true);
    }
  pc_fallback(sc, start);
  return(false);
}

/* -------- int_all_x -------- */
static bool int_all_x_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer expr)
{
  s7_pointer sig;
  s7_function opt;
  sig = s7_procedure_signature(sc, s_func);	  
  if ((is_pair(sig)) &&
      (car(sig) == sc->is_integer_symbol))
    {
      /* fallback on the more general case (all_x_eval, but still guaranteed to be an integer) */
      opt = all_x_optimize(sc, expr);
      if (opt)
	{
	  opc->v2.all_f = opt;
	  opc->v7.fi = opt_unwrap_int;
	  opc->v1.p = expr;
	  return(true);
	}
    }
  return(false);
}

/* -------- set_i_i -------- */
static s7_int opt_set_i_i_f(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  s7_int x;
  o1 = cur_sc->opts[++cur_sc->pc];
  x = o1->v7.fi(o1);
  if (is_mutable(slot_value(o->v1.p)))
    integer(slot_value(o->v1.p)) = x;
  else slot_set_value(o->v1.p, make_integer(cur_sc, x));
  return(x);
}

static s7_int opt_set_i_i_fo(void *p)
{
  opt_info *o = (opt_info *)p;
  s7_int x;
  x = o->v4.i_ii_f(integer(slot_value(o->v3.p)), o->v2.i);
  if (is_mutable(slot_value(o->v1.p)))
    integer(slot_value(o->v1.p)) = x;
  else slot_set_value(o->v1.p, make_integer(cur_sc, x));
  return(x);
}

static bool set_i_i_f_combinable(s7_scheme *sc, opt_info *opc)
{
  if ((sc->pc > 1) &&
      (opc == sc->opts[sc->pc - 2]))
    {
      opt_info *o1;
      o1 = sc->opts[sc->pc - 1];
      if (o1->v7.fi == opt_i_ii_sc)
	{
	  opc->v4.i_ii_f = o1->v3.i_ii_f;
	  opc->v3.p = o1->v1.p;
	  opc->v2.i = o1->v2.i;
	  opc->v7.fi = opt_set_i_i_fo;
	  sc->pc--;
	  return(true);
	}
    }
  return(false);
}

static bool i_syntax_ok(s7_scheme *sc, s7_pointer car_x, int32_t len)
{
  if ((car(car_x) == sc->set_symbol) &&
      (len == 3))
    {
      opt_info *opc;
      opc = alloc_opo(sc, car_x);
      if (is_symbol(cadr(car_x)))  /* (set! i 3) */
	{
	  s7_pointer settee;
	  if ((is_immutable(cadr(car_x))) ||
	      (symbol_has_accessor(cadr(car_x))))
	    return(return_false(sc, car_x, __func__, __LINE__));
	  settee = find_symbol(sc, cadr(car_x));
	  if (is_slot(settee))
	    {
	      opc->v1.p = settee;
	      if ((is_integer(slot_value(settee))) &&
		  (int_optimize(sc, cddr(car_x))))
		{
		  if (!set_i_i_f_combinable(sc, opc))
		    opc->v7.fi = opt_set_i_i_f;
		  return(true);
		}
	    }
	}
      else
	{
	  if ((is_pair(cadr(car_x))) &&    /* if is_pair(settee) get setter */
	      (is_symbol(caadr(car_x))) &&
	      (is_pair(cdadr(car_x))) &&   /* (set! (f . y) z)? */
	      (is_null(cddadr(car_x))))
	    return(opt_int_vector_set(sc, opc, caadr(car_x), cdadr(car_x), cddr(car_x)));
	}
    }
  return(return_false(sc, car_x, __func__, __LINE__));
}

static bool i_implicit_ok(s7_scheme *sc, s7_pointer car_x, int32_t len)
{
  s7_pointer s_slot, head;
  head = car(car_x);
  s_slot = find_symbol(sc, head);
  if ((is_slot(s_slot)) &&
      (len == 2) &&
      (is_int_vector(slot_value(s_slot))) &&
      (vector_rank(slot_value(s_slot)) == 1))
    {
      /* implicit int-vector-ref */
      opt_info *opc;
      opc = alloc_opo(sc, car_x);
      opc->v1.p = s_slot;
      if (is_symbol(cadr(car_x)))
	{
	  s7_pointer slot;
	  slot = find_symbol(sc, cadr(car_x));
	  if ((is_slot(slot)) &&
	      (is_integer(slot_value(slot))))
	    {
	      opc->v7.fi = opt_i_pi_ss;
	      opc->v3.i_pi_f = int_vector_ref_i;
	      opc->v2.p = slot;
	      if ((is_step_end(opc->v2.p)) &&
		  (denominator(slot_value(opc->v2.p)) <= vector_length(slot_value(opc->v1.p))))
		opc->v3.i_pi_f = int_vector_ref_unchecked;
	      return(true);
	    }
	}
      else
	{
	  if (int_optimize(sc, cdr(car_x)))
	    {
	      opc->v7.fi = opt_i_pi_sf;
	      opc->v3.i_pi_f = int_vector_ref_i;
	      return(true);
	    }
	}
    }
  return(return_false(sc, car_x, __func__, __LINE__));
}


/* ------------------------------------- float opts ------------------------------------------- */

static s7_double opt_d_c(void *p) {opt_info *o = (opt_info *)p; return(o->v1.x);}
static s7_double opt_D_s(void *p) {opt_info *o = (opt_info *)p; return(s7_number_to_real(cur_sc, slot_value(o->v1.p)));}
static s7_double opt_d_s(void *p) {opt_info *o = (opt_info *)p; return(real(slot_value(o->v1.p)));}

static bool opt_float_not_pair(s7_scheme *sc, s7_pointer car_x)
{
  opt_info *opc;
  if (is_real(car_x))
    {
      if ((s7_is_ratio(car_x)) ||
	  (!is_opt_real(car_x)))
	return(return_false(sc, car_x, __func__, __LINE__));
      opc = alloc_opo(sc, car_x);
      opc->v1.x = s7_number_to_real(sc, car_x);
      opc->v7.fd = opt_d_c;
      return(true);
    }
  if (is_symbol(car_x))
    {
      s7_pointer p;
      p = find_symbol(sc, car_x);
      if ((is_slot(p)) &&
	  (is_opt_real(slot_value(p))))
	{
	  if (s7_is_ratio(slot_value(p)))
	    return(return_false(sc, car_x, __func__, __LINE__));
	  opc = alloc_opo(sc, car_x);
	  opc->v1.p = p;
	  opc->v7.fd = (is_float(slot_value(p))) ? opt_d_s : opt_D_s;
	  return(true);
	}
    }
  return(return_false(sc, car_x, __func__, __LINE__));
}

/* -------- d -------- */
static s7_double opt_d_f(void *p) {opt_info *o = (opt_info *)p; return(o->v1.d_f());}

static bool d_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func)
{
  s7_d_t func;                   /* (f): (mus-srate) */
  func = s7_d_function(s_func);
  if (func)
    {
      opc->v7.fd = opt_d_f;
      opc->v1.d_f = func;
      return(true);
    }
  return(false);
}

/* -------- d_d -------- */
static s7_double opt_d_d_c(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.d_d_f(o->v1.x));
}

static s7_double opt_d_d_s(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.d_d_f(real(slot_value(o->v1.p))));
}

static s7_double opt_d_d_f(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.d_d_f(o1->v7.fd(o1)));
}

static bool d_d_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x)
{
  s7_d_d_t func;
  int32_t start;
  start = sc->pc;

  func = s7_d_d_function(s_func);
  if (func)
    {
      opc->v3.d_d_f = func;
      if (is_real(cadr(car_x)))
	{
	  opc->v1.x = s7_number_to_real(sc, cadr(car_x));
	  opc->v7.fd = opt_d_d_c;
	  return(true);
	}
      if (is_symbol(cadr(car_x)))
	{
	  opc->v1.p = find_symbol(sc, cadr(car_x));
	  if ((is_slot(opc->v1.p)) &&
	      (!has_methods(slot_value(opc->v1.p))))
	    {
	      if (is_float(slot_value(opc->v1.p)))
		opc->v7.fd = opt_d_d_s;
	      else 
		{
		  if (float_optimize(sc, cdr(car_x)))
		    opc->v7.fd = opt_d_d_f;
		  else return(return_false(sc, car_x, __func__, __LINE__));
		}
	      return(true);
	    }
	}
      else /* is pair arg */
	{
	  if (float_optimize(sc, cdr(car_x)))
	    {
	      opc->v7.fd = opt_d_d_f;
	      return(true);
	    }
	}
      pc_fallback(sc, start);
    }
  return(false);
}

/* -------- d_v -------- */
static s7_double opt_d_v(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.d_v_f(o->v5.obj));
}

static bool d_v_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x)
{
  s7_d_v_t flt_func;
  flt_func = s7_d_v_function(s_func);
  if (flt_func)
    {
      s7_pointer sig;
      sig = s7_procedure_signature(sc, s_func);
      if ((is_pair(sig)) &&
	  (is_symbol(cadr(car_x))))           /* look for (oscil g) */
	{
	  s7_pointer slot, obj, checker;
	  checker = s7_symbol_value(sc, cadr(sig));
	  slot = find_symbol(sc, cadr(car_x));
	  obj = slot_value(slot);
	  if (s7_apply_function(sc, checker, set_plist_1(sc, obj)) == sc->T)
	    {
	      opc->v1.p = slot;
	      opc->v5.obj = (void *)s7_c_object_value(obj);
	      opc->v3.d_v_f = flt_func;
	      opc->v7.fd = opt_d_v;
	      return(true);
	    }
	}
    }
  return(false);
}

/* -------- d_p -------- */
static s7_double opt_d_p_s(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.d_p_f(slot_value(o->v1.p)));
}

static s7_double opt_d_p_f(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.d_p_f(o1->v7.fp(o1)));
}

static bool d_p_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x)
{
  s7_d_p_t dpf;
  int32_t start;
  start = sc->pc;
  dpf = s7_d_p_function(s_func);
  if (dpf)
    {
      opc->v3.d_p_f = dpf;
      if (is_symbol(cadr(car_x)))
	{
	  opc->v1.p = find_symbol(sc, cadr(car_x));
	  if ((is_slot(opc->v1.p)) &&
	      (!has_methods(slot_value(opc->v1.p))))
	    {
	      opc->v7.fd = opt_d_p_s;
	      return(true);
	    }
	  else return(return_false(sc, car_x, __func__, __LINE__));
	}
      if (cell_optimize(sc, cdr(car_x)))
	{
	  opc->v7.fd = opt_d_p_f;
	  return(true);
	}
      pc_fallback(sc, start);
    }
  return(false);
}

/* -------- d_pi -------- */

static s7_double opt_d_pi_sc(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.d_pi_f(slot_value(o->v1.p), o->v2.i));
}

static s7_double opt_d_pi_ss(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.d_pi_f(slot_value(o->v1.p), integer(slot_value(o->v2.p))));
}

static s7_double opt_d_pi_sf(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.d_pi_f(slot_value(o->v1.p), o1->v7.fi(o1)));
}

static bool d_pi_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x)
{
  if (is_symbol(cadr(car_x)))
    {
      s7_d_pi_t ifunc;
      ifunc = s7_d_pi_function(s_func);
      if (ifunc)
	{
	  s7_pointer arg2;
	  int32_t start;
	  start = sc->pc;
	  opc->v1.p = find_symbol(sc, cadr(car_x));
	  if (!is_slot(opc->v1.p)) 
	    return(return_false(sc, car_x, __func__, __LINE__));
	  if ((car(car_x) == sc->float_vector_ref_symbol) &&
	      ((!is_float_vector(slot_value(opc->v1.p))) ||
	       (vector_rank(slot_value(opc->v1.p)) > 1)))
	    return(return_false(sc, car_x, __func__, __LINE__));
	  opc->v3.d_pi_f = ifunc;
	  arg2 = caddr(car_x);
	  if (!is_pair(arg2))
	    {
	      if (is_opt_int(arg2))
		{
		  opc->v2.i = integer(arg2);
		  opc->v7.fd = opt_d_pi_sc;
		  return(true);
		}
	      if (is_symbol(arg2))
		{
		  opc->v2.p = find_symbol(sc, arg2);
		  if ((is_slot(opc->v2.p)) &&
		      (is_integer(slot_value(opc->v2.p))))
		    {
		      opc->v7.fd = opt_d_pi_ss;
		      if ((car(car_x) == sc->float_vector_ref_symbol) &&
			  (is_step_end(opc->v2.p)) &&
			  (denominator(slot_value(opc->v2.p)) <= vector_length(slot_value(opc->v1.p))))
			opc->v3.d_pi_f = float_vector_ref_unchecked;
		      return(true);
		    }
		}
	    }
	  else
	    {
	      if (int_optimize(sc, cddr(car_x)))
		{
		  opc->v7.fd = opt_d_pi_sf;
		  return(true);
		}
	      pc_fallback(sc, start);
	    }
	}
    }
  return(false);
}

/* -------- d_ip -------- */
static s7_double opt_d_ip_ss(void *p)
{
  opt_info *o = (opt_info *)p;
  /* PERHAPS: type check */
  return(o->v3.d_ip_f(integer(slot_value(o->v1.p)), slot_value(o->v2.p)));
}

static bool d_ip_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x)
{
  if ((is_symbol(cadr(car_x))) &&
      (is_symbol(caddr(car_x))))
    {
      s7_d_ip_t pfunc;
      pfunc = s7_d_ip_function(s_func);
      if (pfunc)
	{
	  opc->v3.d_ip_f = pfunc;
	  opc->v1.p = find_symbol(sc, cadr(car_x));
	  opc->v2.p = find_symbol(sc, caddr(car_x));
	  if ((is_slot(opc->v1.p)) &&
	      (is_integer(slot_value(opc->v1.p))) &&
	      (is_slot(opc->v2.p)))
	    {
	      /* (with-sound (:reverb jc-reverb) (fm-violin 0 .1 440 .4 :reverb-amount .5)) */
	      opc->v7.fd = opt_d_ip_ss;
	      return(true);
	    }
	}
    }
  return(false);
}

/* -------- d_pd -------- */
static s7_double opt_d_pd_sf(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.d_pd_f(slot_value(o->v1.p), o1->v7.fd(o1)));
}

static s7_double opt_d_pd_ss(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.d_pd_f(slot_value(o->v1.p), real(slot_value(o->v2.p))));
}

static bool d_pd_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x)
{
  if (is_symbol(cadr(car_x)))
    {
      s7_d_pd_t func;
      func = s7_d_pd_function(s_func);
      if (func)
	{
	  s7_pointer arg2;
	  int32_t start;
	  start = sc->pc;
	  arg2 = caddr(car_x);
	  opc->v3.d_pd_f = func;
	  opc->v1.p = find_symbol(sc, cadr(car_x));
	  if (!is_slot(opc->v1.p)) 
	    return(return_false(sc, car_x, __func__, __LINE__));
	  if (is_symbol(arg2))
	    {
	      opc->v2.p = find_symbol(sc, arg2);
	      if ((is_slot(opc->v2.p)) &&
		  (is_float(slot_value(opc->v2.p))))
		{
		  opc->v7.fd = opt_d_pd_ss;
		  return(true);
		}
	    }
	  if (float_optimize(sc, cddr(car_x)))
	    {
	      opc->v7.fd = opt_d_pd_sf;
	      return(true);
	    }
	  pc_fallback(sc, start);
	}
    }
  return(false);
}

/* -------- d_vd -------- */
static s7_double opt_d_vd_c(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.d_vd_f(o->v5.obj, o->v2.x));
}

static s7_double opt_d_vd_s(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.d_vd_f(o->v5.obj, real(slot_value(o->v2.p))));
}

static s7_double opt_d_vd_f(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.d_vd_f(o->v5.obj, o1->v7.fd(o1)));
}

static s7_double opt_d_vd_o(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.d_vd_f(o->v5.obj, o->v4.d_v_f(o->v6.obj)));
}

static s7_double opt_d_vd_o1(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  cur_sc->pc += 2;
  o1 = cur_sc->opts[cur_sc->pc];
  return(o->v3.d_vd_f(o->v5.obj, o->v4.d_dd_f(real(slot_value(o->v2.p)), o1->v7.fd(o1))));
}

static s7_double opt_d_vd_o2(void *p)
{
  opt_info *o = (opt_info *)p; /* v1.p = v6 obj slot */
  return(o->v4.d_vd_f(o->v6.obj, o->v5.d_vd_f(o->v2.obj, real(slot_value(o->v3.p)))));
} 

static s7_double opt_d_vd_o3(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.d_vd_f(o->v5.obj, o->v4.d_dd_f(o->v6.x, real(slot_value(o->v2.p)))));
}

static s7_double opt_d_vd_ff(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  cur_sc->pc += 2;
  o1 = cur_sc->opts[cur_sc->pc];
  return(o->v3.d_vd_f(o->v5.obj, o->v2.d_vd_f(o->v4.obj, o1->v7.fd(o1))));
}

static s7_double opt_d_dd_cs(void *p);
static s7_double opt_d_dd_sf(void *p);

static bool d_vd_f_combinable(s7_scheme *sc, int32_t start)
{
  opt_info *opc, *o1;
  opc = sc->opts[start - 1];
  o1 = sc->opts[start];
  if (o1->v7.fd == opt_d_v)
    {
      opc->v2.p = o1->v1.p;
      opc->v6.obj = o1->v5.obj;
      opc->v4.d_v_f = o1->v3.d_v_f;
      opc->v7.fd = opt_d_vd_o;
      sc->pc--;
      return(true);
    }
  if (o1->v7.fd == opt_d_vd_s)
    {
      opc->v6.obj = opc->v5.obj;
      opc->v4.d_vd_f = opc->v3.d_vd_f; /* room for symbols */
      opc->v2.obj = o1->v5.obj;
      opc->v5.d_vd_f = o1->v3.d_vd_f;
      opc->v3.p = o1->v2.p;
      opc->v7.fd = opt_d_vd_o2;
      sc->pc--;
      return(true);
    }
  if (o1->v7.fd == opt_d_dd_cs)
    {
      opc->v4.d_dd_f = o1->v3.d_dd_f;
      opc->v6.x = o1->v2.x;
      opc->v2.p = o1->v1.p;
      opc->v7.fd = opt_d_vd_o3;
      sc->pc--;
      return(true);
    }
  if (o1->v7.fd == opt_d_dd_sf)
    {
      opc->v2.p = o1->v1.p;
      opc->v4.d_dd_f = o1->v3.d_dd_f;
      opc->v7.fd = opt_d_vd_o1;
      return(true);
    }
  if (o1->v7.fd == opt_d_vd_f)
    {
      opc->v2.d_vd_f = o1->v3.d_vd_f;
      opc->v4.obj = o1->v5.obj;
      opc->v7.fd = opt_d_vd_ff;
      return(true);
    }
  return(false);
}


static bool d_vd_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x)
{
  if (is_symbol(cadr(car_x)))
    {
      s7_d_vd_t vfunc;
      vfunc = s7_d_vd_function(s_func);
      if (vfunc)
	{
	  s7_pointer sig;
	  sig = s7_procedure_signature(sc, s_func);
	  if ((is_pair(sig)) &&
	      (is_symbol(cadr(sig))))
	    {
	      s7_pointer slot, obj, checker;
	      checker = s7_symbol_value(sc, cadr(sig));
	      slot = find_symbol(sc, cadr(car_x));
	      obj = slot_value(slot);
	      if (s7_apply_function(sc, checker, set_plist_1(sc, obj)) == sc->T)
		{
		  s7_pointer arg2;
		  int32_t start;
		  start = sc->pc;
		  arg2 = caddr(car_x);
		  opc->v3.d_vd_f = vfunc;
		  if (!is_pair(arg2))
		    {
		      opc->v1.p = slot;
		      opc->v5.obj = (void *)s7_c_object_value(obj);
		      if (is_real(arg2))
			{
			  opc->v2.x = s7_number_to_real(sc, arg2);
			  opc->v7.fd = opt_d_vd_c;
			  return(true);
			}
		      opc->v2.p = find_symbol(sc, arg2);
		      if (is_slot(opc->v2.p))
			{
			  if (is_float(slot_value(opc->v2.p)))
			    opc->v7.fd = opt_d_vd_s;
			  else
			    {
			      if (float_optimize(sc, cddr(car_x)))
				{
				  if (!d_vd_f_combinable(sc, start))
				    opc->v7.fd = opt_d_vd_f;
				}
			      else return(return_false(sc, car_x, __func__, __LINE__));
			    }
			  return(true);
			}
		    }
		  else /* is pair arg2 */
		    {
		      if (float_optimize(sc, cddr(car_x)))
			{
			  opc->v1.p = slot;
			  opc->v5.obj = (void *)s7_c_object_value(obj);
			  if (!d_vd_f_combinable(sc, start))
			    opc->v7.fd = opt_d_vd_f;
			  return(true);
			}
		      pc_fallback(sc, start);
		    }
		}
	    }
	}
    }
  return(false);
}

/* -------- d_id -------- */
static s7_double opt_d_id_sf(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.d_id_f(integer(slot_value(o->v1.p)), o1->v7.fd(o1)));
}

static s7_double opt_d_id_sc(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.d_id_f(integer(slot_value(o->v1.p)), o->v2.x));
}

static s7_double opt_d_id_sfo(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v4.d_id_f(integer(slot_value(o->v1.p)), o->v5.d_vd_f(o->v6.obj, real(slot_value(o->v3.p)))));
}

static s7_double opt_d_id_sfo1(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.d_id_f(integer(slot_value(o->v1.p)), o->v5.d_v_f(o->v2.obj)));
}

static bool d_id_sf_combinable(s7_scheme *sc, opt_info *opc)
{
  if ((sc->pc > 1) &&
      (opc == sc->opts[sc->pc - 2]))
    {
      opt_info *o1;
      o1 = sc->opts[sc->pc - 1];
      if (o1->v7.fd == opt_d_vd_s)
	{
	  opc->v4.d_id_f = opc->v3.d_id_f;
	  opc->v2.p = o1->v1.p;
	  opc->v6.obj = o1->v5.obj;
	  opc->v5.d_vd_f = o1->v3.d_vd_f;
	  opc->v3.p = o1->v2.p;
	  opc->v7.fd = opt_d_id_sfo;
	  sc->pc--;
	  return(true);
	}
      if (o1->v7.fd == opt_d_v)
	{
	  opc->v6.p = o1->v1.p;
	  opc->v2.obj = o1->v5.obj;
	  opc->v5.d_v_f = o1->v3.d_v_f;
	  opc->v7.fd = opt_d_id_sfo1;
	  sc->pc--;
	  return(true);
	}
    }
  return(false);
}

static bool d_id_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x)
{
  /* fprintf(stderr, "d_id: %s\n", DISPLAY(car_x)); */
  if (is_symbol(cadr(car_x)))
    {
      s7_d_id_t flt_func;
      flt_func = s7_d_id_function(s_func);
      if (flt_func)
	{
	  int32_t start;
	  start = sc->pc;
	  opc->v3.d_id_f = flt_func;
	  opc->v1.p = find_symbol(sc, cadr(car_x));
	  if ((is_slot(opc->v1.p)) &&
	      (is_integer(slot_value(opc->v1.p))))
	    {
	      if (is_t_real(caddr(car_x)))
		{
		  opc->v7.fd = opt_d_id_sc;
		  opc->v2.x = real(caddr(car_x));
		  return(true);
		}
	      if (float_optimize(sc, cddr(car_x)))
		{
		  if (!d_id_sf_combinable(sc, opc))
		    opc->v7.fd = opt_d_id_sf;
		  return(true);
		}
	    }
	  pc_fallback(sc, start);
	}
    }
  return(false);
}

/* -------- d_dd -------- */

static s7_double opt_d_dd_cc(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.d_dd_f(o->v1.x, o->v2.x));
}

static s7_double opt_d_dd_cs(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.d_dd_f(o->v2.x, real(slot_value(o->v1.p))));
}

static s7_double opt_d_dd_sc(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.d_dd_f(real(slot_value(o->v1.p)), o->v2.x));
}

static s7_double opt_d_dd_ss(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.d_dd_f(real(slot_value(o->v1.p)), real(slot_value(o->v2.p))));
}

static s7_double opt_d_dd_cf(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.d_dd_f(o->v1.x, o1->v7.fd(o1)));
}

static s7_double opt_d_dd_fc(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.d_dd_f(o1->v7.fd(o1), o->v2.x));
}

static s7_double opt_d_dd_sf(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.d_dd_f(real(slot_value(o->v1.p)), o1->v7.fd(o1)));
}

static s7_double opt_d_dd_fs(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.d_dd_f(o1->v7.fd(o1), real(slot_value(o->v1.p))));
}

static s7_double opt_d_dd_ff(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1, *o2;
  s7_double x1;
  o1 = cur_sc->opts[++cur_sc->pc];
  x1 = o1->v7.fd(o1); 
  o2 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.d_dd_f(x1, o2->v7.fd(o2)));
}

static s7_double opt_d_dd_fso(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v4.d_dd_f(o->v5.d_pi_f(slot_value(o->v2.p), integer(slot_value(o->v3.p))), real(slot_value(o->v1.p))));
}

static bool d_dd_fs_combinable(s7_scheme *sc, opt_info *opc)
{
  if ((sc->pc > 1) &&
      (opc == sc->opts[sc->pc - 2]))
    {
      opt_info *o1;
      o1 = sc->opts[sc->pc - 1];
      if (o1->v7.fd == opt_d_pi_ss)
	{
	  opc->v4.d_dd_f = opc->v3.d_dd_f; /* need room for 3 symbols */
	  opc->v2.p = o1->v1.p; 
	  opc->v3.p = o1->v2.p;
	  opc->v5.d_pi_f = o1->v3.d_pi_f;
	  opc->v7.fd = opt_d_dd_fso;
	  sc->pc--;
	  return(true);
	}
    }
  return(false);
}

static s7_double opt_d_dd_ff_o1(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o2;
  s7_double x1;
  x1 = o->v2.d_v_f(o->v1.obj);
  o2 = cur_sc->opts[cur_sc->pc += 2];
  return(o->v3.d_dd_f(x1, o2->v7.fd(o2)));
}

static s7_double opt_d_dd_ff_o2(void *p)
{
  opt_info *o = (opt_info *)p;
  s7_double x1;
  x1 = o->v4.d_v_f(o->v1.obj);
  return(o->v3.d_dd_f(x1, o->v5.d_v_f(o->v2.obj)));
}

static s7_double opt_d_dd_ff_o3(void *p)
{
  opt_info *o = (opt_info *)p;
  s7_double x1;
  x1 = o->v5.d_v_f(o->v1.obj);
  return(o->v4.d_dd_f(x1, o->v6.d_vd_f(o->v2.obj, real(slot_value(o->v3.p)))));
}

static s7_double opt_d_dd_fff(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1, *o2;
  s7_double x1, x2;
  o1 = cur_sc->opts[++cur_sc->pc];
  x1 = o1->v4.d_dd_f(o1->v5.d_pi_f(slot_value(o1->v2.p), integer(slot_value(o1->v3.p))), real(slot_value(o1->v1.p))); /* dd_fso */
  o2 = cur_sc->opts[++cur_sc->pc];
  x2 = o2->v4.d_dd_f(o2->v5.d_pi_f(slot_value(o2->v2.p), integer(slot_value(o2->v3.p))), real(slot_value(o2->v1.p))); /* dd_fso */
  return(o->v3.d_dd_f(x1, x2));
}

static s7_double opt_d_dd_ff_o4(void *p)
{
  opt_info *o = (opt_info *)p;
  s7_double x1;
  x1 = o->v2.d_v_f(o->v1.obj);
  return(o->v3.d_dd_f(x1, o->v8.d_vd_f(o->v5.obj, o->v4.d_v_f(o->v6.obj))));
}

static bool d_dd_ff_combinable(s7_scheme *sc, int32_t start)
{
  opt_info *opc, *o1;
  opc = sc->opts[start - 1];
  o1 = sc->opts[start];
  if (o1->v7.fd == opt_d_v)
    {
      opt_info *o2;
      o2 = sc->opts[start + 1];
      if ((o2->v7.fd == opt_d_v) &&
	  (sc->pc == start + 2))
	{
	  opc->v1.obj = o1->v5.obj;
	  opc->v4.d_v_f = o1->v3.d_v_f;
	  opc->v2.obj = o2->v5.obj;
	  opc->v5.d_v_f = o2->v3.d_v_f;
	  opc->v7.fd = opt_d_dd_ff_o2;
	  sc->pc -= 2;
	}
      else 
	{
	  if ((o2->v7.fd == opt_d_vd_s) &&
	      (sc->pc == start + 2))
	    {
	      opc->v4.d_dd_f = opc->v3.d_dd_f; /* make room for symbols */
	      opc->v1.obj = o1->v5.obj;
	      opc->v5.d_v_f = o1->v3.d_v_f;
	      opc->v2.obj = o2->v5.obj;
	      opc->v6.d_vd_f = o2->v3.d_vd_f;
	      opc->v3.p = o2->v2.p;
	      opc->v7.fd = opt_d_dd_ff_o3;
	      sc->pc -= 2;
	    }
	  else 
	    {
	      if ((o2->v7.fd == opt_d_vd_o) &&
		  (sc->pc == start + 2))
		{
		  opc->v1.obj = o1->v5.obj;
		  opc->v2.d_v_f = o1->v3.d_v_f;
		  opc->v8.d_vd_f = o2->v3.d_vd_f;
		  opc->v4.d_v_f = o2->v4.d_v_f;
		  opc->v5.obj = o2->v5.obj;
		  opc->v6.obj = o2->v6.obj;
		  opc->v7.fd = opt_d_dd_ff_o4;
		  sc->pc -= 2;
		}
	      else
		{
		  opc->v1.obj = o1->v5.obj;
		  opc->v2.d_v_f = o1->v3.d_v_f;
		  opc->v7.fd = opt_d_dd_ff_o1;
		}
	    }
	}
      return(true);
    }
  if (o1->v7.fd == opt_d_dd_fso)
    {
      opt_info *o2;
      o2 = sc->opts[start + 1];
      if (o2->v7.fd == opt_d_dd_fso)
	{
	  opc->v7.fd = opt_d_dd_fff;
	  return(true);
	}
    }
  return(false);
}

static s7_double opt_d_dd_cfo(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.d_dd_f(o->v2.x, o->v4.d_v_f(o->v1.obj)));
}

static s7_double opt_d_dd_cfo1(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.d_dd_f(o->v4.x, o->v5.d_vd_f(o->v6.obj, real(slot_value(o->v2.p)))));
}

static bool d_dd_cf_combinable(s7_scheme *sc, opt_info *opc)
{
  if ((sc->pc > 1) &&
      (opc == sc->opts[sc->pc - 2]))
    {
      opt_info *o1;
      o1 = sc->opts[sc->pc - 1];
      if (o1->v7.fd == opt_d_v)
	{
	  opc->v2.x = opc->v1.x;
	  opc->v6.p = o1->v1.p;
	  opc->v1.obj = o1->v5.obj;
	  opc->v4.d_v_f = o1->v3.d_v_f;
	  opc->v7.fd = opt_d_dd_cfo;
	  sc->pc--;
	  return(true);
	}
      if (o1->v7.fd == opt_d_vd_s)
	{
	  opc->v4.x = opc->v1.x;
	  opc->v1.p = o1->v1.p;
	  opc->v6.obj = o1->v5.obj;
	  opc->v2.p = o1->v2.p;
	  opc->v5.d_vd_f = o1->v3.d_vd_f;
	  opc->v7.fd = opt_d_dd_cfo1;
	  sc->pc--;
	  return(true);
	}
    }
  return(false);
}

static bool d_dd_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x)
{
  s7_d_dd_t func;
  func = s7_d_dd_function(s_func);
  if (func)
    {
      s7_pointer arg1, arg2;
      int32_t start;
      start = sc->pc;
      arg1 = cadr(car_x);
      arg2 = caddr(car_x);
      opc->v3.d_dd_f = func;
      if (is_real(arg1))
	{
	  if (is_real(arg2))
	    {
	      if ((!is_t_real(arg1)) && (!is_t_real(arg2)))
		return(return_false(sc, car_x, __func__, __LINE__));
	      opc->v1.x = s7_number_to_real(sc, arg1);
	      opc->v2.x = s7_number_to_real(sc, arg2);
	      opc->v7.fd = opt_d_dd_cc;
	      return(true);
	    }
	  if (is_symbol(arg2))
	    {
	      opc->v1.p = find_symbol(sc, arg2);
	      if (is_slot(opc->v1.p))
		{
		  if (is_float(slot_value(opc->v1.p)))
		    {
		      opc->v2.x = s7_number_to_real(sc, arg1);
		      opc->v7.fd = opt_d_dd_cs;
		    }
		  else 
		    {
		      opc->v1.x = s7_number_to_real(sc, arg1);     /* cf combine v1.x = arg1 */
		      if (float_optimize(sc, cddr(car_x)))
			{
			  if (!d_dd_cf_combinable(sc, opc))
			    opc->v7.fd = opt_d_dd_cf;
			}
		      else return(return_false(sc, car_x, __func__, __LINE__));
		    }
		  return(true);
		}
	      pc_fallback(sc, start);
	    }
	  else
	    {
	      if (float_optimize(sc, cddr(car_x)))
		{
		  opc->v1.x = s7_number_to_real(sc, arg1);
		  if (!d_dd_cf_combinable(sc, opc))
		    opc->v7.fd = opt_d_dd_cf;
		  return(true);
		}
	      pc_fallback(sc, start);
	    }
	}
      else
	{
	  if (is_symbol(arg1))
	    {
	      opc->v1.p = find_symbol(sc, arg1);
	      if ((is_slot(opc->v1.p)) &&
		  (is_real(slot_value(opc->v1.p))))
		{
		  if (is_float(slot_value(opc->v1.p)))
		    {
		      if (is_real(arg2))
			{
			  opc->v2.x = s7_number_to_real(sc, arg2);
			  opc->v7.fd = opt_d_dd_sc;
			  return(true);
			}
		      if (is_symbol(arg2))
			{
			  opc->v2.p = find_symbol(sc, arg2);
			  if (is_slot(opc->v2.p))
			    {
			      if (is_float(slot_value(opc->v2.p)))
				opc->v7.fd = opt_d_dd_ss;
			      else
				{
				  if (float_optimize(sc, cddr(car_x)))
				    opc->v7.fd = opt_d_dd_sf;
				  else return(return_false(sc, car_x, __func__, __LINE__));
				}
			      return(true);
			    }
			}
		      else
			{
			  if (float_optimize(sc, cddr(car_x)))
			    {
			      opc->v7.fd = opt_d_dd_sf;
			      return(true);
			    }
			  pc_fallback(sc, start);
			}
		    }
		  else
		    {
		      if ((float_optimize(sc, cdr(car_x))) &&
			  (float_optimize(sc, cddr(car_x))))
			{
			  if (!d_dd_ff_combinable(sc, start))
			    opc->v7.fd = opt_d_dd_ff;
			  return(true);
			}
		      pc_fallback(sc, start);
		    }
		}
	      else return(return_false(sc, car_x, __func__, __LINE__)); /* arg1 trouble */
	      /* TODO: d_ddd has a similar bug and check arg2 */
	    }
	  else
	    {
	      if (float_optimize(sc, cdr(car_x)))
		{
		  if (is_real(arg2))
		    {
		      opc->v2.x = s7_number_to_real(sc, arg2);
		      opc->v7.fd = opt_d_dd_fc;
		      return(true);
		    }
		  if (is_symbol(arg2))
		    {
		      opc->v1.p = find_symbol(sc, arg2);
		      if (is_slot(opc->v1.p))
			{
			  if (is_float(slot_value(opc->v1.p)))
			    {
			      if (!d_dd_fs_combinable(sc, opc))
				opc->v7.fd = opt_d_dd_fs;
			    }
			  else
			    {
			      if (float_optimize(sc, cddr(car_x)))
				{
				  if (!d_dd_ff_combinable(sc, start))
				    opc->v7.fd = opt_d_dd_ff;
				}
			      else return(return_false(sc, car_x, __func__, __LINE__));
			    }
			  return(true);
			}
		    }
		  else
		    {
		      if (float_optimize(sc, cddr(car_x)))
			{
			  if (!d_dd_ff_combinable(sc, start))
			    opc->v7.fd = opt_d_dd_ff;
			  return(true);
			}
		    }
		}
	      pc_fallback(sc, start);
	    }
	}
    }
  return(false);
}

/* -------- d_ddd -------- */
static s7_double opt_d_ddd_sss(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v4.d_ddd_f(real(slot_value(o->v1.p)), real(slot_value(o->v2.p)), real(slot_value(o->v3.p))));
}

static s7_double opt_d_ddd_ssf(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v4.d_ddd_f(real(slot_value(o->v1.p)), real(slot_value(o->v2.p)), o1->v7.fd(o1)));
}

static s7_double opt_d_ddd_sff(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  s7_double x1;
  o1 = cur_sc->opts[++cur_sc->pc];
  x1 = o1->v7.fd(o1); 
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v4.d_ddd_f(real(slot_value(o->v1.p)), x1, o1->v7.fd(o1)));
}

static s7_double opt_d_ddd_fff(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1, *o2, *o3;
  s7_double x1, x2;
  o1 = cur_sc->opts[++cur_sc->pc];
  x1 = o1->v7.fd(o1); /* this could involve nested funcs, incrementing pc internally */
  o2 = cur_sc->opts[++cur_sc->pc];
  x2 = o2->v7.fd(o2);
  o3 = cur_sc->opts[++cur_sc->pc];
  return(o->v4.d_ddd_f(x1, x2, o3->v7.fd(o3)));
}

static s7_double opt_d_ddd_fff1(void *p)
{
  opt_info *o = (opt_info *)p;
  s7_double x1, x2, x3;
  x1 = o->v1.d_v_f(o->v2.obj);
  x2 = o->v3.d_v_f(o->v4.obj);
  x3 = o->v5.d_v_f(o->v6.obj);
  return(o->v8.d_ddd_f(x1, x2, x3));
}

static s7_double opt_d_ddd_fff2(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o2, *o3;
  s7_double x1, x2;
  x1 = o->v1.d_v_f(o->v2.obj);
  cur_sc->pc += 2;
  o2 = cur_sc->opts[cur_sc->pc];
  x2 = o2->v7.fd(o2);
  o3 = cur_sc->opts[++cur_sc->pc];
  return(o->v8.d_ddd_f(x1, x2, o3->v7.fd(o3)));
}

static bool d_ddd_fff_combinable(s7_scheme *sc, opt_info *opc, int32_t start)
{
  if (sc->opts[start]->v7.fd == opt_d_v)
    {
      opt_info *o1;
      opc->v8.d_ddd_f = opc->v4.d_ddd_f;
      o1 = sc->opts[start];
      opc->v1.d_v_f = o1->v3.d_v_f;
      opc->v2.obj = o1->v5.obj;
      if ((sc->opts[start + 1]->v7.fd == opt_d_v) &&
	  (sc->opts[start + 2]->v7.fd == opt_d_v))
	{
	  opc->v7.fd = opt_d_ddd_fff1;
	  o1 = sc->opts[start + 1];
	  opc->v3.d_v_f = o1->v3.d_v_f;
	  opc->v4.obj = o1->v5.obj;
	  o1 = sc->opts[start + 2];
	  opc->v5.d_v_f = o1->v3.d_v_f;
	  opc->v6.obj = o1->v5.obj;
	  sc->pc -= 3;
	  return(true);
	}
      opc->v7.fd = opt_d_ddd_fff2;
      return(true);
    }
  return(false);
}

static bool d_ddd_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x)
{
  s7_d_ddd_t f;
  
  f = s7_d_ddd_function(s_func);
  if (f)
    {
      int32_t start;
      s7_pointer arg1, arg2;
      arg1 = cadr(car_x);
      arg2 = caddr(car_x);
      start = sc->pc;
      opc->v4.d_ddd_f = f;
      if (is_symbol(arg1))
	{
	  s7_pointer slot;
	  slot = find_symbol(sc, arg1);
	  opc->v1.p = slot;
	  if ((is_slot(slot)) &&
	      (is_float(slot_value(slot))))
	    {
	      if (is_symbol(arg2))
		{
		  s7_pointer slot;
		  slot = find_symbol(sc, arg2);
		  opc->v2.p = slot;
		  if ((is_slot(slot)) &&
		      (is_float(slot_value(slot))))
		    {
		      s7_pointer arg3;
		      arg3 = cadddr(car_x);
		      if (is_symbol(arg3))
			{
			  s7_pointer slot;
			  slot = find_symbol(sc, arg3);
			  opc->v3.p = slot;
			  if ((is_slot(slot)) &&
			      (is_float(slot_value(slot))))
			    {
			      opc->v7.fd = opt_d_ddd_sss;
			      return(true);
			    }
			}
		      if (float_optimize(sc, cdddr(car_x)))
			{
			  opc->v7.fd = opt_d_ddd_ssf;
			  return(true);
			}
		      pc_fallback(sc, start);
		    }
		}
	      if ((float_optimize(sc, cddr(car_x))) &&
		  (float_optimize(sc, cdddr(car_x))))
		{
		  opc->v7.fd = opt_d_ddd_sff;
		  return(true);
		}
	    }
	  pc_fallback(sc, start);
	}
      if ((float_optimize(sc, cdr(car_x))) &&
	  (float_optimize(sc, cddr(car_x))) &&
	  (float_optimize(sc, cdddr(car_x))))
	{
	  if (!d_ddd_fff_combinable(sc, opc, start))
	    {
	    opc->v7.fd = opt_d_ddd_fff;
	    /* fprintf(stderr, "%s\n", DISPLAY(car_x)); */
	    /* (* (env pulsef) (blackman pulse2) (polywave gen (rand-interp rnd)))
	     *  (* (env e)...) is common = opt_d_v: v3 v5 -> opc 
	     * (+ k (* 2 alpha) -2.0) (* scl ang ang) (- n k 1)
	     */
	    }
	  return(true);
	}
      pc_fallback(sc, start);
    }
  return(false);
}

/* -------- d_pid -------- */
static s7_double opt_d_pid_ssf(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v4.d_pid_f(slot_value(o->v1.p), integer(slot_value(o->v2.p)), o1->v7.fd(o1)));
}

static s7_double opt_d_pid_sss(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v4.d_pid_f(slot_value(o->v1.p), integer(slot_value(o->v2.p)), real(slot_value(o->v3.p))));
}

static s7_double opt_d_pid_sff(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  s7_int pos;
  o1 = cur_sc->opts[++cur_sc->pc];
  pos = o1->v7.fi(o1);
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v4.d_pid_f(slot_value(o->v1.p), pos, o1->v7.fd(o1)));
}

static s7_double opt_d_pid_sso(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v4.d_pid_f(slot_value(o->v1.p), integer(slot_value(o->v2.p)), o->v5.d_v_f(o->v3.obj)));
}

static s7_double opt_d_pid_ss_ss(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v4.d_pid_f(slot_value(o->v1.p), 
		       integer(slot_value(o->v2.p)), 
		       o->v3.d_pi_f(slot_value(o->v5.p), 
				    integer(slot_value(o->v6.p)))));
}

static bool d_pid_ssf_combinable(s7_scheme *sc, opt_info *opc)
{
  if ((sc->pc > 1) &&
      (opc == sc->opts[sc->pc - 2]))
    {
      opt_info *o1;
      o1 = sc->opts[sc->pc - 1];
      if (o1->v7.fd == opt_d_v)
	{
	  opc->v6.p = o1->v1.p;
	  opc->v3.obj = o1->v5.obj;
	  opc->v5.d_v_f = o1->v3.d_v_f;
	  opc->v7.fd = opt_d_pid_sso;
	  sc->pc--;
	  return(true);
	}
      if (o1->v7.fd == opt_d_pi_ss)
	{
	  opc->v3.d_pi_f = o1->v3.d_pi_f;
	  opc->v5.p = o1->v1.p;
	  opc->v6.p = o1->v2.p;
	  opc->v7.fd = opt_d_pid_ss_ss;
	  sc->pc--;
	  return(true);
	}
    }
  return(false);
}

static bool opt_float_vector_set(s7_scheme *sc, opt_info *opc, s7_pointer v, s7_pointer indexp, s7_pointer valp)
{
  s7_pointer settee;
  settee = find_symbol(sc, v);
  if (is_slot(settee))
    {
      opc->v1.p = settee;
      if ((is_float_vector(slot_value(settee))) &&
	  (vector_rank(slot_value(settee)) == 1))
	{
	  opc->v4.d_pid_f = float_vector_set_d;
	  if (is_symbol(car(indexp)))
	    {
	      s7_pointer slot;
	      slot = find_symbol(sc, car(indexp));
	      if ((is_slot(slot)) &&
		  (is_integer(slot_value(slot))))
		{
		  opc->v2.p = slot;
		  if ((is_step_end(opc->v2.p)) &&
		      (denominator(slot_value(opc->v2.p)) <= vector_length(slot_value(settee))))
		    opc->v4.d_pid_f = float_vector_set_unchecked;
		  if (is_symbol(car(valp)))
		    {
		      s7_pointer val_slot;
		      val_slot = find_symbol(sc, car(valp));
		      if ((is_slot(val_slot)) &&
			  (is_float(slot_value(val_slot))))
			{
			  opc->v3.p = val_slot;
			  opc->v7.fd = opt_d_pid_sss;
			  return(true);
			}
		    }
		  if (float_optimize(sc, valp))
		    {
		      if (!d_pid_ssf_combinable(sc, opc))
			opc->v7.fd = opt_d_pid_ssf;
		      return(true);
		    }
		}
	    }
	  else
	    {
	      if ((int_optimize(sc, indexp)) &&
		  (float_optimize(sc, valp)))
		{
		  opc->v7.fd = opt_d_pid_sff;
		  return(true);
		}
	    }
	}
    }
  return(false);
}

static bool d_pid_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x)
{
  s7_d_pid_t f;
  f = s7_d_pid_function(s_func);
  if ((f) &&
      (is_symbol(cadr(car_x))))
    {
      s7_pointer head;
      int32_t start;
      start = sc->pc;

      head = car(car_x);
      opc->v4.d_pid_f = f;
      
      if (head == sc->float_vector_set_symbol)
	return(opt_float_vector_set(sc, opc, cadr(car_x), cddr(car_x), cdddr(car_x)));
      
      opc->v1.p = find_symbol(sc, cadr(car_x));
      if (is_slot(opc->v1.p))
	{
	  if (is_symbol(caddr(car_x)))
	    {
	      opc->v2.p = find_symbol(sc, caddr(car_x));
	      if ((is_slot(opc->v2.p)) &&
		  (is_integer(slot_value(opc->v2.p))))
		{
		  if (is_symbol(cadddr(car_x)))
		    {
		      s7_pointer val_slot;
		      val_slot = find_symbol(sc, cadddr(car_x));
		      if ((is_slot(val_slot)) &&
			  (is_float(slot_value(val_slot))))
			{
			  opc->v3.p = val_slot;
			  opc->v7.fd = opt_d_pid_sss;
			  return(true);
			}
		    }
		  if (float_optimize(sc, cdddr(car_x)))
		    {
		      if (!d_pid_ssf_combinable(sc, opc))
			opc->v7.fd = opt_d_pid_ssf;
		      return(true);
		    }
		  pc_fallback(sc, start);
		}
	    }
	  else
	    {
	      if ((int_optimize(sc, cddr(car_x))) &&
		  (float_optimize(sc, cdddr(car_x))))
		{
		  opc->v7.fd = opt_d_pid_sff;
		  return(true);
		}
	      pc_fallback(sc, start);
	    }
	}
    }
  return(false);
}

/* -------- d_vid -------- */
static s7_double opt_d_vid_ssf(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  /* PERHAPS: check index type as above */
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v4.d_vid_f(o->v5.obj, integer(slot_value(o->v2.p)), o1->v7.fd(o1)));
}

static inline s7_double opt_fmv(void *p)
{
  /* d_vid_ssf -> d_dd_ff_o1 -> d_vd_o1 -> d_dd_ff_o3 */
  opt_info *o = (opt_info *)p;
  opt_info *o1, *o2, *o3;
  s7_double amp_env, index_env, vib;

  o1 = cur_sc->opts[cur_sc->pc + 1];
  o2 = cur_sc->opts[cur_sc->pc + 3];
  o3 = cur_sc->opts[cur_sc->pc += 5];
  amp_env = o1->v2.d_v_f(o1->v1.obj); 
  vib = real(slot_value(o2->v2.p));
  index_env = o3->v5.d_v_f(o3->v1.obj);

  return(o->v4.d_vid_f(o->v5.obj, 
		       integer(slot_value(o->v2.p)),
		       o1->v3.d_dd_f(amp_env,
				     o2->v3.d_vd_f(o2->v5.obj,
						   o2->v4.d_dd_f(vib,
								 o3->v4.d_dd_f(index_env,
									       o3->v6.d_vd_f(o3->v2.obj, vib)))))));
}

static bool d_vid_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x)
{
  if ((is_symbol(cadr(car_x))) &&
      (is_symbol(caddr(car_x))))
    {
      s7_d_vid_t flt;
      flt = s7_d_vid_function(s_func);
      if (flt)
	{
	  s7_pointer sig;
	  opc->v4.d_vid_f = flt;
	  sig = s7_procedure_signature(sc, s_func);
	  if (is_pair(sig))
	    {
	      int32_t start;
	      s7_pointer slot, obj, checker;
	      start = sc->pc;
	      checker = s7_symbol_value(sc, cadr(sig));
	      slot = find_symbol(sc, cadr(car_x));
	      obj = slot_value(slot);
	      if (s7_apply_function(sc, checker, set_plist_1(sc, obj)) == sc->T)
		{
		  opc->v7.fd = opt_d_vid_ssf;
		  opc->v1.p = slot;
		  opc->v5.obj = (void *)s7_c_object_value(obj);
		  opc->v2.p = find_symbol(sc, caddr(car_x));
		  if ((is_slot(opc->v2.p)) &&
		      (is_integer(slot_value(opc->v2.p))) &&
		      (float_optimize(sc, cdddr(car_x))))
		    {
		      opt_info *o1;
		      o1 = sc->opts[start];
		      if (o1->v7.fd == opt_d_dd_ff_o1)
			{
			  o1 = sc->opts[start + 2];
			  if (o1->v7.fd == opt_d_vd_o1)
			    {
			      o1 = sc->opts[start + 4];
			      if (o1->v7.fd == opt_d_dd_ff_o3)
				opc->v7.fd = opt_fmv;
			    }
			}
		      return(true);
		    }
		}
	      pc_fallback(sc, start);
	    }
	}
    }
  return(false);
}
  
/* -------- d_vdd -------- */
static s7_double opt_d_vdd_ff(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1, *o2;
  s7_double x1;
  o1 = cur_sc->opts[++cur_sc->pc];
  x1 = o1->v7.fd(o1);
  o2 = cur_sc->opts[++cur_sc->pc];
  return(o->v4.d_vdd_f(o->v5.obj, x1, o2->v7.fd(o2)));
}

static bool d_vdd_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x)
{
  s7_d_vdd_t flt;
  flt = s7_d_vdd_function(s_func);
  if (flt)
    {
      s7_pointer sig;
      opc->v4.d_vdd_f = flt;
      sig = s7_procedure_signature(sc, s_func);
      if (is_pair(sig))
	{
	  s7_pointer slot, obj, checker;
	  checker = s7_symbol_value(sc, cadr(sig));
	  slot = find_symbol(sc, cadr(car_x));
	  obj = slot_value(slot);
	  if (s7_apply_function(sc, checker, set_plist_1(sc, obj)) == sc->T)
	    {
	      int32_t start;
	      start = sc->pc;
	      if ((float_optimize(sc, cddr(car_x))) &&
		  (float_optimize(sc, cdddr(car_x))))
		{
		  opc->v1.p = slot;
		  opc->v5.obj = (void *)s7_c_object_value(obj);
		  opc->v7.fd = opt_d_vdd_ff;
		  return(true);
		}
	      pc_fallback(sc, start);
	    }
	}
    }
  return(false);
}


/* -------- d_dddd -------- */
static s7_double opt_d_dddd_ffff(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1, *o2, *o3, *o4;
  s7_double x1, x2, x3;
  o1 = cur_sc->opts[++cur_sc->pc];
  x1 = o1->v7.fd(o1);
  o2 = cur_sc->opts[++cur_sc->pc];
  x2 = o2->v7.fd(o2);
  o3 = cur_sc->opts[++cur_sc->pc];
  x3 = o3->v7.fd(o3);
  o4 = cur_sc->opts[++cur_sc->pc];
  return(o->v1.d_dddd_f(x1, x2, x3, o4->v7.fd(o4)));
}

static bool d_dddd_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x)
{
  s7_d_dddd_t f;
  f = s7_d_dddd_function(s_func);
  if (f)
    {
      if ((float_optimize(sc, cdr(car_x))) &&
	  (float_optimize(sc, cddr(car_x))) &&
	  (float_optimize(sc, cdddr(car_x))) &&
	  (float_optimize(sc, cddddr(car_x))))
	{
	  opc->v1.d_dddd_f = f;
	  opc->v7.fd = opt_d_dddd_ffff;
	  return(true);
	}
    }
  return(false);
}

/* -------- d_add|multiply|subtract_any ------- */
static s7_double opt_d_add_any_f(void *p)
{
  opt_info *o = (opt_info *)p;
  s7_double sum = 0.0;
  int32_t i;
  for (i = 0; i < o->v1.i; i++)
    {
      opt_info *o1;
      o1 = cur_sc->opts[++cur_sc->pc];
      sum += o1->v7.fd(o1);
    }
  return(sum);
}

static s7_double opt_d_subtract_any_f(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  s7_double sum;
  int32_t i;
  o1 = cur_sc->opts[++cur_sc->pc];
  sum = o1->v7.fd(o1);
  for (i = 1; i < o->v1.i; i++)
    {
      o1 = cur_sc->opts[++cur_sc->pc];
      sum -= o1->v7.fd(o1);
    }
  return(sum);
}

static s7_double opt_d_add2(void *p)
{
  s7_double sum;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  sum = o1->v7.fd(o1);
  o1 = cur_sc->opts[++cur_sc->pc];
  return(sum + o1->v7.fd(o1));
}

static s7_double opt_d_mul2(void *p)
{
  s7_double sum;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  sum = o1->v7.fd(o1);
  o1 = cur_sc->opts[++cur_sc->pc];
  return(sum * o1->v7.fd(o1));
}

static s7_double opt_d_multiply_any_f(void *p)
{
  opt_info *o = (opt_info *)p;
  s7_double sum = 1.0;
  int32_t i;
  for (i = 0; i < o->v1.i; i++)
    {
      opt_info *o1;
      o1 = cur_sc->opts[++cur_sc->pc];
      sum *= o1->v7.fd(o1);
    }
  return(sum);
}

static bool d_add_any_ok(s7_scheme *sc, opt_info *opc, s7_pointer car_x, int32_t len)
{
  s7_pointer head;
  int32_t start;
  start = sc->pc;
  head = car(car_x);
  if ((head == sc->add_symbol) ||
      (head == sc->multiply_symbol))
    {
      s7_pointer p;
      int32_t cur_len = 0;
      for (p = cdr(car_x); is_pair(p); p = cdr(p))
	{
	  if (is_pair(cdr(p)))
	    {
	      if (!float_optimize(sc, set_plist_1(sc, set_elist_3(sc, head, car(p), cadr(p)))))
		break;
	      cur_len++;
	      p = cdr(p);
	    }
	  else
	    {
	      if (!float_optimize(sc, p))
		break;
	      cur_len++;
	    }
	}
      if (is_null(p))
	{
	  opc->v1.i = cur_len;
	  if (cur_len == 2)
	    opc->v7.fd = (head == sc->add_symbol) ? opt_d_add2 :opt_d_mul2;
	  else opc->v7.fd = (head == sc->add_symbol) ? opt_d_add_any_f : opt_d_multiply_any_f;
	  return(true);
	}
    }
  else
    {
      if (head == sc->subtract_symbol)
	{
	  s7_pointer p;
	  opc->v1.i = (len - 1);
	  for (p = cdr(car_x); is_pair(p); p = cdr(p))
	    if (!float_optimize(sc, p))
	      break;
	  if (is_null(p))
	    {
	      opc->v7.fd = opt_d_subtract_any_f;
	      return(true);
	    }
	}
    }
  pc_fallback(sc, start);
  return(false);
}

/* -------- float_all_x -------- */
static s7_double opt_unwrap_float(void *p)
{
  opt_info *o = (opt_info *)p;
  return(s7_number_to_real(cur_sc, o->v2.all_f(cur_sc, car(o->v1.p))));
}

static bool float_all_x_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer expr)
{
  s7_pointer sig;
  s7_function opt;
  sig = s7_procedure_signature(sc, s_func);	  
  if ((is_pair(sig)) &&
      ((car(sig) == sc->is_float_symbol) ||
       (car(sig) == sc->is_real_symbol)))
    {
      /* fallback on the more general case (all_x_eval, but still guaranteed to be a number) */
      opt = all_x_optimize(sc, expr);
      if (opt)
	{
	  opc->v2.all_f = opt;
	  opc->v7.fd = opt_unwrap_float;
	  opc->v1.p = expr;
	  return(true);
	}
    }
  return(false);
}

/* -------- d_syntax -------- */
static s7_double opt_set_d_d_f(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  s7_double x;
  o1 = cur_sc->opts[++cur_sc->pc];
  x = o1->v7.fd(o1);
  slot_set_value(o->v1.p, make_real(cur_sc, x));
  return(x);
}

static s7_double opt_set_d_d_fm(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  s7_double x;
  o1 = cur_sc->opts[++cur_sc->pc];
  x = o1->v7.fd(o1);
  real(slot_value(o->v1.p)) = x;
  return(x);
}

static bool d_syntax_ok(s7_scheme *sc, s7_pointer car_x, int32_t len)
{
  if ((len == 3) &&
      (car(car_x) == sc->set_symbol))
    {
      opt_info *opc;
      opc = alloc_opo(sc, car_x);
      if (is_symbol(cadr(car_x)))
	{
	  s7_pointer settee;
	  if ((is_immutable(cadr(car_x))) ||
	      (symbol_has_accessor(cadr(car_x))))
	    return(return_false(sc, car_x, __func__, __LINE__));
	  settee = find_symbol(sc, cadr(car_x));
	  if (is_slot(settee))
	    {
	      opc->v1.p = settee;
	      if ((is_t_real(slot_value(settee))) &&
		  (float_optimize(sc, cddr(car_x))))
		{
		  if (is_mutable(slot_value(opc->v1.p)))
		    opc->v7.fd = opt_set_d_d_fm;
		  else opc->v7.fd = opt_set_d_d_f;
		  return(true);
		}
	    }
	}
      else
	{
	  /* if is_pair(settee) get setter */
	  if ((is_pair(cadr(car_x))) &&
	      (is_symbol(caadr(car_x))) &&
	      (is_pair(cdadr(car_x))) &&
	      (is_null(cddadr(car_x))))
	    return(opt_float_vector_set(sc, opc, caadr(car_x), cdadr(car_x), cddr(car_x)));
	}
    }
  return(return_false(sc, car_x, __func__, __LINE__));
}

static bool d_implicit_ok(s7_scheme *sc, s7_pointer car_x, int32_t len)
{
  s7_pointer s_slot;
  s_slot = find_symbol(sc, car(car_x));
  if ((is_slot(s_slot)) &&
      (len == 2) &&
      (is_float_vector(slot_value(s_slot))) &&
      (vector_rank(slot_value(s_slot)) == 1))
    {
      /* implicit float-vector-ref */
      opt_info *opc;
      opc = alloc_opo(sc, car_x);
      opc->v1.p = s_slot;
      if (is_symbol(cadr(car_x)))
	{
	  s7_pointer slot;
	  slot = find_symbol(sc, cadr(car_x));
	  if ((is_slot(slot)) &&
	      (is_integer(slot_value(slot))))
	    {
	      opc->v7.fd = opt_d_pi_ss;
	      opc->v3.d_pi_f = float_vector_ref_d;
	      opc->v2.p = slot;
	      if ((is_step_end(opc->v2.p)) &&
		  (denominator(slot_value(opc->v2.p)) <= vector_length(slot_value(opc->v1.p))))
		opc->v3.d_pi_f = float_vector_ref_unchecked;
	      return(true);
	    }
	}
      else
	{
	  if (int_optimize(sc, cdr(car_x)))
	    {
	      opc->v7.fd = opt_d_pi_sf;
	      opc->v3.d_pi_f = float_vector_ref_d;
	    }
	}
    }
  return(return_false(sc, car_x, __func__, __LINE__));
}


/* -------------------------------- bool opts -------------------------------- */
static bool opt_b_t(void *p)
{
  return(true);
}

static bool opt_b_f(void *p)
{
  return(false);
}

static bool opt_b_s(void *p)
{
  opt_info *o = (opt_info *)p;
  return(slot_value(o->v1.p) != cur_sc->F);
}

static bool opt_bool_not_pair(s7_scheme *sc, s7_pointer car_x)
{
  opt_info *opc;
  s7_pointer p;
  if (!is_symbol(car_x))
    {
      if (!s7_is_boolean(car_x))
	return(return_false(sc, car_x, __func__, __LINE__)); /* i.e. use cell_optimize */
      opc = alloc_opo(sc, car_x);
      opc->v7.fb = ((car_x == sc->F) ? opt_b_f : opt_b_t);
      return(true);
    }
  p = find_symbol(sc, car_x);
  if (is_slot(p))
    {
      if ((has_methods(slot_value(p))) ||
	  (!s7_is_boolean(slot_value(p))))
	return(return_false(sc, car_x, __func__, __LINE__));
      opc = alloc_opo(sc, car_x);
      opc->v1.p = p;
      opc->v7.fb = opt_b_s;
      return(true);
    }
  return(return_false(sc, car_x, __func__, __LINE__));
}

/* -------- b_idp -------- */
static bool opt_b_i_s(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v2.b_i_f(integer(slot_value(o->v1.p))));
}

static bool opt_b_i_f(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v2.b_i_f(o1->v7.fi(o1)));
}

static bool opt_b_d_s(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v2.b_d_f(real(slot_value(o->v1.p))));
}

static bool opt_b_d_f(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v2.b_d_f(o1->v7.fd(o1)));
}

static bool opt_b_p_s(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v2.b_p_f(slot_value(o->v1.p)));
}

static bool opt_b_p_f(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v2.b_p_f(o1->v7.fp(o1)));
}

#if (!WITH_GMP)
static bool opt_zero_mod(void *p)
{
  opt_info *o = (opt_info *)p;
  s7_int x;
  x = integer(slot_value(o->v1.p));
  return((x % o->v2.i) == 0);
}
#endif


static bool b_idp_ok(s7_scheme *sc, s7_pointer s_func, s7_pointer car_x, s7_pointer arg_type)
{
  int32_t cur_index;
  s7_b_p_t bpf;
  opt_info *opc;
  
  opc = alloc_opo(sc, car_x);
  cur_index = sc->pc;
  
  if (arg_type == sc->is_integer_symbol)
    {
      s7_b_i_t bif;
      bif = s7_b_i_function(s_func);
      if (bif)
	{
	  opc->v2.b_i_f = bif;
	  if (is_symbol(cadr(car_x)))
	    {
	      opc->v1.p = find_symbol(sc, cadr(car_x));
	      opc->v7.fb = opt_b_i_s;
	      return(true);
	    }
	  if (int_optimize(sc, cdr(car_x)))
	    {
#if (!WITH_GMP)
	      opt_info *o1;
	      o1 = sc->opts[sc->pc - 1];
	      if ((o1->v7.fi == opt_i_ii_sc) &&
		  (o1->v3.i_ii_f == modulo_i_ii_direct))
		{
		  opc->v7.fb = opt_zero_mod;
		  opc->v1.p = o1->v1.p;
		  opc->v2.i = o1->v2.i;
		  sc->pc--;
		}
	      else 
#endif
		opc->v7.fb = opt_b_i_f;
	      return(true);
	    }
	}
    }
  else
    {
      if (arg_type == sc->is_float_symbol)
	{
	  s7_b_d_t bdf;
	  bdf = s7_b_d_function(s_func);
	  if (bdf)
	    {
	      opc->v2.b_d_f = bdf;
	      if (is_symbol(cadr(car_x)))
		{
		  opc->v1.p = find_symbol(sc, cadr(car_x));
		  opc->v7.fb = opt_b_d_s;
		  return(true);
		}
	      if (float_optimize(sc, cdr(car_x)))
		{
		  opc->v7.fb = opt_b_d_f;
		  return(true);
		}
	    }
	}
    }
  pc_fallback(sc, cur_index);
  
  bpf = s7_b_p_function(s_func);
  if (bpf)
    {
      opc->v2.b_p_f = bpf;
      if (is_symbol(cadr(car_x)))
	{
	  s7_pointer p;
	  p = find_symbol(sc, cadr(car_x));
	  if ((!is_slot(p)) || 
	      (has_methods(slot_value(p))))
	    return(return_false(sc, car_x, __func__, __LINE__));
	  opc->v1.p = p;
	  opc->v7.fb = opt_b_p_s;
	  return(true);
	}
      if (cell_optimize(sc, cdr(car_x)))
	{
	  opc->v7.fb = opt_b_p_f;
	  return(true);
	}
    }
  return(false);
}


/* -------- b_pp -------- */
static bool opt_b_pp_ff(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  s7_pointer p1;
  o1 = cur_sc->opts[++cur_sc->pc];
  p1 = o1->v7.fp(o1);
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.b_pp_f(p1, o1->v7.fp(o1)));
}

static bool opt_b_pp_sf(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.b_pp_f(slot_value(o->v1.p), o1->v7.fp(o1)));
}

static bool opt_b_pp_fs(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.b_pp_f(o1->v7.fp(o1), slot_value(o->v1.p)));
}

static bool opt_b_pp_ss(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.b_pp_f(slot_value(o->v1.p), slot_value(o->v2.p)));
}

static bool opt_b_pp_sc(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.b_pp_f(slot_value(o->v1.p), o->v2.p));
}

static bool opt_b_pp_sfo(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.b_pp_f(slot_value(o->v1.p), o->v4.p_p_f(slot_value(o->v2.p))));
}

static s7_pointer opt_p_p_s(void *p); /* TODO: remove this when p* precedes b* */

static bool b_pp_sf_combinable(s7_scheme *sc, opt_info *opc)
{
  if ((sc->pc > 1) &&
      (opc == sc->opts[sc->pc - 2]))
    {
      opt_info *o1;
      o1 = sc->opts[sc->pc - 1];
      if (o1->v7.fp == opt_p_p_s)
	{
	  opc->v2.p = o1->v1.p; 
	  opc->v4.p_p_f = o1->v2.p_p_f;
	  opc->v7.fb = opt_b_pp_sfo;
	  sc->pc--;
	  return(true);
	}
    }
  return(false);
}

static bool opt_b_pp_ffo(void *p)
{
  opt_info *o = (opt_info *)p;
  s7_pointer b1;
  b1 = o->v4.p_p_f(slot_value(o->v1.p));
  return(o->v3.b_pp_f(b1, o->v5.p_p_f(slot_value(o->v2.p))));
}

static bool b_pp_ff_combinable(s7_scheme *sc, opt_info *opc)
{
  if ((sc->pc > 2) &&
      (opc == sc->opts[sc->pc - 3]))
    {
      opt_info *o1, *o2;
      o1 = sc->opts[sc->pc - 2];
      o2 = sc->opts[sc->pc - 1];
      if ((o1->v7.fp == opt_p_p_s) &&
	  (o2->v7.fp == opt_p_p_s))
	{
	  opc->v1.p = o1->v1.p; 
	  opc->v4.p_p_f = o1->v2.p_p_f;
	  opc->v2.p = o2->v1.p;
	  opc->v5.p_p_f = o2->v2.p_p_f;
	  opc->v7.fb = opt_b_pp_ffo;
	  sc->pc -= 2;
	  return(true);
	}
    }
  return(false);
}

static bool b_pp_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x, s7_pointer arg1, s7_pointer arg2)
{
  if ((is_symbol(arg1)) &&
      (is_symbol(arg2)))
    {
      opc->v1.p = find_symbol(sc, arg1);
      opc->v2.p = find_symbol(sc, arg2);
      if ((is_slot(opc->v1.p)) &&
	  (!has_methods(slot_value(opc->v1.p))) &&
	  (is_slot(opc->v2.p)) &&
	  (!has_methods(slot_value(opc->v2.p))))
	{
	  opc->v7.fb = opt_b_pp_ss;
	  return(true);
	}
    }
  if (is_symbol(arg1))
    {
      opc->v1.p = find_symbol(sc, arg1);
      if ((!is_slot(opc->v1.p)) ||
	  (has_methods(slot_value(opc->v1.p))))
	return(return_false(sc, car_x, __func__, __LINE__));
      if ((!is_symbol(arg2)) && 
	  (!is_pair(arg2)))
	{
	  opc->v2.p = arg2;
	  opc->v7.fb = opt_b_pp_sc;
	  return(true);
	}
      if (cell_optimize(sc, cddr(car_x)))
	{
	  if (!b_pp_sf_combinable(sc, opc))
	    opc->v7.fb = opt_b_pp_sf;
	  return(true);
	}
      return(return_false(sc, car_x, __func__, __LINE__));
    }
  if (is_symbol(arg2))
    {
      if (cell_optimize(sc, cdr(car_x)))
	{
	  opc->v1.p = find_symbol(sc, arg2);
	  if ((!is_slot(opc->v1.p)) ||
	      (has_methods(slot_value(opc->v1.p))))
	    return(return_false(sc, car_x, __func__, __LINE__));
	  opc->v7.fb = opt_b_pp_fs;
	  return(true);
	}
      return(return_false(sc, car_x, __func__, __LINE__));
    }
  
  if ((cell_optimize(sc, cdr(car_x))) &&
      (cell_optimize(sc, cddr(car_x))))
    {
      if (!b_pp_ff_combinable(sc, opc))
	opc->v7.fb = opt_b_pp_ff;
      if ((s7_b_pp_direct_function(s_func)) &&
	  (is_pair(cadr(car_x))) &&
	  (is_pair(caddr(car_x))) &&
	  (caadr(car_x) == caaddr(car_x)))
	{
	  s7_pointer arg_slot, call_sig;
	  call_sig = s7_procedure_signature(sc, s_func);
	  arg_slot = find_symbol(sc, caadr(car_x));
	  if ((is_c_function(slot_value(arg_slot))) &&
	      (cadr(call_sig) == caddr(call_sig)))
	    {
	      s7_pointer arg_sig;
	      arg_sig = s7_procedure_signature(sc, slot_value(arg_slot));
	      if ((is_pair(arg_sig)) &&
		  (car(arg_sig) == cadr(call_sig)))
		opc->v3.b_pp_f = s7_b_pp_direct_function(s_func);
	    }
	}
      return(true);
    }
  return(false);
}


/* -------- b_pi -------- */
static bool opt_b_pi_fs(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v2.b_pi_f(o1->v7.fp(o1), integer(slot_value(o->v1.p))));
}

static bool b_pi_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x, s7_pointer arg2)
{
  s7_b_pi_t bpif;
  bpif = s7_b_pi_function(s_func);
  if (bpif)
    {
      opc->v1.p = find_symbol(sc, arg2); /* slot checked in opt_arg_type */
      if (cell_optimize(sc, cdr(car_x)))
	{
	  opc->v2.b_pi_f = bpif;
	  opc->v7.fb = opt_b_pi_fs;
	  return(true);
	}
    }
  return(false);
}


/* -------- b_dd -------- */
static bool opt_b_dd_ss(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.b_dd_f(real(slot_value(o->v1.p)), real(slot_value(o->v2.p))));
}

static bool opt_b_dd_sc(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.b_dd_f(real(slot_value(o->v1.p)), o->v2.x));
}

static bool opt_b_dd_sf(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.b_dd_f(real(slot_value(o->v1.p)), o1->v7.fd(o1)));
}

static bool opt_b_dd_fs(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.b_dd_f(o1->v7.fd(o1), real(slot_value(o->v1.p))));
}

static bool opt_b_dd_fc(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.b_dd_f(o1->v7.fd(o1), o->v1.x));
}

static bool opt_b_dd_ff(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  s7_double x1;
  o1 = cur_sc->opts[++cur_sc->pc];
  x1 = o1->v7.fd(o1);
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.b_dd_f(x1, o1->v7.fd(o1)));
}

static bool b_dd_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x, s7_pointer arg1, s7_pointer arg2)
{
  s7_b_dd_t bif;
  int32_t cur_index;
  cur_index = sc->pc;
  bif = s7_b_dd_function(s_func);
  if (bif)
    {
      opc->v3.b_dd_f = bif;
      
      if (is_symbol(arg1))
	{
	  opc->v1.p = find_symbol(sc, arg1);
	  if (is_symbol(arg2))
	    {
	      opc->v2.p = find_symbol(sc, arg2);
	      opc->v7.fb = opt_b_dd_ss;
	      return(true);
	    }
	  if (is_real(arg2))
	    {
	      opc->v2.x = s7_number_to_real(cur_sc, arg2);
	      opc->v7.fb = opt_b_dd_sc;
	      return(true);
	    }
	  if (float_optimize(sc, cddr(car_x)))
	    {
	      opc->v7.fb = opt_b_dd_sf;
	      return(true);
	    }
	}
      pc_fallback(sc, cur_index);
      if (float_optimize(sc, cdr(car_x)))
	{
	  if (is_symbol(arg2))
	    {
	      opc->v1.p = find_symbol(sc, arg2);
	      opc->v7.fb = opt_b_dd_fs;
	      return(true);
	    }
	  if (is_real(arg2))
	    {
	      opc->v1.x = s7_number_to_real(cur_sc, arg2);
	      opc->v7.fb = opt_b_dd_fc;
	      return(true);
	    }
	  if (float_optimize(sc, cddr(car_x)))
	    {
	      opc->v7.fb = opt_b_dd_ff;
	      return(true);
	    }
	}
    }
  pc_fallback(sc, cur_index);
  return(false);
}


/* -------- b_ii -------- */
static bool opt_b_ii_ss(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.b_ii_f(integer(slot_value(o->v1.p)), integer(slot_value(o->v2.p))));
}

static bool opt_b_ii_sc(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.b_ii_f(integer(slot_value(o->v1.p)), o->v2.i));
}

static bool opt_b_ii_ff(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  s7_int i1;
  o1 = cur_sc->opts[++cur_sc->pc];
  i1 = o1->v7.fi(o1);
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.b_ii_f(i1, o1->v7.fi(o1)));
}

static bool opt_b_ii_fs(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.b_ii_f(o1->v7.fi(o1), integer(slot_value(o->v2.p))));
}

static bool opt_b_ii_fc(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.b_ii_f(o1->v7.fi(o1), o->v2.i));
}

static bool opt_b_ii_sf(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.b_ii_f(integer(slot_value(o->v1.p)), o1->v7.fi(o1)));
}

static bool b_ii_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x, s7_pointer arg1, s7_pointer arg2)
{
  s7_b_ii_t bif;
  bif = s7_b_ii_function(s_func);
  if (bif)
    {
      opc->v3.b_ii_f = bif;
      if (is_symbol(arg1))
	{
	  opc->v1.p = find_symbol(sc, arg1);
	  if (is_symbol(arg2))
	    {
	      opc->v2.p = find_symbol(sc, arg2);
	      opc->v7.fb = opt_b_ii_ss;
	      return(true);
	    }
	  if (is_opt_int(arg2))
	    {
	      opc->v2.i = integer(arg2);
	      opc->v7.fb = opt_b_ii_sc;
	      return(true);
	    }
	  if (int_optimize(sc, cddr(car_x)))
	    {
	      opc->v7.fb = opt_b_ii_sf;
	      return(true);
	    }
	  return(return_false(sc, car_x, __func__, __LINE__));
	}
      if (is_symbol(arg2))
	{
	  if (int_optimize(sc, cdr(car_x)))
	    {
	      opc->v2.p = find_symbol(sc, arg2);
	      opc->v7.fb = opt_b_ii_fs;
	      return(true);
	    }
	  return(return_false(sc, car_x, __func__, __LINE__));
	}
      if ((is_opt_int(arg2)) &&
	  (int_optimize(sc, cdr(car_x))))
	{
	  opc->v2.i = integer(arg2);
	  opc->v7.fb = opt_b_ii_fc;
	  return(true);
	}
      else
	{
	  if ((int_optimize(sc, cdr(car_x))) &&
	      (int_optimize(sc, cddr(car_x))))
	    {
	      opc->v7.fb = opt_b_ii_ff;
	      return(true);
	    }
	}
    }
  return(false);
}

/* -------- b_or|and -------- */
static bool opt_and_bb(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  if (o1->v7.fb(o1))
    {
      o1 = cur_sc->opts[++cur_sc->pc];
      return(o1->v7.fb(o1));
    }
  cur_sc->pc = o->v1.i;
  return(false);
}

static bool opt_and_bb1(void *p)
{
  opt_info *o = (opt_info *)p;
  if (o->v8.fb(o))
    {
      opt_info *o1;
      cur_sc->pc += 2;
      o1 = cur_sc->opts[cur_sc->pc];
      return(o1->v7.fb(o1));
    }
  cur_sc->pc = o->v4.i;
  return(false);
}

static bool opt_and_any_b(void *p)
{
  opt_info *o = (opt_info *)p;
  int32_t i;
  for (i = 0; i < o->v1.i; i++)
    {
      opt_info *o1;
      o1 = cur_sc->opts[++cur_sc->pc];
      if (!o1->v7.fb(o1))
	{
	  cur_sc->pc = o->v2.i;
	  return(false);
	}
    }
  return(true);
}

static bool opt_or_bb(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  if (o1->v7.fb(o1)) 
    {
      cur_sc->pc = o->v1.i;
      return(true);
    }
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o1->v7.fb(o1));
}

static bool opt_or_bb1(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  if (o->v8.fb(o)) 
    {
      cur_sc->pc = o->v4.i;
      return(true);
    }
  cur_sc->pc += 2;
  o1 = cur_sc->opts[cur_sc->pc];
  return(o1->v7.fb(o1));
}

static bool opt_or_any_b(void *p)
{
  opt_info *o = (opt_info *)p;
  int32_t i;
  for (i = 0; i < o->v1.i; i++)
    {
      opt_info *o1;
      o1 = cur_sc->opts[++cur_sc->pc];
      if (o1->v7.fb(o1))
	{
	  cur_sc->pc = o->v2.i;
	  return(true);
	}
    }
  return(false);
}

static bool opt_b_or_and(s7_scheme *sc, s7_pointer car_x, int32_t len, int32_t is_and)
{
  opt_info *opc;
  opc = alloc_opo(sc, car_x);
  if (len == 3)
    {
      opt_info *o1;
      o1 = sc->opts[sc->pc];
      if ((bool_optimize_nw(sc, cdr(car_x))) &&
	  (bool_optimize_nw(sc, cddr(car_x))))
	{
	  if ((o1->v7.fb == opt_b_dd_ss) ||
	      (o1->v7.fb == opt_b_ii_ss) ||
	      (o1->v7.fb == opt_b_pp_ss))
	    {
	      opc->v4.i = sc->pc - 1;
	      opc->v8.fb = o1->v7.fb;
	      opc->v7.fb = (is_and) ? opt_and_bb1 : opt_or_bb1;
	      opc->v1.p = o1->v1.p;
	      opc->v2.p = o1->v2.p;
	      opc->v3.p = o1->v3.p;
	      return(true);
	    }
	  opc->v7.fb = (is_and) ? opt_and_bb : opt_or_bb;
	  opc->v1.i = sc->pc - 1;
	  return(true);
	}
      return(return_false(sc, car_x, __func__, __LINE__));
    }
  else
    {
      s7_pointer p;
      opc->v1.i = (len - 1);
      for (p = cdr(car_x); is_pair(p); p = cdr(p))
	if (!bool_optimize_nw(sc, p))
	  break;
      if (is_null(p))
	{
	  opc->v7.fb = (is_and) ? opt_and_any_b : opt_or_any_b;
	  opc->v2.i = sc->pc - 1;
	  return(true);
	}
    }
  return(return_false(sc, car_x, __func__, __LINE__));
}

static bool opt_b_and(s7_scheme *sc, s7_pointer car_x, int32_t len) {return(opt_b_or_and(sc, car_x, len, true));}
static bool opt_b_or(s7_scheme *sc, s7_pointer car_x, int32_t len)  {return(opt_b_or_and(sc, car_x, len, false));}


/* ---------------------------------------- cell opts ---------------------------------------- */

static s7_pointer opt_p_c(void *p) {opt_info *o = (opt_info *)p; return(o->v1.p);}
static s7_pointer opt_p_s(void *p) {opt_info *o = (opt_info *)p; return(slot_value(o->v1.p));}

static bool opt_cell_not_pair(s7_scheme *sc, s7_pointer car_x)
{
  s7_pointer p;
  opt_info *opc;
  if (!is_symbol(car_x))
    {
      opc = alloc_opo(sc, car_x);
      opc->v1.p = car_x;
      opc->v7.fp = opt_p_c;
      return(true);
    }
  p = find_symbol(sc, car_x);
  if (is_slot(p))
    {
      if (has_methods(slot_value(p)))
	return(return_false(sc, car_x, __func__, __LINE__));
      opc = alloc_opo(sc, car_x);
      opc->v1.p = p;
      opc->v7.fp = opt_p_s;
      return(true);
    }
  return(return_false(sc, car_x, __func__, __LINE__));
}

/* -------- p -------- */
#define is_opt_safe(P) ((optimize_op(P) >= OP_SAFE_C_S) && ((optimize_op(P) < OP_UNKNOWN) || (optimize_op(P) > OP_UNKNOWN_AA)))

#define cf_call(Sc, Car_x, S_func, Num) \
   (((is_optimized(Car_x)) && (is_opt_safe(Car_x))) ? c_call(Car_x) : c_function_call(c_function_chooser(S_func)(Sc, S_func, Num, Car_x, false)))

static s7_pointer opt_p_f(void *p)  {opt_info *o = (opt_info *)p; return(o->v1.p_f());}
static s7_pointer opt_p_cf(void *p) {opt_info *o = (opt_info *)p; return(o->v1.cf(cur_sc, cur_sc->nil));}

static bool p_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x)
{
  s7_p_t func;
  func = s7_p_function(s_func);
  if (func)
    {
      opc->v1.p_f = func;
      opc->v7.fp = opt_p_f;
      return(true);
    }
  if ((is_safe_procedure(s_func)) &&
      (c_function_required_args(s_func) == 0))
    {
      opc->v1.cf = cf_call(sc, car_x, s_func, 0);
      opc->v7.fp = opt_p_cf;
      return(true);
    }
  return(false);
}

/* -------- p_p -------- */
static s7_pointer opt_p_p_c(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v2.p_p_f(o->v1.p));
}

static s7_pointer opt_p_p_s(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v2.p_p_f(slot_value(o->v1.p)));
}

static s7_pointer opt_p_p_f(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v2.p_p_f(o1->v7.fp(o1)));
}

static s7_pointer opt_p_p_f1(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v2.p_p_f(o->v3.p_p_f(slot_value(o->v1.p))));
}

static bool p_p_f_combinable(s7_scheme *sc, opt_info *opc)
{
  if ((sc->pc > 1) &&
      (opc == sc->opts[sc->pc - 2]))
    {
      opt_info *o1;
      o1 = sc->opts[sc->pc - 1];
      if (o1->v7.fp == opt_p_p_s)
	{
	  opc->v3.p_p_f = o1->v2.p_p_f;
	  opc->v1.p = o1->v1.p;
	  opc->v7.fp = opt_p_p_f1;
	  sc->pc--;
	  return(true);
	}
    }
  return(false);
}

static s7_pointer opt_p_cf_f(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v2.cf(cur_sc, set_plist_1(cur_sc, o1->v7.fp(o1))));
}

static s7_pointer opt_p_cf_s(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v2.cf(cur_sc, set_plist_1(cur_sc, slot_value(o->v1.p))));
}

static bool p_p_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x)
{
  s7_p_p_t ppf;
  int32_t start;
  start = sc->pc;
  ppf = s7_p_p_function(s_func);
  if (ppf)
    {
      opc->v2.p_p_f = ppf;
      if ((ppf == symbol_to_string_p) &&
	  (is_optimized(car_x)) &&
	  (c_call(car_x) == g_symbol_to_string_uncopied))
	opc->v2.p_p_f = symbol_to_string_uncopied_p;
      
      if (is_symbol(cadr(car_x)))
	{
	  opc->v1.p = find_symbol(sc, cadr(car_x));
	  if ((!is_slot(opc->v1.p)) || 
	      (has_methods(slot_value(opc->v1.p))))
	    return(return_false(sc, car_x, __func__, __LINE__));
	  opc->v7.fp = opt_p_p_s;
	  return(true);
	}
      if (!is_pair(cadr(car_x)))
	{
	  opc->v1.p = cadr(car_x);
	  opc->v7.fp = opt_p_p_c;
	  return(true);
	}
      if (cell_optimize(sc, cdr(car_x)))
	{
	  if (!p_p_f_combinable(sc, opc))
	    opc->v7.fp = opt_p_p_f;
	  return(true);
	}
    }
  pc_fallback(sc, start);
  if ((is_safe_procedure(s_func)) &&
      (c_function_required_args(s_func) <= 1) &&
      (c_function_all_args(s_func) >= 1))
    {
      opc->v2.cf = cf_call(sc, car_x, s_func, 1);
      if (is_symbol(cadr(car_x)))
	{
	  opc->v1.p = find_symbol(sc, cadr(car_x));
	  if ((is_slot(opc->v1.p)) &&
	      (!has_methods(slot_value(opc->v1.p))))
	    {
	      opc->v7.fp = opt_p_cf_s;
	      return(true);
	    }
	}
      else
	{
	  if (cell_optimize(sc, cdr(car_x)))
	    {
	      opc->v7.fp = opt_p_cf_f;
	      return(true);
	    }
	}
    }
  return(false);
}

/* -------- p_ii -------- */
static s7_pointer opt_p_ii_ss(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.p_ii_f(integer(slot_value(o->v1.p)), integer(slot_value(o->v2.p))));
}

static s7_pointer opt_p_ii_fs(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.p_ii_f(o1->v7.fi(o1), integer(slot_value(o->v2.p))));
}

static bool p_ii_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x, int32_t pstart)
{
  s7_p_ii_t ifunc;
  ifunc = s7_p_ii_function(s_func);
  if (ifunc)
    {
      s7_pointer arg1, arg2;
      arg1 = cadr(car_x);
      arg2 = caddr(car_x);
      if ((is_symbol(arg1)) &&
	  (is_symbol(arg2)))
	{
	  opc->v1.p = find_symbol(sc, arg1);
	  opc->v2.p = find_symbol(sc, arg2);
	  if ((is_slot(opc->v1.p)) &&
	      (is_opt_int(slot_value(opc->v1.p))) &&
	      (is_slot(opc->v2.p)) &&
	      (is_opt_int(slot_value(opc->v2.p))))
	    {
	      opc->v3.p_ii_f = ifunc;
	      opc->v7.fp = opt_p_ii_ss;
	      return(true);
	    }
	}
      if ((int_optimize(sc, cdr(car_x))) &&
	  (is_symbol(arg2)))
	{
	  opc->v2.p = find_symbol(sc, arg2);
	  if ((is_slot(opc->v2.p)) &&
	      (is_opt_int(slot_value(opc->v2.p))))
	    {
	      opc->v3.p_ii_f = ifunc;
	      opc->v7.fp = opt_p_ii_fs;
	      return(true);
	    }
	}
    }
  pc_fallback(sc, pstart);
  return(false);
}

/* -------- p_dd -------- */
static s7_pointer opt_p_dd_sc(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.p_dd_f(real_to_double(cur_sc, slot_value(o->v1.p), "p_dd"), o->v2.x));
}

static s7_pointer opt_p_dd_cs(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.p_dd_f(o->v2.x, real_to_double(cur_sc, slot_value(o->v1.p), "p_dd")));
}

static bool p_dd_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x, int32_t pstart)
{
  s7_p_dd_t ifunc;
  ifunc = s7_p_dd_function(s_func);
  if (ifunc)
    {
      s7_pointer arg1, arg2;
      arg1 = cadr(car_x);
      arg2 = caddr(car_x);
      if ((is_symbol(arg1)) &&
	  (is_t_real(arg2)))
	{
	  opc->v2.x = real(arg2);
	  opc->v1.p = find_symbol(sc, arg1);
	  if ((is_slot(opc->v1.p)) &&
	      (is_opt_real(slot_value(opc->v1.p))))
	    {
	      opc->v3.p_dd_f = ifunc;
	      opc->v7.fp = opt_p_dd_sc;
	      return(true);
	    }
	}
      if ((is_symbol(arg2)) &&
	  (is_t_real(arg1)))
	{
	  opc->v2.x = real(arg1);
	  opc->v1.p = find_symbol(sc, arg2);
	  if ((is_slot(opc->v1.p)) &&
	      (is_opt_real(slot_value(opc->v1.p))))
	    {
	      opc->v3.p_dd_f = ifunc;
	      opc->v7.fp = opt_p_dd_cs;
	      return(true);
	    }
	}
    }
  pc_fallback(sc, pstart);
  return(false);
}

/* -------- p_pi -------- */
static s7_pointer opt_p_pi_ss(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.p_pi_f(slot_value(o->v1.p), integer(slot_value(o->v2.p))));
}

static s7_pointer opt_p_pi_sc(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.p_pi_f(slot_value(o->v1.p), o->v2.i));
}

static s7_pointer opt_p_pi_sf(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.p_pi_f(slot_value(o->v1.p), o1->v7.fi(o1)));
}

static s7_pointer opt_p_pi_fc(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.p_pi_f(o1->v7.fp(o1), o->v2.i));
}

static bool p_pi_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x)
{
  s7_p_pi_t func;
  func = s7_p_pi_function(s_func);
  if (func)
    {
      s7_pointer obj = NULL, slot, sig, checker = NULL;

      sig = s7_procedure_signature(sc, s_func);
      if ((is_pair(sig)) &&
	  (is_pair(cdr(sig))) &&
	  (is_symbol(cadr(sig))))
	checker = cadr(sig);
      
      /* here we know cadr is a symbol */
      slot = find_symbol(sc, cadr(car_x));
      if ((!is_slot(slot)) || 
	  (has_methods(slot_value(slot))))
	return(return_false(sc, car_x, __func__, __LINE__));
      if ((s7_is_vector(slot_value(slot))) &&
	  (vector_rank(slot_value(slot)) > 1))
	return(return_false(sc, car_x, __func__, __LINE__));

      opc->v3.p_pi_f = func;
      opc->v1.p = slot;

      if ((s7_p_pi_direct_function(s_func)) &&
	  (checker))
	{
	  obj = slot_value(opc->v1.p);
	  if ((is_string(obj)) ||
	      (is_pair(obj)) ||
	      (s7_is_vector(obj)))
	    {
	      if (((is_string(obj)) && (checker == sc->is_string_symbol)) ||
		  ((s7_is_vector(obj)) && (checker == sc->is_vector_symbol)) ||
		  ((is_pair(obj)) && (checker == sc->is_pair_symbol)))
		opc->v3.p_pi_f = s7_p_pi_direct_function(s_func);
	    }
	}
      if (is_symbol(caddr(car_x)))
	{
	  s7_pointer slot;
	  slot = find_symbol(sc, caddr(car_x));
	  if ((is_slot(slot)) &&
	      (is_opt_int(slot_value(slot))))
	    {
	      opc->v7.fp = opt_p_pi_ss;
	      opc->v2.p = slot;
	      if ((obj) &&
		  (is_step_end(slot)))
		switch (type(obj))
		  {
		  case T_VECTOR:
		    if (denominator(slot_value(slot)) <= vector_length(obj))
		      opc->v3.p_pi_f = vector_ref_unchecked; 
		    break;
		  case T_INT_VECTOR:
		    if (denominator(slot_value(slot)) <= vector_length(obj))
		      opc->v3.p_pi_f = int_vector_ref_unchecked_p; 
		    break;
		  case T_FLOAT_VECTOR:
		    if (denominator(slot_value(slot)) <= vector_length(obj))
		      opc->v3.p_pi_f = float_vector_ref_unchecked_p; 
		    break;
		  case T_STRING: 
		    if (denominator(slot_value(slot)) <= string_length(obj))
		      opc->v3.p_pi_f = string_ref_unchecked; 
		    break;
		  }
	      return(true);
	    }
	}
      else
	{
	  if (is_integer(caddr(car_x)))
	    {
	      opc->v2.i = integer(caddr(car_x));
	      opc->v7.fp = opt_p_pi_sc;
	      return(true);
	    }
	  if (int_optimize(sc, cddr(car_x)))
	    {
	      opc->v7.fp = opt_p_pi_sf;
	      return(true);
	    }
	}
    }
  return(false);
}

static s7_pointer opt_p_pi_fco(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.p_pi_f(o->v4.p_p_f(slot_value(o->v1.p)), o->v2.i));
}

static bool p_pi_fc_combinable(s7_scheme *sc, opt_info *opc)
{
  if ((sc->pc > 1) &&
      (opc == sc->opts[sc->pc - 2]))
    {
      opt_info *o1;
      o1 = sc->opts[sc->pc - 1];
      if (o1->v7.fp == opt_p_p_s)
	{
	  opc->v4.p_p_f = o1->v2.p_p_f;
	  opc->v1.p = o1->v1.p;	
	  opc->v7.fp = opt_p_pi_fco;
	  sc->pc--;
	  return(true);
	}
    }
  return(false);
}

/* -------- p_pp -------- */
static s7_pointer opt_p_pp_ss(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.p_pp_f(slot_value(o->v1.p), slot_value(o->v2.p)));
}

static s7_pointer opt_p_pp_sc(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.p_pp_f(slot_value(o->v1.p), o->v2.p));
}

static s7_pointer opt_p_pp_cs(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.p_pp_f(o->v2.p, slot_value(o->v1.p)));
}

static s7_pointer opt_p_pp_sf(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.p_pp_f(slot_value(o->v1.p), o1->v7.fp(o1)));
}

static s7_pointer opt_p_pp_fs(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.p_pp_f(o1->v7.fp(o1), slot_value(o->v1.p)));
}

static s7_pointer opt_p_pp_fc(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.p_pp_f(o1->v7.fp(o1), o->v2.p));
}

static s7_pointer opt_p_pp_cc(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.p_pp_f(o->v1.p, o->v2.p));
}

static s7_pointer opt_p_pp_ff(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  s7_pointer p1;
  o1 = cur_sc->opts[++cur_sc->pc];
  p1 = o1->v7.fp(o1);
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.p_pp_f(p1, o1->v7.fp(o1)));
}

static bool p_pp_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x, int32_t pstart)
{
  s7_p_pp_t func;
  func = s7_p_pp_function(s_func);
  if (func)
    {
      s7_pointer slot, sig, checker = NULL;
      
      sig = s7_procedure_signature(sc, s_func);
      if ((is_pair(sig)) &&
	  (is_pair(cdr(sig))) &&
	  (is_symbol(cadr(sig))))
	checker = cadr(sig);

      if (is_symbol(cadr(car_x)))
	{
	  slot = find_symbol(sc, cadr(car_x));
	  if ((!is_slot(slot)) || 
	      (has_methods(slot_value(slot))))
	    {
	      pc_fallback(sc, pstart);
	      return(return_false(sc, car_x, __func__, __LINE__));
	    }
	  if ((s7_is_vector(slot_value(slot))) &&
	      (vector_rank(slot_value(slot)) > 1))
	    {
	      pc_fallback(sc, pstart);
	      return(return_false(sc, car_x, __func__, __LINE__));
	    }
	  opc->v1.p = slot;
	}
      
      opc->v3.p_pp_f = func;
      if (is_symbol(cadr(car_x)))
	{
	  if ((s7_p_pp_direct_function(s_func)) &&
	      (checker))
	    {
	      checker = s7_symbol_value(sc, checker);
	      if (s7_apply_function(sc, checker, set_plist_1(sc, slot_value(opc->v1.p))) == sc->T)
		opc->v3.p_pp_f = s7_p_pp_direct_function(s_func);
	    }
	  if (is_symbol(caddr(car_x)))
	    {
	      opc->v2.p = find_symbol(sc, caddr(car_x));
	      if ((is_slot(opc->v2.p)) &&
		  (!has_methods(slot_value(opc->v2.p))))
		{
		  opc->v7.fp = opt_p_pp_ss;
		  return(true);
		}
	      pc_fallback(sc, pstart);
	      return(return_false(sc, car_x, __func__, __LINE__));			    
	    }
	  if ((!is_pair(caddr(car_x))) ||
	      (is_proper_quote(sc, caddr(car_x))))
	    {
	      opc->v2.p = (!is_pair(caddr(car_x))) ? caddr(car_x) : cadr(caddr(car_x));
	      opc->v7.fp = opt_p_pp_sc;
	      return(true);
	    }
	  if (cell_optimize(sc, cddr(car_x)))
	    {
	      opc->v7.fp = opt_p_pp_sf;
	      return(true);
	    }
	}
      else
	{
	  if ((!is_pair(cadr(car_x))) ||
	      (is_proper_quote(sc, cadr(car_x))))
	    {
	      opc->v1.p = (!is_pair(cadr(car_x))) ? cadr(car_x) : cadadr(car_x);
	      if ((!is_symbol(caddr(car_x))) &&
		  ((!is_pair(caddr(car_x))) ||
		   (is_proper_quote(sc, caddr(car_x)))))
		{
		  opc->v2.p = (!is_pair(caddr(car_x))) ? caddr(car_x) : cadr(caddr(car_x));
		  opc->v7.fp = opt_p_pp_cc;
		  return(true);
		}
	      if (is_symbol(caddr(car_x)))
		{
		  opc->v2.p = opc->v1.p;
		  opc->v1.p = find_symbol(sc, caddr(car_x));
		  if ((is_slot(opc->v1.p)) &&
		      (!has_methods(slot_value(opc->v1.p))))
		    {
		      opc->v7.fp = opt_p_pp_cs;
		      return(true);
		    }
		  pc_fallback(sc, pstart);
		  return(return_false(sc, car_x, __func__, __LINE__));			    
		}
	    }
	  if (cell_optimize(sc, cdr(car_x)))
	    {
	      if (is_symbol(caddr(car_x)))
		{
		  opc->v1.p = find_symbol(sc, caddr(car_x));
		  if ((is_slot(opc->v1.p)) &&
		      (!has_methods(slot_value(opc->v1.p))))
		    {
		      opc->v7.fp = opt_p_pp_fs;
		      return(true);
		    }
		  pc_fallback(sc, pstart);
		  return(return_false(sc, car_x, __func__, __LINE__));			    
		}
	      if ((!is_pair(caddr(car_x))) ||
		  (is_proper_quote(sc, caddr(car_x))))
		{
		  if (is_t_integer(caddr(car_x)))
		    {
		      s7_p_pi_t ifunc;
		      ifunc = s7_p_pi_function(s_func);
		      if (ifunc)
			{
			  opc->v2.i = integer(caddr(car_x));
			  opc->v3.p_pi_f = ifunc;
			  if (!p_pi_fc_combinable(sc, opc))
			    opc->v7.fp = opt_p_pi_fc;
			  return(true);
			}
		    }
		  opc->v2.p = (!is_pair(caddr(car_x))) ? caddr(car_x) : cadr(caddr(car_x));
		  opc->v7.fp = opt_p_pp_fc;
		  return(true);
		}
	      if (cell_optimize(sc, cddr(car_x)))
		{
		  opc->v7.fp = opt_p_pp_ff;
		  return(true);
		}
	    }
	}
    }
  pc_fallback(sc, pstart);
  return(false);
}

/* -------- p_cf_pp -------- */
static s7_pointer opt_p_cf_ff(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  int32_t tx;
  s7_pointer po2;
  o1 = cur_sc->opts[++cur_sc->pc];
  tx = next_tx(cur_sc);
  cur_sc->t_temps[tx] = o1->v7.fp(o1);
  o1 = cur_sc->opts[++cur_sc->pc];
  po2 = o1->v7.fp(o1);
  return(o->v3.cf(cur_sc, set_plist_2(cur_sc, cur_sc->t_temps[tx], po2)));
}

static s7_pointer opt_p_cf_fs(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  s7_pointer po1;
  o1 = cur_sc->opts[++cur_sc->pc];
  po1 = o1->v7.fp(o1);
  return(o->v3.cf(cur_sc, set_plist_2(cur_sc, po1, slot_value(o->v1.p))));
}

static s7_pointer opt_p_cf_sf(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  s7_pointer po1;
  o1 = cur_sc->opts[++cur_sc->pc];
  po1 = o1->v7.fp(o1);
  return(o->v3.cf(cur_sc, set_plist_2(cur_sc, slot_value(o->v1.p), po1)));
}

static s7_pointer opt_p_cf_sc(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.cf(cur_sc, set_plist_2(cur_sc, slot_value(o->v1.p), o->v2.p)));
}

static s7_pointer opt_p_cf_ss(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.cf(cur_sc, set_plist_2(cur_sc, slot_value(o->v1.p), slot_value(o->v2.p))));
}

static bool p_cf_pp_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x, int32_t pstart)
{
  if ((is_safe_procedure(s_func)) &&
      (c_function_required_args(s_func) <= 2) &&
      (c_function_all_args(s_func) >= 2))
    {
      /* if optimized, we want to use the current c_call (to take advantage of fixups like substring_temp),
       *   but those same fixups are incorrect for this context if op_safe_c_c related.
       */
      opc->v3.cf = cf_call(sc, car_x, s_func, 2);
      if (is_symbol(cadr(car_x)))
	{
	  opc->v1.p = find_symbol(sc, cadr(car_x));
	  if ((is_slot(opc->v1.p)) &&
	      (!has_methods(slot_value(opc->v1.p))))
	    {
	      if (is_symbol(caddr(car_x)))
		{
		  opc->v2.p = find_symbol(sc, caddr(car_x));
		  if ((is_slot(opc->v2.p)) &&
		      (!has_methods(slot_value(opc->v2.p))))
		    {
		      opc->v7.fp = opt_p_cf_ss;
		      return(true);
		    }
		  else 
		    {
		      pc_fallback(sc, pstart);
		      return(return_false(sc, car_x, __func__, __LINE__));
		    }
		}
	      if (!is_pair(caddr(car_x)))
		{
		  opc->v2.p = caddr(car_x);
		  opc->v7.fp = opt_p_cf_sc;
		  return(true);
		}
	      if (cell_optimize(sc, cddr(car_x)))
		{
		  opc->v7.fp = opt_p_cf_sf;
		  return(true);
		}
	    }
	  else 
	    {
	      pc_fallback(sc, pstart);
	      return(return_false(sc, car_x, __func__, __LINE__));
	    }
	}
      if (cell_optimize(sc, cdr(car_x)))
	{
	  if (is_symbol(caddr(car_x)))
	    {
	      opc->v1.p = find_symbol(sc, caddr(car_x));
	      if ((is_slot(opc->v1.p)) &&
		  (!has_methods(slot_value(opc->v1.p))))
		{
		  opc->v7.fp = opt_p_cf_fs;
		  return(true);
		}
	      else 
		{
		  pc_fallback(sc, pstart);
		  return(return_false(sc, car_x, __func__, __LINE__));
		}
	    }
	  if (cell_optimize(sc, cddr(car_x)))
	    {
	      opc->v7.fp = opt_p_cf_ff;
	      return(true);
	    }
	}
    }
  pc_fallback(sc, pstart);
  return(false);
}


/* -------- p_pip --------*/

static s7_pointer opt_p_pip_ssf(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.p_pip_f(slot_value(o->v1.p), integer(slot_value(o->v2.p)), o1->v7.fp(o1)));
}

static s7_pointer opt_p_pip_sss(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v4.p_pip_f(slot_value(o->v1.p), integer(slot_value(o->v2.p)), slot_value(o->v3.p)));
}

static s7_pointer opt_p_pip_ssc(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.p_pip_f(slot_value(o->v1.p), integer(slot_value(o->v2.p)), o->v4.p));
}

static s7_pointer opt_p_pip_c(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.p_pip_f(slot_value(o->v1.p), integer(slot_value(o->v2.p)), o->v5.p_p_f(o->v4.p)));
}

static s7_pointer opt_p_pip_sff(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1, *o2;
  s7_int i1;
  o1 = cur_sc->opts[++cur_sc->pc];
  i1 = o1->v7.fi(o1);
  o2 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.p_pip_f(slot_value(o->v1.p), i1, o2->v7.fp(o2)));
}

static s7_pointer opt_p_pip_sso(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v5.p_pip_f(slot_value(o->v1.p), 
			integer(slot_value(o->v2.p)), 
			o->v6.p_pi_f(slot_value(o->v3.p), 
				       integer(slot_value(o->v4.p)))));
}

static s7_pointer opt_p_pip_ssf1(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[cur_sc->pc += 2];
  return(o->v3.p_pip_f(slot_value(o->v1.p), integer(slot_value(o->v2.p)), o->v4.p_p_f(o1->v7.fp(o1))));
}

#if 0
p_cf_s: (o->v2.cf(cur_sc, set_plist_1(cur_sc, slot_value(o->v1.p))))
p_cf_ss: (o->v3.cf(cur_sc, set_plist_2(cur_sc, slot_value(o->v1.p), slot_value(o->v2.p)))
p_pi_sf: (o->v3.p_pi_f(slot_value(o->v1.p), o1->v7.fi(o1)))
p_pi_ss: (o->v3.p_pi_f(slot_value(o->v1.p), integer(slot_value(o->v2.p))))
i_to_p: s7_make_integer(cur_sc, o->v8.fi(o))
i_i_s: (o->v2.i_i_f(integer(slot_value(o->v1.p)))
p_pp_fc: (o->v3.p_pp_f(o1->v7.fp(o1), o->v2.p)

pip_ssf:  
return(o->v3.p_pip_f(slot_value(o->v1.p), integer(slot_value(o->v2.p)), o1->v7.fp(o1)));

      /* p_pi_sf -> i_i_s (form)
       * p_cf_s, p_cf_ss (hash)
       * p_p_f -> i_to_p (ref), p_pp_fc -> p_pi_ss (ref)
       * i_to_p (map), [p_p_f -> i_to_p (map)]
       * [p_p_c], [p_p_f -> p_p_c (sort)]
       */
#endif

static bool p_pip_ssf_combinable(s7_scheme *sc, opt_info *opc, int32_t start)
{
  opt_info *o1;
  if ((sc->pc > 1) &&
      (opc == sc->opts[sc->pc - 2]))
    {
      o1 = sc->opts[sc->pc - 1];
      if (o1->v7.fp == opt_p_pi_ss)
	{
	  opc->v5.p_pip_f = opc->v3.p_pip_f;
	  opc->v6.p_pi_f = o1->v3.p_pi_f;
	  opc->v3.p = o1->v1.p;
	  opc->v4.p = o1->v2.p;
	  opc->v7.fp = opt_p_pip_sso;
	  sc->pc--;
	  return(true);
	}
      if (o1->v7.fp == opt_p_p_c)
	{
	  opc->v5.p_p_f = o1->v2.p_p_f;
	  opc->v4.p = o1->v1.p;
	  sc->pc--;
	  opc->v7.fp = opt_p_pip_c;
	  return(true);
	}
    }

  o1 = sc->opts[start];
  if (o1->v7.fp == opt_p_p_f)
    {
      opc->v4.p_p_f = o1->v2.p_p_f;
      opc->v7.fp = opt_p_pip_ssf1;
      return(true);
    }
  return(false);
}

static bool p_pip_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x)
{
  s7_p_pip_t func;
  func = s7_p_pip_function(s_func);
  if (func)
    {
      s7_pointer obj = NULL, slot, sig, checker = NULL;

      sig = s7_procedure_signature(sc, s_func);
      if ((is_pair(sig)) &&
	  (is_pair(cdr(sig))) &&
	  (is_symbol(cadr(sig))))
	checker = cadr(sig);
      
      /* here we know cadr is a symbol */
      slot = find_symbol(sc, cadr(car_x));
      if ((!is_slot(slot)) || 
	  (has_methods(slot_value(slot))))
	return(return_false(sc, car_x, __func__, __LINE__));
      if ((s7_is_vector(slot_value(slot))) &&
	  (vector_rank(slot_value(slot)) > 1))
	return(return_false(sc, car_x, __func__, __LINE__));
      opc->v1.p = slot;
      
      opc->v3.p_pip_f = func;
      if ((s7_p_pip_direct_function(s_func)) &&
	  (checker))
	{
	  obj = slot_value(opc->v1.p);
	  if (((is_string(obj)) && (checker == sc->is_string_symbol)) ||
	      ((s7_is_vector(obj)) && (checker == sc->is_vector_symbol)) ||
	      ((is_pair(obj)) && (checker == sc->is_pair_symbol)))
	    opc->v3.p_pip_f = s7_p_pip_direct_function(s_func);
	}
      
      if (is_symbol(caddr(car_x)))
	{
	  s7_pointer slot;
	  int32_t start;
	  start = sc->pc;
	  slot = find_symbol(sc, caddr(car_x));
	  if ((is_slot(slot)) &&
	      (is_opt_int(slot_value(slot))))
	    {
	      opc->v2.p = slot;
	      /* fprintf(stderr, "%d slot %s\n", __LINE__, DISPLAY(slot)); */

	      if ((obj) &&
		  (is_step_end(slot)))
		switch (type(obj))
		  {
		  case T_VECTOR:
		    if (denominator(slot_value(slot)) <= vector_length(obj))
		      opc->v3.p_pip_f = vector_set_unchecked; 
		    break;
		  case T_INT_VECTOR:
		    if (denominator(slot_value(slot)) <= vector_length(obj))
		      opc->v3.p_pip_f = int_vector_set_unchecked_p; 
		    break;
		  case T_FLOAT_VECTOR:
		    if (denominator(slot_value(slot)) <= vector_length(obj))
		      opc->v3.p_pip_f = float_vector_set_unchecked_p; 
		    break;
		  case T_STRING: 
		    if (denominator(slot_value(slot)) <= string_length(obj))
		      opc->v3.p_pip_f = string_set_unchecked; 
		    break;
		  }
	      
	      if (is_symbol(cadddr(car_x)))
		{
		  s7_pointer val_slot;
		  val_slot = find_symbol(sc, cadddr(car_x));
		  if ((is_slot(val_slot)) &&
		      (!has_methods(val_slot)))
		    {
		      opc->v4.p_pip_f = opc->v3.p_pip_f;
		      opc->v3.p = val_slot;
		      opc->v7.fp = opt_p_pip_sss;
		      return(true);
		    }
		}
	      else
		{
		  if ((!is_pair(cadddr(car_x))) ||
		      (is_proper_quote(sc, cadddr(car_x))))
		    {
		      if (!is_pair(cadddr(car_x)))
			opc->v4.p = cadddr(car_x);
		      else opc->v4.p = cadr(cadddr(car_x));
		      opc->v7.fp = opt_p_pip_ssc;
		      return(true);
		    }
		}
	      if (cell_optimize(sc, cdddr(car_x)))
		{
		  if (!p_pip_ssf_combinable(sc, opc, start))
		    opc->v7.fp = opt_p_pip_ssf;
		  return(true);
		}
	    }
	}
      else
	{
	  if ((int_optimize(sc, cddr(car_x))) &&
	      (cell_optimize(sc, cdddr(car_x))))
	    {
	      opc->v7.fp = opt_p_pip_sff;
	      return(true);
	    }
	}
    }
  return(false);
}

/* -------- p_ppi -------- */
static s7_pointer opt_p_ppi_psf(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.p_ppi_f(o->v2.p, slot_value(o->v1.p), o1->v7.fi(o1)));
}

static bool p_ppi_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x)
{
  s7_p_ppi_t ifunc;
  ifunc = s7_p_ppi_function(s_func);
  if (ifunc)
    {
      int32_t start;
      start = sc->pc;
      opc->v3.p_ppi_f = ifunc;
      if ((s7_is_character(cadr(car_x))) &&
	  (is_symbol(caddr(car_x))) &&
	  (int_optimize(sc, cdddr(car_x))))
	{
	  s7_pointer slot;
	  slot = find_symbol(sc, caddr(car_x));
	  if ((is_slot(slot)) &&
	      (!has_methods(slot_value(slot))))
	    {
	      opc->v2.p = cadr(car_x);
	      opc->v1.p = slot;
	      opc->v7.fp = opt_p_ppi_psf;
	      return(true);
	    }
	}
      pc_fallback(sc, start);
    }
  return(false);
}

/* -------- p_ppp -------- */
static s7_pointer opt_p_ppp_ssf(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.p_ppp_f(slot_value(o->v1.p), slot_value(o->v2.p), o1->v7.fp(o1)));
}

static s7_pointer opt_p_ppp_sfs(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.p_ppp_f(slot_value(o->v1.p), o1->v7.fp(o1), slot_value(o->v2.p)));
}

static s7_pointer opt_p_ppp_scs(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.p_ppp_f(slot_value(o->v1.p), o->v4.p, slot_value(o->v2.p)));
}

static s7_pointer opt_p_ppp_sff(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1, *o2;
  s7_pointer po1;
  o1 = cur_sc->opts[++cur_sc->pc];
  po1 = o1->v7.fp(o1);
  o2 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.p_ppp_f(slot_value(o->v1.p), po1, o2->v7.fp(o2)));
}

static s7_pointer opt_p_ppp_sss(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v4.p_ppp_f(slot_value(o->v1.p), slot_value(o->v2.p), slot_value(o->v3.p)));
}

static s7_pointer opt_p_ppp_ssc(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v3.p_ppp_f(slot_value(o->v1.p), slot_value(o->v2.p), o->v4.p));
}

static s7_pointer opt_p_ppp_fff(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  s7_pointer po1, po2;
  o1 = cur_sc->opts[++cur_sc->pc];
  po1 = o1->v7.fp(o1);
  o1 = cur_sc->opts[++cur_sc->pc];
  po2 = o1->v7.fp(o1);
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o->v3.p_ppp_f(po1, po2, o1->v7.fp(o1)));
}

static bool p_ppp_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x)
{
  s7_p_ppp_t func;
  func = s7_p_ppp_function(s_func);
  if (func)
    {
      s7_pointer arg1, arg2, arg3;
      int32_t start;
      s7_pointer sig, checker = NULL;
	  sig = s7_procedure_signature(sc, s_func);
	  if ((is_pair(sig)) &&
	      (is_pair(cdr(sig))) &&
	      (is_symbol(cadr(sig))))
	    checker = cadr(sig);
	  start = sc->pc;

      opc->v3.p_ppp_f = func;
      arg1 = cadr(car_x);
      arg2 = caddr(car_x);
      arg3 = cadddr(car_x);
      if (is_symbol(arg1)) /* dealt with at the top -> p1 */
	{
	  s7_pointer slot;
	  slot = find_symbol(sc, arg1);
	  if ((!is_slot(slot)) || 
	      (has_methods(slot_value(slot))))
	    return(return_false(sc, car_x, __func__, __LINE__));
	  if ((s7_is_vector(slot_value(slot))) &&
	      (vector_rank(slot_value(slot)) > 1))
	    return(return_false(sc, car_x, __func__, __LINE__));
	  opc->v1.p = slot;

	  if ((s7_p_ppp_direct_function(s_func)) &&
	      (checker))
	    {
	      checker = s7_symbol_value(sc, checker);
	      if (s7_apply_function(sc, checker, set_plist_1(sc, slot_value(opc->v1.p))) == sc->T)
		opc->v3.p_ppp_f = s7_p_ppp_direct_function(s_func);
	    }
	  if (is_symbol(arg2))
	    {
	      slot = find_symbol(sc, arg2);
	      if ((is_slot(slot)) &&
		  (!has_methods(slot_value(slot))))
		{
		  opc->v2.p = slot;
		  if (is_symbol(arg3))
		    {
		      slot = find_symbol(sc, arg3);
		      if ((is_slot(slot)) &&
			  (!has_methods(slot_value(slot))))
			{
			  opc->v4.p_ppp_f = opc->v3.p_ppp_f;
			  opc->v3.p = slot;				    
			  opc->v7.fp = opt_p_ppp_sss;
			  return(true);
			}
		    }
		  else
		    {
		      if ((!is_pair(arg3)) ||
			  ((car(arg3) == sc->quote_symbol) &&
			   (is_pair(cdr(arg3))))) /* (quote) as arg3 */
			{
			  if (!is_pair(arg3))
			    opc->v4.p = arg3;
			  else opc->v4.p = cadr(arg3);
			  opc->v7.fp = opt_p_ppp_ssc;
			  return(true);
			}
		    }
		  if (cell_optimize(sc, cdddr(car_x)))
		    {
		      opc->v7.fp = opt_p_ppp_ssf;
		      return(true);
		    }
		  pc_fallback(sc, start);
		}
	    }
	  if ((is_proper_quote(sc, arg2)) &&
	      (is_symbol(arg3)))
	    {
	      s7_pointer val_slot;
	      val_slot = find_symbol(sc, arg3);
	      if ((is_slot(val_slot)) &&
		  (!has_methods(val_slot)))
		{
		  opc->v4.p = cadr(arg2);
		  opc->v2.p = val_slot;
		  opc->v7.fp = opt_p_ppp_scs;
		  return(true);
		}
	    }
	  if (cell_optimize(sc, cddr(car_x)))
	    {
	      if (is_symbol(arg3))
		{
		  s7_pointer val_slot;
		  val_slot = find_symbol(sc, arg3);
		  if ((is_slot(val_slot)) &&
		      (!has_methods(val_slot)))
		    {
		      opc->v2.p = val_slot;
		      opc->v7.fp = opt_p_ppp_sfs;
		      return(true);
		    }
		}
	      if (cell_optimize(sc, cdddr(car_x)))
		{
		  opc->v7.fp = opt_p_ppp_sff;
		  return(true);
		}
	    }
	}
      else
	{
	  if ((cell_optimize(sc, cdr(car_x))) &&
	      (cell_optimize(sc, cddr(car_x))) &&
	      (cell_optimize(sc, cdddr(car_x))))
	    {
	      opc->v7.fp = opt_p_ppp_fff;
	      return(true);
	    }
	}
      pc_fallback(sc, start);
    }
  return(false);
}

/* -------- p_cf_ppp -------- */
static s7_pointer opt_p_cf_ppp(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  int32_t tx1, tx2;
  s7_pointer po3;
  o1 = cur_sc->opts[++cur_sc->pc];
  tx1 = next_tx(cur_sc);
  cur_sc->t_temps[tx1] = o1->v7.fp(o1);
  o1 = cur_sc->opts[++cur_sc->pc];
  tx2 = next_tx(cur_sc);
  cur_sc->t_temps[tx2] = o1->v7.fp(o1);
  o1 = cur_sc->opts[++cur_sc->pc];
  po3 = o1->v7.fp(o1);
  return(o->v2.cf(cur_sc, set_plist_3(cur_sc, cur_sc->t_temps[tx1], cur_sc->t_temps[tx2], po3)));
}

static bool p_cf_ppp_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x)
{
  int32_t start;
  start = sc->pc;
  if ((is_safe_procedure(s_func)) &&
      (c_function_required_args(s_func) <= 3) &&
      (c_function_all_args(s_func) >= 3) &&
      (cell_optimize(sc, cdr(car_x))) &&
      (cell_optimize(sc, cddr(car_x))) &&
      (cell_optimize(sc, cdddr(car_x))))
    {
      opc->v2.cf = cf_call(sc, car_x, s_func, 3);
      opc->v7.fp = opt_p_cf_ppp;
      return(true);
    }
  pc_fallback(sc, start);
  return(false);
}


/* -------- p_cf_any -------- */
static s7_pointer opt_p_cf_any(void *p) 
{
  opt_info *o = (opt_info *)p;
  s7_pointer arg;
  int32_t i, tx;
  tx = next_tx(cur_sc);
  cur_sc->t_temps[tx] = safe_list_if_possible(cur_sc, o->v1.i);
  for (i = 0, arg = cur_sc->t_temps[tx]; i < o->v1.i; i++, arg = cdr(arg))
    {
      opt_info *o1;
      o1 = cur_sc->opts[++cur_sc->pc];
      car(arg) = o1->v7.fp(o1);
    }
  arg = o->v2.cf(cur_sc, cur_sc->t_temps[tx]);
  clear_list_in_use(cur_sc->t_temps[tx]);
  cur_sc->current_safe_list = 0;
  return(arg);
}

static bool p_cf_any_ok(s7_scheme *sc, opt_info *opc, s7_pointer s_func, s7_pointer car_x, int32_t len)
{
  if ((is_safe_procedure(s_func)) &&
      ((int32_t)c_function_required_args(s_func) <= (len - 1)) &&
      ((int32_t)c_function_all_args(s_func) >= (len - 1)))
    {
      s7_pointer p;
      opc->v1.i = (len - 1);
      for (p = cdr(car_x); is_pair(p); p = cdr(p))
	if (!cell_optimize(sc, p))
	  break;
      if (is_null(p))
	{
	  opc->v7.fp = opt_p_cf_any;
	  opc->v2.cf = cf_call(sc, car_x, s_func, len - 1);
	  return(true);
	}
    }
  return(false);
}

/* -------- cell_all_x -------- */

static s7_pointer opt_unwrap_cell(void *p)
{
  opt_info *o = (opt_info *)p;
  return(o->v2.all_f(cur_sc, car(o->v1.p)));
}

static bool cell_all_x_ok(s7_scheme *sc, s7_pointer expr, int32_t start)
{
  s7_function opt;
  opt = all_x_optimize(sc, expr);
  if (opt)
    {
      opt_info *opc;
      pc_fallback(sc, start + 1); /* ?? TODO: are int|float_all_x_ok because all previous checks clean up via pc_fallback? */
      opc = sc->opts[start];
      opc->v2.all_f = opt;
      opc->v7.fp = opt_unwrap_cell;
      opc->v1.p = expr;
      return(true);
    }
  return(false);
}

/* -------- p_implicit -------- */
static bool funcall_optimize(s7_scheme *sc, s7_pointer car_x, s7_pointer s_func);

static bool p_implicit(s7_scheme *sc, s7_pointer car_x, int32_t len)
{
  s7_pointer s_slot;
  s_slot = find_symbol(sc, car(car_x));
  
  if (is_slot(s_slot))
    {
      s7_pointer obj;
      obj = slot_value(s_slot);
      
      if ((is_closure(obj)) && /* need larger opts array if has_optlist removed */
	  (has_optlist(obj))) 
	{
	  /* fprintf(stderr, "%d opt: %s %s\n", __LINE__, DISPLAY(car_x), DISPLAY(obj)); */
	  return(funcall_optimize(sc, car_x, obj));
	}
      
      if (is_sequence(obj))
	{
	  opt_info *opc;
	  opc = alloc_opo(sc, car_x);
	  opc->v1.p = s_slot;
	  if (len == 2)
	    {
	      /* TODO: c-object implicit ref if direct exists (pi case) */
	      switch (type(obj))
		{
		case T_STRING:     opc->v3.p_pi_f = string_ref_p_pi_direct;	  break;
		case T_PAIR:       opc->v3.p_pi_f = list_ref_p_pi_direct;	  break;
		case T_HASH_TABLE: opc->v3.p_pp_f = hash_table_ref_p_pp_direct;   break;
		case T_LET:        opc->v3.p_pp_f = let_ref_p_pp;		  break;
		  
		case T_VECTOR:
		case T_INT_VECTOR:
		case T_FLOAT_VECTOR:
		  if (vector_rank(obj) != 1)
		    return(return_false(sc, car_x, __func__, __LINE__));
		  opc->v3.p_pi_f = vector_ref_p_pi_direct;
		  break;
		  
		case T_C_OBJECT:
		  if (c_object_direct_ref(sc, obj))
		    opc->v3.p_pi_f = c_object_pi_direct;
		  /* this doesn't currently work because in opt_dotimes, safe_stepper is false;
		   *   do_is_safe or whoever can't tell that (obj i) does not affect i!
		   */
		  
		default: 
		  return(return_false(sc, car_x, __func__, __LINE__));
		}
	      if (is_symbol(cadr(car_x)))
		{
		  s7_pointer slot;
		  slot = find_symbol(sc, cadr(car_x));
		  if (is_slot(slot))
		    {
		      opc->v2.p = slot;
		      if ((!is_hash_table(obj)) &&
			  (!is_let(obj)))
			{
			  if (is_opt_int(slot_value(slot)))
			    {
			      opc->v7.fp = opt_p_pi_ss;
			      if ((is_string(obj)) &&
				  (is_step_end(opc->v2.p)) &&
				  (denominator(slot_value(opc->v2.p)) <= string_length(obj)))
				opc->v3.p_pi_f = string_ref_unchecked;
			      else
				{
				  if ((s7_is_vector(obj)) &&
				      (is_step_end(opc->v2.p)) &&
				      (denominator(slot_value(opc->v2.p)) <= vector_length(obj)))
				    opc->v3.p_pi_f = vector_ref_unchecked;
				}
			      return(true);
			    }
			  else return(return_false(sc, car_x, __func__, __LINE__));
			}
		      opc->v7.fp = opt_p_pp_ss;
		      return(true);
		    }
		}
	      else
		{
		  if ((!is_hash_table(obj)) &&
		      (!is_let(obj)))
		    {
		      if (is_integer(cadr(car_x)))
			{
			  opc->v2.i = integer(cadr(car_x));
			  opc->v7.fp = opt_p_pi_sc;
			  return(true);
			}
		      if (int_optimize(sc, cdr(car_x)))
			{
			  opc->v7.fp = opt_p_pi_sf;
			  return(true);
			}
		      return(return_false(sc, car_x, __func__, __LINE__));
		    }
		  if (cell_optimize(sc, cdr(car_x)))
		    {
		      opc->v7.fp = opt_p_pp_sf;
		      return(true);
		    }
		}
	    } /* len==2 */
	  else
	    {
	      if (len > 2)
		{
		  s7_pointer p;
		  opc->v1.i = len;
		  for (p = car_x; is_pair(p); p = cdr(p))
		    if (!cell_optimize(sc, p))
		      break;
		  if (is_null(p))
		    {
		      opc->v7.fp = opt_p_cf_any;
		      switch (type(obj))     /* string can't happen here (no multidimensional strings) */
			{
			case T_PAIR:         opc->v2.cf = g_list_ref;           break;
			case T_HASH_TABLE:   opc->v2.cf = g_hash_table_ref;     break;
			  /* case T_LET:       opc->v2.cf = g_let_ref;            break; */ /* this doesn't handle implicit indices via g_let_ref! apply_let */
			case T_VECTOR:       opc->v2.cf = g_vector_ref;         break;
			case T_INT_VECTOR:   opc->v2.cf = g_int_vector_ref;     break;
			case T_FLOAT_VECTOR: opc->v2.cf = g_float_vector_ref;   break;
			default:             return(return_false(sc, car_x, __func__, __LINE__));
			}
		      return(true);
		    }
		}
	    }
	} /* obj is sequence */
    }
  return(return_false(sc, car_x, __func__, __LINE__));
}
	  
/* -------- p_syntax -------- */
/* -------- cell_quote -------- */
static bool opt_cell_quote(s7_scheme *sc, s7_pointer car_x)
{
  opt_info *opc;
  if (!is_null(cddr(car_x)))
    return(return_false(sc, car_x, __func__, __LINE__));
  opc = alloc_opo(sc, car_x);
  opc->v1.p = cadr(car_x);
  opc->v7.fp = opt_p_c;
  return(true);
}

/* -------- cell_set -------- */
static s7_pointer opt_arg_type(s7_scheme *sc, s7_pointer argp)
{
  s7_pointer arg;
  arg = car(argp);
  if (is_pair(arg))
    {
      if ((is_symbol(car(arg))) &&
	  ((is_global(car(arg))) ||
	   ((is_slot(global_slot(car(arg)))) &&
	    (find_symbol(sc, car(arg)) == global_slot(car(arg))))))
	{
	  s7_pointer a_func;
	  a_func = slot_value(global_slot(car(arg)));
	  if (is_c_function(a_func))
	    {
	      s7_pointer sig;
	      sig = s7_procedure_signature(sc, a_func);
	      if (is_pair(sig))
		{
		  if ((car(sig) == sc->is_integer_symbol) ||
		      ((is_pair(car(sig))) && (direct_memq(sc->is_integer_symbol, car(sig)))))
		    return(sc->is_integer_symbol);
		  if ((car(sig) == sc->is_float_symbol) ||
		      ((is_pair(car(sig))) && (direct_memq(sc->is_float_symbol, car(sig)))))
		    return(sc->is_float_symbol);
		  if ((car(sig) == sc->is_real_symbol) ||
		      (car(sig) == sc->is_number_symbol))
		    {
		      int32_t start;
		      start = sc->pc;
		      if (int_optimize(sc, argp))
			{
			  pc_fallback(sc, start);
			  return(sc->is_integer_symbol);
			}
		      if (float_optimize(sc, argp))
			{
			  pc_fallback(sc, start);
			  return(sc->is_float_symbol);
			}
		      pc_fallback(sc, start);
		    }
		  return(car(sig)); /* we want the function's return type in this context */
		}
	    }
	}
      return(sc->T);
    }
  if (is_symbol(arg))
    {
      s7_pointer slot;
      slot = find_symbol(sc, arg);
      if ((!is_slot(slot)) ||
	  (has_methods(slot_value(slot))))
	return(sc->T);
      return(s7_type_of(slot_value(slot)));
    }
  return(s7_type_of(arg));
}

static s7_pointer opt_set_p_p_f(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  s7_pointer x;
  o1 = cur_sc->opts[++cur_sc->pc];
  x = o1->v7.fp(o1);
  slot_set_value(o->v1.p, x);
  return(x);
}

static s7_pointer opt_set_p_i_s(void *p)
{
  opt_info *o = (opt_info *)p;
  s7_pointer val;
  val = slot_value(o->v2.p);
  if (is_mutable(val))
    val = make_integer(cur_sc, integer(val));
  slot_set_value(o->v1.p, val);
  return(val);
}

static s7_pointer opt_set_p_i_f(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  s7_pointer x;
  o1 = cur_sc->opts[++cur_sc->pc];
  x = make_integer(cur_sc, o1->v7.fi(o1));
  slot_set_value(o->v1.p, x);
  return(x);
}

static s7_pointer opt_set_p_d_f(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  s7_pointer x;
  o1 = cur_sc->opts[++cur_sc->pc];
  x = make_real(cur_sc, o1->v7.fd(o1));
  slot_set_value(o->v1.p, x);
  return(x);
}

static s7_pointer opt_set_p_c(void *p)
{
  opt_info *o = (opt_info *)p;
  slot_set_value(o->v1.p, o->v2.p);
  return(o->v2.p);
}

static s7_pointer opt_set_p_i_fo(void *p)
{
  opt_info *o = (opt_info *)p;
  s7_pointer x;
  x = make_integer(cur_sc, o->v4.i_ii_f(integer(slot_value(o->v2.p)), integer(slot_value(o->v3.p))));
  slot_set_value(o->v1.p, x);
  return(x);
}

static s7_pointer opt_set_p_i_fo1(void *p)
{
  opt_info *o = (opt_info *)p;
  s7_pointer x;
  s7_int i;
  i = o->v4.i_ii_f(integer(slot_value(o->v2.p)), o->v3.i);
  x = make_integer(cur_sc, i);
  slot_set_value(o->v1.p, x);
  return(x);
}

static bool set_p_i_f_combinable(s7_scheme *sc, opt_info *opc)
{
  if ((sc->pc > 1) &&
      (opc == sc->opts[sc->pc - 2]))
    {
      opt_info *o1;
      o1 = sc->opts[sc->pc - 1];
      if (o1->v7.fi == opt_i_ii_ss)
	{
	  opc->v4.i_ii_f = o1->v3.i_ii_f;
	  opc->v2.p = o1->v1.p;
	  opc->v3.p = o1->v2.p;	
	  opc->v7.fp = opt_set_p_i_fo;
	  sc->pc--;
	  return(true);
	}
      if (o1->v7.fi == opt_i_ii_sc)
	{
	  opc->v4.i_ii_f = o1->v3.i_ii_f;
	  opc->v2.p = o1->v1.p;
	  opc->v3.i = o1->v2.i;
	  opc->v7.fp = opt_set_p_i_fo1;
	  sc->pc--;
	  return(true);
	}
    }
  return(false);
}

static bool is_some_number(s7_scheme *sc, s7_pointer tp)
{
  return((tp == sc->is_integer_symbol) ||
	 (tp == sc->is_float_symbol) ||
	 (tp == sc->is_real_symbol) ||
	 (tp == sc->is_complex_symbol) ||
	 (tp == sc->is_number_symbol) ||
	 (tp == sc->is_rational_symbol));
}

static bool opt_cell_set(s7_scheme *sc, s7_pointer car_x)
{
  opt_info *opc;
  opc = alloc_opo(sc, car_x);
  if (is_symbol(cadr(car_x)))
    {
      s7_pointer settee;
      if ((is_immutable_symbol(cadr(car_x))) ||
	  (symbol_has_accessor(cadr(car_x))))
	return(return_false(sc, car_x, __func__, __LINE__));
      settee = find_symbol(sc, cadr(car_x));
      if ((is_slot(settee)) &&
	  (!is_syntax(slot_value(settee))))
	{
	  /* type changes here can confuse the rest of the optimizer */
	  s7_pointer atype, stype;
	  
	  opc->v1.p = settee;
	  stype = s7_type_of(slot_value(settee));
	  if (stype == sc->is_integer_symbol)
	    {
	      if (is_symbol(caddr(car_x)))
		{
		  s7_pointer val_slot;
		  val_slot = find_symbol(sc, caddr(car_x));
		  if ((is_slot(val_slot)) &&
		      (is_opt_int(slot_value(val_slot))))
		    {
		      opc->v2.p = val_slot;
		      opc->v7.fp = opt_set_p_i_s;
		      return(true);
		    }
		}
	      else
		{
		  if (int_optimize(sc, cddr(car_x)))
		    {
		      if (!set_p_i_f_combinable(sc, opc))
			opc->v7.fp = opt_set_p_i_f;
		      return(true);
		    }
		  return(return_false(sc, car_x, __func__, __LINE__));
		}
	    }
	  if (stype == sc->is_float_symbol)
	    {
	      if (is_t_real(caddr(car_x)))
		{
		  opc->v2.p = caddr(car_x);
		  opc->v7.fp = opt_set_p_c;
		  return(true);
		}
	      if (float_optimize(sc, cddr(car_x)))
		{
		  opc->v7.fp = opt_set_p_d_f;
		  return(true);
		}
	      return(return_false(sc, car_x, __func__, __LINE__));
	    }
	  atype = opt_arg_type(sc, cddr(car_x));
	  
	  if ((is_some_number(sc, atype)) && 
	      (!is_some_number(sc, stype))) 
	    return(return_false(sc, car_x, __func__, __LINE__));
	  
	  if (cell_optimize(sc, cddr(car_x)))
	    {
	      opc->v7.fp = opt_set_p_p_f;
	      return(true);
	    }
	}
    }
  else
    {
      if ((is_pair(cadr(car_x))) && 
	  (is_symbol(caadr(car_x))) &&
	  (is_pair(cdadr(car_x))) &&
	  (is_null(cddadr(car_x))))
	{
	  s7_pointer s_slot;
	  s_slot = find_symbol(sc, caadr(car_x));
	  if (is_slot(s_slot))
	    {
	      s7_pointer obj;
	      opc->v1.p = s_slot;
	      obj = slot_value(s_slot);
	      if ((!has_methods(obj)) &&
		  (is_sequence(obj)))
		{
		  s7_pointer index;
		  switch (type(obj))
		    {
		    case T_STRING:     
		      opc->v3.p_pip_f = string_set_p_pip_direct;     
		      break;

		    case T_VECTOR:     
		      opc->v3.p_pip_f = vector_set_p_pip_direct;     
		      break;

		    case T_FLOAT_VECTOR:     
		      if (opt_float_vector_set(sc, opc, caadr(car_x), cdadr(car_x), cddr(car_x)))
			{
			  opc->v8.fd = opc->v7.fd;
			  opc->v7.fp = d_to_p;
			  return(true);
			}
		      return(return_false(sc, car_x, __func__, __LINE__));
		      break;

		    case T_INT_VECTOR:   
		      if (opt_int_vector_set(sc, opc, caadr(car_x), cdadr(car_x), cddr(car_x)))
			{
			  opc->v8.fi = opc->v7.fi;
			  opc->v7.fp = i_to_p;
			  return(true);
			}
		      return(return_false(sc, car_x, __func__, __LINE__));
		      break;

		    case T_PAIR:       
		      opc->v3.p_pip_f = list_set_p_pip_direct;              
		      break;

		    case T_HASH_TABLE:
		      opc->v3.p_ppp_f = hash_table_set_p_ppp_direct; 
		      break;

		    case T_LET:        
		      opc->v3.p_ppp_f = let_set_p_ppp;               
		      break;

		    default:           
		      return(return_false(sc, car_x, __func__, __LINE__));
		    }
		  index = cadadr(car_x);
		  if (is_symbol(index))
		    {
		      s7_pointer slot;
		      int32_t start;
		      start = sc->pc;
		      slot = find_symbol(sc, index);
		      /* fprintf(stderr, "%d slot: %s\n", __LINE__, DISPLAY(slot)); */
		      if ((is_slot(slot)) &&
			  (!has_methods(slot_value(slot))))
			{
			  opc->v2.p = slot;
			  if ((is_opt_int(slot_value(slot))) &&
			      (is_step_end(opc->v2.p)))
			    {
			      if ((is_string(obj)) &&
				  (denominator(slot_value(opc->v2.p)) <= string_length(obj)))
				opc->v3.p_pip_f = string_set_unchecked;
			      else
				{
				  if (s7_is_vector(obj))
				    {
				      if ((s7_is_vector(obj)) &&
					  (denominator(slot_value(opc->v2.p)) <= vector_length(obj)))
					opc->v3.p_pip_f = vector_set_unchecked;
				    }
				}
			    }
			  if (is_symbol(caddr(car_x)))
			    {
			      s7_pointer val_slot;
			      val_slot = find_symbol(sc, caddr(car_x));
			      if ((is_slot(val_slot)) &&
				  (!has_methods(val_slot)))
				{
				  if ((is_string(obj)) || 
				      (s7_is_vector(obj)) ||
				      (is_pair(obj)))
				    {
				      opc->v4.p_pip_f = opc->v3.p_pip_f;
				      opc->v3.p = val_slot;
				      opc->v7.fp = opt_p_pip_sss;
				    }
				  else 
				    {
				      opc->v4.p_ppp_f = opc->v3.p_ppp_f;
				      opc->v3.p = val_slot;
				      opc->v7.fp = opt_p_ppp_sss;
				    }
				  return(true);
				}
			    }
			  else
			    {
			      if ((!is_pair(caddr(car_x))) ||
				  (is_proper_quote(sc, caddr(car_x))))
				{
				  if (!is_pair(caddr(car_x)))
				    opc->v4.p = caddr(car_x);
				  else opc->v4.p = cadr(caddr(car_x));
				  if ((is_string(obj)) || 
				      (s7_is_vector(obj)) ||
				      (is_pair(obj)))
				    opc->v7.fp = opt_p_pip_ssc;
				  else opc->v7.fp = opt_p_ppp_ssc;
				  return(true);
				}
			    }
			  if (cell_optimize(sc, cddr(car_x)))
			    {
			      if ((is_string(obj)) || 
				  (s7_is_vector(obj)) ||
				  (is_pair(obj)))
				{
				  if (!p_pip_ssf_combinable(sc, opc, start))
				    opc->v7.fp = opt_p_pip_ssf;
				}
			      else opc->v7.fp = opt_p_ppp_ssf;
			      return(true);
			    }
			}
		    }
		  else
		    {
		      if ((is_string(obj)) || 
			  (is_pair(obj)) ||
			  (s7_is_vector(obj)))
			{
			  if ((int_optimize(sc, cdadr(car_x))) &&
			      (cell_optimize(sc, cddr(car_x))))
			    {
			      opc->v7.fp = opt_p_pip_sff;
			      return(true);
			    }
			  return(return_false(sc, car_x, __func__, __LINE__));
			}
		      if ((is_proper_quote(sc, cadadr(car_x))) &&
			  (is_symbol(caddr(car_x))))
			{
			  s7_pointer val_slot;
			  val_slot = find_symbol(sc, caddr(car_x));
			  if ((is_slot(val_slot)) &&
			      (!has_methods(val_slot)))
			    {
			      opc->v4.p = cadr(cadadr(car_x));
			      opc->v2.p = val_slot;
			      opc->v7.fp = opt_p_ppp_scs;
			      return(true);
			    }
			}
		      if (cell_optimize(sc, cdadr(car_x)))
			{
			  if (is_symbol(caddr(car_x)))
			    {
			      s7_pointer val_slot;
			      val_slot = find_symbol(sc, caddr(car_x));
			      if ((is_slot(val_slot)) &&
				  (!has_methods(val_slot)))
				{
				  opc->v2.p = val_slot;
				  opc->v7.fp = opt_p_ppp_sfs;
				  return(true);
				}
			    }
			  if (cell_optimize(sc, cddr(car_x)))
			    {
			      opc->v7.fp = opt_p_ppp_sff;
			      return(true);
			    }
			}
		    }
		}
	    }
	}
    }
  return(return_false(sc, car_x, __func__, __LINE__));
}

/* -------- cell_begin -------- */
static s7_pointer opt_begin_p(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  int32_t i, len;
  len = o->v1.i - 1;
  for (i = 0; i < len; i++)
    {
      o1 = cur_sc->opts[++cur_sc->pc];
      o1->v7.fp(o1);
    }
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o1->v7.fp(o1));
}

static s7_pointer opt_begin_p_1(void *p)
{
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  o1->v7.fp(o1);
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o1->v7.fp(o1));
}

static bool opt_cell_begin(s7_scheme *sc, s7_pointer car_x, int32_t len)
{
  opt_info *opc;
  s7_pointer p;
  opc = alloc_opo(sc, car_x);
  for (p = cdr(car_x); is_pair(p); p = cdr(p))
    if (!cell_optimize(sc, p))
      return(return_false(sc, car_x, __func__, __LINE__));
  opc->v1.i = len - 1;
  opc->v7.fp = (len == 3) ? opt_begin_p_1 : opt_begin_p;
  return(true);
}
	      
/* -------- cell_when|unless -------- */
static s7_pointer opt_when_p(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  if (o1->v7.fb(o1))
    {
      int32_t i, len;
      len = o->v1.i - 1;
      for (i = 0; i < len; i++)
	{
	  o1 = cur_sc->opts[++cur_sc->pc];
	  o1->v7.fp(o1);
	}
      o1 = cur_sc->opts[++cur_sc->pc];
      return(o1->v7.fp(o1));
    }
  cur_sc->pc = o->v3.i;
  return(cur_sc->unspecified);
}

static s7_pointer opt_unless_p(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  int32_t i, len;
  o1 = cur_sc->opts[++cur_sc->pc];
  if (o1->v7.fb(o1))
    {
      cur_sc->pc = o->v3.i;
      return(cur_sc->unspecified);
    }
  len = o->v1.i - 1;
  for (i = 0; i < len; i++)
    {
      o1 = cur_sc->opts[++cur_sc->pc];
      o1->v7.fp(o1);
    }
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o1->v7.fp(o1));
}

static bool opt_cell_when(s7_scheme *sc, s7_pointer car_x, int32_t len)	      
{
  s7_pointer p;
  opt_info *opc;
  opc = alloc_opo(sc, car_x);
  if (!bool_optimize(sc, cdr(car_x)))
    return(return_false(sc, car_x, __func__, __LINE__));
  for (p = cddr(car_x); is_pair(p); p = cdr(p))
    if (!cell_optimize(sc, p))
      return(return_false(sc, car_x, __func__, __LINE__));
  opc->v1.i = len - 2;
  opc->v3.i = sc->pc - 1;
  opc->v7.fp = ((car(car_x) == sc->when_symbol) ? opt_when_p : opt_unless_p);
  return(true);
}

/* -------- cell_cond -------- */
static s7_pointer opt_cond(void *p)
{
  opt_info *o = (opt_info *)p;
  o->v2.p = cur_sc->unspecified;
  while (cur_sc->pc < o->v1.i)
    {
      opt_info *o1;
      o1 = cur_sc->opts[++cur_sc->pc];
      o1->v7.fp(o1);
    }
  return(o->v2.p);
}

static s7_pointer opt_cond_clause(void *p)
{
  /* top->p1 gets result, top->i1 is end index, o->v3.i is end of current clause, o->v1.i = body len */
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  if (o1->v7.fb(o1))
    {
      opt_info *top;
      int32_t i, len;
      top = (opt_info *)(o->v5.obj);
      len = o->v1.i - 1;
      for (i = 0; i < len; i++)
	{
	  o1 = cur_sc->opts[++cur_sc->pc];
	  o1->v7.fp(o1);
	}
      o1 = cur_sc->opts[++cur_sc->pc];
      top->v2.p = o1->v7.fp(o1);
      cur_sc->pc = top->v1.i;
      return(top->v2.p);
    }
  cur_sc->pc = o->v3.i;
  return(cur_sc->unspecified);
}

static s7_pointer opt_cond_2(void *p)
{
  /* 2 branches, results 1 expr, else */
  opt_info *o = (opt_info *)p;
  opt_info *o1, *o2;
  s7_pointer res;
  cur_sc->pc += 2;
  o2 = cur_sc->opts[cur_sc->pc];     /* this is the boolean expr of the first clause */
  if (!o2->v7.fb(o2))
    cur_sc->pc = o->v3.i;            /* jump over first clause and #t */
  o1 = cur_sc->opts[++cur_sc->pc];
  res = o1->v7.fp(o1);
  cur_sc->pc = o->v1.i;              /* end of cond index */
  return(res);
}

static bool opt_cell_cond(s7_scheme *sc, s7_pointer car_x)	      
{
  /* top->v2.p gets result, top->v1.i is end index, clause->v3.i is end of current clause, 
   *   clause->v1.i = clause result len, clause->v5.obj = top 
   */
  s7_pointer p, last_clause = NULL;
  opt_info *top;
  int32_t branches = 0, max_blen = 0, start_pc;
  
  top = alloc_opo(sc, car_x);
  start_pc = sc->pc;
  for (p = cdr(car_x); is_pair(p); p = cdr(p), branches++)
    {
      opt_info *opc;
      s7_pointer clause, cp;
      int32_t blen;
      clause = car(p);
      if ((!is_pair(clause)) ||
	  (!is_pair(cdr(clause))) || /* leave the test->result case for later */
	  (cadr(clause) == sc->feed_to_symbol))
	return(return_false(sc, clause, __func__, __LINE__));
      
      last_clause = clause;
      opc = alloc_opo(sc, car_x);
      if ((car(clause) == sc->else_symbol) ||
	  (car(clause) == sc->T))
	{
	  opt_info *opb;
	  opb = alloc_opo(sc, clause);
	  opb->v7.fb = opt_b_t;
	}
      else
	{
	  if (!bool_optimize(sc, clause))
	    return(return_false(sc, clause, __func__, __LINE__));
	}
      for (blen = 0, cp = cdr(clause); is_pair(cp); blen++, cp = cdr(cp))
	if (!cell_optimize(sc, cp))
	  return(return_false(sc, cp, __func__, __LINE__));
      opc->v1.i = blen;
      if (max_blen < blen) max_blen = blen;
      opc->v3.i = sc->pc - 1;
      opc->v5.obj = (void *)top;
      opc->v7.fp = opt_cond_clause;
    }
  top->v1.i = sc->pc - 1;
  top->v7.fp = opt_cond;
  if (branches == 2)
    {
      if ((max_blen == 1) &&
	  ((car(last_clause) == sc->else_symbol) ||
	   (car(last_clause) == sc->T)))
	{
	  opt_info *o1;
	  o1 = sc->opts[start_pc];
	  top->v3.i = o1->v3.i + 2;
	  top->v7.fp = opt_cond_2;
	}
    }
  return(true);
}

/* -------- cell_and|or -------- */
static s7_pointer opt_and_pp(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  if (o1->v7.fp(o1) == cur_sc->F)
    {
      cur_sc->pc = o->v1.i;
      return(cur_sc->F);
    }
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o1->v7.fp(o1));
}

static s7_pointer opt_and_any_p(void *p)
{
  opt_info *o = (opt_info *)p;
  int32_t i;
  s7_pointer val;
  val = cur_sc->T; /* (and) -> #t */
  for (i = 0; i < o->v1.i; i++)
    {
      opt_info *o1;
      o1 = cur_sc->opts[++cur_sc->pc];
      val = o1->v7.fp(o1);
      if (val == cur_sc->F)
	{
	  cur_sc->pc = o->v2.i;
	  return(cur_sc->F);
	}
    }
  return(val);
}

static s7_pointer opt_or_pp(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  s7_pointer val;
  o1 = cur_sc->opts[++cur_sc->pc];
  val = o1->v7.fp(o1);
  if (val != cur_sc->F)
    {
      cur_sc->pc = o->v1.i;
      return(val);
    }
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o1->v7.fp(o1));
}

static s7_pointer opt_or_any_p(void *p)
{
  opt_info *o = (opt_info *)p;
  int32_t i;
  s7_pointer val;
  val = cur_sc->F; /* (or) -> #f */
  for (i = 0; i < o->v1.i; i++)
    {
      opt_info *o1;
      o1 = cur_sc->opts[++cur_sc->pc];
      val = o1->v7.fp(o1);
      if (val != cur_sc->F)
	{
	  cur_sc->pc = o->v2.i;
	  return(val);
	}
    }
  return(cur_sc->F);
}

static bool opt_cell_and(s7_scheme *sc, s7_pointer car_x, int32_t len)
{
  opt_info *opc;
  opc = alloc_opo(sc, car_x);
  if (len == 3)
    {
      opt_info *wrapper;
      int32_t start;
      opc->v7.fp = ((car(car_x) == sc->or_symbol) ? opt_or_pp : opt_and_pp);
      wrapper = sc->opts[sc->pc];
      start = sc->pc;
      if (!cell_optimize(sc, cdr(car_x)))
	{
	  pc_fallback(sc, start);
	  if (!bool_optimize_nw(sc, cdr(car_x)))
	    return(return_false(sc, car_x, __func__, __LINE__));
	  wrapper->v8.fb = wrapper->v7.fb;
	  wrapper->v7.fp = b_to_p;
	}
      wrapper = sc->opts[sc->pc];
      start = sc->pc;
      if (!cell_optimize(sc, cddr(car_x)))
	{
	  pc_fallback(sc, start);
	  if (!bool_optimize_nw(sc, cddr(car_x)))
	    return(return_false(sc, car_x, __func__, __LINE__));
	  wrapper->v8.fb = wrapper->v7.fb;
	  wrapper->v7.fp = b_to_p;
	}
      opc->v1.i = sc->pc - 1;
      return(true);
    }
  else
    {
      if (len > 0)
	{
	  s7_pointer p;
	  opc->v1.i = (len - 1);
	  opc->v7.fp = ((car(car_x) == sc->or_symbol) ? opt_or_any_p : opt_and_any_p);
	  for (p = cdr(car_x); is_pair(p); p = cdr(p))
	    {
	      opt_info *wrapper;
	      int32_t start;
	      wrapper = sc->opts[sc->pc];
	      start = sc->pc;
	      if (!cell_optimize(sc, p))
		{
		  pc_fallback(sc, start);
		  if (!bool_optimize_nw(sc, p))
		    return(return_false(sc, car_x, __func__, __LINE__));
		  wrapper->v8.fb = wrapper->v7.fb;
		  wrapper->v7.fp = b_to_p;
		}
	    }
	  opc->v2.i = sc->pc - 1;
	  return(true);
	}
    }
  return(false);
}

/* -------- cell_if -------- */
static s7_pointer opt_if_bp(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  if (o1->v7.fb(o1))
    {
      o1 = cur_sc->opts[++cur_sc->pc];
      return(o1->v7.fp(o1));
    }
  cur_sc->pc = o->v1.i;
  return(cur_sc->unspecified);
}

static s7_pointer opt_if_bp_f(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  cur_sc->pc += 2;
  o1 = cur_sc->opts[cur_sc->pc];
  if (o->v2.b_p_f(o1->v7.fp(o1)))
    {
      o1 = cur_sc->opts[++cur_sc->pc];
      return(o1->v7.fp(o1));
    }
  cur_sc->pc = o->v1.i;
  return(cur_sc->unspecified);
}

static s7_pointer opt_if_nbp(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  if (!o1->v7.fb(o1))
    {
      o1 = cur_sc->opts[++cur_sc->pc];
      return(o1->v7.fp(o1));
    }
  cur_sc->pc = o->v1.i;
  return(cur_sc->unspecified);
}
/* also b_ii_sf (mac) */

static s7_pointer opt_if_nbp_f(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  cur_sc->pc += 2;
  o1 = cur_sc->opts[cur_sc->pc];
  if (!(o->v2.b_p_f(o1->v7.fp(o1))))
    {
      o1 = cur_sc->opts[++cur_sc->pc];
      return(o1->v7.fp(o1));
    }
  cur_sc->pc = o->v1.i;
  return(cur_sc->unspecified);
}

static s7_pointer opt_if_nbp_s(void *p)
{
  opt_info *o = (opt_info *)p;
  if (!(o->v2.b_p_f(slot_value(o->v3.p))))
    {
      opt_info *o1;
      cur_sc->pc += 2;
      o1 = cur_sc->opts[cur_sc->pc];
      return(o1->v7.fp(o1));
    }
  cur_sc->pc = o->v1.i;
  return(cur_sc->unspecified);
}

static s7_pointer opt_if_nbp_sc(void *p) /* b_pp_sc */
{
  opt_info *o = (opt_info *)p;
  if (!(o->v3.b_pp_f(slot_value(o->v2.p), o->v4.p)))
    {
      opt_info *o1;
      cur_sc->pc += 2;
      o1 = cur_sc->opts[cur_sc->pc];
      return(o1->v7.fp(o1));
    }
  cur_sc->pc = o->v1.i;
  return(cur_sc->unspecified);
}

static s7_pointer opt_if_nbp_ss(void *p) /* b_ii_ss */
{
  opt_info *o = (opt_info *)p;
  if (!(o->v3.b_ii_f(integer(slot_value(o->v2.p)), integer(slot_value(o->v4.p)))))
    {
      opt_info *o1;
      cur_sc->pc += 2;
      o1 = cur_sc->opts[cur_sc->pc];
      return(o1->v7.fp(o1));
    }
  cur_sc->pc = o->v1.i;
  return(cur_sc->unspecified);
}

static s7_pointer opt_if_nbp_fs(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  cur_sc->pc += 2;
  o1 = cur_sc->opts[cur_sc->pc];
  if (!(o->v2.b_pi_f(o1->v7.fp(o1), integer(slot_value(o->v3.p))))) /* b_pi_fs */
    {
      o1 = cur_sc->opts[++cur_sc->pc];
      return(o1->v7.fp(o1));
    }
  cur_sc->pc = o->v1.i;
  return(cur_sc->unspecified);
}

static s7_pointer opt_if_nbp_sf(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  cur_sc->pc += 2;
  o1 = cur_sc->opts[cur_sc->pc];
  if (!(o->v2.b_pp_f(slot_value(o->v3.p), o1->v7.fp(o1))))   /* b_pp_sf */
    {
      o1 = cur_sc->opts[++cur_sc->pc];
      return(o1->v7.fp(o1));
    }
  cur_sc->pc = o->v1.i;
  return(cur_sc->unspecified);
}

static s7_pointer opt_if_bpp(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o1 = cur_sc->opts[++cur_sc->pc];
  if (o1->v7.fb(o1))
    {
      s7_pointer val;
      o1 = cur_sc->opts[++cur_sc->pc];
      val = o1->v7.fp(o1);
      cur_sc->pc = o->v3.i;
      return(val);
    }
  cur_sc->pc = o->v1.i;
  o1 = cur_sc->opts[++cur_sc->pc];
  return(o1->v7.fp(o1));
}

static bool opt_cell_if(s7_scheme *sc, s7_pointer car_x, int32_t len)
{
  opt_info *opc;
  opc = alloc_opo(sc, car_x);
  if (len == 3)
    {
      opt_info *next;
      next = sc->opts[sc->pc];
      if ((is_pair(cadr(car_x))) &&
	  (caadr(car_x) == sc->not_symbol))
	{
	  if ((bool_optimize(sc, cdadr(car_x))) &&
	      (cell_optimize(sc, cddr(car_x))))
	    {
	      opc->v7.fp = opt_if_nbp;
	      opc->v1.i = sc->pc - 1;
	      if (next->v7.fb == opt_b_p_f)
		{
		  opc->v2.b_p_f = next->v2.b_p_f;
		  opc->v7.fp = opt_if_nbp_f;
		}
	      if (next->v7.fb == opt_b_p_s)
		{
		  opc->v2.b_p_f = next->v2.b_p_f;
		  opc->v3.p = next->v1.p;
		  opc->v7.fp = opt_if_nbp_s;
		}
	      if (next->v7.fb == opt_b_pi_fs)
		{
		  opc->v2.b_pi_f = next->v2.b_pi_f;
		  opc->v3.p = next->v1.p;
		  opc->v7.fp = opt_if_nbp_fs;
		}
	      if (next->v7.fb == opt_b_pp_sf)
		{
		  opc->v2.b_pp_f = next->v3.b_pp_f;
		  opc->v3.p = next->v1.p;
		  opc->v7.fp = opt_if_nbp_sf;
		}
	      if (next->v7.fb == opt_b_pp_sc)
		{
		  opc->v3.b_pp_f = next->v3.b_pp_f;
		  opc->v2.p = next->v1.p;
		  opc->v4.p = next->v2.p;
		  opc->v7.fp = opt_if_nbp_sc;
		}
	      if (next->v7.fb == opt_b_ii_ss)
		{
		  opc->v3.b_ii_f = next->v3.b_ii_f;
		  opc->v2.p = next->v1.p;
		  opc->v4.p = next->v2.p;
		  opc->v7.fp = opt_if_nbp_ss;
		}
	      return(true);
	    }
	}
      else
	{
	  if ((bool_optimize(sc, cdr(car_x))) &&
	      (cell_optimize(sc, cddr(car_x))))
	    {
	      opc->v7.fp = opt_if_bp;
	      opc->v1.i = sc->pc - 1;
	      if (next->v7.fb == opt_b_p_f)
		{
		  opc->v2.b_p_f = next->v2.b_p_f;
		  opc->v7.fp = opt_if_bp_f;
		}
	      return(true);
	    }
	}
      return(return_false(sc, car_x, __func__, __LINE__));
    }
  else
    {
      if (len == 4)
	{
	  if ((bool_optimize(sc, cdr(car_x))) &&
	      (cell_optimize(sc, cddr(car_x))))
	    {
	      opc->v7.fp = opt_if_bpp;
	      opc->v1.i = sc->pc - 1;
	      if (cell_optimize(sc, cdddr(car_x)))
		{
		  opc->v3.i = sc->pc - 1;
		  return(true);
		}
	    }
	}
    }
  return(false);
}
  
/* -------- cell_case -------- */
static bool case_memv(s7_scheme *sc, s7_pointer x, s7_pointer y)
{
  s7_pointer z;
  if (is_simple(x))
    {
      for (z = y; is_pair(z); z = cdr(z))
	if (x == car(z))
	  return(true);
      return(false);
    }
  for (z = y; is_pair(z); z = cdr(z))
    if (s7_is_eqv(x, car(z))) return(true);
  return(false);
}

static s7_pointer opt_case(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  o->v2.p = cur_sc->unspecified;
  o1 = cur_sc->opts[++cur_sc->pc];
  o->v4.p = o1->v7.fp(o1);
  while (cur_sc->pc < o->v1.i)
    {
      o1 = cur_sc->opts[++cur_sc->pc];
      o1->v7.fp(o1);
    }
  return(o->v2.p);
}

static s7_pointer opt_case_clause(void *p)
{
  /* top->v2.p gets result, top->i1 is end index, top->v4.p is selector, o->v3.i is end of current clause, o->v1.i = body len */
  opt_info *o = (opt_info *)p;
  opt_info *top;
  top = (opt_info *)(o->v5.obj);
  if ((o->v2.p == cur_sc->else_symbol) ||
      (case_memv(cur_sc, top->v4.p, o->v2.p)))
    {
      opt_info *o1;
      int32_t i, len;
      len = o->v1.i - 1;
      for (i = 0; i < len; i++)
	{
	  o1 = cur_sc->opts[++cur_sc->pc];
	  o1->v7.fp(o1);
	}
      o1 = cur_sc->opts[++cur_sc->pc];
      top->v2.p = o1->v7.fp(o1);
      cur_sc->pc = top->v1.i;
      return(top->v2.p);
    }
  cur_sc->pc = o->v3.i;
  return(cur_sc->unspecified);
}

static bool opt_cell_case(s7_scheme *sc, s7_pointer car_x)
{
  /* top->v2.p gets result, top->v1.i is end index, clause->v3.i is end of current clause, 
   * clause->v1.i = clause result len, clause->v5.obj = top 
   */
  opt_info *top;
  top = alloc_opo(sc, car_x);
  if (cell_optimize(sc, cdr(car_x))) /* selector */
    {
      s7_pointer p;
      for (p = cddr(car_x); is_pair(p); p = cdr(p))
	{
	  opt_info *opc;
	  s7_pointer clause, cp;
	  int32_t blen;
	  clause = car(p);
	  if ((!is_pair(clause)) ||
	      ((!is_pair(car(clause))) && (car(clause) != sc->else_symbol)) ||
	      (!is_pair(cdr(clause))) ||
	      (cadr(clause) == sc->feed_to_symbol))
	    return(return_false(sc, clause, __func__, __LINE__));
	  
	  opc = alloc_opo(sc, car_x);
	  if (car(clause) == sc->else_symbol)
	    {
	      if (!is_null(cdr(p)))
		return(return_false(sc, clause, __func__, __LINE__));
	      opc->v2.p = sc->else_symbol;
	    }
	  else 
	    {
	      if (!s7_is_proper_list(sc, car(clause)))
		return(return_false(sc, clause, __func__, __LINE__));
	      opc->v2.p = car(clause);
	    }
	  
	  for (blen = 0, cp = cdr(clause); is_pair(cp); blen++, cp = cdr(cp))
	    if (!cell_optimize(sc, cp))
	      return(return_false(sc, cp, __func__, __LINE__));
	  if (!is_null(cp))
	    return(return_false(sc, cp, __func__, __LINE__));
	  opc->v1.i = blen;
	  opc->v3.i = sc->pc - 1;
	  opc->v5.obj = (void *)top;
	  opc->v7.fp = opt_case_clause;
	}
      if (!is_null(p))
	return(return_false(sc, p, __func__, __LINE__));
      top->v1.i = sc->pc - 1;
      top->v7.fp = opt_case;
      return(true);
    }
  return(false);
}

/* -------- cell_let_temporarily -------- */
static s7_pointer opt_let_temporarily(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  int32_t tx, i, len;
  s7_pointer result;

  tx = next_tx(cur_sc);
  o1 = cur_sc->opts[++cur_sc->pc];

  o->v4.p = slot_value(o->v1.p);         /* save and protect old value */
  cur_sc->t_temps[tx] = o->v4.p;
  slot_set_value(o->v1.p, o1->v7.fp(o1)); /* set new value */

  len = o->v2.i - 1;
  for (i = 0; i < len; i++)
    {
      o1 = cur_sc->opts[++cur_sc->pc];
      o1->v7.fp(o1);
    }
  o1 = cur_sc->opts[++cur_sc->pc];
  result = o1->v7.fp(o1);

  slot_set_value(o->v1.p, o->v4.p);      /* restore old */
  return(result);
}

static bool opt_cell_let_temporarily(s7_scheme *sc, s7_pointer car_x, int32_t len)
{
  if ((len > 2) &&
      (is_pair(cadr(car_x))) &&
      (is_pair(caadr(car_x))) &&
      (is_null(cdadr(car_x))) &&        /* just one var for now */
      (is_symbol(caaadr(car_x))) &&
      (!is_immutable(caaadr(car_x))) &&
      (!is_syntactic(caaadr(car_x))))
    {
      s7_pointer p;
      opt_info *opc;
      opc = alloc_opo(sc, car_x);
      opc->v1.p = find_symbol(sc, caar(cadr(car_x)));
      if (!is_slot(opc->v1.p))
	return(return_false(sc, car_x, __func__, __LINE__));
      
      if (!cell_optimize(sc, cdar(cadr(car_x))))
	return(return_false(sc, car_x, __func__, __LINE__));
      for (p = cddr(car_x); is_pair(p); p = cdr(p))
	if (!cell_optimize(sc, p))
	  return(return_false(sc, car_x, __func__, __LINE__));
      
      opc->v2.i = len - 2;
      opc->v7.fp = opt_let_temporarily;
      return(true);
    }
  return(false);
}

/* -------- cell_do -------- */
static s7_pointer opt_do_any(void *p)
{
  opt_info *o = (opt_info *)p; /* o->v2.p=frame, o->v1.i=body end index, o->v3.i=body length, o->v4.i(=return length, o->v5.i=end index */
  opt_info *o1, *ostart;
  int32_t loop, i;
  s7_pointer vp, old_e, result;

  old_e = cur_sc->envir;
  cur_sc->envir = o->v2.p;
  
  /* init */
  for (vp = let_slots(o->v2.p); is_slot(vp); vp = next_slot(vp))
    {
      o1 = cur_sc->opts[++cur_sc->pc];
      slot_set_value(vp, o1->v7.fp(o1));
    }

  loop = ++cur_sc->pc;
  ostart = cur_sc->opts[loop];
  while (true)
    {
      /* end */
      /* fprintf(stderr, "frame: %s\n", s7_object_to_c_string(cur_sc, o->v2.p)); */
      if (ostart->v7.fb(ostart))
	break;

      /* body */
      for (i = 0; i < o->v3.i; i++)
	{
	  o1 = cur_sc->opts[++cur_sc->pc];
	  o1->v7.fp(o1);
	}

      /* step (let not let*) */
      for (vp = let_slots(o->v2.p); is_slot(vp); vp = next_slot(vp))
	if (has_stepper(vp))
	  {
	    o1 = cur_sc->opts[++cur_sc->pc];
	    slot_set_pending_value(vp, o1->v7.fp(o1));
	    /* fprintf(stderr, "step %s %s\n", s7_object_to_c_string(cur_sc, vp), s7_object_to_c_string(cur_sc, slot_pending_value(vp))); */
	  }
      for (vp = let_slots(o->v2.p); is_slot(vp); vp = next_slot(vp))
	if (has_stepper(vp))
	  slot_set_value(vp, slot_pending_value(vp));
      
      cur_sc->pc = loop;
    }
  cur_sc->pc = o->v1.i;

  /* result */
  result = cur_sc->T;
  for (i = 0; i < o->v4.i; i++)
    {
      o1 = cur_sc->opts[++cur_sc->pc];
      result = o1->v7.fp(o1);
    }
  cur_sc->pc = o->v5.i;
  cur_sc->envir = old_e;
  return(result);
}

static s7_pointer opt_do_no_vars(void *p)
{
  /* no vars, no return */
  opt_info *o = (opt_info *)p; /* o->v2.p=frame, o->v1.i=body end index, o->v3.i=body length, o->v4.i=return length, o->v5.i=end index */
  opt_info *ostart;
  int32_t loop, i;
  s7_pointer old_e;

  old_e = cur_sc->envir;
  cur_sc->envir = o->v2.p;
  loop = ++cur_sc->pc;
  ostart = cur_sc->opts[loop];
  while (true)
    {
      if (ostart->v7.fb(ostart))
	break;
      for (i = 0; i < o->v3.i; i++)
	{
	  opt_info *o1;
	  o1 = cur_sc->opts[++cur_sc->pc];
	  o1->v7.fp(o1);
	}
      cur_sc->pc = loop;
    }
  cur_sc->pc = o->v5.i;
  cur_sc->envir = old_e;
  return(cur_sc->T);
}

static s7_pointer opt_do_2(void *p)
{
  /* 1 var, no return */
  opt_info *o = (opt_info *)p; /* o->v2.p=frame, o->v1.i=body end index, o->v3.i=body length, o->v4.i=return length, o->v5.i=end index */
  opt_info *o1, *ostart;
  int32_t i, loop;
  s7_pointer vp, old_e;

  old_e = cur_sc->envir;
  cur_sc->envir = o->v2.p;

  vp = let_slots(o->v2.p);  
  o1 = cur_sc->opts[++cur_sc->pc];
  slot_set_value(vp, o1->v7.fp(o1));

  loop = ++cur_sc->pc;
  ostart = cur_sc->opts[loop];
  while (true)
    {
      if (ostart->v7.fb(ostart))
	break;
      for (i = 0; i < o->v3.i; i++)
	{
	  o1 = cur_sc->opts[++cur_sc->pc];
	  o1->v7.fp(o1);
	}
      o1 = cur_sc->opts[++cur_sc->pc];
      slot_set_value(vp, o1->v7.fp(o1));
      cur_sc->pc = loop;
    }
  cur_sc->pc = o->v5.i;
  cur_sc->envir = old_e;
  return(cur_sc->T);
}

static s7_pointer opt_dotimes_2(void *p)
{
  /* 1 var, no return */
  opt_info *o = (opt_info *)p; /* o->v2.p=frame, o->v1.i=body end index, o->v3.i=body length, o->v4.i=return length, o->v5.i=end index, v6.i=end if int32_t */
  opt_info *o1;
  int32_t i, loop;
  s7_int end;
  s7_pointer vp, old_e;

  old_e = cur_sc->envir;
  cur_sc->envir = o->v2.p;

  vp = slot_value(dox_slot1(o->v2.p));
  if (is_slot(dox_slot2_unchecked(o->v2.p)))
    end = integer(slot_value(dox_slot2(o->v2.p)));
  else end = o->v6.i;

  o1 = cur_sc->opts[++cur_sc->pc];
  integer(vp) = integer(o1->v7.fp(o1));

  loop = o->v4.i - 1;
  while (integer(vp) < end)
    {
      cur_sc->pc = loop;
      for (i = 0; i < o->v3.i; i++)
	{
	  o1 = cur_sc->opts[++cur_sc->pc];
	  o1->v7.fp(o1);
	}
      integer(vp)++;
    }
  cur_sc->pc = o->v5.i;
  cur_sc->envir = old_e;
  return(cur_sc->T);
}

static s7_pointer opt_do_simple(void *p)
{
  /* 1 var step by 1, 1 expr, no return */
  opt_info *o = (opt_info *)p; /* o->v2.p=frame, o->v5.i=end index */
  opt_info *o1, *ostart;
  int32_t loop;
  s7_pointer vp, old_e;

  old_e = cur_sc->envir;
  cur_sc->envir = o->v2.p;

  vp = let_slots(o->v2.p);  
  o1 = cur_sc->opts[++cur_sc->pc];
  slot_set_value(vp, o1->v7.fp(o1));

  loop = ++cur_sc->pc;
  ostart = cur_sc->opts[loop];
  while (true)
    {
      if (ostart->v7.fb(ostart))
	break;

      o1 = cur_sc->opts[++cur_sc->pc];
      o1->v7.fp(o1);

      o1 = cur_sc->opts[++cur_sc->pc];
      slot_set_value(vp, o1->v7.fp(o1));
      cur_sc->pc = loop;
    }
  cur_sc->pc = o->v5.i;
  cur_sc->envir = old_e;
  return(cur_sc->T);
}

static s7_pointer opt_do_very_simple(void *p)
{
  /* like simple but step can be direct */
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  s7_int end, loop;
  s7_pointer vp, old_e;
  s7_pointer (*f)(void *p);

  old_e = cur_sc->envir;
  cur_sc->envir = o->v2.p;

  vp = slot_value(dox_slot1(o->v2.p));
  if (is_slot(dox_slot2_unchecked(o->v2.p)))
    end = integer(slot_value(dox_slot2(o->v2.p)));
  else end = o->v3.i;
  o1 = cur_sc->opts[++cur_sc->pc];
  integer(vp) = integer(o1->v7.fp(o1));

  loop = o->v4.i;
  cur_sc->pc = loop;
  o1 = cur_sc->opts[loop]; /* the body */
  f = o1->v7.fp;
  while (integer(vp) < end)
    {
      f(o1);
      cur_sc->pc = loop;
      integer(vp)++;
    }
  cur_sc->pc = o->v5.i;
  cur_sc->envir = old_e;
  return(cur_sc->T);
}

static s7_pointer opt_do_prepackaged(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  s7_int end;
  s7_pointer vp, old_e;

  old_e = cur_sc->envir;
  cur_sc->envir = o->v2.p;

  vp = slot_value(dox_slot1(o->v2.p));
  if (is_slot(dox_slot2_unchecked(o->v2.p)))
    end = integer(slot_value(dox_slot2(o->v2.p)));
  else end = o->v3.i;
  o1 = cur_sc->opts[++cur_sc->pc];
  integer(vp) = integer(o1->v7.fp(o1));
  
  o->v6.p = vp;
  o->v1.i = end;
  o->v8.fp(o);

  cur_sc->pc = o->v5.i;
  cur_sc->envir = old_e;
  return(cur_sc->T);
}

static s7_pointer opt_do_dpnr(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  int32_t loop;
  s7_pointer vp;
  s7_int end;
  s7_double (*f)(void *p);

  end = o->v1.i;
  vp = o->v6.p;
  loop = o->v4.i;
  o1 = cur_sc->opts[loop]; /* the body */
  f = o1->v8.fd;
  while (integer(vp) < end)
    {
      cur_sc->pc = loop;
      f(o1);
      integer(vp)++;
    }
  return(NULL);
}

static s7_pointer opt_do_ipnr(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  int32_t loop;
  s7_pointer vp;
  s7_int end;
  s7_int (*f)(void *p);

  end = o->v1.i;
  vp = o->v6.p;
  loop = o->v4.i;
  o1 = cur_sc->opts[loop]; /* the body */
  f = o1->v8.fi;
  while (integer(vp) < end)
    {
      cur_sc->pc = loop;
      f(o1);
      integer(vp)++;
    }
  return(NULL);
}

static s7_pointer opt_do_ifbp(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  int32_t loop;
  s7_pointer vp;
  s7_int end;
  bool (*f)(void *p);

  end = o->v1.i;
  vp = o->v6.p;
  loop = o->v4.i + 1;
  o1 = cur_sc->opts[loop];
  f = o1->v7.fb;
  while (integer(vp) < end)
    {
      if (f(o1))
	{
	  opt_info *o2;
	  o2 = cur_sc->opts[++cur_sc->pc];
	  o2->v7.fp(o2);
	}
      cur_sc->pc = loop;
      integer(vp)++;
    }
  return(NULL);
}

static s7_pointer opt_do_setpif(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  int32_t loop;
  s7_pointer vp, val;
  s7_int end;

  end = o->v1.i;
  vp = o->v6.p;
  loop = o->v4.i;
  o1 = cur_sc->opts[loop]; 

  val = make_mutable_integer(cur_sc, integer(slot_value(o1->v1.p)));
  slot_set_value(o1->v1.p, val);
  while (integer(vp) < end)
    {
      integer(val) = o1->v4.i_ii_f(integer(slot_value(o1->v2.p)), o1->v3.i);
      integer(vp)++;
    }
  return(NULL);
}


static bool stop_is_safe(s7_scheme *sc, s7_pointer stop, s7_pointer body)
{
  /* this could be folded into the cell_optimize traveral */
  s7_pointer p;
  for (p = body; is_pair(p); p = cdr(p))
    if ((is_pair(car(p))) &&
	(caar(p) == sc->set_symbol) &&
	(is_pair(cdar(p))) &&
	(cadar(p) == stop))
      return(!s7_tree_memq(sc, stop, cdr(p)));
  return(true);
}

static bool do_is_safe(s7_scheme *sc, s7_pointer body, s7_pointer steppers, s7_pointer var_list, bool *has_set);

static bool opt_cell_do(s7_scheme *sc, s7_pointer car_x, int32_t len)
{
  /* TODO:
   *   step by 1/-1 to end opt built-in (mutable index)
   *   need to pass in the end point for unchecked use: var an int, (+ var 1), end of form (= var x) where we can find x -- set stepper and denominator
   *     this is a kludge -- find a better way!
   * and no feed-to
   * see OP_LET above -- get rid of local envir+slots! (longjmp if opts_size overflow also can't work with sc->envir change)
   *   or trap somehow?
   */
  opt_info *opc;
  s7_pointer p, end, frame = NULL, old_e;
  int32_t var_len, body_len, step_len;
  
  if (len < 3)
    return(false);
  
  if (!s7_is_proper_list(sc, cadr(car_x)))
    return(return_false(sc, car_x, __func__, __LINE__));
  var_len = safe_list_length(sc, cadr(car_x));
  step_len = var_len;
  body_len = len - 3;
  end = caddr(car_x);
  if (!is_pair(end))
    return(return_false(sc, car_x, __func__, __LINE__));
  
  old_e = sc->envir;
  opc = alloc_opo(sc, car_x);
  
  new_frame(sc, sc->envir, frame);
  clear_symbol_list(sc);
  for (p = cadr(car_x); is_pair(p); p = cdr(p))
    {
      s7_pointer var;
      var = car(p);
      if ((is_pair(var)) &&
	  (is_symbol(car(var))) &&
	  (is_pair(cdr(var))))
	{
	  s7_pointer sym;
	  sym = car(var);
	  if ((is_immutable_symbol(sym)) ||
	      (symbol_has_accessor(sym)))
	    return(return_false(sc, car_x, __func__, __LINE__));
	  
	  if (symbol_is_in_list(sc, sym))
	    eval_error(sc, "duplicate identifier in do: ~A", var);
	  add_symbol_to_list(sc, sym);

	  if (cell_optimize(sc, cdr(var))) /* opt init in outer env */
	    {
	      s7_pointer slot;
	      add_slot(frame, sym, sc->F);
	      slot = let_slots(frame);
	      if (is_pair(cddr(var)))
		{
		  set_has_stepper(slot);
		  if (!is_null(cdddr(var)))
		    return(return_false(sc, car_x, __func__, __LINE__));
		}
	      else
		{
		  step_len--;
		  if (!is_null(cddr(var)))
		    return(return_false(sc, car_x, __func__, __LINE__));
		}
	      if (is_symbol(cadr(var)))
		slot_set_value(slot, slot_value(find_symbol(sc, cadr(var))));
	      else
		{
		  if (!is_pair(cadr(var)))
		    slot_set_value(slot, cadr(var));
		  else
		    {
		      if (is_proper_quote(sc, cadr(var)))
			slot_set_value(slot, cadadr(var));
		      else
			{
			  s7_pointer sf;
			  sf = find_symbol_checked(sc, caadr(var));
			  if (is_c_function(sf))
			    {
			      s7_pointer sig;
			      sig = s7_procedure_signature(sc, sf);
			      if ((is_pair(sig)) &&
				  ((car(sig) == sc->is_integer_symbol) ||
				   ((is_pair(car(sig))) &&
				    (direct_memq(sc->is_integer_symbol, car(sig))))))
				slot_set_value(slot, small_int(0));
			    }
			}
		    }
		}
	    }
	  else return(return_false(sc, car_x, __func__, __LINE__));
	}
      else return(return_false(sc, car_x, __func__, __LINE__));
    }
  
  if (is_slot(let_slots(frame)))
    let_set_slots(frame, reverse_slots(sc, let_slots(frame)));
  sc->envir = frame;
  push_stack_no_let_no_code(sc, OP_GC_PROTECT, frame);
  
  for (p = cadr(car_x); is_pair(p); p = cdr(p))
    {
      s7_pointer var;
      var = car(p);
      if (is_pair(cddr(var)))
	{
	  s7_pointer init_type;
	  init_type = opt_arg_type(sc, cdr(var));
	  if (((init_type == sc->is_integer_symbol) ||
	       (init_type == sc->is_float_symbol)) &&
	      (opt_arg_type(sc, cddr(var)) != init_type))
	    {
#if OPT_PRINT
	      fprintf(stderr, "init_type: %s, but opt_arg: %s\n", DISPLAY(init_type), DISPLAY(opt_arg_type(sc, cddr(var))));
#endif
	      sc->stack_end -= 4; /* not pop_stack! */
	      sc->envir = old_e;
	      return(return_false(sc, car_x, __func__, __LINE__));
	    }
	}
    }
  
  if (bool_optimize_nw(sc, end))
    {
      int32_t i, body_index;
      s7_pointer p, stop;
      stop = car(end);
      /* fprintf(stderr, "check %s\n", DISPLAY_80(car_x)); */
      if ((is_pair(stop)) &&
	  ((car(stop) == sc->eq_symbol) || (car(stop) == sc->geq_symbol) || (car(stop) == sc->gt_symbol)) &&
	  (is_pair(cdr(stop))) &&
	  (is_symbol(cadr(stop))) &&
	  (is_pair(cddr(stop))) &&
	  (is_null(cdddr(stop))) &&
	  ((is_opt_int(caddr(stop))) || (is_symbol(caddr(stop)))))
	{
	  s7_pointer stop_slot;
	  if (is_symbol(caddr(stop)))
	    {
	      stop_slot = find_symbol(sc, caddr(stop));
	      if ((!is_slot(stop_slot)) ||
		  (!is_opt_int(slot_value(stop_slot))))
		stop_slot = NULL;
	    }
	  else stop_slot = sc->nil;
	  if (stop_slot) 
	    {
	      s7_int lim;
	      bool set_stop = false;
	      s7_pointer slot;
	      
	      if (is_slot(stop_slot))
		lim = integer(slot_value(stop_slot));
	      else lim = integer(caddr(stop));
	      if (car(stop) == sc->gt_symbol) lim++;
	      
	      for (p = cadr(car_x), slot = let_slots(frame); is_pair(p); p = cdr(p), slot = next_slot(slot))
		{
		  /* TODO: put off this decision until it is needed (ref/set) 
		   *    another choice: go from init downto 0: init is lim
		   */
		  if (slot_symbol(slot) == cadr(stop))
		    set_stop = true; /* don't overrule this decision below */
		  if (has_stepper(slot))
		    {
		      s7_pointer var, step;
		      var = car(p);
		      step = caddr(var);
		      /* fprintf(stderr, "    var: %s, stop: %s, step %s\n", DISPLAY(let_slots(frame)), DISPLAY(stop), DISPLAY(step)); */
		      if ((is_opt_int(slot_value(slot))) &&
			  (is_pair(step)) &&
			  (car(var) == cadr(stop)) &&
			  (car(var) == cadr(step)) &&
			  ((car(stop) != sc->eq_symbol) || /* else > protects at least the top */
			   ((caddr(step) == small_int(1)) && (car(step) == sc->add_symbol))))
			{
			  /* fprintf(stderr, "    end: %s\n", DISPLAY(let_slots(frame))); */
			  set_step_end(slot);
			  denominator(slot_value(slot)) = lim;
			}
		    }
		}
	      
	      if (!set_stop)
		{
		  s7_pointer slot;
		  slot = find_symbol(sc, cadr(stop));
		  if ((is_slot(slot)) &&
		      (is_opt_int(slot_value(slot))) &&
		      (stop_is_safe(sc, cadr(stop), cddr(car_x))))
		    {
		      /* fprintf(stderr, "%d %s: %s\n", __LINE__, DISPLAY(slot), DISPLAY(stop)); */
		      set_step_end(slot);
		      denominator(slot_value(slot)) = lim;
		    }
		}
	    }
	}
      
      body_index = sc->pc;
      for (i = 3, p = cdddr(car_x); i < len; i++, p = cdr(p))
	{
	  opt_info *start;
	  start = sc->opts[sc->pc];
	  if (!cell_optimize(sc, p))
	    break;
	  if (start->v7.fp == d_to_p)
	    start->v7.fp = d_to_p_nr;
	  else
	    {
	      if (start->v7.fp == i_to_p)
		start->v7.fp = i_to_p_nr;
	    }
	}
      
      if (!is_null(p))
	{
	  sc->stack_end -= 4;
	  sc->envir = old_e;
	  return(return_false(sc, car_x, __func__, __LINE__));
	}
      
      for (p = cadr(car_x); is_pair(p); p = cdr(p))
	{
	  s7_pointer var;
	  var = car(p);
	  if ((is_pair(cddr(var))) &&
	      (!cell_optimize(sc, cddr(var))))
	    break;
	}
      if (is_null(p))
	{
	  int32_t rtn_len = 0;
	  opc->v1.i = sc->pc - 1;
	  if (!s7_is_list(sc, cdr(end)))
	    {
	      sc->stack_end -= 4;
	      sc->envir = old_e;
	      return(return_false(sc, car_x, __func__, __LINE__));
	    }
	  for (p = cdr(end); is_pair(p); p = cdr(p), rtn_len++)
	    if (!cell_optimize(sc, p))
	      break;
	  if (!is_null(p))
	    {
	      sc->stack_end -= 4;
	      sc->envir = old_e;
	      return(return_false(sc, car_x, __func__, __LINE__));
	    }
	  opc->v2.p = frame;
	  opc->v3.i = len - 3; /* body_len */
	  opc->v4.i = rtn_len;
	  opc->v5.i = sc->pc - 1;
	  sc->envir = old_e;
	  
	  if ((var_len == 0) && (rtn_len == 0))
	    {
	      opc->v7.fp = opt_do_no_vars;
	      return(true);
	    }
	  
	  if ((var_len == 1) && (step_len == 1) && (rtn_len == 0))
	    {
	      s7_pointer ind, ind_step, end, slot, var;
	      bool has_set = false;
	      
	      opc->v7.fp = (body_len == 1) ? opt_do_simple : opt_do_2;

	      /* just a first stab at this
	       *    set|let-set? if not caddr, hash-table|vector|list-set if not cadddr: old code checks !has_set for dotimes
	       *    implicit set similar
	       *    also (+ 1 ind) and (= end ind) and >= and body_len=any but still safe_stepper(s)
	       */
	      var = caadr(car_x);
	      ind = car(var);
	      ind_step = caddr(var);
	      end = caaddr(car_x);
	      slot = let_slots(frame);

	      if ((is_pair(end)) &&                      /* (= i len|100) */
		  (car(end) == sc->eq_symbol) &&
		  (cadr(end) == ind) &&
		  ((is_symbol(caddr(end))) || (is_t_integer(caddr(end)))) &&
		  (is_null(cdddr(end))) &&
		  (is_pair(ind_step)) &&                 /* (+ i 1) */
		  (car(ind_step) == sc->add_symbol) &&
		  (cadr(ind_step) == ind) &&
		  (caddr(ind_step) == small_int(1)) &&
		  (is_null(cdddr(ind_step))) &&
		  (do_is_safe(sc, cdddr(car_x), sc->w = list_1(sc, ind), sc->nil, &has_set)))
		{
		  dox_set_slot1(frame, slot);
		  dox_set_slot2_unchecked(frame, (is_symbol(caddr(end))) ? find_symbol(sc, caddr(end)) : sc->undefined);
		  slot_set_value(slot, make_mutable_integer(sc, integer(slot_value(slot))));
		  opc->v4.i = body_index;
		  if (body_len == 1)
		    {
		      opt_info *o1;
		      opc->v7.fp = opt_do_very_simple;
		      if (is_t_integer(caddr(end)))
			opc->v3.i = integer(caddr(end));
		      
		      o1 = sc->opts[body_index];
		      /* v2, v3, v4, v5 are in use */
		      if (o1->v7.fp == d_to_p_nr)
			{
			  /* snd-test: (do ((k 0 (+ k 1))) ((= k N)) (float-vector-set! rl k (read-sample rd))) */
			  opc->v7.fp = opt_do_prepackaged;
			  opc->v8.fp = opt_do_dpnr;
			}
		      else
			{
			  if (o1->v7.fp == i_to_p_nr)
			    {
			      opc->v7.fp = opt_do_prepackaged;
			      opc->v8.fp = opt_do_ipnr;
			    }
			  else
			    {
			      if (o1->v7.fp == opt_if_bp)
				{
				  opc->v7.fp = opt_do_prepackaged;
				  opc->v8.fp = opt_do_ifbp;
				}
			      else
				{
				  if (o1->v7.fp == opt_set_p_i_fo1)
				    {
				      opc->v7.fp = opt_do_prepackaged;
				      opc->v8.fp = opt_do_setpif;
				    }
				}
			    }
			}
		    }
		  else
		    {
		      opc->v7.fp = opt_dotimes_2;
		      if (is_t_integer(caddr(end)))
			opc->v6.i = integer(caddr(end));
		    }
		}
	    }
	  else opc->v7.fp = opt_do_any;
	  return(true);
	}
    }
  /* fprintf(stderr, "bad: %s\n", DISPLAY_80(car_x)); */
  sc->stack_end -= 4; /* not pop_stack! */
  sc->envir = old_e;
  return(false);
}

static bool p_syntax(s7_scheme *sc, s7_pointer car_x, int32_t len)
{
  opcode_t op;
  s7_pointer func;
  func = slot_value(global_slot(car(car_x)));
  op = (opcode_t)syntax_opcode(func);
  switch (op)
    {
    case OP_QUOTE:
      if (is_pair(cdr(car_x)))
	return(opt_cell_quote(sc, car_x));
      break;
      
    case OP_SET:
      if (len == 3)
	return(opt_cell_set(sc, car_x));
      break;
      
    case OP_BEGIN:
      if (len > 1)
	return(opt_cell_begin(sc, car_x, len));
      break;
      
    case OP_WHEN:
    case OP_UNLESS:
      if (len > 2)
	return(opt_cell_when(sc, car_x, len));
      break;
      
    case OP_COND:
      if (len > 1)
	return(opt_cell_cond(sc, car_x));
      break;
      
    case OP_AND:
    case OP_OR:
      return(opt_cell_and(sc, car_x, len));
      
    case OP_IF:
      return(opt_cell_if(sc, car_x, len));
      
    case OP_CASE:
      if (len > 2)
	return(opt_cell_case(sc, car_x));
      break;
      
    case OP_LET_TEMPORARILY:
      return(opt_cell_let_temporarily(sc, car_x, len));
      
    case OP_DO:
      return(opt_cell_do(sc, car_x, len));
      
    default:
      break;
    }
  /* longjmp(sc->opt_exit, 1); */ /* what good could it do to back up?  But we need to make sure sc->envir isn't clobbered (in op_do??) */
  return(return_false(sc, car_x, __func__, __LINE__));
}
      
/* -------------------------------------------------------------------------------- */

static void start_opts(s7_scheme *sc)
{
  sc->pc = 0;
}

static void pc_fallback(s7_scheme *sc, int32_t new_pc)
{
  sc->pc = new_pc;
}

static void free_optlist(s7_scheme *sc, s7_pointer p)
{
  if (optlist_addr(p) >= 0)
    {
      stofl[++stofl_loc] = optlist_addr(p);
      optlist_addr(p) = -1;
    }
}

static s7_pointer make_optlist(s7_scheme *sc)
{
  s7_pointer x;
  int32_t loc;
  new_cell(sc, x, T_OPTLIST);
  add_optlist(sc, x);
  if (stofl_loc < 0)
    {
      if (!stofl)
	{
	  int32_t i;
	  stored_optlists_size = INITIAL_STORED_OPTLISTS_SIZE;
	  stored_optlists = (s7_pointer *)malloc(INITIAL_STORED_OPTLISTS_SIZE * sizeof(s7_pointer));
	  stofl = (int32_t *)malloc(INITIAL_STORED_OPTLISTS_SIZE * sizeof(int32_t));
	  stofl_loc = INITIAL_STORED_OPTLISTS_SIZE - 1;
	  for (i = 0; i < INITIAL_STORED_OPTLISTS_SIZE; i++)
	    stofl[i] = i;
	}
      else
	{
	  uint32_t k, size, new_size;
	  size = stored_optlists_size;
	  new_size = 2 * size;
	  stored_optlists = (s7_pointer *)realloc(stored_optlists, new_size * sizeof(s7_pointer));
	  stored_optlists_size = new_size;
	  stofl = (int32_t *)realloc(stofl, new_size * sizeof(int32_t));
	  for (k = size; k < new_size; k++)
	    stofl[++stofl_loc] = k;
	}
    }
  loc = stofl[stofl_loc--];
  optlist_addr(x) = loc;
  stored_optlists[loc] = x;
  return(x);
}


static s7_pointer opt_call(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  s7_pointer fp, result, old_e;
  int32_t i;
  s7_pointer x, env;
  uint64_t id;

  id = ++cur_sc->let_number;
  env = o->v2.p;
  let_id(env) = id;
  x = let_slots(env);

  /* arguments */
  if (o->v1.i > 0)
    {
      s7_pointer sym;
      if (o->v1.i == 1)
	{
	  o1 = cur_sc->opts[++cur_sc->pc];
	  slot_set_value(x, o1->v7.fp(o1));
	  sym = slot_symbol(x);
	  symbol_set_local(sym, id, x);
	}
      else
	{
	  int32_t tx;
	  s7_pointer arg;
	  tx = next_tx(cur_sc);
	  cur_sc->t_temps[tx] = safe_list_if_possible(cur_sc, o->v1.i);
	  
	  for (arg = cur_sc->t_temps[tx]; is_pair(arg); arg = cdr(arg))
	    {
	      o1 = cur_sc->opts[++cur_sc->pc];
	      car(arg) = o1->v7.fp(o1);
	    }
	  for (arg = cur_sc->t_temps[tx], fp = x; is_pair(arg); fp = next_slot(fp), arg = cdr(arg))
	    {
	      slot_set_value(fp, car(arg));
	      sym = slot_symbol(fp);
	      symbol_set_local(sym, id, fp);
	    }
	  clear_list_in_use(cur_sc->t_temps[tx]);
	  cur_sc->current_safe_list = 0;
	}
    }

  old_e = cur_sc->envir;
  cur_sc->envir = env;
  cur_sc->pc = o->v4.i - 1; /* preincr below */

  /* body */
  for (i = 0; i < o->v3.i; i++)
    {
      o1 = cur_sc->opts[++cur_sc->pc];
      o1->v7.fp(o1);
    }
  o1 = cur_sc->opts[++cur_sc->pc];
  result = o1->v7.fp(o1);

  cur_sc->envir = old_e;
  cur_sc->pc = o->v5.i;
  return(result);
}

static s7_pointer opt_call_no_env(void *p) {return(opt_call(p));}

static s7_pointer opt_call_1_1(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  s7_pointer result, sym, x, env, old_e;
  uint64_t id;

  id = ++cur_sc->let_number;
  env = o->v2.p;
  let_id(env) = id;
  x = let_slots(env);

  o1 = cur_sc->opts[++cur_sc->pc];
  slot_set_value(x, o1->v7.fp(o1));
  sym = slot_symbol(x);
  symbol_set_local(sym, id, x);

  old_e = cur_sc->envir;
  cur_sc->envir = env;
  cur_sc->pc = o->v4.i;

  o1 = cur_sc->opts[cur_sc->pc];
  result = o1->v7.fp(o1);

  cur_sc->envir = old_e;
  cur_sc->pc = o->v5.i;
  return(result);
}

static s7_pointer opt_call_1_1_no_env(void *p)
{
  opt_info *o = (opt_info *)p;
  opt_info *o1;
  s7_pointer result, e, old_e;

  e = o->v2.p;
  o1 = cur_sc->opts[++cur_sc->pc];
  slot_set_value(let_slots(e), o1->v7.fp(o1));

  old_e = cur_sc->envir;
  cur_sc->envir = e;
  cur_sc->pc = o->v4.i;

  o1 = cur_sc->opts[cur_sc->pc];
  result = o1->v7.fp(o1);

  cur_sc->envir = old_e;
  cur_sc->pc = o->v5.i;
  return(result);
}

static s7_int opt_int_call(void *p)          {return(integer(opt_call(p)));}
static s7_double opt_float_call(void *p)     {return(real(opt_call(p)));}
static bool opt_bool_call(void *p)           {return(opt_call(p) != cur_sc->F);}
static s7_int opt_int_call_1_1(void *p)      {return(integer(opt_call_1_1(p)));}
static s7_double opt_float_call_1_1(void *p) {return(real(opt_call_1_1(p)));}
static bool opt_bool_call_1_1(void *p)       {return(opt_call_1_1(p) != cur_sc->F);}

static bool funcall_optimize(s7_scheme *sc, s7_pointer car_x, s7_pointer s_func)
{
  if (sc->safety > CLM_OPTIMIZATION_SAFETY)
    return(false);

  if (!closure_no_opt(s_func))
    {
      opt_info *opc;
      int32_t i;
      s7_pointer p;

      opc = alloc_opo(sc, car_x);
      /* fprintf(stderr, "func: car_x: %s, opc: %p\n", DISPLAY(car_x), opc); */

      for (i = 0, p = cdr(car_x); is_pair(p); i++, p = cdr(p))
	if (!cell_optimize(sc, p))
	  break;
      if (is_pair(p))
	{
	  if ((has_optlist(s_func)) &&
	      (optlist_len(closure_optlist(s_func)) == 0))
	    clear_has_optlist(s_func);
	  return(return_false(sc, car_x, __func__, __LINE__));
	}

      opc->v1.i = i;
      opc->v2.p = closure_let(s_func);
      opc->v7.fp = opt_call;
      sc->funcalls++;
      
      if (has_optlist(s_func))
	{
	  s7_pointer olst;
	  olst = closure_optlist(s_func);
	  if (optlist_num_args(olst) != opc->v1.i)
	    {
	      /* fprintf(stderr, "wrong argnum: %" PRId64 " %d\n", opc->v1.i, optlist_num_args(olst)); */
	      return(return_false(sc, car_x, __func__, __LINE__));
	    }
	  opc->v3.i = optlist_num_exprs(olst);
	  opc->v4.i = optlist_pc(olst);
	  opc->v5.i = sc->pc - 1;

	  if ((optlist_num_exprs(olst) == 0) && /* omits final */
	      (optlist_num_args(olst) == 1))
	    opc->v7.fp = opt_call_1_1;
	  return(true);
	}
      else
	{
	  if (setjmp(sc->opt_exit) == 0)
	    {
	      s7_pointer op;
	      int32_t start, tx, body_len;

	      tx = next_tx(sc);
	      sc->t_temps[tx] = sc->envir;

	      start = sc->pc;
	      opc->v4.i = sc->pc;

	      body_len = s7_list_length(sc, closure_body(s_func));
	      if (body_len <= 0)
		return(return_false(sc, car_x, __func__, __LINE__));
	      opc->v3.i = body_len - 1;

	      /* to handle recursive calls cleanly, we need the optlist setup right away */
	      op = make_optlist(sc);
	      closure_set_optlist_addr(s_func, optlist_addr(op)); /* optlist_addr set by make_optlist (needed here for GC protection) */
	      optlist_set_num_exprs(op, body_len - 1);
	      optlist_set_num_args(op, opc->v1.i);
	      set_has_optlist(s_func);                            /*    we're optimistic! -- need to unset this if optimization fails */
	      optlist_set_len(op, 0);
	      optlist_set_pc(op, start);                          /* loaded at this position below */

	      sc->envir = closure_let(s_func);
	      for (p = closure_body(s_func); is_pair(p); p = cdr(p))
		if (!cell_optimize(sc, p))
		  break;

	      if (is_null(p))
		{
		  optlist_set_len(op, sc->pc - start);
		  opc->v5.i = sc->pc - 1;
		  sc->envir = sc->t_temps[tx];

		  if ((optlist_num_exprs(op) == 0) &&
		      (optlist_num_args(op) == 1))
		    opc->v7.fp = opt_call_1_1;

		  clear_has_optlist(s_func);
		  sc->envir = sc->t_temps[tx];
		  return(true);
		}
	      clear_has_optlist(s_func);
	      sc->envir = sc->t_temps[tx];
	    }
	}
      set_closure_no_opt(s_func);
    }
  return(return_false(sc, car_x, __func__, __LINE__));
}

static s7_pointer func_returns(s7_scheme *sc, s7_pointer func)
{
  s7_pointer body, last_expr;
  int32_t len;
  body = closure_body(func);
  len = s7_list_length(sc, body);
  if (len <= 0) return(sc->F);
  last_expr = s7_list_ref(sc, body, len - 1);
  if (is_pair(last_expr))
    {
      s7_pointer rtnf;
      rtnf = car(last_expr);
      if (!is_symbol(rtnf)) return(sc->F);
      rtnf = find_symbol_checked(sc, rtnf);
      if (!is_c_function(rtnf)) return(sc->F);
      rtnf = s7_procedure_signature(sc, rtnf);
      if (!is_pair(rtnf)) return(sc->F);
      return(car(rtnf));
    }
  return(s7_type_of(last_expr));
}

static bool returns_float(s7_scheme *sc, s7_pointer func)
{
  s7_pointer sig;
  sig = func_returns(sc, func);
  return((sig) &&
	 ((sig == sc->is_float_symbol) ||
	  ((is_pair(sig)) && (direct_memq(sc->is_float_symbol, sig)))));
}

static bool returns_integer(s7_scheme *sc, s7_pointer func)
{
  s7_pointer sig;
  sig = func_returns(sc, func);
  return((sig) &&
	 ((sig == sc->is_integer_symbol) ||
	  ((is_pair(sig)) && (direct_memq(sc->is_integer_symbol, sig)))));
}

static bool returns_bool(s7_scheme *sc, s7_pointer func)
{
  s7_pointer sig;
  sig = func_returns(sc, func);
  return(sig == sc->is_boolean_symbol);
}

static bool float_optimize(s7_scheme *sc, s7_pointer expr)
{
  s7_pointer car_x, head;
  /* fprintf(stderr, "float_optimize %s safe: %d, pair: %d\n", DISPLAY_80(expr), sc->safety, pair_no_opt(expr)); */
#if (WITH_GMP)
  return(false);
#endif

  car_x = car(expr);
  if (!is_pair(car_x)) /* wrap constants/symbols */
    return(opt_float_not_pair(sc, car_x));

  head = car(car_x);
  if (is_symbol(head))
    {
      /* get func, check sig, check all args */
      s7_pointer s_func;
      int32_t len;

      len = s7_list_length(sc, car_x);
      /* need to check int_opt here */

      if ((is_syntactic(head)) ||
	  (typesflag(car_x) == SYNTACTIC_PAIR))
	return(d_syntax_ok(sc, car_x, len));

      if ((is_global(head)) ||
	  ((is_slot(global_slot(head))) &&
	   (find_symbol(sc, head) == global_slot(head))))
	s_func = slot_value(global_slot(head));
      else return(d_implicit_ok(sc, car_x, len));

      if (is_c_function(s_func))
	{
	  opt_info *opc;
	  opc = alloc_opo(sc, car_x);
	  switch (len)
	    {
	    case 1:
	      if (d_ok(sc, opc, s_func))
		return(true);
	      break;
	      
	    case 2:                            /* (f v) or (f d): (env e) or (abs x) */
	      if ((d_d_ok(sc, opc, s_func, car_x)) ||
		  (d_v_ok(sc, opc, s_func, car_x)) ||
		  (d_p_ok(sc, opc, s_func, car_x)))
		return(true);
	      break;
	      
	    case 3:
	      /* fprintf(stderr, "flt: %s\n", DISPLAY(car_x)); */
	      if ((d_dd_ok(sc, opc, s_func, car_x)) ||
		  (d_vd_ok(sc, opc, s_func, car_x)) ||
		  (d_id_ok(sc, opc, s_func, car_x)) ||
		  (d_pd_ok(sc, opc, s_func, car_x)) ||
		  (d_ip_ok(sc, opc, s_func, car_x)) ||
		  (d_pi_ok(sc, opc, s_func, car_x)))
		return(true);
	      break;
	      
	    case 4:
	      if ((d_ddd_ok(sc, opc, s_func, car_x)) ||
		  (d_pid_ok(sc, opc, s_func, car_x)) ||
		  (d_vid_ok(sc, opc, s_func, car_x)) ||
		  (d_vdd_ok(sc, opc, s_func, car_x)))
		return(true);
	      break;
	      
	    case 5:
	      if (d_dddd_ok(sc, opc, s_func, car_x))
		return(true);
	      break;
	      
	    default:
	      if (d_add_any_ok(sc, opc, car_x, len))
		return(true);
	      break;
	    }
	  
	  if (float_all_x_ok(sc, opc, s_func, expr))
	    return(true);
	}
      else
	{
	  if (is_macro(s_func))
	    {
	      if (!pair_no_opt(expr))
		return(float_optimize(sc, set_plist_1(sc, s7_macroexpand(sc, s_func, cdar(expr)))));
	    }
	  else
	    {
	      if ((!pair_no_opt(car_x)) &&
		  (is_closure(s_func)) &&
		  (is_safe_closure(s_func)) &&
		  (returns_float(sc, s_func)))
		{
		  int32_t start;
		  start = sc->pc;
		  if (funcall_optimize(sc, car_x, s_func))
		    {
		      sc->opts[start]->v7.fd = (sc->opts[start]->v7.fp == opt_call) ? opt_float_call : opt_float_call_1_1;
		      return(true);
		    }
		  set_pair_no_opt(car_x);
		}
	    }
	}
    }
  return(return_false(sc, car_x, __func__, __LINE__));
} 

static bool int_optimize(s7_scheme *sc, s7_pointer expr)
{
  s7_pointer car_x, head;
#if (WITH_GMP)
  return(false);
#endif
  car_x = car(expr);

  if (!is_pair(car_x)) /* wrap constants/symbols */
    return(opt_int_not_pair(sc, car_x));

  head = car(car_x);
  if (is_symbol(head))
    {
      s7_pointer s_func;
      int32_t len;
      len = s7_list_length(sc, car_x);

      if ((is_syntactic(head)) ||
	  (typesflag(car_x) == SYNTACTIC_PAIR))
	return(i_syntax_ok(sc, car_x, len));

      if ((is_global(head)) ||
	  ((is_slot(global_slot(head))) &&
	   (find_symbol(sc, head) == global_slot(head))))
	s_func = slot_value(global_slot(head));
      else return(i_implicit_ok(sc, car_x, len));
	
      if (is_c_function(s_func))
	{
	  opt_info *opc;
	  opc = alloc_opo(sc, car_x);
	  switch (len)
	    {
	    case 2:
	      if (i_idp_ok(sc, opc, s_func, car_x))
		return(true);
	      break;

	    case 3:
	      if ((i_ii_ok(sc, opc, s_func, car_x)) ||
		  (i_pi_ok(sc, opc, s_func, car_x)))
		return(true);
	      break;
	      
	    case 4:
	      if ((i_iii_ok(sc, opc, s_func, car_x)) ||
		  (i_pii_ok(sc, opc, s_func, car_x)))
		return(true);
	      break;
	      
	    default:
	      if (((head == sc->add_symbol) ||
		   (head == sc->multiply_symbol)) &&
		  (i_add_any_ok(sc, opc, car_x)))
		return(true);
	      break;
	    }
	  
	  if (int_all_x_ok(sc, opc, s_func, expr))
	    return(true);
	}
      else
	{
	  if (is_macro(s_func))
	    {
	      if (!pair_no_opt(expr))
		return(int_optimize(sc, set_plist_1(sc, s7_macroexpand(sc, s_func, cdar(expr)))));
	    }
	  else
	    {
	      if ((!pair_no_opt(car_x)) &&
		  (is_closure(s_func)) &&
		  (is_safe_closure(s_func)) &&
		  (returns_integer(sc, s_func)))
		{
		  int32_t start;
		  start = sc->pc;
		  if (funcall_optimize(sc, car_x, s_func))
		    {
		      sc->opts[start]->v7.fi = (sc->opts[start]->v7.fp == opt_call) ? opt_int_call : opt_int_call_1_1;
		      return(true);
		    }
		  set_pair_no_opt(car_x);
		}
	    }
	}
    }
  return(return_false(sc, car_x, __func__, __LINE__));
}


static bool cell_optimize(s7_scheme *sc, s7_pointer expr)
{
  s7_pointer car_x, head;
  /* cell_optimize should also try *-optimize(?? -- this is premature) and wrap the results if cell-opt doesn't work */

  /* fprintf(stderr, "cell_opt %d: %s\n", sc->pc, DISPLAY_80(expr)); */

  car_x = car(expr);
  if (!is_pair(car_x)) /* wrap constants/symbols */
    return(opt_cell_not_pair(sc, car_x));

  head = car(car_x);
  if (is_symbol(head))
    {
      s7_pointer s_func = NULL;
      int32_t len;
      len = s7_list_length(sc, car_x);

      if ((is_syntactic(head)) ||
	  (typesflag(car_x) == SYNTACTIC_PAIR))
	return(p_syntax(sc, car_x, len));

      if ((is_global(head)) ||
	  ((is_slot(global_slot(head))) &&
	   (find_symbol(sc, head) == global_slot(head))))
	s_func = slot_value(global_slot(head));
      else return(p_implicit(sc, car_x, len));

      if (is_c_function(s_func))
	{
	  opt_info *opc;
	  s7_pointer sig;
	  int32_t start;
	  
	  start = sc->pc;
	  sig = s7_procedure_signature(sc, s_func);
	  opc = alloc_opo(sc, car_x);
	  switch (len)
	    {
	    case 1:
	      if (p_ok(sc, opc, s_func, car_x))
		return(true);
	      break;

	    case 2:
	      if (p_p_ok(sc, opc, s_func, car_x))
		return(true);
	      break;
	      
	    case 3:
	      {
		int32_t pstart;
		s7_i_ii_t ifunc;
		pstart = sc->pc;
		/* fprintf(stderr, "cell: %s %p\n", DISPLAY_80(car_x), s7_d_id_function(s_func)); */

		if (is_symbol(cadr(car_x)))
		  {
		    if ((is_pair(sig)) &&
			(is_pair(cdr(sig))) &&
			(is_pair(cddr(sig))) &&
			(caddr(sig) == sc->is_integer_symbol))
		      {
			if (p_pi_ok(sc, opc, s_func, car_x))
			  return(true);
			else
			  {
			    if ((car(sig) == sc->is_float_symbol) ||
				(car(sig) == sc->is_real_symbol))
			      {
				s7_d_pi_t f;
				f = s7_d_pi_function(s_func);
				if (f)
				  {
				    sc->pc--;
				    if (float_optimize(sc, expr))
				      {
					opc->v8.fd = opc->v7.fd;
					opc->v7.fp = d_to_p;
					return(true);
				      }
				  }
			      }
			  }
		      }
		    pc_fallback(sc, pstart);
		  }

		ifunc = s7_i_ii_function(s_func);
		sc->pc--;
		if ((ifunc) &&
		    (int_optimize(sc, expr)))
		  {
		    opc->v8.fi = opc->v7.fi;
		    opc->v7.fp = i_to_p;
		    return(true);
		  }
		pc_fallback(sc, pstart);
		
		if ((p_ii_ok(sc, opc, s_func, car_x, pstart)) ||
		    (p_dd_ok(sc, opc, s_func, car_x, pstart)) ||
		    (p_pp_ok(sc, opc, s_func, car_x, pstart)) ||
		    (p_cf_pp_ok(sc, opc, s_func, car_x, pstart)))
		  return(true);
	      }
	      break;
	      
	    case 4:
	      if (is_symbol(cadr(car_x)))
		{
		  int32_t pstart;
		  pstart = sc->pc;
		  if ((is_pair(sig)) &&
		      (is_pair(cdr(sig))) &&
		      (is_pair(cddr(sig))) &&
		      (caddr(sig) == sc->is_integer_symbol))
		    {
		      if (p_pip_ok(sc, opc, s_func, car_x))
			return(true);
		      else
			{
			  if (((car(sig) == sc->is_float_symbol) ||
			       (car(sig) == sc->is_real_symbol)) &&
			      (s7_d_pid_function(s_func)) &&
			      (d_pid_ok(sc, opc, s_func, car_x)))
			    {
			      opc->v8.fd = opc->v7.fd;
			      opc->v7.fp = d_to_p;
			      return(true);
			    }
			  else
			    {
			      sc->pc--;
			      if ((car(sig) == sc->is_integer_symbol) &&
				  (s7_i_pii_function(s_func)) &&
				  (i_pii_ok(sc, alloc_opo(sc, expr), s_func, car_x)))
				{
				  opc->v8.fi = opc->v7.fi;
				  opc->v7.fp = i_to_p;
				  return(true);
				}
			    }
			}
		    }
		  pc_fallback(sc, pstart);
		}

	      if ((p_ppi_ok(sc, opc, s_func, car_x)) ||
		  (p_ppp_ok(sc, opc, s_func, car_x)) ||
		  (p_cf_ppp_ok(sc, opc, s_func, car_x)))
		return(true);
	      break;
	      
	    default:
	      if (p_cf_any_ok(sc, opc, s_func, car_x, len))
		return(true);
	      break;
	    }
	  
	  if (cell_all_x_ok(sc, expr, start))
	    return(true);
	}
      else
	{
	  if (is_macro(s_func))
	    {
	      if (!pair_no_opt(expr))
		{
		  if (cell_optimize(sc, set_plist_1(sc, s7_macroexpand(sc, s_func, cdar(expr)))))
		    return(true);
		  set_pair_no_opt(expr);
		}
	    }
	  else
	    {
	      /* fprintf(stderr, "cell: %s: %d %d\n", DISPLAY_80(car_x), is_closure(s_func), is_safe_closure(s_func)); */
	      if ((!pair_no_opt(car_x)) &&
		  (is_closure(s_func)) &&
		  (is_safe_closure(s_func)))
		return(funcall_optimize(sc, car_x, s_func));
	      set_pair_no_opt(car_x);
	    }
	}
    }
  return(return_false(sc, car_x, __func__, __LINE__));
}

static bool bool_optimize_nw(s7_scheme *sc, s7_pointer expr)
{
  s7_pointer car_x, head;

  /* fprintf(stderr, "bool_opt_nw: %d %s\n", sc->pc, DISPLAY_80(expr)); */
  car_x = car(expr);
  if (!is_pair(car_x)) /* wrap constants/symbols */
    return(opt_bool_not_pair(sc, car_x));

  head = car(car_x);
  if (is_symbol(head))
    {
      s7_pointer s_func;
      int32_t len;
      len = s7_list_length(sc, car_x);

      if ((is_syntactic(head)) ||
	  (typesflag(car_x) == SYNTACTIC_PAIR))
	{
	  if (head == sc->and_symbol)
	    return(opt_b_and(sc, car_x, len));
	  if (head == sc->or_symbol)
	    return(opt_b_or(sc, car_x, len));
	  return(return_false(sc, car_x, __func__, __LINE__));
	}
      
      if ((is_global(head)) ||
	  ((is_slot(global_slot(head))) &&
	   (find_symbol(sc, head) == global_slot(head))))
	s_func = slot_value(global_slot(head));
      else return(return_false(sc, car_x, __func__, __LINE__));
      
      if (is_c_function(s_func))
	{
	  switch (len)
	    {
	    case 2:
	      return(b_idp_ok(sc, s_func, car_x, opt_arg_type(sc, cdr(car_x))));

	    case 3:
	      {
		s7_b_pp_t bpf;
		bpf = s7_b_pp_function(s_func);
		/* fprintf(stderr, "%s bpf: %s %p\n", DISPLAY_80(car_x), DISPLAY(s_func), bpf); */
		if (bpf)
		  {
		    opt_info *opc;
		    s7_pointer sig1, sig2, arg1, arg2;
		    opc = alloc_opo(sc, car_x);
		    
		    arg1 = cadr(car_x);
		    arg2 = caddr(car_x);
		    sig1 = opt_arg_type(sc, cdr(car_x));
		    sig2 = opt_arg_type(sc, cddr(car_x));

		    /* fprintf(stderr, "%s sigs: %s %s\n", DISPLAY_80(car_x), DISPLAY(sig1), DISPLAY(sig2)); */
		    if (sig2 == sc->is_integer_symbol)
		      {
			int32_t cur_index;
			cur_index = sc->pc;

			if ((sig1 == sc->is_integer_symbol) &&
			    (b_ii_ok(sc, opc, s_func, car_x, arg1, arg2)))
			  return(true);
			pc_fallback(sc, cur_index);

			if ((is_symbol(arg2)) &&
			    (b_pi_ok(sc, opc, s_func, car_x, arg2)))
			  return(true);
			pc_fallback(sc, cur_index);
		      }

		    if ((sig1 == sc->is_float_symbol) &&
			(sig2 == sc->is_float_symbol) &&
			(b_dd_ok(sc, opc, s_func, car_x, arg1, arg2)))
		      return(true);

		    opc->v3.b_pp_f = bpf;
		    return(b_pp_ok(sc, opc, s_func, car_x, arg1, arg2));
		  }
	      }
	      break;

	    default:
	      break;
	    }
	}
      else
	{
	  if (is_macro(s_func))
	    {
#if 0
	      if (!pair_no_opt(expr))
		{
		  return(bool_optimize(sc, set_plist_1(sc, s7_macroexpand(sc, s_func, cdar(expr)))));
		}
#endif
	    }
#if 1
	  else
	    {
	      if ((!pair_no_opt(car_x)) &&
		  (is_closure(s_func)) &&
		  (is_safe_closure(s_func)) &&
		  (returns_bool(sc, s_func)))
		{
		  int32_t start;
		  start = sc->pc;
		  if (funcall_optimize(sc, car_x, s_func))
		    {
#if OPT_PRINT
		      fprintf(stderr, "bool: %s\n", DISPLAY_80(car_x));
#endif
		      sc->opts[start]->v7.fb = (sc->opts[start]->v7.fp == opt_call) ? opt_bool_call : opt_bool_call_1_1;
		      return(true);
		    }
		  set_pair_no_opt(car_x);
		}
	    }
#endif
	}
    }
  return(return_false(sc, car_x, __func__, __LINE__));
}

static bool bool_optimize(s7_scheme *sc, s7_pointer expr)
{
  int32_t start;
  opt_info *wrapper;
  start = sc->pc;
  if (bool_optimize_nw(sc, expr))
    return(true);
  pc_fallback(sc, start);
  wrapper = sc->opts[start];
  if (cell_optimize(sc, expr))
    {
      if (wrapper->v8.fp) /* (when (+ i 1) ...) */
	return(false);
      wrapper->v8.fp = wrapper->v7.fp;
      wrapper->v7.fb = p_to_b;
      return(true);
    }
  return(false);
}

static s7_function s7_bool_optimize(s7_scheme *sc, s7_pointer expr)
{
#if WITH_GMP
  return(NULL);
#endif
  if (sc->safety > CLM_OPTIMIZATION_SAFETY) return(NULL);
#if OPT_PRINT
  fprintf(stderr, "bool opt: %s\n", DISPLAY(expr));
#endif
  if (setjmp(sc->opt_exit) == 0)
    {
      start_opts(sc);
      if (bool_optimize(sc, expr))
	return(opt_bool_any);
      return(all_x_optimize(sc, expr));
    }
  return(NULL);
}

s7_float_function s7_float_optimize(s7_scheme *sc, s7_pointer expr)
{
#if WITH_GMP
  return(NULL);
#endif
  if (sc->safety > CLM_OPTIMIZATION_SAFETY) return(NULL);
#if OPT_PRINT
  fprintf(stderr, "fl opt: %s\n", DISPLAY(expr));
#endif
  if (setjmp(sc->opt_exit) == 0)
    {
      start_opts(sc);
      if (float_optimize(sc, expr))
	return(opt_float_any);
    }
  return(NULL);
}

static s7_function s7_optimize_1(s7_scheme *sc, s7_pointer expr, bool nr)
{
  int32_t old_funcalls, old_unwraps;
#if WITH_GMP
  return(NULL);
#endif
  if ((sc->safety > CLM_OPTIMIZATION_SAFETY) || (pair_no_opt(expr))) 
    return(NULL);
#if OPT_PRINT
  fprintf(stderr, "opt: %s\n", DISPLAY(expr));
#endif

  old_funcalls = sc->funcalls;
  old_unwraps = sc->unwraps;

  if (setjmp(sc->opt_exit) == 0)
    {
      s7_function f;
      start_opts(sc);
      if (!no_int_opt(expr))
	{
	  if (int_optimize(sc, expr))
	    return((nr) ? opt_int_any_nr : opt_wrap_int);
	  pc_fallback(sc, 0);
	  set_no_int_opt(expr);
	}
      if (!no_float_opt(expr))
	{
	  if (float_optimize(sc, expr))
	    return((nr) ? opt_float_any_nr : opt_wrap_float);
	  pc_fallback(sc, 0);
	  set_no_float_opt(expr);
	}
      if (!no_bool_opt(expr))
	{
	  if (bool_optimize_nw(sc, expr))
	    return((nr) ? opt_bool_any_nr : opt_wrap_bool);
	  pc_fallback(sc, 0);
	  set_no_bool_opt(expr);
	}
      if (cell_optimize(sc, expr))
	{
	  /* an experiment */
	  if ((sc->funcalls > old_funcalls) &&
	      (sc->unwraps == old_unwraps))
	    {
	      int32_t i;
	      for (i = 0; i < sc->pc; i++)
		{
		  opt_info *o;
		  o = sc->opts[i];
		  if ((o->v7.fp == opt_call) || 
		      (o->v7.fp == opt_call_1_1))
		    {
		      o->v7.fp = (o->v7.fp == opt_call) ? opt_call_no_env : opt_call_1_1_no_env;
		      let_id(o->v2.p) = ++sc->let_number;
		    }
		}
	    }
	  return((nr) ? opt_cell_any_nr : opt_wrap_cell);
	}
      pc_fallback(sc, 0);
      f = all_x_optimize(sc, expr);
      if (!f)
	set_pair_no_opt(expr);
      return(f);
    }
  return(NULL);
}

s7_function s7_optimize(s7_scheme *sc, s7_pointer expr) {return(s7_optimize_1(sc, expr, false));}
s7_function s7_optimize_nr(s7_scheme *sc, s7_pointer expr) {return(s7_optimize_1(sc, expr, true));}

static s7_pointer g_optimize(s7_scheme *sc, s7_pointer args)
{
  s7_function f;
  s7_pointer code;
  code = car(args);
  f = s7_optimize(sc, code);
  if (f)
    return(f(sc, car(code)));
  return(sc->undefined);
}

static s7_function s7_cell_optimize(s7_scheme *sc, s7_pointer expr, bool nr)
{
  /* fprintf(stderr, "s7_cell_optimize %s safe: %d, pair: %d\n", DISPLAY_80(expr), sc->safety, pair_no_opt(expr)); */
#if WITH_GMP
  return(NULL);
#endif
  if (sc->safety > CLM_OPTIMIZATION_SAFETY) return(NULL);
  if (setjmp(sc->opt_exit) == 0)
    {
      start_opts(sc);
      if (!cell_optimize(sc, expr))
	return(all_x_optimize(sc, expr));
      return((nr) ? opt_cell_any_nr : opt_wrap_cell);
    }
  return(NULL);
}

/* caller: s7_float_optimize(sc, expr, env) to return a function that when called evaluates expr in env
 *   s7_float_optimize returns an s7_float_function (s7_double opt_float_any(s7_scheme *sc, s7_pointer expr) normally)
 *   s7_float_any evaluates the program stored in sc->opts by calling opts[0]->f(opts[0])
 *   each portion of expr resides in an opt_info struct, evalled by calling its "fd" function on itself
 *   fd chooses the basic form of the expr, calling one of the underlying functions in opts[n] such as opt_d_c
 *   finally that calls the actual function such as abs_d
 */


/* -------------------------------------------------------------------------------- */
static void clear_optimizer_fixups(s7_scheme *sc)
{
  hash_entry_t *p, *n;
  for (p = sc->optimizer_fixups; p; p = n)
    {
      n = p->next;
      p->next = hash_free_list;
      hash_free_list = p;
    }
  sc->optimizer_fixups = NULL;
}

static void add_optimizer_fixup(s7_scheme *sc, s7_pointer expr, uint32_t op)
{
  hash_entry_t *p;
#if DEBUGGING
  if (((op & 1) != 0) && (!all_x_function[op])) fprintf(stderr, "no all_x fixup for %s\n", opt_names[op]);
#endif
  p = make_hash_entry(expr, sc->nil, op);
  p->next = sc->optimizer_fixups;
  sc->optimizer_fixups = p;
}

static void handle_optimizer_fixups(s7_scheme *sc)
{
  hash_entry_t *p;
  for (p = sc->optimizer_fixups; p; p = p->next)
    set_optimize_op(p->key, p->raw_hash);
  clear_optimizer_fixups(sc);
}


/* ---------------------------------------- for-each ---------------------------------------- */

static inline s7_pointer make_counter(s7_scheme *sc, s7_pointer iter)
{
  s7_pointer x;
  new_cell(sc, x, T_COUNTER);
  counter_set_result(x, sc->nil);
  counter_set_list(x, iter);     /* iterator -- here it's always either an iterator or a pair */
  counter_set_capture(x, 0);     /* will be capture_let_counter */
  counter_set_let(x, sc->nil);   /* will be the saved env */
  counter_set_slots(x, sc->nil); /* local env slots before body is evalled */
  return(x);
}

static s7_pointer make_iterators(s7_scheme *sc, s7_pointer args)
{
  s7_pointer p;
  sc->temp3 = args;
  sc->z = sc->nil;                                    /* don't use sc->args here -- it needs GC protection until we get the iterators */
  for (p = cdr(args); is_pair(p); p = cdr(p))
    {
      s7_pointer iter;
      iter = car(p);
      if (!is_iterator(car(p)))
	iter = s7_make_iterator(sc, iter);
      sc->z = cons(sc, iter, sc->z);
    }
  sc->temp3 = sc->nil;
  return(safe_reverse_in_place(sc, sc->z));
}

static bool tree_has_setters(s7_scheme *sc, s7_pointer tree)
{
  clear_symbol_list(sc);
  add_symbol_to_list(sc, sc->set_symbol);
  add_symbol_to_list(sc, sc->vector_set_symbol);
  add_symbol_to_list(sc, sc->list_set_symbol);
  add_symbol_to_list(sc, sc->let_set_symbol);
  add_symbol_to_list(sc, sc->hash_table_set_symbol);
  add_symbol_to_list(sc, sc->set_car_symbol);
  add_symbol_to_list(sc, sc->set_cdr_symbol);
  return(tree_set_memq(sc, tree));
}

static s7_pointer g_for_each_closure(s7_scheme *sc, s7_pointer args)
{
  s7_pointer body, f, seq;

  seq = cadr(args);
  if (is_null(seq)) return(sc->unspecified);
  f = car(args);
  body = closure_body(f);
  
  if (!pair_no_opt(body))
    {
      s7_function func;
      s7_pointer slot, old_e, expr, pars;
      
      old_e = sc->envir;
      pars = closure_args(f);
      sc->envir = new_frame_in_env(sc, sc->envir);
      if (is_pair(seq))
	slot = make_slot_1(sc, sc->envir, car(pars), sc->F);
      else
	{
	  if (is_float_vector(seq))
	    slot = make_slot_1(sc, sc->envir, car(pars), real_zero);
	  else slot = make_slot_1(sc, sc->envir, car(pars), (is_int_vector(seq)) ? small_int(0) : sc->F);
	}
      if (is_null(cdr(body)))
	{
	  expr = car(body);
	  func = s7_optimize_nr(sc, body);
	}
      else 
	{
	  expr = cons(sc, sc->begin_symbol, body);
	  func = s7_cell_optimize(sc, cons(sc, expr, sc->nil), true);
	}
      
      if (func)
	{
	  if (is_pair(seq))
	    {
	      s7_pointer x, y;
	      for (x = seq, y = x; is_pair(x); )
		{
		  slot_set_value(slot, car(x));
		  func(sc, expr);
		  x = cdr(x);
		  if (is_pair(x))
		    {
		      slot_set_value(slot, car(x));
		      func(sc, expr);
		      y = cdr(y);
		      x = cdr(x);
		      if (x == y) return(sc->unspecified);
		    }
		}
	      return(sc->unspecified);
	    }
	  
	  if (is_float_vector(seq))
	    {
	      s7_double *vals;
	      s7_int i, len;
	      len = vector_length(seq);
	      vals = float_vector_elements(seq);
	      
	      if ((len > 1000) &&
		  (!tree_has_setters(sc, body)))
		{
		  s7_pointer sv;
		  sv = s7_make_mutable_real(sc, 0.0);
		  slot_set_value(slot, sv);
		  for (i = 0; i < len; i++)
		    {
		      real(sv) = vals[i];
		      func(sc, expr);
		    }
		  return(sc->unspecified);
		}
	      for (i = 0; i < len; i++)
		{
		  slot_set_value(slot, make_real(sc, vals[i]));
		  func(sc, expr);
		}
	      return(sc->unspecified);
	    }
	  
	  /* if no set! vector|list|let|hash-table-set! set-car!|cdr! mutable arg? */
	  if (is_int_vector(seq))
	    {
	      s7_int *vals;
	      s7_int i, len;
	      len = vector_length(seq);
	      vals = int_vector_elements(seq);
	      
	      if ((len > 1000) &&
		  (!tree_has_setters(sc, body)))
		{
		  s7_pointer sv;
		  sv = make_mutable_integer(sc, 0);
		  slot_set_value(slot, sv);
		  for (i = 0; i < len; i++)
		    {
		      integer(sv) = vals[i];
		      func(sc, expr);
		    }
		  return(sc->unspecified);
		}
	      
	      for (i = 0; i < len; i++)
		{
		  slot_set_value(slot, make_integer(sc, vals[i]));
		  func(sc, expr);
		}
	      return(sc->unspecified);
	    }
	  
	  sc->z = seq;
	  if (!is_iterator(sc->z))
	    sc->z = s7_make_iterator(sc, sc->z);
	  seq = sc->z;
	  push_stack_no_let(sc, OP_GC_PROTECT, seq, f);
	  sc->z = sc->nil;
	  while (true)
	    {
	      slot_set_value(slot, s7_iterate(sc, seq));
	      if (iterator_is_at_end(seq))
		{
		  sc->stack_end -= 4;
		  return(sc->unspecified);
		}
	      func(sc, expr);
	    }
	}
      set_pair_no_opt(body);
      sc->envir = old_e;
    }
  
  if ((is_null(cdr(body))) &&
      (is_pair(seq)))
    {
      s7_pointer c;
      c = make_counter(sc, seq);
      counter_set_result(c, seq);
      push_stack(sc, OP_FOR_EACH_2, c, f);
      return(sc->unspecified);
    }
  
  sc->z = seq;
  if (!is_iterator(sc->z))
    sc->z = s7_make_iterator(sc, sc->z);
  push_stack(sc, OP_FOR_EACH_1, make_counter(sc, sc->z), f);
  sc->z = sc->nil;
  return(sc->unspecified);
}

static inline bool for_each_arg_is_null(s7_scheme *sc, s7_pointer args)
{
  s7_pointer p;
  bool got_nil = false;
  for (p = args; is_pair(p); p = cdr(p))
    {
      s7_pointer obj;
      obj = car(p);
      if (!is_mappable(obj))
	{
	  if (is_null(obj)) 
	    got_nil = true;
	  else return(simple_wrong_type_argument_with_type(sc, sc->for_each_symbol, obj, a_sequence_string));
	}
    }
  return(got_nil);
}

static s7_pointer g_for_each(s7_scheme *sc, s7_pointer args)
{
  #define H_for_each "(for-each proc object . objects) applies proc to each element of the objects traversed in parallel. \
Each object can be a list, string, vector, hash-table, or any other sequence."
  #define Q_for_each s7_make_circular_signature(sc, 2, 3, sc->is_unspecified_symbol, sc->is_procedure_symbol, sc->is_sequence_symbol)

  s7_pointer f;
  int32_t len;
  bool arity_ok = false;

  /* try the normal case first */
  f = car(args);                                /* the function */
  len = safe_list_length(sc, cdr(args));

  if (is_closure(f))                            /* not lambda* that might get confused about arg names */
    {
      if ((len == 1) &&
	  (is_pair(closure_args(f))) &&
	  (is_null(cdr(closure_args(f)))))
	arity_ok = true;
    }
  else
    {
      if (!is_applicable(f))
	method_or_bust_with_type(sc, f, sc->for_each_symbol, args, something_applicable_string, 1);
    }

  if ((!arity_ok) &&
      (!s7_is_aritable(sc, f, len)))
    {
      static s7_pointer for_each_args_error = NULL;
      if (!for_each_args_error)
	for_each_args_error = s7_make_permanent_string("for-each ~A: ~A args?");
      return(s7_error(sc, sc->wrong_number_of_args_symbol, set_elist_3(sc, for_each_args_error, f, small_int(len))));
    }

  /* if function is safe c func, do the for-each locally */
  if ((is_safe_procedure(f)) &&
      (is_c_function(f)))
    {
      s7_function func;
      s7_pointer iters;

      if (for_each_arg_is_null(sc, cdr(args))) return(sc->unspecified);

      func = c_function_call(f);
      sc->z = make_iterators(sc, args);
      sc->z = cons(sc, sc->z, make_list(sc, len, sc->nil));
      push_stack_no_let(sc, OP_GC_PROTECT, sc->args, sc->z); /* temporary GC protection */
      if (len == 1)
	{
	  s7_pointer x, y;
	  x = caar(sc->z);
	  y = cdr(sc->z);
	  sc->z = sc->nil;
	  while (true)
	    {
	      set_car(y, s7_iterate(sc, x));
	      if (iterator_is_at_end(x))
		{
		  /* not pop_stack here since that can clobber sc->code et al, and if this for-each call is
		   *   begin treated as safe, c_call(for-each) assumes everywhere that sc->code is left alone.
		   */
		  sc->stack_end -= 4;
		  return(sc->unspecified);
		}
	      func(sc, y);
	    }
	}
      iters = sc->z;
      sc->z = sc->nil;
      while (true)
	{
	  s7_pointer x, y;
	  for (x = car(iters), y = cdr(iters); is_pair(x); x = cdr(x), y = cdr(y))
	    {
	      set_car(y, s7_iterate(sc, car(x)));
	      if (iterator_is_at_end(car(x)))
		{
		  sc->stack_end -= 4;
		  return(sc->unspecified);
		}
	    }
	  func(sc, cdr(iters));
	}
    }

  /* if closure call is straightforward, use OP_FOR_EACH_1 */
  if ((len == 1) &&
      (is_closure(f)) &&                        /* not lambda* that might get confused about arg names */
      (closure_arity_to_int(sc, f) == 1) &&     /* not a rest arg: not is_pair: (lambda (x . args) arg) */
      (!is_immutable_symbol(car(closure_args(f)))))
    return(g_for_each_closure(sc, args));

  if (for_each_arg_is_null(sc, cdr(args))) return(sc->unspecified);

  push_stack(sc, OP_FOR_EACH, cons(sc, make_iterators(sc, args), make_list(sc, len, sc->nil)), f);
  sc->z = sc->nil;
  return(sc->unspecified);
}


/* ---------------------------------------- map ---------------------------------------- */

static s7_pointer slookup(s7_scheme *sc, s7_pointer s) {return(slot_value(s));}

static s7_pointer g_map_closure(s7_scheme *sc, s7_pointer args)
{
  s7_pointer body, f, seq;

  seq = cadr(args);
  if (is_null(seq)) return(sc->nil);
  f = car(args);
  body = closure_body(f);

  /* fprintf(stderr, "f: %p, args: %s, body: %s %s\n", f, DISPLAY(args), DISPLAY(body), DISPLAY(closure_let(f))); */

  if ((is_pair(seq)) &&
      (!pair_no_opt(body)) &&
      (is_optimized(car(body)))) /* for index.scm? */
    {
      s7_function func;
      s7_pointer slot, old_e, expr;
      
      old_e = sc->envir;
      sc->envir = new_frame_in_env(sc, closure_let(f)); 
      slot = make_slot_1(sc, sc->envir, car(closure_args(f)), sc->F);
      
      if (is_null(cdr(body)))
	{
	  expr = car(body);
	  if (is_symbol(expr))
	    {
	      expr = find_symbol(sc, expr);
	      func = slookup;
	    }
	  else func = s7_optimize(sc, body);
	}
      else 
	{
	  expr = cons(sc, sc->begin_symbol, body);
	  func = s7_cell_optimize(sc, cons(sc, expr, sc->nil), false);
	}
      
      if (func)
	{
	  s7_pointer fast, slow, val;
	  
	  val = sc->nil;
	  push_stack_no_let(sc, OP_GC_PROTECT, args, val); /* was sc->args? */
	  for (fast = seq, slow = seq; is_pair(fast); fast = cdr(fast), slow = cdr(slow))
	    {
	      s7_pointer z;
	      slot_set_value(slot, car(fast));
	      z = func(sc, expr);
	      if (z != sc->no_value)
		val = cons(sc, z, val);
	      if (is_pair(cdr(fast)))
		{
		  fast = cdr(fast);
		  if (fast == slow)
		    break;
		  slot_set_value(slot, car(fast));
		  z = func(sc, expr);
		  if (z != sc->no_value)
		    val = cons(sc, z, val);
		}
	    }
	  sc->stack_end -= 4;
	  return(safe_reverse_in_place(sc, val)); 
	}
      set_pair_no_opt(body);
      sc->envir = old_e;
    }

  if ((is_null(cdr(body))) &&
      (is_pair(seq)))
    {
      closure_set_setter(f, seq);
      push_stack(sc, OP_MAP_2, make_counter(sc, seq), f);
      return(sc->unspecified);
    }
  
  sc->z = (!is_iterator(seq)) ? s7_make_iterator(sc, seq) : seq;
  push_stack(sc, OP_MAP_1, make_counter(sc, sc->z), f);
  sc->z = sc->nil;
  return(sc->nil);
}


static s7_pointer g_map(s7_scheme *sc, s7_pointer args)
{
  #define H_map "(map proc object . objects) applies proc to a list made up of the next element of each of its arguments, returning \
a list of the results.  Its arguments can be lists, vectors, strings, hash-tables, or any applicable objects."
  #define Q_map s7_make_circular_signature(sc, 2, 3, sc->is_list_symbol, sc->is_procedure_symbol, sc->is_sequence_symbol)

  static s7_pointer map_args_error = NULL;
  s7_pointer p, f;
  int32_t len;
  bool got_nil = false;

  f = car(args);                                /* the function */
  for (len = 0, p = cdr(args); is_not_null(p); p = cdr(p), len++)
    {
      if (!is_mappable(car(p)))
	{
	  if (is_null(car(p)))
	    got_nil = true;
	  else return(simple_wrong_type_argument_with_type(sc, sc->map_symbol, car(p), a_sequence_string));
	}
    }

  switch (type(f))
    {
    case T_C_FUNCTION:
    case T_C_RST_ARGS_FUNCTION:
      if (((int32_t)c_function_required_args(f) > len) ||
	  ((int32_t)c_function_all_args(f) < len))
	{
	  if (!map_args_error)
	    map_args_error = s7_make_permanent_string("map ~A: ~A args?");
	  return(s7_error(sc, sc->wrong_number_of_args_symbol, set_elist_3(sc, map_args_error, f, small_int(len))));
	}
      
    case T_C_OPT_ARGS_FUNCTION:
    case T_C_ANY_ARGS_FUNCTION:
      /* if function is safe c func, do the map locally */
      if (got_nil) return(sc->nil);
      if (is_safe_procedure(f))
	{
	  s7_function func;
	  func = c_function_call(f);
	  if ((is_pair(cadr(args))) &&
	      (len == 1))
	    {
	      s7_pointer f_args, val, fast, slow;
	      f_args = list_1(sc, sc->F);
	      val = list_1(sc, sc->nil);
	      push_stack_no_let(sc, OP_GC_PROTECT, f_args, val);
	      for (fast = cadr(args), slow = cadr(args); is_pair(fast); fast = cdr(fast), slow = cdr(slow))
		{
		  s7_pointer z;
		  set_car(f_args, car(fast));
		  z = func(sc, f_args);
		  if (z != sc->no_value)
		    set_car(val, cons(sc, z, car(val)));
		  if (is_pair(cdr(fast)))
		    {
		      fast = cdr(fast);
		      if (fast == slow)
			break;
		      set_car(f_args, car(fast));
		      z = func(sc, f_args);
		      if (z != sc->no_value)
			set_car(val, cons(sc, z, car(val)));
		    }
		}
	      sc->stack_end -= 4;
	      return(safe_reverse_in_place(sc, car(val))); 
	    }
	  else
	    {
	      s7_pointer val, val1, old_args, iter_list;
	      sc->z = make_iterators(sc, args);
	      val1 = cons(sc, sc->z, make_list(sc, len, sc->nil));
	      iter_list = sc->z;
	      old_args = sc->args;
	      func = c_function_call(f);
	      push_stack_no_let(sc, OP_GC_PROTECT, val1, val = cons(sc, sc->nil, sc->code)); /* temporary GC protection: need to protect val1, iter_list, val */
	      sc->z = sc->nil;
	      while (true)
		{
		  s7_pointer x, y, z;
		  for (x = iter_list, y = cdr(val1); is_pair(x); x = cdr(x), y = cdr(y))
		    {
		      set_car(y, s7_iterate(sc, car(x)));
		      if (iterator_is_at_end(car(x)))
			{
			  sc->stack_end -= 4;
			  sc->args = old_args;
			  return(safe_reverse_in_place(sc, car(val))); 
			}
		    }
		  z = func(sc, cdr(val1)); /* can this contain multiple-values? */
		  if (z != sc->no_value)
		    set_car(val, cons(sc, z, car(val)));
		}
	    }
	}
      else /* not safe procedure */
	{
	  /* to mimic map values handling elsewhere:
	   *   ((lambda args (format *stderr* "~A~%" (map values args))) (values)):   ()
	   *   ((lambda args (format *stderr* "~A~%" (map values args))) (values #<unspecified>)): #<unspecified> etc
	   */
	  if ((f == slot_value(global_slot(sc->values_symbol))) &&
	      (len == 1) &&
	      (!has_methods(cadr(args)))) /* iterator should be ok here -- object_to_list can handle it */
	    {
	      p = object_to_list(sc, cadr(args));
	      if (p != cadr(args))
		return(p);
	    }
	}
      break;
      
      case T_CLOSURE:
	{
	  int32_t fargs;
	  fargs = closure_arity_to_int(sc, f);
	  /* if closure call is straightforward, use OP_MAP_1 */
	  if ((len == 1) &&
	      (fargs == 1) &&
	      (!is_immutable_symbol(car(closure_args(f)))))
	    {
	      /* g_map_closure here if not s7_tree_memq 'map takes more time than it saves */
	      if (got_nil) return(sc->nil);
	      /* don't go to OP_MAP_2 here! It assumes no recursion */
	      sc->z = (!is_iterator(cadr(args))) ? s7_make_iterator(sc, cadr(args)) : cadr(args);
	      push_stack(sc, OP_MAP_1, make_counter(sc, sc->z), f);
	      sc->z = sc->nil;
	      return(sc->nil);
	    }
	  if ((fargs > len) ||
	      ((fargs < len) &&
	       ((fargs >= 0) ||
		(abs(fargs) > len))))
	    {
	      if (!map_args_error)
		map_args_error = s7_make_permanent_string("map ~A: ~A args?");
	      return(s7_error(sc, sc->wrong_number_of_args_symbol, set_elist_3(sc, map_args_error, f, small_int(len))));
	    }
	  if (got_nil) return(sc->nil);
	}
	break;

    default:
      if (!is_applicable(f))
	method_or_bust_with_type(sc, f, sc->map_symbol, args, something_applicable_string, 1);

      if ((!is_pair(f)) &&
	  (!s7_is_aritable(sc, f, len)))
	{
	  if (!map_args_error)
	    map_args_error = s7_make_permanent_string("map ~A: ~A args?");
	  return(s7_error(sc, sc->wrong_number_of_args_symbol, set_elist_3(sc, map_args_error, f, small_int(len))));
	}
      if (got_nil) return(sc->nil);
      break;
    }

  sc->z = make_iterators(sc, args);
  push_stack(sc, OP_MAP, make_counter(sc, sc->z), f);
  sc->z = sc->nil;
  return(sc->nil);
}


/* -------------------------------- multiple-values -------------------------------- */

static s7_pointer splice_in_values(s7_scheme *sc, s7_pointer args)
{
  int64_t top;
  s7_pointer x;
  top = s7_stack_top(sc) - 1; /* stack_end - stack_start: if this is negative, we're in big trouble */

  switch (stack_op(sc->stack, top))
    {
      /* the normal case -- splice values into caller's args */
    case OP_EVAL_ARGS1:
    case OP_EVAL_ARGS2:
    case OP_EVAL_ARGS3:
    case OP_EVAL_ARGS4:
      /* code = args yet to eval in order, args = evalled args reversed
       *
       * it's not safe to simply reverse args and tack the current stacked args onto its (new) end,
       *   setting stacked args to cdr of reversed-args and returning car because the list (args)
       *   can be some variable's value in a macro expansion via ,@ and reversing it in place
       *   (all this to avoid consing), clobbers the variable's value.
       */
      for (x = args; is_not_null(cdr(x)); x = cdr(x))
	stack_args(sc->stack, top) = cons(sc, car(x), stack_args(sc->stack, top));
      return(car(x));
      
      /* in the next set, the main evaluator branches blithely assume no multiple-values,
       *   and if it happens anyway, we vector to a different branch here
       */
    case OP_EVAL_ARGS_AAP_1:
      vector_element(sc->stack, top) = (s7_pointer)OP_EVAL_ARGS_AAP_MV;
      return(args);
      
    case OP_C_AP_1:
    case OP_EVAL_ARGS_P_2:
    case OP_EVAL_ARGS_P_4:
      vector_element(sc->stack, top) = (s7_pointer)OP_EVAL_ARGS_P_2_MV;
      return(args);
      
    case OP_EVAL_ARGS_P_3:
      vector_element(sc->stack, top) = (s7_pointer)OP_EVAL_ARGS_P_3_MV;
      return(args);
      
    case OP_C_P_1:
    case OP_SAFE_C_P_1:
    case OP_NOT_P_1:
      vector_element(sc->stack, top) = (s7_pointer)OP_C_P_2;
      return(args);
      
    case OP_SAFE_CLOSURE_P_1:
    case OP_CLOSURE_P_1:
      vector_element(sc->stack, top) = (s7_pointer)OP_CLOSURE_P_MV;
      return(args);

    case OP_SAFE_CLOSURE_AP_1:
    case OP_CLOSURE_AP_1:
      vector_element(sc->stack, top) = (s7_pointer)OP_CLOSURE_AP_MV;
      return(args);

    case OP_SAFE_CLOSURE_PA_1:
    case OP_CLOSURE_PA_1:
      vector_element(sc->stack, top) = (s7_pointer)OP_CLOSURE_PA_MV;
      return(args);
      
    case OP_SAFE_C_PP_1:
      vector_element(sc->stack, top) = (s7_pointer)OP_SAFE_C_PP_3;
      return(args);
      
    case OP_SAFE_C_PP_5:
      vector_element(sc->stack, top) = (s7_pointer)OP_SAFE_C_PP_6;
      return(args);
      
    case OP_EVAL_ARGS5:
      /* code = previous arg saved, args = ante-previous args reversed
       *   we'll take value->code->args and reverse in args5
       *   if one value, return it, else
       *      put code onto args, splice as above until there are 2 left
       *      set code to first and value to last
       */
      if (is_null(args))
	return(sc->unspecified);
      
      if (is_null(cdr(args)))
	return(car(args));
      
      stack_args(sc->stack, top) = cons(sc, stack_code(sc->stack, top), stack_args(sc->stack, top));
      for (x = args; is_not_null(cddr(x)); x = cdr(x))
	stack_args(sc->stack, top) = cons(sc, car(x), stack_args(sc->stack, top));
      stack_code(sc->stack, top) = car(x);
      return(cadr(x));
      
      /* look for errors here rather than glomming up the set! and let code */
    case OP_SET_SAFE:
    case OP_SET1:                                             /* (set! var (values 1 2 3)) */
      set_multiple_value(args);
      eval_error(sc, "can't set! some variable to ~S", args);
      
    case OP_SET_PAIR_P_1:
      set_multiple_value(args);
      eval_error(sc, "too many values to set! ~S", args);
      
    case OP_LET1:                                             /* (let ((var (values 1 2 3))) ...) */
    case OP_LET_ONE_1:
    case OP_LET_Z_1:
      set_multiple_value(args);
      eval_error_with_caller(sc, "~A: can't bind some variable to ~S", sc->let_symbol, args);
      /* "some variable" is ugly, but the actual name is tricky to find at this point --
       *   it's in main_stack_args, but finding the right one is a mess.  It isn't sc->code.
       */
      
    case OP_LET_STAR1:
      set_multiple_value(args);
      eval_error_with_caller(sc, "~A: can't bind some variable to ~S", sc->let_star_symbol, args);
      
    case OP_LETREC1:
      set_multiple_value(args);
      eval_error_with_caller(sc, "~A: can't bind some variable to ~S", sc->letrec_symbol, args);

    case OP_LETREC_STAR1:
      set_multiple_value(args);
      eval_error_with_caller(sc, "~A: can't bind some variable to ~S", sc->letrec_star_symbol, args);
      
      /* handle 'and' and 'or' specially */
    case OP_AND1:
      for (x = args; is_not_null(cdr(x)); x = cdr(x))
	if (car(x) == sc->F)
	  return(sc->F);
      return(car(x));
      
    case OP_OR1:
      for (x = args; is_not_null(cdr(x)); x = cdr(x))
	if (car(x) != sc->F)
	  return(car(x));
      return(car(x));
      
    case OP_BARRIER:
      pop_stack(sc);
      return(splice_in_values(sc, args));
      
    case OP_BEGIN1:
      /* here we have a values call with nothing to splice into.  So flush it...
       *   otherwise the multiple-values bit gets set in some innocent list and never unset:
       *   :(let ((x '((1 2)))) (eval `(apply apply values x)) x)
       *   ((values 1 2))
       * other cases: (+ 1 (begin (values 5 6) (values 2 3)) 4) -> 10 -- the (5 6) is dropped
       *              (let () (values 1 2 3) 4) but (+ (let () (values 1 2))) -> 3
       */
      return(args);
      
    case OP_CATCH:
    case OP_CATCH_1:
    case OP_CATCH_2:
      /* (+ (catch #t (lambda () (values 3 4)) (lambda args args))) */
      pop_stack(sc);
      return(splice_in_values(sc, args));
      
    case OP_EXPANSION:
      /* we get here if a reader-macro (define-expansion) returned multiple values.
       *    these need to be read in order into the current reader lists (we'll assume OP_READ_LIST is next in the stack.
       *    and that it will be expecting the next arg entry in sc->value).
       */
      top -= 4;
      for (x = args; is_not_null(cdr(x)); x = cdr(x))
	stack_args(sc->stack, top) = cons(sc, car(x), stack_args(sc->stack, top));
      pop_stack(sc);               /* need GC protection in loop above, so do this afterwards */
      return(car(x));              /* sc->value from OP_READ_LIST point of view */
      
    default:
      break;
    }

  /* let it meander back up the call chain until someone knows where to splice it */
  set_multiple_value(args);
  return(args);
}


s7_pointer s7_values(s7_scheme *sc, s7_pointer args)
{
  #define H_values "(values obj ...) splices its arguments into whatever list holds it (its 'continuation')"
  #define Q_values s7_make_circular_signature(sc, 1, 2, sc->values_symbol, sc->T)

  if (is_null(args))         /* ((lambda () (let ((x 1)) (set! x (boolean? (values)))))) */
    return(sc->no_value);

  if (is_null(cdr(args)))
    return(car(args));

  return(splice_in_values(sc, args));
}

#define g_values s7_values

static s7_pointer values_p(void) {return(cur_sc->no_value);}
static s7_pointer values_p_p(s7_pointer p) {return(p);}


/* -------------------------------- quasiquote -------------------------------- */

static s7_pointer g_list_values(s7_scheme *sc, s7_pointer args)
{
  #define H_list_values "(list-values ...) returns its arguments in a list (internal to quasiquote)"
  #define Q_list_values s7_make_circular_signature(sc, 1, 2, sc->is_list_symbol, sc->T)

  s7_pointer x, y, px;

  if (sc->no_values == 0)
    return(args);

  for (x = args; is_pair(x); x = cdr(x))
    if (car(x) == sc->no_value)
      break;

  if (is_null(x))
    return(args);

  /* this is not maximally efficient, but it's not important:
   *   we've hit the rare special case where (apply-values ())) needs to be ignored
   *   in the splicing process (i.e. the arglist acts as if the thing never happened)
   *   (list-values (apply-values ())) -> (), also (list-values (apply-values)) -> ()
   */
  px = sc->nil;
  for (x = args, y = args; is_pair(y); y = cdr(y))
    if (car(y) != sc->no_value)
      {
	set_car(x, car(y));
	px = x;
	x = cdr(x);
      }

  if ((is_not_null(y)) &&
      (y != sc->no_value))
    set_cdr(x, cdr(y));
  else
    {
      sc->no_values--;
      if (is_null(px))
	return(sc->nil);
      set_cdr(px, sc->nil);
    }
  return(args);
}


static s7_pointer g_apply_values(s7_scheme *sc, s7_pointer args)
{
  #define H_apply_values "(apply-values var) applies values to var.  This is an internal function."
  #define Q_apply_values s7_make_signature(sc, 2, sc->T, sc->is_list_symbol)
  s7_pointer x;

  if (is_null(args))
    {
      sc->no_values++;
      return(sc->no_value);
    }
  if (is_null(cdr(args)))
    x = car(args);
  else x = apply_list_star(sc, args);

  if (!s7_is_proper_list(sc, x))
    return(apply_list_error(sc, args));
  if (is_null(x))
    {
      sc->no_values++;
      return(sc->no_value);
    }
  return(g_values(sc, x));
}


/* (apply values ...) replaces (unquote_splicing ...)
 *
 * (define-macro (hi a) `(+ 1 ,a) == (list '+ 1 a)
 * (define-macro (hi a) ``(+ 1 ,,a) == (list list '+ 1 (list quote a)))
 *
 * (define-macro (hi a) `(+ 1 ,@a) == (list '+ 1 (apply values a))
 * (define-macro (hi a) ``(+ 1 ,,@a) == (list list '+ 1 (apply values a))
 *
 * this is not the same as CL's quasiquote; for example:
 *   [1]> (let ((a 1) (b 2)) `(,a ,@b))
 *   (1 . 2)
 *   in s7 this is an error.
 *
 * also in CL the target of ,@ can apparently be a circular list
 */

static bool is_simple_code(s7_scheme *sc, s7_pointer form)
{
  /* TODO: perhaps tree_memq here? the cycle check below is inadequate, and the last check looks dumb */
  s7_pointer tmp, sp;
  for (tmp = form, sp = form; is_pair(tmp); tmp = cdr(tmp), sp = cdr(sp))
    {
      if (is_pair(car(tmp)))
	{
	  if ((tmp == car(tmp)) || /* try to protect against #1=(#1) -- do we actually need cyclic_sequences here? */
	      (!is_simple_code(sc, car(tmp))))
	    return(false);
	}
      else
	{
	  if ((car(tmp) == sc->unquote_symbol) ||
	      ((is_null(car(tmp))) && (is_null(cdr(tmp)))))
	    return(false);
	}
      tmp = cdr(tmp);
      if (!is_pair(tmp)) break;
      if (tmp == sp) return(false);
      if (is_pair(car(tmp)))
	{
	  if ((tmp == car(tmp)) || /* try to protect against #1=(#1) -- do we actually need cyclic_sequences here? */
	      (!is_simple_code(sc, car(tmp))))
	    return(false);
	}
      else
	{
	  if ((car(tmp) == sc->unquote_symbol) ||
	      ((is_null(car(tmp))) && (is_null(cdr(tmp)))))
	    return(false);
	}
    }
  return(is_null(tmp));
}


static s7_pointer g_quasiquote_1(s7_scheme *sc, s7_pointer form)
{
  #define H_quasiquote "(quasiquote arg) is the same as `arg.  If arg is a list, it can contain \
comma (\"unquote\") and comma-atsign (\"apply values\") to pre-evaluate portions of the list. \
unquoted expressions are evaluated and plugged into the list, apply-values evaluates the expression \
and splices the resultant list into the outer list. `(1 ,(+ 1 1) ,@(list 3 4)) -> (1 2 3 4)."

  /* fprintf(stderr, "g_quasiquote_1: form: %s %s\n", DISPLAY(form), type_name(sc, form, NO_ARTICLE)); */
  if (!is_pair(form))
    {
      if ((is_symbol(form)) &&
	  (!is_keyword(form)))
	return(list_2(sc, sc->quote_symbol, form));
      /* things that evaluate to themselves don't need to be quoted. */
      return(form);
    }

  if (car(form) == sc->unquote_symbol)
    {
      if (is_not_null(cddr(form)))
	eval_error(sc, "unquote: too many arguments, ~S", form);
      return(cadr(form));
    }

  /* it's a list, so return the list with each element handled as above.
   *    we try to support dotted lists which makes the code much messier.
   */

  /* if no element of the list is a list or unquote, just return the original quoted */
  if (is_simple_code(sc, form))
    return(list_2(sc, sc->quote_symbol, form));

  {
    int32_t len, i;
    uint32_t loc;
    s7_pointer orig, bq, old_scw;
    bool dotted = false;

    len = s7_list_length(sc, form);
    if (len == 0)
      {
	/* a circular form, apparently */
	return(list_2(sc, sc->quote_symbol, form));
      }
    if (len < 0)
      {
	len = -len;
	dotted = true;
      }

    old_scw = sc->w;
    loc = s7_gc_protect(sc, old_scw);

    sc->w = sc->nil;
    for (i = 0; i <= len; i++)
      sc->w = cons(sc, sc->nil, sc->w);

    set_car(sc->w, sc->list_values_symbol);

    if (!dotted)
      {
	for (orig = form, bq = cdr(sc->w), i = 0; i < len; i++, orig = cdr(orig), bq = cdr(bq))
	  {
	    if ((is_pair(cdr(orig))) &&            /* this was is_pair(orig) which seems to be always the case */
		(cadr(orig) == sc->unquote_symbol))
	      {
		/* `(1 . ,(+ 1 1)) -> '(1 unquote (+ 1 1)) -> '(1 . 2)
		 * `(1 . ,@'((2 3))) -> (1 unquote (apply-values '((2 3)))) -> (append (list-values 1) (apply-values '((2 3)))) -> '(1 2 3)
		 * this used to be `(1 . ,@('(2 3))).
		 *     This now becomes (1 unquote (apply-values ('(2 3)))) -> (append (list-values 1) (apply-values ('(2 3)))) -> error
		 * `(1 . (,@'(2 3))) works in both cases, and `(1 . (,(+ 1 1)))
		 */
		set_car(bq, g_quasiquote_1(sc, car(orig)));
		set_cdr(bq, sc->nil);
		sc->w = list_3(sc, sc->append_symbol, sc->w, caddr(orig));
		break;
	      }
	    else set_car(bq, g_quasiquote_1(sc, car(orig)));
	  }
      }
    else
      {
	/* `(1 2 . 3) */
	len--;
	for (orig = form, bq = cdr(sc->w), i = 0; i < len; i++, orig = cdr(orig), bq = cdr(bq))
	  set_car(bq, g_quasiquote_1(sc, car(orig)));
	set_car(bq, g_quasiquote_1(sc, car(orig)));

	sc->w = list_3(sc, sc->append_symbol, sc->w, g_quasiquote_1(sc, cdr(orig)));
	/* quasiquote might quote a symbol in cdr(orig), so it's not completely pointless */
      }

    bq = sc->w;
    sc->w = old_scw;
    s7_gc_unprotect_at(sc, loc);
    return(bq);
  }
}


static s7_pointer g_quasiquote(s7_scheme *sc, s7_pointer args)
{
  /* this is for explicit quasiquote support, not the backquote stuff in macros */
  return(g_quasiquote_1(sc, car(args)));
}



/* ---------------- reader funcs for eval ---------------- */

static void back_up_stack(s7_scheme *sc)
{
  opcode_t top_op;
  top_op = stack_op(sc->stack, s7_stack_top(sc) - 1);
  if (top_op == OP_READ_DOT)
    {
      pop_stack(sc);
      top_op = stack_op(sc->stack, s7_stack_top(sc) - 1);
    }
  if ((top_op == OP_READ_VECTOR) ||
      (top_op == OP_READ_BYTE_VECTOR) ||
      (top_op == OP_READ_INT_VECTOR) ||
      (top_op == OP_READ_FLOAT_VECTOR))
   {
      pop_stack(sc);
      top_op = stack_op(sc->stack, s7_stack_top(sc) - 1);
    }
  if (top_op == OP_READ_QUOTE)
    pop_stack(sc);
}


static token_t read_sharp(s7_scheme *sc, s7_pointer pt)
{
  int32_t c;
  /* inchar can return EOF, so it can't be used directly as an index into the digits array */
  c = inchar(pt);
  switch (c)
    {
    case EOF:
      s7_error(sc, sc->read_error_symbol, set_elist_1(sc, s7_make_string_wrapper(sc, "unexpected '#' at end of input")));
      break;

    case '(':
      sc->w = small_int(1);
      return(TOKEN_VECTOR);

    case 'i':
      if (read_sharp(sc, pt) == TOKEN_VECTOR)
	return(TOKEN_INT_VECTOR);
      backchar('i', pt);
      break;

    case 'r':
      if (read_sharp(sc, pt) == TOKEN_VECTOR)
	return(TOKEN_FLOAT_VECTOR);
      backchar('r', pt);
      break;

    case '0': case '1': case '2': case '3': case '4': case '5': case '6': case '7': case '8': case '9':
      {
	/* here we can get an overflow: #1231231231231232131D()
	 *   and we can't shrug it off:
	 *   :#2147483649123D()
	 *   ;#nD(...) dimensions argument 1, -2147483647, is out of range (must be 1 or more)
	 * but
	 *   :#2147483649123D()
	 *   creates a vector with 512 dimensions!
	 * ndims in the vector struct is an uint32_t, so we'll complain if it goes over short max for now
	 */
	s7_int dims;
	int32_t d, loc = 0;
	sc->strbuf[loc++] = c;
	dims = digits[c];

	while (true)
	  {
	    s7_int dig;
	    d = inchar(pt);
	    if (d == EOF)
	      s7_error(sc, sc->read_error_symbol, set_elist_1(sc, s7_make_string_wrapper(sc, "unexpected end of input while reading #n...")));

	    dig = digits[d];
	    if (dig >= 10) break;
	    dims = dig + (dims * 10);
	    if ((dims <= 0) ||
		(dims > S7_SHORT_MAX))
	      s7_error(sc, sc->read_error_symbol, set_elist_2(sc, s7_make_string_wrapper(sc, "overflow while reading #nD: ~A"), make_integer(sc, dims)));
	    sc->strbuf[loc++] = d;
	  }
	sc->strbuf[loc++] = d;
	if ((d == 'D') || (d == 'd'))
	  {
	    d = inchar(pt);
	    if (d == EOF)
	      s7_error(sc, sc->read_error_symbol, set_elist_1(sc, s7_make_string_wrapper(sc, "unexpected end of input while reading #nD...")));
	    sc->strbuf[loc++] = d;
	    if (d == '(')
	      {
		sc->w = make_integer(sc, dims);
		return(TOKEN_VECTOR);
	      }
	  }

	/* try to back out */
	for (d = loc - 1; d > 0; d--)
	  backchar(sc->strbuf[d], pt);
      }
      break;

    case 'u':
      {
	int32_t d;
	d = inchar(pt);
	if (d == EOF)
	  s7_error(sc, sc->read_error_symbol, set_elist_1(sc, s7_make_string_wrapper(sc, "unexpected end of input while reading #u...")));
	if (d == '8')
	  {
	    d = inchar(pt);
	    if (d == EOF)
	      s7_error(sc, sc->read_error_symbol, set_elist_1(sc, s7_make_string_wrapper(sc, "unexpected end of input while reading #u8...")));
	    if (d == '(')
	      return(TOKEN_BYTE_VECTOR);
	    backchar(d, pt);
	    backchar('8', pt);
	  }
	else backchar(d, pt);
      }
      break;

    case ':':  /* turn #: into : -- this is for compatibility with Guile, sigh.
		*   I just noticed that Rick is using this -- I'll just leave it alone.
		*   but that means : readers need to handle this case specially.
		* I don't think #! is special anymore -- maybe remove that code?
		*/
      sc->strbuf[0] = ':';
      return(TOKEN_ATOM);

      /* block comments in #! ... !# */
      /* this is needed when an input file is treated as a script:
	 #!/home/bil/cl/snd
	 !#
	 (format #t "a test~%")
	 (exit)
      * but very often the closing !# is omitted which is too bad
      */
    case '!':
      {
	char last_char;
	s7_pointer reader;
	
	/* make it possible to override #! handling */
	for (reader = slot_value(sc->sharp_readers); is_pair(reader); reader = cdr(reader))
	  if (s7_character(caar(reader)) == '!')
	    {
	      sc->strbuf[0] = c;
	      return(TOKEN_SHARP_CONST); /* next stage notices any errors */
	    }

	/* not #! as block comment (for Guile I guess) */
	last_char = ' ';
	while ((c = inchar(pt)) != EOF)
	  {
	    if ((c == '#') &&
		(last_char == '!'))
	      break;
	    last_char = c;
	  }
	if (c == EOF)
	  s7_error(sc, sc->read_error_symbol, set_elist_1(sc, s7_make_string_wrapper(sc, "unexpected end of input while reading #!")));
	return(token(sc));
      }

      /* block comments in #| ... |#
       *   since we ignore everything until the |#, internal semicolon comments are ignored,
       *   meaning that ;|# is as effective as |#
       */
    case '|':
      {
	if (is_file_port(pt))
	  {
	    char last_char;
	    last_char = ' ';
	    while (true)
	      {
		c = fgetc(port_file(pt));
		if (c == EOF)
		  s7_error(sc, sc->read_error_symbol, set_elist_1(sc, s7_make_string_wrapper(sc, "unexpected end of input while reading #|")));
		if ((c == '#') &&
		    (last_char == '|'))
		  break;
		last_char = c;
		if (c == '\n')
		  port_line_number(pt)++;
	      }
	    return(token(sc));
	  }
	else
	  {
	    const char *str, *orig_str, *p, *pend;

	    orig_str = (const char *)(port_data(pt) + port_position(pt));
	    pend = (const char *)(port_data(pt) + port_data_size(pt));
	    str = orig_str;

	    while (true)
	      {
		p = strchr(str, (int)'|');
		if ((!p) || (p >= pend))
		  {
		    port_position(pt) = port_data_size(pt);
		    s7_error(sc, sc->read_error_symbol, set_elist_1(sc, s7_make_string_wrapper(sc, "unexpected end of input while reading #|")));
		  }
		if (p[1] == '#')
		  break;
		str = (const char *)(p + 1);
	      }
	    port_position(pt) += (p - orig_str + 2);

	    /* now count newline inside the comment */
	    str = (const char *)orig_str;
	    pend = p;
	    while (true)
	      {
		p = strchr(str, (int)'\n');
		if ((p) && (p < pend))
		  {
		    port_line_number(pt)++;
		    str = (char *)(p + 1);
		  }
		else break;
	      }
	    return(token(sc));
	  }
      }
    }
  sc->strbuf[0] = c;
  return(TOKEN_SHARP_CONST); /* next stage notices any errors */
}


static token_t read_comma(s7_scheme *sc, s7_pointer pt)
{
  int32_t c;
  /* here we probably should check for symbol names that start with "@":
     :(define-macro (hi @foo) `(+ ,@foo 1))
     hi
     :(hi 2)
     ;foo: unbound variable
     but
     :(define-macro (hi .foo) `(+ ,.foo 1))
     hi
     :(hi 2)
     3
     and ambiguous:
     :(define-macro (hi @foo . foo) `(list ,@foo))
     what about , @foo -- is the space significant?  We accept ,@ foo.
  */

  if ((c = inchar(pt)) == '@')
    return(TOKEN_AT_MARK);

  if (c == EOF)
    {
      sc->strbuf[0] = ',';  /* was '@' which doesn't make any sense */
      return(TOKEN_COMMA);  /* was TOKEN_ATOM, which also doesn't seem sensible */
    }

  backchar(c, pt);
  return(TOKEN_COMMA);
}


static token_t read_dot(s7_scheme *sc, s7_pointer pt)
{
  int32_t c;
  c = inchar(pt);
  if (c != EOF)
    {
      backchar(c, pt);

      if ((!char_ok_in_a_name[c]) && (c != 0))
	return(TOKEN_DOT);
    }
  else
    {
      sc->strbuf[0] = '.';
      return(TOKEN_DOT);
    }
  sc->strbuf[0] = '.';
  return(TOKEN_ATOM);  /* i.e. something that can start with a dot like a number */
}


static token_t token(s7_scheme *sc)
{
  int32_t c;
  c = port_read_white_space(sc->input_port)(sc, sc->input_port);
  switch (c)
    {
    case '(':  return(TOKEN_LEFT_PAREN);
    case ')':  return(TOKEN_RIGHT_PAREN);
    case '.':  return(read_dot(sc, sc->input_port));
    case '\'': return(TOKEN_QUOTE);
    case ';':  return(port_read_semicolon(sc->input_port)(sc, sc->input_port));
    case '"':  return(TOKEN_DOUBLE_QUOTE);
    case '`':  return(TOKEN_BACK_QUOTE);
    case ',':  return(read_comma(sc, sc->input_port));
    case '#':  return(read_sharp(sc, sc->input_port));
    case '\0':
    case EOF:  return(TOKEN_EOF);
    default:
      sc->strbuf[0] = c; /* every TOKEN_ATOM return goes to port_read_name, so we save a backchar/inchar shuffle by starting the read here */
      return(TOKEN_ATOM);
    }
}

static int32_t read_x_char(s7_scheme *sc, int32_t i, s7_pointer pt)
{
  /* possible "\xn...;" char (write creates these things, so we have to read them)
   *   but we could have crazy input like "\x -- with no trailing double quote
   */
  while (true)
    {
      int32_t d1, d2, c;
      c = inchar(pt);
      if (c == '"')
	{
	  backchar(c, pt);
	  return(i);
	}
      if (c == ';') return(i);
      if (c == EOF) 
	{
	  read_error(sc, "#<eof> in midst of hex-char");
	  return(i);
	}
      d1 = digits[c];
      if (d1 >= 16)
	{
	  sc->strbuf[i++] = c; /* just go on -- maybe a special char is not intended */
	  return(i);
	}
      c = inchar(pt);
      if (c == '"')
	{
	  sc->strbuf[i++] = d1;
	  backchar(c, pt);
	  return(i);
	}
      if (c == EOF)
	{
	  read_error(sc, "#<eof> in midst of hex-char");
	  return(i);
	}
      if (c == ';')
	{
	  sc->strbuf[i++] = d1;
	  return(i);
	}
      d2 = digits[c];
      if (d2 >= 16)
	{
	  sc->strbuf[i++] = c; /* just go on -- maybe a special char is not intended */
	  return(i);
	}
      sc->strbuf[i++] = 16 * d1 + d2;
    }
  return(i);
}

static s7_pointer unknown_string_constant(s7_scheme *sc, int32_t c)
{
  /* check *read-error-hook* */
  if (hook_has_functions(sc->read_error_hook))
    {
      s7_pointer result;
      result = s7_call(sc, sc->read_error_hook, list_2(sc, sc->F, s7_make_character(sc, (unsigned char)c)));
      if (s7_is_character(result))
	return(result);
    }
  return(sc->T);
}

static s7_pointer read_string_constant(s7_scheme *sc, s7_pointer pt)
{
  /* sc->F => error
   *   no check needed here for bad input port and so on
   */
  uint32_t i = 0;

  if (is_string_port(pt))
    {
      /* try the most common case first */
      char *s, *start, *end;
      start = (char *)(port_data(pt) + port_position(pt));
      if (*start == '"')
	{
	  port_position(pt)++;
	  return(make_empty_string(sc, 0, 0));
	}

      end = (char *)(port_data(pt) + port_data_size(pt));
      s = strpbrk(start, "\"\n\\");
      if ((!s) || (s >= end))                     /* can this read a huge string constant from a file? */
	{
	  if (start == end)
	    sc->strbuf[0] = '\0';
	  else memcpy((void *)(sc->strbuf), (void *)start, (end - start > 8) ? 8 : (end - start));
	  sc->strbuf[8] = '\0';
	  return(sc->F);
	}
      if (*s == '"')
	{
	  int32_t len;
	  len = s - start;
	  port_position(pt) += (len + 1);
	  return(s7_make_string_with_length(sc, start, len));
	}

      for (; s < end; s++)
	{
	  if (*s == '"')                         /* switch here no faster */
	    {
	      int32_t len;
	      len = s - start;
	      port_position(pt) += (len + 1);
	      return(s7_make_string_with_length(sc, start, len));
	    }
	  else
	    {
	      if (*s == '\\')
		{
		  /* all kinds of special cases here (resultant string is not the current string), so drop to loop below (setting "i") */
		  uint32_t len;
		  len = (uint32_t)(s - start);
		  if (len > 0)
		    {
		      if (len >= sc->strbuf_size)
			resize_strbuf(sc, len);
		      /* for (i = 0; i < len; i++) sc->strbuf[i] = port_data(pt)[port_position(pt)++]; */
		      memcpy((void *)(sc->strbuf), (void *)(port_data(pt) + port_position(pt)), len);
		      port_position(pt) += len;
		    }
		  i = len;
		  break;
		}
	      else
		{
		  if (*s == '\n')
		    port_line_number(pt)++;
		}
	    }
	}
    }

  while (true)
    {
      /* splitting this check out and duplicating the loop was slower?!? */
      int32_t c;
      c = port_read_character(pt)(sc, pt);

      switch (c)
	{
	case '\n':
	  port_line_number(pt)++;
	  sc->strbuf[i++] = c;
	  break;

	case EOF:
	  sc->strbuf[(i > 8) ? 8 : i] = '\0';
	  return(sc->F);

	case '"':
	  return(s7_make_string_with_length(sc, sc->strbuf, i));

	case '\\':
	  c = inchar(pt);

	  switch (c)
	    {
	    case EOF:
	      sc->strbuf[(i > 8) ? 8 : i] = '\0';
	      return(sc->F);
	      
	    case '\\': case '"': case '|':
	      sc->strbuf[i++] = c;
              break;
	      
	    case 'n': sc->strbuf[i++] = '\n'; break;
	    case 't': sc->strbuf[i++] = '\t'; break;
	    case 'r': sc->strbuf[i++] = '\r'; break;
	    case '/': sc->strbuf[i++] = '/';  break;
	    case 'b': sc->strbuf[i++] = 8;    break;
	    case 'f': sc->strbuf[i++] = 12;   break;
	      
	    case 'x':
	      i = read_x_char(sc, i, pt);
              break;
	      
            default:	      /* if (!is_white_space(c)) */ /* changed 8-Apr-12 */
	      if ((c != '\n') && (c != '\r')) /* i.e. line continuation via #\\ at end of line */
		{
		  s7_pointer result;
		  result = unknown_string_constant(sc, c);
		  if (s7_is_character(result))
		    sc->strbuf[i++] = character(result);
		  else return(result);
		}
	      /* #f here would give confusing error message "end of input", so return #t=bad backslash.
	       *     this is not optimal. It's easy to forget that backslash needs to be backslashed.
	       *
	       * the white_space business half-implements Scheme's \<newline>...<eol>... or \<space>...<eol>...
	       *   feature -- the characters after \ are flushed if they're all white space and include a newline.
	       *   (string->number "1\   2") is 12??  Too bizarre.
	       */
	    }
	  break;

	default:
	  sc->strbuf[i++] = c;
	  break;
	}

      if (i >= sc->strbuf_size)
	resize_strbuf(sc, i);
    }
}


static s7_pointer read_expression(s7_scheme *sc)
{
  while (true)
    {
      int32_t c;
      switch (sc->tok)
	{
	case TOKEN_EOF:
	  return(sc->eof_object);

	case TOKEN_BYTE_VECTOR:
	  push_stack_no_let_no_code(sc, OP_READ_BYTE_VECTOR, sc->nil); /* assume 1-dim for now */
	  sc->tok = TOKEN_LEFT_PAREN;
	  break;

	case TOKEN_INT_VECTOR:
	  push_stack_no_let_no_code(sc, OP_READ_INT_VECTOR, sc->w);
	  sc->tok = TOKEN_LEFT_PAREN;
	  break;

	case TOKEN_FLOAT_VECTOR:
	  push_stack_no_let_no_code(sc, OP_READ_FLOAT_VECTOR, sc->w);
	  sc->tok = TOKEN_LEFT_PAREN;
	  break;

	case TOKEN_VECTOR:         /* already read #( -- TOKEN_VECTOR is triggered by #( */
	  push_stack_no_let_no_code(sc, OP_READ_VECTOR, sc->w);   /* sc->w is the dimensions */
	  /* fall through */

	case TOKEN_LEFT_PAREN:
	  sc->tok = token(sc);

	  if (sc->tok == TOKEN_RIGHT_PAREN)
	    return(sc->nil);

	  if (sc->tok == TOKEN_DOT)
	    {
	      back_up_stack(sc);
	      do {c = inchar(sc->input_port);} while ((c != ')') && (c != EOF));
	      return(read_error(sc, "stray dot after '('?"));         /* (car '( . )) */
	    }

	  if (sc->tok == TOKEN_EOF)
	    return(missing_close_paren_error(sc));

	  push_stack_no_let_no_code(sc, OP_READ_LIST, sc->nil);
	  /* here we need to clear args, but code is ignored */

	  check_stack_size(sc);
	  break;

	case TOKEN_QUOTE:
	  push_stack_no_let_no_code(sc, OP_READ_QUOTE, sc->nil);
	  sc->tok = token(sc);
	  break;

	case TOKEN_BACK_QUOTE:
	  sc->tok = token(sc);
	  push_stack_no_let_no_code(sc, OP_READ_QUASIQUOTE, sc->nil);
	  break;

	case TOKEN_COMMA:
	  push_stack_no_let_no_code(sc, OP_READ_UNQUOTE, sc->nil);
	  sc->tok = token(sc);
	  switch (sc->tok)
	    {
	    case TOKEN_EOF: 
	      pop_stack(sc);
	      return(read_error(sc, "stray comma at the end of the input?")); 

	    case TOKEN_RIGHT_PAREN:
	      pop_stack(sc);
	      {
		char *str;
		str = current_input_string(sc, sc->input_port);
		if (str)
		  {
		    char *msg;
		    int32_t len;
		    msg = (char *)malloc(128 * sizeof(char));
		    len = snprintf(msg, 128, "at \"...%s...\", stray comma before ')'?", str);
		    free (str);
		    return(s7_error(sc, sc->read_error_symbol, set_elist_1(sc, make_string_uncopied_with_length(sc, msg, len))));
		  }
		return(read_error(sc, "stray comma before ')'?"));         /* '("a" "b",) */
	      }

	    default:
	      break;
	    }
	  break;

	case TOKEN_AT_MARK:
	  push_stack_no_let_no_code(sc, OP_READ_APPLY_VALUES, sc->nil);
	  sc->tok = token(sc);
	  break;

	case TOKEN_ATOM:
	  return(port_read_name(sc->input_port)(sc, sc->input_port));
	  /* If reading list (from lparen), this will finally get us to op_read_list */

	case TOKEN_DOUBLE_QUOTE:
	  sc->value = read_string_constant(sc, sc->input_port);
	  if (sc->value == sc->F)                                   /* can happen if input code ends in the middle of a string */
	    return(string_read_error(sc, "end of input encountered while in a string"));
	  if (sc->value == sc->T)
	    return(read_error(sc, "unknown backslash usage -- perhaps you meant two backslashes?"));
	  if (sc->safety > IMMUTABLE_VECTOR_SAFETY) set_immutable(sc->value);
	  return(sc->value);

	case TOKEN_SHARP_CONST:
	  sc->value = port_read_sharp(sc->input_port)(sc, sc->input_port);

	  /* here we need the following character and form
	   *   strbuf[0] == '#', false above = # case, not an atom
	   */
	  if (is_null(sc->value))
	    {
	      return(read_error(sc, "undefined # expression"));
	      /* a read error here seems draconian -- this unknown constant doesn't otherwise get in our way
	       *   but how to alert the caller to the problem without stopping the read?
	       */
	    }
	  return(sc->value);

	case TOKEN_DOT:                                             /* (catch #t (lambda () (+ 1 . . )) (lambda args 'hiho)) */
	  back_up_stack(sc);
	  do {c = inchar(sc->input_port);} while ((c != ')') && (c != EOF));
	  return(read_error(sc, "stray dot in list?"));             /* (+ 1 . . ) */

	case TOKEN_RIGHT_PAREN:                                     /* (catch #t (lambda () '(1 2 . )) (lambda args 'hiho)) */
	  back_up_stack(sc);
	  return(read_error(sc, "unexpected close paren"));         /* (+ 1 2)) or (+ 1 . ) */
	}
    }
  /* we never get here */
  return(sc->nil);
}



/* ---------------- *unbound-variable-hook* ---------------- */

static s7_pointer loaded_library(s7_scheme *sc, const char *file)
{
  s7_pointer p;
  for (p = slot_value(sc->libraries); is_pair(p); p = cdr(p))
    if (local_strcmp(file, string_value(caar(p))))
      return(cdar(p));
  return(sc->nil);
}

static s7_pointer find_closure_let(s7_scheme *sc, s7_pointer cur_env)
{
  s7_pointer e;
  for (e = cur_env; is_let(e); e = outlet(e))
    if (is_funclet(e))
      return(e);
  return(sc->nil);
}

static s7_pointer unbound_variable(s7_scheme *sc, s7_pointer sym)
{
  /* this always occurs in a context where we're trying to find anything, so I'll move a couple of those checks here
   */
  if (has_ref_fallback(sc->envir)) /* an experiment -- see s7test (with-let *db* (+ int32_t (length str))) */
    apply_known_method(sc, sc->envir, sc->let_ref_fallback_symbol, sc->w = list_2(sc, sc->envir, sym));
  /* but if the thing we want to hit this fallback happens to exist at a higher level, oops... */

  if (sym == sc->unquote_symbol)
    eval_error(sc, "unquote (',') occurred outside quasiquote: ~S", current_code(sc));

  if (sym == sc->__func___symbol) /* __func__ is a sort of symbol macro */
    {
      s7_pointer env;
      env = find_closure_let(sc, sc->envir);
      if (is_let(env))
	{
	  /* for C-defined things like hooks and dilambda, let_file and let_line are 0 */
	  if ((let_file(env) > 0) && 
	      (let_file(env) < (s7_int)sc->file_names_top) && /* let_file(env) might be > int32_t */
	      (let_line(env) > 0))
	    return(list_3(sc, funclet_function(env), sc->file_names[let_file(env)], make_integer(sc, let_line(env))));
	  return(funclet_function(env));
	}
      return(sc->undefined);
    }

  if (safe_strcmp(symbol_name(sym), "|#"))
    return(read_error(sc, "unmatched |#"));

  /* check *autoload*, autoload_names, then *unbound-variable-hook*
   */
  if ((sc->autoload_names) ||
      (is_hash_table(sc->autoload_table)) ||
      (hook_has_functions(sc->unbound_variable_hook)))
    {
      s7_pointer result, cur_code, value, code, args, cur_env, x, z;
      /* sc->args and sc->code are pushed on the stack by s7_call, then
       *   restored by eval, so they are normally protected, but sc->value and current_code(sc) are
       *   not protected (yet).  We need current_code(sc) so that the possible eventual error
       *   call can tell where the error occurred, and we need sc->value because it might
       *   be awaiting addition to sc->args in e.g. OP_EVAL_ARGS5, and then be clobbered
       *   by the hook function.  (+ 1 asdf) will end up evaluating (+ asdf asdf) if sc->value
       *   is not protected.  We also need to save/restore sc->envir in case s7_load is called.
       */

      args = sc->args;
      code = sc->code;
      value = sc->value;
      cur_code = current_code(sc);
      cur_env = sc->envir;
      result = sc->undefined;
      x = sc->x;
      z = sc->z;
      sc->temp7 = cons(sc, code, cons(sc, args, cons(sc, value, cons(sc, cur_code, cons(sc, x, cons(sc, z, sc->nil)))))); /* not s7_list (debugger checks) */

      if (!is_pair(cur_code))
	{
	  /* isolated typo perhaps -- no pair to hold the position info, so make one.
	   *   current_code(sc) is GC-protected, so this should be safe.
	   */
	  cur_code = cons(sc, sym, sc->nil);     /* the error will say "(sym)" which is not too misleading */
	  pair_set_line(cur_code, remember_location(port_line_number(sc->input_port), port_file_number(sc->input_port)));
	  set_has_line_number(cur_code);
	}

#if (!DISABLE_AUTOLOAD)
      /* check sc->autoload_names */
      if ((sc->is_autoloading) &&
	  (sc->autoload_names))
	{
	  const char *file;
	  bool loaded = false;
	  file = find_autoload_name(sc, sym, &loaded, true);
	  if ((file) && (!loaded))
	    {
	      s7_pointer e;
	      /* if we've already loaded this file, we can get the library (e) from a table [(file lib) ...]
	       * here it was possible to get caught in a loop:
	       *   change file, reload, unbound var seen, check autoload, it says "load file"... (where file does not get added to *libraries*)
	       *   so the "loaded" arg tries to catch such cases
	       */
	      e = loaded_library(sc, file);
	      if (!is_let(e))
		e = s7_load(sc, file);
	      result = s7_symbol_value(sc, sym); /* calls find_symbol, does not trigger unbound_variable search */
	      if ((result == sc->undefined) &&
		  (is_let(e)))
		{
		  result = s7_let_ref(sc, e, sym);
		  /* I think to be consistent we should add '(sym . result) to the global env */
		  if (result != sc->undefined)
		    s7_define(sc, sc->nil, sym, result);
		}
	    }
	}
#endif

      if (result == sc->undefined)
	{
#if (!DISABLE_AUTOLOAD)
	  /* check the *autoload* hash table */
	  if ((sc->is_autoloading) &&
	      (is_hash_table(sc->autoload_table)))
	    {
	      s7_pointer val;
	      /* it was possible to get in a loop here: missing paren in x.scm, checks last symbol, sees
	       *   autoload sym -> x.scm, loads x.scm, missing paren...
	       */
	      val = s7_hash_table_ref(sc, sc->autoload_table, sym);
	      if (is_string(val))                /* val should be a filename. *load-path* is searched if necessary. */
		s7_load(sc, string_value(val));
	      else
		{
		  if (is_closure(val))           /* val should be a function of one argument, the current (calling) environment */
		    s7_call(sc, val, s7_cons(sc, sc->envir, sc->nil));
		}
	      result = s7_symbol_value(sc, sym); /* calls find_symbol, does not trigger unbound_variable search */
	    }
#endif

	  /* check *unbound-variable-hook* */
	  if ((result == sc->undefined) &&
	      (hook_has_functions(sc->unbound_variable_hook)))
	    {
	      /* (let () (set! (hook-functions *unbound-variable-hook*) (list (lambda (v) _asdf_))) _asdf_) */
	      s7_pointer old_hook;

	      old_hook = sc->unbound_variable_hook;
	      set_car(sc->z2_1, old_hook);
	      sc->unbound_variable_hook = sc->error_hook;      /* avoid the infinite loop mentioned above */
	      result = s7_call(sc, old_hook, list_1(sc, sym)); /* not s7_apply_function */
	      sc->unbound_variable_hook = old_hook;
	    }
	}

      sc->value = _NFre(value);
      set_current_code(sc, cur_code);
      sc->args = args;
      sc->code = code;
      sc->envir = cur_env;
      sc->x = x;
      sc->z = z;
      sc->temp7 = sc->nil;

      if ((result != sc->undefined) &&
	  (result != sc->unspecified))
	return(result);
    }
  eval_error(sc, "~A: unbound variable", sym);
}


static s7_pointer assign_syntax(s7_scheme *sc, const char *name, opcode_t op, s7_pointer min_args, s7_pointer max_args, const char *doc)
{
  s7_pointer x, syn;
  uint64_t hash;
  uint32_t loc;

  hash = raw_string_hash((const unsigned char *)name, safe_strlen(name));
  loc = hash % SYMBOL_TABLE_SIZE;
  x = new_symbol(sc, name, safe_strlen(name), hash, loc);

  syn = alloc_pointer();
  unheap(syn);
  set_type(syn, T_SYNTAX | T_SYNTACTIC | T_DONT_EVAL_ARGS | T_GLOBAL);
  syntax_opcode(syn) = op;
  syntax_set_symbol(syn, x);
  syntax_min_args(syn) = integer(min_args);
  syntax_max_args(syn) = ((max_args == max_arity) ? -1 : integer(max_args));
  syntax_documentation(syn) = s7_make_permanent_string(doc);

  set_global_slot(x, permanent_slot(x, syn));
  set_initial_slot(x, permanent_slot(x, syn));
  set_local_slot(x, global_slot(x));
  typeflag(x) = SYNTACTIC_TYPE;  /* symbol syntactic etc */
  symbol_set_local(x, 0LL, sc->nil);
  symbol_set_ctr(x, 0;)
  return(x);
}

static s7_pointer assign_internal_syntax(s7_scheme *sc, const char *name, opcode_t op)
{
  s7_pointer x, str, syn;
  s7_pointer symbol, old_syn;

  symbol = s7_make_symbol(sc, name);
  old_syn = slot_value(global_slot(symbol));
  str = s7_make_permanent_string(name);

  x = alloc_pointer();
  unheap(x);
  set_type(x, T_SYMBOL);
  symbol_set_name_cell(x, str);
  symbol_set_local(x, 0LL, sc->nil);
  symbol_set_ctr(x, 0);

  syn = alloc_pointer();
  heap_location(syn) = heap_location(old_syn);
  set_type(syn, T_SYNTAX | T_SYNTACTIC | T_DONT_EVAL_ARGS | T_GLOBAL);
  syntax_opcode(syn) = op;
  syntax_set_symbol(syn, symbol);
  syntax_min_args(syn) = syntax_min_args(old_syn);
  syntax_max_args(syn) = syntax_max_args(old_syn);
  syntax_documentation(syn) = syntax_documentation(old_syn);

  set_global_slot(x, permanent_slot(x, syn));
  set_initial_slot(x, permanent_slot(x, syn));
  set_local_slot(x, global_slot(x));
  typeflag(x) = SYNTACTIC_TYPE;
  return(x);
}


static s7_pointer g_pair_line_number(s7_scheme *sc, s7_pointer args)
{
  s7_pointer p;
  #define H_pair_line_number "(pair-line-number pair) returns the line number at which it read 'pair', or #f if no such number is available"
  #define Q_pair_line_number s7_make_signature(sc, 2, s7_make_signature(sc, 2, sc->is_integer_symbol, sc->is_boolean_symbol), sc->is_pair_symbol)
  
  p = car(args);
  if (!is_pair(p))
    method_or_bust_one_arg(sc, p, sc->pair_line_number_symbol, set_plist_1(sc, p), T_PAIR);

  if (has_line_number(p))
    {
      uint32_t x;
      x = pair_line(p);
      return(make_integer(sc, remembered_line_number(x)));
    }
  return(sc->F); /* was 0 21-Mar-17 */
}


static s7_pointer g_pair_filename(s7_scheme *sc, s7_pointer args)
{
  #define H_pair_filename "(pair-filename pair) returns the name of the file containing 'pair'"
  #define Q_pair_filename s7_make_signature(sc, 2, s7_make_signature(sc, 2, sc->is_string_symbol, sc->is_boolean_symbol), sc->is_pair_symbol)
  s7_pointer p;
  p = car(args);

  if (!is_pair(p))
    {
      check_method(sc, p, sc->pair_filename_symbol, args);
      return(simple_wrong_type_argument(sc, sc->pair_filename_symbol, p, T_PAIR));	
    }
  if (has_line_number(p))
    {
      int32_t x;
      x = pair_line(p);
      return(remembered_file_name(x));
    }
  return(sc->F);
}


static s7_pointer is_pair_car_s, is_pair_cdr_s, is_pair_cadr_s;
static s7_pointer g_is_pair_car_s(s7_scheme *sc, s7_pointer args)
{
  s7_pointer val;
  val = find_symbol_unchecked(sc, cadar(args));
  if (is_pair(val))                                               /* (define (tst) (let ((a 123)) (pair? (car a)))) */
    return(make_boolean(sc, is_pair(car(val))));
  return(g_is_pair(sc, set_plist_1(sc, g_car(sc, set_plist_1(sc, val)))));
}

static s7_pointer g_is_pair_cdr_s(s7_scheme *sc, s7_pointer args)
{
  s7_pointer val;
  val = find_symbol_unchecked(sc, cadar(args));
  if (is_pair(val))
    return(make_boolean(sc, is_pair(cdr(val))));
  return(g_is_pair(sc, set_plist_1(sc, g_cdr(sc, set_plist_1(sc, val)))));
}

static s7_pointer g_is_pair_cadr_s(s7_scheme *sc, s7_pointer args)
{
  s7_pointer val;
  val = find_symbol_unchecked(sc, cadar(args));
  if ((is_pair(val)) && 
      (is_pair(cdr(val))))
    return(make_boolean(sc, is_pair(cadr(val))));
  return(g_is_pair(sc, set_plist_1(sc, g_cadr(sc, set_plist_1(sc, val)))));
}

static s7_pointer is_pair_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
  if (!ops) return(f);
  if (is_h_safe_c_s(cadr(expr)))
    {
      s7_function g;
      g = c_callee(cadr(expr));
      if (g == g_car)
	{
	  set_optimize_op(expr, HOP_SAFE_C_C);
	  return(is_pair_car_s);
	}
      if (g == g_cdr)
	{
	  set_optimize_op(expr, HOP_SAFE_C_C);
	  return(is_pair_cdr_s);
	}
      if (g == g_cadr)
	{
	  set_optimize_op(expr, HOP_SAFE_C_C);
	  return(is_pair_cadr_s);
	}
    }
  return(f);
}

static s7_pointer is_null_cdr;
static s7_pointer g_is_null_cdr(s7_scheme *sc, s7_pointer args)
{
  s7_pointer val;
  val = find_symbol_unchecked(sc, cadar(args));
  if (!is_pair(val))
    return(g_is_null(sc, set_plist_1(sc, g_cdr(sc, set_plist_1(sc, val)))));
  return(make_boolean(sc, is_null(cdr(val))));
}

static s7_pointer is_null_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
  if (!ops) return(f);
  if (is_h_safe_c_s(cadr(expr)))
    {
      s7_function g;
      g = c_callee(cadr(expr));
      if (g == g_cdr)
	{
	  set_optimize_op(expr, HOP_SAFE_C_C);
	  return(is_null_cdr);
	}
    }
  return(f);
}

static s7_pointer format_allg, format_allg_no_column, format_just_newline;
static s7_pointer g_format_allg(s7_scheme *sc, s7_pointer args)
{
  return(g_format_1(sc, args));
}

static s7_pointer g_format_just_newline(s7_scheme *sc, s7_pointer args)
{
  s7_pointer pt, str;

  pt = car(args);
  str = cadr(args);

  if (is_null(pt)) pt = sc->output_port;
  if (pt == sc->F)
    return(s7_make_string_with_length(sc, string_value(str), string_length(str)));

  if (pt == sc->T)
    {
      if (sc->output_port != sc->F)
	port_write_string(sc->output_port)(sc, string_value(str), string_length(str), sc->output_port);
      return(s7_make_string_with_length(sc, string_value(str), string_length(str)));
    }

  if ((!is_output_port(pt)) ||
      (port_is_closed(pt)))
    method_or_bust_with_type(sc, pt, sc->format_symbol, args, a_format_port_string, 1);

  port_write_string(pt)(sc, string_value(str), string_length(str), pt);
  return(sc->F);
}


static s7_pointer g_format_allg_no_column(s7_scheme *sc, s7_pointer args)
{
  s7_pointer pt, str;
  pt = car(args);
  if (is_null(pt)) pt = sc->output_port;

  if (!((s7_is_boolean(pt)) ||
	((is_output_port(pt)) &&             /* (current-output-port) or call-with-open-file arg, etc */
	 (!port_is_closed(pt)))))
    method_or_bust_with_type(sc, pt, sc->format_symbol, args, a_format_port_string, 1);

  str = cadr(args);
  sc->format_column = 0;
  return(format_to_port_1(sc, (pt == sc->T) ? sc->output_port : pt,
			  string_value(str), cddr(args), NULL,
			  !is_output_port(pt),   /* i.e. is boolean port so we're returning a string */
			  false,                 /* we checked in advance that it is not columnized */
			  string_length(str),
			  str));
}


static s7_pointer format_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
  if (args > 1)
    {
      s7_pointer port, str_arg;
      port = cadr(expr);
      str_arg = caddr(expr);
      if ((!is_string(port)) &&
	  (is_string(str_arg)))
	{
	  if ((ops) && (args == 2))
	    {
	      int32_t len;
	      char *orig;
	      const char *p;
	      
	      orig = string_value(str_arg);
	      p = strchr((const char *)orig, (int)'~');
	      if (!p)
		{
		  if (s7_is_boolean(port))
		    set_optimize_op(expr, HOP_SAFE_C_C);
		  return(format_just_newline); /* "just_newline" actually just outputs the control string -- see fixup below */
		}
	      
	      len = string_length(str_arg);
	      if ((len > 1) &&
		  (orig[len - 1] == '%') &&
		  ((p - orig) == len - 2))
		{
		  orig[len - 2] = '\n';
		  orig[len - 1] = '\0';
		  string_length(str_arg) = len - 1;
		  if (s7_is_boolean(port))
		    set_optimize_op(expr, HOP_SAFE_C_C);
		  return(format_just_newline);
		}
	    }
	  /* this used to worry about optimized expr and particular cases -- why? I can't find a broken case */
	  if (!is_columnizing(string_value(str_arg)))
	    return(format_allg_no_column);
	  return(format_allg);
	}
    }
  return(f);
}

static s7_pointer is_eq_car, is_eq_car_q, is_eq_caar_q;
static s7_pointer g_is_eq_car(s7_scheme *sc, s7_pointer args)
{
  s7_pointer lst, val;
  lst = find_symbol_unchecked(sc, cadar(args));
  val = find_symbol_unchecked(sc, cadr(args));
  if (is_pair(lst))
    return(make_boolean(sc, car(lst) == val));
  return(g_is_eq(sc, set_plist_2(sc, g_car(sc, set_plist_1(sc, lst)), val)));
}

static s7_pointer g_is_eq_car_q(s7_scheme *sc, s7_pointer args)
{
  s7_pointer lst;
  lst = find_symbol_unchecked(sc, opt_sym2(args));
  if (is_pair(lst))
    return(make_boolean(sc, car(lst) == opt_sym3(args)));
  return(g_is_eq(sc, set_plist_2(sc, g_car(sc, set_plist_1(sc, lst)), opt_sym3(args))));
}

static s7_pointer g_is_eq_caar_q(s7_scheme *sc, s7_pointer args)
{
  /* (eq? (caar x) 'y), but x is not guaranteed to be list(list) */
  s7_pointer lst;
  lst = find_symbol_unchecked(sc, opt_sym2(args));
  if ((!is_pair(lst)) || (!is_pair(car(lst))))
    return(g_is_eq(sc, set_plist_2(sc, g_caar(sc, set_plist_1(sc, lst)), opt_sym3(args))));
  return(make_boolean(sc, caar(lst) == opt_sym3(args)));
}

static s7_pointer is_eq_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
  if (!ops) return(f);
  if (is_h_safe_c_s(cadr(expr)))
    {
      if (((optimize_op(expr) == HOP_SAFE_C_opSq_S) ||
	   ((is_h_safe_c_c(expr)) &&
	    (is_symbol(caddr(expr))))) &&
	  (c_callee(cadr(expr)) == g_car))
	{
	  set_optimize_op(expr, HOP_SAFE_C_C);
	  return(is_eq_car);
	}
      if (is_proper_quote(sc, caddr(expr)))
	{
	  if (c_callee(cadr(expr)) == g_car)
	    {
	      set_optimize_op(expr, HOP_SAFE_C_C);
	      set_opt_sym2(cdr(expr), cadr(cadr(expr)));
	      set_opt_sym3(cdr(expr), cadr(caddr(expr)));
	      return(is_eq_car_q);
	    }
	  if (c_callee(cadr(expr)) == g_caar)
	    {
	      set_optimize_op(expr, HOP_SAFE_C_C);
	      set_opt_sym2(cdr(expr), cadr(cadr(expr)));
	      set_opt_sym3(cdr(expr), cadr(caddr(expr)));
	      return(is_eq_caar_q);
	    }
	}
    }
  return(f);
}


static s7_pointer not_is_pair_s, not_is_symbol_s, not_is_null_s, not_is_number_s;
static s7_pointer not_is_zero_s, not_is_eq_sq, not_is_eq_ss;

static s7_pointer g_not_is_pair_s(s7_scheme *sc, s7_pointer args)   {check_boolean_not_method(sc, is_pair,         sc->is_pair_symbol, args);}
static s7_pointer g_not_is_null_s(s7_scheme *sc, s7_pointer args)   {check_boolean_not_method(sc, is_null,         sc->is_null_symbol, args);}
static s7_pointer g_not_is_symbol_s(s7_scheme *sc, s7_pointer args) {check_boolean_not_method(sc, is_symbol,       sc->is_symbol_symbol, args);}
static s7_pointer g_not_is_number_s(s7_scheme *sc, s7_pointer args) {check_boolean_not_method(sc, s7_is_number,    sc->is_number_symbol, args);}
static s7_pointer g_not_is_zero_s(s7_scheme *sc, s7_pointer args)   {check_boolean_not_method(sc, s7_is_zero,      sc->is_zero_symbol, args);}

/* eq? does not check for methods */
static s7_pointer g_not_is_eq_sq(s7_scheme *sc, s7_pointer args)
{
  return(make_boolean(sc, find_symbol_unchecked(sc, cadar(args)) != opt_con2(args)));       /* cadr(caddr(car(args))) */
}

static s7_pointer g_not_is_eq_ss(s7_scheme *sc, s7_pointer args)
{
  return(make_boolean(sc, find_symbol_unchecked(sc, cadar(args)) != find_symbol_unchecked(sc, opt_sym3(args)))); /* caddr(car(args)) */
}

/* here the method finder is in either car or cdr */
static s7_pointer not_is_pair_car_s;
static s7_pointer g_not_is_pair_car_s(s7_scheme *sc, s7_pointer args)
{
  s7_pointer val;
  val = find_symbol_unchecked(sc, cadr(cadar(args)));
  if (!is_pair(val))
    return(g_not(sc, set_plist_1(sc, g_is_pair(sc, set_plist_1(sc, g_car(sc, set_plist_1(sc, val)))))));
  return(make_boolean(sc, !is_pair(car(val))));
}

static s7_pointer not_c_c;
static s7_pointer g_not_c_c(s7_scheme *sc, s7_pointer args)
{
  /* args: ( (null? l) ) */
  return(make_boolean(sc, is_false(sc, c_call(car(args))(sc, cdar(args)))));
}

static s7_pointer not_chooser(s7_scheme *sc, s7_pointer g, int32_t args, s7_pointer expr, bool ops)
{
  if (!ops) return(g);
  if (is_optimized(cadr(expr))) /* cadr(expr) might be a symbol, for example; is_optimized includes is_pair */
    {
      if (is_h_safe_c_s(cadr(expr)))
	{
	  s7_function f;
	  f = c_callee(cadr(expr));

	  if (f == g_is_pair)
	    {
	      set_optimize_op(expr, HOP_SAFE_C_C);
	      return(not_is_pair_s);
	    }
	  if (f == g_is_null)
	    {
	      set_optimize_op(expr, HOP_SAFE_C_C);
	      return(not_is_null_s);
	    }
	  if (f == g_is_symbol)
	    {
	      set_optimize_op(expr, HOP_SAFE_C_C);
	      return(not_is_symbol_s);
	    }
	  /* g_is_number is c_function_call(slot_value(global_slot(sc->is_number_symbol)))
	   *   so if this is changed (via openlet??) the latter is perhaps better??
	   * but user might have (#_number? e), so we can't change later and catch this.
	   */
	  if ((f == g_is_number) || (f == g_is_complex))
	    {
	      set_optimize_op(expr, HOP_SAFE_C_C);
	      return(not_is_number_s);
	    }

	  if (f == g_is_zero)
	    {
	      set_optimize_op(expr, HOP_SAFE_C_C);
	      return(not_is_zero_s);
	    }
	}
      if ((optimize_op(cadr(expr)) == HOP_SAFE_C_SQ) &&
	  (c_callee(cadr(expr)) == g_is_eq))
	{
	  set_optimize_op(expr, HOP_SAFE_C_C);
	  set_opt_con2(cdr(expr), cadr(caddr(cadr(expr))));
	  return(not_is_eq_sq);
	}
      if ((optimize_op(cadr(expr)) == HOP_SAFE_C_SS) || (optimize_op(cadr(expr)) == HOP_SAFE_C_LL))
	{
	  if (c_callee(cadr(expr)) == g_is_eq)
	    {
	      set_optimize_op(expr, HOP_SAFE_C_C);
	      set_opt_sym3(cdr(expr), caddr(cadr(expr)));
	      return(not_is_eq_ss);
	    }
	}
      if (is_h_safe_c_c(cadr(expr)))
	{
	  set_optimize_op(expr, HOP_SAFE_C_C);
	  if (c_callee(cadr(expr)) == g_is_pair_car_s)
	    return(not_is_pair_car_s);
	  return(not_c_c);
	}
    }
  return(g);
}


static s7_pointer vector_ref_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
  if (args == 2)
    {
      if (ops) 
	{
	  s7_pointer arg1, arg2;
	  arg1 = cadr(expr);
	  arg2 = caddr(expr);
	  if (is_symbol(arg1))
	    {
	      if (((optimize_op(expr) == HOP_SAFE_C_SC) ||
		   (is_h_safe_c_c(expr))) &&
		  (s7_is_integer(arg2)) &&
		  (s7_integer(arg2) >= 0))
		{
		  set_optimize_op(expr, HOP_SAFE_C_C);
		  switch (s7_integer(arg2)) /* (might be big int) */
		    {
		    case 0: return(vector_ref_ic_0);
		    case 1: return(vector_ref_ic_1);
		    case 2: return(vector_ref_ic_2);
		    case 3: return(vector_ref_ic_3);
		    default: return(vector_ref_ic);
		    }
		}
	      if ((is_immutable_symbol(arg1)) &&
		  (is_slot(local_slot(arg1))))
		{
		  s7_pointer v;
		  v = slot_value(local_slot(arg1));
		  if ((is_normal_vector(v)) &&
		      (vector_rank(v) == 1))
		    return(vector_ref_2_direct);
		}
	    }
	}
      /* vector_ref_sub1 was not worth the code, and few other easily optimized expressions happen here */
      return(vector_ref_2);
    }
  return(f);
}


static s7_pointer vector_set_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
  if (args == 3)
    {
      if (ops)
	{
	  s7_pointer arg1, arg2, arg3;
	  
	  arg1 = cadr(expr);
	  arg2 = caddr(expr);
	  arg3 = cadddr(expr);
	  
	  if (((optimize_op(expr) == HOP_SAFE_C_SCS) ||
	       ((is_h_safe_c_c(expr)) &&
		(is_symbol(arg1)) &&
		(is_symbol(arg3)))) &&
	      (s7_is_integer(arg2)) &&
	      (s7_integer(arg2) >= 0))
	    {
	      set_optimize_op(expr, HOP_SAFE_C_C);
	      return(vector_set_ic);
	    }
	}
      return(vector_set_3);
    }
  return(f);
}


static s7_pointer list_set_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
  if ((args == 3) &&
      (s7_is_integer(caddr(expr))) &&
      (s7_integer(caddr(expr)) >= 0) &&
      (s7_integer(caddr(expr)) < sc->max_list_length))
    return(list_set_ic);
  return(f);
}

static s7_pointer hash_table_ref_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
 if (args == 2)
    {
      /* choosers are run even in non-hop non-safe-symbol contexts, so any that look for a symbol
       *   need some way to ensure it is safe before changing to (say) hash_table_ref_ss which
       *   assumes that (coming from op_safe_c_ss normally).  But we get here more than once on
       *   the same expression somehow, so we have to recognize the initial case (h_safe_c_ss),
       *   then kludge up thge returning case (h_safe_c_c).  Using is_slot(find_symbol()) is
       *   no good because in context we use the current walker's "e" env to see the symbol,
       *   no the running environment.
       */
      if (ops)
	{
	  if ((optimize_op(expr) == HOP_SAFE_C_SS) || (optimize_op(expr) == HOP_SAFE_C_LL) ||
	      ((is_h_safe_c_c(expr)) &&
	       (is_symbol(cadr(expr))) &&
	       (is_symbol(caddr(expr)))))
	    {
	      set_optimize_op(expr, HOP_SAFE_C_C);
	      return(hash_table_ref_ss);
	    }
	  if (((optimize_op(expr) == HOP_SAFE_C_S_opSq) ||
	       ((is_h_safe_c_c(expr)) && (is_symbol(cadr(expr))))) &&
	      (is_h_safe_c_s(caddr(expr))) &&
	      (c_callee(caddr(expr)) == g_car))
	    {
	      set_optimize_op(expr, HOP_SAFE_C_C);
	      set_opt_sym3(cdr(expr), cadr(caddr(expr)));
	      return(hash_table_ref_car);
	    }
	}
      return(hash_table_ref_2);
    }
  return(f);
}


#if (!WITH_GMP)
/* an experiment in non-safe_c_c overrides for (a (b,.))
 *   expt is disabled by setting its c_function to cons
 *   modulo is rerouted by setting its function to exptmod
 *   this appears to work even in the clm-opt world because
 *   cons/exptmod are called there as well.
 */
static s7_int exptmod_1(s7_int a, s7_int b, s7_int n)
{
  s7_int d;
  if (b == 0) return(1);
  d = exptmod_1((a * a) % n, b >> 1, n);
  if ((b & 1) == 0) return(d);
  return((a * d) % n);
}

static s7_pointer exptmod;
static s7_pointer g_exptmod(s7_scheme *sc, s7_pointer args)
{
  s7_pointer a, b, n;
  a = caar(args);
  b = cdar(args);
  /* free_cell(sc, car(args)); */ /* seems to be safe */
  n = cadr(args);
  if ((!is_t_integer(a)) || (integer(a) < 0) ||
      (!is_t_integer(b)) || (integer(b) < 0) ||
      (!is_t_integer(n)) || (integer(n) <= 0))                                   /* floating exception if n==0 */
    return(g_modulo(sc, set_plist_2(sc, g_expt(sc, set_elist_2(sc, a, b)), n))); /* fallback on (modulo (expt a b) n) */
  return(make_integer(sc, exptmod_1(integer(a), integer(b), integer(n))));
}

static s7_pointer modulo_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
  if (!ops) return(f);
  if ((args == 2) &&
      ((optimize_op(expr) == HOP_SAFE_C_SC) ||
       ((is_h_safe_c_c(expr)) &&
	(is_symbol(cadr(expr))))) &&
      (is_integer(caddr(expr))) &&
      (integer(caddr(expr)) > 1))
    {
      set_optimize_op(expr, HOP_SAFE_C_C);
      return(mod_si);
    }
  
  if ((is_pair(cadr(expr))) &&
      (caadr(expr) == sc->expt_symbol))
    {
      set_c_function(cadr(expr), slot_value(initial_slot(sc->cons_symbol)));
      return(exptmod);
    }

  return(f);
}
#endif

static s7_pointer add_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
  /* (+ s f) (+ (* s s) s) (+ s s) (+ s (* s s))
   */
#if (!WITH_GMP)
  if (args == 2)
    {
      if (ops)
	{
	  s7_pointer arg1, arg2;
	  arg1 = cadr(expr);
	  arg2 = caddr(expr);
	  
	  if (arg1 == small_int(1))
	    return(add_1s);
	  
	  if (arg2 == small_int(1))                          /* (+ ... 1) */
	    {
	      if ((optimize_op(expr) == HOP_SAFE_C_SC) ||    /* (+ x 1) */
		  ((is_h_safe_c_c(expr)) && (is_symbol(arg1))))
		{
		  set_optimize_op(expr, HOP_SAFE_C_C);
		  if (is_local_symbol(cdr(expr)))
		    return(add_cl1);
		  /* fprintf(stderr, "%s\n", DISPLAY(expr)); */
		  return(add_cs1);
		}
	      return(add_s1);
	    }
	  if (s7_is_integer(arg2))
	    {
	      if ((optimize_op(expr) == HOP_SAFE_C_SC) ||    /* (+ x 123) */
		  ((is_h_safe_c_c(expr)) && (is_symbol(arg1))))
		{
		  set_optimize_op(expr, HOP_SAFE_C_C);
		  return(add_si);
		}
	    }
	  
	  if ((is_t_real(arg2)) &&
	      ((optimize_op(expr) == HOP_SAFE_C_SC) ||       /* (+ x 1.0) */
	       ((is_h_safe_c_c(expr)) && (is_symbol(arg1)))))
	    {
	      set_optimize_op(expr, HOP_SAFE_C_C);
	      return(add_sf);
	    }
	  
	  if (is_t_real(arg1))
	    {
	      if ((optimize_op(expr) == HOP_SAFE_C_CS) ||    /* (+ 1.0 x) */
		  ((is_h_safe_c_c(expr)) && (is_symbol(arg2))))
		{
		  set_optimize_op(expr, HOP_SAFE_C_C);
		  return(add_fs);
		}
	      if ((is_h_safe_c_c(arg2)) &&
		  (c_callee(arg2) == g_multiply_sf))
		{
		  set_optimize_op(expr, HOP_SAFE_C_C);
		  return(add_f_sf);
		}
	    }
	}
      return(add_2);
    }
#endif
  return(f);
}


static s7_pointer multiply_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
#if (!WITH_GMP)
  if (args == 2)
    {
      if (ops)
	{
	  s7_pointer arg1, arg2;
	  arg1 = cadr(expr);
	  arg2 = caddr(expr);
	  
	  if (is_symbol(arg1))
	    {
	      if ((optimize_op(expr) == HOP_SAFE_C_SC) ||
		  (is_h_safe_c_c(expr)))
		{
		  if (s7_is_integer(arg2))
		    {
		      set_optimize_op(expr, HOP_SAFE_C_C);
		      return(multiply_si);
		    }
		  if (is_t_real(arg2))
		    {
		      set_optimize_op(expr, HOP_SAFE_C_C);
		      return(multiply_sf);
		    }
		}
	      if ((arg1 == arg2) &&
		  ((optimize_op(expr) == HOP_SAFE_C_SS) || (optimize_op(expr) == HOP_SAFE_C_LL) ||    /* (* x c) */
		   (is_h_safe_c_c(expr))))
		{
		  set_optimize_op(expr, HOP_SAFE_C_C);
		  return(sqr_ss);
		}
	    }

	  if (is_symbol(arg2))
	    {
	      if ((optimize_op(expr) == HOP_SAFE_C_CS) ||
		  (is_h_safe_c_c(expr)))
		{
		  if (s7_is_integer(arg1))
		    {
		      set_optimize_op(expr, HOP_SAFE_C_C);
		      return(multiply_is);
		    }
		  if (is_t_real(arg1))
		    {
		      set_optimize_op(expr, HOP_SAFE_C_C);
		      return(multiply_fs);
		    }
		}
	    }
	}
      /* fprintf(stderr, "%s\n", DISPLAY_80(expr)); */
      return(multiply_2);
    }
#endif
  return(f);
}


static s7_pointer subtract_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
#if (!WITH_GMP)
  if (args == 1)
    return(subtract_1);

  if (args == 2)
    {
      if (ops)
	{
	  s7_pointer arg1, arg2;
	  arg1 = cadr(expr);
	  arg2 = caddr(expr);
	  
	  if (arg2 == small_int(1))
	    {
	      if ((optimize_op(expr) == HOP_SAFE_C_SC) ||    /* (- x 1) */
		  ((is_h_safe_c_c(expr)) && (is_symbol(arg1))))
		{
		  set_optimize_op(expr, HOP_SAFE_C_C);
		  if (is_local_symbol(cdr(expr)))
		    return(subtract_cl1);
		  return(subtract_cs1);
		}
	      return(subtract_s1);
	    }
	  
	  if (is_t_real(arg2))
	    {
	      if ((optimize_op(expr) == HOP_SAFE_C_SC) ||    /* (- x 1.0) */
		  ((is_h_safe_c_c(expr)) && (is_symbol(arg1))))
		{
		  set_optimize_op(expr, HOP_SAFE_C_C);
		  return(subtract_sf);
		}
	    }
	  
	  if (is_t_real(arg1))
	    {
	      if ((optimize_op(expr) == HOP_SAFE_C_CS) ||    /* (- 1.0 x) */
		  ((is_h_safe_c_c(expr)) && (is_symbol(arg2))))
		{
		  set_optimize_op(expr, HOP_SAFE_C_C);
		  return(subtract_fs);
		}
	    }
	  
	  if (s7_is_integer(arg2))
	    {
	      if ((optimize_op(expr) == HOP_SAFE_C_SC) ||    /* (- x 123) */
		  ((is_h_safe_c_c(expr)) && (is_symbol(arg1))))
		{
		  set_optimize_op(expr, HOP_SAFE_C_C);
		  return(subtract_csn);
		}
	    }
	  
	  if (is_t_real(arg2))
	    return(subtract_2f);
	}
      return(subtract_2);
    }
#endif
  return(f);
}


static s7_pointer divide_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
#if (!WITH_GMP)
  if (args == 1)
    return(invert_1);
  if (ops)
    {
      if (args == 2)
	{
	  s7_pointer arg1;
	  arg1 = cadr(expr);
	  if ((is_t_real(arg1)) &&
	      (real(arg1) == 1.0))
	    return(divide_1r);
	}
    }
#endif
  return(f);
}

#if (!WITH_GMP)
static s7_pointer equal_chooser(s7_scheme *sc, s7_pointer ur_f, int32_t args, s7_pointer expr, bool ops)
{
  if (args == 2)
    {
      if (ops)
	{
	  s7_pointer arg1, arg2;
	  arg1 = cadr(expr);
	  arg2 = caddr(expr);
	  
	  if (s7_is_integer(arg2))
	    {
	      if (is_safely_optimized(arg1))
		{
		  s7_function f;
		  f = c_callee(arg1);
		  if (f == g_length)
		    {
		      if (is_h_safe_c_s(arg1))
			{
			  set_optimize_op(expr, HOP_SAFE_C_C);
			  return(equal_length_ic);
			}
		    }
		}
	      if ((optimize_op(expr) == HOP_SAFE_C_SC) ||
		  ((is_h_safe_c_c(expr)) && (is_symbol(arg1))))
		{
		  set_optimize_op(expr, HOP_SAFE_C_C);
		  return(equal_s_ic);
		}
	    }
	}
      return(equal_2);
    }
  return(ur_f);
}

static s7_pointer less_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
  if (args == 2)
    {
      if (ops)
	{
	  s7_pointer arg2;
	  arg2 = caddr(expr);
	  if (is_integer(arg2))
	    {
	      if (is_h_safe_c_s(cadr(expr)))
		{
		  s7_function f;
		  f = c_callee(cadr(expr));
		  if (f == g_length)
		    {
		      set_optimize_op(expr, HOP_SAFE_C_C);
		      return(less_length_ic);
		    }
		}
	      if (integer(arg2) == 0)
		return(less_s0);
	      
	      if ((integer(arg2) < s7_int32_max) &&
		  (integer(arg2) > s7_int32_min))
		return(less_s_ic);
	    }
	}
      return(less_2);
    }
  return(f);
}


static s7_pointer leq_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
  if (args == 2)
    {
      if (ops)
	{
	  s7_pointer arg2;
	  arg2 = caddr(expr);
	  if ((is_integer(arg2)) &&
	      (integer(arg2) < s7_int32_max) &&
	      (integer(arg2) > s7_int32_min))
	    return(leq_s_ic);
	}
      return(leq_2);
    }
  return(f);
}


static s7_pointer greater_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
  if (args == 2)
    {
      if (ops)
	{
	  s7_pointer arg2;
	  arg2 = caddr(expr);
	  
	  if ((is_integer(arg2)) &&
	      (integer(arg2) < s7_int32_max) &&
	      (integer(arg2) > s7_int32_min))
	    return(greater_s_ic);
	  
	  if ((is_t_real(arg2)) &&
	      (real(arg2) < s7_int32_max) &&
	      (real(arg2) > s7_int32_min))
	    return(greater_s_fc);
	}
      return(greater_2);
    }
  return(f);
}


static s7_pointer geq_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
  if (args == 2)
    {
      if (ops)
	{
	  s7_pointer arg2;
	  arg2 = caddr(expr);
	  if ((is_integer(arg2)) &&
	      (integer(arg2) < s7_int32_max) &&
	      (integer(arg2) > s7_int32_min))
	    return(geq_s_ic);
	  if ((is_t_real(arg2)) &&
	      (real(arg2) < s7_int32_max) &&
	      (real(arg2) > s7_int32_min))
	    return(geq_s_fc);
	}
      return(geq_2);
    }
  return(f);
}
#endif
/* end (!WITH_GMP) */

static bool returns_char(s7_scheme *sc, s7_pointer arg)
{
  /* also if arg is immutable symbol + value is char */
  if (s7_is_character(arg)) return(true);
  if ((is_h_optimized(arg)) &&
      (is_safe_c_op(optimize_op(arg))) && /* make sure opt_cfunc has been set */
      (is_c_function(opt_cfunc(arg))))
    {
      s7_pointer sig;
      sig = c_function_signature(opt_cfunc(arg));
      return((sig) && 
	     (is_pair(sig)) &&
	     (car(sig) == sc->is_char_symbol));
    }
  return(false);
}

static s7_pointer char_equal_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
  if (args == 2)
    {
      if (ops)
	{
	  s7_pointer arg1, arg2;
	  arg1 = cadr(expr);
	  arg2 = caddr(expr);
	  if ((returns_char(sc, arg1)) && (returns_char(sc, arg2)))
	    return(simple_char_eq);
	  
	  if (((optimize_op(expr) == HOP_SAFE_C_SC) ||
	       ((is_h_safe_c_c(expr)) && (is_symbol(arg1)))) &&
	      (s7_is_character(arg2)))
	    {
	      set_optimize_op(expr, HOP_SAFE_C_C);
	      return(char_equal_s_ic);
	    }
	}
      return(char_equal_2);
    }
  return(f);
}

static s7_pointer char_less_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
  if (args == 2)
    return(char_less_2);
  return(f);
}

static s7_pointer char_greater_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
  if (args == 2)
    return(char_greater_2);
  return(f);
}

static void check_for_substring_temp(s7_scheme *sc, s7_pointer expr)
{
  s7_pointer p, np = NULL, ap = NULL, sp = NULL, arg;
  int32_t pairs = 0;
  /* a bit tricky -- accept temp only if there's just one inner expression and it calls substring */
  for (p = cdr(expr); is_pair(p); p = cdr(p))
    {
      arg = car(p);
      if (is_pair(arg))
	{
	  pairs++;
	  if ((is_symbol(car(arg))) &&
	      (is_safely_optimized(arg)))
	    {
	      if (c_callee(arg) == g_substring)
		np = arg;
	      else
		{
		  if (c_callee(arg) == g_number_to_string)
		    sp = arg;
		  else
		    {
		      if (c_callee(arg) == g_string_append)
			ap = arg;
		      else
			{
			  if (c_callee(arg) == g_symbol_to_string)
			    set_c_function(arg, symbol_to_string_uncopied);
			  else
			    {
			      if ((c_callee(arg) == g_read_line) &&
				  (is_pair(cdr(arg))))
				set_c_function(arg, read_line_uncopied);
			    }}}}}}}
  if (pairs == 1)
    {
      if (np)
	set_c_function(np, substring_to_temp);
      else
	{
	  if (sp)
	    set_c_function(sp, number_to_string_temp);
	  else
	    {
	      if (ap)
		{
		  for (p = ap; is_pair(p); p = cdr(p))
		    {
		      /* make sure there are no embedded uses of the temp string */
		      arg = car(p);
		      if ((is_pair(arg)) &&
			  (is_safely_optimized(arg)))
			{
			  if (c_callee(arg) == g_substring_to_temp)
			    set_c_function(arg, slot_value(global_slot(sc->substring_symbol)));
			  else
			    {
			      if (c_callee(arg) == g_string_append_to_temp)
				set_c_function(arg, slot_value(global_slot(sc->string_append_symbol)));
			    }
			}
		    }
		  set_c_function(ap, string_append_to_temp);
		}
	    }
	}
    }
}

static s7_pointer char_position_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
  if (!ops) return(f);
  if (((args == 2) || (args == 3)) &&
      (s7_is_character(cadr(expr))))
     return(char_position_csi);
  return(f);
}

static s7_pointer string_equal_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
  check_for_substring_temp(sc, expr);
  if (args == 2)
    return(string_equal_2);
  return(f);
}

static s7_pointer string_less_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
  check_for_substring_temp(sc, expr);
  if (args == 2)
    return(string_less_2);
  return(f);
}

static s7_pointer string_greater_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
  check_for_substring_temp(sc, expr);
  if (args == 2)
    return(string_greater_2);
  return(f);
}

static s7_pointer string_to_symbol_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
  check_for_substring_temp(sc, expr);
  return(f);
}


static s7_pointer string_ref_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
  check_for_substring_temp(sc, expr);
  return(f);
}

static s7_pointer string_set_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
  return(f);
}

static s7_pointer string_append_chooser(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops)
{
  check_for_substring_temp(sc, expr);
  return(f);
}


static s7_pointer or_n, or_2, or_3;
static s7_pointer g_or_n(s7_scheme *sc, s7_pointer args)
{
  s7_pointer p;
  for (p = args; is_pair(p); p = cdr(p))
    {
      s7_pointer x;
      x = c_call(p)(sc, car(p));
      if (is_true(sc, x))
	return(x);
    }
  return(sc->F);
}

static s7_pointer g_or_2(s7_scheme *sc, s7_pointer args)
{
  s7_pointer p;
  p = c_call(args)(sc, car(args));
  if (p != sc->F) return(p);
  p = cdr(args);
  return(c_call(p)(sc, car(p)));
}

static s7_pointer g_or_3(s7_scheme *sc, s7_pointer args)
{
  s7_pointer p;
  p = c_call(args)(sc, car(args));
  if (p != sc->F) return(p);
  p = cdr(args);
  p = c_call(p)(sc, car(p));
  if (p != sc->F) return(p);
  p = cddr(args);
  return(c_call(p)(sc, car(p)));
}

static s7_pointer and_n, and_2, and_sc, and_3;
static s7_pointer g_and_n(s7_scheme *sc, s7_pointer args)
{
  s7_pointer p, x = sc->T;

  for (p = args; is_pair(p); p = cdr(p))
    {
      x = c_call(p)(sc, car(p));
      if (is_false(sc, x))
	return(x);
    }
  return(x);
}

static s7_pointer g_and_2(s7_scheme *sc, s7_pointer args)
{
  if (c_call(args)(sc, car(args)) == sc->F)
    return(sc->F);
  return(c_call(cdr(args))(sc, cadr(args)));
}

static s7_pointer g_and_3(s7_scheme *sc, s7_pointer args)
{
  s7_pointer p;
  if (c_call(args)(sc, car(args)) == sc->F)
    return(sc->F);
  p = cdr(args);
  if (c_call(p)(sc, car(p)) == sc->F)
    return(sc->F);
  p = cdr(p);
  return(c_call(p)(sc, car(p)));
}

static s7_pointer g_and_sc(s7_scheme *sc, s7_pointer args)
{
   s7_pointer p;
   p = car(args);
   set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(p)));
   p = c_call(p)(sc, sc->t1_1);
   if (p == sc->F) return(p);
   p = cadr(args);
   return(c_call(p)(sc, cdr(p)));
}

static s7_pointer if_x1;
static s7_pointer g_if_x1(s7_scheme *sc, s7_pointer args)
{
  s7_pointer p;
  if (is_true(sc, c_call(args)(sc, car(args))))
    p = cdr(args);
  else return(sc->unspecified);
  return(c_call(p)(sc, car(p)));
}

static s7_pointer if_x2;
static s7_pointer g_if_x2(s7_scheme *sc, s7_pointer args)
{
  s7_pointer p;
  if (is_true(sc, c_call(args)(sc, car(args))))
    p = cdr(args);
  else p = cddr(args);
  return(c_call(p)(sc, car(p)));
}


static s7_pointer if_not_x1;
static s7_pointer g_if_not_x1(s7_scheme *sc, s7_pointer args)
{
  s7_pointer p;
  if (is_true(sc, c_call(args)(sc, cadar(args))))
    return(sc->unspecified);
  p = cdr(args);
  return(c_call(p)(sc, car(p)));
}

static s7_pointer if_not_x2;
static s7_pointer g_if_not_x2(s7_scheme *sc, s7_pointer args)
{
  s7_pointer p;
  if (is_false(sc, c_call(args)(sc, cadar(args))))
    p = cdr(args);
  else p = cddr(args);
  return(c_call(p)(sc, car(p)));
}

static s7_pointer if_x_qq;
static s7_pointer g_if_x_qq(s7_scheme *sc, s7_pointer args)
{
  if (is_true(sc, c_call(args)(sc, car(args))))
    return(cadadr(args));
  return(cadr(caddr(args)));
}

static s7_pointer if_x_qa;
static s7_pointer g_if_x_qa(s7_scheme *sc, s7_pointer args)
{
  if (is_true(sc, c_call(args)(sc, car(args))))
    return(cadadr(args));
  return(c_call(cddr(args))(sc, caddr(args)));
}
static s7_pointer or_s_direct;
static s7_pointer g_or_s_direct(s7_scheme *sc, s7_pointer args)
{
  s7_pointer p;
  set_car(sc->t1_1, find_symbol_unchecked(sc, cadar(args)));
  for (p = args; is_pair(p); p = cdr(p))
    {
      s7_pointer x;
      x = c_call(car(p))(sc, sc->t1_1);
      if (is_true(sc, x))
	return(x);
    }
  return(sc->F);
}


static s7_pointer and_s_direct;
static s7_pointer g_and_s_direct(s7_scheme *sc, s7_pointer args)
{
  s7_pointer p, x = sc->T;
  set_car(sc->t1_1, find_symbol_unchecked(sc, cadar(args)));
  for (p = args; is_pair(p); p = cdr(p))
    {
      x = c_call(car(p))(sc, sc->t1_1);
      if (is_false(sc, x))
	return(x);
    }
  return(x);
}


static s7_pointer make_function_with_class(s7_scheme *sc, s7_pointer cls, const char *name, s7_function f, 
					   int32_t required_args, int32_t optional_args, bool rest_arg, const char *doc)
{
  s7_pointer uf;
  /* the "safe_function" business here doesn't matter -- this is after the optimizer decides what is safe */
  uf = s7_make_safe_function(sc, name, f, required_args, optional_args, rest_arg, doc);
  s7_function_set_class(uf, cls);
  return(uf);
}


static s7_pointer set_function_chooser(s7_scheme *sc, s7_pointer sym, s7_pointer (*chooser)(s7_scheme *sc, s7_pointer f, int32_t args, s7_pointer expr, bool ops))
{
  s7_pointer f;
  f = slot_value(global_slot(sym));
#ifndef WITHOUT_CHOOSERS
  c_function_chooser(f) = chooser;
#endif
  return(f);
}


static void init_choosers(s7_scheme *sc)
{
  s7_pointer f;

  /* + */
  f = set_function_chooser(sc, sc->add_symbol, add_chooser);
  sc->add_class = c_function_class(f);

  add_2 = make_function_with_class(sc, f, "+", g_add_2, 2, 0, false, "+ opt");
  add_1s = make_function_with_class(sc, f, "+", g_add_1s, 2, 0, false, "+ opt");
  add_s1 = make_function_with_class(sc, f, "+", g_add_s1, 2, 0, false, "+ opt");
  add_cs1 = make_function_with_class(sc, f, "+", g_add_cs1, 2, 0, false, "+ opt");
  add_cl1 = make_function_with_class(sc, f, "+", g_add_cl1, 2, 0, false, "+ opt");
  add_si = make_function_with_class(sc, f, "+", g_add_si, 2, 0, false, "+ opt");
  add_sf = make_function_with_class(sc, f, "+", g_add_sf, 2, 0, false, "+ opt");
  add_fs = make_function_with_class(sc, f, "+", g_add_fs, 2, 0, false, "+ opt");
  add_f_sf = make_function_with_class(sc, f, "+", g_add_f_sf, 2, 0, false, "+ opt");

  /* - */
  f = set_function_chooser(sc, sc->subtract_symbol, subtract_chooser);
  sc->subtract_class = c_function_class(f);
  subtract_1 = make_function_with_class(sc, f, "-", g_subtract_1, 1, 0, false, "- opt");
  subtract_2 = make_function_with_class(sc, f, "-", g_subtract_2, 2, 0, false, "- opt");
  subtract_s1 = make_function_with_class(sc, f, "-", g_subtract_s1, 2, 0, false, "- opt");
  subtract_cs1 = make_function_with_class(sc, f, "-", g_subtract_cs1, 2, 0, false, "- opt");
  subtract_cl1 = make_function_with_class(sc, f, "-", g_subtract_cl1, 2, 0, false, "- opt");
  subtract_csn = make_function_with_class(sc, f, "-", g_subtract_csn, 2, 0, false, "- opt");
  subtract_sf = make_function_with_class(sc, f, "-", g_subtract_sf, 2, 0, false, "- opt");
  subtract_2f = make_function_with_class(sc, f, "-", g_subtract_2f, 2, 0, false, "- opt");
  subtract_fs = make_function_with_class(sc, f, "-", g_subtract_fs, 2, 0, false, "- opt");

  /* * */
  f = set_function_chooser(sc, sc->multiply_symbol, multiply_chooser);
  sc->multiply_class = c_function_class(f);
#if (!WITH_GMP)
  multiply_2 = make_function_with_class(sc, f, "*", g_multiply_2, 2, 0, false, "* opt");
  multiply_is = make_function_with_class(sc, f, "*", g_multiply_is, 2, 0, false, "* opt");
  multiply_si = make_function_with_class(sc, f, "*", g_multiply_si, 2, 0, false, "* opt");
  multiply_fs = make_function_with_class(sc, f, "*", g_multiply_fs, 2, 0, false, "* opt");
  multiply_sf = make_function_with_class(sc, f, "*", g_multiply_sf, 2, 0, false, "* opt");

  sqr_ss = make_function_with_class(sc, f, "*", g_sqr_ss, 2, 0, false, "* opt");
#endif

  /* / */
  f = set_function_chooser(sc, sc->divide_symbol, divide_chooser);
#if (!WITH_GMP)
  invert_1 = make_function_with_class(sc, f, "/", g_invert_1, 1, 0, false, "/ opt");
  divide_1r = make_function_with_class(sc, f, "/", g_divide_1r, 2, 0, false, "/ opt");

  /* modulo */
  f = set_function_chooser(sc, sc->modulo_symbol, modulo_chooser);
  mod_si = make_function_with_class(sc, f, "modulo", g_mod_si, 2, 0, false, "modulo opt");
  exptmod = make_function_with_class(sc, f, "modulo", g_exptmod, 3, 0, false, "modulo opt");

  /* = */
  f = set_function_chooser(sc, sc->eq_symbol, equal_chooser);
  sc->equal_class = c_function_class(f);
  equal_s_ic = make_function_with_class(sc, f, "=", g_equal_s_ic, 2, 0, false, "= opt");
  equal_length_ic = make_function_with_class(sc, f, "=", g_equal_length_ic, 2, 0, false, "= opt");
  equal_2 = make_function_with_class(sc, f, "=", g_equal_2, 2, 0, false, "= opt");

  /* < */
  f = set_function_chooser(sc, sc->lt_symbol, less_chooser);
  less_s_ic = make_function_with_class(sc, f, "<", g_less_s_ic, 2, 0, false, "< opt");
  less_s0 = make_function_with_class(sc, f, "<", g_less_s0, 2, 0, false, "< opt");
  less_2 = make_function_with_class(sc, f, "<", g_less_2, 2, 0, false, "< opt");
  less_length_ic = make_function_with_class(sc, f, "<", g_less_length_ic, 2, 0, false, "< opt");

  /* > */
  f = set_function_chooser(sc, sc->gt_symbol, greater_chooser);
  greater_s_ic = make_function_with_class(sc, f, ">", g_greater_s_ic, 2, 0, false, "> opt");
  greater_s_fc = make_function_with_class(sc, f, ">", g_greater_s_fc, 2, 0, false, "> opt");
  greater_2 = make_function_with_class(sc, f, ">", g_greater_2, 2, 0, false, "> opt");

  /* <= */
  f = set_function_chooser(sc, sc->leq_symbol, leq_chooser);
  leq_s_ic = make_function_with_class(sc, f, "<=", g_leq_s_ic, 2, 0, false, "<= opt");
  leq_2 = make_function_with_class(sc, f, "<=", g_leq_2, 2, 0, false, "<= opt");

  /* >= */
  f = set_function_chooser(sc, sc->geq_symbol, geq_chooser);
  geq_s_ic = make_function_with_class(sc, f, ">=", g_geq_s_ic, 2, 0, false, ">= opt");
  geq_s_fc = make_function_with_class(sc, f, ">=", g_geq_s_fc, 2, 0, false, ">= opt");
  geq_2 = make_function_with_class(sc, f, ">=", g_geq_2, 2, 0, false, ">= opt");

  /* random */
  f = set_function_chooser(sc, sc->random_symbol, random_chooser);
  random_ic = make_function_with_class(sc, f, "random", g_random_ic, 1, 0, false, "random opt");
  random_rc = make_function_with_class(sc, f, "random", g_random_rc, 1, 0, false, "random opt");
#endif

  /* char=? */
  f = set_function_chooser(sc, sc->char_eq_symbol, char_equal_chooser);
  simple_char_eq = make_function_with_class(sc, f, "char=?", g_simple_char_eq, 2, 0, false, "char=? opt");
  char_equal_s_ic = make_function_with_class(sc, f, "char=?", g_char_equal_s_ic, 2, 0, false, "char=? opt");
  char_equal_2 = make_function_with_class(sc, f, "char=?", g_char_equal_2, 2, 0, false, "char=? opt");

  /* char>? */
  f = set_function_chooser(sc, sc->char_gt_symbol, char_greater_chooser);
  char_greater_2 = make_function_with_class(sc, f, "char>?", g_char_greater_2, 2, 0, false, "char>? opt");

  /* char<? */
  f = set_function_chooser(sc, sc->char_lt_symbol, char_less_chooser);
  char_less_2 = make_function_with_class(sc, f, "char<?", g_char_less_2, 2, 0, false, "char<? opt");

  /* char-position */
  f = set_function_chooser(sc, sc->char_position_symbol, char_position_chooser);
  char_position_csi = make_function_with_class(sc, f, "char-position", g_char_position_csi, 2, 1, false, "char-position opt");

  /* string->symbol */
  set_function_chooser(sc, sc->string_to_symbol_symbol, string_to_symbol_chooser);

  /* string=? */
  f = set_function_chooser(sc, sc->string_eq_symbol, string_equal_chooser);
  string_equal_2 = make_function_with_class(sc, f, "string=?", g_string_equal_2, 2, 0, false, "string=? opt");

  /* substring */
  substring_to_temp = s7_make_function(sc, "substring", g_substring_to_temp, 2, 1, false, "substring opt");
  s7_function_set_class(substring_to_temp, slot_value(global_slot(sc->substring_symbol)));

  /* number->string */
  number_to_string_temp = s7_make_function(sc, "number->string", g_number_to_string_temp, 1, 1, false, "number->string opt");
  s7_function_set_class(number_to_string_temp, slot_value(global_slot(sc->number_to_string_symbol)));

  /* string>? */
  f = set_function_chooser(sc, sc->string_gt_symbol, string_greater_chooser);
  string_greater_2 = make_function_with_class(sc, f, "string>?", g_string_greater_2, 2, 0, false, "string>? opt");

  /* string<? */
  f = set_function_chooser(sc, sc->string_lt_symbol, string_less_chooser);
  string_less_2 = make_function_with_class(sc, f, "string<?", g_string_less_2, 2, 0, false, "string<? opt");

  /* string-ref */
  set_function_chooser(sc, sc->string_ref_symbol, string_ref_chooser);

  /* string-set! */
  set_function_chooser(sc, sc->string_set_symbol, string_set_chooser);

  /* string-append */
  f = set_function_chooser(sc, sc->string_append_symbol, string_append_chooser);
  string_append_to_temp = make_function_with_class(sc, f, "string-append", g_string_append_to_temp, 0, 0, true, "string-append opt");

  /* symbol->string */
  f = slot_value(global_slot(sc->symbol_to_string_symbol));
  symbol_to_string_uncopied = s7_make_function(sc, "symbol->string", g_symbol_to_string_uncopied, 1, 0, false, "symbol->string opt");
  s7_function_set_class(symbol_to_string_uncopied, f);

  /* vector-ref */
  f = set_function_chooser(sc, sc->vector_ref_symbol, vector_ref_chooser);
  vector_ref_ic = make_function_with_class(sc, f, "vector-ref", g_vector_ref_ic, 2, 0, false, "vector-ref opt");
  vector_ref_ic_0 = make_function_with_class(sc, f, "vector-ref", g_vector_ref_ic_0, 1, 0, false, "vector-ref opt");
  vector_ref_ic_1 = make_function_with_class(sc, f, "vector-ref", g_vector_ref_ic_1, 1, 0, false, "vector-ref opt");
  vector_ref_ic_2 = make_function_with_class(sc, f, "vector-ref", g_vector_ref_ic_2, 1, 0, false, "vector-ref opt");
  vector_ref_ic_3 = make_function_with_class(sc, f, "vector-ref", g_vector_ref_ic_3, 1, 0, false, "vector-ref opt");
  vector_ref_2 = make_function_with_class(sc, f, "vector-ref", g_vector_ref_2, 2, 0, false, "vector-ref opt");
  vector_ref_2_direct = make_function_with_class(sc, f, "vector-ref", g_vector_ref_2_direct, 2, 0, false, "vector-ref opt");

  /* vector-set! */
  f = set_function_chooser(sc, sc->vector_set_symbol, vector_set_chooser);
  vector_set_ic = make_function_with_class(sc, f, "vector-set!", g_vector_set_ic, 3, 0, false, "vector-set! opt");
  vector_set_3 = make_function_with_class(sc, f, "vector-set!", g_vector_set_3, 3, 0, false, "vector-set! opt");

  /* list-set! */
  f = set_function_chooser(sc, sc->list_set_symbol, list_set_chooser);
  list_set_ic = make_function_with_class(sc, f, "list-set!", g_list_set_ic, 3, 0, false, "list-set! opt");

  /* hash-table-ref */
  f = set_function_chooser(sc, sc->hash_table_ref_symbol, hash_table_ref_chooser);
  hash_table_ref_2 = make_function_with_class(sc, f, "hash-table-ref", g_hash_table_ref_2, 2, 0, false, "hash-table-ref opt");
  hash_table_ref_ss = make_function_with_class(sc, f, "hash-table-ref", g_hash_table_ref_ss, 2, 0, false, "hash-table-ref opt");
  hash_table_ref_car = make_function_with_class(sc, f, "hash-table-ref", g_hash_table_ref_car, 2, 0, false, "hash-table-ref opt");

  /* format */
  f = set_function_chooser(sc, sc->format_symbol, format_chooser);
  format_allg = make_function_with_class(sc, f, "format", g_format_allg, 1, 0, true, "format opt");
  format_allg_no_column = make_function_with_class(sc, f, "format", g_format_allg_no_column, 1, 0, true, "format opt");
  format_just_newline = make_function_with_class(sc, f, "format", g_format_just_newline, 2, 0, false, "format opt");

  /* not */
  f = set_function_chooser(sc, sc->not_symbol, not_chooser);
  not_is_pair_s = make_function_with_class(sc, f, "not", g_not_is_pair_s, 1, 0, false, "not opt");
  not_is_null_s = make_function_with_class(sc, f, "not", g_not_is_null_s, 1, 0, false, "not opt");
  not_is_symbol_s = make_function_with_class(sc, f, "not", g_not_is_symbol_s, 1, 0, false, "not opt");
  not_is_number_s = make_function_with_class(sc, f, "not", g_not_is_number_s, 1, 0, false, "not opt");
  not_is_zero_s = make_function_with_class(sc, f, "not", g_not_is_zero_s, 1, 0, false, "not opt");
  not_is_eq_ss = make_function_with_class(sc, f, "not", g_not_is_eq_ss, 1, 0, false, "not opt");
  not_is_eq_sq = make_function_with_class(sc, f, "not", g_not_is_eq_sq, 1, 0, false, "not opt");
  not_is_pair_car_s = make_function_with_class(sc, f, "not", g_not_is_pair_car_s, 1, 0, false, "not opt");
  not_c_c = make_function_with_class(sc, f, "not", g_not_c_c, 1, 0, false, "not opt");

  /* pair? */
  f = set_function_chooser(sc, sc->is_pair_symbol, is_pair_chooser);
  is_pair_car_s = make_function_with_class(sc, f, "pair?", g_is_pair_car_s, 1, 0, false, "pair? opt");
  is_pair_cdr_s = make_function_with_class(sc, f, "pair?", g_is_pair_cdr_s, 1, 0, false, "pair? opt");
  is_pair_cadr_s = make_function_with_class(sc, f, "pair?", g_is_pair_cadr_s, 1, 0, false, "pair? opt");

  /* null? */
  f = set_function_chooser(sc, sc->is_null_symbol, is_null_chooser);
  is_null_cdr = make_function_with_class(sc, f, "null?", g_is_null_cdr, 1, 0, false, "null? opt");

  /* eq? */
  f = set_function_chooser(sc, sc->is_eq_symbol, is_eq_chooser);
  is_eq_car = make_function_with_class(sc, f, "eq?", g_is_eq_car, 2, 0, false, "eq? opt");
  is_eq_car_q = make_function_with_class(sc, f, "eq?", g_is_eq_car_q, 2, 0, false, "eq? opt");
  is_eq_caar_q = make_function_with_class(sc, f, "eq?", g_is_eq_caar_q, 2, 0, false, "eq? opt");

  /* member */
  f = set_function_chooser(sc, sc->member_symbol, member_chooser);
  member_ss = make_function_with_class(sc, f, "member", g_member_ss, 2, 0, false, "member opt");
  member_sq = make_function_with_class(sc, f, "member", g_member_sq, 2, 0, false, "member opt");

  /* memq */
  f = set_function_chooser(sc, sc->memq_symbol, memq_chooser);
  /* is pure-s7, use member here */
  memq_2 = make_function_with_class(sc, f, "memq", g_memq_2, 2, 0, false, "memq opt");
  memq_3 = make_function_with_class(sc, f, "memq", g_memq_3, 2, 0, false, "memq opt");
  memq_4 = make_function_with_class(sc, f, "memq", g_memq_4, 2, 0, false, "memq opt");
  memq_any = make_function_with_class(sc, f, "memq", g_memq_any, 2, 0, false, "memq opt");
  memq_car = make_function_with_class(sc, f, "memq", g_memq_car, 2, 0, false, "memq opt");
  memq_car_2 = make_function_with_class(sc, f, "memq", g_memq_car_2, 2, 0, false, "memq opt");

  /* read-line */
  read_line_uncopied = s7_make_function(sc, "read-line", g_read_line_uncopied, 1, 1, false, "read-line opt");
  s7_function_set_class(read_line_uncopied, slot_value(global_slot(sc->read_line_symbol)));

  /* write-string */
  set_function_chooser(sc, sc->write_string_symbol, write_string_chooser);

  /* eval-string */
  set_function_chooser(sc, sc->eval_string_symbol, eval_string_chooser);
  
  /* inlet */
  f = set_function_chooser(sc, sc->inlet_symbol, inlet_chooser);
  simple_inlet = make_function_with_class(sc, f, "inlet", g_simple_inlet, 0, 0, true, "inlet opt");

  /* let-ref */
  f = set_function_chooser(sc, sc->let_ref_symbol, let_ref_chooser);
  lint_let_ref = make_function_with_class(sc, f, "let-ref", g_lint_let_ref, 2, 0, false, "let-ref opt");
  local_lint_let_ref = make_function_with_class(sc, f, "let-ref", g_local_lint_let_ref, 2, 0, false, "let-ref opt");

  /* let-set */
  f = set_function_chooser(sc, sc->let_set_symbol, let_set_chooser);
  lint_let_set = make_function_with_class(sc, f, "let-set!", g_lint_let_set, 3, 0, false, "let-set! opt");
  local_lint_let_set = make_function_with_class(sc, f, "let-set!", g_local_lint_let_set, 3, 0, false, "let-set! opt");

  or_n = s7_make_function(sc, "or", g_or_n, 0, 0, true, "or opt");
  or_2 = s7_make_function(sc, "or", g_or_2, 2, 0, false, "or opt");
  or_3 = s7_make_function(sc, "or", g_or_3, 3, 0, false, "or opt");
  and_n = s7_make_function(sc, "and", g_and_n, 0, 0, true, "and opt");
  and_2 = s7_make_function(sc, "and", g_and_2, 2, 0, false, "and opt");
  and_3 = s7_make_function(sc, "and", g_and_3, 3, 0, false, "and opt");
  and_sc = s7_make_function(sc, "and", g_and_sc, 2, 0, false, "and opt");
  if_x1 = s7_make_function(sc, "if", g_if_x1, 2, 0, false, "if opt");
  if_x2 = s7_make_function(sc, "if", g_if_x2, 3, 0, false, "if opt");
  if_not_x1 = s7_make_function(sc, "if", g_if_not_x1, 2, 0, false, "if opt");
  if_not_x2 = s7_make_function(sc, "if", g_if_not_x2, 3, 0, false, "if opt");
  if_x_qq = s7_make_function(sc, "if", g_if_x_qq, 3, 0, false, "if opt");
  if_x_qa = s7_make_function(sc, "if", g_if_x_qa, 3, 0, false, "if opt");

  or_s_direct = s7_make_function(sc, "or", g_or_s_direct, 0, 0, true, "or opt");
  and_s_direct = s7_make_function(sc, "and", g_and_s_direct, 0, 0, true, "and opt");
}


#define choose_c_function(Sc, Expr, Func, Args) do {clear_overlay(Expr); set_c_function(Expr, c_function_chooser(Func)(Sc, Func, Args, Expr, true));} while (0)

static opt_t optimize_thunk(s7_scheme *sc, s7_pointer expr, s7_pointer func, int32_t hop)
{
  if (is_immutable_symbol(car(expr)))
    hop = 1;

  if (is_closure(func))
    {
      if (is_null(closure_args(func)))                 /* no rest arg funny business */
	{
	  if (is_safe_closure(func))
	    {
	      s7_pointer body;
	      body = closure_body(func);
	      set_unsafe_optimize_op(expr, hop + OP_SAFE_THUNK);
	      if (is_null(cdr(body)))
		{
		  if (is_optimized(car(body)))
		    set_unsafe_optimize_op(expr, hop + OP_SAFE_THUNK_E);
		  else
		    {
		      if ((is_pair(car(body))) &&
			  (is_syntactic(caar(body))))
			{
			  set_optimize_op(expr, hop + OP_SAFE_THUNK_P);
			  if (typesflag(car(body)) != SYNTACTIC_PAIR)
			    {
			      pair_set_syntax_op(car(body), symbol_syntax_op(caar(body)));
			      set_syntactic_pair(car(body));
			    }
			}
		    }
		}
	    }
	  else set_unsafe_optimize_op(expr, hop + OP_THUNK);
	  set_opt_lambda(expr, func);
	}
      return(OPT_F);                                    /* false (OPT_F) because currently the C_PP stuff assumes safe procedure calls */
    }

  if (is_c_function(func))
    {
      if (c_function_required_args(func) != 0)
	return(OPT_F);

      if ((is_safe_procedure(func)) ||
	  (c_function_call(func) == g_list) ||          /* (list) is safe */
	  (c_function_call(func) == g_values))          /* (values) is safe */
	{
	  set_safe_optimize_op(expr, hop + OP_SAFE_C_C);
	  choose_c_function(sc, expr, func, 0);
	  return(OPT_T);
	}
      return(OPT_F);
    }
  return(OPT_F);
}


static opt_t optimize_func_dotted_args(s7_scheme *sc, s7_pointer expr, s7_pointer func, int32_t hop, int32_t args, int32_t pairs, int32_t symbols, int32_t quotes, int32_t bad_pairs, s7_pointer e)
{
  if (all_x_count(sc, expr) == args)
    {
      set_unsafely_optimized(expr);
      annotate_args(sc, cdr(expr), e);
      set_arglist_length(expr, small_int(args));
      set_optimize_op(expr, hop + OP_CLOSURE_ANY_ALL_X);
      set_opt_lambda(expr, func);
      /* fprintf(stderr, "expr: %s, func: %s, opt: %s\n", DISPLAY(expr), DISPLAY(func), DISPLAY(opt_lambda(expr))); */
      return(OPT_F);
    }
  return(OPT_F);
}

static int32_t combine_ops(s7_scheme *sc, s7_pointer func, s7_pointer expr, combine_op_t cop, s7_pointer e1, s7_pointer e2)
{
  int32_t arg_op;
  s7_pointer arg;

  switch (cop)
    {
    case E_C_P:
      arg_op = op_no_hop(e1);
      switch (arg_op)
	{
	case OP_SAFE_C_S: case OP_SAFE_C_L: case OP_SAFE_CAR_S: case OP_SAFE_CDR_S: case OP_SAFE_CADR_S:
	case OP_SAFE_IS_PAIR_S: case OP_SAFE_IS_NULL_S: case OP_SAFE_IS_SYMBOL_S:
	  return(OP_SAFE_C_opSq);
	case OP_SAFE_C_C:      return(OP_SAFE_C_opCq); /* this includes the multi-arg C_C cases */
	case OP_SAFE_C_LL:
	case OP_SAFE_C_SS:     return(OP_SAFE_C_opSSq);
	case OP_SAFE_C_SQ:     return(OP_SAFE_C_opSQq);
	case OP_SAFE_C_QS:     return(OP_SAFE_C_opQSq);
	case OP_SAFE_C_SC:     return(OP_SAFE_C_opSCq);
	case OP_SAFE_C_CS:     return(OP_SAFE_C_opCSq);
	case OP_SAFE_C_opSq:   return(OP_SAFE_C_op_opSq_q);
	case OP_SAFE_C_S_opSq: return(OP_SAFE_C_op_S_opSq_q);
	case OP_SAFE_C_opSq_S: return(OP_SAFE_C_op_opSq_S_q);
	case OP_SAFE_C_A:      return(OP_SAFE_C_opAq);
	case OP_SAFE_C_AA:     return(OP_SAFE_C_opAAq);
	case OP_SAFE_C_AAA:    return(OP_SAFE_C_opAAAq);
	}
      return(OP_SAFE_C_Z); /* this splits out to A in optimize_func_one_arg */

    case E_C_SP:
      arg = e2;
      arg_op = op_no_hop(arg);
      switch (arg_op)
	{
	case OP_SAFE_C_S: case OP_SAFE_C_L: case OP_SAFE_CAR_S: case OP_SAFE_CDR_S: case OP_SAFE_CADR_S:
	case OP_SAFE_IS_PAIR_S: case OP_SAFE_IS_NULL_S: case OP_SAFE_IS_SYMBOL_S:
	  set_opt_sym1(cdr(expr), cadr(arg));
	  return(OP_SAFE_C_S_opSq);

	case OP_SAFE_C_C:
	  set_opt_pair1(cdr(expr), cdr(arg));
	  return(OP_SAFE_C_S_opCq);

	case OP_SAFE_C_SC:
	  set_opt_sym1(cdr(expr), cadr(arg));
	  set_opt_con2(cdr(expr), caddr(arg));
	  return(OP_SAFE_C_S_opSCq);

	case OP_SAFE_C_CS:	  /* expr is (* a (- 1 b)), e2 is (- 1 b) */
	  set_opt_con1(cdr(expr), cadr(arg));
	  set_opt_sym2(cdr(expr), caddr(arg));
	  return(OP_SAFE_C_S_opCSq);

	case OP_SAFE_C_LL:
	case OP_SAFE_C_SS:	  /* (* a (- b c)) */
	  set_opt_sym1(cdr(expr), cadr(arg));
	  set_opt_sym2(cdr(expr), caddr(arg));
	  return(OP_SAFE_C_S_opSSq);

	case OP_SAFE_C_opSSq_S:	    return(OP_SAFE_C_S_op_opSSq_Sq);
	case OP_SAFE_C_S_opSSq:	    return(OP_SAFE_C_S_op_S_opSSqq);
	case OP_SAFE_C_S_opSq:	    return(OP_SAFE_C_S_op_S_opSqq);
	case OP_SAFE_C_opSSq_opSSq: return(OP_SAFE_C_S_op_opSSq_opSSqq);
	case OP_SAFE_C_SZ:	    return(OP_SAFE_C_S_opSZq);
	case OP_SAFE_C_A:	    return(OP_SAFE_C_S_opAq);
	case OP_SAFE_C_AA:	    return(OP_SAFE_C_S_opAAq);
	case OP_SAFE_C_CSA:
	case OP_SAFE_C_SCA:
	case OP_SAFE_C_SAS:
	case OP_SAFE_C_SSA:
	case OP_SAFE_C_AAA:
	  return(OP_SAFE_C_S_opAAAq);
	}
      return(OP_SAFE_C_SZ);

    case E_C_PS:
      arg = e1;
      arg_op = op_no_hop(arg);
      switch (arg_op)
	{
	case OP_SAFE_C_S: case OP_SAFE_C_L: case OP_SAFE_CAR_S: case OP_SAFE_CDR_S: case OP_SAFE_CADR_S:
	case OP_SAFE_IS_PAIR_S: case OP_SAFE_IS_NULL_S: case OP_SAFE_IS_SYMBOL_S:
	  return(OP_SAFE_C_opSq_S);
	case OP_SAFE_C_LL:
	case OP_SAFE_C_SS:    return(OP_SAFE_C_opSSq_S);
	case OP_SAFE_C_C:     return(OP_SAFE_C_opCq_S);
	case OP_SAFE_C_CS:    return(OP_SAFE_C_opCSq_S);
	case OP_SAFE_C_SC:    return(OP_SAFE_C_opSCq_S);
	case OP_SAFE_C_opSq:  return(OP_SAFE_C_op_opSq_q_S);
	case OP_SAFE_C_opSSq: return(OP_SAFE_C_op_opSSq_q_S);
	}
      return(OP_SAFE_C_ZS);

    case E_C_PC:
      arg = e1;
      arg_op = op_no_hop(arg);
      switch (arg_op)
	{
	case OP_SAFE_C_S: case OP_SAFE_C_L: case OP_SAFE_CAR_S: case OP_SAFE_CDR_S: case OP_SAFE_CADR_S:
	case OP_SAFE_IS_PAIR_S: case OP_SAFE_IS_NULL_S: case OP_SAFE_IS_SYMBOL_S:
	  return(OP_SAFE_C_opSq_C);
	case OP_SAFE_C_C:     return(OP_SAFE_C_opCq_C);
	case OP_SAFE_C_CS:    return(OP_SAFE_C_opCSq_C);
	case OP_SAFE_C_SC:    return(OP_SAFE_C_opSCq_C);
	case OP_SAFE_C_opSq:  return(OP_SAFE_C_op_opSq_q_C);
	case OP_SAFE_C_opSSq: return(OP_SAFE_C_op_opSSq_q_C);
	case OP_SAFE_C_LL:   
	case OP_SAFE_C_SS:    return(OP_SAFE_C_opSSq_C);
	}
      return(OP_SAFE_C_ZC);

    case E_C_CP:
      arg= e2;
      arg_op = op_no_hop(arg);
      switch (arg_op)
	{
	case OP_SAFE_C_C:
	  set_opt_pair1(cdr(expr), cdr(arg));
	  return(OP_SAFE_C_C_opCq);

	case OP_SAFE_C_S: case OP_SAFE_C_L: case OP_SAFE_CAR_S: case OP_SAFE_CDR_S: case OP_SAFE_CADR_S:
	case OP_SAFE_IS_PAIR_S: case OP_SAFE_IS_NULL_S: case OP_SAFE_IS_SYMBOL_S:
	  set_opt_sym1(cdr(expr), cadr(arg));
	  return(OP_SAFE_C_C_opSq);

	case OP_SAFE_C_CS:
	  set_opt_con1(cdr(expr), cadr(arg));
	  set_opt_sym2(cdr(expr), caddr(arg));
	  return(OP_SAFE_C_C_opCSq);

	case OP_SAFE_C_SC:
	  set_opt_sym1(cdr(expr), cadr(arg));
	  set_opt_con2(cdr(expr), caddr(arg));
	  return(OP_SAFE_C_C_opSCq);

	case OP_SAFE_C_LL:
	case OP_SAFE_C_SS:
	    set_opt_sym1(cdr(expr), cadr(arg));
	    /* set_opt_sym2(cdr(expr), caddr(arg)); -- we're clobbering the old C_SS -> C_C_opSSq opt_sym2 setting */
	    if ((is_real(e1)) &&
		(symbol_id(car(arg)) == 0) && 
		(s7_d_pd_function(slot_value(global_slot(car(arg))))))
	      {
		s7_p_dd_t fp;
		fp = s7_p_dd_function(func);
		if (fp) 
		  {
		    /* direct_x_c_c_opssq calls number_to_real on e1 */
		    set_opt_direct_x(cddr(expr), s7_d_pd_function(slot_value(global_slot(car(arg)))));
		    set_opt_direct_x_call(cdr(expr), fp);
		    set_direct_x_opt(expr);
		  }
	      }
	    return(OP_SAFE_C_C_opSSq);

	case OP_SAFE_C_S_opCq:
	  return(OP_SAFE_C_C_op_S_opCqq);
	}
      return(OP_SAFE_C_CZ);

    case E_C_PP:
      arg = e2;
      arg_op = op_no_hop(arg);
      switch (arg_op)
	{
	case OP_SAFE_C_S: case OP_SAFE_C_L: case OP_SAFE_CAR_S: case OP_SAFE_CDR_S: case OP_SAFE_CADR_S:
	case OP_SAFE_IS_PAIR_S: case OP_SAFE_IS_NULL_S: case OP_SAFE_IS_SYMBOL_S:
	  if (is_safe_c_s(e1))
	    {
	      s7_p_dd_t fp;
	      fp = s7_p_dd_function(func);
	      if (fp) 
		{
		  if ((symbol_id(car(e1)) == 0) && (s7_d_p_function(slot_value(global_slot(car(e1))))) &&
		      (symbol_id(car(e2)) == 0) && (s7_d_p_function(slot_value(global_slot(car(e2))))))
		    {
		      set_opt_direct_x(cdr(expr), s7_d_p_function(slot_value(global_slot(car(e1)))));
		      set_opt_direct_x(cddr(expr), s7_d_p_function(slot_value(global_slot(car(e2)))));
		      set_opt_direct_x_call(cdr(expr), fp);
		      set_direct_x_opt(expr);
		    }
		}
	      return(OP_SAFE_C_opSq_opSq);
	    }
	  if ((optimize_op_match(e1, OP_SAFE_C_SS)) || (optimize_op_match(e1, OP_SAFE_C_LL)))
	    return(OP_SAFE_C_opSSq_opSq);
	  if (optimize_op_match(e1, OP_SAFE_C_C))
	    return(OP_SAFE_C_opCq_opSq);
	  break;

	case OP_SAFE_C_C:
	  if (optimize_op_match(e1, OP_SAFE_C_C))
	    return(OP_SAFE_C_opCq_opCq);
	  if ((optimize_op_match(e1, OP_SAFE_C_SS)) || (optimize_op_match(e1, OP_SAFE_C_LL)))
	    return(OP_SAFE_C_opSSq_opCq);
	  if (optimize_op_match(e1, OP_SAFE_C_S))
	    return(OP_SAFE_C_opSq_opCq);
	  break;

	case OP_SAFE_C_SC:
	  if (optimize_op_match(e1, OP_SAFE_C_SC))
	    return(OP_SAFE_C_opSCq_opSCq);
	  break;

	case OP_SAFE_C_LL:
	case OP_SAFE_C_SS:
	  if (optimize_op_match(e1, OP_SAFE_C_C))
	    return(OP_SAFE_C_opCq_opSSq);
	  if ((optimize_op_match(e1, OP_SAFE_C_SS)) || (optimize_op_match(e1, OP_SAFE_C_LL)))
	    return(OP_SAFE_C_opSSq_opSSq);
	  if (is_safe_c_s(e1))
	    return(OP_SAFE_C_opSq_opSSq);
	  break;
	}
      return(OP_SAFE_C_ZZ);

    default:
      break;
    }
  return(OP_NO_OP);
}

static void annotate_args(s7_scheme *sc, s7_pointer args, s7_pointer e)
{
  s7_pointer p;
  for (p = args; is_pair(p); p = cdr(p))
    {
#if DEBUGGING
      s7_function allx;
      allx = all_x_eval(sc, p, e, (s7_is_list(sc, e)) ? pair_symbol_is_safe : let_symbol_is_safe);
      if (!allx)
	{
	  fprintf(stderr, "%s: is_all_x_safe: %d, all_x_eval null?\n", DISPLAY(p), is_all_x_safe(sc, car(p)));
	  abort();
	}
      set_x_call(p, allx);
#else
      set_x_call(p, all_x_eval(sc, p, e, (s7_is_list(sc, e)) ? pair_symbol_is_safe : let_symbol_is_safe));
#endif
    }
}

static void annotate_arg(s7_scheme *sc, s7_pointer arg, s7_pointer e)
{
  /* if sc->envir is sc->nil, we're at the top-level, but the global_slot check should suffice for that */
  set_x_call(arg, all_x_eval(sc, arg, e, (s7_is_list(sc, e)) ? pair_symbol_is_safe : let_symbol_is_safe));
#if DEBUGGING
  if (!c_call(arg))
    abort();
#endif
}


static void opt_generator(s7_scheme *sc, s7_pointer func, s7_pointer expr, int32_t hop)
{
  /* this is an optimization aimed at generators.  So we might as well go all out... */
  if (is_global(car(expr))) /* not a function argument for example */
    {
      s7_pointer body;
      body = closure_body(func);
      if ((s7_list_length(sc, body) == 2) &&
	  (caar(body) == sc->let_set_symbol) &&
	  (is_optimized(car(body))) &&
	  (optimize_op(car(body)) == HOP_SAFE_C_SQS) &&
	  (caadr(body) == sc->with_let_symbol) &&
	  (is_symbol(cadadr(body))))
	{
	  s7_pointer args;
	  args = closure_args(func);
	  if ((cadadr(body) == car(args)) &&
	      (is_pair(cdr(args))) &&
	      (is_pair(cadr(args))) &&
	      (cadddr(car(body)) == caadr(closure_args(func))))
	    {
	      if (is_global(car(expr))) hop = 1; /* it's my party... */
	      set_optimize_op(expr, hop + OP_SAFE_CLOSURE_STAR_S0);
	      set_opt_sym1(cdr(expr), cadr(caddar(body)));
	      set_opt_pair2(cdr(expr), cddadr(body));
	    }
	}
    }
}

static bool is_lambda(s7_scheme *sc, s7_pointer sym)
{
  return((sym == sc->lambda_symbol) && (symbol_id(sym) == 0));
  /* symbol_id==0 means it has never been rebound (T_GLOBAL might not be set for initial stuff) */
}


static opt_t optimize_func_one_arg(s7_scheme *sc, s7_pointer expr, s7_pointer func, int32_t hop, int32_t pairs, int32_t symbols, int32_t quotes, int32_t bad_pairs, s7_pointer e)
{
  s7_pointer arg1;
  /* very often, expr is already optimized, quoted stuff is counted under "bad_pairs"! as well as quotes */
  /* fprintf(stderr, "opt 1: %s, hop: %d, pairs: %d, symbols: %d, quotes: %d, bad: %d\n", DISPLAY_80(expr), hop, pairs, symbols, quotes, bad_pairs); */

  arg1 = cadr(expr);
  if ((bad_pairs == quotes) &&
      (is_immutable_symbol(car(expr))))
    hop = 1;

  if ((symbols == 1) &&
      (!is_keyword(arg1)) &&
      (!pair_symbol_is_safe(sc, arg1, e)) &&
      (!is_slot(find_symbol(sc, arg1))))
    {
      /* wrap the bad arg in a check symbol lookup */
      if (s7_is_aritable(sc, func, 1))
	{
	  set_x_call_direct(cdr(expr), all_x_unsafe_s); /* was set_c_call 21-May-17 */
	  set_arglist_length(expr, small_int(1));
	  if (is_c_function(func))
	    {
	      set_safe_optimize_op(expr, hop + ((is_safe_procedure(func)) ? OP_SAFE_C_A : OP_C_A));
	      clear_overlay(expr);
	      set_c_function(expr, func);
	      return(OPT_T);
	    }
	  if ((is_closure(func)) &&
	      (!arglist_has_rest(sc, closure_args(func))))
	    {
	      set_unsafely_optimized(expr);
	      /* fprintf(stderr, "%d: %s %d\n", __LINE__, DISPLAY(func), is_local_symbol(expr)); */
	      set_optimize_op(expr, hop + ((is_safe_closure(func)) ? OP_SAFE_CLOSURE_A : OP_CLOSURE_A));
	      set_opt_lambda(expr, func);
	      return(OPT_F);
	    }
	  if ((is_closure_star(func)) &&
	      (has_simple_arg_defaults(closure_body(func))) &&
	      (closure_star_arity_to_int(sc, func) >= 1) &&
	      (!arglist_has_rest(sc, closure_args(func))))
	    {
	      set_unsafely_optimized(expr);
	      set_optimize_op(expr, ((is_safe_closure(func)) ? OP_SAFE_CLOSURE_STAR_A : OP_CLOSURE_STAR_A));
	      set_opt_lambda(expr, func);
	    }
	}
      return(OPT_F);
    }

  if (((is_c_function(func)) &&
       (c_function_required_args(func) <= 1) &&
       (c_function_all_args(func) >= 1)) ||
      ((is_c_function_star(func)) &&
       (c_function_all_args(func) == 1))) /* surely no need to check key here? */
    {
      bool func_is_safe;
      func_is_safe = is_safe_procedure(func);

      if (pairs == 0)
	{
	  if (func_is_safe)                  /* safe c function */
	    {
	      set_safe_optimize_op(expr, hop + ((symbols == 0) ? OP_SAFE_C_C : ((is_local_symbol(cdr(expr))) ? OP_SAFE_C_L : OP_SAFE_C_S)));
	      choose_c_function(sc, expr, func, 1);

	      /* these are border-line useless -- lint uses cxr_s */
	      if (symbols == 1)
		{
		  if (c_call(expr) == g_car)
		    add_optimizer_fixup(sc, expr, (uint32_t)(hop + OP_SAFE_CAR_S));
		  if (c_call(expr) == g_cdr)
		    add_optimizer_fixup(sc, expr, (uint32_t)(hop + OP_SAFE_CDR_S));
		  if (c_call(expr) == g_cadr)
		    add_optimizer_fixup(sc, expr, (uint32_t)(hop + OP_SAFE_CADR_S));
		  if (c_call(expr) == g_is_pair)
		    add_optimizer_fixup(sc, expr, (uint32_t)(hop + OP_SAFE_IS_PAIR_S));
		  if (c_call(expr) == g_is_null)
		    add_optimizer_fixup(sc, expr, (uint32_t)(hop + OP_SAFE_IS_NULL_S));
		  if (c_call(expr) == g_is_symbol)
		    add_optimizer_fixup(sc, expr, (uint32_t)(hop + OP_SAFE_IS_SYMBOL_S));
		}
	      return(OPT_T);
	    }
	  else                               /* c function is not safe */
	    {
	      if (symbols == 0)
		{
		  set_unsafely_optimized(expr);
		  set_optimize_op(expr, hop + OP_C_A);
		  annotate_arg(sc, cdr(expr), e);
		  set_arglist_length(expr, small_int(1));
		}
	      else
		{
		  set_unsafely_optimized(expr);
		  if (c_function_call(func) == g_read)
		    set_optimize_op(expr, hop + OP_READ_S);
		  else set_optimize_op(expr, hop + OP_C_S);
		}
	      choose_c_function(sc, expr, func, 1);
	      return(OPT_F);
	    }
	}
      else /* pairs == 1 */
	{
	  if (bad_pairs == 0)
	    {
	      if (func_is_safe)
		{
		  int32_t op;
		  op = combine_ops(sc, func, expr, E_C_P, arg1, NULL);
		  set_safe_optimize_op(expr, hop + op);
		  /* fallback is Z */
		  if (!hop)
		    {
		      clear_hop(arg1);
		    }
		  else
		    {
		      if ((op == OP_SAFE_C_Z) &&
			  (is_all_x_op(optimize_op(arg1))))
			{
			  /* this is confusing!  this is much faster than safe_c_z, but
			   *   the parallel let_z|a case seems to claim that z is faster.
			   */
			  set_optimize_op(expr, hop + OP_SAFE_C_A);
			  annotate_arg(sc, cdr(expr), e);
			}
		    }
		  choose_c_function(sc, expr, func, 1);
		  return(OPT_T);
		}
	      if (is_all_x_op(optimize_op(arg1)))
		{
		  set_unsafe_optimize_op(expr, hop + OP_C_A);
		  annotate_arg(sc, cdr(expr), e);
		  set_arglist_length(expr, small_int(1));
		  choose_c_function(sc, expr, func, 1);
		  return(OPT_F);
		}
	    }
	  else /* bad_pairs == 1 */
	    {
	      if (quotes == 1)
		{
		  annotate_arg(sc, cdr(expr), e);
		  set_arglist_length(expr, small_int(1));
                 if (func_is_safe)
                   {
                     set_safe_optimize_op(expr, hop + OP_SAFE_C_A);
                     choose_c_function(sc, expr, func, 1);
                     return(OPT_T);
                   }
		  set_unsafe_optimize_op(expr, hop + OP_C_A);
		  choose_c_function(sc, expr, func, 1);
		  return(OPT_F);
		}
	      else  /* quotes == 0 */
		{
		  if (!func_is_safe)
		    {
		      s7_pointer lambda_expr;
		      lambda_expr = arg1;
		      if ((is_pair(lambda_expr)) &&
			  (is_lambda(sc, car(lambda_expr))) && /* check for stuff like (define (f) (eval (lambda 2))) */
			  (is_pair(cdr(lambda_expr))) &&
			  (is_pair(cddr(lambda_expr))))
			{
			  if ((c_function_call(func) == g_call_with_exit) &&
			      (is_pair(cadr(lambda_expr))) &&
			      (is_null(cdadr(lambda_expr))) &&
			      (is_symbol(caadr(lambda_expr))))
			    {
			      set_unsafe_optimize_op(expr, hop + OP_CALL_WITH_EXIT);
			      choose_c_function(sc, expr, func, 1);
			      set_opt_pair2(expr, cdr(lambda_expr));
			      return(OPT_F);
			    }
			}
		      set_unsafe_optimize_op(expr, hop + ((is_optimized(arg1)) ? OP_C_Z : OP_C_P));
		      choose_c_function(sc, expr, func, 1);
		      return(OPT_F);
		    }
		  else
		    {
		      set_unsafe_optimize_op(expr, hop + ((car(expr) == sc->not_symbol) ? OP_NOT_P : OP_SAFE_C_P));
		      choose_c_function(sc, expr, func, 1);
		      return(OPT_F);
		    }
		}
	    }
	}
      
      if (func_is_safe)
	set_unsafe_optimize_op(expr, hop + ((is_optimized(arg1)) ? OP_SAFE_C_Z : ((car(expr) == sc->not_symbol) ? OP_NOT_P : OP_SAFE_C_P)));
      else set_unsafe_optimize_op(expr, hop + ((is_optimized(arg1)) ? OP_C_Z : OP_C_P));
      choose_c_function(sc, expr, func, 1);
      return(OPT_F);

      return((is_optimized(expr)) ? OPT_T : OPT_F);
    }

  if (is_closure(func))
    {
      bool safe_case;
      s7_pointer body;
      int32_t arit;
      arit = closure_arity_to_int(sc, func);
      if (arit != 1)
	{
	  if ((arit == -1) &&
	      (is_symbol(closure_args(func))))
	    return(optimize_func_dotted_args(sc, expr, func, hop, 1, pairs, symbols, quotes, bad_pairs, e));
	  return(OPT_F);
	}

      safe_case = is_safe_closure(func);
      body = closure_body(func);

      if (is_immutable(func)) hop = 1;
      if (pairs == 0)
	{
	  if (symbols == 1)
	    {
	      if (safe_case)
		{
		  set_optimize_op(expr, hop + OP_SAFE_CLOSURE_S);
		  if (is_null(cdr(body)))
		    {
		      s7_pointer bexpr;
		      bexpr = car(body);
		      if ((is_pair(bexpr)) &&
			  (is_h_safe_c_c(bexpr)))
			set_optimize_op(expr, hop + OP_SAFE_CLOSURE_S_C);
		      else
			{
			  if ((!is_optimized(bexpr)) &&
			      (is_pair(bexpr)) &&
			      (is_syntactic(car(bexpr))))
			    {
			      set_optimize_op(expr, hop + OP_SAFE_CLOSURE_S_P);
			      if (typesflag(bexpr) != SYNTACTIC_PAIR)
				{
				  pair_set_syntax_op(bexpr, symbol_syntax_op(car(bexpr)));
				  set_syntactic_pair(bexpr);
				}
			    }
			}
		    }
		}
	      else set_optimize_op(expr, hop + OP_CLOSURE_S); 
	      set_opt_sym2(expr, arg1);
	    }
	  else 
	    {
	      set_optimize_op(expr, hop + ((safe_case) ? OP_SAFE_CLOSURE_C : OP_CLOSURE_C));
	      set_opt_con2(expr, arg1);
	    }
	  set_opt_lambda(expr, func);
	  set_unsafely_optimized(expr);
	  /* fprintf(stderr, "%s: %s\n", DISPLAY_80(expr), opt_names[optimize_op(expr)]); */
	  return(OPT_F);
	}
      else /* pairs == 1 */
	{
	  if (bad_pairs == quotes)
	    {
	      if ((is_optimized(arg1)) &&
		  (is_all_x_op(optimize_op(arg1))))
		{
		  set_unsafely_optimized(expr);
		  annotate_arg(sc, cdr(expr), e);
		  set_arglist_length(expr, small_int(1));
		  if (safe_case)
		    {
		      /* fprintf(stderr, "%d: %s %d\n", __LINE__, DISPLAY(car(expr)), is_local_symbol(expr)); */
		      set_optimize_op(expr, hop + OP_SAFE_CLOSURE_A);
		      if ((is_pair(car(body))) &&
			  (is_null(cdr(body))) &&
			  (is_h_safe_c_c(car(body))))
			set_optimize_op(expr, hop + OP_SAFE_CLOSURE_A_C);
		    }
		  else 
		    {
		      set_optimize_op(expr, hop + OP_CLOSURE_A);
		      if ((is_null(cdr(body))) &&
			  (is_pair(car(body))) &&
			  (is_syntactic(caar(body))))
			{
			  set_optimize_op(expr, hop + OP_CLOSURE_A_P);
			  if (typesflag(car(body)) != SYNTACTIC_PAIR)
			    {
			      pair_set_syntax_op(car(body), symbol_syntax_op(caar(body)));
			      set_syntactic_pair(car(body));
			    }
			}
		    }
		  set_opt_lambda(expr, func);
		  return(OPT_F);
		}
	    }
	}

      if (all_x_count(sc, expr) == 1)
	{
	  set_unsafely_optimized(expr);
	  if ((safe_case) &&
	      (is_pair(car(body))) &&
	      (is_h_safe_c_c(car(body))))
	    set_optimize_op(expr, hop + OP_SAFE_CLOSURE_A_C);
	  else set_unsafe_optimize_op(expr, hop + ((safe_case ? OP_SAFE_CLOSURE_A : OP_CLOSURE_A)));
	  annotate_arg(sc, cdr(expr), e);
	  set_opt_lambda(expr, func);
	  set_arglist_length(expr, small_int(1));
	  return(OPT_F);
	}

      set_optimize_op(expr, hop + ((safe_case) ? OP_SAFE_CLOSURE_P : OP_CLOSURE_P));
      set_opt_lambda(expr, func);
      set_unsafely_optimized(expr);
      return(OPT_F);  /* don't check is_optimized here for OPT_T */
    }

  if (is_closure_star(func))
    {
      bool safe_case;
#if 0
      fprintf(stderr, "%s: %d %d %d %d %d\n", DISPLAY(expr), 
	      has_simple_arg_defaults(closure_body(func)), 
	      is_null(closure_args(func)), 
	      is_safe_closure(func),
	      arglist_has_rest(sc, closure_args(func)), 
	      all_x_count(sc, expr));
#endif
      if ((!has_simple_arg_defaults(closure_body(func))) ||
	  (is_null(closure_args(func))))
	return(OPT_F);
      safe_case = is_safe_closure(func);

      if (symbols == 1)
	{
	  set_unsafely_optimized(expr);
	  annotate_arg(sc, cdr(expr), e);
	  set_arglist_length(expr, small_int(1));
	  if (safe_case)
	    {
	      set_optimize_op(expr, hop + OP_SAFE_CLOSURE_STAR_A);
	      if (closure_star_arity_to_int(sc, func) == 2)
		{
		  s7_pointer defarg2;
		  defarg2 = cadr(closure_args(func));
		  if ((is_pair(defarg2)) &&
		      (s7_is_zero(cadr(defarg2))))
		    opt_generator(sc, func, expr, hop);
		}
	    }
	  else set_optimize_op(expr, hop + OP_CLOSURE_STAR_A);
	  set_opt_lambda(expr, func);
	  /* set_opt_sym2(expr, arg1); */
	  return(OPT_F);
	}

      if ((!arglist_has_rest(sc, closure_args(func))) &&
	  (all_x_count(sc, expr) == 1))
	{
	  set_unsafe_optimize_op(expr, hop + ((safe_case) ? OP_SAFE_CLOSURE_STAR_A : OP_CLOSURE_STAR_A));
	  annotate_arg(sc, cdr(expr), e);
	  set_opt_lambda(expr, func);
	  set_arglist_length(expr, small_int(1));
	  return(OPT_F);
	}
      return((is_optimized(expr)) ? OPT_T : OPT_F);
    }

  if ((is_c_function_star(func)) &&
      (all_x_count(sc, expr) == 1))
    {
      set_optimized(expr);
      set_optimize_op(expr, hop + OP_SAFE_C_STAR_A);
      annotate_args(sc, cdr(expr), e);
      set_arglist_length(expr, small_int(1));
      clear_overlay(expr);
      set_c_function(expr, func);
      return(OPT_T);
    }

  if ((s7_is_vector(func)) &&
      (is_all_x_safe(sc, arg1)))
    {
      if ((is_immutable_symbol(car(expr))) &&
	  (vector_rank(func) == 1) &&
	  (is_normal_vector(func)))
	set_unsafe_optimize_op(expr, hop + OP_CVECTOR_A);
      else set_unsafe_optimize_op(expr, hop + OP_VECTOR_A);
      annotate_arg(sc, cdr(expr), e);
      set_arglist_length(expr, small_int(1));
      return(OPT_T);
    }
  /* unknown_* is set later */
  return((is_optimized(expr)) ? OPT_T : OPT_F);
}


static bool let_memq(s7_scheme *sc, s7_pointer symbol, s7_pointer symbols)
{
  s7_pointer x;
  for (x = symbols; is_pair(x); x = cdr(x))
    {
      if (car(x) == symbol)
	return(true);
    }
  return(false);
}


static s7_pointer find_uncomplicated_symbol(s7_scheme *sc, s7_pointer symbol, s7_pointer e)
{
  s7_pointer x;
  int64_t id;

  if ((symbol_is_in_list(sc, symbol)) &&
      (let_memq(sc, symbol, e)))   /* it's probably a local variable reference */
    return(sc->nil);

  if (is_global(symbol))
    return(global_slot(symbol));

  /* see 59108 (OP_DEFINE_* in optimize_syntax) -- keyword version of name is used if a definition is
   *   contingent on some run-time decision, so we're looking here for local defines that might not happen.
   *   s7test.scm has a test case using acos.
   */
  if ((has_keyword(symbol)) &&
      (symbol_is_in_list(sc, symbol_to_keyword(sc, symbol))))
    return(sc->nil);

  id = symbol_id(symbol);
  for (x = sc->envir; id < let_id(x); x = outlet(x));
  for (; is_let(x); x = outlet(x))
    {
      s7_pointer y;
      if (let_id(x) == id)
	return(local_slot(symbol));

      for (y = let_slots(x); is_slot(y); y = next_slot(y))
	if (slot_symbol(y) == symbol)
	  return(y);
    }

  return(global_slot(symbol)); /* it's no longer global perhaps (local definition now inaccessible) */
}


static bool unsafe_is_safe(s7_scheme *sc, s7_pointer func, s7_pointer arg1, s7_pointer arg2, s7_pointer arg3, s7_pointer e)
{
  s7_pointer f = NULL;                     /* arg3 if member|assoc */
  if (!arg3) return(true);
  f = arg3;
  if (!is_symbol(f)) return(false);
  f = find_uncomplicated_symbol(sc, f, e); /* how to catch local c-funcs here? */
  if (is_slot(f))
    {
      f = slot_value(f);
      return((is_c_function(f)) &&
	     (is_safe_procedure(f)));
    }
  return(false);
}

static void check_lambda(s7_scheme *sc);

static opt_t optimize_func_two_args(s7_scheme *sc, s7_pointer expr, s7_pointer func, int32_t hop, int32_t pairs, int32_t symbols, int32_t quotes, int32_t bad_pairs, s7_pointer e)
{
  s7_pointer arg1, arg2;

  arg1 = cadr(expr);
  arg2 = caddr(expr);
  if ((bad_pairs == quotes) &&
      (is_immutable_symbol(car(expr))))
    hop = 1;

  if (((is_symbol(arg1)) &&
       (!pair_symbol_is_safe(sc, arg1, e)) &&
       (!is_slot(find_symbol(sc, arg1)))) ||
      ((is_symbol(arg2)) &&
       (!pair_symbol_is_safe(sc, arg2, e)) &&
       (!is_slot(find_symbol(sc, arg2)))))
    {
      /* wrap bad args */
      if ((is_all_x_safe(sc, arg1)) &&
	  (is_all_x_safe(sc, arg2)) &&
	  (s7_is_aritable(sc, func, 2)))
	{
	  annotate_args(sc, cdr(expr), e);
	  set_arglist_length(expr, small_int(2));
	  if (is_c_function(func))
	    {
	      set_safe_optimize_op(expr, hop + ((is_safe_procedure(func)) ? OP_SAFE_C_AA : OP_C_AA));
	      set_c_function(expr, func);
	      return(OPT_T);
	    }
	  if ((is_closure(func)) &&
	      (!arglist_has_rest(sc, closure_args(func))))
	    {
	      set_unsafely_optimized(expr);
	      set_optimize_op(expr, hop + ((is_safe_closure(func)) ? OP_SAFE_CLOSURE_AA : OP_CLOSURE_AA));
	      set_opt_lambda(expr, func);
	      return(OPT_F);
	    }
	  if ((is_closure_star(func)) &&
	      (has_simple_arg_defaults(closure_body(func))) &&
	      (!arglist_has_rest(sc, closure_args(func))))
	    {
	      set_unsafely_optimized(expr);
	      if (closure_star_arity_to_int(sc, func) == 2)
		set_optimize_op(expr, ((is_safe_closure(func)) ? OP_SAFE_CLOSURE_STAR_AA : OP_CLOSURE_STAR_AA));
	      else set_optimize_op(expr, ((is_safe_closure(func)) ? OP_SAFE_CLOSURE_STAR_ALL_X : OP_CLOSURE_STAR_ALL_X));
	      set_opt_lambda(expr, func);
	    }
	}
      return(OPT_F);
    }
  /* end of bad symbol wrappers */

  if ((is_c_function(func) &&
       (c_function_required_args(func) <= 2) &&
       (c_function_all_args(func) >= 2)) ||
      ((is_c_function_star(func)) &&
       (c_function_all_args(func) == 2) &&
       (!is_keyword(arg1))))
    {
      /* this is a mess */
      bool func_is_safe;
      func_is_safe = is_safe_procedure(func);
      if (pairs == 0)
	{
	  if ((func_is_safe) || (is_maybe_safe(func)))
	    {
	      /* another case here: set-car! and set-cdr! are safe if symbols==1 and arg1 is the symbol (i.e. arg2 is a constant) */
	      if (symbols == 0)
		set_optimize_op(expr, hop + OP_SAFE_C_C);
	      else
		{
		  if (symbols == 2) /* these two symbols are almost never the same, (sqrt (+ (* x x) (* y y))) */
		    {
		      if ((is_local_symbol(cdr(expr))) &&
			  (is_local_symbol(cddr(expr))))
			set_optimize_op(expr, hop + OP_SAFE_C_LL); 
		      else set_optimize_op(expr, hop + OP_SAFE_C_SS); 
		    }
		  else set_optimize_op(expr, hop + ((is_symbol(arg1)) ? OP_SAFE_C_SC : OP_SAFE_C_CS));
		}
	      set_optimized(expr);
	      choose_c_function(sc, expr, func, 2);
	      return(OPT_T);
	    }
	  set_unsafely_optimized(expr);
	  if (symbols == 2)
	    {
	      if (c_function_call(func) == g_apply)
		{
		  set_optimize_op(expr, hop + OP_APPLY_SS);
		  set_opt_cfunc(expr, func); /* not quite set_c_function */
		  set_opt_sym2(expr, arg2);
		}
	      else
		{
		  set_optimize_op(expr, hop + OP_C_SS);
		  choose_c_function(sc, expr, func, 2);
		}
	    }
	  else
	    {
	      set_optimize_op(expr, hop + OP_C_AA);
	      annotate_args(sc, cdr(expr), e);
	      set_arglist_length(expr, small_int(2));
	      choose_c_function(sc, expr, func, 2);
	      if (is_safe_procedure(opt_cfunc(expr)))
		{
		  clear_unsafe(expr);
		  /* symbols can be 0..2 here, no pairs */
		  set_optimized(expr);
		  if (symbols == 1)
		    {
		      if (is_symbol(arg1))
			set_optimize_op(expr, hop + OP_SAFE_C_SC);
		      else set_optimize_op(expr, hop + OP_SAFE_C_CS);
		    }
		  else
		    {
		      if (symbols == 2)
			{
			  if ((is_local_symbol(cdr(expr))) &&
			      (is_local_symbol(cddr(expr))))
			    set_optimize_op(expr, hop + OP_SAFE_C_LL);
			  else set_optimize_op(expr, hop + OP_SAFE_C_SS);
			}
		      else set_optimize_op(expr, hop + OP_SAFE_C_C);
		    }
		  return(OPT_T);
		}
	    }
	  return(OPT_F);
	}

      /* pairs != 0 */
      if ((bad_pairs == 0) &&
	  (pairs == 2))
	{
	  if ((func_is_safe) || (is_maybe_safe(func)))
	    {
	      int32_t op;
	      op = combine_ops(sc, func, expr, E_C_PP, arg1, arg2);
	      set_safe_optimize_op(expr, hop + op);
	      /* fallback here is ZZ */
	      if (!hop)
		{
		  clear_hop(arg1);
		  clear_hop(arg2);
		}
	      else
		{
		  if (op == OP_SAFE_C_ZZ)
		    {
		      if (is_all_x_safe(sc, arg1))
			{
			  if (is_all_x_safe(sc, arg2))
			    {
			      set_optimize_op(expr, hop + OP_SAFE_C_AA);
			      annotate_args(sc, cdr(expr), e);
			      set_arglist_length(expr, small_int(2));
			    }
			  else
			    {
			      if (optimize_op(arg1) == HOP_SAFE_C_C)
				set_optimize_op(expr, hop + OP_SAFE_C_opCq_Z);
			      else
				{
				  set_optimize_op(expr, hop + OP_SAFE_C_AZ);
				  annotate_arg(sc, cdr(expr), e);
				  set_arglist_length(expr, small_int(2));
				}
			    }
			}
		      else
			{
			  if (is_all_x_safe(sc, arg2))
			    {
			      set_optimize_op(expr, hop + OP_SAFE_C_ZA);
			      annotate_arg(sc, cddr(expr), e);
			      set_arglist_length(expr, small_int(2));
			    }
			}
		    }
		}
	      choose_c_function(sc, expr, func, 2); /* this might change the op to safe_c_c, so it has to be last */
	      return(OPT_T);
	    }
	}

      if ((bad_pairs == 0) &&
	  (pairs == 1))
	{
	  if ((func_is_safe) || (is_maybe_safe(func)))
	    {
	      combine_op_t orig_op;
	      int32_t op;

	      if (is_pair(arg1))
		{
		  if (is_symbol(arg2))
		    orig_op = E_C_PS;
		  else orig_op = E_C_PC;
		  op = combine_ops(sc, func, expr, orig_op, arg1, arg2);
		  if (!hop) clear_hop(arg1);
		}
	      else
		{
		  if (is_symbol(arg1))
		    orig_op = E_C_SP;
		  else orig_op = E_C_CP;
		  op = combine_ops(sc, func, expr, orig_op, arg1, arg2);
		  if (!hop) clear_hop(arg2);
		}

	      if ((((op == OP_SAFE_C_SZ) || (op == OP_SAFE_C_CZ)) &&
		   (is_all_x_op(optimize_op(arg2)))) ||			   
		  (((op == OP_SAFE_C_ZS) || (op == OP_SAFE_C_ZC)) &&
		   (is_all_x_op(optimize_op(arg1)))))
		{
		  set_safe_optimize_op(expr, hop + OP_SAFE_C_AA);
		  annotate_arg(sc, cdr(expr), e);
		  annotate_arg(sc, cddr(expr), e);
		}
	      else set_safe_optimize_op(expr, hop + op);
	      choose_c_function(sc, expr, func, 2);
	      return(OPT_T);
	    }

	  if (symbols == 1)
	    {
	      if (is_symbol(arg1))
		{
		  if (is_safe_c_s(arg2))
		    {
		      set_unsafe_optimize_op(expr, hop + OP_C_S_opSq);
		      set_opt_sym1(cdr(expr), cadr(arg2));
		      choose_c_function(sc, expr, func, 2);
		      return(OPT_F);
		    }
		  if (optimize_op_match(arg2, OP_SAFE_C_C))
		    {
		      set_unsafe_optimize_op(expr, hop + OP_C_S_opCq);
		      set_opt_pair1(cdr(expr), cdr(arg2));
		      choose_c_function(sc, expr, func, 2);
		      return(OPT_F);
		    }
		}
	    }
	}

      if ((bad_pairs == 1) && (quotes == 1))
	{
	  if ((func_is_safe) || (is_maybe_safe(func)))
	    {
	      if  (symbols == 1)
		{
		  set_optimized(expr);
		  if (is_symbol(arg1))
		    {
		      set_opt_con1(cdr(expr), cadr(caddr(expr)));
		      set_optimize_op(expr, hop + OP_SAFE_C_SQ);
		    }
		  else set_optimize_op(expr, hop + OP_SAFE_C_QS);
		  choose_c_function(sc, expr, func, 2);
		  return(OPT_T);
		}
	      else
		{
		  if (pairs == 1)
		    {
		      /* Q must be 1, symbols = 0, pairs = 1 (the quote), so this must be CQ or QC? */
		      set_optimized(expr);
		      if (is_pair(arg1))
			set_optimize_op(expr, hop + OP_SAFE_C_QC);
		      else set_optimize_op(expr, hop + OP_SAFE_C_CQ);
		      choose_c_function(sc, expr, func, 2);
		      return(OPT_T);
		    }
		}
	    }
	  else
	    {
	      if (pairs == 1)
		{
		  set_unsafe_optimize_op(expr, hop + OP_C_AA);
		  annotate_args(sc, cdr(expr), e);
		  set_arglist_length(expr, small_int(2));
		  choose_c_function(sc, expr, func, 2);
		  return(OPT_F);
		}
	    }
	}

      if (quotes == 2)
	{
	  if ((func_is_safe) || (is_maybe_safe(func)))
	    {
	      set_safe_optimize_op(expr, hop + OP_SAFE_C_QQ);
	      choose_c_function(sc, expr, func, 2);
	      return(OPT_T);
	    }
	  set_unsafe_optimize_op(expr, hop + OP_C_AA);
	  annotate_args(sc, cdr(expr), e);
	  set_arglist_length(expr, small_int(2));
	  choose_c_function(sc, expr, func, 2);
	  return(OPT_F);
	}

      if ((pairs == 1) &&
	  (quotes == 0) &&
	  ((func_is_safe) || (is_maybe_safe(func))))
	{
	  if (symbols == 1)
	    {
	      set_optimized(expr);
	      if (is_symbol(arg1))
		{
		  if ((bad_pairs == 0) || (is_optimized(arg2))) /* bad_pair && h_optimized happens a lot */
		    {
		      set_optimize_op(expr, hop + OP_SAFE_C_SZ);
		      choose_c_function(sc, expr, func, 2);
		      /* if hop is on, is it the case that opt1 is unused?  where besides c_function_is_ok is it referenced?
		       *    some like add_ss_1ss use opt1(cdr(...)) which is safe here I think because cadr is a symbol
		       *    it's used in the choosers to detect e.g. temp funcs
		       */
		      return(OPT_T);
		    }
		  set_unsafe(expr);
		  set_optimize_op(expr, hop + OP_SAFE_C_SP);
		  choose_c_function(sc, expr, func, 2);
		  return(OPT_F);
		}

	      /* arg2 is a symbol */
	      if ((bad_pairs == 0) || (is_optimized(arg1)))
		{
		  set_optimize_op(expr, hop + OP_SAFE_C_ZS);
		  choose_c_function(sc, expr, func, 2);
		  return(OPT_T);
		}
	      /* unknowns get here: (* amp (amps 0))
	       * also list: (make-polywave pitch (list 1 0.93 2 0.07))
	       * and (* vol (granulate gen))
	       */
	      set_unsafe(expr);
	      set_optimize_op(expr, hop + OP_SAFE_C_PS);
	      choose_c_function(sc, expr, func, 2);
	      return(OPT_F);
	    }
	  if (symbols == 0)
	    {
	      set_optimized(expr);
	      if (is_pair(arg1))
		{
		  if ((bad_pairs == 0) || (is_optimized(arg2)))
		    {
		      set_optimize_op(expr, hop + OP_SAFE_C_ZC);
		      choose_c_function(sc, expr, func, 2);
		      return(OPT_T);
		    }
		  else
		    {
		      set_unsafe(expr);
		      set_optimize_op(expr, hop + OP_SAFE_C_PC);
		      choose_c_function(sc, expr, func, 2);
		      return(OPT_F);
		    }
		}
	      else
		{
		  if ((bad_pairs == 0) || (is_optimized(arg1)))
		    {
		      set_optimize_op(expr, hop + OP_SAFE_C_CZ);
		      choose_c_function(sc, expr, func, 2);
		      return(OPT_T);
		    }
		  else
		    {
		      set_unsafe(expr);
		      set_optimize_op(expr, hop + OP_SAFE_C_CP);
		      choose_c_function(sc, expr, func, 2);
		      return(OPT_F);
		    }
		}
	    }
	}

      if ((pairs == 2) &&
	  ((func_is_safe) || (is_maybe_safe(func))))
	{
	  if ((bad_pairs == 1) &&
	      (is_safe_c_s(arg1)))
	    {
	      /* unsafe func here won't work unless we check that later and make the new arg list (for list-values etc)
	       *   (and it has to be the last pair else the unknown_g stuff can mess up)
	       */
	      if (car(arg2) == sc->quote_symbol)
		{
		  if (!is_pair(cdr(arg2)))
		    return(OPT_OOPS);
		  set_safe_optimize_op(expr, hop + OP_SAFE_C_opSq_Q);
		  set_opt_con1(cdr(expr), cadadr(expr));
		  set_opt_con2(cdr(expr), cadr(caddr(expr)));
		  choose_c_function(sc, expr, func, 2);
		  return(OPT_T);
		}
	      set_unsafe_optimize_op(expr, hop + OP_SAFE_C_opSq_P);
	      choose_c_function(sc, expr, func, 2);
	      return(OPT_F);
	    }
	  else
	    {
	      if (quotes == 0)
		{
		  set_unsafely_optimized(expr);
		  if (is_all_x_safe(sc, arg1))
		    {
		      set_optimize_op(expr, hop + ((is_optimized(arg2)) ? OP_SAFE_C_AZ : OP_SAFE_C_AP));
		      annotate_arg(sc, cdr(expr), e);
		    }
		  else set_optimize_op(expr, hop + OP_SAFE_C_PP);
		  choose_c_function(sc, expr, func, 2);
		  return(OPT_F);
		}
	      else
		{
		  if (quotes == 1)
		    {
		      if (car(arg1) == sc->quote_symbol)
			set_optimize_op(expr, hop + OP_SAFE_C_QP);
		      else 
			{
			  set_optimize_op(expr, hop + ((is_optimized(arg1)) ? OP_SAFE_C_ZQ : OP_SAFE_C_PQ));
			  set_opt_con2(cdr(expr), cadr(caddr(expr)));
			}
		      set_unsafely_optimized(expr);
		      choose_c_function(sc, expr, func, 2);
		      return(OPT_F);
		    }
		}
	    }
	}

      if (func_is_safe)
	{
	  if (all_x_count(sc, expr) == 2)
	    {
	      set_safe_optimize_op(expr, hop + OP_SAFE_C_AA);
	      annotate_args(sc, cdr(expr), e);
	      set_arglist_length(expr, small_int(2));
	      choose_c_function(sc, expr, func, 2);
	      return(OPT_T);
	    }
	}

      if ((pairs == 1) &&
	  (symbols == 1) &&
	  (quotes == 0) &&
	  (!func_is_safe) &&
	  ((!is_pair(arg1)) ||
	   ((is_optimized(arg1)) &&
	    (is_all_x_op(optimize_op(arg1))))))
	{
	  annotate_arg(sc, cdr(expr), e);
	  set_unsafe_optimize_op(expr, hop + OP_C_AP);
	  choose_c_function(sc, expr, func, 2);
	  return(OPT_F);
	}

      if ((!func_is_safe) &&
	  ((!is_pair(arg2)) ||
	   ((is_optimized(arg2)) &&
	    (is_all_x_op(optimize_op(arg2))))))
	{
	  if ((is_pair(arg1)) &&
	      (car(arg1) == sc->lambda_symbol))
	    {
	      s7_pointer code;
	      annotate_arg(sc, cddr(expr), e);
	      set_unsafe_optimize_op(expr, hop + OP_C_FA);
	      code = sc->code;               /* save old -- not of direct interest here -- just avoiding unexpected clobberage */
	      sc->code = cdadr(expr);
	      check_lambda(sc);
	      sc->code = code;
	      choose_c_function(sc, expr, func, 2);
	      if (((c_call(expr) == g_for_each) || (c_call(expr) == g_map)) &&
		  (is_pair(cadr(arg1))) &&
		  (is_null(cdadr(arg1))) &&
		  (is_pair(cddr(arg1))) &&
		  (is_null(cdddr(arg1))) &&
		  (!is_immutable_symbol(caadr(arg1))))
		{
		  set_c_call(expr, (c_call(expr) == g_for_each) ? g_for_each_closure : g_map_closure);
		  set_unsafe_optimize_op(expr, HOP_C_FA_1);
		}
	      return(OPT_F);
	    }
	  
	  if ((!is_pair(arg1)) ||
	      ((is_optimized(arg1)) &&
	       (is_all_x_op(optimize_op(arg1)))))
	    {
	      set_unsafe_optimize_op(expr, hop + OP_C_AA);
	      annotate_args(sc, cdr(expr), e);
	      set_arglist_length(expr, small_int(2));
	      choose_c_function(sc, expr, func, 2);
	      return(OPT_F);
	    }
	}
      return((is_optimized(expr)) ? OPT_T : OPT_F);
    }

  if (is_closure(func))
    {
      int32_t arit;
      arit = closure_arity_to_int(sc, func);
      if (arit != 2)
	{
	  if ((arit == -1) &&
	      (is_symbol(closure_args(func))))
	    return(optimize_func_dotted_args(sc, expr, func, hop, 2, pairs, symbols, quotes, bad_pairs, e));
	  return(OPT_F);
	}
      if (is_immutable(func)) hop = 1;

      /* safe_closures with 1-expr c_c h-optimized body with hop=1 happen only in tgen (and not enough there to warrant any effort) */

      if ((pairs == 0) &&
	  (symbols >= 1))
	{
	  set_unsafely_optimized(expr);
	  if (symbols == 2)
	    {
	      if (is_safe_closure(func))
		set_optimize_op(expr, hop + ((is_null(cdr(closure_body(func)))) ? OP_SAFE_CLOSURE_SS : OP_SAFE_CLOSURE_SS_B));
	      else
		{
		  s7_pointer body;
		  body = closure_body(func);
		  if ((is_null(cdr(body))) &&
		      (is_pair(car(body))) &&
		      (is_syntactic(caar(body))))
		    {
		      set_optimize_op(expr, hop + OP_CLOSURE_SS_P);
		      if (typesflag(car(body)) != SYNTACTIC_PAIR)
			{
			  pair_set_syntax_op(car(body), symbol_syntax_op(caar(body)));
			  set_syntactic_pair(car(body));
			}
		    }
		  else set_optimize_op(expr, hop + OP_CLOSURE_SS);
		}
	      set_opt_sym2(expr, arg2);
	    }
	  else
	    {
	      if (is_symbol(arg1))
		{
		  set_optimize_op(expr, hop + ((is_safe_closure(func) ? OP_SAFE_CLOSURE_SC : OP_CLOSURE_SC)));
		  set_opt_con2(expr, arg2);
		}
	      else 
		{
		  set_optimize_op(expr, hop + ((is_safe_closure(func) ? OP_SAFE_CLOSURE_CS : OP_CLOSURE_CS)));
		  set_opt_sym2(expr, arg2);
		}
	    }
	  set_opt_lambda(expr, func);
	  return(OPT_F);
	}
      
      if ((!arglist_has_rest(sc, closure_args(func))) &&
	  (all_x_count(sc, expr) == 2))
	{
	  set_unsafely_optimized(expr);
	  if (is_safe_closure(func))
	    {
	      if (is_symbol(arg1))
		set_optimize_op(expr, hop + OP_SAFE_CLOSURE_SA);
	      else set_optimize_op(expr, hop + OP_SAFE_CLOSURE_AA);
	    }
	  else set_optimize_op(expr, hop + OP_CLOSURE_AA);
	  annotate_args(sc, cdr(expr), e);
	  set_opt_lambda(expr, func);
	  set_arglist_length(expr, small_int(2));
	  return(OPT_F);
	}
      
      if (is_all_x_safe(sc, arg1))
	{
	  set_unsafely_optimized(expr);
	  annotate_arg(sc, cdr(expr), e);
	  set_optimize_op(expr, hop + ((is_safe_closure(func)) ? OP_SAFE_CLOSURE_AP : OP_CLOSURE_AP));
	  set_opt_lambda(expr, func);
	  return(OPT_F);
	}

      if ((is_pair(arg1)) &&
	  (car(arg1) == sc->lambda_symbol) &&
	  ((!is_pair(arg2)) ||
	   ((is_optimized(arg2)) &&
	    (is_all_x_op(optimize_op(arg2))))) &&
	  (is_null(cdr(closure_body(func)))))
	{
	  s7_pointer code;
	  annotate_arg(sc, cddr(expr), e);
	  set_unsafe_optimize_op(expr, hop + OP_CLOSURE_FA);
	  code = sc->code;
	  sc->code = cdadr(expr);
	  check_lambda(sc);
	  clear_safe_closure(cdr(sc->code)); /* otherwise we need to fixup the local let for the optimizer */
	  sc->code = code;
	  set_opt_lambda(expr, func);
	  return(OPT_F);
	}

      if (is_all_x_safe(sc, arg2))
	{
	  set_unsafely_optimized(expr);
	  annotate_arg(sc, cddr(expr), e);
	  set_optimize_op(expr, hop + ((is_safe_closure(func)) ? OP_SAFE_CLOSURE_PA : OP_CLOSURE_PA));
	  set_opt_lambda(expr, func);
	  return(OPT_F);
	}

      return((is_optimized(expr)) ? OPT_T : OPT_F);
    }

  if (is_closure_star(func))
    {
      if (!has_simple_arg_defaults(closure_body(func)))
	return(OPT_F);

      if ((!arglist_has_rest(sc, closure_args(func))) &&
	  (all_x_count(sc, expr) == 2))
	{
	  set_unsafely_optimized(expr);
	  if (closure_star_arity_to_int(sc, func) == 2)
	    set_optimize_op(expr, hop + ((is_safe_closure(func)) ? OP_SAFE_CLOSURE_STAR_AA : OP_CLOSURE_STAR_AA));
	  else set_optimize_op(expr, hop + ((is_safe_closure(func)) ? OP_SAFE_CLOSURE_STAR_ALL_X : OP_CLOSURE_STAR_ALL_X));
	  annotate_args(sc, cdr(expr), e);
	  set_opt_lambda(expr, func);
	  set_arglist_length(expr, small_int(2));
	  return(OPT_F);
	}
    }

  if ((is_c_function_star(func)) &&
      (all_x_count(sc, expr) == 2))
    {
      set_optimized(expr);
      set_optimize_op(expr, hop + OP_SAFE_C_STAR_AA);
      annotate_args(sc, cdr(expr), e);
      set_arglist_length(expr, small_int(2));
      clear_overlay(expr);
      set_c_function(expr, func);
      return(OPT_T);
    }

  return((is_optimized(expr)) ? OPT_T : OPT_F);
}

static opt_t optimize_func_three_args(s7_scheme *sc, s7_pointer expr, s7_pointer func, int32_t hop, int32_t pairs, int32_t symbols, int32_t quotes, int32_t bad_pairs, s7_pointer e)
{
  s7_pointer arg1, arg2, arg3;

  arg1 = cadr(expr);
  arg2 = caddr(expr);
  arg3 = cadddr(expr);

  if (((is_symbol(arg1)) &&
       (!pair_symbol_is_safe(sc, arg1, e)) &&
       (!is_slot(find_symbol(sc, arg1)))) ||
      ((is_symbol(arg2)) &&
       (!pair_symbol_is_safe(sc, arg2, e)) &&
       (!is_slot(find_symbol(sc, arg2)))) ||
      ((is_symbol(arg3)) &&
       (!pair_symbol_is_safe(sc, arg3, e)) &&
       (!is_slot(find_symbol(sc, arg3)))))
    {
      /* wrap bad args */
      if ((is_all_x_safe(sc, arg1)) &&
	  (is_all_x_safe(sc, arg2)) &&
	  (is_all_x_safe(sc, arg3)) &&
	  (s7_is_aritable(sc, func, 3)))
	{
	  annotate_args(sc, cdr(expr), e);
	  set_arglist_length(expr, small_int(3));
	  if (is_c_function(func))
	    {
	      set_safe_optimize_op(expr, hop + ((is_safe_procedure(func)) ? OP_SAFE_C_AAA : OP_C_ALL_X));
	      set_c_function(expr, func);
	      return(OPT_T);
	    }
	  if ((is_closure(func)) &&
	      (closure_arity_to_int(sc, func) == 3) &&
	      (!arglist_has_rest(sc, closure_args(func))))
	    {
	      set_unsafely_optimized(expr);
	      set_optimize_op(expr, hop + ((is_safe_closure(func)) ? OP_SAFE_CLOSURE_ALL_X : OP_CLOSURE_ALL_X));
	      set_opt_lambda(expr, func);
	      return(OPT_F);
	    }
	  if ((is_closure_star(func)) &&
	      (has_simple_arg_defaults(closure_body(func))) &&
	      (closure_star_arity_to_int(sc, func) >= 3) &&
	      (!arglist_has_rest(sc, closure_args(func))))
	    {
	      set_unsafely_optimized(expr);
	      set_optimize_op(expr, ((is_safe_closure(func)) ? OP_SAFE_CLOSURE_STAR_ALL_X : OP_CLOSURE_STAR_ALL_X));
	      set_opt_lambda(expr, func);
	    }
	}
      return(OPT_F);
    }
  /* end of bad symbol wrappers */

  if ((bad_pairs == quotes) &&
      (is_immutable_symbol(car(expr))))
    hop = 1;

  if ((is_c_function(func) &&
       (c_function_required_args(func) <= 3) &&
       (c_function_all_args(func) >= 3)) ||
      ((is_c_function_star(func)) &&
       (c_function_all_args(func) == 3) &&
       (!is_keyword(arg1)) &&
       (!is_keyword(arg2))))
    {
      if ((is_safe_procedure(func)) ||
	  ((is_maybe_safe(func)) && 
	   (unsafe_is_safe(sc, func, arg1, arg2, arg3, e))))
	{
	  if (pairs == 0)
	    {
	      set_optimized(expr);
	      if (symbols == 0)
		set_optimize_op(expr, hop + OP_SAFE_C_C);
	      else
		{
		  if (symbols == 3)
		    {
		      set_optimize_op(expr, hop + OP_SAFE_C_SSS);
		      set_opt_sym1(cdr(expr), arg2);
		      set_opt_sym2(cdr(expr), arg3);
		    }
		  else
		    {
		      if (symbols == 2)
			{
			  if (!is_symbol(arg1))
			    {
			      set_optimize_op(expr, hop + OP_SAFE_C_CSS);
			      set_opt_sym1(cdr(expr), arg2);
			      set_opt_sym2(cdr(expr), arg3);
			    }
			  else
			    {
			      if (!is_symbol(arg3))
				{
				  set_opt_con2(cdr(expr), arg3);
				  if (is_keyword(arg2))
				    {
				      set_opt_con1(cdr(expr), arg2);
				      set_optimize_op(expr, hop + OP_SAFE_C_SCC);
				    }
				  else 
				    {
				      set_opt_sym1(cdr(expr), arg2);
				      set_optimize_op(expr, hop + OP_SAFE_C_SSC);
				    }
				}
			      else 
				{
				  set_opt_con1(cdr(expr), arg2);
				  set_opt_sym2(cdr(expr), arg3);
				  set_optimize_op(expr, hop + OP_SAFE_C_SCS);
				}
			    }
			}
		      else
			{
			  if (is_symbol(arg1))
			    {
			      set_opt_con1(cdr(expr), arg2);
			      set_opt_con2(cdr(expr), arg3);
			      set_optimize_op(expr, hop + OP_SAFE_C_SCC);
			    }
			  else
			    {
			      if (is_symbol(arg2))
				{
				  set_opt_sym1(cdr(expr), arg2);
				  set_opt_con2(cdr(expr), arg3);
				  set_optimize_op(expr, hop + OP_SAFE_C_CSC);
				}
			      else
				{
				  set_optimize_op(expr, hop + OP_SAFE_C_AAA); /* fallback on all_x_c and s here -- a kludge */
				  annotate_args(sc, cdr(expr), e);
				  set_arglist_length(expr, small_int(3));
				}
			    }
			}
		    }
		}
	      choose_c_function(sc, expr, func, 3);
	      return(OPT_T);
	    }

	  /* pairs != 0 */
	  if (all_x_count(sc, expr) == 3)
	    {
	      set_optimized(expr);
	      if (quotes == 1)
		{
		  if ((symbols == 2) &&
		      (is_symbol(arg1)) &&
		      (is_symbol(arg3)))
		    {
		      set_opt_con1(cdr(expr), cadr(arg2));
		      set_opt_sym2(cdr(expr), arg3);
		      set_optimize_op(expr, hop + OP_SAFE_C_SQS);
		      choose_c_function(sc, expr, func, 3);
		      return(OPT_T);
		    }
		  if ((symbols == 1) &&
		      (is_symbol(arg3)) &&
		      (is_proper_quote(sc, arg2)) &&
		      (is_safe_c_s(arg1)))
		    {
		      set_safe_optimize_op(expr, hop + OP_SAFE_C_opSq_QS);
		      set_opt_con1(cdr(expr), cadr(arg2));
		      set_opt_sym2(cdr(expr), arg3);
		      choose_c_function(sc, expr, func, 3);
		      return(OPT_T);
		    }
		}
	      annotate_args(sc, cdr(expr), e);
	      set_arglist_length(expr, small_int(3));
	      set_optimize_op(expr, hop + OP_SAFE_C_AAA);

	      if (pairs == 1)
		{
		  if (symbols == 1)
		    {
		      if (is_pair(arg3))
			{
			  if (is_symbol(arg2))
			    set_optimize_op(expr, hop + OP_SAFE_C_CSA);
			  else set_optimize_op(expr, hop + OP_SAFE_C_SCA);
			}
		    }
		  else
		    {
		      if ((symbols == 2) && (is_symbol(arg1)))
			set_optimize_op(expr, hop + ((is_symbol(arg2)) ? OP_SAFE_C_SSA : OP_SAFE_C_SAS));
		    }
		}
	      choose_c_function(sc, expr, func, 3);
	      return(OPT_T);
	    }

	  if (bad_pairs == 0)
	    {
	      if ((symbols == 2) &&
		  (is_symbol(arg1)) &&
		  (is_symbol(arg2)))
		{
		  annotate_arg(sc, cdr(expr), e);
		  annotate_arg(sc, cddr(expr), e);
		  set_optimize_op(expr, hop + OP_SAFE_C_AAZ);
		}
	      else
		{
		  /* use either X or Z in all 8 choices */
		  if ((!is_pair(arg1)) ||
		      (is_all_x_op(optimize_op(arg1))))
		    {
		      annotate_arg(sc, cdr(expr), e);
		      if ((!is_pair(arg2)) ||
			  (is_all_x_op(optimize_op(arg2))))
			{
			  set_optimize_op(expr, hop + OP_SAFE_C_AAZ); /* here last can't be A because we checked for that above */
			  annotate_arg(sc, cddr(expr), e);
			}
		      else
			{
			  if ((!is_pair(arg3)) ||
			      (is_all_x_op(optimize_op(arg3))))
			    {
			      set_optimize_op(expr, hop + OP_SAFE_C_AZA);
			      annotate_arg(sc, cdddr(expr), e);
			    }
			  else set_optimize_op(expr, hop + OP_SAFE_C_AZZ);
			}
		    }
		  else
		    {
		      if ((!is_pair(arg2)) ||
			  (is_all_x_op(optimize_op(arg2))))
			{
			  annotate_arg(sc, cddr(expr), e);
			  if ((!is_pair(arg3)) ||
			      (is_all_x_op(optimize_op(arg3))))
			    {
			      set_optimize_op(expr, hop + OP_SAFE_C_ZAA);
			      annotate_arg(sc, cdddr(expr), e);
			    }
			  else set_optimize_op(expr, hop + OP_SAFE_C_ZAZ);
			}
		      else
			{
			  if ((!is_pair(arg3)) ||
			      (is_all_x_op(optimize_op(arg3))))
			    {
			      set_optimize_op(expr, hop + OP_SAFE_C_ZZA);
			      annotate_arg(sc, cdddr(expr), e);
			    }
			  else 
			    {
			      set_opt_pair2(cdr(expr), arg2);
			      set_opt_pair1(cdr(expr), arg3);
			      set_optimize_op(expr, hop + OP_SAFE_C_ZZZ);
			    }
			}
		    }
		}
	      set_optimized(expr);
	      choose_c_function(sc, expr, func, 3);
	      set_arglist_length(expr, small_int(3));
	      return(OPT_T);
	    }

	  if ((is_all_x_safe(sc, arg1)) &&
	      (is_all_x_safe(sc, arg2)))
	    {
	      annotate_arg(sc, cdr(expr), e);
	      annotate_arg(sc, cddr(expr), e);
	      set_unsafe_optimize_op(expr, hop + OP_SAFE_C_AAP);
	      choose_c_function(sc, expr, func, 3);
	      return(OPT_F);
	    }
	}
      else /* func is not safe */
	{
	  if (all_x_count(sc, expr) == 3)
	    {
	      set_optimized(expr);
	      if ((symbols == 2) &&
		  (pairs == 0) &&
		  (is_symbol(arg1)) &&
		  (is_symbol(arg3)))
		set_optimize_op(expr, hop + OP_C_SCS);
	      else
		{
		  annotate_args(sc, cdr(expr), e);
		  set_arglist_length(expr, small_int(3));
		  set_optimize_op(expr, hop + OP_C_ALL_X);
		}
	      choose_c_function(sc, expr, func, 3);
	      if (optimize_op(expr) != HOP_SAFE_C_C) /* did chooser fix it up? */
		{
		  set_unsafe(expr);
		  return(OPT_F);
		}
	      return(OPT_T);
	    }

	  /* (define (hi) (catch #t (lambda () 1) (lambda args 2)))
	   *   first arg list must be (), second a symbol
	   */
	  if (c_function_call(func) == g_catch)
	    {
	      if (((bad_pairs == 2) && (!is_pair(arg1))) ||
		  ((bad_pairs == 3) && (car(arg1) == sc->quote_symbol)))
		{
		  s7_pointer body_lambda, error_lambda;
		  body_lambda = arg2;
		  error_lambda = arg3;

		  if ((is_pair(body_lambda)) &&
		      (is_lambda(sc, car(body_lambda))) &&
		      (is_pair(error_lambda)) &&
		      (is_lambda(sc, car(error_lambda))) &&
		      (is_null(cadr(body_lambda))) &&
		      (is_not_null(cddr(body_lambda))) &&
		      (((is_symbol(cadr(error_lambda))) &&              /* (lambda args ... */
			(!is_immutable_symbol(cadr(error_lambda)))) ||
		       ((is_pair(cadr(error_lambda))) &&                /* (lambda (type info) ... */
			(is_pair(cdadr(error_lambda))) && 
			(is_null(cddadr(error_lambda))))) &&
		      (is_not_null(cddr(error_lambda))))
		    {
		      s7_pointer error_result;
		      error_result = caddr(error_lambda);
		      set_unsafely_optimized(expr);
		      if ((arg1 == sc->T) &&
			  (is_null(cdddr(error_lambda))) &&
			  ((!is_symbol(error_result)) ||                  /* (lambda args #f) */
			   ((is_pair(cadr(error_lambda))) &&
			    (error_result == caadr(error_lambda)))) &&    /* (lambda (type info) type) */
			  ((!is_pair(error_result)) ||                    
			   (car(error_result) == sc->quote_symbol) ||     /* (lambda args 'a) */
			   ((car(error_result) == sc->car_symbol) &&
			    (cadr(error_result) == cadr(error_lambda))))) /* (lambda args (car args) -> error-type */
			{
			  set_optimize_op(expr, hop + OP_C_CATCH_ALL);
			  set_c_function(expr, func);
			  set_opt_con2(expr, error_result);
			  set_opt_pair1(cdr(expr), cddr(body_lambda));
			  if ((is_null(cdddr(body_lambda))) &&
			      (is_optimized(caddr(body_lambda))))
			    set_optimize_op(expr, hop + OP_C_CATCH_ALL_Z);
			    
			}
		      else
			{
			  set_optimize_op(expr, hop + OP_C_CATCH);
			  choose_c_function(sc, expr, func, 3);
			}
		      return(OPT_F);
		    }
		}
	    }
	}
      return((is_optimized(expr)) ? OPT_T : OPT_F);
    }

  /* not c func */
  if (is_closure(func))
    {
      int32_t arit;
      arit = closure_arity_to_int(sc, func);
      if (arit != 3)
	{
	  if ((arit == -1) &&
	      (is_symbol(closure_args(func))))
	    return(optimize_func_dotted_args(sc, expr, func, hop, 3, pairs, symbols, quotes, bad_pairs, e));
	  return(OPT_F);
	}

      if (is_immutable(func)) hop = 1;

      if ((symbols == 3) &&
	  (!is_safe_closure(func)))
	{
	  set_unsafely_optimized(expr);
	  set_opt_lambda(expr, func);
	  set_arglist_length(expr, small_int(3));
	  set_optimize_op(expr, hop + ((is_null(cdr(closure_body(func)))) ? OP_CLOSURE_ALL_S_P : OP_CLOSURE_ALL_S));
	  return(OPT_F);
	}

      if (all_x_count(sc, expr) == 3)
	{
	  if (is_safe_closure(func))
	    {
	      if (is_symbol(arg1))
		set_optimize_op(expr, hop + OP_SAFE_CLOSURE_SAA);
	      else set_optimize_op(expr, hop + OP_SAFE_CLOSURE_ALL_X);
	    }
	  else set_optimize_op(expr, hop + OP_CLOSURE_ALL_X);
	  set_unsafely_optimized(expr);
	  annotate_args(sc, cdr(expr), e);
	  set_opt_lambda(expr, func);
	  set_arglist_length(expr, small_int(3));
	  return(OPT_F);
	}
    }

  if (is_closure_star(func))
    {
      if ((!has_simple_arg_defaults(closure_body(func))) ||
	  (closure_star_arity_to_int(sc, func) < 3) ||
	  (arglist_has_rest(sc, closure_args(func))))
	return(OPT_F);

      if (all_x_count(sc, expr) == 3)
	{
	  set_unsafe_optimize_op(expr, hop + ((is_safe_closure(func) ? OP_SAFE_CLOSURE_STAR_ALL_X : OP_CLOSURE_STAR_ALL_X)));
	  annotate_args(sc, cdr(expr), e);
	  set_opt_lambda(expr, func);
	  set_arglist_length(expr, small_int(3));
	  return(OPT_F);
	}
    }

  if ((is_c_function_star(func)) &&
      (all_x_count(sc, expr) == 3))
    {
      set_optimized(expr);
      set_optimize_op(expr, hop + OP_SAFE_C_STAR_ALL_X);
      annotate_args(sc, cdr(expr), e);
      set_arglist_length(expr, small_int(3));
      clear_overlay(expr);
      set_c_function(expr, func);
      return(OPT_T);
    }

  if (bad_pairs > quotes) return(OPT_F);
  return((is_optimized(expr)) ? OPT_T : OPT_F);
}

static bool symbols_are_safe(s7_scheme *sc, s7_pointer args, s7_pointer e)
{
  s7_pointer p;
  for (p = args; is_pair(p); p = cdr(p))
    {
      s7_pointer arg;
      arg = car(p);
      if ((is_symbol(arg)) &&
	  (!pair_symbol_is_safe(sc, arg, e)) &&
	  (!is_slot(find_symbol(sc, arg))))
	return(false);
    }
  return(true);
}

static opt_t optimize_func_many_args(s7_scheme *sc, s7_pointer expr, s7_pointer func, int32_t hop, int32_t args, int32_t pairs, int32_t symbols, int32_t quotes, int32_t bad_pairs, s7_pointer e)
{
  bool func_is_closure;

  if (bad_pairs > quotes) return(OPT_F);
  if ((bad_pairs == quotes) &&
      (is_immutable_symbol(car(expr))))
    hop = 1;

  if ((is_c_function(func)) &&
      (c_function_required_args(func) <= (uint32_t)args) &&
      (c_function_all_args(func) >= (uint32_t)args))
    {
      if (is_safe_procedure(func))
	{
	  if (pairs == 0)
	    {
	      if (symbols == 0)
		{
		  set_safe_optimize_op(expr, hop + OP_SAFE_C_C);
		  choose_c_function(sc, expr, func, args);
		  return(OPT_T);
		}
	      if ((symbols == args) &&
		  (args < GC_TRIGGER_SIZE))
		{
		  if (symbols_are_safe(sc, cdr(expr), e))
		    set_safe_optimize_op(expr, hop + OP_SAFE_C_ALL_S);
		  else
		    {
		      set_optimized(expr);
		      if (args == 4)
			set_optimize_op(expr, hop + OP_SAFE_C_AAAA);
		      else set_optimize_op(expr, hop + OP_SAFE_C_ALL_X);
		      annotate_args(sc, cdr(expr), e);
		    }
		  set_arglist_length(expr, make_permanent_integer(args));
		  choose_c_function(sc, expr, func, args);
		  return(OPT_T);
		}
	    }

	  if ((args < GC_TRIGGER_SIZE) &&
	      (all_x_count(sc, expr) == args))
	    {
	      set_optimized(expr);
	      if (args == 4)
		set_optimize_op(expr, hop + OP_SAFE_C_AAAA);
	      else set_optimize_op(expr, hop + OP_SAFE_C_ALL_X);
	      annotate_args(sc, cdr(expr), e);
	      set_arglist_length(expr, make_permanent_integer(args));
	      choose_c_function(sc, expr, func, args);
	      return(OPT_T);
	    }
	}
      else /* c_func is not safe */
	{
	  if ((args < GC_TRIGGER_SIZE) &&
	      (all_x_count(sc, expr) == args))
	    {
	      set_unsafe_optimize_op(expr, hop + OP_C_ALL_X);
	      annotate_args(sc, cdr(expr), e);
	      set_arglist_length(expr, make_permanent_integer(args));
	      choose_c_function(sc, expr, func, args);
	      return(OPT_F);
	    }
	}
      return((is_optimized(expr)) ? OPT_T : OPT_F);
    }

  func_is_closure = is_closure(func);
  if (func_is_closure)
    {
      int32_t arit;
      arit = closure_arity_to_int(sc, func);
      if (arit != args)
	{
	  if ((arit == -1) &&
	      (is_symbol(closure_args(func))))
	    return(optimize_func_dotted_args(sc, expr, func, hop, args, pairs, symbols, quotes, bad_pairs, e));
	  return(OPT_F);
	}

      if (is_immutable(func)) hop = 1;

      if ((pairs == 0) &&
	  /* ((symbols == args) || (symbols == 0)) && */ /* i.e. all args are either symbols or constants? */
	  (args < GC_TRIGGER_SIZE))
	{
	  bool safe_case;
	  set_unsafely_optimized(expr);
	  safe_case = is_safe_closure(func);
	  set_unsafe_optimize_op(expr, hop + ((safe_case) ? OP_SAFE_CLOSURE_ALL_X : OP_CLOSURE_ALL_X));
	  annotate_args(sc, cdr(expr), e);
	  set_arglist_length(expr, make_permanent_integer(args));
	  set_opt_lambda(expr, func);

	  if ((!safe_case) &&
	      (symbols == args) &&
	      (symbols_are_safe(sc, cdr(expr), e)))
	    set_optimize_op(expr, hop + ((is_null(cdr(closure_body(func)))) ? OP_CLOSURE_ALL_S_P : OP_CLOSURE_ALL_S));
	  return(OPT_F);
	}
    }

  if ((is_closure_star(func)) &&
      ((!has_simple_arg_defaults(closure_body(func))) ||
       (closure_star_arity_to_int(sc, func) < args)))
    return(OPT_F);

  if ((is_c_function_star(func)) &&
      (all_x_count(sc, expr) == args))
    {
      set_optimized(expr);
      set_optimize_op(expr, hop + OP_SAFE_C_STAR_ALL_X);
      annotate_args(sc, cdr(expr), e);
      set_arglist_length(expr, make_permanent_integer(args));
      clear_overlay(expr);
      set_c_function(expr, func);
      return(OPT_T);
    }

  if (args < GC_TRIGGER_SIZE)
    {
      if (((func_is_closure) ||
	   (is_closure_star(func))) &&
	  (!arglist_has_rest(sc, closure_args(func))) &&
	  (all_x_count(sc, expr) == args))
	{
	  set_unsafely_optimized(expr);
	  if (func_is_closure)
	    set_optimize_op(expr, hop + ((is_safe_closure(func)) ? OP_SAFE_CLOSURE_ALL_X : OP_CLOSURE_ALL_X));
	  else set_optimize_op(expr, hop + ((is_safe_closure(func)) ? OP_SAFE_CLOSURE_STAR_ALL_X : OP_CLOSURE_STAR_ALL_X));
	  annotate_args(sc, cdr(expr), e);
	  set_arglist_length(expr, make_permanent_integer(args));
	  set_opt_lambda(expr, func);
	  return(OPT_F);
	}
    }
  return((is_optimized(expr)) ? OPT_T : OPT_F);
}

static opt_t optimize_expression(s7_scheme *sc, s7_pointer expr, int32_t hop, s7_pointer e);

static opt_t optimize_syntax(s7_scheme *sc, s7_pointer expr, s7_pointer func, int32_t hop, s7_pointer e)
{
  opcode_t op;
  s7_pointer p, body;
  /* TODO: those not allowed need to be passed back and treated as unsafe symbols (find_uncomplicated_symbol etc) */

  op = (opcode_t)syntax_opcode(func);
  sc->w = e;
  body = cdr(expr);

  switch (op)
    {
    case OP_QUOTE:
    case OP_MACROEXPAND:
      return(OPT_F);

    case OP_LET:
    case OP_LETREC:
      {
	s7_pointer vars;
	if (is_symbol(cadr(expr)))
	  {
	    if (!is_pair(cddr(expr))) /* (let name . x) */
	      return(OPT_F);
	    vars = caddr(expr);
	    body = cdddr(expr);
	  }
	else
	  {
	    vars = cadr(expr);
	    body = cddr(expr);
	    if (is_null(vars))
	      e = cons(sc, sc->nil, e);
	  }
	for (p = vars; is_pair(p); p = cdr(p))
	  {
	    s7_pointer var;
	    var = car(p);
	    if ((!is_pair(var)) ||
		(!is_symbol(car(var))) ||
		(!is_pair(cdr(var))))
	      return(OPT_OOPS);
	    if ((is_pair(cadr(var))) &&
		(!is_checked(cadr(var))) &&
		(optimize_expression(sc, cadr(var), hop, e) == OPT_OOPS))
	      return(OPT_OOPS);
	  }
	e = collect_variables(sc, vars, e);
	if (is_symbol(cadr(expr)))
	  {
	    e = cons(sc, add_symbol_to_list(sc, cadr(expr)), e);
	    sc->w = e;
	  }
      }
      break;
	  
    case OP_LET_STAR:
    case OP_LETREC_STAR:
      {
	s7_pointer vars;
	if (is_symbol(cadr(expr)))
	  {
	    if (!is_pair(cddr(expr))) /* (let name . x) */
	      return(OPT_F);
	    vars = caddr(expr);
	    body = cdddr(expr);
	  }
	else
	  {
	    vars = cadr(expr);
	    body = cddr(expr);
	    if (is_null(vars))
	      e = cons(sc, sc->nil, e);
	  }
	for (p = vars; is_pair(p); p = cdr(p))
	  {
	    s7_pointer var;
	    var = car(p);
	    if ((!is_pair(var)) ||
		(!is_symbol(car(var))) ||
		(!is_pair(cdr(var))))
	      return(OPT_OOPS);

	    if ((is_pair(cadr(var))) &&
		(!is_checked(cadr(var))) &&
		(optimize_expression(sc, cadr(var), hop, e) == OPT_OOPS))
	      return(OPT_OOPS);

	    e = cons(sc, add_symbol_to_list(sc, car(var)), e);
	    sc->w = e;
	  }
	if (is_symbol(cadr(expr)))
	  {
	    e = cons(sc, add_symbol_to_list(sc, cadr(expr)), e);
	    sc->w = e;
	  }
      }
      break;
      
    case OP_LET_TEMPORARILY:
      {
	s7_pointer vars;
	vars = cadr(expr);
	body = cddr(expr);
	for (p = vars; is_pair(p); p = cdr(p))
	  {
	    s7_pointer var;
	    var = car(vars);
	    if ((is_pair(var)) &&
		(is_pair(cdr(var))) &&
		(is_pair(cadr(var))) &&
		(!is_checked(cadr(var))))
	      {
		if (optimize_expression(sc, cadr(var), hop, e) == OPT_OOPS)
		  return(OPT_OOPS);
	      }
	  }
      }
      break;

    case OP_DO:
      {
	s7_pointer vars;

	vars = cadr(expr);
	if (is_null(vars))
	  e = cons(sc, sc->nil, e);
	body = cddr(expr);

	for (p = vars; is_pair(p); p = cdr(p))
	  {
	    s7_pointer var;
	    var = car(p);
	    if ((!is_pair(var)) ||
		(!is_symbol(car(var))) ||
		(!is_pair(cdr(var))))
	      return(OPT_OOPS);
	    if ((is_pair(cadr(var))) &&
		(!is_checked(cadr(var))) &&
		(optimize_expression(sc, cadr(var), hop, e) == OPT_OOPS)) /* the init field -- locals are not defined yet */
	      return(OPT_OOPS);
	  }
	e = collect_variables(sc, vars, e);
	for (p = vars; is_pair(p); p = cdr(p))
	  {
	    s7_pointer var;
	    var = cddar(p);
	    if ((is_pair(var)) &&
		(is_pair(car(var))) &&
		(!is_checked(car(var))) &&
		(optimize_expression(sc, car(var), hop, e) == OPT_OOPS)) /* the step field -- locals are defined */
	      return(OPT_OOPS);
	  }
      }
      break;

    case OP_DEFINE_MACRO:
    case OP_DEFINE_MACRO_STAR:
    case OP_DEFINE_BACRO:
    case OP_DEFINE_BACRO_STAR:
    case OP_DEFINE_CONSTANT:
    case OP_DEFINE_EXPANSION:
    case OP_DEFINE:
    case OP_DEFINE_STAR:
      /* fprintf(stderr, "  %s before: %s\n", DISPLAY(car(expr)), DISPLAY(e)); */
      /* define adds a name to the incoming env (e), the added name is inserted into e after the first, so the caller
       *   can flush added symbols by maintaining its own pointer into the list if blockers set the car.
       * the list is used both to see local symbols and to catch "complicated" functions (find_uncomplicated_symbol).
       * In cases like (if expr (define...)) we can't tell at this level whether the define takes place, so
       *   its name should not be in "e", but it needs to be marked for find_uncomplicated_symbol in a way
       *   that can be distinguished from members of "e".  So in that (rare) case, we use the associated keyword.
       *   Then find_uncomplicated_symbol can use has_keyword to tell if the keyword search is needed.
       * An oddity: if we're not defining a function, the name is not defined in the body:
       *   (define x (+ x 1)) the inner x is not the x-being-defined.
       */
      {
	s7_pointer name_args;
	name_args = cadr(expr);
	body = cddr(expr);
	if (is_pair(name_args))
	  {
	    if (is_symbol(car(name_args)))
	      {
		add_symbol_to_list(sc, car(name_args));
		if (is_pair(e))
		  {
		    if (car(e) != sc->key_rest_symbol)
		      set_cdr(e, cons(sc, car(name_args), cdr(e))); /* export it */
		    else add_symbol_to_list(sc, symbol_to_keyword(sc, car(name_args)));
		  }
		else e = cons(sc, car(name_args), e);
	      }
	    e = collect_parameters(sc, cdr(name_args), e);
	  }
	else 
	  {
	    if (is_symbol(name_args))
	      {
		/* actually if this is defining a function, the name should probably be included in the local env 
		 *   but that's next-to-impossible to guarantee unless it's (define x (lambda...)) of course.
		 */
		sc->temp9 = e;
		for (p = body; is_pair(p); p = cdr(p))
		  if ((is_pair(car(p))) && 
		      (!is_checked(car(p))) && /* ((typeflag & (0xff | T_CHECKED)) == T_PAIR) is not faster */
		      (optimize_expression(sc, car(p), hop, e) == OPT_OOPS))
		    return(OPT_OOPS);
		sc->temp9 = sc->nil;

		add_symbol_to_list(sc, name_args);
		if (is_pair(e))
		  {
		    if (car(e) != sc->key_rest_symbol)
		      set_cdr(e, cons(sc, name_args, cdr(e)));     /* export it */
		    else add_symbol_to_list(sc, symbol_to_keyword(sc, name_args));
		  }
		else e = cons(sc, name_args, e);
		return(OPT_F);
	      }
	  }
      }
      break;

    case OP_LAMBDA:
    case OP_LAMBDA_STAR:
      if (is_null(cadr(expr)))
	e = cons(sc, sc->nil, e);
      e = collect_parameters(sc, cadr(expr), e);
      body = cddr(expr);
      break;

    case OP_SET:
      if ((is_pair(cadr(expr))) &&
	  (caadr(expr) == sc->outlet_symbol))
	return(OPT_OOPS);
      body = cddr(expr); 
      break;

    case OP_WITH_LET:
      if (sc->safety > NO_SAFETY)
	hop = 0;
      e = sc->nil;
      /* we can't trust anything here, so hop ought to be off.  For example,
       *    (define (hi)
       *      (let ((e (sublet (curlet)
       *                 (cons 'abs (lambda (a) (- a 1))))))
       *        (with-let e (abs -1))))
       * returns 1 if hop is 1, but -2 outside the function body.
       */
      break;

    case OP_CASE:
      if ((is_pair(cadr(expr))) &&
	  (optimize_expression(sc, cadr(expr), hop, e) == OPT_OOPS))
	return(OPT_OOPS);
      for (p = cddr(expr); is_pair(p); p = cdr(p))
	if ((is_pair(car(p))) &&
	    (is_pair(cdar(p))) &&
	    (optimize_expression(sc, cdar(p), hop, e) == OPT_OOPS))
	  return(OPT_OOPS);
      return(OPT_F);
      break;

    case OP_COND: /* split opt is necessary: (cond (lambda (x) ...)) */
      for (p = cdr(expr); is_pair(p); p = cdr(p))
	if (is_pair(car(p)))
	  {
	    s7_pointer test, rest;
	    test = caar(p);
	    rest = cdar(p);
	    e = cons(sc, sc->key_rest_symbol, e);
	    if (((is_pair(test)) && (optimize_expression(sc, test, hop, e) == OPT_OOPS)) ||
		((is_pair(rest)) && (optimize_expression(sc, rest, hop, e) == OPT_OOPS)))
	      return(OPT_OOPS);
	  }
      return(OPT_F);
      break;

    case OP_IF:
    case OP_WHEN:
    case OP_UNLESS:
    case OP_OR:
    case OP_AND:
      e = cons(sc, sc->key_rest_symbol, e);
      break;

    default:
      break;
    }

  sc->temp9 = e;
  for (p = body; is_pair(p); p = cdr(p))
    if ((is_pair(car(p))) && 
	(!is_checked(car(p))) && /* ((typeflag & (0xff | T_CHECKED)) == T_PAIR) is not faster */
	(optimize_expression(sc, car(p), hop, e) == OPT_OOPS))
      {
	sc->temp9 = sc->nil;
	return(OPT_OOPS);
      }
  sc->temp9 = sc->nil;

  if ((hop == 1) &&
      (symbol_id(car(expr)) == 0))
    {
      if ((op == OP_IF) || (op == OP_OR) || (op == OP_AND))
	{
	  bool happy = true;
	  for (p = cdr(expr); (happy) && (is_pair(p)); p = cdr(p))
	    happy = is_all_x_safe(sc, car(p));

	  if ((happy) &&       /* all_x* will work */
	      (is_null(p)))    /* catch the syntax error later: (or #f . 2) etc */
	    {
	      int32_t args, symbols = 0, pairs = 0, rest = 0;
	      s7_pointer sym = NULL;
	      bool c_s_is_ok = true;

	      for (args = 0, p = cdr(expr); is_pair(p); p = cdr(p), args++)
		{
		  if (is_symbol(car(p)))
		    symbols++;
		  else
		    {
		      if (!is_pair(car(p)))
			rest++;
		      else
			{
			  pairs++;
			  if ((c_s_is_ok) &&
			      ((!is_h_safe_c_s(car(p))) ||
			       ((sym) && (sym != cadar(p)))))
			    c_s_is_ok = false;
			  else 
			    {
			      if (is_pair(cdar(p)))
				sym = cadar(p);
			      else sym = sc->unspecified;
			    }
			}
		    }
		}

	      if ((op == OP_IF) &&
		  ((args < 2) || (args > 3))) /* syntax error */
		return(OPT_F);

	      if ((pairs == args) &&
		  (c_s_is_ok))
		{
		  if (op == OP_OR)
		    {
		      set_safe_optimize_op(expr, hop + OP_SAFE_C_C);
		      set_c_function(expr, or_s_direct);
		    }
		  else
		    {
		      if (op == OP_AND)
			{
			  set_safe_optimize_op(expr, hop + OP_SAFE_C_C);
			  set_c_function(expr, and_s_direct);
			}
		    }
		  return(OPT_F);
		}

	      set_safe_optimize_op(expr, hop + OP_SAFE_C_C);

	      for (p = cdr(expr); is_pair(p); p = cdr(p))
		set_x_call(p, all_x_eval(sc, p, e, pair_symbol_is_safe));

	      if (op == OP_OR)
		{
		  if (s7_list_length(sc, cdr(expr)) == 2)
		    {
		      set_c_function(expr, or_2);
		      add_optimizer_fixup(sc, expr, (uint32_t)(hop + OP_SAFE_C_OR2));
		    }
		  else 
		    {
		      if (s7_list_length(sc, cdr(expr)) == 3)
			set_c_function(expr, or_3);
		      else set_c_function(expr, or_n);
		    }
		}
	      else
		{
		  if (op == OP_AND)
		    {
		      if (s7_list_length(sc, cdr(expr)) == 2)
			{
			  if ((c_call(cdr(expr)) == all_x_c_s) &&
			      (c_call(cddr(expr)) == all_x_c_c))
			    set_c_function(expr, and_sc);
			  else 
			    {
			      set_c_function(expr, and_2);
			      add_optimizer_fixup(sc, expr, (uint32_t)(hop + OP_SAFE_C_AND2));
			    }
			}
		      else
			{
			  if (s7_list_length(sc, cdr(expr)) == 3)
			    set_c_function(expr, and_3);
			  else set_c_function(expr, and_n);
			}
		    }
		  else
		    {
		      s7_pointer test, b1, b2;
		      test = cdr(expr);
		      b1 = cdr(test);
		      b2 = cdr(b1);
		      if ((c_call(b1) == all_x_q) &&
			  (is_pair(b2)))
			{
			  if (c_call(b2) == all_x_q)
			    set_c_function(expr, if_x_qq);
			  else set_c_function(expr, if_x_qa);
			}
		      else
			{
			  if ((is_pair(car(test))) &&
			      (caar(test) == sc->not_symbol) &&
			      (is_all_x_safe(sc, cadar(test))))
			    {
			      set_x_call(test, all_x_eval(sc, cdar(test), e, pair_symbol_is_safe));
			      if (is_null(b2))
				set_c_function(expr, if_not_x1);
			      else set_c_function(expr, if_not_x2);
			    }
			  else
			    {
			      if (is_null(b2))
				set_c_function(expr, if_x1);
			      else set_c_function(expr, if_x2);
			    }
			}
		    }
		}
	      return(OPT_T);
	    }
	  /* else we could check other if cases here (test is often all_x_safe)
	   */
	}
    }
  return(OPT_F);
}


static opt_t optimize_expression(s7_scheme *sc, s7_pointer expr, int32_t hop, s7_pointer e)
{
  s7_pointer car_expr;

  /* fprintf(stderr, "opt-expr %d %s %s\n", hop, DISPLAY_80(expr), DISPLAY(e)); */
  /* if (is_checked(expr)) return(OPT_T); */

  set_checked(expr);
  car_expr = car(expr);

  if (is_symbol(car_expr))
    {
      s7_pointer func;
      if (is_syntactic(car_expr))
	{
	  if (!is_pair(cdr(expr)))
	    return(OPT_OOPS);
	  return(optimize_syntax(sc, expr, _TSyn(slot_value(global_slot(car_expr))), hop, e));
	}

      func = find_uncomplicated_symbol(sc, car_expr, e);
      if (is_slot(func))
	{
	  func = slot_value(func);
	  if (is_syntax(func))   /* 12-8-16 was is_syntactic, but that is only appropriate above -- here we have the value */
	    {
	      if (!is_pair(cdr(expr)))
		return(OPT_OOPS);
	      return(optimize_syntax(sc, expr, func, hop, e));  /* e can be extended via set-cdr! here */
	    }
	  /* we miss implicit indexing here because at this time, the data are not set */
	  if ((is_procedure(func)) ||
	      /* (is_c_function(func)) || */
	      (is_safe_procedure(func))) /* built-in applicable objects like vectors */
	    {
	      int32_t pairs = 0, symbols = 0, args = 0, bad_pairs = 0, quotes = 0, orig_hop;
	      s7_pointer p;

	      orig_hop = hop;
	      if ((!is_immutable(car_expr)) &&     /* can't depend on opt1 here because it might not be global, or might be redefined locally */
		  ((is_any_closure(func)) ||
		   ((!is_global(car_expr)) &&
		    ((!is_slot(global_slot(car_expr))) ||
		     (slot_value(global_slot(car_expr)) != func)))))
		{
		  /* (let () (define (f2 a) (+ a 1)) (define (f1 a) (f2 a)) (define (f2 a) (- a)) (f1 12))
		   * (let () (define (f2 a) (+ a 1)) (define (f1 a) (f2 a)) (define (f2 a) (- a 1)) (f1 12))
		   * and similar define* cases
		   */

		  hop = 0;
		  /* this is very tricky!  See s7test for some cases.  Basically, we need to protect a recursive call
		   *   of the current function being optimized from being confused with some previous definition
		   *   of the same name.  But method lists have global names so the global bit is off even though the
		   *   thing is actually a safe global.  But no closure can be considered safe in the hop sense --
		   *   even a global function might be redefined at any time, and previous uses of it in other functions
		   *   need to reflect its new value.
		   *   So, closures are always checked, but built-in functions are used as if never redefined until that redefinition.
		   * Syntax handling is already impure in s7, so the special handling of built-in functions doesn't
		   *   offend me much.  Consider each a sort of reader macro until someone redefines it -- previous
		   *   uses may not be affected because they might have been optimized away -- the result depends on the
		   *   current optimizer.
		   * Another case (from K Matheussen):
		   *   (define (call-func func arg1 arg2) (define (call) (func arg1 arg2)) (call)) (call-func + 1 2.5) (call-func - 5 2)
		   *   when we get here originally "func" is +, hop=1, but just checking for !is_global(car_expr) is
		   *   not good enough -- if we load mockery.scm, nothing is global!
		   * Yet another case (define (test-abs) (define (abs x) (+ x 1)) (format *stderr* "abs ~A~%" (abs -1)))
		   *   when optimize_syntax sees the (define abs ...), it inserts abs into e so that the caller's e is extended (set-cdr!)
		   *   so that find_uncomplicated_symbol above will be unhappy when we reach (abs -1) as the format arg!
		   *   This can be confused if lambda is redefined at some point, but...
		   */
		}
	      /* but if we make a recursive call on a func, we've obviously already looked up that function, and
	       *   if it has not been shadowed, then we don't need to check it -- so the hop bit should be on
	       *   for that one case.
	       */

	      for (p = cdr(expr); is_pair(p); p = cdr(p), args++) /* check the args (the calling expression) */
		{
		  s7_pointer car_p;
		  car_p = car(p);
		  if (is_symbol(car_p))
		    symbols++;
		  else
		    {
		      if (is_pair(car_p))
			{
			  pairs++;
			  if (!is_checked(car_p))
			    {
			      opt_t res;
			      res = optimize_expression(sc, car_p, orig_hop, e);
			      if (res == OPT_F)
				{
				  bad_pairs++;
				  if ((is_proper_quote(sc, car_p)) &&
				      (is_null(cddr(car_p))))
				    quotes++;
				}
			      else
				{
				  if (res == OPT_OOPS)
				    return(OPT_OOPS);
				}
			    }
			  else
			    {
			      if ((!is_optimized(car_p)) ||
				  (is_unsafe(car_p)))
				{
				  bad_pairs++;
				  if ((is_proper_quote(sc, car_p)) &&
				      (is_null(cddr(car_p))))
				    quotes++;
				}
			    }
			}
		    }
		}
	      if (is_null(p))                    /* if not null, dotted list of args? */
		{
		  switch (args)
		    {
		    case 0:  return(optimize_thunk(sc, expr, func, hop));
		    case 1:  return(optimize_func_one_arg(sc, expr, func, hop, pairs, symbols, quotes, bad_pairs, e));
		    case 2:  return(optimize_func_two_args(sc, expr, func, hop, pairs, symbols, quotes, bad_pairs, e));
		    case 3:  return(optimize_func_three_args(sc, expr, func, hop, pairs, symbols, quotes, bad_pairs, e));
		    default: return(optimize_func_many_args(sc, expr, func, hop, args, pairs, symbols, quotes, bad_pairs, e));
		    }
		}
	      return(OPT_F);
	    }
	}
      else
	{
	  if ((sc->undefined_identifier_warnings) &&
	      (func == sc->undefined) &&           /* car_expr is not in e or global */
	      (symbol_tag(car_expr) == 0))         /*    and we haven't looked it up earlier */
	    {
	      s7_pointer p;
	      p = sc->input_port;
	      if ((is_input_port(p)) &&
		  (port_file(p) != stdin) &&
		  (!port_is_closed(p)) &&
		  (port_filename(p)))
		s7_warn(sc, 1024, "%s might be undefined (%s %u)\n", DISPLAY(car_expr), port_filename(p), port_line_number(p));
	      else s7_warn(sc, 1024, "; %s might be undefined\n", DISPLAY(car_expr));
	      symbol_set_tag(car_expr, 1);             /* one warning is enough */
	    }
	  /* we need local definitions and func args in e?  also check is_symbol case below
	   */
	}

      /* car_expr is a symbol but it's not a known procedure or a "safe" case = vector etc */
      {
	/* else maybe it's something like a let variable binding: (sqrtfreq (sqrt frequency)) */
	s7_pointer p;
	int32_t len = 0, pairs = 0, symbols = 0, quotes = 0;

	for (p = cdr(expr); is_pair(p); p = cdr(p), len++)
	  {
	    s7_pointer car_p;
	    car_p = car(p);
	    if (is_pair(car_p))
	      {
		pairs++;
		if ((hop != 0) && (car(car_p) == sc->quote_symbol))
		  quotes++;
		if (!is_checked(car_p))
		  {
		    if (optimize_expression(sc, car_p, hop, e) == OPT_OOPS)
		      return(OPT_OOPS);
		  }
	      }
	    else
	      {
		if (is_symbol(car_p))
		  symbols++;
	      }
	  }
	
	if ((is_null(p)) &&              /* (+ 1 . 2) */
	    (!is_optimized(expr)))
	  {
	    /* len=0 case is almost entirely arglists */
	    set_opt_con1(expr, sc->gc_nil);
	    clear_overlay(expr);

	    if (pairs == 0)
	      {
		if (len == 0)
		  {
		    /* hoping to catch object application here, as in readers in Snd */
		    set_unsafe_optimize_op(expr, OP_UNKNOWN);
		    return(OPT_F);
		  }

		if (len == 1)
		  {
		    if (car_expr != sc->quote_symbol) /* !! quote can be redefined locally, unsetting the T_SYNTACTIC flag -- can this happen elsewhere? */
		      {
			set_unsafe_optimize_op(expr, OP_UNKNOWN_G);
			/* hooboy -- we get here in let bindings...
			 * to save access to the caller, we'd need to pass it as an arg to optimize_expression
			 */
		      }
		    return(OPT_F);
		  }

		if (len == 2)
		  {
		    set_unsafely_optimized(expr);
		    set_optimize_op(expr, OP_UNKNOWN_GG);
		    return(OPT_F);
		  }

		if ((len >= 3) &&
		    (len == symbols))
		  {
		    set_unsafe_optimize_op(expr, OP_UNKNOWN_ALL_S);
		    set_arglist_length(expr, make_permanent_integer(len));
		    return(OPT_F);
		  }
	      }
	    else /* pairs != 0 */
	      {
		s7_pointer arg1;
		arg1 = cadr(expr);
		if (pairs == 1)
		  {
		    if (len == 1)
		      {
			if (car(expr) == sc->quote_symbol)
			  {
			    set_unsafe_optimize_op(expr, HOP_SAFE_QUOTE); /* see also below */
			    set_c_function(expr, slot_value(global_slot(sc->cadr_symbol)));
			    return(OPT_F);
			  }

			if (is_all_x_safe(sc, arg1))
			  {
			    set_arglist_length(expr, small_int(1));
			    annotate_arg(sc, cdr(expr), e);
			    set_unsafe_optimize_op(expr, OP_UNKNOWN_A);
			    return(OPT_F);
			  }
		      }
		    else
		      {
			if (len == 2)
			  {
			    if ((is_all_x_safe(sc, arg1)) &&
				(is_all_x_safe(sc, caddr(expr))))
			      {
				set_arglist_length(expr, small_int(2));
				set_unsafe_optimize_op(expr, OP_UNKNOWN_AA);
				return(OPT_F);
			      }
			  }
		      }
		  }

		if ((len == 2) &&
		    (is_all_x_safe(sc, arg1)) &&
		    (is_all_x_safe(sc, caddr(expr))))
		  {
		    set_arglist_length(expr, small_int(2));
		    set_unsafe_optimize_op(expr, OP_UNKNOWN_AA);
		    return(OPT_F);
		  }

		if ((all_x_count(sc, expr) == len) &&
		    (len < GC_TRIGGER_SIZE))
		  {
		    if ((len == 1) &&
			(car(expr) == sc->quote_symbol))
		      {
			set_safe_optimize_op(expr, HOP_SAFE_QUOTE);
			set_c_function(expr, slot_value(global_slot(sc->cadr_symbol)));
			return(OPT_T);
		      }

		    set_unsafe_optimize_op(expr, (len == 1) ? OP_UNKNOWN_A : OP_UNKNOWN_ALL_X);
		    set_arglist_length(expr, make_permanent_integer(len));
		    if (len == 1)
		      annotate_arg(sc, cdr(expr), e);
		    return(OPT_F);
		  }
	      }
	  }
      }
    }
  else
    {
      /* car(expr) is not a symbol, but there might be interesting stuff here */
      /* (define (hi a) (case 1 ((1) (if (> a 2) a 2)))) */
      s7_pointer p;

      for (p = expr; is_pair(p); p = cdr(p))
	{
	  if ((is_pair(car(p))) && (!is_checked(car(p))))
	    if (optimize_expression(sc, car(p), hop, e) == OPT_OOPS)
	      return(OPT_OOPS);
	}

      if ((is_pair(cdr(expr))) &&
	  (is_null(cddr(expr))) &&
	  (is_pair(car_expr)) &&
	  (car(car_expr) == sc->if_symbol) &&
	  (is_pair(cdr(car_expr))) &&
	  (is_pair(cddr(car_expr))) &&
	  (is_symbol(caddr(car_expr))) &&
	  (is_pair(cdddr(car_expr))) &&
	  (is_symbol(cadddr(car_expr))) &&
	  ((!is_pair(cadr(car_expr))) || (is_all_x_safe(sc, cadr(car_expr)))) &&
	  ((!is_pair(cadr(expr))) || (is_all_x_safe(sc, cadr(expr)))))
	{
	  s7_pointer ptrue;
	  ptrue = find_uncomplicated_symbol(sc, caddr(car_expr), e);
	  if (is_slot(ptrue))
	    {
	      ptrue = slot_value(ptrue);
	      if ((is_c_function(ptrue)) &&
		  (is_safe_procedure(ptrue)))
		{
		  s7_pointer pfalse;
		  pfalse = find_uncomplicated_symbol(sc, cadddr(car_expr), e);
		  if (is_slot(pfalse))
		    {
		      pfalse = slot_value(pfalse);
		      if ((is_c_function(pfalse)) &&
			  (is_safe_procedure(pfalse)))
			{
			  set_opt_con1(expr, ptrue);
			  set_opt_con2(expr, pfalse);
			  set_safe_optimize_op(expr, hop + OP_SAFE_IFA_SS_A);
			  annotate_arg(sc, cdr(car_expr), e);
			  annotate_arg(sc, cdr(expr), e);
			  return(OPT_T);
			}
		    }
		}
	    }
	}
    }
  return(OPT_F);
}


static opt_t optimize(s7_scheme *sc, s7_pointer code, int32_t hop, s7_pointer e)
{
  s7_pointer x;
  /* fprintf(stderr, "optimize: %s %s\n", DISPLAY_80(code), DISPLAY(e)); */
  if (sc->safety > ALL_OPTIMIZATION_SAFETY) return(OPT_F);
  for (x = code; (is_pair(x)) && (!is_checked(x)); x = cdr(x))
    {
      set_checked(x);
      if ((is_pair(car(x))) && (!is_checked(car(x))))
	{
	  /* fprintf(stderr, "   e in: %s\n", DISPLAY(e)); */
	  if (optimize_expression(sc, car(x), hop, e) == OPT_OOPS)
	    return(OPT_OOPS);
	  /* fprintf(stderr, "   e out: %s\n", DISPLAY(e)); */
	}
    }
  handle_optimizer_fixups(sc);
  if ((!is_null(x)) &&
      (!is_pair(x)))
    eval_error_no_return(sc, sc->syntax_error_symbol, "stray dot in function body: ~S", code);
  return(OPT_F);
}


/* ---------------------------------------- error checks ---------------------------------------- */

#define goto_START 0
#define goto_BEGIN1 1
#define fall_through 2
#define goto_DO_END_CLAUSES 3
#define goto_SAFE_DO_END_CLAUSES 4
#define goto_OPT_EVAL 5
#define goto_START_WITHOUT_POP_STACK 6
#define goto_EVAL 7
#define goto_APPLY 8
#define goto_EVAL_ARGS 9
#define goto_DO_UNCHECKED 10

static s7_pointer check_lambda_args(s7_scheme *sc, s7_pointer args, int32_t *arity)
{
  s7_pointer x;
  int32_t i;
  /* fprintf(stderr, "check lambda args %s\n", DISPLAY(args)); */

  if ((!is_pair(args)) && (!is_null(args)))
    {
      if (s7_is_constant(args))                       /* (lambda :a ...) */
	eval_error(sc, "lambda parameter '~S is a constant", args); /* not ~A here, (lambda #\null do) for example */

      /* we currently accept (lambda i i . i) (lambda quote i)  (lambda : : . #()) (lambda : 1 . "")
       *   at this level, but when the lambda form is evaluated, it will trigger an error.
       */
      if (is_symbol(args))
	set_local(args);

      if (arity) (*arity) = -1;
      return(sc->F);
    }

  for (i = 0, x = args; is_pair(x); i++, x = cdr(x))
    {
      s7_pointer car_x;
      car_x = car(x);
      if (s7_is_constant(car_x))                       /* (lambda (pi) pi), constant here means not a symbol */
	{
	  if (is_pair(car_x))                          /* (lambda ((:hi . "hi") . "hi") 1) */
	    eval_error(sc, "lambda parameter '~S is a pair (perhaps you want define* or lambda*?)", car_x);
	  eval_error(sc, "lambda parameter '~S is a constant", car_x);
	}
      if (symbol_is_in_arg_list(car_x, cdr(x)))       /* (lambda (a a) ...) or (lambda (a . a) ...) */
	eval_error(sc, "lambda parameter '~S is used twice in the parameter list", car_x);
      set_local(car_x);
    }
  if (is_not_null(x))
    {
      if (s7_is_constant(x))                             /* (lambda (a . 0.0) a) or (lambda (a . :b) a) */
	eval_error(sc, "lambda :rest parameter '~S is a constant", x);
      i = -i - 1;
    }

  if (arity) (*arity) = i;
  return(sc->F);
}


static s7_pointer check_lambda_star_args(s7_scheme *sc, s7_pointer args, int32_t *arity)
{
  s7_pointer top, v, w;
  int32_t i;

  if (!s7_is_list(sc, args))
    {
      if (s7_is_constant(args))                                  /* (lambda* :a ...) */
	eval_error(sc, "lambda* parameter '~S is a constant", args);
      if (is_symbol(args))
	set_local(args);
      if (arity) (*arity) = -1;
      return(args);
    }

  top = args;
  v = args;
  for (i = 0, w = args; is_pair(w); i++, v = w, w = cdr(w))
    {
      s7_pointer car_w;
      car_w = car(w);
      if (is_pair(car_w))
	{
	  if (s7_is_constant(car(car_w)))                            /* (lambda* ((:a 1)) ...) */
	    eval_error(sc, "lambda* parameter '~A is a constant", car(car_w));
	  if (symbol_is_in_arg_list(caar(w), cdr(w)))             /* (lambda* ((a 1) a) ...) */
	    eval_error(sc, "lambda* parameter '~A is used twice in the argument list", car(car_w));

	  if (!is_pair(cdr(car_w)))                                  /* (lambda* ((a . 0.0)) a) */
	    {
	      if (is_null(cdr(car_w)))                               /* (lambda* ((a)) ...) */
		eval_error(sc, "lambda* parameter default value missing? '~A", car_w);
	      eval_error(sc, "lambda* parameter is a dotted pair? '~A", car_w);
	    }
	  else
	    {
	      if ((is_pair(cadr(car_w))) &&                          /* (lambda* ((a (quote . -1))) ...) */
		  (s7_list_length(sc, cadr(car_w)) < 0))
		eval_error(sc, "lambda* parameter default value is improper? ~A", car_w);
	    }

	  if (is_not_null(cddr(car_w)))                              /* (lambda* ((a 0.0 'hi)) a) */
	    eval_error(sc, "lambda* parameter has multiple default values? '~A", car_w);

	  set_local(car(car_w));
	}
      else
	{
	  if (car_w != sc->key_rest_symbol)
	    {
	      if (s7_is_constant(car_w))
		{
		  if (car_w == sc->key_allow_other_keys_symbol)
		    {
		      if (is_not_null(cdr(w)))                /* (lambda* (:allow-other-keys x) x) */
			eval_error(sc, ":allow-other-keys should be the last parameter: ~A", args);
		      if (w == top)
			eval_error(sc, ":allow-other-keys can't be the only parameter: ~A", args);
		      set_allow_other_keys(top);
		      set_cdr(v, sc->nil);
		    }
		  else                                        /* (lambda* (pi) ...) */
		    eval_error(sc, "lambda* parameter '~A is a constant", car_w);
		}
	      if (symbol_is_in_arg_list(car_w, cdr(w)))       /* (lambda* (a a) ...) or (lambda* (a . a) ...) */
		eval_error(sc, "lambda* parameter '~A is used twice in the argument list", car_w);

	      if (!is_keyword(car_w)) set_local(car_w);
	    }
	  else
	    {
	      if (!is_pair(cdr(w)))                               /* (lambda* (:rest) ...) */
		eval_error(sc, "lambda* :rest parameter missing? ~A", w);
	      if (!is_symbol(cadr(w)))                         /* (lambda* (:rest (a 1)) ...) */
		{
		  if (!is_pair(cadr(w)))                          /* (lambda* (:rest 1) ...) */
		    eval_error(sc, "lambda* :rest parameter is not a symbol? ~A", w);
		  eval_error(sc, "lambda* :rest parameter can't have a default value. ~A", w);
		}
	      else
		{
		  if (is_immutable_symbol(cadr(w)))
		    return(s7_error(sc, sc->wrong_type_arg_symbol, set_elist_2(sc, s7_make_string_wrapper(sc, "can't bind an immutable object: ~S"), w)));
		}
	      set_local(cadr(w));
	    }
	}
    }
  if (is_not_null(w))
    {
      if (s7_is_constant(w))                             /* (lambda* (a . 0.0) a) or (lambda* (a . :b) a) */
	eval_error(sc, "lambda* :rest parameter '~A is a constant", w);
      if (is_symbol(w))
	set_local(w);
      i = -1;
    }
  if (arity) (*arity) = i;
  return(top);
}

/* sort(etc) lambda form is ok -- maybe unsafe? there are about a dozen of these
 * in tsort does do init (vc i) cause trouble (vc is arg) -- not localizable?
 */

typedef enum {UNSAFE_BODY=0, SAFE_BODY=1, VERY_SAFE_BODY=2} body_t;
static body_t min_body(body_t b1, body_t b2) {return((b1 < b2) ? b1 : b2);}

typedef struct slist {s7_pointer sym; struct slist *next;} slist;
static slist *syms_free_list = NULL;

static inline slist *add_sym(s7_scheme *sc, s7_pointer symbol, slist *lst)
{
  slist *top;
  if (syms_free_list)
    {
      top = syms_free_list;
      syms_free_list = syms_free_list->next;
    }
  else top = (slist *)malloc(sizeof(slist));
  top->sym = add_symbol_to_list(sc, symbol);
  top->next = lst;
  return(top);
}

static bool memq_sym(s7_scheme *sc, s7_pointer symbol, slist *top)
{
  slist *p;
  if (symbol_is_in_list(sc, symbol))
    {
      for (p = top; p; p = p->next)
	if (p->sym == symbol)
	  return(true);
    }
  return(false);
}

/* #define cancel_sym(Sc, Symbol, Top) cancel_sym_1(Sc, Symbol, Top, x, __LINE__) */
/* static void cancel_sym_1(s7_scheme *sc, s7_pointer symbol, slist *top, s7_pointer x, int32_t line) */
static void cancel_sym(s7_scheme *sc, s7_pointer symbol, slist *top)
{
  slist *p;
  if (symbol_is_in_list(sc, symbol))
    {
      for (p = top; p; p = p->next)
	if (p->sym == symbol)
	  p->sym = sc->gc_nil;
    }
}

static void free_syms(slist *top)
{
  if (top)
    {
      slist *p;
      for (p = top; p->next; p = p->next);
      p->next = syms_free_list;
      syms_free_list = top;
    }
}

#define WITH_SYMS_PRINT 0
#if WITH_SYMS_PRINT
static void display_syms(s7_scheme *sc, slist *top)
{
  if (top)
    {
      slist *p;
      fprintf(stderr, "<slist:");
      for (p = top; p; p = p->next)
	fprintf(stderr, " %s", DISPLAY(p->sym));
      fprintf(stderr, ">");
    }
  else fprintf(stderr, "<slist>");
}
#endif

static slist *split_slist(slist *top, slist *main_args)
{
  slist *p;
  if (top == main_args)
    return(NULL);
  for (p = top; p->next != main_args; p = p->next);
  p->next = NULL;
  return(top);
}


static inline void set_all_locals(s7_scheme *sc, s7_pointer tree, slist *args)
{
  s7_pointer p;
#if WITH_SYMS_PRINT
  fprintf(stderr, "set_all_locals: %s: ", DISPLAY_80(tree));
  display_syms(sc, args);
  fprintf(stderr, "\n");
#endif
  for (p = tree; is_pair(p); p = cdr(p))
    {
      s7_pointer cp;
      cp = car(p);
      if (is_symbol(cp))
	{
	  if ((memq_sym(sc, cp, args)) || 
	      ((is_immutable(cp)) &&                         /* immutable (by itself) would work except for tricky cases like with-let (no local_slot! -- changed) */
	       (unchecked_type(local_slot(cp)) == T_SLOT) && /* local_slot might be a free cell (so debugging free cell check is annoying) */
	       ((is_number(slot_value(local_slot(cp)))) ||
		(is_sequence(slot_value(local_slot(cp)))))))
	    set_local_symbol(p);
	}
      else 
	{
	  if (is_pair(cp))
	    set_all_locals(sc, cp, args);
	}
    }
}

static body_t body_is_safe(s7_scheme *sc, s7_pointer func, s7_pointer body, slist *main_args, bool at_end);
#define SAFE_FORM_PRINT 0

static body_t form_is_safe(s7_scheme *sc, s7_pointer func, s7_pointer x, slist *main_args, bool at_end)
{
  s7_pointer expr;
  body_t result = VERY_SAFE_BODY;

  if (!is_pair(x)) return(result);
  expr = car(x);
  if (is_syntactic_symbol(expr))
    {
      if (!is_pair(cdr(x))) return(UNSAFE_BODY);
      switch (symbol_syntax_op(expr))
	{
	case OP_OR:
	case OP_AND:
	case OP_BEGIN:
	case OP_WITH_BAFFLE:
	  return(body_is_safe(sc, func, cdr(x), main_args, at_end));
	  
	case OP_MACROEXPAND:
	  return(UNSAFE_BODY);
	  
	case OP_QUOTE:
	  return(VERY_SAFE_BODY);
	  
	case OP_IF:
	  if (!is_pair(cddr(x))) return(UNSAFE_BODY);
	  if (is_pair(cadr(x)))
	    {
	      result = form_is_safe(sc, func, cadr(x), main_args, false);
	      if (result == UNSAFE_BODY) 
		return(UNSAFE_BODY);
	    }
	  if (is_pair(caddr(x)))
	    {
	      result = min_body(result, form_is_safe(sc, func, caddr(x), main_args, at_end));
	      if (result == UNSAFE_BODY) 
		return(UNSAFE_BODY);
	    }
	  if ((is_pair(cdddr(x))) &&
	      (is_pair(cadddr(x))))
	    return(min_body(result, form_is_safe(sc, func, cadddr(x), main_args, at_end)));
	  return(result);
	  break;
	  
	case OP_WHEN:
	case OP_UNLESS:
	  if (!is_pair(cddr(x))) return(UNSAFE_BODY);
	  if (is_pair(cadr(x)))
	    {
	      result = form_is_safe(sc, func, cadr(x), main_args, false);
	      if (result == UNSAFE_BODY) 
		return(UNSAFE_BODY);
	    }
	  return(min_body(result, body_is_safe(sc, func, cddr(x), main_args, at_end)));
	  break;

	case OP_COND:
	  {
	    bool follow = false;
	    s7_pointer sp, p;
	    for (p = cdr(x), sp = x; is_pair(p); p = cdr(p))
	      {
		s7_pointer ex;
		ex = car(p);
		if (!is_pair(ex)) 
		  return(UNSAFE_BODY);
		if (is_pair(car(ex)))
		  {
		    result = min_body(result, form_is_safe(sc, func, car(ex), main_args, false));
		    if (result == UNSAFE_BODY) 
		      return(UNSAFE_BODY);
		  }
		if (is_pair(cdr(ex)))
		  {
		    result = min_body(result, body_is_safe(sc, func, cdr(ex), main_args, at_end));
		    if (result == UNSAFE_BODY) 
		      return(UNSAFE_BODY);
		  }
		if (follow) {sp = cdr(sp); if (p == sp) return(UNSAFE_BODY);}
		follow = (!follow);
	      }
	    if (is_not_null(p))
	      return(UNSAFE_BODY);
	    return(result);
	  }
	  break;
	  
	case OP_CASE:
	  {
	    bool follow = false;
	    s7_pointer sp, p;
	    if (!is_pair(cddr(x))) return(UNSAFE_BODY);
	    if (is_pair(cadr(x)))
	      {
		result = form_is_safe(sc, func, cadr(x), main_args, false);
		if (result == UNSAFE_BODY) 
		  return(UNSAFE_BODY);
	      }
	    sp = cdr(x);
	    p = cdr(sp);
	    for (; is_pair(p); p = cdr(p))
	      {
		if (!is_pair(car(p))) return(UNSAFE_BODY);
		if (is_pair(cdar(p)))
		  {
		    result = min_body(result, body_is_safe(sc, func, cdar(p), main_args, at_end)); /* null cdar(p) ok here */
		    if (result == UNSAFE_BODY) 
		      return(UNSAFE_BODY);
		  }
		if (follow) {sp = cdr(sp); if (p == sp) return(UNSAFE_BODY);}
		follow = (!follow);
	      }
	    return(result);
	  }
	  break;
	  
	case OP_SET:
	  /* if we set func, we have to make sure we abandon the tail call scan:
	   * (let () (define (hi a) (let ((v (vector 1 2 3))) (set! hi v) (hi a))) (hi 1))
	   */
	  if (!is_pair(cddr(x))) return(UNSAFE_BODY);
	  if (cadr(x) == func)
	    return(UNSAFE_BODY);
	  
	  /* car(x) is set!, cadr(x) is settee or obj, caddr(x) is val */
	  if (is_pair(caddr(x)))
	    {
	      result = form_is_safe(sc, func, caddr(x), main_args, false);
	      if (result == UNSAFE_BODY) 
		return(UNSAFE_BODY);
	    }
	  if (is_pair(cadr(x)))
	    return(min_body(result, form_is_safe(sc, func, cadr(x), main_args, false)));
	  return(result);

	  /* not OP_DEFINE even in simple cases (safe_closure assumes constant funclet) */

	case OP_WITH_LET:
	  if (!is_pair(cddr(x))) return(UNSAFE_BODY);
	  if (is_pair(cadr(x)))
	    return(UNSAFE_BODY);
	  return(min_body(body_is_safe(sc, sc->F, cddr(x), main_args, at_end), SAFE_BODY));
	  /* shadowing can happen in with-let -- symbols are global so local_slots are shadowable */
	  break;

	case OP_LET_TEMPORARILY:
	  {
	    s7_pointer p;
	    if (!is_pair(cadr(x))) return(UNSAFE_BODY);
	    for (p = cadr(x); is_pair(p); p = cdr(p))
	      {
		if ((!is_pair(car(p))) ||
		    (!is_pair(cdar(p))))
		  return(UNSAFE_BODY);
		result = min_body(result, form_is_safe(sc, sc->F, cadar(p), main_args, false));
		if (result == UNSAFE_BODY)
		  return(UNSAFE_BODY);
	      }
	    return(min_body(result, body_is_safe(sc, sc->F, cddr(x), main_args, at_end)));
	  }

	  /* in the name binders, we first have to check that "func" actually is the same thing as the caller's func */
	case OP_LET:
	case OP_LET_STAR:
	case OP_LETREC:
	case OP_LETREC_STAR:
	  {
	    bool follow = false;
	    s7_pointer vars, body, let_name, sp;
	    slist *top, *locals;
	    top = main_args;
	    vars = cadr(x);
	    body = cddr(x);
	    if (is_symbol(vars))
	      {
		if (!is_pair(body)) return(UNSAFE_BODY); /* (let name . res) */
		if (vars == func)                        /* named let shadows caller */
		  {
		    free_syms(split_slist(top, main_args));
		    return(UNSAFE_BODY);
		  }
		let_name = vars;
		vars = caddr(x);
		body = cdddr(x);
	      }
	    else let_name = func;

	    for (sp = NULL; is_pair(vars); vars = cdr(vars))
	      {
		s7_pointer let_var, var_name;
		
		let_var = car(vars);
		if ((!is_pair(let_var)) ||
		    (!is_pair(cdr(let_var))))
		  {
		    free_syms(split_slist(top, main_args));
		    return(UNSAFE_BODY);
		  }
		var_name = car(let_var);
		if ((!is_symbol(var_name)) ||
		    (var_name == let_name) || /* let var shadows caller */
		    (var_name == func))
		  {
		    free_syms(split_slist(top, main_args));
		    return(UNSAFE_BODY); 
		  }

		cancel_sym(sc, var_name, top);
		top = add_sym(sc, var_name, top);
		
		if (is_pair(cadr(let_var)))
		  {
		    result = min_body(result, form_is_safe(sc, let_name, cadr(let_var), top, false));
		    if (result == UNSAFE_BODY)
		      {
			free_syms(split_slist(top, main_args));
			return(UNSAFE_BODY);
		      }
		  }
		follow = (!follow);
		if (follow)
		  {
		    if (!sp)
		      sp = vars;
		    else
		      {
			sp = cdr(sp); 
			if (vars == sp) return(UNSAFE_BODY);
		      }
		  }
	      }
	    result = min_body(result, body_is_safe(sc, let_name, body, top, at_end));
	    locals = split_slist(top, main_args);
	    if (locals)
	      {
		/* fprintf(stderr, "form_is_safe %d %s: %d\n", __LINE__, DISPLAY_80(body), result); */
		if (result == VERY_SAFE_BODY)
		  set_all_locals(sc, body, locals);
		free_syms(locals);
	      }
	    return(result);
	  }
	  break;

	case OP_DO:	  /* (do (...) (...) ...) */
	  {
	    slist *top, *locals;
	    top = main_args;
	    if (!is_pair(cddr(x))) return(UNSAFE_BODY);
	    if (is_pair(cadr(x)))
	      {
		bool follow = false;
		s7_pointer vars, sp;
		vars = cadr(x);
		sp = vars;
		for (; is_pair(vars); vars = cdr(vars))
		  {
		    s7_pointer do_var;
		    do_var = car(vars);
		    if ((!is_pair(do_var)) ||
			(!is_pair(cdr(do_var))) ||   /* (do ((a . 1) (b . 2)) ...) */
			(car(do_var) == func) ||
			(!is_symbol(car(do_var))))
		      {
			free_syms(split_slist(top, main_args));
			return(UNSAFE_BODY);
		      }
		    
		    cancel_sym(sc, car(do_var), top);
		    top = add_sym(sc, car(do_var), top);
		    
		    result = min_body(result, form_is_safe(sc, func, cadr(do_var), top, false));
		    if (is_pair(cddr(do_var)))
		      result = min_body(result, form_is_safe(sc, func, caddr(do_var), top, false));
		    if (result == UNSAFE_BODY) 
		      {
			free_syms(split_slist(top, main_args));
			return(UNSAFE_BODY);
		      }
		    if (sp != vars)
		      {
			if (follow) {sp = cdr(sp); if (vars == sp) return(UNSAFE_BODY);}
			follow = (!follow);
		      }
		  }
	      }
	    if (is_pair(caddr(x)))
	      result = min_body(result, body_is_safe(sc, func, caddr(x), top, at_end));
	    result = min_body(result, body_is_safe(sc, func, cdddr(x), top, false));

	    locals = split_slist(top, main_args);
	    if (locals)
	      {
		/* fprintf(stderr, "form_is_safe %d %s: %d\n", __LINE__, DISPLAY_80(x), result); */
		if (result == VERY_SAFE_BODY)
		  {
		    s7_pointer vars;
		    for (vars = cadr(x); is_pair(vars); vars = cdr(vars))
		      {
			s7_pointer do_var;
			do_var = car(vars);
			if (is_pair(cddr(do_var)))
			  set_all_locals(sc, caddr(do_var), locals);
		      }
		    set_all_locals(sc, cddr(x), locals);
		  }
		free_syms(locals);
	      }

	    return(result);
	  }
	  break;
	  
	  /* define and friends are not safe: (define (a) (define b 3)...) tries to put b in the current env,
	   *   but in a safe func, that's a constant.  See s7test L 1865 for an example.
	   */
	  /* TODO: how to tell that the temp env can be freed? -- check results for func/macro/let? 
	   *   not captured? but how to catch that?
	   */
	default:
	  /* try to catch weird cases like:
	   * (let () (define (hi1 a) (define (hi1 b) (+ b 1)) (hi1 a)) (hi1 1))
	   * (let () (define (hi1 a) (define (ho1 b) b) (define (hi1 b) (+ b 1)) (hi1 a)) (hi1 1))
	   */
	  return(UNSAFE_BODY);
	}
    }
  else /* car(x) is not syntactic ?? */
    {
      if (expr == func) /* try to catch tail call, expr is car(x) */
	{
	  bool follow = false;
	  s7_pointer sp, p;
	  p = cdr(x);
	  sp = x;
	  for (; is_pair(p); p = cdr(p))
	    {
	      if (is_pair(car(p)))
		{
		  if (caar(p) == func)    /* func called as arg, so not tail call */
		    return(UNSAFE_BODY);
		  result = min_body(result, form_is_safe(sc, func, car(p), main_args, false));
		  if (result == UNSAFE_BODY)
		    return(UNSAFE_BODY);
		}
	      else
		{
		  if (car(p) == func)
		    return(UNSAFE_BODY);
		}
	      if (follow) {sp = cdr(sp); if (p == sp) return(UNSAFE_BODY);}
	      follow = (!follow);
	    }
	  if ((at_end) && (is_null(p))) /* tail call, so safe */
	    return(result); 
	  return(UNSAFE_BODY);
	}
      
      if (is_symbol(expr)) /* expr=car(x) */
	{
	  s7_pointer f, f_slot;
	  bool c_safe;

	  if (memq_sym(sc, expr, main_args)) /* it's defined locally (so don't use find_symbol) */
	    return(UNSAFE_BODY);

	  f_slot = find_symbol(sc, expr);
	  if (!is_slot(f_slot))
	    return(UNSAFE_BODY);
	  f = slot_value(f_slot);
	  c_safe = (is_c_function(f)) && (is_safe_or_scope_safe_procedure(f));
	  result = ((is_sequence(f)) || 
		    ((c_safe) && (is_global(expr)))) ? VERY_SAFE_BODY : SAFE_BODY;
#if SAFE_FORM_PRINT
	  fprintf(stderr, "%d: %s: %d %d %s (scope: %d)\n", __LINE__, DISPLAY(f), result, is_global(expr), describe_type_bits(sc, f), c_safe);
#endif
	  if ((c_safe) ||
	      ((is_any_closure(f)) && (is_safe_closure(f))) ||
	      (is_sequence(f)))
	    {
	      bool follow = false;
	      s7_pointer sp, p;
	      p = cdr(x);
	      sp = x;
	      for (; is_pair(p); p = cdr(p))
		{
		  if ((is_pair(car(p))) &&
		      (caar(p) != sc->quote_symbol))
		    {
		      if (caar(p) == func)
			return(UNSAFE_BODY);

		      if ((is_c_function(f)) && (is_scope_safe(f)) &&
			  (caar(p) == sc->lambda_symbol))
			{
			  s7_pointer largs, lbody, q;
			  slist *top, *locals;
			  body_t lresult;

			  if (!is_pair(cdar(p))) /* (lambda . /) */
			    return(UNSAFE_BODY);
			  largs = cadar(p);
			  top = main_args;
			  lbody = cddar(p);
			  for (q = largs; is_pair(q); q = cdr(q))
			    {
			      if (!is_symbol(car(q)))
				{
				  free_syms(split_slist(top, main_args));
				  return(UNSAFE_BODY); 
				}
			      cancel_sym(sc, car(q), top);
			      top = add_sym(sc, car(q), top);
			    }
			  lresult = body_is_safe(sc, func, lbody, top, false);
			  locals = split_slist(top, main_args);
			  if (locals)
			    {
#if SAFE_FORM_PRINT
			      fprintf(stderr, "%d: %s result: %d\n", __LINE__, DISPLAY(lbody), result);
#endif
			      if (lresult == VERY_SAFE_BODY)
				set_all_locals(sc, lbody, locals);
			      free_syms(locals);
			    }
			  result = min_body(result, lresult);
			}
		      else 
			{
			  result = min_body(result, form_is_safe(sc, func, car(p), main_args, false));
#if SAFE_FORM_PRINT
			  fprintf(stderr, "%d: %s %d\n", __LINE__, DISPLAY_80(car(p)), result);
#endif
			}
#if SAFE_FORM_PRINT
		      fprintf(stderr, "%d: %s: %d\n", __LINE__, DISPLAY_80(car(p)), result);
#endif
		      if (result == UNSAFE_BODY)
			return(UNSAFE_BODY);
		    }
		  else
		    {
#if SAFE_FORM_PRINT
		      fprintf(stderr, "not a pair: %s\n", DISPLAY(car(p)));
#endif
		      if (car(p) == func)
			return(UNSAFE_BODY);
		    }
		  if (follow) {sp = cdr(sp); if (p == sp) return(UNSAFE_BODY);}
		  follow = (!follow);
		}
	      if (!is_null(p))
		return(UNSAFE_BODY);
	      return(result);
	    }
	}
      return(UNSAFE_BODY);
    }
  return(result);
}

static body_t body_is_safe(s7_scheme *sc, s7_pointer func, s7_pointer body, slist *main_args, bool at_end)
{
  bool follow = false;
  s7_pointer p, sp;
  body_t result = VERY_SAFE_BODY;
  for (p = body, sp = body; is_pair(p); p = cdr(p))
    {
      if (is_pair(car(p)))
	{
	  result = min_body(result, form_is_safe(sc, func, car(p), main_args, (at_end) && (is_null(cdr(p)))));
	  if (result == UNSAFE_BODY) return(UNSAFE_BODY);
	}
      if (p != body)
	{
	  if (follow) {sp = cdr(sp); if (p == sp) return(UNSAFE_BODY);}
	  follow = (!follow);
	}
    }
  if (!is_null(p))
    return(UNSAFE_BODY);
#if SAFE_FORM_PRINT
  fprintf(stderr, "%d: %d\n", __LINE__, result);
#endif
  return(result);
}

static s7_pointer optimize_lambda(s7_scheme *sc, bool unstarred_lambda, s7_pointer func, s7_pointer args, s7_pointer body)
{
  int32_t len;

  len = s7_list_length(sc, body);
  if (len < 0)                /* (define (hi) 1 . 2) */
    eval_error_with_caller(sc, "~A: function body messed up, ~A", (unstarred_lambda) ? sc->lambda_symbol : sc->lambda_star_symbol, sc->code);

  if (len > 0)  /* i.e. not circular */
    {
      body_t result;
      slist *sargs = NULL;
      s7_pointer p;

      clear_symbol_list(sc);
      for (p = args; is_pair(p); p = cdr(p))
	sargs = add_sym(sc, (is_symbol(car(p))) ? car(p) : caar(p), sargs);
      if (!is_null(p))
	sargs = add_sym(sc, p, sargs);
      result = body_is_safe(sc, func, body, sargs, true);
#if SAFE_FORM_PRINT
      fprintf(stderr, "%s: %d\n", DISPLAY_80(body), result);
#endif
      if (result == VERY_SAFE_BODY)
	{
	  if ((is_symbol(func)) &&
	      ((is_let(sc->envir)) || (symbol_id(func) == 0)))
	    sargs = add_sym(sc, func, sargs);
	  /* this is really tricky.  If a function has been defined earlier locally, it has no global slot.
	   *   If it is later defined globally, make_slot creates the global slot but does not touch the
	   *   (possibly free) local slot (it can't because as s7test 24981 shows, both can be in play).
	   *   If we than use the local_symbol mark set here to call (say) OP_LCLOSURE_A, it looks at
	   *   the local_slot.  If that is invalid (normally a free_cell), oops... So the code above,
	   *   checks for a global function that has been previously defined locally.  
	   */
#if SAFE_FORM_PRINT
	  fprintf(stderr, "set_locals: %s\n", DISPLAY_80(body));
#endif
	  set_all_locals(sc, body, sargs);
	}

      free_syms(sargs);
      clear_symbol_list(sc);  /* tracks locals */

      if (is_symbol(func))    /* func can be sc->gc_nil (see check_lambda and check_lambda_star) */
 	{
 	  s7_pointer lst;
 	  lst = cons(sc, add_symbol_to_list(sc, func), sc->nil);
	  sc->temp10 = lst;
 	  if (optimize(sc, body, 1, collect_parameters(sc, args, lst)) == OPT_OOPS)
	    clear_all_optimizations(sc, body);
	  sc->temp10 = sc->nil;
	  free_cell(sc, lst);
 	}
      else 
	{
	  /* fprintf(stderr, "%s[%d]: %s\n", __func__, __LINE__, DISPLAY(body)); */
	  if (optimize(sc, body, 1, collect_parameters(sc, args, sc->nil)) == OPT_OOPS)
	    clear_all_optimizations(sc, body);
	}

      /* if the body is safe, we can optimize the calling sequence */
      if (!arglist_has_rest(sc, args))
	{
	  if (!unstarred_lambda)
	    {
	      s7_pointer p;
	      bool happy = true;
	      /* check default vals -- if none is an expression or symbol, set simple args */
	      for (p = args; is_pair(p); p = cdr(p))
		{
		  s7_pointer arg;
		  arg = car(p);
		  if ((is_pair(arg)) &&                   /* has default value */
		      (is_pair(cdr(arg))) &&              /*   is not a ridiculous improper list */
		      ((is_symbol(cadr(arg))) ||          /*   if default value might involve eval in any way, it isn't simple */
                      ((is_pair(cadr(arg))) &&            /*   pair as default only ok if it is (quote ...) */
                       (caadr(arg) !=  sc->quote_symbol))))
		    {
		      happy = false;
		      break;
		    }
		}
	      if (happy)
		set_simple_arg_defaults(body);
	    }
	  if (result != UNSAFE_BODY)
	    {
#if SAFE_FORM_PRINT
	      fprintf(stderr, "%s is safe\n", DISPLAY_80(body));
#endif
	    set_safe_closure(body);
	  /* this bit is set on the function itself in make_closure and friends */
	    }
	}
      /* else fprintf(stderr, "rest: %d %s\n", result, DISPLAY(body)); */
    }
  return(NULL);
}

static void check_lambda(s7_scheme *sc)
{
  /* code is a lambda form minus the "lambda": ((a b) (+ a b)) */
  /* this includes unevaluated symbols (direct symbol table refs) in macro arg list */
  s7_pointer code, body;

  code = sc->code;
  if (!is_pair(code))                                 /* (lambda) or (lambda . 1) */
    eval_error_no_return(sc, sc->syntax_error_symbol, "lambda: no args? ~A", current_code(sc));

  body = cdr(code);
  if (!is_pair(body))                                 /* (lambda #f) */
    eval_error_no_return(sc, sc->syntax_error_symbol, "lambda: no body? ~A", code);
  
  /* in many cases, this is a no-op -- we already checked at define */
  check_lambda_args(sc, car(code), NULL);
  clear_symbol_list(sc); /* tracks locals to this function */
  
  /* look for (define f (let (...) (lambda ...))) and treat as equivalent to (define (f ...)...)
   *   one problem the hop=0 fixes is that safe closures assume the old frame exists, so we need to check for define below
   *   I wonder about apply define...
   */
  if ((main_stack_op(sc) == OP_DEFINE1) ||
      (((sc->stack_end - sc->stack_start) > 4) &&
       (((opcode_t)(sc->stack_end[-5])) == OP_DEFINE1) &&  /* surely if define is ok, so is define dilambda? 16-Apr-16 */
       (sc->op_stack_now > sc->op_stack) &&
       ((*(sc->op_stack_now - 1)) == (s7_pointer)slot_value(global_slot(sc->dilambda_symbol)))))
    optimize_lambda(sc, true, sc->gc_nil, car(code), body); /* why was lambda the func? */
  else 
    {
      /* fprintf(stderr, "%s[%d]: %s %s\n", __func__, __LINE__, DISPLAY_80(body), (sc->op_stack_now > sc->op_stack) ? DISPLAY(*(sc->op_stack_now - 1)) : "??"); */
      if (optimize(sc, body, 0,
		   /* ((sc->op_stack_now > sc->op_stack) && (is_c_function((*(sc->op_stack_now - 1)))) && (is_scope_safe((*(sc->op_stack_now - 1))))) ? 1 : 0, */
		   /* this works except when someone resets outlet(curlet) after defining a local function! */
		   collect_parameters(sc, car(sc->code), sc->nil)) == OPT_OOPS)
	clear_all_optimizations(sc, body);
    }

  if ((is_overlaid(code)) &&
      (has_opt_back(code)))
    pair_set_syntax_symbol(code, sc->lambda_unchecked_symbol);
}

static void check_lambda_star(s7_scheme *sc)
{
  if ((!is_pair(sc->code)) ||
      (!is_pair(cdr(sc->code))))                                          /* (lambda*) or (lambda* #f) */
    eval_error_no_return(sc, sc->syntax_error_symbol, "lambda*: no args or no body? ~A", sc->code);
  
  set_car(sc->code, check_lambda_star_args(sc, car(sc->code), NULL));
  clear_symbol_list(sc);
  
  if ((sc->safety > NO_SAFETY) ||
      (main_stack_op(sc) != OP_DEFINE1))
    {
      /* fprintf(stderr, "%s[%d]: %s\n", __func__, __LINE__, DISPLAY(cdr(sc->code))); */
      if (optimize(sc, cdr(sc->code), 0, collect_parameters(sc, car(sc->code), sc->nil)) == OPT_OOPS)
	clear_all_optimizations(sc, cdr(sc->code));
    }
  else optimize_lambda(sc, false, sc->gc_nil, car(sc->code), cdr(sc->code));
  
  if ((is_overlaid(sc->code)) &&
      (has_opt_back(sc->code)))
    pair_set_syntax_symbol(sc->code, sc->lambda_star_unchecked_symbol);
}


static s7_pointer check_case(s7_scheme *sc)
{
  bool keys_simple = true, has_feed_to = false, keys_single = true, bodies_simple = true;
  int32_t key_type = T_FREE;
  s7_pointer x;

  if (!is_pair(sc->code))                                            /* (case) or (case . 1) */
    eval_error(sc, "case has no selector:  ~A", sc->code);
  if (!is_pair(cdr(sc->code)))                                       /* (case 1) or (case 1 . 1) */
    eval_error(sc, "case has no clauses?:  ~A", sc->code);
  if (!is_pair(cadr(sc->code)))                                      /* (case 1 1) */
    eval_error(sc, "case clause is not a list? ~A", sc->code);
  set_opt_any3(sc->code, sc->unspecified);

  for (x = cdr(sc->code); is_pair(x); x = cdr(x))
    {
      s7_pointer y, car_x;
      if ((!is_pair(x)) ||                                           /* (case 1 ((2) 1) . 1) */
	  (!is_pair(car(x))))
	eval_error(sc, "case clause ~A messed up", x);
      car_x = car(x);

      if (!s7_is_list(sc, cdr(car_x)))                               /* (case 1 ((1))) */
	eval_error(sc, "case clause result messed up: ~A", car_x);

      if ((bodies_simple) &&
	  ((is_null(cdr(car_x))) || (!is_null(cddr(car_x)))))
	bodies_simple = false;
	
      y = car(car_x);
      if (!is_pair(y))
	{
	  if ((y != sc->else_symbol) &&                              /* (case 1 (2 1)) */
	      ((!is_symbol(y)) ||
	       (s7_symbol_value(sc, y) != sc->else_symbol)))         /* "proper list" below because: (case 1 (() 2) ... */
	    eval_error(sc, "case clause key list ~A is not a proper list or 'else'", y);
	  if (is_not_null(cdr(x)))                                  /* (case 1 (else 1) ((2) 1)) */
	    eval_error(sc, "case 'else' clause, ~A, is not the last clause", x);
	  if (is_null(cdr(car_x)))                                  /* (else) so return selector */
	    {
	      /* opt_any3?? */
	    }
	  else
	    {
	      if (is_pair(cddr(car_x)))
		{
		  set_opt_any3(sc->code, cdr(car_x));
		  /* fprintf(stderr, "else not simple\n"); */
		  bodies_simple = false;
		}
	      else
		{
		  if ((bodies_simple) && 
		      (keys_single))
		    set_opt_any3(sc->code, cadr(car_x));
		  else set_opt_any3(sc->code, cdr(car_x));
		  set_opt_clause(x, cadr(car_x));
		}
	    }
	}
      else
	{
	  if (!is_simple(car(y)))
	    keys_simple = false;
	  if (!is_null(cdr(y)))
	    keys_single = false;
	  if (key_type == T_FREE)
	    key_type = type(car(y));
	  else
	    {
	      if (key_type != type(car(y)))
		key_type = NUM_TYPES;
	    }

	  for (y = cdr(y); is_pair(y); y = cdr(y))
	    {
	      if (!is_simple(car(y)))
		keys_simple = false;
	      if (key_type != type(car(y)))
		key_type = NUM_TYPES;
	    }
	  if (!is_null(y))                                        /* (case () ((1 . 2) . hi) . hi) */
	    eval_error(sc, "case key list is improper? ~A", x);
	}
      y = car_x;
      if (!s7_is_proper_list(sc, cdr(y)))
	eval_error(sc, "case: stray dot? ~A", y);
      if ((is_pair(cdr(y))) &&
	  (cadr(y) == sc->feed_to_symbol) &&
	  (s7_symbol_value(sc, sc->feed_to_symbol) == sc->undefined))
	{
	  has_feed_to = true;
	  if (!is_pair(cddr(y)))                                  /* (case 1 (else =>)) */
	    eval_error(sc, "case: '=>' target missing?  ~A", y);
	  if (is_pair(cdddr(y)))                                  /* (case 1 (else => + - *)) */
	    eval_error(sc, "case: '=>' has too many targets: ~A", y);
	}
    }
  if (is_not_null(x))                                             /* (case x ((1 2)) . 1) */
    eval_error(sc, "case: stray dot? ~A", sc->code);

  if ((is_overlaid(sc->code)) &&
      (has_opt_back(sc->code)))
    {
      if ((keys_single) &&
	  (bodies_simple))
	{
	  for (x = cdr(sc->code); is_not_null(x); x = cdr(x))
	    {
	      set_opt_any2(x, caar(x));
	      if (is_pair(opt_any2(x)))
		{
		  set_opt_any2(x, car(opt_any2(x)));
		  if (is_pair(cdar(x)))
		    set_opt_clause(x, cadar(x));
		}
	    }
	}
      else
	{
	  for (x = cdr(sc->code); is_not_null(x); x = cdr(x))
	    {
	      set_opt_any2(x, caar(x));
	      if ((is_pair(opt_any2(x))) &&
		  (is_pair(cdar(x))))
		set_opt_clause(x, cadar(x));
	    }
      }
      
      pair_set_syntax_symbol(sc->code, sc->case_unchecked_symbol);
      if ((has_feed_to) ||
	  (!bodies_simple) ||  /* x_x_g g=general keys or bodies */
	  (!keys_single))
	{
	  if (!keys_simple)  /* x_g_g (no int32_t case here) */
	    {
	      if (is_symbol(car(sc->code)))
		pair_set_syntax_symbol(sc->code, sc->case_s_g_g_symbol);
	      else
		{
		  if (is_all_x_safe(sc, car(sc->code)))
		    {
		      pair_set_syntax_symbol(sc->code, sc->case_a_g_g_symbol);
		      set_x_call(sc->code, all_x_eval(sc, sc->code, sc->envir, let_symbol_is_safe));
		    }
		  else pair_set_syntax_symbol(sc->code, sc->case_p_g_g_symbol);
		}
	    }
	  else             /* x_e_g */
	    {
	      if (is_symbol(car(sc->code)))
		pair_set_syntax_symbol(sc->code, sc->case_s_e_g_symbol);
	      else
		{
		  if (is_all_x_safe(sc, car(sc->code)))
		    {
		      pair_set_syntax_symbol(sc->code, sc->case_a_e_g_symbol);
		      set_x_call(sc->code, all_x_eval(sc, sc->code, sc->envir, let_symbol_is_safe));
		    }
		  else pair_set_syntax_symbol(sc->code, sc->case_p_e_g_symbol);
		}
	    }
	}
      else                /* x_x_s */
	{
	  if (!keys_simple)  /* x_g|i_s */
	    {
	      if (is_symbol(car(sc->code)))
		pair_set_syntax_symbol(sc->code, (key_type == T_INTEGER) ? sc->case_s_i_s_symbol: sc->case_s_g_s_symbol);
	      else
		{
		  if (is_all_x_safe(sc, car(sc->code)))
		    {
		      pair_set_syntax_symbol(sc->code, (key_type == T_INTEGER) ? sc->case_a_i_s_symbol : sc->case_a_g_s_symbol);
		      set_x_call(sc->code, all_x_eval(sc, sc->code, sc->envir, let_symbol_is_safe));
		    }
		  else pair_set_syntax_symbol(sc->code, (key_type == T_INTEGER) ? sc->case_p_i_s_symbol : sc->case_p_g_s_symbol);
		}
	    }
	  else             /* x_e_s */
	    {
	      if (is_symbol(car(sc->code)))
		pair_set_syntax_symbol(sc->code, sc->case_s_e_s_symbol);
	      else
		{
		  if (is_all_x_safe(sc, car(sc->code)))
		    {
		      pair_set_syntax_symbol(sc->code, sc->case_a_e_s_symbol);
		      set_x_call(sc->code, all_x_eval(sc, sc->code, sc->envir, let_symbol_is_safe));
		    }
		  else pair_set_syntax_symbol(sc->code, sc->case_p_e_s_symbol);
		}
	    }
	}
    }
  return(sc->code);
}

static s7_pointer check_let_one_var(s7_scheme *sc, s7_pointer start)
{
  s7_pointer binding;

  pair_set_syntax_symbol(sc->code, sc->let_one_symbol);
  binding = car(start);
  /* all-x-able body happens a few times, but to avoid the new frame we'd need to ensure it was safe etc */
  if (is_pair(cadr(binding)))
    {
      if (is_h_optimized(cadr(binding)))
	{
	  if (is_null(cddr(sc->code)))                   /* one statement body */
	    {
	      set_opt_sym2(cdr(sc->code), car(binding)); /* these don't collide -- cdr(code) and code */
	      set_opt_pair2(sc->code, cadr(binding));
	      pair_set_syntax_symbol(sc->code, sc->let_z_symbol);

	      if ((is_h_safe_c_s(cadr(binding))) &&
		  (is_pair(cadr(sc->code))))             /* one body expr is a pair */
		{
 		  pair_set_syntax_symbol(sc->code, sc->let_opsq_p_symbol);

		  set_opt_sym2(sc->code, cadadr(binding));
		  if ((!is_optimized(cadr(sc->code))) &&
		      (is_syntactic_symbol(caadr(sc->code))))
		    {
		      /* the is_optimized check here and in other parallel cases protects against cases like:
		       *   (define (hi) (let ((e #f)) (let ((val (not e))) (if (boolean? val) val e)))) (hi)
		       * where the "(if...)" part is optimized as safe_c_s before we get here.  If we simply
		       * pair_set_syntax_op(cadr(sc->code)) as below, the optimization bit is on, but the
		       * apparent optimize_op (op) is now safe_c_qq! So eval ejects it and it is handled by the
		       * explicit ("trailers") code.
		       * I think the optimize bit is now turned off by pair_set_syntax_op.
		       */
		      pair_set_syntax_op(cadr(sc->code), symbol_syntax_op(caadr(sc->code)));
		    }
		  return(sc->code);
		}

	      if (optimize_op(cadr(binding)) == HOP_SAFE_C_SS)
		{
		  set_opt_pair2(sc->code, cadr(binding));
		  pair_set_syntax_symbol(sc->code, sc->let_opssq_e_symbol);
		  if (c_call(cadr(binding)) == g_assq)
		    pair_set_syntax_symbol(sc->code, sc->let_opassq_e_symbol);
		  set_opt_sym3(sc->code, caddr(cadr(binding)));
		  return(sc->code);
		}
	      if (is_all_x_safe(sc, cadr(binding)))
		{
		  pair_set_syntax_symbol(sc->code, sc->let_a_symbol);
		  annotate_arg(sc, cdr(binding), sc->envir);
		  if (is_optimized(cadr(sc->code)))
		    pair_set_syntax_symbol(sc->code, sc->let_a_z_symbol);
		  return(sc->code);
		}
	    }
	  
	  if (is_h_safe_c_s(cadr(binding)))
	    {
	      pair_set_syntax_symbol(sc->code, (caadr(binding) == sc->car_symbol) ? sc->let_car_symbol : sc->let_opsq_symbol);
	      set_opt_sym2(sc->code, cadadr(binding));
	      return(sc->code);
	    }

	  set_opt_pair2(sc->code, cadr(binding));
	  if (optimize_op(cadr(binding)) == HOP_SAFE_C_SS)
	    {
	      pair_set_syntax_symbol(sc->code, sc->let_opssq_symbol);
	      set_opt_sym3(sc->code, caddr(cadr(binding)));
	    }
	  else
	    {
	      if (is_h_safe_c_c(cadr(binding)))
		{
		  set_opt_sym3(sc->code, car(binding));
		  pair_set_syntax_symbol(sc->code, sc->let_opcq_symbol);
		}
	    }
	}
    }
  else
    {
      s7_pointer p;
      p = cadaar(sc->code);                    /* sc->code is of the form '(((x y))...) */
      set_opt_sym3(sc->code, caaar(sc->code));
      if (is_symbol(p))
	{
	  set_opt_sym2(sc->code, p);
	  pair_set_syntax_symbol(sc->code, ((is_null(cddr(sc->code))) && (is_optimized(cadr(sc->code)))) ? sc->let_s_z_symbol : sc->let_s_symbol);
	}
      else 
	{
	  set_opt_con2(sc->code, p);
	  pair_set_syntax_symbol(sc->code, sc->let_c_symbol);
	}
    }
  return(sc->code);
}


static s7_pointer check_let(s7_scheme *sc)
{
  s7_pointer x, start;
  bool named_let;
  int32_t vars;

  if (!is_pair(sc->code))               /* (let . 1) */
    {
      if (is_null(sc->code))            /* (let) */
	eval_error(sc, "let has no variables or body: ~A", sc->code);
      eval_error(sc, "let form is an improper list? ~A", sc->code);
    }

  if (!is_pair(cdr(sc->code)))          /* (let () ) */
    eval_error(sc, "let has no body: ~A", sc->code);

  if ((!s7_is_list(sc, car(sc->code))) && /* (let 1 ...) */
      (!is_symbol(car(sc->code))))
    eval_error(sc, "let variable list is messed up or missing: ~A", sc->code);

  /* we accept these (other schemes complain, but I can't see why -- a no-op is the user's business!):
   *   (let () (define (hi) (+ 1 2)))
   *   (let () (begin (define x 3)))
   *   (let () 3 (begin (define x 3)))
   *   (let () (define x 3))
   *   (let () (if #t (define (x) 3)))
   *
   * similar cases:
   *   (case 0 ((0) (define (x) 3) (x)))
   *   (cond (0 (define (x) 3) (x)))
   *   (and (define (x) x) 1)
   *   (begin (define (x y) y) (x (define (x y) y)))
   *   (if (define (x) 1) 2 3)
   *   (do () ((define (x) 1) (define (y) 2)))
   *
   * but we can get some humorous results:
   *   (let ((x (lambda () 3))) (if (define (x) 4) (x) 0)) -> 4
   */

  named_let = (is_symbol(car(sc->code)));

  if (named_let)
    {
      if (!s7_is_list(sc, cadr(sc->code)))      /* (let hi #t) */
	eval_error(sc, "let variable list is messed up: ~A", sc->code);
      if (is_null(cddr(sc->code)))              /* (let hi () ) */
	eval_error(sc, "named let has no body: ~A", sc->code);
      if (!is_pair(cddr(sc->code)))             /* (let hi () . =>) */
	eval_error(sc, "named let stray dot? ~A", sc->code);
      if (is_immutable_symbol(car(sc->code)))
	return(s7_error(sc, sc->wrong_type_arg_symbol, set_elist_2(sc, s7_make_string_wrapper(sc, "can't bind an immutable object: ~S"), sc->code)));
      set_local(car(sc->code));
      start = cadr(sc->code);
    }
  else start = car(sc->code);

  clear_symbol_list(sc);
  for (vars = 0, x = start; is_pair(x); vars++, x = cdr(x))
    {
      s7_pointer y, carx;

      carx = car(x);

      if ((!is_pair(carx)) || (is_null(cdr(carx))))  /* (let ((x)) ...) or (let ((x 1) . (y 2)) ...) */
	eval_error(sc, "let variable declaration, but no value?: ~A", x);

      if (!(is_pair(cdr(carx))))                     /* (let ((x . 1))...) */
	eval_error(sc, "let variable declaration is not a proper list?: ~A", x);

      if (is_not_null(cddr(carx)))                   /* (let ((x 1 2 3)) ...) */
	eval_error(sc, "let variable declaration has more than one value?: ~A", x);

      /* currently if the extra value involves a read error, we get a kind of panicky-looking message:
       *   (let ((x . 2 . 3)) x)
       *   ;let variable declaration has more than one value?: (x error error "stray dot?: ...  ((x . 2 . 3)) x) ..")
       */

      y = car(carx);
      if (!(is_symbol(y)))
	eval_error(sc, "bad variable ~S in let", carx);

      if (is_immutable_symbol(y))
	return(s7_error(sc, sc->wrong_type_arg_symbol,	set_elist_2(sc, s7_make_string_wrapper(sc, "can't bind an immutable object: ~S"), x)));

      /* check for name collisions -- not sure this is required by Scheme */
      if (symbol_is_in_list(sc, y))
	eval_error(sc, "duplicate identifier in let: ~A", y);
      add_symbol_to_list(sc, y);
      set_local(y);
    }

  /* (let ('1) quote) -> 1 */

  if (is_not_null(x))                  /* (let* ((a 1) . b) a) */
    eval_error(sc, "let var list improper?: ~A", sc->code);

  if ((is_overlaid(sc->code)) &&
      (has_opt_back(sc->code)))
    {
      if (named_let)
	{
	  if (is_null(start))
	    pair_set_syntax_symbol(sc->code, sc->named_let_no_vars_symbol);
	  else pair_set_syntax_symbol(sc->code, sc->named_let_symbol);

	  /* this is (let name ...) so the initial values need to be removed from the closure arg list */
	  if (vars > 0)
	    {
	      s7_pointer ex, exp;
	      sc->args = safe_list_if_possible(sc, vars);
	      for (ex = start, exp = sc->args; is_pair(ex); ex = cdr(ex), exp = cdr(exp))
		car(exp) = caar(ex);
	      optimize_lambda(sc, true, car(sc->code), sc->args, cddr(sc->code));
	      clear_list_in_use(sc->args);
	      sc->current_safe_list = 0;
	    }
	  else optimize_lambda(sc, true, car(sc->code), sc->nil, cddr(sc->code));
	  return(sc->code);
	}

      if (is_null(start))
	pair_set_syntax_symbol(sc->code, sc->let_no_vars_symbol);
      else
	{
	  pair_set_syntax_symbol(sc->code, sc->let_unchecked_symbol);
	  if (is_null(cdr(start)))                                    /* one binding */
	    check_let_one_var(sc, start);
	  else
	    {
	      if (vars < GC_TRIGGER_SIZE)
		{
		  s7_pointer p, op;

		  op = sc->nil;
		  for (p = start; is_pair(p); p = cdr(p))
		    {
		      s7_pointer x;
		      x = car(p);
		      if (is_pair(cadr(x)))
			{
			  if (is_h_safe_c_s(cadr(x)))
			    {
			      if ((op == sc->nil) || (op == sc->let_all_opsq_symbol))
				op = sc->let_all_opsq_symbol;
			      else op = sc->let_all_x_symbol;
			    }
			  else
			    {
			      if (is_all_x_safe(sc, cadr(x)))
				op = sc->let_all_x_symbol;
			      else
				{
				  op = sc->let_unchecked_symbol;
				  break;
				}
			    }
			}
		      else
			{
			  if (is_symbol(cadr(x)))
			    {
			      if ((op == sc->nil) || (op == sc->let_all_s_symbol))
				op = sc->let_all_s_symbol;
			      else op = sc->let_all_x_symbol;
			    }
			  else
			    {
			      if ((op == sc->nil) || (op == sc->let_all_c_symbol))
				op = sc->let_all_c_symbol;
			      else op = sc->let_all_x_symbol;
			    }
			}
		    }
		  pair_set_syntax_symbol(sc->code, op);
		}
	      else pair_set_syntax_symbol(sc->code, sc->let_unchecked_symbol);
	    }
	}
      if (pair_syntax_symbol(sc->code) == sc->let_all_x_symbol)
	{
	  s7_pointer p;
	  for (p = start; is_pair(p); p = cdr(p))
	    set_x_call(cdar(p), all_x_eval(sc, cdar(p), sc->envir, let_symbol_is_safe));
	  /* all-x-able body does not happen much */
	}
    }
  return(sc->code);
}


static s7_pointer check_let_star(s7_scheme *sc)
{
  s7_pointer vars;
  bool named_let;

  if (!is_pair(sc->code))                           /* (let* . 1) */
    eval_error(sc, "let* variable list is messed up: ~A", sc->code);
  if (!is_pair(cdr(sc->code)))                      /* (let*) */
    eval_error(sc, "let* variable list is messed up: ~A", sc->code);

  named_let = (is_symbol(car(sc->code)));

  if (named_let)
    {
      if (!s7_is_list(sc, cadr(sc->code)))          /* (let* hi #t) */
	eval_error(sc, "let* variable list is messed up: ~A", sc->code);
      if (is_null(cddr(sc->code)))                  /* (let* hi () ) */
	eval_error(sc, "named let* has no body: ~A", sc->code);
      if (!is_pair(cddr(sc->code)))                 /* (let* hi () . =>) */
	eval_error(sc, "named let* stray dot? ~A", sc->code);
      if (is_immutable_symbol(car(sc->code)))
	return(s7_error(sc, sc->wrong_type_arg_symbol,	set_elist_2(sc, s7_make_string_wrapper(sc, "can't bind an immutable object: ~S"), sc->code)));
      set_local(car(sc->code));
      if ((!is_null(cadr(sc->code))) &&
	  (!is_pair(cadr(sc->code))))               /* (let* hi x ... ) */
	eval_error(sc, "named let* variable declaration value is missing: ~A", sc->code);
    }
  else
    {
      if ((!is_null(car(sc->code))) &&
	  (!is_pair(car(sc->code))))                /* (let* x ... ) */
	eval_error(sc, "let* variable declaration value is missing: ~A", sc->code);
    }

  for (vars = ((named_let) ? cadr(sc->code) : car(sc->code)); is_pair(vars); vars = cdr(vars))
    {
      s7_pointer var_and_val, var;
      var_and_val = car(vars);

      if (!is_pair(var_and_val))                    /* (let* (3) ... */
	eval_error(sc, "let* variable list is messed up? ~A", var_and_val);

      if (is_null(cdr(var_and_val)))                /* (let* ((x)) ...) */
	eval_error(sc, "let* variable declaration, but no value?: ~A", var_and_val);

      if (!(is_pair(cdr(var_and_val))))             /* (let* ((x . 1))...) */
	eval_error(sc, "let* variable declaration is not a proper list?: ~A", var_and_val);

      if (!is_null(cddr(var_and_val)))              /* (let* ((c 1 2)) ...) */
	eval_error(sc, "let* variable declaration has more than one value?: ~A", var_and_val);

      var = car(var_and_val);

      if (!(is_symbol(var)))                        /* (let* ((3 1)) 1) */
	eval_error(sc, "bad variable ~S in let*", var);

      if (is_immutable_symbol(var))                 /* (let* ((pi 3)) ...) */
	return(s7_error(sc, sc->wrong_type_arg_symbol,	set_elist_2(sc, s7_make_string_wrapper(sc, "can't bind an immutable object: ~S"), var_and_val)));

      /* currently (let* ((a 1) (a (+ a 1))) a) is 2, not an error. */
      set_local(var);
    }
  if (!is_null(vars))
    eval_error(sc, "let* variable list is not a proper list?: ~A", vars);

  if ((is_overlaid(sc->code)) &&
      (has_opt_back(sc->code)))
    {
      if (named_let)
	{
	  if (is_null(cadr(sc->code)))
	    pair_set_syntax_symbol(sc->code, sc->named_let_no_vars_symbol);
	  else
	    {
	      pair_set_syntax_symbol(sc->code, sc->named_let_star_symbol);
	      set_opt_con2(sc->code, cadr(caadr(sc->code)));
	    }
	  return(sc->code);
	}

      pair_set_syntax_symbol(sc->code, sc->let_star_unchecked_symbol);
      if (is_null(car(sc->code)))
	pair_set_syntax_symbol(sc->code, sc->let_no_vars_symbol); /* (let* () ...) */
      else
	{
	  if (is_null(cdar(sc->code)))
	    check_let_one_var(sc, car(sc->code)); /* (let* ((var...))...) -> (let ((var...))...) */
	  else  /* more than one entry */
	    {
	      s7_pointer p, op;
	      clear_symbol_list(sc);
	      op = sc->let_star_all_x_symbol;
	      set_opt_con2(sc->code, cadaar(sc->code));
	      for (p = car(sc->code); is_pair(p); p = cdr(p))
		{
		  s7_pointer x;
		  x = car(p);
		  add_symbol_to_list(sc, car(x));
		  if (is_pair(cadr(x)))
		    {
		      if ((!is_all_x_safe(sc, cadr(x))) &&
			  (caadr(x) != sc->quote_symbol))
			{
			  if (p == car(sc->code))
			    op = sc->let_star_a2_symbol;
			  else 
			    {
			      /* fprintf(stderr, "%s\n", DISPLAY_80(cadr(x))); */
			      op = sc->let_star2_symbol;
			      break;
			    }
			}
		    }
		}
	      pair_set_syntax_symbol(sc->code, op);
	    }
	}

      if ((pair_syntax_symbol(sc->code) == sc->let_all_x_symbol) ||
	  (pair_syntax_symbol(sc->code) == sc->let_star_all_x_symbol) ||
	  (pair_syntax_symbol(sc->code) == sc->let_star_a2_symbol))
	{
	  s7_pointer p;
	  clear_symbol_list(sc);
	  if (pair_syntax_symbol(sc->code) == sc->let_star_a2_symbol)
	    {
	      p = cdar(sc->code);
	      add_symbol_to_list(sc, caar(p));
	    }
	  else p = car(sc->code);
	  for (; is_pair(p); p = cdr(p))
	    {
	      set_x_call(cdar(p), all_x_eval(sc, cdar(p), sc->envir, let_star_symbol_is_safe));
	      add_symbol_to_list(sc, caar(p));
	    }
	}
      /* most optimized cases here not caught via all_x* are safe_closure calls etc */
    }
  return(sc->code);
}


static s7_pointer check_letrec(s7_scheme *sc, bool letrec)
{
  s7_pointer x, caller;
  caller = (letrec) ? sc->letrec_symbol : sc->letrec_star_symbol;
  if ((!is_pair(sc->code)) ||                 /* (letrec . 1) */
      (!is_pair(cdr(sc->code))) ||            /* (letrec) */
      (!s7_is_list(sc, car(sc->code))))       /* (letrec 1 ...) */
    eval_error_with_caller(sc, "~A: variable list is messed up: ~A", caller, sc->code);

  clear_symbol_list(sc);
  for (x = car(sc->code); is_not_null(x); x = cdr(x))
    {
      s7_pointer y, carx;
      if (!is_pair(x))                        /* (letrec ((a 1) . 2) ...) */
	eval_error_with_caller(sc, "~A: improper list of variables? ~A", caller, sc->code);

      carx = car(x);
      if ((!is_pair(carx)) ||                 /* (letrec (1 2) #t) */
	  (!(is_symbol(car(carx)))))
	eval_error_with_caller(sc, "~A: bad variable ~S", caller, carx);

      y = car(carx);
      if (is_immutable_symbol(y))
	return(s7_error(sc, sc->wrong_type_arg_symbol,	set_elist_2(sc, s7_make_string_wrapper(sc, "can't bind an immutable object: ~S"), x)));

      if (!is_pair(cdr(carx)))                /* (letrec ((x . 1))...) */
	{
	  if (is_null(cdr(carx)))             /* (letrec ((x)) x) -- perhaps this is legal? */
	    eval_error_with_caller(sc, "~A: variable declaration has no value?: ~A", caller, carx);
	  eval_error_with_caller(sc, "~A: variable declaration is not a proper list?: ~A", caller, carx);
	}
      if (is_not_null(cddr(carx)))            /* (letrec ((x 1 2 3)) ...) */
	eval_error_with_caller(sc, "~A: variable declaration has more than one value?: ~A", caller, carx);

      /* check for name collisions -- this is needed in letrec* else which of the two legit values
       *    does our "rec" refer to, so to speak.
       */
      if (symbol_is_in_list(sc, y))
	eval_error_with_caller(sc, "~A: duplicate identifier: ~A", caller, y);
      add_symbol_to_list(sc, y);
      set_local(y);
    }

  if ((is_overlaid(sc->code)) &&
      (has_opt_back(sc->code)))
    pair_set_syntax_symbol(sc->code, (letrec) ? sc->letrec_unchecked_symbol : sc->letrec_star_unchecked_symbol);

  return(sc->code);
}


static s7_pointer check_let_temporarily(s7_scheme *sc)
{
  s7_pointer x;

  if ((!is_pair(sc->code)) ||                 /* (let-temporarily . 1) */
      (!s7_is_list(sc, car(sc->code))))       /* (let-temporarily 1 ...) */
    eval_error(sc, "let-temporarily: variable list is messed up: ~A", sc->code);
  /* cdr(sc->code) = body can be nil */

  for (x = car(sc->code); is_not_null(x); x = cdr(x))
    {
      s7_pointer carx;
      if (!is_pair(x))                        /* (let-temporarily ((a 1) . 2) ...) */
	eval_error(sc, "let-temporarily: improper list of variables? ~A", sc->code);

      carx = car(x);
      if (!is_pair(carx))                     /* (let-temporarily (1 2) #t) */
	eval_error(sc, "let-temporarily: bad variable ~S", carx);

      if (is_symbol(car(carx)))
	{
	  if (is_immutable_symbol(car(carx))) /* (let-temporarily ((pi 3)) ...) */
	    return(s7_error(sc, sc->wrong_type_arg_symbol, set_elist_2(sc, s7_make_string_wrapper(sc, "can't bind an immutable object: ~S"), x)));
	  if (is_syntactic(car(carx)))        /* (let-temporarily ((if 3)) ...) */
	    return(s7_error(sc, sc->wrong_type_arg_symbol, set_elist_2(sc, s7_make_string_wrapper(sc, "can't set! ~S"), car(carx))));
	}

      if (!is_pair(cdr(carx)))                /* (let-temporarily ((x . 1))...) */
	eval_error(sc, "let-temporarily: variable declaration value is messed up: ~S", carx);

      if (is_not_null(cddr(carx)))            /* (let-temporarily ((x 1 2 3)) ...) */
	eval_error(sc, "let-temporarily: variable declaration has more than one value?: ~A", carx);
    }
  
  if ((is_overlaid(sc->code)) &&
      (has_opt_back(sc->code)))
    pair_set_syntax_symbol(sc->code, sc->let_temporarily_unchecked_symbol);

  return(sc->code);
}


static inline s7_pointer check_quote(s7_scheme *sc, s7_pointer code)
{
  if (!is_pair(code))                    /* (quote . -1) */
    {
      if (is_null(code))
	eval_error(sc, "quote: not enough arguments: ~A", code);
      eval_error(sc, "quote: stray dot?: ~A", code);
    }
  if (is_not_null(cdr(code)))             /* (quote . (1 2)) or (quote 1 1) */
    eval_error(sc, "quote: too many arguments ~A", code);
#if 0
  /* the problem here is set-cdr! */
  if ((is_overlaid(code)) &&
      (has_opt_back(code)))
    {
      pair_set_syntax_symbol(code, sc->quote_unchecked_symbol);
    }
#endif
  return(code);
}


static s7_pointer check_and(s7_scheme *sc)
{
  /* this and check_or and check_if may not be called -- optimize_syntax can short-circuit it to return all_x* choices */
  s7_pointer p;
  bool all_pairs;

  if (is_null(sc->code))
    return(sc->code);

  all_pairs = is_pair(sc->code);
  for (p = sc->code; is_pair(p); p = cdr(p))
    {
      if (!is_pair(car(p)))
	all_pairs = false;
    }

  if (is_not_null(p))                                    /* (and . 1) (and #t . 1) */
    eval_error(sc, "and: stray dot?: ~A", sc->code);

  if ((is_overlaid(sc->code)) &&
      (has_opt_back(sc->code)))
    {
      if (all_pairs)
	{
	  bool any_nils = false, any_ps = false;
	  for (p = sc->code; is_pair(p); p = cdr(p))
	    {
	      s7_function callee;
	      callee = all_x_eval(sc, p, sc->envir, let_symbol_is_safe);  /* c_callee can be nil! */
	      if (!callee) 
		any_nils = true;
	      set_x_call_checked(p, callee);
	      if (!is_optimized(car(p)))
		any_ps = true;
	      /* fprintf(stderr, "%s: %p %d\n", DISPLAY_80(car(p)), callee, is_optimized(car(p))); */
	    }
	  if ((c_callee(sc->code)) &&
	      (is_pair(cdr(sc->code))) &&
	      (is_null(cddr(sc->code))))
	    pair_set_syntax_symbol(sc->code, (any_nils) ? ((any_ps) ? sc->and_ap_symbol : sc->and_az_symbol) : sc->and_safe_aa_symbol);
	  else pair_set_syntax_symbol(sc->code, (any_nils) ? sc->and_p_symbol : sc->and_safe_p_symbol);
	}
      else pair_set_syntax_symbol(sc->code, sc->and_unchecked_symbol);
    }
  return(sc->code);
}


static s7_pointer check_or(s7_scheme *sc)
{
  s7_pointer p;
  bool all_pairs;

  if (is_null(sc->code))
    return(sc->code);

  all_pairs = is_pair(sc->code);
  for (p = sc->code; is_pair(p); p = cdr(p))
    {
      if (!is_pair(car(p)))
	all_pairs = false;
    }

  if (is_not_null(p))
    eval_error(sc, "or: stray dot?: ~A", sc->code);

  if ((is_overlaid(sc->code)) &&
      (has_opt_back(sc->code)))
    {
      if (all_pairs)
	{
	  s7_pointer ep;
	  bool any_nils = false;
	  for (ep = sc->code; is_pair(ep); ep = cdr(ep))
	    {
	      s7_function callee;
	      callee = all_x_eval(sc, ep, sc->envir, let_symbol_is_safe);
	      if (!callee) any_nils = true;
	      set_x_call_checked(ep, callee);
	    }
	  if ((c_callee(sc->code)) &&
	      (is_pair(cdr(sc->code))) &&
	      (is_null(cddr(sc->code))))
	    pair_set_syntax_symbol(sc->code, (any_nils) ? sc->or_ap_symbol : sc->or_safe_aa_symbol);
	  else pair_set_syntax_symbol(sc->code, (any_nils) ? sc->or_p_symbol : sc->or_safe_p_symbol);
	}
      else pair_set_syntax_symbol(sc->code, sc->or_unchecked_symbol);
    }
  return(sc->code);
}


#define choose_if_opt(Op, One, Reversed, Not) ((One) ? ((Reversed) ? Op ## _r_symbol : ((Not) ? Op ## _n_symbol : Op ## _p_symbol)) :  ((Not) ? Op ## _n_n_symbol : Op ## _p_p_symbol))

static void set_if_opts(s7_scheme *sc, bool one_branch, bool reversed)
{
  s7_pointer test;
  bool not_case = false;
  test = car(sc->code);
  if ((!reversed) && 
      (is_pair(test)) &&
      (car(test) == sc->not_symbol))
    {
      if (!is_pair(cdr(test))) return;
      not_case = true;
      test = cadr(test);
    }
  if (is_pair(test))
    {
      if (is_h_optimized(test))
	{
	  if (is_h_safe_c_c(test))
	    {
	      if (c_callee(test) == g_and_2)
		{
		  pair_set_syntax_symbol(sc->code, choose_if_opt(sc->if_and2, one_branch, reversed, not_case));
		  set_opt_pair2(sc->code, cdr(test));
		  set_opt_pair3(sc->code, cddr(test));
		  return;
		}
	      if (c_callee(test) == g_and_3)
		{
		  pair_set_syntax_symbol(sc->code, choose_if_opt(sc->if_and3, one_branch, reversed, not_case));
		  set_opt_pair2(sc->code, cdr(test));
		  set_opt_pair3(sc->code, cddr(test));
		  return;
		}
	      if (c_callee(test) == g_or_2)
		{
		  pair_set_syntax_symbol(sc->code, choose_if_opt(sc->if_or2, one_branch, reversed, not_case));
		  set_opt_pair2(sc->code, cdr(test));
		  set_opt_pair3(sc->code, cddr(test));
		  return;
		}
	      set_opt_pair2(sc->code, cdr(test));
	      pair_set_syntax_symbol(sc->code, choose_if_opt(sc->if_c, one_branch, reversed, not_case));
	      return;
	    }

	  if (is_h_safe_c_s(test))
	    {
	      uint8_t typ;
	      typ = symbol_type(car(test));
	      if (typ > 0)
		{
		  pair_set_syntax_symbol(sc->code, choose_if_opt(sc->if_is_type_s, one_branch, reversed, not_case));
		  set_opt_con3(sc->code, typ);
		}
	      else pair_set_syntax_symbol(sc->code, choose_if_opt(sc->if_cs, one_branch, reversed, not_case));
	      set_opt_sym2(sc->code, cadr(test));
	      return;
	    }

	  if (optimize_op(test) == HOP_SAFE_C_SQ)
	    {
	      pair_set_syntax_symbol(sc->code, choose_if_opt(sc->if_csq, one_branch, reversed, not_case));
	      set_opt_con2(sc->code, cadr(caddr(test)));
	      set_opt_sym3(sc->code, cadr(test));
	      return;
	    }
	  if ((optimize_op(test) == HOP_SAFE_C_SS) || (optimize_op(test) == HOP_SAFE_C_LL))
	    {
	      pair_set_syntax_symbol(sc->code, choose_if_opt(sc->if_css, one_branch, reversed, not_case));
	      set_opt_sym2(sc->code, caddr(test));
	      set_opt_sym3(sc->code, cadr(test));
	      return;
	    }
	  if ((optimize_op(test) == HOP_SAFE_C_SC) &&
	      (!is_local_symbol(cdr(test))))
	    {
	      pair_set_syntax_symbol(sc->code, choose_if_opt(sc->if_csc, one_branch, reversed, not_case));
	      set_opt_con2(sc->code, caddr(test));
	      set_opt_sym3(sc->code, cadr(test));
	      return;
	    }
	  if (optimize_op(test) == HOP_SAFE_C_S_opCq)
	    {
	      pair_set_syntax_symbol(sc->code, choose_if_opt(sc->if_s_opcq, one_branch, reversed, not_case));
	      set_opt_pair2(sc->code, caddr(test));
	      set_opt_sym3(sc->code, cadr(test));
	      return;
	    }
	  if (optimize_op(test) == HOP_SAFE_C_opSq)
	    {
	      uint8_t typ;
	      typ = symbol_type(car(test));
	      if (typ > 0)
		{
		  pair_set_syntax_symbol(sc->code, choose_if_opt(sc->if_is_type_opsq, one_branch, reversed, not_case));
		  set_opt_sym2(sc->code, cadadr(test));
		  set_opt_con3(sc->code, typ);
		}
	      else
		{
		  pair_set_syntax_symbol(sc->code, choose_if_opt(sc->if_opsq, one_branch, reversed, not_case));
		  set_opt_pair2(sc->code, cadr(test));
		  set_opt_sym3(sc->code, cadadr(test));
		}
	      return;
	    }
	  if (is_all_x_safe(sc, test))
	    {
	      pair_set_syntax_symbol(sc->code, choose_if_opt(sc->if_a, one_branch, reversed, not_case));
	      if (not_case)
		set_x_call(cdar(sc->code), all_x_eval(sc, cdar(sc->code), sc->envir, let_symbol_is_safe));
	      else set_x_call(sc->code, all_x_eval(sc, sc->code, sc->envir, let_symbol_is_safe));
	    }
	  else
	    {
	      pair_set_syntax_symbol(sc->code, choose_if_opt(sc->if_z, one_branch, reversed, not_case));
	    }
	}
      else
	{
	  pair_set_syntax_symbol(sc->code, choose_if_opt(sc->if_p, one_branch, reversed, not_case));
	  if (is_syntactic_symbol(car(test)))
	    {
	      pair_set_syntax_op(test, symbol_syntax_op(car(test)));
	      
	      if ((symbol_syntax_op(car(test)) == OP_AND) ||
		  (symbol_syntax_op(car(test)) == OP_OR))
		{
		  opcode_t new_op;
		  s7_pointer old_code;
		  old_code = sc->code;
		  sc->code = cdr(test);
		  if (symbol_syntax_op(car(test)) == OP_AND) check_and(sc); else check_or(sc);
		  new_op = symbol_syntax_op(car(test));
		  sc->code = old_code;
		  if ((new_op == OP_AND_P) || (new_op == OP_AND_AP) || (new_op == OP_AND_AZ) || 
		      (new_op == OP_AND_SAFE_P) || (new_op == OP_AND_SAFE_AA))
		    pair_set_syntax_symbol(sc->code, choose_if_opt(sc->if_andp, one_branch, reversed, not_case));
		  else
		    {
		      if ((new_op == OP_OR_P) || (new_op == OP_OR_AP) || 
			  (new_op == OP_OR_SAFE_P) || (new_op == OP_OR_SAFE_AA))
			pair_set_syntax_symbol(sc->code, choose_if_opt(sc->if_orp, one_branch, reversed, not_case));
		    }
		}
	    }
	}
    }
  else /* test is symbol or constant, but constant here is nutty */
    {
      if (is_safe_symbol(test))
	pair_set_syntax_symbol(sc->code, choose_if_opt(sc->if_s, one_branch, reversed, not_case));
    }
}


/* (cond <> (else <>)) only happens in old-fashioned code, so set_if_opts covers if/when/unless but not cond 
 * g_and_3 and g_or_3 are slightly slower here??
 */

static s7_pointer check_if(s7_scheme *sc)
{
  s7_pointer cdr_code;

  if (!is_pair(sc->code))                                /* (if) or (if . 1) */
    eval_error(sc, "(if): if needs at least 2 expressions: ~A", sc->code);

  cdr_code = cdr(sc->code);
  if (!is_pair(cdr_code))                                /* (if 1) */
    eval_error(sc, "(if ~A): if needs another clause", car(sc->code));

  if (is_pair(cdr(cdr_code)))
    {
      if (is_not_null(cddr(cdr_code)))                   /* (if 1 2 3 4) */
	eval_error(sc, "too many clauses for if: ~A", sc->code);
    }
  else
    {
      if (is_not_null(cdr(cdr_code)))                    /* (if 1 2 . 3) */
	eval_error(sc, "if: ~A has improper list?", sc->code);
    }

  if ((is_overlaid(sc->code)) &&
      (has_opt_back(sc->code)))
    {
      pair_set_syntax_symbol(sc->code, sc->if_unchecked_symbol);
      set_if_opts(sc, is_null(cdr(cdr_code)), false);
    }
  return(sc->code);
}


static s7_pointer check_when(s7_scheme *sc)
{
  if (!is_pair(sc->code))                                            /* (when) or (when . 1) */
    eval_error(sc, "when has no expression or body:  ~A", sc->code);
  if (!is_pair(cdr(sc->code)))                                       /* (when 1) or (when 1 . 1) */
    eval_error(sc, "when has no body?:  ~A", sc->code);
  else
    {
      if (!s7_is_proper_list(sc, cddr(sc->code)))
	eval_error(sc, "when: stray dot?", sc->code);
    }

  if ((is_overlaid(sc->code)) &&
      (has_opt_back(sc->code)))
    {
      pair_set_syntax_symbol(sc->code, sc->when_unchecked_symbol);
      if (is_null(cddr(sc->code)))
	set_if_opts(sc, true, false);
      else
	{
	  if (is_safe_symbol(car(sc->code)))
	    pair_set_syntax_symbol(sc->code, sc->when_s_symbol);
	  else
	    {
	      if (is_all_x_safe(sc, car(sc->code)))
		{
		  pair_set_syntax_symbol(sc->code, sc->when_a_symbol);
		  set_x_call(sc->code, all_x_eval(sc, sc->code, sc->envir, let_symbol_is_safe));
		}
	      else
		{
		  if (is_pair(cdr(sc->code)))
		    pair_set_syntax_symbol(sc->code, sc->when_p_symbol);
		}
	    }
	}
    }
  return(sc->code);
}


static s7_pointer check_unless(s7_scheme *sc)
{
  if (!is_pair(sc->code))                                            /* (unless) or (unless . 1) */
    eval_error(sc, "unless has no expression or body:  ~A", sc->code);
  if (!is_pair(cdr(sc->code)))                                       /* (unless 1) or (unless 1 . 1) */
    eval_error(sc, "unless has no body?:  ~A", sc->code);
  else
    {
      if (!s7_is_proper_list(sc, cddr(sc->code)))
	eval_error(sc, "unless: stray dot?", sc->code);
    }

  if ((is_overlaid(sc->code)) &&
      (has_opt_back(sc->code)))
    {
      pair_set_syntax_symbol(sc->code, sc->unless_unchecked_symbol);
      if (is_null(cddr(sc->code)))
	set_if_opts(sc, true, true);
      else
	{
	  if (is_safe_symbol(car(sc->code)))
	    pair_set_syntax_symbol(sc->code, sc->unless_s_symbol);
	  else
	    {
	      if (is_all_x_safe(sc, car(sc->code)))
		{
		  pair_set_syntax_symbol(sc->code, sc->unless_a_symbol);
		  set_x_call(sc->code, all_x_eval(sc, sc->code, sc->envir, let_symbol_is_safe));
		}
	    }
	}
    }
  return(sc->code);
}


static s7_pointer check_define(s7_scheme *sc)
{
  s7_pointer func, caller;
  bool starred;
  int32_t arity = CLOSURE_ARITY_NOT_SET;

  starred = (sc->cur_op == OP_DEFINE_STAR);
  if (starred)
    {
      caller = sc->define_star_symbol;
      sc->cur_op = OP_DEFINE_STAR_UNCHECKED;
    }
  else 
    {
      if (sc->cur_op == OP_DEFINE)
	caller = sc->define_symbol;
      else caller = sc->define_constant_symbol;
    }

  if (!is_pair(sc->code))
    eval_error_with_caller(sc, "~A: nothing to define? ~A", caller, sc->code);     /* (define) */

  if (!is_pair(cdr(sc->code)))
    {
      if (is_null(cdr(sc->code)))
	eval_error_with_caller(sc, "~A: no value? ~A", caller, sc->code);          /* (define var) */
      eval_error_with_caller(sc, "~A: stray dot? ~A", caller, sc->code);            /* (define var . 1) */
    }
  if (!is_pair(car(sc->code)))
    {
      if (is_not_null(cddr(sc->code)))                                           /* (define var 1 . 2) */
	eval_error_with_caller_and_print_limit(sc, "~A: more than one value? ~A", caller, sc->code); /* (define var 1 2) */
      if (starred)
	eval_error(sc, "define* is restricted to functions: (define* ~{~S~^ ~})", sc->code);

      func = car(sc->code);
      if (!is_symbol(func))                                                      /* (define 3 a) */
	eval_error_with_caller(sc, "~A: define a non-symbol? ~S", caller, func);
      if (is_keyword(func))                                                      /* (define :hi 1) */
	eval_error_with_caller(sc, "~A ~A: keywords are constants", caller, func);
      if (is_syntactic(func))                                                    /* (define and a) */
	{
	  if (sc->safety > NO_SAFETY)
	    s7_warn(sc, 128, "%s: syntactic keywords tend to behave badly if redefined", DISPLAY(func));
	  set_local(func);
	}
      if ((is_pair(cadr(sc->code))) &&               /* look for (define sym (lambda ...)) and treat it like (define (sym ...)...) */
	  ((caadr(sc->code) == sc->lambda_symbol) ||
	   (caadr(sc->code) == sc->lambda_star_symbol)) &&
	  (symbol_id(caadr(sc->code)) == 0))
	{
	  /* not is_global here because that bit might not be set for initial symbols (why not? -- redef as method etc) */
	  if (!is_pair(cdadr(sc->code)))                                         /* (define x (lambda . 1)) */
	    eval_error_with_caller(sc, "~A: stray dot? ~A", caller, sc->code);
	  if (caadr(sc->code) == sc->lambda_star_symbol)
	    check_lambda_star_args(sc, cadadr(sc->code), &arity);
	  else check_lambda_args(sc, cadadr(sc->code), &arity);
	  optimize_lambda(sc, caadr(sc->code) == sc->lambda_symbol, func, cadadr(sc->code), cddr(cadr(sc->code)));
	}
    }
  else
    {
      func = caar(sc->code);
      if (!is_symbol(func))                                                      /* (define (3 a) a) */
	eval_error_with_caller(sc, "~A: define a non-symbol? ~S", caller, func);
      if (is_syntactic(func))                                                    /* (define (and a) a) */
	{
	  if (sc->safety > NO_SAFETY)
	    s7_warn(sc, 128, "%s: syntactic keywords tend to behave badly if redefined", DISPLAY(func));
	  set_local(func);
	}
      if (starred)
	set_cdar(sc->code, check_lambda_star_args(sc, cdar(sc->code), &arity));
      else check_lambda_args(sc, cdar(sc->code), &arity);
      optimize_lambda(sc, !starred, func, cdar(sc->code), cdr(sc->code));
    }

  if ((is_overlaid(sc->code)) &&
      (has_opt_back(sc->code)))
    {
      if (sc->cur_op == OP_DEFINE)
	{
	  if ((is_pair(car(sc->code))) &&
	      (!symbol_has_accessor(func)) &&
	      (!is_immutable_symbol(func)))
	    pair_set_syntax_symbol(sc->code, sc->define_funchecked_symbol);
	  else pair_set_syntax_symbol(sc->code, sc->define_unchecked_symbol);
	}
      else 
	{
	  if (starred)
	    pair_set_syntax_symbol(sc->code, sc->define_star_unchecked_symbol);
	  else pair_set_syntax_symbol(sc->code, sc->define_constant_unchecked_symbol); 
	}
    }
  return(sc->code);
}

static int32_t define_unchecked_ex(s7_scheme *sc)
{
  if (sc->cur_op == OP_DEFINE_STAR_UNCHECKED)
    {
      s7_pointer x;
      uint64_t typ;
      if (is_safe_closure(cdr(sc->code)))
	typ = T_CLOSURE_STAR | T_SAFE_CLOSURE;
      else typ = T_CLOSURE_STAR;
      new_cell(sc, x, typ);
      closure_set_args(x, cdar(sc->code));
      closure_set_body(x, cdr(sc->code));
      closure_set_let(x, sc->envir);
      closure_arity(x) = CLOSURE_ARITY_NOT_SET;
      closure_set_setter(x, sc->F);
      sc->capture_let_counter++;
      sc->value = x;
      sc->code = caar(sc->code);
      return(fall_through);
    }

  if (!is_pair(car(sc->code)))
    {
      s7_pointer x;
      x = car(sc->code);
      sc->code = cadr(sc->code);
      if (is_pair(sc->code))
	{
	  push_stack(sc, OP_DEFINE1, sc->nil, x);
	  return(goto_EVAL);
	}
      
      if (is_symbol(sc->code))
	sc->value = find_global_symbol_checked(sc, sc->code);
      else sc->value = sc->code;
      sc->code = x;
    }
  else
    {
      s7_pointer x,args;
      /* a closure.  If we called this same code earlier (a local define), the only thing
       *   that is new here is the environment -- we can't blithely save the closure object
       *   in opt2 somewhere, and pick it up the next time around (since call/cc might take
       *   us back to the previous case).  We also can't re-use opt2(sc->code) because opt2
       *   is not cleared in the gc.
       */
      args = cdar(sc->code);
      make_closure_with_let(sc, x, args, cdr(sc->code), sc->envir, (is_null(args)) ? 0 : CLOSURE_ARITY_NOT_SET);
      sc->value = _NFre(x);
      sc->code = caar(sc->code);
    }
  return(fall_through);
}

static s7_pointer make_funclet(s7_scheme *sc, s7_pointer new_func, s7_pointer func_name, s7_pointer outer_env)
{
  s7_pointer new_env;
  new_cell_no_check(sc, new_env, T_LET | T_FUNCLET);
  let_id(new_env) = ++sc->let_number;
  let_set_slots(new_env, sc->nil);
  set_outlet(new_env, outer_env);
  closure_set_let(new_func, new_env);
  funclet_set_function(new_env, func_name); /* __func__ returns at least funclet_function */
  if (is_safe_closure(new_func))
    {
      s7_pointer arg;
      for (arg = closure_args(new_func); is_pair(arg); arg = cdr(arg))
	{
	  if (is_pair(car(arg)))
	    make_slot_1(sc, new_env, caar(arg), sc->nil);
	  else make_slot_1(sc, new_env, car(arg), sc->nil);
	}
      let_set_slots(new_env, reverse_slots(sc, let_slots(new_env)));
    }
  return(new_env);
}

static void define_funchecked(s7_scheme *sc)
{
  s7_pointer new_func, code;
  code = sc->code;
  sc->value = caar(code);

  new_cell(sc, new_func, T_CLOSURE | T_COPY_ARGS);
  closure_set_args(new_func, cdar(code));
  closure_set_body(new_func, cdr(code));
  closure_set_setter(new_func, sc->F);
  closure_arity(new_func) = CLOSURE_ARITY_NOT_SET;
  sc->capture_let_counter++;
  
  if (is_safe_closure(cdr(code)))
    {
      set_safe_closure(new_func);
      make_funclet(sc, new_func, sc->value, sc->envir);
    }
  else closure_set_let(new_func, sc->envir);
  /* unsafe closures created by other functions do not support __func__ */

  if (is_let(sc->envir))
    add_slot(sc->envir, sc->value, new_func);
  set_local(sc->value);
  sc->value = new_func;
}

		    
static s7_pointer check_define_macro(s7_scheme *sc, opcode_t op)
{
  s7_pointer x, y, caller;
  caller = sc->define_macro_symbol;
  switch (op)
    {
    case OP_DEFINE_MACRO:      caller = sc->define_macro_symbol;      break;
    case OP_DEFINE_MACRO_STAR: caller = sc->define_macro_star_symbol; break;
    case OP_DEFINE_BACRO:      caller = sc->define_bacro_symbol;      break;
    case OP_DEFINE_BACRO_STAR: caller = sc->define_bacro_star_symbol; break;
    case OP_DEFINE_EXPANSION:  caller = sc->define_expansion_symbol;  break;
    }

  if (!is_pair(sc->code))                                               /* (define-macro . 1) */
    eval_error_with_caller(sc, "~A name missing (stray dot?): ~A", caller, sc->code);
  if (!is_pair(car(sc->code)))                                          /* (define-macro a ...) */
    return(wrong_type_argument_with_type(sc, caller, 1, car(sc->code), s7_make_string_wrapper(sc, "a list: (name ...)")));
  /* not car(opt_back(sc->code)) to get the caller (e.g. 'define-bacro) because opt_back might not be set: (apply define-macro '(1)) */

  x = caar(sc->code);
  if (!is_symbol(x))
    eval_error_with_caller(sc, "~A: ~S is not a symbol?", caller, x);
  if (dont_eval_args(x))                                               /* (define-macro (quote a) quote) */
    {
      if (sc->safety > NO_SAFETY)
	s7_warn(sc, 128, "%s: syntactic keywords tend to behave badly if redefined", DISPLAY(x));
      set_local(x);
    }
  if (is_immutable_symbol(x))
    eval_error_with_caller(sc, "~A: ~S is immutable", caller, x);

  if (!is_pair(cdr(sc->code)))                /* (define-macro (...)) */
    eval_error_with_caller(sc, "~A ~A, but no body?", caller, x);

  y = cdar(sc->code);            /* the arglist */
  if ((!s7_is_list(sc, y)) &&
      (!is_symbol(y)))
    return(s7_error(sc, sc->syntax_error_symbol,                                      /* (define-macro (mac . 1) ...) */
		    set_elist_3(sc, s7_make_string_wrapper(sc, "macro ~A argument list is ~S?"), x, y)));

  for ( ; is_pair(y); y = cdr(y))
    if ((!is_symbol(car(y))) &&
	((op == OP_DEFINE_MACRO) || (op == OP_DEFINE_BACRO) || (op == OP_DEFINE_EXPANSION)))
      return(s7_error(sc, sc->syntax_error_symbol,                                    /* (define-macro (mac 1) ...) */
		      set_elist_3(sc, s7_make_string_wrapper(sc, "define-macro ~A argument name is not a symbol: ~S"), x, y)));

  if ((op == OP_DEFINE_MACRO_STAR) || (op == OP_DEFINE_BACRO_STAR))
    set_cdar(sc->code, check_lambda_star_args(sc, cdar(sc->code), NULL));
  else check_lambda_args(sc, cdar(sc->code), NULL);

  return(sc->code);
}

static int32_t expansion_ex(s7_scheme *sc)
{
  int64_t loc;
  s7_pointer caller;
  
  /* read-time macro expansion:
   *   (define-macro (hi a) (format #t "hi...") `(+ ,a 1))
   *   (define (ho b) (+ 1 (hi b)))
   * here sc->value is: (ho b), (hi b), (+ 1 (hi b)), (define (ho b) (+ 1 (hi b)))
   * but... first we can't tell for sure at this point that "hi" really is a macro
   *   (letrec ((hi ... (hi...))) will be confused about the second hi,
   *   or (call/cc (lambda (hi) (hi 1))) etc.
   * second, figuring out that we're quoted is not easy -- we have to march all the
   * way to the bottom of the stack looking for op_read_quote or op_read_vector
   *    #(((hi)) 2) or '(((hi)))
   * or op_read_list with args not equal (quote) or (macroexapand)
   *    '(hi 3) or (macroexpand (hi 3) or (quote (hi 3))
   * and those are only the problems I noticed!
   *
   * The hardest of these problems involve shadowing, so Rick asked for "define-expansion"
   *   which is like define-macro, but the programmer guarantees that the macro
   *   name will not be shadowed.
   *
   * to make expansion recognition fast here, define-expansion sets the T_EXPANSION
   *   bit in the symbol as well as the value:
   *   set_type(sc->code, T_EXPANSION | T_SYMBOL)
   * but this can lead to confusion because the expansion name is now globally identified as an expansion.
   *    (let () (define-expansion (ex1 a) `(+ ,a 1)) (display (ex1 3)))
   *    (define (ex1 b) (* b 2)) (display (ex1 3))
   * since this happens at the top level, the first line is evaluated, ex1 becomes an expansion.
   * but the reader has no idea about lets and whatnot, so in the second line, ex1 is still an expansion
   * to the reader, so it sees (define (+ b 1) ...) -- error!  To support tail-calls, there's no
   * way in eval to see the let close, so we can't clear the expansion flag when the let is done.
   * But we don't want define-expansion to mimic define-constant (via T_IMMUTABLE) because programs
   * like lint need to cancel reader-cond (for example).  So, we allow an expansion to be redefined,
   * and check here that the expander symbol still refers to an expansion.
   *
   * but in (define (ex1 b) b), the reader doesn't know we're in a define call (or it would be
   *   a bother to notice), so to redefine an expansion, first (set! ex1 #f) or (define ex1 #f),
   *   then (define (ex1 b) b).
   *
   * This is a mess!  Maybe we should insist that expansions are always global.
   *
   * run-time expansion and splicing into the code as in CL won't work in s7 because macros
   *   are first-class objects.  For example (define (f m) (m 1)), call it with a macro, say `(+ ,arg 1),
   *   and in CL-style, you'd now have the body (+ ,arg 1) or maybe even 2, now call f with a function,
   *   or some other macro -- oops!
   */
  
  loc = s7_stack_top(sc) - 1;
  if (is_pair(stack_args(sc->stack, loc)))
    caller = car(stack_args(sc->stack, loc)); /* this can be garbage */
  else caller = sc->F;
  if ((loc >= 3) &&
      (stack_op(sc->stack, loc) != OP_READ_QUOTE) &&             /* '(hi 1) for example */
      (stack_op(sc->stack, loc) != OP_READ_VECTOR) &&            /* #(reader-cond) for example */
      (caller != sc->quote_symbol) &&          /* (quote (hi 1)) */
      (caller != sc->macroexpand_symbol) &&    /* (macroexpand (hi 1)) */
      (caller != sc->define_expansion_symbol)) /* (define-expansion ...) being reloaded/redefined */
    {
      s7_pointer symbol, slot;
      /* we're playing fast and loose with sc->envir in the reader, so here we need a disaster check */
#if DEBUGGING
      if (unchecked_type(sc->envir) != T_LET) sc->envir = sc->nil;
#else
      if (!is_let(sc->envir)) sc->envir = sc->nil;
#endif
      symbol = car(sc->value);

      if ((symbol_id(symbol) == 0) ||
	  (sc->envir == sc->nil))
	slot = global_slot(symbol);
      else slot = find_symbol(sc, symbol);

      if (is_slot(slot))
	sc->code = slot_value(slot);
      else sc->code = sc->undefined;

      if ((!is_macro(sc->code)) || (!is_expansion(sc->code)))
	clear_expansion(symbol);
      else
	{
	  sc->args = copy_list(sc, cdr(sc->value));
	  return(goto_APPLY);
	}
    }
  return(fall_through);
}

static s7_pointer check_with_let(s7_scheme *sc)
{
  if (!is_pair(sc->code))                            /* (with-let . "hi") */
    eval_error(sc, "with-let takes an environment argument: ~A", sc->code);
  if (!is_pair(cdr(sc->code)))                       /* (with-let e) -> an error? */
    eval_error(sc, "with-let body is messed up: ~A", sc->code);
  if ((!is_pair(cddr(sc->code))) &&
      (!is_null(cddr(sc->code))))
    eval_error(sc, "with-let body has stray dot? ~A", sc->code);

  if ((is_overlaid(sc->code)) &&
      (has_opt_back(sc->code)))
    {
      pair_set_syntax_symbol(sc->code, sc->with_let_unchecked_symbol);
      if ((is_symbol(car(sc->code))) &&
	  (is_pair(cadr(sc->code))))
	pair_set_syntax_symbol(sc->code, sc->with_let_s_symbol);
    }
  return(sc->code);
}


static s7_pointer check_cond(s7_scheme *sc)
{
  bool has_feed_to = false;
  s7_pointer x;
  
  if (!is_pair(sc->code))                                             /* (cond) or (cond . 1) */
    eval_error(sc, "cond, but no body: ~A", sc->code);

  for (x = sc->code; is_pair(x); x = cdr(x))
    {
      if (!is_pair(car(x)))                                           /* (cond 1) or (cond (#t 1) 3) */
	eval_error(sc, "every clause in cond must be a list: ~A", car(x));
      else
	{
	  s7_pointer y;
	  y = car(x);
	  if (!s7_is_proper_list(sc, cdr(y)))
	    eval_error(sc, "stray dot? ~A", y);
	  if (is_pair(cdr(y)))
	    {
	      if ((cadr(y) == sc->feed_to_symbol) &&
		  (s7_symbol_value(sc, sc->feed_to_symbol) == sc->undefined))
		{
		  has_feed_to = true;
		  if (!is_pair(cddr(y)))                                  /* (cond (#t =>)) or (cond (#t => . 1)) */
		    eval_error(sc, "cond: '=>' target missing?  ~A", x);
		  if (is_pair(cdddr(y)))                                  /* (cond (1 => + abs)) */
		    eval_error(sc, "cond: '=>' has too many targets: ~A", x);
		}
	    }
	  else
	    {
	      if (!is_null(cdr(y)))              /* (cond (1 . 2)) */
		eval_error(sc, "cond: stray dot? ~A", sc->code);
	    }
	  /* currently we accept:
	   *     (cond (1 2) (=> . =>)) and all variants thereof, e.g. (cond (1 2) (=> 1 . 2) (1 2)) or
	   *     (cond (1) (=>)) but Guile accepts this?
	   *     (cond (1) (1 =>))
	   * amusing (correct) case: (cond (1 => "hi")) -> #\i
	   */
	}
    }
  if (is_not_null(x))                                             /* (cond ((1 2)) . 1) */
    eval_error(sc, "cond: stray dot? ~A", sc->code);

  if ((is_overlaid(sc->code)) &&
      (has_opt_back(sc->code)))
    {
      if (has_feed_to)
	{
	  pair_set_syntax_symbol(sc->code, (is_optimized(caar(sc->code))) ? sc->cond_unchecked_z_symbol : sc->cond_unchecked_symbol);
	  if (is_null(cdr(sc->code)))
	    {
	      s7_pointer expr, f;
	      expr = car(sc->code);
	      f = caddr(expr);
	      if ((is_pair(f)) &&
		  (car(f) == sc->lambda_symbol) &&
		  (is_pair(cdr(f))) &&            /* not => (lambda) */
		  (is_pair(cddr(f))) &&
		  (is_null(cdddr(f))))
		{
		  s7_pointer arg;
		  arg = cadr(f);
		  if ((is_pair(arg)) &&
		      (is_null(cdr(arg))) &&
		      (is_symbol(car(arg))))
		    {
		      /* (define (hi) (cond (#t => (lambda (s) s)))) */
		      set_opt_lambda2(sc->code, caddar(sc->code));  /* (lambda ...) above */
		      pair_set_syntax_symbol(sc->code, sc->cond_feed_symbol);
		    }
		}
	    }
	}
      else
	{
	  s7_pointer p;
	  bool xopt = true;
	  pair_set_syntax_symbol(sc->code, sc->cond_simple_symbol);

	  for (p = sc->code; xopt && (is_pair(p)); p = cdr(p))
	    xopt = is_all_x_safe(sc, caar(p)); /* caar=test */
	  if (xopt)
	    {
	      int32_t i;
	      bool eopt = true;
	      pair_set_syntax_symbol(sc->code, sc->cond_all_x_symbol);
	      for (i = 0, p = sc->code; is_pair(p); i++, p = cdr(p))
		set_x_call(car(p), all_x_eval(sc, car(p), (is_null(sc->envir)) ? sc->rootlet : sc->envir, let_symbol_is_safe));
	      if (i == 2)
		pair_set_syntax_symbol(sc->code, sc->cond_all_x_2_symbol);

	      for (p = sc->code; eopt && (is_pair(p)); p = cdr(p))
		eopt = ((!is_pair(cdar(p))) || ((!is_pair(cddar(p))) && ((!is_pair(cadar(p))) || (is_optimized(cadar(p))))));
	      if (eopt)
		pair_set_syntax_symbol(sc->code, sc->cond_all_x_z_symbol);
	    }
	}
    }
  return(sc->code);
}

static int32_t feed_to(s7_scheme *sc)
{
  /* old form (pre 6-June-16): this causes a double evaluation:
   *   (let ((x 'y) (y 32)) (cond ((values x y) => list))) -> '(32 32)
   *   but it should be '(y 32)
   * it's also extremely slow: make/eval a list?!
   *
   *   if (is_multiple_value(sc->value)) 
   *     sc->code = cons(sc, cadr(sc->code), multiple_value(sc->value));
   *   else sc->code = list_2(sc, cadr(sc->code), list_2(sc, sc->quote_symbol, sc->value));
   *   goto EVAL;
   */
  if (is_multiple_value(sc->value))                             /* (cond ((values 1 2) => +)) */
    {
      sc->args = multiple_value(sc->value);
      clear_multiple_value(sc->args);
    }
  else 
    {
      if (is_symbol(cadr(sc->code)))
	{
	  s7_pointer func;
	  func = find_symbol_checked(sc, cadr(sc->code));            /* car is => */
	  if ((is_c_function(func)) &&
	      (is_safe_procedure(func)))
	    {
	      if (((int32_t)c_function_required_args(func) <= 1) &&
		  ((int32_t)c_function_all_args(func) >= 1))
		{
		  sc->value = c_function_call(func)(sc, set_plist_1(sc, sc->value));
		  return(goto_START);
		}
	    }
	  sc->code = func;
	  sc->args = list_1(sc, sc->value);
	  return(goto_APPLY);
	}
      else sc->args = list_1(sc, sc->value);
    }

  /* need to evaluate the target function */
  push_stack(sc, OP_COND1_1, sc->args, sc->code);
  sc->code = cadr(sc->code);
  return(goto_EVAL);
}

static void set_dilambda_opt(s7_scheme *sc, s7_pointer opt, s7_pointer expr)
{
  s7_pointer func;
  func = find_symbol_checked(sc, car(expr));
  if ((is_closure(func)) &&
      (is_safe_closure(closure_setter(func))))
    {
      s7_pointer setter;
      setter = closure_setter(func);
      pair_set_syntax_symbol(sc->code, opt);
      if ((!(is_let(closure_let(setter)))) ||
	  (!(is_funclet(closure_let(setter)))))
	make_funclet(sc, setter, car(expr), closure_let(setter));
    }
}


static s7_pointer check_set(s7_scheme *sc)
{
  if (!is_pair(sc->code))
    {
      if (is_null(sc->code))                                             /* (set!) */
	eval_error(sc, "set!: not enough arguments: ~A", sc->code);
      eval_error(sc, "set!: stray dot? ~A", sc->code);           /* (set! . 1) */
    }
  if (!is_pair(cdr(sc->code)))
    {
      if (is_null(cdr(sc->code)))                                         /* (set! var) */
	eval_error(sc, "set!: not enough arguments: ~A", sc->code);
      eval_error(sc, "set!: stray dot? ~A", sc->code);           /* (set! var . 1) */
    }
  if (is_not_null(cddr(sc->code)))                                       /* (set! var 1 2) */
    eval_error(sc, "~A: too many arguments to set!", sc->code);

  /* cadr (the value) has not yet been evaluated */

  if (is_pair(car(sc->code)))
    {
      if (is_pair(caar(sc->code)))
	{
	  if (!s7_is_list(sc, cdar(sc->code)))                          /* (set! ('(1 2) . 0) 1) */
	    eval_error(sc, "improper list of args to set!: ~A", sc->code);
	}
      if (!s7_is_proper_list(sc, car(sc->code)))                        /* (set! ("hi" . 1) #\a) or (set! (#(1 2) . 1) 0) */
	eval_error(sc, "set! target is an improper list: (set! ~A ...)", car(sc->code));
    }
  else
    {
      if (!is_symbol(car(sc->code)))                                 /* (set! 12345 1) */
	eval_error(sc, "set! can't change ~S", car(sc->code));
      else
	{
	  if (is_immutable_symbol(car(sc->code)))                    /* (set! pi 3) */
	    eval_error(sc, "set!: can't alter immutable object: ~S", car(sc->code));
	}
    }

  if ((is_overlaid(sc->code)) &&
      (has_opt_back(sc->code)))
    {
      if (is_pair(car(sc->code)))
	{
	  /* here we have (set! (...) ...) */
	  s7_pointer inner, value;
	  inner = car(sc->code);
	  value = cadr(sc->code);

	  pair_set_syntax_symbol(sc->code, sc->set_unchecked_symbol);
	  if (is_symbol(car(inner)))
	    {
	      if ((is_null(cdr(inner))) &&
		  (!is_pair(value)) &&
		  (is_global(car(inner))) &&
		  (is_c_function(slot_value(global_slot(car(inner))))) &&
		  (c_function_required_args(slot_value(global_slot(car(inner)))) == 0))
		pair_set_syntax_symbol(sc->code, sc->set_pws_symbol);
	      else
		{
		  if ((is_pair(cdr(inner))) &&
		      (!is_pair(cddr(inner)))) /* we check cddr(sc->code) above */
		    {
		      if (!is_pair(cadr(inner)))
			{
			  /* (set! (f s) ...) */
			  if (!is_pair(value))
			    {
			      pair_set_syntax_symbol(sc->code, sc->set_pair_symbol);
			      if (is_symbol(car(inner)))
				set_dilambda_opt(sc, sc->set_dilambda_symbol, inner);
			    }
			  else
			    {
			      pair_set_syntax_symbol(sc->code, sc->set_pair_p_symbol);
			      /* splice_in_values protects us here from values */
			      if (is_optimized(value))
				{
				  pair_set_syntax_symbol(sc->code, sc->set_pair_z_symbol);
				  if (is_all_x_safe(sc, value))
				    {
				      s7_pointer obj;
				      annotate_arg(sc, cdr(sc->code), sc->envir);
				      pair_set_syntax_symbol(sc->code, sc->set_pair_za_symbol);
				      obj = find_symbol_checked(sc, car(inner));
				      if ((is_c_function(obj)) &&
					  (is_c_function(c_function_setter(obj))))
					pair_set_syntax_symbol(sc->code, sc->set_pair_a_symbol);
				    }
				  else
				    {
				      if (is_symbol(car(inner)))
					set_dilambda_opt(sc, sc->set_dilambda_z_symbol, inner);
				    }
				}
			    }
			}
		      else
			{
			  if ((caadr(inner) == sc->quote_symbol) &&
			      (is_symbol(car(inner))) &&
			      ((is_symbol(value)) || (is_all_x_safe(sc, value))))
			    {
			      if (is_safe_symbol(value))
				pair_set_syntax_symbol(sc->code, sc->set_let_s_symbol);
			      else
				{
				  pair_set_syntax_symbol(sc->code, sc->set_let_all_x_symbol);
				  set_x_call(cdr(sc->code), all_x_eval(sc, cdr(sc->code), sc->envir, let_symbol_is_safe));
				}
			    }
			}
		    }
		}
	    }
	}
      else pair_set_syntax_symbol(sc->code, sc->set_normal_symbol);

      if (is_symbol(car(sc->code)))
	{
	  s7_pointer settee, value;
	  settee = car(sc->code);
	  value = cadr(sc->code);

	  if ((!symbol_has_accessor(settee)) &&
	      (!is_syntactic(settee)))
	    {
	      if (is_symbol(value))
		{
		  if (is_local_symbol(cdr(sc->code)))
		    pair_set_syntax_symbol(sc->code, sc->set_symbol_l_symbol);
		  else
		    {
		      if (is_slot(find_symbol(sc, value)))
			pair_set_syntax_symbol(sc->code, sc->set_symbol_s_symbol);
		    }
		}
	      else
		{
		  if (!is_pair(value))
		    pair_set_syntax_symbol(sc->code, sc->set_symbol_c_symbol);
		  else
		    {
		      if (car(value) == sc->quote_symbol)
			pair_set_syntax_symbol(sc->code, sc->set_symbol_q_symbol);
		      else
			{
			  /* if cadr(cadr) == car, or cdr(cadr) not null and cadr(cadr) == car, and cddr(cadr) == null,
			   *   it's (set! <var> (<op> <var> val)) or (<op> val <var>) or (<op> <var>)
			   *   in the set code, we get the slot as usual, then in case 1 above,
			   *   car(sc->t2_1) = slot_value(slot), car(sc->t2_2) = increment, call <op>, set slot_value(slot)
			   *
			   * this can be done in all combined cases where a symbol is repeated (do in particular)
			   */

			  /* (define (hi) (let ((x 1)) (set! x (+ x 1))))
			   *   but the value might be values:
			   *   (let () (define (hi) (let ((x 0)) (set! x (values 1 2)) x)) (catch #t hi (lambda a a)) (hi))
			   *   which is caught in splice_in_values
			   */
			  pair_set_syntax_symbol(sc->code, sc->set_symbol_p_symbol);
			  if (is_h_safe_c_s(value))
			    {
			      pair_set_syntax_symbol(sc->code, (is_local_symbol(cdr(value))) ? sc->set_symbol_oplq_symbol : sc->set_symbol_opsq_symbol);
			      set_opt_sym2(sc->code, cadr(value));
			    }
			  else
			    {
			      if (is_optimized(value))
				{
				  pair_set_syntax_symbol(sc->code, sc->set_symbol_z_symbol);
				  if (is_h_safe_c_c(value))
				    {
				      pair_set_syntax_symbol(sc->code, sc->set_symbol_opcq_symbol);
				      /* opt1 here points back? */
				      set_opt_pair2(sc->code, cdr(value));
				    }
				  else
				    {
				      /* most of these special cases probably don't matter */
				      if ((optimize_op(value) == HOP_SAFE_C_SS) || (optimize_op(value) == HOP_SAFE_C_LL))
					{
					  bool locals;
					  locals = ((is_local_symbol(cdr(value))) && (is_local_symbol(cddr(value))));
					  if (settee == cadr(value))
					    pair_set_syntax_symbol(sc->code, sc->increment_ss_symbol);
					  else pair_set_syntax_symbol(sc->code, (locals) ? sc->set_symbol_opllq_symbol : sc->set_symbol_opssq_symbol);
					  set_opt_pair2(sc->code, cdr(value));
					}
				      else
					{
					  if (optimize_op(value) == HOP_SAFE_C_SSS)
					    {
					      if ((settee == cadr(value)) &&
						  (car(value) == sc->add_symbol))
						pair_set_syntax_symbol(sc->code, sc->increment_sss_symbol);
					      else pair_set_syntax_symbol(sc->code, sc->set_symbol_opsssq_symbol);
					      set_opt_pair2(sc->code, cdr(value));
					    }
					  else
					    {
					      if (is_all_x_safe(sc, value)) /* value = cadr(sc->code) */
						{
						  pair_set_syntax_symbol(sc->code, sc->set_symbol_a_symbol);
						  annotate_arg(sc, cdr(sc->code), sc->envir);
						}
					      if ((is_callable_c_op(optimize_op(value))) &&
						  (is_pair(cdr(value))) &&
						  (settee == cadr(value)) &&
						  (!is_null(cddr(value))))
						{
						  if (is_null(cdddr(value)))
						    {
						      if (is_all_x_safe(sc, caddr(value)))
							{
							  /* this appears to give a slight savings over the SZ case */
							  pair_set_syntax_symbol(sc->code, sc->increment_sa_symbol);
							  annotate_arg(sc, cddr(value), sc->envir); /* this sets c_callee(arg) */
							  set_opt_pair2(sc->code, cddr(value));
							}
						      else
							{
							  if (is_optimized(caddr(value)))
							    {
							      pair_set_syntax_symbol(sc->code, sc->increment_sz_symbol);
							      set_opt_pair2(sc->code, caddr(value));
							    }
							}
						    }
						  else
						    {
						      if ((is_null(cddddr(value))) &&
							  (is_all_x_safe(sc, caddr(value))) &&
							  (is_all_x_safe(sc, cadddr(value))))
							{
							  pair_set_syntax_symbol(sc->code, sc->increment_saa_symbol);
							  annotate_arg(sc, cddr(value), sc->envir);
							  annotate_arg(sc, cdddr(value), sc->envir);
							  set_opt_pair2(sc->code, cddr(value));
							}
						    }
						}
					    }
					}
				    }
				}
			    }
			  if ((is_h_optimized(value)) &&
			      (!is_unsafe(value)) &&                   /* is_unsafe(value) can happen! */
			      (is_not_null(cdr(value))))               /* (set! x (y)) */
			    {
			      if (is_not_null(cddr(value)))
				{
				  if ((caddr(value) == small_int(1)) &&
				      (cadr(value) == settee))
				    {
				      if ((opt_cfunc(value) == add_s1) ||
					  (opt_cfunc(value) == add_cs1) ||
					  (opt_cfunc(value) == add_cl1))
					pair_set_syntax_symbol(sc->code, sc->increment_1_symbol);
				      else
					{
					  if ((opt_cfunc(value) == subtract_s1) ||
					      (opt_cfunc(value) == subtract_cs1) ||
					      (opt_cfunc(value) == subtract_cl1))
					    pair_set_syntax_symbol(sc->code, sc->decrement_1_symbol);
					}
				    }
				  else
				    {
				      if ((cadr(value) == small_int(1)) &&
					  (caddr(value) == settee) &&
					  (opt_cfunc(value) == add_1s))
					pair_set_syntax_symbol(sc->code, sc->increment_1_symbol);
				      else
					{
					  if ((settee == caddr(value)) &&
					      (is_safe_symbol(cadr(value))) &&
					      (caadr(sc->code) == sc->cons_symbol))
					    {
					      pair_set_syntax_symbol(sc->code, sc->set_cons_symbol);
					      set_opt_sym2(sc->code, cadr(value));
					    }
					}
				    }
				}
			    }
			}
		    }
		}
	    }
	}
    }
  return(sc->code);
}

static bool set_pair_p_3(s7_scheme *sc, s7_pointer obj, s7_pointer arg, s7_pointer value)
{
  /* fprintf(stderr, "p_3: %s\n", DISPLAY(obj)); */
  if (is_slot(obj))
    obj = slot_value(obj);
  else eval_error(sc, "no generalized set for ~A", caar(sc->code));
  
  switch (type(obj))
    {
    case T_C_OBJECT:
      set_car(sc->t2_1, arg);
      set_car(sc->t2_2, value);
      sc->value = (*(c_object_set(sc, obj)))(sc, obj, sc->t2_1);
      break;
      
      /* some of these are wasteful -- we know the object type! (list hash-table) */
    case T_INT_VECTOR:
    case T_FLOAT_VECTOR:
    case T_VECTOR:
#if WITH_GMP
      set_car(sc->t3_1, obj);
      set_car(sc->t3_2, arg);
      set_car(sc->t3_3, value);
      sc->value = g_vector_set(sc, sc->t3_1);
#else
      if (vector_rank(obj) > 1)
	{
	  set_car(sc->t3_1, obj);
	  set_car(sc->t3_2, arg);
	  set_car(sc->t3_3, value);
	  sc->value = g_vector_set(sc, sc->t3_1);
	}
      else
	{
	  s7_int index;

	  if (!is_integer(arg))
	    eval_type_error(sc, "vector-set!: index must be an integer: ~S", sc->code);
	  index = integer(arg);
	  if (index < 0)
	    eval_range_error(sc, "vector-set!: index must not be negative: ~S", sc->code);
	  if (index >= vector_length(obj))
	    eval_range_error(sc, "vector-set!: index must be less than vector length: ~S", sc->code);
	  vector_setter(obj)(sc, obj, index, value);
	  sc->value = _NFre(value);
	}
#endif
      break;
      
    case T_STRING:
#if WITH_GMP
      set_car(sc->t3_1, obj);
      set_car(sc->t3_2, arg);
      set_car(sc->t3_3, value);
      sc->value = g_string_set(sc, sc->t3_1);
#else
      {
	s7_int index;
	if (!is_integer(arg))
	  eval_type_error(sc, "string-set!: index must be an integer: ~S", sc->code);
	index = integer(arg);
	if (index < 0)
	  eval_range_error(sc, "string-set!: index must not be negative: ~S", sc->code);
	if (index >= string_length(obj))
	  eval_range_error(sc, "string-set!: index must be less than string length: ~S", sc->code);
	if (s7_is_character(value))
	  {
	    string_value(obj)[index] = (char)s7_character(value);
	    sc->value = _NFre(value);
	  }
	else
	  {
	    if ((is_byte_vector(obj)) &&
		(s7_is_integer(value)))
	      {
		int32_t ic;
		ic = s7_integer(value);
		if ((ic < 0) || (ic > 255))
		  eval_type_error(sc, "string-set!: value must be a character: ~S", sc->code);
		string_value(obj)[index] = (char)ic;
		sc->value = _NFre(value);
	      }
	    else eval_type_error(sc, "string-set!: value must be a character: ~S", sc->code);
	  }
      }
#endif
      break;
      
    case T_PAIR:
      set_car(sc->t3_1, obj);
      set_car(sc->t3_2, arg);
      set_car(sc->t3_3, value);
      sc->value = g_list_set(sc, sc->t3_1);
      break;
      
    case T_HASH_TABLE:
      sc->value = s7_hash_table_set(sc, obj, arg, value);
      break;
      
    case T_LET:
      sc->value = s7_let_set(sc, obj, arg, value);
      break;
      
    case T_C_OPT_ARGS_FUNCTION:
    case T_C_RST_ARGS_FUNCTION:
    case T_C_ANY_ARGS_FUNCTION:                       /* (let ((lst (list 1 2))) (set! (list-ref lst 1) 2) lst) */
    case T_C_FUNCTION:
    case T_C_FUNCTION_STAR:
      /* obj here is a c_function, but its setter could be a closure and vice versa below */
      if (is_procedure_or_macro(c_function_setter(obj)))
	{
	  if (is_c_function(c_function_setter(obj)))
	    {
	      set_car(sc->t2_1, arg);
	      set_car(sc->t2_2, value);
	      sc->value = c_function_call(c_function_setter(obj))(sc, sc->t2_1);
	    }
	  else
	    {
	      sc->code = c_function_setter(obj);
	      if (needs_copied_args(sc->code))
		sc->args = list_2(sc, arg, value);
	      else sc->args = set_plist_2(sc, arg, value);
	      return(true); /* goto APPLY; */
	    }
	}
      else eval_error(sc, "no generalized set for ~A", obj);
      break;
      
    case T_MACRO:   case T_MACRO_STAR:
    case T_BACRO:   case T_BACRO_STAR:
    case T_CLOSURE: case T_CLOSURE_STAR:
      if (is_procedure_or_macro(closure_setter(obj)))
	{
	  if (is_c_function(closure_setter(obj)))
	    {
	      set_car(sc->t2_1, arg);
	      set_car(sc->t2_2, value);
	      sc->value = c_function_call(closure_setter(obj))(sc, sc->t2_1);
	    }
	  else
	    {
	      sc->code = closure_setter(obj);
	      if (needs_copied_args(sc->code))
		sc->args = list_2(sc, arg, value);
	      else sc->args = set_plist_2(sc, arg, value);
	      return(true); /* goto APPLY; */
	    }
	}
      else eval_error(sc, "no generalized set for ~A", obj);
      break;
      
    default:                                         /* (set! (1 2) 3) */
      eval_error(sc, "no generalized set for ~A", obj);
    }
  return(false);
}


static bool safe_stepper(s7_scheme *sc, s7_pointer expr, s7_pointer vars)
{
  /* for now, just look for stepper as last element of any list
   *    any embedded set is handled by do-is-safe, so we don't need to descend into the depths
   */
  s7_pointer p;
  if (direct_memq(cadr(expr), vars))
    return(false);

  for (p = cdr(expr); is_pair(cdr(p)); p = cdr(p));

  if (is_pair(p))
    {
      if ((is_optimized(p)) &&
	  ((optimize_op(p) & 1) != 0) &&
	  (is_safe_c_op(optimize_op(p))))
	return(true);

      if (direct_memq(car(p), vars))
	return(false);
    }
  else
    {
      if (direct_memq(p, vars))
	return(false);
    }
  return(true);
}

static int32_t set_pair_ex(s7_scheme *sc)
{
  s7_pointer caar_code, cx;

  caar_code = caar(sc->code);
  if (is_pair(caar_code))
    {
      push_stack(sc, OP_SET2, cdar(sc->code), cdr(sc->code));
      sc->code = caar_code;
      return(goto_EVAL);
    }
  
  if (is_symbol(caar_code))
    {
      /* this was cx = s7_symbol_value(sc, caar_code) but the function call overhead is noticeable */
      cx = find_symbol(sc, caar_code);
      if (is_slot(cx))
	cx = slot_value(cx);
      else eval_error_no_return(sc, sc->syntax_error_symbol, "no generalized set for ~A", caar_code);
    }
  else cx = caar_code;
  
  /* code here is the accessor and the value without the "set!": ((window-width) 800) */
  /*    (set! (hi 0) (* 2 3)) -> ((hi 0) (* 2 3)) */
  
  /* for these kinds of objects, some Schemes restrict set!
   * (list-set! '(1 2 3) 1 32) is accepted but does it make sense?
   * (set-car! '(1 . 2) 32)
   * (string-set! "hiho" 1 #\z)
   * (vector-set! #(1 2 3) 1 32)
   * (let ((x (lambda () "hiho"))) (string-set! (x) 1 #\a))
   * (let ((x (lambda () #(1 2 3)))) (vector-set! (x) 1 32))
   * (let ((str "hiho")) (string-set! str 1 #\x) str)
   * (let ((v #(1 2 3))) (vector-set! v 1 32) v)
   * (let ((x (lambda () "hiho"))) (string-set! (x) 1 #\x) (x))
   *
   *   It seems weird that we can reach into both the function body, and its closure:
   *   (let ((xx (let ((x '(1 2 3))) (lambda () x)))) (list-set! (xx) 1 32) (xx)) -> '(1 32 3)
   *
   * (let* ((x '(1 2)) (y (list x)) (z (car y))) (list-set! z 1 32) (list x y z))
   * ((1 32) ((1 32)) (1 32))
   *
   * (string-set! (symbol->string 'symbol->string) 1 #\X) -> error currently also in Guile "string is read-only"
   * (setf (elt (symbol-name 'xyz) 1) #\X) -> error in CL "read-only string"
   */
  /* for gmp case, indices need to be decoded via s7_integer, not just integer */
  
  switch (type(cx))
    {
    case T_C_OBJECT:
      {
	s7_pointer settee, index, val;
	
	if (is_null(cdr(sc->code)))
	  s7_wrong_number_of_args_error(sc, "no value for object-set!: ~S", sc->code);
	if (!is_null(cddr(sc->code)))
	  s7_wrong_number_of_args_error(sc, "too many values for object-set!: ~S", sc->code);
	
	settee = car(sc->code);
	if ((is_null(cdr(settee))) ||
	    (!is_null(cddr(settee))))
	  {
	    /* no-index or multi-index case -- use slow version.
	     *  TODO: ambiguity here -- is (set! (obj a b) v) actually (set! ((obj a) b) v)?
	     *  perhaps look at setter? c-object-set takes 1 arg -- is this a bug?
	     */
	    push_op_stack(sc, sc->object_set_function);
	    if (is_null(cdr(settee)))
	      {
		push_stack(sc, OP_EVAL_ARGS1, list_1(sc, cx), cddr(sc->code));
		sc->code = cadr(sc->code);
	      }
	    else
	      {
		push_stack(sc, OP_EVAL_ARGS1, list_1(sc, cx), s7_append(sc, cddr(settee), cdr(sc->code)));
		sc->code = cadr(settee);
	      }
	    return(goto_EVAL);
	  }
	
	index = cadr(settee);
	if (!is_pair(index))
	  {
	    if (is_symbol(index))
	      index = find_symbol_checked(sc, index);
	    
	    val = cadr(sc->code);
	    if (!is_pair(val))
	      {
		if (is_symbol(val))
		  val = find_symbol_checked(sc, val);
		set_car(sc->t2_1, index);
		set_car(sc->t2_2, val);
		sc->value = (*(c_object_set(sc, cx)))(sc, cx, sc->t2_1);
		return(goto_START);
	      }
	    push_op_stack(sc, sc->object_set_function);
	    sc->args = list_2(sc, index, cx);
	    sc->code = cdr(sc->code);
	    return(goto_EVAL_ARGS);
	  }
	else
	  {
	    push_stack(sc, OP_EVAL_ARGS1, list_1(sc, cx), cdr(sc->code));
	    push_op_stack(sc, sc->object_set_function);
	    sc->code = cadr(settee);
	  }
	return(goto_EVAL);
      }
      break;
      
    case T_INT_VECTOR:
    case T_FLOAT_VECTOR:
    case T_VECTOR:
      {
	/* cx is the vector, sc->code is expr without the set! */
	/*  args have not been evaluated! */
	
	s7_pointer settee, index, val;
	
	if (is_null(cdr(sc->code)))
	  s7_wrong_number_of_args_error(sc, "no value for vector-set!: ~S", sc->code);
	if (!is_null(cddr(sc->code)))
	  s7_wrong_number_of_args_error(sc, "too many values for vector-set!: ~S", sc->code);
	
	settee = car(sc->code);
	if (is_null(cdr(settee)))
	  s7_wrong_number_of_args_error(sc, "no index for vector-set!: ~S", sc->code);

	if ((!is_null(cddr(settee))) &&
	    (is_normal_vector(cx)))
	  {
	    push_stack(sc, OP_SET2, cddr(settee), cdr(sc->code));
	    sc->code = list_2(sc, car(settee), cadr(settee));
	    return(goto_EVAL);
	  }

	if ((!is_null(cddr(settee))) ||
	    (vector_rank(cx) > 1))
	  {
	    /* multi-index case -- use slow version */
	    push_op_stack(sc, sc->vector_set_function);
	    push_stack(sc, OP_EVAL_ARGS1, list_1(sc, cx), s7_append(sc, cddr(settee), cdr(sc->code)));
	    sc->code = cadr(settee);
	    return(goto_EVAL);
	  }
	
	index = cadr(settee);
	if (!is_pair(index))
	  {
	    s7_int ind;
	    
	    if (is_symbol(index))
	      index = find_symbol_checked(sc, index);
	    if (!s7_is_integer(index))
	      eval_error_no_return(sc, sc->wrong_type_arg_symbol, "vector-set!: index must be an integer: ~S", sc->code);
	    ind = s7_integer(index);
	    if ((ind < 0) ||
		(ind >= vector_length(cx)))
	      out_of_range(sc, sc->vector_set_symbol, small_int(2), index, (ind < 0) ? its_negative_string : its_too_large_string);
	    val = cadr(sc->code);
	    if (!is_pair(val))
	      {
		if (is_symbol(val))
		  val = find_symbol_checked(sc, val);
		vector_setter(cx)(sc, cx, ind, val);
		sc->value = _NFre(val);
		return(goto_START);
	      }
	    push_op_stack(sc, sc->vector_set_function);
	    sc->args = list_2(sc, index, cx);
	    sc->code = cdr(sc->code);
	    return(goto_EVAL_ARGS);
	  }
	else
	  {
	    /* here the index calc might be trivial -- (+ i 1) or (- j 1) but this branch hardly ever happens
	     */
	    push_stack(sc, OP_EVAL_ARGS1, list_1(sc, cx), cdr(sc->code));
	    push_op_stack(sc, sc->vector_set_function);
	    sc->code = cadr(settee);
	  }
      }
      break;
      
    case T_STRING:
      {
	/* sc->code = cons(sc, sc->string_set_function, s7_append(sc, car(sc->code), cdr(sc->code)));
	 *
	 * here only one index makes sense, and it is required, so
	 *   (set! ("str") #\a), (set! ("str" . 1) #\a) and (set! ("str" 1 2) #\a)
	 *   are all errors (but see below!).
	 */
	s7_pointer settee, index, val;
	
	if (is_null(cdr(sc->code)))
	  s7_wrong_number_of_args_error(sc, "no value for string-set!: ~S", sc->code);
	if (!is_null(cddr(sc->code)))
	  s7_wrong_number_of_args_error(sc, "too many values for string-set!: ~S", sc->code);
	
	settee = car(sc->code);
	if (is_null(cdr(settee))) /* there's an index: (set! (str i) #\a), code is ((str 0) #\1) */
	  s7_wrong_number_of_args_error(sc, "no index for string-set!: ~S", sc->code);
	if (!is_null(cddr(settee)))
	  s7_wrong_number_of_args_error(sc, "too many indices for string-set!: ~S", sc->code);
	
	/* if there's one index (the standard case), and it is not a pair, and there's one value (also standard)
	 *    and it is not a pair, let's optimize this thing!
	 *    cx is what we're setting, cadar is the index, cadr is the new value
	 */
	index = cadr(settee);
	if (!is_pair(index))
	  {
	    s7_int ind;
	    
	    if (is_symbol(index))
	      index = find_symbol_checked(sc, index);
	    if (!s7_is_integer(index))
	      eval_error_no_return(sc, sc->wrong_type_arg_symbol, "string-set!: index must be an integer: ~S", sc->code);
	    ind = s7_integer(index);
	    if ((ind < 0) ||
		(ind >= string_length(cx)))
	      out_of_range(sc, sc->string_set_symbol, small_int(2), index, (ind < 0) ? its_negative_string : its_too_large_string);
	    
	    val = cadr(sc->code);
	    if (!is_pair(val))
	      {
		if (is_symbol(val))
		  val = find_symbol_checked(sc, val);
		if (s7_is_character(val))
		  {
		    string_value(cx)[ind] = character(val);
		    sc->value = val;
		    return(goto_START);
		  }
		else
		  {
		    if ((is_byte_vector(cx)) &&
			(s7_is_integer(val)))
		      {
			int32_t ic;
			ic = s7_integer(val);
			if ((ic < 0) || (ic > 255))
			  eval_error_no_return(sc, sc->wrong_type_arg_symbol, "string-set!: value must be a character: ~S", sc->code);
			string_value(cx)[ind] = (char)ic;
			sc->value = val;
			return(goto_START);
		      }
		  }
		eval_error_no_return(sc, sc->wrong_type_arg_symbol, "string-set!: value must be a character: ~S", sc->code);
	      }
	    push_op_stack(sc, sc->string_set_function);
	    sc->args = list_2(sc, index, cx);
	    sc->code = cdr(sc->code);
	    return(goto_EVAL_ARGS);
	  }
	else
	  {
	    push_stack(sc, OP_EVAL_ARGS1, list_1(sc, cx), cdr(sc->code));
	    push_op_stack(sc, sc->string_set_function);
	    sc->code = cadar(sc->code);
	  }
      }
      break;
      
    case T_PAIR:
      /* code: ((lst 1) 32) from (let ((lst (list 1 2 3))) (set! (lst 1) 32)) */
      {
	s7_pointer settee, index, val;
	
	if (is_null(cdr(sc->code)))
	  s7_wrong_number_of_args_error(sc, "no value for list-set!: ~S", sc->code);
	if (!is_null(cddr(sc->code)))
	  s7_wrong_number_of_args_error(sc, "too many values for list-set!: ~S", sc->code);

	settee = car(sc->code);
	if (is_null(cdr(settee)))
	  s7_wrong_number_of_args_error(sc, "no index for list-set!: ~S", sc->code);

	if (!is_null(cddr(settee)))
	  {
	    /* split (set! (a b c...) v) into (set! ((a b) c ...) v), eval (a b), return 
	     *    (let ((L (list (list 1 2)))) (set! (L 0 0) 3) L)
	     */
	    push_stack(sc, OP_SET2, cddr(settee), cdr(sc->code));
	    sc->code = list_2(sc, car(settee), cadr(settee));
	    return(goto_EVAL);
	  }

	index = cadr(settee);
	val = cadr(sc->code);
	
	if ((is_pair(index)) ||
	    (is_pair(val)))
	  {
	    push_op_stack(sc, sc->list_set_function);
	    push_stack(sc, OP_EVAL_ARGS1, list_1(sc, cx), s7_append(sc, cddr(settee), cdr(sc->code)));
	    sc->code = index;
	    return(goto_EVAL);
	  }
	
	if (is_symbol(index))
	  index = find_symbol_checked(sc, index);
	if (is_symbol(val))
	  val = find_symbol_checked(sc, val);
	
	set_car(sc->t2_1, index);
	set_car(sc->t2_2, val);
	sc->value = g_list_set_1(sc, cx, sc->t2_1, 2);
	return(goto_START);
      }
      break;
      
      
    case T_HASH_TABLE:
      {
	s7_pointer settee, key;
	
	if (is_null(cdr(sc->code)))
	  s7_wrong_number_of_args_error(sc, "no value for hash-table-set!: ~S", sc->code);
	if (!is_null(cddr(sc->code)))
	  s7_wrong_number_of_args_error(sc, "too many values for hash-table-set!: ~S", sc->code);
	
	settee = car(sc->code);
	if (is_null(cdr(settee)))
	  s7_wrong_number_of_args_error(sc, "no key for hash-table-set!: ~S", sc->code);

	if (!is_null(cddr(settee)))
	  {
	    push_stack(sc, OP_SET2, cddr(settee), cdr(sc->code));
	    sc->code = list_2(sc, car(settee), cadr(settee));
	    return(goto_EVAL);
	  }
	
	key = cadr(settee);
	if (!is_pair(key))
	  {
	    s7_pointer val;
	    if (is_symbol(key))
	      key = find_symbol_checked(sc, key);
	    val = cadr(sc->code);
	    if (!is_pair(val))
	      {
		if (is_symbol(val))
		  val = find_symbol_checked(sc, val);
		sc->value = s7_hash_table_set(sc, cx, key, val);
		return(goto_START);
	      }
	    push_op_stack(sc, sc->hash_table_set_function);
	    sc->args = list_2(sc, key, cx);
	    sc->code = cdr(sc->code);
	    return(goto_EVAL_ARGS);
	  }
	else
	  {
	    push_stack(sc, OP_EVAL_ARGS1, list_1(sc, cx), cdr(sc->code));
	    push_op_stack(sc, sc->hash_table_set_function);
	    sc->code = cadar(sc->code);
	  }
      }
      break;
      
      
    case T_LET:
      /* sc->code = cons(sc, sc->let_set_function, s7_append(sc, car(sc->code), cdr(sc->code))); */
      {
	s7_pointer settee, key;
	/* code: ((gen 'input) input) from (set! (gen 'input) input)
	 */
	
	if (is_null(cdr(sc->code)))
	  s7_wrong_number_of_args_error(sc, "no value for let-set!: ~S", sc->code);
	if (!is_null(cddr(sc->code)))
	  s7_wrong_number_of_args_error(sc, "too many values for let-set!: ~S", sc->code);
	
	settee = car(sc->code);
	if (is_null(cdr(settee)))
	  s7_wrong_number_of_args_error(sc, "no identifier for let-set!: ~S", sc->code);
	
	if (!is_null(cddr(settee)))
	  {
	    push_stack(sc, OP_SET2, cddr(settee), cdr(sc->code));
	    sc->code = list_2(sc, car(settee), cadr(settee));
	    return(goto_EVAL);
	  }

	key = cadr(settee);
	if (is_proper_quote(sc, key))
	  {
	    s7_pointer val;
	    key = cadr(key);
	    val = cadr(sc->code);
	    if (!is_pair(val))
	      {
		if (is_symbol(val))
		  val = find_symbol_checked(sc, val);
		sc->value = s7_let_set(sc, cx, key, val);
		return(goto_START);
	      }
	    push_op_stack(sc, sc->let_set_function);
	    sc->args = list_2(sc, key, cx);
	    sc->code = cdr(sc->code);
	    return(goto_EVAL_ARGS);
	  }
	else
	  {
	    push_stack(sc, OP_EVAL_ARGS1, list_1(sc, cx), cdr(sc->code));
	    push_op_stack(sc, sc->let_set_function);
	    sc->code = cadar(sc->code);
	  }
      }
      break;
      
      
    case T_C_MACRO:
    case T_C_OPT_ARGS_FUNCTION:
    case T_C_RST_ARGS_FUNCTION:
    case T_C_ANY_ARGS_FUNCTION:                       /* (let ((lst (list 1 2))) (set! (list-ref lst 0) 2) lst) */
    case T_C_FUNCTION:
    case T_C_FUNCTION_STAR:
      /* perhaps it has a setter */
      if (is_procedure(c_function_setter(cx)))
	{
	  /* here the setter can be anything, so we need to check the needs_copied_args bit 
	   *    (set! ((dilambda / (let ((x 3)) (lambda (y) (+ x y))))) 3)!
	   */
	  /* sc->code = cons(sc, c_function_setter(cx), s7_append(sc, cdar(sc->code), cdr(sc->code))); */
	  if (is_pair(cdar(sc->code)))
	    {
	      if ((is_symbol(cadr(sc->code))) &&
		  (is_symbol(cadar(sc->code))))
		{
		  if (is_null(cddar(sc->code)))
		    {
		      if (needs_copied_args(c_function_setter(cx)))
			sc->args = list_2(sc, find_symbol_checked(sc, cadar(sc->code)), find_symbol_checked(sc, cadr(sc->code)));
		      else
			{
			  set_car(sc->t2_1, find_symbol_checked(sc, cadar(sc->code)));
			  set_car(sc->t2_2, find_symbol_checked(sc, cadr(sc->code)));
			  sc->args = sc->t2_1;
			}
		      sc->code = c_function_setter(cx);
		      return(goto_APPLY); /* check arg num etc */
		    }
		  if ((is_symbol(caddar(sc->code))) &&
		      (is_null(cdddar(sc->code))))
		    {
		      if (needs_copied_args(c_function_setter(cx)))
			sc->args = list_3(sc, 
					  find_symbol_checked(sc, cadar(sc->code)), 
					  find_symbol_checked(sc, caddar(sc->code)), 
					  find_symbol_checked(sc, cadr(sc->code)));
		      else
			{
			  set_car(sc->t3_1, find_symbol_checked(sc, cadar(sc->code)));
			  set_car(sc->t3_2, find_symbol_checked(sc, caddar(sc->code)));
			  set_car(sc->t3_3, find_symbol_checked(sc, cadr(sc->code)));
			  sc->args = sc->t3_1;
			}
		      sc->code = c_function_setter(cx);
		      return(goto_APPLY); /* check arg num etc */
		    }
		}
	      push_op_stack(sc, c_function_setter(cx));
	      push_stack(sc, OP_EVAL_ARGS1, sc->nil, s7_append(sc, cddar(sc->code), cdr(sc->code)));
	      sc->code = cadar(sc->code);
	    }
	  else
	    {
	      if ((is_null(cddr(sc->code))) &&
		  (!is_pair(cadr(sc->code))))
		{
		  if (needs_copied_args(c_function_setter(cx)))
		    sc->args = list_1(sc, (is_symbol(cadr(sc->code))) ? find_symbol_checked(sc, cadr(sc->code)) : cadr(sc->code));
		  else
		    {
		      if (is_symbol(cadr(sc->code)))
			set_car(sc->t1_1, find_symbol_checked(sc, cadr(sc->code)));
		      else set_car(sc->t1_1, cadr(sc->code));
		      sc->args = sc->t1_1;
		    }
		  sc->code = c_function_setter(cx);
		  return(goto_APPLY); /* check arg num etc */
		}
	      push_op_stack(sc, c_function_setter(cx));
	      push_stack(sc, OP_EVAL_ARGS1, sc->nil, cddr(sc->code));
	      sc->code = cadr(sc->code);
	    }
	}
      else
	{
	  if (is_any_macro(c_function_setter(cx)))
	    {
	      if (is_null(cdar(sc->code)))
		sc->args = copy_list(sc, cdr(sc->code));
	      else sc->args = s7_append(sc, cdar(sc->code), copy_list(sc, cdr(sc->code)));
	      /* append copies except for its last arg, but for macros, we have to copy everything, hence the extra copy_list */
	      sc->code = c_function_setter(cx);
	      return(goto_APPLY);
	    }
	  else eval_error_no_return(sc, sc->syntax_error_symbol, "no generalized set for ~A", caar_code);
	}
      break;
      
    case T_MACRO:   case T_MACRO_STAR:
    case T_BACRO:   case T_BACRO_STAR:
    case T_CLOSURE: case T_CLOSURE_STAR:
      {
	s7_pointer setter;
	setter = closure_setter(cx);
	if (is_procedure(setter))          /* appears to be caar_code */
	  {
	    /* (set! (o g) ...), here cx = o, sc->code = ((o g) ...) */
	    push_op_stack(sc, setter);
	    if (is_null(cdar(sc->code)))
	      {
		push_stack(sc, OP_EVAL_ARGS1, sc->nil, cddr(sc->code));
		sc->code = cadr(sc->code);
	      }
	    else
	      {
		if (is_null(cddar(sc->code)))
		  push_stack(sc, OP_EVAL_ARGS1, sc->nil, cdr(sc->code));
		else push_stack(sc, OP_EVAL_ARGS1, sc->nil, s7_append(sc, cddar(sc->code), cdr(sc->code)));
		sc->code = cadar(sc->code);
	      }
	  }
	else
	  {
	    if (is_any_macro(setter))
	      {
		if (is_null(cdar(sc->code)))
		  sc->args = copy_list(sc, cdr(sc->code));
		else sc->args = s7_append(sc, cdar(sc->code), copy_list(sc, cdr(sc->code)));
		sc->code = setter;
		return(goto_APPLY);
	      }
	    else eval_error_no_return(sc, sc->syntax_error_symbol, "no generalized set for ~A", caar_code);
	  }
      }
      break;
      
    case T_ITERATOR:     /* not sure this makes sense */
      {
	s7_pointer setter;
	setter = iterator_sequence(cx);
	if ((is_any_closure(setter)) || (is_any_macro(setter)))
	  setter = closure_setter(iterator_sequence(cx));
	else setter = sc->F;
	if (is_procedure(setter))
	  {
	    push_op_stack(sc, setter);
	    push_stack(sc, OP_EVAL_ARGS1, sc->nil, sc->nil);
	    sc->code = cadr(sc->code);    /* the (as yet unevaluated) value, incoming code was ((obj) val) */
	  }
	else
	  {
	    if (is_any_macro(setter))
	      {
		sc->args = list_1(sc, cadr(sc->code));
		sc->code = setter;
		return(goto_APPLY);
	      }
	    else eval_error_no_return(sc, sc->syntax_error_symbol, "no generalized set for ~A", caar_code);
	  }
      }
      break;

    case T_SYNTAX:
      if (cx == slot_value(global_slot(sc->with_let_symbol)))
	{
	  /* (set! (with-let a b) x), cx = with-let, sc->code = ((with-let a b) x)
	   *   a and x are in the current env, b is in a, we need to evaluate a and x, then
	   *   call (with-let a-value (set! b x-value))
	   */
	  sc->args = cdar(sc->code);
	  sc->code = cadr(sc->code);
	  push_stack(sc, OP_SET_WITH_LET_1, sc->args, sc->code);
	  return(goto_EVAL);
	}
      /* else fall through */
      
    default:                                         /* (set! (1 2) 3) */
      eval_error_no_return(sc, sc->syntax_error_symbol, "no generalized set for ~A", caar_code);
    }
  return(goto_EVAL);
}

static void activate_let(s7_scheme *sc, s7_pointer e)
{
  if (!is_let(e))                    /* (with-let . "hi") */
    eval_error_no_return(sc, sc->wrong_type_arg_symbol, "with-let takes an environment argument: ~A", e);
  if (e == sc->rootlet)
    sc->envir = sc->nil;                             /* (with-let (rootlet) ...) */
  else
    {
      s7_pointer p;
      set_with_let_let(e);
      let_id(e) = ++sc->let_number;
      sc->envir = e;
      for (p = let_slots(e); is_slot(p); p = next_slot(p))
	{
	  s7_pointer sym;
	  sym = slot_symbol(p);
	  if (symbol_id(sym) != sc->let_number)
	    symbol_set_local(sym, sc->let_number, p);
	}
    }
}

static bool tree_match(s7_pointer tree)
{
  if (is_symbol(tree))
    return(is_matched_symbol(tree));
  if (is_pair(tree))
    return((tree_match(car(tree))) || (tree_match(cdr(tree))));
  return(false);
}
#define DO_PRINT 0

static bool do_is_safe(s7_scheme *sc, s7_pointer body, s7_pointer steppers, s7_pointer var_list, bool *has_set)
{
  /* here any (unsafe?) closure or jumping-op (call/cc) or shadowed variable is trouble */
  s7_pointer p;
  if (DO_PRINT) fprintf(stderr, "%s\n", DISPLAY_80(body));

  for (p = body; is_pair(p); p = cdr(p))
    {
      s7_pointer expr;
      expr = car(p);
      if (DO_PRINT) fprintf(stderr, "    %s\n", DISPLAY_80(expr));
      if (is_pair(expr))
	{
	  s7_pointer x;
	  x = car(expr);
	  if (is_symbol(x))
	    {
	      if (is_syntactic(x))
		{
		  opcode_t op;
		  s7_pointer func, vars;
		  func = slot_value(global_slot(x));
		  op = (opcode_t)syntax_opcode(func);
		  switch (op)
		    {
		    case OP_MACROEXPAND:
		      {if (DO_PRINT) fprintf(stderr, "%d\n", __LINE__); return(false);}

		    case OP_QUOTE:
		      break;

		    case OP_LET:
		    case OP_LET_STAR:
		    case OP_LETREC:
		    case OP_LETREC_STAR:
		      {
			s7_pointer nv;
			nv = var_list;

			if (!is_pair(cdr(expr)))
			  {if (DO_PRINT) fprintf(stderr, "%d\n", __LINE__); return(false);}
			if (!s7_is_list(sc, cadr(expr)))
			  {if (DO_PRINT) fprintf(stderr, "%d\n", __LINE__); return(false);}

			for (vars = cadr(expr); is_pair(vars); vars = cdr(vars))
			  {
			    s7_pointer var;
			    if (!is_pair(car(vars)))
			      {if (DO_PRINT) fprintf(stderr, "%d\n", __LINE__); return(false);}
			    var = caar(vars);
			    if (direct_memq(var, ((op == OP_LET) || (op == OP_LETREC)) ? nv : var_list))
			      {if (DO_PRINT) fprintf(stderr, "%d\n", __LINE__); return(false);}
			    nv = cons(sc, var, nv);
			    sc->x = nv;
			  }
			sc->x = sc->nil;
			if (!do_is_safe(sc, cddr(expr), steppers, nv, has_set))
			  {if (DO_PRINT) fprintf(stderr, "%d\n", __LINE__); return(false);}
			break;
		      }

		    case OP_DO:
		      {
			s7_pointer nv;
			if (DO_PRINT) fprintf(stderr, "do loop\n");
			nv = var_list;
			for (vars = cadr(expr); is_pair(vars); vars = cdr(vars))
			  {
			    s7_pointer var;
			    if (!is_pair(car(vars)))
			      {if (DO_PRINT) fprintf(stderr, "%d\n", __LINE__); return(false);}
			    var = caar(vars);
			    if ((direct_memq(var, nv)) ||
				(direct_memq(var, steppers)))
			      {if (DO_PRINT) fprintf(stderr, "%d\n", __LINE__); return(false);}
			  
			    nv = cons(sc, var, nv);
			    sc->x = nv;
			    if ((is_pair(cdar(vars))) &&
				(!do_is_safe(sc, cdar(vars), steppers, nv, has_set)))
			      {
				sc->x = sc->nil;
				{if (DO_PRINT) fprintf(stderr, "%d, step %s\n", __LINE__, DISPLAY_80(cdar(vars))); return(false);}
			      }
			  }
			sc->x = sc->nil;
			if ((!do_is_safe(sc, caddr(expr), steppers, nv, has_set)) ||
			    (!do_is_safe(sc, cdddr(expr), steppers, nv, has_set)))
			  {if (DO_PRINT) fprintf(stderr, "%d\n", __LINE__); return(false);}
			if (DO_PRINT) fprintf(stderr, "do is ok\n");
			break;
		      }

		    case OP_SET:
		      {
			s7_pointer settee;
			settee = cadr(expr);
			if (!is_symbol(settee))             /* (set! (...) ...) which is tricky due to setter functions/macros */
			  {
			    s7_pointer setv;
			    if ((!is_pair(settee)) ||
				(!is_symbol(car(settee))))
			      {if (DO_PRINT) fprintf(stderr, "%d\n", __LINE__); return(false);}
			    setv = find_symbol_unexamined(sc, car(settee));
			    if (!((setv) &&
				  ((is_sequence(setv)) ||
				   ((is_c_function(setv)) && 
				    (is_safe_procedure(c_function_setter(setv)))))))
			      {if (DO_PRINT) fprintf(stderr, "%d\n", __LINE__); return(false);}
			    (*has_set) = true;
			  }
			else
			  {
			    if ((is_pair(cadr(sc->code))) &&
				(is_pair(caadr(sc->code))))
			      {
				bool res;
				set_match_symbol(settee);
				res = tree_match(caadr(sc->code)); /* (set! end ...) in some fashion */
				clear_match_symbol(settee);
				if (res) {if (DO_PRINT) fprintf(stderr, "%d\n", __LINE__); return(false);}
			      }
			    
			    if (!direct_memq(cadr(expr), var_list)) /* is some non-local variable being set? */
			      (*has_set) = true;
			  }
			if (!do_is_safe(sc, cddr(expr), steppers, var_list, has_set))
			  {if (DO_PRINT) fprintf(stderr, "%d\n", __LINE__); return(false);}
			if (!safe_stepper(sc, expr, steppers))  /* is step var's value used as the stored value by set!? */
			  {if (DO_PRINT) fprintf(stderr, "%d\n", __LINE__); return(false);}
		      }
		      break;

		    case OP_LET_TEMPORARILY:
		      {
			s7_pointer lp;
			if ((!is_pair(cdr(expr))) ||
			    (!is_pair(cadr(expr))) ||
			    (!is_pair(cddr(expr))))
			  return(false);
			for (lp = cadr(expr); is_pair(lp); lp = cdr(lp))
			  if ((!is_pair(car(lp))) ||
			      (!is_pair(cdar(lp))) ||
			      (!do_is_safe(sc, cdar(lp), steppers, var_list, has_set)))
			    return(false);
			if (!do_is_safe(sc, cddr(expr), steppers, var_list, has_set))
			  return(false);
			break;
		      }
		      
		    case OP_COND:
		      {
			s7_pointer cp;
			for (cp = cdr(expr); is_pair(cp); cp = cdr(cp))
			  if (!do_is_safe(sc, car(cp), steppers, var_list, has_set))
			    return(false);
			break;
		      }

		    case OP_CASE:
		      {
			s7_pointer cp;
			if (!is_pair(cdr(expr))) return(false);
			if (!do_is_safe(sc, cadr(expr), steppers, var_list, has_set))
			  return(false);
			for (cp = cddr(expr); is_pair(cp); cp = cdr(cp))
			  if ((!is_pair(car(cp))) ||      /* (case x #(123)...) */
			      (!do_is_safe(sc, cdar(cp), steppers, var_list, has_set)))
			    return(false);
			break;
		      }

		    case OP_IF:
		    case OP_WHEN:
		    case OP_UNLESS:
		    case OP_AND:
		    case OP_OR:
		    case OP_BEGIN:
		      if (!do_is_safe(sc, cdr(expr), steppers, var_list, has_set))
			{if (DO_PRINT) fprintf(stderr, "%d\n", __LINE__); return(false);}
		      break;

		    case OP_WITH_LET:
		      return(true);

		    default:
		      {if (DO_PRINT) fprintf(stderr, "%d\n", __LINE__); return(false);}
		    }
		} /* is_syntactic(x=car(expr)) */
	      else
		{
		  /* is a macro, we'll eventually expand it (if *_optimize), but that requires a symbol lookup here and s7_macroexpand */
		  if ((!is_optimized(expr)) ||
		      (!do_is_safe(sc, cdr(expr), steppers, var_list, has_set)))
		    {if (DO_PRINT) fprintf(stderr, "%d, opt: %d\n", __LINE__, is_optimized(expr)); return(false);}
		  else
		    {
		      if (is_setter(x))           /* "setter" includes stuff like cons and vector -- x is a symbol */
			{
			  /* (hash-table-set! ht i 0) -- caddr is being saved, so this is not safe
			   *   similarly (vector-set! v 0 i) etc
			   */
			  if (is_null(cdr(expr)))                         /* (vector) for example */
			    {
			      return((x == sc->vector_symbol) ||
				     (x == sc->list_symbol) ||
				     (x == sc->string_symbol));
			    }
			  if (!direct_memq(cadr(expr), var_list))         /* non-local is being changed */
			    {
			      if ((direct_memq(cadr(expr), steppers)) ||  /* stepper is being set? */
				  (!is_pair(cddr(expr))) ||
				  (!is_pair(cdddr(expr))) ||
				  (is_pair(cddddr(expr))) ||
				  ((x == sc->hash_table_set_symbol) &&
				   (is_symbol(caddr(expr))) && 
				   (direct_memq(caddr(expr), steppers))) ||
				  ((is_symbol(cadddr(expr))) && 
				   (direct_memq(cadddr(expr), steppers))) ||
				  ((is_pair(cadddr(expr))) &&
				   (tree_set_memq_b_pp(steppers, cadddr(expr)))))
				(*has_set) = true;
			    }
			  if (!do_is_safe(sc, cddr(expr), steppers, var_list, has_set))
			    {if (DO_PRINT) fprintf(stderr, "%d\n", __LINE__); return(false);}
			  if (!safe_stepper(sc, expr, steppers))
			    {if (DO_PRINT) fprintf(stderr, "%d\n", __LINE__); return(false);}
			}
		    }
		}
	    } /* is_symbol(x=car(expr)) */
	  else
	    {
	      {if (DO_PRINT) fprintf(stderr, "%d, %s not a symbol\n", __LINE__, DISPLAY_80(x)); return(false);}
	      /* car(expr) ("x") is not a symbol: ((mus-data loc) chan) for example
	       *   but that's actually safe since it's just in effect vector-ref
	       *   there are several examples in dlocsig: ((group-speakers group) i) etc
	       */
	    }
	}
    }
  return(true);
}

static bool preserves_type(s7_scheme *sc, uint32_t x)
{
  return((x == sc->add_class) || 
	 (x == sc->subtract_class) || 
	 (x == sc->multiply_class));
}

static bool is_simple_expression(s7_scheme *sc, s7_pointer x)
{
  return((!is_pair(x)) ||
	 ((is_optimized(x)) &&
	  (is_all_x_safe(sc, x)) &&
	  (car(x) != sc->quote_symbol)));
}
	
static bool tree_has_definers(s7_scheme *sc, s7_pointer tree)
{
  clear_symbol_list(sc);
  add_symbol_to_list(sc, sc->define_symbol);
  add_symbol_to_list(sc, sc->define_macro_symbol);
  add_symbol_to_list(sc, sc->define_bacro_symbol);
  add_symbol_to_list(sc, sc->define_star_symbol);
  add_symbol_to_list(sc, sc->define_macro_star_symbol);
  add_symbol_to_list(sc, sc->define_bacro_star_symbol);
  add_symbol_to_list(sc, sc->define_expansion_symbol);
  add_symbol_to_list(sc, sc->define_constant_symbol);
  add_symbol_to_list(sc, sc->varlet_symbol);
  add_symbol_to_list(sc, sc->provide_symbol);             /* local *features*! */
  return(tree_set_memq(sc, tree));
}

static s7_pointer check_do(s7_scheme *sc)
{
  s7_pointer x;
  
  if ((!is_pair(sc->code)) ||                             /* (do . 1) */
      ((!is_pair(car(sc->code))) &&                       /* (do 123) */
       (is_not_null(car(sc->code)))))                     /* (do () ...) is ok */
    eval_error(sc, "do: var list is not a list: ~S", sc->code);

  if (!is_pair(cdr(sc->code)))                            /* (do () . 1) */
    eval_error(sc, "do body is messed up: ~A", sc->code);

  if ((!is_pair(cadr(sc->code))) &&                       /* (do ((i 0)) 123) */
      (is_not_null(cadr(sc->code))))                      /* no end-test? */
    eval_error(sc, "do: end-test and end-value list is not a list: ~A", sc->code);

  if (is_pair(car(sc->code)))
    {
      clear_symbol_list(sc);
      for (x = car(sc->code); is_pair(x); x = cdr(x))
	{
	  s7_pointer y;
	  y = car(x);
	  if (!(is_pair(y)))                             /* (do (4) (= 3)) */
	    eval_error(sc, "do: variable name missing? ~A", sc->code);

	  if (!is_symbol(car(y)))                        /* (do ((3 2)) ()) */
	    eval_error(sc, "do step variable: ~S is not a symbol?", y);

	  if (is_immutable_symbol(car(y)))               /* (do ((pi 3 (+ pi 1))) ((= pi 4)) pi) */
	    eval_error(sc, "do step variable: ~S is immutable", y);

	  if (is_pair(cdr(y)))
	    {
	      if (!is_pair(cddr(y)))
		{
		  if (is_not_null(cddr(y)))             /* (do ((i 0 . 1)) ...) */
		    eval_error(sc, "do: step variable info is an improper list?: ~A", x);
		}
	      else
		{
		  if (is_not_null(cdddr(y)))            /* (do ((i 0 1 (+ i 1))) ...) */
		    eval_error(sc, "do: step variable info has extra stuff after the increment: ~A", x);
		}
	    }
	  else eval_error(sc, "do: step variable has no initial value: ~A", x);
	  set_local(car(y));

	  if (symbol_is_in_list(sc, car(y)))            /* (do ((i 0 (+ i 1)) (i 2))...) */
	    eval_error(sc, "duplicate identifier in do: ~A", x);
	  add_symbol_to_list(sc, car(y));
	}
      if (is_not_null(x))                               /* (do ((i 0 i) . 1) ((= i 1))) */
	eval_error(sc, "do: list of variables is improper: ~A", sc->code);
    }

  if (is_pair(cadr(sc->code)))
    {
      for (x = cadr(sc->code); is_pair(x); x = cdr(x));
      if (is_not_null(x))
	eval_error(sc, "stray dot in do end section? ~A", sc->code);
    }

  for (x = cddr(sc->code); is_pair(x); x = cdr(x));
  if (is_not_null(x))
    eval_error(sc, "stray dot in do body? ~A", sc->code);

  if ((is_overlaid(sc->code)) &&
      (has_opt_back(sc->code)))
    {
      s7_pointer vars, end, body, p;
      int32_t nvars, nsteps = 0;

      pair_set_syntax_symbol(sc->code, sc->do_unchecked_symbol);

      end = cadr(sc->code);
      if (!is_pair(end))              /* () as end+result -- uncommon, normally not optimizable anyway -- use (#f) instead */
	return(sc->code);             
      if (is_simple_expression(sc, car(end)))
	set_x_call(end, all_x_eval(sc, end, sc->envir, let_symbol_is_safe));
      else return(sc->code);          /* if end is not allxable, give up */

      vars = car(sc->code);
      if (is_null(vars))
	{
	  pair_set_syntax_symbol(sc->code, sc->do_no_vars_symbol);
	  return(sc->nil);
	}
      
      /* an annoying kludge -- define in the body can clobber the step expressions set up below! 
       *      (let ((x 2)) (do ((i 0 (+ i x))) ((= i 4)) (define x 1) (display i)) (newline)) -- steps by 1
       */
      body = cddr(sc->code);
      if (tree_has_definers(sc, body))
	return(sc->code);

      /* TODO: set up init/step choices */
      for (nvars = 0, p = vars; is_pair(p); nvars++, p = cdr(p))
	if (is_pair(cddar(p)))
	  {
	    x = car(p);
	    nsteps++;
	  }
      /* 1/1 2/2 1/n n/m.  0/n almost never (only make-index where currently cell_optimize has no chance) */
#if 1
      if ((nvars > 1) &&
	  (nsteps == 1))
	{
	  /* temporary */
	  for (nvars = 0, p = vars; is_pair(p); nvars++, p = cdr(p))
	    {
	      s7_pointer v;
	      v = car(p);
	      if (is_simple_expression(sc, cadr(v)))
		set_x_call(cdr(v), all_x_eval(sc, cdr(v), sc->envir, let_symbol_is_safe));
	      else return(sc->code);
	    }

	  if (!is_simple_expression(sc, caddr(x)))  /* x is the stepper */
	    return(sc->code);
	  if (is_pair(caddr(x)))
	    {
	      set_x_call(cddr(x), all_x_eval(sc, cddr(x), sc->envir, let_symbol_is_safe));
	      if (c_callee(cddr(x)) == all_x_c_add1)
		{
		  if ((c_callee(end) == all_x_c_ss) &&
		      (caar(end) == sc->eq_symbol) &&
		      (cadar(end) == car(x)))
		    {
		      bool has_set = false;
		      if ((is_null(cdr(body))) &&
			  (do_is_safe(sc, body, sc->w = list_1(sc, car(x)), sc->nil, &has_set)))
			{
			  if (!has_set)
			    {
			      pair_set_syntax_symbol(sc->code, sc->dotimes_one_step_symbol); /* safe dotimes */
			      return(sc->nil);
			    }
			}
		    }
		}
	    }
	  
	  /* inits all non-pair, or all allx, or any */
	  /* steppers allx, +1 -1 cdr */
	}
#endif

      /* 1/1 is of course the biggy -- can 1/n share its code? */
      /* need to start: do_op_1step_no_opt[_1] = set up frame goto check end, then push _1, set code, goto begin1, step via allx
       * :set frame non-steppers, then stepper at front, goto CHECK
       * _1: step, CHECK: check end, push, code=, go begin1
       *
       * then add op_do_1step_add1: like do_no_vars but frame setup is more complicated, and need
       *   all the safe-stepper mutable ints and so on 
       */ 

      /* are inits allxable, are steppers allxable, +/- by 1 by int, cdr, are there shadowing problems
       * (=|>= end lim) where end is stepper or (null? stp)
       * is body safe, 1 expr
       * return null or constant or 1-expr
       */

      /* -------------------------------------------------------------------------------- */
      /* old version from here */

      /* (define (hi) (do ((i 0 (+ i 1))) ((= i 3)) (display i)) (newline)) */
      /* (define (hi) (do ((i 1.5 (+ i 1))) ((= i 2.5)) (display i) (newline)))
       *   in OP_SAFE_DOTIMES, for example, if init value is not an integer, it goes to OP_SIMPLE_DO
       * remaining optimizable cases: we can step by 1 and use = for end, and yet simple_do(_p) calls the functions
       * geq happens as often as =, and -1 as step
       * also cdr as step to is_null as end
       * also do body is optimized expr: vector_set_3 via hop_safe_c_sss for example or (vset v i (vref w i))
       */
      if ((is_pair(end)) && (is_pair(car(end))) &&
	  (is_pair(vars)) && (is_null(cdr(vars))) &&
	  (is_pair(body)))
	{
	  /* loop has one step variable, and normal-looking end test
	   */
	  s7_pointer v;
	  v = car(vars);
	  if ((is_pair(car(body))) &&
	      (is_symbol(caar(body))) &&
	      (safe_list_length(sc, v) == 3) &&
	      ((!is_pair(cadr(v))) ||
	       (is_h_safe_c_c(cadr(v)))))
	    {
	      s7_pointer step_expr;
	      step_expr = caddr(v);

	      if ((is_optimized(step_expr)) &&
		  (((optimize_op(step_expr) == HOP_SAFE_C_SC) && (car(v) == cadr(step_expr))) ||
		   ((is_h_safe_c_c(step_expr)) && 
		    (is_pair(cdr(step_expr))) &&         /* ((v 0 (+))) */
		    (car(v) == cadr(step_expr)) &&
		    ((opt_cfunc(step_expr) == add_cs1) || (opt_cfunc(step_expr) == add_cl1) || 
		     (opt_cfunc(step_expr) == subtract_cs1) || (opt_cfunc(step_expr) == subtract_cl1))) ||
		   ((optimize_op(step_expr) == HOP_SAFE_C_CS) && (car(v) == caddr(step_expr)))))
		{
		  /* step var is (var const|symbol (op var const)|(op const var))
		   */
		  end = car(end);

		  if ((is_optimized(end)) &&
		      (car(v) == cadr(end)) &&
		      (is_pair(cddr(end))) &&      /* end: (zero? n) */
		      (cadr(end) != caddr(end)) &&
#if (!WITH_GMP)
		      ((opt_any1(end) == equal_s_ic) ||
		       (optimize_op(end) == HOP_SAFE_C_SS) || (optimize_op(end) == HOP_SAFE_C_LL) ||
		       (optimize_op(end) == HOP_SAFE_C_SC))
#else
		      ((optimize_op(end) == HOP_SAFE_C_SS) || (optimize_op(end) == HOP_SAFE_C_LL) ||
		       (optimize_op(end) == HOP_SAFE_C_SC))
#endif
		      )
		    {
		      /* end var is (op var const|symbol) using same var as step
		       *   so at least we can use SIMPLE_DO
		       */
		      bool has_set = false, one_line;
		      one_line = ((is_null(cdr(body))) && (is_pair(car(body))));

		      if ((opt_cfunc(step_expr) == add_cs1) || (opt_cfunc(step_expr) == add_cl1))
			{
			  set_c_function(step_expr, add_s1);
			  set_optimize_op(step_expr, HOP_SAFE_C_SC);
			}
		      if ((opt_cfunc(step_expr) == subtract_cs1) || (opt_cfunc(step_expr) == subtract_cl1))
			{
			  set_c_function(step_expr, subtract_s1);
			  set_optimize_op(step_expr, HOP_SAFE_C_SC);
			}
#if (!WITH_GMP)
		      if (opt_cfunc(end) == equal_s_ic)
			{
			  set_c_function(end, equal_2);
			  set_optimize_op(end, HOP_SAFE_C_SC);
			}
#endif

		      pair_set_syntax_symbol(sc->code, sc->simple_do_symbol);              /* simple_do: 1 var easy step/end */

		      if ((one_line) &&
			  ((!is_optimized(car(body))) || (op_no_hop(car(body)) != OP_SAFE_C_C)) &&
			  (is_syntactic_symbol(caar(body))) &&
			  (s7_is_integer(caddr(step_expr))) &&
			  (s7_integer(caddr(step_expr)) == 1) &&
			  (c_function_class(opt_cfunc(step_expr)) == sc->add_class) &&
			  /* we check above that (car(v) == cadr(step_expr))
			   *    and that         (car(v) == cadr(end))
			   */
			  ((c_function_class(opt_cfunc(end)) == sc->equal_class) ||
			   (opt_cfunc(end) == geq_2)))
			{
			  pair_set_syntax_op(car(body), symbol_syntax_op(caar(body)));
			  set_opt_pair2(sc->code, caddr(caar(sc->code)));
			  pair_set_syntax_symbol(sc->code, sc->dotimes_p_symbol);           /* dotimes_p: simple + syntax body + 1 expr */
			  /* 5 bench, -60 gen, 653 all, 1423 snd-test */
			}

		      /* now look for the very common dotimes case
		       */
		      if ((((s7_is_integer(caddr(step_expr))) &&
			    (s7_integer(caddr(step_expr)) == 1)) ||
			   ((s7_is_integer(cadr(step_expr))) &&
			    (s7_integer(cadr(step_expr)) == 1))) &&
			  (c_function_class(opt_cfunc(step_expr)) == sc->add_class) &&
			  ((c_function_class(opt_cfunc(end)) == sc->equal_class) ||
			   (opt_cfunc(end) == geq_2)))
			{
			  if (do_is_safe(sc, body, sc->w = list_1(sc, car(v)), sc->nil, &has_set))
			    {
			      /* we're stepping by +1 and going to =
			       *   the final integer check has to wait until run time (symbol value dependent)
			       */
			      pair_set_syntax_symbol(sc->code, sc->safe_do_symbol);          /* safe_do: body is safe, step by 1 */
			      if ((!has_set) &&
				  (c_function_class(opt_cfunc(end)) == sc->equal_class))
				pair_set_syntax_symbol(sc->code, sc->safe_dotimes_symbol);   /* safe_dotimes: end is = */
			    }
			}
		      return(sc->nil); 
		    }
		}
	    }
	}

      /* we get here if there is more than one local var or anything "non-simple" about the rest
       */
      /* (define (hi) (do ((i 0 (+ i 1))) ((= i 3)) (display i)) (newline))
       * (define (hi) (do ((i 0 (+ i 1)) (j 1 (+ j 1))) ((= i 3)) (display j))(newline))
       */
      for (p = vars; is_pair(p); p = cdr(p))
	{
	  s7_pointer var;
	  var = car(p);
	  
	  if ((!is_all_x_safe(sc, cadr(var))) ||
	      ((is_pair(cddr(var))) &&
	       (!is_all_x_safe(sc, caddr(var)))))
	    {
	      s7_pointer q;
#if 0
	      fprintf(stderr, "%s bad %s %s\n", DISPLAY(var), 
		      (is_optimized(cadr(var))) ? opt_names[optimize_op(cadr(var))] : "?", 
		      ((is_pair(cddr(var))) && (is_optimized(caddr(var)))) ? opt_names[optimize_op(caddr(var))] : "?");
#endif
	      for (q = vars; q != p; q = cdr(q))
		clear_match_symbol(caar(q));
	      return(sc->code);
	    }
	  set_match_symbol(car(var));
	}
      /* we want to use the pending_value slot for other purposes, so make sure
       *   the current val is not referred to in any trailing step exprs.  The inits
       *   are ok because at init-time, the new frame is not connected.
       * another tricky case: current var might be used in previous step expr(!)
       *   and worse, the loop env can be changed by a top-level define in the body,
       *   clobbering the step_expression accessors in dox_step!
       */
      for (p = vars; is_pair(p); p = cdr(p))
	{
	  s7_pointer var, val;
	  var = car(p);
	  val = cddr(var);
	  if (is_pair(val))
	    {
	      var = car(var);
	      clear_match_symbol(var); /* ignore current var */
	      if (tree_match(car(val)))
		{
		  s7_pointer q;
		  for (q = vars; is_pair(q); q = cdr(q))
		    clear_match_symbol(caar(q));
		  return(sc->code);
		}
	      set_match_symbol(var);
	    }
	}
      for (p = vars; is_pair(p); p = cdr(p))
	clear_match_symbol(caar(p));
  
      /* end and steps look ok! */
      pair_set_syntax_symbol(sc->code, sc->dox_symbol);                                     /* dox: vars/end are allxable */

      /* each step expr is safe so not an explicit set!
       *   the symbol_is_safe check in all_x_eval needs to see the do envir, not the caller's
       *   but that means the is_all_x_safe check above also needs to use the local env?
       */
      for (p = vars; is_pair(p); p = cdr(p))
	{
	  s7_pointer var;
	  var = car(p);
	  if (is_pair(cdr(var))) 
	    set_x_call(cdr(var), all_x_eval(sc, cdr(var), sc->envir, let_symbol_is_safe)); /* init val */
	  if (is_pair(cddr(var))) 
	    {
	      s7_pointer step_expr;
	      step_expr = caddr(var);
	      set_x_call(cddr(var), all_x_eval(sc, cddr(var), vars, do_symbol_is_safe)); /* sets opt2(cddr(var)), not opt1 */

	      if ((is_pair(step_expr)) &&
		  (car(step_expr) != sc->quote_symbol) &&     /* opt_cfunc(==opt1) might not be set in this case (sigh) */
		  ((preserves_type(sc, c_function_class(opt_cfunc(step_expr)))) || /* add etc */
		   (car(step_expr) == sc->cdr_symbol) ||
		   (car(step_expr) == sc->cddr_symbol) ||	
		   ((is_pair(cadr(var))) &&
		    (is_pair(s7_procedure_signature(sc, c_function_base(opt_cfunc(step_expr))))) &&
		    (car(s7_procedure_signature(sc, c_function_base(opt_cfunc(step_expr)))) != sc->T) &&
		    (caadr(var) == car(step_expr)))))
		/* i.e. accept char-position as init/step, but not iterate */
		set_safe_stepper(cddr(var));
	    }
	}
      /* there are only a couple of cases in snd-test where a multi-statement do body is completely all-x-able */
      return(sc->nil);
    }
  return(sc->code);
}

static s7_pointer make_do_frame(s7_scheme *sc)
{
  int64_t id;
  s7_pointer frame, vars;

  new_frame(sc, sc->envir, frame);           /* new frame is not tied into the symbol lookup process yet */
  sc->temp11 = frame;
  id = let_id(frame);

  for (vars = car(sc->code); is_pair(vars); vars = cdr(vars))
    {
      s7_pointer v, slot;
      v = car(vars);
      new_cell_no_check(sc, slot, T_SLOT);
      slot_set_symbol(slot, car(v));
      slot_set_value(slot, sc->F);
      set_next_slot(slot, let_slots(frame)); /* GC protect it right away */
      let_set_slots(frame, slot);
      symbol_set_local(slot_symbol(slot), id, slot);
      slot_set_value(slot, c_call(cdr(v))(sc, cadr(v)));
      slot_set_expression(slot, cddr(v));
      if (is_pair(cddr(v)))
	dox_set_slot1(frame, slot);
    }

  sc->temp11 = sc->nil;
  return(frame);
}

static void update_steppers(s7_scheme *sc)
{
  /* TODO: probably need to use pending_expr here etc */
  s7_pointer v;
  for (v = let_slots(sc->envir); is_slot(v); v = next_slot(v))
    {
      s7_pointer step_expr;
      step_expr = slot_expression(v);
      if (!is_null(step_expr))
	slot_set_value(v, c_call(step_expr)(sc, car(step_expr)));
    }
}

static bool has_safe_steppers(s7_scheme *sc, s7_pointer frame)
{
  s7_pointer slot;
  for (slot = let_slots(frame); is_slot(slot); slot = next_slot(slot))
    {
      s7_pointer step_expr, val;
      val = slot_value(slot);
      step_expr = slot_expression(slot);

      if (!is_pair(step_expr))
	{
	  if ((is_null(step_expr)) ||
	      (type(step_expr) == type(val)))
	    set_safe_stepper(slot);
	  else clear_safe_stepper(slot);
	}
      else
	{
	  if (is_safe_stepper(step_expr))
	    {
	      if (is_t_integer(val)) /* a temporary kludge */
		{
		  sc->pc = 0;
		  if (int_optimize(sc, step_expr))
		    set_safe_stepper(slot);
		  else clear_safe_stepper(slot);
		}
	      else 
		{
		  if (is_real(val)) /* a temporary kludge */
		    {
		      sc->pc = 0;
		      if (float_optimize(sc, step_expr))
			set_safe_stepper(slot);
		      else clear_safe_stepper(slot);
		    }
		  else set_safe_stepper(slot);  /* ?? shouldn't this check types ?? */
		}
	    }
	}
      
      if (!is_safe_stepper(slot))
	return(false);
    }
  return(true);
}


static int32_t dox_ex(s7_scheme *sc)
{
  /* any number of steppers using dox exprs, end also dox, body and end result arbitrary.
   *    since all these exprs are local, we don't need to jump until the body
   */
  int64_t id;
  s7_pointer frame, vars, slot, code, end, endp;
  s7_function endf;

  /* fprintf(stderr, "dox_ex: %d %s\n", is_unsafe_do(sc->code), DISPLAY_80(sc->code)); */

#if 0
  /* teq tmac index tref tlet tcopy tauto tform tmap titer(much changed) tsort toss-up, 
   * lt(1) tall tgen(1.5) thash(much slower -- reader? yes--the let(*) -- set has type[char-position?] trouble), call(1)
   * fft faster(1.5)
   */
  if (!pair_no_opt(sc->code))
    {
      endf = s7_optimize(sc, cons_unchecked(sc, cons(sc, sc->do_symbol, sc->code), sc->nil));
      if (endf)
	{
	  sc->value = endf(sc, sc->code);
	  sc->code = sc->nil;
	  return(goto_SAFE_DO_END_CLAUSES);
	}
      set_pair_no_opt(sc->code);
    }
#endif

  new_frame(sc, sc->envir, frame);   /* new frame is not tied into the symbol lookup process yet */
  sc->temp11 = frame;
  for (vars = car(sc->code); is_pair(vars); vars = cdr(vars))
    {
      s7_pointer expr, val;
      expr = cadar(vars);
      if (is_pair(expr))
	{
	  if (car(expr) == sc->quote_symbol)
	    val = cadr(expr);
	  else val = c_call(cdar(vars))(sc, expr);
	}
      else
	{
	  if (is_symbol(expr))
	    val = find_symbol_checked(sc, expr); /* symbol as init? */
	  else val = expr;
	}
      new_cell_no_check(sc, slot, T_SLOT);
      slot_set_symbol(slot, caar(vars));
      slot_set_value(slot, val);
      slot_set_expression(slot, cddar(vars));
      set_next_slot(slot, let_slots(frame));
      let_set_slots(frame, slot);
    }
  
  sc->envir = frame;
  sc->temp11 = sc->nil;
  id = let_id(frame);
  for (slot = let_slots(frame); is_slot(slot); slot = next_slot(slot))
    symbol_set_local(slot_symbol(slot), id, slot);
  
  end = cadr(sc->code);
  endp = car(end);
  endf = c_callee(end);
  if (is_true(sc, sc->value = endf(sc, endp)))
    {
      /* if no end result exprs, we return nil, but others probably #<unspecified>
       *    (let ((x (do ((i 0 (+ i 1))) (#t)))) x) -> ()
       */
      sc->code = cdr(end);
      return(goto_DO_END_CLAUSES);
    }
	    
  code = cddr(sc->code);
  if (is_null(code)) /* no body? */
    {
      s7_pointer slots;
      if (endf == all_x_c_c)
	{
	  endf = c_callee(endp);
	  endp = cdr(endp);
	}

      slots = let_slots(sc->envir);
      if ((is_null(next_slot(slots))) && (is_pair(slot_expression(slots))))
	{
	  s7_function f;
	  s7_pointer a;

	  f = c_callee(slot_expression(slots));
	  a = car(slot_expression(slots));
	  if (f == all_x_c_c)
	    {
	      f = c_callee(a);
	      a = cdr(a);
	    }

	  while (true) /* thash titer */
	    {
	      slot_set_value(slots, f(sc, a));
	      if (is_true(sc, sc->value = endf(sc, endp)))
		{
		  sc->code = cdr(end);
		  return(goto_DO_END_CLAUSES);
		}
	    }
	}
      else
	{
	  while (true)
	    {
	      s7_pointer slt;
	      for (slt = slots; is_slot(slt); slt = next_slot(slt))
		if (is_pair(slot_expression(slt)))
		  slot_set_value(slt, c_call(slot_expression(slt))(sc, car(slot_expression(slt))));
	      if (is_true(sc, sc->value = endf(sc, endp)))
		{
		  sc->code = cdr(end); 
		  return(goto_DO_END_CLAUSES);
		}
	    }
	}
    }
  else /* there is a body */
    {
      if (!is_unsafe_do(sc->code))
	{
	  s7_pointer slots;
	  slots = let_slots(sc->envir);
	  /* is let activated? also multiexpr body  and other all_x? */

	  if ((is_null(cdr(code))) &&
	      (is_pair(car(code))))
	    {
	      s7_pointer lcode;
	      s7_function body = NULL;
	      lcode = car(code);

	      if ((!pair_no_opt(code)) &&
		  (has_safe_steppers(sc, sc->envir)))
		{
		  body = s7_optimize_nr(sc, code);
		  if (!body) 
		    set_pair_no_opt(code);
		}
	      if (!body)
		{
		  if (is_all_x_safe(sc, lcode))
		    body = all_x_eval(sc, code, sc->envir, let_symbol_is_safe);
		}
	      if (body)
		{
		  while (true)
		    {
		      s7_pointer slot;
		      body(sc, lcode);
		      for (slot = slots; is_slot(slot); slot = next_slot(slot))
			if (is_pair(slot_expression(slot)))
			  slot_set_value(slot, c_call(slot_expression(slot))(sc, car(slot_expression(slot))));
		      if (is_true(sc, sc->value = endf(sc, endp)))
			{
			  sc->code = cdr(end);
			  return(goto_DO_END_CLAUSES);
			}
		    }
		}
	    }
	  else /* more than one expr */
	    {
	      s7_pointer p;
	      bool use_opts = false;
	      int32_t body_len = 0;
	      p = code;

	      if ((!pair_no_opt(code)) &&
		  (has_safe_steppers(sc, sc->envir)))
		{
		  if (setjmp(sc->opt_exit) == 0)
		    {
		      start_opts(sc);
		      for (; is_pair(p); p = cdr(p), body_len++)
			if (!cell_optimize(sc, p))
			  {
			    set_pair_no_opt(code);
			    p = code;
			    break;
			  }
		      use_opts = is_null(p);
		    }
		}

	      if (p == code)
		{
		  for (; is_pair(p); p = cdr(p))
		    if (!is_all_x_safe(sc, car(p)))
		      break;
		}

	      if (is_null(p))
		{
		  int32_t i;
		  if (!use_opts)
		    annotate_args(sc, code, sc->envir);

		  while (true)
		    {
		      s7_pointer slot;
		      if (use_opts)
			{
			  sc->pc = 0;
			  for (i = 0; i < body_len; i++)
			    {
			      opt_info *o;
			      o = sc->opts[sc->pc];
			      o->v7.fp(o);
			      sc->pc++;
			    }
			}
		      else
			{
			  for (p = code; is_pair(p); p = cdr(p))
			    c_call(p)(sc, car(p));
			}

		      for (slot = slots; is_slot(slot); slot = next_slot(slot))
			if (is_pair(slot_expression(slot)))
			  slot_set_value(slot, c_call(slot_expression(slot))(sc, car(slot_expression(slot))));
		      if (is_true(sc, sc->value = endf(sc, endp)))
			{
			  sc->code = cdr(end);
			  return(goto_DO_END_CLAUSES);
			}
		    }
		}
	    }
	  set_unsafe_do(sc->code);
	}
    }

  if ((is_null(cdr(code))) && /* one expr */
      (is_pair(car(code))))
    {
      code = car(code);
      
      if ((typesflag(code) == SYNTACTIC_PAIR) ||
	  (typesflag(car(code)) == SYNTACTIC_TYPE))
	{
	  push_stack_no_args(sc, OP_DOX_STEP_P, sc->code);
	  
	  if (typesflag(code) == SYNTACTIC_PAIR)
	    sc->cur_op = (opcode_t)pair_syntax_op(code);
	  else
	    {
	      sc->cur_op = (opcode_t)symbol_syntax_op(car(code));
	      pair_set_syntax_op(code, sc->cur_op);
	      set_syntactic_pair(code);
	    }
	  sc->code = cdr(code);
	  return(goto_START_WITHOUT_POP_STACK);
	}
    }
  return(fall_through);
}

	  
static int32_t simple_do_ex(s7_scheme *sc, s7_pointer code)
{
  s7_pointer body, step_expr, step_var, ctr, end;
  s7_function stepf, endf;
  s7_function func;

  /* fprintf(stderr, "%s: %s\n", __func__, DISPLAY_80(opt_pair2(code))); */

  body = car(opt_pair2(code));

#if DEBUGGING
  if (!is_symbol(car(body)))
    {
      fprintf(stderr, "simple_do_ex car(body) not symbol: %s\n", DISPLAY_80(code));
      return(fall_through);
    }
#endif
    
  if (!pair_no_opt(opt_pair2(code)))
    {
      func = s7_optimize_nr(sc, opt_pair2(code));
      if (!func) 
	{
	  set_pair_no_opt(opt_pair2(code));
	  return(fall_through);
	}
    }
  else return(fall_through);

  /* func must be set */
  step_expr = caddr(caar(code));
  stepf = c_callee(step_expr);
  endf = c_callee(caadr(code));
  ctr = dox_slot1(sc->envir);
  end = dox_slot2(sc->envir);
  step_var = caddr(step_expr);

#if (!WITH_GMP)
  if (((stepf == g_subtract_s1) && (endf == g_less_s0)) ||
      ((stepf == g_add_s1) && (endf == g_equal_2)))
    set_safe_stepper(ctr);
#endif
  
  if ((stepf == g_add_s1) &&
      (is_integer(slot_value(ctr))) &&
      (endf == g_equal_2) &&
      (is_integer(slot_value(end))))
    {
      s7_int i, start, stop;
      start = integer(slot_value(ctr));
      stop = integer(slot_value(end));
	  
      if (func == opt_cell_any_nr)
	{
	  opt_info *o;
	  s7_pointer (*fp)(void *o);
	  cur_sc = sc;
	  o = sc->opts[0];
	  fp = o->v7.fp;
	  for (i = start; i < stop; i++)
	    {
	      slot_set_value(ctr, make_integer(sc, i));
	      sc->pc = 0;
	      fp(o);
	    }
	}
      else
	{
	  for (i = start; i < stop; i++)
	    {
	      slot_set_value(ctr, make_integer(sc, i));
	      func(sc, body);
	    }
	}

      sc->value = sc->T;
      sc->code = cdadr(code);
      return(goto_DO_END_CLAUSES);
    }

  while (true)
    {
      func(sc, body);
	  
      set_car(sc->t2_1, slot_value(ctr));
      set_car(sc->t2_2, step_var);
      slot_set_value(ctr, stepf(sc, sc->t2_1));
	  
      set_car(sc->t2_1, slot_value(ctr));
      set_car(sc->t2_2, slot_value(end));
      if (is_true(sc, sc->value = endf(sc, sc->t2_1)))
	{
	  sc->code = cdadr(code);
	  return(goto_DO_END_CLAUSES);
	}
    }
  return(fall_through);
}

static bool opt_dotimes(s7_scheme *sc, s7_pointer code, s7_pointer scc, bool safe_step)
{
  int32_t i, body_len;
  s7_int end;

  if (sc->safety > CLM_OPTIMIZATION_SAFETY) return(false);

  if (safe_step)
    set_safe_stepper(sc->args);
  else set_safe_stepper(dox_slot1(sc->envir));
  body_len = s7_list_length(sc, code);

  /* fprintf(stderr, "opt_dotimes: %s %d %d %d\n", DISPLAY(code), safe_step, body_len, pair_no_opt(code)); */

  /* I think safe_step means the stepper is completely unproblematic */
  if (body_len == 1) /* && (safe_step)) */
    {
      s7_function func;

      if (pair_no_opt(code)) return(false);
      func = s7_optimize_nr(sc, code);
      if (!func) 
	{
	  set_pair_no_opt(code);
	  return(false);
	}

      end = denominator(slot_value(sc->args));
      if (safe_step)
	{
	  s7_pointer stepper;
	  slot_set_value(sc->args, stepper = make_mutable_integer(sc, integer(slot_value(sc->args))));
	  if ((func == opt_float_any_nr) ||
	      (func == opt_cell_any_nr))
	    {
	      opt_info *o;
	      cur_sc = sc;
	      o = sc->opts[0];
	      if (func == opt_float_any_nr)
		{
		  s7_double (*fd)(void *o);
		  fd = o->v7.fd;
		  if ((fd == opt_d_id_sf) && /* by far the most common case in clm: (outa i ...) etc */
		      (is_slot(o->v1.p)) &&
		      (stepper == slot_value(o->v1.p)))
		    {
		      opt_info *o1;
		      s7_d_id_t f0;
		      f0 = o->v3.d_id_f;
		      o1 = sc->opts[1];
		      fd = o1->v7.fd;
		      for (; integer(stepper) < end; integer(stepper)++)
			{
			  sc->pc = 1;
			  f0(integer(stepper), fd(o1));
			}
		    }
		  else
		    {
		      for (; integer(stepper) < end; integer(stepper)++)
			{
			  sc->pc = 0;
			  fd(o);
			}
		    }
		}
	      else
		{
		  s7_pointer (*fp)(void *o);
		  fp = o->v7.fp;
		  for (; integer(stepper) < end; integer(stepper)++)
		    {
		      sc->pc = 0;
		      fp(o);
		    }
		}
	    }
	  else
	    {
	      for (; integer(stepper) < end; integer(stepper)++)
		func(sc, car(code)); 
	    }
	}
      else
	{
	  s7_int step;
	  s7_pointer step_slot, end_slot;
	  step_slot = dox_slot1(sc->envir);
	  end_slot = dox_slot2(sc->envir);
	  
	  if (func == opt_cell_any_nr)
	    {
	      opt_info *o;
	      s7_pointer (*fp)(void *o);
	      cur_sc = sc;
	      o = sc->opts[0];
	      fp = o->v7.fp;
	      while (true)
		{
		  sc->pc = 0;
		  fp(o);
		  /* TODO: are these type checks needed now? -- could check setter+integer etc in cell_optimize */
		  if (!is_integer(slot_value(step_slot)))
		    {
		      slot_set_value(step_slot, g_add_1s(sc, set_plist_2(sc, small_int(1), slot_value(step_slot))));
		      if (is_true(sc, g_equal_2(sc, set_plist_2(sc, slot_value(step_slot), slot_value(end_slot))))) break;
		    }
		  else
		    {
		      step = s7_integer(slot_value(step_slot)) + 1;
		      slot_set_value(step_slot, make_integer(sc, step));
		      if (step == s7_integer(slot_value(end_slot))) break;
		    }
		}
	    }
	  else
	    {
	      while (true)
		{
		  func(sc, car(code));
		  if (!is_integer(slot_value(step_slot)))
		    {
		      slot_set_value(step_slot, g_add_1s(sc, set_plist_2(sc, small_int(1), slot_value(step_slot))));
		      if (is_true(sc, g_equal_2(sc, set_plist_2(sc, slot_value(step_slot), slot_value(end_slot))))) break;
		    }
		  else
		    {
		      step = s7_integer(slot_value(step_slot)) + 1;
		      slot_set_value(step_slot, make_integer(sc, step));
		      if (step == s7_integer(slot_value(end_slot))) break;
		    }
		}
	    }
	}
      sc->value = sc->T;
      sc->code = cdadr(scc);
      return(true);
    }

  if (setjmp(sc->opt_exit) == 0)
    {
      s7_pointer p;
      start_opts(sc);
      for (p = code; is_pair(p); p = cdr(p))
	if (!float_optimize(sc, p))
	  return(false);
    }
  else return(false);
  /* TODO: here and in do_let, generalize the body exprs */

  end = denominator(slot_value(sc->args));
  if (safe_step)
    {
      s7_pointer stepper;
      slot_set_value(sc->args, stepper = make_mutable_integer(sc, integer(slot_value(sc->args))));
      for (; integer(stepper) < end; integer(stepper)++)
	{
	  sc->pc = 0;
	  for (i = 0; i < body_len; i++)
	    {
              sc->opts[sc->pc]->v7.fd(sc->opts[sc->pc]);
	      sc->pc++;
	    }
	}
    }
  else
    {
      s7_pointer step_slot, end_slot;
      step_slot = dox_slot1(sc->envir);
      end_slot = dox_slot2(sc->envir);
      while (true)
	{
	  s7_int step;
	  sc->pc = 0;
	  for (i = 0; i < body_len; i++)
	    {
	      sc->opts[sc->pc]->v7.fd(sc->opts[sc->pc]);
	      sc->pc++;
	    }
	  
	  step = s7_integer(slot_value(step_slot)) + 1;
	  slot_set_value(step_slot, make_integer(sc, step));
	  if (step == s7_integer(slot_value(end_slot))) break;
	}
    }
  sc->value = sc->T;
  sc->code = cdadr(scc);
  return(true);
}
 

static int32_t do_let(s7_scheme *sc, s7_pointer step_slot, s7_pointer scc, bool safe_case)
{
  s7_pointer let_body, p = NULL, let_vars, let_code;
  bool let_star;
  s7_pointer old_e, stepper;
  int32_t body_len, var_len;

  if (sc->safety > CLM_OPTIMIZATION_SAFETY) return(fall_through);
  /* fprintf(stderr, "let_ok: %s\n", DISPLAY(scc)); */

  let_code = caddr(scc);
  let_body = cddr(let_code);
  body_len = s7_list_length(sc, let_body);
  let_star = (symbol_syntax_op(car(let_code)) == OP_LET_STAR);
  let_vars = cadr(let_code);
  set_safe_stepper(step_slot);
  stepper = slot_value(step_slot);

  old_e = sc->envir;
  sc->envir = new_frame_in_env(sc, sc->envir);

  if (setjmp(sc->opt_exit) == 0)
    {
      start_opts(sc);
      for (var_len = 0, p = let_vars; (is_pair(p)) && (is_pair(cdar(p))); var_len++, p = cdr(p))
	{
	  s7_pointer expr;
	  expr = cdar(p);
	  if (!float_optimize(sc, expr))   /* each of these needs to set the associated variable */
	    {
	      sc->envir = old_e;
	      return(fall_through);
	    }
	  if (let_star)
	    make_slot_1(sc, sc->envir, caar(p), s7_make_mutable_real(sc, 1.5));
	}
    }
  else return(fall_through);

  if (!let_star)
    for (p = let_vars; is_pair(p); p = cdr(p))
      make_slot_1(sc, sc->envir, caar(p), s7_make_mutable_real(sc, 1.5));

  for (p = let_body; is_pair(p); p = cdr(p))
    {
      if (!float_optimize(sc, p))
	{
	  sc->envir = old_e;
	  return(fall_through);
	}
    }

  if (is_null(p))
    {
      s7_int k, end;

      end = denominator(stepper);
      let_set_slots(sc->envir, reverse_slots(sc, let_slots(sc->envir)));

      if ((var_len == 1) && (body_len == 1))
	{
	  s7_pointer ip, xp;
	  int32_t pc2;
	  opt_info *first, *second;
	  s7_double (*f1)(void *p);
	  s7_double (*f2)(void *p);
	  xp = slot_value(let_slots(sc->envir));
	  ip = slot_value(step_slot);
	  first = sc->opts[0];
	  f1 = first->v7.fd;
	  integer(ip) = numerator(stepper);
	  sc->pc = 0;
	  set_real(xp, f1(first));
	  pc2 = ++sc->pc;
	  second = sc->opts[pc2];
	  f2 = second->v7.fd;
	  f2(second);
	  if (f2 == opt_fmv)
	    {
	      for (k = numerator(stepper) + 1; k < end; k++)
		{
		  integer(ip) = k;
		  sc->pc = 0;
		  set_real(xp, f1(first));
		  sc->pc = pc2;
		  opt_fmv(second);
		}
	    }
	  else
	    {
	      for (k = numerator(stepper) + 1; k < end; k++)
		{
		  integer(ip) = k;
		  sc->pc = 0;
		  set_real(xp, f1(first));
		  sc->pc = pc2;
		  f2(second);
		}
	    }
	}
      else
	{
	  for (k = numerator(stepper); k < end; k++)
	    {
	      int32_t i;
	      integer(slot_value(step_slot)) = k;
	      sc->pc = 0;
	      for (p = let_slots(sc->envir); is_slot(p); p = next_slot(p))
		{
		  set_real(slot_value(p), sc->opts[sc->pc]->v7.fd(sc->opts[sc->pc]));
		  sc->pc++;
		}
	      for (i = 0; i < body_len; i++)
		{
		  sc->opts[sc->pc]->v7.fd(sc->opts[sc->pc]);
		  sc->pc++;
		}
	      /* fprintf(stderr, "%s\n", DISPLAY(sc->envir)); */
	    }
	}
      sc->envir = old_e;
      sc->value = sc->T;
      sc->code = cdadr(scc);
      return(goto_SAFE_DO_END_CLAUSES);
    }
  sc->envir = old_e;
  return(fall_through);
}

static bool dotimes(s7_scheme *sc, s7_pointer code, bool safe_case)
{
  s7_pointer body;
  body = caddr(code);
  /* here we assume one expr in body */

  if (((typesflag(body) == SYNTACTIC_PAIR) ||
       (typesflag(car(body)) == SYNTACTIC_TYPE)) &&
      ((symbol_syntax_op(car(body)) == OP_LET) ||
       (symbol_syntax_op(car(body)) == OP_LET_STAR)))
    return(do_let(sc, sc->args, code, safe_case) == goto_SAFE_DO_END_CLAUSES);

  return(opt_dotimes(sc, cddr(code), code, safe_case));
}


static int32_t safe_dotimes_ex(s7_scheme *sc)
{
  s7_pointer init_val;

  /* fprintf(stderr, "%s: %s\n", __func__, DISPLAY_80(sc->code)); */

  init_val = cadr(caar(sc->code));
  if (is_symbol(init_val))
    init_val = find_symbol_checked(sc, init_val);
  else
    {
      if (is_pair(init_val))
	init_val = c_call(init_val)(sc, cdr(init_val));
    }
  if (s7_is_integer(init_val))
    {
      s7_pointer end_expr, end_val, code;
      
      code = sc->code;
      end_expr = caadr(code);
      end_val = caddr(end_expr);
      if (is_symbol(end_val))
	end_val = find_symbol_checked(sc, end_val);
      
      if (s7_is_integer(end_val))
	{
	  sc->code = cddr(code);
	  sc->envir = new_frame_in_env(sc, sc->envir);
	  sc->args = make_slot_1(sc, sc->envir, caaar(code), make_mutable_integer(sc, s7_integer(init_val)));
	  
	  denominator(slot_value(sc->args)) = s7_integer(end_val);
	  /* fprintf(stderr, "set %s end: %s\n", DISPLAY(sc->args), DISPLAY(sc->code)); */
	  set_step_end(sc->args);  /* safe_dotimes step is by 1 */

	  /* (define (hi) (do ((i 1 (+ 1 i))) ((= i 1) i))) -- we need the frame even if the loop is not evaluated */

	  /* safe_dotimes: if null (or constant) body, set step=end and quit (here end is (= step lim) so nothing can happen)
	   *   TODO: break out this case in check_do
	   *   if 1-expr body look for syntactic case, if let(*) goto do_let, else opt_dotimes
	   *       if they are unhappy, got safe_dotimes_step_p
	   *   TODO: another case for check_do
	   *   else goto opt_dotimes then safe_dotimes_step_o
	   *   if multi-line body, check opt_dotimes, then safe_dotimes_step
	   */

	  if ((is_null(sc->code)) ||
	      ((!is_pair(car(sc->code))) &&
	       (is_null(cdr(sc->code)))))
	    {
	      numerator(slot_value(sc->args)) = s7_integer(end_val);
	      sc->value = sc->T;
	      sc->code = cdadr(code);
	      return(goto_SAFE_DO_END_CLAUSES);
	    }
	  
	  if (s7_integer(init_val) == s7_integer(end_val))
	    {
	      sc->value = sc->T;
	      sc->code = cdadr(code);
	      return(goto_SAFE_DO_END_CLAUSES);
	    }
	  
	  if ((is_null(cdr(sc->code))) && 
	      (is_pair(car(sc->code))))
	    {
	      sc->code = car(sc->code);
	      set_opt_pair2(code, sc->code); /* is_pair above */
	      
	      if ((typesflag(sc->code) == SYNTACTIC_PAIR) ||
		  (typesflag(car(sc->code)) == SYNTACTIC_TYPE))
		{
		  if (!is_unsafe_do(code))
		    {
		      if (dotimes(sc, code, true))
			return(goto_SAFE_DO_END_CLAUSES);
		      set_unsafe_do(code);
		    }
		  push_stack(sc, OP_SAFE_DOTIMES_STEP_P, sc->args, code);
		  if (typesflag(sc->code) == SYNTACTIC_PAIR)
		    sc->cur_op = (opcode_t)pair_syntax_op(sc->code);
		  else
		    {
		      sc->cur_op = (opcode_t)symbol_syntax_op(car(sc->code));
		      pair_set_syntax_op(sc->code, sc->cur_op);
		      set_syntactic_pair(sc->code);
		    }
		  sc->code = cdr(sc->code);
		  return(goto_START_WITHOUT_POP_STACK);
		}
	      else /* car not syntactic? */
		{
		  if ((!is_unsafe_do(code)) &&
		      (opt_dotimes(sc, cddr(code), code, true)))
		    return(goto_SAFE_DO_END_CLAUSES);
		  set_unsafe_do(code);

		  /* matters in tcopy 50 */
		  push_stack(sc, OP_SAFE_DOTIMES_STEP_O, sc->args, code);
		  return(goto_OPT_EVAL);
		}
	      /* impossible? but make sure in any case we're set up for begin */
	      sc->code = cddr(code);			
	    }
	  
	  /* multi-line body */
	  if ((!is_unsafe_do(code)) &&
	      (opt_dotimes(sc, sc->code, code, true)))
	    return(goto_SAFE_DO_END_CLAUSES);
	  set_unsafe_do(code);

	  set_opt_pair2(code, sc->code);
	  push_stack(sc, OP_SAFE_DOTIMES_STEP, sc->args, code);
	  return(goto_BEGIN1);
	}
    }
  return(fall_through);
}

static int32_t safe_do_ex(s7_scheme *sc)
{
  /* body is safe, step = +1, end is =, but stepper and end might be set (or at least indirectly exported) in the body:
   *    (let ((lst ())) (do ((i 0 (+ i 1))) ((= i 10)) (let ((j (min i 100))) (set! lst (cons j lst)))) lst)
   *  however, we're very restrictive about this in check_do and do_is_safe; even this is considered trouble:
   *    (let ((x 0)) (do ((i i (+ i 1))) ((= i 7)) (set! x (+ x i))) x)
   * but end might not be an integer -- need to catch this earlier.
   */
  s7_pointer end, init_val, end_val, code;

  /* inits, if not >= opt_dotimes else safe_do_step
   */

  /* fprintf(stderr, "%s: %s\n", __func__, DISPLAY_80(sc->code)); */
  code = sc->code;

  init_val = cadaar(code);
  if (is_symbol(init_val))
    init_val = find_symbol_checked(sc, init_val);
  else
    {
      if (is_pair(init_val))
	init_val = c_call(init_val)(sc, cdr(init_val));
    }
  
  end = caddr(caadr(code));
  if (is_symbol(end))
    end_val = find_symbol_checked(sc, end);
  else end_val = end;

  if ((!s7_is_integer(init_val)) || (!s7_is_integer(end_val)))
    {
      pair_set_syntax_symbol(sc->code, sc->do_unchecked_symbol);
      return(goto_DO_UNCHECKED);
    }

  /* (let ((sum 0)) (define (hi) (do ((i 10 (+ i 1))) ((= i 10) i) (set! sum (+ sum i)))) (hi)) */
  sc->envir = new_frame_in_env(sc, sc->envir);
  dox_set_slot1(sc->envir, make_slot_1(sc, sc->envir, caaar(code), init_val)); /* define the step var -- might be needed in the end clauses */

  if ((s7_integer(init_val) == s7_integer(end_val)) ||
      ((s7_integer(init_val) > s7_integer(end_val)) &&
       (opt_cfunc(caadr(code)) == geq_2)))
    {
      sc->value = sc->T;
      sc->code = cdadr(code);
      return(goto_SAFE_DO_END_CLAUSES);
    }

  if (is_symbol(end))
    dox_set_slot2(sc->envir, find_symbol(sc, end));
  else dox_set_slot2(sc->envir, make_slot_1(sc, sc->envir, sc->dox_slot_symbol, end));
  sc->args = dox_slot2(sc->envir);  /* the various safe steps assume sc->args is the end slot */

  /* fprintf(stderr, "%s %d %d\n", DISPLAY(sc->code), is_unsafe_do(sc->code), is_optimized(caadr(code))); */
  if ((!is_unsafe_do(sc->code)) &&
      ((!is_optimized(caadr(code))) ||
       (opt_cfunc(caadr(code)) != geq_2)))
    {
      if (opt_dotimes(sc, cddr(sc->code), sc->code, false))
	return(goto_SAFE_DO_END_CLAUSES);
      set_unsafe_do(sc->code);
    }

  sc->code = cddr(code);
  set_unsafe_do(sc->code);
  set_opt_pair2(code, sc->code);
  push_stack(sc, OP_SAFE_DO_STEP, sc->args, code);
  return(goto_BEGIN1);
}

static int32_t dotimes_p_ex(s7_scheme *sc)
{
  s7_pointer init, end, code, init_val, end_val, slot;
  /* (do ... (set! args ...)) -- one line, syntactic */

  /* fprintf(stderr, "%s: %s\n", __func__, DISPLAY_80(sc->code)); */

  code = sc->code;
  init = cadaar(code);
  if (is_symbol(init))
    init_val = find_symbol_checked(sc, init);
  else
    {
      if (is_pair(init))
	init_val = c_call(init)(sc, cdr(init));
      else init_val = init;
    }
  sc->value = init_val;

  set_opt_pair2(code, caadr(code));
  end = caddr(opt_pair2(code));
  if (is_symbol(end))
    {
      slot = find_symbol(sc, end);
      end_val = slot_value(slot);
    }
  else 
    {
      slot = make_slot(sc, sc->dox_slot_symbol, end);
      end_val = end;
    }

  if ((!s7_is_integer(init_val)) || (!s7_is_integer(end_val)))
    {
      pair_set_syntax_symbol(sc->code, sc->do_unchecked_symbol);
      return(goto_DO_UNCHECKED);
    }

  sc->envir = new_frame_in_env(sc, sc->envir);
  dox_set_slot1(sc->envir, make_slot_1(sc, sc->envir, caaar(code), init_val));
  dox_set_slot2(sc->envir, slot);
  if (!is_symbol(end))
    {
      next_slot(slot) = let_slots(sc->envir);
      let_slots(sc->envir) = slot;
    }
  set_car(sc->t2_1, slot_value(dox_slot1(sc->envir)));
  set_car(sc->t2_2, slot_value(dox_slot2(sc->envir)));
  if (is_true(sc, sc->value = c_call(caadr(code))(sc, sc->t2_1)))
    {
      sc->code = cdadr(code);
      return(goto_DO_END_CLAUSES);
    }

  if ((!is_unsafe_do(code)) &&
      (opt_cfunc(caadr(code)) != geq_2))
    {
      s7_pointer old_args, old_init;

      old_args = sc->args; 
      old_init = slot_value(dox_slot1(sc->envir));
      sc->args = dox_slot1(sc->envir);  /* used in opt_dotimes */
      slot_set_value(sc->args, make_mutable_integer(sc, integer(slot_value(dox_slot1(sc->envir)))));
      denominator(slot_value(sc->args)) = integer(slot_value(dox_slot2(sc->envir)));
      set_step_end(sc->args);  /* dotimes step is by 1 */
      
      if (dotimes(sc, code, false))
	return(goto_DO_END_CLAUSES);

      slot_set_value(sc->args, old_init);
      sc->args = old_args;
      set_unsafe_do(code);
    }

  push_stack(sc, OP_DOTIMES_STEP_P, sc->args, code);
  sc->code = caddr(code);
  return(goto_EVAL);
}
	  
static int32_t do_init_ex(s7_scheme *sc)
{
  s7_pointer x, y, z;
  /* fprintf(stderr, "%s: %s\n", __func__, DISPLAY_80(sc->code)); */
  while (true)
    {
      sc->args = cons(sc, sc->value, sc->args);    /* code will be last element (first after reverse) */
      if (is_pair(sc->code))
	{
	  /* here sc->code is a list like: ((i 0 (+ i 1)) ...) so cadar gets the init value. */
	  s7_pointer init;
	  init = cadar(sc->code);
	  if (is_pair(init))
	    {
	      push_stack(sc, OP_DO_INIT, sc->args, cdr(sc->code));
	      sc->code = init;
	      return(goto_EVAL);
	    }
	  if (is_symbol(init))
	    sc->value = find_symbol_checked(sc, init);
	  else sc->value = init;
	  sc->code = cdr(sc->code);
	}
      else break;
    }
  
  /* all the initial values are now in the args list */
  sc->args = safe_reverse_in_place(sc, sc->args);
  sc->code = car(sc->args);                       /* saved at the start */
  z = sc->args;
  sc->args = cdr(sc->args);                       /* init values */
  
  /* sc->envir = new_frame_in_env(sc, sc->envir); */
  /* sc->args was cons'd above, so it should be safe to reuse it as the new frame */
  sc->envir = reuse_as_let(sc, z, sc->envir);
  
  /* run through sc->code and sc->args adding '( caar(car(code)) . car(args) ) to sc->envir,
   *    also reuse the value cells as the new frame slots.
   */
  sc->value = sc->nil;
  y = sc->args;
  for (x = car(sc->code); is_not_null(y); x = cdr(x))
    {
      s7_pointer sym, args;
      sym = caar(x);
      args = cdr(y);
      reuse_as_slot(y, sym, unchecked_car(y));
      set_next_slot(y, let_slots(sc->envir));
      let_set_slots(sc->envir, y);
      symbol_set_local(sym, let_id(sc->envir), y);
      
      if (is_not_null(cddar(x)))                /* else no incr expr, so ignore it henceforth */
	{
	  s7_pointer p;
	  p = cons(sc, caddar(x), sc->gc_nil);  /* this is where we store the new value */
	  set_opt_slot1(p, y);
	  sc->value = cons_unchecked(sc, p, sc->value);
	}
      y = args;
    }
  sc->args = cons(sc, sc->value = safe_reverse_in_place(sc, sc->value), cadr(sc->code));
  sc->code = cddr(sc->code);
  
  /* here args is a list of 2 or 3 lists, first is (list (list (var . binding) incr-expr init-value) ...), second is end-expr, third can be result expr
   *   so for (do ((i 0 (+ i 1))) ((= i 3) (+ i 1)) ...) args is ((((i . 0) (+ i 1) 0 #f)) (= i 3) (+ i 1))
   */
  return(fall_through);
}
/* -------------------------------------------------------------------------------- */
  
  
static inline bool closure_is_ok_1(s7_scheme *sc, s7_pointer code, uint16_t type, int32_t args)
{
  s7_pointer f;
  f = find_symbol_unexamined(sc, car(code));
  if ((f == opt_lambda_unchecked(code)) ||
      ((f) &&
       (typesflag(f) == type) &&
       ((closure_arity(f) == args) || (closure_arity_to_int(sc, f) == args)) &&
       (set_opt_lambda(code, f))))
    return(true);
  sc->last_function = f;
  return(false);
}

static bool closure_star_is_ok_1(s7_scheme *sc, s7_pointer code, uint16_t type, int32_t args)
{
  s7_pointer val; 
  val = find_symbol_unexamined(sc, car(code));
  if ((val == opt_lambda_unchecked(code)) ||
      ((val) &&
       (typesflag(val) == (uint16_t)type) &&
       ((closure_arity(val) >= args) || 
	(closure_star_arity_to_int(sc, val) >= args)) &&
       (set_opt_lambda(code, val))))
    return(true);
  sc->last_function = val;
  return(false);
}

/* it is almost never the case that we already have the value and can see it in the current environment directly,
 * but once found, the value usually matches the current (opt_lambda(code))
 *
 * (_val_) is needed below because car(code) might be undefined (with-let can cause this confusion),
 *   and find_symbol_unchecked returns NULL in that case.
 */

#define closure_is_ok(Sc, Code, Type, Args)			\
  (((symbol_ctr(car(Code)) == 1) &&				\
    (unchecked_type(local_slot(car(Code))) == T_SLOT) && 	\
    (slot_value(local_slot(car(Code))) == opt_lambda_unchecked(Code))) ||	\
   (closure_is_ok_1(Sc, Code, Type, Args)))

#define closure_is_equal(Sc, Code)				\
  (((symbol_ctr(car(Code)) == 1) &&				\
    (unchecked_type(local_slot(car(Code))) == T_SLOT) &&  	\
    (slot_value(local_slot(car(Code))) == opt_lambda_unchecked(Code))) ||	\
   ((sc->last_function = find_symbol_unexamined(Sc, car(Code))) == opt_lambda_unchecked(Code)))

#define closure_star_is_ok(Sc, Code, Type, Args)		\
  (((symbol_ctr(car(Code)) == 1) &&				\
    (unchecked_type(local_slot(car(Code))) == T_SLOT) && 	\
    (slot_value(local_slot(car(Code))) == opt_lambda_unchecked(Code))) ||	\
   (closure_star_is_ok_1(Sc, Code, Type, Args)))


#define MATCH_UNSAFE_CLOSURE      (T_CLOSURE)
#define MATCH_SAFE_CLOSURE        (T_CLOSURE      | T_SAFE_CLOSURE)
#define MATCH_UNSAFE_CLOSURE_STAR (T_CLOSURE_STAR)
#define MATCH_SAFE_CLOSURE_STAR   (T_CLOSURE_STAR | T_SAFE_CLOSURE)

/* since T_HAS_METHODS is on if there might be methods, this can protect us from that case */


/* unknown ops */

static int32_t fixup_unknown_op(s7_scheme *sc, s7_pointer code, s7_pointer func, int32_t op)
{
  /* sc arg used if debugging */
  set_optimize_op(code, op);
  set_opt_lambda(code, func); /* opt_lambda works here because it is the only checked case, but ideally we'd split out all the cases via switch (op) */
  return(goto_OPT_EVAL);
}

static int32_t unknown_ex(s7_scheme *sc, s7_pointer f)
{
  s7_pointer code;
  int32_t hop;
  
  code = sc->code;
  hop = (is_immutable_symbol(car(code))) ? 1 : 0;
  /* fprintf(stderr, "%s: local: %d, hop: %d\n", DISPLAY(code), is_local_symbol(code), hop); */

  switch (type(f))
    {
    case T_CLOSURE:
      if ((!has_methods(f)) &&
	  (is_null(closure_args(f))))
	{
	  if (is_safe_closure(f))
	    {
	      s7_pointer body;
	      body = closure_body(f);
	      set_optimize_op(code, hop + OP_SAFE_THUNK);
	      if (is_null(cdr(body)))
		{
		  if (is_optimized(car(body)))
		    set_optimize_op(code, hop + OP_SAFE_THUNK_E);
		  else
		    {
		      if ((is_pair(car(body))) &&
			  (is_syntactic_symbol(caar(body))))
			{
			  set_optimize_op(code, hop + ((is_local_symbol(code)) ? OP_SAFE_LTHUNK_P : OP_SAFE_THUNK_P));
			  if (typesflag(car(body)) != SYNTACTIC_PAIR)
			    {
			      pair_set_syntax_op(car(body), symbol_syntax_op(caar(body)));
			      set_syntactic_pair(car(body));
			    }
			}
		    }
		}
	      set_opt_lambda(code, f);
	      return(goto_OPT_EVAL);
	    }
	  return(fixup_unknown_op(sc, code, f, hop + OP_THUNK));
	}
      /* we can't ignore the recheck here (i.e. set the hop bit) because the closure, even if a global can be set later:
       *   (begin (define *x* #f) (define (test) (display (*x*))) (define (setx n) (set! *x* (lambda () n))) (setx 1) (test) (setx 2) (test))
       * this is a case where the name matters (we need a pristine global), so it's easily missed.
       */
      break;
      
    case T_GOTO:     return(fixup_unknown_op(sc, code, f, hop + OP_GOTO));
    case T_ITERATOR: return(fixup_unknown_op(sc, code, f, hop + OP_ITERATE));

    default:
      return(fixup_unknown_op(sc, code, f, OP_S));
    }
  return(fall_through);
}

static int32_t unknown_g_ex(s7_scheme *sc, s7_pointer f)
{
  s7_pointer code;
  bool sym_case;
  int32_t hop;

  code = sc->code;
  hop = (is_immutable_symbol(car(code))) ? 1 : 0;
  /* fprintf(stderr, "unknown_g_ex: %s, local: %d, hop: %d\n", DISPLAY_80(sc->code), is_local_symbol(sc->code), hop); */

  sym_case = is_symbol(cadr(code));
  if ((sym_case) &&
      (!is_slot(find_symbol(sc, cadr(code)))))
    return(fall_through);
  
  switch (type(f))
    {
    case T_C_FUNCTION:
    case T_C_RST_ARGS_FUNCTION:
      if ((c_function_required_args(f) > 1) ||
	  (c_function_all_args(f) == 0))
	break;
      
    case T_C_OPT_ARGS_FUNCTION:
    case T_C_ANY_ARGS_FUNCTION:
      if (sym_case)
	{
	  if (is_safe_procedure(f))
	    set_optimize_op(code, hop + OP_SAFE_C_S);
	  else set_optimize_op(code, hop + OP_C_S);
	  set_c_function(code, f);
	  return(goto_OPT_EVAL);
	}
      else
	{
	  if (is_safe_procedure(f))
	    {
	      set_optimize_op(code, hop + OP_SAFE_C_C);
	      set_c_function(code, f);
	      return(goto_OPT_EVAL);
	    }
	}
      break;
      
    case T_CLOSURE:
      if ((!has_methods(f)) &&
	  (closure_arity_to_int(sc, f) == 1))
	{
	  if (sym_case)
	    {
	      set_opt_sym2(code, cadr(code));
	      if (is_safe_closure(f))
		{
		  s7_pointer body;
		  set_optimize_op(code, hop + OP_SAFE_CLOSURE_S);
		  body = closure_body(f);
		  if ((is_null(cdr(body))) &&
		      (!is_optimized(car(body))) && /* might be h_safe_c_c->if_x2 or whatever */
		      (is_pair(car(body))) &&
		      (is_syntactic_symbol(caar(body))))
		    {
		      set_optimize_op(code, hop + (((is_local_symbol(code)) && (is_local_symbol(cdr(code)))) ? OP_SAFE_LCLOSURE_L_P : OP_SAFE_CLOSURE_S_P));
		      if (typesflag(car(body)) != SYNTACTIC_PAIR)
			{
			  pair_set_syntax_op(car(body), symbol_syntax_op(caar(body)));
			  set_syntactic_pair(car(body));
			}
		    }
		}
	      else set_optimize_op(code, hop + OP_CLOSURE_S);
	    }
	  else 
	    {
	      set_optimize_op(code, hop + ((is_safe_closure(f)) ? OP_SAFE_CLOSURE_C : OP_CLOSURE_C));
	      set_opt_con2(code, cadr(code));
	    }
	  set_opt_lambda(code, f);
	  return(goto_OPT_EVAL);
	}
      break;
      
    case T_CLOSURE_STAR:
      if ((!has_methods(f)) &&
	  (has_simple_arg_defaults(closure_body(f))) &&
	  (closure_star_arity_to_int(sc, f) >= 1))
	{
	  annotate_arg(sc, cdr(code), sc->envir);
	  set_arglist_length(code, small_int(1));
	  return(fixup_unknown_op(sc, code, f, hop + ((is_safe_closure(f)) ? OP_SAFE_CLOSURE_STAR_A : OP_CLOSURE_STAR_A)));
	}
      break;

    case T_GOTO:
      return(fixup_unknown_op(sc, code, f, hop + ((sym_case) ? OP_GOTO_S : OP_GOTO_C)));
      
    case T_INT_VECTOR:
    case T_FLOAT_VECTOR:
    case T_VECTOR:
      if ((sym_case) ||
	  (is_integer(cadr(code))))      /* (v 4/3) */
	{
	  annotate_arg(sc, cdr(code), sc->envir);
	  return(fixup_unknown_op(sc, code, f, hop + OP_VECTOR_A));
	}
      break;
      
    case T_STRING:
      annotate_arg(sc, cdr(code), sc->envir);
      return(fixup_unknown_op(sc, code, f, hop + OP_STRING_A));
      
    case T_PAIR:
      annotate_arg(sc, cdr(code), sc->envir);
      return(fixup_unknown_op(sc, code, f, hop + OP_PAIR_A));
      
    case T_C_OBJECT:
      if (s7_is_aritable(sc, f, 1))
	{
	  annotate_arg(sc, cdr(code), sc->envir);
	  return(fixup_unknown_op(sc, code, f, hop + OP_C_OBJECT_A));
	}
      break;
      
    case T_LET:
      annotate_arg(sc, cdr(code), sc->envir);
      return(fixup_unknown_op(sc, code, f, hop + OP_ENVIRONMENT_A));
      
    case T_HASH_TABLE:
      annotate_arg(sc, cdr(code), sc->envir);
      return(fixup_unknown_op(sc, code, f, hop + OP_HASH_TABLE_A));

    case T_CONTINUATION:
      annotate_arg(sc, cdr(code), sc->envir);
      return(fixup_unknown_op(sc, code, f, hop + OP_CONTINUATION_A));
      
    default:
      break;
    }
  return(fixup_unknown_op(sc, code, f, (sym_case) ? OP_S_S : OP_S_C));
}
		  
static int32_t unknown_gg_ex(s7_scheme *sc, s7_pointer f)
{
  bool s1, s2;
  int32_t hop;
  s7_pointer code;
  
  code = sc->code;
  hop = (is_immutable_symbol(car(code))) ? 1 : 0;
  s1 = is_symbol(cadr(code));
  s2 = is_symbol(caddr(code));
  
  if ((s1) &&
      (!is_slot(find_symbol(sc, cadr(code)))))
    return(fall_through);
  if ((s2) &&
      (!is_slot(find_symbol(sc, caddr(code)))))
    return(fall_through);
  
  switch (type(f))
    {
    case T_C_FUNCTION:
    case T_C_RST_ARGS_FUNCTION:
      if ((c_function_required_args(f) > 2) ||
	  (c_function_all_args(f) < 2))
	break;
      
    case T_C_OPT_ARGS_FUNCTION:
    case T_C_ANY_ARGS_FUNCTION:
      if (is_safe_procedure(f))
	{
	  if (s1) 
	    {
	      if (s2)
		set_optimize_op(code, hop + OP_SAFE_C_SS);
	      else set_optimize_op(code, hop + OP_SAFE_C_SC);
	    }
	  else set_optimize_op(code, hop + ((s2) ? OP_SAFE_C_CS : OP_SAFE_C_C));
	}
      else
	{
	  set_optimize_op(code, hop + OP_C_ALL_X);
	  annotate_args(sc, cdr(code), sc->envir);
	}
      set_arglist_length(code, small_int(2));
      set_c_function(code, f);
      return(goto_OPT_EVAL);
      
    case T_CLOSURE:
      if (has_methods(f)) break;
      if (closure_arity_to_int(sc, f) == 2)
	{
	  if (s1)
	    {
	      if (is_safe_closure(f))
		set_optimize_op(code, hop + ((s2) ? ((is_null(cdr(closure_body(f)))) ? OP_SAFE_CLOSURE_SS : OP_SAFE_CLOSURE_SS_B) : OP_SAFE_CLOSURE_SC));
	      else 
		{
		  if (!s2)
		    set_optimize_op(code, hop + OP_CLOSURE_SC);
		  else
		    {
		      s7_pointer body;
		      body = closure_body(f);
		      if ((is_null(cdr(body))) &&
			  (is_pair(car(body))) &&
			  (is_syntactic(caar(body))))
			{
			  set_optimize_op(code, hop + OP_CLOSURE_SS_P);
			  if (typesflag(car(body)) != SYNTACTIC_PAIR)
			    {
			      pair_set_syntax_op(car(body), symbol_syntax_op(caar(body)));
			      set_syntactic_pair(car(body));
			    }
			}
		      else set_optimize_op(code, hop + OP_CLOSURE_SS);
		    }
		}
	    }
	  else
	    {
	      if (!s2) break;
	      set_optimize_op(code, hop + ((is_safe_closure(f)) ? OP_SAFE_CLOSURE_CS : OP_CLOSURE_CS));
	    }
	  if (s2) set_opt_sym2(code, caddr(code)); else set_opt_con2(code, caddr(code));
	  set_opt_lambda(code, f);
	  return(goto_OPT_EVAL);
	}
      break;

    case T_CLOSURE_STAR:
      if ((!has_methods(f)) &&
	  (has_simple_arg_defaults(closure_body(f))) &&
	  (closure_star_arity_to_int(sc, f) >= 2))
	{
	  annotate_args(sc, cdr(code), sc->envir);
	  set_arglist_length(code, small_int(2));
	  if (closure_star_arity_to_int(sc, f) == 2)
	    return(fixup_unknown_op(sc, code, f, hop + ((is_safe_closure(f)) ? OP_SAFE_CLOSURE_STAR_AA : OP_CLOSURE_STAR_AA)));
	  return(fixup_unknown_op(sc, code, f, hop + ((is_safe_closure(f)) ? OP_SAFE_CLOSURE_STAR_ALL_X : OP_CLOSURE_STAR_ALL_X)));
	}
      break;

    default:
      break;
    }
  return(fall_through);
}

static int32_t unknown_all_s_ex(s7_scheme *sc, s7_pointer f)
{
  s7_pointer code, arg;
  int32_t num_args, hop;

  code = sc->code;
  num_args = integer(arglist_length(code));
  for (arg = cdr(code); is_pair(arg); arg = cdr(arg))
    if (/* (!is_symbol(car(arg))) ||  */            /* can't happen?? */
	(!is_slot(find_symbol(sc, car(arg)))))
      return(fall_through);

  hop = (is_immutable_symbol(car(code))) ? 1 : 0;

  switch (type(f))
    {
    case T_C_FUNCTION:
    case T_C_RST_ARGS_FUNCTION:
      if (((int32_t)c_function_required_args(f) > num_args) ||
	  ((int32_t)c_function_all_args(f) < num_args))
	break;
      
    case T_C_OPT_ARGS_FUNCTION:
    case T_C_ANY_ARGS_FUNCTION:
      if (is_safe_procedure(f))
	{
	  if (num_args == 3)
	    {
	      set_optimize_op(code, hop + OP_SAFE_C_SSS);
	      set_opt_sym1(cdr(code), caddr(code));
	      set_opt_sym2(cdr(code), cadddr(code));
	    }
	  else set_optimize_op(code, hop + OP_SAFE_C_ALL_S);
	}
      else
	{
	  set_optimize_op(code, hop + OP_C_ALL_X);
	  annotate_args(sc, cdr(code), sc->envir);
	}
      set_c_function(code, f);
      return(goto_OPT_EVAL);
      
    case T_CLOSURE:
      if ((!has_methods(f)) &&
	  (closure_arity_to_int(sc, f) == num_args))
	{
	  annotate_args(sc, cdr(code), sc->envir);
	  return(fixup_unknown_op(sc, code, f, hop + ((is_safe_closure(f)) ? OP_SAFE_CLOSURE_ALL_X : 
						      ((is_null(cdr(closure_body(f)))) ? OP_CLOSURE_ALL_S_P : OP_CLOSURE_ALL_S))));
	}
      break;

    case T_CLOSURE_STAR:
      if ((!has_methods(f)) &&
	  (has_simple_arg_defaults(closure_body(f))) &&
	  (closure_star_arity_to_int(sc, f) >= num_args))
	{
	  annotate_args(sc, cdr(code), sc->envir);
	  return(fixup_unknown_op(sc, code, f, hop + ((is_safe_closure(f)) ? OP_SAFE_CLOSURE_STAR_ALL_X : OP_CLOSURE_STAR_ALL_X)));
	}
      break;
      
    default:
      break;
    }
  return(fall_through);
}

static int32_t unknown_a_ex(s7_scheme *sc, s7_pointer f)
{
  s7_pointer code;
  int32_t hop;

  code = sc->code;
  hop = (is_immutable_symbol(car(code))) ? 1 : 0;

#if DEBUGGING
  if (!has_all_x(cdr(code)))
    fprintf(stderr, "unknown_a_ex missing _a support? %s\n", DISPLAY_80(code));
#endif

  switch (type(f))
    {
    case T_C_FUNCTION:
    case T_C_RST_ARGS_FUNCTION:
      if ((c_function_required_args(f) > 1) ||
	  (c_function_all_args(f) == 0))
	break;

    case T_C_OPT_ARGS_FUNCTION:
    case T_C_ANY_ARGS_FUNCTION:
      set_optimize_op(code, hop + ((is_safe_procedure(f)) ? OP_SAFE_C_A : OP_C_A));
      set_c_function(code, f);
      return(goto_OPT_EVAL);
      
    case T_CLOSURE:
      /* fprintf(stderr, "unknown_a_ex: %s: %s %d %s\n", DISPLAY(code), DISPLAY(f), is_safe_closure(f), opt_names[optimize_op(code)]); */
      if ((!has_methods(f)) &&
	  (closure_arity_to_int(sc, f) == 1))
	{
	  if (is_safe_closure(f))
	    set_optimize_op(code, hop + (((is_local_symbol(code)) ? 
					  ((is_null(cdr(closure_body(f)))) ? OP_SAFE_LCLOSURE_A_P : OP_SAFE_LCLOSURE_A) : OP_SAFE_CLOSURE_A)));
	  else 
	    {
	      set_optimize_op(code, hop + OP_CLOSURE_A);
	      if (!is_global(car(code)))
		{
		  s7_pointer body;
		  body = closure_body(f);
		  if ((is_null(cdr(body))) &&
		      (is_pair(car(body))) &&
		      (is_syntactic_symbol(caar(body))))
		    {
		      set_optimize_op(code, hop + OP_CLOSURE_A_P);
		      if (typesflag(car(body)) != SYNTACTIC_PAIR)
			{
			  pair_set_syntax_op(car(body), symbol_syntax_op(caar(body)));
			  set_syntactic_pair(car(body));
			}
		    }
		}
	    }
	  set_opt_lambda(code, f);
	  return(goto_OPT_EVAL);
	}
      break;
      
    case T_CLOSURE_STAR:
      if ((!has_methods(f)) &&
	  (has_simple_arg_defaults(closure_body(f))) &&
	  (closure_star_arity_to_int(sc, f) >= 1))
	return(fixup_unknown_op(sc, code, f, hop + ((is_safe_closure(f)) ? OP_SAFE_CLOSURE_STAR_A : OP_CLOSURE_STAR_A)));
      break;
      
    case T_INT_VECTOR:
    case T_FLOAT_VECTOR:
    case T_VECTOR:
      return(fixup_unknown_op(sc, code, f, hop + OP_VECTOR_A));
      
    case T_STRING:     return(fixup_unknown_op(sc, code, f, hop + OP_STRING_A));
    case T_PAIR:       return(fixup_unknown_op(sc, code, f, hop + OP_PAIR_A));
    case T_C_OBJECT:   return(fixup_unknown_op(sc, code, f, hop + OP_C_OBJECT_A));
    case T_LET:        return(fixup_unknown_op(sc, code, f, hop + (((is_pair(cadr(code))) && (caadr(code) == sc->quote_symbol)) ? OP_ENVIRONMENT_Q : OP_ENVIRONMENT_A)));
    case T_HASH_TABLE: return(fixup_unknown_op(sc, code, f, hop + OP_HASH_TABLE_A));
    case T_GOTO:       return(fixup_unknown_op(sc, code, f, hop + OP_GOTO_A));
      
    default:
      /* macro, continuation */
      break;
    }
  return(fixup_unknown_op(sc, code, f, OP_S_A)); /* closure with methods etc */
  /* return(fall_through); */
}

static int32_t unknown_aa_ex(s7_scheme *sc, s7_pointer f)
{
  s7_pointer code;
  int32_t hop;
  /* fprintf(stderr, "unknown_aa: %d %s %s\n", type(f), DISPLAY(f), DISPLAY(sc->code)); */
  
  code = sc->code;
  hop = (is_immutable_symbol(car(code))) ? 1 : 0;
  set_arglist_length(code, small_int(2));
  annotate_args(sc, cdr(code), sc->envir);
  
  switch (type(f))
    {
    case T_C_FUNCTION:
    case T_C_RST_ARGS_FUNCTION:
      if ((c_function_required_args(f) > 2) ||
	  (c_function_all_args(f) < 2))
	break;
      
    case T_C_OPT_ARGS_FUNCTION:
    case T_C_ANY_ARGS_FUNCTION:
      set_optimize_op(code, hop + ((is_safe_procedure(f)) ? OP_SAFE_C_AA : OP_C_ALL_X));
      set_c_function(code, f);
      return(goto_OPT_EVAL);
      
    case T_CLOSURE:
      if ((!has_methods(f)) &&
	  (closure_arity_to_int(sc, f) == 2))
	{
	  set_optimize_op(code, hop + ((is_safe_closure(f)) ? OP_SAFE_CLOSURE_AA : OP_CLOSURE_AA));
	  set_opt_lambda(code, f);
	  return(goto_OPT_EVAL);
	}
      break;

    case T_CLOSURE_STAR:
      if ((!has_methods(f)) &&
	  (has_simple_arg_defaults(closure_body(f))))
	{
	  set_arglist_length(code, small_int(2));
	  if (closure_star_arity_to_int(sc, f) == 2)
	    return(fixup_unknown_op(sc, code, f, hop + ((is_safe_closure(f)) ? OP_SAFE_CLOSURE_STAR_AA : OP_CLOSURE_STAR_AA)));
	  return(fixup_unknown_op(sc, code, f, hop + ((is_safe_closure(f)) ? OP_SAFE_CLOSURE_STAR_ALL_X : OP_CLOSURE_STAR_ALL_X)));
	}
      break;

    default:
      break;
    }
  return(fall_through);
}

static int32_t unknown_all_x_ex(s7_scheme *sc, s7_pointer f)
{
  s7_pointer code;
  int32_t num_args, hop;

  code = sc->code;
  hop = (is_immutable_symbol(car(code))) ? 1 : 0;
  num_args = integer(arglist_length(code));

  switch (type(f))
    {
    case T_C_FUNCTION:
    case T_C_RST_ARGS_FUNCTION:
      if (((int32_t)c_function_required_args(f) > num_args) ||
	  ((int32_t)c_function_all_args(f) < num_args))
	break;
      
    case T_C_OPT_ARGS_FUNCTION:
    case T_C_ANY_ARGS_FUNCTION:
      if (is_safe_procedure(f))
	set_optimize_op(code, hop + ((num_args == 3) ? OP_SAFE_C_AAA : OP_SAFE_C_ALL_X));
      else set_optimize_op(code, hop + OP_C_ALL_X);
      annotate_args(sc, cdr(code), sc->envir);
      set_c_function(code, f);
      return(goto_OPT_EVAL);
	  
    case T_CLOSURE:
      if ((!has_methods(f)) &&
	  (closure_arity_to_int(sc, f) == num_args))
	{
	  annotate_args(sc, cdr(code), sc->envir);
	  if (is_safe_closure(f))
	    {
	      if ((is_symbol(cadr(code))) &&
		  (num_args == 3))
		set_optimize_op(code, hop + OP_SAFE_CLOSURE_SAA);
	      else set_optimize_op(code, hop + OP_SAFE_CLOSURE_ALL_X);
	    }
	  else set_optimize_op(code, hop + OP_CLOSURE_ALL_X);
	  set_opt_lambda(code, f);
	  return(goto_OPT_EVAL);
	}
      break;

    case T_CLOSURE_STAR:
      if ((!has_methods(f)) &&
	  (has_simple_arg_defaults(closure_body(f))) &&
	  (closure_star_arity_to_int(sc, f) >= num_args))
	{
	  set_arglist_length(code, small_int(num_args));
	  annotate_args(sc, cdr(code), sc->envir);
	  return(fixup_unknown_op(sc, code, f, hop + ((is_safe_closure(f)) ? OP_SAFE_CLOSURE_STAR_ALL_X : OP_CLOSURE_STAR_ALL_X)));
	}
      break;

    default:
      break;
    }
  return(fall_through);
}


static void unwind_output_ex(s7_scheme *sc)
{
  bool is_file;
  is_file = is_file_port(sc->code);
  
  if ((is_output_port(sc->code)) &&
      (!port_is_closed(sc->code)))
    s7_close_output_port(sc, sc->code); /* may call fflush */
  
  if ((is_output_port(sc->args)) &&
      (!port_is_closed(sc->args)))
    sc->output_port = sc->args;
  
  if ((is_file) &&
      (is_multiple_value(sc->value)))
    sc->value = splice_in_values(sc, multiple_value(sc->value));
}

static void unwind_input_ex(s7_scheme *sc)
{
  if ((is_input_port(sc->code)) &&
      (!port_is_closed(sc->code)))
    s7_close_input_port(sc, sc->code);
  
  if ((is_input_port(sc->args)) &&
      (!port_is_closed(sc->args)))
    sc->input_port = sc->args;
  
  if (is_multiple_value(sc->value))
    sc->value = splice_in_values(sc, multiple_value(sc->value));
}

static int32_t dynamic_wind_ex(s7_scheme *sc)
{
  if (dynamic_wind_state(sc->code) == DWIND_INIT)
    {
      dynamic_wind_state(sc->code) = DWIND_BODY;
      push_stack(sc, OP_DYNAMIC_WIND, sc->nil, sc->code);
      sc->code = dynamic_wind_body(sc->code);
      sc->args = sc->nil;
      return(goto_APPLY);
    }
  else
    {
      if (dynamic_wind_state(sc->code) == DWIND_BODY)
	{
	  dynamic_wind_state(sc->code) = DWIND_FINISH;
	  if (dynamic_wind_out(sc->code) != sc->F)
	    {
	      push_stack(sc, OP_DYNAMIC_WIND, sc->value, sc->code);
	      sc->code = dynamic_wind_out(sc->code);
	      sc->args = sc->nil;
	      return(goto_APPLY);
	    }
	  else
	    {
	      if (is_multiple_value(sc->value))
		sc->value = splice_in_values(sc, multiple_value(sc->value));
	      return(goto_START);
	    }
	}
      if (is_multiple_value(sc->args))       /* (+ 1 (dynamic-wind (lambda () #f) (lambda () (values 2 3 4)) (lambda () #f)) 5) */
	sc->value = splice_in_values(sc, multiple_value(sc->args));
      else sc->value = sc->args;             /* value saved above */
    }
  return(goto_START);
}

static int32_t read_s_ex(s7_scheme *sc)
{
  /* another lint opt */
  s7_pointer port, code;

  code = sc->code;
  port = find_symbol_unchecked(sc, cadr(code));
  
  if (!is_input_port(port)) /* was also not stdin */
    {
      sc->value = g_read(sc, list_1(sc, port));
      return(goto_START);
    }
  /* I guess this port_is_closed check is needed because we're going down a level below */
  if (port_is_closed(port))
    simple_wrong_type_argument_with_type(sc, sc->read_symbol, port, an_open_port_string);
  
  if (is_function_port(port))
    sc->value = (*(port_input_function(port)))(sc, S7_READ, port);
  else
    {
      if ((is_string_port(port)) &&
	  (port_data_size(port) <= port_position(port)))
	sc->value = sc->eof_object;
      else
	{
	  push_input_port(sc, port);
	  push_stack_op(sc, OP_READ_DONE); /* this stops the internal read process so we only get one form */
	  sc->tok = token(sc);
	  switch (sc->tok)
	    {
	    case TOKEN_EOF:	      return(goto_START);
	    case TOKEN_RIGHT_PAREN:   read_error(sc, "unexpected close paren");
	    case TOKEN_COMMA:	      read_error(sc, "unexpected comma");
	    default:
	      sc->value = read_expression(sc);
	      sc->current_line = port_line_number(sc->input_port);  /* this info is used to track down missing close parens */
	      sc->current_file = port_filename(sc->input_port);
	    }
	}
    }
  /* equally read-done and read-list here */
  return(goto_START);
}
		  
static void eval_string_1_ex(s7_scheme *sc)
{
  if ((sc->tok != TOKEN_EOF) &&
      (port_position(sc->input_port) < port_data_size(sc->input_port))) /* ran past end somehow? */
    {
      unsigned char c;
      while (white_space[c = port_data(sc->input_port)[port_position(sc->input_port)++]])
	if (c == '\n')
	  port_line_number(sc->input_port)++;

      if (c != 0)
	{
	  backchar(c, sc->input_port);
	  push_stack_no_args(sc, OP_EVAL_STRING_1, sc->value);
	  push_stack_op_let(sc, OP_READ_INTERNAL);
	}
      else push_stack_op_let(sc, OP_EVAL_STRING_2);
    }
  else push_stack_op_let(sc, OP_EVAL_STRING_2);
  sc->code = sc->value;
}

static int32_t string_a_ex(s7_scheme *sc)
{
  s7_int index;
  s7_pointer s, x, code;
  code = sc->code;

  s = find_symbol_unchecked(sc, car(code));
  x = c_call(cdr(code))(sc, cadr(code));
  if (!is_string(s))
    {
      sc->last_function = s;
      return(fall_through);
    }
  if (!s7_is_integer(x))
    {
      sc->value = string_ref_1(sc, s, cons(sc, x, sc->nil));
      return(goto_START);
    }
  index = s7_integer(x);
  if ((index < string_length(s)) &&
      (index >= 0))
    {
      if (is_byte_vector(s))
	sc->value = small_int((unsigned char)string_value(s)[index]);
      else sc->value = s7_make_character(sc, ((unsigned char *)string_value(s))[index]);
      return(goto_START);
    }
  sc->value = string_ref_1(sc, s, x);
  return(goto_START);
}

static int32_t vector_a_ex(s7_scheme *sc)
{
  s7_pointer v, x, code;

  code = sc->code;
  v = find_symbol_unchecked(sc, car(code));
  if (!s7_is_vector(v))
    {
      sc->last_function = v;
      return(fall_through);
    }
  x = c_call(cdr(code))(sc, cadr(code));
  if (s7_is_integer(x))
    {
      if (vector_rank(v) == 1)
	{
	  s7_int index;
	  index = s7_integer(x);
	  if ((index < vector_length(v)) &&
	      (index >= 0))
	    {
	      if (is_float_vector(v))
		sc->value = make_real(sc, float_vector_element(v, index));
	      else sc->value = vector_getter(v)(sc, v, index);
	      return(goto_START);
	    }
	}
    }
  sc->value = vector_ref_1(sc, v, cons(sc, x, sc->nil));
  return(goto_START);
}

static int32_t constant_vector_a_ex(s7_scheme *sc)
{
  s7_pointer v, x, code;

  code = sc->code;
  v = local_symbol_value(car(code));
  x = c_call(cdr(code))(sc, cadr(code));
  if (s7_is_integer(x))
    {
      s7_int index;
      index = s7_integer(x);
      if ((index < vector_length(v)) &&
	  (index >= 0))
	{
	  sc->value = vector_element(v, index);
	  return(goto_START);
	}
    }
  sc->value = vector_ref_1(sc, v, set_plist_1(sc, x));
  return(goto_START);
}

static void increment_1_ex(s7_scheme *sc)
{
  /* ([set!] ctr (+ ctr 1)) */
  s7_pointer val, y;
  
  y = find_symbol(sc, car(sc->code));
  if (!is_slot(y))
    eval_error_no_return(sc, sc->wrong_type_arg_symbol, "set! ~A: unbound variable", car(sc->code));
  
  val = slot_value(y);
  switch (type(val))
    {
    case T_INTEGER:
      sc->value = make_integer(sc, integer(val) + 1);  /* this can't be optimized to treat y's value as a mutable integer */
      break;
      
    case T_RATIO:
      new_cell(sc, sc->value, T_RATIO);
      numerator(sc->value) = numerator(val) + denominator(val);
      denominator(sc->value) = denominator(val);
      break;
      
    case T_REAL:
      sc->value = make_real(sc, real(val) + 1.0);
      break;
      
    case T_COMPLEX:
      new_cell(sc, sc->value, T_COMPLEX);
      set_real_part(sc->value, real_part(val) + 1.0);
      set_imag_part(sc->value, imag_part(val));
      break;
      
    default:
      sc->value = g_add(sc, set_plist_2(sc, val, small_int(1)));
      break;
    }
  slot_set_value(y, sc->value);
}

static void decrement_1_ex(s7_scheme *sc)
{
  /* ([set!] ctr (- ctr 1)) */
  s7_pointer val, y;
  y = find_symbol(sc, car(sc->code));
  if (!is_slot(y))
    eval_error_no_return(sc, sc->wrong_type_arg_symbol, "set! ~A: unbound variable", car(sc->code));
  val = slot_value(y);
  switch (type(val))
    {
    case T_INTEGER:
      sc->value = make_integer(sc, integer(val) - 1);
      break;
      
    case T_RATIO:
      new_cell(sc, sc->value, T_RATIO);
      numerator(sc->value) = numerator(val) - denominator(val);
      denominator(sc->value) = denominator(val);
      break;
      
    case T_REAL:
      sc->value = make_real(sc, real(val) - 1.0);
      break;
      
    case T_COMPLEX:
      new_cell(sc, sc->value, T_COMPLEX);
      set_real_part(sc->value, real_part(val) - 1.0);
      set_imag_part(sc->value, imag_part(val));
      break;
      
    default:
      sc->value = g_subtract(sc, set_plist_2(sc, val, small_int(1)));
      break;
    }
  slot_set_value(y, sc->value);
}

static void set_pws_ex(s7_scheme *sc)
{
  /* ([set!] (save-dir) "/home/bil/zap/snd") */
  s7_pointer obj;
  
  obj = caar(sc->code);
  if (is_symbol(obj))
    {
      obj = find_symbol(sc, obj);
      if (is_slot(obj))
	obj = slot_value(obj);
      else eval_error_no_return(sc, sc->syntax_error_symbol, "no generalized set for ~A", caar(sc->code));
    }
  
  if ((is_c_function(obj)) &&
      (is_procedure(c_function_setter(obj))))
    {
      s7_pointer value;
      value = cadr(sc->code);
      if (is_symbol(value))
	value = find_symbol_checked(sc, value);
      
      set_car(sc->t1_1, value);
      sc->value = c_function_call(c_function_setter(obj))(sc, sc->t1_1);
    }
  else eval_error_no_return(sc, sc->syntax_error_symbol, "no generalized set for ~A", obj);
}


/* -------------------------------- apply functions -------------------------------- */

static void apply_c_function(s7_scheme *sc) 	                    /* -------- C-based function -------- */
{
  uint32_t len;
  len = safe_list_length(sc, sc->args);
  if (len < c_function_required_args(sc->code))
    s7_error(sc, sc->wrong_number_of_args_symbol, set_elist_3(sc, sc->not_enough_arguments_string, sc->code, sc->args));
  if (c_function_all_args(sc->code) < len)
    s7_error(sc, sc->wrong_number_of_args_symbol, set_elist_3(sc, sc->too_many_arguments_string, sc->code, sc->args));
  sc->value = c_function_call(sc->code)(sc, sc->args);
}

static void apply_c_opt_args_function(s7_scheme *sc)                /* -------- C-based function that has n optional arguments -------- */
{
  uint32_t len;
  len = safe_list_length(sc, sc->args);
  if (c_function_all_args(sc->code) < len)
    s7_error(sc, sc->wrong_number_of_args_symbol, set_elist_3(sc, sc->too_many_arguments_string, sc->code, sc->args));
  sc->value = c_function_call(sc->code)(sc, sc->args);
}

static void apply_c_rst_args_function(s7_scheme *sc)                /* -------- C-based function that has n required args, then any others -------- */
{
  uint32_t len;
  len = safe_list_length(sc, sc->args);
  if (len < c_function_required_args(sc->code))
    s7_error(sc, sc->wrong_number_of_args_symbol, set_elist_3(sc, sc->not_enough_arguments_string, sc->code, sc->args));
  sc->value = c_function_call(sc->code)(sc, sc->args);
  /* sc->code here need not match sc->code before the function call (map for example) */
}

static void apply_c_any_args_function(s7_scheme *sc)                /* -------- C-based function that can take any number of arguments -------- */
{
  sc->value = c_function_call(sc->code)(sc, sc->args);
}

static void apply_c_macro(s7_scheme *sc)  	                    /* -------- C-based macro -------- */
{
  int32_t len;
  /* fprintf(stderr, "apply_c_macro: args: %s %s\n", DISPLAY(sc->args), type_name(sc, car(sc->args), NO_ARTICLE)); */
  len = safe_list_length(sc, sc->args);
  
  if (len < (int32_t)c_macro_required_args(sc->code))
    s7_error(sc, sc->wrong_number_of_args_symbol, set_elist_3(sc, sc->not_enough_arguments_string, sc->code, sc->args));
  
  if ((int32_t)c_macro_all_args(sc->code) < len)
    s7_error(sc, sc->wrong_number_of_args_symbol, set_elist_3(sc, sc->too_many_arguments_string, sc->code, sc->args));
  
  /* fprintf(stderr, "code: %s, args: %s\n", DISPLAY(sc->code), DISPLAY(sc->args)); */
  sc->code = c_macro_call(sc->code)(sc, sc->args);
  if (is_multiple_value(sc->code)) /* can this happen? s7_values splices before returning, and `(values ...) is handled later */
    {
#if DEBUGGING
      fprintf(stderr, "%d unexpected mv code: %s\n", __LINE__, DISPLAY(sc->code));
#endif
      push_stack(sc, OP_EVAL_MACRO_MV, sc->nil, cdr(sc->code));
      sc->code = car(sc->code);
    }
}

static void apply_syntax(s7_scheme *sc)                            /* -------- syntactic keyword as applicable object -------- */
{                                                                  /* current reader-cond macro uses this via (map quote ...) */
  int32_t len;                                                     /*    ((apply lambda '((x) (+ x 1))) 4) */
  if (is_pair(sc->args))
    {
      len = s7_list_length(sc, sc->args);
      if (len == 0) eval_error_no_return(sc, sc->syntax_error_symbol, "attempt to evaluate a circular list: ~A", sc->args);
    }
  else len = 0;
  
  if (len < syntax_min_args(sc->code))
    s7_error(sc, sc->wrong_number_of_args_symbol, set_elist_3(sc, sc->not_enough_arguments_string, sc->code, sc->args));
  
  if ((syntax_max_args(sc->code) < len) &&
      (syntax_max_args(sc->code) != -1))
    s7_error(sc, sc->wrong_number_of_args_symbol, set_elist_3(sc, sc->too_many_arguments_string, sc->code, sc->args));

  sc->cur_op = (opcode_t)syntax_opcode(sc->code);         /* (apply begin '((define x 3) (+ x 2))) */
  /* I used to have elaborate checks here for embedded circular lists, but now i think that is the caller's problem */
  sc->code = sc->args;
}

static void apply_vector(s7_scheme *sc)                            /* -------- vector as applicable object -------- */
{
  /* sc->code is the vector, sc->args is the list of indices */
  if (is_null(sc->args))                  /* (#2d((1 2) (3 4))) */
    s7_wrong_number_of_args_error(sc, "not enough args for vector-ref: ~A", sc->args);
  
  if ((is_null(cdr(sc->args))) &&
      (s7_is_integer(car(sc->args))) &&
      (vector_rank(sc->code) == 1))
    {
      s7_int index;
      index = s7_integer(car(sc->args));
      if ((index >= 0) &&
	  (index < vector_length(sc->code)))
	sc->value = vector_getter(sc->code)(sc, sc->code, index);
      else out_of_range(sc, sc->vector_ref_symbol, small_int(2), car(sc->args), (index < 0) ? its_negative_string : its_too_large_string);
    }
  else sc->value = vector_ref_1(sc, sc->code, sc->args);
}

static void apply_string(s7_scheme *sc)                            /* -------- string as applicable object -------- */
{
  if ((is_pair(sc->args)) &&
      (is_null(cdr(sc->args))))
    {
      if (s7_is_integer(car(sc->args)))
	{
	  s7_int index;                  /* not int: ("abs" most-negative-fixnum) */
	  index = s7_integer(car(sc->args));
	  if ((index >= 0) &&
	      (index < string_length(sc->code)))
	    {
	      if (is_byte_vector(sc->code))
		sc->value = small_int((unsigned char)(string_value(sc->code))[index]);
	      else sc->value = s7_make_character(sc, ((unsigned char *)string_value(sc->code))[index]);
	      return;
	    }
	}
      sc->value = string_ref_1(sc, sc->code, car(sc->args));
      return;
    }
  s7_error(sc, sc->wrong_number_of_args_symbol, 
	   set_elist_3(sc, (is_null(sc->args)) ? sc->not_enough_arguments_string : sc->too_many_arguments_string, sc->code, sc->args));
}

static int32_t apply_pair(s7_scheme *sc)                              /* -------- list as applicable object -------- */
{
  if (is_multiple_value(sc->code))                                  /* ((values 1 2 3) 0) */
    {
      /* car of values can be anything, so conjure up a new expression, and apply again */
      sc->x = multiple_value(sc->code);                             /* ((values + 1 2) 3) */
      sc->code = car(sc->x);
#if 0
      /* this can't happen? */
      if (!s7_is_proper_list(sc, cdr(sc->x)))
	s7_error(sc, sc->syntax_error_symbol, set_elist_2(sc, s7_make_string_wrapper(sc, "values arglist is a dotted list: ~A"), sc->x));
#endif
      sc->args = s7_append(sc, cdr(sc->x), sc->args);
      sc->x = sc->nil;
      return(goto_APPLY);
    }
  if (is_null(sc->args))
    s7_wrong_number_of_args_error(sc, "not enough args for list-ref (via list as applicable object): ~A", sc->args);
  sc->value = list_ref_1(sc, sc->code, car(sc->args));            /* (L 1) */
  if (!is_null(cdr(sc->args)))
    sc->value = implicit_index(sc, sc->value, cdr(sc->args));     /* (L 1 2) */
  return(goto_START);
}

static void apply_hash_table(s7_scheme *sc)                        /* -------- hash-table as applicable object -------- */
{
  if (is_null(sc->args))
    s7_wrong_number_of_args_error(sc, "not enough args for hash-table-ref (via hash table as applicable object): ~A", sc->args);
  sc->value = s7_hash_table_ref(sc, sc->code, car(sc->args));
  if (!is_null(cdr(sc->args)))
    sc->value = implicit_index(sc, sc->value, cdr(sc->args));
}

static void apply_let(s7_scheme *sc)                               /* -------- environment as applicable object -------- */
{
  if (is_null(sc->args))
    sc->value = s7_let_ref(sc, sc->code, sc->F);         /* why #f and not ()? both are ok in s7test */
  else 
    {
      sc->value = s7_let_ref(sc, sc->code, car(sc->args));
      if (is_pair(cdr(sc->args)))
	sc->value = implicit_index(sc, sc->value, cdr(sc->args));
    }
  /*    (let ((v #(1 2 3))) (let ((e (curlet))) ((e 'v) 1))) -> 2
   * so (let ((v #(1 2 3))) (let ((e (curlet))) (e 'v 1))) -> 2
   */
}

static void apply_iterator(s7_scheme *sc)                          /* -------- iterator as applicable object -------- */
{
  if (!is_null(sc->args))
    s7_wrong_number_of_args_error(sc, "too many args for iterator: ~A", sc->args);
  sc->value = s7_iterate(sc, sc->code);
}

static void apply_lambda(s7_scheme *sc)                            /* -------- normal function (lambda), or macro -------- */
{             /* load up the current args into the ((args) (lambda)) layout [via the current environment] */
  s7_pointer x, z, e;
  uint64_t id;
  e = sc->envir;
  id = let_id(e);
  for (x = closure_args(sc->code), z = _TLst(sc->args); is_pair(x); x = cdr(x)) /* closure_args can be a symbol, for example */
    {
      s7_pointer sym, args;
      /* reuse the value cells as the new frame slots */
      
      if (is_null(z))
	{
	  s7_pointer name, ccode;
	  name = closure_name(sc, sc->code);
	  ccode = current_code(sc);
	  s7_error(sc, sc->wrong_number_of_args_symbol, set_elist_3(sc, sc->not_enough_arguments_string, (name == ccode) ? sc->code : name, ccode));
	}
      /* now that args are being reused as slots, the error message can't use sc->args,
       *  so fallback on current_code(sc) in this section.
       *  But that can be #f, and closure_name can be confusing in this context, so we need a better error message!
       */

      sym = car(x);
      args = cdr(z);
      reuse_as_slot(z, sym, unchecked_car(z));
      symbol_set_local(sym, id, z);
      set_next_slot(z, let_slots(e));
      let_set_slots(e, z);
      z = args;
    }

  if (is_null(x))
    {
      if (is_not_null(z))
	{
	  s7_pointer name, ccode;
	  name = closure_name(sc, sc->code);
	  ccode = current_code(sc);
	  s7_error(sc, sc->wrong_number_of_args_symbol, set_elist_3(sc, sc->too_many_arguments_string, (name == ccode) ? sc->code : name, ccode));
	}
    }
  else
    {
      sc->temp6 = z; /* the rest arg */
      make_slot_1(sc, sc->envir, x, z);
      sc->temp6 = sc->nil;
    }
  sc->code = closure_body(sc->code);
}


/* lambda* */
static s7_pointer lambda_star_argument_set_value(s7_scheme *sc, s7_pointer sym, s7_pointer val)
{
  s7_pointer x;

  for (x = let_slots(sc->envir) /* presumably the arglist */; is_slot(x); x = next_slot(x))
    if (slot_symbol(x) == sym)
      {
	/* x is our binding (symbol . value) */
	if (is_not_checked_slot(x))
	  set_checked_slot(x); /* this is a special use of this bit, I think */
	else return(s7_error(sc, sc->wrong_type_arg_symbol, set_elist_4(sc, parameter_set_twice_string, closure_name(sc, sc->code), sym, sc->args)));
	slot_set_value(x, val);
	return(val);
      }
  return(sc->no_value);
}


static s7_pointer lambda_star_set_args(s7_scheme *sc)
{
  /* sc->code is a closure: (args body envir)
   * (define* (hi a (b 1)) (+ a b))
   * (procedure-source hi) -> (lambda* (a (b 1)) (+ a b))
   *
   * so rather than spinning through the args binding names to values in the
   *   procedure's new environment (as in the usual closure case above),
   *   we scan the current args, and match against the
   *   template in the car of the closure, binding as we go.
   *
   * for each actual arg, if it's not a keyword that matches a member of the
   *   template, bind it to its current (place-wise) arg, else bind it to
   *   that arg.  If it's :rest bind the next arg to the trailing args at this point.
   *   All args can be accessed by their name as a keyword.
   *
   * all args are optional, any arg with no default value defaults to #f.
   *   but the rest arg should default to ().
   * I later decided to add two warnings: if a parameter is set twice and if
   *   an unknown keyword is seen in a keyword position and there is no rest arg.
   */

  bool allow_other_keys;
  s7_pointer lx, cx, zx, code, args;

  /* get the current args, re-setting args that have explicit values */
  code = sc->code;
  args = sc->args;
  cx = closure_args(code);
  allow_other_keys = ((is_pair(cx)) && (allows_other_keys(cx)));
  lx = sc->args;

  zx = sc->nil;
  while ((is_pair(cx)) &&
	 (is_pair(lx)))
    {
      if (car(cx) == sc->key_rest_symbol)           /* the rest arg */
	{
	  /* next arg is bound to trailing args from this point as a list */
	  zx = sc->key_rest_symbol;
	  cx = cdr(cx);
	  lambda_star_argument_set_value(sc, car(cx), lx); /* default arg not allowed here (see check_lambda_star_args) */
	  lx = cdr(lx);
	  cx = cdr(cx);
	}
      else
	{
	  /* mock-symbols introduce an ambiguity here; if the mock symbol's value is a keyword, is that
	   *   intended to be used as an argument name or value? 
	   * this applies to any evaluated arg that returns a keyword.  
	   * 22-May-17: decided to use the value (i.e. treat a keyword as a keyword)
	   */
	  s7_pointer car_lx;
	  car_lx = car(lx);
	  if (has_methods(car_lx))
	    car_lx = check_values(sc, car_lx, lx);
	  if (is_keyword(car_lx))
	    {
	      s7_pointer sym;
	      if (!is_pair(cdr(lx)))
		return(s7_error(sc, sc->wrong_type_arg_symbol,
				set_elist_4(sc, s7_make_string_wrapper(sc, "~A: keyword argument's value is missing: ~S in ~S"),
					    closure_name(sc, code), lx, args)));
	      sym = keyword_symbol(car_lx);
	      
	      if (lambda_star_argument_set_value(sc, sym, cadr(lx)) == sc->no_value)
		{
		  /* if default value is a key, go ahead and use this value.
		   *    (define* (f (a :b)) a) (f :c)
		   * this has become much trickier than I anticipated...
		   */
		  if (allow_other_keys)
		    {
		      /* in CL: (defun hi (&key (a 1) &allow-other-keys) a) (hi :b :a :a 3) -> 3
		       * in s7: (define* (hi (a 1) :allow-other-keys) a)    (hi :b :a :a 3) -> 3
		       */
		      lx = cddr(lx);
		      continue;
		    }
		  else
		    {
		      return(s7_error(sc, sc->wrong_type_arg_symbol,
				      set_elist_4(sc, s7_make_string_wrapper(sc, "~A: unknown key: ~S in ~S"),
						  closure_name(sc, code), lx, args)));
		    }
		}
	      lx = cddr(lx);
	    }
	  else                                  /* not a key/value pair */
	    {
	      /* this is always a positional (i.e. direct) change, but the closure_args are in the
	       *   definition order whereas currently the environment slots are in reverse order.
	       */
	      if (is_pair(car(cx)))
		lambda_star_argument_set_value(sc, caar(cx), car(lx));
	      else lambda_star_argument_set_value(sc, car(cx), car(lx));
	      
	      lx = cdr(lx);
	    }
	  cx = cdr(cx);
	}
    }

  /* (let () (define* (hi (a 1) :allow-other-keys) a) (hi :a 2 32)) */
  /* (let () (define* (f (a :b)) a) (list (f) (f 1) (f :c) (f :a :c) (f :a 1) (f))) */

  /* check for trailing args with no :rest arg */
  if (is_not_null(lx))
    {
      if ((is_not_null(cx)) ||
	  (zx == sc->key_rest_symbol))
	{
	  if (is_symbol(cx))
	    make_slot_1(sc, sc->envir, cx, lx);
	}
      else
	{
	  if (!allow_other_keys)                       /* ((lambda* (a) a) :a 1 2) */
	    return(s7_error(sc, sc->wrong_number_of_args_symbol, set_elist_3(sc, sc->too_many_arguments_string, closure_name(sc, code), args)));
	  else
	    {
	      /* check trailing args for repeated keys or keys with no values or values with no keys */
	      while (is_pair(lx))
		{
		  if ((!is_keyword(car(lx))) ||     /* ((lambda* (a :allow-other-keys) a) :a 1 :b 2 3) */
		      (!is_pair(cdr(lx))))          /* ((lambda* (a :allow-other-keys) a) :a 1 :b) */
		    return(s7_error(sc, sc->wrong_type_arg_symbol,
				    set_elist_3(sc, s7_make_string_wrapper(sc, "~A: not a key/value pair: ~S"), closure_name(sc, code), lx)));
		  /* errors not caught?
		   *    ((lambda* (a :allow-other-keys) a) :a 1 :a 2)
		   *    ((lambda* (:allow-other-keys ) #f) :b :a :a :b)
		   */
		  lx = cddr(lx);
		}
	    }
	}
    }
  return(sc->nil);
}

static int32_t lambda_star_default(s7_scheme *sc)
{
  while (true)
    {
      s7_pointer z;
      z = sc->args;
      if (is_slot(z))
	{
	  if (slot_value(z) == sc->undefined)
	    {
	      if (is_closure_star(sc->code))
		{
		  s7_pointer val;
		  val = slot_expression(z);
		  if (is_symbol(val))
		    {
		      slot_set_value(z, find_symbol_checked(sc, val));
		      if (slot_value(z) == sc->undefined)
			{
			  /* the current environment here contains the function parameters which
			   *   defaulted to #<undefined> earlier in apply_lambda_star,
			   *   so (define (f f) (define* (f (f f)) f) (f)) (f 0) looks for the
			   *   default f, finds itself currently undefined, and raises an error!
			   *   So, before claiming it is unbound, we need to check outlet as well.
			   *   But in the case above, the inner define* shadows the caller's
			   *   parameter before checking the default arg values, so the default f
			   *   refers to the define* -- I'm not sure this is a bug.  It means 
			   *   that (define* (f (a f)) a) returns f: (equal? f (f)) -> #t, so
			   *   any outer f needs an extra let and endless outlets:
			   *   (let ((f 3)) (let () (define* (f (a ((outlet (outlet (outlet (curlet)))) 'f))) a) (f))) -> 3
			   *   We want the shadowing once the define* is done, so the current mess is simplest.
			   */
			  slot_set_value(z, s7_symbol_local_value(sc, val, outlet(sc->envir)));
			  if (slot_value(z) == sc->undefined)
			    eval_error_no_return(sc, sc->syntax_error_symbol, "lambda* defaults: ~A is unbound", slot_symbol(z));
			}
		    }
		  else
		    {
		      if (is_pair(val))
			{
			  if (car(val) == sc->quote_symbol)
			    {
			      if ((!is_pair(cdr(val))) ||      /* (lambda* ((a (quote))) a) or (lambda* ((a (quote 1 1))) a) etc */
				  (is_pair(cddr(val))))
				eval_error_no_return(sc, sc->syntax_error_symbol, "lambda* default: ~A is messed up", val);
			      slot_set_value(z, cadr(val));
			    }
			  else
			    {
			      push_stack(sc, OP_LAMBDA_STAR_DEFAULT, sc->args, sc->code);
			      sc->code = val;
			      return(goto_EVAL);
			    }
			}
		      else slot_set_value(z, val);
		    }
		}
	      else slot_set_value(z, slot_expression(z));
	    }
	  sc->args = slot_pending_value(z);
	}
      else break;
    }
  return(fall_through);
}

static int32_t apply_lambda_star(s7_scheme *sc) 	                  /* -------- define* (lambda*) -------- */
{
  /* to check for and fixup unset args from defaults, we need to traverse the slots in left-to-right order
   *   but they are stored backwards in the environment, so use pending_value as a back-pointer.
   * We have to build the environment before calling lambda_star_set_args because keywords can
   *   cause any arg to be set at any point in the arg list.
   *
   * the frame-making step below could be precalculated, but where to store it?
   */
  s7_pointer z, top, nxt;
  top = NULL;
  nxt = NULL;

  for (z = closure_args(sc->code); is_pair(z); z = cdr(z))
    {
      s7_pointer car_z;
      car_z = car(z);
      if (is_pair(car_z))           /* arg has a default value */
	{
	  s7_pointer val;
	  val = cadr(car_z);
	  if ((!is_pair(val)) &&
	      (!is_symbol(val)))
	    make_slot_1(sc, sc->envir, car(car_z), val);
	  else
	    {
	      s7_pointer y;
	      add_slot(sc->envir, car(car_z), sc->undefined);
	      y = let_slots(sc->envir);
	      slot_set_expression(y, cadr(car_z));
	      slot_set_pending_value(y, sc->nil);
	      if (!top)
		{
		  top = y;
		  nxt = top;
		}
	      else
		{
		  slot_set_pending_value(nxt, y);
		  nxt = y;
		}
	    }
	}
      else
	{
	  if (!is_keyword(car_z))
	    make_slot_1(sc, sc->envir, car_z, sc->F);
	  else
	    {
	      if (car_z == sc->key_rest_symbol)
		{
		  make_slot_1(sc, sc->envir, cadr(z), sc->nil);
		  z = cdr(z);
		}
	    }
	}
    }
  if (is_symbol(z))
    make_slot_1(sc, sc->envir, z, sc->nil);
  lambda_star_set_args(sc);                     /* load up current arg vals */
  
  if (top)
    {
      /* get default values, which may involve evaluation */
      push_stack(sc, OP_LAMBDA_STAR_DEFAULT, sc->args, sc->code); /* op is just a placeholder (don't use OP_BARRIER here) */
      sc->args = top;
      if (lambda_star_default(sc) == goto_EVAL) return(goto_EVAL);
      pop_stack_no_op(sc);              /* get original args and code back */
    }
  sc->code = closure_body(sc->code);
  return(goto_BEGIN1);
}


static inline void apply_continuation(s7_scheme *sc)               /* -------- continuation ("call/cc") -------- */
{
  if (!call_with_current_continuation(sc))
    {
      static s7_pointer cc_err = NULL; 
      if (!cc_err) cc_err = s7_make_permanent_string("continuation can't jump into with-baffle");
      s7_error(sc, sc->baffled_symbol, set_elist_1(sc, cc_err));
    }
}

static void apply_c_object(s7_scheme *sc)                          /* -------- applicable (new-type) object -------- */
{
  sc->value = (*(c_object_ref(sc, sc->code)))(sc, sc->code, sc->args);
}


/* -------------------------------------------------------------------------------- */

static int32_t define1_ex(s7_scheme *sc)
{	  
  /* sc->code is the symbol being defined, sc->value is its value
   *   if sc->value is a closure, car is of the form ((args...) body...)
   *   so the doc string if any is (cadr (car value))
   *   and the arg list gives the number of optional args up to the dot
   */
  
  /* it's not possible to expand and replace macros at this point without evaluating
   *   the body.  Just as examples, say we have a macro "mac",
   *   (define (hi) (call/cc (lambda (mac) (mac 1))))
   *   (define (hi) (quote (mac 1))) or macroexpand etc
   *   (define (hi mac) (mac 1)) assuming mac here is a function passed as an arg,
   * etc...
   */
  
  /* the immutable constant check needs to wait until we have the actual new value because
   *   we want to ignore the rebinding (not raise an error) if it is the existing value.
   *   This happens when we reload a file that calls define-constant.
   */
  /* sc->code is a symbol at this point */
  if (is_immutable_symbol(sc->code))                                  /* (define pi 3) or (define (pi a) a) */
    {
      s7_pointer x;
      
      if (is_slot(global_slot(sc->code)))
	x = global_slot(sc->code); 
      else x = find_symbol(sc, sc->code);  /* local_slot can be free even if sc->code is immutable (local constant now defunct) */
      if (!is_slot(x))
	{
	  if ((is_slot(local_slot(sc->code))) &&
	      (type(sc->value) == unchecked_type(slot_value(local_slot(sc->code)))))
	    x = local_slot(sc->code);
	}
      if (!((is_slot(x)) &&
	    (type(sc->value) == unchecked_type(slot_value(x))) &&
	    (s7_is_morally_equal(sc, sc->value, slot_value(x)))))    /* if value is unchanged, just ignore this (re)definition */
	eval_error_no_return(sc, sc->syntax_error_symbol, "define: ~S is immutable", sc->code);   /*   can't use s7_is_equal because value might be NaN, etc */
    }
  if (symbol_has_accessor(sc->code))
    {
      s7_pointer x;
      x = find_symbol(sc, sc->code);
      if ((is_slot(x)) &&
	  (slot_has_accessor(x)))
	{
	  sc->value = bind_accessed_symbol(sc, OP_DEFINE_WITH_ACCESSOR, sc->code, sc->value);
	  if (sc->value == sc->no_value)
	    return(goto_APPLY);
	  /* if all goes well, OP_DEFINE_WITH_ACCESSOR will jump to DEFINE2 */
	}
    }
  return(fall_through);
}

static void define2_ex(s7_scheme *sc)
{
  if ((is_any_closure(sc->value)) &&
      ((!(is_let(closure_let(sc->value)))) ||
       (!(is_funclet(closure_let(sc->value))))))  /* otherwise it's (define f2 f1) or something similar */
    {
      s7_pointer new_func, new_env;
      new_func = sc->value;
      new_env = make_funclet(sc, new_func, sc->code, closure_let(new_func));
      /* this should happen only if the closure* default values do not refer in any way to
       *   the enclosing environment (else we can accidentally shadow something that happens
       *   to share an argument name that is being used as a default value -- kinda dumb!).
       *   I think I'll check this before setting the safe_closure bit.
       */

      if ((port_filename(sc->input_port)) &&
	  (port_file(sc->input_port) != stdin))
	{
	  /* unbound_variable will be called if __func__ is encountered, and will return this info as if __func__ had some meaning */
	  let_set_file(new_env, port_file_number(sc->input_port));
	  let_set_line(new_env, port_line_number(sc->input_port));
	}
      else
	{
	  let_set_file(new_env, 0);
	  let_set_line(new_env, 0);
	}
      
      /* add the newly defined thing to the current environment */
      if (is_let(sc->envir))
	{
	  add_slot(sc->envir, sc->code, new_func);
	  set_local(sc->code);
	  /* so funchecked is always local already -- perhaps reset below? */
	}
      else s7_make_slot(sc, sc->envir, sc->code, new_func); 
      sc->value = new_func; /* 25-Jul-14 so define returns the value not the name */
    }
  else
    {
      s7_pointer lx;
      /* add the newly defined thing to the current environment */
      lx = find_local_symbol(sc, sc->code, sc->envir);
      if (is_slot(lx))
	slot_set_value_with_hook(lx, sc->value);
      else s7_make_slot(sc, sc->envir, sc->code, sc->value);
    }
}


/* ---------------------------------------- */

#define c_function_is_ok_cadr(Sc, P) ((c_function_is_ok(Sc, P)) && (c_function_is_ok(Sc, cadr(P))))
#define c_function_is_ok_caddr(Sc, P) ((c_function_is_ok(Sc, P)) && (c_function_is_ok(Sc, caddr(P))))
#define c_function_is_ok_cadr_caddr(Sc, P) ((c_function_is_ok(Sc, P)) && (c_function_is_ok(Sc, cadr(P))) && (c_function_is_ok(Sc, caddr(P))))

#if WITH_GCC
  #define indirect_c_function_is_ok(Sc, X) ({s7_pointer _X_; _X_ = X; (((optimize_op(_X_) & 0x1) != 0) || (c_function_is_ok(Sc, _X_)));})
  #define indirect_cq_function_is_ok(Sc, X) ({s7_pointer _X_; _X_ = X; ((!is_optimized(_X_)) || ((optimize_op(_X_) & 0x1) != 0) || (c_function_is_ok(Sc, _X_)));})
#else
  #define indirect_c_function_is_ok(Sc, X) (((optimize_op(X) & 0x1) != 0) || (c_function_is_ok(Sc, X)))
  #define indirect_cq_function_is_ok(Sc, X) ((!is_optimized(X)) || ((optimize_op(X) & 0x1) != 0) || (c_function_is_ok(Sc, X)))
#endif


#if WITH_PROFILE
#define profile_location(p)        p->file_and_line
#define profile_set_location(p, N) p->file_and_line = N

static void profile(s7_scheme *sc, s7_pointer expr)
{
  /* I tried using SIGPROF and a tick counter below (in addition to the line counter), but the added info
   *   did not seem very useful.
   */
  if ((sc->free_heap_top - sc->free_heap) < 32)
    gc(sc);

  if (is_null(sc->profile_info))
    {
      sc->profile_info = s7_make_hash_table(sc, 65536);
      s7_gc_protect(sc, sc->profile_info);
    }
  if ((is_pair(expr)) &&
      (profile_location(expr) > 0))
    {
      s7_pointer val, key;
      key = s7_make_integer(sc, profile_location(expr)); /* file + line */
      val = s7_hash_table_ref(sc, sc->profile_info, key);
      if (val == sc->F)
	{
	  bool old_short_print;
	  s7_pointer env;
	  int32_t tx1, tx2;

	  old_short_print = sc->short_print;
	  sc->short_print = true;

	  env = find_closure_let(sc, sc->envir);

	  tx1 = next_tx(sc);
	  tx2 = next_tx(sc);
	  sc->t_temps[tx1] = g_object_to_string(sc, set_plist_3(sc, expr, sc->T, small_int(120)));
	  sc->t_temps[tx2] = (is_let(env)) ? g_object_to_string(sc, set_plist_1(sc, funclet_function(env))) : sc->nil;

	  s7_hash_table_set(sc, sc->profile_info, key,
			    cons(sc, make_mutable_integer(sc, 1), 
				 cons(sc, sc->t_temps[tx1], sc->t_temps[tx2])));
	  sc->t_temps[tx1] = sc->nil;
	  sc->t_temps[tx2] = sc->nil;
	  sc->short_print = old_short_print;
	}
      /* can't save the actual expr here -- it can be stepped on */
      else integer(car(val))++;
    }
}

static s7_pointer g_profile_line_number(s7_scheme *sc, s7_pointer args)
{
  #define H_profile_line_number "(profile-line-number obj) returns the line number at which the profiler read obj"
  #define Q_profile_line_number s7_make_signature(sc, 2, sc->is_integer_symbol, sc->T)
  return(make_integer(sc, remembered_line_number(integer(car(args)))));
}

static s7_pointer g_profile_filename(s7_scheme *sc, s7_pointer args)
{
  #define H_profile_filename "(profile-filename obj) returns the name of the file containing obj"
  #define Q_profile_filename s7_make_signature(sc, 2, sc->is_string_symbol, sc->T)
  return(remembered_file_name(integer(car(args))));
}
#endif


/* -------------------------------- eval -------------------------------- */

#if WITH_GCC
#undef new_cell
#if (!DEBUGGING)
#define new_cell(Sc, Obj, Type)			\
  do {						\
    if (Sc->free_heap_top <= Sc->free_heap_trigger) {try_to_call_gc(Sc); if ((Sc->begin_hook) && (call_begin_hook(Sc))) return(Sc->F);} \
    Obj = (*(--(Sc->free_heap_top)));					\
    set_type(Obj, Type);						\
    } while (0)
#else
#define new_cell(Sc, Obj, Type)			\
  do {						\
    if ((Sc->free_heap_top <= Sc->free_heap_trigger) || (for_any_other_reason(sc, __LINE__))) {last_gc_line = __LINE__; last_gc_func = __func__; try_to_call_gc(Sc); if ((Sc->begin_hook) && (call_begin_hook(Sc))) return(Sc->F);} \
    Obj = (*(--(Sc->free_heap_top))); \
    Obj->alloc_line = __LINE__;	Obj->alloc_func = __func__; Obj->debugger_bits = 0;	\
    set_type(Obj, Type);	      \
    } while (0)
#endif
#endif

#if WITH_GMP
#define global_add big_add
#else
#define global_add g_add
#endif

static s7_pointer check_for_cyclic_code(s7_scheme *sc, s7_pointer code)
{
  if (cyclic_sequences(sc, code, false) == sc->T)
    eval_error(sc, "attempt to evaluate a circular list: ~A", code);
  resize_stack(sc);
  return(sc->F);
}

#if WITH_PROFILE
  static s7_pointer profile_at_start = NULL;
#endif

static s7_pointer eval(s7_scheme *sc, opcode_t first_op)
{
  sc->cur_op = first_op;

  /* this procedure can be entered recursively (via s7_call for example), so it's no place for a setjmp
   *   I don't think the recursion can hurt our continuations because s7_call is coming from hooks and
   *   callbacks that are implicit in our stack.
   */

  goto START_WITHOUT_POP_STACK;
  /* this ugly two-step is faster than other ways of writing this code */
  while (true) 
    {
    START:
      pop_stack(sc);
      
    START_WITHOUT_POP_STACK:
      /* fprintf(stderr, "%s (%d) %s\n", op_names[sc->cur_op], (int)(sc->cur_op), DISPLAY_80(sc->code)); */

#if WITH_PROFILE
      profile_at_start = sc->code;
      profile(sc, sc->code);
#endif

      switch (sc->cur_op)
	{
	case OP_NO_OP:
	case OP_GC_PROTECT:
	  break;

	case OP_READ_INTERNAL:
	  /* if we're loading a file, and in the file we evaluate something like:
	   *    (let ()
	   *      (set-current-input-port (open-input-file "tmp2.r5rs"))
	   *      (close-input-port (current-input-port)))
	   *    ... (with no reset of input port to its original value)
	   * the load process tries to read the loaded string, but the sc->input_port is now closed,
	   * and the original is inaccessible!  So we get a segfault in token.  We don't want to put
	   * a port_is_closed check there because token only rarely is in this danger.  I think this
	   * is the only place where we can be about to call token, and someone has screwed up our port.
	   *
	   * We can't call read_error here because it assumes the input string is ok!
	   */
	  
	  if (port_is_closed(sc->input_port))
	    return(s7_error(sc, sc->read_error_symbol, set_elist_1(sc, s7_make_string_wrapper(sc, "our input port got clobbered!"))));
	  
	  sc->tok = token(sc);
	  switch (sc->tok)
	    {
	    case TOKEN_EOF:
	      {
		/* (eval-string "'a ; b") gets here with 'a -> a, so we need to squelch the pending eval.
		 *   another approach would read-ahead in eval_string_1_ex, but this seems less messy.
		 */
		int64_t top;
		top = s7_stack_top(sc) - 1;
		if (stack_op(sc->stack, top) == OP_EVAL_STRING_1)
		  vector_element(sc->stack, top) = (s7_pointer)OP_EVAL_STRING_2;
	      }
	      break;
	      
	    case TOKEN_RIGHT_PAREN:
	      read_error(sc, "unexpected close paren");
	      
	    case TOKEN_COMMA:
	      read_error(sc, "unexpected comma");
	      
	    default:
	      sc->value = read_expression(sc);
	      sc->current_line = port_line_number(sc->input_port);  /* this info is used to track down missing close parens */
	      sc->current_file = port_filename(sc->input_port);
	      break;
	    }
	  break;
	  
	  
	  /* (read p) from scheme
	   *    "p" becomes current input port for eval's duration, then pops back before returning value into calling expr
	   */
	case OP_READ_DONE:
	  pop_input_port(sc);
	  
	  if (sc->tok == TOKEN_EOF)
	    sc->value = sc->eof_object;
	  sc->current_file = NULL; /* this is for error handling */
	  break;
	  
	  
	  /* load("file"); from C (g_load) -- assume caller will clean up
	   *   read and evaluate exprs until EOF that matches (stack reflects nesting)
	   */
	case OP_LOAD_RETURN_IF_EOF:  /* loop here until eof (via push stack below) */
	  if (sc->tok != TOKEN_EOF)
	    {
	      push_stack_op_let(sc, OP_LOAD_RETURN_IF_EOF);
	      push_stack_op_let(sc, OP_READ_INTERNAL);
	      sc->code = sc->value;
	      goto EVAL;             /* we read an expression, now evaluate it, and return to read the next */
	    }
	  sc->current_file = NULL;
	  return(sc->F);
	  
	  
	  /* (load "file") in scheme
	   *    read and evaluate all exprs, then upon EOF, close current and pop input port stack
	   */
	case OP_LOAD_CLOSE_AND_POP_IF_EOF:
	  if (sc->tok != TOKEN_EOF)
	    {
	      push_stack_op_let(sc, OP_LOAD_CLOSE_AND_POP_IF_EOF); /* was push args, code */
	      if ((!is_string_port(sc->input_port)) ||
		  (port_position(sc->input_port) < port_data_size(sc->input_port)))
		{
		  push_stack_op_let(sc, OP_READ_INTERNAL);
		}
	      else sc->tok = TOKEN_EOF;
	      sc->code = sc->value;
	      goto EVAL;             /* we read an expression, now evaluate it, and return to read the next */
	    }
	  s7_close_input_port(sc, sc->input_port);
	  pop_input_port(sc);
	  sc->current_file = NULL;
	  
	  if (is_multiple_value(sc->value))                    /* (load "file") where "file" is (values 1 2 3) */
	    sc->value = splice_in_values(sc, multiple_value(sc->value));
	  break;
	  
	 	  
	  /* -------------------- sort! (heapsort, done directly so that call/cc in the sort function will work correctly) -------------------- */
          #define SORT_N integer(vector_element(sc->code, 0))
          #define SORT_K integer(vector_element(sc->code, 1))
          #define SORT_J integer(vector_element(sc->code, 2))
          #define SORT_K1 integer(vector_element(sc->code, 3))
          #define SORT_CALLS integer(vector_element(sc->code, 4))
          #define SORT_STOP integer(vector_element(sc->code, 5))
          #define SORT_DATA(K) vector_element(car(sc->args), K)
          #define SORT_LESSP cadr(sc->args)
	  
	HEAPSORT:
	  {
	    s7_int n, j, k;
	    s7_pointer lx;
	    n = SORT_N;
	    k = SORT_K1;
	    
	    if ((n == k) || (k > ((s7_int)(n / 2)))) /* k == n == 0 is the first case */
	      goto START;
	    
	    if (sc->safety > NO_SAFETY)
	      {
		SORT_CALLS++;
		if (SORT_CALLS > SORT_STOP)
		  eval_range_error(sc, "sort! is caught in an infinite loop, comparison: ~S", SORT_LESSP);
	      }
	    j = 2 * k;
	    SORT_J = j;
	    if (j < n)
	      {
		push_stack(sc, OP_SORT1, sc->args, sc->code);
		lx = SORT_LESSP; /* cadr of sc->args */
		if (needs_copied_args(lx))
		  sc->args = list_2(sc, SORT_DATA(j), SORT_DATA(j + 1));
		else
		  {
		    set_car(sc->t2_1, SORT_DATA(j));
		    set_car(sc->t2_2, SORT_DATA(j + 1));
		    sc->args = sc->t2_1;
		  }
		sc->code = lx;
		goto APPLY;
	      }
	    else sc->value = sc->F;
	  }
	  
	case OP_SORT1:
	  {
	    s7_int j, k;
	    s7_pointer lx;
	    k = SORT_K1;
	    j = SORT_J;
	    if (is_true(sc, sc->value))
	      {
		j = j + 1;
		SORT_J = j;
	      }
	    push_stack(sc, OP_SORT2, sc->args, sc->code);
	    lx = SORT_LESSP;
	    if (needs_copied_args(lx))
	      sc->args = list_2(sc, SORT_DATA(k), SORT_DATA(j));
	    else
	      {
		set_car(sc->t2_1, SORT_DATA(k));
		set_car(sc->t2_2, SORT_DATA(j));
		sc->args = sc->t2_1;
	      }
	    sc->code = lx;
	    goto APPLY;
	  }
	  
	case OP_SORT2:
	  {
	    s7_int j, k;
	    k = SORT_K1;
	    j = SORT_J;
	    if (is_true(sc, sc->value))
	      {
		s7_pointer lx;
		lx = SORT_DATA(j);
		SORT_DATA(j) = SORT_DATA(k);
		SORT_DATA(k) = lx;
	      }
	    else goto START;
	    SORT_K1 = SORT_J;
	    goto HEAPSORT;
	  }
	  
	case OP_SORT:
	  /* coming in sc->args is sort args (data less?), sc->code = #(n k 0 ...)
	   * here we call the inner loop until k <= 0 [the local k! -- this is tricky because scheme passes args by value]
	   */
	  {
	    s7_int k;
	    k = SORT_K;
	    if (k > 0)
	      {
		SORT_K = k - 1;
		SORT_K1 = k - 1;
		push_stack(sc, OP_SORT, sc->args, sc->code);
		goto HEAPSORT;
	      }
	    /* else fall through */
	  }

	case OP_SORT3:
	  {
	    s7_int n;
	    s7_pointer lx;
	    n = SORT_N;
	    if (n <= 0)
	      {
		sc->value = car(sc->args);
		goto START;
	      }
	    lx = SORT_DATA(0);
	    SORT_DATA(0) = SORT_DATA(n);
	    SORT_DATA(n) = lx;
	    SORT_N = n - 1;
	    SORT_K1 = 0;
	    push_stack(sc, OP_SORT3, sc->args, sc->code);
	    goto HEAPSORT;
	  }
	  
	case OP_SORT_PAIR_END:       /* sc->value is the sort vector which needs to be copied into the original list */
	  sc->value = vector_into_list(sc->value, car(sc->args));
	  free_cell(sc, sc->args);
	  break;
	  
	case OP_SORT_VECTOR_END:     /* sc->value is the sort (s7_pointer) vector which needs to be copied into the original (double/int) vector */
	  sc->value = vector_into_fi_vector(sc->value, car(sc->args));
	  free_cell(sc, sc->args);
	  break;
	  
	case OP_SORT_STRING_END:
	  sc->value = vector_into_string(sc->value, car(sc->args));
	  free_cell(sc, sc->args); 
	  break;
	  
	  /* batcher networks:
	   *    ((0 2) (0 1) (1 2))
	   *    ((0 2) (1 3) (0 1) (2 3) (1 2))
	   *    etc -- see batcher in s7test.scm (from Doug Hoyte)
	   * but since it has to be done here by hand, it turns into too much code, 3 is:
	   *    < l0 l2 ?
	   *    no goto L1
	   *    < l0 l1 ?
	   *    no  return 1 0 2
	   *    < l1 l2?
	   *    yes return 0 1 2 (direct)
	   *    no  return 0 2 1
	   *  L1:
	   *    < l0 l1 ?
	   *    yes return 2 0 1
	   *    < l1 l2 ?
	   *    yes return 1 2 0
	   *    no  return 2 1 0
	   * since each "<" op above goes to OP_APPLY, we have ca 5 labels, and ca 25-50 lines
	   */
	  
	  /* -------------------------------- MAP -------------------------------- */
	case OP_MAP_GATHER_1:
	  if (sc->value != sc->no_value)
	    {
	      if (is_multiple_value(sc->value))
		counter_set_result(sc->args, revappend(sc, multiple_value(sc->value), counter_result(sc->args)));
	      else counter_set_result(sc->args, cons(sc, sc->value, counter_result(sc->args)));
	    }
	  
	case OP_MAP_1:
	  {
	    s7_pointer x, args, code, p;
	    code = sc->code;
	    args = sc->args;
	    p = counter_list(args);
	    x = s7_iterate(sc, p);
	    
	    if (iterator_is_at_end(p))
	      {
		sc->value = safe_reverse_in_place(sc, counter_result(args));
		/* an experiment */	  
		free_cell(sc, sc->args);
		sc->args = sc->nil; 
		goto START;
	      }
	    push_stack(sc, OP_MAP_GATHER_1, args, code);
	    if (counter_capture(args) != sc->capture_let_counter)
	      {
		new_frame_with_slot(sc, closure_let(code), sc->envir, car(closure_args(code)), x);
		counter_set_let(args, sc->envir);
		counter_set_slots(args, let_slots(sc->envir));
		counter_set_capture(args, sc->capture_let_counter);
	      }
	    else 
	      {
		/* the counter_slots field saves the original local let slot(s) representing the function
		 *   argument.  If the function has internal defines, they get added to the front of the
		 *   slots list, but old_frame_with_slot (maybe stupidly) assumes only the one original
		 *   slot exists when it updates its symbol_id from the (possibly changed) let_id.  So,
		 *   a subsequent reference to the parameter name causes "unbound variable", or a segfault
		 *   if the check has been optimized away.  I think each function call should start with
		 *   the original let slots, so counter_slots saves that pointer, and resets it here.
		 */
		let_set_slots(counter_let(args), counter_slots(args));
		sc->envir = old_frame_with_slot(sc, counter_let(args), x);
	      }
	    sc->code = _TPair(closure_body(code));
	    goto BEGIN1;
	  }
	  
	case OP_MAP_GATHER:
	  if (sc->value != sc->no_value)                   /* (map (lambda (x) (values)) (list 1)) */
	    {
	      if (is_multiple_value(sc->value))            /* (map (lambda (x) (if (odd? x) (values x (* x 20)) (values))) (list 1 2 3 4)) */
		counter_set_result(sc->args, revappend(sc, multiple_value(sc->value), counter_result(sc->args)));
	      /* not append_in_place here because sc->value has the multiple-values bit set */
	      else counter_set_result(sc->args, cons(sc, sc->value, counter_result(sc->args)));
	    }
	  
	case OP_MAP:
	  {
	    s7_pointer y, iterators;
	    iterators = counter_list(sc->args);
	    sc->x = sc->nil;                     /* can't use preset args list here (as in for-each): (map list '(a b c)) */
	    for (y = iterators; is_pair(y); y = cdr(y))
	      {
		s7_pointer x;
		x = s7_iterate(sc, car(y));
		if (iterator_is_at_end(car(y)))
		  {
		    sc->value = safe_reverse_in_place(sc, counter_result(sc->args));
		    free_cell(sc, sc->args); 
		    sc->args = sc->nil;
		    goto START;
		  }
		sc->x = cons(sc, x, sc->x);
	      }
	    sc->x = safe_reverse_in_place(sc, sc->x);
	    push_stack(sc, OP_MAP_GATHER, sc->args, sc->code);
	    sc->args = sc->x;
	    sc->x = sc->nil;
	    
	    if (needs_copied_args(sc->code))
	      sc->args = copy_list(sc, sc->args);
	    goto APPLY;
	  }
	  
	case OP_MAP_GATHER_2:
	case OP_MAP_GATHER_3:
	  if (sc->value != sc->no_value)
	    {
	      if (is_multiple_value(sc->value))
		counter_set_result(sc->args, revappend(sc, multiple_value(sc->value), counter_result(sc->args)));
	      else counter_set_result(sc->args, cons(sc, sc->value, counter_result(sc->args)));
	    }
	case OP_MAP_2:
	  {
	    s7_pointer x, c, p, code;
	    code = sc->code;
	    c = sc->args;
	    p = counter_list(c);
	    if (!is_pair(p))
	      {
		sc->value = safe_reverse_in_place(sc, counter_result(c));
		free_cell(sc, sc->args);
		sc->args = sc->nil; 
		goto START;
	      }
	    x = car(p);
	    counter_set_list(c, cdr(p));

	    if (sc->cur_op == OP_MAP_GATHER_3)
	      {
		closure_set_setter(code, cdr(closure_setter(code)));
		/* this depends on code (the function) being non-recursive, else closure_setter gets stepped on */
		if (closure_setter(code) == counter_list(c))
		  {
		    sc->value = safe_reverse_in_place(sc, counter_result(c));
		    free_cell(sc, c);
		    sc->args = sc->nil;
		    goto START;
		  }
		push_stack(sc, OP_MAP_GATHER_2, c, code);
	      }
	    else push_stack(sc, OP_MAP_GATHER_3, c, code);

	    if (counter_capture(c) != sc->capture_let_counter)
	      {
		new_frame_with_slot(sc, closure_let(code), sc->envir, car(closure_args(code)), x);
		counter_set_let(c, sc->envir);
		counter_set_slots(c, let_slots(sc->envir));
		counter_set_capture(c, sc->capture_let_counter);
	      }
	    else 
	      {
		let_set_slots(counter_let(c), counter_slots(c));
		sc->envir = old_frame_with_slot(sc, counter_let(c), x);
	      }
	    sc->code = car(closure_body(code));
	    goto EVAL;
	  }
	  
	  /* -------------------------------- FOR-EACH -------------------------------- */
	case OP_FOR_EACH:
	  {
	    s7_pointer x, y, iterators, saved_args;
	    iterators = car(sc->args);
	    saved_args = cdr(sc->args);
	    for (x = saved_args, y = iterators; is_pair(x); x = cdr(x), y = cdr(y))
	      {
		set_car(x, s7_iterate(sc, car(y)));
		if (iterator_is_at_end(car(y)))
		  {
		    sc->value = sc->unspecified;
		    free_cell(sc, sc->args);
		    sc->args = sc->nil;
		    goto START;
		  }
	      }
	    push_stack(sc, OP_FOR_EACH, sc->args, sc->code);
	    sc->args = saved_args;
	    if (needs_copied_args(sc->code))
	      sc->args = copy_list(sc, sc->args);
	    goto APPLY;
	  }
	  
	  /* for-each et al remake the local frame, but that's only needed if the local env is exported,
	   *   and that can only happen through make-closure in various guises and curlet.
	   *   owlet captures, but it would require a deliberate error to use it in this context.
	   *   c_objects call object_set_let but that requires a prior curlet or sublet.  So we have
	   *   sc->capture_let_counter that is incremented every time an environment is captured, then
	   *   here we save that ctr, call body, on rerun check ctr, if it has not changed we are safe and can reuse frame.
	   */

	case OP_FOR_EACH_1:
	  {
	    s7_pointer code, counter, p, arg;
	    counter = sc->args;
	    p = counter_list(counter);
	    arg = s7_iterate(sc, p);
	    if (iterator_is_at_end(p))
	      {
		sc->value = sc->unspecified;
		free_cell(sc, counter);
		sc->args = sc->nil;
		goto START;
	      }
	    code = sc->code;
	    if (counter_capture(counter) != sc->capture_let_counter)
	      {
		new_frame_with_slot(sc, closure_let(code), sc->envir, car(closure_args(code)), arg);
		counter_set_let(counter, sc->envir);
		counter_set_slots(counter, let_slots(sc->envir));
		counter_set_capture(counter, sc->capture_let_counter);
	      }
	    else 
	      {
		let_set_slots(counter_let(counter), counter_slots(counter));
		sc->envir = old_frame_with_slot(sc, counter_let(counter), arg);
	      }
	    push_stack(sc, OP_FOR_EACH_1, counter, code);
	    sc->code = _TPair(closure_body(code));
	    goto BEGIN1;
	  }

	case OP_FOR_EACH_3:
	case OP_FOR_EACH_2:
	  {
	    s7_pointer code, c, lst, arg;
	    c = sc->args; /* the counter */
	    lst = counter_list(c);
	    if (!is_pair(lst))  /* '(1 2 . 3) as arg? -- counter_list can be anything here */
	      {
		sc->value = sc->unspecified;
		free_cell(sc, c);
		sc->args = sc->nil;
		goto START;
	      }
	    code = _TClo(sc->code);
	    arg = car(lst);
	    counter_set_list(c, cdr(lst));
	    if (sc->cur_op == OP_FOR_EACH_3)
	      {
		counter_set_result(c, cdr(counter_result(c)));
		if (counter_result(c) == counter_list(c))
		  {
		    sc->value = sc->unspecified;
		    free_cell(sc, c);
		    sc->args = sc->nil;
		    goto START;
		  }
		push_stack(sc, OP_FOR_EACH_2, c, code);
	      }
	    else push_stack(sc, OP_FOR_EACH_3, c, code);
	    if (counter_capture(c) != sc->capture_let_counter)
	      {
		new_frame_with_slot(sc, closure_let(code), sc->envir, car(closure_args(code)), arg);
		counter_set_let(c, sc->envir);
		counter_set_slots(c, let_slots(sc->envir));
		counter_set_capture(c, sc->capture_let_counter);
	      }
	    else 
	      {
		let_set_slots(counter_let(c), counter_slots(c));
		sc->envir = old_frame_with_slot(sc, counter_let(c), arg);
	      }
	    sc->code = car(closure_body(code));
	    goto EVAL;
	  }

	  
	  /* -------------------------------- MEMBER -------------------------------- */
	case OP_MEMBER_IF:
	case OP_MEMBER_IF1:
	  /* code=func, args = (list original args) with opt_fast->position in cadr (the list), value = result of comparison
	   */
	  if (sc->value != sc->F)                      /* previous comparison was not #f -- return list */
	    {
	      sc->value = opt_fast(sc->args);
	      goto START;
	    }
	  if (!is_pair(cdr(opt_fast(sc->args))))           /* no more args -- return #f */
	    {
	      sc->value = sc->F;
	      goto START;
	    }
	  set_opt_fast(sc->args, cdr(opt_fast(sc->args)));     /* cdr down arg list */
	  
	  if (sc->cur_op == OP_MEMBER_IF1)
	    {
	      /* circular list check */
	      if (opt_fast(sc->args) == opt_slow(sc->args))
		{
		  sc->value = sc->F;
		  goto START;
		}
	      set_opt_slow(sc->args, cdr(opt_slow(sc->args)));  /* cdr down the slow list (check for circular list) */
	      push_stack(sc, OP_MEMBER_IF, sc->args, sc->code);
	    }
	  else push_stack(sc, OP_MEMBER_IF1, sc->args, sc->code);

	  if (needs_copied_args(sc->code))
	    sc->args = list_2(sc, caar(sc->args), car(opt_fast(sc->args)));
	  else sc->args = set_plist_2(sc, caar(sc->args), car(opt_fast(sc->args)));
	  goto APPLY;
	  
	  
	  /* -------------------------------- ASSOC -------------------------------- */
	case OP_ASSOC_IF:
	case OP_ASSOC_IF1:
	  /* code=func, args=(list args) with f/opt_fast=list, value=result of comparison
	   *   (assoc 3 '((1 . a) (2 . b) (3 . c) (4 . d)) =)
	   */
	  if (sc->value != sc->F)            /* previous comparison was not #f -- return (car list) */
	    {
	      sc->value = car(opt_fast(sc->args));
	      goto START;
	    }
	  if (!is_pair(cdr(opt_fast(sc->args))))     /* (assoc 3 '((1 . 2) . 3) =) or nil */
	    {
	      sc->value = sc->F;
	      goto START;
	    }
	  set_opt_fast(sc->args, cdr(opt_fast(sc->args)));  /* cdr down arg list */
	  
	  if (sc->cur_op == OP_ASSOC_IF1)
	    {
	      /* circular list check */
	      if (opt_fast(sc->args) == opt_slow(sc->args))
		{
		  sc->value = sc->F;
		  goto START;
		}
	      set_opt_slow(sc->args, cdr(opt_slow(sc->args)));  /* cdr down the slow list */
	      push_stack(sc, OP_ASSOC_IF, sc->args, sc->code);
	    }
	  else push_stack(sc, OP_ASSOC_IF1, sc->args, sc->code);
	  
	  if (!is_pair(car(opt_fast(sc->args))))     /* (assoc 1 '((2 . 2) 3) =) -- we access caaadr below */
	    eval_type_error(sc, "assoc: second arg is not an alist: ~S", sc->args);
	  /* not sure about this -- we could simply skip the entry both here and in g_assoc
	   *   (assoc 1 '((2 . 2) 3)) -> #f
	   *   (assoc 1 '((2 . 2) 3) =) -> error currently
	   */
	  if (needs_copied_args(sc->code))
	    sc->args = list_2(sc, caar(sc->args), caar(opt_fast(sc->args)));
	  else sc->args = set_plist_2(sc, caar(sc->args), caar(opt_fast(sc->args)));
	  goto APPLY;
	  
	  
	  /* -------------------------------- DO -------------------------------- */

#define do_all_x_end(Code) \
  do {s7_pointer End; End = Code; sc->value = c_call(End)(sc, car(End)); if (is_true(sc, sc->value)) {sc->code = cdr(End); goto DO_END_CLAUSES;}} while (0)

	DO_NO_VARS:
	case OP_DO_NO_VARS:
	  {
	    s7_pointer p;
	    int32_t i;
	    sc->pc = 0;
	    for (i = 0, p = cddr(sc->code); is_pair(p); i++, p = cdr(p))
	      if (!cell_optimize(sc, p))
		break;
	    if (is_null(p))
	      {
		s7_pointer end;
		end = cadr(sc->code);
		sc->envir = new_frame_in_env(sc, sc->envir);
		if (i == 1)
		  {
		    while (true)
		      {
			do_all_x_end(end);
			sc->pc = 0;
			sc->opts[0]->v7.fp(sc->opts[0]);
		      }
		  }
		else
		  {
		    if (i == 0) /* null body! */
		      {
			s7_function endf;
			s7_pointer endp;
			endf = c_call(end);
			endp = car(end);
			while (!is_true(sc, sc->value = endf(sc, endp))); /* the assignment is (normally) in the noise */
			sc->code = cdr(end);
			goto DO_END_CLAUSES;
		      }
		    else
		      {
			while (true)
			  {
			    int32_t k;
			    do_all_x_end(end);
			    sc->pc = -1;
			    for (k = 0; k < i; k++)
			      {
				opt_info *o;
				o = sc->opts[++sc->pc];
				o->v7.fp(o);
			      }
			  }
		      }
		  }
	      }
	    /* back out of do_no_vars */
	    pair_set_syntax_symbol(sc->code, sc->do_no_vars_no_opt_symbol);
	    /* fall through */
	  }
	    
	case OP_DO_NO_VARS_NO_OPT:
	  sc->envir = new_frame_in_env(sc, sc->envir);

	case OP_DO_NO_VARS_NO_OPT_1:
	  do_all_x_end(cadr(sc->code));
	  push_stack_no_args(sc, OP_DO_NO_VARS_NO_OPT_1, sc->code);
	  sc->code = _TPair(cddr(sc->code));
	  goto BEGIN1;


	DOTIMES_ONE_STEP:
	case OP_DOTIMES_ONE_STEP:
	  {
	    s7_pointer end, stepper, end_slot;
	    s7_int lim = 0; /* make the compiler happy */

	    /* fprintf(stderr, "dotimes one step %s\n", DISPLAY(sc->code)); */

	    sc->envir = make_do_frame(sc);
	    /* fprintf(stderr, "frame: %s\n", DISPLAY(sc->envir)); */

	    end = cadr(sc->code);
	    do_all_x_end(end);

	    stepper = dox_slot1(sc->envir);
	    end_slot = find_symbol(sc, caddar(end)); /* can't trust local slot here (local_symbol?) */
	    if (is_t_integer(slot_value(end_slot)))
	      {
		/* TODO: need to make sure step not set */
		lim = integer(slot_value(end_slot));
		denominator(slot_value(stepper)) = lim;
		set_step_end(stepper);
	      }

	    if (s7_optimize_nr(sc, cddr(sc->code)))
	      {
		s7_pointer step_expr;
		step_expr = car(slot_expression(stepper));
		/* fprintf(stderr, "slot: %s, end: %s %s\n", DISPLAY(stepper), DISPLAY(end), DISPLAY(slot_expression(stepper))); */

		if ((is_slot(stepper)) &&
		    (c_callee(end) == all_x_c_ss) &&
		    (c_callee(car(end)) == g_equal_2) &&
		    (cadar(end) == slot_symbol(stepper)) &&
		    (is_t_integer(slot_value(stepper))) &&
		    /* (is_safe_stepper(stepper)) && */
		    (caar(end) = sc->add_symbol) &&
		    (cadar(end) == cadr(step_expr)) &&
		    (is_t_integer(caddr(step_expr))) &&
		    (integer(caddr(step_expr)) == 1) &&
		    (is_step_end(stepper)))
		  {
		    opt_info *o;
		    s7_pointer (*fp)(void *o);
		    s7_pointer val;
		    o = sc->opts[0];
		    fp = o->v7.fp;
		    val = make_mutable_integer(sc, integer(slot_value(stepper)));
		    slot_set_value(stepper, val);
		    for (; integer(val) < lim; integer(val)++)
		      {
			sc->pc = 0;
			fp(o);
		      }
		    sc->value = sc->T;
		    sc->code = cdr(end);
		    goto SAFE_DO_END_CLAUSES; /* no multiple-values here */
		  }
		  
		while (true)
		  {
		    sc->pc = 0;
		    sc->opts[0]->v7.fp(sc->opts[0]);
		    update_steppers(sc);
		    /* TODO: get the stepper and handle direct in more than just +1 (do_let|dotimes) */
		    do_all_x_end(end);
		  }
	      }
	  }
	  push_stack_no_args(sc, OP_DOX_STEP, sc->code);
	  sc->code = _TPair(cddr(sc->code));
	  goto BEGIN1;


	  /* -------------------------------- */
	SAFE_DOTIMES:
	case OP_SAFE_DOTIMES:
	  {
	    int32_t choice;
	    choice = safe_dotimes_ex(sc);
	    if (choice == goto_SAFE_DO_END_CLAUSES) goto SAFE_DO_END_CLAUSES;
	    if (choice == goto_BEGIN1) goto BEGIN1;
	    if (choice == goto_OPT_EVAL) goto OPT_EVAL;
	    if (choice == goto_START_WITHOUT_POP_STACK) goto START_WITHOUT_POP_STACK;
	    pair_set_syntax_symbol(sc->code, sc->simple_do_symbol);
	    goto SIMPLE_DO;
	  }
	  
	case OP_SAFE_DOTIMES_STEP_P:
	  {
	    s7_pointer arg;
	    arg = slot_value(sc->args);
	    numerator(arg)++;
	    if (numerator(arg) == denominator(arg))
	      {
		sc->value = sc->T;
		sc->code = cdadr(sc->code);
		goto DO_END_CLAUSES;
	      }
	    push_stack(sc, OP_SAFE_DOTIMES_STEP_P, sc->args, sc->code);
	    sc->code = opt_pair2(sc->code);
	    sc->cur_op = (opcode_t)pair_syntax_op(sc->code);
	    sc->code = cdr(sc->code);
	    goto START_WITHOUT_POP_STACK;
	  }
	  
	case OP_SAFE_DOTIMES_STEP_O:
	  {
	    s7_pointer arg;
	    arg = slot_value(sc->args);
	    numerator(arg)++;
	    if (numerator(arg) == denominator(arg))
	      {
		sc->value = sc->T;
		sc->code = cdadr(sc->code);
		goto DO_END_CLAUSES;
	      }
	    push_stack(sc, OP_SAFE_DOTIMES_STEP_O, sc->args, sc->code);
	    sc->code = _TPair(opt_pair2(sc->code));
	    goto OPT_EVAL;
	  }

	case OP_SAFE_DOTIMES_STEP:
	  {
	    s7_pointer arg;
	    arg = slot_value(sc->args);
	    numerator(arg)++;
	    if (numerator(arg) == denominator(arg))
	      {
		sc->value = sc->T;
		sc->code = cdadr(sc->code);
		goto DO_END_CLAUSES;
	      }
	    push_stack(sc, OP_SAFE_DOTIMES_STEP, sc->args, sc->code);
	    
	    arg = opt_pair2(sc->code);
	    /* here we know the body has more than one form */
	    push_stack_no_args(sc, OP_BEGIN1, cdr(arg));
	    sc->code = car(arg);
	    goto EVAL;
	  }
	  
	SAFE_DO:
	case OP_SAFE_DO:
	  {
	    int32_t choice;
	    choice = safe_do_ex(sc);
	    if (choice == goto_SAFE_DO_END_CLAUSES) goto SAFE_DO_END_CLAUSES;
	    if (choice == goto_EVAL) goto EVAL;
	    if (choice == goto_DO_UNCHECKED) goto DO_UNCHECKED;
	    goto BEGIN1;
	  }

	case OP_SAFE_DO_STEP:
	  {
	    s7_int step, end;
	    s7_pointer args, code, slot;
	    
	    args = sc->envir;
	    code = sc->code;
	    slot = dox_slot1(args);
	    
	    step = s7_integer(slot_value(slot)) + 1;
	    slot_set_value(slot, make_integer(sc, step));
	    end = s7_integer(slot_value(dox_slot2(args)));

	    if ((step == end) ||
		((step > end) &&
		 (opt_cfunc(caadr(code)) == geq_2)))
		{
		  sc->value = sc->T;
		  sc->code = cdadr(code);
		  goto DO_END_CLAUSES;
		}
	    push_stack(sc, OP_SAFE_DO_STEP, sc->args, code);
	    sc->code = _TPair(opt_pair2(code));
	    goto BEGIN1;
	  }
	  
	SIMPLE_DO:
	case OP_SIMPLE_DO:
	  {
	    /* body might not be safe in this case, but the step and end exprs are easy
	     *   "not safe" merely means we hit something that the optimizer can't specialize
	     */
	    /* simple_do: set up local env, check end (c_c?), goto simple_do_ex
	     *   if latter gets s7_optimize, run locally, else goto simple_do_step.
	     *   but is not 1 expr body, etc -- just goto simple_do_step,
	     *   TODO: make this decision in check_do
	     */
	    s7_pointer init, end, code;

	    code = sc->code;
	    sc->envir = new_frame_in_env(sc, sc->envir);
	    init = cadaar(code);
	    if (is_symbol(init))
	      sc->value = find_symbol_checked(sc, init);
	    else
	      {
		if (is_pair(init))
		  sc->value = c_call(init)(sc, cdr(init));
		else sc->value = init;
	      }
	    dox_set_slot1(sc->envir, make_slot_1(sc, sc->envir, caaar(code), sc->value));
	    end = caddr(caadr(code));
	    if (is_symbol(end))
	      dox_set_slot2(sc->envir, find_symbol(sc, end));
	    else dox_set_slot2(sc->envir, make_slot_1(sc, sc->envir, sc->dox_slot_symbol, end));
	    
	    set_car(sc->t2_1, slot_value(dox_slot1(sc->envir)));
	    set_car(sc->t2_2, slot_value(dox_slot2(sc->envir)));
	    sc->value = c_call(caadr(code))(sc, sc->t2_1);
	    if (is_true(sc, sc->value))
	      {
		sc->code = cdadr(code);
		goto DO_END_CLAUSES;
	      }
	    
	    set_opt_pair2(code, cddr(code));
	    if ((is_null(cdr(opt_pair2(code)))) &&      /* one expr in body */
		(is_pair(car(opt_pair2(code)))) &&      /*   and it is a pair */
		(is_symbol(cadr(caddr(caar(code))))) && /* caar=(i 0 (+ i 1)), caddr=(+ i 1), so this is apparently checking that the stepf is reasonable? */
		(is_t_integer(caddr(caddr(caar(code))))))
	      {
		int32_t choice;
		choice = simple_do_ex(sc, code);
		if (choice == goto_START) goto START;
		if (choice == goto_BEGIN1) goto BEGIN1;
		if (choice == goto_DO_END_CLAUSES) goto DO_END_CLAUSES;
	      }
	    /* fprintf(stderr, "%s\n", DISPLAY(code)); */
	    push_stack(sc, OP_SIMPLE_DO_STEP, sc->args, code);
	    sc->code = _TPair(opt_pair2(code));
	    goto BEGIN1;
	  }
	  
	case OP_SIMPLE_DO_STEP:
	  {
	    s7_pointer step, ctr, end, code;
	    ctr = dox_slot1(sc->envir);
	    end = dox_slot2(sc->envir);
	    code = sc->code;
	    
	    step = caddr(caar(code));
	    if (is_symbol(cadr(step)))
	      {
		set_car(sc->t2_1, slot_value(ctr));
		set_car(sc->t2_2, caddr(step));
	      }
	    else
	      {
		set_car(sc->t2_2, slot_value(ctr));
		set_car(sc->t2_1, cadr(step));
	      }
	    slot_set_value(ctr, c_call(step)(sc, sc->t2_1));
	    
	    set_car(sc->t2_1, slot_value(ctr));
	    set_car(sc->t2_2, slot_value(end));
	    sc->value = c_call(caadr(code))(sc, sc->t2_1);
	    if (is_true(sc, sc->value))
	      {
		sc->code = cdadr(code);
		goto DO_END_CLAUSES;
	      }
	    push_stack(sc, OP_SIMPLE_DO_STEP, sc->args, code);
#if DEBUGGING
	    if (sc->cur_op != OP_SIMPLE_DO_STEP)
	      fprintf(stderr, "simple_do_step: %s\n", op_names[sc->cur_op]);
#endif
	    sc->code = _TPair(opt_pair2(code));
	    goto BEGIN1;
	  }
	  
	DOTIMES_P:
	case OP_DOTIMES_P:
	  {
	    int32_t choice;
	    choice = dotimes_p_ex(sc);
	    if (choice == goto_DO_END_CLAUSES) goto DO_END_CLAUSES;
	    if (choice == goto_DO_UNCHECKED) goto DO_UNCHECKED;
	    goto EVAL;
	  }
	  
	case OP_DOTIMES_STEP_P:
	  {
	    s7_pointer ctr, now, end, code, end_test;
	    
	    code = sc->code;
	    ctr = dox_slot1(sc->envir);
	    now = slot_value(ctr);
	    end = slot_value(dox_slot2(sc->envir));
	    end_test = opt_pair2(code);
	    
	    if (is_integer(now))
	      {
		slot_set_value(ctr, make_integer(sc, integer(now) + 1));
		now = slot_value(ctr);
		if (is_integer(end))
		  {
		    if ((integer(now) == integer(end)) ||
			((integer(now) > integer(end)) &&
			 (opt_cfunc(end_test) == geq_2)))
			{
			  sc->value = sc->T;
			  sc->code = cdadr(code);
			  goto DO_END_CLAUSES;
			}
		  }
		else
		  {
		    set_car(sc->t2_1, now);
		    set_car(sc->t2_2, end);
		    sc->value = c_call(end_test)(sc, sc->t2_1);
		    if (is_true(sc, sc->value))
		      {
			sc->code = cdadr(code);
			goto DO_END_CLAUSES;
		      }
		  }
	      }
	    else
	      {
		set_car(sc->t1_1, now);
		slot_set_value(ctr, g_add_s1(sc, sc->t1_1));
		/* (define (hi) (let ((x 0.0) (y 1.0)) (do ((i y (+ i 1))) ((= i 6)) (do ((i i (+ i 1))) ((>= i 7)) (set! x (+ x i)))) x)) */
		set_car(sc->t2_1, slot_value(ctr));
		set_car(sc->t2_2, end);
		sc->value = c_call(end_test)(sc, sc->t2_1);
		if (is_true(sc, sc->value))
		  {
		    sc->code = cdadr(code);
		    goto DO_END_CLAUSES;
		  }
	      }
	    push_stack(sc, OP_DOTIMES_STEP_P, sc->args, code);
	    code = caddr(code);
	    set_current_code(sc, code);
	    sc->cur_op = (opcode_t)pair_syntax_op(code);
	    sc->code = cdr(code);
	    goto START_WITHOUT_POP_STACK;
	  }
	  
	DOX:
	case OP_DOX:
	  {
	    int32_t choice;
	    choice = dox_ex(sc);
	    if (choice == goto_DO_END_CLAUSES) goto DO_END_CLAUSES;
	    if (choice == goto_SAFE_DO_END_CLAUSES) goto SAFE_DO_END_CLAUSES;
	    if (choice == goto_START) goto START;
	    if (choice == goto_BEGIN1) goto BEGIN1;
	    if (choice == goto_START_WITHOUT_POP_STACK) goto START_WITHOUT_POP_STACK;

	    push_stack_no_args(sc, OP_DOX_STEP, sc->code);
	    sc->code = _TPair(cddr(sc->code));
	    goto BEGIN1;
	  }
	  
	case OP_DOX_STEP:
	  {
	    s7_pointer slot, end;
	    for (slot = let_slots(sc->envir); is_slot(slot); slot = next_slot(slot))
	      if (is_pair(slot_expression(slot)))
		slot_set_value(slot, c_call(slot_expression(slot))(sc, car(slot_expression(slot))));
	    end = cadr(sc->code);
	    sc->value = c_call(end)(sc, car(end));
	    if (is_true(sc, sc->value))
	      {
		sc->code = cdr(end);
		goto DO_END_CLAUSES;
	      }
	    push_stack_no_args(sc, OP_DOX_STEP, sc->code);
	    sc->code = _TPair(cddr(sc->code));
	    goto BEGIN1;
	  }
	  
	case OP_DOX_STEP_P:
	  {
	    s7_pointer slot, end;
	    
	    for (slot = let_slots(sc->envir); is_slot(slot); slot = next_slot(slot))
	      if (is_pair(slot_expression(slot)))
		slot_set_value(slot, c_call(slot_expression(slot))(sc, car(slot_expression(slot))));
	    
	    end = cadr(sc->code);
	    sc->value = c_call(end)(sc, car(end));
	    if (is_true(sc, sc->value))
	      {
		sc->code = cdr(end);
		goto DO_END_CLAUSES;
	      }
	    push_stack_no_args(sc, OP_DOX_STEP_P, sc->code);
	    sc->code = caddr(sc->code);
	    sc->cur_op = (opcode_t)pair_syntax_op(sc->code);
	    sc->code = cdr(sc->code);
	    goto START_WITHOUT_POP_STACK;
	  }
	  
	  /* we could use slot_pending_value, slot_expression, not this extra list, but the list seems simpler. */
        #define DO_VAR_SLOT(P) opt_slot1(P)
        #define DO_VAR_NEW_VALUE(P) cdr(P)
        #define DO_VAR_SET_NEW_VALUE(P, Val) set_cdar(P, Val)
        #define DO_VAR_STEP_EXPR(P) car(P)
	  
	DO_STEP:
	case OP_DO_STEP:
	  /* increment all vars, return to endtest
	   *   these are also updated in parallel at the end, so we gather all the incremented values first
	   *
	   * here we know car(sc->args) is not null, args is the list of steppable vars,
	   *   any unstepped vars in the do var section are not in this list, so
	   *   (do ((i 0 (+ i 1)) (j 2)) ...)
	   *   arrives here with sc->args:
	   *   '(((+ i 1) . 0))
	   */
	  push_stack(sc, OP_DO_END, sc->args, sc->code);
	  sc->args = car(sc->args);                /* the var data lists */
	  sc->code = sc->args;                     /* save the top of the list */
	  
	DO_STEP1:
	  /* on each iteration, each arg incr expr is evaluated and the value placed in caddr while we cdr down args
	   * finally args is nil...
	   */
	  if (is_null(sc->args))
	    {
	      s7_pointer x;
	      
	      for (x = sc->code; is_not_null(x); x = cdr(x))
		slot_set_value(DO_VAR_SLOT(car(x)), DO_VAR_NEW_VALUE(car(x)));
	      
	      /* some schemes rebind here, rather than reset, but that is expensive,
	       *    and only matters once in a blue moon (closure over enclosed lambda referring to a do var)
	       *    and the caller can easily mimic the correct behavior in that case by adding a let or using a named let,
	       *    making the rebinding explicit.
	       *
	       * Hmmm... I'll leave this alone, but there are other less cut-and-dried cases:
	       *   (let ((j (lambda () 0))
	       *         (k 0))
	       *     (do ((i (j) (j))
	       *          (j (lambda () 1) (lambda () (+ i 1)))) ; bind here hits different "i" than set!
	       *         ((= i 3) k)
	       *       (set! k (+ k i))))
	       *   is it 6 or 3?
	       *
	       * if we had a way to tell that there were no lambdas in the do expression, would that
	       *   guarantee that set was ok?  Here's a bad case:
	       *   (let ((f #f))
	       *     (do ((i 0 (+ i 1)))
	       *         ((= i 3))
	       *       (let () ; so that the define is ok
	       *         (define (x) i)
	       *         (if (= i 1) (set! f x))))
	       *    (f))
	       * s7 says 3, guile says 1.
	       *
	       * I wonder if what they're actually talking about is a kind of shared value problem.  If we
	       *   set the value directly (not the cdr(binding) but, for example, integer(cdr(binding))), then
	       *   every previous reference gets changed as a side-effect.  In the current code, we're "binding"
	       *   the value in the sense that on each step, a new value is assigned to the step variable.
	       *   In the "direct" case, (let ((v #(0 0 0))) (do ((i 0 (+ i 1))) ((= i 3) v) (set! (v i) i))
	       *   would return #(3 3 3).
	       *
	       * if sc->capture_let_counter changes, would it be sufficient to simply make a new slot?
	       *   I think not; the closure retains the current env chain, not the slots, so we need a new env.
	       */
	      
	      /* sc->value = sc->nil; */
	      pop_stack_no_op(sc);
	      goto DO_END;
	    }
	  push_stack(sc, OP_DO_STEP2, sc->args, sc->code);
	  
	  /* here sc->args is a list like (((i . 0) (+ i 1) 0) ...)
	   *   so sc->code becomes (+ i 1) in this case
	   */
	  sc->code = DO_VAR_STEP_EXPR(car(sc->args));
	  goto EVAL;
	  
	case OP_DO_STEP2:
	  DO_VAR_SET_NEW_VALUE(sc->args, sc->value);  /* save current value */
	  sc->args = cdr(sc->args);                   /* go to next step var */
	  goto DO_STEP1;
	  
	case OP_DO:    /* sc->code is the stuff after "do" */
	  if (is_null(check_do(sc)))
	    {
	      s7_pointer op;
	      op = car(opt_back(sc->code));
	      if (op == sc->dox_symbol)	         goto DOX;
	      if (op == sc->safe_dotimes_symbol) goto SAFE_DOTIMES;
	      if (op == sc->dotimes_p_symbol)    goto DOTIMES_P;
	      if (op == sc->safe_do_symbol)	 goto SAFE_DO;

	      if (op == sc->do_no_vars_symbol)   goto DO_NO_VARS;
	      if (op == sc->dotimes_one_step_symbol)   goto DOTIMES_ONE_STEP;
	      goto SIMPLE_DO;
	    }
	  
	DO_UNCHECKED:
	case OP_DO_UNCHECKED:
	  if (is_null(car(sc->code)))                           /* (do () ...) -- (let ((i 0)) (do () ((= i 1)) (set! i 1))) */
	    {
	      sc->envir = new_frame_in_env(sc, sc->envir);
	      sc->args = cons_unchecked(sc, sc->nil, cadr(sc->code));
	      sc->code = cddr(sc->code);
	      goto DO_END;
	    }
	  /* eval each init value, then set up the new frame (like let, not let*) */
	  sc->args = sc->nil;                             /* the evaluated var-data */
	  sc->value = sc->code;                           /* protect it */
	  sc->code = car(sc->code);                       /* the vars */
	  
	case OP_DO_INIT:
	  if (do_init_ex(sc) == goto_EVAL)
	    goto EVAL;
	  /* fall through */
	  
	DO_END:
	case OP_DO_END:
	  /* here vars have been init'd or incr'd
	   *    args = (list var-data end-expr return-expr-if-any)
	   *      if (do ((i 0 (+ i 1))) ((= i 3) 10)),            args: (vars (= i 3) 10)
	   *      if (do ((i 0 (+ i 1))) ((= i 3))),               args: (vars (= i 3)) and result expr is () == (begin)
	   *      if (do ((i 0 (+ i 1))) (#t 10 12)),              args: (vars #t 10 12), result: ([begin] 10 12) -> 12
	   *      if (call-with-exit (lambda (r) (do () () (r)))), args: '(())
	   *    code = body
	   */
	  
	  if (is_not_null(cdr(sc->args)))
	    {
	      push_stack(sc, OP_DO_END1, sc->args, sc->code);
	      sc->code = cadr(sc->args);               /* evaluate the end expr */
	      goto EVAL;
	    }
	  else
	    {
	    DO_END2:
	      if (is_pair(sc->code))
		{
		  if (is_null(car(sc->args)))
		    push_stack(sc, OP_DO_END, sc->args, sc->code);
		  else push_stack(sc, OP_DO_STEP, sc->args, sc->code);
		  goto BEGIN1;
		}
	      else
		{ 
		  if (is_null(car(sc->args))) /* no steppers */
		    goto DO_END;
		  goto DO_STEP;
		}
	    }
	  
	case OP_DO_END1:
	  /* sc->value is the result of end-test evaluation */
	  if (is_true(sc, sc->value))
	    {
	      /* we're done -- deal with result exprs
	       *   if there isn't an end test, there also isn't a result (they're in the same list)
	       */
	      sc->code = _TLst(cddr(sc->args));   /* result expr (a list -- implicit begin) */
	      free_cell(sc, sc->args);
	      sc->args = sc->nil;
	      if (is_null(sc->code))
		{
		  if (is_multiple_value(sc->value))  /* (define (f) (+ 1 (do ((i 2 (+ i 1))) ((values i (+ i 1)))))) -> 6 */
		    sc->value = splice_in_values(sc, multiple_value(sc->value));
		  /* similarly, if the result is a multiple value:
		   * (define (f) (+ 1 (do ((i 2 (+ i 1))) ((= i 3) (values i (+ i 1)))))) -> 8
		   */
		  goto START;
		}
	      /* might be => here as in cond and case */
	      if (is_null(cdr(sc->code)))
		{
		  sc->code = car(sc->code);
		  goto EVAL;
		}
	      if ((car(sc->code) == sc->feed_to_symbol) &&
		  (s7_symbol_value(sc, sc->feed_to_symbol) == sc->undefined))
		{
		  int32_t res;
		  res = feed_to(sc);
		  if (res == goto_START) goto START;
		  if (res == goto_APPLY) goto APPLY;
		  goto EVAL;
		}
	      push_stack_no_args(sc, OP_BEGIN1, cdr(sc->code));
	      sc->code = car(sc->code);
	      goto EVAL;
	    }
	  goto DO_END2;
	  
	SAFE_DO_END_CLAUSES:
	  if (is_null(sc->code)) /* I don't think multiple values (as test result) can happen here -- all safe do loops involve counters by 1 to some integer end */
	    goto START;
	  goto DO_END_CODE;
	  
	DO_END_CLAUSES:
	  if (is_null(sc->code))
	    {
	      if (is_multiple_value(sc->value))
		sc->value = splice_in_values(sc, multiple_value(sc->value));
	      goto START;
	    }

	DO_END_CODE:
	  if (is_pair(cdr(sc->code)))
	    {
	      if ((car(sc->code) == sc->feed_to_symbol) &&
		  (s7_symbol_value(sc, sc->feed_to_symbol) == sc->undefined))
		{
		  int32_t res;
		  res = feed_to(sc);
		  if (res == goto_START) goto START;
		  if (res == goto_APPLY) goto APPLY;
		  goto EVAL;
		}
	      push_stack_no_args(sc, OP_BEGIN1, cdr(sc->code));
	      sc->code = car(sc->code);
	      goto EVAL;
	    }
	  sc->code = car(sc->code);
	  if (is_pair(sc->code))
	    goto EVAL;
	  if (is_symbol(sc->code))
	    sc->value = find_symbol_checked(sc, sc->code);
	  else sc->value = sc->code;
	  goto START;
	  
	  
	  /* -------------------------------- BEGIN -------------------------------- */
	case OP_BEGIN:
	  if (!s7_is_proper_list(sc, sc->code))    /* proper list includes () */
	    eval_error(sc, "unexpected dot? ~A", sc->code);
	  if (is_null(sc->code))                   /* (begin) -> () */
	    {
	      sc->value = sc->nil;
	      goto START;
	    }
	  if ((is_overlaid(sc->code)) &&
	      (has_opt_back(sc->code)))
	    pair_set_syntax_symbol(sc->code, sc->begin1_symbol);
	  
	case OP_BEGIN1:
	  if ((sc->begin_hook) && (call_begin_hook(sc))) return(sc->F);
	BEGIN1:
	  if (is_pair(cdr(_TPair(sc->code))))
	    push_stack_no_args(sc, OP_BEGIN1, cdr(sc->code));
	  sc->code = car(sc->code);
	  /* goto EVAL; */
	  
	  
	EVAL:
	case OP_EVAL:

	  /* main part of evaluation
	   *   at this point, it's sc->code we care about; sc->args is not set yet.
	   */
	  /* fprintf(stderr, "    eval: %s\n", DISPLAY_80(sc->code)); */

	  if (typesflag(sc->code) == SYNTACTIC_PAIR)  /* xor is not faster here, and bit check is slower */
	    {
#if WITH_PROFILE
	      if (sc->code != profile_at_start)
		profile(sc, sc->code);
#endif
	      set_current_code(sc, sc->code);         /* in case an error occurs, this helps tell us where we are */
	      sc->cur_op = (opcode_t)pair_syntax_op(sc->code);
	      sc->code = cdr(sc->code);
	      goto START_WITHOUT_POP_STACK;	      
	    }
	  
	OPT_EVAL_CHECKED:
	  if (is_optimized(sc->code))
	    {
	      s7_pointer code;
	      /* fprintf(stderr, "    %s %s\n", opt_names[optimize_op(sc->code)], DISPLAY_80(sc->code)); */

	    OPT_EVAL:
	     /* fprintf(stderr, "opt_eval: %s %s\n", opt_names[optimize_op(sc->code)], DISPLAY_80(sc->code)); */

#if WITH_PROFILE
	      if (sc->code != profile_at_start)
		profile(sc, sc->code);
#endif
	      code = sc->code;
	      set_current_code(sc, code);
	      
	      /* it is only slightly faster to use labels as values (computed gotos) here.  The two big switch statements
	       *   at START and below replaced by goto *table[optimize_op(code)] (etc) saves about 1/4 of the switch time
	       *   (according to callgrind).  Since this won't work in MS C++, I'd need 2 versions via macros. 
	       *   It would probably help to store the label (rather than jump through an array of labels), but
	       *   that adds a lot of complexity in the optimizer.
	       *   The switch time is only a small portion of the time spent in this function, so the overall
	       *   savings is much less than 1% -- sometimes there is none at all (the rest of the code appears to
	       *   be slower in the computed goto case -- some timing tests are actually slower with gotos).  Also 
	       *   the lack of a default case means an error => segfault rather than printing some useless error message.
	       */
	      switch (optimize_op(code))
		{
		  /* -------------------------------------------------------------------------------- */
		case OP_SAFE_QUOTE:
		case HOP_SAFE_QUOTE:
		  /* this is for some very special cases that would otherwise fall into unknown_a->trailers 
		   *   all_x_eval assumes it is available even if never used here
		   */
		  sc->value = cadr(code);
		  goto START;
		  

		case OP_SAFE_C_C:
		  if (!c_function_is_ok(sc, code)) {if (is_null(cddr(code))) {set_optimize_op(code, OP_S_C); goto OPT_EVAL;} break;}
		  /* break = fall into the "trailers" section where optimizations are cleared */
		case HOP_SAFE_C_C:
		  sc->value = c_call(code)(sc, cdr(code)); /* this includes all safe calls where all args are constants */
		  goto START;

		case OP_SAFE_C_AND2:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_AND2:
		  code = cdr(code);
		  if (is_false(sc, c_call(code)(sc, car(code))))
		    {
		      sc->value = sc->F; 
		      goto START;
		    }
		  code = cdr(code);
		  sc->value = c_call(code)(sc, car(code));
		  goto START;

		case OP_SAFE_C_OR2:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_OR2:
		  code = cdr(code);
		  sc->value = c_call(code)(sc, car(code));
		  if (is_true(sc, sc->value))
		    goto START;
		  code = cdr(code);
		  sc->value = c_call(code)(sc, car(code));
		  goto START;

		case OP_SAFE_C_S:
		  if (!c_function_is_ok(sc, code)) {set_optimize_op(code, OP_S_S); goto OPT_EVAL;}
		case HOP_SAFE_C_S:
		  set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(code)));
		  sc->value = c_call(code)(sc, sc->t1_1);
		  goto START;
		  
		case OP_SAFE_C_L:
		  if (!c_function_is_ok(sc, code)) {set_optimize_op(code, OP_S_S); goto OPT_EVAL;}
		case HOP_SAFE_C_L:
		  set_car(sc->t1_1, local_symbol_value(cadr(code)));
		  sc->value = c_call(code)(sc, sc->t1_1);
		  goto START;
		  
		case OP_SAFE_CAR_S:
		  if (!c_function_is_ok(sc, code)) {set_optimize_op(code, OP_S_S); goto OPT_EVAL;}
		case HOP_SAFE_CAR_S:
		  {
		    s7_pointer val;
		    val = find_symbol_unchecked(sc, cadr(code));
		    sc->value = (is_pair(val)) ? car(val) : g_car(sc, set_plist_1(sc, val));
		    goto START;
		  }
		  
		case OP_SAFE_CDR_S:
		  if (!c_function_is_ok(sc, code)) {set_optimize_op(code, OP_S_S); goto OPT_EVAL;}
		case HOP_SAFE_CDR_S:
		  {
		    s7_pointer val;
		    val = find_symbol_unchecked(sc, cadr(code));
		    sc->value = (is_pair(val)) ? cdr(val) : g_cdr(sc, set_plist_1(sc, val));
		    goto START;
		  }
		  
		case OP_SAFE_CADR_S:
		  if (!c_function_is_ok(sc, code)) {set_optimize_op(code, OP_S_S); goto OPT_EVAL;}
		case HOP_SAFE_CADR_S:
		  {
		    s7_pointer val;
		    val = find_symbol_unchecked(sc, cadr(code));
		    sc->value = ((is_pair(val)) && (is_pair(cdr(val)))) ? cadr(val) : g_cadr(sc, set_plist_1(sc, val));
		    goto START;
		  }
		  
		case OP_SAFE_IS_PAIR_S:
		  if (!c_function_is_ok(sc, code)) {set_optimize_op(code, OP_S_S); goto OPT_EVAL;}
		case HOP_SAFE_IS_PAIR_S:
		  {
		    s7_pointer val;
		    val = find_symbol_unchecked(sc, cadr(code));
		    eval_boolean_method(sc, is_pair, sc->is_pair_symbol, val);
		    goto START;
		  }

		case OP_SAFE_IS_NULL_S:
		  if (!c_function_is_ok(sc, code)) {set_optimize_op(code, OP_S_S); goto OPT_EVAL;}
		case HOP_SAFE_IS_NULL_S:
		  {
		    s7_pointer val;
		    val = find_symbol_unchecked(sc, cadr(code));
		    eval_boolean_method(sc, is_null, sc->is_null_symbol, val);
		    goto START;
		  }

		case OP_SAFE_IS_SYMBOL_S:  
		  if (!c_function_is_ok(sc, code)) {set_optimize_op(code, OP_S_S); goto OPT_EVAL;}
		case HOP_SAFE_IS_SYMBOL_S:
		  {
		    s7_pointer val;
		    val = find_symbol_unchecked(sc, cadr(code));
		    eval_boolean_method(sc, is_symbol, sc->is_symbol_symbol, val);
		    goto START;
		  }

		case OP_SAFE_C_SS:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_SS:
		  {
		    s7_pointer val, args;
		    args = cdr(code);
		    val = find_symbol_unchecked(sc, car(args));
		    set_car(sc->t2_2, find_symbol_unchecked(sc, cadr(args)));
		    set_car(sc->t2_1, val);
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_LL:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_LL:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    set_car(sc->t2_1, local_symbol_value(car(args)));
		    set_car(sc->t2_2, local_symbol_value(cadr(args)));
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_ALL_S:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_ALL_S:
		  {
		    s7_pointer args, p;

		    sc->args = safe_list_if_possible(sc, integer(arglist_length(code)));
		    for (args = cdr(code), p = sc->args; is_pair(args); args = cdr(args), p = cdr(p))
		      set_car(p, find_symbol_unchecked(sc, car(args)));
		    clear_list_in_use(sc->args);
		    sc->current_safe_list = 0;
		    sc->value = c_call(code)(sc, sc->args);
		    goto START;
		  }
		  
		case OP_SAFE_C_SC:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_SC:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    set_car(sc->t2_1, find_symbol_unchecked(sc, car(args)));
		    set_car(sc->t2_2, cadr(args));
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_CS:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_CS:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    set_car(sc->t2_2, find_symbol_unchecked(sc, cadr(args)));
		    set_car(sc->t2_1, car(args));
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_SQ:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_SQ:
		  set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(code)));
		  set_car(sc->t2_2, opt_con1(cdr(code)));
		  sc->value = c_call(code)(sc, sc->t2_1);
		  goto START;
		  
		case OP_SAFE_C_QS:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_QS:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    set_car(sc->t2_2, find_symbol_unchecked(sc, cadr(args)));
		    set_car(sc->t2_1, cadr(car(args)));
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_QQ:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_QQ:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    set_car(sc->t2_1, cadr(car(args)));
		    set_car(sc->t2_2, cadadr(args));
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_CQ:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_CQ:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    set_car(sc->t2_1, car(args));
		    set_car(sc->t2_2, cadadr(args));
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_QC:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_QC:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    set_car(sc->t2_1, cadr(car(args)));
		    set_car(sc->t2_2, cadr(args));
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_Z:
		  if (!c_function_is_ok(sc, code)) break;
		  /* I think c_function_is_ok of cadr here and below is redundant -- they'll be checked when Z is
		   *    because we cleared the hop bit after combine_ops.
		   */
		case HOP_SAFE_C_Z:
		  check_stack_size(sc);
		  push_stack_no_args(sc, OP_SAFE_C_P_1, code);
		  sc->code = _TPair(cadr(code));
		  goto OPT_EVAL_CHECKED;

		case OP_SAFE_C_P:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_P:
		  push_stack_no_args(sc, OP_SAFE_C_P_1, code);
		  sc->code = _TPair(cadr(code));
		  goto EVAL;
		  
		case OP_NOT_P:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_NOT_P:
		  push_stack_no_args(sc, OP_NOT_P_1, code);
		  sc->code = _TPair(cadr(code));
		  goto EVAL;
		  
		case OP_SAFE_C_CZ:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_CZ:
		  check_stack_size(sc);
		  /* it's possible in a case like this to overflow the stack -- s7test has a deeply
		   *   nested expression involving (+ c (+ c (+ ... ))) all opt'd as safe_c_cz -- if we're close
		   *   to the stack end at the start, it runs off the end.  Normally the stack increase in
		   *   the reader protects us, but a call/cc can replace the original stack with a much smaller one.
		   * How to minimize the cost of this check?
		   */
		  push_stack(sc, OP_EVAL_ARGS_P_2, cadr(code), code);
		  sc->code = _TPair(caddr(code));
		  goto OPT_EVAL_CHECKED;
		  
		case OP_SAFE_C_ZC:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_ZC:
		  check_stack_size(sc);
		  push_stack(sc, OP_EVAL_ARGS_P_4, caddr(code), code);
		  sc->code = _TPair(cadr(code));
		  goto OPT_EVAL_CHECKED;
		  
		case OP_SAFE_C_SZ:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_SZ:
		  /* about 10% appear to be local symbols */
		  check_stack_size(sc);
		  push_stack(sc, OP_EVAL_ARGS_P_2, find_symbol_unchecked(sc, cadr(code)), code);
		  sc->code = _TPair(caddr(code));	            /* splitting out the all_x cases here and elsewhere saves nothing */
		  goto OPT_EVAL_CHECKED;
		  
		case OP_SAFE_C_ZS:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_ZS:
		  check_stack_size(sc);
		  push_stack_no_args(sc, OP_EVAL_ARGS_P_3, code);
		  sc->code = _TPair(cadr(code));
		  goto OPT_EVAL_CHECKED;
		  
		case OP_SAFE_C_opAq:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_opAq:
		  {
		    s7_pointer arg;
		    arg = cadr(code);
		    set_car(sc->a1_1, c_call(cdr(arg))(sc, cadr(arg)));
		    set_car(sc->t1_1, c_call(arg)(sc, sc->a1_1));
		    sc->value = c_call(code)(sc, sc->t1_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_opAAq:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_opAAq:
		  {
		    s7_pointer arg, val;
		    arg = cadr(code);
		    val = c_call(cdr(arg))(sc, cadr(arg));
		    set_car(sc->a2_2, c_call(cddr(arg))(sc, caddr(arg)));
		    set_car(sc->a2_1, val);
		    set_car(sc->t1_1, c_call(arg)(sc, sc->a2_1));
		    sc->value = c_call(code)(sc, sc->t1_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_opAAAq:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_opAAAq:
		  {
		    s7_pointer arg, val1, val2;
		    arg = cadr(code);
		    val1 = c_call(cdr(arg))(sc, cadr(arg));
		    val2 = c_call(cddr(arg))(sc, caddr(arg));
		    set_car(sc->a3_3, c_call(cdddr(arg))(sc, cadddr(arg)));
		    set_car(sc->a3_1, val1);
		    set_car(sc->a3_2, val2);
		    set_car(sc->t1_1, c_call(arg)(sc, sc->a3_1));
		    sc->value = c_call(code)(sc, sc->t1_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_S_opAq:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_S_opAq:
		  {
		    s7_pointer arg;
		    arg = caddr(code);
		    set_car(sc->a1_1, c_call(cdr(arg))(sc, cadr(arg)));
		    set_car(sc->t2_2, c_call(arg)(sc, sc->a1_1));
		    set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(code)));
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_S_opAAq:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_S_opAAq:
		  {
		    s7_pointer arg, val1;
		    arg = caddr(code);
		    val1 = c_call(cdr(arg))(sc, cadr(arg));
		    set_car(sc->a2_2, c_call(cddr(arg))(sc, caddr(arg)));
		    set_car(sc->a2_1, val1);
		    set_car(sc->t2_2, c_call(arg)(sc, sc->a2_1));
		    set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(code)));
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_S_opAAAq:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_S_opAAAq:
		  {
		    s7_pointer arg, p, val1, val2;
		    p = caddr(code);
		    arg = cdr(p);
		    val1 = c_call(arg)(sc, car(arg));
		    arg = cdr(arg);
		    val2 = c_call(arg)(sc, car(arg));
		    arg = cdr(arg);
		    set_car(sc->a3_3, c_call(arg)(sc, car(arg)));
		    set_car(sc->a3_1, val1);
		    set_car(sc->a3_2, val2);
		    set_car(sc->t2_2, c_call(p)(sc, sc->a3_1));
		    set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(code)));
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_S_opSZq:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_S_opSZq:
		  push_stack(sc, OP_SAFE_C_SZ_SZ, find_symbol_unchecked(sc, cadr(caddr(code))), code);
		  sc->code = _TPair(caddr(caddr(code)));
		  goto OPT_EVAL_CHECKED;
		  
		case OP_SAFE_C_AZ:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_AZ:
		  {
		    s7_pointer val;
		    val = c_call(cdr(code))(sc, cadr(code));
		    push_stack(sc, OP_EVAL_ARGS_P_2, val, code);
		    sc->code = _TPair(caddr(code));
		    goto OPT_EVAL_CHECKED;
		  }
		  
		case OP_SAFE_C_ZA:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_ZA:
		  /* here we can't use ZS order because we sometimes assume left->right arg evaluation (binary-io.scm for example) */
		  push_stack_no_args(sc, OP_SAFE_C_ZA_1, code);
		  sc->code = _TPair(cadr(code));
		  goto OPT_EVAL_CHECKED;
		  
		case OP_SAFE_C_ZZ:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_ZZ:
		  /* most of the component Z's here are very complex:
		   *    264600: (+ (* even-amp (oscil (vector-ref evens k) (+ even-freq val))) (* odd-amp...
		   */
		  push_stack_no_args(sc, OP_SAFE_C_ZZ_1, code);
		  sc->code = _TPair(cadr(code));
		  goto OPT_EVAL_CHECKED;
		  
		case OP_SAFE_C_opCq_Z:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_opCq_Z:
		  {
		    s7_pointer val;
		    val = c_call(cadr(code))(sc, cdadr(code));
		    push_stack(sc, OP_EVAL_ARGS_P_2, val, code);
		    sc->code = _TPair(caddr(code));
		    goto OPT_EVAL_CHECKED;
		  }
		  
		case OP_SAFE_C_ZAA:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_ZAA:
		  {
		    s7_pointer val;
		    val = c_call(cddr(code))(sc, caddr(code));
		    push_stack(sc, OP_SAFE_C_ZAA_1, val, code);
		    sc->code = _TPair(cadr(code));
		    goto OPT_EVAL_CHECKED;
		  }
		  
		case OP_SAFE_C_AZA:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_AZA:
		  {
		    s7_pointer val;
		    val = c_call(cdr(code))(sc, cadr(code));
		    push_stack(sc, OP_SAFE_C_AZA_1, val, code);
		    sc->code = _TPair(caddr(code));
		    /* mostly stuff like h_safe_c_aaa */
		    goto OPT_EVAL_CHECKED;
		  }
		  
		case OP_SAFE_C_AAZ:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_AAZ:
		  {
		    s7_pointer val, op_val;
		    op_val = c_call(cdr(code))(sc, cadr(code));
		    val = c_call(cddr(code))(sc, caddr(code));
		    push_op_stack(sc, op_val);
		    push_stack(sc, OP_SAFE_C_AAZ_1, val, code);
		    sc->code = _TPair(cadddr(code));
		    goto OPT_EVAL_CHECKED;
		  }
		  
		case OP_SAFE_C_ZZA:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_ZZA:
		  push_stack_no_args(sc, OP_SAFE_C_ZZA_1, code);
		  sc->code = _TPair(cadr(code));
		  goto OPT_EVAL_CHECKED;
		  
		case OP_SAFE_C_ZAZ:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_ZAZ:
		  push_stack(sc, OP_SAFE_C_ZAZ_1, sc->nil, code);
		  sc->code = _TPair(cadr(code));
		  goto OPT_EVAL_CHECKED;
		  
		case OP_SAFE_C_AZZ:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_AZZ:
		  {
		    s7_pointer val;
		    val = c_call(cdr(code))(sc, cadr(code));
		    push_stack(sc, OP_SAFE_C_AZZ_1, val, code);
		    sc->code = _TPair(caddr(code));
		    goto OPT_EVAL_CHECKED;
		  }
		  
		case OP_SAFE_C_ZZZ:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_ZZZ:
		  push_stack_no_args(sc, OP_SAFE_C_ZZZ_1, code);
		  sc->code = _TPair(cadr(code));
		  goto OPT_EVAL_CHECKED;

		case OP_SAFE_C_A:
		  if (!c_function_is_ok(sc, code)) 
		    {
		      if (unknown_a_ex(sc, find_symbol_checked(sc, car(code))) == goto_OPT_EVAL)
			{
			  if (op_no_hop(sc->code) == OP_SAFE_C_A)
			    {
			      set_car(sc->a1_1, c_call(cdr(code))(sc, cadr(code)));
			      sc->value = c_call(code)(sc, sc->a1_1);
			      goto START;
			    }
			}
		      else set_optimize_op(code, OP_S_A);
		      goto OPT_EVAL;
		    }
		case HOP_SAFE_C_A:
		  set_car(sc->a1_1, c_call(cdr(code))(sc, cadr(code)));
		  sc->value = c_call(code)(sc, sc->a1_1);
		  goto START;
		  
		case OP_SAFE_C_AA:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_AA:
		  {
		    s7_pointer val;
		    val = c_call(cdr(code))(sc, cadr(code));
		    sc->temp4 = val;
		    set_car(sc->a2_2, c_call(cddr(code))(sc, caddr(code)));
		    set_car(sc->a2_1, val);
		    sc->value = c_call(code)(sc, sc->a2_1);
		    sc->temp4 = sc->nil;
		    goto START;
		  }
		  
		case OP_SAFE_C_AAA:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_AAA:
		  {
		    s7_pointer arg, val1, val2;
		    arg = cdr(code);
		    val1 = c_call(arg)(sc, car(arg));
		    sc->temp4 = val1;
		    arg = cdr(arg);
		    val2 = c_call(arg)(sc, car(arg));
		    sc->temp11 = val2;
		    arg = cdr(arg);
		    set_car(sc->t3_3, c_call(arg)(sc, car(arg)));
		    set_car(sc->t3_1, val1);
		    set_car(sc->t3_2, val2);
		    sc->value = c_call(code)(sc, sc->t3_1);
		    sc->temp4 = sc->nil;
		    sc->temp11 = sc->nil;
		    goto START;
		  }
		  
		case OP_SAFE_C_SSA:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_SSA:
		  {
		    s7_pointer arg, val1, val2;
		    arg = cdr(code);
		    val1 = find_symbol_unchecked(sc, car(arg));
		    arg = cdr(arg);
		    val2 = find_symbol_unchecked(sc, car(arg));
		    arg = cdr(arg);
		    set_car(sc->a3_3, c_call(arg)(sc, car(arg)));
		    set_car(sc->a3_1, val1);
		    set_car(sc->a3_2, val2);
		    sc->value = c_call(code)(sc, sc->a3_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_SAS:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_SAS:
		  {
		    s7_pointer arg;
		    arg = cdr(code);
		    set_car(sc->a3_1, find_symbol_unchecked(sc, car(arg)));
		    arg = cdr(arg);
		    set_car(sc->a3_2, c_call(arg)(sc, car(arg)));
		    arg = cdr(arg);
		    set_car(sc->a3_3, find_symbol_unchecked(sc, car(arg)));
		    sc->value = c_call(code)(sc, sc->a3_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_CSA:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_CSA:
		  {
		    s7_pointer arg;
		    arg = cdr(code);
		    set_car(sc->a3_1, car(arg));
		    arg = cdr(arg);
		    set_car(sc->a3_2, find_symbol_unchecked(sc, car(arg)));
		    arg = cdr(arg);
		    set_car(sc->a3_3, c_call(arg)(sc, car(arg)));
		    sc->value = c_call(code)(sc, sc->a3_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_SCA:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_SCA:
		  {
		    s7_pointer arg;
		    arg = cdr(code);
		    set_car(sc->a3_1, find_symbol_unchecked(sc, car(arg)));
		    arg = cdr(arg);
		    set_car(sc->a3_2, car(arg));
		    arg = cdr(arg);
		    set_car(sc->a3_3, c_call(arg)(sc, car(arg)));
		    sc->value = c_call(code)(sc, sc->a3_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_AAAA:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_AAAA:
		  {
		    s7_pointer arg, val1, val2, val3;
		    arg = cdr(code);
		    val1 = c_call(arg)(sc, car(arg));
		    sc->temp4 = val1;
		    arg = cdr(arg);
		    val2 = c_call(arg)(sc, car(arg));
		    sc->temp11 = val2;
		    arg = cdr(arg);
		    val3 = c_call(arg)(sc, car(arg));
		    sc->temp5 = val3;
		    arg = cdr(arg);
		    set_car(sc->a4_4, c_call(arg)(sc, car(arg)));
		    set_car(sc->a4_1, val1);
		    set_car(sc->a4_2, val2);
		    set_car(sc->a4_3, val3);
		    sc->value = c_call(code)(sc, sc->a4_1);
		    sc->temp4 = sc->nil;
		    sc->temp5 = sc->nil;
		    sc->temp11 = sc->nil;
		    goto START;
		  }
		  
		case OP_SAFE_C_ALL_X:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_ALL_X:
		  {
		    s7_pointer args, p;
		    
		    sc->args = safe_list_if_possible(sc, integer(arglist_length(code)));
		    for (args = cdr(code), p = sc->args; is_pair(args); args = cdr(args), p = cdr(p))
		      set_car(p, c_call(args)(sc, car(args)));
		    clear_list_in_use(sc->args);
		    sc->current_safe_list = 0;
		    
		    sc->value = c_call(code)(sc, sc->args);
		    /* we can't release a temp here:
		     *   (define (hi) (vector 14800 14020 (oscil os) (* 1/3 14800) 14800 (* 1/2 14800))) (hi) where os returns non-zero:
		     *   #(14800 14020 <output-string-port> 14800/3 14800 7400)
		     */
		    goto START;
		  }
		  
		case OP_SAFE_C_SQS:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_SQS:
		  {
		    /* (let-set! gen 'fm fm);  many of these are handled in safe_closure_star_s0 */
		    s7_pointer val1, args;
		    args = cdr(code);
		    val1 = find_symbol_unchecked(sc, car(args));
		    set_car(sc->t3_3, find_symbol_unchecked(sc, opt_sym2(args)));
		    set_car(sc->t3_2, opt_con1(args));
		    set_car(sc->t3_1, val1);
		    sc->value = c_call(code)(sc, sc->t3_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_SCS:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_SCS:
		  {
		    /* (define (hi) (let ((x 32) (lst '(0 1))) (list-set! lst 0 x) x)) */
		    s7_pointer val1, args;
		    args = cdr(code);
		    
		    val1 = find_symbol_unchecked(sc, car(args));
		    set_car(sc->t3_3, find_symbol_unchecked(sc, opt_sym2(args)));
		    set_car(sc->t3_2, opt_con1(args));
		    set_car(sc->t3_1, val1);
		    sc->value = c_call(code)(sc, sc->t3_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_SSC:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_SSC:
		  {
		    /* (define (hi) (let ((v #(0 1 2)) (i 0)) (vector-set! v i 1) v)) */
		    s7_pointer val1, args;
		    args = cdr(code);
		    
		    val1 = find_symbol_unchecked(sc, car(args));
		    set_car(sc->t3_2, find_symbol_unchecked(sc, opt_sym1(args)));
		    set_car(sc->t3_3, opt_con2(args));
		    set_car(sc->t3_1, val1);
		    sc->value = c_call(code)(sc, sc->t3_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_SCC:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_SCC:
		  {
		    /* (make-env E :length 100) */
		    s7_pointer args;
		    args = cdr(code);
		    
		    set_car(sc->t3_1, find_symbol_unchecked(sc, car(args)));
		    set_car(sc->t3_2, opt_con1(args));
		    set_car(sc->t3_3, opt_con2(args));
		    sc->value = c_call(code)(sc, sc->t3_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_CSC:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_CSC:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    
		    set_car(sc->t3_2, find_symbol_unchecked(sc, opt_sym1(args)));
		    set_car(sc->t3_1, car(args));
		    set_car(sc->t3_3, opt_con2(args));
		    sc->value = c_call(code)(sc, sc->t3_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_CSS:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_CSS:
		  {
		    s7_pointer val1, args;
		    args = cdr(code);
		    
		    val1 = find_symbol_unchecked(sc, opt_sym2(args));
		    set_car(sc->t3_2, find_symbol_unchecked(sc, opt_sym1(args)));
		    set_car(sc->t3_3, val1);
		    set_car(sc->t3_1, car(args));
		    sc->value = c_call(code)(sc, sc->t3_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_SSS:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_SSS:
		  {
		    s7_pointer val1, val2, args;
		    args = cdr(code);
		    
		    val1 = find_symbol_unchecked(sc, car(args));
		    val2 = find_symbol_unchecked(sc, opt_sym1(args));
		    set_car(sc->t3_3, find_symbol_unchecked(sc, opt_sym2(args)));
		    set_car(sc->t3_1, val1);
		    set_car(sc->t3_2, val2);
		    sc->value = c_call(code)(sc, sc->t3_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_opCq:
		  if (!c_function_is_ok_cadr(sc, code)) break;
		case HOP_SAFE_C_opCq:
		  set_car(sc->t1_1, c_call(cadr(code))(sc, cdadr(code))); /* OP_SAFE_C_C can involve any number of ops */
		  sc->value = c_call(code)(sc, sc->t1_1);
		  goto START;
		  
		case OP_SAFE_C_opSq:
		  if (!c_function_is_ok_cadr(sc, code)) break;
		case HOP_SAFE_C_opSq:
		  {
		    s7_pointer args;
		    args = cadr(code);
		    set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(args)));
		    set_car(sc->t1_1, c_call(args)(sc, sc->t1_1));
		    sc->value = c_call(code)(sc, sc->t1_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_op_opSq_q:
		  if ((!c_function_is_ok(sc, code)) || (!c_function_is_ok(sc, cadr(code))) || (!c_function_is_ok(sc, cadadr(code)))) break;
		case HOP_SAFE_C_op_opSq_q:
		  {
		    s7_pointer outer, args;
		    outer = cadr(code);
		    args = cadr(outer);
		    set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(args)));
		    set_car(sc->t1_1, c_call(args)(sc, sc->t1_1));
		    set_car(sc->t1_1, c_call(outer)(sc, sc->t1_1));
		    sc->value = c_call(code)(sc, sc->t1_1);
		    goto START;
		  }
		  /* op_op_opSq_q_q and op_opSq_q_op_opSq_q got almost no hits */
		  
		case OP_SAFE_C_op_S_opSq_q:
		  if ((!c_function_is_ok(sc, code)) || (!c_function_is_ok(sc, cadr(code))) || (!c_function_is_ok(sc, caddr(cadr(code))))) break;
		case HOP_SAFE_C_op_S_opSq_q:
		  {
		    /* (exp (* r (cos x))) */
		    s7_pointer outer, args;
		    outer = cadr(code);
		    args = caddr(outer);
		    set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(args)));
		    set_car(sc->t2_2, c_call(args)(sc, sc->t1_1));
		    set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(outer)));
		    set_car(sc->t1_1, c_call(outer)(sc, sc->t2_1));
		    sc->value = c_call(code)(sc, sc->t1_1);
		    goto START;
		  }

		case OP_SAFE_C_op_opSq_S_q:
		  if ((!c_function_is_ok(sc, code)) || (!c_function_is_ok(sc, cadr(code))) || (!c_function_is_ok(sc, cadadr(code)))) break;
		case HOP_SAFE_C_op_opSq_S_q:
		  {
		    /* (exp (* (cos x) r)) */
		    s7_pointer outer, args;
		    outer = cadr(code);
		    args = cadr(outer);
		    set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(args)));
		    set_car(sc->t2_1, c_call(args)(sc, sc->t1_1));
		    set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(outer)));
		    set_car(sc->t1_1, c_call(outer)(sc, sc->t2_1));
		    sc->value = c_call(code)(sc, sc->t1_1);
		    goto START;
		  }

		case OP_SAFE_C_PS:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_PS:
		  push_stack_no_args(sc, OP_EVAL_ARGS_P_3, code); /* gotta wait in this case */
		  sc->code = cadr(code);
		  goto EVAL;
		  
		case OP_SAFE_C_PC:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_PC:
		  push_stack(sc, OP_EVAL_ARGS_P_4, caddr(code), code);
		  sc->code = cadr(code);
		  goto EVAL;
		  
		case OP_SAFE_C_PQ:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_PQ:
		  push_stack(sc, OP_EVAL_ARGS_P_4, opt_con2(cdr(code)), code); /* was P_5, but that's the same as P_4 */
		  sc->code = cadr(code);
		  goto EVAL;
		  
		case OP_SAFE_C_ZQ:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_ZQ:
		  push_stack(sc, OP_EVAL_ARGS_P_4, opt_con2(cdr(code)), code);
		  sc->code = cadr(code);
		  goto OPT_EVAL_CHECKED;
		  
		case OP_SAFE_C_SP:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_SP:
		  check_stack_size(sc);
		  push_stack(sc, OP_EVAL_ARGS_P_2, find_symbol_unchecked(sc, cadr(code)), code);
		  sc->code = caddr(code);
		  goto EVAL;
		  
		case OP_SAFE_C_AP:
		  if ((!c_function_is_ok(sc, code)) || (!c_function_is_ok(sc, cadr(code)))) break;
		case HOP_SAFE_C_AP:
		  {
		    s7_pointer val;
		    check_stack_size(sc);
		    val = c_call(cdr(code))(sc, cadr(code));
		    push_stack(sc, OP_EVAL_ARGS_P_2, val, code);
		    sc->code = caddr(code);
		    goto EVAL;
		  }
		  
		case OP_SAFE_C_CP:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_CP:
		  check_stack_size(sc);
		  push_stack(sc, OP_EVAL_ARGS_P_2, cadr(code), code);
		  sc->code = caddr(code);
		  goto EVAL;
		  
		case OP_SAFE_C_QP:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_QP:
		  check_stack_size(sc);
		  push_stack(sc, OP_EVAL_ARGS_P_2, cadadr(code), code);
		  sc->code = caddr(code);
		  goto EVAL;
		  
		case OP_SAFE_C_PP:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_PP:
		  check_stack_size(sc);
		  push_stack_no_args(sc, OP_SAFE_C_PP_1, code);
		  sc->code = cadr(code);
		  goto EVAL;
		  
		case OP_SAFE_C_AAP:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_AAP:
		  {
		    s7_pointer val;
		    check_stack_size(sc);
		    val = c_call(cdr(code))(sc, cadr(code));
		    push_stack(sc, OP_EVAL_ARGS_AAP_1, val, code);
		    sc->code = cadddr(code);
		    goto EVAL;
		  }
		  
		case OP_SAFE_C_opSSq:
		  if (!c_function_is_ok_cadr(sc, code)) break;
		case HOP_SAFE_C_opSSq:
		  {
		    s7_pointer args, val1;
		    args = cadr(code);
		    val1 = find_symbol_unchecked(sc, cadr(args));
		    set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(args)));
		    set_car(sc->t2_1, val1);
		    set_car(sc->t1_1, c_call(args)(sc, sc->t2_1));
		    sc->value = c_call(code)(sc, sc->t1_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_opSCq:
		  if (!c_function_is_ok_cadr(sc, code)) break;
		case HOP_SAFE_C_opSCq:
		  {
		    s7_pointer args;
		    args = cadr(code);
		    set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(args)));
		    set_car(sc->t2_2, caddr(args));
		    set_car(sc->t1_1, c_call(args)(sc, sc->t2_1));
		    sc->value = c_call(code)(sc, sc->t1_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_opCSq:
		  if (!c_function_is_ok_cadr(sc, code)) break;
		case HOP_SAFE_C_opCSq:
		  {
		    s7_pointer args;
		    args = cadr(code);
		    set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(args)));
		    set_car(sc->t2_1, cadr(args));
		    set_car(sc->t1_1, c_call(args)(sc, sc->t2_1));
		    sc->value = c_call(code)(sc, sc->t1_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_opSQq:
		  if (!c_function_is_ok_cadr(sc, code)) break;
		case HOP_SAFE_C_opSQq:
		  {
		    s7_pointer args;
		    args = cadr(code);
		    set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(args)));
		    set_car(sc->t2_2, cadr(caddr(args)));
		    set_car(sc->t1_1, c_call(args)(sc, sc->t2_1));
		    sc->value = c_call(code)(sc, sc->t1_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_opQSq:
		  if (!c_function_is_ok_cadr(sc, code)) break;
		case HOP_SAFE_C_opQSq:
		  {
		    s7_pointer args;
		    args = cadr(code);
		    set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(args)));
		    set_car(sc->t2_1, cadadr(args));
		    set_car(sc->t1_1, c_call(args)(sc, sc->t2_1));
		    sc->value = c_call(code)(sc, sc->t1_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_S_opSq:
		  if (!c_function_is_ok_caddr(sc, code)) break;
		case HOP_SAFE_C_S_opSq:
		  {
		    s7_pointer args, val;
		    args = cdr(code);
		    val = find_symbol_unchecked(sc, car(args));
		    set_car(sc->t1_1, find_symbol_unchecked(sc, opt_sym1(args)));
		    set_car(sc->t2_2, c_call(cadr(args))(sc, sc->t1_1));
		    set_car(sc->t2_1, val);
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_S_opCq:
		  if (!c_function_is_ok_caddr(sc, code))break;
		case HOP_SAFE_C_S_opCq:
		  {
		    s7_pointer args, val;
		    args = cdr(code);
		    val = find_symbol_unchecked(sc, car(args));
		    set_car(sc->t2_2, c_call(cadr(args))(sc, opt_pair1(args))); /* any number of constants here */
		    set_car(sc->t2_1, val);
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_C_opSq:
		  if (!c_function_is_ok_caddr(sc, code)) break;
		case HOP_SAFE_C_C_opSq:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    set_car(sc->t1_1, find_symbol_unchecked(sc, opt_sym1(args)));
		    set_car(sc->t2_2, c_call(cadr(args))(sc, sc->t1_1));
		    set_car(sc->t2_1, car(args));
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_C_opCq:
		  if (!c_function_is_ok_caddr(sc, code)) break;
		case HOP_SAFE_C_C_opCq:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    set_car(sc->t2_2, c_call(cadr(args))(sc, opt_pair1(args))); /* any # of args */
		    set_car(sc->t2_1, car(args));
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_C_opCSq:
		  if (!c_function_is_ok_caddr(sc, code)) break;
		case HOP_SAFE_C_C_opCSq:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    set_car(sc->t2_2, find_symbol_unchecked(sc, opt_sym2(args)));
		    set_car(sc->t2_1, opt_con1(args));
		    set_car(sc->t2_2, c_call(cadr(args))(sc, sc->t2_1));
		    set_car(sc->t2_1, car(args));
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_C_opSSq:
		  if (!c_function_is_ok_caddr(sc, code)) break;
		case HOP_SAFE_C_C_opSSq:
		  {
		    s7_pointer args, val;
		    args = cdr(code);
		    val = find_symbol_unchecked(sc, opt_sym1(args));
		    set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(cadr(args)))); /* opt_sym2(args) */
		    set_car(sc->t2_1, val);
		    set_car(sc->t2_2, c_call(cadr(args))(sc, sc->t2_1));
		    set_car(sc->t2_1, car(args));
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_opCSq_C:
		  if (!c_function_is_ok_cadr(sc, code)) break;
		case HOP_SAFE_C_opCSq_C:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(car(args))));
		    set_car(sc->t2_1, cadr(car(args)));
		    set_car(sc->t2_1, c_call(car(args))(sc, sc->t2_1));
		    set_car(sc->t2_2, cadr(args));
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_opSSq_C:
		  if (!c_function_is_ok_cadr(sc, code)) break;
		case HOP_SAFE_C_opSSq_C:
		  {
		    s7_pointer args, val;
		    args = cdr(code);
		    val = find_symbol_unchecked(sc, cadr(car(args)));
		    set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(car(args))));
		    set_car(sc->t2_1, val);
		    set_car(sc->t2_1, c_call(car(args))(sc, sc->t2_1));
		    set_car(sc->t2_2, cadr(args));
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_opSSq_S:
		  if (!c_function_is_ok_cadr(sc, code)) break;
		case HOP_SAFE_C_opSSq_S:
		  {
		    s7_pointer args, val, val1;
		    args = cdr(code);
		    val = find_symbol_unchecked(sc, cadr(car(args)));
		    val1 = find_symbol_unchecked(sc, cadr(args));
		    set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(car(args))));
		    set_car(sc->t2_1, val);
		    set_car(sc->t2_1, c_call(car(args))(sc, sc->t2_1));
		    set_car(sc->t2_2, val1);
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_op_opSSq_q_C:
		  if ((!c_function_is_ok(sc, code)) || (!c_function_is_ok(sc, cadr(code))) || (!c_function_is_ok(sc, cadadr(code)))) break;
		case HOP_SAFE_C_op_opSSq_q_C:
		  {
		    /* code: (> (magnitude (- old new)) 0.001) */
		    s7_pointer arg;
		    arg = cadadr(code);
		    set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(arg)));
		    set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(arg)));
		    set_car(sc->t1_1, c_call(arg)(sc, sc->t2_1));
		    set_car(sc->t2_1, c_call(cadr(code))(sc, sc->t1_1));
		    set_car(sc->t2_2, caddr(code));
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_op_opSSq_q_S:
		  if ((!c_function_is_ok(sc, code)) || (!c_function_is_ok(sc, cadr(code))) || (!c_function_is_ok(sc, cadadr(code)))) break;
		case HOP_SAFE_C_op_opSSq_q_S:
		  {
		    /* code: (> (magnitude (- old new)) s) */
		    s7_pointer arg;
		    arg = cadadr(code);
		    set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(arg)));
		    set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(arg)));
		    set_car(sc->t1_1, c_call(arg)(sc, sc->t2_1));
		    set_car(sc->t2_1, c_call(cadr(code))(sc, sc->t1_1));
		    set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(code)));
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_op_opSq_q_C:
		  if ((!c_function_is_ok(sc, code)) || (!c_function_is_ok(sc, cadr(code))) || (!c_function_is_ok(sc, cadadr(code)))) break;
		case HOP_SAFE_C_op_opSq_q_C:
		  {
		    s7_pointer arg;
		    arg = cadadr(code);
		    set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(arg)));
		    set_car(sc->t1_1, c_call(arg)(sc, sc->t1_1));
		    set_car(sc->t2_1, c_call(cadr(code))(sc, sc->t1_1));
		    set_car(sc->t2_2, caddr(code));
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_op_opSq_q_S:
		  if ((!c_function_is_ok(sc, code)) || (!c_function_is_ok(sc, cadr(code))) || (!c_function_is_ok(sc, cadadr(code)))) break;
		case HOP_SAFE_C_op_opSq_q_S:
		  {
		    s7_pointer arg;
		    arg = cadadr(code);
		    set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(arg)));
		    set_car(sc->t1_1, c_call(arg)(sc, sc->t1_1));
		    set_car(sc->t2_1, c_call(cadr(code))(sc, sc->t1_1));
		    set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(code)));
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_S_op_opSSq_Sq:
		  if ((!c_function_is_ok(sc, code)) || (!c_function_is_ok(sc, caddr(code))) || (!c_function_is_ok(sc, cadr(caddr(code))))) break;
		case HOP_SAFE_C_S_op_opSSq_Sq:
		  {
		    /* (let () (define (hi a b c d) (+ a (* (- b c) d))) (define (ho) (hi 1 2 3 4)) (ho))
		     *   or actually... (oscil fmosc1 (+ (* fm1-rat vib) fuzz))
		     *      and that is then packaged as opCq...: (* (env indf1) (oscil fmosc1 (+ (* fm1-rat vib) fuzz)))
		     *      and that is then (+ ...)
		     *   but now this is handled in clm2xen.c
		     */
		    s7_pointer args, val, val1;
		    args = caddr(code);                            /* (* (- b c) d) */
		    val1 = cadr(args);
		    val = find_symbol_unchecked(sc, cadr(val1));                  /* b */
		    set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(val1)));       /* c */
		    set_car(sc->t2_1, val);
		    val = find_symbol_unchecked(sc, caddr(args));                 /* d */
		    set_car(sc->t2_1, c_call(val1)(sc, sc->t2_1));    /* (- b c) */
		    set_car(sc->t2_2, val);
		    set_car(sc->t2_2, c_call(args)(sc, sc->t2_1));    /* (* ...) */
		    set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(code)));        /* a */
		    sc->value = c_call(code)(sc, sc->t2_1);        /* (+ ...) */
		    goto START;
		  }
		  
		case OP_SAFE_C_S_op_S_opSqq:
		  if ((!c_function_is_ok(sc, code)) || (!c_function_is_ok(sc, caddr(code))) || (!c_function_is_ok(sc, caddr(caddr(code))))) break;
		case HOP_SAFE_C_S_op_S_opSqq:
		  {
		    /* (let () (define (hi a c d) (+ a (* d (- c)))) (define (ho) (hi 1 3 4)) (ho)) */
		    s7_pointer args, val, val1;
		    args = caddr(code); 
		    val1 = caddr(args);
		    val = find_symbol_unchecked(sc, cadr(args)); 
		    set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(val1))); 
		    set_car(sc->t2_2, c_call(val1)(sc, sc->t1_1));
		    set_car(sc->t2_1, val);
		    set_car(sc->t2_2, c_call(args)(sc, sc->t2_1));
		    set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(code))); 
		    sc->value = c_call(code)(sc, sc->t2_1); 
		    goto START;
		  }
		  
		case OP_SAFE_C_S_op_S_opSSqq:
		  if ((!c_function_is_ok(sc, code)) || (!c_function_is_ok(sc, caddr(code))) || (!c_function_is_ok(sc, caddr(caddr(code))))) break;
		case HOP_SAFE_C_S_op_S_opSSqq:
		  {
		    /* (let () (define (hi a b c d) (+ a (* d (- b c)))) (define (ho) (hi 1 2 3 4)) (ho)) */
		    s7_pointer args, val, val1;
		    args = caddr(code);                            /* (* d (- b c)) */
		    val1 = caddr(args);
		    val = find_symbol_unchecked(sc, cadr(val1));                  /* b */
		    set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(val1)));       /* c */
		    set_car(sc->t2_1, val);
		    val = find_symbol_unchecked(sc, cadr(args));                  /* d */
		    set_car(sc->t2_2, c_call(val1)(sc, sc->t2_1));    /* (- b c) */
		    set_car(sc->t2_1, val);
		    set_car(sc->t2_2, c_call(args)(sc, sc->t2_1));    /* (* ...) */
		    set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(code)));        /* a */
		    sc->value = c_call(code)(sc, sc->t2_1);        /* (+ ...) */
		    goto START;
		  }
		  
		case OP_SAFE_C_S_op_opSSq_opSSqq:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_S_op_opSSq_opSSqq:
		  {
		    /* (* s (f3 (f1 a b) (f2 c d))) */
		    s7_pointer args, op1, op2;
		    int32_t tx;
		    tx = next_tx(sc);
		    
		    args = caddr(code);
		    op1 = cadr(args);
		    op2 = caddr(args);
		    
		    set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(op1)));
		    set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(op1)));
		    sc->t_temps[tx] = c_call(op1)(sc, sc->t2_1);
		    
		    set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(op2)));
		    set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(op2)));
		    set_car(sc->t2_2, c_call(op2)(sc, sc->t2_1));
		    
		    set_car(sc->t2_1, sc->t_temps[tx]);
		    set_car(sc->t2_2, c_call(args)(sc, sc->t2_1));
		    set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(code)));
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_opSCq_S:
		  if (!c_function_is_ok_cadr(sc, code)) break;
		case HOP_SAFE_C_opSCq_S:
		  {
		    s7_pointer args, val1;
		    args = cdr(code);
		    val1 = find_symbol_unchecked(sc, cadr(args));
		    set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(car(args))));
		    set_car(sc->t2_2, caddr(car(args)));
		    set_car(sc->t2_1, c_call(car(args))(sc, sc->t2_1));
		    set_car(sc->t2_2, val1);
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_opSCq_C:
		  if (!c_function_is_ok_cadr(sc, code)) break;
		case HOP_SAFE_C_opSCq_C:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(car(args))));
		    set_car(sc->t2_2, caddr(car(args)));
		    set_car(sc->t2_1, c_call(car(args))(sc, sc->t2_1));
		    set_car(sc->t2_2, cadr(args));
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_opCSq_S:
		  if (!c_function_is_ok_cadr(sc, code)) break;
		case HOP_SAFE_C_opCSq_S:
		  {
		    s7_pointer args, val1;
		    args = cdr(code);
		    val1 = find_symbol_unchecked(sc, cadr(args));
		    set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(car(args))));
		    set_car(sc->t2_1, cadr(car(args)));
		    set_car(sc->t2_1, c_call(car(args))(sc, sc->t2_1));
		    set_car(sc->t2_2, val1);
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_S_opSCq:
		  if (!c_function_is_ok_caddr(sc, code)) break;
		case HOP_SAFE_C_S_opSCq:
		  {
		    s7_pointer val1, args;
		    args = cdr(code);
		    val1 = find_symbol_unchecked(sc, car(args));
		    set_car(sc->t2_1, find_symbol_unchecked(sc, opt_sym1(args)));
		    set_car(sc->t2_2, opt_con2(args));
		    set_car(sc->t2_2, c_call(cadr(args))(sc, sc->t2_1));
		    set_car(sc->t2_1, val1);
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_C_opSCq:
		  if (!c_function_is_ok_caddr(sc, code)) break;
		case HOP_SAFE_C_C_opSCq:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    set_car(sc->t2_1, find_symbol_unchecked(sc, opt_sym1(args)));
		    set_car(sc->t2_2, opt_con2(args));
		    set_car(sc->t2_2, c_call(cadr(args))(sc, sc->t2_1));
		    set_car(sc->t2_1, car(args));
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_S_opSSq:
		  if (!c_function_is_ok_caddr(sc, code)) break;
		case HOP_SAFE_C_S_opSSq:
		  {
		    /* (* a (- b c)) */
		    s7_pointer val1, val2, args;
		    args = cdr(code);
		    val1 = find_symbol_unchecked(sc, car(args));
		    val2 = find_symbol_unchecked(sc, opt_sym1(args));
		    set_car(sc->t2_2, find_symbol_unchecked(sc, opt_sym2(args)));
		    set_car(sc->t2_1, val2);
		    set_car(sc->t2_2, c_call(cadr(args))(sc, sc->t2_1));
		    set_car(sc->t2_1, val1);
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_S_opCSq:
		  if (!c_function_is_ok_caddr(sc, code)) break;
		case HOP_SAFE_C_S_opCSq:
		  {
		    /* (* a (- 1 b)) or (logand a (ash 1 b)) */
		    s7_pointer val1, args;
		    args = cdr(code);
		    val1 = find_symbol_unchecked(sc, car(args));                     /* a */
		    set_car(sc->t2_2, find_symbol_unchecked(sc, opt_sym2(args)));    /* b */
		    set_car(sc->t2_1, opt_con1(args));                               /* 1 */
		    set_car(sc->t2_2, c_call(cadr(args))(sc, sc->t2_1));             /* (- 1 b) */
		    set_car(sc->t2_1, val1);
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_opSq_S:
		  if (!c_function_is_ok_cadr(sc, code)) break;
		case HOP_SAFE_C_opSq_S:
		  {
		    s7_pointer args;
		    int32_t tx;
		    tx = next_tx(sc);
		    args = cdr(code);
		    set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(car(args))));
		    sc->t_temps[tx] = c_call(car(args))(sc, sc->t1_1);
		    set_car(sc->t2_2, find_symbol_unchecked(sc, cadr(args)));
		    set_car(sc->t2_1, sc->t_temps[tx]);
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_opSq_P:
		  if (!c_function_is_ok_cadr(sc, code)) break;
		case HOP_SAFE_C_opSq_P:
		  {
		    s7_pointer args, val;
		    args = cadr(code);
		    check_stack_size(sc);
		    set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(args)));
		    val = c_call(args)(sc, sc->t1_1);
		    push_stack(sc, OP_EVAL_ARGS_P_2, val, sc->code);
		    sc->code = caddr(code);
		    goto EVAL;
		  }
		  
		case OP_SAFE_C_opSq_Q:
		  if (!c_function_is_ok_cadr(sc, code)) break;
		case HOP_SAFE_C_opSq_Q:
		  {
		    s7_pointer arg1;  /* (let-ref (cdr v) 'x) */
		    arg1 = cadr(code);
		    set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(arg1)));
		    set_car(sc->t2_1, c_call(arg1)(sc, sc->t1_1));
		    set_car(sc->t2_2, opt_con2(cdr(code)));               /* cadr(caddr(code)) */
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_opSq_QS:
		  if (!c_function_is_ok_cadr(sc, code)) break;
		case HOP_SAFE_C_opSq_QS:
		  {
		    s7_pointer arg1, arg3;  /* (let-set! (cdr v) 'x y) */
		    arg1 = cadr(code);
		    arg3 = find_symbol_unchecked(sc, opt_sym2(cdr(code)));  /* cadddr(code) */
		    set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(arg1)));
		    set_car(sc->t3_1, c_call(arg1)(sc, sc->t1_1));
		    set_car(sc->t3_2, opt_con1(cdr(code)));                 /* cadr(caddr(code)) */
		    set_car(sc->t3_3, arg3);
		    sc->value = c_call(code)(sc, sc->t3_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_opCq_S:
		  if (!c_function_is_ok_cadr(sc, code)) break;
		case HOP_SAFE_C_opCq_S:
		  {
		    s7_pointer args, val;
		    args = cdr(code);
		    val = find_symbol_unchecked(sc, cadr(args));
		    set_car(sc->t2_1, c_call(car(args))(sc, cdar(args)));
		    set_car(sc->t2_2, val);
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_opCq_C:
		  if (!c_function_is_ok_cadr(sc, code)) break;
		case HOP_SAFE_C_opCq_C:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    set_car(sc->t2_1, c_call(car(args))(sc, cdar(args)));
		    set_car(sc->t2_2, cadr(args)); /* the second C stands for 1 arg? */
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_opSq_C:
		  if (!c_function_is_ok_cadr(sc, code)) break;
		case HOP_SAFE_C_opSq_C:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(car(args))));
		    set_car(sc->t2_1, c_call(car(args))(sc, sc->t1_1));
		    set_car(sc->t2_2, cadr(args));
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_C_op_S_opCqq:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_C_op_S_opCqq:
		  {
		    /* (define (hi a) (< 1.0 (+ a (* a 2)))) */
		    s7_pointer args, arg1, arg2;
		    args = cdr(code);     /* C_op_S_opCqq */
		    arg1 = cadr(args);    /* op_S_opCqq */
		    arg2 = caddr(arg1);   /* opCq */
		    set_car(sc->t2_2, c_call(arg2)(sc, cdr(arg2)));
		    set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(arg1)));
		    set_car(sc->t2_2, c_call(arg1)(sc, sc->t2_1));
		    set_car(sc->t2_1, car(args));
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_opSq_opSq:
		  if (!c_function_is_ok_cadr_caddr(sc, code)) break;
		case HOP_SAFE_C_opSq_opSq:
		  {
		    s7_pointer args;
		    int32_t tx;
		    tx = next_tx(sc);
		    args = cdr(code);
		    set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(car(args))));
		    sc->t_temps[tx] = c_call(car(args))(sc, sc->t1_1);
		    args = cadr(args);
		    set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(args)));
		    set_car(sc->t2_2, c_call(args)(sc, sc->t1_1));
		    set_car(sc->t2_1, sc->t_temps[tx]);
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_opCq_opCq:
		  if (!c_function_is_ok_cadr_caddr(sc, code)) break;
		case HOP_SAFE_C_opCq_opCq:
		  {
		    s7_pointer args;
		    int32_t tx;
		    tx = next_tx(sc);
		    args = cdr(code);
		    sc->t_temps[tx] = c_call(car(args))(sc, cdar(args));
		    set_car(sc->t2_2, c_call(cadr(args))(sc, cdadr(args))); /* this can clobber sc->t2_1! */
		    set_car(sc->t2_1, sc->t_temps[tx]);
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_opSq_opCq:
		  if (!c_function_is_ok_cadr_caddr(sc, code)) break;
		case HOP_SAFE_C_opSq_opCq:
		  {
		    s7_pointer args;
		    int32_t tx;
		    tx = next_tx(sc);
		    args = cdr(code);
		    set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(car(args))));
		    sc->t_temps[tx] = c_call(car(args))(sc, sc->t1_1);
		    args = cadr(args);
		    set_car(sc->t2_2, c_call(args)(sc, cdr(args)));
		    set_car(sc->t2_1, sc->t_temps[tx]);
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_opCq_opSq:
		  if (!c_function_is_ok_cadr_caddr(sc, code)) break;
		case HOP_SAFE_C_opCq_opSq:
		  {
		    s7_pointer args;
		    int32_t tx;
		    tx = next_tx(sc);
		    args = cdr(code);
		    sc->t_temps[tx] = c_call(car(args))(sc, cdar(args));
		    set_car(sc->t1_1, find_symbol_unchecked(sc, cadadr(args)));
		    set_car(sc->t2_2, c_call(cadr(args))(sc, sc->t1_1));
		    set_car(sc->t2_1, sc->t_temps[tx]);
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_opCq_opSSq:
		  if (!c_function_is_ok_cadr_caddr(sc, code)) break;
		case HOP_SAFE_C_opCq_opSSq:
		  {
		    s7_pointer args;
		    int32_t tx;
		    tx = next_tx(sc);
		    /* code: (/ (+ bn 1) (+ bn an)) */
		    args = cdr(code);
		    sc->t_temps[tx] = c_call(car(args))(sc, cdar(args));
		    args = cdr(args);
		    set_car(sc->t2_1, find_symbol_unchecked(sc, cadar(args)));
		    set_car(sc->t2_2, find_symbol_unchecked(sc, caddar(args)));
		    set_car(sc->t2_2, c_call(car(args))(sc, sc->t2_1));
		    set_car(sc->t2_1, sc->t_temps[tx]);
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_opSCq_opSCq:
		  if (!c_function_is_ok_cadr_caddr(sc, code)) break;
		case HOP_SAFE_C_opSCq_opSCq:
		  {
		    s7_pointer args, val2;
		    int32_t tx;
		    tx = next_tx(sc);
		    args = cdr(code);
		    val2 = find_symbol_unchecked(sc, cadadr(args));
		    set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(car(args))));
		    set_car(sc->t2_2, caddr(car(args)));
		    sc->t_temps[tx] = c_call(car(args))(sc, sc->t2_1);
		    set_car(sc->t2_1, val2);
		    set_car(sc->t2_2, caddr(cadr(args)));
		    set_car(sc->t2_2, c_call(cadr(args))(sc, sc->t2_1));
		    set_car(sc->t2_1, sc->t_temps[tx]);
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_opSSq_opSSq:
		  if (!c_function_is_ok_cadr_caddr(sc, code)) break;
		case HOP_SAFE_C_opSSq_opSSq:
		  {
		    s7_pointer args, val3, val4;
		    int32_t tx;
		    tx = next_tx(sc);
		    args = cdr(code);
		    val3 = find_symbol_unchecked(sc, caddr(car(args)));
		    val4 = find_symbol_unchecked(sc, caddr(cadr(args)));
		    
		    set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(car(args))));
		    set_car(sc->t2_2, val3);
		    sc->t_temps[tx] = c_call(car(args))(sc, sc->t2_1);
		    set_car(sc->t2_1, find_symbol_unchecked(sc, cadadr(args)));
		    set_car(sc->t2_2, val4);
		    set_car(sc->t2_2, c_call(cadr(args))(sc, sc->t2_1));
		    set_car(sc->t2_1, sc->t_temps[tx]);
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_opSSq_opSq:
		  if (!c_function_is_ok_cadr_caddr(sc, code)) break;
		case HOP_SAFE_C_opSSq_opSq:
		  {
		    s7_pointer args, val3;
		    int32_t tx;
		    tx = next_tx(sc);
		    args = cdr(code);
		    val3 = find_symbol_unchecked(sc, caddr(car(args)));
		    set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(car(args))));
		    set_car(sc->t2_2, val3);
		    sc->t_temps[tx] = c_call(car(args))(sc, sc->t2_1);
		    set_car(sc->t1_1, find_symbol_unchecked(sc, cadadr(args)));
		    set_car(sc->t2_2, c_call(cadr(args))(sc, sc->t1_1));
		    set_car(sc->t2_1, sc->t_temps[tx]);
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_opSq_opSSq:
		  if (!c_function_is_ok_cadr_caddr(sc, code)) break;
		case HOP_SAFE_C_opSq_opSSq:
		  {
		    s7_pointer args;
		    int32_t tx;
		    tx = next_tx(sc);
		    args = cdr(code);
		    set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(car(args))));
		    sc->t_temps[tx] = c_call(car(args))(sc, sc->t1_1);
		    set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(cadr(args))));
		    set_car(sc->t2_1, find_symbol_unchecked(sc, cadadr(args)));
		    set_car(sc->t2_2, c_call(cadr(args))(sc, sc->t2_1));
		    set_car(sc->t2_1, sc->t_temps[tx]);
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_opSSq_opCq:
		  if (!c_function_is_ok_cadr_caddr(sc, code)) break;
		case HOP_SAFE_C_opSSq_opCq:
		  {
		    s7_pointer arg1, arg2, val3;
		    int32_t tx;
		    tx = next_tx(sc);
		    arg1 = cadr(code);
		    arg2 = caddr(code);
		    val3 = find_symbol_unchecked(sc, caddr(arg1));
		    set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(arg1)));
		    set_car(sc->t2_2, val3);
		    sc->t_temps[tx] = c_call(arg1)(sc, sc->t2_1);
		    set_car(sc->t2_2, c_call(arg2)(sc, cdr(arg2)));
		    set_car(sc->t2_1, sc->t_temps[tx]);
		    sc->value = c_call(code)(sc, sc->t2_1);
		    goto START;
		  }
		  
		case OP_SAFE_IFA_SS_A: /* ((if a s s) a) I think */
		case HOP_SAFE_IFA_SS_A:
		  {
		    s7_function f;
		    f = c_function_call((is_true(sc, c_call(cdar(code))(sc, cadar(code)))) ? opt_con1(code) : opt_con2(code));
		    sc->value = f(sc, set_plist_1(sc, c_call(cdr(code))(sc, cadr(code))));
		    goto START;
		  }
		  

		  /* -------------------------------------------------------------------------------- */
		case OP_C_S:
		  if (!c_function_is_ok(sc, code)) {set_optimize_op(code, OP_S_S); goto OPT_EVAL;}
		case HOP_C_S:
		  sc->args = list_1(sc, find_symbol_unchecked(sc, cadr(code)));
		  sc->value = c_call(code)(sc, sc->args);
		  goto START;
		  
		case OP_READ_S:
		  if (!c_function_is_ok(sc, code)) {set_optimize_op(code, OP_S_S); goto OPT_EVAL;}
		case HOP_READ_S:
		  read_s_ex(sc);
		  goto START;

		case OP_C_A:
		  if (!c_function_is_ok(sc, code)) {set_optimize_op(code, OP_S_A); goto OPT_EVAL;}
		case HOP_C_A:
		  sc->args = list_1(sc, c_call(cdr(code))(sc, cadr(code)));
		  sc->value = c_call(code)(sc, sc->args);
		  goto START;
		  
		case OP_C_Z:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_C_Z:
		  push_stack_no_args(sc, OP_C_P_1, code);
		  sc->code = _TPair(cadr(code));
		  goto OPT_EVAL_CHECKED;
		  
		case OP_C_P:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_C_P:
		  push_stack_no_args(sc, OP_C_P_1, code);
		  sc->code = _TPair(cadr(code));
		  goto EVAL;
		  
		case OP_C_SS:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_C_SS:
		  sc->args = list_2(sc, find_symbol_unchecked(sc, cadr(code)), find_symbol_unchecked(sc, caddr(code)));
		  sc->value = c_call(code)(sc, sc->args);
		  goto START;
		  
		case OP_C_AP:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_C_AP:
		  {
		    s7_pointer val;
		    /*   op_c_ap_1 sends us to apply which calls check_stack_size I think */
		    val = c_call(cdr(code))(sc, cadr(code));
		    push_stack(sc, OP_C_AP_1, val, code);
		    sc->code = caddr(code);
		    goto EVAL;
		  }

		case OP_C_FA: /* op_c_fs was not faster if all_x_s below */
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_C_FA:
		  sc->code = cdadr(code);
		  make_closure_with_let(sc, sc->value, car(sc->code), cdr(sc->code), sc->envir, CLOSURE_ARITY_NOT_SET);
		  /* sc->value=new closure cell, car=args, cdr=body */
		  sc->args = list_2(sc, sc->value, c_call(cddr(code))(sc, caddr(code)));
		  sc->value = c_call(code)(sc, sc->args);
		  goto START;

		case OP_C_FA_1: 
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_C_FA_1:
		  sc->code = cdadr(code);
		  make_closure_with_let(sc, sc->value, car(sc->code), cdr(sc->code), sc->envir, CLOSURE_ARITY_NOT_SET);
		  sc->value = c_call(code)(sc, set_plist_2(sc, sc->value, c_call(cddr(code))(sc, caddr(code))));
		  set_plist_2(sc, sc->nil, sc->nil); /* hooboy -- GC protects plists */
		  goto START;

		case OP_C_AA:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_C_AA:
		  sc->code = c_call(cdr(code))(sc, cadr(code));
		  sc->value = c_call(cddr(code))(sc, caddr(code));
		  sc->args = list_2(sc, sc->code, sc->value);
		  sc->value = c_call(code)(sc, sc->args);
		  goto START;
		  
		case OP_APPLY_SS:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_APPLY_SS:
		  sc->code = find_symbol_unchecked(sc, cadr(code));      /* global search here was slower */
		  sc->args = find_symbol_unchecked(sc, opt_sym2(code));  /* is this right if code=macro? */
		  if (!s7_is_proper_list(sc, sc->args))                  /* (apply + #f) etc */
		    return(apply_list_error(sc, sc->args));
		  if (needs_copied_args(sc->code))
		    sc->args = copy_list(sc, sc->args);
		  goto APPLY;
		  
		case OP_C_S_opSq:
		  if ((!c_function_is_ok(sc, code)) || (!indirect_c_function_is_ok(sc, caddr(code)))) break;
		case HOP_C_S_opSq:
		  {
		    s7_pointer args, val;
		    args = cdr(code);
		    val = find_symbol_unchecked(sc, car(args));
		    set_car(sc->t1_1, find_symbol_unchecked(sc, opt_sym1(args)));
		    sc->args = list_2(sc, val, c_call(cadr(args))(sc, sc->t1_1));
		    sc->value = c_call(code)(sc, sc->args);
		    goto START;
		  }
		  
		case OP_C_S_opCq:
		  if ((!c_function_is_ok(sc, code)) || (!indirect_c_function_is_ok(sc, caddr(code)))) break;
		case HOP_C_S_opCq:
		  {
		    s7_pointer args, val;
		    int32_t tx;
		    tx = next_tx(sc);
		    args = cdr(code);
		    sc->t_temps[tx] = find_symbol_unchecked(sc, car(args));
		    val = c_call(cadr(args))(sc, opt_pair1(args));
		    sc->args = list_2(sc, sc->t_temps[tx], val);
		    sc->value = c_call(code)(sc, sc->args);
		    goto START;
		  }
		  
		case OP_C_SCS:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_C_SCS:
		  {
		    s7_pointer a1, a2;
		    a1 = cdr(code);
		    a2 = cdr(a1);
		    sc->args = list_3(sc, find_symbol_unchecked(sc, car(a1)), car(a2), find_symbol_unchecked(sc, cadr(a2)));
		    sc->value = c_call(code)(sc, sc->args);
		    goto START;
		  }
		  
		case OP_C_ALL_X:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_C_ALL_X:
		  { /* (set-cdr! lst ()) */
		    s7_pointer args, p, new_args;
		    new_args = make_list(sc, integer(arglist_length(code)), sc->nil);
		    sc->args = new_args;      
		    /* GC protect? all_x stuff below can clobber sc->args:
		     *    (catch #f (vector-ref #(1 2) 0 1.0+1.0i) (vector-ref #(1 2) 0 1.0+1.0i))
		     */
		    for (args = cdr(code), p = new_args; is_pair(args); args = cdr(args), p = cdr(p))
		      set_car(p, c_call(args)(sc, car(args)));
		    sc->value = c_call(code)(sc, new_args);
		    goto START;
		  }
		  
		case OP_CALL_WITH_EXIT:
		  if (!c_function_is_ok(sc, code)) break;
		  check_lambda_args(sc, cadadr(code), NULL);
		case HOP_CALL_WITH_EXIT:
		  {
		    s7_pointer go, args;
		    args = opt_pair2(code);
		    go = make_goto(sc);
		    push_stack(sc, OP_DEACTIVATE_GOTO, go, code); /* code arg is ignored, but perhaps this is safer in GC? */
		    new_frame_with_slot(sc, sc->envir, sc->envir, caar(args), go);
		    sc->code = _TPair(cdr(args));
		    goto BEGIN1;
		  }
		  
		case OP_C_CATCH:
		  if (!c_function_is_ok(sc, code)) break;
		  check_lambda_args(sc, cadr(cadddr(code)), NULL);
		case HOP_C_CATCH:
		  {
		    /* (catch #t (lambda () (set! ("hi") #\a)) (lambda args args))
		     *    code is (catch #t (lambda () ....) (lambda args ....))
		     */
		    s7_pointer p, f, args, tag;
		    args = cddr(code);

		    /* defer making the error lambda */
		    /* check catch tag */
		    f = cadr(code);
		    if (!is_pair(f))                     /* (catch #t ...) or (catch sym ...) */
		      {
			if (is_symbol(f))
			  tag = find_symbol_checked(sc, f);
			else tag = f;
		      }
		    else tag = cadr(f);                  /* (catch 'sym ...) */
		    
		    new_cell(sc, p, T_CATCH);            /* the catch object sitting on the stack */
		    catch_tag(p) = tag;
		    catch_goto_loc(p) = s7_stack_top(sc);
		    catch_op_loc(p) = (int32_t)(sc->op_stack_now - sc->op_stack);
		    catch_handler(p) = cdadr(args);      /* not yet a closure... */
		    
		    push_stack(sc, OP_CATCH_1, code, p); /* code ignored here, except by GC */
		    new_frame(sc, sc->envir, sc->envir);
		    sc->code = _TPair(cddar(args));
		    goto BEGIN1;
		  }
		  
		case OP_C_CATCH_ALL:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_C_CATCH_ALL:                    /* (catch #t (lambda () ...) (lambda args #f) */
		  {
		    s7_pointer p;
		    new_frame(sc, sc->envir, sc->envir);
		    p = sc->envir;
		    catch_all_set_goto_loc(p, s7_stack_top(sc));
		    catch_all_set_op_loc(p, sc->op_stack_now - sc->op_stack);
		    push_stack(sc, OP_CATCH_ALL, opt_con2(code), code);
		    sc->code = _TPair(opt_pair1(cdr(code)));       /* the body of the first lambda */
		    goto BEGIN1;
		  }
		  
		case OP_C_CATCH_ALL_Z:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_C_CATCH_ALL_Z:
		  {
		    s7_pointer p;
		    new_frame(sc, sc->envir, sc->envir);
		    p = sc->envir;
		    catch_all_set_goto_loc(p, s7_stack_top(sc));
		    catch_all_set_op_loc(p, sc->op_stack_now - sc->op_stack);
		    push_stack(sc, OP_CATCH_ALL, opt_con2(code), code);
		    sc->code = _TPair(car(opt_pair1(cdr(code))));
		    goto OPT_EVAL_CHECKED;
		  }


		  /* -------------------------------------------------------------------------------- */
		  /* unknown* fallback on these
		   */
		case OP_S:
		case HOP_S:
		  sc->code = find_symbol_unchecked(sc, car(code));
		  sc->args = sc->nil;
		  goto APPLY;

		case OP_S_C:
		case HOP_S_C:
		  sc->code = find_symbol_unchecked(sc, car(code));
		  if (dont_eval_args(sc->code))
		    sc->args = cdr(code);
		  else 
		    {
		      sc->args = sc->t1_1;
		      set_car(sc->t1_1, cadr(code));
		    }
		  if (needs_copied_args(sc->code))
		    sc->args = copy_list(sc, sc->args);
		  goto APPLY;
	      
		case OP_S_S:
		case HOP_S_S:
		  sc->code = find_symbol_unchecked(sc, car(code));
		  if (dont_eval_args(sc->code))
		    sc->args = cdr(code);
		  else 
		    {
		      sc->args = sc->t1_1;
		      set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(code)));
		    }
		  if (needs_copied_args(sc->code))
		    sc->args = copy_list(sc, sc->args);
		  goto APPLY;
	      
		case OP_S_A:
		case HOP_S_A:
		  sc->code = find_symbol_unchecked(sc, car(code));
		  if (dont_eval_args(sc->code))
		    sc->args = cdr(code);
		  else 
		    {
		      set_car(sc->a1_1, c_call(cdr(code))(sc, cadr(code)));
		      sc->args = sc->a1_1;
		    }
		  if (needs_copied_args(sc->code))
		    sc->args = copy_list(sc, sc->args);
		  goto APPLY;


		case OP_SAFE_C_STAR_A:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_STAR_A:
		  set_car(sc->a1_1, c_call(cdr(code))(sc, cadr(code)));
		  sc->args = sc->a1_1;
		  sc->code = opt_cfunc(code);
		  apply_c_function_star(sc);
		  goto START;

		case OP_SAFE_C_STAR_AA:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_STAR_AA:
		  set_car(sc->a2_1, c_call(cdr(code))(sc, cadr(code)));
		  set_car(sc->a2_2, c_call(cddr(code))(sc, caddr(code)));
		  sc->args = sc->a2_1;
		  sc->code = opt_cfunc(code);
		  apply_c_function_star(sc);
		  goto START;

		case OP_SAFE_C_STAR_ALL_X:
		  if (!c_function_is_ok(sc, code)) break;
		case HOP_SAFE_C_STAR_ALL_X:
		  {
		    s7_pointer args, p;
		    sc->args = safe_list_if_possible(sc, integer(arglist_length(code)));
		    for (args = cdr(code), p = sc->args; is_pair(args); args = cdr(args), p = cdr(p))
		      set_car(p, c_call(args)(sc, car(args)));
		    clear_list_in_use(sc->args);
		    sc->current_safe_list = 0;
		    sc->code = opt_cfunc(code);
		    apply_c_function_star(sc);
		    goto START;
		  }


		  /* -------------------------------------------------------------------------------- */
		case OP_THUNK:
		  if (!closure_is_ok(sc, code, MATCH_UNSAFE_CLOSURE, 0)) {if (unknown_ex(sc, sc->last_function) == goto_OPT_EVAL) goto OPT_EVAL; break;}
		case HOP_THUNK:
		  check_stack_size(sc);
		  /* this recursion check is consistent with the other unsafe closure calls, but we're probably in big trouble:
		   *   (letrec ((a (lambda () (cons 1 (b)))) (b (lambda () (a)))) (b))
		   * unfortunately the alternative is a segfault when we wander off the end of the stack.
		   *
		   * It seems that we could use the hop bit here (since it is always off) to choose between BEGIN1 and OPT_EVAL or EVAL,
		   *   but the EVAL choice gains nothing in time, and the OPT_EVAL choice is too tricky -- it is a two-level optimization,
		   *   so if the inner (car(closure_body)) gets unopt'd for some reason, the outer HOP_THUNK never finds
		   *   out, and peculiar things start to happen.  (Also, is_h_optimized would need to be smarter).
		   */
		  new_frame(sc, closure_let(opt_lambda(code)), sc->envir);
		  sc->code = _TPair(closure_body(opt_lambda(code)));
		  goto BEGIN1;
		  
		case OP_SAFE_THUNK:
		  if (!closure_is_ok(sc, code, MATCH_SAFE_CLOSURE, 0)) {if (unknown_ex(sc, sc->last_function) == goto_OPT_EVAL)	goto OPT_EVAL; break;}
		case HOP_SAFE_THUNK: /* no frame needed */
		  /* (let ((x 1)) (let () (define (f) x) (let ((x 0)) (define (g) (set! x 32) (f)) (g)))) */
		  sc->envir = closure_let(opt_lambda(code));
		  sc->code = _TPair(closure_body(opt_lambda(code)));
		  goto BEGIN1;

		case OP_SAFE_THUNK_E:
		  if (!closure_is_equal(sc, code)) {if (unknown_ex(sc, sc->last_function) == goto_OPT_EVAL) goto OPT_EVAL; break;}
		case HOP_SAFE_THUNK_E:
		  sc->envir = closure_let(opt_lambda(code));
		  sc->code = _TPair(car(closure_body(opt_lambda(code))));
		  goto OPT_EVAL;

		case OP_SAFE_THUNK_P:
		  if (!closure_is_equal(sc, code)) {if (unknown_ex(sc, sc->last_function) == goto_OPT_EVAL) goto OPT_EVAL; break;}
		  goto SAFE_LTHUNK_P;
		case OP_SAFE_LTHUNK_P:
		  set_opt_lambda(code, local_symbol_value(car(code)));
		SAFE_LTHUNK_P:
		case HOP_SAFE_THUNK_P:
		case HOP_SAFE_LTHUNK_P:
		  sc->envir = closure_let(opt_lambda(code));
		  sc->code = car(closure_body(opt_lambda(code)));
		  sc->cur_op = (opcode_t)pair_syntax_op(sc->code);
		  sc->code = cdr(sc->code);
		  goto START_WITHOUT_POP_STACK;
		  
		case OP_SAFE_CLOSURE_S:
		  if (!closure_is_ok(sc, code, MATCH_SAFE_CLOSURE, 1)) {if (unknown_g_ex(sc, sc->last_function) == goto_OPT_EVAL) goto OPT_EVAL; break;}
		case HOP_SAFE_CLOSURE_S:
		  /* since a tail call is safe, we can't change the current env's let_id until
		   *   after we do the lookup -- it might be the current func's arg, and we're
		   *   about to call the same func.
		   */
		  sc->envir = old_frame_with_slot(sc, closure_let(opt_lambda(code)), find_symbol_unchecked(sc, opt_sym2(code)));
		  sc->code = _TPair(closure_body(opt_lambda(code)));
		  goto BEGIN1;
		  
		case OP_SAFE_CLOSURE_S_C:
		  /* here and below the closure has to be the original -- matches are no good */
		  if (!closure_is_equal(sc, code)) {if (unknown_g_ex(sc, sc->last_function) == goto_OPT_EVAL) goto OPT_EVAL; break;}
		case HOP_SAFE_CLOSURE_S_C:
		  sc->envir = old_frame_with_slot(sc, closure_let(opt_lambda(code)), find_symbol_unchecked(sc, opt_sym2(code)));
		  sc->code = car(closure_body(opt_lambda(code)));
		  sc->value = c_call(sc->code)(sc, cdr(sc->code));
		  goto START;

		case OP_SAFE_CLOSURE_S_P:
		  if (!closure_is_equal(sc, code)) {if (unknown_g_ex(sc, sc->last_function) == goto_OPT_EVAL) goto OPT_EVAL; break;}
		case HOP_SAFE_CLOSURE_S_P:
		  sc->envir = old_frame_with_slot(sc, closure_let(opt_lambda(code)), find_symbol_unchecked(sc, opt_sym2(code)));
		  sc->code = car(closure_body(opt_lambda(code)));
		  sc->cur_op = (opcode_t)pair_syntax_op(sc->code);
		  sc->code = cdr(sc->code);
		  goto START_WITHOUT_POP_STACK;

		case OP_SAFE_LCLOSURE_L_P:
		  set_opt_lambda(code, local_symbol_value(car(code)));		  
		case HOP_SAFE_LCLOSURE_L_P:
		  sc->envir = old_frame_with_slot(sc, closure_let(opt_lambda(code)), local_symbol_value(opt_sym2(code)));
		  sc->code = car(closure_body(opt_lambda(code)));
		  sc->cur_op = (opcode_t)pair_syntax_op(sc->code);
		  sc->code = cdr(sc->code);
		  goto START_WITHOUT_POP_STACK;

		case OP_SAFE_CLOSURE_C:
		  if (!closure_is_ok(sc, code, MATCH_SAFE_CLOSURE, 1)) {if (unknown_g_ex(sc, sc->last_function) == goto_OPT_EVAL) goto OPT_EVAL; break;}
		case HOP_SAFE_CLOSURE_C:
		  sc->envir = old_frame_with_slot(sc, closure_let(opt_lambda(code)), cadr(code));
		  sc->code = _TPair(closure_body(opt_lambda(code)));
		  goto BEGIN1;
		  
		case OP_SAFE_CLOSURE_P:
		  if (!closure_is_ok(sc, code, MATCH_SAFE_CLOSURE, 1)) 
		    {
		      if ((has_all_x(cdr(code))) &&
			  (unknown_a_ex(sc, sc->last_function) == goto_OPT_EVAL))
			goto OPT_EVAL; 
		      break;
		    }
		case HOP_SAFE_CLOSURE_P:
		  push_stack(sc, OP_SAFE_CLOSURE_P_1, sc->args, sc->code);
		  sc->code = cadr(code);
		  goto EVAL;

		case OP_SAFE_CLOSURE_A:
		  if (!closure_is_ok(sc, code, MATCH_SAFE_CLOSURE, 1)) {if (unknown_a_ex(sc, sc->last_function) == goto_OPT_EVAL) goto OPT_EVAL; break;}
		case HOP_SAFE_CLOSURE_A:
		  sc->envir = old_frame_with_slot(sc, closure_let(opt_lambda(code)), c_call(cdr(code))(sc, cadr(code)));
		  sc->code = _TPair(closure_body(opt_lambda(code)));
		  goto BEGIN1;

		case OP_SAFE_LCLOSURE_A:
		  set_opt_lambda(code, slot_value(local_slot(car(code))));
		case HOP_SAFE_LCLOSURE_A:
		  sc->envir = old_frame_with_slot(sc, closure_let(opt_lambda(code)), c_call(cdr(code))(sc, cadr(code)));
		  sc->code = _TPair(closure_body(opt_lambda(code)));
		  goto BEGIN1;

		case OP_SAFE_LCLOSURE_A_P:
		  set_opt_lambda(code, slot_value(local_slot(car(code))));
		case HOP_SAFE_LCLOSURE_A_P:
		  sc->envir = old_frame_with_slot(sc, closure_let(opt_lambda(code)), c_call(cdr(code))(sc, cadr(code)));
		  sc->code = car(closure_body(opt_lambda(code)));
		  goto EVAL;

		case OP_SAFE_CLOSURE_A_C:
		  if (!closure_is_equal(sc, code)) {if (unknown_a_ex(sc, sc->last_function) == goto_OPT_EVAL) goto OPT_EVAL; break;}
		case HOP_SAFE_CLOSURE_A_C:
		  sc->envir = old_frame_with_slot(sc, closure_let(opt_lambda(code)), c_call(cdr(code))(sc, cadr(code)));
		  sc->code = car(closure_body(opt_lambda(code)));
		  sc->value = c_call(sc->code)(sc, cdr(sc->code));
		  goto START;

		case OP_SAFE_CLOSURE_AP:
		  if (!closure_is_ok(sc, code, MATCH_SAFE_CLOSURE, 2)) break;
		case HOP_SAFE_CLOSURE_AP:
		  {
		    s7_pointer val;
		    val = c_call(cdr(code))(sc, cadr(code));
		    push_stack(sc, OP_SAFE_CLOSURE_AP_1, val, sc->code);
		    sc->code = caddr(code);
		    goto EVAL;
		  }

		case OP_SAFE_CLOSURE_PA:
		  if (!closure_is_ok(sc, code, MATCH_SAFE_CLOSURE, 2)) break;
		case HOP_SAFE_CLOSURE_PA:
		  {
		    s7_pointer val;
		    val = c_call(cddr(code))(sc, caddr(code));
		    push_stack(sc, OP_SAFE_CLOSURE_PA_1, val, sc->code);
		    sc->code = cadr(code);
		    goto EVAL;
		  }

		case OP_SAFE_CLOSURE_SS:
		  if (!closure_is_ok(sc, code, MATCH_SAFE_CLOSURE, 2)) {if (unknown_gg_ex(sc, sc->last_function) == goto_OPT_EVAL) goto OPT_EVAL; break;}
		case HOP_SAFE_CLOSURE_SS:
		  sc->envir = old_frame_with_two_slots(sc, closure_let(opt_lambda(code)), 
						       find_symbol_unchecked(sc, cadr(code)), 
						       find_symbol_unchecked(sc, opt_sym2(code)));
		  sc->code = car(closure_body(opt_lambda(code)));
		  goto EVAL;
		  
		case OP_SAFE_CLOSURE_SS_B:
		  if (!closure_is_ok(sc, code, MATCH_SAFE_CLOSURE, 2)) {if (unknown_gg_ex(sc, sc->last_function) == goto_OPT_EVAL) goto OPT_EVAL; break;}
		case HOP_SAFE_CLOSURE_SS_B:
		  sc->envir = old_frame_with_two_slots(sc, closure_let(opt_lambda(code)), 
						       find_symbol_unchecked(sc, cadr(code)), 
						       find_symbol_unchecked(sc, opt_sym2(code)));
		  sc->code = closure_body(opt_lambda(code));
		  push_stack_no_args(sc, OP_BEGIN1, _TPair(cdr(sc->code)));
		  sc->code = car(sc->code);
		  goto EVAL;
		  
		case OP_SAFE_CLOSURE_SC:
		  if (!closure_is_ok(sc, code, MATCH_SAFE_CLOSURE, 2)) {if (unknown_gg_ex(sc, sc->last_function) == goto_OPT_EVAL) goto OPT_EVAL; break;}
		case HOP_SAFE_CLOSURE_SC:
		  sc->envir = old_frame_with_two_slots(sc, closure_let(opt_lambda(code)), find_symbol_unchecked(sc, cadr(code)), opt_con2(code));
		  sc->code = _TPair(closure_body(opt_lambda(code)));
		  goto BEGIN1;
		  
		case OP_SAFE_CLOSURE_CS:
		  if (!closure_is_ok(sc, code, MATCH_SAFE_CLOSURE, 2)) {if (unknown_gg_ex(sc, sc->last_function) == goto_OPT_EVAL) goto OPT_EVAL; break;}
		case HOP_SAFE_CLOSURE_CS:
		  sc->envir = old_frame_with_two_slots(sc, closure_let(opt_lambda(code)), cadr(code), find_symbol_unchecked(sc, opt_sym2(code)));
		  sc->code = _TPair(closure_body(opt_lambda(code)));
		  goto BEGIN1;
		  
		case OP_SAFE_CLOSURE_SA:
		  if (!closure_is_ok(sc, code, MATCH_SAFE_CLOSURE, 2)) {if (unknown_aa_ex(sc, sc->last_function) == goto_OPT_EVAL) goto OPT_EVAL; break;}
		case HOP_SAFE_CLOSURE_SA:
		  {
		    s7_pointer args;
		    args = cddr(code);
		    args = c_call(args)(sc, car(args));
		    sc->envir = old_frame_with_two_slots(sc, closure_let(opt_lambda(code)), find_symbol_unchecked(sc, cadr(code)), args);
		    sc->code = _TPair(closure_body(opt_lambda(code)));
		    goto BEGIN1;
		  }

		case OP_SAFE_CLOSURE_AA:
		  if (!closure_is_ok(sc, code, MATCH_SAFE_CLOSURE, 2)) {if (unknown_aa_ex(sc, sc->last_function) == goto_OPT_EVAL) goto OPT_EVAL; break;}
		case HOP_SAFE_CLOSURE_AA:
		  {
		    s7_pointer args, z;
		    int32_t tx;
		    tx = next_tx(sc);
		    args = cdr(code);
		    sc->t_temps[tx] = c_call(args)(sc, car(args));
		    args = cdr(args);
		    z = c_call(args)(sc, car(args));
		    sc->envir = old_frame_with_two_slots(sc, closure_let(opt_lambda(code)), sc->t_temps[tx], z);
		    sc->code = _TPair(closure_body(opt_lambda(code)));
		    goto BEGIN1;
		  }
		  
		case OP_SAFE_CLOSURE_SAA:
		  if (!closure_is_ok(sc, code, MATCH_SAFE_CLOSURE, 3)) break;
		case HOP_SAFE_CLOSURE_SAA:
		  {
		    s7_pointer args, z;
		    int32_t tx;
		    tx = next_tx(sc);
		    args = cddr(code);
		    sc->t_temps[tx] = c_call(args)(sc, car(args));
		    args = cdr(args);
		    z = c_call(args)(sc, car(args));
		    sc->envir = old_frame_with_three_slots(sc, closure_let(opt_lambda(code)), find_symbol_unchecked(sc, cadr(code)), sc->t_temps[tx], z);
		    sc->code = _TPair(closure_body(opt_lambda(code)));
		    goto BEGIN1;
		  }
		  
		case OP_SAFE_CLOSURE_ALL_X:
		  if (!closure_is_ok(sc, code, MATCH_SAFE_CLOSURE, integer(arglist_length(code)))) break;
		case HOP_SAFE_CLOSURE_ALL_X:
		  {
		    s7_pointer args, p, env, x, z;
		    uint64_t id;

		    sc->args = safe_list_if_possible(sc, integer(arglist_length(code)));
		    for (args = cdr(code), p = sc->args; is_pair(args); args = cdr(args), p = cdr(p))
		      set_car(p, c_call(args)(sc, car(args)));
		    clear_list_in_use(sc->args);
		    sc->current_safe_list = 0;
		    sc->code = opt_lambda(code);
		    
		    id = ++sc->let_number;
		    env = closure_let(sc->code);
		    let_id(env) = id;
		    
		    for (x = let_slots(env), z = sc->args; is_slot(x); x = next_slot(x), z = cdr(z))
		      {
			slot_set_value(x, car(z));
			symbol_set_local(slot_symbol(x), id, x);
		      }
		    sc->envir = env;
		    sc->code = closure_body(sc->code);
		    
		    if (is_pair(cdr(sc->code)))
		      {
			push_stack_no_args(sc, OP_BEGIN1, cdr(sc->code));
			sc->code = car(sc->code);
		      }
		    else
		      {
			sc->code = car(sc->code);
			if (is_optimized(sc->code))
			  goto OPT_EVAL;
		      }
		    goto EVAL;
		  }
		  
		  
		  /* -------------------------------------------------------------------------------- */
		  
		case OP_SAFE_CLOSURE_STAR_A:
		  if (!closure_star_is_ok(sc, code, MATCH_SAFE_CLOSURE_STAR, 1)) {if (unknown_a_ex(sc, sc->last_function) == goto_OPT_EVAL) goto OPT_EVAL; break;}		  
		case HOP_SAFE_CLOSURE_STAR_A:
		  {
		    s7_pointer p, x, val;
		    val = c_call(cdr(code))(sc, cadr(code));
		    if (is_keyword(val))
		      s7_error(sc, sc->wrong_type_arg_symbol,
			       set_elist_4(sc, s7_make_string_wrapper(sc, "~A: keyword argument's value is missing: ~S in ~S"),
					   closure_name(sc, opt_lambda(code)), val, sc->args));
		    sc->envir = old_frame_with_slot(sc, closure_let(opt_lambda(code)), val);
		    /* that sets the first arg to the passed symbol value; now set default values, if any */

		    /* fill_safe_closure_star(sc, next_slot(let_slots(closure_let(opt_lambda(code)))), cdr(closure_args(opt_lambda(code)))); */
		    p = cdr(closure_args(opt_lambda(code)));
		    x = next_slot(let_slots(closure_let(opt_lambda(code))));
		    for (; is_pair(p); p = cdr(p), x = next_slot(x))
		      {
			if (is_pair(car(p)))
			  {
			    s7_pointer defval;
			    defval = cadar(p);
			    if (is_pair(defval))
			      slot_set_value(x, cadr(defval));
			    else slot_set_value(x, defval);
			  }
			else slot_set_value(x, sc->F);
			symbol_set_local(slot_symbol(x), let_id(sc->envir), x);
		      }
		    sc->code = _TPair(closure_body(opt_lambda(sc->code)));
		    goto BEGIN1;
		  }
		  
		case OP_SAFE_CLOSURE_STAR_AA:
		  if (!closure_star_is_ok(sc, code, MATCH_SAFE_CLOSURE_STAR, 2)) break;
		case HOP_SAFE_CLOSURE_STAR_AA:
		  {
		    /* here closure_arity == 2 and we have 2 args, so no need to worry about trailing defaults */
		    s7_pointer arg1, arg2, clet, cargs;
		    clet = closure_let(opt_lambda(code));
		    cargs = closure_args(opt_lambda(code));
		    arg1 = c_call(cdr(code))(sc, cadr(code));
		    arg2 = c_call(cddr(code))(sc, caddr(code));

		    if (is_keyword(arg1))
		      {
			if (keyword_symbol(arg1) == slot_symbol(let_slots(clet)))
			  {
			    arg1 = arg2;
			    arg2 = cadr(cargs);
			    if (is_pair(arg2)) arg2 = cadr(arg2); else arg2 = sc->F;
			  }
			else
			  {
			    if (keyword_symbol(arg1) == slot_symbol(next_slot(let_slots(clet))))
			      {
				arg1 = car(cargs);
				if (is_pair(arg1)) arg1 = cadr(arg1); else arg1 = sc->F;
			      }
			    else
			      s7_error(sc, sc->wrong_type_arg_symbol,
				       set_elist_4(sc, s7_make_string_wrapper(sc, "~A: unknown keyword argument: ~S in ~S"),
						   closure_name(sc, opt_lambda(code)), arg1, code));
			  }
		      }
		    else
		      {
			if (is_keyword(arg2))
			  s7_error(sc, sc->wrong_type_arg_symbol,
				   set_elist_4(sc, s7_make_string_wrapper(sc, "~A: keyword argument's value is missing: ~S in ~S"),
					       closure_name(sc, opt_lambda(code)), arg2, code));
		      }
		    sc->envir = old_frame_with_two_slots(sc, clet, arg1, arg2);
		    sc->code = _TPair(closure_body(opt_lambda(code)));
		    goto BEGIN1;
		  }
		  
		case OP_SAFE_CLOSURE_STAR_S0:
		  /* an old (probably now unnecessary) kludge to speed up generators.scm */
		  if (!closure_is_equal(sc, code)) {if (unknown_g_ex(sc, sc->last_function) == goto_OPT_EVAL) goto OPT_EVAL; break;}
		case HOP_SAFE_CLOSURE_STAR_S0:
		  /* here we know we have (let-set! arg1 'name arg2) (with-env arg1 ...) as the safe closure body.
		   *   since no errors can come from the first, there's no need for the procedure env.
		   *   so do the set and with-env by hand, leaving with the env body.
		   */
		  {
		    s7_pointer e;
		    e = find_symbol_unchecked(sc, cadr(code));         /* S of S0 above */
		    if (e == sc->rootlet)
		      sc->envir = sc->nil;
		    else
		      {
			if (!is_let(e))
			  eval_type_error(sc, "with-let takes an environment argument: ~A", e);
			sc->envir = e;
			set_with_let_let(e);
		      }
		    
		    if (e != sc->rootlet)
		      {
			s7_pointer p;
			let_id(e) = ++sc->let_number;
			for (p = let_slots(e); is_slot(p); p = next_slot(p))
			  {
			    s7_pointer sym;
			    sym = slot_symbol(p);
			    symbol_set_local(sym, sc->let_number, p);
			  }
			slot_set_value(local_slot(opt_sym1(cdr(code))), real_zero); /* "arg2" above */
		      }
		    sc->code = _TPair(opt_pair2(cdr(code)));
		    goto BEGIN1;
		  }
		  
		case OP_SAFE_CLOSURE_STAR_ALL_X:
		  if (!closure_star_is_ok(sc, code, MATCH_SAFE_CLOSURE_STAR, integer(arglist_length(code)))) break;
		case HOP_SAFE_CLOSURE_STAR_ALL_X:
		  {
		    s7_pointer p, old_args;
		    sc->w = cdr(code);               /* args aren't evaluated yet */
		    sc->args = make_list(sc, integer(arglist_length(code)), sc->F);
		    for (p = sc->args, old_args = sc->w; is_pair(p); p = cdr(p), old_args = cdr(old_args))
		      set_car(p, c_call(old_args)(sc, car(old_args)));
		    sc->w = sc->nil;
		    sc->code = opt_lambda(code);
		    check_stack_size(sc);
		    sc->envir = new_frame_in_env(sc, closure_let(sc->code));
		    /* sc->envir = closure_let(sc->code); */
		    if (apply_lambda_star(sc) == goto_EVAL) goto EVAL;
		    goto BEGIN1;
		  }
		  

		  /* -------------------------------------------------------------------------------- */
		  
		case OP_GOTO:
		  set_opt_goto(code, find_symbol_checked(sc, car(code)));
		  if (!is_goto(opt_goto(code)))	{if (unknown_ex(sc, opt_goto(code)) == goto_OPT_EVAL) goto OPT_EVAL; break;}
		case HOP_GOTO:
		  sc->args = sc->nil;
		  sc->code = opt_goto(code);
		  call_with_exit(sc);
		  goto START;
		  
		case OP_GOTO_C:
		  /* call-with-exit repeat use internally is very rare, so let's just look it up */
		  set_opt_goto(code, find_symbol_checked(sc, car(code)));
		  if (!is_goto(opt_goto(code))) {if (unknown_g_ex(sc, opt_goto(code)) == goto_OPT_EVAL) goto OPT_EVAL; break;}
		case HOP_GOTO_C:
		  /* (return #t) -- recognized via OP_UNKNOWN_G, opt_goto(code) is the function [parallels OP_CLOSURE_C] */
		  sc->args = cdr(code);
		  sc->code = opt_goto(code);
		  call_with_exit(sc);
		  goto START;
		  
		case OP_GOTO_S:
		  set_opt_goto(code, find_symbol_unchecked(sc, car(code)));
		  if (!is_goto(opt_goto(code))) {if (unknown_g_ex(sc, opt_goto(code)) == goto_OPT_EVAL) goto OPT_EVAL; break;}		  
		case HOP_GOTO_S:
		  sc->args = list_1(sc, find_symbol_unchecked(sc, cadr(code)));
		  /* I think this needs listification because call_with_exit might call dynamic unwinders etc. */
		  sc->code = opt_goto(code);
		  call_with_exit(sc);
		  goto START;
		  
		case OP_GOTO_A:
		  set_opt_goto(code, find_symbol_checked(sc, car(code)));
		  if (!is_goto(opt_goto(code))) {if (unknown_a_ex(sc, opt_goto(code)) == goto_OPT_EVAL) goto OPT_EVAL; break;}		  
		case HOP_GOTO_A:
		  sc->args = list_1(sc, c_call(cdr(code))(sc, cadr(code)));
		  sc->code = opt_goto(code);
		  call_with_exit(sc);
		  goto START;
		  /* for T_CONTINUATION, set sc->args to list_1(sc, ...) as in goto (and code?), then call_with_current_continuation */


		case OP_CLOSURE_ANY_ALL_X:
		  if (!closure_is_ok(sc, code, MATCH_UNSAFE_CLOSURE, -1)) break;
		case HOP_CLOSURE_ANY_ALL_X:
		  {
		    s7_pointer p, old_args;
		    check_stack_size(sc);
		    sc->w = cdr(code);               /* args aren't evaluated yet */
		    sc->args = make_list(sc, integer(arglist_length(code)), sc->F);
		    for (p = sc->args, old_args = sc->w; is_pair(p); p = cdr(p), old_args = cdr(old_args))
		      set_car(p, c_call(old_args)(sc, car(old_args)));
		    sc->w = sc->nil;
		    sc->code = opt_lambda(code);
		    new_frame_with_slot(sc, closure_let(sc->code), sc->envir, closure_args(sc->code), sc->args);
		    sc->code = _TPair(closure_body(sc->code));
		    goto BEGIN1;
		  }
		  
		  
		case OP_CLOSURE_C:
		  if (!closure_is_ok(sc, code, MATCH_UNSAFE_CLOSURE, 1)) {if (unknown_g_ex(sc, sc->last_function) == goto_OPT_EVAL) goto OPT_EVAL; break;}		  
		case HOP_CLOSURE_C:
		  check_stack_size(sc);
		  code = opt_lambda(code);
		  new_frame_with_slot(sc, closure_let(code), sc->envir, car(closure_args(code)), cadr(sc->code));
		  sc->code = _TPair(closure_body(code));
		  goto BEGIN1;

		case OP_CLOSURE_P:
		  if (!closure_is_ok(sc, code, MATCH_UNSAFE_CLOSURE, 1)) {if (unknown_g_ex(sc, sc->last_function) == goto_OPT_EVAL) goto OPT_EVAL; break;}		  
		case HOP_CLOSURE_P:
		  push_stack(sc, OP_CLOSURE_P_1, sc->args, sc->code);
		  sc->code = cadr(code);
		  goto EVAL;
		  
		case OP_CLOSURE_A:
		  if (!closure_is_ok(sc, code, MATCH_UNSAFE_CLOSURE, 1)) {if (unknown_a_ex(sc, sc->last_function) == goto_OPT_EVAL) goto OPT_EVAL; fprintf(stderr, "closure_a gives up\n"); break;}		  
		case HOP_CLOSURE_A:
		  sc->value = c_call(cdr(code))(sc, cadr(code));
		  check_stack_size(sc);
		  code = opt_lambda(code);
		  new_frame_with_slot(sc, closure_let(code), sc->envir, car(closure_args(code)), sc->value);
		  sc->code = _TPair(closure_body(code));
		  goto BEGIN1;
		  
		case OP_CLOSURE_A_P:
		  if (!closure_is_equal(sc, code)) {if (unknown_a_ex(sc, sc->last_function) == goto_OPT_EVAL) goto OPT_EVAL; break;}		  
		case HOP_CLOSURE_A_P:
		  sc->value = c_call(cdr(code))(sc, cadr(code));
		  check_stack_size(sc);
		  code = opt_lambda(code);
		  new_frame_with_slot(sc, closure_let(code), sc->envir, car(closure_args(code)), sc->value);
		  code = car(closure_body(code));
		  sc->cur_op = (opcode_t)pair_syntax_op(code);
		  sc->code = cdr(code);
		  goto START_WITHOUT_POP_STACK;
		  
		case OP_CLOSURE_S:
		  if (!closure_is_ok(sc, code, MATCH_UNSAFE_CLOSURE, 1)) {if (unknown_g_ex(sc, sc->last_function) == goto_OPT_EVAL) goto OPT_EVAL; break;}		  
		case HOP_CLOSURE_S:
		  sc->value = find_symbol_unchecked(sc, opt_sym2(code));
		  check_stack_size(sc);
		  code = opt_lambda(code);
		  new_frame_with_slot(sc, closure_let(code), sc->envir, car(closure_args(code)), sc->value);
		  sc->code = _TPair(closure_body(code));
		  goto BEGIN1;
		  
		case OP_CLOSURE_SS:
		  if (!closure_is_ok(sc, code, MATCH_UNSAFE_CLOSURE, 2)) {if (unknown_gg_ex(sc, sc->last_function) == goto_OPT_EVAL) goto OPT_EVAL; break;}		  
		case HOP_CLOSURE_SS:
		  {
		    s7_pointer f, args;
		    check_stack_size(sc);
		    f = opt_lambda(sc->code);
		    args = closure_args(f);
		    new_frame_with_two_slots(sc, closure_let(f), sc->envir, 
					     car(args), find_symbol_unchecked(sc, cadr(code)),
					     cadr(args), find_symbol_unchecked(sc, opt_sym2(code)));
		    sc->code = _TPair(closure_body(f));
		    goto BEGIN1;
		  }

		case OP_CLOSURE_SS_P:
		  if (!closure_is_equal(sc, code)) {if (unknown_gg_ex(sc, sc->last_function) == goto_OPT_EVAL) goto OPT_EVAL; break;}		  
		case HOP_CLOSURE_SS_P:
		  {
		    s7_pointer func, args;
		    check_stack_size(sc);
		    func = opt_lambda(sc->code); 
		    args = closure_args(func);
		    new_frame_with_two_slots(sc, closure_let(func), sc->envir, 
					     car(args), find_symbol_unchecked(sc, cadr(code)),
					     cadr(args), find_symbol_unchecked(sc, opt_sym2(code)));
		    sc->code = car(closure_body(func));
		    sc->cur_op = (opcode_t)pair_syntax_op(sc->code);
		    sc->code = cdr(sc->code);
		    goto START_WITHOUT_POP_STACK;	      
		  }
		  
		case OP_CLOSURE_SC:
		  if (!closure_is_ok(sc, code, MATCH_UNSAFE_CLOSURE, 2)) {if (unknown_gg_ex(sc, sc->last_function) == goto_OPT_EVAL) goto OPT_EVAL; break;}		  
		case HOP_CLOSURE_SC:
		  {
		    s7_pointer f, args;
		    check_stack_size(sc);
		    f = opt_lambda(sc->code);
		    args = closure_args(f);
		    new_frame_with_two_slots(sc, closure_let(f), sc->envir, car(args), find_symbol_unchecked(sc, cadr(code)), cadr(args), opt_con2(code));
		    sc->code = _TPair(closure_body(f));
		    goto BEGIN1;
		  }
		  
		case OP_CLOSURE_CS:
		  if (!closure_is_ok(sc, code, MATCH_UNSAFE_CLOSURE, 2)) {if (unknown_gg_ex(sc, sc->last_function) == goto_OPT_EVAL) goto OPT_EVAL; break;}		  
		case HOP_CLOSURE_CS:
		  {
		    s7_pointer f, args;
		    check_stack_size(sc);
		    f = opt_lambda(sc->code);
		    args = closure_args(f);
		    new_frame_with_two_slots(sc, closure_let(f), sc->envir, car(args), cadr(code), cadr(args), find_symbol_unchecked(sc, opt_sym2(code)));
		    sc->code = _TPair(closure_body(f));
		    goto BEGIN1;
		  }
		  
		case OP_CLOSURE_AA:
		  if (!closure_is_ok(sc, code, MATCH_UNSAFE_CLOSURE, 2)) {if (unknown_aa_ex(sc, sc->last_function) == goto_OPT_EVAL) goto OPT_EVAL; break;}		  
		case HOP_CLOSURE_AA:
		  {
		    s7_pointer f, args, a_args;
		    int32_t tx;
		    check_stack_size(sc);
		    tx = next_tx(sc);
		    a_args = cdr(code);
		    sc->t_temps[tx] = c_call(a_args)(sc, car(a_args));
		    f = opt_lambda(sc->code);
		    args = closure_args(f);
		    new_frame_with_two_slots(sc, closure_let(f), sc->envir, car(args), sc->t_temps[tx], cadr(args), c_call(cdr(a_args))(sc, cadr(a_args)));
		    sc->code = _TPair(closure_body(f));
		    goto BEGIN1;
		  }

		case OP_CLOSURE_AP:
		  if (!closure_is_ok(sc, code, MATCH_UNSAFE_CLOSURE, 2)) break;
		case HOP_CLOSURE_AP:
		  {
		    s7_pointer val;
		    val = c_call(cdr(code))(sc, cadr(code)); /* don't use this as a push_stack argument */
		    push_stack(sc, OP_CLOSURE_AP_1, val, code);
		    sc->code = caddr(code);
		    goto EVAL;
		  }

		case OP_CLOSURE_PA:
		  if (!closure_is_ok(sc, code, MATCH_UNSAFE_CLOSURE, 2)) break;
		case HOP_CLOSURE_PA:
		  {
		    s7_pointer val;
		    val = c_call(cddr(code))(sc, caddr(code)); /* don't use this as a push_stack argument */
		    push_stack(sc, OP_CLOSURE_PA_1, val, code);
		    sc->code = cadr(code);
		    goto EVAL;
		  }

		case OP_CLOSURE_FA:
		  if (!closure_is_ok(sc, code, MATCH_UNSAFE_CLOSURE, 2)) break;
		case HOP_CLOSURE_FA:
		  {
		    s7_pointer farg, larg, aarg, func, func_args;
		    farg = cdadr(code);
		    aarg = c_call(cddr(code))(sc, caddr(code));
		    make_closure_with_let(sc, larg, car(farg), cdr(farg), sc->envir, CLOSURE_ARITY_NOT_SET);
		    check_stack_size(sc);
		    func = opt_lambda(sc->code);         /* outer func */
		    func_args = closure_args(func);
		    new_frame_with_two_slots(sc, closure_let(func), sc->envir, car(func_args), larg, cadr(func_args), aarg);
		    sc->code = car(closure_body(func));
		    goto EVAL;
		  }

		case OP_CLOSURE_ALL_S:
		  if (!closure_is_ok(sc, code, MATCH_UNSAFE_CLOSURE, integer(arglist_length(code))))
		    {
		      if (unknown_all_s_ex(sc, sc->last_function) == goto_OPT_EVAL)
			goto OPT_EVAL;		       
		      break;
		    }
		case HOP_CLOSURE_ALL_S:
		  {
		    s7_pointer args, p, func, e;
		    /* in this case, we have just lambda (not lambda*), and no dotted arglist,
		     *   and no accessed symbols in the arglist, and we know the arglist matches the parameter list.
		     */
		    check_stack_size(sc);
		    func = opt_lambda(code);
		    /* we need to get the slot names from the current function, but the values from the calling environment */
		    new_frame(sc, closure_let(func), e);
		    sc->z = e;
		    for (p = closure_args(func), args = cdr(code); is_pair(p); p = cdr(p), args = cdr(args))
		      add_slot(e, car(p), find_symbol_unchecked(sc, car(args)));
		    sc->envir = e;
		    sc->z = sc->nil;
		    sc->code = _TPair(closure_body(func));
		    goto BEGIN1;
		  }
		  
		case OP_CLOSURE_ALL_S_P:
		  if (!closure_is_equal(sc, code)) {if (unknown_all_s_ex(sc, sc->last_function) == goto_OPT_EVAL) goto OPT_EVAL; break;}		  
		case HOP_CLOSURE_ALL_S_P:
		  {
		    s7_pointer args, p, func, e;
		    check_stack_size(sc);
		    func = opt_lambda(code);
		    new_frame(sc, closure_let(func), e);
		    sc->z = e;
		    for (p = closure_args(func), args = cdr(code); is_pair(p); p = cdr(p), args = cdr(args))
		      add_slot(e, car(p), find_symbol_unchecked(sc, car(args)));
		    sc->envir = e;
		    sc->z = sc->nil;
		    sc->code = car(closure_body(func));
		    goto EVAL;
		  }
		  
		case OP_CLOSURE_ALL_X:
		  check_stack_size(sc);
		  if (!closure_is_ok(sc, code, MATCH_UNSAFE_CLOSURE, integer(arglist_length(code)))) 
		    {
		      if (unknown_all_x_ex(sc, sc->last_function) == goto_OPT_EVAL)
			goto OPT_EVAL;		       
		      break;
		    }
		case HOP_CLOSURE_ALL_X:
		  {
		    s7_pointer args, p, func, e;
		    func = opt_lambda(code);
		    new_frame(sc, closure_let(func), e);
		    sc->z = e;
		    for (p = closure_args(func), args = cdr(code); is_pair(p); p = cdr(p), args = cdr(args))
		      {
			s7_pointer val;
			val = c_call(args)(sc, car(args));
			add_slot_checked(e, car(p), val); /* can't use add_slot here -- all_x_c_* hit trigger? */
		      }
		    sc->envir = e;
		    sc->z = sc->nil;
		    sc->code = _TPair(closure_body(func));
		    goto BEGIN1;
		  }
		  /* -------------------------------------------------------------------------------- */
		  
		case OP_CLOSURE_STAR_A:
		  if (!closure_star_is_ok(sc, code, MATCH_UNSAFE_CLOSURE_STAR, 1)) {if (unknown_a_ex(sc, sc->last_function) == goto_OPT_EVAL) goto OPT_EVAL; break;}
		case HOP_CLOSURE_STAR_A:
		  {
		    s7_pointer val, p;
		    val = c_call(cdr(code))(sc, cadr(code));
		    if (is_keyword(val))
		      s7_error(sc, sc->wrong_type_arg_symbol,
			       set_elist_4(sc, s7_make_string_wrapper(sc, "~A: keyword argument's value is missing: ~S in ~S"),
					   closure_name(sc, opt_lambda(code)), val, code));
		    sc->args = list_1(sc, val);

		    /* fill_closure_star(sc, cdr(closure_args(opt_lambda(code)))); */
		    p = cdr(closure_args(opt_lambda(code)));
		    for (; is_pair(p); p = cdr(p))
		      {
			if (is_pair(car(p)))
			  {
			    s7_pointer defval;
			    defval = cadar(p);
			    if (is_pair(defval))
			      sc->args = cons(sc, cadr(defval), sc->args);
			    else sc->args = cons(sc, defval, sc->args);
			  }
			else sc->args = cons(sc, sc->F, sc->args);
		      }
		    sc->args = safe_reverse_in_place(sc, sc->args);
		    sc->code = opt_lambda(sc->code);

		    /* unsafe_closure_star(sc) */
		    {
		      s7_pointer x, z, e;
		      uint64_t id;
		      
		      new_frame(sc, closure_let(sc->code), sc->envir);
		      e = sc->envir;
		      id = let_id(e);
		      
		      for (x = closure_args(sc->code), z = sc->args; is_pair(x); x = cdr(x))
			{
			  s7_pointer sym, args;
			  if (is_pair(car(x)))
			    sym = caar(x);
			  else sym = car(x);
			  args = cdr(z);
			  reuse_as_slot(z, sym, unchecked_car(z));
			  symbol_set_local(sym, id, z);
			  set_next_slot(z, let_slots(e));
			  let_set_slots(e, z);
			  z = args;
			}
		      sc->code = _TPair(closure_body(sc->code));
		    }
		    goto BEGIN1;
		  }

		case OP_CLOSURE_STAR_AA:
		  /* in the AA case closure_arity==2 and args=2 so we could simplify the all_x cases below */
		case OP_CLOSURE_STAR_ALL_X:
		  if (!closure_star_is_ok(sc, code, MATCH_UNSAFE_CLOSURE_STAR, integer(arglist_length(code))))
		    {
		      if (unknown_all_x_ex(sc, sc->last_function) == goto_OPT_EVAL)
			goto OPT_EVAL;		       
		      break;
		    }
		case HOP_CLOSURE_STAR_AA:
		case HOP_CLOSURE_STAR_ALL_X:
		  {
		    s7_pointer p, old_args;
		    sc->w = cdr(code);               /* args aren't evaluated yet */
		    sc->args = make_list(sc, integer(arglist_length(code)), sc->F);
		    for (p = sc->args, old_args = sc->w; is_pair(p); p = cdr(p), old_args = cdr(old_args))
		      set_car(p, c_call(old_args)(sc, car(old_args)));
		    sc->w = sc->nil;
		    sc->code = opt_lambda(code);
		    check_stack_size(sc);
		    sc->envir = new_frame_in_env(sc, closure_let(sc->code));
		    if (apply_lambda_star(sc) == goto_EVAL) goto EVAL;
		    goto BEGIN1;
		  }
		  
		  
		  /* -------------------------------------------------------------------------------- */
		case OP_UNKNOWN:
		case HOP_UNKNOWN:
		  if (unknown_ex(sc, find_symbol_checked(sc, car(code))) == goto_OPT_EVAL)
		    goto OPT_EVAL;
		  break;
		  
		case OP_UNKNOWN_G:
		case HOP_UNKNOWN_G:
		  if (unknown_g_ex(sc, find_symbol_checked(sc, car(code))) == goto_OPT_EVAL)
		    goto OPT_EVAL;
		  break;

		case OP_UNKNOWN_GG:
		case HOP_UNKNOWN_GG:
		  if (unknown_gg_ex(sc, find_symbol_checked(sc, car(code))) == goto_OPT_EVAL)
		    goto OPT_EVAL;
		  break;
		  
		case OP_UNKNOWN_ALL_S:
		case HOP_UNKNOWN_ALL_S:
		  if (unknown_all_s_ex(sc, find_symbol_checked(sc, car(code))) == goto_OPT_EVAL)
		    goto OPT_EVAL;
		  break;
		  
		case OP_UNKNOWN_A:
		case HOP_UNKNOWN_A:
		  if (unknown_a_ex(sc, find_symbol_checked(sc, car(code))) == goto_OPT_EVAL)
		    goto OPT_EVAL;
		  break;

		case OP_UNKNOWN_AA:
		case HOP_UNKNOWN_AA:
		  if (unknown_aa_ex(sc, find_symbol_checked(sc, car(code))) == goto_OPT_EVAL)
		    goto OPT_EVAL;
		  break;
		  
		case OP_UNKNOWN_ALL_X:
		case HOP_UNKNOWN_ALL_X:
		  if (unknown_all_x_ex(sc, find_symbol_checked(sc, car(code))) == goto_OPT_EVAL)
		    goto OPT_EVAL;
		  break;
		  /* -------------------------------------------------------------------------------- */

		  
		case OP_VECTOR_A:
		case HOP_VECTOR_A:
		  if (vector_a_ex(sc) == goto_START) goto START;
		  if (unknown_a_ex(sc, sc->last_function) == goto_OPT_EVAL) goto OPT_EVAL;
		  break;

		case OP_CVECTOR_A:
		case HOP_CVECTOR_A:
		  if (constant_vector_a_ex(sc) == goto_START) goto START;
		  if (unknown_a_ex(sc, sc->last_function) == goto_OPT_EVAL) goto OPT_EVAL;
		  break;

		case OP_STRING_A:
		case HOP_STRING_A:
		  if (string_a_ex(sc) == goto_START) goto START;
		  if (unknown_a_ex(sc, sc->last_function) == goto_OPT_EVAL) goto OPT_EVAL;
		  break;
		  
		case OP_HASH_TABLE_A:
		case HOP_HASH_TABLE_A:
		  {
		    s7_pointer s;
		    s = find_symbol_unchecked(sc, car(code));
		    if (!is_hash_table(s)) 
		      {
			if (unknown_a_ex(sc, s) == goto_OPT_EVAL) goto OPT_EVAL;
			break;
		      }
		    sc->value = s7_hash_table_ref(sc, s, c_call(cdr(code))(sc, cadr(code)));
		    goto START;
		  }
		  
		case OP_CONTINUATION_A:
		case HOP_CONTINUATION_A:
		  {
		    s7_pointer s;
		    s = find_symbol_unchecked(sc, car(code));
		    if (!is_continuation(s)) 
		      {
			if (unknown_a_ex(sc, s) == goto_OPT_EVAL) goto OPT_EVAL;
			break;
		      }
		    sc->code = s;
		    sc->args = set_plist_1(sc, c_call(cdr(code))(sc, cadr(code)));
		    apply_continuation(sc);
		    goto START;
		  }
		  
		case OP_ITERATE:
		case HOP_ITERATE:
		  {
		    s7_pointer s;
		    s = find_symbol_unchecked(sc, car(code));
		    if (!is_iterator(s)) 
		      {
			if (unknown_ex(sc, s) == goto_OPT_EVAL) goto OPT_EVAL;
			break;
		      }
		    sc->value = (iterator_next(s))(sc, s);
		    goto START;
		  }
		  
		case OP_ENVIRONMENT_Q:
		case HOP_ENVIRONMENT_Q:
		  {
		    s7_pointer s;
		    s = find_symbol_unchecked(sc, car(code));
		    if (!is_let(s)) 
		      {
			if (has_all_x(cdr(code)))
			  {
			    if (unknown_a_ex(sc, s) == goto_OPT_EVAL) goto OPT_EVAL;
			  }
			break;
		      }
		    sc->value = g_let_ref(sc, set_qlist_2(sc, s, cadadr(code)));
		    goto START;
		  }
		  
		case OP_ENVIRONMENT_A:
		case HOP_ENVIRONMENT_A:
		  {
		    s7_pointer s;
		    s = find_symbol_unchecked(sc, car(code));
		    if (!is_let(s)) 
		      {
			if (unknown_a_ex(sc, s) == goto_OPT_EVAL) goto OPT_EVAL;
			break;
		      }
		    sc->value = g_let_ref(sc, set_qlist_2(sc, s, c_call(cdr(code))(sc, cadr(code))));
		    goto START;
		  }
		  
		case OP_PAIR_A:
		case HOP_PAIR_A:
		  {
		    s7_pointer s, x;
		    s = find_symbol_unchecked(sc, car(code));
		    if (!is_pair(s)) 
		      {
			if (unknown_a_ex(sc, s) == goto_OPT_EVAL) goto OPT_EVAL;
			break;
		      }
		    x = c_call(cdr(code))(sc, cadr(code));
		    sc->value = list_ref_1(sc, s, x);
		    goto START;
		  }
		  
		case OP_C_OBJECT_A:
		case HOP_C_OBJECT_A:
		  {
		    s7_pointer c;
		    c = find_symbol_unchecked(sc, car(code));
		    if (!is_c_object(c)) 
		      {
			if (unknown_a_ex(sc, c) == goto_OPT_EVAL) goto OPT_EVAL;
			break;
		      }
		    set_car(sc->t1_1, c_call(cdr(code))(sc, cadr(code)));
		    sc->value = (*(c_object_ref(sc, c)))(sc, c, sc->t1_1);
		    goto START;
		  }
		  
		default:
		  fprintf(stderr, "bad op in opt_eval: op %u, is_opt: %d, %s\n", optimize_op(code), is_optimized(code), DISPLAY_80(code));
		  break;
		}
	      
	      /* else cancel all the optimization info -- someone stepped on our symbol */
	      /* there is a problem with this -- if the caller still insists on goto OPT_EVAL, for example,
	       *   we get here over and over.  (let ((x (list (car y))))...) where list is redefined away.
	       */
#if 0
	      if ((is_optimized(code)) &&
		  (is_symbol(car(code))))
		{
		  s7_pointer val;
		  val = find_symbol_unchecked(sc, car(code));
		  if (val)
		    fprintf(stderr, "opt trail: %s %s %s\n", opt_names[optimize_op(code)], type_name(sc, val, NO_ARTICLE), DISPLAY_80(code)); 
		}
#endif
	      /* gg/ss -> ss etc
	       * have a translation table, only c_c looks like trouble -- needs to be unwound
	       *
	       * some things are never opt'd?
	       *   (* (expt (polywave osc fm) k) norm) ??? this is tgen...
	       *   (fm-violin (* i 0.001) 0.01 440 0.001) tall
	       *   clm defins are using car(body)=string as documentation!
	       *   all list-values, define 
	       * lg:
	       *  trail: safe_c_aa c-function (fnc (car lst) (cdr lst))
	       *  trail: safe_closure_all_x function (f args form env)
	       */
	      clear_all_optimizations(sc, code);
	      /* and fall into the normal evaluator */
	    }
	  
	  {
	    s7_pointer code, carc;
	    code = sc->code;
	    if (is_pair(code))
	      {
#if WITH_PROFILE
		if (sc->code != profile_at_start)
		  profile(sc, code);
#endif
		/* fprintf(stderr, "trail: %s\n", DISPLAY_80(sc->code)); */

		set_current_code(sc, code);
		carc = car(code);
		
		if (typesflag(carc) == SYNTACTIC_TYPE)
		  {
		    set_syntactic_pair(code);      /* leave other bits (T_LINE_NUMBER) intact */
		    set_car(code, syntax_symbol(slot_value(initial_slot(carc)))); /* clear possible optimization confusion */
		    sc->cur_op = (opcode_t)symbol_syntax_op(carc);
		    pair_set_syntax_op(code, sc->cur_op);
		    sc->code = cdr(code);
		    goto START_WITHOUT_POP_STACK;
		  }
		
		/* -------------------------------------------------------------------------------- */
		/* trailers */

		if (is_symbol(carc))
		  {
		    /* car is a symbol, sc->code a list */
		    sc->value = find_global_symbol_checked(sc, carc);
		    sc->code = cdr(code);
		    /* drop into eval args */
		  }
		else
		  {
		    /* very uncommon case: car is either itself a pair or some non-symbol */
		    if (is_pair(carc))
		      {
			/* evaluate the inner list but that list can be circular: carc: #1=(#1# #1#)!
			 *   and the cycle can be well-hidden -- #1=((#1 2) . 2) and other such stuff
			 */
			if (sc->stack_end >= sc->stack_resize_trigger)
			  check_for_cyclic_code(sc, code);
			push_stack(sc, OP_EVAL_ARGS, sc->nil, cdr(code));
			if (typesflag(car(carc)) == SYNTACTIC_TYPE)
			  /* was checking for is_syntactic here but that can be confused by successive optimizer passes:
			   *  (define (hi) (((lambda () list)) 1 2 3)) etc
			   */
			  {
			    if ((car(carc) == sc->quote_symbol) &&        /* ('and #f) */
				((!is_pair(cdr(carc))) ||                 /* ((quote . #\h) (2 . #\i)) ! */
				 (is_syntactic(cadr(carc)))))
			      return(apply_error(sc, (is_pair(cdr(carc))) ? cadr(carc) : carc, cdr(code)));
			    sc->cur_op = (opcode_t)symbol_syntax_op(car(carc));
			    sc->code = cdr(carc);
			    goto START_WITHOUT_POP_STACK;
			  }
			
			push_stack(sc, OP_EVAL_ARGS, sc->nil, cdr(carc));
			sc->code = car(carc);
			goto EVAL;
		      }
		    else
		      {
			/* car must be the function to be applied */
			sc->value = _NFre(carc);
			sc->code = cdr(code);
			/* drop into OP_EVAL_ARGS */
		      }
		  }
	      }
	    else /* sc->code is not a pair */
	      {
		if (is_symbol(code))
		  sc->value = find_symbol_checked(sc, code);
		else sc->value = _NFre(code);
		goto START;
	      }
	  }

	  /* sc->value is car=something applicable
	   * sc->code = rest of expression
	   * sc->args is nil (set by the drop-through cases above -- perhaps clearer to bring that down?)
	   */
	case OP_EVAL_ARGS:
	  if (dont_eval_args(sc->value))
	    {
	      if (is_any_macro(sc->value))
		{
		  /* macro expansion */
		  sc->args = copy_list_with_arglist_error(sc, sc->code);
		  sc->code = sc->value;
		  goto APPLY;                      /* not UNSAFE_CLOSURE because it might be a bacro */
		}
	      /* (define progn begin) (progn (display "hi") (+ 1 23)) */
	      if (!is_syntax(sc->value))
		eval_error(sc, "attempt to evaluate: ~A?", sc->code);
	      sc->cur_op = (opcode_t)syntax_opcode(sc->value);
	      goto START_WITHOUT_POP_STACK;
	    }
	  
	  /* sc->value is the func
	   *   we don't have to delay lookup of the func because arg evaluation order is not specified, so
	   *     (let ((func +)) (func (let () (set! func -) 3) 2))
	   *   can return 5.
	   */
	  /* if (is_null(sc->code)) {sc->code = sc->value; goto APPLY;}
	   *   this is hit very rarely so it costs more than it saves
	   */
	  
	  push_op_stack(sc, sc->value);
	  if (sc->op_stack_now >= sc->op_stack_end)
	    resize_op_stack(sc);
	  
	  sc->args = sc->nil;
	  goto EVAL_ARGS;
	  /* moving eval_args up here (to avoid this goto) was slightly slower, probably by chance. */
	  
	case OP_EVAL_ARGS5:
	  /* sc->value is the last arg, sc->code is the previous */
	  {
	    s7_pointer x, y;
	    new_cell(sc, x, T_PAIR);
	    new_cell_no_check(sc, y, T_PAIR);
	    set_car(x, sc->code);
	    set_cdr(x, sc->args);
	    set_car(y, sc->value);
	    set_cdr(y, x);
	    sc->args = safe_reverse_in_place(sc, y);
	    sc->code = pop_op_stack(sc);
	    goto APPLY;
	  }
	  
	case OP_EVAL_ARGS2:          /* sc->value is the last arg, [so if is_null(cdr(sc->code) and current is pair, push args2] */
	  {
	    s7_pointer x;
	    sc->code = pop_op_stack(sc);
	    new_cell(sc, x, T_PAIR);
	    set_car(x, sc->value);
	    set_cdr(x, sc->args);
	    if (!is_null(sc->args))
	      sc->args = safe_reverse_in_place(sc, x);
	    else sc->args = x;
	    goto APPLY;
	  }
	  
	  /* tricky cases here all involve values (i.e. multiple-values) */
	case OP_EVAL_ARGS_P_2:
	  set_car(sc->t2_1, sc->args);
	  set_car(sc->t2_2, sc->value);
	  sc->value = c_call(sc->code)(sc, sc->t2_1);
	  break;
	  
	case OP_EVAL_ARGS_P_2_MV:
	  sc->args = cons(sc, sc->args, sc->value);
	  sc->code = c_function_base(opt_cfunc(sc->code));
	  goto APPLY;
	  
	case OP_EVAL_ARGS_P_3:
	  set_car(sc->t2_2, find_symbol_unchecked(sc, caddr(sc->code)));
	  /* we have to wait because we say the evaluation order is left to right (in lambda*)
	   *   and the first arg's evaluation might change the value of the second arg.
	   */
	  set_car(sc->t2_1, sc->value);
	  sc->value = c_call(sc->code)(sc, sc->t2_1);
	  break;
	  
	case OP_EVAL_ARGS_P_3_MV:      /* (define (hi a) (+ (values 1 2) a)) */
	  sc->w = sc->value;
	  sc->args = cons(sc, find_symbol_unchecked(sc, caddr(sc->code)), sc->w);
	  sc->code = c_function_base(opt_cfunc(sc->code));
	  goto APPLY;
	  
	case OP_EVAL_ARGS_P_4:
	  set_car(sc->t2_1, sc->value);
	  set_car(sc->t2_2, sc->args);
	  sc->value = c_call(sc->code)(sc, sc->t2_1);
	  break;
	  
	case OP_EVAL_ARGS_AAP_1:	  /* from OP_SAFE_C_AAP */
	  {
	    s7_pointer val1, val3;
	    val1 = sc->args;
	    val3 = sc->value;
	    set_car(sc->t3_2, c_call(cddr(sc->code))(sc, caddr(sc->code)));
	    set_car(sc->t3_1, val1);
	    set_car(sc->t3_3, val3);
	    sc->value = c_call(sc->code)(sc, sc->t3_1);
	    break;
	  }
	  
	case OP_EVAL_ARGS_AAP_MV:
	  {
	    s7_pointer val1, val2, val3;
	    val1 = sc->args;
	    val3 = sc->value;
	    val2 = c_call(cddr(sc->code))(sc, caddr(sc->code));
	    sc->args = cons(sc, val1, cons(sc, val2, val3));
	    sc->code = c_function_base(opt_cfunc(sc->code));
	    goto APPLY;
	  }

	case OP_SAFE_C_SZ_SZ:
	  /* S_opSZq actually, in (nominal second, only actual) SZ, S=args, Z=value,
	   *   SZ from the SP combiner for SZ
	   */
	  set_car(sc->t2_1, sc->args);
	  set_car(sc->t2_2, sc->value);
	  set_car(sc->t2_2, c_call(caddr(sc->code))(sc, sc->t2_1));
	  set_car(sc->t2_1, find_symbol_unchecked(sc, cadr(sc->code)));
	  sc->value = c_call(sc->code)(sc, sc->t2_1);
	  break;
	  
	case OP_SAFE_C_ZA_1:
	  set_car(sc->t2_2, c_call(cddr(sc->code))(sc, caddr(sc->code)));
	  set_car(sc->t2_1, sc->value);
	  sc->value = c_call(sc->code)(sc, sc->t2_1);
	  break;
	  
	case OP_SAFE_C_ZZ_1:
	  push_stack(sc, OP_EVAL_ARGS_P_2, sc->value, sc->code);
	  sc->code = _TPair(caddr(sc->code));
	  goto OPT_EVAL;
	  
	case OP_SAFE_C_ZAA_1:
	  {
	    s7_pointer val1, val2;
	    val1 = sc->value;
	    val2 = sc->args;
	    set_car(sc->a3_3, c_call(cdddr(sc->code))(sc, cadddr(sc->code)));
	    set_car(sc->a3_1, val1);
	    set_car(sc->a3_2, val2);
	    sc->value = c_call(sc->code)(sc, sc->a3_1);
	    break;
	  }
	  
	case OP_SAFE_C_AZA_1:
	  set_car(sc->t3_3, c_call(cdddr(sc->code))(sc, cadddr(sc->code)));
	  set_car(sc->t3_2, sc->value);
	  set_car(sc->t3_1, sc->args);
	  sc->value = c_call(sc->code)(sc, sc->t3_1);
	  break;
	  
	case OP_SAFE_C_AAZ_1:
	  set_car(sc->t3_1, pop_op_stack(sc));
	  set_car(sc->t3_2, sc->args);
	  set_car(sc->t3_3, sc->value);
	  sc->value = c_call(sc->code)(sc, sc->t3_1);
	  break;
	  
	case OP_SAFE_C_ZZA_1:
	  push_op_stack(sc, sc->value);
	  push_stack(sc, OP_SAFE_C_ZZA_2, sc->args, sc->code);
	  sc->code = _TPair(caddr(sc->code));
	  goto OPT_EVAL_CHECKED;
	  
	case OP_SAFE_C_ZZA_2:
	  set_car(sc->a3_1, pop_op_stack(sc));
	  set_car(sc->a3_2, sc->value);
	  set_car(sc->a3_3, c_call(cdddr(sc->code))(sc, cadddr(sc->code)));
	  sc->value = c_call(sc->code)(sc, sc->a3_1);
	  break;
	  
	case OP_SAFE_C_ZAZ_1:
	  {
	    s7_pointer val;
	    push_op_stack(sc, sc->value);
	    val = c_call(cddr(sc->code))(sc, caddr(sc->code));
	    push_stack(sc, OP_SAFE_C_AAZ_1, val,  sc->code);
	    sc->code = _TPair(cadddr(sc->code));
	    goto OPT_EVAL_CHECKED;
	  }

	case OP_SAFE_C_AZZ_1:
	  push_op_stack(sc, sc->value);
	  push_stack(sc, OP_SAFE_C_AZZ_2, sc->args, sc->code);
	  sc->code = _TPair(cadddr(sc->code));
	  goto OPT_EVAL_CHECKED;
	  
	case OP_SAFE_C_AZZ_2:
	  set_car(sc->t3_1, sc->args);
	  set_car(sc->t3_2, pop_op_stack(sc));
	  set_car(sc->t3_3, sc->value);
	  sc->value = c_call(sc->code)(sc, sc->t3_1);
	  break;
	  
	case OP_SAFE_C_ZZZ_1:
	  push_stack(sc, OP_SAFE_C_ZZZ_2, sc->value, sc->code);
	  sc->code = opt_pair2(cdr(sc->code));
	  goto OPT_EVAL_CHECKED;
	  
	case OP_SAFE_C_ZZZ_2:
	  push_op_stack(sc, sc->value);
	  push_stack(sc, OP_SAFE_C_AZZ_2, sc->args, sc->code);
	  sc->code = opt_pair1(cdr(sc->code));
	  goto OPT_EVAL_CHECKED;
	  
	case OP_EVAL_ARGS3:  /* sc->value is the next-to-last arg, and we know the last arg is not a list (so values can't mess us up!) */
	  {
	    s7_pointer x, y, val;
	    
	    val = sc->code;
	    if (is_symbol(val))
	      val = find_symbol_checked(sc, val);
	    
	    new_cell(sc, x, T_PAIR);
	    new_cell_no_check(sc, y, T_PAIR);
	    set_car(x, sc->value);
	    set_cdr(x, sc->args);
	    set_car(y, val);
	    set_cdr(y, x);
	    sc->args = safe_reverse_in_place(sc, y);
	    sc->code = pop_op_stack(sc);
	    goto APPLY;
	  }
	  
	case OP_EVAL_ARGS4:
	  /* sc->code is a pair, and either cdr(sc->code) is not null or car(sc->code) is a pair
	   *
	   * (#f #f) (env #f) etc.  args is very often nil here, so we're looking at 3 simple args
	   *   or even just 2 in some cases: (+ req opt) with value 2 and args ()
	   */
	  {
	    s7_pointer x;
	    new_cell(sc, x, T_PAIR);
	    set_car(x, sc->value);
	    set_cdr(x, sc->args);
	    sc->args = x;          /* all the others reverse -- why not this case? -- reverse is at end? (below) */
	    goto EVAL_ARGS_PAIR;
	  }
	  
	case OP_EVAL_ARGS1:
	  {
	    s7_pointer x;
	    new_cell(sc, x, T_PAIR);
	    set_car(x, sc->value);
	    set_cdr(x, sc->args);
	    sc->args = x;
	  }
	  
	EVAL_ARGS:
	  /* first time, value = op, args = nil, code is args */
	  if (is_pair(sc->code))  /* evaluate current arg -- must check for pair here, not sc->nil (improper list as args) */
	    {
	      s7_pointer car_code;
	      
	    EVAL_ARGS_PAIR:
	      car_code = car(sc->code);
	      
	      /* switch statement here is much slower for some reason */
	      if (is_pair(car_code))
		{
		  if (sc->stack_end >= sc->stack_resize_trigger)
		    check_for_cyclic_code(sc, sc->code);

		  /* all 3 of these push_stacks can result in stack overflow, see above 64065 */
		  if (is_null(cdr(sc->code)))
		    push_stack(sc, OP_EVAL_ARGS2, sc->args, sc->nil);
		  else
		    {
		      if (!is_pair(cdr(sc->code)))            /* (= 0 '(1 . 2) . 3) */
			improper_arglist_error(sc);
		      
		      if ((is_null(cddr(sc->code))) &&
			  (!is_pair(cadr(sc->code))))
			push_stack(sc, OP_EVAL_ARGS3, sc->args, cadr(sc->code));
		      else push_stack(sc, OP_EVAL_ARGS4, sc->args, cdr(sc->code));
		    }
		  sc->code = car_code;
		  if (is_optimized(sc->code))
		    goto OPT_EVAL;
		  goto EVAL;
		}
	      
	      /* car(sc->code) is not a pair */
	      if (is_pair(cdr(sc->code)))
		{
		  sc->code = cdr(sc->code);
		  if (is_symbol(car_code))
		    sc->value = find_symbol_checked(sc, car_code);
		  else sc->value = _NFre(car_code);
		  /* sc->value is the current arg's value, sc->code is pointing to the next */
		  
		  /* cdr(sc->code) may not be a pair or nil here!
		   *   (eq? #f . 1) -> sc->code is 1
		   */
		  if (is_null(cdr(sc->code)))
		    {
		      s7_pointer x, y, val;
		      /* we're at the last arg, sc->value is the previous one, not yet saved in the args list */
		      car_code = car(sc->code);
		      if (is_pair(car_code))
			{
			  if (sc->stack_end >= sc->stack_resize_trigger)
			    check_for_cyclic_code(sc, sc->code);
			  push_stack(sc, OP_EVAL_ARGS5, sc->args, sc->value);
			  sc->code = car_code;
			  goto EVAL;
			}
		      
		      /* get the last arg */
		      if (is_symbol(car_code))
			val = find_symbol_checked(sc, car_code);
		      else val = car_code;
		      sc->temp4 = val;
		      
		      /* get the current arg, which is not a list */
		      sc->code = pop_op_stack(sc);
		      new_cell(sc, x, T_PAIR);
		      new_cell_no_check(sc, y, T_PAIR);
		      set_car(x, sc->value);
		      set_cdr(x, sc->args);
		      set_car(y, val);
		      set_cdr(y, x);
		      sc->args = safe_reverse_in_place(sc, y);
		      /* drop into APPLY */
		    }
		  else
		    {
		      /* here we know sc->code is a pair, cdr(sc->code) is not null
		       *   sc->value is the previous arg's value
		       */
		      s7_pointer x;
		      new_cell(sc, x, T_PAIR);
		      set_car(x, sc->value);
		      set_cdr(x, sc->args);
		      sc->args = x;
		      goto EVAL_ARGS_PAIR;
		    }
		}
	      else
		{
		  /* here we've reached the last arg (sc->code == nil), it is not a pair */
		  s7_pointer x, val;
		  
		  if (!is_null(cdr(sc->code)))
		    improper_arglist_error(sc);
		  
		  sc->code = pop_op_stack(sc);
		  if (is_symbol(car_code))
		    val = find_symbol_checked(sc, car_code); /* this has to precede the set_type below */
		  else val = car_code;
		  sc->temp4 = val;
		  new_cell(sc, x, T_PAIR);
		  set_car(x, val);
		  set_cdr(x, sc->args);
		  
		  if (!is_null(sc->args))
		    sc->args = safe_reverse_in_place(sc, x);
		  else sc->args = x;
		  /* drop into APPLY */
		}
	    }
	  else                       /* got all args -- go to apply */
	    {
	      if (is_not_null(sc->code))
		improper_arglist_error(sc);
	      else
		{
		  sc->code = pop_op_stack(sc);
		  sc->args = safe_reverse_in_place(sc, sc->args);
		  /* we could omit the arg reversal in many cases, but lots of code assumes the args are in order;
		   *   adding a bit for this in the type field saves some time in s7test (many + and * tests), but costs
		   *   about the same time in other cases, so it's not a clear win.
		   */
		}
	    }
	  
	  /* turning this into a call on an array of functions was not a complete disaster, but tauto.scm was ~1.5% slower.
	   *   the array-index overhead is the same as the current switch statement's, but there was also the boolean+jump overhead,
	   *   and the function-local overhead currently otherwise 0 (I assume because the compiler can simply plug it in here).
	   */

	APPLY:
	case OP_APPLY:
	  /* fprintf(stderr, "apply %s to %s\n", DISPLAY(sc->code), DISPLAY(sc->args)); */
	  switch (type(sc->code))
	    {
	    case T_C_FUNCTION:          apply_c_function(sc);           goto START;
	    case T_C_ANY_ARGS_FUNCTION: apply_c_any_args_function(sc);  goto START;
	    case T_C_FUNCTION_STAR:     apply_c_function_star(sc);      goto START;
	    case T_C_OPT_ARGS_FUNCTION: apply_c_opt_args_function(sc);  goto START;
	    case T_C_RST_ARGS_FUNCTION: apply_c_rst_args_function(sc);  goto START;
	    case T_C_MACRO:  	        apply_c_macro(sc);	        goto EVAL;
	    case T_CONTINUATION:        apply_continuation(sc);         goto START;
	    case T_GOTO:	        call_with_exit(sc);	        goto START;
	    case T_C_OBJECT:	        apply_c_object(sc);	        goto START;
	    case T_INT_VECTOR:
	    case T_FLOAT_VECTOR:
	    case T_VECTOR: 	        apply_vector(sc);	        goto START;
	    case T_STRING:	        apply_string(sc);	        goto START;
	    case T_HASH_TABLE:	        apply_hash_table(sc);           goto START;
	    case T_ITERATOR:	        apply_iterator(sc);	        goto START;	      
	    case T_LET:	                apply_let(sc);	                goto START;
	    case T_SYNTAX:	        apply_syntax(sc);	        goto START_WITHOUT_POP_STACK;
	    case T_PAIR:	        
	      if (apply_pair(sc) == goto_APPLY) goto APPLY; 
	      goto START;

	    case T_MACRO:
	      /* this is not from the reader, so treat expansions here as normal macros */
	      push_stack_op_let(sc, OP_EVAL_MACRO);
	      new_frame(sc, closure_let(sc->code), sc->envir);
	      goto APPLY_LAMBDA;
	      
	    case T_BACRO: 
	      push_stack_op_let(sc, OP_EVAL_MACRO);
	      new_frame(sc, sc->envir, sc->envir);       /* like let* -- we'll be adding macro args, so might as well sequester things here */
	      goto APPLY_LAMBDA;
	      
	    case T_CLOSURE:
	      check_stack_size(sc);
	      new_frame(sc, closure_let(sc->code), sc->envir);
	      goto APPLY_LAMBDA;
	      
	    case T_MACRO_STAR:
	      push_stack_op_let(sc, OP_EVAL_MACRO);
	      new_frame(sc, closure_let(sc->code), sc->envir);
	      if (apply_lambda_star(sc) == goto_EVAL) goto EVAL;
	      goto BEGIN1;
	      
	    case T_BACRO_STAR:
	      push_stack_op_let(sc, OP_EVAL_MACRO);
	      new_frame(sc, sc->envir, sc->envir);
	      if (apply_lambda_star(sc) == goto_EVAL) goto EVAL;
	      goto BEGIN1;
	      
	    case T_CLOSURE_STAR:
	      check_stack_size(sc);
	      sc->envir = new_frame_in_env(sc, closure_let(sc->code));
	      if (apply_lambda_star(sc) == goto_EVAL) goto EVAL;
	      goto BEGIN1;

	    default:
	      return(apply_error(sc, sc->code, sc->args));
	    }

	APPLY_LAMBDA:
	case OP_APPLY_LAMBDA:
	  apply_lambda(sc);
	  goto BEGIN1;

	case OP_LAMBDA_STAR_DEFAULT:
	  /* sc->args is the current closure arg list position, sc->value is the default expression's value */
	  slot_set_value(sc->args, sc->value);
	  sc->args = slot_pending_value(sc->args);
	  if (lambda_star_default(sc) == goto_EVAL) goto EVAL;
	  pop_stack_no_op(sc);
	  sc->code = _TPair(closure_body(sc->code));
	  goto BEGIN1;
	  
	case OP_MACROEXPAND_1:
	  sc->args = _TLst(cdar(sc->code));
	  sc->code = sc->value;
	  goto MACROEXPAND;
	  
	case OP_MACROEXPAND:
	  /* mimic APPLY above, but don't push OP_EVAL_MACRO or OP_EXPANSION
	   *   (define-macro (mac a) `(+ ,a 1)) (macroexpand (mac 3)), sc->code: ((mac 3))
	   */
	  if ((!is_pair(sc->code)) ||
	      (!is_pair(car(sc->code))))
	    eval_error(sc, "macroexpand argument is not a macro call: ~A", sc->code);
	  if (!is_null(cdr(sc->code)))
	    eval_error(sc, "macroexpand: too many arguments: ~A", sc->code);
	  if (!s7_is_proper_list(sc, car(sc->code)))
	    eval_error(sc, "macroexpand: improper arg list: ~A", car(sc->code));
	  /* this seems to happen when it should not? */
	  
	  if (is_pair(caar(sc->code)))                            /* (macroexpand ((symbol->value 'mac) (+ 1 2))) */
	    {
	      push_stack_no_args(sc, OP_MACROEXPAND_1, sc->code);
	      sc->code = caar(sc->code);
	      goto EVAL;
	    }

	  sc->args = _TLst(cdar(sc->code));
	  if (!is_symbol(caar(sc->code)))
	    eval_error(sc, "macroexpand argument is not a macro call: ~A", sc->code);
	  sc->code = find_symbol_checked(sc, caar(sc->code));
	  
	MACROEXPAND:
	  switch (type(sc->code))
	    {
	    case T_MACRO:
	      new_frame(sc, closure_let(sc->code), sc->envir);
	      goto APPLY_LAMBDA;
	      
	    case T_BACRO:
	      new_frame(sc, sc->envir, sc->envir);
	      goto APPLY_LAMBDA;
	      
	    case T_MACRO_STAR:
	      new_frame(sc, closure_let(sc->code), sc->envir);
	      if (apply_lambda_star(sc) == goto_EVAL) goto EVAL;
	      goto BEGIN1;
	      
	    case T_BACRO_STAR:
	      new_frame(sc, sc->envir, sc->envir);
	      if (apply_lambda_star(sc) == goto_EVAL) goto EVAL;
	      goto BEGIN1;
	      
	    case T_C_MACRO:
	      sc->value = c_macro_call(sc->code)(sc, sc->args);
	      goto START;
	    }
	  eval_error(sc, "macroexpand argument is not a macro call: ~A", sc->args);
	  
	case OP_QUOTE:
	case OP_QUOTE_UNCHECKED:
	  /* I think a quoted list in another list can be applied to a function, come here and
	   *   be changed to unchecked, set-cdr! or something clobbers the argument so we get
	   *   here on the next time around with the equivalent of (quote . 0) if unchecked
	   * so set-cdr! of constant -- if marked immutable, we could catch this case and clear.
	   */
	  check_quote(sc, sc->code);
	  sc->value = car(sc->code);
	  break;
	  
	case OP_DEFINE_FUNCHECKED:
	  define_funchecked(sc);
	  break;
	  
	case OP_DEFINE_CONSTANT1:
	  sc->code = car(sc->code);
	  if (is_pair(sc->code)) sc->code = car(sc->code); /* (define-constant (ex3 a)...) */
	  if (is_symbol(sc->code))
	    set_immutable(sc->code);
	  break;
	  
	case OP_DEFINE_CONSTANT:
	  if ((!is_pair(sc->code)) || (!is_pair(cdr(sc->code)))) /* (define-constant) */
	    eval_error(sc, "define-constant: not enough arguments: ~S", sc->code);

	  if ((is_symbol(car(sc->code))) &&                /* (define-constant abs abs): "abs will not be touched" */
	      (car(sc->code) == cadr(sc->code)) &&
	      (symbol_id(car(sc->code)) == 0) &&           /* else (let iter ... (define-constant iter iter) ...) -> segfault on later calls */
	      (is_null(cddr(sc->code))))
	    {
	      set_immutable(car(sc->code));
	      sc->value = find_symbol_checked(sc, car(sc->code));
	      goto START;
	    }
	  push_stack_no_args(sc, OP_DEFINE_CONSTANT1, sc->code);
	  
	case OP_DEFINE_STAR:
	case OP_DEFINE:
	  check_define(sc);
	  
	case OP_DEFINE_CONSTANT_UNCHECKED:
	case OP_DEFINE_STAR_UNCHECKED:
	case OP_DEFINE_UNCHECKED:
	  if (define_unchecked_ex(sc) == goto_EVAL) goto EVAL;

	case OP_DEFINE1:
	  if (define1_ex(sc) == goto_APPLY) goto APPLY;
	  
	case OP_DEFINE_WITH_ACCESSOR:
	  define2_ex(sc);
	  break;
	  

	case OP_EVAL_STRING_2:
	  s7_close_input_port(sc, sc->input_port);
	  pop_input_port(sc);
	  if (is_multiple_value(sc->value))
	    sc->value = splice_in_values(sc, multiple_value(sc->value));
	  break;
	  
	case OP_EVAL_STRING_1:
	  eval_string_1_ex(sc);
	  goto EVAL;
	  
	  	  
	  /* -------------------------------- SET! -------------------------------- */
	case OP_SET_PAIR_P:
	  /* ([set!] (car a) (cadr a)) */
	  /* here the pair can't generate multiple values, or if it does, it's an error (caught below)
	   *  splice_in_values will notice the OP_SET_PAIR_P_1 and complain.
	   * (let () (define (hi) (let ((str "123")) (set! (str 0) (values #\a)) str)) (hi) (hi)) is "a23"
	   * (let () (define (hi) (let ((str "123")) (set! (str 0) (values #\a #\b)) str)) (hi) (hi)) is an error from the first call (caught elsewhere)
	   * (let () (define (hi) (let ((str "123")) (set! (str 0) (values #\a #\b)) str)) (catch #t hi (lambda a a)) (hi)) is an error from the second call
	   * (let ((v (make-vector '(2 3) 0))) (set! (v (values 0 1)) 23) v) -> #2D((0 23 0) (0 0 0))
	   */
	  push_stack_no_args(sc, OP_SET_PAIR_P_1, sc->code);
	  sc->code = cadr(sc->code);
	  goto EVAL;
	  
	case OP_SET_PAIR_Z:
	  push_stack_no_args(sc, OP_SET_PAIR_P_1, sc->code);
	  sc->code = _TPair(cadr(sc->code));
	  goto OPT_EVAL_CHECKED;
	  
	case OP_SET_PAIR_A:
	  {
	    s7_pointer obj, val;
	    obj = find_symbol_checked(sc, caar(sc->code));
	    val = c_call(cdr(sc->code))(sc, cadr(sc->code)); /* this call can step on sc->tx_x */
	    set_car(sc->t2_1, cadar(sc->code));              /* might be a constant: (set! (mus-sound-srate "oboe.snd") 12345) */
	    if (is_symbol(car(sc->t2_1)))
	      set_car(sc->t2_1, find_symbol_checked(sc, cadar(sc->code)));
	    set_car(sc->t2_2, val);
	    sc->value = c_function_call(c_function_setter(obj))(sc, sc->t2_1);
	  }
	  break;
	  
	case OP_SET_LET_S:       /* (set! (*s7* 'print-length) i) */
	  if (set_pair_p_3(sc, find_symbol(sc, caar(sc->code)), cadr(cadar(sc->code)), find_symbol_unchecked(sc, cadr(sc->code))))
	    goto APPLY;
	  break;
	    
	case OP_SET_LET_ALL_X:  /* (set! (hook 'result) 123) or (set! (H 'c) 32) */
	  if (set_pair_p_3(sc, find_symbol(sc, caar(sc->code)), cadr(cadar(sc->code)), c_call(cdr(sc->code))(sc, cadr(sc->code))))
	    goto APPLY;
	  break;
	    
	case OP_SET_PAIR_ZA:    /* unknown setter pair, but value is easy */
	  sc->value = c_call(cdr(sc->code))(sc, cadr(sc->code));
	  /* fall through */
	    
	case OP_SET_PAIR_P_1:
	  {
	    /* car(sc->code) is a pair, caar(code) is the object with a setter, it has one (safe) argument, and one safe value to set
	     *   (set! (str i) #\a) in a function (both inner things need to be symbols (or the second can be a quoted symbol) to get here)
	     *   the inner list is a proper list, with no embedded list at car.
	     */
	    s7_pointer arg, value;
	    value = sc->value;
	    arg = cadar(sc->code);
	    if (is_symbol(arg))
	      arg = find_symbol_checked(sc, arg);
	    else
	      {
		if (is_pair(arg))
		  arg = cadr(arg); /* can only be (quote ...) in this case */
	      }
	    if (set_pair_p_3(sc, find_symbol(sc, caar(sc->code)), arg, value))
	      goto APPLY;
	  }
	  break;
	    
	case OP_SET_DILAMBDA_Z:
	  push_stack_no_args(sc, OP_SET_DILAMBDA_Z_1, sc->code);
	  sc->code = _TPair(cadr(sc->code));
	  goto OPT_EVAL_CHECKED;

	case OP_SET_DILAMBDA_Z_1:
	  {
	    s7_pointer obj, func, arg, value;
	    value = sc->value;
	    
	    arg = cadar(sc->code);
	    if (is_symbol(arg))
	      arg = find_symbol_checked(sc, arg);
	    else
	      {
		if (is_pair(arg))
		  arg = cadr(arg); /* can only be (quote ...) in this case */
	      }
	    obj = find_symbol(sc, caar(sc->code));
	    func = slot_value(obj);
	    if ((is_closure(func)) &&
		(is_safe_closure(closure_setter(func))))
	      {
		sc->code = closure_setter(func);
		sc->envir = old_frame_with_two_slots(sc, closure_let(sc->code), arg, value);
		sc->code = _TPair(closure_body(sc->code));
		goto BEGIN1;
	      }
	    /* fallback on set-pair */
	    if (set_pair_p_3(sc, obj, arg, value))
	      goto APPLY;
	  }
	  break;

	case OP_SET_DILAMBDA:
	  {
	    /* ([set!] (dilambda-setter g) s) */
	    s7_pointer obj, func, arg, value;
	    value = cadr(sc->code);
	    if (is_symbol(value))
	      value = find_symbol_checked(sc, value);
	    
	    arg = cadar(sc->code);
	    if (is_symbol(arg))
	      arg = find_symbol_checked(sc, arg);
	    else
	      {
		if (is_pair(arg))
		  arg = cadr(arg); /* can only be (quote ...) in this case */
	      }
	    obj = find_symbol(sc, caar(sc->code));
	    func = slot_value(obj);
	    if ((is_closure(func)) &&
		(is_safe_closure(closure_setter(func))))
	      {
		sc->code = closure_setter(func);
		sc->envir = old_frame_with_two_slots(sc, closure_let(sc->code), arg, value);
		sc->code = _TPair(closure_body(sc->code));
		goto BEGIN1;
	      }
	    /* fallback on set-pair */
	    if (set_pair_p_3(sc, obj, arg, value))
	      goto APPLY;
	  }
	  break;
	    
	case OP_SET_PAIR:
	  {
	    /* ([set!] (procedure-setter g) s) or ([set!] (str 0) #\a) */
	    s7_pointer obj, arg, value;
	    value = cadr(sc->code);
	    if (is_symbol(value))
	      value = find_symbol_checked(sc, value);
	    
	    arg = cadar(sc->code);
	    if (is_symbol(arg))
	      arg = find_symbol_checked(sc, arg);
	    else
	      {
		if (is_pair(arg))
		  arg = cadr(arg); /* can only be (quote ...) in this case */
	      }
	    obj = caar(sc->code);
	    if (is_symbol(obj))
	      obj = find_symbol(sc, obj);
	    if (set_pair_p_3(sc, obj, arg, value))
	      goto APPLY;
	  }
	  break;
	  
	  /* this is (set! (getter) val) where getter is a global c_function (a built-in pws) and val is not a pair */
	case OP_SET_PWS:       /* (set! (mus-clipping) #f) */
	  set_pws_ex(sc);
	  break;
	  
	case OP_INCREMENT_1:
	  increment_1_ex(sc);
	  break;
	  
	case OP_DECREMENT_1:
	  decrement_1_ex(sc);
	  break;
	  
        #define SET_CASE(Op, Code)					\
	  case Op:							\
	    {								\
	      s7_pointer lx;						\
	      lx = find_symbol(sc, _TSet(car(sc->code)));		\
	      if (!is_slot(lx)) eval_type_error(sc, "set! ~A: unbound variable", sc->code); \
	      Code;							\
	      sc->value = slot_value(lx);				\
	      goto START;						\
	    }
	  
	    SET_CASE(OP_SET_SYMBOL_C, slot_set_value(lx, cadr(sc->code)))
	    
	    SET_CASE(OP_SET_SYMBOL_Q, slot_set_value(lx, cadadr(sc->code)))
	    
	    SET_CASE(OP_SET_SYMBOL_A, slot_set_value(lx, c_call(cdr(sc->code))(sc, cadr(sc->code))))
	    
	    SET_CASE(OP_SET_SYMBOL_L, slot_set_value(lx, local_symbol_value(cadr(sc->code))))
	    SET_CASE(OP_SET_SYMBOL_S, slot_set_value(lx, find_symbol_unchecked(sc, cadr(sc->code))))
	    
	    SET_CASE(OP_SET_CONS, slot_set_value(lx, cons(sc, find_symbol_unchecked(sc, opt_sym2(sc->code)), slot_value(lx))))  /* ([set!] bindings (cons v bindings)) */
	    
	    SET_CASE(OP_SET_SYMBOL_opCq, slot_set_value(lx, c_call(cadr(sc->code))(sc, opt_pair2(sc->code))))
	    
	    /* here we know the symbols do not have accessors, at least at optimization time */
	    SET_CASE(OP_SET_SYMBOL_opSq,
		     do {						\
		       set_car(sc->t1_1, find_symbol_unchecked(sc, opt_sym2(sc->code))); \
		       slot_set_value(lx, c_call(cadr(sc->code))(sc, sc->t1_1)); \
		     } while (0))
	    SET_CASE(OP_SET_SYMBOL_opLq,
		     do {						\
		       set_car(sc->t1_1, local_symbol_value(opt_sym2(sc->code))); \
		       slot_set_value(lx, c_call(cadr(sc->code))(sc, sc->t1_1)); \
		     } while (0))
	    
	    SET_CASE(OP_SET_SYMBOL_opSSq,
		     do {						\
		       set_car(sc->t2_1, find_symbol_unchecked(sc, car(opt_pair2(sc->code)))); \
		       set_car(sc->t2_2, find_symbol_unchecked(sc, cadr(opt_pair2(sc->code)))); \
		       slot_set_value(lx, c_call(cadr(sc->code))(sc, sc->t2_1)); \
		     } while (0))
	    SET_CASE(OP_SET_SYMBOL_opLLq,
		     do {						\
		       set_car(sc->t2_1, local_symbol_value(car(opt_pair2(sc->code)))); \
		       set_car(sc->t2_2, local_symbol_value(cadr(opt_pair2(sc->code)))); \
		       slot_set_value(lx, c_call(cadr(sc->code))(sc, sc->t2_1)); \
		     } while (0))
	    
	    SET_CASE(OP_SET_SYMBOL_opSSSq,
		     do {						\
		       set_car(sc->t3_1, find_symbol_unchecked(sc, car(opt_pair2(sc->code)))); \
		       set_car(sc->t3_2, find_symbol_unchecked(sc, opt_sym1(opt_pair2(sc->code)))); \
		       set_car(sc->t3_3, find_symbol_unchecked(sc, opt_sym2(opt_pair2(sc->code)))); \
		       slot_set_value(lx, c_call(cadr(sc->code))(sc, sc->t3_1)); \
		     } while (0))
	    
	    SET_CASE(OP_INCREMENT_SS,                                       /* ([set!] x (+ x i)) */
		     do {						\
		       set_car(sc->t2_1, slot_value(lx));		\
		       set_car(sc->t2_2, find_symbol_unchecked(sc, cadr(opt_pair2(sc->code)))); \
		       slot_set_value(lx, c_call(cadr(sc->code))(sc, sc->t2_1)); \
		     } while (0))
	    
	    SET_CASE(OP_INCREMENT_SSS,                                      /* ([set!] x (+ x y z)) -- nearly always involves reals */
		     do {						\
		       s7_pointer x1; s7_pointer x2; s7_pointer x3;	\
		       x1 = slot_value(lx);				\
		       x2 = find_symbol_unchecked(sc, opt_sym1(opt_pair2(sc->code))); \
		       x3 = find_symbol_unchecked(sc, opt_sym2(opt_pair2(sc->code))); \
		       if ((is_t_real(x1)) && (is_t_real(x2)) && (is_t_real(x3))) \
			 slot_set_value(lx, make_real(sc, real(x1) + real(x2) + real(x3))); \
		       else {						\
		         set_car(sc->t3_1, x1); set_car(sc->t3_2, x2); set_car(sc->t3_3, x3); \
			 slot_set_value(lx, global_add(sc, sc->t3_1));	\
		       }						\
		     } while (0))
	    
	    SET_CASE(OP_INCREMENT_SA,
		     do {						\
		       s7_pointer arg;					\
		       arg = opt_pair2(sc->code);				\
		       set_car(sc->t2_2, c_call(arg)(sc, car(arg)));	\
		       set_car(sc->t2_1, slot_value(lx));		\
		       slot_set_value(lx, c_call(cadr(sc->code))(sc, sc->t2_1)); \
		     } while (0))
	    
	    SET_CASE(OP_INCREMENT_SAA,                                      /* (set! sum (+ sum (expt k i) (expt (- k) i))) -- oops */
		     do {					\
		       s7_pointer arg;					\
		       arg = opt_pair2(sc->code); /* cddr(value) */		\
		       set_car(sc->a3_3, c_call(cdr(arg))(sc, cadr(arg))); \
		       set_car(sc->a3_2, c_call(arg)(sc, car(arg)));	\
		       set_car(sc->a3_1, slot_value(lx));		\
		       slot_set_value(lx, c_call(cadr(sc->code))(sc, sc->a3_1)); \
		     } while (0))
	    
	case OP_SET_SAFE:
	  {
	    s7_pointer lx;
	    lx = find_symbol(sc, _TSet(sc->code)); /* SET_CASE above looks for car(sc->code) */
	    if (!is_slot(lx)) eval_type_error(sc, "set! ~A: unbound variable", sc->code);
	    slot_set_value(lx, sc->value);
	    sc->value = slot_value(lx);
	  }
	  break;
	  
	case OP_SET_SYMBOL_P:      /* ([set!] f (lambda () 1)) */
	  push_stack_no_args(sc, OP_SET_SAFE, car(sc->code));
	  sc->code = cadr(sc->code);
	  goto EVAL;
	  
	case OP_SET_SYMBOL_Z:
	  /* ([set!] sum (+ sum n)) */
	  push_stack_no_args(sc, OP_SET_SAFE, car(sc->code));
	  sc->code = _TPair(cadr(sc->code));
	  goto OPT_EVAL_CHECKED;
	  
	case OP_INCREMENT_SZ:
	  {
	    s7_pointer sym;
	    sym = find_symbol(sc, car(sc->code));
	    if (is_slot(sym))
	      {
		push_stack(sc, OP_INCREMENT_SZ_1, sym, sc->code);
		sc->code = _TPair(opt_pair2(sc->code)); /* caddr(cadr(sc->code)); */
		goto OPT_EVAL_CHECKED;
	      }
	    eval_type_error(sc, "set! ~A: unbound variable", sc->code);
	  }
	  
	case OP_INCREMENT_SZ_1:
	  set_car(sc->t2_1, slot_value(sc->args));
	  set_car(sc->t2_2, sc->value);
	  sc->value = c_call(cadr(sc->code))(sc, sc->t2_1);
	  slot_set_value(sc->args, sc->value);
	  break;
	  
	case OP_SET2:
	  if (is_pair(sc->value))
	    {
	      /* (let ((L '((1 2 3)))) (set! ((L 0) 1) 32) L)
	       * (let ((L '(((1 2 3))))) (set! ((L 0) 0 1) 32) L)
	       * any deeper nesting was handled already by the first eval
	       *   set! looks at its first argument, if it's a symbol, it sets the associated value,
	       *   if it's a list, it looks at the car of that list to decide which setter to call,
	       *   if it's a list of lists, it passes the embedded lists to eval, then looks at the
	       *   car of the result.  This means that we can do crazy things like:
	       *   (let ((x '(1)) (y '(2))) (set! ((if #t x y) 0) 32) x)
	       *
	       * the other args need to be evaluated (but not the list as if it were code):
	       *   (let ((L '((1 2 3))) (index 1)) (set! ((L 0) index) 32) L)
	       */
	      
	      if (!s7_is_proper_list(sc, sc->args))                              /* (set! ('(1 2) 1 . 2) 1) */
		eval_error(sc, "set! target arguments are an improper list: ~A", sc->args);
	      
	      /* in all of these cases, we might need to GC protect the temporary lists */
	      
	      if (is_multiple_value(sc->value))
		sc->code = cons(sc, sc->set_symbol, s7_append(sc, multiple_value(sc->value), s7_append(sc, sc->args, sc->code))); /* drop into OP_SET */
	      else
		{
		  if (sc->args != sc->nil)
		    {
		      push_op_stack(sc, sc->list_set_function);
		      push_stack(sc, OP_EVAL_ARGS1, list_1(sc, sc->value), s7_append(sc, cdr(sc->args), sc->code));
		      sc->code = car(sc->args);
		    }
		  else eval_error(sc, "list set!: not enough arguments: ~S", sc->code);
		  goto EVAL;
		}
	    }
	  else
	    {
	      if (s7_is_vector(sc->value))
		{
		  /* (let ((L #(#(1 2 3) #(4 5 6)))) (set! ((L 1) 0) 32) L)
		   * bad case when args is nil: (let ((L #(#(1 2 3) #(4 5 6)))) (set! ((L 1)) 32) L)
		   */
		  if (sc->args != sc->nil)
		    {
		      push_op_stack(sc, sc->vector_set_function);
		      push_stack(sc, OP_EVAL_ARGS1, list_1(sc, sc->value), s7_append(sc, cdr(sc->args), sc->code));
		      sc->code = car(sc->args);
		    }
		  else eval_error(sc, "vector set!: not enough arguments: ~S", sc->code);
		  goto EVAL;
		}
	      sc->code = cons_unchecked(sc, cons(sc, sc->value, sc->args), sc->code);
	    }
	  /* fall through */
	  
	case OP_SET:                                                             /* entry for set! */
	  check_set(sc);
	  
	case OP_SET_UNCHECKED:
	  if (is_pair(car(sc->code)))                                            /* has accessor */
	    {
	      int32_t choice;
	      choice = set_pair_ex(sc);
	      if (choice == goto_EVAL) goto EVAL;
	      if (choice == goto_START) goto START;
	      if (choice == goto_APPLY) goto APPLY;
	      goto EVAL_ARGS;
	    }
	  /* fall through */
	  
	case OP_SET_NORMAL:
	  {
	    s7_pointer x;
	    x = cadr(sc->code);
	    if (is_pair(x))
	      {
		push_stack_no_args(sc, OP_SET1, car(sc->code));
		sc->code = x;
		goto EVAL;
	      }
	    
	    if (is_symbol(x))
	      sc->value = find_symbol_checked(sc, x);
	    else sc->value = _NFre(x);
	    sc->code = car(sc->code);
	  }
	  
	case OP_SET1:
	  {
	    s7_pointer lx;
	    /* if unbound variable hook here, we need the binding, not the current value */
	    lx = find_symbol(sc, _TSet(sc->code));
	    if (is_slot(lx))
	      {
		if (slot_has_accessor(lx))
		  {
		    s7_pointer func;
		    func = slot_accessor(lx);
		    if (is_procedure_or_macro(func))
		      {
			if (is_c_function(func))
			  {
			    set_car(sc->t2_1, sc->code);
			    set_car(sc->t2_2, sc->value);
			    sc->value = c_function_call(func)(sc, sc->t2_1);
			    if (sc->value == sc->error_symbol) /* backwards compatibility... (but still used I think in g_features_set) */
			      return(s7_error(sc, sc->error_symbol, set_elist_3(sc, s7_make_string_wrapper(sc, "can't set ~S to ~S"), car(sc->t2_1), car(sc->t2_2))));
			  }
			else
			  {
			    sc->args = list_2(sc, sc->code, sc->value);
			    push_stack(sc, OP_SET_WITH_ACCESSOR, sc->args, lx); /* op, args, code */
			    sc->code = func;
			    goto APPLY;
			  }
		      }
		  }
		else
		  {
		    if (is_syntax(slot_value(lx)))
		      eval_error(sc, "can't set! ~A", sc->code);
		  }
		slot_set_value(lx, sc->value);
		goto START;
	      }
	    eval_type_error(sc, "set! ~A: unbound variable", sc->code);
	  }
	  
	case OP_SET_WITH_ACCESSOR:
	  if (sc->value == sc->error_symbol) /* backwards compatibility... */
	    return(s7_error(sc, sc->error_symbol, set_elist_2(sc, s7_make_string_wrapper(sc, "can't set ~S"), sc->code)));
	  slot_set_value(sc->code, sc->value);
	  break;

	case OP_SET_WITH_LET_1:
	  {
	    s7_pointer e, b, x;
	    /* from the T_SYNTAX branch of set_pair_ex: (set! (with-let e b) x) as in let-temporarily
	     *   here sc->value is the new value for the settee = x, args has the (as yet unevaluated) let and settee-expression.
	     *   'b above can be a pair = generalized set in the 'e environment.
	     */
	    if (!is_pair(sc->args))                /* (set! (with-let) ...) */
	      eval_error(sc, "set! (with-let)? ~A", sc->cur_code);
	    if (!is_pair(cdr(sc->args)))           /* (set! (with-let e) ...) */
	      eval_error(sc, "set! (with-let ...) has no symbol to set? ~A", sc->cur_code);
	    e = car(sc->args);
	    b = cadr(sc->args);
	    x = sc->value;
	    if (is_symbol(e))
	      {
		if (is_symbol(b))
		  {
		    e = find_symbol_checked(sc, e); /* the let */
		    if (!is_let(e))
		      return(wrong_type_argument_with_type(sc, sc->let_set_symbol, 1, e, a_let_string));
		    sc->value = let_set_1(sc, e, b, x);
		    goto START;
		  }
		sc->value = find_symbol_checked(sc, e);
		sc->code = list_2(sc, b, ((is_symbol(x)) || (is_pair(x))) ? set_plist_2(sc, sc->quote_symbol, x) : x);
		goto SET_WITH_LET;
	      }
	    else
	      {
		sc->code = e;                       /* 'e above, an expression we need to evaluate */
		sc->args = list_2(sc, b, x);        /* can't reuse sc->args here via set-car! etc */
		push_stack(sc, OP_SET_WITH_LET_2, sc->args, sc->code);
		goto EVAL;
	      }
	  }

	case OP_SET_WITH_LET_2:
	  {
	    s7_pointer b, x;
	    /* here sc->value = let = 'e, args = '(b x) where 'b might be a pair */
	    if (!is_let(sc->value))
	      return(wrong_type_argument_with_type(sc, sc->let_set_symbol, 1, sc->value, a_let_string));
	    b = car(sc->args);
	    x = cadr(sc->args);
	    if (is_symbol(b))   /* b is a symbol -- everything else is ready so call let-set! */
	      {
		sc->value = let_set_1(sc, sc->value, b, x);
		goto START;
	      }
	    if ((is_symbol(x)) || (is_pair(x)))
	      sc->code = list_2(sc, b, ((is_symbol(x)) || (is_pair(x))) ? set_plist_2(sc, sc->quote_symbol, x) : x);
	    else sc->code = sc->args;
	  }

	SET_WITH_LET:
	  activate_let(sc, sc->value);  /* this activates sc->value, so the set! will happen in that environment */
	  if (is_pair(car(sc->code)))
	    {
	      int32_t choice;
	      choice = set_pair_ex(sc);
	      if (choice == goto_EVAL) goto EVAL;
	      if (choice == goto_START) goto START;
	      if (choice == goto_APPLY) goto APPLY;
	      goto EVAL_ARGS;
	    }
	  return(s7_error(sc, sc->error_symbol, set_elist_2(sc, s7_make_string_wrapper(sc, "can't set ~S"), sc->args)));	  
	  
	  
	  /* -------------------------------- IF -------------------------------- */
	case OP_IF:
	  check_if(sc);
	  
	case OP_IF_UNCHECKED:
	  push_stack_no_args(sc, OP_IF1, cdr(sc->code));
	  sc->code = car(sc->code);
	  goto EVAL;
	  
	case OP_IF1:
	  if (is_true(sc, sc->value))
	    sc->code = car(sc->code);
	  else sc->code = unchecked_car(cdr(sc->code));   /* even pre-optimization, (if #f #f) ==> #<unspecified> because unique_car(sc->nil) = sc->unspecified */
	  if (is_pair(sc->code))
	    goto EVAL;
	  if (is_symbol(sc->code))
	    sc->value = find_symbol_checked(sc, sc->code);
	  else sc->value = sc->code;
	  break;
	  
	  /* adding if*_z and if*_z_z made no difference */
        #define IF_CASE(Op, Code, Not_Code)						\
	  case Op ## _P:   Code {sc->code = cadr(sc->code); goto EVAL;} else {sc->value = sc->unspecified; goto START;} \
	  case Op ## _R:   Code {sc->value = sc->unspecified; goto START;} else {sc->code = cadr(sc->code); goto EVAL;} \
  	  case Op ## _P_P: Code {sc->code = cadr(sc->code); goto EVAL;} else {sc->code = caddr(sc->code); goto EVAL;} \
	  case Op ## _N:   Not_Code {sc->code = cadr(sc->code); goto EVAL;} else {sc->value = sc->unspecified; goto START;} \
  	  case Op ## _N_N: Not_Code {sc->code = cadr(sc->code); goto EVAL;} else {sc->code = caddr(sc->code); goto EVAL;}
	  
	  IF_CASE(OP_IF_S, 
	    if (is_true(sc, find_symbol_unchecked(sc, car(sc->code)))),
	    if (is_false(sc, find_symbol_unchecked(sc, cadar(sc->code)))))
	    
	  IF_CASE(OP_IF_A, 
		  if (is_true(sc, c_call(sc->code)(sc, car(sc->code)))),	
		  if (is_false(sc, c_call(cdar(sc->code))(sc, cadar(sc->code)))))
   
	  IF_CASE(OP_IF_C, 
		  if (is_true(sc, c_call(car(sc->code))(sc, opt_pair2(sc->code)))),
		  if (is_false(sc, c_call(cadar(sc->code))(sc, opt_pair2(sc->code)))))

	  IF_CASE(OP_IF_IS_TYPE_S,
		  if (type(find_symbol_unchecked(sc, opt_sym2(sc->code))) == opt_con3(sc->code)),
		  if (type(find_symbol_unchecked(sc, opt_sym2(sc->code))) != opt_con3(sc->code)))
	    
	  IF_CASE(OP_IF_IS_TYPE_opSq, 
		  set_car(sc->t1_1, find_symbol_unchecked(sc, opt_sym2(sc->code))); \
		  if (type(c_call(cadar(sc->code))(sc, sc->t1_1)) == opt_con3(sc->code)),
		  set_car(sc->t1_1, find_symbol_unchecked(sc, opt_sym2(sc->code))); \
		  if (type(c_call(cadr(cadar(sc->code)))(sc, sc->t1_1)) != opt_con3(sc->code)))
	    
	  IF_CASE(OP_IF_CS, 
		  set_car(sc->t1_1, find_symbol_unchecked(sc, opt_sym2(sc->code))); if (is_true(sc, c_call(car(sc->code))(sc, sc->t1_1))),
		  set_car(sc->t1_1, find_symbol_unchecked(sc, opt_sym2(sc->code))); if (is_false(sc, c_call(cadar(sc->code))(sc, sc->t1_1))))
	    
	  IF_CASE(OP_IF_CSQ, 
		  set_car(sc->t2_1, find_symbol_unchecked(sc, opt_sym3(sc->code))); \
		  set_car(sc->t2_2, opt_con2(sc->code));		\
		  if (is_true(sc, c_call(car(sc->code))(sc, sc->t2_1))),
		  set_car(sc->t2_1, find_symbol_unchecked(sc, opt_sym3(sc->code))); \
		  set_car(sc->t2_2, opt_con2(sc->code));		\
		  if (is_false(sc, c_call(cadar(sc->code))(sc, sc->t2_1))))
	    
	  IF_CASE(OP_IF_CSS, 
		  set_car(sc->t2_1, find_symbol_unchecked(sc, opt_sym3(sc->code))); \
		  set_car(sc->t2_2, find_symbol_unchecked(sc, opt_sym2(sc->code)));
		  if (is_true(sc, c_call(car(sc->code))(sc, sc->t2_1))),
		  set_car(sc->t2_1, find_symbol_unchecked(sc, opt_sym3(sc->code))); \
		  set_car(sc->t2_2, find_symbol_unchecked(sc, opt_sym2(sc->code)));
		  if (is_false(sc, c_call(cadar(sc->code))(sc, sc->t2_1))))
	    
	  IF_CASE(OP_IF_CSC, 
		  set_car(sc->t2_1, find_symbol_unchecked(sc, opt_sym3(sc->code))); \
		  set_car(sc->t2_2, opt_con2(sc->code));		\
		  if (is_true(sc, c_call(car(sc->code))(sc, sc->t2_1))),
		  set_car(sc->t2_1, find_symbol_unchecked(sc, opt_sym3(sc->code))); \
		  set_car(sc->t2_2, opt_con2(sc->code));		\
		  if (is_false(sc, c_call(cadar(sc->code))(sc, sc->t2_1))))
	    
	  IF_CASE(OP_IF_S_opCq, 
		  set_car(sc->t2_2, c_call(opt_pair2(sc->code))(sc, cdr(opt_pair2(sc->code)))); \
		  set_car(sc->t2_1, find_symbol_unchecked(sc, opt_sym3(sc->code))); \
		  if (is_true(sc, c_call(car(sc->code))(sc, sc->t2_1))),
		  set_car(sc->t2_2, c_call(opt_pair2(sc->code))(sc, cdr(opt_pair2(sc->code)))); \
		  set_car(sc->t2_1, find_symbol_unchecked(sc, opt_sym3(sc->code))); \
		  if (is_false(sc, c_call(cadar(sc->code))(sc, sc->t2_1))))

	  IF_CASE(OP_IF_opSq, 
		  set_car(sc->t1_1, find_symbol_unchecked(sc, opt_sym3(sc->code))); \
		  set_car(sc->t1_1, c_call(opt_pair2(sc->code))(sc, sc->t1_1));     \
		  if (is_true(sc, c_call(car(sc->code))(sc, sc->t1_1))),
		  set_car(sc->t1_1, find_symbol_unchecked(sc, opt_sym3(sc->code))); \
		  set_car(sc->t1_1, c_call(opt_pair2(sc->code))(sc, sc->t1_1));     \
		  if (is_false(sc, c_call(cadar(sc->code))(sc, sc->t1_1))))
	    
	  IF_CASE(OP_IF_AND2, 
		  if ((is_true(sc, c_call(opt_pair2(sc->code))(sc, car(opt_pair2(sc->code))))) && \
		      (is_true(sc, c_call(opt_pair3(sc->code))(sc, car(opt_pair3(sc->code)))))),
		  if ((is_false(sc, c_call(opt_pair2(sc->code))(sc, car(opt_pair2(sc->code))))) || \
		      (is_false(sc, c_call(opt_pair3(sc->code))(sc, car(opt_pair3(sc->code)))))))

	  IF_CASE(OP_IF_AND3, 
		  if ((is_true(sc, c_call(opt_pair2(sc->code))(sc, car(opt_pair2(sc->code))))) && \
		      (is_true(sc, c_call(opt_pair3(sc->code))(sc, car(opt_pair3(sc->code))))) && \
		      (is_true(sc, c_call(cdr(opt_pair3(sc->code)))(sc, cadr(opt_pair3(sc->code)))))),
		  if ((is_false(sc, c_call(opt_pair2(sc->code))(sc, car(opt_pair2(sc->code))))) || \
		      (is_false(sc, c_call(opt_pair3(sc->code))(sc, car(opt_pair3(sc->code))))) || \
		      (is_false(sc, c_call(cdr(opt_pair3(sc->code)))(sc, cadr(opt_pair3(sc->code)))))))

	  IF_CASE(OP_IF_OR2, 
		  if ((is_true(sc, c_call(opt_pair2(sc->code))(sc, car(opt_pair2(sc->code))))) || \
		      (is_true(sc, c_call(opt_pair3(sc->code))(sc, car(opt_pair3(sc->code)))))),
		  if ((is_false(sc, c_call(opt_pair2(sc->code))(sc, car(opt_pair2(sc->code))))) && \
		      (is_false(sc, c_call(opt_pair3(sc->code))(sc, car(opt_pair3(sc->code)))))))

	case OP_IF_P_P: push_stack_no_args(sc, OP_IF_PP, cadr(sc->code)); sc->code = car(sc->code); goto EVAL;
	case OP_IF_P_N: push_stack_no_args(sc, OP_IF_PR, cadr(sc->code)); sc->code = cadar(sc->code); goto EVAL;
	case OP_IF_P_R: push_stack_no_args(sc, OP_IF_PR, cadr(sc->code)); sc->code = car(sc->code); goto EVAL;
	case OP_IF_P_P_P: push_stack_no_args(sc, OP_IF_PPP, cdr(sc->code)); sc->code = car(sc->code); goto EVAL;
	case OP_IF_P_N_N: push_stack_no_args(sc, OP_IF_PRR, cdr(sc->code)); sc->code = cadar(sc->code); goto EVAL;
	  
	case OP_IF_Z_P: push_stack_no_args(sc, OP_IF_PP, cadr(sc->code)); sc->code = _TPair(car(sc->code)); goto OPT_EVAL;
	case OP_IF_Z_N: push_stack_no_args(sc, OP_IF_PR, cadr(sc->code)); sc->code = cadar(sc->code); goto OPT_EVAL;
	case OP_IF_Z_R: push_stack_no_args(sc, OP_IF_PR, cadr(sc->code)); sc->code = _TPair(car(sc->code)); goto OPT_EVAL;
	case OP_IF_Z_P_P: push_stack_no_args(sc, OP_IF_PPP, cdr(sc->code)); sc->code = _TPair(car(sc->code)); goto OPT_EVAL;
	case OP_IF_Z_N_N: push_stack_no_args(sc, OP_IF_PRR, cdr(sc->code)); sc->code = cadar(sc->code); goto OPT_EVAL;
	  
	case OP_IF_ANDP_P: push_stack_no_args(sc, OP_IF_PP, cadr(sc->code)); sc->code = cdar(sc->code); goto AND_P;
	case OP_IF_ANDP_N: push_stack_no_args(sc, OP_IF_PR, cadr(sc->code)); sc->code = cdadar(sc->code); goto AND_P;
	case OP_IF_ANDP_R: push_stack_no_args(sc, OP_IF_PR, cadr(sc->code)); sc->code = cdar(sc->code); goto AND_P;
	case OP_IF_ANDP_P_P: push_stack_no_args(sc, OP_IF_PPP, cdr(sc->code)); sc->code = cdar(sc->code); goto AND_P;
	case OP_IF_ANDP_N_N: push_stack_no_args(sc, OP_IF_PRR, cdr(sc->code)); sc->code = cdadar(sc->code); goto AND_P;
	  
	case OP_IF_ORP_P: push_stack_no_args(sc, OP_IF_PP, cadr(sc->code)); sc->code = cdar(sc->code); goto OR_P;
	case OP_IF_ORP_N: push_stack_no_args(sc, OP_IF_PR, cadr(sc->code)); sc->code = cdadar(sc->code); goto OR_P;
	case OP_IF_ORP_R: push_stack_no_args(sc, OP_IF_PR, cadr(sc->code)); sc->code = cdar(sc->code); goto OR_P;
	case OP_IF_ORP_P_P: push_stack_no_args(sc, OP_IF_PPP, cdr(sc->code)); sc->code = cdar(sc->code); goto OR_P;
	case OP_IF_ORP_N_N: push_stack_no_args(sc, OP_IF_PRR, cdr(sc->code)); sc->code = cdadar(sc->code); goto OR_P;
	  
	case OP_IF_PP:
	  if (is_true(sc, sc->value))
	    goto EVAL;
	  sc->value = sc->unspecified;
	  break;

	case OP_IF_PPP:
	  if (is_true(sc, sc->value))
	    sc->code = car(sc->code);
	  else sc->code = cadr(sc->code);
	  goto EVAL;
	  
	case OP_IF_PR:
	  if (is_false(sc, sc->value))
	    goto EVAL;
	  sc->value = sc->unspecified;
	  break;

	case OP_IF_PRR:
	  if (is_false(sc, sc->value))
	    sc->code = car(sc->code);
	  else sc->code = cadr(sc->code);
	  goto EVAL;
	  
	case OP_WHEN_PP:
	  if (is_true(sc, sc->value))
	    {
	      push_stack_no_args(sc, OP_BEGIN1, _TPair(cdr(sc->code)));
	      sc->code = car(sc->code);
	      goto EVAL;
	    }
	  sc->value = sc->unspecified;
	  break;
	  
	case OP_COND_FEED:
	  /* (cond (expr => p)) where p is (lambda (s) ...) -- see check_cond */
	  push_stack_no_args(sc, OP_COND_FEED_1, sc->code);
	  sc->code = caar(sc->code);
	  goto EVAL;
	  
	case OP_COND_FEED_1:
	  if (is_true(sc, sc->value))
	    {
	      if (is_multiple_value(sc->value))
		sc->code = cons(sc, opt_lambda2(sc->code), multiple_value(sc->value));
	      else
		{
		  new_frame_with_slot(sc, sc->envir, sc->envir, caadr(opt_lambda2(sc->code)), sc->value);
		  sc->code = caddr(opt_lambda2(sc->code));
		}
	      goto EVAL;
	    }
	  sc->value = sc->unspecified; /* it's cond -- perhaps push as sc->args above; this was nil until 21-Feb-17! */
	  break;
	  
	case OP_WHEN:
	  check_when(sc);
	  
	case OP_WHEN_UNCHECKED:
	  push_stack_no_args(sc, OP_WHEN1, cdr(sc->code));
	  sc->code = car(sc->code);
	  goto EVAL;
	  
	case OP_WHEN1:
	  if (is_true(sc, sc->value)) 
	    goto BEGIN1;
	  sc->value = sc->unspecified;
	  break;
	  
	case OP_WHEN_S:
	  if (is_true(sc, find_symbol_unchecked(sc, car(sc->code))))
	    {
	      push_stack_no_args(sc, OP_BEGIN1, _TPair(cddr(sc->code)));
	      sc->code = cadr(sc->code);
	      goto EVAL;
	    }
	  sc->value = sc->unspecified;
	  break;
	  
	case OP_WHEN_A:
	  if (is_true(sc, c_call(sc->code)(sc, car(sc->code))))
	    {
	      push_stack_no_args(sc, OP_BEGIN1, _TPair(cddr(sc->code)));
	      sc->code = cadr(sc->code);
	      goto EVAL;
	    }
	  sc->value = sc->unspecified;
	  break;

        case OP_WHEN_P:
	  push_stack_no_args(sc, OP_WHEN_PP, cdr(sc->code));
	  sc->code = car(sc->code);
	  goto EVAL;

	case OP_UNLESS:
	  check_unless(sc);
	  
	case OP_UNLESS_UNCHECKED:
	  push_stack_no_args(sc, OP_UNLESS1, cdr(sc->code));
	  sc->code = car(sc->code);
	  goto EVAL;
	  
	case OP_UNLESS1:
	  if (is_false(sc, sc->value)) 
	    goto BEGIN1;
	  sc->value = sc->unspecified;
	  break;
	  
	case OP_UNLESS_S:
	  if (is_false(sc, find_symbol_unchecked(sc, car(sc->code))))
	    {
	      push_stack_no_args(sc, OP_BEGIN1, _TPair(cddr(sc->code)));
	      sc->code = cadr(sc->code);
	      goto EVAL;
	    }
	  sc->value = sc->unspecified;
	  break;
	  
	case OP_UNLESS_A:
	  if (is_false(sc, c_call(sc->code)(sc, car(sc->code))))
	    {
	      push_stack_no_args(sc, OP_BEGIN1, _TPair(cddr(sc->code)));
	      sc->code = cadr(sc->code);
	      goto EVAL;
	    }
	  sc->value = sc->unspecified;
	  break;
	  
	case OP_SAFE_C_P_1:
	  set_car(sc->t1_1, sc->value);
	  sc->value = c_call(sc->code)(sc, sc->t1_1);
	  break;
	  
	case OP_NOT_P_1:
	  sc->value = ((sc->value == sc->F) ? sc->T : sc->F);
	  break;

	case OP_SAFE_CLOSURE_P_1:
	  sc->envir = old_frame_with_slot(sc, closure_let(opt_lambda(sc->code)), sc->value);
	  sc->code = _TPair(closure_body(opt_lambda(sc->code)));
	  goto BEGIN1;

	case OP_SAFE_CLOSURE_AP_1:
	  sc->envir = old_frame_with_two_slots(sc, closure_let(opt_lambda(sc->code)), sc->args, sc->value);
	  sc->code = _TPair(closure_body(opt_lambda(sc->code)));
	  goto BEGIN1;

	case OP_SAFE_CLOSURE_PA_1:
	  sc->envir = old_frame_with_two_slots(sc, closure_let(opt_lambda(sc->code)), sc->value, sc->args);
	  sc->code = _TPair(closure_body(opt_lambda(sc->code)));
	  goto BEGIN1;

	case OP_CLOSURE_P_1:
	  check_stack_size(sc);
	  sc->code = opt_lambda(sc->code);
	  new_frame_with_slot(sc, closure_let(sc->code), sc->envir, car(closure_args(sc->code)), sc->value);
	  sc->code = _TPair(closure_body(sc->code));
	  goto BEGIN1;
	  
	case OP_SAFE_C_PP_1:
	  /* unless multiple values from last call (first arg), sc->args == sc->nil because we pushed that.
	   *   we get here only from OP_SAFE_C_PP.
	   *
	   * currently splice_in_values changes the operator so if we get here, sc->value is the result of the first arg
	   *
	   * safe_c_pp -> 1, but if mv, -> 3
	   * 1: -> 2, if mv -> 4
	   * 2: done (both normal)
	   * 3: -> 5, but if mv, -> 6
	   * 4: done (1 normal, 2 mv)
	   * 5: done (1 mv, 2 normal)
	   * 6: done (both mv)
	   *
	   * I think safe_c_ppp would require 18 branches (or maybe just collect the args and concatenate at the end?)
	   */
	  push_stack(sc, OP_EVAL_ARGS_P_2, sc->value, sc->code); /* mv -> 3 */
	  sc->code = caddr(sc->code);
	  if (is_optimized(sc->code))
	    goto OPT_EVAL;
	  goto EVAL;

	case OP_SAFE_C_PP_3:
	  /* we get here if the first arg returned multiple values */
	  push_stack(sc, OP_SAFE_C_PP_5, sc->value, sc->code);
	  sc->code = caddr(sc->code);
	  if (is_optimized(sc->code))
	    goto OPT_EVAL;
	  goto EVAL;
	  
	case OP_SAFE_C_PP_5:
	  /* 1 mv, 2, normal */
	  sc->args = s7_append(sc, sc->args, list_1(sc, sc->value));
	  sc->code = c_function_base(opt_cfunc(sc->code));
	  goto APPLY;
	  
	case OP_SAFE_C_PP_6:
	  /* both mv */
	  sc->args = s7_append(sc, sc->args, sc->value);
	  /*
	   * c_call(sc->code) here is g_add_2, but we have any number of args from a values call
	   *   the original (unoptimized) function is (hopefully) c_function_base(opt_cfunc(sc->code))?
	   *   (let () (define (ho a) (+ a 2)) (define (hi) (+ (ho 1) (ho 2))) (hi)) -> 7
	   *   (let () (define (ho a) (+ a 2)) (define (hi) (+ (ho 1) (values 3 4))) (hi)) -> 10
	   *   (let () (define (ho a) (+ a 2)) (define (hi) (+ (values 3 4) (ho 1))) (hi)) -> 10
	   *   (let () (define (hi) (+ (values 1 2) (values 3 4))) (hi)) -> 10
	   */
	  sc->code = c_function_base(opt_cfunc(sc->code));
	  goto APPLY;
	  
	case OP_C_P_1:
	  sc->value = c_call(sc->code)(sc, list_1(sc, sc->value));
	  break;
	  
	case OP_C_P_2:
	  /* op_c_p_1 -> mv case: (define (hi) (format (values #f "~A ~D" 1 2))) */
	  sc->code = c_function_base(opt_cfunc(sc->code)); /* see comment above */
	  sc->args = copy_list(sc, sc->value);
	  goto APPLY;
	  
	case OP_CLOSURE_AP_1:
	  /* sc->value is presumably the "P" argument value, "A" is sc->args */
	  check_stack_size(sc);
	  sc->code = opt_lambda(sc->code);
	  new_frame_with_two_slots(sc, closure_let(sc->code), sc->envir, 
				   car(closure_args(sc->code)), sc->args, 
				   cadr(closure_args(sc->code)), sc->value);
	  sc->code = _TPair(closure_body(sc->code));
	  goto BEGIN1;
	  
	case OP_CLOSURE_AP_MV:	  /* here we got multiple values */
	  sc->code = opt_lambda(sc->code);
	  sc->args = cons(sc, sc->args, copy_list(sc, sc->value));
	  goto APPLY;
	  
	case OP_CLOSURE_PA_1:
	  check_stack_size(sc);
	  sc->code = opt_lambda(sc->code);
	  new_frame_with_two_slots(sc, closure_let(sc->code), sc->envir, 
				   car(closure_args(sc->code)), sc->value, 
				   cadr(closure_args(sc->code)), sc->args);
	  sc->code = _TPair(closure_body(sc->code));
	  goto BEGIN1;
	  
	case OP_CLOSURE_PA_MV:
	  sc->code = opt_lambda(sc->code);
	  sc->args = s7_append(sc, copy_list(sc, sc->value), cons(sc, sc->args, sc->nil));
	  goto APPLY;
	  
	case OP_CLOSURE_P_MV:
	  sc->code = opt_lambda(sc->code);
	  sc->args = copy_list(sc, sc->value);
	  goto APPLY;
	  
	case OP_C_AP_1:
	  sc->value = c_call(sc->code)(sc, list_2(sc, sc->args, sc->value));
	  break;

	  /* -------------------------------- LET -------------------------------- */
	case OP_LET_NO_VARS:
	  new_frame(sc, sc->envir, sc->envir);
	  sc->code = _TPair(cdr(sc->code));         /* ignore the () */
	  goto BEGIN1;
	  
	case OP_NAMED_LET_NO_VARS:
	  {
	    s7_pointer body;
	    new_frame(sc, sc->envir, sc->envir);
	    body = cddr(sc->code);
	    sc->args = make_closure(sc, sc->nil, body, T_CLOSURE | T_COPY_ARGS | ((is_safe_closure(body)) ? T_SAFE_CLOSURE : 0), 0);
	    /* sc->args is a temp here */
	    make_slot_1(sc, sc->envir, car(sc->code), sc->args);
	    sc->code = _TPair(body);
	    goto BEGIN1;
	  }
	  
	case OP_LET_C:
	  {
	  /* one var, init is constant, incoming sc->code is '(((var val))...)! 
	   *   somehow we can get here from make-hook (let ((result #<unspecified>))...) with opt_sym3 and opt_con2 unset??
	   *   opt_sym3 can be clobbered by a subsequent opt_back apparently?? 
	   */
	    s7_pointer binding;
	    binding = _TPair(caar(sc->code));
	    new_frame_with_slot(sc, sc->envir, sc->envir, car(binding), cadr(binding));
	    sc->code = _TPair(cdr(sc->code));
	    goto BEGIN1;
	  }
	  
	case OP_LET_S:
	  {
	  /* one var, init is symbol, incoming sc->code is '(((var sym))...) */
	    s7_pointer binding;
	    binding = _TPair(caar(sc->code));
	    sc->value = find_symbol_checked(sc, cadr(binding));
	    new_frame_with_slot(sc, sc->envir, sc->envir, car(binding), sc->value);
	    sc->code = _TPair(cdr(sc->code));
	    goto BEGIN1;
	  }
	  
	case OP_LET_S_Z:
	  {
	    s7_pointer binding;
	    binding = _TPair(caar(sc->code));
	    sc->value = find_symbol_checked(sc, cadr(binding));
	    new_frame_with_slot(sc, sc->envir, sc->envir, car(binding), sc->value);
	    sc->code = _TPair(cadr(sc->code));
	    goto OPT_EVAL_CHECKED;
	  }
	  
	case OP_LET_opSq:
	  {
	    s7_pointer binding;
	    binding = _TPair(caar(sc->code));
	    set_car(sc->t1_1, find_symbol_unchecked(sc, opt_sym2(sc->code)));
	    sc->value = c_call(cadr(binding))(sc, sc->t1_1);
	    new_frame_with_slot(sc, sc->envir, sc->envir, car(binding), sc->value);
	    push_stack_no_args(sc, OP_BEGIN1, _TPair(cddr(sc->code)));
	    sc->code = cadr(sc->code);
	    goto EVAL;
	  }
	  
	case OP_LET_CAR:
	  {
	    s7_pointer binding, val;
	    binding = _TPair(caar(sc->code));
	    val = find_symbol_unchecked(sc, opt_sym2(sc->code));
	    sc->value = (is_pair(val)) ? car(val) : g_car(sc, set_plist_1(sc, val));
	    new_frame_with_slot(sc, sc->envir, sc->envir, car(binding), sc->value);
	    push_stack_no_args(sc, OP_BEGIN1, _TPair(cddr(sc->code)));
	    sc->code = cadr(sc->code);
	    goto EVAL;
	  }
	  
	case OP_LET_opSq_P:
	  {
	    s7_pointer binding;
	    binding = _TPair(caar(sc->code));
	    set_car(sc->t1_1, find_symbol_unchecked(sc, opt_sym2(sc->code)));
	    sc->value = c_call(cadr(binding))(sc, sc->t1_1);
	    new_frame_with_slot(sc, sc->envir, sc->envir, car(binding), sc->value);
	    sc->code = cadr(sc->code);
	    goto EVAL;
	  }
	  
	case OP_LET_opCq:      /* one var, init is safe_c_c */
	  sc->value = c_call(opt_pair2(sc->code))(sc, cdr(opt_pair2(sc->code)));
	  new_frame_with_slot(sc, sc->envir, sc->envir, opt_sym3(sc->code), sc->value);
	  sc->code = _TPair(cdr(sc->code));
	  goto BEGIN1;
	  
	case OP_LET_opSSq:      /* one var, init is safe_c_ss */
	  {
	    s7_pointer largs, in_val;
	    largs = _TPair(opt_pair2(sc->code));                              /* cadr(caar(sc->code)); */
	    in_val = find_symbol_unchecked(sc, cadr(largs));
	    set_car(sc->t2_2, find_symbol_unchecked(sc, opt_sym3(sc->code))); /* caddr(largs)); */
	    set_car(sc->t2_1, in_val);
	    sc->value = c_call(largs)(sc, sc->t2_1);
	    new_frame_with_slot(sc, sc->envir, sc->envir, caaar(sc->code), sc->value);
	    sc->code = _TPair(cdr(sc->code));
	    goto BEGIN1;
	  }

	case OP_LET_opSSq_E: 
	  {
	    s7_pointer largs, in_val;
	    largs = _TPair(opt_pair2(sc->code));                              /* cadr(caar(sc->code)); */
	    in_val = find_symbol_unchecked(sc, cadr(largs));
	    set_car(sc->t2_2, find_symbol_unchecked(sc, opt_sym3(sc->code))); /* caddr(largs)); */
	    set_car(sc->t2_1, in_val);
	    sc->value = c_call(largs)(sc, sc->t2_1);
	    new_frame_with_slot(sc, sc->envir, sc->envir, caaar(sc->code), sc->value);
	    sc->code = cadr(sc->code);
	    goto EVAL;
	  }
	  
	case OP_LET_opaSSq_E: 
	  {
	    s7_pointer in_val, lst;
	    in_val = find_symbol_unchecked(sc, cadr(opt_pair2(sc->code)));
	    lst = find_symbol_unchecked(sc, opt_sym3(sc->code)); 
	    if (is_pair(lst))
	      sc->value = s7_assq(sc, in_val, lst);
	    else
	      {
		if (is_null(lst))
		  sc->value = sc->F;
		else sc->value = g_assq(sc, set_plist_2(sc, in_val, lst));
	      }
	    new_frame_with_slot(sc, sc->envir, sc->envir, caaar(sc->code), sc->value);
	    sc->code = cadr(sc->code);
	    goto EVAL;
	  }
	  
	case OP_LET_Z:
	  push_stack(sc, OP_LET_Z_1, opt_sym2(cdr(sc->code)), cadr(sc->code));
	  sc->code = _TPair(opt_pair2(sc->code));
	  goto OPT_EVAL;
	  
	case OP_LET_Z_1:
	  new_frame_with_slot(sc, sc->envir, sc->envir, sc->args, sc->value);
	  goto EVAL;
	  
	case OP_LET_A:
	  {
	    s7_pointer binding;
	    binding = caar(sc->code);
	    new_frame_with_slot(sc, sc->envir, sc->envir, car(binding), c_call(cdr(binding))(sc, cadr(binding)));
	    sc->code = cadr(sc->code);
	    goto EVAL;
	  }
	  
	case OP_LET_A_Z:
	  {
	    s7_pointer binding;
	    binding = caar(sc->code);
	    new_frame_with_slot(sc, sc->envir, sc->envir, car(binding), c_call(cdr(binding))(sc, cadr(binding)));
	    sc->code = cadr(sc->code);
	    goto OPT_EVAL_CHECKED;
	  }
	  
	case OP_LET_ONE:	  /* one var */
	  {
	    s7_pointer p;
	    p = caar(sc->code);
	    sc->value = cadr(p);
	    if (is_pair(sc->value))
	      {
		push_stack(sc, OP_LET_ONE_1, car(p), cdr(sc->code)); /* args code */
		sc->code = sc->value;
		goto EVAL;
	      }
	    if (is_symbol(sc->value))
	      sc->value = find_symbol_checked(sc, sc->value);
	    sc->code = _TPair(cdr(sc->code));
	    sc->args = car(p);
	    /* drop through */
	  }
	  
	case OP_LET_ONE_1:
	  new_frame_with_slot(sc, sc->envir, sc->envir, sc->args, sc->value);
	  goto BEGIN1;
	  
	case OP_LET_ALL_C:
	  {
	    s7_pointer p;
	    new_frame(sc, sc->envir, sc->envir);
	    for (p = car(sc->code); is_pair(p); p = cdr(p))
	      add_slot(sc->envir, caar(p), cadar(p));
	    sc->code = _TPair(cdr(sc->code));
	    goto BEGIN1;
	  }
	  
	case OP_LET_ALL_S:
	  /* n vars, all inits are symbols.  We need to GC-protect the new frame-list as it is being
	   *    created without tying the new frame into sc->envir until the end.
	   */
	  {
	    s7_pointer p, frame;
	    frame = make_simple_let(sc);
	    sc->args = frame;
	    for (p = car(sc->code); is_pair(p); p = cdr(p))
	      add_slot(frame, caar(p), find_symbol_checked(sc, cadar(p)));
	    sc->let_number++;
	    sc->envir = frame;
	    sc->code = _TPair(cdr(sc->code));
	    goto BEGIN1;
	  }
	  
	case OP_LET_ALL_opSq:
	  {
	    s7_pointer p, frame;
	    frame = make_simple_let(sc);
	    sc->args = frame;
	    for (p = car(sc->code); is_pair(p); p = cdr(p))
	      {
		s7_pointer cp;
		cp = cadar(p);
		set_car(sc->t1_1, find_symbol_unchecked(sc, cadr(cp)));
		add_slot(frame, caar(p), c_call(cp)(sc, sc->t1_1));
	      }
	    sc->let_number++;
	    sc->envir = frame;
	    sc->code = _TPair(cdr(sc->code));
	    goto BEGIN1;
	  }
	  
	  /* it is possible to save the frame+slots in a copied symbol+syntax pair, then reuse them
	   *  on every call here, but the savings in GC+allocation+setup is less than the cost in
	   *  marking the saved stuff past its actual life!  (If the code is removed from the heap,
	   *  the frame has to be saved on the permanent_objects list).
	   * But if all lookups are local, no frame/slots are needed.
	   */
	case OP_LET_ALL_X:
	  {
	    s7_pointer p, frame;
	    frame = make_simple_let(sc);
	    sc->args = frame;
	    for (p = car(sc->code); is_pair(p); p = cdr(p))
	      {
		s7_pointer arg;
		arg = cdar(p);
		arg = c_call(arg)(sc, car(arg));
		add_slot(frame, caar(p), arg);
	      }
	    sc->let_number++;
	    sc->envir = frame;
	    sc->code = _TPair(cdr(sc->code));
	    goto BEGIN1;
	  }
	  
	case OP_NAMED_LET:
	  sc->args = sc->nil;
	  sc->value = sc->code;
	  sc->code = cadr(sc->code);
	  goto LET1;
	  
	case OP_LET_UNCHECKED:
	  /* not named, but has vars */
	  {
	    s7_pointer x;
	    new_cell(sc, x, T_PAIR);
	    set_car(x, sc->code);
	    set_cdr(x, sc->nil);
	    sc->args = x;
	    sc->code = car(sc->code);
	    goto LET1A;
	  }
	  
	case OP_LET:
	  /* sc->code is everything after the let: (let ((a 1)) a) so sc->code is (((a 1)) a) */
	  /*   car can be either a list or a symbol ("named let") */
	  {
	    bool named_let;
	    check_let(sc);
	    sc->args = sc->nil;
	    sc->value = sc->code;
	    named_let = is_symbol(car(sc->code));
	    
	    sc->code = (named_let) ? cadr(sc->code) : car(sc->code);
	    if (is_null(sc->code))                    /* (let [name] () ...):  no bindings, so skip that step */
	      {
		sc->code = sc->value;
		new_frame(sc, sc->envir, sc->envir);
		if (named_let)
		  {
		    s7_pointer body;
		    body = cddr(sc->code);
		    sc->x = make_closure(sc, sc->nil, body, T_CLOSURE | T_COPY_ARGS | ((is_safe_closure(body)) ? T_SAFE_CLOSURE : 0), 0);
		    /* args = () in new closure, see NAMED_LET_NO_VARS above */
		    /* if this is a safe closure, we can build its env in advance and name it (a thunk in this case) */
		    set_funclet(closure_let(sc->x));
		    funclet_set_function(closure_let(sc->x), car(sc->code));
		    make_slot_1(sc, sc->envir, car(sc->code), sc->x);
		    sc->code = _TPair(body);
		    sc->x = sc->nil;
		  }
		else sc->code = _TPair(cdr(sc->code));
		goto BEGIN1;
	      }
	  }
	  
	LET1:
	case OP_LET1:
	  {
	    s7_pointer x, y;
	    
	    new_cell(sc, x, T_PAIR);
	    set_car(x, sc->value); /* the first time (now handled above), this saves the entire let body across the evaluations -- we pick it up later */
	    set_cdr(x, sc->args);
	    sc->args = x;
	    
	    if (is_pair(sc->code))
	      {
	      LET1A:
		x = cadar(sc->code);
		if (is_pair(x))
		  {
		    push_stack(sc, OP_LET1, sc->args, cdr(sc->code));
		    sc->code = x;
		    if (is_optimized(x))
		      goto OPT_EVAL;
		    goto EVAL;
		    /* this push_stack/goto can't be optimized away via a local optimize_op case statement
		     *    because any c_call can trigger an embedded call on the evaluator (for example,
		     *    open-sound involves both hooks, and s7_load if the corresponding .scm code exists),
		     *    so we have to protect sc->code and sc->args via the stack.  (I subsequently added
		     *    some protection here, but debugging this is hard, and the gain is not huge).
		     */
		  }
		if (is_symbol(x))
		  sc->value = find_symbol_checked(sc, x);
		else sc->value = _NFre(x);
		sc->code = cdr(sc->code);
		goto LET1;
	      }
	    
	    x = safe_reverse_in_place(sc, sc->args);
	    sc->code = car(x); /* restore the original form */
	    y = cdr(x);        /* use sc->args as the new frame */
	    sc->y = y;
	    sc->envir = reuse_as_let(sc, x, sc->envir);
	    
	    {
	      bool named_let;
	      named_let = is_symbol(car(sc->code));
	      if (named_let)
		{
		  /* we need to check the current environment for ridiculous cases like
		   *    (let hiho ((hiho 4)) hiho) -- I guess hiho is 4
		   * it's possible to package the entire named-let in all_x_* (see rc-a-s7.c), but
		   *    it's complicated code, and gains about 1/4 total compute time.  The overhead
		   *    is in eval -- goto BEGIN1;, all the eval switches, etc -- probably 500 of the 700
		   *    can be regained directly.
		   */
		  s7_pointer let_name, body;
		  int32_t n;
		  let_name = car(sc->code);
		  body = cddr(sc->code);
		  sc->envir = new_frame_in_env(sc, sc->envir);
		  
		  sc->w = sc->nil;
		  for (n = 0, x = cadr(sc->code); is_pair(x); n++, x = cdr(x))
		    sc->w = cons(sc, caar(x), sc->w);
		  
		  if (is_safe_closure(body))
		    {
		      s7_pointer arg, new_env;
		      sc->x = make_closure(sc, sc->w = safe_reverse_in_place(sc, sc->w), body, T_CLOSURE | T_SAFE_CLOSURE | T_COPY_ARGS, n);
		      new_env = new_frame_in_env(sc, sc->envir);
		      closure_set_let(sc->x, new_env);
		      for (arg = closure_args(sc->x); is_pair(arg); arg = cdr(arg))
			make_slot_1(sc, new_env, car(arg), sc->nil);
		      let_set_slots(new_env, reverse_slots(sc, let_slots(new_env)));
		    }
		  else sc->x = make_closure(sc, sc->w = safe_reverse_in_place(sc, sc->w), body, T_CLOSURE | T_COPY_ARGS, n);
		  make_slot_1(sc, sc->envir, let_name, sc->x);
		  /* sc->x = sc->nil; */
		  
		  sc->envir = new_frame_in_env(sc, sc->envir);
		  for (x = cadr(sc->code); is_not_null(y); x = cdr(x))
		    {
		      s7_pointer sym, args;
		      /* reuse the value cells as the new frame slots */
		      
		      sym = caar(x);
		      if (sym == let_name) let_name = sc->nil;
		      args = cdr(y);
		      reuse_as_slot(y, sym, unchecked_car(y)); /* y=slot, sym=symbol, car(y)=value */
		      set_next_slot(y, let_slots(sc->envir));
		      let_set_slots(sc->envir, y);
		      symbol_set_local(sym, let_id(sc->envir), y);
		      y = args;
		    }

		  sc->code = _TPair(body);
		  sc->w = sc->nil;
		  sc->x = sc->nil;
		}
	      else
		{
		  s7_pointer e;
		  uint64_t id;
		  
		  e = sc->envir;
		  id = let_id(e);
		  
		  for (x = car(sc->code); is_not_null(y); x = cdr(x))
		    {
		      s7_pointer sym, args;
		      /* reuse the value cells as the new frame slots */
		      
		      sym = caar(x);
		      args = cdr(y);
		      reuse_as_slot(y, sym, unchecked_car(y));
		      symbol_set_local(sym, id, y);
		      set_next_slot(y, let_slots(e));
		      let_set_slots(e, y);
		      
		      y = args;
		    }
		  sc->code = _TPair(cdr(sc->code));
		}
	    }
	    sc->y = sc->nil;
	    goto BEGIN1;
	  }
	  
	  
	  /* -------------------------------- LET* -------------------------------- */
	case OP_LET_STAR_ALL_X:
	  {
	    s7_pointer p, e = NULL;
	    for (p = car(sc->code); is_pair(p); p = cdr(p))
	      {
		s7_pointer arg;
		arg = cdar(p);
		arg = c_call(arg)(sc, car(arg));
		/* since each value is all_x safe, there are no internal closures over the on-going stack of lets here (so use one frame) */
		if (e)
		  make_slot_1(sc, e, caar(p), arg);
		else
		  {
		    new_frame_with_slot(sc, sc->envir, e, caar(p), arg);
		    sc->envir = e;
		  }
	      }
	    sc->code = _TPair(cdr(sc->code));
	    goto BEGIN1;
	  }
	  
	case OP_LET_STAR_A2:
	  push_stack(sc, OP_LET_STAR_A, sc->code, car(sc->code));
	  sc->code = opt_con2(sc->code);
	  goto EVAL;

	case OP_LET_STAR_A:
	  {
	    s7_pointer p, old_code;
	    old_code = sc->args;
	    new_frame_with_slot(sc, sc->envir, sc->envir, caar(sc->code), sc->value);
	    for (p = cdr(sc->code); is_pair(p); p = cdr(p))
	      {
		s7_pointer arg;
		arg = cdar(p);
		arg = c_call(arg)(sc, car(arg));
		make_slot_1(sc, sc->envir, caar(p), arg);
	      }
	    sc->code = _TPair(cdr(old_code));
	    goto BEGIN1;
	  }
	  
	case OP_NAMED_LET_STAR:
	  push_stack(sc, OP_LET_STAR1, sc->code, cadr(sc->code));
	  sc->code = opt_con2(sc->code);
	  goto EVAL;
	  
	case OP_LET_STAR2:
	  push_stack(sc, OP_LET_STAR1, sc->code, car(sc->code));
	  sc->code = opt_con2(sc->code);
	  goto EVAL;
	  
	case OP_LET_STAR:
	  check_let_star(sc);
	  
	case OP_LET_STAR_UNCHECKED:
	  if (is_symbol(car(sc->code)))
	    {
	      s7_pointer cx;
	      cx = car(sc->code);
	      sc->value = cdr(sc->code);
	      if (is_null(car(sc->value)))
		{
		  sc->envir = new_frame_in_env(sc, sc->envir);
		  sc->code = _TPair(cdr(sc->value));
		  make_slot_1(sc, sc->envir, cx, make_closure(sc, sc->nil, sc->code, T_CLOSURE_STAR | ((is_safe_closure(sc->code)) ? T_SAFE_CLOSURE : 0), 0));
		  goto BEGIN1;
		}
	    }
	  else
	    {
	      if (is_null(car(sc->code)))
		{
		  sc->envir = new_frame_in_env(sc, sc->envir);
		  sc->code = _TPair(cdr(sc->code));
		  goto BEGIN1;
		}
	    }
	  
	  if (is_symbol(car(sc->code)))
	    {
	      push_stack(sc, OP_LET_STAR1, sc->code, cadr(sc->code));
	      sc->code = cadr(caadr(sc->code));
	    }
	  else
	    {
	      push_stack(sc, OP_LET_STAR1, sc->code, car(sc->code));
	      /* args is the let body, saved for later, code is the list of vars+initial-values */
	      sc->code = cadr(caar(sc->code));
	      /* caar(code) = first var/val pair, we've checked that all these guys are legit, so cadr of that is the value */
	    }
	  goto EVAL;
	  
	case OP_LET_STAR1:    /* let* -- calculate parameters */
	  /* we can't skip (or reuse) this new frame -- we have to imitate a nested let, otherwise
	   *   (let ((f1 (lambda (arg) (+ arg 1))))
	   *     (let* ((x 32)
	   *            (f1 (lambda (arg) (f1 (+ x arg)))))
	   *       (f1 1)))
	   * will hang.
	   * To get around this requires find_symbol or s7_tree_memq in check_let_star, 
	   *   both (much) more expensive than making a useless frame!.
	   */
	  while (true)
	    {
	      new_frame_with_slot(sc, sc->envir, sc->envir, caar(sc->code), sc->value);
	      sc->code = cdr(sc->code);
	      if (is_pair(sc->code))
		{
		  s7_pointer x;
		  x = cadar(sc->code);
		  if (is_pair(x))
		    {
		      push_stack(sc, OP_LET_STAR1, sc->args, sc->code);
		      sc->code = x;
		      if (is_optimized(x))
			goto OPT_EVAL;
		      goto EVAL;
		    }
		  if (is_symbol(x))
		    sc->value = find_symbol_checked(sc, x);
		  else sc->value = _NFre(x);
		}
	      else break;
	    }
	  sc->code = sc->args; /* original sc->code set in push_stack above */
	  if (is_symbol(car(sc->code)))
	    {
	      /* now we need to declare the new function */
	      s7_pointer body, args;
	      body = cddr(sc->code);
	      args = cadr(sc->code);
	      make_slot_1(sc, sc->envir, car(sc->code), make_closure(sc, args, body, 
								     T_CLOSURE_STAR | ((is_safe_closure(body)) ? T_SAFE_CLOSURE : 0),
								     (is_null(args)) ? 0 : CLOSURE_ARITY_NOT_SET));
	      sc->code = body;
	    }
	  else sc->code = _TPair(cdr(sc->code));
	  goto BEGIN1;
	  
	  
	  /* -------------------------------- LETREC -------------------------------- */
	case OP_LETREC:
	  check_letrec(sc, true);
	  
	case OP_LETREC_UNCHECKED:
	  /* get all local vars and set to #<undefined>
	   * get parallel list of values
	   * eval each member of values list with env still full of #<undefined>'s
	   * assign each value to its variable
	   * eval body
	   *
	   * which means that (letrec ((x x)) x) is not an error!
	   * but this assumes the environment is not changed by evaluating the exprs?
	   * (letrec ((a (define b 1))) b) -- if let, the define takes place in the calling env, not the current env
	   * (letrec ((f1 (lambda (x) (f2 (* 2 x))))) (define (f2 y) (- y 1)) (f1 3)) -> 5 (Guile says unbound f2)
	   *
	   * I think I need to check here that slot_pending_value is set (using the is_checked bit below).
	   */
	  sc->envir = new_frame_in_env(sc, sc->envir);
	  if (is_pair(car(sc->code)))
	    {
	      s7_pointer x;
	      for (x = car(sc->code); is_not_null(x); x = cdr(x))
		{
		  s7_pointer slot;
		  slot = make_slot_1(sc, sc->envir, caar(x), sc->undefined);
		  slot_set_pending_value(slot, sc->undefined);
		  slot_set_expression(slot, cadar(x));
		  set_checked_slot(slot);
		}
	      sc->args = let_slots(sc->envir);
	      push_stack(sc, OP_LETREC1, sc->args, sc->code);
	      sc->code = slot_expression(sc->args);
	      goto EVAL;
	    }
	  sc->code = _TPair(cdr(sc->code));
	  goto BEGIN1;
	  
	case OP_LETREC1:
	  slot_set_pending_value(sc->args, sc->value);
	  sc->args = next_slot(sc->args);
	  if (is_slot(sc->args))
	    {
	      push_stack(sc, OP_LETREC1, sc->args, sc->code);
	      sc->code = slot_expression(sc->args);
	      goto EVAL;
	    }
	  else
	    {
	      s7_pointer slot;
	      for (slot = let_slots(sc->envir); is_slot(slot); slot = next_slot(slot))
		if (is_checked_slot(slot))
		  slot_set_value(slot, slot_pending_value(slot));
	      sc->code = _TPair(cdr(sc->code));
	      goto BEGIN1;
	    }
	  
	  
	  /* -------------------------------- LETREC* -------------------------------- */
	case OP_LETREC_STAR:
	  check_letrec(sc, false);
	  
	case OP_LETREC_STAR_UNCHECKED:
	  /* get all local vars and set to #<undefined>
	   * eval each member of values list and assign immediately, as in let*
	   * eval body
	   */
	  sc->envir = new_frame_in_env(sc, sc->envir);
	  if (is_pair(car(sc->code)))
	    {
	      s7_pointer x, p, q;
	      for (x = car(sc->code); is_not_null(x); x = cdr(x))
		{
		  s7_pointer slot;
		  slot = make_slot_1(sc, sc->envir, caar(x), sc->undefined);
		  slot_set_expression(slot, cadar(x));
		}
	      /* these are reversed, and for letrec*, they need to be in order, so... (reverse_in_place on the slot list) */
	      p = let_slots(sc->envir);
	      x = sc->nil;
	      while (is_slot(p))
		{
		  q = next_slot(p);
		  set_next_slot(p, x);
		  x = p;
		  p = q;
		}
	      let_set_slots(sc->envir, x);
	      sc->args = let_slots(sc->envir);
	      push_stack(sc, OP_LETREC_STAR1, sc->args, sc->code);
	      sc->code = slot_expression(sc->args);
	      goto EVAL;
	    }
	  sc->code = _TPair(cdr(sc->code));
	  goto BEGIN1;
	  
	case OP_LETREC_STAR1:
	  {
	    s7_pointer slot;
	    slot = sc->args;
	    slot_set_value(slot, sc->value);
	    slot = next_slot(slot);
	    if (is_slot(slot))
	      {
		push_stack(sc, OP_LETREC_STAR1, slot, sc->code);
		sc->code = slot_expression(slot);
		goto EVAL;
	      }
	    else
	      {
		sc->code = _TPair(cdr(sc->code));
		goto BEGIN1;
	      }
	  }


	  /* -------------------------------- LET-TEMPORARILY -------------------------------- */
	case OP_LET_TEMPORARILY:
	  check_let_temporarily(sc);
	  
	case OP_LET_TEMP_UNCHECKED:
	  push_stack(sc, OP_GC_PROTECT, sc->args = list_4(sc, car(sc->code), sc->nil, sc->nil, sc->nil), sc->code);
	  /* sc->args: varlist, settees, old_values, new_values */
	  goto LET_TEMP_INIT1;

	  case OP_LET_TEMP_INIT1:
	    caddr(sc->args) = cons(sc, sc->value, caddr(sc->args));

	  LET_TEMP_INIT1:
	    while (is_pair(car(sc->args)))
	      {
		/* eval car, add result to old-vals list, if any vars undefined, error */
		s7_pointer binding, settee, new_value;
		binding = caar(sc->args);
		settee = car(binding);
		new_value = cadr(binding);
		cadr(sc->args) = cons(sc, settee, cadr(sc->args));
		cadddr(sc->args) = cons(sc, new_value, cadddr(sc->args));
		car(sc->args) = cdar(sc->args);
		if (is_symbol(settee))
		  caddr(sc->args) = cons(sc, find_symbol_checked(sc, settee), caddr(sc->args));
		else
		  {
		    if (is_pair(settee))
		      {
			push_stack(sc, OP_LET_TEMP_INIT1, sc->args, sc->code);
			sc->code = settee;
			goto EVAL;
		      }
		    else caddr(sc->args) = cons(sc, new_value, caddr(sc->args));
		  }
	      }
	    car(sc->args) = cadr(sc->args);

	case OP_LET_TEMP_INIT2:
	  /* now eval set car new-val, cadr=settees, cadddr= new_values */
	  while (is_pair(car(sc->args)))
	    {
	      s7_pointer settee, new_value;
	      settee = caar(sc->args);
	      new_value = car(cadddr(sc->args));
	      cadddr(sc->args) = cdr(cadddr(sc->args));
	      car(sc->args) = cdar(sc->args);
	      if ((!is_symbol(settee)) ||
		  (symbol_has_accessor(settee)) ||
		  (is_pair(new_value)))
		{
		  push_stack(sc, OP_LET_TEMP_INIT2, sc->args, sc->code);
		  sc->code = list_3(sc, sc->set_symbol, settee, new_value);
		  goto EVAL;
		}
	      if (is_symbol(new_value))
		new_value = find_symbol_checked(sc, new_value);
	      slot_set_value(find_symbol(sc, settee), new_value);
	    }

	  car(sc->args) = cadr(sc->args);
	  pop_stack(sc);
	  push_stack(sc, OP_LET_TEMP_DONE, sc->args, sc->code);
	  sc->code = cdr(sc->code);
	  if (is_pair(sc->code))
	    goto BEGIN1;
	  else sc->value = sc->nil; /* so (let-temporarily (<vars)) -> () like begin I guess */

	case OP_LET_TEMP_DONE:
	  push_stack(sc, OP_GC_PROTECT, sc->args, sc->value);
	  
	case OP_LET_TEMP_DONE1:
	  while (is_pair(car(sc->args)))
	    {
	      s7_pointer settee, old_value;
	      settee = caar(sc->args);
	      old_value = caaddr(sc->args);
	      caddr(sc->args) = cdaddr(sc->args);
	      car(sc->args) = cdar(sc->args);
	      if ((!is_symbol(settee)) ||
		  (symbol_has_accessor(settee)))
		{
		  push_stack(sc, OP_LET_TEMP_DONE1, sc->args, sc->code);
		  if ((is_pair(old_value)) || (is_symbol(old_value)))
		    sc->code = list_3(sc, sc->set_symbol, settee, list_2(sc, sc->quote_symbol, old_value));
		  else sc->code = list_3(sc, sc->set_symbol, settee, old_value);
		  goto EVAL;
		}
	      slot_set_value(find_symbol(sc, settee), old_value);
	    }
	  pop_stack(sc);
	  sc->value = sc->code;
	  if (is_multiple_value(sc->value))
	    sc->value = splice_in_values(sc, multiple_value(sc->value));
	  break;

	  
	  /* -------------------------------- COND -------------------------------- */
	case OP_COND:
	  check_cond(sc);
	  
	case OP_COND_UNCHECKED:
	  push_stack_no_args(sc, OP_COND1, sc->code);
	  sc->code = caar(sc->code);
	  goto EVAL;
	  
	case OP_COND_UNCHECKED_Z:
	  push_stack_no_args(sc, OP_COND1, sc->code);
	  sc->code = caar(sc->code);
	  goto OPT_EVAL_CHECKED;
	  
	case OP_COND1:
	  if (is_true(sc, sc->value))
	    {
	      COND1:
	      sc->code = cdar(sc->code);
	      if (is_pair(sc->code))
		{
		  if (is_null(cdr(sc->code)))
		    {
		      sc->code = car(sc->code);
		      goto EVAL;
		    }
		  /* check_cond catches stray dots */
		  if ((car(sc->code) == sc->feed_to_symbol) &&
		      (s7_symbol_value(sc, sc->feed_to_symbol) == sc->undefined))
		    {
		      int32_t res;
		      res = feed_to(sc);
		      if (res == goto_START) goto START;
		      if (res == goto_APPLY) goto APPLY;
		      goto EVAL;
		    }
		  push_stack_no_args(sc, OP_BEGIN1, _TPair(cdr(sc->code)));
		  sc->code = car(sc->code);
		  goto EVAL;
		}
	      if (is_null(sc->code))
		{
		  if (is_multiple_value(sc->value))                             /* (+ 1 (cond ((values 2 3)))) */
		    sc->value = splice_in_values(sc, multiple_value(sc->value));
		  /* no result clause, so return test, (cond (#t)) -> #t, (cond ((+ 1 2))) -> 3 */
		  goto START;
		}
	    }
	  sc->code = cdr(sc->code);
	  if (is_null(sc->code))
	    {
	      sc->value = sc->unspecified; /* changed 31-Dec-15 */
	      /* r7rs sez the value if no else clause is unspecified, and this choice makes cond consistent with if and case,
	       *   and rewrite choices between the three are simpler if they are consistent.
	       */
	      goto START;
	    }

	  if ((caar(sc->code) == sc->else_symbol) &&
	      (symbol_id(sc->else_symbol) == 0))
	    {
	      sc->value = sc->else_symbol;
	      goto COND1;
	    }

	  push_stack_no_args(sc, OP_COND1, sc->code);
	  sc->code = caar(sc->code);
	  goto EVAL;

	case OP_COND1_1:
	  sc->code = sc->value;
	  goto APPLY;
	  
	COND_SIMPLE:
	case OP_COND_SIMPLE: /* no => */
	  sc->value = caar(sc->code);
	  if (is_pair(sc->value))
	    {
	      push_stack_no_args(sc, OP_COND1_SIMPLE, sc->code);
	      sc->code = sc->value;
	      goto EVAL;
	    }
	  if ((is_symbol(sc->value)) &&
	      ((sc->value != sc->else_symbol) || (symbol_id(sc->else_symbol) != 0)))
	    sc->value = find_symbol_checked(sc, sc->value);
	  
	case OP_COND1_SIMPLE:
	  if (is_true(sc, sc->value))
	    {
	      sc->code = _TLst(cdar(sc->code));
	      if (is_null(sc->code))
		{
		  if (is_multiple_value(sc->value))
		    sc->value = splice_in_values(sc, multiple_value(sc->value));
		  goto START;
		}
	      goto BEGIN1;
	    }
	  sc->code = cdr(sc->code);
	  if (is_null(sc->code))
	    {
	      sc->value = sc->unspecified; 
	      goto START;
	    }
	  goto COND_SIMPLE;
	  
	case OP_COND_ALL_X:
	  {
	    s7_pointer p;
	    for (p = sc->code; is_pair(p); p = cdr(p))
	      {
		sc->value = c_call(car(p))(sc, caar(p));
		if (is_true(sc, sc->value))
		  {
		    sc->code = _TLst(cdar(p));
		    if (is_null(sc->code))
		      goto START;
		    goto BEGIN1; 
		  }
	      }
	    sc->value = sc->unspecified; 
	  }
	  break;
	  
	case OP_COND_ALL_X_2:
	  {
	    s7_pointer p;
	    p = sc->code;
	    sc->value = c_call(car(p))(sc, caar(p));
	    if (!is_true(sc, sc->value))
	      {
		p = cdr(p);
		sc->value = c_call(car(p))(sc, caar(p));
		if (!is_true(sc, sc->value))
		  {
		    sc->value = sc->unspecified; 
		    goto START;
		  }
	      }
	    sc->code = _TLst(cdar(p));
	    if (is_null(sc->code))
	      goto START;
	    goto BEGIN1;
	  }
	  
	case OP_COND_ALL_X_Z:
	  {
	    s7_pointer p;
	    for (p = sc->code; is_pair(p); p = cdr(p))
	      {
		sc->value = c_call(car(p))(sc, caar(p));
		if (is_true(sc, sc->value))
		  {
		    sc->code = _TLst(cdar(p));
		    if (is_null(sc->code))
		      goto START;
		    sc->code = car(sc->code);
		    goto OPT_EVAL_CHECKED;
		  }
	      }
	    sc->value = sc->unspecified; 
	  }
	  break;

	  
	  /* -------------------------------- AND -------------------------------- */
	case OP_AND:
	  check_and(sc);
	  if (is_null(sc->code))
	    {
	      sc->value = sc->T;
	      goto START;
	    }
	  goto AND1;
	  
	case OP_AND1:
	  if ((is_false(sc, sc->value)) ||
	      (is_null(sc->code)))
	    goto START;
	  
	AND1:
	case OP_AND_UNCHECKED:
	  {
	    s7_pointer p;
	    p = car(sc->code);
	    if (!is_pair(p))
	      {
		if (is_symbol(p))
		  sc->value = find_global_symbol_checked(sc, p);
		else sc->value = p;
		
		if ((is_false(sc, sc->value)) ||
		    (is_null(cdr(sc->code))))
		  goto START;
		
		sc->code = cdr(sc->code);
		goto AND1;
	      }
	    
	    if (is_not_null(cdr(sc->code)))
	      push_stack_no_args(sc, OP_AND1, cdr(sc->code));
	    sc->code = p;
	    if (is_optimized(p))
	      goto OPT_EVAL;
	    goto EVAL;
	  }
	  
	case OP_AND_P1:
	  if ((is_false(sc, sc->value)) ||
	      (is_null(sc->code)))
	    goto START;
	  /* fall through */
	  
	AND_P:
	case OP_AND_P:
	  if (c_callee(sc->code)) /* all c_callee's are set via all_x_eval which can return nil */
	    {
	      sc->value = c_call(sc->code)(sc, car(sc->code));
	      if (is_false(sc, sc->value))
		goto START;
	      sc->code = cdr(sc->code);
	      if (is_null(sc->code))
		goto START;
	      goto AND_P;
	    }
	  else
	    {
	      if (is_not_null(cdr(sc->code)))
		push_stack_no_args(sc, OP_AND_P1, cdr(sc->code));
	      sc->code = car(sc->code);
	      goto EVAL;
	    }
	  
	AND_SAFE_P:            /* all branches all_x_safe */
	case OP_AND_SAFE_P:
	  sc->value = c_call(sc->code)(sc, car(sc->code));
	  if (is_false(sc, sc->value))
	    goto START;
	  sc->code = cdr(sc->code);
	  if (is_null(sc->code))
	    goto START;
	  goto AND_SAFE_P;

	case OP_AND_AP:
	  /* we know c_callee is set on sc->code, and there are only two branches */
	  if (is_false(sc, c_call(sc->code)(sc, car(sc->code))))
	    {
	      sc->value = sc->F;
	      goto START;
	    }
	  sc->code = cadr(sc->code);
	  goto EVAL;

	case OP_AND_AZ:
	  if (is_false(sc, c_call(sc->code)(sc, car(sc->code))))
	    {
	      sc->value = sc->F;
	      goto START;
	    }
	  sc->code = cadr(sc->code);
	  goto OPT_EVAL_CHECKED;

	case OP_AND_SAFE_AA:
	  /* we know both c_callee's are set */
	  sc->value = c_call(sc->code)(sc, car(sc->code));
	  if (is_false(sc, sc->value))
	    goto START;
	  sc->value = c_call(cdr(sc->code))(sc, cadr(sc->code));
	  goto START;

	  
	  /* -------------------------------- OR -------------------------------- */
	case OP_OR:
	  check_or(sc);
	  if (is_null(sc->code))
	    {
	      sc->value = sc->F;
	      goto START;
	    }
	  goto OR1;
	  
	case OP_OR1:
	  if ((is_true(sc, sc->value)) ||
	      (is_null(sc->code)))
	    goto START;
	  
	OR1:
	case OP_OR_UNCHECKED:
	  if (!is_pair(car(sc->code)))
	    {
	      sc->value = car(sc->code);
	      if (is_symbol(sc->value))
		sc->value = find_symbol_checked(sc, sc->value);
	      
	      if ((is_true(sc, sc->value)) ||
		  (is_null(cdr(sc->code))))
		goto START;
	      
	      sc->code = cdr(sc->code);
	      goto OR1;
	    }
	  
	  if (is_not_null(cdr(sc->code)))
	    push_stack_no_args(sc, OP_OR1, cdr(sc->code));
	  sc->code = car(sc->code);
	  goto EVAL;
	  
	case OP_OR_P1:
	  if ((is_true(sc, sc->value)) ||
	      (is_null(sc->code)))
	    goto START;
	  /* fall through */
	  
	OR_P:
	case OP_OR_P:
	  if (c_callee(sc->code))
	    {
	      sc->value = c_call(sc->code)(sc, car(sc->code));
	      if (is_true(sc, sc->value))
		goto START;
	      sc->code = cdr(sc->code);
	      if (is_null(sc->code))
		goto START;
	      goto OR_P;
	    }
	  else
	    {
	      if (is_not_null(cdr(sc->code)))
		push_stack_no_args(sc, OP_OR_P1, cdr(sc->code));
	      sc->code = car(sc->code);
	      goto EVAL;
	    }
	  
	OR_SAFE_P:
	case OP_OR_SAFE_P:
	  sc->value = c_call(sc->code)(sc, car(sc->code));
	  if (is_true(sc, sc->value))
	    goto START;
	  sc->code = cdr(sc->code);
	  if (is_null(sc->code))
	    goto START;
	  goto OR_SAFE_P;

	case OP_OR_AP:
	  /* we know c_callee is set on sc->code, and there are only two branches */
	  sc->value = c_call(sc->code)(sc, car(sc->code));
	  if (is_true(sc, sc->value))
	    goto START;
	  sc->code = cadr(sc->code);
	  goto EVAL;

	case OP_OR_SAFE_AA:
	  sc->value = c_call(sc->code)(sc, car(sc->code));
	  if (is_true(sc, sc->value))
	    goto START;
	  sc->value = c_call(cdr(sc->code))(sc, cadr(sc->code));
	  goto START;

	  /* by going direct without a push_stack on the last one we get tail calls,
	   *   but if the last arg (also in "and" above) is "values", there is a slight
	   *   inconsistency: the values are returned and spliced into the caller if trailing, but
	   *   are spliced into the "or" if not trailing, so
	   *
	   *   (+ 10 (or (values 1 2) #f))
	   *   11
	   *   (+ 10 (or #f (values 1 2)))
	   *   13
	   *   (+ 10 (or (or #f (values 1 2)) #f))
	   *   11
	   *
	   * The tail recursion is more important.  This behavior matches that of "begin" -- if the
	   * values statement is last, it splices into the next outer arglist.
	   */
	  
	  
	  /* -------------------------------- macro evaluation -------------------------------- */
	case OP_EVAL_MACRO:    /* after (scheme-side) macroexpansion, evaluate the resulting expression */
	  /*
	   * (define-macro (hi a) `(+ ,a 1))
	   * (hi 2)
	   * here with value: (+ 2 1)
	   */
	  if (is_multiple_value(sc->value))
	    {
	      /* a normal macro's result is evaluated (below) and its value replaces the macro invocation,
	       *   so if a macro returns multiple values, evaluate each one, then replace the macro
	       *   invocation with (apply values evaluated-results-in-a-list).  We need to save the
	       *   new list of results, and where we are in the macro's output list, so code=macro output,
	       *   args=new list.  If it returns (values), should we use #<unspecified>?  I think that
	       *   happens now without generating a multiple_value object:
	       *       (define-macro (hi) (values)) (hi) -> #<unspecified>
	       *
	       *   (define-macro (ho) (values '(+ 1 2) '(* 3 4))) (+ 1 (ho) 3) -> 19
	       *   (define-macro (ha) (values '(define a 1) '(define b 2))) (let () (ha) (+ a b)) -> 3
	       */
	      push_stack(sc, OP_EVAL_MACRO_MV, sc->nil, cdr(sc->value));
	      sc->code = car(sc->value);
	    }
	  else sc->code = sc->value;
	  goto EVAL;
	  
	case OP_EVAL_MACRO_MV:
	  if (is_null(sc->code)) /* end of values list */
	    {
	      sc->value = splice_in_values(sc, multiple_value(safe_reverse_in_place(sc, cons(sc, sc->value, sc->args))));
	      goto START;
	    }
	  push_stack(sc, OP_EVAL_MACRO_MV, cons(sc, sc->value, sc->args), cdr(sc->code));
	  sc->code = car(sc->code);
	  goto EVAL;
	  
	case OP_EXPANSION:
	  /* fprintf(stderr, "expansion: %s\n", DISPLAY(sc->value)); */
	  /* after the expander has finished, if a list was returned, we need to add some annotations.
	   *   if the expander returned (values), the list-in-progress vanishes! (This mimics map and *#readers*).
	   */
	  if (sc->value == sc->no_value)
	    sc->stack_end[-1] = (s7_pointer)OP_READ_NEXT;
	  else 
	    {
	      if (is_pair(sc->value))
		annotate_expansion(sc->value);
	    }
	  break;
	  
	case OP_DEFINE_MACRO_WITH_ACCESSOR:
	  if (sc->value == sc->error_symbol) /* backwards compatibility... */
	    return(s7_error(sc, sc->error_symbol, set_elist_3(sc, s7_make_string_wrapper(sc, "can't define-macro ~S to ~S"), car(sc->args), cadr(sc->args))));
	  sc->code = sc->value;
	  if ((!is_pair(sc->code)) ||
	      (!is_pair(car(sc->code))) ||
	      (!is_symbol(caar(sc->code))))
	    eval_error(sc, "define-macro: ~S does not look like a macro?", sc->code);
	  sc->value = make_macro(sc, sc->cur_op);
	  break;
	  
	case OP_DEFINE_BACRO:
	case OP_DEFINE_BACRO_STAR:
	case OP_DEFINE_EXPANSION:
	case OP_DEFINE_MACRO:
	case OP_DEFINE_MACRO_STAR:
	  check_define_macro(sc, sc->cur_op);
	  if (symbol_has_accessor(caar(sc->code)))
	    {
	      s7_pointer x;
	      x = find_symbol(sc, caar(sc->code));
	      if ((is_slot(x)) &&
		  (slot_has_accessor(x)))
		{
		  sc->value = bind_accessed_symbol(sc, OP_DEFINE_MACRO_WITH_ACCESSOR, caar(sc->code), sc->code);
		  if (sc->value == sc->no_value)
		    goto APPLY;
		  sc->code = sc->value;
		}
	    }
	  sc->value = make_macro(sc, sc->cur_op);
	  break;
	  
	case OP_LAMBDA:
	  check_lambda(sc);
	  
	case OP_LAMBDA_UNCHECKED:     /* pre-calculating type/arity in check_lambda was slower?? */
	  make_closure_with_let(sc, sc->value, car(sc->code), cdr(sc->code), sc->envir, CLOSURE_ARITY_NOT_SET);  /* sc->value=new closure cell, car=args, cdr=body */
	  break;
	  
	case OP_LAMBDA_STAR:
	  check_lambda_star(sc);
	  
	case OP_LAMBDA_STAR_UNCHECKED:
	  sc->value = make_closure(sc, car(sc->code), cdr(sc->code), T_CLOSURE_STAR | ((is_safe_closure(cdr(sc->code))) ? T_SAFE_CLOSURE : 0), CLOSURE_ARITY_NOT_SET);
	  break;
	  
	  
	  /* -------------------------------- CASE -------------------------------- */
	case OP_CASE:      /* case, car(sc->code) is the selector */
	  check_case(sc);
	  
	case OP_CASE_UNCHECKED:
	  {
	    s7_pointer carc;
	    carc = car(sc->code);
	    if (!is_pair(carc))
	      {
		if (is_symbol(carc))
		  sc->value = find_symbol_checked(sc, carc);
		else sc->value = carc;
		/* fall through */
	      }
	    else
	      {
		push_stack_no_args(sc, OP_CASE_G_G, sc->code);
		sc->code = carc;
		goto EVAL;
	      }
	  }
	  
	CASE_G_G:
	case OP_CASE_G_G:
	  {
	    s7_pointer x, y;
	    sc->code = cdr(sc->code);
	    if (is_simple(sc->value))
	      {
		for (x = sc->code; is_pair(x); x = cdr(x))
		  {
		    y = caar(x);
		    if (!is_pair(y))
		      goto ELSE_CASE;
		    do {
		      if (car(y) == sc->value)
			goto ELSE_CASE;
		      y = cdr(y);
		    } while (is_pair(y));
		  }
	      }
	    else
	      {
		for (x = sc->code; is_pair(x); x = cdr(x))
		  {
		    y = caar(x);
		    if (!is_pair(y))
		      goto ELSE_CASE;
		    for (; is_pair(y); y = cdr(y))
		      if (s7_is_eqv(car(y), sc->value))
			goto ELSE_CASE;
		  }
	      }
	    /* x is the entire matching clause, (case 2 ((2) 3)), x: (((2) 3)) */
	  ELSE_CASE:
	    if (is_not_null(x))
	      {
		sc->code = _TLst(cdar(x));
		if (is_null(sc->code))  /* sc->value is already the selector */
		  goto START;
		
		if (is_null(cdr(sc->code)))
		  {
		    sc->code = car(sc->code);
		    goto EVAL;
		  }
		/* check for => */
		if ((car(sc->code) == sc->feed_to_symbol) &&
		    (s7_symbol_value(sc, sc->feed_to_symbol) == sc->undefined))
		  {
		    int32_t res;
		    res = feed_to(sc);
		    if (res == goto_START) goto START;
		    if (res == goto_APPLY) goto APPLY;
		    goto EVAL;

		    /* sc->code = list_2(sc, cadr(sc->code), list_2(sc, sc->quote_symbol, sc->value)); */
		  }
		goto BEGIN1;
	      }
	    
	    /* no match found */
	    sc->value = sc->unspecified; /* this was sc->nil but the spec says case value is unspecified if no clauses match */
	  }
	  break;

	  /* selector A, key type: E=eq (symbol/char), I=integer, G=any, S=single keys and single bodies */
	case OP_CASE_A_E_S:
	  sc->value = c_call(sc->code)(sc, car(sc->code));
	  goto CASE_E_S;

	case OP_CASE_A_I_S:
	  sc->value = c_call(sc->code)(sc, car(sc->code));
	  goto CASE_I_S;

	case OP_CASE_A_E_G:
	  sc->value = c_call(sc->code)(sc, car(sc->code));
	  goto CASE_E_G;

	case OP_CASE_A_G_G:
	  sc->value = c_call(sc->code)(sc, car(sc->code));
	  goto CASE_G_G;

	case OP_CASE_S_G_S:
	  sc->value = find_symbol_checked(sc, car(sc->code));
	  goto CASE_G_S;

	case OP_CASE_S_G_G:
	  sc->value = find_symbol_checked(sc, car(sc->code));
	  goto CASE_G_G;

	  /* selector = any */
	case OP_CASE_P_E_S:
	  push_stack_no_args(sc, OP_CASE_E_S, sc->code);
	  sc->code = car(sc->code);
	  goto EVAL;

	case OP_CASE_P_I_S:
	  push_stack_no_args(sc, OP_CASE_I_S, sc->code);
	  sc->code = car(sc->code);
	  goto EVAL;

	case OP_CASE_P_G_S:
	  push_stack_no_args(sc, OP_CASE_G_S, sc->code);
	  sc->code = car(sc->code);
	  goto EVAL;

	case OP_CASE_P_E_G:
	  push_stack_no_args(sc, OP_CASE_E_G, sc->code);
	  sc->code = car(sc->code);
	  goto EVAL;

	case OP_CASE_P_G_G:
	  push_stack_no_args(sc, OP_CASE_G_G, sc->code);
	  sc->code = car(sc->code);
	  goto EVAL;

	case OP_CASE_S_E_S:
	  sc->value = find_symbol_checked(sc, car(sc->code));
	  /* goto CASE_E_S; */

	CASE_E_S:
	case OP_CASE_E_S:
	  {
	    s7_pointer x, selector;
	    selector = sc->value;
	    if (is_simple(selector))
	      {
	        for (x = cdr(sc->code); is_pair(x); x = cdr(x))
		  if (opt_any2(x) == selector)
		    {
 	              sc->code = opt_clause(x);
		      goto EVAL;
		    }
	      }
	    sc->code = opt_any3(sc->code);
	    goto EVAL;
	  }
	  break;

	case OP_CASE_S_I_S:
	  sc->value = find_symbol_checked(sc, car(sc->code));
	  /* goto CASE_I_S; */

	CASE_I_S:
	case OP_CASE_I_S:
	  {
	    s7_pointer x, selector, else_clause;
	    selector = sc->value;
	    else_clause = opt_any3(sc->code);
	    if (else_clause != sc->unspecified)
	      {
		if (is_integer(selector))
		  {
		    s7_int val;
		    val = integer(selector);
		    for (x = cdr(sc->code); is_pair(x); x = cdr(x))
		      {
			if (is_integer(opt_any2(x)))
			  {
			    if (integer(opt_any2(x)) == val)
			      {
				sc->code = opt_clause(x);
				goto EVAL;
			      }
			  }
			else break;
		      }
		  }
		sc->code = else_clause;
		goto EVAL;
	      }
	    else
	      {
		if (is_integer(selector))
		  {
		    s7_int val;
		    val = integer(selector);
		    for (x = cdr(sc->code); is_pair(x); x = cdr(x))
		      {
			if (integer(opt_any2(x)) == val)
			  {
			    sc->code = opt_clause(x);
			    goto EVAL;
			  }
		      }
		  }
		sc->value = sc->unspecified;
	      }
	  }
	  break;

	case OP_CASE_A_G_S:
	  sc->value = c_call(sc->code)(sc, car(sc->code));
	  /* goto CASE_G_S; */

	CASE_G_S:
	case OP_CASE_G_S:
	  {
	    s7_pointer x, selector;
	    selector = sc->value;
	    for (x = cdr(sc->code); is_pair(x); x = cdr(x))
	      if (s7_is_eqv(opt_any2(x), selector))
		{
		  sc->code = opt_clause(x);
		  goto EVAL;
		}
	    sc->code = opt_any3(sc->code);
	    goto EVAL;
	  }
	  break;

	case OP_CASE_S_E_G:
	  sc->value = find_symbol_checked(sc, car(sc->code));
	  /* goto CASE_E_G; */

	CASE_E_G:
	case OP_CASE_E_G:
	  {
	    s7_pointer x, y, selector;
	    selector = sc->value;
	    if (is_simple(selector))
	      {
		for (x = cdr(sc->code); is_pair(x); x = cdr(x))
		  {
		    y = opt_any2(x);
		    if (!is_pair(y))
		      goto ELSE_CASE_1;
		    do {
		      if (car(y) == selector)
			goto ELSE_CASE_1;
		      y = cdr(y);
		    } while (is_pair(y));
		  }
	      }
	    else 
	      {
		sc->code = opt_any3(sc->code);
		if (is_pair(sc->code))
		  goto ELSE_CASE_2;
		goto START;
	      }
	    /* x is the entire matching clause, (case 2 ((2) 3)), x: (((2) 3)) */
	  ELSE_CASE_1:
	    if (is_not_null(x))
	      {
		sc->code = _TLst(cdar(x));
		if (is_null(sc->code))  /* sc->value is already the selector */
		  goto START;

	      ELSE_CASE_2:
		if (is_null(cdr(sc->code)))
		  {
		    sc->code = car(sc->code);
		    goto EVAL;
		  }
		/* check for => */
		if ((car(sc->code) == sc->feed_to_symbol) &&
		    (s7_symbol_value(sc, sc->feed_to_symbol) == sc->undefined))
		  {
		    int32_t res;
		    res = feed_to(sc);
		    if (res == goto_START) goto START;
		    if (res == goto_APPLY) goto APPLY;
		    goto EVAL;
		    /* sc->code = list_2(sc, cadr(sc->code), list_2(sc, sc->quote_symbol, sc->value)); */
		  }
		push_stack_no_args(sc, OP_BEGIN1, _TPair(cdr(sc->code)));
		sc->code = car(sc->code);
		goto EVAL;
	      }
	    sc->value = sc->unspecified;
	  }
	  break;

	case OP_ERROR_QUIT:
	  if (sc->stack_end <= sc->stack_start)
	    stack_reset(sc); /* sets stack_end to stack_start, then pushes op_barrier and op_eval_done */
	  return(sc->F);
	  break;

	case OP_ERROR_HOOK_QUIT:
	  sc->error_hook = sc->code;  /* restore old value */
	  
	  /* now mimic the end of the normal error handler.  Since this error hook evaluation can happen
	   *   in an arbitrary s7_call nesting, we can't just return from the current evaluation --
	   *   we have to jump to the original (top-level) call.  Otherwise '#<unspecified> or whatever
	   *   is simply treated as the (non-error) return value, and the higher level evaluations
	   *   get confused.
	   */
	  stack_reset(sc);                                 /* is this necessary? */
	  push_stack_op(sc, OP_ERROR_QUIT);                /* added 3-Dec-16: try to make sure we actually exit! */
	  sc->cur_op = OP_ERROR_QUIT;
	  if (sc->longjmp_ok) longjmp(sc->goto_start, ERROR_QUIT_JUMP);
#if DEBUGGING
	  fprintf(stderr, "%d: op_error_hook_quit did not jump, returns %s\n", __LINE__, DISPLAY(sc->value));
#endif
	  return(sc->value); /* not executed I hope */
	  
	case OP_EVAL_DONE:
	  /* this is the "time to quit" operator */
	  return(sc->F);
	  break;
	  
	case OP_BARRIER:
	case OP_CATCH_ALL:
	case OP_CATCH:
	case OP_CATCH_1:
	case OP_CATCH_2:
	  break;
	  
	case OP_DEACTIVATE_GOTO:
	  call_exit_active(sc->args) = false;      /* as we leave the call-with-exit body, deactivate the exiter */
	  break;
	  
	case OP_GET_OUTPUT_STRING:          /* from get-output-string -- return a new string */
	  sc->value = s7_make_string_with_length(sc, (const char *)port_data(sc->code), port_position(sc->code));
	  break;
	  
	case OP_GET_OUTPUT_STRING_1:        /* from call-with-output-string and with-output-to-string -- return the port string directly */
	  if ((!is_output_port(sc->code)) ||
	      (port_is_closed(sc->code)))
	    simple_wrong_type_argument_with_type(sc, sc->with_output_to_string_symbol, sc->code, s7_make_string_wrapper(sc, "an open string output port"));
	  
	  if (port_position(sc->code) >= port_data_size(sc->code))
	    resize_port_data(sc->code, port_position(sc->code) + 1); /* need room for the trailing #\null */
	  sc->value = make_string_uncopied_with_length(sc, (char *)port_data(sc->code), port_position(sc->code));
	  string_value(sc->value)[port_position(sc->code)] = 0;
	  port_data(sc->code) = NULL;
	  port_data_size(sc->code) = 0;
	  port_needs_free(sc->code) = false;
	  /* fall through */
	  
	case OP_UNWIND_OUTPUT:
	  unwind_output_ex(sc);
	  break;

	case OP_UNWIND_INPUT:
	  unwind_input_ex(sc);
	  break;
	  
	case OP_DYNAMIC_WIND:
	  if (dynamic_wind_ex(sc) == goto_APPLY) goto APPLY;
	  break;
	  
	  
	  /* -------------------------------- with-let --------------------------------
	   *
	   * the extra set! to pull in args, or fixup the outlet is annoying, but
	   *   but with-let is hard to do right -- what if env is chained as in class/objects?
	   * also, currently a mock-let is an error -- perhaps add the method checks?
	   *   but unless 'values, that would require a 'with-let method (it's not a function)
	   */
	case OP_WITH_LET_S:
	  {
	    s7_pointer e;
	    e = find_symbol_checked(sc, car(sc->code));
	    if (e == sc->rootlet)
	      sc->envir = sc->nil;
	    else
	      {
		s7_pointer p;
		if (!is_let(e))
		  eval_type_error(sc, "with-let takes an environment argument: ~A", e);
		set_with_let_let(e);
		let_id(e) = ++sc->let_number;
		sc->envir = e;
		/* if the let in question has 10,000 names (e.g. *gtk*) this loop (which can't be avoided currently)
		 *   will be noticeable in a few cases.  So, instead of saying (with-let *gtk* ...) use something
		 *   equivalent to (with-let (sublet *gtk*) ...) which is cleaner anyway.  (In my timing tests, even
		 *   when pounding on this one block, the loop only amounts to 1% of the time.  Normally it's
		 *   negligible).
		 */
		for (p = let_slots(e); is_slot(p); p = next_slot(p))
		  {
		    s7_pointer sym;
		    sym = slot_symbol(p);
		    if (symbol_id(sym) != sc->let_number)
		      symbol_set_local(sym, sc->let_number, p);
		  }
	      }
	    sc->code = _TPair(cdr(sc->code));
	    goto BEGIN1;
	  }
	  
	case OP_WITH_LET:
	  check_with_let(sc);
	  
	case OP_WITH_LET_UNCHECKED:
	  sc->value = car(sc->code);
	  if (!is_pair(sc->value))
	    {
	      if (is_symbol(sc->value))
		sc->value = find_symbol_checked(sc, sc->value);
	      sc->code = cdr(sc->code);
	      
	      if (!is_pair(sc->code))
		{
		  if (!is_let(sc->value))            /* (with-let e abs) */
		    eval_type_error(sc, "with-let takes an environment argument: ~A", sc->value);
		  if (is_symbol(sc->code))
		    sc->value = s7_symbol_local_value(sc, sc->code, sc->value);
		  else sc->value = sc->code;
		  goto START;
		}
	      /* else fall through */
	    }
	  else
	    {
	      push_stack_no_args(sc, OP_WITH_LET1, cdr(sc->code));
	      sc->code = sc->value;                          /* eval env arg */
	      goto EVAL;
	    }
	  
	case OP_WITH_LET1:
	  activate_let(sc, sc->value);
	  goto BEGIN1;
	  
	case OP_WITH_BAFFLE:
	  if (!s7_is_proper_list(sc, sc->code))
	    eval_error(sc, "with-baffle: unexpected dot? ~A", sc->code);
	  
	  if ((!is_null(sc->code)) &&
	      (is_overlaid(sc->code)) &&
	      (has_opt_back(sc->code)))
	    pair_set_syntax_symbol(sc->code, sc->with_baffle_unchecked_symbol);
	  
	case OP_WITH_BAFFLE_UNCHECKED:
	  if (is_null(sc->code))
	    {
	      sc->value = sc->nil;
	      goto START;
	    }
	  new_frame(sc, sc->envir, sc->envir);
	  make_slot_1(sc, sc->envir, sc->baffle_symbol, make_baffle(sc));
	  goto BEGIN1;
	  
	  
	  /* -------------------------------- the reader -------------------------------- */
	POP_READ_LIST:
	  /* push-stack OP_READ_LIST is always no_code and op is always OP_READ_LIST (and not used), sc->envir is apparently not needed here
	   */
	  sc->stack_end -= 4;
	  sc->args = sc->stack_end[2];
	  
	READ_LIST:
	case OP_READ_LIST:        /* sc->args is sc->nil at first */
	  {
	    s7_pointer x;
	    new_cell(sc, x, T_PAIR);
	    set_car(x, sc->value);
	    set_cdr(x, sc->args);
	    sc->args = x;
#if WITH_PROFILE
	    profile_set_location(x, remember_location(port_line_number(sc->input_port), port_file_number(sc->input_port)));
#endif
	  }
	  
	case OP_READ_NEXT:       /* this is 75% of the token calls, so expanding it saves lots of time */
	  {
	    int32_t c;
	    s7_pointer pt;
	    
	    pt = sc->input_port;
	    c = port_read_white_space(pt)(sc, pt);
	    
	  READ_C:
	    switch (c)
	      {
	      case '(':
		c = port_read_white_space(pt)(sc, pt);  /* sc->tok = token(sc) */
		switch (c)
		  {
		  case '(':  sc->tok = TOKEN_LEFT_PAREN;                break;
		  case ')':  sc->value = sc->nil; goto READ_LIST;       /* was tok = TOKEN_RIGHT_PAREN */
		  case '.':  sc->tok = read_dot(sc, pt);                break;
		  case '\'': sc->tok = TOKEN_QUOTE;                     break;
		  case ';':  sc->tok = port_read_semicolon(pt)(sc, pt); break;
		  case '"':  sc->tok = TOKEN_DOUBLE_QUOTE;              break;
		  case '`':  sc->tok = TOKEN_BACK_QUOTE;                break;
		  case ',':  sc->tok = read_comma(sc, pt);              break;
		  case '#':  sc->tok = read_sharp(sc, pt);              break;
		  case '\0': case EOF: sc->tok = TOKEN_EOF;             break;
		    
		  default:
		    {
		      s7_pointer x;
		      sc->strbuf[0] = c;
		      push_stack_no_let_no_code(sc, OP_READ_LIST, sc->args);
		      check_stack_size(sc);
		      sc->value = port_read_name(pt)(sc, pt);
		      new_cell(sc, x, T_PAIR);
		      set_car(x, sc->value);
		      set_cdr(x, sc->nil);
		      sc->args = x;
#if WITH_PROFILE
		      profile_set_location(x, remember_location(port_line_number(sc->input_port), port_file_number(sc->input_port)));
#endif
		      c = port_read_white_space(pt)(sc, pt);
		      goto READ_C;
		    }
		  }
		
		if (sc->tok == TOKEN_ATOM)
		  {
		    s7_pointer x;
		    push_stack_no_let_no_code(sc, OP_READ_LIST, sc->args);
		    check_stack_size(sc);
		    sc->value = port_read_name(pt)(sc, pt);
		    new_cell(sc, x, T_PAIR);
		    set_car(x, sc->value);
		    set_cdr(x, sc->nil);
		    sc->args = x;
#if WITH_PROFILE
		    profile_set_location(x, remember_location(port_line_number(sc->input_port), port_file_number(sc->input_port)));
#endif
		    c = port_read_white_space(pt)(sc, pt);
		    goto READ_C;
		  }
		
		if (sc->tok == TOKEN_RIGHT_PAREN)
		  {
		    sc->value = sc->nil;
		    goto READ_LIST;
		  }
		
		if (sc->tok == TOKEN_DOT)
		  {
		    do {c = inchar(pt);} while ((c != ')') && (c != EOF));
		    return(read_error(sc, "stray dot after '('?"));      /* (car '( . )) */
		  }
		
		if (sc->tok == TOKEN_EOF)
		  return(missing_close_paren_error(sc));
		
		push_stack_no_let_no_code(sc, OP_READ_LIST, sc->args);
		push_stack_no_let_no_code(sc, OP_READ_LIST, sc->nil);
		check_stack_size(sc);
		sc->value = read_expression(sc);
		if (main_stack_op(sc) == OP_READ_LIST) goto POP_READ_LIST;
		goto START;
		
	      case ')':
		sc->tok = TOKEN_RIGHT_PAREN;
		break;
		
	      case '.':
		sc->tok = read_dot(sc, pt); /* dot or atom */
		break;
		
	      case '\'':
		sc->tok = TOKEN_QUOTE;
		push_stack_no_let_no_code(sc, OP_READ_LIST, sc->args);
		sc->value = read_expression(sc);
		goto START;
		
	      case ';':
		sc->tok = port_read_semicolon(pt)(sc, pt);
		break;
		
	      case '"':
		sc->tok = TOKEN_DOUBLE_QUOTE;
		sc->value = read_string_constant(sc, pt);
		if (sc->value == sc->F)                                /* can happen if input code ends in the middle of a string */
		  return(string_read_error(sc, "end of input encountered while in a string"));
		if (sc->value == sc->T)
		  return(read_error(sc, "unknown backslash usage -- perhaps you meant two backslashes?"));
		if (sc->safety > IMMUTABLE_VECTOR_SAFETY) set_immutable(sc->value);
		goto READ_LIST;
		
	      case '`':
		sc->tok = TOKEN_BACK_QUOTE;
		push_stack_no_let_no_code(sc, OP_READ_LIST, sc->args);
		sc->value = read_expression(sc);
		if (main_stack_op(sc) == OP_READ_LIST) goto POP_READ_LIST;
		goto START;
		
	      case ',':
		sc->tok = read_comma(sc, pt); /* at_mark or comma */
		push_stack_no_let_no_code(sc, OP_READ_LIST, sc->args);
		sc->value = read_expression(sc);
		goto START;
		
	      case '#':
		sc->tok = read_sharp(sc, pt);
		break;
		
	      case '\0':
	      case EOF:
		return(missing_close_paren_error(sc));
	      
	      default:
		sc->strbuf[0] = c;
		sc->value = port_read_name(pt)(sc, pt);
		goto READ_LIST;
	      }
	  }
	  
	READ_TOK:
	  switch (sc->tok)
	    {
	    case TOKEN_RIGHT_PAREN:
	      /* sc->args can't be null here */
	      sc->value = safe_reverse_in_place(sc, sc->args);
	      if (is_symbol(car(sc->value)))
		{
		  pair_set_line(sc->value, remember_location(port_line_number(sc->input_port), port_file_number(sc->input_port)));
		  set_has_line_number(sc->value);	      /* sc->input_port above can't be nil(?) -- it falls back on stdin now */
		  
		  if ((is_expansion(car(sc->value))) &&
		      (expansion_ex(sc) == goto_APPLY))
		    {
		      push_stack(sc, OP_EXPANSION, sc->nil, sc->gc_nil);
		      new_frame(sc, closure_let(sc->code), sc->envir);
		      goto APPLY_LAMBDA;
		    }
		  if (is_pair(cdr(sc->value)))
		    {
		      set_opt_back(sc->value);
		      set_overlay(cdr(sc->value));
		    }
		}
	      break;
	      
	    case TOKEN_EOF:      /* can't happen, I believe */
	      return(missing_close_paren_error(sc));
	      
	    case TOKEN_ATOM:
	      sc->value = port_read_name(sc->input_port)(sc, sc->input_port);
	      goto READ_LIST;
	      
	    case TOKEN_SHARP_CONST:
	      sc->value = port_read_sharp(sc->input_port)(sc, sc->input_port);
	      if (is_null(sc->value))
		return(read_error(sc, "undefined # expression"));
	      if (sc->value == sc->no_value)
		{
		  /* (set! *#readers* (cons (cons #\; (lambda (s) (read) (values))) *#readers*))
		   * (+ 1 #;(* 2 3) 4)
		   * so we need to get the next token, act on it without any assumptions about read list
		   */
		  sc->tok = token(sc);
		  goto READ_TOK;
		}
	      goto READ_LIST;
	      
	    case TOKEN_DOUBLE_QUOTE:
	      sc->value = read_string_constant(sc, sc->input_port);
	      if (sc->value == sc->F)                                /* can happen if input code ends in the middle of a string */
		return(string_read_error(sc, "end of input encountered while in a string"));
	      if (sc->value == sc->T)
		return(read_error(sc, "unknown backslash usage -- perhaps you meant two backslashes?"));
	      if (sc->safety > IMMUTABLE_VECTOR_SAFETY) set_immutable(sc->value);
	      goto READ_LIST;
	      
	    case TOKEN_DOT:
	      push_stack_no_let_no_code(sc, OP_READ_DOT, sc->args);
	      sc->tok = token(sc);
	      sc->value = read_expression(sc);
	      break;
	      
	    default:
	      /* by far the main case here is TOKEN_LEFT_PAREN, but it doesn't save anything to move it to this level */
	      push_stack_no_let_no_code(sc, OP_READ_LIST, sc->args);
	      sc->value = read_expression(sc);
	      /* check for op_read_list here and explicit pop_stack are slower */
	      break;
	    }
	  if (main_stack_op(sc) == OP_READ_LIST) goto POP_READ_LIST;
	  break;
	  
	case OP_READ_DOT:
	  if (token(sc) != TOKEN_RIGHT_PAREN)
	    {
	      back_up_stack(sc);
	      read_error(sc, "stray dot?");            /* (+ 1 . 2 3) or (list . ) */
	    }
	  /* args = previously read stuff, value = thing just after the dot and before the ')':
	   *   (list 1 2 . 3)
	   *   value: 3, args: (2 1 list)
	   *   '(1 . 2)
	   *   value: 2, args: (1)
	   *
	   * but we also get here in a lambda arg list:
	   *   (lambda (a b . c) #f)
	   *   value: c, args: (b a)
	   *
	   * so we have to leave any error checks until later, I guess
	   *   -- in eval_args1, if we end with non-pair-not-nil then
	   *      something is fishy
	   */
	  sc->value = reverse_in_place(sc, sc->value, sc->args);
	  pair_set_dotted(sc->value);
	  if (main_stack_op(sc) == OP_READ_LIST) goto POP_READ_LIST;
	  break;
	  
	case OP_READ_QUOTE:
	  /* can't check for sc->value = sc->nil here because we want ''() to be different from '() */
	  if ((sc->safety > IMMUTABLE_VECTOR_SAFETY) &&
	      ((is_pair(sc->value)) || (s7_is_vector(sc->value)) || (is_string(sc->value))))
	    set_immutable(sc->value);
	  sc->value = list_2(sc, sc->quote_symbol, sc->value);
	  set_opt_back(sc->value);
	  set_overlay(cdr(sc->value));
	  if (main_stack_op(sc) == OP_READ_LIST) goto POP_READ_LIST;
	  break;
	  
	case OP_READ_QUASIQUOTE:
	  /* this was pushed when the backquote was seen, then eventually we popped back to it */
	  sc->value = g_quasiquote_1(sc, sc->value);
	  /* doing quasiquote at read time means there are minor inconsistencies in
	   *    various combinations or quote/' and quasiquote/`.  A quoted ` will expand
	   *    but quoted quasiquote will not (` can't be redefined, but quasiquote can).
	   *    see s7test.scm for examples.
	   */
	  if (main_stack_op(sc) == OP_READ_LIST) goto POP_READ_LIST;
	  break;
	  
	case OP_READ_VECTOR:
	  if (is_dotted_pair(sc->value))            /* #(1 . 2) */
	    return(read_error(sc, "vector constant data is not a proper list"));
	  sc->v = sc->value;
	  if (sc->args == small_int(1))             /* sc->args was sc->w earlier from read_sharp */
	    sc->value = g_vector(sc, sc->value);
	  else sc->value = g_multivector(sc, s7_integer(sc->args), sc->value);
	  /* here and below all of the sc->value list can be freed, but my tests showed no speed up even in large cases */
	  free_vlist(sc, sc->v);
	  if (sc->safety > IMMUTABLE_VECTOR_SAFETY) set_immutable(sc->value);
	  if (main_stack_op(sc) == OP_READ_LIST) goto POP_READ_LIST;
	  break;

	case OP_READ_INT_VECTOR:
	  if (is_dotted_pair(sc->value))
	    return(read_error(sc, "vector constant data is not a proper list"));
	  sc->v = sc->value;
	  if (sc->args == small_int(1))             /* sc->args was sc->w earlier from read_sharp */
	    sc->value = g_int_vector(sc, sc->value);
	  else sc->value = g_int_multivector(sc, s7_integer(sc->args), sc->value);
	  free_vlist(sc, sc->v);
	  if (sc->safety > IMMUTABLE_VECTOR_SAFETY) set_immutable(sc->value);
	  if (main_stack_op(sc) == OP_READ_LIST) goto POP_READ_LIST;
	  break;

	case OP_READ_FLOAT_VECTOR:
	  if (is_dotted_pair(sc->value))
	    return(read_error(sc, "vector constant data is not a proper list"));
	  sc->v = sc->value;
	  if (sc->args == small_int(1))             /* sc->args was sc->w earlier from read_sharp */
	    sc->value = g_float_vector(sc, sc->value);
	  else sc->value = g_float_multivector(sc, s7_integer(sc->args), sc->value);
	  free_vlist(sc, sc->v);
	  if (sc->safety > IMMUTABLE_VECTOR_SAFETY) set_immutable(sc->value);
	  if (main_stack_op(sc) == OP_READ_LIST) goto POP_READ_LIST;
	  break;	  

	case OP_READ_BYTE_VECTOR:
	  if (is_dotted_pair(sc->value))
	    return(read_error(sc, "byte-vector constant data is not a proper list"));
	  sc->v = sc->value;
	  sc->value = g_byte_vector(sc, sc->value);
	  free_vlist(sc, sc->v);
	  if (sc->safety > IMMUTABLE_VECTOR_SAFETY) set_immutable(sc->value);
	  if (main_stack_op(sc) == OP_READ_LIST) goto POP_READ_LIST;
	  break;
	  
	case OP_READ_UNQUOTE:
	  /* here if sc->value is a constant, the unquote is pointless (should we complain?) */
	  if ((is_pair(sc->value)) ||
	      (is_symbol(sc->value)))
	    sc->value = list_2(sc, sc->unquote_symbol, sc->value);
	  if (main_stack_op(sc) == OP_READ_LIST) goto POP_READ_LIST;
	  break;
	  
	case OP_READ_APPLY_VALUES:
	  sc->value = list_2(sc, sc->unquote_symbol, list_2(sc, sc->apply_values_symbol, sc->value)); 
	  if (main_stack_op(sc) == OP_READ_LIST) goto POP_READ_LIST;
	  break;
	  
	default:
	  fprintf(stderr, "unknown operator: %" PRIdPTR " in %s\n", sc->cur_op, DISPLAY(current_code(sc)));
#if DEBUGGING
	  fprintf(stderr, "stack size: %u\n", sc->stack_size);
	  if (sc->stack_end < sc->stack_start) 
	    fprintf(stderr, "%sstack underflow%s\n", BOLD_TEXT, UNBOLD_TEXT);
	  if (sc->stack_end >= sc->stack_start + sc->stack_size) 
	    fprintf(stderr, "%sstack overflow%s\n", BOLD_TEXT, UNBOLD_TEXT);
	  abort();
#endif
	  return(sc->F);
	}
    }
  return(sc->F);
}

#if WITH_GCC
#undef new_cell
#if (!DEBUGGING)
#define new_cell(Sc, Obj, Type)			\
  do {									\
    if (Sc->free_heap_top <= Sc->free_heap_trigger) try_to_call_gc(Sc); \
    Obj = (*(--(Sc->free_heap_top)));					\
    set_type(Obj, Type);						\
  } while (0)
#else
#define new_cell(Sc, Obj, Type)			\
  do {									\
    if ((Sc->free_heap_top <= Sc->free_heap_trigger) || (for_any_other_reason(sc, __LINE__))) {last_gc_line = __LINE__; last_gc_func = __func__; try_to_call_gc(Sc);} \
    Obj = (*(--(Sc->free_heap_top)));					\
    Obj->alloc_line = __LINE__;	Obj->alloc_func = __func__; Obj->debugger_bits = 0;		\
    set_type(Obj, Type);						\
  } while (0)
#endif
#endif


/* needed in s7_gmp_init and s7_init, initialized in s7_init before we get to gmp */
static s7_pointer pl_bt, pl_p, pl_bc, pcl_bc, pcl_bs, pl_bn, pl_sf, pcl_bt, pcl_e, pcl_i, pcl_r, pcl_n, pcl_s, pcl_v, pcl_f, pcl_c, pl_tl;




/* -------------------------------- multiprecision arithmetic -------------------------------- */

#if WITH_GMP
static mp_prec_t mpc_precision = DEFAULT_BIGNUM_PRECISION; /* global for libs */
static mp_prec_t mpc_set_default_precision(mp_prec_t prec) {mpc_precision = prec; return(prec);}

#define mpc_init(Z) mpc_init2(Z, mpc_precision)

static void mpc_init_set(mpc_ptr z, mpc_ptr y, mpc_rnd_t rnd)
{
  mpc_init(z);
  mpc_set(z, y, rnd);
}


mpfr_t *s7_big_real(s7_pointer x)    {return(&big_real(x));}
mpz_t  *s7_big_integer(s7_pointer x) {return(&big_integer(x));}
mpq_t  *s7_big_ratio(s7_pointer x)   {return(&big_ratio(x));}
mpc_t  *s7_big_complex(s7_pointer x) {return(&big_complex(x));}

static char *mpfr_to_string(mpfr_t val, int32_t radix)
{
  char *str, *tmp, *str1;
  mp_exp_t expptr;
  int32_t i, len, ep;
  
  if (mpfr_zero_p(val))
    return(copy_string("0.0"));
  
  if (mpfr_nan_p(val))
    return(copy_string("nan.0"));
  if (mpfr_inf_p(val))
    {
      if (mpfr_signbit(val) == 0)
	return(copy_string("inf.0"));
      return(copy_string("-inf.0"));
    }
  
  str1 = mpfr_get_str(NULL, &expptr, radix, 0, val, GMP_RNDN);
  
  /* 0 -> full precision, but it's too hard to make this look like C formatted output.
   *  :(format #f "~,3F" pi)
   *  "3.141592653589793238462643383279502884195E0"
   *  :(format #f "~,3F" 1.1234567890123)  ; not a bignum
   *  "1.123"
   *  :(format #f "~,3F" 1.12345678901234) ; a bignum
   *  "1.123456789012339999999999999999999999999E0"
   * but we don't know the exponent or the string length until after we call mpfr_get_str.
   */
  str = str1;
  ep = (int32_t)expptr;
  len = safe_strlen(str);
  
  /* remove trailing 0's */
  for (i = len - 1; i > 3; i--)
    if (str[i] != '0')
      break;
  if (i < len - 1)
    str[i + 1] = '\0';
  
  len += 64;
  tmp = (char *)malloc(len * sizeof(char));
  
  if (str[0] == '-')
    snprintf(tmp, len, "-%c.%s%c%d", str[1], (char *)(str + 2), (radix <= 10) ? 'E' : '@', ep - 1);
  else snprintf(tmp, len, "%c.%s%c%d", str[0], (char *)(str + 1), (radix <= 10) ? 'E' : '@', ep - 1);
  
  mpfr_free_str(str1);
  return(tmp);
}


static char *mpc_to_string(mpc_t val, int32_t radix, use_write_t use_write)
{
  char *rl, *im, *tmp;
  int32_t len;
  mpfr_t a, b;
  
  mpfr_init(a);
  mpc_real(a, val, GMP_RNDN);
  rl = mpfr_to_string(a, radix);
  mpfr_init(b);
  mpc_imag(b, val, GMP_RNDN);
  im = mpfr_to_string(b, radix);
  
  len = safe_strlen(rl) + safe_strlen(im) + 128;
  tmp = (char *)malloc(len * sizeof(char));
  
  if (use_write == USE_READABLE_WRITE)
    snprintf(tmp, len, "(complex %s %s)", rl, im);
  else snprintf(tmp, len, "%s%s%si", rl, (im[0] == '-') ? "" : "+", im);
  
  free(rl);
  free(im);
  return(tmp);
}


static char *big_number_to_string_with_radix(s7_pointer p, int32_t radix, int32_t width, int32_t *nlen, use_write_t use_write)
{
  char *str = NULL;
  
  switch (type(p))
    {
    case T_BIG_INTEGER: str = mpz_get_str(NULL, radix, big_integer(p));        break;
    case T_BIG_RATIO:   str = mpq_get_str(NULL, radix, big_ratio(p));          break;
    case T_BIG_REAL:    str = mpfr_to_string(big_real(p), radix);              break;
    default:            str = mpc_to_string(big_complex(p), radix, use_write); break;
    }
  
  if (width > 0)
    {
      int32_t len;
      len = safe_strlen(str);
      if (width > len)
	{
	  int32_t spaces;
	  str = (char *)realloc(str, (width + 1) * sizeof(char));
	  spaces = width - len;
	  str[width] = '\0';
	  memmove((void *)(str + spaces), (void *)str, len);
	  memset((void *)str, (int)' ', spaces);
	  (*nlen) = width;
	}
      else (*nlen) = len;
    }
  else (*nlen) = safe_strlen(str);
  return(str);
}


static bool s7_is_one_or_big_one(s7_pointer p)
{
  bool result = false;
  
  if (!is_big_number(p))
    return(s7_is_one(p));
  
  if (is_t_big_integer(p))
    {
      mpz_t n;
      mpz_init_set_si(n, 1);
      result = (mpz_cmp(n, big_integer(p)) == 0);
      mpz_clear(n);
    }
  else
    {
      if (is_t_big_real(p))
	{
	  mpfr_t n;
	  mpfr_init_set_d(n, 1.0, GMP_RNDN);
	  result = (mpfr_cmp(n, big_real(p)) == 0);
	  mpfr_clear(n);
	}
    }
  return(result);
}


static s7_pointer string_to_big_integer(s7_scheme *sc, const char *str, int32_t radix)
{
  s7_pointer x;
  
  new_cell(sc, x, T_BIG_INTEGER);
  add_bigint(sc, x);
  mpz_init_set_str(big_integer(x), (str[0] == '+') ? (const char *)(str + 1) : str, radix);
  return(x);
}


static s7_pointer mpz_to_big_integer(s7_scheme *sc, mpz_t val)
{
  s7_pointer x;
  
  new_cell(sc, x, T_BIG_INTEGER);
  add_bigint(sc, x);
  mpz_init_set(big_integer(x), val);
  return(x);
}


s7_pointer s7_make_big_integer(s7_scheme *sc, mpz_t *val)
{
  return(mpz_to_big_integer(sc, *val));
}


static s7_pointer string_to_big_ratio(s7_scheme *sc, const char *str, int32_t radix)
{
  s7_pointer x;
  mpq_t n;
  
  mpq_init(n);
  mpq_set_str(n, str, radix);
  mpq_canonicalize(n);
  
  if (mpz_cmp_ui(mpq_denref(n), 1) == 0)
    x = mpz_to_big_integer(sc, mpq_numref(n));
  else
    {
      new_cell(sc, x, T_BIG_RATIO);
      add_bigratio(sc, x);
      mpq_init(big_ratio(x));
      mpq_set_num(big_ratio(x), mpq_numref(n));
      mpq_set_den(big_ratio(x), mpq_denref(n));
    }
  mpq_clear(n);
  return(x);
}


static s7_pointer mpq_to_big_ratio(s7_scheme *sc, mpq_t val)
{
  s7_pointer x;
  
  new_cell(sc, x, T_BIG_RATIO);
  add_bigratio(sc, x);
  mpq_init(big_ratio(x));
  mpq_set_num(big_ratio(x), mpq_numref(val));
  mpq_set_den(big_ratio(x), mpq_denref(val));
  return(x);
}


s7_pointer s7_make_big_ratio(s7_scheme *sc, mpq_t *val)
{
  return(mpq_to_big_ratio(sc, *val));
}


static s7_pointer mpz_to_big_ratio(s7_scheme *sc, mpz_t val)
{
  s7_pointer x;
  
  new_cell(sc, x, T_BIG_RATIO);
  add_bigratio(sc, x);
  mpq_init(big_ratio(x));
  mpq_set_num(big_ratio(x), val);
  return(x);
}


static s7_pointer make_big_integer_or_ratio(s7_scheme *sc, s7_pointer z)
{
  if (mpz_cmp_ui(mpq_denref(big_ratio(z)), 1) == 0)
    return(mpz_to_big_integer(sc, mpq_numref(big_ratio(z))));
  return(z);
}


static s7_pointer string_to_big_real(s7_scheme *sc, const char *str, int32_t radix)
{
  s7_pointer x;
  
  new_cell(sc, x, T_BIG_REAL);
  add_bigreal(sc, x);
  mpfr_init_set_str(big_real(x), str, radix, GMP_RNDN);
  return(x);
}

static void mpz_init_set_s7_int(mpz_t n, s7_int uval);

static s7_pointer s7_number_to_big_real(s7_scheme *sc, s7_pointer p)
{
  s7_pointer x;
  
  new_cell(sc, x, T_BIG_REAL);
  add_bigreal(sc, x);
  
  switch (type(p))
    {
    case T_INTEGER:
      if (sizeof(s7_int) == sizeof(long int))
	mpfr_init_set_si(big_real(x), integer(p), GMP_RNDN);
      else mpfr_init_set_ld(big_real(x), (long double)integer(p), GMP_RNDN);
      break;
      
    case T_RATIO:
      /* here we can't use fraction(number(p)) even though that uses long double division because
       *   there are lots of int64_t ratios that will still look the same.
       *   We have to do the actual bignum divide by hand.
       */
      {
	mpq_t rat;
	mpz_t n1, d1;
	
	mpz_init_set_s7_int(n1, numerator(p));
	mpz_init_set_s7_int(d1, denominator(p));
	mpq_init(rat);
	
	mpq_set_num(rat, n1);
	mpq_set_den(rat, d1);
	mpq_canonicalize(rat);
	mpfr_init_set_q(big_real(x), rat, GMP_RNDN);
	
	mpz_clear(n1);
	mpz_clear(d1);
	mpq_clear(rat);
      }
      break;
      
    default:
      mpfr_init_set_d(big_real(x), s7_real(p), GMP_RNDN);
      break;
    }
  return(x);
}


static s7_pointer mpz_to_big_real(s7_scheme *sc, mpz_t val)
{
  s7_pointer x;
  
  new_cell(sc, x, T_BIG_REAL);
  add_bigreal(sc, x);
  mpfr_init_set_z(big_real(x), val, GMP_RNDN);
  return(x);
}


static s7_pointer mpq_to_big_real(s7_scheme *sc, mpq_t val)
{
  s7_pointer x;
  
  new_cell(sc, x, T_BIG_REAL);
  add_bigreal(sc, x);
  mpfr_init_set_q(big_real(x), val, GMP_RNDN);
  return(x);
}


static s7_pointer mpfr_to_big_real(s7_scheme *sc, mpfr_t val)
{
  s7_pointer x;
  
  new_cell(sc, x, T_BIG_REAL);
  add_bigreal(sc, x);
  mpfr_init_set(big_real(x), val, GMP_RNDN);
  
  return(x);
}


s7_pointer s7_make_big_real(s7_scheme *sc, mpfr_t *val)
{
  return(mpfr_to_big_real(sc, *val));
}


static s7_pointer big_pi(s7_scheme *sc)
{
  s7_pointer x;
  
  new_cell(sc, x, T_BIG_REAL);
  add_bigreal(sc, x);
  mpfr_init(big_real(x));
  mpfr_const_pi(big_real(x), GMP_RNDN);
  return(x);
}


static s7_pointer s7_number_to_big_complex(s7_scheme *sc, s7_pointer p)
{
  s7_pointer x;

  new_cell(sc, x, T_BIG_COMPLEX);
  add_bignumber(sc, x);
  mpc_init(big_complex(x));

  switch (type(p))
    {
    case T_INTEGER:
      if (sizeof(s7_int) == sizeof(long int))
	mpc_set_si(big_complex(x), integer(p), MPC_RNDNN);
      else mpc_set_d(big_complex(x), (double)integer(p), MPC_RNDNN);
      break;

    case T_RATIO:
      /* can't use fraction here */
      {
	mpfr_t temp;
	mpq_t rat;
	mpz_t n1, d1;

	mpz_init_set_s7_int(n1, numerator(p));
	mpz_init_set_s7_int(d1, denominator(p));
	mpq_init(rat);

	mpq_set_num(rat, n1);
	mpq_set_den(rat, d1);
	mpq_canonicalize(rat);
	mpfr_init_set_q(temp, rat, GMP_RNDN);
	mpc_set_fr(big_complex(x), temp, MPC_RNDNN);

	mpz_clear(n1);
	mpz_clear(d1);
	mpq_clear(rat);
	mpfr_clear(temp);
      }
      break;

    case T_REAL:
      mpc_set_d(big_complex(x), s7_real(p), MPC_RNDNN);
      break;

    default:
      mpc_set_d_d(big_complex(x), real_part(p), imag_part(p), MPC_RNDNN);
      break;
    }
  return(x);
}


static s7_pointer make_big_real_or_complex(s7_scheme *sc, s7_pointer z)
{
  double ipart;

  ipart = mpfr_get_d(mpc_imagref(big_complex(z)), GMP_RNDN);
  /* not mpfr_cmp_ui to 0 here because that misleads us when imag_part is NaN or inf */
  if (ipart == 0.0)
    return(mpfr_to_big_real(sc, mpc_realref(big_complex(z))));
  return(z);
}


static s7_pointer mpz_to_big_complex(s7_scheme *sc, mpz_t val)
{
  mpfr_t temp;
  s7_pointer x;

  new_cell(sc, x, T_BIG_COMPLEX);
  add_bignumber(sc, x);
  mpc_init(big_complex(x));
  mpfr_init_set_z(temp, val, GMP_RNDN);
  mpc_set_fr(big_complex(x), temp, MPC_RNDNN);

  mpfr_clear(temp);
  return(x);
}


static s7_pointer mpq_to_big_complex(s7_scheme *sc, mpq_t val)
{
  mpfr_t temp;
  s7_pointer x;

  new_cell(sc, x, T_BIG_COMPLEX);
  add_bignumber(sc, x);
  mpc_init(big_complex(x));
  mpfr_init_set_q(temp, val, GMP_RNDN);
  mpc_set_fr(big_complex(x), temp, MPC_RNDNN);

  mpfr_clear(temp);
  return(x);
}


static s7_pointer mpfr_to_big_complex(s7_scheme *sc, mpfr_t val)
{
  s7_pointer x;

  new_cell(sc, x, T_BIG_COMPLEX);
  add_bignumber(sc, x);
  mpc_init(big_complex(x));
  mpc_set_fr(big_complex(x), val, MPC_RNDNN);
  return(x);
}


static s7_pointer mpc_to_big_complex(s7_scheme *sc, mpc_t val)
{
  s7_pointer x;

  new_cell(sc, x, T_BIG_COMPLEX);
  add_bignumber(sc, x);
  mpc_init(big_complex(x));
  mpc_set(big_complex(x), val, MPC_RNDNN);
  return(x);
}


s7_pointer s7_make_big_complex(s7_scheme *sc, mpc_t *val)
{
  return(mpc_to_big_complex(sc, *val));
}


static s7_pointer make_big_complex(s7_scheme *sc, mpfr_t rl, mpfr_t im)
{
  /* there is no mpc_get_str equivalent, so we need to split up str,
   *   use make_big_real to get the 2 halves, then mpc_init, then
   *   mpc_set_fr_fr.
   */
  s7_pointer x;

  new_cell(sc, x, T_BIG_COMPLEX);
  add_bignumber(sc, x);
  mpc_init(big_complex(x));
  mpc_set_fr_fr(big_complex(x), rl ,im, MPC_RNDNN);
  return(x);
}


/* gmp.h mpz_init_set_si the "si" part is "signed long int", so in 64-bit machines, s7_int already fits (if it's int64_t).  
 *   I guess we can catch the 4-byte long int32_t (since no configure script) by noticing that sizeof(s7_int) == sizeof(long int)?
 */

static void mpz_init_set_s7_int(mpz_t n, s7_int uval)
{
  if (sizeof(s7_int) == sizeof(long int))
    mpz_init_set_si(n, uval);
  else
    {
      /* int64_t to gmp mpz_t */
      bool need_sign;
      int64_t val;
      val = (int64_t)uval;
      /* handle one special case (sigh) */
      if (val == s7_int_min)
	mpz_init_set_str(n, "-9223372036854775808", 10);
      else
	{
	  need_sign = (val < 0);
	  if (need_sign) val = -val;
	  mpz_init_set_si(n, val >> 32);
	  mpz_mul_2exp(n, n, 32);
	  mpz_add_ui(n, n, (uint32_t)(val & 0xffffffff));
	  if (need_sign) mpz_neg(n, n);
	}
    }
}


static s7_pointer s7_int_to_big_integer(s7_scheme *sc, s7_int val)
{
  s7_pointer x;

  new_cell(sc, x, T_BIG_INTEGER);
  add_bigint(sc, x);
  mpz_init_set_s7_int(big_integer(x), val);
  return(x);
}


static s7_int big_integer_to_s7_int(mpz_t n)
{
  int64_t high, low;
  mpz_t x;
  bool need_sign = false;

  if (mpz_fits_slong_p(n))
    return(mpz_get_si(n));

  if ((cur_sc->safety > NO_SAFETY) &&
      (sizeof(s7_int) == sizeof(long int)))
    {
      char *str;
      str = mpz_get_str(NULL, 10, n);
      s7_warn(cur_sc, 256, "can't convert %s to s7_int\n", str);
      free(str);
    }

  mpz_init_set(x, n);
  if (mpz_cmp_ui(x, 0) < 0)
    {
      need_sign = true;
      mpz_neg(x, x);
    }
  low = mpz_get_ui(x);
  if (low == s7_int_min)
    return(s7_int_min);

  mpz_fdiv_q_2exp(x, x, 32);
  high = mpz_get_ui(x);
  mpz_clear(x);
  if (need_sign)
    return(-(low + (high << 32)));
  return(low + (high << 32));
}


static mpq_t *s7_ints_to_mpq(s7_int num, s7_int den)
{
  /* den here always comes from denominator(x) so it is not negative */
  mpq_t *n;
  n = (mpq_t *)malloc(sizeof(mpq_t));
  mpq_init(*n);
  if (sizeof(s7_int) == sizeof(long int))
    mpq_set_si(*n, num, den);
  else
    {
      mpz_t n1, d1;
      mpz_init_set_s7_int(n1, num);
      mpz_init_set_s7_int(d1, den);
      mpq_set_num(*n, n1);
      mpq_set_den(*n, d1);
      mpq_canonicalize(*n);
      mpz_clear(n1);
      mpz_clear(d1);
    }
  return(n);
}


static mpfr_t *s7_double_to_mpfr(s7_double val)
{
  mpfr_t *n;
  n = (mpfr_t *)malloc(sizeof(mpfr_t));
  mpfr_init_set_d(*n, val, GMP_RNDN);
  return(n);
}


static mpc_t *s7_doubles_to_mpc(s7_double rl, s7_double im)
{
  mpc_t *n;
  n = (mpc_t *)malloc(sizeof(mpc_t));
  mpc_init(*n);
  mpc_set_d_d(*n, rl, im, MPC_RNDNN);
  return(n);
}


static s7_pointer s7_ratio_to_big_ratio(s7_scheme *sc, s7_int num, s7_int den)
{
  /* den here always comes from denominator(x) or some positive constant so it is not negative */
  s7_pointer x;
  new_cell(sc, x, T_BIG_RATIO);
  add_bigratio(sc, x);
  mpq_init(big_ratio(x));
  if (sizeof(s7_int) == sizeof(long int))
    mpq_set_si(big_ratio(x), num, den);
  else
    {
      mpz_t n1, d1;
      mpz_init_set_s7_int(n1, num);
      mpz_init_set_s7_int(d1, den);
      mpq_set_num(big_ratio(x), n1);
      mpq_set_den(big_ratio(x), d1);
      mpq_canonicalize(big_ratio(x));
      mpz_clear(n1);
      mpz_clear(d1);
    }
  return(x);
}


static bool big_numbers_are_eqv(s7_pointer a, s7_pointer b)
{
  bool result;
  /* either or both can be big here, but not neither */

  if (s7_is_integer(a))
    {
      mpz_t a1, b1;
      if (!(s7_is_integer(b))) return(false);

      if ((is_big_number(a)) && (is_big_number(b)))
	return(mpz_cmp(big_integer(a), big_integer(b)) == 0);

      if (is_big_number(a))
	mpz_init_set(a1, big_integer(a));
      else mpz_init_set_s7_int(a1, s7_integer(a));

      if (is_big_number(b))
	mpz_init_set(b1, big_integer(b));
      else mpz_init_set_s7_int(b1, s7_integer(b));
      result = (mpz_cmp(a1, b1) == 0);

      mpz_clear(a1);
      mpz_clear(b1);
      return(result);
    }

  if (s7_is_ratio(a))
    {
      mpq_t *a1, *b1;
      if (!s7_is_ratio(b)) return(false);

      if ((is_big_number(a)) && (is_big_number(b)))
	return(mpq_cmp(big_ratio(a), big_ratio(b)) == 0);

      if (is_big_number(a))
	a1 = &big_ratio(a);
      else a1 = s7_ints_to_mpq(numerator(a), denominator(a));
      if (is_big_number(b))
	b1 = &big_ratio(b);
      else b1 = s7_ints_to_mpq(numerator(b), denominator(b));

      result = (mpq_cmp(*a1, *b1) == 0);

      if (!is_big_number(a))
	{
	  mpq_clear(*a1);
	  free(a1);
	}
      if (!is_big_number(b))
	{
	  mpq_clear(*b1);
	  free(b1);
	}
      return(result);
    }

  if (s7_is_real(a))
    {
      mpfr_t *a1, *b1;

      /* s7_is_real is not finicky enough here -- (eqv? 1.0 1) should return #f */
      if (is_big_number(b))
	{
	  if (type(b) != T_BIG_REAL)
	    return(false);
	}
      else
	{
	  if (type(b) != T_REAL)
	    return(false);
	}

      if ((is_big_number(a)) && (is_big_number(b)))
	return(mpfr_equal_p(big_real(a), big_real(b)));

      if (is_big_number(a))
	a1 = &big_real(a);
      else a1 = s7_double_to_mpfr(s7_real(a));

      if (is_big_number(b))
	b1 = &big_real(b);
      else b1 = s7_double_to_mpfr(s7_real(b));

      result = (mpfr_cmp(*a1, *b1) == 0);

      if (!is_big_number(a))
	{
	  mpfr_clear(*a1);
	  free(a1);
	}
      if (!is_big_number(b))
	{
	  mpfr_clear(*b1);
	  free(b1);
	}
      return(result);
    }

  if (s7_is_complex(a))
    {
      mpc_t *a1, *b1;
      /* s7_is_complex is not finicky enough here */

      if ((type(b) != T_BIG_COMPLEX) &&
	  (type(b) != T_COMPLEX))
	return(false);

      /* (eqv? (bignum "1+i") 1+1i) */
      if ((is_big_number(a)) && (is_big_number(b)))
	return(mpc_cmp(big_complex(a), big_complex(b)) == 0);

      if (is_big_number(a))
	a1 = &big_complex(a);
      else a1 = s7_doubles_to_mpc(real_part(a), imag_part(a));

      if (is_big_number(b))
	b1 = &big_complex(b);
      else b1 = s7_doubles_to_mpc(real_part(b), imag_part(b));

      result = (mpc_cmp(*a1, *b1) == 0);

      if (!is_big_number(a))
	{
	  mpc_clear(*a1);
	  free(a1);
	}
      if (!is_big_number(b))
	{
	  mpc_clear(*b1);
	  free(b1);
	}
      return(result);
    }
  return(false);
}


static s7_pointer string_to_either_integer(s7_scheme *sc, const char *str, int32_t radix)
{
  s7_int val;
  bool overflow = false;

  val = string_to_integer(str, radix, &overflow);
  if (!overflow)
    return(make_integer(sc, val));

  return(string_to_big_integer(sc, str, radix));
}


static s7_pointer string_to_either_ratio(s7_scheme *sc, const char *nstr, const char *dstr, int32_t radix)
{
  s7_int n, d;
  bool overflow = false;

  /* gmp segfaults if passed a bignum/0 so this needs to check first that
   *   the denominator is not 0 before letting gmp screw up.  Also, if the
   *   first character is '+', gmp returns 0!
   */
  d = string_to_integer(dstr, radix, &overflow);
  if (!overflow)
    {
      if (d == 0)
	return(real_NaN);

      n = string_to_integer(nstr, radix, &overflow);
      if (!overflow)
	return(s7_make_ratio(sc, n, d));
    }
  if (nstr[0] == '+')
    return(string_to_big_ratio(sc, (const char *)(nstr + 1), radix));
  return(string_to_big_ratio(sc, nstr, radix));
}


static s7_pointer string_to_either_real(s7_scheme *sc, const char *str, int32_t radix)
{
  bool overflow = false;
  s7_double val;

  val = string_to_double_with_radix((char *)str, radix, &overflow);
  if (!overflow)
    return(make_real(sc, val));

  return(string_to_big_real(sc, str, radix));
}


static s7_pointer string_to_either_complex_1(s7_scheme *sc, char *q, char *slash1, char *ex1, bool has_dec_point1, int32_t radix, s7_double *d_rl)
{
  bool overflow = false;
  /* there's a real problem here -- we don't want to promote s7_double .1 to a bignum because
   *    its low order digits are garbage, causing (rationalize .1 0) to return 3602879701896397/36028797018963968
   *    no matter what the bignum-precision.  But we can't just fallback on gmp's reader because (for example)
   *    it reads 1/2+i or 1+0/0i as 1.0.  Also format gets screwed up.  And string->number signals an error
   *    where it should return #f.  I wonder what to do.
   */
  if ((has_dec_point1) ||
      (ex1))
    {
      (*d_rl) = string_to_double_with_radix(q, radix, &overflow);
      if (overflow)
	return(string_to_big_real(sc, q, radix));
    }
  else
    {
      if (slash1)
	{
	  s7_int n, d;

	  /* q can include the slash and denominator */
	  n = string_to_integer(q, radix, &overflow);
	  if (overflow)
	    return(string_to_big_ratio(sc, q, radix));
	  else
	    {
	      d = string_to_integer(slash1, radix, &overflow);
	      if (!overflow)
		(*d_rl) = (s7_double)n / (s7_double)d;
	      else return(string_to_big_ratio(sc, q, radix));
	    }
	}
      else
	{
	  s7_int val;

	  val = string_to_integer(q, radix, &overflow);
	  if (overflow)
	    return(string_to_big_integer(sc, q, radix));
	  (*d_rl) = (s7_double)val;
	}
    }
  if ((*d_rl) == -0.0) (*d_rl) = 0.0;
  return(NULL);
}


static s7_pointer string_to_either_complex(s7_scheme *sc,
					   char *q, char *slash1, char *ex1, bool has_dec_point1,
					   char *plus, char *slash2, char *ex2, bool has_dec_point2,
					   int32_t radix, int32_t has_plus_or_minus)
{
  /* this can be just about anything involving 2 real/ratio/int portions, +/- in between and 'i' at the end */
  double d_rl = 0.0, d_im = 0.0;
  s7_pointer p_rl = NULL, p_im = NULL, result;
  mpfr_t m_rl, m_im;

  p_rl = string_to_either_complex_1(sc, q, slash1, ex1, has_dec_point1, radix, &d_rl);
  p_im = string_to_either_complex_1(sc, plus, slash2, ex2, has_dec_point2, radix, &d_im);

  if (d_im == 0.0)
    {
      /* 1.0+0.0000000000000000000000000000i */
      if ((!p_im) ||
	  (s7_is_zero(p_im)))
	{
	  if (!p_rl)
	    return(make_real(sc, d_rl));
	  return(p_rl);
	}
    }

  if ((!p_rl) && (!p_im))
    return(s7_make_complex(sc, d_rl, (has_plus_or_minus == -1) ? (-d_im) : d_im));

  if (p_rl)
    mpfr_init_set(m_rl, big_real(promote_number(sc, T_BIG_REAL, p_rl)), GMP_RNDN);
  else mpfr_init_set_d(m_rl, d_rl, GMP_RNDN);

  if (p_im)
    mpfr_init_set(m_im, big_real(promote_number(sc, T_BIG_REAL, p_im)), GMP_RNDN);
  else mpfr_init_set_d(m_im, d_im, GMP_RNDN);

  if (has_plus_or_minus == -1)
    mpfr_neg(m_im, m_im, GMP_RNDN);

  result = make_big_complex(sc, m_rl, m_im);

  mpfr_clear(m_rl);
  mpfr_clear(m_im);
  return(result);
}


static int32_t big_type_to_result_type(int32_t cur_type, int32_t next_type)
{
  if ((cur_type == T_BIG_COMPLEX) ||
      (cur_type == T_COMPLEX) ||
      (next_type == T_BIG_COMPLEX))
    return(T_BIG_COMPLEX);

  if ((cur_type == T_BIG_REAL) ||
      (cur_type == T_REAL) ||
      (next_type == T_BIG_REAL))
    return(T_BIG_REAL);

  if ((cur_type == T_BIG_RATIO) ||
      (cur_type == T_RATIO) ||
      (next_type == T_BIG_RATIO))
    return(T_BIG_RATIO);

  return(T_BIG_INTEGER);
}


static int32_t normal_type_to_result_type(int32_t cur_type, int32_t next_type)
{
  if (cur_type > T_COMPLEX)
    next_type += 4;
  if (cur_type > next_type)
    return(cur_type);
  return(next_type);
}


static s7_pointer promote_number_1(s7_scheme *sc, int32_t type, s7_pointer x, bool copy)
{
  /* x can be any number -- need to convert it to the current result type */

  switch (type)
    {
    case T_BIG_INTEGER:
      if (is_big_number(x))
	{
	  if (copy)
	    return(mpz_to_big_integer(sc, big_integer(x)));
	  return(x);                       /* can only be T_BIG_INTEGER here */
	}
      return(s7_int_to_big_integer(sc, s7_integer(x))); /* can only be integer here */

    case T_BIG_RATIO:
      if (is_big_number(x))
	{
	  if (is_t_big_ratio(x))
	    {
	      if (copy)
		return(mpq_to_big_ratio(sc, big_ratio(x)));
	      return(x);
	    }
	  return(mpz_to_big_ratio(sc, big_integer(x)));
	}
      if (is_t_integer(x))
	return(s7_ratio_to_big_ratio(sc, integer(x), 1));
      return(s7_ratio_to_big_ratio(sc, numerator(x), denominator(x)));

    case T_BIG_REAL:
      if (is_big_number(x))
	{
	  if (is_t_big_real(x))
	    {
	      if (copy)
		return(mpfr_to_big_real(sc, big_real(x)));
	      return(x);
	    }
	  if (is_t_big_ratio(x))
	    return(mpq_to_big_real(sc, big_ratio(x)));
	  return(mpz_to_big_real(sc, big_integer(x)));
	}
      return(s7_number_to_big_real(sc, x));

    default:
      if (is_big_number(x))
	{
	  if (is_t_big_complex(x))
	    {
	      if (copy)
		return(mpc_to_big_complex(sc, big_complex(x)));
	      return(x);
	    }
	  if (is_t_big_real(x))
	    return(mpfr_to_big_complex(sc, big_real(x)));
	  if (is_t_big_ratio(x))
	    return(mpq_to_big_complex(sc, big_ratio(x)));
	  return(mpz_to_big_complex(sc, big_integer(x)));
	}
      return(s7_number_to_big_complex(sc, x));
    }
  return(sc->nil);
}


static s7_pointer promote_number(s7_scheme *sc, int32_t type, s7_pointer x)
{
  return(promote_number_1(sc, type, x, false));
}


static s7_pointer to_big(s7_scheme *sc, s7_pointer x)
{
  if (is_big_number(x))
    return(x);
  switch (type(x))
    {
    case T_INTEGER: return(s7_int_to_big_integer(sc, integer(x)));
    case T_RATIO:   return(s7_ratio_to_big_ratio(sc, numerator(x), denominator(x)));
    case T_REAL:    return(s7_number_to_big_real(sc, x));
    default:        return(s7_number_to_big_complex(sc, x));
    }
}


static s7_pointer copy_and_promote_number(s7_scheme *sc, int32_t type, s7_pointer x)
{
  return(promote_number_1(sc, type, x, true));
}


void s7_vector_fill(s7_scheme *sc, s7_pointer vec, s7_pointer obj)
{
  /* if the same bignum object is assigned to each element, different vector elements
   *    are actually the same -- we need to make a copy of obj for each one
   */
  if ((is_normal_vector(vec)) && (is_big_number(obj)))
    {
      uint32_t gc_loc;
      s7_int i, len;
      s7_pointer *tp;

      len = vector_length(vec);
      tp = (s7_pointer *)(vector_elements(vec));

      /* we'll be calling new_cell below, hence the GC, so make sure the elements are markable,
       *   and the vector itself is GC protected (we can be called within make-vector).
       */
      gc_loc = s7_gc_protect(sc, vec);
      vector_fill(sc, vec, sc->nil);

      switch (type(obj))
	{
	case T_BIG_INTEGER: for (i = 0; i < len; i++) tp[i] = mpz_to_big_integer(sc, big_integer(obj)); break;
	case T_BIG_RATIO:   for (i = 0; i < len; i++) tp[i] = mpq_to_big_ratio(sc, big_ratio(obj));     break;
	case T_BIG_REAL:    for (i = 0; i < len; i++) tp[i] = mpfr_to_big_real(sc, big_real(obj));      break;
	default:  	    for (i = 0; i < len; i++) tp[i] = mpc_to_big_complex(sc, big_complex(obj)); break;
	}
      s7_gc_unprotect_at(sc, gc_loc);
    }
  else vector_fill(sc, vec, obj);
}


static s7_pointer big_bignum(s7_scheme *sc, s7_pointer args)
{
  #define H_bignum "(bignum val (radix 10)) returns a multiprecision version of the string 'val'"
  #define Q_bignum s7_make_signature(sc, 3, sc->is_bignum_symbol, sc->is_number_symbol, sc->is_integer_symbol)
  s7_pointer p;

  p = g_string_to_number_1(sc, args, sc->bignum_symbol);
  if (is_false(sc, p))                                       /* (bignum "1/3.0") */
    s7_error(sc, make_symbol(sc, "bignum-error"), 
	     set_elist_2(sc, s7_make_string_wrapper(sc, "bignum argument does not represent a number: ~S"), car(args)));

  switch (type(p))
    {
    case T_INTEGER:
      return(promote_number(sc, T_BIG_INTEGER, p));

    case T_RATIO:
      return(promote_number(sc, T_BIG_RATIO, p));

      /* we can't use promote_number here because it propagates C-double inaccuracies
       *    (rationalize (bignum "0.1") 0) should return 1/10 not 3602879701896397/36028797018963968
       */
    case T_REAL:
      if (is_NaN(real(p))) return(p);
      return(string_to_big_real(sc, string_value(car(args)), (is_pair(cdr(args))) ? s7_integer(cadr(args)) : 10));

    case T_COMPLEX:
      return(promote_number(sc, T_BIG_COMPLEX, p));

    default:
      return(p);
    }
}


bool s7_is_bignum(s7_pointer obj)
{
  return(is_big_number(obj));
}


static s7_pointer big_is_bignum(s7_scheme *sc, s7_pointer args)
{
  #define H_is_bignum "(bignum? obj) returns #t if obj is a multiprecision number."
  #define Q_is_bignum pl_bt
  return(s7_make_boolean(sc, is_big_number(car(args))));
}

#define get_result_type(Sc, Type, P) \
  ((is_number(P)) ? normal_type_to_result_type(Type, type(p)) : ((is_big_number(P)) ? big_type_to_result_type(Type, type(p)) : result_type_via_method(Sc, Type, P)))

static int32_t result_type_via_method(s7_scheme *sc, int32_t result_type, s7_pointer p)
{
  s7_pointer f;
  if (!has_methods(p)) return(-1);

  f = find_method(sc, find_let(sc, p), sc->is_integer_symbol);
  if ((f != sc->undefined) &&
      (is_true(sc, s7_apply_function(sc, f, cons(sc, p, sc->nil)))))
    return(big_type_to_result_type(result_type, T_BIG_INTEGER));

  f = find_method(sc, find_let(sc, p), sc->is_rational_symbol);
  if ((f != sc->undefined) &&
      (is_true(sc, s7_apply_function(sc, f, cons(sc, p, sc->nil)))))
    return(big_type_to_result_type(result_type, T_BIG_RATIO));

  f = find_method(sc, find_let(sc, p), sc->is_real_symbol);
  if ((f != sc->undefined) &&
      (is_true(sc, s7_apply_function(sc, f, cons(sc, p, sc->nil)))))
    return(big_type_to_result_type(result_type, T_BIG_REAL));

  /* might be a number, but not complex (quaternion) */
  f = find_method(sc, find_let(sc, p), sc->is_complex_symbol);
  if ((f != sc->undefined) &&
      (is_true(sc, s7_apply_function(sc, f, cons(sc, p, sc->nil)))))
    return(big_type_to_result_type(result_type, T_BIG_COMPLEX));

  return(-1);
}


static s7_pointer big_add(s7_scheme *sc, s7_pointer args)
{
  int32_t result_type = T_INTEGER;
  s7_pointer x, result;

  if (is_null(args))
    return(small_int(0));

  if ((is_null(cdr(args))) && (s7_is_number(car(args))))
    return(car(args));

  for (x = args; is_not_null(x); x = cdr(x))
    {
      s7_pointer p;
      p = car(x);
      result_type = get_result_type(sc, result_type, p);
      if (result_type < 0)
	return(g_add(sc, args));
    }

  if (result_type < T_BIG_INTEGER)
    return(g_add(sc, args));
  if (!s7_is_number(car(args)))
    check_method(sc, car(args), sc->add_symbol, args);

  result = copy_and_promote_number(sc, result_type, car(args));

  for (x = cdr(args); is_not_null(x); x = cdr(x))
    {
      s7_pointer arg;
      if (!s7_is_number(car(x)))
	check_method(sc, car(x), sc->add_symbol, cons(sc, result, x));

      arg = promote_number(sc, result_type, car(x));

      switch (result_type)
	{
	case T_BIG_INTEGER: mpz_add(big_integer(result), big_integer(result), big_integer(arg));             break;
	case T_BIG_RATIO:   mpq_add(big_ratio(result), big_ratio(result), big_ratio(arg));                   break;
	case T_BIG_REAL:    mpfr_add(big_real(result), big_real(result), big_real(arg), GMP_RNDN);           break;
	case T_BIG_COMPLEX: mpc_add(big_complex(result), big_complex(result), big_complex(arg), MPC_RNDNN);  break;
	}
    }

  switch (result_type)
      {
      case T_BIG_RATIO:	  return(make_big_integer_or_ratio(sc, result));
      case T_BIG_COMPLEX: return(make_big_real_or_complex(sc, result));
      }
  return(result);
}


static s7_pointer big_negate(s7_scheme *sc, s7_pointer args)
{
  /* assume cdr(args) is nil and we're called from subtract, so check for big num else call g_subtract */
  s7_pointer p, x;

  p = car(args);
  switch (type(p))
    {
    case T_BIG_INTEGER:
      x = mpz_to_big_integer(sc, big_integer(p));
      mpz_neg(big_integer(x), big_integer(x));
      return(x);

    case T_BIG_RATIO:
      x = mpq_to_big_ratio(sc, big_ratio(p));
      mpq_neg(big_ratio(x), big_ratio(x));
      return(x);

    case T_BIG_REAL:
      x = mpfr_to_big_real(sc, big_real(p));
      mpfr_neg(big_real(x), big_real(x), GMP_RNDN);
      return(x);

    case T_BIG_COMPLEX:
      x = mpc_to_big_complex(sc, big_complex(p));
      mpc_neg(big_complex(x), big_complex(x), MPC_RNDNN);
      return(x);

    case T_INTEGER:
      if (integer(p) == s7_int_min)
	{
	  x = s7_int_to_big_integer(sc, integer(p));
	  mpz_neg(big_integer(x), big_integer(x));
	  return(x);
	}
      return(make_integer(sc, -integer(p)));

    case T_RATIO:
      return(s7_make_ratio(sc, -numerator(p), denominator(p)));

    case T_REAL:
      return(make_real(sc, -real(p)));

    default:
      return(s7_make_complex(sc, -real_part(p), -imag_part(p)));
    }
}


static s7_pointer big_subtract(s7_scheme *sc, s7_pointer args)
{
  int32_t result_type = T_INTEGER;
  s7_pointer x, result;

  if (!s7_is_number(car(args)))
    method_or_bust_with_type(sc, car(args), sc->subtract_symbol, args, a_number_string, 1);

  if (is_null(cdr(args)))
    return(big_negate(sc, args));

  for (x = args; is_not_null(x); x = cdr(x))
    {
      s7_pointer p;
      p = car(x);
      result_type = get_result_type(sc, result_type, p);
      if (result_type < 0)
	return(g_subtract(sc, args));
    }

  if (result_type < T_BIG_INTEGER)
    return(g_subtract(sc, args));

  if (!s7_is_number(car(args)))
    check_method(sc, car(args), sc->subtract_symbol, args);

  result = copy_and_promote_number(sc, result_type, car(args));

  for (x = cdr(args); is_not_null(x); x = cdr(x))
    {
      s7_pointer arg;
      if (!s7_is_number(car(x)))
	check_method(sc, car(x), sc->subtract_symbol, cons(sc, result, x));

      arg = promote_number(sc, result_type, car(x));

      switch (result_type)
	{
	case T_BIG_INTEGER:  mpz_sub(big_integer(result), big_integer(result), big_integer(arg));	     break;
	case T_BIG_RATIO:    mpq_sub(big_ratio(result),   big_ratio(result),   big_ratio(arg));	             break;
	case T_BIG_REAL:     mpfr_sub(big_real(result),   big_real(result),    big_real(arg), GMP_RNDN);     break;
	case T_BIG_COMPLEX:  mpc_sub(big_complex(result), big_complex(result), big_complex(arg), MPC_RNDNN); break;
	}
    }

  switch (result_type)
    {
      case T_BIG_RATIO:	  return(make_big_integer_or_ratio(sc, result));
      case T_BIG_COMPLEX: return(make_big_real_or_complex(sc, result));
    }
  return(result);
}


static s7_pointer big_multiply(s7_scheme *sc, s7_pointer args)
{
  int32_t result_type = T_INTEGER;
  s7_pointer x, result;

  if (is_null(args))
    return(small_int(1));

  if ((is_null(cdr(args))) && (s7_is_number(car(args))))
    return(car(args));

  for (x = args; is_not_null(x); x = cdr(x))
    {
      s7_pointer p;
      p = car(x);
      result_type = get_result_type(sc, result_type, p);
      if (result_type < 0)
	return(g_multiply(sc, args));
    }

  if (result_type < T_BIG_INTEGER)
    return(g_multiply(sc, args));

  if (!s7_is_number(car(args)))
    check_method(sc, car(args), sc->multiply_symbol, args);

  result = copy_and_promote_number(sc, result_type, car(args));

  for (x = cdr(args); is_not_null(x); x = cdr(x))
    {
      s7_pointer arg;
      if (!s7_is_number(car(x)))
	check_method(sc, car(x), sc->multiply_symbol, cons(sc, result, x));

      arg = promote_number(sc, result_type, car(x));
      switch (result_type)
	{
	case T_BIG_INTEGER: mpz_mul(big_integer(result), big_integer(result), big_integer(arg));	    break;
	case T_BIG_RATIO:   mpq_mul(big_ratio(result),   big_ratio(result),   big_ratio(arg));	            break;
	case T_BIG_REAL:    mpfr_mul(big_real(result),   big_real(result),    big_real(arg), GMP_RNDN);     break;
	case T_BIG_COMPLEX: mpc_mul(big_complex(result), big_complex(result), big_complex(arg), MPC_RNDNN); break;
	}
    }

  switch (result_type)
    {
    case T_BIG_RATIO:   return(make_big_integer_or_ratio(sc, result));
    case T_BIG_COMPLEX: return(make_big_real_or_complex(sc, result));
    }
  return(result);
}


static s7_pointer big_invert(s7_scheme *sc, s7_pointer args)
{
  /* assume cdr(args) is nil and we're called from divide, so check for big num else call g_divide */
  s7_pointer p, x;

  p = car(args);
  if (s7_is_zero(p))
    return(division_by_zero_error(sc, sc->divide_symbol, p));

  switch (type(p))
    {
    case T_INTEGER:
      if (integer(p) == s7_int_min)
	{
	  mpz_t n1, d1;

	  new_cell(sc, x, T_BIG_RATIO);
	  add_bigratio(sc, x);

	  mpz_init_set_s7_int(n1, 1);
	  mpz_init_set_s7_int(d1, s7_int_min);
	  mpq_set_num(big_ratio(x), n1);
	  mpq_set_den(big_ratio(x), d1);
	  mpq_canonicalize(big_ratio(x));
	  mpz_clear(n1);
	  mpz_clear(d1);

	  return(x);
	}
      return(s7_make_ratio(sc, 1, integer(p)));      /* a already checked, not 0 */

    case T_RATIO:
      return(s7_make_ratio(sc, denominator(p), numerator(p)));

    case T_REAL:
      return(make_real(sc, 1.0 / real(p)));

    case T_COMPLEX:
      {
	s7_double r2, i2, den;
	r2 = real_part(p);
	i2 = imag_part(p);
	den = (r2 * r2 + i2 * i2);
	return(s7_make_complex(sc, r2 / den, -i2 / den));
      }

    case T_BIG_INTEGER:
      /* p might be 1 or -1 */
      {
	mpz_t n;

	mpz_init_set_si(n, 1);
	if (mpz_cmp(n, big_integer(p)) == 0)
	  {
	    mpz_clear(n);
	    return(small_int(1));
	  }
	mpz_set_si(n, -1);
	if (mpz_cmp(n, big_integer(p)) == 0)
	  {
	    mpz_clear(n);
	    return(minus_one);
	  }

	new_cell(sc, x, T_BIG_RATIO);
	add_bigratio(sc, x);
	mpq_init(big_ratio(x));

	mpz_set_ui(n, 1);
	mpq_set_num(big_ratio(x), n);
	mpz_clear(n);

	mpq_set_den(big_ratio(x), big_integer(p));
	mpq_canonicalize(big_ratio(x));
	return(x);
      }

    case T_BIG_RATIO:
      {
	mpz_t n;

	mpz_init_set_si(n, 1);
	if (mpz_cmp(n, mpq_numref(big_ratio(p))) == 0)
	  {
	    mpz_clear(n);
	    return(mpz_to_big_integer(sc, mpq_denref(big_ratio(p))));
	  }
	mpz_set_si(n, -1);
	if (mpz_cmp(n, mpq_numref(big_ratio(p))) == 0)
	  {
	    mpz_clear(n);
	    x = mpz_to_big_integer(sc, mpq_denref(big_ratio(p)));
	    mpz_neg(big_integer(x), big_integer(x));
	    return(x);
	  }
	mpz_clear(n);

	new_cell(sc, x, T_BIG_RATIO);
	add_bigratio(sc, x);

	mpq_init(big_ratio(x));
	mpq_set_num(big_ratio(x), mpq_denref(big_ratio(p)));
	mpq_set_den(big_ratio(x), mpq_numref(big_ratio(p)));
	mpq_canonicalize(big_ratio(x));
	return(x);
      }

    case T_BIG_REAL:
      x = mpfr_to_big_real(sc, big_real(p));
      mpfr_ui_div(big_real(x), 1, big_real(x), GMP_RNDN);
      return(x);

    default:
      x = mpc_to_big_complex(sc, big_complex(p));
      mpc_ui_div(big_complex(x), 1, big_complex(x), MPC_RNDNN);
      return(x);
    }
}


static s7_pointer big_divide(s7_scheme *sc, s7_pointer args)
{
  int32_t result_type = T_INTEGER;
  s7_pointer x, divisor, result;

  if (!s7_is_number(car(args)))
    method_or_bust_with_type(sc, car(args), sc->divide_symbol, args, a_number_string, 1);

  if (is_null(cdr(args)))
    return(big_invert(sc, args));

  for (x = args; is_not_null(x); x = cdr(x))
    {
      s7_pointer p;
      p = car(x);
      /* if divisor is 0, gmp throws an exception and halts s7!
       *   I don't think we can trap gmp errors, and the abort is built into the library code.
       */
      result_type = get_result_type(sc, result_type, p);
      if (result_type < 0)
	return(g_divide(sc, args));

      if ((x != args) &&
	  (s7_is_zero(p)))
	return(division_by_zero_error(sc, sc->divide_symbol, args));
    }

  if (result_type < T_BIG_INTEGER)
    return(g_divide(sc, args));

  if (!s7_is_number(car(args)))
    check_method(sc, car(args), sc->divide_symbol, args);

  if (!s7_is_number(cadr(args)))
    check_method(sc, cadr(args), sc->divide_symbol, args);

  divisor = copy_and_promote_number(sc, result_type, cadr(args));

  for (x = cddr(args); is_not_null(x); x = cdr(x))
    {
      s7_pointer arg;
      if (!s7_is_number(car(x)))
	{
	  s7_pointer func;
	  if ((has_methods(car(x))) && ((func = find_method(sc, find_let(sc, car(x)), sc->multiply_symbol)) != sc->undefined))
	    {
	      divisor = s7_apply_function(sc, func, cons(sc, divisor, x));
	      break;
	    }
	}

      arg = promote_number(sc, result_type, car(x));
      switch (result_type)
	{
	case T_BIG_INTEGER: mpz_mul(big_integer(divisor), big_integer(divisor), big_integer(arg));	      break;
	case T_BIG_RATIO:   mpq_mul(big_ratio(divisor),   big_ratio(divisor),   big_ratio(arg));	      break;
	case T_BIG_REAL:    mpfr_mul(big_real(divisor),   big_real(divisor),    big_real(arg), GMP_RNDN);     break;
	case T_BIG_COMPLEX: mpc_mul(big_complex(divisor), big_complex(divisor), big_complex(arg), MPC_RNDNN); break;
	}
    }

  if (s7_is_zero(divisor))
    return(division_by_zero_error(sc, sc->divide_symbol, args));

  /* it's possible for the divisor to be the wrong type here (if complex multiply -> real for example */
  divisor = promote_number_1(sc, result_type, divisor, false);

  result = copy_and_promote_number(sc, result_type, car(args));

  switch (result_type)
    {
    case T_BIG_INTEGER:
      {
	new_cell(sc, x, T_BIG_RATIO);
	add_bigratio(sc, x);

	mpq_init(big_ratio(x));
	mpq_set_num(big_ratio(x), big_integer(result));
	mpq_set_den(big_ratio(x), big_integer(divisor));
	mpq_canonicalize(big_ratio(x));

	if (mpz_cmp_ui(mpq_denref(big_ratio(x)), 1) == 0)
	  return(mpz_to_big_integer(sc, mpq_numref(big_ratio(x))));
	return(x);
      }

    case T_BIG_RATIO:
      mpq_div(big_ratio(result), big_ratio(result), big_ratio(divisor));
      return(make_big_integer_or_ratio(sc, result));

    case T_BIG_REAL:
      mpfr_div(big_real(result), big_real(result), big_real(divisor), GMP_RNDN);
      break;

    case T_BIG_COMPLEX:
      mpc_div(big_complex(result), big_complex(result), big_complex(divisor), MPC_RNDNN);
      return(make_big_real_or_complex(sc, result));
    }
  return(result);
}


static s7_pointer big_abs(s7_scheme *sc, s7_pointer args)
{
  #define H_abs "(abs x) returns the absolute value of the real number x"
  #define Q_abs s7_make_signature(sc, 2, sc->is_real_symbol, sc->is_real_symbol)  
  
  s7_pointer p, x;

  p = car(args);
  switch (type(p))
    {
    case T_INTEGER:
      if (integer(p) < 0)
	{
	  if (integer(p) == s7_int_min)
	    {
	      x = s7_int_to_big_integer(sc, integer(p));
	      mpz_neg(big_integer(x), big_integer(x));
	      return(x);
	    }
	  return(make_integer(sc, -integer(p)));
	}
      return(p);

    case T_RATIO:
      if (numerator(p) < 0)
	return(s7_make_ratio(sc, -numerator(p), denominator(p)));
      return(p);

    case T_REAL:
      if (real(p) < 0.0)
	return(make_real(sc, -real(p)));
      return(p);

    case T_BIG_INTEGER:
      x = mpz_to_big_integer(sc, big_integer(p));
      mpz_abs(big_integer(x), big_integer(x));
      return(x);

    case T_BIG_RATIO:
      x = mpq_to_big_ratio(sc, big_ratio(p));
      mpq_abs(big_ratio(x), big_ratio(x));
      return(x);

    case T_BIG_REAL:
      x = mpfr_to_big_real(sc, big_real(p));
      mpfr_abs(big_real(x), big_real(x), GMP_RNDN);
      return(x);

    default:
      method_or_bust_one_arg(sc, p, sc->abs_symbol, args, T_REAL);
    }
}


static s7_pointer big_magnitude(s7_scheme *sc, s7_pointer args)
{
  #define H_magnitude "(magnitude z) returns the magnitude of z"
  #define Q_magnitude s7_make_signature(sc, 2, sc->is_real_symbol, sc->is_number_symbol)

  s7_pointer p;

  p = car(args);
  if (!s7_is_number(p))
    method_or_bust_with_type_one_arg(sc, p, sc->magnitude_symbol, args, a_number_string);

  if (is_t_big_complex(p))
    {
      mpfr_t n;
      mpfr_init(n);
      mpc_abs(n, big_complex(p), GMP_RNDN);
      p = mpfr_to_big_real(sc, n);
      mpfr_clear(n);
      return(p);
    }

  if (is_t_complex(p))
    return(make_real(sc, hypot(imag_part(p), real_part(p))));

  return(big_abs(sc, args));
}

static s7_pointer big_angle(s7_scheme *sc, s7_pointer args)
{
  #define H_angle "(angle z) returns the angle of z"
  #define Q_angle s7_make_signature(sc, 2, sc->is_real_symbol, sc->is_number_symbol)

  s7_pointer p;

  p = car(args);
  switch (type(p))
    {
    case T_INTEGER:
      if (integer(p) < 0)
	return(real_pi);
      return(small_int(0));

    case T_RATIO:
      if (numerator(p) < 0)
	return(real_pi);
      return(small_int(0));

    case T_REAL:
      if (is_NaN(real(p))) return(p);
      if (real(p) < 0.0)
	return(real_pi);
      return(real_zero);

    case T_COMPLEX:
      return(make_real(sc, atan2(imag_part(p), real_part(p))));

    case T_BIG_INTEGER:
      if (mpz_cmp_ui(big_integer(p), 0) >= 0)
	return(small_int(0));
      return(big_pi(sc));

    case T_BIG_RATIO:
      if (mpq_cmp_ui(big_ratio(p), 0, 1) >= 0)
	return(small_int(0));
      return(big_pi(sc));

    case T_BIG_REAL:
      {
	double x;
	x = mpfr_get_d(big_real(p), GMP_RNDN);
	/* mpfr_get_d returns inf or -inf if the arg is too large for a double */
	if (is_NaN(x)) return(p);
	if (x >= 0.0)
	  return(real_zero);
	return(big_pi(sc));
      }

    case T_BIG_COMPLEX:
      {
	s7_pointer x;
	new_cell(sc, x, T_BIG_REAL);
	add_bigreal(sc, x);
	mpfr_init(big_real(x));
	mpc_arg(big_real(x), big_complex(p), GMP_RNDN);
	return(x);
      }

    default:
      method_or_bust_with_type_one_arg(sc, p, sc->angle_symbol, args, a_number_string);
    }
}


static s7_pointer c_big_complex(s7_scheme *sc, s7_pointer args)
{
  #define H_complex "(complex x1 x2) returns a complex number with real-part x1 and imaginary-part x2"
  #define Q_complex s7_make_signature(sc, 3, sc->is_number_symbol, sc->is_real_symbol, sc->is_real_symbol)

  s7_pointer p0, p1, p;
  mpfr_t rl, im;
  double x;

  p0 = car(args);
  if (!s7_is_real(p0))
    method_or_bust(sc, p0, sc->complex_symbol, args, T_REAL, 1);

  p1 = cadr(args);
  if (!s7_is_real(p1))
    method_or_bust(sc, p1, sc->complex_symbol, args, T_REAL, 2);

  if ((!is_big_number(p1)) && (real_to_double(sc, p1, "complex") == 0.0)) /* imag-part is not bignum and is 0.0 */
    return(p0);

  mpfr_init_set(im, big_real(promote_number(sc, T_BIG_REAL, p1)), GMP_RNDN);
  x = mpfr_get_d(im, GMP_RNDN);
  if (x == 0.0)                                   /* imag-part is bignum 0.0 */
    {
      mpfr_clear(im);
      return(p0);
    }

  mpfr_init_set(rl, big_real(promote_number(sc, T_BIG_REAL, p0)), GMP_RNDN);

  new_cell(sc, p, T_BIG_COMPLEX);
  add_bignumber(sc, p);
  mpc_init(big_complex(p));
  mpc_set_fr_fr(big_complex(p), rl, im, MPC_RNDNN);

  mpfr_clear(rl);
  mpfr_clear(im);
  return(p);
}


/* (make-polar 0 (real-part (log 0))) = 0? or nan? */

#if (!WITH_PURE_S7)
static s7_pointer big_make_polar(s7_scheme *sc, s7_pointer args)
{
  #define H_make_polar "(make-polar mag ang) returns a complex number with magnitude mag and angle ang"
  #define Q_make_polar s7_make_signature(sc, 3, sc->is_number_symbol, sc->is_real_symbol, sc->is_real_symbol)

  s7_pointer p0, p1, p;
  mpfr_t ang, mag, rl, im;
  double x, y;

  p0 = car(args);
  if (!s7_is_real(p0))
    method_or_bust(sc, p0, sc->make_polar_symbol, args, T_REAL, 1);

  p1 = cadr(args);
  if (!s7_is_real(p1))
    method_or_bust(sc, p1, sc->make_polar_symbol, args, T_REAL, 2);

  mpfr_init_set(ang, big_real(promote_number(sc, T_BIG_REAL, p1)), GMP_RNDN);
  y = mpfr_get_d(ang, GMP_RNDN);

  if (is_NaN(y))
    {
      mpfr_clear(ang);
      return(real_NaN);
    }

  mpfr_init_set(mag, big_real(promote_number(sc, T_BIG_REAL, p0)), GMP_RNDN);
  x = mpfr_get_d(mag, GMP_RNDN);

  if (is_NaN(x))
    {
      mpfr_clear(ang);
      mpfr_clear(mag);
      return(real_NaN);
    }

  if ((x == 0.0) || (y == 0.0))
    {
      mpfr_clear(ang);
      mpfr_clear(mag);
      return(p0);
    }

  mpfr_init_set(im, ang, GMP_RNDN);
  mpfr_sin(im, im, GMP_RNDN);
  mpfr_mul(im, im, mag, GMP_RNDN);

  x = mpfr_get_d(im, GMP_RNDN);
  if (x == 0.0)
    {
      mpfr_clear(im);
      mpfr_clear(ang);
      mpfr_clear(mag);
      return(p0);
    }

  mpfr_init_set(rl, ang, GMP_RNDN);
  mpfr_cos(rl, rl, GMP_RNDN);
  mpfr_mul(rl, rl, mag, GMP_RNDN);

  new_cell(sc, p, T_BIG_COMPLEX);
  add_bignumber(sc, p);
  mpc_init(big_complex(p));
  mpc_set_fr_fr(big_complex(p), rl, im, MPC_RNDNN);

  mpfr_clear(rl);
  mpfr_clear(im);
  mpfr_clear(ang);
  mpfr_clear(mag);
  return(p);
}
#endif


static s7_pointer big_log(s7_scheme *sc, s7_pointer args)
{
  #define H_log "(log z1 (z2 e)) returns log(z1) / log(z2) where z2 (the base) defaults to e: (log 8 2) = 3"
  #define Q_log pcl_n

  /* either arg can be big, second is optional */
  s7_pointer p0, p1 = NULL, p;

  p0 = car(args);
  if (!s7_is_number(p0))
    method_or_bust_with_type(sc, p0, sc->log_symbol, args, a_number_string, 1);

  if (is_not_null(cdr(args)))
    {
      p1 = cadr(args);
      if (!s7_is_number(p1))
	method_or_bust_with_type(sc, p1, sc->log_symbol, args, a_number_string, 2);
    }

  if ((s7_is_real(p0)) &&
      ((!p1) || (s7_is_real(p1))))
    {
      double x, y = 0.0;

      p0 = promote_number(sc, T_BIG_REAL, p0);
      x = mpfr_get_d(big_real(p0), GMP_RNDN);
      if (is_NaN(x))
	return(real_NaN);

      if (p1)
	{
	  p1 = promote_number(sc, T_BIG_REAL, p1);
	  y = mpfr_get_d(big_real(p1), GMP_RNDN);

	  /* we can't check y here for 1.0 (check for 0.0 apparently is ok):
	   *  :(log 100.0 (+ 1.0 (bignum "1e-16")))
	   *  ;log base, argument 2, 1.000000000000000100000000000000000000002E0, is out of range (can't be 0.0 or 1.0)
	   *  :(= 1.0 (+ 1.0 (bignum "1e-16")))
	   *  #f
	   */
	  if (is_NaN(y))
	    return(real_NaN);
	  if (y == 0.0)
	    return(out_of_range(sc, sc->log_symbol, small_int(2), p1, s7_make_string_wrapper(sc, "argument can't be 0.0")));
	}
      if (x == 0.0)
	return(s7_make_complex(sc, -INFINITY, M_PI));

      if ((x > 0.0) && (y >= 0.0))
	{
	  mpfr_t n, base;

	  mpfr_init_set(n, big_real(p0), GMP_RNDN);
	  mpfr_log(n, n, GMP_RNDN);

	  if (!p1)
	    {
	      /* presumably log is safe with regard to real-part overflow giving a bogus int? */
	      if ((s7_is_rational(car(args))) &&
		  (mpfr_integer_p(n) != 0))
		{
		  new_cell(sc, p, T_BIG_INTEGER);
		  add_bigint(sc, p);
		  mpz_init(big_integer(p));
		  mpfr_get_z(big_integer(p), n, GMP_RNDN);
		}
	      else p = mpfr_to_big_real(sc, n);
	      mpfr_clear(n);
	      return(p);
	    }

	  mpfr_init_set(base, big_real(p1), GMP_RNDN);
	  mpfr_log(base, base, GMP_RNDN);
	  mpfr_div(n, n, base, GMP_RNDN);
	  mpfr_clear(base);

	  if ((s7_is_rational(car(args))) &&
	      (s7_is_rational(cadr(args))) &&
	      (mpfr_integer_p(n) != 0))
	    {
	      new_cell(sc, p, T_BIG_INTEGER);
	      add_bigint(sc, p);
	      mpz_init(big_integer(p));
	      mpfr_get_z(big_integer(p), n, GMP_RNDN);
	    }
	  else p = mpfr_to_big_real(sc, n);
	  mpfr_clear(n);
	  return(p);
	}
    }

  p0 = promote_number(sc, T_BIG_COMPLEX, p0);
  if (p1) p1 = promote_number(sc, T_BIG_COMPLEX, p1);
  {
    mpc_t n, base;
    double x;

    mpc_init(n);
    mpc_set(n, big_complex(p0), MPC_RNDNN);
    mpc_log(n, n, MPC_RNDNN);
    if (!p1)
      {
	p = mpc_to_big_complex(sc, n);
	mpc_clear(n);
	return(p);
      }

    mpc_init(base);
    mpc_set(base, big_complex(p1), MPC_RNDNN);
    mpc_log(base, base, MPC_RNDNN);
    mpc_div(n, n, base, MPC_RNDNN);
    mpc_clear(base);

    x = mpfr_get_d(mpc_imagref(n), GMP_RNDN);
    if (x == 0.0)
      p = mpfr_to_big_real(sc, mpc_realref(n));
    else p = mpc_to_big_complex(sc, n);

    mpc_clear(n);
    return(p);
  }
}


static s7_pointer big_sqrt(s7_scheme *sc, s7_pointer args)
{
  /* real >= 0 -> real, else complex */
  #define H_sqrt "(sqrt z) returns the square root of z"
  #define Q_sqrt pcl_n  

  s7_pointer p;

  p = car(args);
  if (!s7_is_number(p))
    method_or_bust_with_type_one_arg(sc, p, sc->sqrt_symbol, args, a_number_string);
  p = to_big(sc, p);

  /* if big integer, try to return int32_t if perfect square */
  if (is_t_big_integer(p))
    {
      if (mpz_cmp_ui(big_integer(p), 0) < 0)
	p = promote_number(sc, T_BIG_COMPLEX, p);
      else
	{
	  mpz_t n, rem;

	  mpz_init(rem);
	  mpz_init_set(n, big_integer(p));
	  mpz_sqrtrem(n, rem, n);

	  if (mpz_cmp_ui(rem, 0) == 0)
	    {
	      p = mpz_to_big_integer(sc, n);
	      mpz_clear(n);
	      mpz_clear(rem);
	      return(p);
	    }
	  mpz_clear(n);
	  mpz_clear(rem);
	  p = promote_number(sc, T_BIG_REAL, p);
	}
    }

  /* if big ratio, check both num and den for squares */
  if (is_t_big_ratio(p))
    {
      if (mpq_cmp_ui(big_ratio(p), 0, 1) < 0)
	p = promote_number(sc, T_BIG_COMPLEX, p);
      else
	{
	  mpz_t n1, rem;
	  mpz_init(rem);
	  mpz_init_set(n1, mpq_numref(big_ratio(p)));
	  mpz_sqrtrem(n1, rem, n1);

	  if (mpz_cmp_ui(rem, 0) == 0)
	    {
	      mpz_t d1;
	      mpz_init_set(d1, mpq_denref(big_ratio(p)));
	      mpz_sqrtrem(d1, rem, d1);

	      if (mpz_cmp_ui(rem, 0) == 0)
		{
		  mpq_t n;
		  mpq_init(n);
		  mpq_set_num(n, n1);
		  mpq_set_den(n, d1);
		  mpq_canonicalize(n);
		  p = mpq_to_big_ratio(sc, n);
		  mpz_clear(n1);
		  mpz_clear(d1);
		  mpz_clear(rem);
		  mpq_clear(n);
		  return(p);
		}
	      mpz_clear(d1);
	    }

	  mpz_clear(n1);
	  mpz_clear(rem);
	  p = promote_number(sc, T_BIG_REAL, p);
	}
    }

  /* if real and not negative, use mpfr_sqrt */
  if (is_t_big_real(p))
    {
      if (mpfr_cmp_ui(big_real(p), 0) < 0)
	p = promote_number(sc, T_BIG_COMPLEX, p);
      else
	{
	  mpfr_t n;
	  mpfr_init_set(n, big_real(p), GMP_RNDN);
	  mpfr_sqrt(n, n, GMP_RNDN);
	  p = mpfr_to_big_real(sc, n);
	  mpfr_clear(n);
	  return(p);
	}
    }

  /* p is a big number, so it must be complex at this point */
  {
    mpc_t n;
    mpc_init(n);
    mpc_set(n, big_complex(p), MPC_RNDNN);
    mpc_sqrt(n, n, MPC_RNDNN);
    p = mpc_to_big_complex(sc, n);
    mpc_clear(n);
    return(p);
  }
}


/* (define (diff f a) (magnitude (- (f a) (f (bignum (number->string a))))))
 * (sin 1e15+1e15i) hangs in mpc 0.8.2, but appears to be fixed in the current svn sources
 */

enum {TRIG_NO_CHECK, TRIG_TAN_CHECK, TRIG_TANH_CHECK};

static s7_pointer big_trig(s7_scheme *sc, s7_pointer args,
			   int32_t (*mpfr_trig)(mpfr_ptr, mpfr_srcptr, mpfr_rnd_t),
			   int32_t (*mpc_trig)(mpc_ptr, mpc_srcptr, mpc_rnd_t),
			   int32_t tan_case, s7_pointer sym)
/* these declarations mimic the mpfr.h and mpc.h declarations.  It seems to me that
 *   they ought to be:
 *			   int32_t (*mpfr_trig)(mpfr_t rop, mpfr_t op, mp_rnd_t rnd),
 *			   void (*mpc_trig)(mpc_t rop, mpc_t op, mpc_rnd_t rnd))
 */
{
  s7_pointer p;
  p = car(args);

  /* I think here we should always promote to bignum (otherwise, for example, (exp 800) -> inf)
   */
  if (!s7_is_number(p))
    method_or_bust_with_type_one_arg(sc, p, sym, args, a_number_string);
  if (s7_is_real(p))
    {
      mpfr_t n;
      mpfr_init_set(n, big_real(promote_number(sc, T_BIG_REAL, p)), GMP_RNDN);
      mpfr_trig(n, n, GMP_RNDN);
      /* it's confusing to check for ints here via mpfr_integer_p because it
       *   is dependent on the precision!  (exp 617/5) returns an integer if
       *   precision is 128, but a float if 512.
       */
      p = mpfr_to_big_real(sc, n);
      mpfr_clear(n);
      return(p);
    }

  if (!is_big_number(p))
    p = promote_number(sc, T_BIG_COMPLEX, p);

  if (tan_case == TRIG_TAN_CHECK)
    {
      if ((MPC_INEX_IM(mpc_cmp_si_si(big_complex(p), 1, 350))) > 0)
	return(s7_make_complex(sc, 0.0, 1.0));
      if ((MPC_INEX_IM(mpc_cmp_si_si(big_complex(p), 1, -350))) < 0)
	return(s7_make_complex(sc, 0.0, -1.0));
    }

  if (tan_case == TRIG_TANH_CHECK)
    {
      if ((MPC_INEX_RE(mpc_cmp_si_si(big_complex(p), 350, 1))) > 0)
	return(real_one);
      if ((MPC_INEX_RE(mpc_cmp_si_si(big_complex(p), -350, 1))) < 0)
	return(make_real(sc, -1.0));
    }

  {
    mpc_t n;
    double ix;

    mpc_init(n);
    mpc_trig(n, big_complex(p), MPC_RNDNN);
    /* (sin (bignum "1e15+1e15i")) causes mpc to hang (e9 is ok, but e10 hangs)
     *   (sin (bignum "0+1e10i")) -> 0+inf (sin (bignum "1+1e10i")) hangs
     *
     * before comparing imag-part to 0, we need to look for NaN and inf, else:
     *    (sinh 0+0/0i) -> 0.0
     *    (sinh (log 0.0)) -> inf.0
     */

    ix = mpfr_get_d(mpc_imagref(n), GMP_RNDN);
    if (ix == 0.0)
      {
	mpfr_t z;

	mpfr_init_set(z, mpc_realref(n), GMP_RNDN);
	p = mpfr_to_big_real(sc, z);
	mpfr_clear(z);
      }
    else p = mpc_to_big_complex(sc, n);
    mpc_clear(n);
    return(p);
  }
}


static s7_pointer big_sin(s7_scheme *sc, s7_pointer args)
{
  #define H_sin "(sin z) returns sin(z)"
  #define Q_sin pcl_n  

  return(big_trig(sc, args, mpfr_sin, mpc_sin, TRIG_NO_CHECK, sc->sin_symbol));
}


static s7_pointer big_cos(s7_scheme *sc, s7_pointer args)
{
  #define H_cos "(cos z) returns cos(z)"
  #define Q_cos pcl_n  

  return(big_trig(sc, args, mpfr_cos, mpc_cos, TRIG_NO_CHECK, sc->cos_symbol));
}


s7_pointer s7_cos(s7_scheme *sc, s7_pointer x)
{
  return(big_cos(sc, cons(sc, x, sc->nil)));
}


static s7_pointer big_tan(s7_scheme *sc, s7_pointer args)
{
  #define H_tan "(tan z) returns tan(z)"
  #define Q_tan pcl_n  

  return(big_trig(sc, args, mpfr_tan, mpc_tan, TRIG_TAN_CHECK, sc->tan_symbol));
}


static s7_pointer big_sinh(s7_scheme *sc, s7_pointer args)
{
  #define H_sinh "(sinh z) returns sinh(z)"
  #define Q_sinh pcl_n  

  /* currently (sinh 0+0/0i) -> 0.0? */
  return(big_trig(sc, args, mpfr_sinh, mpc_sinh, TRIG_NO_CHECK, sc->sinh_symbol));
}


static s7_pointer big_cosh(s7_scheme *sc, s7_pointer args)
{
  #define H_cosh "(cosh z) returns cosh(z)"
  #define Q_cosh pcl_n  

  return(big_trig(sc, args, mpfr_cosh, mpc_cosh, TRIG_NO_CHECK, sc->cosh_symbol));
}


static s7_pointer big_tanh(s7_scheme *sc, s7_pointer args)
{
  #define H_tanh "(tanh z) returns tanh(z)"
  #define Q_tanh pcl_n  

  return(big_trig(sc, args, mpfr_tanh, mpc_tanh, TRIG_TANH_CHECK, sc->tanh_symbol));
}


static s7_pointer big_exp(s7_scheme *sc, s7_pointer args)
{
  #define H_exp "(exp z) returns e^z, (exp 1) is 2.718281828459"
  #define Q_exp pcl_n  

  return(big_trig(sc, args, mpfr_exp, mpc_exp, TRIG_NO_CHECK, sc->exp_symbol));
}


static s7_pointer big_expt(s7_scheme *sc, s7_pointer args)
{
  #define H_expt "(expt z1 z2) returns z1^z2"
  #define Q_expt pcl_n

  s7_pointer x, y, p;

  /* see comment under g_expt
   *  if (is_not_null(cddr(args)))
   *    return(big_expt(sc, set_plist_2(sc, car(args), big_expt(sc, cdr(args)))));
   */

  x = car(args);
  if (!s7_is_number(x))
    method_or_bust_with_type(sc, x, sc->expt_symbol, args, a_number_string, 1);

  y = cadr(args);
  if (!s7_is_number(y))
    method_or_bust_with_type(sc, y, sc->expt_symbol, args, a_number_string, 2);

  if (s7_is_zero(x))
    {
      if ((s7_is_integer(x)) &&
	  (s7_is_integer(y)) &&
	  (s7_is_zero(y)))
	return(small_int(1));

      if (s7_is_real(y))
	{
	  if (s7_is_negative(y))
	    return(division_by_zero_error(sc, sc->expt_symbol, args));
	}
      else
	{
	  if (s7_is_negative(g_real_part(sc, cdr(args))))
	    return(division_by_zero_error(sc, sc->expt_symbol, args));
	}

      if ((s7_is_rational(x)) &&
	  (s7_is_rational(y)))
	return(small_int(0));
      return(real_zero);
    }

  if (s7_is_integer(y))
    {
      s7_int yval;
      yval = s7_integer(y);
      if (yval == 0)
	{
	  if (s7_is_rational(x))
	    return(small_int(1));
	  return(real_one);
	}

      if (yval == 1)
	return(x);

      if (!is_big_number(x))
	{
	  if ((s7_is_one(x)) || (s7_is_zero(x)))
	    return(x);
	}

      if ((yval < s7_int32_max) &&
	  (yval > s7_int32_min))
	{
	  /* from here yval can fit in an uint32_t
	   *    (protect against gmp exception if for example (expt 1/9223372036854775807 -9223372036854775807)
	   */
	  if (s7_is_integer(x))
	    {
	      mpz_t n;
	      mpq_t r;

	      x = promote_number(sc, T_BIG_INTEGER, x);
	      mpz_init_set(n, big_integer(x));
	      if (yval >= 0)
		{
		  mpz_pow_ui(n, n, (uint32_t)yval);
		  p = mpz_to_big_integer(sc, n);
		  mpz_clear(n);
		  return(p);
		}

	      mpz_pow_ui(n, n, (uint32_t)(-yval));
	      mpq_init(r);
	      mpq_set_z(r, n);
	      mpq_inv(r, r);
	      if (mpz_cmp_ui(mpq_denref(r), 1) == 0)
		{
		  mpz_t z;
		  mpz_init_set(z, mpq_numref(r));
		  mpq_clear(r);
		  mpz_clear(n);
		  p = mpz_to_big_integer(sc, z);
		  mpz_clear(z);
		  return(p);
		}
	      mpz_clear(n);
	      p = mpq_to_big_ratio(sc, r);
	      mpq_clear(r);
	      return(p);
	    }

	  if (s7_is_ratio(x)) /* here y is an integer */
	    {
	      mpz_t n, d;
	      mpq_t r;

	      x = promote_number(sc, T_BIG_RATIO, x);
	      mpz_init_set(n, mpq_numref(big_ratio(x)));
	      mpz_init_set(d, mpq_denref(big_ratio(x)));
	      mpq_init(r);
	      if (yval >= 0)
		{
		  mpz_pow_ui(n, n, (uint32_t)yval);
		  mpz_pow_ui(d, d, (uint32_t)yval);
		  mpq_set_num(r, n);
		  mpq_set_den(r, d);
		}
	      else
		{
		  yval = -yval;
		  mpz_pow_ui(n, n, (uint32_t)yval);
		  mpz_pow_ui(d, d, (uint32_t)yval);
		  mpq_set_num(r, d);
		  mpq_set_den(r, n);
		  mpq_canonicalize(r);
		}
	      mpz_clear(n);
	      mpz_clear(d);
	      if (mpz_cmp_ui(mpq_denref(r), 1) == 0)
		{
		  mpz_t z;
		  mpz_init_set(z, mpq_numref(r));
		  mpq_clear(r);
		  p = mpz_to_big_integer(sc, z);
		  mpz_clear(z);
		  return(p);
		}
	      p = mpq_to_big_ratio(sc, r);
	      mpq_clear(r);
	      return(p);
	    }

	  if (s7_is_real(x))
	    {
	      mpfr_t z;
	      x = promote_number(sc, T_BIG_REAL, x);
	      mpfr_init_set(z, big_real(x), GMP_RNDN);
	      mpfr_pow_si(z, z, yval, GMP_RNDN);
	      p = mpfr_to_big_real(sc, z);
	      mpfr_clear(z);
	      return(p);
	    }
	}
    }

  if ((is_t_ratio(y)) &&              /* not s7_is_ratio which accepts bignums */
      (numerator(y) == 1))
    {
      if (denominator(y) == 2)
	return(big_sqrt(sc, args));

      if ((s7_is_real(x)) &&
	  (denominator(y) == 3))
	{
	  mpfr_t z;
	  mpfr_init_set(z, big_real(promote_number(sc, T_BIG_REAL, x)), GMP_RNDN);
	  mpfr_cbrt(z, z, GMP_RNDN);
	  p = mpfr_to_big_real(sc, z);
	  mpfr_clear(z);
	  return(p);
	}
    }

  if ((s7_is_real(x)) &&
      (s7_is_real(y)) &&
      (s7_is_positive(x)))
    {
      mpfr_t z;
      mpfr_init_set(z, big_real(promote_number(sc, T_BIG_REAL, x)), GMP_RNDN);
      mpfr_pow(z, z, big_real(promote_number(sc, T_BIG_REAL, y)), GMP_RNDN);
      p = mpfr_to_big_real(sc, z);
      mpfr_clear(z);
      return(p);
    }

  {
    mpc_t cy;
    mpc_t z;

    x = promote_number(sc, T_BIG_COMPLEX, x);
    y = promote_number(sc, T_BIG_COMPLEX, y);

    mpc_init(z);
    mpc_set(z, big_complex(x), MPC_RNDNN);

    if (mpc_cmp_si_si(z, 0, 0) == 0)
      {
	mpc_clear(z);
	return(small_int(0));
      }

    if (mpc_cmp_si_si(z, 1, 0) == 0)
      {
	mpc_clear(z);
	return(small_int(1));
      }

    mpc_init(cy);
    mpc_set(cy, big_complex(y), MPC_RNDNN);
    mpc_pow(z, z, cy, MPC_RNDNN);
    mpc_clear(cy);

    if (mpfr_cmp_ui(mpc_imagref(z), 0) == 0)
      {
	mpfr_t n;
	if ((s7_is_rational(car(args))) &&
	    (s7_is_rational(cadr(args))) &&
	    (mpfr_integer_p(mpc_realref(z)) != 0))
	  {
	    /* mpfr_integer_p can be confused: (expt 2718/1000 (bignum "617/5")) returns an int32_t if precision=128, float if 512 */
	    /*   so first make sure we're within (say) 31 bits */
	    mpfr_t zi;
	    mpfr_init_set_ui(zi, s7_int32_max, GMP_RNDN);
	    if (mpfr_cmpabs(mpc_realref(z), zi) < 0)
	      {
		mpz_t k;
		mpz_init(k);
		mpfr_get_z(k, mpc_realref(z), GMP_RNDN);
		mpc_clear(z);
		mpfr_clear(zi);
		p = mpz_to_big_integer(sc, k);
		mpz_clear(k);
		return(p);
	      }
	    mpfr_clear(zi);
	  }

	mpfr_init_set(n, mpc_realref(z), GMP_RNDN);
	mpc_clear(z);
	p = mpfr_to_big_real(sc, n);
	mpfr_clear(n);
	return(p);
      }

    p = mpc_to_big_complex(sc, z);
    mpc_clear(z);
    return(p);
  }
}


static s7_pointer big_asinh(s7_scheme *sc, s7_pointer args)
{
  #define H_asinh "(asinh z) returns asinh(z)"
  #define Q_asinh pcl_n  

  s7_pointer p;

  p = car(args);
  if (!s7_is_number(p))
    method_or_bust_with_type_one_arg(sc, p, sc->asinh_symbol, args, a_number_string);

  if (s7_is_real(p))
    {
      mpfr_t n;
      p = promote_number(sc, T_BIG_REAL, p);
      mpfr_init_set(n, big_real(p), GMP_RNDN);
      mpfr_asinh(n, n, GMP_RNDN);
      p = mpfr_to_big_real(sc, n);
      mpfr_clear(n);
      return(p);
    }

  {
    mpc_t n;
    p = promote_number(sc, T_BIG_COMPLEX, p);
    mpc_init(n);
    mpc_set(n, big_complex(p), MPC_RNDNN);
    mpc_asinh(n, n, MPC_RNDNN);
    p = mpc_to_big_complex(sc, n);
    mpc_clear(n);
    return(p);
  }
}


static s7_pointer big_acosh(s7_scheme *sc, s7_pointer args)
{
  #define H_acosh "(acosh z) returns acosh(z)"
  #define Q_acosh pcl_n  

  s7_pointer p;
  double x;
  mpc_t n;

  p = car(args);
  if (!s7_is_number(p))
    method_or_bust_with_type_one_arg(sc, p, sc->acosh_symbol, args, a_number_string);
  p = promote_number(sc, T_BIG_COMPLEX, p);

  mpc_init(n);
  mpc_set(n, big_complex(p), MPC_RNDNN);
  mpc_acosh(n, n, MPC_RNDNN);

  x = mpfr_get_d(mpc_imagref(n), GMP_RNDN);
  if (x == 0.0)
    p = mpfr_to_big_real(sc, mpc_realref(n));
  else p = mpc_to_big_complex(sc, n);
  mpc_clear(n);
  return(p);
}


static s7_pointer big_atanh(s7_scheme *sc, s7_pointer args)
{
  #define H_atanh "(atanh z) returns atanh(z)"
  #define Q_atanh pcl_n  

  s7_pointer p;

  p = car(args);
  if (!s7_is_number(p))
    method_or_bust_with_type_one_arg(sc, p, sc->atanh_symbol, args, a_number_string);

  if (s7_is_real(p))
    {
      bool ok;
      mpfr_t temp;
      p = promote_number(sc, T_BIG_REAL, p);
      mpfr_init_set_ui(temp, 1, GMP_RNDN);
      ok = (mpfr_cmpabs(big_real(p), temp) < 0);
      mpfr_clear(temp);
      if (ok)
	{
	  mpfr_t n;
	  mpfr_init_set(n, big_real(p), GMP_RNDN);
	  mpfr_atanh(n, n, GMP_RNDN);
	  p = mpfr_to_big_real(sc, n);
	  mpfr_clear(n);
	  return(p);
	}
    }

  {
    mpc_t n;
    p = promote_number(sc, T_BIG_COMPLEX, p);
    mpc_init(n);
    mpc_set(n, big_complex(p), MPC_RNDNN);
    mpc_atanh(n, n, MPC_RNDNN);
    p = mpc_to_big_complex(sc, n);
    mpc_clear(n);
    return(p);
  }
}


static s7_pointer big_atan(s7_scheme *sc, s7_pointer args)
{
  #define H_atan "(atan z) returns atan(z), (atan y x) returns atan(y/x)"
  #define Q_atan s7_make_signature(sc, 3, sc->is_number_symbol, sc->is_number_symbol, sc->is_real_symbol)  

  s7_pointer p0, p1 = NULL, p;

  p0 = car(args);
  if (!s7_is_number(p0))
    method_or_bust_with_type_one_arg(sc, p0, sc->atan_symbol, args, a_number_string);

  if (is_not_null(cdr(args)))
    {
      p1 = cadr(args);
      if (!s7_is_real(p1))
	method_or_bust(sc, p1, sc->atan_symbol, args, T_REAL, 2);

      if (!s7_is_real(p0))
	return(wrong_type_argument(sc, sc->atan_symbol, 1, p0, T_REAL));

      p1 = promote_number(sc, T_BIG_REAL, p1);
    }

  if (s7_is_real(p0))
    {
      mpfr_t n;
      p0 = promote_number(sc, T_BIG_REAL, p0);
      mpfr_init_set(n, big_real(p0), GMP_RNDN);
      if (!p1)
	mpfr_atan(n, n, GMP_RNDN);
      else mpfr_atan2(n, n, big_real(p1), GMP_RNDN);
      p = mpfr_to_big_real(sc, n);
      mpfr_clear(n);
      return(p);
    }

  {
    mpc_t n;
    p0 = promote_number(sc, T_BIG_COMPLEX, p0);
    mpc_init_set(n, big_complex(p0), MPC_RNDNN);
    mpc_atan(n, n, MPC_RNDNN);
    p = mpc_to_big_complex(sc, n);
    mpc_clear(n);
    return(p);
  }
}


static s7_pointer big_acos(s7_scheme *sc, s7_pointer args)
{
  #define H_acos "(acos z) returns acos(z); (cos (acos 1)) = 1"
  #define Q_acos pcl_n  

  s7_pointer p;

  p = car(args);
  if (!s7_is_number(p))
    method_or_bust_with_type_one_arg(sc, p, sc->acos_symbol, args, a_number_string);

  if (s7_is_real(p))
    {
      bool ok;
      mpfr_t temp;
      mpfr_t n;
      p = promote_number(sc, T_BIG_REAL, p);
      mpfr_init_set(n, big_real(p), GMP_RNDN);
      mpfr_init_set_ui(temp, 1, GMP_RNDN);
      ok = (mpfr_cmpabs(n, temp) <= 0);
      mpfr_clear(temp);
      if (ok)
	{
	  mpfr_acos(n, n, GMP_RNDN);
	  p = mpfr_to_big_real(sc, n);
	  mpfr_clear(n);
	  return(p);
	}
      mpfr_clear(n);
    }

  {
    mpc_t n;
    p = promote_number(sc, T_BIG_COMPLEX, p);
    mpc_init_set(n, big_complex(p), MPC_RNDNN);
    mpc_acos(n, n, MPC_RNDNN);
    p = mpc_to_big_complex(sc, n);
    mpc_clear(n);
    return(p);
  }
}


static s7_pointer big_asin(s7_scheme *sc, s7_pointer args)
{
  #define H_asin "(asin z) returns asin(z); (sin (asin 1)) = 1"
  #define Q_asin pcl_n  

  s7_pointer p;

  p = car(args);
  if (!s7_is_number(p))
    method_or_bust_with_type_one_arg(sc, p, sc->asin_symbol, args, a_number_string);

  if (s7_is_real(p))
    {
      bool ok;
      mpfr_t temp;
      mpfr_t n;
      p = promote_number(sc, T_BIG_REAL, p);
      mpfr_init_set(n, big_real(p), GMP_RNDN);
      mpfr_init_set_ui(temp, 1, GMP_RNDN);
      ok = (mpfr_cmpabs(n, temp) <= 0);
      mpfr_clear(temp);
      if (ok)
	{
	  mpfr_asin(n, n, GMP_RNDN);
	  p = mpfr_to_big_real(sc, n);
	  mpfr_clear(n);
	  return(p);
	}
      mpfr_clear(n);
    }

  {
    mpc_t n;
    p = promote_number(sc, T_BIG_COMPLEX, p);
    mpc_init_set(n, big_complex(p), MPC_RNDNN);
    mpc_asin(n, n, MPC_RNDNN);
    p = mpc_to_big_complex(sc, n);
    mpc_clear(n);
    return(p);
  }
}


static s7_pointer big_lognot(s7_scheme *sc, s7_pointer args)
{
  if (is_t_big_integer(car(args)))
    {
      s7_pointer p;
      mpz_t n;
      mpz_init(n);
      mpz_com(n, big_integer(car(args)));
      p = mpz_to_big_integer(sc, n);
      mpz_clear(n);
      return(p);
    }
  return(g_lognot(sc, args));
}


#if (!WITH_PURE_S7)
static s7_pointer big_integer_length(s7_scheme *sc, s7_pointer args)
{
  if (is_t_big_integer(car(args)))
    {
      s7_pointer result;
      mpfr_t n;
      mpfr_init_set_z(n, big_integer(car(args)), GMP_RNDN);
      if (mpfr_cmp_ui(n, 0) < 0)
	mpfr_neg(n, n, GMP_RNDN);
      else mpfr_add_ui(n, n, 1, GMP_RNDN);
      mpfr_log2(n, n, GMP_RNDU);
      result = make_integer(sc, mpfr_get_si(n, GMP_RNDU));
      mpfr_clear(n);
      return(result);
    }
  return(g_integer_length(sc, args));
}
#endif


static s7_pointer big_ash(s7_scheme *sc, s7_pointer args)
{
  s7_pointer p0, p1;

  p0 = car(args);
  p1 = cadr(args);
  /* here, as in expt, there are cases like (ash 1 63) which need to be handled as bignums
   *   so there's no way to tell when it's safe to drop into g_ash instead.
   */
  if ((s7_is_integer(p0)) && /* this includes bignum ints... */
      (s7_is_integer(p1)))
    {
      mpz_t n;
      s7_int shift;
      s7_pointer p;
      bool p0_is_big;
      int32_t p0_compared_to_zero = 0;

      p0_is_big = is_big_number(p0);
      if (p0_is_big)
	p0_compared_to_zero = mpz_cmp_ui(big_integer(p0), 0);
      else
	{
	  if (s7_integer(p0) > 0)
	    p0_compared_to_zero = 1;
	  else
	    {
	      if (s7_integer(p0) < 0)
		p0_compared_to_zero = -1;
	      else p0_compared_to_zero = 0;
	    }
	}

      if (p0_compared_to_zero == 0)
	return(small_int(0));

      if (is_big_number(p1))
	{
	  if (!mpz_fits_sint_p(big_integer(p1)))
	    {
	      if (mpz_cmp_ui(big_integer(p1), 0) > 0)
		return(out_of_range(sc, sc->ash_symbol, small_int(2), p1, its_too_large_string));

	      /* here if p0 is negative, we need to return -1 */
	      if (p0_compared_to_zero == 1)
		return(small_int(0));
	      return(minus_one);
	    }
	  shift = mpz_get_si(big_integer(p1));
	}
      else
	{
	  shift = s7_integer(p1);
	  if (shift < s7_int32_min)
	    {
	      if (p0_compared_to_zero == 1)
		return(small_int(0));
	      return(minus_one);
	    }
	}

      mpz_init_set(n, big_integer(promote_number(sc, T_BIG_INTEGER, p0)));
      if (shift > 0)     /* left */
	mpz_mul_2exp(n, n, shift);
      else
	{
	  if (shift < 0) /* right */
	    mpz_fdiv_q_2exp(n, n, (uint32_t)(-shift));
	}
      p = mpz_to_big_integer(sc, n);
      mpz_clear(n);
      return(p);
    }
  return(g_ash(sc, args));
}


static bool is_integer_via_method(s7_scheme *sc, s7_pointer p)
{
  if (s7_is_integer(p))
    return(true);
  if (has_methods(p))
    {
      s7_pointer f;
      f = find_method(sc, find_let(sc, p), sc->is_integer_symbol);
      if (f != sc->undefined)
	return(is_true(sc, s7_apply_function(sc, f, cons(sc, p, sc->nil))));
    }
  return(false);
}

static s7_pointer big_bits(s7_scheme *sc, s7_pointer args, s7_pointer sym, int32_t start, s7_function g_bits,
			   void (*mpz_bits)(mpz_ptr, mpz_srcptr, mpz_srcptr))
{
  s7_pointer x, lst;
  bool use_bigs = false;
  for (x = args; is_not_null(x); x = cdr(x))
    {
      if (!is_integer_via_method(sc, car(x)))
	return(wrong_type_argument(sc, sym, position_of(x, args), car(x), T_INTEGER));
      if (!use_bigs) use_bigs = (type(car(x)) != T_INTEGER);
    }
  if (use_bigs)
    {
      mpz_t n;
      mpz_init_set_si(n, 0);
      if (start == -1)
	mpz_sub_ui(n, n, 1);
      for (x = args; is_not_null(x); x = cdr(x))
	{
	  s7_pointer i;
	  i = car(x);
	  switch (type(i))
	    {
	    case T_BIG_INTEGER:
	      mpz_bits(n, n, big_integer(i));
	      break;

	    case T_INTEGER:
	      mpz_bits(n, n, big_integer(s7_int_to_big_integer(sc, integer(i))));
	      break;

	    default:
	      /* we know it's an integer of some sort, but what about the method */
	      lst = cons(sc, mpz_to_big_integer(sc, n), x);
	      mpz_clear(n);
	      method_or_bust(sc, i, sym, lst, T_INTEGER, position_of(x, args));
	    }
	}
      x = mpz_to_big_integer(sc, n);
      mpz_clear(n);
      return(x);
    }
  return(g_bits(sc, args));
}


static s7_pointer big_logand(s7_scheme *sc, s7_pointer args)
{
  if (is_null(args))
    return(minus_one);
  return(big_bits(sc, args, sc->logand_symbol, -1, g_logand, mpz_and));
}


static s7_pointer big_logior(s7_scheme *sc, s7_pointer args)
{
  if (is_null(args))
    return(small_int(0));
  return(big_bits(sc, args, sc->logior_symbol, 0, g_logior, mpz_ior));
}


static s7_pointer big_logxor(s7_scheme *sc, s7_pointer args)
{
  if (is_null(args))
    return(small_int(0));
  return(big_bits(sc, args, sc->logxor_symbol, 0, g_logxor, mpz_xor));
}


static s7_pointer big_rationalize(s7_scheme *sc, s7_pointer args)
{
  #define H_rationalize "(rationalize x err) returns the ratio with lowest denominator within err of x"
  #define Q_rationalize s7_make_signature(sc, 3, sc->is_rational_symbol, sc->is_real_symbol, sc->is_real_symbol)

  /* currently (rationalize 1/0 1e18) -> 0
   * remember to pad with many trailing zeros:
   *
   *    : (rationalize 0.1 0)
   *    3602879701896397/36028797018963968
   *    :(rationalize 0.1000000000000000 0)
   *    1/10
   *
   * perhaps gmp number reader used if gmp -- could this be the trailing zeros problem?  (why is the non-gmp case ok?)
   *         also the bignum function is faking it.
   *         (rationalize (real-part (bignum "0.1+i")) 0) -> 3602879701896397/36028797018963968
   *
   * a confusing case:
   *      > (rationalize 5925563891587147521650777143.74135805596e05)
   *   should be 148139097289678688041269428593533951399/250000
   *   but that requires more than 128 bits of bignum-precision.
   */

  s7_pointer p0, p1 = NULL, p;
  mpfr_t error, ux, x0, x1;
  mpz_t i, i0, i1;
  double xx;

  p0 = car(args);
  if (!s7_is_real(p0))
    method_or_bust(sc, p0, sc->rationalize_symbol, args, T_REAL, 1);

  /* p0 can be exact, but we still have to check it for simplification */
  if (is_not_null(cdr(args)))
    {
      double err_x;
      p1 = cadr(args);
      if (!s7_is_real(p1))              /* (rationalize (expt 2 60) -) */
	method_or_bust(sc, p1, sc->rationalize_symbol, args, T_REAL, 2);

      if (is_big_number(p1))
	mpfr_init_set(error, big_real(promote_number(sc, T_BIG_REAL, p1)), GMP_RNDN);
      else mpfr_init_set_d(error, real_to_double(sc, p1, "rationalize"), GMP_RNDN);

      err_x = mpfr_get_d(error, GMP_RNDN);
      if (is_NaN(err_x))
	{
	  mpfr_clear(error);
	  return(out_of_range(sc, sc->rationalize_symbol, small_int(2), cadr(args), its_nan_string));
	}
      if (mpfr_inf_p(error) != 0)
	{
	  mpfr_clear(error);
	  return(small_int(0));
	}
      mpfr_abs(error, error, GMP_RNDN);
    }
  else mpfr_init_set_d(error, sc->default_rationalize_error, GMP_RNDN);

  if (is_big_number(p0))
    mpfr_init_set(ux, big_real(promote_number(sc, T_BIG_REAL, p0)), GMP_RNDN);
  else mpfr_init_set_d(ux, real_to_double(sc, p0, "rationalize"), GMP_RNDN);

  xx = mpfr_get_d(ux, GMP_RNDN);
  if (is_NaN(xx))
    {
      mpfr_clear(ux);
      mpfr_clear(error);
      return(out_of_range(sc, sc->rationalize_symbol, small_int(1), car(args), its_nan_string));
    }
  if (mpfr_inf_p(ux) != 0)
    {
      mpfr_clear(ux);
      mpfr_clear(error);
      return(out_of_range(sc, sc->rationalize_symbol, small_int(1), car(args), its_infinite_string));
    }

  mpfr_init_set(x0, ux, GMP_RNDN);        /* x0 = ux - error */
  mpfr_sub(x0, x0, error, GMP_RNDN);
  mpfr_init_set(x1, ux, GMP_RNDN);        /* x1 = ux + error */
  mpfr_add(x1, x1, error, GMP_RNDN);
  mpz_init(i);
  mpfr_get_z(i, x0, GMP_RNDU);            /* i = ceil(x0) */

  if (mpfr_cmp_ui(error, 1) >= 0)         /* if (error >= 1.0) */
    {
      mpz_t n;

      if (mpfr_cmp_ui(x0, 0) < 0)                /* if (x0 < 0) */
	{
	  if (mpfr_cmp_ui(x1, 0) < 0)            /*   if (x1 < 0) */
	    {
	      mpz_init(n);
	      mpfr_get_z(n, x1, GMP_RNDD);       /*     num = floor(x1) */
	    }
	  else mpz_init_set_ui(n, 0);            /*   else num = 0 */
	}
      else mpz_init_set(n, i);                   /* else num = i */

      mpz_clear(i);
      mpfr_clear(ux);
      mpfr_clear(x0);
      mpfr_clear(x1);
      mpfr_clear(error);
      p = mpz_to_big_integer(sc, n);
      mpz_clear(n);
      return(p);
    }

  if (mpfr_cmp_z(x1, i) >= 0)                /* if (x1 >= i) */
    {
      mpz_t n;

      if (mpz_cmp_ui(i, 0) >= 0)            /* if (i >= 0) */
	mpz_init_set(n, i);                 /*   num = i */
      else
	{
	  mpz_init(n);
	  mpfr_get_z(n, x1, GMP_RNDD);      /* else num = floor(x1) */
	}

      mpz_clear(i);
      mpfr_clear(ux);
      mpfr_clear(x0);
      mpfr_clear(x1);
      mpfr_clear(error);
      p = mpz_to_big_integer(sc, n);
      mpz_clear(n);
      return(p);
    }

  {
    mpz_t p0, q0, r, r1, p1, q1, old_p1, old_q1;
    mpfr_t val, e0, e1, e0p, e1p, old_e0, old_e1, old_e0p;

    mpz_init(i0);
    mpz_init(i1);
    mpfr_get_z(i0, x0, GMP_RNDD);            /* i0 = floor(x0) */
    mpfr_get_z(i1, x1, GMP_RNDU);            /* i1 = ceil(x1) */

    mpz_init_set(p0, i0);                    /* p0 = i0 */
    mpz_init_set_ui(q0, 1);                  /* q0 = 1 */
    mpz_init_set(p1, i1);                    /* p1 = i1 */
    mpz_init_set_ui(q1, 1);                  /* q1 = 1 */
    mpfr_init(e0);
    mpfr_init(e1);
    mpfr_init(e0p);
    mpfr_init(e1p);
    mpfr_sub_z(e0, x0, i1, GMP_RNDN);        /* e0 = i1 - x0 */
    mpfr_neg(e0, e0, GMP_RNDN);
    mpfr_sub_z(e1, x0, i0, GMP_RNDN);        /* e1 = x0 - i0 */
    mpfr_sub_z(e0p, x1, i1, GMP_RNDN);       /* e0p = i1 - x1 */
    mpfr_neg(e0p, e0p, GMP_RNDN);
    mpfr_sub_z(e1p, x1, i0, GMP_RNDN);       /* e1p = x1 - i0 */

    mpfr_init(val);

    mpfr_init(old_e0);
    mpfr_init(old_e1);
    mpfr_init(old_e0p);

    mpz_init(r);
    mpz_init(r1);
    mpz_init(old_p1);
    mpz_init(old_q1);

    while (true)
      {
	mpfr_set_z(val, p0, GMP_RNDN);
	mpfr_div_z(val, val, q0, GMP_RNDN);  /* val = p0/q0 */

	if (((mpfr_cmp(x0, val) <= 0) &&      /* if ((x0 <= val) && (val <= x1)) */
	     (mpfr_cmp(val, x1) <= 0)) ||
	    (mpfr_cmp_ui(e1, 0) == 0) ||
	    (mpfr_cmp_ui(e1p, 0) == 0))
	  /* these last 2 are probably not needed -- they protect against running out of bits in the non-gmp case above */
	  {
	    mpq_t q;
	    mpq_init(q);
	    mpq_set_num(q, p0);            /* return(p0/q0) */
	    mpq_set_den(q, q0);

	    mpz_clear(i);
	    mpz_clear(i0);
	    mpz_clear(i1);
	    mpfr_clear(ux);
	    mpfr_clear(x0);
	    mpfr_clear(x1);
	    mpfr_clear(error);

	    mpz_clear(p0);
	    mpz_clear(q0);
	    mpz_clear(r);
	    mpz_clear(r1);
	    mpz_clear(p1);
	    mpz_clear(q1);
	    mpz_clear(old_p1);
	    mpz_clear(old_q1);

	    mpfr_clear(val);
	    mpfr_clear(e0);
	    mpfr_clear(e1);
	    mpfr_clear(e0p);
	    mpfr_clear(e1p);
	    mpfr_clear(old_e0);
	    mpfr_clear(old_e1);
	    mpfr_clear(old_e0p);

	    p = mpq_to_big_ratio(sc, q);
	    mpq_clear(q);
	    return(p);
	  }

	mpfr_div(val, e0, e1, GMP_RNDN);
	mpfr_get_z(r, val, GMP_RNDD);           /* r = floor(e0/e1) */
	mpfr_div(val, e0p, e1p, GMP_RNDN);
	mpfr_get_z(r1, val, GMP_RNDU);          /* r1 = ceil(e0p/e1p) */
	if (mpz_cmp(r1, r) < 0)                 /* if (r1 < r) */
	  mpz_set(r, r1);                       /*   r = r1 */

	mpz_set(old_p1, p1);                    /* old_p1 = p1 */
	mpz_set(p1, p0);                        /* p1 = p0 */
	mpz_set(old_q1, q1);                    /* old_q1 = q1 */
	mpz_set(q1, q0);                        /* q1 = q0 */

	mpfr_set(old_e0, e0, GMP_RNDN);         /* old_e0 = e0 */
	mpfr_set(e0, e1p, GMP_RNDN);            /* e0 = e1p */
	mpfr_set(old_e0p, e0p, GMP_RNDN);       /* old_e0p = e0p */
	mpfr_set(e0p, e1, GMP_RNDN);            /* e0p = e1 */
	mpfr_set(old_e1, e1, GMP_RNDN);         /* old_e1 = e1 */

	mpz_mul(p0, p0, r);                     /* p0 = old_p1 + r * p0 */
	mpz_add(p0, p0, old_p1);

	mpz_mul(q0, q0, r);                     /* q0 = old_q1 + r * q0 */
	mpz_add(q0, q0, old_q1);

	mpfr_mul_z(e1, e1p, r, GMP_RNDN);       /* e1 = old_e0p - r * e1p */
	mpfr_sub(e1, old_e0p, e1, GMP_RNDN);

	mpfr_mul_z(e1p, old_e1, r, GMP_RNDN);   /* e1p = old_e0 - r * old_e1 */
	mpfr_sub(e1p, old_e0, e1p, GMP_RNDN);
      }
  }
}

#if (!WITH_PURE_S7)
static s7_pointer big_exact_to_inexact(s7_scheme *sc, s7_pointer args)
{
  #define H_exact_to_inexact "(exact->inexact num) converts num to an inexact number; (exact->inexact 3/2) = 1.5"
  #define Q_exact_to_inexact pcl_r

  s7_pointer p;

  p = car(args);
  if (!s7_is_number(p))   /* apparently (exact->inexact 1+i) is not an error */
    method_or_bust_with_type_one_arg(sc, p, sc->exact_to_inexact_symbol, args, a_number_string);

  if (!s7_is_rational(p))
    return(p);

  return(promote_number(sc, T_BIG_REAL, to_big(sc, p)));
}


static s7_pointer big_inexact_to_exact(s7_scheme *sc, s7_pointer args)
{
  #define H_inexact_to_exact "(inexact->exact num) converts num to an exact number; (inexact->exact 1.5) = 3/2"
  #define Q_inexact_to_exact s7_make_signature(sc, 2, sc->is_rational_symbol, sc->is_real_symbol)

  s7_pointer p;
  p = car(args);

  if (s7_is_rational(p))
    return(p);

  if (!s7_is_real(p))
    method_or_bust_one_arg(sc, p, sc->inexact_to_exact_symbol, args, T_REAL);
  return(big_rationalize(sc, args));
}
#endif

static s7_pointer big_convert_to_int(s7_scheme *sc, s7_pointer args, s7_pointer sym,
				     void (*div_func)(mpz_ptr, mpz_srcptr, mpz_srcptr),
				     mp_rnd_t mode)
{
  /* we can't go to the normal (non-gmp) functions here */
  s7_pointer p;
  mpz_t n;

  p = car(args);
  if (!s7_is_real(p))
    method_or_bust_one_arg(sc, p, sym, args, T_REAL);

  if (s7_is_integer(p))
    return(p);

  p = to_big(sc, p);
  if (is_t_big_ratio(p))
    {
      /* apparently we have to do the divide by hand */
      mpz_t d;
      mpz_init_set(n, mpq_numref(big_ratio(p)));
      mpz_init_set(d, mpq_denref(big_ratio(p)));
      div_func(n, n, d);
      mpz_clear(d);
    }
  else
    {
      if ((g_is_nan(sc, args) == sc->T) ||
	  (g_is_infinite(sc, args)) == sc->T)
	return(simple_out_of_range(sc, sym, p, (g_is_nan(sc, args) == sc->T) ? its_nan_string : its_infinite_string));

      mpz_init(n);
      mpfr_get_z(n, big_real(p), mode);
    }
  p = mpz_to_big_integer(sc, n);
  mpz_clear(n);
  return(p);
}


static s7_pointer big_floor(s7_scheme *sc, s7_pointer args)
{
  #define H_floor "(floor x) returns the integer closest to x toward -inf"
  #define Q_floor s7_make_signature(sc, 2, sc->is_integer_symbol, sc->is_real_symbol)

  return(big_convert_to_int(sc, args, sc->floor_symbol, mpz_fdiv_q, GMP_RNDD));
}


static s7_pointer big_ceiling(s7_scheme *sc, s7_pointer args)
{
  #define H_ceiling "(ceiling x) returns the integer closest to x toward inf"
  #define Q_ceiling s7_make_signature(sc, 2, sc->is_integer_symbol, sc->is_real_symbol)

  return(big_convert_to_int(sc, args, sc->ceiling_symbol, mpz_cdiv_q, GMP_RNDU));
}


static s7_pointer big_truncate(s7_scheme *sc, s7_pointer args)
{
  #define H_truncate "(truncate x) returns the integer closest to x toward 0"
  #define Q_truncate s7_make_signature(sc, 2, sc->is_integer_symbol, sc->is_real_symbol)

  return(big_convert_to_int(sc, args, sc->truncate_symbol, mpz_tdiv_q, GMP_RNDZ));
}


static s7_pointer big_round(s7_scheme *sc, s7_pointer args)
{
  #define H_round "(round x) returns the integer closest to x"
  #define Q_round s7_make_signature(sc, 2, sc->is_integer_symbol, sc->is_real_symbol)

  s7_pointer p;
  mpz_t n;

  p = car(args);
  if (!s7_is_real(p))
    method_or_bust_one_arg(sc, p, sc->round_symbol, args, T_REAL);

  if (s7_is_integer(p))
    return(p);

  p = to_big(sc, p);
  if (is_t_big_integer(p))
    return(p);

  if (is_t_big_ratio(p))
    {
      int32_t rnd;
      mpz_t rm;
      mpz_init_set(n, mpq_numref(big_ratio(p)));
      mpz_init(rm);
      mpz_fdiv_qr(n, rm, n, mpq_denref(big_ratio(p)));
      mpz_mul_ui(rm, rm, 2);
      rnd = mpz_cmpabs(rm, mpq_denref(big_ratio(p)));
      mpz_fdiv_q(rm, rm, mpq_denref(big_ratio(p)));
      if (rnd > 0)
	mpz_add(n, n, rm);
      else
	{
	  if (rnd == 0)
	    {
	      if (mpz_odd_p(n))
		mpz_add_ui(n, n, 1);
	    }
	}
      mpz_clear(rm);
      p = mpz_to_big_integer(sc, n);
      mpz_clear(n);
      return(p);
    }

  if ((g_is_nan(sc, args) == sc->T) ||
      (g_is_infinite(sc, args)) == sc->T)
    return(simple_out_of_range(sc, sc->round_symbol, p, (g_is_nan(sc, args) == sc->T) ? its_nan_string : its_infinite_string));

  {
    int32_t cmp_res;
    mpz_t fl, ce;
    mpfr_t x, dfl, dce;
    mpfr_init_set(x, big_real(p), GMP_RNDN);
    mpz_init(fl);
    mpfr_get_z(fl, x, GMP_RNDD);           /* fl = floor(x) */
    mpz_init(ce);
    mpfr_get_z(ce, x, GMP_RNDU);           /* ce = ceil(x) */
    mpfr_init(dfl);
    mpfr_sub_z(dfl, x, fl, GMP_RNDN);      /* dfl = x - fl */
    mpfr_init(dce);
    mpfr_sub_z(dce, x, ce, GMP_RNDN);      /* dce = -(ce - x) */
    mpfr_neg(dce, dce, GMP_RNDN);          /*    and reversed */
    cmp_res = mpfr_cmp(dfl, dce);
    if (cmp_res > 0)                       /* if (dfl > dce) return(ce) */
      mpz_init_set(n, ce);
    else
      {
	if (cmp_res < 0)                   /* if (dfl < dce) return(fl) */
	  mpz_init_set(n, fl);
	else
	  {
	    if (mpz_even_p(fl))
	      mpz_init_set(n, fl);         /* if (mod(fl, 2) == 0) return(fl) */
	    else mpz_init_set(n, ce);      /* else return(ce) */
	  }
      }
    mpz_clear(fl);
    mpz_clear(ce);
    mpfr_clear(dfl);
    mpfr_clear(dce);
    mpfr_clear(x);
    p = mpz_to_big_integer(sc, n);
    mpz_clear(n);
    return(p);
  }
}


static s7_pointer big_quotient(s7_scheme *sc, s7_pointer args)
{
  #define H_quotient "(quotient x1 x2) returns the integer quotient of x1 and x2; (quotient 4 3) = 1"
  #define Q_quotient pcl_r

  s7_pointer x, y, p;
  x = car(args);
  y = cadr(args);

  if (!s7_is_real(x))
    method_or_bust(sc, x, sc->quotient_symbol, args, T_REAL, 1);

  if (!s7_is_real(y))
    method_or_bust(sc, y, sc->quotient_symbol, args, T_REAL, 2);

  if ((s7_is_integer(x)) &&
      (s7_is_integer(y)))
    {
      mpz_t n;
      x = to_big(sc, x);
      y = to_big(sc, y);

      if (s7_is_zero(y))
	return(division_by_zero_error(sc, sc->quotient_symbol, args));

      mpz_init_set(n, big_integer(x));
      mpz_tdiv_q(n, n, big_integer(y));

      p = mpz_to_big_integer(sc, n);
      mpz_clear(n);
      return(p);
    }
  return(big_truncate(sc, set_plist_1(sc, big_divide(sc, args))));
}


static s7_pointer big_remainder(s7_scheme *sc, s7_pointer args)
{
  #define H_remainder "(remainder x1 x2) returns the integer remainder of x1 and x2; (remainder 10 3) = 1"
  #define Q_remainder pcl_r

  s7_pointer x, y, p;
  x = car(args);
  y = cadr(args);

  if (!s7_is_real(x))
    method_or_bust(sc, x, sc->remainder_symbol, args, T_REAL, 1);

  if (!s7_is_real(y))
    method_or_bust(sc, y, sc->remainder_symbol, args, T_REAL, 2);

  if ((s7_is_integer(x)) &&
      (s7_is_integer(y)))
    {
      mpz_t n;
      x = to_big(sc, x);
      y = to_big(sc, y);

      if (s7_is_zero(y))
	return(division_by_zero_error(sc, sc->remainder_symbol, args));

      mpz_init_set(n, big_integer(x));
      mpz_tdiv_r(n, n, big_integer(y));

      p = mpz_to_big_integer(sc, n);
      mpz_clear(n);
      return(p);
    }
  return(big_subtract(sc,
          list_2(sc, x,
           big_multiply(sc,
            set_plist_2(sc, y,
             big_quotient(sc, args))))));
}


static s7_pointer big_modulo(s7_scheme *sc, s7_pointer args)
{
  #define H_modulo "(modulo x1 x2) returns x1 mod x2; (modulo 4 3) = 1.  The arguments can be real numbers."
  #define Q_modulo pcl_r

  s7_pointer a, b, p;

  a = car(args);
  if (!s7_is_real(a))
    method_or_bust(sc, a, sc->modulo_symbol, args, T_REAL, 1);

  b = cadr(args);
  if (!s7_is_real(b))
    method_or_bust(sc, b, sc->modulo_symbol, args, T_REAL, 2);

  a = to_big(sc, a);
  b = to_big(sc, b);

  if ((s7_is_integer(a)) &&
      (s7_is_integer(b)))
    {
      s7_pointer x, y;
      int32_t cy, cz;
      mpz_t n;

      y = promote_number(sc, T_BIG_INTEGER, b);
      if (mpz_cmp_ui(big_integer(y), 0) == 0)
	return(a);

      x = promote_number(sc, T_BIG_INTEGER, a);
      /* mpz_mod is too tricky here */

      mpz_init_set(n, big_integer(x));
      mpz_fdiv_r(n, n, big_integer(y));
      cy = mpz_cmp_ui(big_integer(y), 0);
      cz = mpz_cmp_ui(n, 0);
      if (((cy < 0) && (cz > 0)) ||
	  ((cy > 0) && (cz < 0)))
	mpz_add(n, n, big_integer(y));

      p = mpz_to_big_integer(sc, n);
      mpz_clear(n);
      return(p);
    }
  return(big_subtract(sc,
          list_2(sc, a,
           big_multiply(sc,
            list_2(sc, b,
    	     big_floor(sc,
              set_plist_1(sc,
               big_divide(sc,
	        set_plist_2(sc, a, b)))))))));
}


static int32_t big_real_scan_args(s7_scheme *sc, s7_pointer args)
{
  int32_t i, result_type = T_INTEGER;
  s7_pointer arg;

  for (i = 1, arg = args; is_not_null(arg); i++, arg = cdr(arg))
    {
      s7_pointer p;
      p = car(arg);
      if (!is_real_via_method(sc, p))
	return(-i);
      result_type = get_result_type(sc, result_type, p);
    }
  return(result_type);
}


static s7_pointer big_max(s7_scheme *sc, s7_pointer args)
{
  int32_t result_type;
  s7_pointer x, result, arg;

  result_type = big_real_scan_args(sc, args);
  if (result_type < 0)
    return(wrong_type_argument(sc, sc->max_symbol, -result_type, s7_list_ref(sc, args, -1 - result_type), T_REAL));

  if (result_type < T_BIG_INTEGER)
    return(g_max(sc, args));

  if (!s7_is_number(car(args)))
    check_method(sc, car(args), sc->max_symbol, args);

  result = promote_number(sc, result_type, car(args));

  for (x = cdr(args); is_not_null(x); x = cdr(x))
    {
      if (!s7_is_number(car(x)))
	check_method(sc, car(x), sc->max_symbol, cons(sc, result, x));

      arg = promote_number(sc, result_type, car(x));
      switch (result_type)
	{
	case T_BIG_INTEGER: if (mpz_cmp(big_integer(result), big_integer(arg)) < 0) result = arg; break;
	case T_BIG_RATIO:   if (mpq_cmp(big_ratio(result), big_ratio(arg)) < 0)	    result = arg; break;
	case T_BIG_REAL:    if (mpfr_cmp(big_real(result), big_real(arg)) < 0)	    result = arg; break;
	}
    }
  if (result_type == T_BIG_RATIO) /* maybe actual result was an int32_t */
    {
      if (mpz_cmp_ui(mpq_denref(big_ratio(result)), 1) == 0)
	{
	  mpz_t n;
	  s7_pointer p;
	  mpz_init_set(n, mpq_numref(big_ratio(result)));
	  p = mpz_to_big_integer(sc, n);
	  mpz_clear(n);
	  return(p);
	}
    }
  return(result);
}


static s7_pointer big_min(s7_scheme *sc, s7_pointer args)
{
  int32_t result_type;
  s7_pointer x, result, arg;

  result_type = big_real_scan_args(sc, args);
  if (result_type < 0)
    return(wrong_type_argument(sc, sc->min_symbol, -result_type, s7_list_ref(sc, args, -1 - result_type), T_REAL));

  if (result_type < T_BIG_INTEGER)
    return(g_min(sc, args));

  if (!s7_is_number(car(args)))
    check_method(sc, car(args), sc->min_symbol, args);

  result = promote_number(sc, result_type, car(args));

  for (x = cdr(args); is_not_null(x); x = cdr(x))
    {
      if (!s7_is_number(car(x)))
	check_method(sc, car(x), sc->min_symbol, cons(sc, result, x));

      arg = promote_number(sc, result_type, car(x));
      switch (result_type)
	{
	case T_BIG_INTEGER: if (mpz_cmp(big_integer(result), big_integer(arg)) > 0)  result = arg; break;
	case T_BIG_RATIO:   if (mpq_cmp(big_ratio(result), big_ratio(arg)) > 0)      result = arg; break;
	case T_BIG_REAL:    if (mpfr_cmp(big_real(result), big_real(arg)) > 0)       result = arg; break;
	}
    }
  if (result_type == T_BIG_RATIO) /* maybe actual result was an int32_t */
    {
      if (mpz_cmp_ui(mpq_denref(big_ratio(result)), 1) == 0)
	{
	  mpz_t n;
	  s7_pointer p;
	  mpz_init_set(n, mpq_numref(big_ratio(result)));
	  p = mpz_to_big_integer(sc, n);
	  mpz_clear(n);
	  return(p);
	}
    }
  return(result);
}


static s7_pointer big_less(s7_scheme *sc, s7_pointer args)
{
  #define H_less "(< x1 ...) returns #t if its arguments are in increasing order"
  #define Q_less s7_make_circular_signature(sc, 1, 2, sc->is_boolean_symbol, sc->is_real_symbol)

  int32_t result_type;
  s7_pointer x, previous, current;

  result_type = big_real_scan_args(sc, args);
  if (result_type < 0)
    return(wrong_type_argument(sc, sc->lt_symbol, -result_type, s7_list_ref(sc, args, -1 - result_type), T_REAL));

  /* don't try to use g_less here */
  if (result_type < T_BIG_INTEGER)
    result_type += 4;

  if (!s7_is_number(car(args)))
    check_method(sc, car(args), sc->lt_symbol, args);

  previous = promote_number(sc, result_type, car(args));

  for (x = cdr(args); is_not_null(x); x = cdr(x))
    {
      if (!s7_is_number(car(x)))
	check_method(sc, car(x), sc->lt_symbol, cons(sc, previous, x));

      current = promote_number(sc, result_type, car(x));
      switch (result_type)
	{
	case T_BIG_INTEGER: if (mpz_cmp(big_integer(previous), big_integer(current)) >= 0) return(sc->F); break;
	case T_BIG_RATIO:   if (mpq_cmp(big_ratio(previous),   big_ratio(current)) >= 0) return(sc->F);	  break;
	case T_BIG_REAL:    if (mpfr_cmp(big_real(previous),   big_real(current)) >= 0) return(sc->F);	  break;
	}
      previous = current;
    }
  return(sc->T);
}


static s7_pointer big_less_or_equal(s7_scheme *sc, s7_pointer args)
{
  #define H_less_or_equal "(<= x1 ...) returns #t if its arguments are in increasing order"
  #define Q_less_or_equal s7_make_circular_signature(sc, 1, 2, sc->is_boolean_symbol, sc->is_real_symbol)

  int32_t result_type;
  s7_pointer x, previous, current;

  result_type = big_real_scan_args(sc, args);
  if (result_type < 0)
    return(wrong_type_argument(sc, sc->leq_symbol, -result_type, s7_list_ref(sc, args, -1 - result_type), T_REAL));

  if (result_type < T_BIG_INTEGER)
    result_type += 4;

  if (!s7_is_number(car(args)))
    check_method(sc, car(args), sc->leq_symbol, args);

  previous = promote_number(sc, result_type, car(args));

  for (x = cdr(args); is_not_null(x); x = cdr(x))
    {
      if (!s7_is_number(car(x)))
	check_method(sc, car(x), sc->leq_symbol, cons(sc, previous, x));

      current = promote_number(sc, result_type, car(x));
      switch (result_type)
	{
	case T_BIG_INTEGER: if (mpz_cmp(big_integer(previous), big_integer(current)) > 0) return(sc->F);  break;
	case T_BIG_RATIO:   if (mpq_cmp(big_ratio(previous),   big_ratio(current)) > 0) return(sc->F);	  break;
	case T_BIG_REAL:    if (mpfr_cmp(big_real(previous),   big_real(current)) > 0) return(sc->F);	  break;
	}
      previous = current;
    }
  return(sc->T);
}


static s7_pointer big_greater(s7_scheme *sc, s7_pointer args)
{
  #define H_greater "(> x1 ...) returns #t if its arguments are in decreasing order"
  #define Q_greater s7_make_circular_signature(sc, 1, 2, sc->is_boolean_symbol, sc->is_real_symbol)

  int32_t result_type;
  s7_pointer x, previous, current;

  result_type = big_real_scan_args(sc, args);
  if (result_type < 0)
    return(wrong_type_argument(sc, sc->gt_symbol, -result_type, s7_list_ref(sc, args, -1 - result_type), T_REAL));

  if (result_type < T_BIG_INTEGER)
    result_type += 4;

  if (!s7_is_number(car(args)))
    check_method(sc, car(args), sc->gt_symbol, args);

  previous = promote_number(sc, result_type, car(args));

  for (x = cdr(args); is_not_null(x); x = cdr(x))
    {
      if (!s7_is_number(car(x)))
	check_method(sc, car(x), sc->gt_symbol, cons(sc, previous, x));
      current = promote_number(sc, result_type, car(x));
      switch (result_type)
	{
	case T_BIG_INTEGER: if (mpz_cmp(big_integer(previous), big_integer(current)) <= 0) return(sc->F); break;
	case T_BIG_RATIO:   if (mpq_cmp(big_ratio(previous),   big_ratio(current)) <= 0) return(sc->F);	  break;
	case T_BIG_REAL:    if (mpfr_cmp(big_real(previous),   big_real(current)) <= 0) return(sc->F);	  break;
	}
      previous = current;
    }
  return(sc->T);
}


static s7_pointer big_greater_or_equal(s7_scheme *sc, s7_pointer args)
{
  #define H_greater_or_equal "(>= x1 ...) returns #t if its arguments are in decreasing order"
  #define Q_greater_or_equal s7_make_circular_signature(sc, 1, 2, sc->is_boolean_symbol, sc->is_real_symbol)

  int32_t result_type;
  s7_pointer x, previous, current;

  result_type = big_real_scan_args(sc, args);
  if (result_type < 0)
    return(wrong_type_argument(sc, sc->geq_symbol, -result_type, s7_list_ref(sc, args, -1 - result_type), T_REAL));

  if (result_type < T_BIG_INTEGER)
    result_type += 4;

  if (!s7_is_number(car(args)))
    check_method(sc, car(args), sc->geq_symbol, args);
  previous = promote_number(sc, result_type, car(args));

  for (x = cdr(args); is_not_null(x); x = cdr(x))
    {
      if (!s7_is_number(car(x)))
	check_method(sc, car(x), sc->geq_symbol, cons(sc, previous, x));
      current = promote_number(sc, result_type, car(x));
      switch (result_type)
	{
	case T_BIG_INTEGER: if (mpz_cmp(big_integer(previous), big_integer(current)) < 0) return(sc->F);  break;
	case T_BIG_RATIO:   if (mpq_cmp(big_ratio(previous),   big_ratio(current)) < 0) return(sc->F);	  break;
	case T_BIG_REAL:    if (mpfr_cmp(big_real(previous),   big_real(current)) < 0) return(sc->F);	  break;
	}
      previous = current;
    }
  return(sc->T);
}


static s7_pointer big_equal(s7_scheme *sc, s7_pointer args)
{
  #define Q_equal s7_make_circular_signature(sc, 1, 2, sc->is_boolean_symbol, sc->is_number_symbol)

  /* this is morally-equal? for bignums, the other case goes through big_numbers_are_eqv */
  int32_t result_type = T_INTEGER;
  s7_pointer x, y, result;
  bool got_nan = false;

  for (x = args; is_not_null(x); x = cdr(x))
    {
      s7_pointer p;
      p = car(x);
      if (!s7_is_number(p))
	{
	  check_method(sc, car(args), sc->eq_symbol, x);
	  return(wrong_type_argument_with_type(sc, sc->eq_symbol, position_of(x, args), p, a_number_string));
	}

      result_type = get_result_type(sc, result_type, p);
      if (!got_nan)
	got_nan = (((is_t_real(p)) && (is_NaN(real(p)))) ||  /* (= (bignum "3") 1/0) */
		   ((is_t_complex(p)) && ((is_NaN(real_part(p))) || (is_NaN(imag_part(p))))));
    }
  if (got_nan) return(sc->F); /* put this off until here so that non-numbers anywhere in the arg list will raise an error */

  if (result_type < T_BIG_INTEGER)
    return(g_equal(sc, args));

  result = promote_number(sc, result_type, car(args));
  for (y = cdr(args); is_not_null(y); y = cdr(y))
    {
      s7_pointer arg;
      arg = promote_number(sc, result_type, car(y));
      switch (result_type)
	{
	case T_BIG_INTEGER: 
	  if (mpz_cmp(big_integer(result), big_integer(arg)) != 0) return(sc->F);  
	  break;

	case T_BIG_RATIO:   
	  if (mpq_cmp(big_ratio(result),   big_ratio(arg)) != 0)   return(sc->F);  
	  break;

	case T_BIG_REAL: 
	  {
	    mpfr_t *a1;
	    a1 = s7_double_to_mpfr(sc->morally_equal_float_epsilon);
	    if (mpfr_cmp(big_real(big_abs(sc, set_plist_1(sc, big_subtract(sc, set_plist_2(sc, result, arg))))), *a1) > 0)
	      return(sc->F);
	  }
	  break;

	case T_BIG_COMPLEX: 
	  {
	    mpfr_t *a1;
	    a1 = s7_double_to_mpfr(sc->morally_equal_float_epsilon);
	    if (mpfr_cmp(big_real(big_magnitude(sc, set_plist_1(sc, big_subtract(sc, set_plist_2(sc, result, arg))))), *a1) > 0)
	      return(sc->F);
	  }
	  break;
	}
    }
  return(sc->T);
}


static s7_pointer big_gcd(s7_scheme *sc, s7_pointer args)
{
  #define H_gcd "(gcd ...) returns the greatest common divisor of its rational arguments"
  #define Q_gcd pcl_f

  bool rats = false;
  s7_pointer x, lst;

  if (is_null(args)) return(small_int(0));

  for (x = args; is_pair(x); x = cdr(x))
    {
      if (!is_rational_via_method(sc, car(x)))
	return(wrong_type_argument_with_type(sc, sc->gcd_symbol, position_of(x, args), car(x), a_rational_string));
      if (!rats)
	rats = (!is_integer_via_method(sc, car(x)));
    }

  if (is_null(cdr(args)))    /* (gcd -2305843009213693951/4611686018427387903) */
    return(big_abs(sc, args));

  if (!rats)
    {
      mpz_t n;
      mpz_init(n);
      for (x = args; is_not_null(x); x = cdr(x))
	{
	  if (!s7_is_number(car(x)))
	    {
	      lst = cons(sc, mpz_to_big_integer(sc, n), x);
	      mpz_clear(n);
	      method_or_bust(sc, car(x), sc->gcd_symbol, lst, T_INTEGER, position_of(x, args));
	    }
	  mpz_gcd(n, n, big_integer(promote_number(sc, T_BIG_INTEGER, car(x))));
	  if (mpz_cmp_ui(n, 1) == 0)
	    {
	      mpz_clear(n);
	      return(small_int(1));
	    }
	}
      x = mpz_to_big_integer(sc, n);
      mpz_clear(n);
      return(x);
    }

  {
    s7_pointer rat;
    mpq_t q;
    mpz_t n, d;

    if (!s7_is_number(car(args)))
      check_method(sc, car(args), sc->gcd_symbol, args);

    rat = promote_number(sc, T_BIG_RATIO, car(args));
    mpz_init_set(n, mpq_numref(big_ratio(rat)));
    mpz_init_set(d, mpq_denref(big_ratio(rat)));
    for (x = cdr(args); is_not_null(x); x = cdr(x))
      {
	if (!s7_is_number(car(x)))
	  {
	    mpq_init(q);
	    mpq_set_num(q, n);
	    mpq_set_den(q, d);
	    lst = cons(sc, mpq_to_big_ratio(sc, q), x);
	    mpz_clear(n);
	    mpz_clear(d);
	    mpq_clear(q);
	    method_or_bust_with_type(sc, car(x), sc->gcd_symbol, lst, a_rational_string, position_of(x, args));
	  }
	rat = promote_number(sc, T_BIG_RATIO, car(x));
	mpz_gcd(n, n, mpq_numref(big_ratio(rat)));
	mpz_lcm(d, d, mpq_denref(big_ratio(rat)));
      }
    if (mpz_cmp_ui(d, 1) == 0)
      {
	rat = mpz_to_big_integer(sc, n);
	mpz_clear(n);
	mpz_clear(d);
	return(rat);
      }

    mpq_init(q);
    mpq_set_num(q, n);
    mpq_set_den(q, d);
    mpz_clear(n);
    mpz_clear(d);

    x = mpq_to_big_ratio(sc, q);
    mpq_clear(q);
    return(x);
  }
}


static s7_pointer big_lcm(s7_scheme *sc, s7_pointer args)
{
  #define H_lcm "(lcm ...) returns the least common multiple of its rational arguments"
  #define Q_lcm pcl_f

  s7_pointer x, lst;
  bool rats = false;

  if (is_null(args)) return(small_int(1));

  for (x = args; is_pair(x); x = cdr(x))
    {
      if (!is_rational_via_method(sc, car(x)))
	return(wrong_type_argument_with_type(sc, sc->lcm_symbol, position_of(x, args), car(x), a_rational_string));
      if (!rats)
	rats = (!is_integer_via_method(sc, car(x)));
    }

   if (is_null(cdr(args)))    /* (lcm -2305843009213693951/4611686018427387903) */
      return(big_abs(sc, args));

   if (!rats)
    {
      mpz_t n;
      mpz_init(n);
      mpz_set_ui(n, 1);
      for (x = args; is_not_null(x); x = cdr(x))
	{
	  if (!s7_is_number(car(x)))
	    {
	      lst = cons(sc, mpz_to_big_integer(sc, n), x);
	      mpz_clear(n);
	      method_or_bust(sc, car(x), sc->lcm_symbol, lst, T_INTEGER, position_of(x, args));
	    }
	  mpz_lcm(n, n, big_integer(promote_number(sc, T_BIG_INTEGER, car(x))));
	  if (mpz_cmp_ui(n, 0) == 0)
	    {
	      mpz_clear(n);
	      return(small_int(0));
	    }
	}
      x = mpz_to_big_integer(sc, n);
      mpz_clear(n);
      return(x);
    }

  {
    s7_pointer rat;
    mpq_t q;
    mpz_t n, d;

    if (!s7_is_number(car(args)))
      check_method(sc, car(args), sc->lcm_symbol, args);

    rat = promote_number(sc, T_BIG_RATIO, car(args));
    mpz_init_set(n, mpq_numref(big_ratio(rat)));
    if (mpz_cmp_ui(n, 0) == 0)
      {
	mpz_clear(n);
	return(small_int(0));
      }

    mpz_init_set(d, mpq_denref(big_ratio(rat)));
    for (x = cdr(args); is_not_null(x); x = cdr(x))
      {
	if (!s7_is_number(car(x)))
	  {
	    mpq_init(q);
	    mpq_set_num(q, n);
	    mpq_set_den(q, d);
	    lst = cons(sc, mpq_to_big_ratio(sc, q), x);
	    mpz_clear(n);
	    mpz_clear(d);
	    mpq_clear(q);
	    method_or_bust_with_type(sc, car(x), sc->lcm_symbol, lst, a_rational_string, position_of(x, args));
	  }

	rat = promote_number(sc, T_BIG_RATIO, car(x));
	mpz_lcm(n, n, mpq_numref(big_ratio(rat)));
	if (mpz_cmp_ui(n, 0) == 0)
	  {
	    mpz_clear(n);
	    mpz_clear(d);
	    return(small_int(0));
	  }
	mpz_gcd(d, d, mpq_denref(big_ratio(rat)));
      }

    if (mpz_cmp_ui(d, 1) == 0)
      {
	rat = mpz_to_big_integer(sc, n);
	mpz_clear(n);
	mpz_clear(d);
	return(rat);
      }

    mpq_init(q);
    mpq_set_num(q, n);
    mpq_set_den(q, d);
    mpz_clear(n);
    mpz_clear(d);
    x = mpq_to_big_ratio(sc, q);
    mpq_clear(q);
    return(x);
  }
}


static s7_pointer set_bignum_precision(s7_scheme *sc, int32_t precision)
{
  mp_prec_t bits;
  if (precision <= 1)                   /* (set! (*s7* 'bignum-precision) 1) causes mpfr to segfault! (also 0 and -1) */
    return(s7_out_of_range_error(sc, "set! (*s7* 'bignum-precision)", 0, make_integer(sc, precision), "has to be greater than 1"));

  bits = (mp_prec_t)precision;
  mpfr_set_default_prec(bits);
  mpc_set_default_precision(bits);
  s7_symbol_set_value(sc, sc->pi_symbol, big_pi(sc));
  return(sc->F);
}


static s7_pointer big_random_state(s7_scheme *sc, s7_pointer args)
{
  #define H_random_state "(random-state seed) returns a new random number state initialized with 'seed'. \
Pass this as the second argument to 'random' to get a repeatable random number sequence:\n\
    (let ((seed (random-state 1234))) (random 1.0 seed))"
  #define Q_random_state s7_make_circular_signature(sc, 1, 2, sc->is_random_state_symbol, sc->is_integer_symbol)

  s7_pointer r, seed;
  seed = car(args);
  if (!s7_is_integer(seed))
    method_or_bust_one_arg(sc, seed, sc->random_state_symbol, args, T_INTEGER);

  if (type(seed) != T_BIG_INTEGER)
    seed = promote_number(sc, T_BIG_INTEGER, seed);

  new_cell(sc, r, T_RANDOM_STATE);
  gmp_randinit_default(random_gmp_state(r));
  gmp_randseed(random_gmp_state(r), big_integer(seed));
  return(r);
}


static s7_pointer big_random(s7_scheme *sc, s7_pointer args)
{
  #define H_random "(random num (state #f)) returns a random number between 0 and num (0 if num=0)."
  #define Q_random s7_make_signature(sc, 3, sc->is_number_symbol, sc->is_number_symbol, sc->is_random_state_symbol)
  s7_pointer num, state, x;

  num = car(args);
  if (!s7_is_number(num))
    method_or_bust_with_type(sc, num, sc->random_symbol, args, a_number_string, 1);

  state = sc->default_rng;
  if (is_not_null(cdr(args)))
    {
      state = cadr(args);
      if (!is_random_state(state))
	return(wrong_type_argument_with_type(sc, sc->random_symbol, 2, state, a_random_state_object_string));
    }

  if (s7_is_zero(num))
    return(num);

  if (!is_big_number(num))
    {
      switch (type(num))
	{
	case T_INTEGER: num = promote_number(sc, T_BIG_INTEGER, num);  break;
	case T_RATIO:   num = promote_number(sc, T_BIG_RATIO, num);    break;
	case T_REAL:    num = promote_number(sc, T_BIG_REAL, num);     break;
	default:        num = promote_number(sc, T_BIG_COMPLEX, num);  break;
	}
    }

  switch (type(num))
    {
    case T_BIG_INTEGER:
      {
	mpz_t n;
	mpz_init(n);
	mpz_urandomm(n, random_gmp_state(state), big_integer(num));
	
	/* this does not work if num is a negative number -- you get positive results.
	 *   so check num for sign, and negate result if necessary.
	 */
	if (mpz_cmp_ui(big_integer(num), 0) < 0)
	  mpz_neg(n, n);
	
	x = mpz_to_big_integer(sc, n);
	mpz_clear(n);
	return(x);
      }
      
    case T_BIG_RATIO:
      {
	mpfr_t n, e;
	mpfr_t rat;
	
	mpfr_init_set_ui(n, 1, GMP_RNDN);
	mpfr_urandomb(n, random_gmp_state(state));
	mpfr_init_set_q(rat, big_ratio(num), GMP_RNDN);
	mpfr_mul(n, n, rat, GMP_RNDN);
	
	mpfr_init_set_str(e, "0.0000001", 10, GMP_RNDN);
	mpfr_mul(e, e, rat, GMP_RNDN);
	mpfr_clear(rat);
	/* as in g_random, small ratios are a problem because the error term (sc->default_rationalize_error = 1e-12 here)
	 *   clobbers everything to 0.
	 */
	x = big_rationalize(sc, set_plist_2(sc, mpfr_to_big_real(sc, n), mpfr_to_big_real(sc, e)));
	mpfr_clear(n);
	mpfr_clear(e);
	return(x);
      }
      
    case T_BIG_REAL:
      {
	mpfr_t n;
	mpfr_init_set_ui(n, 1, GMP_RNDN);
	mpfr_urandomb(n, random_gmp_state(state));
	mpfr_mul(n, n, big_real(num), GMP_RNDN);
	x = mpfr_to_big_real(sc, n);
	mpfr_clear(n);
	return(x);
      }
      
    case T_BIG_COMPLEX:
      {
	mpc_t n;
	mpc_init(n);
	mpc_urandom(n, random_gmp_state(state));
	mpfr_mul(mpc_realref(n), mpc_realref(n), mpc_realref(big_complex(num)), GMP_RNDN);
	mpfr_mul(mpc_imagref(n), mpc_imagref(n), mpc_imagref(big_complex(num)), GMP_RNDN);
	x = mpc_to_big_complex(sc, n);
	mpc_clear(n);
	return(x);
      }
    }
  return(sc->F); /* make the compiler happy */
}

s7_double s7_random(s7_scheme *sc, s7_pointer state)
{
  s7_pointer p;
  p = big_random(sc, set_plist_1(sc, (state) ? state : sc->default_rng));
  return((s7_double)mpfr_get_d(big_real(p), GMP_RNDN));
}


static void s7_gmp_init(s7_scheme *sc)
{
  #define big_defun(Scheme_Name, C_Name, Req, Opt, Rst) s7_define_typed_function(sc, Scheme_Name, big_ ## C_Name, Req, Opt, Rst, H_ ## C_Name, Q_ ## C_Name)
  #define c_big_defun(Scheme_Name, C_Name, Req, Opt, Rst) s7_define_typed_function(sc, Scheme_Name, c_big_ ## C_Name, Req, Opt, Rst, H_ ## C_Name, Q_ ## C_Name)

  sc->add_symbol =              big_defun("+",                add,              0, 0, true);
  sc->subtract_symbol =         big_defun("-",                subtract,         1, 0, true);
  sc->multiply_symbol =         big_defun("*",                multiply,         0, 0, true);
  sc->divide_symbol =           big_defun("/",                divide,           1, 0, true);
  sc->max_symbol =              big_defun("max",              max,              1, 0, true);
  sc->min_symbol =              big_defun("min",              min,              1, 0, true);
  sc->lt_symbol =               big_defun("<",                less,             2, 0, true);
  sc->leq_symbol =              big_defun("<=",               less_or_equal,    2, 0, true);
  sc->gt_symbol =               big_defun(">",                greater,          2, 0, true);
  sc->geq_symbol =              big_defun(">=",               greater_or_equal, 2, 0, true);
  sc->eq_symbol =               big_defun("=",                equal,            2, 0, true);
  sc->rationalize_symbol =      big_defun("rationalize",      rationalize,      1, 1, false);
#if (!WITH_PURE_S7)
  sc->exact_to_inexact_symbol = big_defun("exact->inexact",   exact_to_inexact, 1, 0, false);
  sc->inexact_to_exact_symbol = big_defun("inexact->exact",   inexact_to_exact, 1, 0, false);
  sc->integer_length_symbol =   big_defun("integer-length",   integer_length,   1, 0, false);
  sc->make_rectangular_symbol = c_big_defun("make-rectangular", complex,        2, 0, false);
  sc->make_polar_symbol =       big_defun("make-polar",       make_polar,       2, 0, false);
#endif
  sc->floor_symbol =            big_defun("floor",            floor,            1, 0, false);
  sc->ceiling_symbol =          big_defun("ceiling",          ceiling,          1, 0, false);
  sc->truncate_symbol =         big_defun("truncate",         truncate,         1, 0, false);
  sc->round_symbol =            big_defun("round",            round,            1, 0, false);
  sc->quotient_symbol =         big_defun("quotient",         quotient,         2, 0, false);
  sc->remainder_symbol =        big_defun("remainder",        remainder,        2, 0, false);
  sc->modulo_symbol =           big_defun("modulo",           modulo,           2, 0, false);
  sc->gcd_symbol =              big_defun("gcd",              gcd,              0, 0, true);
  sc->lcm_symbol =              big_defun("lcm",              lcm,              0, 0, true);
  sc->complex_symbol =          c_big_defun("complex",        complex,          2, 0, false);
  sc->magnitude_symbol =        big_defun("magnitude",        magnitude,        1, 0, false);
  sc->angle_symbol =            big_defun("angle",            angle,            1, 0, false);
  sc->abs_symbol =              big_defun("abs",              abs,              1, 0, false);
  sc->lognot_symbol =           big_defun("lognot",           lognot,           1, 0, false);
  sc->logior_symbol =           big_defun("logior",           logior,           0, 0, true);
  sc->logxor_symbol =           big_defun("logxor",           logxor,           0, 0, true);
  sc->logand_symbol =           big_defun("logand",           logand,           0, 0, true);
  sc->ash_symbol =              big_defun("ash",              ash,              2, 0, false);
  sc->exp_symbol =              big_defun("exp",              exp,              1, 0, false);
  sc->expt_symbol =             big_defun("expt",             expt,             2, 0, false);
  sc->log_symbol =              big_defun("log",              log,              1, 1, false);
  sc->sqrt_symbol =             big_defun("sqrt",             sqrt,             1, 0, false);
  sc->sin_symbol =              big_defun("sin",              sin,              1, 0, false);
  sc->cos_symbol =              big_defun("cos",              cos,              1, 0, false);
  sc->tan_symbol =              big_defun("tan",              tan,              1, 0, false);
  sc->asin_symbol =             big_defun("asin",             asin,             1, 0, false);
  sc->acos_symbol =             big_defun("acos",             acos,             1, 0, false);
  sc->atan_symbol =             big_defun("atan",             atan,             1, 1, false);
  sc->sinh_symbol =             big_defun("sinh",             sinh,             1, 0, false);
  sc->cosh_symbol =             big_defun("cosh",             cosh,             1, 0, false);
  sc->tanh_symbol =             big_defun("tanh",             tanh,             1, 0, false);
  sc->asinh_symbol =            big_defun("asinh",            asinh,            1, 0, false);
  sc->acosh_symbol =            big_defun("acosh",            acosh,            1, 0, false);
  sc->atanh_symbol =            big_defun("atanh",            atanh,            1, 0, false);

  sc->random_symbol =           big_defun("random",           random,           1, 1, false);
  sc->random_state_symbol =     big_defun("random-state",     random_state,     1, 1, false);

  sc->is_bignum_symbol =        big_defun("bignum?",          is_bignum,        1, 0, false); /* needed by Q_bignum below */
  sc->bignum_symbol =           big_defun("bignum",           bignum,           1, 1, false);

  sc->bignum_precision = DEFAULT_BIGNUM_PRECISION;
  mpfr_set_default_prec((mp_prec_t)DEFAULT_BIGNUM_PRECISION);
  mpc_set_default_precision((mp_prec_t)DEFAULT_BIGNUM_PRECISION);

  s7_symbol_set_value(sc, sc->pi_symbol, big_pi(sc));

  /* if these fixnum limits were read as strings, they'd be bignums in the gmp case,
   *   so for consistency make the symbolic versions bignums as well.
   */
  s7_symbol_set_value(sc, make_symbol(sc, "most-positive-fixnum"), s7_int_to_big_integer(sc, s7_integer(s7_name_to_value(sc, "most-positive-fixnum"))));
  s7_symbol_set_value(sc, make_symbol(sc, "most-negative-fixnum"), s7_int_to_big_integer(sc, s7_integer(s7_name_to_value(sc, "most-negative-fixnum"))));

  s7_provide(sc, "gmp");
}

#endif
/* WITH_GMP */



/* -------------------------------- *s7* environment -------------------------------- */

static void init_s7_let(s7_scheme *sc)
{
  sc->stack_top_symbol =                     s7_make_symbol(sc, "stack-top");
  sc->stack_size_symbol =                    s7_make_symbol(sc, "stack-size");
  sc->stacktrace_defaults_symbol =           s7_make_symbol(sc, "stacktrace-defaults");
  sc->heap_size_symbol =                     s7_make_symbol(sc, "heap-size");
  sc->free_heap_size_symbol =                s7_make_symbol(sc, "free-heap-size");
  sc->gc_freed_symbol =                      s7_make_symbol(sc, "gc-freed");
  sc->gc_protected_objects_symbol =          s7_make_symbol(sc, "gc-protected-objects");
  set_immutable(sc->gc_protected_objects_symbol);

  sc->input_ports_symbol =                   s7_make_symbol(sc, "input-ports");
  sc->output_ports_symbol =                  s7_make_symbol(sc, "output-ports");
  sc->strings_symbol =                       s7_make_symbol(sc, "strings");
  sc->gensyms_symbol =                       s7_make_symbol(sc, "gensyms");
  sc->vectors_symbol =                       s7_make_symbol(sc, "vectors");
  sc->hash_tables_symbol =                   s7_make_symbol(sc, "hash-tables");
  sc->continuations_symbol =                 s7_make_symbol(sc, "continuations");

  sc->c_objects_symbol =                     s7_make_symbol(sc, "c-objects");
  sc->file_names_symbol =                    s7_make_symbol(sc, "file-names");
  sc->rootlet_size_symbol =                  s7_make_symbol(sc, "rootlet-size");
  sc->c_types_symbol =                       s7_make_symbol(sc, "c-types");
  sc->safety_symbol =                        s7_make_symbol(sc, "safety");
  sc->undefined_identifier_warnings_symbol = s7_make_symbol(sc, "undefined-identifier-warnings");
  sc->gc_stats_symbol =                      s7_make_symbol(sc, "gc-stats");
  sc->max_stack_size_symbol =                s7_make_symbol(sc, "max-stack-size");
  sc->cpu_time_symbol =                      s7_make_symbol(sc, "cpu-time");
  sc->catches_symbol =                       s7_make_symbol(sc, "catches");
  sc->exits_symbol =                         s7_make_symbol(sc, "exits");
  sc->stack_symbol =                         s7_make_symbol(sc, "stack");
  sc->max_string_length_symbol =             s7_make_symbol(sc, "max-string-length");
  sc->max_list_length_symbol =               s7_make_symbol(sc, "max-list-length");
  sc->max_vector_length_symbol =             s7_make_symbol(sc, "max-vector-length");
  sc->max_vector_dimensions_symbol =         s7_make_symbol(sc, "max-vector-dimensions");
  sc->default_hash_table_length_symbol =     s7_make_symbol(sc, "default-hash-table-length");
  sc->initial_string_port_length_symbol =    s7_make_symbol(sc, "initial-string-port-length");
  sc->default_rationalize_error_symbol =     s7_make_symbol(sc, "default-rationalize-error");
  sc->default_random_state_symbol =          s7_make_symbol(sc, "default-random-state");
  sc->morally_equal_float_epsilon_symbol =   s7_make_symbol(sc, "morally-equal-float-epsilon");
  sc->hash_table_float_epsilon_symbol =      s7_make_symbol(sc, "hash-table-float-epsilon");
  sc->print_length_symbol =                  s7_make_symbol(sc, "print-length");
  sc->bignum_precision_symbol =              s7_make_symbol(sc, "bignum-precision");
  sc->memory_usage_symbol =                  s7_make_symbol(sc, "memory-usage");
  sc->float_format_precision_symbol =        s7_make_symbol(sc, "float-format-precision");
  sc->history_symbol =                       s7_make_symbol(sc, "history");
  sc->history_size_symbol =                  s7_make_symbol(sc, "history-size");
  sc->profile_info_symbol =                  s7_make_symbol(sc, "profile-info");
  sc->autoloading_symbol =                   s7_make_symbol(sc, "autoloading?");
}

#ifdef __linux__
  #include <sys/resource.h>
#endif

static s7_pointer describe_memory_usage(s7_scheme *sc)
{
  /* heap, permanent, stack?, doc strings, sigs, c_func structs (and ports etc), mx_alloc, output bufs,
   *   sinc_tables, c-objects, rc_data, strbuf/tmpbuf[reallocs], autoload tables, hash_entrys, symbol_table,
   *   small_ints?=permanent
   */
  int32_t i, syms = 0, len, n;
  s7_pointer x;
  gc_list *gp;
  char buf[1024];

#ifdef __linux__
  struct rusage info;
  struct timeval ut;
  getrusage(RUSAGE_SELF, &info);
  ut = info.ru_utime;
  n = snprintf(buf, 1024, "process size: %" PRId64 ", time: %ld.%d\n", (s7_int)(info.ru_maxrss * 1024), ut.tv_sec, (int)floor(ut.tv_usec / 1000.0));
  port_write_string(sc->output_port)(sc, buf, n, sc->output_port);
#endif

  n = snprintf(buf, 1024, "heap: %" PRId64 " (%" PRId64 " bytes)", sc->heap_size, (s7_int)(sc->heap_size * (sizeof(s7_pointer) + sizeof(s7_cell))));
  port_write_string(sc->output_port)(sc, buf, n, sc->output_port);

  {
    int64_t k;
    int32_t ts[NUM_TYPES];
    for (i = 0; i < NUM_TYPES; i++) ts[i] = 0;
    for (k = 0; k < sc->heap_size; k++)
      ts[unchecked_type(sc->heap[k])]++;
    for (i = 0; i < NUM_TYPES; i++)
      {
	if ((i % 10) == 0) 
	  {
	    n = snprintf(buf, 1024, "\n ");
	    port_write_string(sc->output_port)(sc, buf, n, sc->output_port);
	  }
	
	if (ts[i] > 100)
	  n = snprintf(buf, 1024, " (%s): %d", type_name_from_type(sc, i, NO_ARTICLE), ts[i]);
	else n = snprintf(buf, 1024, " %d", ts[i]);
	port_write_string(sc->output_port)(sc, buf, n, sc->output_port);
      }
    n = snprintf(buf, 1024, "\n");
    port_write_string(sc->output_port)(sc, buf, n, sc->output_port);
  }
  n = snprintf(buf, 1024, "permanent cells: %d (%" PRId64 " bytes)\n", permanent_cells, (s7_int)(permanent_cells * sizeof(s7_cell)));
  port_write_string(sc->output_port)(sc, buf, n, sc->output_port);
  n = snprintf(buf, 1024, "gc protected size: %u, unused: %d\n", sc->protected_objects_size, sc->gpofl_loc);
  port_write_string(sc->output_port)(sc, buf, n, sc->output_port);
      
  for (i = 0; i < SYMBOL_TABLE_SIZE; i++)
    for (x = vector_element(sc->symbol_table, i); is_not_null(x); x = cdr(x))
      syms++;
  n = snprintf(buf, 1024, "symbol table: %d (%d symbols, %" PRId64 " bytes)\n", SYMBOL_TABLE_SIZE, syms, 
	       (s7_int)(SYMBOL_TABLE_SIZE * sizeof(s7_pointer) + syms * 3 * sizeof(s7_cell)));
  port_write_string(sc->output_port)(sc, buf, n, sc->output_port);

  n = snprintf(buf, 1024, "stack: %u (%" PRId64 " bytes, current top: %ld)\n", sc->stack_size, (s7_int)(sc->stack_size * sizeof(s7_pointer)), (long int)s7_stack_top(sc));
  port_write_string(sc->output_port)(sc, buf, n, sc->output_port);
  n = snprintf(buf, 1024, "c_functions: %d (%d bytes)\n", c_functions, (int)(c_functions * sizeof(c_proc_t)));
  port_write_string(sc->output_port)(sc, buf, n, sc->output_port);

  len = 0;
  gp = sc->strings;
  for (i = 0; i < (int32_t)(gp->loc); i++)
    len += string_length(gp->list[i]);
  syms = gp->loc;
  gp = sc->strings1;
  for (i = 0; i < (int32_t)(gp->loc); i++)
    len += string_length(gp->list[i]);
  n = snprintf(buf, 1024, "strings: %u, %d bytes\n", syms + gp->loc, len); /* also doc strings, permanent strings, etc */
  port_write_string(sc->output_port)(sc, buf, n, sc->output_port);

  {
    int32_t hs;
    hash_entry_t *p;
    for (hs = 0, p = hash_free_list; p; p = (hash_entry_t *)(p->next), hs++);

    len = 0;
    gp = sc->hash_tables;
    for (i = 0; i < (int32_t)(gp->loc); i++)
      len += (hash_table_mask(gp->list[i]) + 1);
    
    n = snprintf(buf, 1024, "hash tables: %" PRId32 " (entries in use: %d, free: %d)\n", (int32_t)(gp->loc), len, hs);
    port_write_string(sc->output_port)(sc, buf, n, sc->output_port);
  }
  {
    int64_t len = 0, flen = 0, ilen = 0;
    gp = sc->vectors;
    for (i = 0; i < (int32_t)(gp->loc); i++)
      {
	s7_pointer v;
	v = gp->list[i];
	if (is_float_vector(v))
	  flen += vector_length(v);
	else
	  {
	    if (is_int_vector(v))
	      ilen += vector_length(v);
	    else len += vector_length(v);
	  }
      }
    n = snprintf(buf, 1024, "vectors: %u (%" PRId64 " %" PRId64 " %" PRId64 ")\n", sc->vectors->loc, len, flen, ilen);
    port_write_string(sc->output_port)(sc, buf, n, sc->output_port);
  }
  {
    int64_t len = 0, flen = 0;
    gp = sc->input_ports;
    for (i = 0; i < (int32_t)(gp->loc); i++)
      {
	s7_pointer v;
	v = gp->list[i];
	if (port_data(v))
	  {
	    flen += port_data_size(v);
	    if (port_needs_free(v))
	      len += port_data_size(v);
	  }
      }
    n = snprintf(buf, 1024, "input ports: %d (%" PRId64 " %" PRId64 ")\n", sc->input_ports->loc, len, flen);
    port_write_string(sc->output_port)(sc, buf, n, sc->output_port);
  }
  {
    int32_t fs;
    uint32_t cc_stacks;
    port_t *p;
    for (fs = 0, p = sc->port_heap; p; p = (port_t *)(p->next), fs++);
    gp = sc->continuations;
    for (i = 0, cc_stacks = 0; i < (int32_t)gp->loc; i++)
      if (is_continuation(gp->list[i]))
	cc_stacks += continuation_stack_size(gp->list[i]);

    n = snprintf(buf, 1024, "output ports: %u, free port: %d\ncontinuations: %u (total stack: %u), c_objects: %u, gensyms: %u, setters: %u, optlists: %u\n",
	    sc->output_ports->loc, fs, 
	    gp->loc, cc_stacks,
	    sc->c_objects->loc, sc->gensyms->loc, sc->setters_loc, sc->optlists->loc);
    port_write_string(sc->output_port)(sc, buf, n, sc->output_port);
  }
  return(sc->F);
}

static s7_pointer g_s7_let_ref_fallback(s7_scheme *sc, s7_pointer args)
{
  s7_pointer sym;

  sym = cadr(args);
  if (!is_symbol(sym))
    return(simple_wrong_type_argument(sc, sc->let_ref_symbol, sym, T_SYMBOL));

  if (sym == sc->print_length_symbol)                                    /* print-length */
    return(s7_make_integer(sc, sc->print_length));

  if (sym == sc->stack_top_symbol)                                       /* stack-top = how many frames active (4 stack entries per frame) */
    return(s7_make_integer(sc, (sc->stack_end - sc->stack_start) / 4));
  if (sym == sc->stack_size_symbol)                                      /* stack-size (max so far) */
    return(s7_make_integer(sc, sc->stack_size));
  if (sym == sc->max_stack_size_symbol)                                  /* max-stack-size */
    return(s7_make_integer(sc, sc->max_stack_size));
  if (sym == sc->stacktrace_defaults_symbol)                             /* stacktrace-defaults (used to be *stacktrace*) */
    return(sc->stacktrace_defaults);

  if (sym == sc->symbol_table_symbol)                                    /* symbol-table (the raw vector) */
    return(sc->symbol_table);
  if (sym == sc->rootlet_size_symbol)                                    /* rootlet-size */
    return(s7_make_integer(sc, sc->rootlet_entries));
  if (sym == sc->safety_symbol)                                          /* safety */
    return(s7_make_integer(sc, sc->safety));
  if (sym == sc->autoloading_symbol)                                     /* autoloading? */
    return(s7_make_boolean(sc, sc->is_autoloading));
  if (sym == sc->undefined_identifier_warnings_symbol)                   /* undefined-identifier-warnings */
    return(s7_make_boolean(sc, sc->undefined_identifier_warnings));
  if (sym == sc->cpu_time_symbol)                                        /* cpu-time */
    return(s7_make_real(sc, (double)clock() / (double)CLOCKS_PER_SEC));
  if (sym == sc->catches_symbol)                                         /* catches */
    return(active_catches(sc));
  if (sym == sc->exits_symbol)                                           /* exits */
    return(active_exits(sc));
  if (sym == sc->stack_symbol)                                           /* stack */
    return(stack_entries(sc, sc->stack, s7_stack_top(sc)));

  if (sym == sc->heap_size_symbol)                                       /* heap-size */
    return(s7_make_integer(sc, sc->heap_size));
  if (sym == sc->free_heap_size_symbol)                                  /* free-heap-size (number of unused cells in the heap) */
    return(s7_make_integer(sc, sc->free_heap_top - sc->free_heap));
  if (sym == sc->gc_freed_symbol)                                        /* gc-freed = how many cells freed during last GC sweep */
    return(s7_make_integer(sc, sc->gc_freed));
  if (sym == sc->gc_protected_objects_symbol)                            /* gc-protected-objects */
    return(sc->protected_objects);
  if (sym == sc->gc_stats_symbol)                                        /* gc-stats */
    return(make_integer(sc, sc->gc_stats));

  if (sym == sc->default_rationalize_error_symbol)                       /* default-rationalize-error */
    return(make_real(sc, sc->default_rationalize_error));
  if (sym == sc->default_random_state_symbol)                            /* default-random-state */
    return(sc->default_rng);

  if (sym == sc->history_symbol)                                         /* history (eval history circular buffer) */
    return(sc->cur_code);
  if (sym == sc->history_size_symbol)                                    /* history-size (eval history circular buffer size) */
    return(s7_make_integer(sc, sc->history_size));
  if (sym == sc->profile_info_symbol)                                    /* profile-info -- profiling data hash-table */
    return(sc->profile_info);
  if (sym == sc->max_list_length_symbol)                                 /* max-list-length (as arg to make-list) */
    return(s7_make_integer(sc, sc->max_list_length));
  if (sym == sc->max_vector_length_symbol)                               /* max-vector-length (as arg to make-vector and make-hash-table) */
    return(s7_make_integer(sc, sc->max_vector_length));
  if (sym == sc->max_vector_dimensions_symbol)                           /* max-vector-dimensions (make-vector) */
    return(s7_make_integer(sc, sc->max_vector_dimensions));
  if (sym == sc->max_string_length_symbol)                               /* max-string-length (as arg to make-string and read-string) */
    return(s7_make_integer(sc, sc->max_string_length));
  if (sym == sc->default_hash_table_length_symbol)                       /* default size for make-hash-table */
    return(s7_make_integer(sc, sc->default_hash_table_length));
  if (sym == sc->morally_equal_float_epsilon_symbol)                     /* morally-equal-float-epsilon */
    return(s7_make_real(sc, sc->morally_equal_float_epsilon));
  if (sym == sc->hash_table_float_epsilon_symbol)                        /* hash-table-float-epsilon */
    return(s7_make_real(sc, sc->hash_table_float_epsilon));
  if (sym == sc->initial_string_port_length_symbol)                      /* initial-string-port-length */
    return(s7_make_integer(sc, sc->initial_string_port_length));

  if (sym == sc->input_ports_symbol)                                     /* input-ports */
    return(make_vector_wrapper(sc, sc->input_ports->loc, sc->input_ports->list));
  if (sym == sc->output_ports_symbol)                                    /* output-ports */
    return(make_vector_wrapper(sc, sc->output_ports->loc, sc->output_ports->list));
  if (sym == sc->strings_symbol)                                         /* strings */
    return(make_vector_wrapper(sc, sc->strings->loc, sc->strings->list));
  if (sym == sc->gensyms_symbol)                                         /* gensyms */
    return(make_vector_wrapper(sc, sc->gensyms->loc, sc->gensyms->list));
  if (sym == sc->vectors_symbol)                                         /* vectors */
    return(make_vector_wrapper(sc, sc->vectors->loc, sc->vectors->list));
  if (sym == sc->hash_tables_symbol)                                     /* hash-tables */
    return(make_vector_wrapper(sc, sc->hash_tables->loc, sc->hash_tables->list));
  if (sym == sc->continuations_symbol)                                   /* continuations */
    return(make_vector_wrapper(sc, sc->continuations->loc, sc->continuations->list));
  if (sym == sc->c_objects_symbol)                                       /* c-objects */
    return(make_vector_wrapper(sc, sc->c_objects->loc, sc->c_objects->list));

  if (sym == sc->file_names_symbol)                                      /* file-names (loaded files) */
    return(make_vector_wrapper(sc, sc->file_names_top, sc->file_names));
  if (sym == sc->c_types_symbol)                                         /* c-types */
    {
      s7_pointer res;
      int32_t i;
      sc->w = sc->nil;
      for (i = 0; i < sc->num_c_object_types; i++)                       /* c-object type (tag) is i */
	sc->w = cons(sc, sc->c_object_types[i]->scheme_name, sc->w);
      res = safe_reverse_in_place(sc, sc->w);                            /*  so car(types) has tag 0 */
      sc->w = sc->nil;
      return(res);
    }

  if (sym == sc->bignum_precision_symbol)                                /* bignum-precision */
    return(s7_make_integer(sc, sc->bignum_precision));
  if (sym == sc->float_format_precision_symbol)                          /* float-format-precision */
    return(s7_make_integer(sc, float_format_precision));
  if (sym == sc->memory_usage_symbol)                                    /* memory-usage */
    return(describe_memory_usage(sc));

  /* sc->unlet is a scheme vector of slots -- not very useful at the scheme level */
  return(sc->undefined);
}

static s7_pointer g_s7_let_set_fallback(s7_scheme *sc, s7_pointer args)
{
  s7_pointer sym, val;

  sym = cadr(args);
  if (!is_symbol(sym))
    return(simple_wrong_type_argument(sc, sc->let_set_symbol, sym, T_SYMBOL));

  val = caddr(args);

  if ((sym == sc->print_length_symbol) ||
      (sym == sc->max_vector_length_symbol) ||
      (sym == sc->max_vector_dimensions_symbol) ||
      (sym == sc->max_list_length_symbol) ||
      (sym == sc->history_size_symbol) ||
      (sym == sc->max_string_length_symbol) ||
      (sym == sc->default_hash_table_length_symbol) ||
      (sym == sc->float_format_precision_symbol) ||
      (sym == sc->bignum_precision_symbol) ||
      (sym == sc->initial_string_port_length_symbol))
    {
      s7_int iv;

      if (!s7_is_integer(val))
	return(simple_wrong_type_argument(sc, sym, val, T_INTEGER));

      iv = s7_integer(val);         /* might be bignum if gmp */
      
      if ((iv < 0) ||               /* only print-length and float-format-precision can be 0, none can be negative */
	  ((iv == 0) && 
	   ((sym != sc->print_length_symbol) &&
	    (sym != sc->float_format_precision_symbol))))
	return(simple_out_of_range(sc, sym, val, s7_make_string_wrapper(sc, "should be a positive integer")));
      
      if (sym == sc->print_length_symbol)               {sc->print_length = iv;               return(val);}
      if (sym == sc->default_hash_table_length_symbol)  {sc->default_hash_table_length = iv;  return(val);}
      if (sym == sc->max_vector_length_symbol)          {sc->max_vector_length = iv;          return(val);}
      if (sym == sc->max_vector_dimensions_symbol)      {sc->max_vector_dimensions = iv;      return(val);}
      if (sym == sc->max_list_length_symbol)            {sc->max_list_length = iv;            return(val);}
      if (sym == sc->max_string_length_symbol)          {sc->max_string_length = iv;          return(val);}
      if (sym == sc->float_format_precision_symbol)     {float_format_precision = iv;         return(val);}
      if (sym == sc->initial_string_port_length_symbol) {sc->initial_string_port_length = iv; return(val);}
      if (sym == sc->history_size_symbol)
	{
#if WITH_HISTORY
	  s7_pointer p1, p2;
	  if (iv > sc->true_history_size)
	    {
	      /* splice in the new cells, reattach the circles */
	      s7_pointer next1, next2;
	      next1 = cdr(sc->eval_history1);
	      next2 = cdr(sc->eval_history2);
	      set_cdr(sc->eval_history1, permanent_list(sc, iv - sc->true_history_size));
	      set_cdr(sc->eval_history2, permanent_list(sc, iv - sc->true_history_size));
	      for (p1 = sc->eval_history1, p2 = sc->eval_history2; is_pair(cdr(p1)); p1 = cdr(p1), p2 = cdr(p2));
	      set_cdr(p1, next1);
	      set_cdr(p2, next2);
	      sc->true_history_size = iv;
	    }
	  sc->history_size = iv;
	  /* clear out both bufffers to avoid GC confusion */
	  for (p1 = sc->eval_history1, p2 = sc->eval_history2; ; p2 = cdr(p2))
	    {
	      set_car(p1, sc->nil);
	      set_car(p2, sc->nil);
	      p1 = cdr(p1);
	      if (p1 == sc->eval_history1) break;
	    }
#else
	  sc->history_size = iv;
#endif
	  return(val);
	}
      if (sym == sc->bignum_precision_symbol)
	{
	  if (s7_is_integer(val))
	    {
	      sc->bignum_precision = s7_integer(val);
#if WITH_GMP
	      set_bignum_precision(sc, sc->bignum_precision);
#endif
	    }
	}
      return(val);
    }

  if (sym == sc->gc_stats_symbol)
    {
      if (s7_is_boolean(val)) {sc->gc_stats = ((val == sc->T) ? GC_STATS : 0); return(val);}
      if (s7_is_integer(val)) {sc->gc_stats = s7_integer(val); return(val);}
      return(simple_wrong_type_argument(sc, sym, val, T_BOOLEAN));
    }

  if (sym == sc->max_stack_size_symbol)
    {
      if (s7_is_integer(val))
	{
	  s7_int size;
	  size = s7_integer(val);
	  if (size >= INITIAL_STACK_SIZE)
	    {
	      sc->max_stack_size = (uint32_t)size;
	      return(val);
	    }
	  return(simple_out_of_range(sc, sym, val, s7_make_string_wrapper(sc, "should be greater than the initial stack size (512)")));
	}
      return(simple_wrong_type_argument(sc, sym, val, T_INTEGER));
    }

  if (sym == sc->safety_symbol)
    {
      if (s7_is_integer(val)) {sc->safety = s7_integer(val); return(val);}
      return(simple_wrong_type_argument(sc, sym, val, T_INTEGER));
    }

  if (sym == sc->heap_size_symbol)
    {
      if (s7_is_integer(val)) 
	{
	  s7_int size;
	  size = s7_integer(val);
	  if (size >= sc->heap_size * 2)
	    resize_heap_to(sc, size);
	  return(val);
	}
      return(simple_wrong_type_argument(sc, sym, val, T_INTEGER));
    }

  if (sym == sc->autoloading_symbol)
    {
      if (s7_is_boolean(val)) {sc->is_autoloading = s7_boolean(sc, val); return(val);}
      return(simple_wrong_type_argument(sc, sym, val, T_BOOLEAN));
    }

  if (sym == sc->undefined_identifier_warnings_symbol)
    {
      if (s7_is_boolean(val)) {sc->undefined_identifier_warnings = s7_boolean(sc, val); return(val);}
      return(simple_wrong_type_argument(sc, sym, val, T_BOOLEAN));
    }

  if (sym == sc->morally_equal_float_epsilon_symbol)
    {
      if (s7_is_real(val)) {sc->morally_equal_float_epsilon = s7_real(val); return(val);}
      return(simple_wrong_type_argument(sc, sym, val, T_REAL));
    }

  if (sym == sc->hash_table_float_epsilon_symbol)
    {
      if (s7_is_real(val)) {sc->hash_table_float_epsilon = s7_real(val); return(val);}
      return(simple_wrong_type_argument(sc, sym, val, T_REAL));
    }

  if (sym == sc->default_rationalize_error_symbol)
    {
      if (s7_is_real(val)) {sc->default_rationalize_error = real_to_double(sc, val, "set! default-rationalize-error"); return(val);}
      return(simple_wrong_type_argument(sc, sym, val, T_REAL));
    }

  if (sym == sc->default_random_state_symbol)
    {
      if (is_random_state(val))
	{
#if (!WITH_GMP)
	  random_seed(sc->default_rng) = random_seed(val);
	  random_carry(sc->default_rng) = random_carry(val);
#endif
	  return(val);
	}
      return(wrong_type_argument_with_type(sc, sym, 1, val, a_random_state_object_string));
    }

  if (sym == sc->stacktrace_defaults_symbol)
    {
      if (!is_pair(val))                
	return(simple_wrong_type_argument(sc, sym, val, T_PAIR));
      if (s7_list_length(sc, val) != 5) 
	return(simple_wrong_type_argument_with_type(sc, sym, val, s7_make_string_wrapper(sc, "a list with 5 entries")));
      if (!is_integer(car(val)))        
	return(wrong_type_argument_with_type(sc, sym, 1, car(val), s7_make_string_wrapper(sc, "an integer (stack frames)")));
      if (!is_integer(cadr(val)))       
	return(wrong_type_argument_with_type(sc, sym, 2, cadr(val), s7_make_string_wrapper(sc, "an integer (cols-for-data)")));
      if (!is_integer(caddr(val)))      
	return(wrong_type_argument_with_type(sc, sym, 3, caddr(val), s7_make_string_wrapper(sc, "an integer (line length)")));
      if (!is_integer(cadddr(val)))     
	return(wrong_type_argument_with_type(sc, sym, 4, cadddr(val), s7_make_string_wrapper(sc, "an integer (comment position)")));
      if (!s7_is_boolean(s7_list_ref(sc, val, 4))) 
	return(wrong_type_argument_with_type(sc, sym, 5, s7_list_ref(sc, val, 4), s7_make_string_wrapper(sc, "a boolean (treat-data-as-comment)")));
      sc->stacktrace_defaults = copy_list(sc, val);
      return(val);
    }

  if ((sym == sc->cpu_time_symbol) || (sym == sc->profile_info_symbol) ||
      (sym == sc->free_heap_size_symbol) || (sym == sc->gc_freed_symbol) || (sym == sc->gc_protected_objects_symbol) ||
      (sym == sc->file_names_symbol) || (sym == sc->c_types_symbol) || (sym == sc->catches_symbol) || (sym == sc->exits_symbol) || 
      (sym == sc->rootlet_size_symbol) || (sym == sc->stack_top_symbol) || (sym == sc->stack_size_symbol))
    return(s7_error(sc, sc->error_symbol, set_elist_2(sc, s7_make_string_wrapper(sc, "can't set (*s7* '~S)"), sym)));
      
  return(sc->undefined);
}


/* some procedure-signature support functions */

static s7_pointer g_is_float(s7_scheme *sc, s7_pointer args)
{
  #define H_is_float "(float? x) returns #t is x is real and not rational."
  #define Q_is_float pl_bt
  s7_pointer p;
  p = car(args);
  return(make_boolean(sc, is_float(p)));
}

static bool is_float_b(s7_pointer p) {return(is_float(p));}


static s7_pointer g_is_proper_list(s7_scheme *sc, s7_pointer args)
{
  #define H_is_proper_list "(proper-list? x) returns #t is x is a list that is neither circular nor dotted."
  #define Q_is_proper_list pl_bt
  s7_pointer p;
  p = car(args);
  return(make_boolean(sc, s7_is_proper_list(sc, p)));
}

static bool is_proper_list_b(s7_pointer p) {return(s7_is_proper_list(cur_sc, p));}


#if (!MS_WINDOWS)
/* gdb stacktrace decoding */

static bool is_decodable(s7_scheme *sc, s7_pointer p)
{
  int32_t i;
  s7_pointer x;
  s7_pointer *tp, *heap_top;

  if ((void *)p == (void *)sc) return(false);

  /* check basic constants */
  if ((p == sc->nil) || (p == sc->T) || (p == sc->F) || (p == sc->eof_object) || (p == sc->else_symbol) || (p == sc->rootlet) ||
      (p == sc->undefined) || (p == sc->unspecified) || (p == sc->no_value) || (p == sc->gc_nil) ||
      (p == sc->t1_1) || (p == sc->t2_1) || (p == sc->t3_1) || (p == sc->a1_1) || (p == sc->a2_1) || (p == sc->a3_1) || (p == sc->a4_1))
    return(true);

  /* check symbol-table */
  for (i = 0; i < SYMBOL_TABLE_SIZE; i++)
    for (x = vector_element(sc->symbol_table, i); is_not_null(x); x = cdr(x))
      {
	s7_pointer sym;
	sym = car(x);
	if ((sym == p) ||
	    ((is_global(sym)) && (is_slot(global_slot(sym))) && (p == slot_value(global_slot(sym)))))
	  return(true);
      }

  for (i = 0; i < NUM_CHARS; i++) if (p == chars[i]) return(true);
  for (i = 0; i <= NUM_SMALL_INTS; i++) if (p == small_ints[i]) return(true);
  /* also real_one and friends, sc->safe_lists, tmp_strs? p|elist? */

  /* check the heap */
  tp = sc->heap;
  heap_top = (s7_pointer *)(sc->heap + sc->heap_size);
  while (tp < heap_top)
    if (p == (*tp++))
      return(true);

  return(false);
}

char *s7_decode_bt(void)
{
  FILE *fp;
  fp = fopen("gdb.txt", "r");
  if (fp)
    {
      int64_t i, size;
      size_t bytes;
      bool in_quotes = false;
      unsigned char *bt;
      s7_scheme *sc;
      sc = cur_sc;

      fseek(fp, 0, SEEK_END);
      size = ftell(fp);
      rewind(fp);

      bt = (unsigned char *)malloc((size + 1) * sizeof(unsigned char));
      bytes = fread(bt, sizeof(unsigned char), size, fp);
      if (bytes != (size_t)size)
	{
	  fclose(fp);
	  free(bt);
	  return((char *)" oops ");
	}
      bt[size] = '\0';
      fclose(fp);

      for (i = 0; i < size; i++)
	{
	  fputc(bt[i], stdout);
	  if ((bt[i] == '"') && ((i == 0) || (bt[i - 1] != '\\')))
	    in_quotes = (!in_quotes);
	  else
	    {
	      if ((!in_quotes) && (i < size - 8))
		{
		  if ((bt[i] == '=') &&
		      (((bt[i + 1] == '0') && (bt[i + 2] == 'x')) ||
		       ((bt[i + 1] == ' ') && (bt[i + 2] == '0') && (bt[i + 3] == 'x'))))
		    {
		      void *vp;
		      int32_t vals;
		      vals = sscanf((const char *)(bt + i + 1), "%p", &vp);
		      if (vals == 1)
			{
			  int32_t k;
			  for (k = i + ((bt[i + 2] == 'x') ? 3 : 4); (k < size) && (IS_DIGIT(bt[k], 16)); k++);
			  if ((bt[k] != ' ') || (bt[k + 1] != '"'))
			    {
			      s7_pointer p;
			      p = (s7_pointer)vp;
			      if ((is_decodable(sc, p)) &&
				  (!is_free(p)))
				{
				  if (bt[i + 1] == ' ') fputc(' ', stdout);
				  i = k - 1;
				  if (s7_is_valid(sc, p))
				    {
				      char *str;
				      str = s7_object_to_c_string(sc, p);
				      fprintf(stdout, "%s%s%s", BOLD_TEXT, str, UNBOLD_TEXT);
				      free(str);
				    }
				  else 
				    {
				      if (is_free(p))
					fprintf(stderr, "%p: %sfree cell%s", p, BOLD_TEXT, UNBOLD_TEXT);
				      else fprintf(stderr, "%p: %sunprintable?%s", p, BOLD_TEXT, UNBOLD_TEXT);
				    }
				}
			    }
			}
		    }
		}
	    }
	}
      free(bt);
    }
  return((char *)"");
}
#endif



/* -------------------------------- initialization -------------------------------- */

static s7_pointer make_unique_object(const char* name, uint64_t typ)
{
  s7_pointer p;
  p = alloc_pointer();
  set_type(p, typ | T_IMMUTABLE);
  unique_name_length(p) = safe_strlen(name);
  unique_name(p) = copy_string_with_length(name, unique_name_length(p));
  unheap(p);
  return(p);
}

s7_scheme *s7_init(void)
{
  int32_t i;
  s7_scheme *sc;
  s7_pointer sym;
  static bool already_inited = false;

#if (!MS_WINDOWS)
  setlocale(LC_NUMERIC, "C"); /* use decimal point in floats */
#endif
  
  if (!already_inited)
    {
      init_types();
      init_ctables();
      init_mark_functions();
      init_display_functions();
      init_length_functions();
      init_equals();
      init_hash_maps();
      init_pows();
      init_uppers();
      all_x_function_init();
      init_catchers();
      /* sizeof(__float128) == sizeof(long double) so how to distinguish them for printf (L vs Q)? */
      /* if (sizeof(s7_double) >= 16) float_format_g = "%.*Qg"; */      /* __float128 */
      if (sizeof(s7_double) > 8)
	float_format_g = "%.*Lg";   /* long double (80-bit precision?) */
      else float_format_g = "%.*g"; /* float and double */
    }

  sc = (s7_scheme *)calloc(1, sizeof(s7_scheme)); /* malloc is not recommended here */
  cur_sc = sc;                                    /* for gdb/debugging and clm optimizer */
  sc->gc_off = true;                              /* sc->args and so on are not set yet, so a gc during init -> segfault */
  sc->gc_stats = 0;
  init_gc_caches(sc);
  init_string_free_lists(sc);

  sc->longjmp_ok = false;
  sc->setjmp_loc = NO_SET_JUMP;

  if (sizeof(s7_int) == 4)
    sc->max_vector_length = (1 << 24);
  else sc->max_vector_length = (1LL << 32);
  sc->max_string_length = 1073741824; /* 1 << 30 */
  sc->max_list_length = 1073741824;
  sc->max_vector_dimensions = 512;

  sc->strbuf_size = INITIAL_STRBUF_SIZE;
  sc->strbuf = (char *)calloc(sc->strbuf_size, sizeof(char));
  sc->tmpbuf = (char *)calloc(TMPBUF_SIZE, sizeof(char));
  sc->print_width = sc->max_string_length;
  sc->short_print = false;

  sc->initial_string_port_length = 128;
  sc->format_depth = -1;
  sc->slash_str_size = 0;
  sc->slash_str = NULL;

  sc->singletons = (s7_pointer *)calloc(256, sizeof(s7_pointer));
  sc->read_line_buf = NULL;
  sc->read_line_buf_size = 0;

  sc->nil =         make_unique_object("()",             T_NIL);
  sc->gc_nil =      make_unique_object("#<nil>",         T_NIL);        /* ?? perhaps a unique type for this? */
  sc->T =           make_unique_object("#t",             T_BOOLEAN);
  sc->F =           make_unique_object("#f",             T_BOOLEAN);
  sc->eof_object =  make_unique_object("#<eof>",         T_EOF_OBJECT);
  sc->undefined =   make_unique_object("#<undefined>",   T_UNDEFINED);
  sc->unspecified = make_unique_object("#<unspecified>", T_UNSPECIFIED);
  sc->no_value =    make_unique_object("#<unspecified>", T_UNSPECIFIED);

  unique_car(sc->nil) = sc->unspecified;
  unique_cdr(sc->nil) = sc->unspecified;
  /* this is mixing two different s7_cell structs, cons and envr, but luckily
   *    envr has two initial s7_pointer fields, equivalent to car and cdr, so
   *    let_id which is the same as opt1 is unaffected.  To get the names
   *    built-in, I'll append unique_name and unique_name_length fields to
   *    the envr struct.
   */
  let_id(sc->nil) = -1;
  unique_cdr(sc->unspecified) = sc->unspecified;
  unique_cdr(sc->undefined) = sc->undefined;
  /* this way find_symbol of an undefined symbol returns #<undefined> not #<unspecified> */

  sc->temp_cell_1 = permanent_cons(sc->nil, sc->nil, T_PAIR | T_IMMUTABLE);
  sc->temp_cell = permanent_cons(sc->temp_cell_1, sc->nil, T_PAIR | T_IMMUTABLE);
  sc->temp_cell_2 = permanent_cons(sc->nil, sc->nil, T_PAIR | T_IMMUTABLE);

  sc->t1_1 = permanent_cons(sc->nil, sc->nil, T_PAIR | T_IMMUTABLE);

  sc->t2_2 = permanent_cons(sc->nil, sc->nil, T_PAIR | T_IMMUTABLE);
  sc->t2_1 = permanent_cons(sc->nil, sc->t2_2, T_PAIR | T_IMMUTABLE);
  sc->z2_2 = permanent_cons(sc->nil, sc->nil, T_PAIR | T_IMMUTABLE);
  sc->z2_1 = permanent_cons(sc->nil, sc->z2_2, T_PAIR | T_IMMUTABLE);

  sc->t3_3 = permanent_cons(sc->nil, sc->nil, T_PAIR | T_IMMUTABLE);
  sc->t3_2 = permanent_cons(sc->nil, sc->t3_3, T_PAIR | T_IMMUTABLE);
  sc->t3_1 = permanent_cons(sc->nil, sc->t3_2, T_PAIR | T_IMMUTABLE);

  sc->a4_4 = permanent_cons(sc->nil, sc->nil, T_PAIR | T_IMMUTABLE);
  sc->a4_3 = permanent_cons(sc->nil, sc->a4_4, T_PAIR | T_IMMUTABLE);
  sc->a4_2 = permanent_cons(sc->nil, sc->a4_3, T_PAIR | T_IMMUTABLE);
  sc->a4_1 = permanent_cons(sc->nil, sc->a4_2, T_PAIR | T_IMMUTABLE);

  sc->a1_1 = sc->a4_4;
  sc->a2_1 = sc->a4_3;
  sc->a2_2 = sc->a4_4;
  sc->a3_1 = sc->a4_2;
  sc->a3_2 = sc->a4_3;
  sc->a3_3 = sc->a4_4;

  /* sc->safe_lists = (s7_pointer *)calloc(NUM_SAFE_LISTS, sizeof(s7_pointer)); */
  for (i = 1; i < NUM_SAFE_LISTS; i++)
    sc->safe_lists[i] = sc->nil; /* permanent_list(sc, i); */
  sc->current_safe_list = 0;

  sc->input_port_stack = sc->nil;
  sc->code = sc->nil;
#if WITH_HISTORY
  sc->eval_history1 = permanent_list(sc, DEFAULT_HISTORY_SIZE);
  sc->eval_history2 = permanent_list(sc, DEFAULT_HISTORY_SIZE);
  {
    s7_pointer p1, p2;
    for (p1 = sc->eval_history1, p2 = sc->eval_history2; is_pair(cdr(p1)); p1 = cdr(p1), p2 = cdr(p2));
    set_cdr(p1, sc->eval_history1);
    set_cdr(p2, sc->eval_history2);
    sc->cur_code = sc->eval_history1;
    sc->using_history1 = true;
  }
#else
  sc->cur_code = sc->F;
#endif
  sc->args = sc->nil;
  sc->value = sc->nil;
  sc->v = sc->nil;
  sc->w = sc->nil;
  sc->x = sc->nil;
  sc->y = sc->nil;
  sc->z = sc->nil;

  sc->temp1 = sc->nil;
  sc->temp2 = sc->nil;
  sc->temp3 = sc->nil;
  sc->temp4 = sc->nil;
  sc->temp5 = sc->nil;
  sc->temp6 = sc->nil;
  sc->temp7 = sc->nil;
  sc->temp8 = sc->nil;
  sc->temp9 = sc->nil;
  sc->temp10 = sc->nil;
  sc->temp11 = sc->nil;
  /* sc->t_temps = (s7_pointer *)malloc(T_TEMPS_SIZE * sizeof(s7_pointer)); */
  for (i = 0; i < T_TEMPS_SIZE; i++) sc->t_temps[i] = sc->nil;
  sc->t_temp_ctr = 0;

  sc->begin_hook = NULL;
  sc->autoload_table = sc->nil;
  sc->autoload_names = NULL;
  sc->autoload_names_sizes = NULL;
  sc->autoloaded_already = NULL;
  sc->autoload_names_loc = 0;
  sc->is_autoloading = true;

  sc->port_heap = NULL;
  sc->permanent_objects = NULL;

  sc->heap_size = INITIAL_HEAP_SIZE;
  if ((sc->heap_size % 32) != 0)
    sc->heap_size = 32 * (int64_t)ceil((double)(sc->heap_size) / 32.0);
  sc->heap = (s7_pointer *)malloc(sc->heap_size * sizeof(s7_pointer));

  sc->free_heap = (s7_cell **)malloc(sc->heap_size * sizeof(s7_cell *));
  sc->free_heap_top = (s7_cell **)(sc->free_heap + INITIAL_HEAP_SIZE);
  sc->free_heap_trigger = (s7_cell **)(sc->free_heap + GC_TRIGGER_SIZE);
  sc->previous_free_heap_top = sc->free_heap_top;

  {
    s7_cell *cells;
    cells = (s7_cell *)calloc(INITIAL_HEAP_SIZE, sizeof(s7_cell));
    for (i = 0; i < INITIAL_HEAP_SIZE; i++)
      {
	sc->heap[i] = &cells[i];
 	sc->free_heap[i] = sc->heap[i];
 	heap_location(sc->heap[i]) = i;
	i++;
	sc->heap[i] = &cells[i];
 	sc->free_heap[i] = sc->heap[i];
 	heap_location(sc->heap[i]) = i;
     }
  }

  /* this has to precede s7_make_* allocations */
  sc->protected_objects_size = INITIAL_PROTECTED_OBJECTS_SIZE;
  sc->gpofl = (uint32_t *)malloc(INITIAL_PROTECTED_OBJECTS_SIZE * sizeof(uint32_t));
  sc->gpofl_loc = INITIAL_PROTECTED_OBJECTS_SIZE - 1;
  sc->protected_objects = s7_make_vector(sc, INITIAL_PROTECTED_OBJECTS_SIZE);

  sc->protected_accessors_size = INITIAL_PROTECTED_OBJECTS_SIZE;
  sc->protected_accessors_loc = 0;
  sc->protected_accessors = s7_make_vector(sc, INITIAL_PROTECTED_OBJECTS_SIZE);

  for (i = 0; i < INITIAL_PROTECTED_OBJECTS_SIZE; i++)
    {
      vector_element(sc->protected_objects, i) = sc->gc_nil;
      vector_element(sc->protected_accessors, i) = sc->gc_nil;
      sc->gpofl[i] = i;
    }

  sc->stack = s7_make_vector(sc, INITIAL_STACK_SIZE);
  sc->stack_start = vector_elements(sc->stack);
  sc->stack_end = sc->stack_start;
  sc->stack_size = INITIAL_STACK_SIZE;
  sc->stack_resize_trigger = (s7_pointer *)(sc->stack_start + sc->stack_size / 2);
  set_type(sc->stack, T_STACK);
  sc->max_stack_size = (1 << 30);

  initialize_op_stack(sc);

  /* keep the symbol table out of the heap */
  sc->symbol_table = (s7_pointer)calloc(1, sizeof(s7_cell));
  set_type(sc->symbol_table, T_VECTOR);
  vector_length(sc->symbol_table) = SYMBOL_TABLE_SIZE;
  vector_elements(sc->symbol_table) = (s7_pointer *)malloc(SYMBOL_TABLE_SIZE * sizeof(s7_pointer));
  vector_getter(sc->symbol_table) = default_vector_getter;
  vector_setter(sc->symbol_table) = default_vector_setter;
  s7_vector_fill(sc, sc->symbol_table, sc->nil);
  unheap(sc->symbol_table);

  sc->tmp_strs = (s7_pointer *)malloc(2 * sizeof(s7_pointer));
  for (i = 0; i < 2; i++)
    {
      s7_pointer p;
      p = alloc_pointer();
      sc->tmp_strs[i] = p;
      unheap(p);
      set_type(p, T_STRING | T_SAFE_PROCEDURE);
      string_hash(p) = 0;
      string_needs_free(p) = false;
      string_length(p) = 0;
      string_value(p) = (char *)malloc(INITIAL_TMP_STR_SIZE * sizeof(char));
      string_temp_true_length(p) = INITIAL_TMP_STR_SIZE;
    }

  {
    opt_info *os;
    os = (opt_info *)calloc(OPTS_SIZE, sizeof(opt_info));
    for (i = 0; i < OPTS_SIZE; i++)
      sc->opts[i] = &os[i];
  }
  sc->funcalls = 0;
  sc->unwraps = 0;
      
#if WITH_MULTITHREAD_CHECKS
  sc->lock_count = 0;
  {
    pthread_mutexattr_t attr;
    pthread_mutexattr_init(&attr);
    pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE);
    pthread_mutex_init(&sc->lock, &attr);
  }
#endif
  sc->c_object_types = NULL;
  sc->c_object_types_size = 0;
  sc->num_c_object_types = 0;

  sc->typnam = NULL;
  sc->typnam_len = 0;
  sc->help_arglist = NULL;
  sc->default_rationalize_error = 1.0e-12;
  sc->hash_table_float_epsilon = 1.0e-12;
  sc->morally_equal_float_epsilon = 1.0e-15;
  sc->default_hash_table_length = 8;
  sc->gensym_counter = 0;
  sc->capture_let_counter = 0;
  sc->f_class = 0;
  sc->add_class = 0;
  sc->equal_class = 0;
  sc->let_number = 0;
  sc->format_column = 0;
  sc->file_names = NULL;
  sc->file_names_size = 0;
  sc->file_names_top = -1;
  sc->no_values = 0;
  sc->s7_call_line = 0;
  sc->s7_call_file = NULL;
  sc->s7_call_name = NULL;
  sc->safety = NO_SAFETY;
  sc->print_length = 8;
  sc->history_size = DEFAULT_HISTORY_SIZE;
  sc->true_history_size = DEFAULT_HISTORY_SIZE;
  sc->profile_info = sc->nil;
  sc->baffle_ctr = 0;
  sc->syms_tag = 0;
  sc->class_name_symbol = make_symbol(sc, "class-name");
  sc->circle_info = init_circle_info();
  sc->fdats = (format_data **)calloc(8, sizeof(format_data *));
  sc->num_fdats = 8;
  sc->plist_1 = permanent_list(sc, 1);
  sc->plist_2 = permanent_list(sc, 2);
  sc->plist_3 = permanent_list(sc, 3);
  sc->qlist_2 = permanent_list(sc, 2);
  sc->elist_1 = permanent_list(sc, 1);
  sc->elist_2 = permanent_list(sc, 2);
  sc->elist_3 = permanent_list(sc, 3);
  sc->elist_4 = permanent_list(sc, 4);
  sc->elist_5 = permanent_list(sc, 5);
  sc->direct_str = s7_make_permanent_string(NULL);
  sc->undefined_identifier_warnings = false;
  sc->wrap_only = make_wrap_only(sc);
  sc->dox_slot_symbol = s7_make_symbol(sc, "(dox_slot)");

  sc->rootlet = s7_make_vector(sc, ROOTLET_SIZE);
  set_type(sc->rootlet, T_LET | T_SAFE_PROCEDURE);
  sc->rootlet_entries = 0;
  for (i = 0; i < ROOTLET_SIZE; i++)
    vector_element(sc->rootlet, i) = sc->nil;
  sc->envir = sc->nil;
  sc->shadow_rootlet = sc->nil;

  if (!already_inited)
    {
      /* keep the small_ints out of the heap */
      small_ints = (s7_pointer *)malloc((NUM_SMALL_INTS + 1) * sizeof(s7_pointer));
      {
	s7_cell *cells;
	cells = (s7_cell *)calloc((NUM_SMALL_INTS + 1), sizeof(s7_cell));
	for (i = 0; i <= NUM_SMALL_INTS; i++)
	  {
	    s7_pointer p;
	    small_ints[i] = &cells[i];
	    p = small_ints[i];
	    typeflag(p) = T_IMMUTABLE | T_INTEGER;
	    unheap(p);
	    integer(p) = i;
	  }
      }

      real_zero = make_permanent_real(0.0);
      real_one = make_permanent_real(1.0);
      real_NaN = make_permanent_real(NAN);
      real_infinity = make_permanent_real(INFINITY);
      real_minus_infinity = make_permanent_real(-INFINITY);
      real_pi = make_permanent_real(3.1415926535897932384626433832795029L); /* M_PI is not good enough for s7_double = long double */
      arity_not_set = make_permanent_integer_unchecked(CLOSURE_ARITY_NOT_SET);
      max_arity = make_permanent_integer_unchecked(MAX_ARITY);
      minus_one = make_permanent_integer_unchecked(-1);
      minus_two = make_permanent_integer_unchecked(-2);
      /* prebuilt null string is tricky mainly because it overlaps #u8() */

      /* keep the characters out of the heap */
      chars = (s7_pointer *)malloc((NUM_CHARS + 1) * sizeof(s7_pointer));
      chars[0] = sc->eof_object;
      chars++;                    /* now chars[EOF] == chars[-1] == sc->eof_object */
      {
	s7_cell *cells;
	cells = (s7_cell *)calloc(NUM_CHARS, sizeof(s7_cell));
	for (i = 0; i < NUM_CHARS; i++)
	  {
	    s7_pointer cp;
	    unsigned char c;

	    c = (unsigned char)i;
	    cp = &cells[i];
	    typeflag(cp) = T_IMMUTABLE | T_CHARACTER;
	    unheap(cp);
	    character(cp) = c;
	    upper_character(cp) = (unsigned char)toupper(i);
	    is_char_alphabetic(cp) = (bool)isalpha(i);
	    is_char_numeric(cp) = (bool)isdigit(i);
	    is_char_whitespace(cp) = white_space[i];
	    is_char_uppercase(cp) = (((bool)isupper(i)) || ((i >= 192) && (i < 208)));
	    is_char_lowercase(cp) = (bool)islower(i);
	    chars[i] = cp;

            #define make_character_name(C, S) strncat((char *)(&(character_name(C))), S, character_name_length(C) = strlen(S))
	    switch (c)
	      {
	      case ' ':	       make_character_name(cp, "#\\space");     break;
	      case '\n':       make_character_name(cp, "#\\newline");   break;
	      case '\r':       make_character_name(cp, "#\\return");    break;
	      case '\t':       make_character_name(cp, "#\\tab");       break;
	      case '\0':       make_character_name(cp, "#\\null");      break;
	      case (char)0x1b: make_character_name(cp, "#\\escape");    break;
	      case (char)0x7f: make_character_name(cp, "#\\delete");    break;
	      case (char)7:    make_character_name(cp, "#\\alarm");     break;
	      case (char)8:    make_character_name(cp, "#\\backspace"); break;
	      default:
		{
                  #define P_SIZE 12
		  int32_t len;
		  if ((c < 32) || (c >= 127))
		    len = snprintf((char *)(&(character_name(cp))), P_SIZE, "#\\x%x", c);
		  else len = snprintf((char *)(&(character_name(cp))), P_SIZE, "#\\%c", c);
		  character_name_length(cp) = len;
		  break;
		}
	      }
	  }
      }
    }

  make_standard_ports(sc);

  sc->syn_docs = (s7_pointer *)calloc(OP_MAX_DEFINED, sizeof(s7_pointer));
  #define quote_help             "(quote obj) returns obj unevaluated.  'obj is an abbreviation for (quote obj)."
  #define if_help                "(if expr true-stuff optional-false-stuff) evaluates expr, then if it is true, evaluates true-stuff; otherwise, \
                                      if optional-false-stuff exists, it is evaluated."
  #define when_help              "(when expr ...) evaluates expr, and if it is true, evaluates each form in its body, returning the value of the last"
  #define unless_help            "(unless expr ...) evaluates expr, and if it is false, evaluates each form in its body, returning the value of the last"
  #define begin_help             "(begin ...) evaluates each form in its body, returning the value of the last one"
  #define set_help               "(set! variable value) sets the value of variable to value."
  #define let_help               "(let ((var val)...) ...) binds each variable to its initial value, then evaluates its body,\
                                      returning the value of the last form.  The let variables are local to it, and \
                                      are not available for use until all have been initialized."
  #define let_star_help          "(let* ((var val)...) ...) binds each variable to its initial value, then evaluates its body, \
                                      returning the value of the last form.  The let* variables are local to it, and are available immediately."
  #define letrec_help            "(letrec ((var (lambda ...)))...) is like let, but var can refer to itself in its value \
                                      (i.e. you can define local recursive functions)"
  #define letrec_star_help       "(letrec* ((var val))...) is like letrec, but successive bindings are handled as in let*"
  #define cond_help              "(cond (expr clause...)...) is like if..then.  Each expr is evaluated in order, and if one is not #f, \
                                      the associated clauses are evaluated, whereupon cond returns."
  #define and_help               "(and expr expr ...) evaluates each of its arguments in order, quitting (and returning #f) \
                                      as soon as one of them returns #f.  If all are non-#f, it returns the last value."
  #define or_help                "(or expr expr ...) evaluates each of its argments in order, quitting as soon as one of them is not #f.  \
                                      If all are #f, or returns #f."
  #define case_help              "(case val ((key...) clause...)...) looks for val in the various lists of keys, and if a \
                                      match is found (via eqv?), the associated clauses are evaluated, and case returns."
  #define do_help                "(do (vars...) (loop control and return value) ...) is a do-loop."
  #define lambda_help            "(lambda args ...) returns a function."
  #define lambda_star_help       "(lambda* args ...) returns a function; the args list can have default values, \
                                      the parameters themselves can be accessed via keywords."
  #define define_help            "(define var val) assigns val to the variable (symbol) var.  (define (func args) ...) is \
                                      shorthand for (define func (lambda args ...))"
  #define define_star_help       "(define* (func args) ...) defines a function with optional/keyword arguments."
  #define define_constant_help   "(define-constant var val) defines var to be a constant (it can't be set or bound), with the value val."
  #define define_macro_help      "(define-macro (mac args) ...) defines mac to be a macro."
  #define define_macro_star_help "(define-macro* (mac args) ...) defines mac to be a macro with optional/keyword arguments."
  #define define_expansion_help  "(define-expansion (mac args) ...) defines mac to be a read-time macro."
  #define define_bacro_help      "(define-bacro (mac args) ...) defines mac to be a bacro."
  #define define_bacro_star_help "(define-bacro* (mac args) ...) defines mac to be a bacro with optional/keyword arguments."
  #define with_baffle_help       "(with-baffle ...) evaluates its body in a context that is safe from outside interference."
  #define macroexpand_help       "(macroexpand macro-call) returns the result of the expansion phase of evaluating the macro call."
  #define with_let_help          "(with-let env ...) evaluates its body in the environment env."
  #define let_temporarily_help   "(let-temporarily ((var value)...) . body) sets each var to its new value, evals body, then returns each var to its original value."

  sc->quote_symbol =             assign_syntax(sc, "quote",           OP_QUOTE,             small_int(1), small_int(1), quote_help);
  sc->if_symbol =                assign_syntax(sc, "if",              OP_IF,                small_int(2), small_int(3), if_help);
  sc->when_symbol =              assign_syntax(sc, "when",            OP_WHEN,              small_int(2), max_arity,    when_help);
  sc->unless_symbol =            assign_syntax(sc, "unless",          OP_UNLESS,            small_int(2), max_arity,    unless_help);
  sc->begin_symbol =             assign_syntax(sc, "begin",           OP_BEGIN,             small_int(0), max_arity,    begin_help);
  sc->set_symbol =               assign_syntax(sc, "set!",            OP_SET,               small_int(2), small_int(2), set_help);
  sc->let_symbol =               assign_syntax(sc, "let",             OP_LET,               small_int(2), max_arity,    let_help);
  sc->let_star_symbol =          assign_syntax(sc, "let*",            OP_LET_STAR,          small_int(2), max_arity,    let_star_help);
  sc->letrec_symbol =            assign_syntax(sc, "letrec",          OP_LETREC,            small_int(2), max_arity,    letrec_help);
  sc->letrec_star_symbol =       assign_syntax(sc, "letrec*",         OP_LETREC_STAR,       small_int(2), max_arity,    letrec_star_help);
  sc->cond_symbol =              assign_syntax(sc, "cond",            OP_COND,              small_int(1), max_arity,    cond_help);
  sc->and_symbol =               assign_syntax(sc, "and",             OP_AND,               small_int(0), max_arity,    and_help);
  sc->or_symbol =                assign_syntax(sc, "or",              OP_OR,                small_int(0), max_arity,    or_help);
  sc->case_symbol =              assign_syntax(sc, "case",            OP_CASE,              small_int(2), max_arity,    case_help);
  sc->do_symbol =                assign_syntax(sc, "do",              OP_DO,                small_int(2), max_arity,    do_help); /* 2 because body can be null */
  sc->lambda_symbol =            assign_syntax(sc, "lambda",          OP_LAMBDA,            small_int(2), max_arity,    lambda_help);
  sc->lambda_star_symbol =       assign_syntax(sc, "lambda*",         OP_LAMBDA_STAR,       small_int(2), max_arity,    lambda_star_help);
  sc->define_symbol =            assign_syntax(sc, "define",          OP_DEFINE,            small_int(2), max_arity,    define_help);
  sc->define_star_symbol =       assign_syntax(sc, "define*",         OP_DEFINE_STAR,       small_int(2), max_arity,    define_star_help);
  sc->define_constant_symbol =   assign_syntax(sc, "define-constant", OP_DEFINE_CONSTANT,   small_int(2), max_arity,    define_constant_help);
  sc->define_macro_symbol =      assign_syntax(sc, "define-macro",    OP_DEFINE_MACRO,      small_int(2), max_arity,    define_macro_help);
  sc->define_macro_star_symbol = assign_syntax(sc, "define-macro*",   OP_DEFINE_MACRO_STAR, small_int(2), max_arity,    define_macro_star_help);
  sc->define_expansion_symbol =  assign_syntax(sc, "define-expansion",OP_DEFINE_EXPANSION,  small_int(2), max_arity,    define_expansion_help);
  sc->define_bacro_symbol =      assign_syntax(sc, "define-bacro",    OP_DEFINE_BACRO,      small_int(2), max_arity,    define_bacro_help);
  sc->define_bacro_star_symbol = assign_syntax(sc, "define-bacro*",   OP_DEFINE_BACRO_STAR, small_int(2), max_arity,    define_bacro_star_help);
  sc->with_baffle_symbol =       assign_syntax(sc, "with-baffle",     OP_WITH_BAFFLE,       small_int(1), max_arity,    with_baffle_help);
  sc->macroexpand_symbol =       assign_syntax(sc, "macroexpand",     OP_MACROEXPAND,       small_int(1), small_int(1), macroexpand_help);
  sc->with_let_symbol =          assign_syntax(sc, "with-let",        OP_WITH_LET,          small_int(1), max_arity,    with_let_help);
  sc->let_temporarily_symbol =   assign_syntax(sc, "let-temporarily", OP_LET_TEMPORARILY,   small_int(2), max_arity,    let_temporarily_help);
  set_immutable(sc->with_let_symbol);

  sc->quote_unchecked_symbol =       assign_internal_syntax(sc, "quote",       OP_QUOTE_UNCHECKED);
  sc->begin1_symbol =                assign_internal_syntax(sc, "begin",       OP_BEGIN1);
  sc->with_baffle_unchecked_symbol = assign_internal_syntax(sc, "with-baffle", OP_WITH_BAFFLE_UNCHECKED);
  sc->let_unchecked_symbol =         assign_internal_syntax(sc, "let",         OP_LET_UNCHECKED);
  sc->let_star_unchecked_symbol =    assign_internal_syntax(sc, "let*",        OP_LET_STAR_UNCHECKED);
  sc->letrec_unchecked_symbol =      assign_internal_syntax(sc, "letrec",      OP_LETREC_UNCHECKED);
  sc->letrec_star_unchecked_symbol = assign_internal_syntax(sc, "letrec*",     OP_LETREC_STAR_UNCHECKED);
  sc->let_no_vars_symbol =           assign_internal_syntax(sc, "let",         OP_LET_NO_VARS);
  sc->let_c_symbol =                 assign_internal_syntax(sc, "let",         OP_LET_C);
  sc->let_s_symbol =                 assign_internal_syntax(sc, "let",         OP_LET_S);
  sc->let_s_z_symbol =               assign_internal_syntax(sc, "let",         OP_LET_S_Z);
  sc->let_all_c_symbol =             assign_internal_syntax(sc, "let",         OP_LET_ALL_C);
  sc->let_all_s_symbol =             assign_internal_syntax(sc, "let",         OP_LET_ALL_S);
  sc->let_all_x_symbol =             assign_internal_syntax(sc, "let",         OP_LET_ALL_X);
  sc->let_star_all_x_symbol =        assign_internal_syntax(sc, "let*",        OP_LET_STAR_ALL_X);
  sc->let_star_a2_symbol =           assign_internal_syntax(sc, "let*",        OP_LET_STAR_A2);
  sc->let_opcq_symbol =              assign_internal_syntax(sc, "let",         OP_LET_opCq);
  sc->let_opssq_symbol =             assign_internal_syntax(sc, "let",         OP_LET_opSSq);
  sc->let_opssq_e_symbol =           assign_internal_syntax(sc, "let",         OP_LET_opSSq_E);
  sc->let_opassq_e_symbol =          assign_internal_syntax(sc, "let",         OP_LET_opaSSq_E);
  sc->let_car_symbol =               assign_internal_syntax(sc, "let",         OP_LET_CAR);
  sc->let_opsq_symbol =              assign_internal_syntax(sc, "let",         OP_LET_opSq);
  sc->let_opsq_p_symbol =            assign_internal_syntax(sc, "let",         OP_LET_opSq_P);
  sc->let_one_symbol =               assign_internal_syntax(sc, "let",         OP_LET_ONE);
  sc->let_z_symbol =                 assign_internal_syntax(sc, "let",         OP_LET_Z);
  sc->let_a_symbol =                 assign_internal_syntax(sc, "let",         OP_LET_A);
  sc->let_a_z_symbol =               assign_internal_syntax(sc, "let",         OP_LET_A_Z);
  sc->let_all_opsq_symbol =          assign_internal_syntax(sc, "let",         OP_LET_ALL_opSq);
  sc->named_let_no_vars_symbol =     assign_internal_syntax(sc, "let",         OP_NAMED_LET_NO_VARS);
  sc->named_let_symbol =             assign_internal_syntax(sc, "let",         OP_NAMED_LET);
  sc->named_let_star_symbol =        assign_internal_syntax(sc, "let*",        OP_NAMED_LET_STAR);
  sc->let_star2_symbol =             assign_internal_syntax(sc, "let*",        OP_LET_STAR2);
  sc->with_let_unchecked_symbol =    assign_internal_syntax(sc, "with-let",    OP_WITH_LET_UNCHECKED);
  sc->with_let_s_symbol =            assign_internal_syntax(sc, "with-let",    OP_WITH_LET_S);
  sc->let_temporarily_unchecked_symbol = assign_internal_syntax(sc, "let-temporarily", OP_LET_TEMP_UNCHECKED);
  sc->case_unchecked_symbol =        assign_internal_syntax(sc, "case",        OP_CASE_UNCHECKED);
  sc->cond_unchecked_symbol =        assign_internal_syntax(sc, "cond",        OP_COND_UNCHECKED);

  sc->do_unchecked_symbol =          assign_internal_syntax(sc, "do",          OP_DO_UNCHECKED);
  sc->dotimes_p_symbol =             assign_internal_syntax(sc, "do",          OP_DOTIMES_P);
  sc->simple_do_symbol =             assign_internal_syntax(sc, "do",          OP_SIMPLE_DO);
  sc->safe_dotimes_symbol =          assign_internal_syntax(sc, "do",          OP_SAFE_DOTIMES);
  sc->safe_do_symbol =               assign_internal_syntax(sc, "do",          OP_SAFE_DO);
  sc->dox_symbol =                   assign_internal_syntax(sc, "do",          OP_DOX);
  sc->do_no_vars_symbol =            assign_internal_syntax(sc, "do",          OP_DO_NO_VARS);
  sc->do_no_vars_no_opt_symbol =     assign_internal_syntax(sc, "do",          OP_DO_NO_VARS_NO_OPT);
  sc->dotimes_one_step_symbol =      assign_internal_syntax(sc, "do",          OP_DOTIMES_ONE_STEP);

  sc->lambda_unchecked_symbol =      assign_internal_syntax(sc, "lambda",      OP_LAMBDA_UNCHECKED);
  sc->lambda_star_unchecked_symbol = assign_internal_syntax(sc, "lambda*",     OP_LAMBDA_STAR_UNCHECKED);
  sc->define_unchecked_symbol =      assign_internal_syntax(sc, "define",      OP_DEFINE_UNCHECKED);
  sc->define_funchecked_symbol =     assign_internal_syntax(sc, "define",      OP_DEFINE_FUNCHECKED);
  sc->define_star_unchecked_symbol = assign_internal_syntax(sc, "define*",     OP_DEFINE_STAR_UNCHECKED);
  sc->define_constant_unchecked_symbol = assign_internal_syntax(sc, "define-constant", OP_DEFINE_CONSTANT_UNCHECKED);
  sc->set_unchecked_symbol =         assign_internal_syntax(sc, "set!",        OP_SET_UNCHECKED);
  sc->set_symbol_c_symbol =          assign_internal_syntax(sc, "set!",        OP_SET_SYMBOL_C);
  sc->set_symbol_l_symbol =          assign_internal_syntax(sc, "set!",        OP_SET_SYMBOL_L);
  sc->set_symbol_s_symbol =          assign_internal_syntax(sc, "set!",        OP_SET_SYMBOL_S);
  sc->set_symbol_q_symbol =          assign_internal_syntax(sc, "set!",        OP_SET_SYMBOL_Q);
  sc->set_symbol_opsq_symbol =       assign_internal_syntax(sc, "set!",        OP_SET_SYMBOL_opSq);
  sc->set_symbol_oplq_symbol =       assign_internal_syntax(sc, "set!",        OP_SET_SYMBOL_opLq);
  sc->set_symbol_opssq_symbol =      assign_internal_syntax(sc, "set!",        OP_SET_SYMBOL_opSSq);
  sc->set_symbol_opllq_symbol =      assign_internal_syntax(sc, "set!",        OP_SET_SYMBOL_opLLq);
  sc->set_symbol_opsssq_symbol =     assign_internal_syntax(sc, "set!",        OP_SET_SYMBOL_opSSSq);
  sc->set_symbol_opcq_symbol =       assign_internal_syntax(sc, "set!",        OP_SET_SYMBOL_opCq);
  sc->set_symbol_p_symbol =          assign_internal_syntax(sc, "set!",        OP_SET_SYMBOL_P);
  sc->set_symbol_z_symbol =          assign_internal_syntax(sc, "set!",        OP_SET_SYMBOL_Z);
  sc->set_symbol_a_symbol =          assign_internal_syntax(sc, "set!",        OP_SET_SYMBOL_A);
  sc->set_normal_symbol =            assign_internal_syntax(sc, "set!",        OP_SET_NORMAL);
  sc->set_pws_symbol =               assign_internal_syntax(sc, "set!",        OP_SET_PWS);
  sc->set_pair_symbol =              assign_internal_syntax(sc, "set!",        OP_SET_PAIR);
  sc->set_dilambda_symbol =          assign_internal_syntax(sc, "set!",        OP_SET_DILAMBDA);
  sc->set_dilambda_z_symbol =        assign_internal_syntax(sc, "set!",        OP_SET_DILAMBDA_Z);
  sc->set_pair_p_symbol =            assign_internal_syntax(sc, "set!",        OP_SET_PAIR_P);
  sc->set_pair_z_symbol =            assign_internal_syntax(sc, "set!",        OP_SET_PAIR_Z);
  sc->set_pair_a_symbol =            assign_internal_syntax(sc, "set!",        OP_SET_PAIR_A);
  sc->set_pair_za_symbol =           assign_internal_syntax(sc, "set!",        OP_SET_PAIR_ZA);
  sc->set_let_s_symbol =             assign_internal_syntax(sc, "set!",        OP_SET_LET_S);
  sc->set_let_all_x_symbol =         assign_internal_syntax(sc, "set!",        OP_SET_LET_ALL_X);
  sc->increment_1_symbol =           assign_internal_syntax(sc, "set!",        OP_INCREMENT_1);
  sc->increment_ss_symbol =          assign_internal_syntax(sc, "set!",        OP_INCREMENT_SS);
  sc->increment_sss_symbol =         assign_internal_syntax(sc, "set!",        OP_INCREMENT_SSS);
  sc->increment_sz_symbol =          assign_internal_syntax(sc, "set!",        OP_INCREMENT_SZ);
  sc->increment_sa_symbol =          assign_internal_syntax(sc, "set!",        OP_INCREMENT_SA);
  sc->increment_saa_symbol =         assign_internal_syntax(sc, "set!",        OP_INCREMENT_SAA);
  sc->decrement_1_symbol =           assign_internal_syntax(sc, "set!",        OP_DECREMENT_1);
  sc->set_cons_symbol =              assign_internal_syntax(sc, "set!",        OP_SET_CONS);
  sc->and_unchecked_symbol =         assign_internal_syntax(sc, "and",         OP_AND_UNCHECKED);
  sc->and_p_symbol =                 assign_internal_syntax(sc, "and",         OP_AND_P);
  sc->and_ap_symbol =                assign_internal_syntax(sc, "and",         OP_AND_AP);
  sc->and_az_symbol =                assign_internal_syntax(sc, "and",         OP_AND_AZ);
  sc->and_safe_p_symbol =            assign_internal_syntax(sc, "and",         OP_AND_SAFE_P);
  sc->and_safe_aa_symbol =           assign_internal_syntax(sc, "and",         OP_AND_SAFE_AA);
  sc->or_unchecked_symbol =          assign_internal_syntax(sc, "or",          OP_OR_UNCHECKED);
  sc->or_p_symbol =                  assign_internal_syntax(sc, "or",          OP_OR_P);
  sc->or_ap_symbol =                 assign_internal_syntax(sc, "or",          OP_OR_AP);
  sc->or_safe_p_symbol =             assign_internal_syntax(sc, "or",          OP_OR_SAFE_P);
  sc->or_safe_aa_symbol =            assign_internal_syntax(sc, "or",          OP_OR_SAFE_AA);
  sc->if_unchecked_symbol =          assign_internal_syntax(sc, "if",          OP_IF_UNCHECKED);

  sc->cond_unchecked_z_symbol =      assign_internal_syntax(sc, "cond",        OP_COND_UNCHECKED_Z);
  sc->cond_simple_symbol =           assign_internal_syntax(sc, "cond",        OP_COND_SIMPLE);
  sc->cond_feed_symbol =             assign_internal_syntax(sc, "cond",        OP_COND_FEED);
  sc->cond_all_x_z_symbol =          assign_internal_syntax(sc, "cond",        OP_COND_ALL_X_Z);
  sc->cond_all_x_symbol =            assign_internal_syntax(sc, "cond",        OP_COND_ALL_X);
  sc->cond_all_x_2_symbol =          assign_internal_syntax(sc, "cond",        OP_COND_ALL_X_2);

  #define assign_if(Sym, Op) \
    sc->if_ ## Sym ## _p_symbol =    assign_internal_syntax(sc, "if",          OP_IF_ ## Op ## _P); \
    sc->if_ ## Sym ## _p_p_symbol =  assign_internal_syntax(sc, "if",          OP_IF_ ## Op ## _P_P); \
    sc->if_ ## Sym ## _n_symbol =    assign_internal_syntax(sc, "if",          OP_IF_ ## Op ## _N); \
    sc->if_ ## Sym ## _n_n_symbol =  assign_internal_syntax(sc, "if",          OP_IF_ ## Op ## _N_N); \
    sc->if_ ## Sym ## _r_symbol =    assign_internal_syntax(sc, "if",          OP_IF_ ## Op ## _R);
  
  assign_if(p, P)
  assign_if(andp, ANDP)
  assign_if(orp, ORP)
  assign_if(and2, AND2)
  assign_if(and3, AND3)
  assign_if(or2, OR2)
  assign_if(z, Z)
  assign_if(s, S)
  assign_if(a, A)
  assign_if(c, C)
  assign_if(cs, CS)
  assign_if(csq, CSQ)
  assign_if(css, CSS)
  assign_if(csc, CSC)
  assign_if(s_opcq, S_opCq)
  assign_if(opsq, opSq)
  assign_if(is_type_s, IS_TYPE_S)
  assign_if(is_type_opsq, IS_TYPE_opSq)
    /* CLL and CLC happen a few times, but make no difference */

  #define assign_case(Sym, Op) \
    sc->case_ ## Sym ## _e_s_symbol = assign_internal_syntax(sc, "case", OP_CASE_ ## Op ## _E_S); \
    sc->case_ ## Sym ## _i_s_symbol = assign_internal_syntax(sc, "case", OP_CASE_ ## Op ## _I_S); \
    sc->case_ ## Sym ## _g_s_symbol = assign_internal_syntax(sc, "case", OP_CASE_ ## Op ## _G_S); \
    sc->case_ ## Sym ## _e_g_symbol = assign_internal_syntax(sc, "case", OP_CASE_ ## Op ## _E_G); \
    sc->case_ ## Sym ## _g_g_symbol = assign_internal_syntax(sc, "case", OP_CASE_ ## Op ## _G_G);
  assign_case(a, A)
  assign_case(s, S)
  assign_case(p, P)

  sc->when_s_symbol =                assign_internal_syntax(sc, "when",        OP_WHEN_S);
  sc->when_a_symbol =                assign_internal_syntax(sc, "when",        OP_WHEN_A);
  sc->when_p_symbol =                assign_internal_syntax(sc, "when",        OP_WHEN_P);
  sc->unless_s_symbol =              assign_internal_syntax(sc, "unless",      OP_UNLESS_S);
  sc->unless_a_symbol =              assign_internal_syntax(sc, "unless",      OP_UNLESS_A);
  sc->when_unchecked_symbol =        assign_internal_syntax(sc, "when",        OP_WHEN_UNCHECKED);
  sc->unless_unchecked_symbol =      assign_internal_syntax(sc, "unless",      OP_UNLESS_UNCHECKED);

  sc->documentation_symbol =  make_symbol(sc, "documentation");
  sc->signature_symbol =      make_symbol(sc, "signature");

#if WITH_IMMUTABLE_UNQUOTE
  /* this code solves the various unquote redefinition troubles
   * if "," -> "(unquote...)" in the reader, (let (, (lambda (x) (+ x 1))) ,,,,1) -> 5
   *   in s7, this requires a quote: (let (, (lambda (x) (+ x 1))) ,,,,'1)
   */
  sc->unquote_symbol =              make_symbol(sc, ",");
  set_immutable(sc->unquote_symbol);
#else
  sc->unquote_symbol =              make_symbol(sc, "unquote");
#endif

  sc->feed_to_symbol =              make_symbol(sc, "=>");
  sc->baffle_symbol =               make_symbol(sc, "(baffle)");
  sc->body_symbol =                 make_symbol(sc, "body");
  sc->error_symbol =                make_symbol(sc, "error");
  sc->read_error_symbol =           make_symbol(sc, "read-error");
  sc->string_read_error_symbol =    make_symbol(sc, "string-read-error");
  sc->syntax_error_symbol =         make_symbol(sc, "syntax-error");
  sc->wrong_type_arg_symbol =       make_symbol(sc, "wrong-type-arg");
  sc->wrong_number_of_args_symbol = make_symbol(sc, "wrong-number-of-args");
  sc->format_error_symbol =         make_symbol(sc, "format-error");
  sc->out_of_range_symbol =         make_symbol(sc, "out-of-range");
  sc->no_catch_symbol =             make_symbol(sc, "no-catch");
  sc->io_error_symbol =             make_symbol(sc, "io-error");
  sc->missing_method_symbol =       make_symbol(sc, "missing-method");
  sc->invalid_escape_function_symbol = make_symbol(sc, "invalid-escape-function");
  sc->baffled_symbol =              make_symbol(sc, "baffled!");

  sc->key_allow_other_keys_symbol = s7_make_keyword(sc, "allow-other-keys");
  sc->key_rest_symbol =             s7_make_keyword(sc, "rest");
  sc->key_readable_symbol =         s7_make_keyword(sc, "readable");

  sc->value_symbol =                s7_make_symbol(sc, "value");
  sc->type_symbol =                 s7_make_symbol(sc, "type");
  sc->else_symbol =                 s7_make_symbol(sc, "else");
  s7_make_slot(sc, sc->nil, sc->else_symbol, sc->else_symbol);
  sc->__func___symbol =             s7_make_symbol(sc, "__func__");

  sc->owlet = init_owlet(sc);

  sc->err_wrap1 = make_permanent_string_wrapper();
  sc->err_wrap2 = make_permanent_string_wrapper();
  sc->wrong_type_arg_info = permanent_list(sc, 6);
  set_car(sc->wrong_type_arg_info, s7_make_permanent_string("~A argument ~D, ~S, is ~A but should be ~A"));

  sc->simple_wrong_type_arg_info = permanent_list(sc, 5);
  set_car(sc->simple_wrong_type_arg_info, s7_make_permanent_string("~A argument, ~S, is ~A but should be ~A"));

  sc->out_of_range_info = permanent_list(sc, 5);
  set_car(sc->out_of_range_info, s7_make_permanent_string("~A argument ~D, ~S, is out of range (~A)"));

  sc->simple_out_of_range_info = permanent_list(sc, 4);
  set_car(sc->simple_out_of_range_info, s7_make_permanent_string("~A argument, ~S, is out of range (~A)"));

  sc->too_many_arguments_string = s7_make_permanent_string("~A: too many arguments: ~A");
  sc->not_enough_arguments_string = s7_make_permanent_string("~A: not enough arguments: ~A");
  sc->division_by_zero_error_string = s7_make_permanent_string("~A: division by zero, ~S");
  sc->division_by_zero_symbol = make_symbol(sc, "division-by-zero");
  sc->missing_method_string = s7_make_permanent_string("missing ~S method in ~S");

  if (!already_inited)
    init_car_a_list();

  for (i = 0; i < NUM_TYPES; i++)
    {
      const char *str;
      str = type_name_from_type(sc, i, INDEFINITE_ARTICLE);
      if (str)
	prepackaged_type_names[i] = s7_make_permanent_string(str);
      else prepackaged_type_names[i] = sc->F;
    }
  /* unset built-ins: T_STACK (can't happen), T_C_OBJECT (want actual name), T_INPUT|OUTPUT_PORT (want string|file|etc included) */

  sc->gc_off = false;

  #define defun(Scheme_Name, C_Name, Req, Opt, Rst) s7_define_typed_function(sc, Scheme_Name, g_ ## C_Name, Req, Opt, Rst, H_ ## C_Name, Q_ ## C_Name)
  #define unsafe_defun(Scheme_Name, C_Name, Req, Opt, Rst) s7_define_unsafe_typed_function(sc, Scheme_Name, g_ ## C_Name, Req, Opt, Rst, H_ ## C_Name, Q_ ## C_Name)

  /* we need the sc->IS_* symbols first for the procedure signature lists */
  sc->is_boolean_symbol = make_symbol(sc, "boolean?");
  pl_bt = s7_make_signature(sc, 2, sc->is_boolean_symbol, sc->T);

  sc->is_symbol_symbol =             defun("symbol?",		is_symbol,		1, 0, false); symbol_type(sc->is_symbol_symbol) = T_SYMBOL;
  sc->is_syntax_symbol =             defun("syntax?",		is_syntax,		1, 0, false); symbol_type(sc->is_syntax_symbol) = T_SYNTAX;
  sc->is_gensym_symbol =             defun("gensym?",		is_gensym,		1, 0, false); 
  sc->is_keyword_symbol =            defun("keyword?",		is_keyword,		1, 0, false);
  sc->is_let_symbol =                defun("let?",		is_let,			1, 0, false); symbol_type(sc->is_let_symbol) = T_LET;
  sc->is_openlet_symbol =            defun("openlet?",		is_openlet,		1, 0, false); 
  sc->is_iterator_symbol =           defun("iterator?",	        is_iterator,		1, 0, false); symbol_type(sc->is_iterator_symbol) = T_ITERATOR;
  sc->is_constant_symbol =           defun("constant?",	        is_constant,		1, 0, false);
  sc->is_macro_symbol =              defun("macro?",		is_macro,		1, 0, false);
  sc->is_c_pointer_symbol =          defun("c-pointer?",	is_c_pointer,		1, 1, false); symbol_type(sc->is_c_pointer_symbol) = T_C_POINTER;
  sc->is_c_object_symbol =           defun("c-object?",	        is_c_object,		1, 0, false); symbol_type(sc->is_c_object_symbol) = T_C_OBJECT;
  sc->is_input_port_symbol =         defun("input-port?",	is_input_port,		1, 0, false); symbol_type(sc->is_input_port_symbol) = T_INPUT_PORT;
  sc->is_output_port_symbol =        defun("output-port?",	is_output_port,		1, 0, false); symbol_type(sc->is_output_port_symbol) = T_OUTPUT_PORT;
  sc->is_eof_object_symbol =         defun("eof-object?",	is_eof_object,		1, 0, false); symbol_type(sc->is_eof_object_symbol) = T_EOF_OBJECT;
  sc->is_integer_symbol =            defun("integer?",		is_integer,		1, 0, false); 
  sc->is_number_symbol =             defun("number?",		is_number,		1, 0, false);
  sc->is_real_symbol =               defun("real?",		is_real,		1, 0, false);
  sc->is_complex_symbol =            defun("complex?",		is_complex,		1, 0, false);
  sc->is_rational_symbol =           defun("rational?",	        is_rational,		1, 0, false);
  sc->is_random_state_symbol =       defun("random-state?",	is_random_state,	1, 0, false); symbol_type(sc->is_random_state_symbol) = T_RANDOM_STATE;
  sc->is_char_symbol =               defun("char?",		is_char,		1, 0, false); symbol_type(sc->is_char_symbol) = T_CHARACTER;
  sc->is_string_symbol =             defun("string?",		is_string,		1, 0, false); symbol_type(sc->is_string_symbol) = T_STRING;
  sc->is_list_symbol =               defun("list?",		is_list,		1, 0, false);
  sc->is_pair_symbol =               defun("pair?",		is_pair,		1, 0, false); symbol_type(sc->is_pair_symbol) = T_PAIR;
  sc->is_vector_symbol =             defun("vector?",		is_vector,		1, 0, false); 
  sc->is_float_vector_symbol =       defun("float-vector?",	is_float_vector,	1, 0, false); symbol_type(sc->is_float_vector_symbol) = T_FLOAT_VECTOR;
  sc->is_int_vector_symbol =         defun("int-vector?",	is_int_vector,		1, 0, false); symbol_type(sc->is_int_vector_symbol) = T_INT_VECTOR;
  sc->is_byte_vector_symbol =        defun("byte-vector?",	is_byte_vector,		1, 0, false);
  sc->is_hash_table_symbol =         defun("hash-table?",	is_hash_table,		1, 0, false); symbol_type(sc->is_hash_table_symbol) = T_HASH_TABLE;
  sc->is_continuation_symbol =       defun("continuation?",	is_continuation,	1, 0, false); symbol_type(sc->is_continuation_symbol) = T_CONTINUATION;
  sc->is_procedure_symbol =          defun("procedure?",	is_procedure,		1, 0, false);
  sc->is_dilambda_symbol =           defun("dilambda?",	        is_dilambda,		1, 0, false);
  /* set above */                    defun("boolean?",		is_boolean,		1, 0, false); symbol_type(sc->is_boolean_symbol) = T_BOOLEAN;
  sc->is_float_symbol =              defun("float?",            is_float,               1, 0, false);
  sc->is_proper_list_symbol =        defun("proper-list?",      is_proper_list,         1, 0, false);
  sc->is_sequence_symbol =           defun("sequence?",	        is_sequence,		1, 0, false);
  sc->is_null_symbol =               defun("null?",		is_null,		1, 0, false); symbol_type(sc->is_null_symbol) = T_NIL;

  /* these are for signatures */
  sc->is_unspecified_symbol = s7_make_symbol(sc, "unspecified?");
  sc->is_undefined_symbol = s7_make_symbol(sc, "undefined?");
  sc->is_integer_or_real_at_end_symbol = s7_make_symbol(sc, "integer:real?");
  sc->is_integer_or_any_at_end_symbol =  s7_make_symbol(sc, "integer:any?");

  pl_p =   s7_make_signature(sc, 2, sc->T, sc->is_pair_symbol);
  pl_tl =  s7_make_signature(sc, 3, s7_make_signature(sc, 2, sc->is_pair_symbol, sc->is_boolean_symbol), sc->T, sc->is_list_symbol); /* memq and memv signature */
  pl_bc =  s7_make_signature(sc, 2, sc->is_boolean_symbol, sc->is_char_symbol);
  pl_bn =  s7_make_signature(sc, 2, sc->is_boolean_symbol, sc->is_number_symbol);
  pl_sf =  s7_make_signature(sc, 3, sc->T, sc->is_string_symbol, sc->is_procedure_symbol);
  pcl_bt = s7_make_circular_signature(sc, 1, 2, sc->is_boolean_symbol, sc->T);
  pcl_bc = s7_make_circular_signature(sc, 1, 2, sc->is_boolean_symbol, sc->is_char_symbol);
  pcl_bs = s7_make_circular_signature(sc, 1, 2, sc->is_boolean_symbol, sc->is_string_symbol);

  pcl_i =  s7_make_circular_signature(sc, 0, 1, sc->is_integer_symbol);
  pcl_r =  s7_make_circular_signature(sc, 0, 1, sc->is_real_symbol);
  pcl_f =  s7_make_circular_signature(sc, 0, 1, sc->is_rational_symbol);
  pcl_n =  s7_make_circular_signature(sc, 0, 1, sc->is_number_symbol);
  pcl_s =  s7_make_circular_signature(sc, 0, 1, sc->is_string_symbol);
  pcl_v =  s7_make_circular_signature(sc, 0, 1, sc->is_vector_symbol);
  pcl_c =  s7_make_circular_signature(sc, 0, 1, sc->is_char_symbol);
  pcl_e =  s7_make_circular_signature(sc, 0, 1, s7_make_signature(sc, 4, sc->is_let_symbol, sc->is_procedure_symbol, sc->is_macro_symbol, sc->is_c_object_symbol));

  sc->values_symbol = make_symbol(sc, "values");

  sc->gensym_symbol =                defun("gensym",		gensym,			0, 1, false);
  sc->symbol_table_symbol =          defun("symbol-table",	symbol_table,		0, 0, false);
  sc->symbol_to_string_symbol =      defun("symbol->string",	symbol_to_string,	1, 0, false);
  sc->string_to_symbol_symbol =      defun("string->symbol",	string_to_symbol,	1, 0, false);
  sc->symbol_symbol =                defun("symbol",		symbol,			1, 0, true);
  sc->symbol_to_value_symbol =       defun("symbol->value",	symbol_to_value,	1, 1, false);
  sc->symbol_to_dynamic_value_symbol = defun("symbol->dynamic-value", symbol_to_dynamic_value, 1, 0, false);
  s7_typed_dilambda(sc, "symbol-access", g_symbol_access, 1, 1, g_symbol_set_access,	2, 1, H_symbol_access, Q_symbol_access, NULL);
  sc->symbol_access_symbol = make_symbol(sc, "symbol-access");

  sc->string_to_keyword_symbol =     defun("string->keyword",	string_to_keyword,      1, 0, false);
  sc->symbol_to_keyword_symbol =     defun("symbol->keyword",	symbol_to_keyword,	1, 0, false);
  sc->keyword_to_symbol_symbol =     defun("keyword->symbol",	keyword_to_symbol,	1, 0, false);

  sc->outlet_symbol =                defun("outlet",		outlet,			1, 0, false);
  sc->rootlet_symbol =               defun("rootlet",		rootlet,		0, 0, false);
  sc->curlet_symbol =                defun("curlet",		curlet,			0, 0, false);
  sc->unlet_symbol =                 defun("unlet",		unlet,			0, 0, false);
  set_immutable(sc->unlet_symbol);
  /* unlet (and with-let) don't actually need to be immutable, but s7.html says they are... */
  sc->sublet_symbol =                defun("sublet",		sublet,			1, 0, true);
  sc->varlet_symbol =                unsafe_defun("varlet",	varlet,			1, 0, true);
  sc->cutlet_symbol =                unsafe_defun("cutlet",	cutlet,			1, 0, true);
  sc->inlet_symbol =                 defun("inlet",		inlet,			0, 0, true);
  sc->owlet_symbol =                 defun("owlet",		owlet,			0, 0, false);
  sc->coverlet_symbol =              defun("coverlet",		coverlet,		1, 0, false);
  sc->openlet_symbol =               defun("openlet",		openlet,		1, 0, false);
  sc->let_ref_symbol =               defun("let-ref",		let_ref,		2, 0, false);
  sc->let_set_symbol =               defun("let-set!",		let_set,		3, 0, false);
  sc->let_ref_fallback_symbol = make_symbol(sc, "let-ref-fallback");
  sc->let_set_fallback_symbol = make_symbol(sc, "let-set!-fallback");

  sc->make_iterator_symbol =         defun("make-iterator",	make_iterator,		1, 1, false);
  sc->iterate_symbol =               defun("iterate",		iterate,		1, 0, false);
  sc->iterator_sequence_symbol =     defun("iterator-sequence", iterator_sequence,	1, 0, false);
  sc->iterator_is_at_end_symbol =    defun("iterator-at-end?",  iterator_is_at_end,	1, 0, false);

  sc->is_provided_symbol =           defun("provided?",	        is_provided,		1, 0, false);
  sc->provide_symbol =               defun("provide",		provide,		1, 0, false);
  sc->is_defined_symbol =            defun("defined?",		is_defined,		1, 2, false);

  sc->c_pointer_symbol =             defun("c-pointer",	        c_pointer,		1, 2, false);

  sc->port_line_number_symbol =      defun("port-line-number",  port_line_number,	0, 1, false);
  sc->port_filename_symbol =         defun("port-filename",	port_filename,		0, 1, false);
  sc->pair_line_number_symbol =      defun("pair-line-number",  pair_line_number,	1, 0, false);
  sc->pair_filename_symbol =         defun("pair-filename",     pair_filename,	        1, 0, false);
#if WITH_PROFILE
                                     defun("profile-line-number", profile_line_number,	1, 0, false);
                                     defun("profile-filename",    profile_filename,     1, 0, false);
#endif
  sc->is_port_closed_symbol =        defun("port-closed?",	is_port_closed,		1, 0, false);

  sc->current_input_port_symbol =    defun("current-input-port", current_input_port,	0, 0, false);
  sc->current_output_port_symbol =   defun("current-output-port", current_output_port,  0, 0, false);
  sc->current_error_port_symbol =    defun("current-error-port", current_error_port,    0, 0, false);
                                     defun("set-current-error-port", set_current_error_port, 1, 0, false);
#if (!WITH_PURE_S7)
  sc->let_to_list_symbol =           defun("let->list",	         let_to_list,		1, 0, false);
                                     defun("set-current-input-port", set_current_input_port, 1, 0, false);
                                     defun("set-current-output-port", set_current_output_port, 1, 0, false);
  sc->is_char_ready_symbol =         defun("char-ready?",	is_char_ready,		0, 1, false); /* the least-used scheme function */
#endif

  sc->close_input_port_symbol =      defun("close-input-port",   close_input_port,	1, 0, false);
  sc->close_output_port_symbol =     defun("close-output-port",  close_output_port,	1, 0, false);
  sc->flush_output_port_symbol =     defun("flush-output-port",  flush_output_port,	0, 1, false);
  sc->open_input_file_symbol =       defun("open-input-file",    open_input_file,	1, 1, false);
  sc->open_output_file_symbol =      defun("open-output-file",   open_output_file,	1, 1, false);
  sc->open_input_string_symbol =     defun("open-input-string",  open_input_string,	1, 0, false);
  sc->open_output_string_symbol =    defun("open-output-string", open_output_string,	0, 0, false);
  sc->get_output_string_symbol =     defun("get-output-string",  get_output_string,	1, 1, false);

  sc->newline_symbol =               defun("newline",		newline,		0, 1, false);
  sc->write_symbol =                 defun("write",		write,			1, 1, false);
  sc->display_symbol =               defun("display",		display,		1, 1, false);
  sc->read_char_symbol =             defun("read-char",	        read_char,		0, 1, false);
  sc->peek_char_symbol =             defun("peek-char",	        peek_char,		0, 1, false);
  sc->write_char_symbol =            defun("write-char",	write_char,		1, 1, false);
  sc->write_string_symbol =          defun("write-string",	write_string,		1, 3, false);
  sc->read_byte_symbol =             defun("read-byte",	        read_byte,		0, 1, false);
  sc->write_byte_symbol =            defun("write-byte",	write_byte,		1, 1, false);
  sc->read_line_symbol =             defun("read-line",	        read_line,		0, 2, false);
  sc->read_string_symbol =           defun("read-string",	read_string,		1, 1, false);
  sc->read_symbol =                  unsafe_defun("read",	read,			0, 1, false);
  /* read can't be safe because it messes with the stack, expecting to be all by itself in the call sequence 
   *   (not embedded in OP_SAFE_C_opSq for example) -- that is, it pushes OP_READ_INTERNAL, then returns
   *   expecting goto START, which would be nonsense if arg=c_call(read) -> c_call(arg).
   *   a safe procedure leaves its argument list alone, does not push anything on the stack,
   *   and leaves sc->code|args unscathed (c_call assumes that is the case).  The stack part can
   *   be hidden: if a c_function calls s7_apply_function (lambda passed as arg as in some clm gens)
   *   then is called with args that use all_x*, and the lambda func does the same, the two calls
   *   can step on each other.  
   */

  sc->call_with_input_string_symbol = unsafe_defun("call-with-input-string", call_with_input_string, 2, 0, false);
  sc->call_with_input_file_symbol =   unsafe_defun("call-with-input-file",   call_with_input_file,   2, 0, false);
  sc->with_input_from_string_symbol = unsafe_defun("with-input-from-string", with_input_from_string, 2, 0, false);
  sc->with_input_from_file_symbol =   unsafe_defun("with-input-from-file",   with_input_from_file,   2, 0, false);

  sc->call_with_output_string_symbol = unsafe_defun("call-with-output-string", call_with_output_string, 1, 0, false);
  sc->call_with_output_file_symbol =   unsafe_defun("call-with-output-file",   call_with_output_file,   2, 0, false);
  sc->with_output_to_string_symbol =   unsafe_defun("with-output-to-string",   with_output_to_string,   1, 0, false);
  sc->with_output_to_file_symbol =     unsafe_defun("with-output-to-file",     with_output_to_file,     2, 0, false);

#if WITH_SYSTEM_EXTRAS
  sc->is_directory_symbol =          defun("directory?",	is_directory,		1, 0, false);
  sc->file_exists_symbol =           defun("file-exists?",	file_exists,		1, 0, false);
  sc->delete_file_symbol =           defun("delete-file",	delete_file,		1, 0, false);
  sc->getenv_symbol =                defun("getenv",		getenv,			1, 0, false);
  sc->system_symbol =                defun("system",		system,			1, 1, false);
#if (!MS_WINDOWS)
  sc->directory_to_list_symbol =     defun("directory->list",   directory_to_list,	1, 0, false);
  sc->file_mtime_symbol =            defun("file-mtime",	file_mtime,		1, 0, false);
#endif
#endif

  sc->real_part_symbol =             defun("real-part",	        real_part,		1, 0, false);
  sc->imag_part_symbol =             defun("imag-part",	        imag_part,		1, 0, false);
  sc->numerator_symbol =             defun("numerator",	        numerator,		1, 0, false);
  sc->denominator_symbol =           defun("denominator",	denominator,		1, 0, false);
  sc->is_even_symbol =               defun("even?",		is_even,		1, 0, false);
  sc->is_odd_symbol =                defun("odd?",		is_odd,			1, 0, false);
  sc->is_zero_symbol =               defun("zero?",		is_zero,		1, 0, false);
  sc->is_positive_symbol =           defun("positive?",	        is_positive,		1, 0, false);
  sc->is_negative_symbol =           defun("negative?",	        is_negative,		1, 0, false);
  sc->is_infinite_symbol =           defun("infinite?",	        is_infinite,		1, 0, false);
  sc->is_nan_symbol =                defun("nan?",		is_nan,			1, 0, false);

#if (!WITH_GMP)
  sc->complex_symbol =               defun("complex",	        complex,	        2, 0, false);
  sc->magnitude_symbol =             defun("magnitude",	        magnitude,		1, 0, false);
  sc->angle_symbol =                 defun("angle",		angle,			1, 0, false);
  sc->rationalize_symbol =           defun("rationalize",	rationalize,		1, 1, false);
  sc->abs_symbol =                   defun("abs",		abs,			1, 0, false);
  sc->exp_symbol =                   defun("exp",		exp,			1, 0, false);
  sc->log_symbol =                   defun("log",		log,			1, 1, false);
  sc->sin_symbol =                   defun("sin",		sin,			1, 0, false);
  sc->cos_symbol =                   defun("cos",		cos,			1, 0, false);
  sc->tan_symbol =                   defun("tan",		tan,			1, 0, false);
  sc->asin_symbol =                  defun("asin",		asin,			1, 0, false);
  sc->acos_symbol =                  defun("acos",		acos,			1, 0, false);
  sc->atan_symbol =                  defun("atan",		atan,			1, 1, false);
  sc->sinh_symbol =                  defun("sinh",		sinh,			1, 0, false);
  sc->cosh_symbol =                  defun("cosh",		cosh,			1, 0, false);
  sc->tanh_symbol =                  defun("tanh",		tanh,			1, 0, false);
  sc->asinh_symbol =                 defun("asinh",		asinh,			1, 0, false);
  sc->acosh_symbol =                 defun("acosh",		acosh,			1, 0, false);
  sc->atanh_symbol =                 defun("atanh",		atanh,			1, 0, false);
  sc->sqrt_symbol =                  defun("sqrt",		sqrt,			1, 0, false);
  sc->expt_symbol =                  defun("expt",		expt,			2, 0, false);
  sc->floor_symbol =                 defun("floor",		floor,			1, 0, false);
  sc->ceiling_symbol =               defun("ceiling",		ceiling,		1, 0, false);
  sc->truncate_symbol =              defun("truncate",		truncate,		1, 0, false);
  sc->round_symbol =                 defun("round",		round,			1, 0, false);
  sc->lcm_symbol =                   defun("lcm",		lcm,			0, 0, true);
  sc->gcd_symbol =                   defun("gcd",		gcd,			0, 0, true);
  sc->add_symbol =                   defun("+",		        add,			0, 0, true);
  sc->subtract_symbol =              defun("-",		        subtract,		1, 0, true);
  sc->multiply_symbol =              defun("*",		        multiply,		0, 0, true);
  sc->divide_symbol =                defun("/",		        divide,			1, 0, true);
  sc->max_symbol =                   defun("max",		max,			1, 0, true);
  sc->min_symbol =                   defun("min",		min,			1, 0, true);
  sc->quotient_symbol =              defun("quotient",		quotient,		2, 0, false);
  sc->remainder_symbol =             defun("remainder",	        remainder,		2, 0, false);
  sc->modulo_symbol =                defun("modulo",		modulo,			2, 0, false);
  sc->eq_symbol =                    defun("=",		        equal,			2, 0, true);
  sc->lt_symbol =                    defun("<",		        less,			2, 0, true);
  sc->gt_symbol =                    defun(">",		        greater,		2, 0, true);
  sc->leq_symbol =                   defun("<=",		less_or_equal,		2, 0, true);
  sc->geq_symbol =                   defun(">=",		greater_or_equal,	2, 0, true);
  sc->logior_symbol =                defun("logior",		logior,			0, 0, true);
  sc->logxor_symbol =                defun("logxor",		logxor,			0, 0, true);
  sc->logand_symbol =                defun("logand",		logand,			0, 0, true);
  sc->lognot_symbol =                defun("lognot",		lognot,			1, 0, false);
  sc->ash_symbol =                   defun("ash",		ash,			2, 0, false);
  sc->random_state_symbol =          defun("random-state",      random_state,	        1, 1, false);
  sc->random_symbol =                defun("random",		random,			1, 1, false);
#if (!WITH_PURE_S7)
  sc->inexact_to_exact_symbol =      defun("inexact->exact",	inexact_to_exact,	1, 0, false);
  sc->exact_to_inexact_symbol =      defun("exact->inexact",	exact_to_inexact,	1, 0, false);
  sc->integer_length_symbol =        defun("integer-length",	integer_length,		1, 0, false);
  sc->make_polar_symbol =            defun("make-polar",	make_polar,		2, 0, false);
  sc->make_rectangular_symbol =      defun("make-rectangular",  complex,	        2, 0, false);
#endif
#endif /* !gmp */

  sc->logbit_symbol =                defun("logbit?",		logbit,			2, 0, false);
  sc->integer_decode_float_symbol =  defun("integer-decode-float", integer_decode_float, 1, 0, false);
#if (!WITH_PURE_S7)
  sc->is_exact_symbol =              defun("exact?",		is_exact,		1, 0, false);
  sc->is_inexact_symbol =            defun("inexact?",		is_inexact,		1, 0, false);
#endif
  sc->random_state_to_list_symbol =  defun("random-state->list", random_state_to_list, 0, 1, false);

  sc->number_to_string_symbol =      defun("number->string",	number_to_string,	1, 1, false);
  sc->string_to_number_symbol =      defun("string->number",	string_to_number,	1, 1, false);

  sc->char_upcase_symbol =           defun("char-upcase",	char_upcase,		1, 0, false);
  sc->char_downcase_symbol =         defun("char-downcase",	char_downcase,		1, 0, false);
  sc->char_to_integer_symbol =       defun("char->integer",	char_to_integer,	1, 0, false);
  sc->integer_to_char_symbol =       defun("integer->char",	integer_to_char,	1, 0, false);

  sc->is_char_upper_case_symbol =    defun("char-upper-case?",  is_char_upper_case,	1, 0, false);
  sc->is_char_lower_case_symbol =    defun("char-lower-case?",  is_char_lower_case,	1, 0, false);
  sc->is_char_alphabetic_symbol =    defun("char-alphabetic?",  is_char_alphabetic,	1, 0, false);
  sc->is_char_numeric_symbol =       defun("char-numeric?",	is_char_numeric,	1, 0, false);
  sc->is_char_whitespace_symbol =    defun("char-whitespace?",  is_char_whitespace,	1, 0, false);

  sc->char_eq_symbol =               defun("char=?",		chars_are_equal,	2, 0, true);
  sc->char_lt_symbol =               defun("char<?",		chars_are_less,		2, 0, true);
  sc->char_gt_symbol =               defun("char>?",		chars_are_greater,	2, 0, true);
  sc->char_leq_symbol =              defun("char<=?",		chars_are_leq,		2, 0, true);
  sc->char_geq_symbol =              defun("char>=?",		chars_are_geq,		2, 0, true);
  sc->char_position_symbol =         defun("char-position",	char_position,		2, 1, false);
  sc->string_position_symbol =       defun("string-position",	string_position,	2, 1, false);

  sc->make_string_symbol =           defun("make-string",	make_string,		1, 1, false);
  sc->string_ref_symbol =            defun("string-ref",	string_ref,		2, 0, false);
  sc->string_set_symbol =            defun("string-set!",	string_set,		3, 0, false);

  sc->string_eq_symbol =             defun("string=?",		strings_are_equal,	2, 0, true);
  sc->string_lt_symbol =             defun("string<?",		strings_are_less,	2, 0, true);
  sc->string_gt_symbol =             defun("string>?",		strings_are_greater,	2, 0, true);
  sc->string_leq_symbol =            defun("string<=?",	        strings_are_leq,	2, 0, true);
  sc->string_geq_symbol =            defun("string>=?",	        strings_are_geq,	2, 0, true);

#if (!WITH_PURE_S7)
  sc->char_ci_eq_symbol =            defun("char-ci=?",	        chars_are_ci_equal,	2, 0, true);
  sc->char_ci_lt_symbol =            defun("char-ci<?",	        chars_are_ci_less,	2, 0, true);
  sc->char_ci_gt_symbol =            defun("char-ci>?",	        chars_are_ci_greater,	2, 0, true);
  sc->char_ci_leq_symbol =           defun("char-ci<=?",	chars_are_ci_leq,	2, 0, true);
  sc->char_ci_geq_symbol =           defun("char-ci>=?",	chars_are_ci_geq,	2, 0, true);
  sc->string_ci_eq_symbol =          defun("string-ci=?",	strings_are_ci_equal,	2, 0, true);
  sc->string_ci_lt_symbol =          defun("string-ci<?",	strings_are_ci_less,	2, 0, true);
  sc->string_ci_gt_symbol =          defun("string-ci>?",	strings_are_ci_greater, 2, 0, true);
  sc->string_ci_leq_symbol =         defun("string-ci<=?",	strings_are_ci_leq,	2, 0, true);
  sc->string_ci_geq_symbol =         defun("string-ci>=?",	strings_are_ci_geq,	2, 0, true);
  sc->string_copy_symbol =           defun("string-copy",	string_copy,		1, 0, false);
  sc->string_fill_symbol =           defun("string-fill!",	string_fill,		2, 2, false);
  sc->list_to_string_symbol =        defun("list->string",	list_to_string,		1, 0, false);
  sc->string_length_symbol =         defun("string-length",	string_length,		1, 0, false);
  sc->string_to_list_symbol =        defun("string->list",	string_to_list,		1, 2, false);
#endif

  sc->string_downcase_symbol =       defun("string-downcase",	string_downcase,	1, 0, false);
  sc->string_upcase_symbol =         defun("string-upcase",	string_upcase,		1, 0, false);
  sc->string_append_symbol =         defun("string-append",	string_append,		0, 0, true);
  sc->substring_symbol =             defun("substring",	        substring,		2, 1, false);
  sc->string_symbol =                defun("string",		string,			0, 0, true);
  sc->object_to_string_symbol =      defun("object->string",	object_to_string,	1, 2, false);
  sc->format_symbol =                defun("format",		format,			1, 0, true);
  /* this was unsafe, but was that due to the (ill-advised) use of temp_call_2 in the arg lists? */
  sc->object_to_let_symbol =         defun("object->let",	object_to_let,	        1, 0, false);

  sc->cons_symbol =                  defun("cons",		cons,			2, 0, false);
  sc->car_symbol =                   defun("car",		car,			1, 0, false);
  sc->cdr_symbol =                   defun("cdr",		cdr,			1, 0, false);
  sc->set_car_symbol =               defun("set-car!",		set_car,		2, 0, false);
  sc->set_cdr_symbol =               unsafe_defun("set-cdr!",	set_cdr,		2, 0, false);
  sc->caar_symbol =                  defun("caar",		caar,			1, 0, false);
  sc->cadr_symbol =                  defun("cadr",		cadr,			1, 0, false);
  sc->cdar_symbol =                  defun("cdar",		cdar,			1, 0, false);
  sc->cddr_symbol =                  defun("cddr",		cddr,			1, 0, false);
  sc->caaar_symbol =                 defun("caaar",		caaar,			1, 0, false);
  sc->caadr_symbol =                 defun("caadr",		caadr,			1, 0, false);
  sc->cadar_symbol =                 defun("cadar",		cadar,			1, 0, false);
  sc->cdaar_symbol =                 defun("cdaar",		cdaar,			1, 0, false);
  sc->caddr_symbol =                 defun("caddr",		caddr,			1, 0, false);
  sc->cdddr_symbol =                 defun("cdddr",		cdddr,			1, 0, false);
  sc->cdadr_symbol =                 defun("cdadr",		cdadr,			1, 0, false);
  sc->cddar_symbol =                 defun("cddar",		cddar,			1, 0, false);
  sc->caaaar_symbol =                defun("caaaar",		caaaar,			1, 0, false);
  sc->caaadr_symbol =                defun("caaadr",		caaadr,			1, 0, false);
  sc->caadar_symbol =                defun("caadar",		caadar,			1, 0, false);
  sc->cadaar_symbol =                defun("cadaar",		cadaar,			1, 0, false);
  sc->caaddr_symbol =                defun("caaddr",		caaddr,			1, 0, false);
  sc->cadddr_symbol =                defun("cadddr",		cadddr,			1, 0, false);
  sc->cadadr_symbol =                defun("cadadr",		cadadr,			1, 0, false);
  sc->caddar_symbol =                defun("caddar",		caddar,			1, 0, false);
  sc->cdaaar_symbol =                defun("cdaaar",		cdaaar,			1, 0, false);
  sc->cdaadr_symbol =                defun("cdaadr",		cdaadr,			1, 0, false);
  sc->cdadar_symbol =                defun("cdadar",		cdadar,			1, 0, false);
  sc->cddaar_symbol =                defun("cddaar",		cddaar,			1, 0, false);
  sc->cdaddr_symbol =                defun("cdaddr",		cdaddr,			1, 0, false);
  sc->cddddr_symbol =                defun("cddddr",		cddddr,			1, 0, false);
  sc->cddadr_symbol =                defun("cddadr",		cddadr,			1, 0, false);
  sc->cdddar_symbol =                defun("cdddar",		cdddar,			1, 0, false);

  sc->assq_symbol =                  defun("assq",		assq,			2, 0, false);
  sc->assv_symbol =                  defun("assv",		assv,			2, 0, false);
  sc->assoc_symbol =                 unsafe_defun("assoc",	assoc,			2, 1, false); 
  sc->memq_symbol =                  defun("memq",		memq,			2, 0, false);
  sc->memv_symbol =                  defun("memv",		memv,			2, 0, false);
  sc->member_symbol =                unsafe_defun("member",	member,			2, 1, false); 

  sc->list_symbol =                  defun("list",		list,			0, 0, true);
  sc->list_ref_symbol =              defun("list-ref",		list_ref,		2, 0, true);
  sc->list_set_symbol =              defun("list-set!",	        list_set,		3, 0, true);
  sc->list_tail_symbol =             defun("list-tail",	        list_tail,		2, 0, false);
  sc->make_list_symbol =             defun("make-list",  	make_list,		1, 1, false);

  sc->length_symbol =                defun("length",		length,			1, 0, false);
  sc->copy_symbol =                  defun("copy",		copy,			1, 3, false);
  sc->fill_symbol =                  defun("fill!",		fill,			2, 2, false);
  sc->reverse_symbol =               defun("reverse",		reverse,		1, 0, false);
  sc->reverseb_symbol =              defun("reverse!",		reverse_in_place,	1, 0, false);
  sc->sort_symbol =                  unsafe_defun("sort!",	sort,			2, 0, false); 
  sc->append_symbol =                defun("append",		append,			0, 0, true);

#if (!WITH_PURE_S7)
  sc->vector_append_symbol =         defun("vector-append",	vector_append,		0, 0, true);
  sc->list_to_vector_symbol =        defun("list->vector",	list_to_vector,		1, 0, false);
  sc->vector_fill_symbol =           defun("vector-fill!",	vector_fill,		2, 2, false);
  sc->vector_length_symbol =         defun("vector-length",	vector_length,		1, 0, false);
  sc->vector_to_list_symbol =        defun("vector->list",	vector_to_list,		1, 2, false);
#else
  sc->vector_append_symbol = sc->append_symbol;
  sc->vector_fill_symbol = sc->fill_symbol;
  sc->string_fill_symbol = sc->fill_symbol;
#endif
  sc->vector_ref_symbol =            defun("vector-ref",	vector_ref,		2, 0, true);
  sc->vector_set_symbol =            defun("vector-set!",	vector_set,		3, 0, true);
  sc->vector_dimensions_symbol =     defun("vector-dimensions", vector_dimensions,	1, 0, false);
  sc->make_vector_symbol =           defun("make-vector",	make_vector,		1, 1, false);
  sc->make_shared_vector_symbol =    defun("make-shared-vector", make_shared_vector,	2, 1, false);
  sc->vector_symbol =                defun("vector",		vector,			0, 0, true);
  set_setter(sc->vector_symbol); /* like cons, I guess */
  sc->vector_function = slot_value(global_slot(sc->vector_symbol));

  sc->float_vector_symbol =          defun("float-vector",	float_vector,		0, 0, true);
  sc->make_float_vector_symbol =     defun("make-float-vector", make_float_vector,	1, 1, false);
  sc->float_vector_set_symbol =      defun("float-vector-set!", float_vector_set,	3, 0, true);
  sc->float_vector_ref_symbol =      defun("float-vector-ref",  float_vector_ref,	2, 0, true);

  sc->int_vector_symbol =            defun("int-vector",	int_vector,		0, 0, true);
  sc->make_int_vector_symbol =       defun("make-int-vector",	make_int_vector,	1, 1, false);
  sc->int_vector_set_symbol =        defun("int-vector-set!",	int_vector_set,		3, 0, true);
  sc->int_vector_ref_symbol =        defun("int-vector-ref",	int_vector_ref,		2, 0, true);

  sc->string_to_byte_vector_symbol = defun("string->byte-vector", string_to_byte_vector, 1, 0, false);
  sc->byte_vector_symbol =           defun("byte-vector",	byte_vector,		0, 0, true);
  sc->make_byte_vector_symbol =      defun("make-byte-vector",  make_byte_vector,	1, 1, false);
  sc->byte_vector_ref_symbol =       defun("byte-vector-ref",	byte_vector_ref,        2, 0, false);
  sc->byte_vector_set_symbol =       defun("byte-vector-set!",	byte_vector_set,	3, 0, false);

  sc->hash_table_symbol =            defun("hash-table",	hash_table,		0, 0, true);
  sc->hash_table_star_symbol =       defun("hash-table*",	hash_table_star,	0, 0, true);
  sc->make_hash_table_symbol =       defun("make-hash-table",	make_hash_table,	0, 2, false);
  sc->hash_table_ref_symbol =        defun("hash-table-ref",	hash_table_ref,		2, 0, true);
  sc->hash_table_set_symbol =        defun("hash-table-set!",	hash_table_set,		3, 0, false);
  sc->hash_table_entries_symbol =    defun("hash-table-entries", hash_table_entries,	1, 0, false);

  sc->cyclic_sequences_symbol =      defun("cyclic-sequences",  cyclic_sequences,	1, 0, false);
  sc->call_cc_symbol =               unsafe_defun("call/cc",	call_cc,		1, 0, false);
  sc->call_with_current_continuation_symbol = unsafe_defun("call-with-current-continuation", call_cc, 1, 0, false);
  sc->call_with_exit_symbol =        unsafe_defun("call-with-exit", call_with_exit,	1, 0, false);

  sc->load_symbol =                  unsafe_defun("load",	load,			1, 1, false);
  sc->autoload_symbol =              unsafe_defun("autoload",	autoload,		2, 0, false);
  sc->eval_symbol =                  unsafe_defun("eval",	eval,			1, 1, false);
  sc->eval_string_symbol =           unsafe_defun("eval-string", eval_string,		1, 1, false);
  sc->apply_symbol =                 unsafe_defun("apply",	apply,			1, 0, true);
  sc->apply_function = slot_value(global_slot(sc->apply_symbol));
  set_type(sc->apply_function, type(sc->apply_function) | T_COPY_ARGS);
  /* (let ((x '((1 2) 3 4))) (catch #t (lambda () (apply apply apply x)) (lambda args 'error)) x) should not mess up x! */

  sc->for_each_symbol =              unsafe_defun("for-each",	for_each,		2, 0, true); 
  sc->map_symbol =                   unsafe_defun("map",	map,			2, 0, true); 
  sc->dynamic_wind_symbol =          unsafe_defun("dynamic-wind", dynamic_wind,	3, 0, false);
  /* sc->values_symbol = */          unsafe_defun("values",	values,			0, 0, true);
  sc->catch_symbol =                 unsafe_defun("catch",	catch,			3, 0, false);
  sc->throw_symbol =                 unsafe_defun("throw",	throw,			1, 0, true);
  sc->error_symbol =                 unsafe_defun("error",	error,			0, 0, true);
  /* it's faster to leave error/throw unsafe than to set needs_copied_args and use s7_define_safe_function because copy_list overwhelms any other savings */
  sc->stacktrace_symbol =            defun("stacktrace",	stacktrace,		0, 5, false);

  sc->apply_values_symbol =          unsafe_defun("apply-values", apply_values, 0, 1, false);
  set_immutable(sc->apply_values_symbol);
  sc->apply_values_function = slot_value(global_slot(sc->apply_values_symbol));

  sc->list_values_symbol =           unsafe_defun("list-values", list_values, 0, 0, true); /* see comment above */
  set_immutable(sc->list_values_symbol);
  
  sc->procedure_documentation_symbol = defun("procedure-documentation", procedure_documentation, 1, 0, false);
  sc->procedure_signature_symbol =   defun("procedure-signature", procedure_signature,	1, 0, false);
  sc->help_symbol =                  defun("help",		help,			1, 0, false);
  sc->procedure_source_symbol =      defun("procedure-source",  procedure_source,	1, 0, false);
  sc->funclet_symbol =               defun("funclet",		funclet,		1, 0, false);
  sc->dilambda_symbol =              defun("dilambda",          dilambda,               2, 0, false);
  s7_typed_dilambda(sc, "procedure-setter", g_procedure_setter, 1, 0, g_procedure_set_setter, 2, 0, H_procedure_setter, Q_procedure_setter, NULL);

  sc->arity_symbol =                 defun("arity",		arity,			1, 0, false);
  sc->is_aritable_symbol =           defun("aritable?",	        is_aritable,		2, 0, false);

  sc->not_symbol =                   defun("not",		not,			1, 0, false);
  sc->is_eq_symbol =                 defun("eq?",		is_eq,			2, 0, false);
  sc->is_eqv_symbol =                defun("eqv?",		is_eqv,			2, 0, false);
  sc->is_equal_symbol =              defun("equal?",		is_equal,		2, 0, false);
  sc->is_morally_equal_symbol =      defun("morally-equal?",	is_morally_equal,	2, 0, false);
  sc->type_of_symbol =               defun("type-of",		type_of,		1, 0, false);

  sc->gc_symbol =                    defun("gc",		gc,			0, 1, false);
  sc->s7_version_symbol =            defun("s7-version",	s7_version,		0, 0, false);
                                     defun("emergency-exit",	emergency_exit,		0, 1, false);
  sc->exit_symbol =                  defun("exit",		exit,			0, 1, false);
#if DEBUGGING
  s7_define_function(sc, "abort", g_abort, 0, 0, true, "drop into gdb I hope");
#endif
  s7_define_function(sc, "s7-optimize", g_optimize, 1, 0, false, "short-term debugging aid");

  sym = s7_define_function(sc, "(c-object set)", g_internal_c_object_set, 1, 0, true, "internal object setter redirection");
  sc->object_set_function = slot_value(global_slot(sym));

  set_scope_safe(slot_value(global_slot(sc->call_with_input_string_symbol)));
  set_scope_safe(slot_value(global_slot(sc->call_with_input_file_symbol)));
  set_scope_safe(slot_value(global_slot(sc->call_with_output_string_symbol)));
  set_scope_safe(slot_value(global_slot(sc->call_with_output_file_symbol)));
  set_scope_safe(slot_value(global_slot(sc->with_input_from_string_symbol)));
  set_scope_safe(slot_value(global_slot(sc->with_input_from_file_symbol)));
  set_scope_safe(slot_value(global_slot(sc->with_output_to_string_symbol)));
  set_scope_safe(slot_value(global_slot(sc->with_output_to_file_symbol)));
  set_scope_safe(slot_value(global_slot(sc->set_cdr_symbol)));
  set_maybe_safe(slot_value(global_slot(sc->assoc_symbol)));
  set_scope_safe(slot_value(global_slot(sc->assoc_symbol)));
  set_maybe_safe(slot_value(global_slot(sc->member_symbol))); 
  set_scope_safe(slot_value(global_slot(sc->member_symbol)));
  set_scope_safe(slot_value(global_slot(sc->sort_symbol)));
  set_scope_safe(slot_value(global_slot(sc->call_with_exit_symbol)));
  set_scope_safe(slot_value(global_slot(sc->for_each_symbol)));
  set_scope_safe(slot_value(global_slot(sc->map_symbol)));
  set_scope_safe(slot_value(global_slot(sc->dynamic_wind_symbol)));
  set_scope_safe(slot_value(global_slot(sc->catch_symbol)));
  set_scope_safe(slot_value(global_slot(sc->throw_symbol)));
  set_scope_safe(slot_value(global_slot(sc->error_symbol)));
  set_scope_safe(slot_value(global_slot(sc->apply_values_symbol)));
  set_scope_safe(slot_value(global_slot(sc->list_values_symbol)));

  defun("tree-leaves",   tree_leaves,   1, 0, false);
  defun("tree-memq",     tree_memq,     2, 0, false);
  defun("tree-set-memq", tree_set_memq, 2, 0, false);
  sc->tree_count_symbol = defun("tree-count", tree_count, 2, 1, false);


  /* -------- *features* -------- */
  sc->features_symbol = s7_define_variable(sc, "*features*", sc->nil);
  s7_symbol_set_access(sc, sc->features_symbol, s7_make_function(sc, "(set *features*)", g_features_set, 2, 0, false, "*features* accessor"));

  /* -------- *load-path* -------- */
  sc->load_path_symbol = s7_define_variable_with_documentation(sc, "*load-path*", sc->nil, 
			   "*load-path* is a list of directories (strings) that the load function searches if it is passed an incomplete file name");
  s7_symbol_set_access(sc, sc->load_path_symbol, s7_make_function(sc, "(set *load-path*)", g_load_path_set, 2, 0, false, "*load-path* accessor"));

#ifdef CLOAD_DIR
  sc->cload_directory_symbol = s7_define_variable(sc, "*cload-directory*", s7_make_string(sc, (char *)CLOAD_DIR));
  s7_add_to_load_path(sc, (const char *)CLOAD_DIR);
#else
  sc->cload_directory_symbol = s7_define_variable(sc, "*cload-directory*", make_empty_string(sc, 0, 0));
#endif
  s7_symbol_set_access(sc, sc->cload_directory_symbol, s7_make_function(sc, "(set *cload-directory*)", g_cload_directory_set, 2, 0, false, 
                           "*cload-directory* accessor"));


  /* -------- *autoload* --------
   * this pretends to be a hash-table or environment, but it's actually a function
   */
  sc->autoloader_symbol = s7_define_function(sc, "*autoload*", g_autoloader, 1, 0, false, H_autoloader);
  sym = s7_define_variable(sc, "*libraries*", sc->nil);
  sc->libraries = global_slot(sym);

  s7_autoload(sc, make_symbol(sc, "cload.scm"),       s7_make_permanent_string("cload.scm"));
  s7_autoload(sc, make_symbol(sc, "lint.scm"),        s7_make_permanent_string("lint.scm"));
  s7_autoload(sc, make_symbol(sc, "stuff.scm"),       s7_make_permanent_string("stuff.scm"));
  s7_autoload(sc, make_symbol(sc, "mockery.scm"),     s7_make_permanent_string("mockery.scm"));
  s7_autoload(sc, make_symbol(sc, "write.scm"),       s7_make_permanent_string("write.scm"));
  s7_autoload(sc, make_symbol(sc, "repl.scm"),        s7_make_permanent_string("repl.scm"));
  s7_autoload(sc, make_symbol(sc, "r7rs.scm"),        s7_make_permanent_string("r7rs.scm"));

  s7_autoload(sc, make_symbol(sc, "libc.scm"),        s7_make_permanent_string("libc.scm"));
  s7_autoload(sc, make_symbol(sc, "libm.scm"),        s7_make_permanent_string("libm.scm"));
  s7_autoload(sc, make_symbol(sc, "libdl.scm"),       s7_make_permanent_string("libdl.scm"));
  s7_autoload(sc, make_symbol(sc, "libgsl.scm"),      s7_make_permanent_string("libgsl.scm"));
  s7_autoload(sc, make_symbol(sc, "libgdbm.scm"),     s7_make_permanent_string("libgdbm.scm"));
  s7_autoload(sc, make_symbol(sc, "libutf8proc.scm"), s7_make_permanent_string("libutf8proc.scm"));

  sc->require_symbol = s7_define_macro(sc, "require", g_require, 0, 0, true, H_require);
  sc->stacktrace_defaults = s7_list(sc, 5, small_int(3), small_int(45), small_int(80), small_int(45), sc->T);

  /* -------- *#readers* -------- */
  sym = s7_define_variable(sc, "*#readers*", sc->nil);
  sc->sharp_readers = global_slot(sym);
  s7_symbol_set_access(sc, sym, s7_make_function(sc, "(set *#readers*)", g_sharp_readers_set, 2, 0, false, "*#readers* accessor"));

  /* sigh... I don't like these! */
  s7_define_constant(sc, "nan.0", real_NaN);
  s7_define_constant(sc, "-nan.0", real_NaN);
  s7_define_constant(sc, "inf.0", real_infinity);
  s7_define_constant(sc, "-inf.0", real_minus_infinity);

  /* *features* */
  s7_provide(sc, "s7");
  s7_provide(sc, "s7-" S7_VERSION);
  s7_provide(sc, "ratio");

#if WITH_PURE_S7
  s7_provide(sc, "pure-s7");
#endif
#if WITH_EXTRA_EXPONENT_MARKERS
  s7_provide(sc, "dfls-exponents");
#endif
#if HAVE_OVERFLOW_CHECKS
  s7_provide(sc, "overflow-checks");
#endif
#if WITH_SYSTEM_EXTRAS
  s7_provide(sc, "system-extras");
#endif
#if WITH_IMMUTABLE_UNQUOTE
  s7_provide(sc, "immutable-unquote");
#endif
#if DEBUGGING
  s7_provide(sc, "debugging");
#endif
#if WITH_PROFILE
  s7_provide(sc, "profiling");
#endif
#if HAVE_COMPLEX_NUMBERS
  s7_provide(sc, "complex-numbers");
#endif
#if WITH_C_LOADER
  s7_provide(sc, "dlopen");
#endif
#if (!DISABLE_AUTOLOAD)
  s7_provide(sc, "autoload");
#endif

#ifdef __APPLE__
  s7_provide(sc, "osx");
#endif
#ifdef __linux__
  s7_provide(sc, "linux");
#endif
#ifdef __OpenBSD__
  s7_provide(sc, "openbsd");
#endif
#ifdef __NetBSD__
  s7_provide(sc, "netbsd");
#endif
#ifdef __FreeBSD__
  s7_provide(sc, "freebsd");
#endif
#if MS_WINDOWS
  s7_provide(sc, "windows");
#endif
#ifdef __bfin__
  s7_provide(sc, "blackfin");
#endif
#ifdef __ANDROID__
  s7_provide(sc, "android");
#endif
#ifdef __CYGWIN__
  s7_provide(sc, "cygwin");
#endif
#ifdef __hpux
  s7_provide(sc, "hpux");
#endif
#if defined(__sun) && defined(__SVR4)
  s7_provide(sc, "solaris");
#endif
#ifdef __SUNPRO_C
  s7_provide(sc, "sunpro_c");
#endif
#ifdef __MINGW32__
  s7_provide(sc, "mingw");
#endif

  sc->vector_set_function = slot_value(global_slot(sc->vector_set_symbol));
  set_setter(sc->vector_set_symbol);
  /* not float-vector-set! here */

  sc->list_set_function = slot_value(global_slot(sc->list_set_symbol));
  set_setter(sc->list_set_symbol);

  sc->hash_table_set_function = slot_value(global_slot(sc->hash_table_set_symbol));
  set_setter(sc->hash_table_set_symbol);

  sc->let_set_function = slot_value(global_slot(sc->let_set_symbol));
  set_setter(sc->let_set_symbol);

  set_setter(sc->cons_symbol); /* (this blocks an over-eager do loop optimization -- see do-test-15 in s7test) */

  sc->string_set_function = slot_value(global_slot(sc->string_set_symbol));
  set_setter(sc->string_set_symbol);

  set_setter(sc->set_car_symbol);
  set_setter(sc->set_cdr_symbol);

#if (!WITH_PURE_S7)
  set_setter(s7_make_symbol(sc, "set-current-input-port"));
  set_setter(s7_make_symbol(sc, "set-current-output-port"));
  s7_function_set_setter(sc, "current-input-port",  "set-current-input-port");
  s7_function_set_setter(sc, "current-output-port", "set-current-output-port");
#endif

  set_setter(s7_make_symbol(sc, "set-current-error-port"));
  s7_function_set_setter(sc, "current-error-port",  "set-current-error-port");
  /* despite the similar names, current-error-port is different from the other two, and a setter is needed
   *    in scheme because error and warn send output to it by default.  It is not a "dynamic variable" unlike
   *    the other two.  In the input/output cases, setting the port can only cause confusion.
   *    current-error-port should simply be an s7 variable with a name like *error-port* and an accessor to
   *    ensure its new value, if any, is an output port.
   */

  s7_function_set_setter(sc, "car",              "set-car!");
  s7_function_set_setter(sc, "cdr",              "set-cdr!");
  s7_function_set_setter(sc, "hash-table-ref",   "hash-table-set!");
  s7_function_set_setter(sc, "vector-ref",       "vector-set!");
  s7_function_set_setter(sc, "float-vector-ref", "float-vector-set!");
  s7_function_set_setter(sc, "int-vector-ref",   "int-vector-set!");
  s7_function_set_setter(sc, "list-ref",         "list-set!");
  s7_function_set_setter(sc, "let-ref",          "let-set!");
  s7_function_set_setter(sc, "string-ref",       "string-set!");
  c_function_set_setter(slot_value(global_slot(sc->outlet_symbol)), s7_make_function(sc, "(set! outlet)", g_set_outlet, 2, 0, false, "outlet setter"));
  c_function_set_setter(slot_value(global_slot(sc->port_line_number_symbol)), s7_make_function(sc, "(set! port-line-number)", g_set_port_line_number, 1, 1, false, "port line setter"));

  {
    int32_t i, top;
#if WITH_GMP
    #define S7_LOG_LLONG_MAX 36.736800
    #define S7_LOG_LONG_MAX  16.6355322
#else
    /* actually not safe = (log (- (expt 2 63) 1)) and (log (- (expt 2 31) 1))
     *   (using 63 and 31 bits)
     */
    #define S7_LOG_LLONG_MAX 43.668274
    #define S7_LOG_LONG_MAX  21.487562
    sc->bignum_precision = DEFAULT_BIGNUM_PRECISION;
#endif

    top = sizeof(s7_int);
    s7_int32_max = (top == 8) ? S7_LONG_MAX : S7_SHORT_MAX;
    s7_int32_min = (top == 8) ? S7_LONG_MIN : S7_SHORT_MIN;
    s7_int_bits = (top == 8) ? 63 : 31;
    s7_int_digits = (top == 8) ? 18 : 8;

    s7_int_max = (top == 8) ? S7_LLONG_MAX : S7_LONG_MAX;
    s7_int_min = (top == 8) ? S7_LLONG_MIN : S7_LONG_MIN;

    s7_int_digits_by_radix[0] = 0;
    s7_int_digits_by_radix[1] = 0;

    for (i = 2; i < 17; i++)
      s7_int_digits_by_radix[i] = (int32_t)(floor(((top == 8) ? S7_LOG_LLONG_MAX : S7_LOG_LONG_MAX) / log((double)i)));

    s7_define_constant(sc, "most-positive-fixnum", make_permanent_integer_unchecked((top == 8) ? s7_int_max : ((top == 4) ? S7_LONG_MAX : S7_SHORT_MAX)));
    s7_define_constant(sc, "most-negative-fixnum", make_permanent_integer_unchecked((top == 8) ? s7_int_min : ((top == 4) ? S7_LONG_MIN : S7_SHORT_MIN)));

    if (top == 4) sc->default_rationalize_error = 1.0e-6;
    s7_define_constant(sc, "pi", real_pi);
    sc->pi_symbol = s7_make_symbol(sc, "pi");

    sc->objstr_max_len = s7_int_max;
    {
      s7_pointer p;
      new_cell(sc, p, T_RANDOM_STATE);
#if WITH_GMP
      {
	mpz_t seed;
	mpz_init_set_ui(seed, (uint32_t)time(NULL));
	gmp_randinit_default(random_gmp_state(p));
	gmp_randseed(random_gmp_state(p), seed);
	mpz_clear(seed);
      }
#else
      random_seed(p) = (uint64_t)time(NULL);
      random_carry(p) = 1675393560;
#endif
      sc->default_rng = p;
    }
    
    for (i = 0; i < 10; i++) sc->singletons[(unsigned char)'0' + i] = small_int(i);
    sc->singletons[(unsigned char)'+'] = sc->add_symbol;
    sc->singletons[(unsigned char)'-'] = sc->subtract_symbol;
    sc->singletons[(unsigned char)'*'] = sc->multiply_symbol;
    sc->singletons[(unsigned char)'/'] = sc->divide_symbol;
    sc->singletons[(unsigned char)'<'] = sc->lt_symbol;
    sc->singletons[(unsigned char)'>'] = sc->gt_symbol;
    sc->singletons[(unsigned char)'='] = sc->eq_symbol;
  }

#if WITH_GMP
  s7_gmp_init(sc);
#endif

  init_choosers(sc);
  init_typers(sc);

#if DEBUGGING
  s7_define_safe_function(sc, "local-symbol?",   g_is_local_symbol,   1, 0, false, "an experiment");
#endif

  /* -------------------------------------------------------------------------------- */
  s7_set_d_pi_function(slot_value(global_slot(sc->float_vector_ref_symbol)), float_vector_ref_d);
  s7_set_d_pid_function(slot_value(global_slot(sc->float_vector_set_symbol)), float_vector_set_d);

  s7_set_i_pi_function(slot_value(global_slot(sc->int_vector_ref_symbol)), int_vector_ref_i);
  s7_set_i_pii_function(slot_value(global_slot(sc->int_vector_set_symbol)), int_vector_set_i);
  s7_set_i_pi_function(slot_value(global_slot(sc->byte_vector_ref_symbol)), byte_vector_ref_i);
  s7_set_i_pii_function(slot_value(global_slot(sc->byte_vector_set_symbol)), byte_vector_set_i);

  s7_set_p_pi_function(slot_value(global_slot(sc->vector_ref_symbol)), vector_ref_p_pi);
  s7_set_p_pip_function(slot_value(global_slot(sc->vector_set_symbol)), vector_set_p_pip);
  s7_set_p_pi_direct_function(slot_value(global_slot(sc->vector_ref_symbol)), vector_ref_p_pi_direct);
  s7_set_p_pip_direct_function(slot_value(global_slot(sc->vector_set_symbol)), vector_set_p_pip_direct);

  s7_set_p_pi_function(slot_value(global_slot(sc->list_ref_symbol)), list_ref_p_pi);
  s7_set_p_pip_function(slot_value(global_slot(sc->list_set_symbol)), list_set_p_pip);
  s7_set_p_pi_direct_function(slot_value(global_slot(sc->list_ref_symbol)), list_ref_p_pi_direct);
  s7_set_p_pip_direct_function(slot_value(global_slot(sc->list_set_symbol)), list_set_p_pip_direct);

  s7_set_p_pp_function(slot_value(global_slot(sc->let_ref_symbol)), let_ref_p_pp);
  s7_set_p_ppp_function(slot_value(global_slot(sc->let_set_symbol)), let_set_p_ppp);

  s7_set_p_pi_function(slot_value(global_slot(sc->string_ref_symbol)), string_ref_p_pi);
  s7_set_p_pip_function(slot_value(global_slot(sc->string_set_symbol)), string_set_p_pip);
  s7_set_p_pi_direct_function(slot_value(global_slot(sc->string_ref_symbol)), string_ref_p_pi_direct);
  s7_set_p_pip_direct_function(slot_value(global_slot(sc->string_set_symbol)), string_set_p_pip_direct);

  s7_set_p_pp_function(slot_value(global_slot(sc->hash_table_ref_symbol)), hash_table_ref_p_pp);
  s7_set_p_ppp_function(slot_value(global_slot(sc->hash_table_set_symbol)), hash_table_set_p_ppp);
  s7_set_p_pp_direct_function(slot_value(global_slot(sc->hash_table_ref_symbol)), hash_table_ref_p_pp_direct);
  s7_set_p_ppp_direct_function(slot_value(global_slot(sc->hash_table_set_symbol)), hash_table_set_p_ppp_direct);

#if (!WITH_GMP)
  s7_set_p_ii_function(slot_value(global_slot(sc->complex_symbol)), complex_p_ii);
  s7_set_p_p_function(slot_value(global_slot(sc->random_symbol)), random_p_p);
#endif

  s7_set_p_p_function(slot_value(global_slot(sc->car_symbol)), car_p_p);
  s7_set_p_pp_function(slot_value(global_slot(sc->set_car_symbol)), set_car_p_pp);
  s7_set_p_p_function(slot_value(global_slot(sc->cdr_symbol)), cdr_p_p);
  s7_set_p_pp_function(slot_value(global_slot(sc->set_cdr_symbol)), set_cdr_p_pp);
  s7_set_p_p_function(slot_value(global_slot(sc->caar_symbol)), caar_p_p);
  s7_set_p_p_function(slot_value(global_slot(sc->cadr_symbol)), cadr_p_p);
  s7_set_p_p_function(slot_value(global_slot(sc->cdar_symbol)), cdar_p_p);
  s7_set_p_p_function(slot_value(global_slot(sc->cddr_symbol)), cddr_p_p);

  s7_set_p_p_function(slot_value(global_slot(sc->string_to_symbol_symbol)), string_to_symbol_p_p);
  s7_set_p_p_function(slot_value(global_slot(sc->symbol_to_string_symbol)), symbol_to_string_p);
  s7_set_p_function(slot_value(global_slot(sc->newline_symbol)), newline_p);
  s7_set_p_p_function(slot_value(global_slot(sc->newline_symbol)), newline_p_p);
  s7_set_p_p_function(slot_value(global_slot(sc->display_symbol)), display_p_p);
  s7_set_p_pp_function(slot_value(global_slot(sc->display_symbol)), display_p_pp);
  s7_set_p_p_function(slot_value(global_slot(sc->write_symbol)), write_p_p);
  s7_set_p_pp_function(slot_value(global_slot(sc->write_symbol)), write_p_pp);
  s7_set_p_p_function(slot_value(global_slot(sc->write_char_symbol)), write_char_p_p);
  s7_set_p_pp_function(slot_value(global_slot(sc->write_char_symbol)), write_char_p_pp);
  s7_set_i_p_function(slot_value(global_slot(sc->port_line_number_symbol)), port_line_number_i_p);
  s7_set_p_p_function(slot_value(global_slot(sc->number_to_string_symbol)), number_to_string_p);
  s7_set_p_p_function(slot_value(global_slot(sc->number_to_string_symbol)), number_to_string_p_p);
  s7_set_p_pp_function(slot_value(global_slot(sc->number_to_string_symbol)), number_to_string_p_pp);
  s7_set_p_pp_function(slot_value(global_slot(sc->cons_symbol)), cons_p_pp);
  s7_set_p_function(slot_value(global_slot(sc->s7_version_symbol)), s7_version_p);
  s7_set_p_function(slot_value(global_slot(sc->open_output_string_symbol)), open_output_string_p);
  s7_set_p_ppi_function(slot_value(global_slot(sc->char_position_symbol)), char_position_p_ppi);
  s7_set_p_pp_function(slot_value(global_slot(sc->append_symbol)), append_p_pp);
  s7_set_p_ppp_function(slot_value(global_slot(sc->append_symbol)), append_p_ppp);
  s7_set_p_function(slot_value(global_slot(sc->values_symbol)), values_p);
  s7_set_p_p_function(slot_value(global_slot(sc->values_symbol)), values_p_p);
  s7_set_p_pp_function(slot_value(global_slot(sc->member_symbol)), member_p_pp);
  s7_set_p_pp_function(slot_value(global_slot(sc->assoc_symbol)), assoc_p_pp);

#if (!WITH_GMP)
  s7_set_i_i_function(slot_value(global_slot(sc->abs_symbol)), abs_i_i);
  s7_set_d_d_function(slot_value(global_slot(sc->abs_symbol)), abs_d_d);
  s7_set_d_d_function(slot_value(global_slot(sc->exp_symbol)), exp_d_d);
  s7_set_d_d_function(slot_value(global_slot(sc->sin_symbol)), sin_d_d);
  s7_set_d_d_function(slot_value(global_slot(sc->cos_symbol)), cos_d_d);
  s7_set_d_d_function(slot_value(global_slot(sc->tan_symbol)), tan_d_d);
  s7_set_d_d_function(slot_value(global_slot(sc->sinh_symbol)), sinh_d_d);
  s7_set_d_d_function(slot_value(global_slot(sc->cosh_symbol)), cosh_d_d);
  s7_set_d_d_function(slot_value(global_slot(sc->tanh_symbol)), tanh_d_d);
  s7_set_d_d_function(slot_value(global_slot(sc->random_symbol)), random_d_d);
  s7_set_d_d_function(slot_value(global_slot(sc->round_symbol)), round_d_d);
  s7_set_i_i_function(slot_value(global_slot(sc->round_symbol)), round_i_i);
  s7_set_d_d_function(slot_value(global_slot(sc->floor_symbol)), floor_d_d);
  s7_set_i_i_function(slot_value(global_slot(sc->floor_symbol)), floor_i_i);
  s7_set_d_d_function(slot_value(global_slot(sc->truncate_symbol)), truncate_d_d);
  s7_set_i_i_function(slot_value(global_slot(sc->truncate_symbol)), truncate_i_i);
  s7_set_d_d_function(slot_value(global_slot(sc->ceiling_symbol)), ceiling_d_d);
  s7_set_i_i_function(slot_value(global_slot(sc->ceiling_symbol)), ceiling_i_i);
#endif

  s7_set_d_d_function(slot_value(global_slot(sc->add_symbol)), add_d_d);
  s7_set_d_d_function(slot_value(global_slot(sc->subtract_symbol)), subtract_d_d);
  s7_set_d_d_function(slot_value(global_slot(sc->multiply_symbol)), multiply_d_d);
  s7_set_d_d_function(slot_value(global_slot(sc->divide_symbol)), divide_d_d);

  s7_set_d_dd_function(slot_value(global_slot(sc->add_symbol)), add_d_dd);
  s7_set_d_dd_function(slot_value(global_slot(sc->subtract_symbol)), subtract_d_dd);
  s7_set_d_dd_function(slot_value(global_slot(sc->multiply_symbol)), multiply_d_dd);
  s7_set_d_id_function(slot_value(global_slot(sc->multiply_symbol)), multiply_d_id);
  s7_set_d_id_function(slot_value(global_slot(sc->add_symbol)), add_d_id);
  s7_set_d_id_function(slot_value(global_slot(sc->subtract_symbol)), sub_d_id);
  s7_set_d_dd_function(slot_value(global_slot(sc->divide_symbol)), divide_d_dd);
#if (!WITH_GMP)
  s7_set_d_dd_function(slot_value(global_slot(sc->atan_symbol)), atan_d_dd);
  s7_set_d_dd_function(slot_value(global_slot(sc->quotient_symbol)), quotient_d_dd);
  s7_set_d_dd_function(slot_value(global_slot(sc->remainder_symbol)), remainder_d_dd);
  s7_set_d_dd_function(slot_value(global_slot(sc->modulo_symbol)), modulo_d_dd);
#endif
  
  s7_set_d_ddd_function(slot_value(global_slot(sc->add_symbol)), add_d_ddd);
  s7_set_d_ddd_function(slot_value(global_slot(sc->subtract_symbol)), subtract_d_ddd);
  s7_set_d_ddd_function(slot_value(global_slot(sc->multiply_symbol)), multiply_d_ddd);
  s7_set_d_ddd_function(slot_value(global_slot(sc->divide_symbol)), divide_d_ddd);

  s7_set_d_dddd_function(slot_value(global_slot(sc->add_symbol)), add_d_dddd);
  s7_set_d_dddd_function(slot_value(global_slot(sc->subtract_symbol)), subtract_d_dddd);
  s7_set_d_dddd_function(slot_value(global_slot(sc->multiply_symbol)), multiply_d_dddd);
  s7_set_d_dddd_function(slot_value(global_slot(sc->divide_symbol)), divide_d_dddd);
  s7_set_p_ii_function(slot_value(global_slot(sc->divide_symbol)), divide_p_ii);

  s7_set_p_pp_function(slot_value(global_slot(sc->add_symbol)), add_p_pp);
  s7_set_p_pp_function(slot_value(global_slot(sc->subtract_symbol)), subtract_p_pp);
  s7_set_d_dd_function(slot_value(global_slot(sc->max_symbol)), max_d_dd);
  s7_set_d_dd_function(slot_value(global_slot(sc->min_symbol)), min_d_dd);
  s7_set_i_ii_function(slot_value(global_slot(sc->max_symbol)), max_i_ii);
  s7_set_i_ii_function(slot_value(global_slot(sc->min_symbol)), min_i_ii);
#if (!WITH_GMP)
  s7_set_d_p_function(slot_value(global_slot(sc->real_part_symbol)), real_part_d_p);
  s7_set_d_p_function(slot_value(global_slot(sc->imag_part_symbol)), imag_part_d_p);
  s7_set_d_p_function(slot_value(global_slot(sc->magnitude_symbol)), magnitude_d_p);
  s7_set_p_pp_function(slot_value(global_slot(sc->multiply_symbol)), multiply_p_pp);
  s7_set_p_dd_function(slot_value(global_slot(sc->multiply_symbol)), mul_p_dd);
  s7_set_p_dd_function(slot_value(global_slot(sc->add_symbol)), add_p_dd);
  s7_set_p_dd_function(slot_value(global_slot(sc->subtract_symbol)), sub_p_dd);
  s7_set_p_ii_function(slot_value(global_slot(sc->multiply_symbol)), mul_p_ii);
  s7_set_p_ii_function(slot_value(global_slot(sc->add_symbol)), add_p_ii);
  s7_set_p_ii_function(slot_value(global_slot(sc->subtract_symbol)), sub_p_ii);
  s7_set_d_p_function(slot_value(global_slot(sc->angle_symbol)), angle_d_p);
  s7_set_i_d_function(slot_value(global_slot(sc->round_symbol)), round_i_d);
  s7_set_i_d_function(slot_value(global_slot(sc->floor_symbol)), floor_i_d);
  s7_set_i_p_function(slot_value(global_slot(sc->floor_symbol)), floor_i_p);
  s7_set_i_d_function(slot_value(global_slot(sc->truncate_symbol)), truncate_i_d);
  s7_set_i_d_function(slot_value(global_slot(sc->ceiling_symbol)), ceiling_i_d);
  s7_set_i_i_function(slot_value(global_slot(sc->random_symbol)), random_i_i);
  s7_set_i_ii_function(slot_value(global_slot(sc->quotient_symbol)), quotient_i_ii);
  s7_set_i_ii_function(slot_value(global_slot(sc->remainder_symbol)), remainder_i_ii);
  s7_set_i_ii_function(slot_value(global_slot(sc->modulo_symbol)), modulo_i_ii);
#endif
  s7_set_i_i_function(slot_value(global_slot(sc->subtract_symbol)), subtract_i_i);
  s7_set_i_i_function(slot_value(global_slot(sc->lognot_symbol)), lognot_i_i);

  s7_set_i_ii_function(slot_value(global_slot(sc->add_symbol)), add_i_ii);
  s7_set_i_iii_function(slot_value(global_slot(sc->add_symbol)), add_i_iii);
  s7_set_i_ii_function(slot_value(global_slot(sc->subtract_symbol)), subtract_i_ii);
  s7_set_i_iii_function(slot_value(global_slot(sc->subtract_symbol)), subtract_i_iii);
  s7_set_i_ii_function(slot_value(global_slot(sc->multiply_symbol)), multiply_i_ii);
  s7_set_i_iii_function(slot_value(global_slot(sc->multiply_symbol)), multiply_i_iii);
  s7_set_i_ii_function(slot_value(global_slot(sc->ash_symbol)), ash_i_ii);
  s7_set_i_ii_function(slot_value(global_slot(sc->logior_symbol)), logior_i_ii);
  s7_set_i_ii_function(slot_value(global_slot(sc->logxor_symbol)), logxor_i_ii);
  s7_set_i_ii_function(slot_value(global_slot(sc->logand_symbol)), logand_i_ii);

#if (!WITH_PURE_S7)
  s7_set_p_pp_function(slot_value(global_slot(sc->vector_append_symbol)), vector_append_p_pp);
  s7_set_p_ppp_function(slot_value(global_slot(sc->vector_append_symbol)), vector_append_p_ppp);
  s7_set_i_i_function(slot_value(global_slot(sc->integer_length_symbol)), integer_length_i_i);
  s7_set_i_p_function(slot_value(global_slot(sc->string_length_symbol)), string_length_i);
  s7_set_i_p_function(slot_value(global_slot(sc->vector_length_symbol)), vector_length_i);
  s7_set_p_p_function(slot_value(global_slot(sc->vector_to_list_symbol)), vector_to_list_p_p);
  s7_set_b_p_function(slot_value(global_slot(sc->is_exact_symbol)), s7_is_rational);
  s7_set_b_p_function(slot_value(global_slot(sc->is_char_ready_symbol)), is_char_ready_b_p);
  s7_set_b_p_function(slot_value(global_slot(sc->is_exact_symbol)), is_exact_b);
  s7_set_b_p_function(slot_value(global_slot(sc->is_inexact_symbol)), is_inexact_b);
#endif
  s7_set_i_p_function(slot_value(global_slot(sc->numerator_symbol)), numerator_i);
  s7_set_i_p_function(slot_value(global_slot(sc->denominator_symbol)), denominator_i);
  s7_set_i_p_function(slot_value(global_slot(sc->char_to_integer_symbol)), char_to_integer_i);
  s7_set_i_p_function(slot_value(global_slot(sc->hash_table_entries_symbol)), hash_table_entries_i);
  s7_set_i_p_function(slot_value(global_slot(s7_make_symbol(sc, "tree-leaves"))), tree_leaves_i);

  s7_set_b_p_function(slot_value(global_slot(sc->is_boolean_symbol)), s7_is_boolean);
  s7_set_b_p_function(slot_value(global_slot(sc->is_byte_vector_symbol)), s7_is_byte_vector);
  s7_set_b_p_function(slot_value(global_slot(sc->is_c_object_symbol)), s7_is_c_object);
  s7_set_b_p_function(slot_value(global_slot(sc->is_char_symbol)), s7_is_character);
  s7_set_b_p_function(slot_value(global_slot(sc->is_complex_symbol)), s7_is_complex);
  s7_set_b_p_function(slot_value(global_slot(sc->is_constant_symbol)), s7_is_constant);
  s7_set_b_p_function(slot_value(global_slot(sc->is_continuation_symbol)), s7_is_continuation);
  s7_set_b_p_function(slot_value(global_slot(sc->is_c_pointer_symbol)), s7_is_c_pointer);
  s7_set_b_p_function(slot_value(global_slot(sc->is_dilambda_symbol)), s7_is_dilambda);
  s7_set_b_p_function(slot_value(global_slot(sc->is_eof_object_symbol)), s7_is_eof_object);
  s7_set_b_p_function(slot_value(global_slot(sc->is_even_symbol)), is_even_b);
  s7_set_b_p_function(slot_value(global_slot(sc->is_float_symbol)), is_float_b);
  s7_set_b_p_function(slot_value(global_slot(sc->is_float_vector_symbol)), s7_is_float_vector);
  s7_set_b_p_function(slot_value(global_slot(sc->is_gensym_symbol)), s7_is_gensym);
  s7_set_b_p_function(slot_value(global_slot(sc->is_hash_table_symbol)), s7_is_hash_table);
  s7_set_b_p_function(slot_value(global_slot(sc->is_infinite_symbol)), is_infinite_b);
  s7_set_b_p_function(slot_value(global_slot(sc->is_input_port_symbol)), is_input_port_b);
  s7_set_b_p_function(slot_value(global_slot(sc->is_integer_symbol)), s7_is_integer);
  s7_set_b_p_function(slot_value(global_slot(sc->is_int_vector_symbol)), s7_is_int_vector);
  s7_set_b_p_function(slot_value(global_slot(sc->is_keyword_symbol)), s7_is_keyword);
  s7_set_b_p_function(slot_value(global_slot(sc->is_let_symbol)), s7_is_let);
  s7_set_b_p_function(slot_value(global_slot(sc->is_list_symbol)), is_list_b);
  s7_set_b_p_function(slot_value(global_slot(sc->is_macro_symbol)), is_macro_b);
  s7_set_b_p_function(slot_value(global_slot(sc->is_nan_symbol)), is_nan_b);
  s7_set_b_p_function(slot_value(global_slot(sc->is_null_symbol)), is_null_b);
  s7_set_b_p_function(slot_value(global_slot(sc->is_number_symbol)), s7_is_number);
  s7_set_b_p_function(slot_value(global_slot(sc->is_odd_symbol)), is_odd_b);
  s7_set_b_p_function(slot_value(global_slot(sc->is_output_port_symbol)), is_output_port_b);
  s7_set_b_p_function(slot_value(global_slot(sc->is_pair_symbol)), s7_is_pair);
  s7_set_b_p_function(slot_value(global_slot(sc->is_port_closed_symbol)), is_port_closed_b);
  s7_set_b_p_function(slot_value(global_slot(sc->is_procedure_symbol)), s7_is_procedure);
  s7_set_b_p_function(slot_value(global_slot(sc->is_proper_list_symbol)), is_proper_list_b);
  s7_set_b_p_function(slot_value(global_slot(sc->is_random_state_symbol)), is_random_state_b);
  s7_set_b_p_function(slot_value(global_slot(sc->is_rational_symbol)), s7_is_rational);
  s7_set_b_p_function(slot_value(global_slot(sc->is_real_symbol)), s7_is_real);
  s7_set_b_p_function(slot_value(global_slot(sc->is_sequence_symbol)), is_sequence_b);
  s7_set_b_p_function(slot_value(global_slot(sc->is_string_symbol)), s7_is_string);
  s7_set_b_p_function(slot_value(global_slot(sc->is_symbol_symbol)), s7_is_symbol);
  s7_set_b_p_function(slot_value(global_slot(sc->is_syntax_symbol)), s7_is_syntax);
  s7_set_b_p_function(slot_value(global_slot(sc->is_vector_symbol)), s7_is_vector);
  s7_set_b_p_function(slot_value(global_slot(sc->is_iterator_symbol)), is_iterator_b);
  s7_set_b_p_function(slot_value(global_slot(sc->is_char_alphabetic_symbol)), is_char_alphabetic_b);
  s7_set_b_p_function(slot_value(global_slot(sc->is_char_lower_case_symbol)), is_char_lower_case_b);
  s7_set_b_p_function(slot_value(global_slot(sc->is_char_numeric_symbol)), is_char_numeric_b);
  s7_set_b_p_function(slot_value(global_slot(sc->is_char_upper_case_symbol)), is_char_upper_case_b);
  s7_set_b_p_function(slot_value(global_slot(sc->is_char_whitespace_symbol)), is_char_whitespace_b);
  s7_set_b_p_function(slot_value(global_slot(sc->is_openlet_symbol)), s7_is_openlet);
  s7_set_b_p_function(slot_value(global_slot(sc->iterator_is_at_end_symbol)), iterator_is_at_end_b);
  s7_set_b_p_function(slot_value(global_slot(sc->is_zero_symbol)), is_zero_b);
  s7_set_b_p_function(slot_value(global_slot(sc->is_negative_symbol)), is_negative_b);
  s7_set_b_p_function(slot_value(global_slot(sc->is_positive_symbol)), is_positive_b);
  s7_set_b_p_function(slot_value(global_slot(sc->not_symbol)), not_b);
  s7_set_b_p_function(slot_value(global_slot(sc->is_provided_symbol)), is_provided_b);
  s7_set_b_p_function(slot_value(global_slot(sc->is_defined_symbol)), is_defined_b_p);
  s7_set_b_pp_function(slot_value(global_slot(sc->is_defined_symbol)), is_defined_b_pp);
  s7_set_b_pp_function(slot_value(global_slot(s7_make_symbol(sc, "tree-memq"))), tree_memq_b_pp);
  s7_set_b_pp_function(slot_value(global_slot(s7_make_symbol(sc, "tree-set-memq"))), tree_set_memq_b_pp);

  s7_set_p_p_function(slot_value(global_slot(sc->is_pair_symbol)), is_pair_p_p);
  s7_set_p_p_function(slot_value(global_slot(sc->integer_to_char_symbol)), integer_to_char_p_p);

#if WITH_SYSTEM_EXTRAS
  s7_set_b_p_function(slot_value(global_slot(sc->is_directory_symbol)), is_directory_b);
  s7_set_b_p_function(slot_value(global_slot(sc->file_exists_symbol)), file_exists_b);
#endif

  s7_set_b_i_function(slot_value(global_slot(sc->is_even_symbol)), is_even_i);
  s7_set_b_i_function(slot_value(global_slot(sc->is_odd_symbol)), is_odd_i);
  s7_set_b_i_function(slot_value(global_slot(sc->is_zero_symbol)), is_zero_i);
  s7_set_b_d_function(slot_value(global_slot(sc->is_zero_symbol)), is_zero_d);
  s7_set_b_i_function(slot_value(global_slot(sc->is_positive_symbol)), is_positive_i);
  s7_set_b_d_function(slot_value(global_slot(sc->is_positive_symbol)), is_positive_d);
  s7_set_b_i_function(slot_value(global_slot(sc->is_negative_symbol)), is_negative_i);
  s7_set_b_d_function(slot_value(global_slot(sc->is_negative_symbol)), is_negative_d);

  s7_set_b_ii_function(slot_value(global_slot(sc->logbit_symbol)), logbit_b_ii);
  s7_set_b_ii_function(slot_value(global_slot(sc->eq_symbol)), req_b_ii);
  s7_set_b_ii_function(slot_value(global_slot(sc->lt_symbol)), lt_b_ii);
  s7_set_b_ii_function(slot_value(global_slot(sc->leq_symbol)), leq_b_ii);
  s7_set_b_ii_function(slot_value(global_slot(sc->gt_symbol)), gt_b_ii);
  s7_set_b_ii_function(slot_value(global_slot(sc->geq_symbol)), geq_b_ii);
  s7_set_b_dd_function(slot_value(global_slot(sc->eq_symbol)), req_b_dd);
  s7_set_b_dd_function(slot_value(global_slot(sc->lt_symbol)), lt_b_dd);
  s7_set_b_dd_function(slot_value(global_slot(sc->leq_symbol)), leq_b_dd);
  s7_set_b_dd_function(slot_value(global_slot(sc->gt_symbol)), gt_b_dd);
  s7_set_b_dd_function(slot_value(global_slot(sc->geq_symbol)), geq_b_dd);

#if (!WITH_GMP)
  s7_set_b_pp_function(slot_value(global_slot(sc->logbit_symbol)), logbit_b_pp);
  s7_set_p_pi_function(slot_value(global_slot(sc->eq_symbol)), equal_p_pi);
  s7_set_p_pp_function(slot_value(global_slot(sc->eq_symbol)), equal_p_pp);
  s7_set_p_ii_function(slot_value(global_slot(sc->eq_symbol)), equal_p_ii);
  s7_set_p_dd_function(slot_value(global_slot(sc->eq_symbol)), equal_p_dd);

  s7_set_p_pp_function(slot_value(global_slot(sc->lt_symbol)), lt_p_pp);
  s7_set_p_pi_function(slot_value(global_slot(sc->lt_symbol)), lt_p_pi);
  s7_set_p_ii_function(slot_value(global_slot(sc->lt_symbol)), lt_p_ii);
  s7_set_p_dd_function(slot_value(global_slot(sc->lt_symbol)), lt_p_dd);

  s7_set_p_pp_function(slot_value(global_slot(sc->leq_symbol)), leq_p_pp);
  s7_set_p_pi_function(slot_value(global_slot(sc->leq_symbol)), leq_p_pi);
  s7_set_p_dd_function(slot_value(global_slot(sc->leq_symbol)), leq_p_dd);
  s7_set_p_ii_function(slot_value(global_slot(sc->leq_symbol)), leq_p_ii);

  s7_set_p_pp_function(slot_value(global_slot(sc->gt_symbol)), gt_p_pp);
  s7_set_p_pi_function(slot_value(global_slot(sc->gt_symbol)), gt_p_pi);
  s7_set_p_dd_function(slot_value(global_slot(sc->gt_symbol)), gt_p_dd);
  s7_set_p_ii_function(slot_value(global_slot(sc->gt_symbol)), gt_p_ii);

  s7_set_p_pp_function(slot_value(global_slot(sc->geq_symbol)), geq_p_pp);
  s7_set_p_pi_function(slot_value(global_slot(sc->geq_symbol)), geq_p_pi);
  s7_set_p_ii_function(slot_value(global_slot(sc->geq_symbol)), geq_p_ii);
  s7_set_p_dd_function(slot_value(global_slot(sc->geq_symbol)), geq_p_dd);

  s7_set_b_pp_function(slot_value(global_slot(sc->eq_symbol)), req_b_pp);
  s7_set_b_pp_function(slot_value(global_slot(sc->lt_symbol)), lt_b_pp);
  s7_set_b_pp_function(slot_value(global_slot(sc->leq_symbol)), leq_b_pp);
  s7_set_b_pp_function(slot_value(global_slot(sc->gt_symbol)), gt_b_pp);
  s7_set_b_pp_function(slot_value(global_slot(sc->geq_symbol)), geq_b_pp);

  s7_set_b_pi_function(slot_value(global_slot(sc->eq_symbol)), req_b_pi);
  s7_set_b_pi_function(slot_value(global_slot(sc->lt_symbol)), lt_b_pi);
  s7_set_b_pi_function(slot_value(global_slot(sc->leq_symbol)), leq_b_pi);
  s7_set_b_pi_function(slot_value(global_slot(sc->gt_symbol)), gt_b_pi);
  s7_set_b_pi_function(slot_value(global_slot(sc->geq_symbol)), geq_b_pi);
#endif

  s7_set_b_pp_function(slot_value(global_slot(sc->is_eq_symbol)), s7_is_eq);
  s7_set_b_pp_function(slot_value(global_slot(sc->is_eqv_symbol)), s7_is_eqv);
  s7_set_b_pp_function(slot_value(global_slot(sc->is_equal_symbol)), is_equal_b_pp);
  s7_set_b_pp_function(slot_value(global_slot(sc->is_morally_equal_symbol)), is_morally_equal_b_pp);
  s7_set_p_pp_function(slot_value(global_slot(sc->is_equal_symbol)), is_equal_p_pp);
  s7_set_p_pp_function(slot_value(global_slot(sc->is_morally_equal_symbol)), is_morally_equal_p_pp);
  s7_set_b_pp_function(slot_value(global_slot(sc->char_lt_symbol)), char_lt_b);
  s7_set_b_pp_function(slot_value(global_slot(sc->char_leq_symbol)), char_leq_b);
  s7_set_b_pp_function(slot_value(global_slot(sc->char_gt_symbol)), char_gt_b);
  s7_set_b_pp_function(slot_value(global_slot(sc->char_geq_symbol)), char_geq_b);
  s7_set_b_pp_function(slot_value(global_slot(sc->char_eq_symbol)), char_eq_b);
  s7_set_b_pp_function(slot_value(global_slot(sc->string_lt_symbol)), string_lt_b);
  s7_set_b_pp_function(slot_value(global_slot(sc->string_leq_symbol)), string_leq_b);
  s7_set_b_pp_function(slot_value(global_slot(sc->string_gt_symbol)), string_gt_b);
  s7_set_b_pp_function(slot_value(global_slot(sc->string_geq_symbol)), string_geq_b);
  s7_set_b_pp_function(slot_value(global_slot(sc->string_eq_symbol)), string_eq_b);

  s7_set_b_pp_direct_function(slot_value(global_slot(sc->char_lt_symbol)), char_lt_b_direct);
  s7_set_b_pp_direct_function(slot_value(global_slot(sc->char_leq_symbol)), char_leq_b_direct);
  s7_set_b_pp_direct_function(slot_value(global_slot(sc->char_gt_symbol)), char_gt_b_direct);
  s7_set_b_pp_direct_function(slot_value(global_slot(sc->char_geq_symbol)), char_geq_b_direct);
  s7_set_b_pp_direct_function(slot_value(global_slot(sc->char_eq_symbol)), char_eq_b_direct);
  s7_set_b_pp_direct_function(slot_value(global_slot(sc->string_lt_symbol)), string_lt_b_direct);
  s7_set_b_pp_direct_function(slot_value(global_slot(sc->string_leq_symbol)), string_leq_b_direct);
  s7_set_b_pp_direct_function(slot_value(global_slot(sc->string_gt_symbol)), string_gt_b_direct);
  s7_set_b_pp_direct_function(slot_value(global_slot(sc->string_geq_symbol)), string_geq_b_direct);
  s7_set_b_pp_direct_function(slot_value(global_slot(sc->string_eq_symbol)), string_eq_b_direct);

  s7_set_b_pp_function(slot_value(global_slot(sc->is_aritable_symbol)), is_aritable_b_pp);
#if (!WITH_PURE_S7)
  s7_set_b_pp_function(slot_value(global_slot(sc->char_ci_lt_symbol)), char_ci_lt_b);
  s7_set_b_pp_function(slot_value(global_slot(sc->char_ci_leq_symbol)), char_ci_leq_b);
  s7_set_b_pp_function(slot_value(global_slot(sc->char_ci_gt_symbol)), char_ci_gt_b);
  s7_set_b_pp_function(slot_value(global_slot(sc->char_ci_geq_symbol)), char_ci_geq_b);
  s7_set_b_pp_function(slot_value(global_slot(sc->char_ci_eq_symbol)), char_ci_eq_b);
  s7_set_b_pp_function(slot_value(global_slot(sc->string_ci_lt_symbol)), string_ci_lt_b);
  s7_set_b_pp_function(slot_value(global_slot(sc->string_ci_leq_symbol)), string_ci_leq_b);
  s7_set_b_pp_function(slot_value(global_slot(sc->string_ci_gt_symbol)), string_ci_gt_b);
  s7_set_b_pp_function(slot_value(global_slot(sc->string_ci_geq_symbol)), string_ci_geq_b);
  s7_set_b_pp_function(slot_value(global_slot(sc->string_ci_eq_symbol)), string_ci_eq_b);

  s7_set_b_pp_direct_function(slot_value(global_slot(sc->char_ci_lt_symbol)), char_ci_lt_b_direct);
  s7_set_b_pp_direct_function(slot_value(global_slot(sc->char_ci_leq_symbol)), char_ci_leq_b_direct);
  s7_set_b_pp_direct_function(slot_value(global_slot(sc->char_ci_gt_symbol)), char_ci_gt_b_direct);
  s7_set_b_pp_direct_function(slot_value(global_slot(sc->char_ci_geq_symbol)), char_ci_geq_b_direct);
  s7_set_b_pp_direct_function(slot_value(global_slot(sc->char_ci_eq_symbol)), char_ci_eq_b_direct);
  s7_set_b_pp_direct_function(slot_value(global_slot(sc->string_ci_lt_symbol)), string_ci_lt_b_direct);
  s7_set_b_pp_direct_function(slot_value(global_slot(sc->string_ci_leq_symbol)), string_ci_leq_b_direct);
  s7_set_b_pp_direct_function(slot_value(global_slot(sc->string_ci_gt_symbol)), string_ci_gt_b_direct);
  s7_set_b_pp_direct_function(slot_value(global_slot(sc->string_ci_geq_symbol)), string_ci_geq_b_direct);
  s7_set_b_pp_direct_function(slot_value(global_slot(sc->string_ci_eq_symbol)), string_ci_eq_b_direct);
#endif


  /* -------------------------------------------------------------------------------- */
  s7_define_macro(sc, "quasiquote", g_quasiquote, 1, 0, false, H_quasiquote);

#if (!WITH_PURE_S7)
  s7_eval_c_string(sc, "(define-macro (defmacro name args . body) (cons 'define-macro (cons (cons name args) body)))");
  s7_eval_c_string(sc, "(define-macro (defmacro* name args . body) (cons 'define-macro* (cons (cons name args) body)))");

  s7_eval_c_string(sc, "(define-macro (call-with-values producer consumer) (list consumer (list producer)))");   
  /* (call-with-values (lambda () (values 1 2 3)) +) */

  s7_eval_c_string(sc, "(define-macro (multiple-value-bind vars expression . body)                            \n\
                          (list (cons 'lambda (cons vars body)) expression))");

  s7_eval_c_string(sc, "(define-macro (cond-expand . clauses)                                                 \n\
                          (letrec ((traverse (lambda (tree)                                                   \n\
		                               (if (pair? tree)                                               \n\
			                            (cons (traverse (car tree))                               \n\
				                          (case (cdr tree) ((())) (else => traverse)))        \n\
			                            (if (memq tree '(and or not else)) tree                   \n\
			                                (and (symbol? tree) (provided? tree)))))))            \n\
                            (cons 'cond (map (lambda (clause)                                                 \n\
		                               (cons (traverse (car clause))                                  \n\
			                             (case (cdr clause) ((()) '(#f)) (else))))                \n\
		                             clauses))))");
#endif

  s7_eval_c_string(sc, "(define-expansion (reader-cond . clauses)                                             \n\
                          (call-with-exit                                                                     \n\
                            (lambda (return)                                                                  \n\
                              (for-each                                                                       \n\
                                (lambda (clause)                                                              \n\
	                          (let ((val (eval (car clause))))                                            \n\
                                    (when val                                                                 \n\
                                      (return (if (null? (cdr clause))                                        \n\
                                                  val                                                         \n\
                                                  (if (eq? (cadr clause) '=>)                                 \n\
                                                      ((eval (caddr clause)) val)                             \n\
	                                              (if (null? (cddr clause))                               \n\
                                                          (cadr clause)                                       \n\
                                                          (apply values (map quote (cdr clause))))))))))      \n\
                                clauses)                                                                      \n\
                              (values))))"); /* this is not redundant */

  s7_eval_c_string(sc, "(define make-hook                                                                     \n\
                          (let ((signature '(procedure?))                                                     \n\
                                (documentation \"(make-hook . pars) returns a new hook (a function) that passes the parameters to its function list.\")) \n\
                            (lambda hook-args                                                                 \n\
                              (let ((body ()))                                                                \n\
                                (apply lambda* hook-args                                                      \n\
                                  (copy '(let ((result #<unspecified>))                                       \n\
                                           (let ((hook (curlet)))                                             \n\
                                             (for-each (lambda (hook-function) (hook-function hook)) body)    \n\
                                             result))                                                         \n\
                                        :readable)                                                            \n\
                                  ())))))");

  s7_eval_c_string(sc, "(define hook-functions                                                                \n\
                          (let ((signature '(list? procedure?))                                               \n\
                                (documentation \"(hook-functions hook) gets or sets the list of functions associated with the hook\")) \n\
                            (dilambda                                                                         \n\
                              (lambda (hook)                                                                  \n\
                                ((funclet hook) 'body))                                                       \n\
                              (lambda (hook lst)                                                              \n\
                                (if (do ((p lst (cdr p)))                                                     \n\
                                        ((not (and (pair? p)                                                  \n\
                                                   (procedure? (car p))                                       \n\
                                                   (aritable? (car p) 1)))                                    \n\
                                         (null? p)))                                                          \n\
                                    (set! ((funclet hook) 'body) lst)                                         \n\
                                    (error 'wrong-type-arg \"hook-functions must be a list of functions, each accepting one argument: ~S\" lst))))))");


  /* -------- *unbound-variable-hook* -------- */
  sc->unbound_variable_hook = s7_eval_c_string(sc, "(make-hook 'variable)");
  s7_define_constant_with_documentation(sc, "*unbound-variable-hook*", sc->unbound_variable_hook, 
					"*unbound-variable-hook* functions are called when an unbound variable is encountered, passed (hook 'variable).");

  /* -------- *missing-close-paren-hook* -------- */
  sc->missing_close_paren_hook = s7_eval_c_string(sc, "(make-hook)");
  s7_define_constant_with_documentation(sc, "*missing-close-paren-hook*", sc->missing_close_paren_hook, 
					"*missing-close-paren-hook* functions are called when the reader thinks a close paren is missing");

  /* -------- *load-hook* -------- */
  sc->load_hook = s7_eval_c_string(sc, "(make-hook 'name)");
  s7_define_constant_with_documentation(sc, "*load-hook*", sc->load_hook, 
					"*load-hook* functions are invoked by load, passing the to-be-loaded filename as (hook 'name)");

  /* -------- *error-hook* -------- */
  sc->error_hook = s7_eval_c_string(sc, "(make-hook 'type 'data)");
  s7_define_constant_with_documentation(sc, "*error-hook*", sc->error_hook, 
					"*error-hook* functions are called in the error handler, passed (hook 'type) and (hook 'data).");

  /* -------- *read-error-hook* -------- */
  sc->read_error_hook = s7_eval_c_string(sc, "(make-hook 'type 'data)");
  s7_define_constant_with_documentation(sc, "*read-error-hook*", sc->read_error_hook, 
					"*read-error-hook* functions are called by the reader if it is unhappy, passing the current program string as (hook 'data).");

  /* -------- *rootlet-redefinition-hook* -------- */
  sc->rootlet_redefinition_hook = s7_eval_c_string(sc, "(make-hook 'name 'value)");
  s7_define_constant_with_documentation(sc, "*rootlet-redefinition-hook*", sc->rootlet_redefinition_hook, 
					"*rootlet-redefinition-hook* functions are called when a top-level variable's value is changed, (hook 'name 'value).");
  /* first parameter was originally 'symbol, but that collides with the built-in symbol function */

  s7_define_constant(sc, "*s7*",
    s7_openlet(sc, s7_inlet(sc,
       s7_list(sc, 2,
	       s7_cons(sc, sc->let_ref_fallback_symbol, s7_make_function(sc, "s7-let-ref", g_s7_let_ref_fallback, 2, 0, false, "*s7* reader")),
	       s7_cons(sc, sc->let_set_fallback_symbol, s7_make_function(sc, "s7-let-set", g_s7_let_set_fallback, 3, 0, false, "*s7* writer"))))));


#if (!DISABLE_DEPRECATED)
  s7_eval_c_string(sc, "(begin                                                    \n\
                          (define global-environment         rootlet)             \n\
                          (define current-environment        curlet)              \n\
                          (define make-procedure-with-setter dilambda)            \n\
                          (define procedure-with-setter?     dilambda?)           \n\
                          (define make-random-state          random-state)        \n\
                          (define make-complex               complex)             \n\
                          (define make-keyword               string->keyword)     \n\
                          (define (procedure-arity obj) (let ((c (arity obj))) (list (car c) (- (cdr c) (car c)) (> (cdr c) 100000)))))");
#endif
#if DEBUGGING
  if (strcmp(opt_names[HOP_SAFE_C_AAP], "h_safe_c_aap") != 0)
    fprintf(stderr, "opt_name: %s\n", opt_names[HOP_SAFE_C_AAP]);
  if (strcmp(op_names[OP_SET_WITH_LET_2], "set_with_let_2") != 0)
    fprintf(stderr, "op_name: %s\n", op_names[OP_SET_WITH_LET_2]);
#endif

  /* fprintf(stderr, "size: %d, max op: %d, opt: %d\n", (int)sizeof(s7_cell), OP_MAX_DEFINED, OPT_MAX_DEFINED); */
  /* 64 bit machine: size: 56 [size 80 if gmp, 136 if debugging], op: 425, opt: 442 */

  if (sizeof(void *) > sizeof(s7_int))
    fprintf(stderr, "s7_int is too small: it has %d bytes, but void* has %d\n", (int)sizeof(s7_int), (int)sizeof(void *));

  save_unlet(sc);
  init_s7_let(sc);          /* set up *s7* */
  already_inited = true;

  return(sc);
}


/* -------------------------------- repl -------------------------------- */

#ifndef USE_SND
  #define USE_SND 0
#endif
#ifndef WITH_MAIN
  #define WITH_MAIN 0
#endif

#if (WITH_MAIN && (!USE_SND))

int main(int argc, char **argv)
{
  s7_scheme *sc;

  sc = s7_init();
  if (argc == 2)
    {
      fprintf(stderr, "load %s\n", argv[1]);
      s7_load(sc, argv[1]);
    }
  else 
    {
#if (!MS_WINDOWS)
      s7_load(sc, "repl.scm");              /* this is libc dependent */
      s7_eval_c_string(sc, "((*repl* 'run))");
#else
      while (1)                             /* a minimal repl -- taken from s7.html */
	{
	  char buffer[512];
	  char response[1024];
	  fprintf(stdout, "\n> ");
	  fgets(buffer, 512, stdin);
	  if ((buffer[0] != '\n') || (strlen(buffer) > 1))
	    { 
	      sprintf(response, "(write %s)", buffer);
	      s7_eval_c_string(sc, response); 
	    }
	}
#endif
    }
  return(0);
}

/* in Linux:  gcc s7.c -o repl -DWITH_MAIN -I. -g3 -ldl -lm -Wl,-export-dynamic
 * in *BSD:   gcc s7.c -o repl -DWITH_MAIN -I. -g3 -lm -Wl,-export-dynamic
 * in OSX:    gcc s7.c -o repl -DWITH_MAIN -I. -g3 -lm
 *   (clang also needs LDFLAGS="-Wl,-export-dynamic" in Linux)
 */
#endif


/* --------------------------------------------------------------------
 *
 * new snd version: snd.h configure.ac HISTORY.Snd NEWS barchive, /usr/ccrma/web/html/software/snd/index.html
 *
 * s7:
 * if profile, use line/file num to get at hashed count? and use that to annotate pp output via [count]-symbol pre-rewrite
 *   (profile-count file line)?
 *
 * gtk_box_pack* has changed -- many uses!
 * gtk4: no draw signal -- need to set the draw func
 * musglyphs gtk version is broken (probably cairo_t confusion -- make/free-cairo are obsolete for example)
 *   the problem is less obvious:
 *     "The window 0x5555564dab00 already has a drawing context. You cannot call gdk_window_begin_draw_frame() without calling gdk_window_end_draw_frame() first."
 *   and the stupid thing segfaults.  This is called in make-cairo, end in free-cairo. 
 *
 * Snd:
 * dac loop [need start/end of loop in dac_info, reader goes to start when end reached (requires rebuffering)
 *   looper does not stop/restart -- just keep going]
 *   play_selection_1 could put ends somewhere, set ends to NO_END_SPECIFIED, dac_loop_sample can
 *   use begs/other-ends to get loop points, so free_dac_info does not need to restart the loop(?)
 *   If start/end selection changed while playing, are these loop points updated?
 *
 * gtk gl: I can't see how to switch gl in and out as in the motif version -- I guess I need both gl_area and drawing_area
 * the old mus-audio-* code needs to use play or something, especially bess*
 * snd+gtk+script->eps fails??  Also why not make a graph in the no-gui case? t415.scm.
 * remove as many edpos args as possible, and num+bool->num
 * snd namespaces: dac, edits, fft, gxcolormaps, mix, region, snd.  for snd-mix, tie-ins are in place
 *
 * libgtk:
 *   callback funcs need calling check -- 5 list as fields of c-pointer?
 *   several more special funcs
 *
 * test opt_sizes escape in sort et al -- perhaps save sc->envir, make sure it is ok if optimize fails
 * opt let? opt_float_begin in s7_float_optimize for map-channel in snd?
 * check glob/libc.scm in openbsd -- some problem loading libc_s7.so (it works in snd, not in repl?)
 * libc needs many type checks
 * is_type replacing is_symbol etc [all_x_is_*_s if_is_* safe_is_*]
 *   is_type_car|cdr|a in all 3 cases
 *   need symbol->type-checker-recog->type -- symbol_type: object.sym.type
 * maybe pass \u... through in read_constant_string unchanged, or read in s7??  no worse than \x..;
 *
 * immutable sequence as bit 25? == elements can't be set, immutable let=no slot added/deleted, so values changed
 *   then immutable-let access->offsets (type?), auto-copy-on-write? or (constant-copied ...)?
 *   currently the define-constant cases are inconsistent
 *   slot-setter, symbol-setter, let-setter?
 *   (define fnc (let ((setter...)) (lambda ...))) (define lt (let ((setter (lambda..))) (inlet...)))? -- and all the others like documentation/tester
 *     but where is this outlet saved esp. if vector? for let, a new union.
 *     vect get/set should be s7_function (multi-dim indices for example)
 *     string a new union as long as not associated with symbol, hash-table has extensible dproc
 *     pair can use opt fields, c_pointer has room already, immutable slot can use expr/pending_value
 *     so no problem at the s7_cell level, how to handle (eg) (let ((lt (let ((setter (lambda...))) (proc-creates-let...))))...)
 *       but how to access these?
 *       (define str (let ((documentation "a string")) "asdf")) -- (documentation str) is not in curlet
 *     and how to recognize immutable values/symbols/bindings etc?
 *   (define x (let ((documentation "help for x")) (* pi 3))) -- there is room for this but it's currently the number print name
 *   add constant = set completely-immutable-value bit, but since callable anytime, every seq-set always has to check it
 *   literator for lambda?
 *   c_object type table entries should also be s7_function, reported by object->let perhaps
 *     wrappers in the meantime?
 *
 * symbol 8-bits->cycling number, let tracks range? inserts ordered? [see above]
 *   could opt recognize large heavily-used lets and use this?
 *
 * ex lint for specific ques: turn off lint-format, seek all calls of f global|local|with a specific arg etc
 *   (requires lint on set of files, then specialize report-usage)
 *   does this give who-calls?
 *   or change to new call, reporting changes etc
 *
 * ----------------------------------------------------------------------------
 *
 *           12  |  13  |  14  |  15  ||  16  | 17.4  17.5  17.6  17.7  17.8
 * tmac          |      |      |      || 9052 |  615   259   261   261
 * index    44.3 | 3291 | 1725 | 1276 || 1255 | 1158  1111  1058  1053
 * tref          |      |      | 2372 || 2125 | 1375  1231  1125  1109
 * tauto     265 |   89 |  9   |  8.4 || 2993 | 3255  3254  1772  1378
 * teq           |      |      | 6612 || 2777 | 2129  1978  1988  1921
 * s7test   1721 | 1358 |  995 | 1194 || 2926 | 2645  2356  2215  2172
 * tlet     5318 | 3701 | 3712 | 3700 || 4006 | 3616  2527  2436  2436
 * lint          |      |      |      || 4041 | 3376  3114  3003  2726
 * lg            |      |      |      || 211  | 161   149   144   134.9
 * tform         |      |      | 6816 || 3714 | 3530  3361  3295  2746
 * tcopy         |      |      | 13.6 || 3183 | 3404  3229  3092  3071
 * tmap          |      |      |  9.3 || 5279 |       3939  3387  3386
 * tfft          |      | 15.5 | 16.4 || 17.3 | 4901  4008  3963  3964
 * tsort         |      |      |      || 8584 | 4869  4080  4010  4012
 * titer         |      |      |      || 5971 | 5224  4768  4707  4562
 * bench         |      |      |      || 7012 | 6378  6327  5934  5106
 * thash         |      |      | 50.7 || 8778 | 8488  8057  7550  7537
 * tgen          |   71 | 70.6 | 38.0 || 12.6 | 12.4  12.6  11.7  11.9
 * tall       90 |   43 | 14.5 | 12.7 || 17.9 | 20.4  18.6  17.7  17.8
 * calls     359 |  275 | 54   | 34.7 || 43.7 | 42.5  41.1  39.7  39.4
 *                                    || 145  | 135   132   93.2  89.4
 * 
 * ----------------------------------------------------------------------------
 */
